/**
 * Compiler written in Node.js.
 * Recursively traverses folders and compiles a bunch of files into a single output file.
 */

// Arguments used by the program.
var args = require('minimist')(process.argv.slice(2));
// Framework for pretty printing.
var columnify = require('columnify');
// Framework for colorful output.
var colors = require('colors/safe');
// Internal utils for object debugging.
var util = require('util');
// Path util for constructing file paths.
var path = require('path');
// Filesystem util required for working with files.
var fs = require('fs');

// The working directory of the terminal invoking this script.
var exeDir = path.resolve(process.cwd());
// The name of the per-directory configuration file.
var folderConfigFile = "_aBuild.json";

/**
 * Iterates over the passed array of elements which may have embedded arrays, and applies the function to each element.
 * @param {function} fun The function to apply. Will be passed the current element.
 * @param {array} elems The elements to iterate over.
 */
function recArrayTraverse(fun, elems) {
    for (var i = 0; i < elems.length; i++) {
        var elem = elems[i];
        if (elem && elem.constructor === Array) {
            recArrayTraverse(fun, elem);
        } else {
            fun(elem);
        }
    }
}

/**
 * Logs the passed strings as errors, and then terminates the program.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function error(...strings) {
    if (!args.q) {
        recArrayTraverse(elem => console.log(colors.red("ERROR: " + elem)), strings);
    }
    process.exit(1);
}

/**
 * Logs the passed strings as warnings.
 * Does nothing if output is disabled.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function warn(...strings) {
    if (args.q) {
        return;
    }
    recArrayTraverse(elem => console.log(colors.yellow("WARNING: " + elem)), strings);
}

/**
 * Logs the passed strings.
 * Does nothing if output is disabled.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function log(...strings) {
    if (args.q) {
        return;
    }
    recArrayTraverse(elem => console.log(elem), strings);
}

/**
 * Logs the passed strings as debug messages.
 * Does nothing if debugging or output in general is disabled.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function debug(...strings) {
    if (args.q || !args.debug) {
        return;
    }
    recArrayTraverse(elem => console.log(colors.green("DEBUG: " + elem)), strings);
}

/**
 * Prints the help dialogue.
 */
function printHelp() {
    var argumentHelp = {
        "targets": "Everything without a flag is a filename or a folder to be traversed by aCompiler. No filenames means aCompiler will start in the current directory. Targets must always be specified first.",
        "--ext extension": "Will only target files with the specified extension. Does not override manually specified targets. Must start with a dot.",
        "--out filename": "Will output results to the specified file (creating it if need be). Without --out, the output will be printed in the console.",
        "-h": "Displays this help message. Does not run aCompiler.",
        "-q": "Runs aCompiler silently (with no output).",
        "--debug": "Displays debug information while running."
    };
    log("aCompiler is a compiler for Kobold Adventure.");
    log("It recursively traverses folders and compiles files into a single output file.");
    log("");
    log(columnify(argumentHelp, {columns: ['FLAG', 'EXPLANATION'], config: {FLAG: {minWidth: 20}, EXPLANATION: {maxWidth: process.stdout.columns - 40}}}));
}

/**
 * Checks if the passed parameter value is a string.
 * If it's not, the program exits with an error.
 * @param {string} toCheck The variable to check.
 * @param {string} flag Which flag is being checked. Produces a more informative error message.
 */
function stringOrExit(toCheck, flag) {
    if (typeof toCheck !== "string") {
        error("", `Parameter ${flag} must be a string value.`, `Example: ${flag} test`, "");
    }
}

/**
 * Gathers and returns all of the initial targets for aCompiler by inspecting the command line parameters.
 * @returns {array} An array of file and folder names to target.
 */
function buildInitialTargets() {
    debug("", "Building initial targets.");
    var input = args._;
    var output = [];

    if (input && input.length) {
        debug("Found targets in arguments.");
        for (var i = 0; i < input.length; i++) {
            output.push(path.resolve(input[i]));
        }
    } else {
        debug("No targets in arguments.", "Using current directory.");
        output.push(exeDir);
    }

    debug("Initial targets:", output, "");
    return output;
}

/**
 * Returns true if the target is an existing file or folder, false otherwise.
 * @param {string} file A path to a file or folder.
 * @returns {boolean} True if the target exists, false otherwise.
 */
function exists(file) {
    return fs.existsSync(file);
}

/**
 * Exits the program with an error message if the passed file does not exist.
 * @param {string} file A path to a file or folder.
 */
function existsOrExit(file) {
    if (!exists(file)) {
        error("", "File not found:", file, "");
    }
}

/**
 * Gets filestats for the specified file or folder.
 * @param {string} file The path to the file or folder to get stats for.
 * @returns {fs.Stats} The file's stats.
 */
function fileStats(file) {
    return fs.lstatSync(file);
}

/**
 * Returns true if the file associated with these stats is indeed a file.
 * @param {fs.Stats} stats The file's stats.
 * @returns {boolean} True if the file is actually a file, false otherwise.
 */
function isFile(stats) {
    return stats.isFile();
}

/**
 * Returns true if the file associated with these stats is a folder.
 * @param {fs.Stats} stats The file's stats.
 * @returns {boolean} True if the file is a folder, false otherwise.
 */
function isFolder(stats) {
    return stats.isDirectory();
}

/**
 * Reads the _aBuild.json file of the specified folder (if there is one) and returns it converted into an object.
 * Does nothing if there is no buildfile. Crashes the program is there is a build file, but it's malformed.
 * @param {string} target A path to a folder.
 * @returns {object} An object describing the aBuild file.
 */
function readBuildFile(target) {
    var aBuildPath = path.resolve(path.join(target, folderConfigFile));
    if (!exists(aBuildPath)) {
        return {};
    }
    log("Reading " + aBuildPath);
    var aBuild = fs.readFileSync(aBuildPath, 'utf8');
    try {
        aBuild = JSON.parse(aBuild);
    } catch (e) {
        error("", "Failed to parse build file " + aBuildPath, "Make sure the JSON file is well-structured.", "");
    }
    debug("Parsed: " + util.inspect(aBuild));
    return aBuild;
}

/**
 * Alters the constructed targets-object for the specified target-directory by reading its _aBuild.json file.
 * Does nothing if there is no _aBuild.json file.
 * @param {string} target A path to a folder.
 * @param {object} targets An object containing the reqs and the targets for the specified folder. May be altered by this function.
 */
function folderSpecificBuildFile(target, targets) {
    var aBuild = readBuildFile(target);
    var includes = aBuild.inc && aBuild.inc.constructor === Array ? aBuild.inc : [];
    var excludes = aBuild.exc && aBuild.exc.constructor === Array ? aBuild.exc : [];
    excludes.push(folderConfigFile); // Ensure the aBuild file is not included in the output unless specifically included.
    for (var i = 0; i < includes.length; i++) {
        var filePath = path.resolve(path.join(target, includes[i]));
        targets.reqs.push(filePath);
        excludes.push(filePath); // Make sure it's no longer in targets.
    }
    for (var i = 0; i < excludes.length; i++) {
        var filePath = path.resolve(path.join(target, excludes[i]));
        var index = targets.targets.indexOf(filePath);
        index >= 0 && targets.targets.splice(index, 1);
    }
    if (aBuild.skip) {
        targets.targets = [];
    }
}

/**
 * Builds and returns an object containing the reqs and the targets for the specified folder.
 * @param {string} target A path to a folder.
 * @returns {object} An object containing the reqs and the targets for the specified folder.
 */
function buildTargetsForFolder(target) {
    var output = {reqs: [], targets: []};
    var files = fs.readdirSync(target);
    for (var i = 0; i < files.length; i++) {
        var absoluteFile = path.resolve(path.join(target, files[i]));
        output.targets.push(absoluteFile);
    }
    folderSpecificBuildFile(target, output);
    return output;
}

/**
 * Compiles a single file, if it has not yet been encountered.
 * Prints a warning if the file has already been compiled.
 * @param {string} target A path to the file to compile.
 * @returns {string} The compiled file. The empty string if the file has already been compiled.
 */
function compileFile(target) {
    var output = "";
    if (compile.visited.includes(target)) {
        warn("File already compiled: " + target);
        return output;
    }
    compile.visited.push(target);
    log("Compiling " + target);
    var contents = fs.readFileSync(target, 'utf8');
    var relative = target.replace(exeDir, "").substr(1);
    output += "// " + relative + "\n\n";
    output += contents.trim();
    output += "\n\n// END " + relative + "\n\n";
    return output;
}

/**
 * Compiles a single folder, if it has not yet been encountered.
 * Prints a warning if the folder has already been compiled.
 * Compiles by creating a new list of reqs and targets and recursively calling the compile function with these new lists.
 * @param {string} target A path to the folder to compile.
 * @returns {string} The compiled folder. The empty string if the folder has already been compiled.
 */
function compileFolder(target) {
    if (compile.visited.includes(target)) {
        warn("File already compiled: " + target);
        return "";
    }
    compile.visited.push(target);
    debug("Compiling folder " + target);
    var targets = buildTargetsForFolder(target);
    return compile(targets.reqs, targets.targets);
}

/**
 * Compiles the passed required target, unless it has already been compiled.
 * Returns the compiled string.
 * Prints an error and exits the program if the file or folder can not be found.
 * @param {string} req A single required file or directory.
 * @returns {string} The compiled directory or file.
 */
function compileReq(req) {
    var output = "";
    existsOrExit(req);
    var stats = fileStats(req);
    if (isFolder(stats)) {
        output += compileFolder(req);
    } else if (isFile(stats)) {
        output += compileFile(req);
    } else {
        warn("Skipped because neither folder nor file: " + req);
    }
    return output;
}

/**
 * Compiles the passed target, unless it is filtered out by settings or unless it has already been compiled.
 * Returns the compiled string.
 * Prints an error and exits the program if the file or folder can not be found.
 * @param {string} target A single file or directory.
 * @returns {string} The compiled directory or file.
 */
function compileTarget(target) {
    var output = "";
    existsOrExit(target);
    var stats = fileStats(target);
    if (isFolder(stats)) {
        output += compileFolder(target);
    } else if (isFile(stats)) {
        if (!args.ext || path.extname(target) === args.ext) {
            output += compileFile(target);
        }
    } else {
        warn("Skipped because neither folder nor file: " + target);
    }
    return output;
}

/**
 * Recursively compiles all the specified required targets.
 * Returns the compiled string.
 * @param {array} reqs All the required targets to compile. These may be directories or files. Directories will be recursively scanned.
 *                     Files are guaranteed inclusion in the output, regardless of settings.
 * @returns {string} The compiled file as a single large string.
 */
function compileReqs(reqs) {
    var output = "";
    for (var i = 0; i < reqs.length; i++) {
        output += compileReq(reqs[i]);
    }
    return output;
}

/**
 * Recursively compiles all the specified targets.
 * Returns the compiled string.
 * @param {array} targets All the targets to compile. These may be directories or files. Directories will be recursively scanned.
 *                     Files may be omitted from the output depending on settings.
 * @returns {string} The compiled file as a single large string.
 */
function compileTargets(targets) {
    var output = "";
    for (var i = 0; i < targets.length; i++) {
        output += compileTarget(targets[i]);
    }
    return output;
}

/**
 * Recursively compiles all the specified targets.
 * Returns the compiled string.
 * initCompile() should always be called before the initial call to this function.
 * @param {array} reqs All the required targets to compile. These may be directories or files. Directories will be recursively scanned.
 *                     Files are guaranteed inclusion in the output, regardless of settings.
 * @param {array} targets All the targets to compile. These may be directories or files. Directories will be recursively scanned.
 *                     Files may be omitted from the output depending on settings.
 * @returns {string} The compiled file as a single large string.
 */
function compile(reqs, targets) {
    var output = "";
    output += compileReqs(reqs);
    output += compileTargets(targets);
    return output;
}

/**
 * Initializes things needed for compilation.
 * Should always be called before the initial call to compile.
 */
function initCompile() {
    compile.visited = [];
    log("Compiler initialized.");
}

if (args.h) {
    printHelp();
    process.exit(0);
}

log(colors.cyan.underline("aCompiler"));
debug("", "----------------------", "Debug logging started.", "----------------------", "");
debug("Arguments:", util.inspect(args), "");

if (args.ext) {
    stringOrExit(args.ext, "--ext");
}

if (args.out) {
    stringOrExit(args.out, "--out");
}

initCompile();
console.time("Duration");
log("Commencing compilation.");
var compiled = compile(buildInitialTargets(), []);
log("Compilation complete.");
console.timeEnd("Duration");

if (args.out) {
    var outputFileName = path.resolve(args.out);
    log("Exporting to " + outputFileName);
    fs.writeFileSync(outputFileName, compiled);
    log("Output file successfully created.");
} else {
    log(compiled);
}

