/**
 * A single color to be used in themes.
 * @param {string} color The color either in RGB or in HEX. Hex should be a string (with or without # sign). RGB should be an {r, g, b} object or a 3-value array.
 */
function ThemeColor(color) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    var minColor = 0;
    var maxColor = 255;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Forces all colors to remain within color-bounds.
     */
    function boundsForce() {
        that.r = Math.max(Math.min(that.r, maxColor), minColor);
        that.g = Math.max(Math.min(that.g, maxColor), minColor);
        that.b = Math.max(Math.min(that.b, maxColor), minColor);
    }

    /**
     * Converts the passed hexadecimal string to a color within bounds.
     * If the color is just a single digit, adds a second digit that is the same as the first digit.
     * @param {string} hex A hexidecimal string representing a color.
     * @returns {integer} The color, within (min, max) bounds.
     */
    function hexToCappedColor(hex) {
        var chars = hex.length;
        if (chars === 1) {
            return hexToCappedColor(hex + hex);
        }
        var max = Math.pow(16, chars);
        var ratio = (maxColor + 1) / max;
        var number = parseInt(hex, 16);
        if (isNaN(number)) {
            return 0;
        }
        return Math.floor(number * ratio);
    }

    /**
     * Converts the passed hexadecimal strings to an {r, g, b} object within color bounds.
     * @param {string} hexR Red hexadecimal string.
     * @param {string} hexG Green hexadecimal string.
     * @param {string} hexB Blue hexadecimal string.
     * @returns {object} {r, g, b} object.
     */
    function hexToRgb(hexR, hexG, hexB) {
        var output = {};
        output.r = hexToCappedColor(hexR);
        output.g = hexToCappedColor(hexG);
        output.b = hexToCappedColor(hexB);
        return output;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.r = 0;
    this.g = 0;
    this.b = 0;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Returns a textual hexadecimal representation of this color, starting with #.
     * @returns {string} A textual hexadecimal representation of this color, starting with #.
     */
    this.toHex = function () {
        var red = that.r.toString(16);
        var green = that.g.toString(16);
        var blue = that.b.toString(16);
        red = red.length > 1 ? red : "0" + red;
        green = green.length > 1 ? green : "0" + green;
        blue = blue.length > 1 ? blue : "0" + blue;
        return `#${red}${green}${blue}`;
    };

    /**
     * Returns a textual rgb representation of this color, starting with rgb( and ending with ).
     * @returns {string} A textual rgb representation of this color, starting with rgb( and ending with ).
     */
    this.toRgb = function () {
        return `rgb(${that.r}, ${that.g}, ${that.b})`;
    };

    /**
     * Returns a textual rgba representation of this color, starting with rgba( and ending with ).
     * @param {double} a The alpha value, between 0 and 1. 
     * @returns {string} A textual rgba representation of this color, starting with rgba( and ending with ).
     */
    this.toRgba = function (a) {
        return `rgba(${that.r}, ${that.g}, ${that.b}, ${a})`;
    };

    /**
     * Sets this ThemeColor's colors to the values of a hexadecimal string.
     * @param {string} hex A string (with or without # sign)
     */
    this.fromHex = function (hex) {
        var hexValue = hex;
        if (hexValue.startsWith("#")) {
            hexValue = hexValue.substr(1);
        }
        var charsPerColor = Math.floor(hexValue.length / 3);
        if (!charsPerColor) {
            that.reset();
            return;
        }
        var red = hexValue.substr(0, charsPerColor);
        var green = hexValue.substr(charsPerColor, charsPerColor);
        var blue = hexValue.substr(charsPerColor * 2, charsPerColor);
        that.fromRgb(hexToRgb(red, green, blue));
    };

    /**
     * Sets this ThemeColor's colors to the values of an {r, g, b} object.
     * @param {object} rgb {r, g, b} object containing color values (max 255).
     */
    this.fromRgb = function (rgb) {
        that.r = rgb.r || 0;
        that.g = rgb.g || 0;
        that.b = rgb.b || 0;
        boundsForce();
    };

    /**
     * Resets all colors to zero.
     */
    this.reset = function () {
        that.r = 0;
        that.g = 0;
        that.b = 0;
    };

    /**
     * Dampens this color by the specified factor between zero and one.
     * @param {number} factor The factor by which to dampen the color. Should be between zero and one. Above one, will amplify the color instead.
     */
    this.dampen = function (factor) {
        that.fromRgb({r: Math.round(that.r * factor), g: Math.round(that.g * factor), b: Math.round(that.b * factor)});
    };

    // Init
    if (color) {
        if (typeof color === "string") {
            that.fromHex(color);
        } else {
            if (Array.isArray(color)) {
                color = {r: color[0], g: color[1], b: color[2]};
            }
            that.fromRgb(color);
        }
    }
}
