

// SCENE Introduction

// scenes\Introduction\IntroductionController.js

/**
 * Controller for the Introduction scene.
 * @param {string} sceneId The id of the scene.
 */
function IntroductionController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;
    var genderChoice = "gender";
    var nameChoice = "name";
    var sanityLower = "sanity";
    var pressedContinue = "continue";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets the kobold's gender to the selected one.
     * @param {GenderEnum} choice The selected gender.
     */
    function genderChoiceGender(choice) {
        global.kobold.body.setGender(choice);
        if (choice == GenderEnum.MALE) {
            global.kobold.body.growDick(SizeEnum.SMALL);
        }
        if (choice == GenderEnum.FEMALE) {
            global.kobold.body.growBreasts(SizeEnum.SMALL);
        }
    }

    /**
     * Adds stats to the kobold depending on the chosen gender.
     * @param {GenderEnum} choice The selected gender.
     */
    function genderChoiceStats(choice) {
        if (choice == GenderEnum.MALE) {
            global.kobold.stats.increaseStat(StatsEnum.STRENGTH, 1);
            global.kobold.stats.increaseStat(StatsEnum.ENDURANCE, 1);
        }
        if (choice == GenderEnum.FEMALE) {
            global.kobold.stats.increaseStat(StatsEnum.AGILITY, 1);
            global.kobold.stats.increaseStat(StatsEnum.CHARISMA, 1);
        }
    }

    /**
     * Adds equipment to the kobold depending on the chosen gender.
     * @param {GenderEnum} choice The selected gender.
     */
    function genderChoiceEquipment(choice) {
        if (choice == GenderEnum.MALE) {
            global.kobold.inventory.equipment.wear(new LoinCloth());
        }
        if (choice == GenderEnum.FEMALE) {
            global.kobold.inventory.equipment.wear(new LinenPants());
            global.kobold.inventory.equipment.wear(new ChestWrap());
        }
    }

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Makes the gender choice, setting the kobold's gender and altering the kobold's history.
     * @param {GenderEnum} choice The selected gender.
     */
    this.makeGenderChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[genderChoice] = choice;
        global.kobold.history.set(id, history);
        genderChoiceGender(choice);
        genderChoiceStats(choice);
        genderChoiceEquipment(choice);
    };

    /**
     * Fetches the gender choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     */
    this.getGenderChoice = function () {
        var history = global.kobold.history.get(id);
        return history[genderChoice];
    };

    /**
     * Returns true if the gender choice has already been made. False otherwise.
     */
    this.hasMadeGenderChoice = function () {
        return global.kobold.history.has(id, genderChoice);
    };

    /**
     * Makes the name choice, setting the kobold's name.
     * @param {string} name The entered name.
     */
    this.makeNameChoice = function (name) {
        var history = global.kobold.history.get(id);
        history[nameChoice] = name;
        global.kobold.history.set(id, history);
        global.kobold.name = name;
    };

    /**
     * Fetches the name choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     */
    this.getNameChoice = function () {
        var history = global.kobold.history.get(id);
        return history[nameChoice];
    };

    /**
     * Returns true if the name choice has already been made. False otherwise.
     */
    this.hasMadeNameChoice = function () {
        return global.kobold.history.has(id, nameChoice);
    };
    
    /**
     * Presses the continue button.
     */
    this.pressContinue = function () {
        var history = global.kobold.history.get(id);
        history[pressedContinue] = true;
        global.kobold.history.set(id, history);
    };
    
    /**
     * Returns true if the continue button has been pressed.
     */
    this.hasContinued = function () {
        return global.kobold.history.has(id, pressedContinue);
    };

    /**
     * Lowers the kobold's sanity, if it hasn't been lowered yet.
     * @returns {string} HTML describing the change.
     */
    this.thirdPartMentalEffects = function () {
        if (!global.kobold.history.has(id, sanityLower)) {
            global.kobold.status.mental.decrease(15, "Deadline worries");
            var history = global.kobold.history.get(id);
            history[sanityLower] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([new Change("Distraught", "Sanity", -15)]);
    };
}

// END scenes\Introduction\IntroductionController.js

// scenes\Introduction\IntroductionFirstPart.js

/**
 * The first part of the introduction.
 * @param {IntroductionController} controller The controller.
 * @param {IntroductionSecondPart} secondPart The next part to load once finished.
 */
function IntroductionFirstPart(controller, secondPart) {
    // PRIVATE MEMBERS \\       var private = "test";
    var ctrl = controller;
    var nextPart = secondPart;
    var start =  `  <div class="title fancy">
                         introduction
                    </div>
                    <div class="text">
                        You are a <span class="bold">kobold</span>. You are small, flexible, nimble and rather stealthy when you want
                        to be. On the other hand you are naive, not very strong, kind of a pushover and 
                        generally not very good in a fight. You make up for your lack of combat skills
                        by a driven determination not to fall prey to the many dangers of this world.
                    </div>

                    <div class="text">
                        The area you live in is mostly predominated by humans and anthros, who see you as a lesser species. 
                        Kobolds officially have no rights, so the sight of one of your kind getting raped, 
                        beaten, sold as a slave or even killed in broad daylight is, unfortunately for you, not so uncommon.
                    </div>

                    <div class="text">
                        Luckily, you have managed to avoid the worst so far, by making your home in the outstretched 
                        woods surrounding the human capital of <span class="bold">Varanar</span>. While the city itself is 
                        a melting pot for all walks of life, the forest isn't nearly as diverse.
                    </div>
    
                    <div class="text">
                        Still, there's plenty to be afraid of. 
                        Wolves, bobcats, forest wyverns, gryphons, you name it.
                        Heck, even something as docile as a deer, when provoked, could pose a threat to a small kobold like you.
                        To make matters worse, there are rumors going around of a hulking mass of tentacles skulking through the darkness, 
                        violating who or <span class="italic">what</span>ever it gets in its grip.
                    </div>

                    <div class="text">
                        On top of the feral dangers, there are various bandits and outlaws 
                        which have made these woods their home. Many of them band together, forming
                        fickle alliances that often end in backstabbing and betrayal, all in pursuit
                        of coin. They prey on unsuspecting victims, to rob them of their 
                        riches, their decency, their life or all three.
                    </div>
    
                    <div class="text">
                        Some of these groups are comprised exclusively of kobolds.
                        One of them is the <span class="bold">Three-Tailed Rat Clan</span>.
                        Based in Varanar, this organized gang of ruthless, thick-scaled thugs is lead by 
                        the only non-kobold member, a man known as <span class="bold">The Rat</span>.
                        Their main source of income is extortion, coercing protection money from people
                        living in the more lawless regions of the land.
                        <span class="bold">Regions like the forest. And people like you.</span>
                    </div>`;
    
    var genderChoiceHeader =       `<div class="choicecontainer">
                                        <div class="inlinechoiceheader">
                                            You are:
                                        </div>`;
    
    var genderChoiceButtonMid =        `<hr class="inlinechoiceseparator">`;
    
    var genderChoiceFooter =    `</div>`;
    
    var genderChoiceButtonMaleId = "male";
    var genderChoiceButtonFemaleId = "female";
    
    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    
     /**
     * Builds and returns the gender choice.
     * The starting choice may be disabled if it has already been made.
     */
    function buildGenderChoice() {
        var choice = genderChoiceHeader;
        choice += buildGenderChoiceMaleButton();
        choice += genderChoiceButtonMid;
        choice += buildGenderChoiceFemaleButton();
        choice += genderChoiceFooter;
        return choice;
    }
    
    /**
     * Returns the male choice button.
     */
    function buildGenderChoiceMaleButton() {
        var disabled = ctrl.hasMadeGenderChoice() ? "disabled" : "";
        var selected = ctrl.hasMadeGenderChoice() && ctrl.getGenderChoice() == GenderEnum.MALE ? "selected" : "";
        return `<button class="button inlinechoice ${selected}" id="${genderChoiceButtonMaleId}" ${disabled}>
                    Male
                </button>`;
    }
    
    /**
     * Returns the female choice button.
     */
    function buildGenderChoiceFemaleButton() {
        var disabled = ctrl.hasMadeGenderChoice() ? "disabled" : "";
        var selected = ctrl.hasMadeGenderChoice() && ctrl.getGenderChoice() == GenderEnum.FEMALE ? "selected" : "";
        return `<button class="button inlinechoice ${selected}" id="${genderChoiceButtonFemaleId}" ${disabled}>
                    Female
                </button>`;
    }
        
    /**
     * Registers a click listener for the male gender choice button.
     */
    function registerMaleGenderChoiceListener() {
        $("#" + genderChoiceButtonMaleId).click(function(){
            $(this).addClass("selected");
            genderChoiceClick(GenderEnum.MALE);
        });
    }
    
    /**
     * Registers a click listener for the female gender choice button.
     */
    function registerFemaleGenderChoiceListener() {
        $("#" + genderChoiceButtonFemaleId).click(function(){
            $(this).addClass("selected");
            genderChoiceClick(GenderEnum.FEMALE);
        });
    }
    
    /**
     * Executed when a gender choice button is clicked.
     * @param {GenderEnum} choice The selected gender.
     */
    function genderChoiceClick(choice) {
        disableButtons();
        ctrl.makeGenderChoice(choice);
        var secondPart = nextPart.build(true);
        global.ui.appendBeforeFinisher(secondPart, global.ui.genericHeaderId);
        $("#" + nextPart.id).fadeIn(400);
        nextPart.registerListeners();
    }
    
    /**
     * Disables both gender choice buttons.
     */
    function disableButtons() {
        var maleButton = $("#" + genderChoiceButtonMaleId);
        var femaleButton = $("#" + genderChoiceButtonFemaleId);
        maleButton.unbind();
        femaleButton.unbind();
        maleButton.prop('disabled', true);
        femaleButton.prop('disabled', true);
    }
    
    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
    
    /**
     * Builds and returns this part.
     */
    this.build = function() {
        return start + buildGenderChoice();
    };
    
    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function() {
        if(ctrl.hasMadeGenderChoice()){
            return;
        }
        registerMaleGenderChoiceListener();
        registerFemaleGenderChoiceListener();
    };
}

// END scenes\Introduction\IntroductionFirstPart.js

// scenes\Introduction\IntroductionSecondPart.js

/**
 * The second part of the introduction.
 * @param {IntroductionController} controller The controller.
 * @param {IntroductionThirdPart} thirdPart The next part to load once finished.
 */
function IntroductionSecondPart(controller, thirdPart) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    var nextPart = thirdPart;
    var nameInputId = "namefield";
    var submitNameButtonId = "namebutton";

    var maleDescription = `<div class="text">
                                You are male. You're a bit stronger and tougher than the females of your kind, 
                                yet still pathetically weak compared to anything that isn't a kobold. 
                                You’re wearing nothing but a small, simple ${global.kobold.inventory.createItemWithTooltip(new LoinCloth())}, 
                                which covers the cockslit between your legs quite nice. Your scaly butt, 
                                however, is completely exposed. The only thing covering your virginal rear is your tail.
                            </div>`;

    var femaleDescription = `<div class="text">
                                You are female. You’re not quite as strong as the males of your kind, but you make 
                                up for it by being even lighter, even more nimble and just a little bit more flexible. 
                                Enemies you face might be more concerned with fucking something as small and cute as 
                                yourself rather than trying to hurt you. This may be a good thing, yet if you don’t 
                                watch out you might end up a cum-filled lizard before long. 
                            </div>
                            <div class="text">
                                You wear a baggy pair of brown ${global.kobold.inventory.createItemWithTooltip(new LinenPants())}, 
                                and a matching ${global.kobold.inventory.createItemWithTooltip(new ChestWrap())} which covers up
                                your small breasts quite nicely. Your leg-wear has seen better days. Small holes and
                                cuts dot the fabric in every which way, some large enough to show just a glimmer of your
                                scales beneath. At least your groin and rear are still suitably covered.
                            </div>`;

    var postDescription = `<div class="text">
                                <span class="sneakytext bold">
                                    You can check your attributes in the stats tab, as well as get a physical description of your body.
                                    Your equipment can be found in the inventory tab.
                                    The names of most pieces of equipment can be moused over to reveal a short description of the item.
                                    Your current health, sanity, stamina and sexual exertion may be viewed in the status tab.
                                </span>
                            </div>
                            <div class="text">
                                You used to have a comfortable wool sweater that had kept you warm for many winters, 
                                yet you lost it just a few days ago whilst running from a rather unpleasant group of humans.
                                The soft fabric got caught on a thornbush. Fearing for your life, you slipped out of the 
                                sweater instead of trying to pry it loose. You can't quite remember where all of this happened.
                                The forest can be a confusing place, at times.
                            </div>
                            `;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the enclosing div's starting tag.
     * @param {bool} invisible True if the entire section should be invisible.
     * @returns {string} The starting tag.
     */
    function enclosingDivStart(invisible) {
        var style = invisible ? 'style="display:none;"' : "";
        return `<div id="${that.id}" ${style} >`;
    }

    /**
     * Builds and returns the gender-specific description part.
     * @returns {string} The gender-specific description part.
     */
    function genderDescription() {
        if (ctrl.getGenderChoice() == GenderEnum.MALE) {
            return maleDescription;
        } else {
            return femaleDescription;
        }
    }

    /**
     * Builds and returns the end of the enclosing division.
     * @returns {string} The end of the enclosing division.
     */
    function enclosingDivEnd() {
        return "</div>";
    }

    /**
     * Builds and returns the name selection section.
     * @returns {string} The name selection section.
     */
    function nameSelection() {
        var nameChoiceMade = ctrl.hasMadeNameChoice();
        var disabled = nameChoiceMade ? "disabled" : "";
        var value = nameChoiceMade ? ctrl.getNameChoice() : "Moe";
        var nameSelect = `<div class="choicecontainer">`;
        nameSelect += `<div class="inlinechoiceheader">Your name is:</div>`;
        nameSelect += `<input id="${nameInputId}" class="textinput" maxlength="20" value="${value}" ${disabled}>`;
        nameSelect += `<button id="${submitNameButtonId}" class="button textinputbutton" ${disabled}>Confirm</button>`;
        nameSelect += `</div>`;
        return nameSelect;
    }

    /**
     * Registers a listener disabling the name submit button if no name has been entered.
     */
    function registerNameInputChangeListener() {
        var input = $("#" + nameInputId);
        var button = $("#" + submitNameButtonId);
        input.on('input', function () {
            var value = input.val();
            if (value.length === 0) {
                button.prop('disabled', true);
            } else {
                button.prop('disabled', false);
            }
        });
    }

    /**
     * Registers a listener capable of submitting the name choice.
     */
    function registerNameSubmitListener() {
        var input = $("#" + nameInputId);
        var button = $("#" + submitNameButtonId);
        button.click(function () {
            var value = input.val();
            if (value.length === 0) {
                return;
            }
            button.prop('disabled', true);
            input.prop('disabled', true);
            ctrl.makeNameChoice(value);
            var thirdPart = nextPart.build(true);
            global.ui.appendBeforeFinisher(thirdPart, global.ui.genericHeaderId);
            $("#" + nextPart.id).fadeIn(400);
            nextPart.registerListeners();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "part2";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.hasMadeGenderChoice()) {
            return "";
        }
        var part = enclosingDivStart(invisible);
        part += genderDescription();
        part += postDescription;
        part += nameSelection();
        part += enclosingDivEnd();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.hasMadeGenderChoice()) {
            return;
        }
        if (ctrl.hasMadeNameChoice()) {
            return;
        }
        registerNameInputChangeListener();
        registerNameSubmitListener();
    };
}

// END scenes\Introduction\IntroductionSecondPart.js

// scenes\Introduction\IntroductionThirdPart.js

/**
 * The third part of the introduction.
 * @param {IntroductionController} controller The controller.
 */
function IntroductionThirdPart(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    var nextScene = "Forest";

    var continueButtonId = "continueButton";
    var contentStart = `<div class="text">`;
    var contentMiddle = `   You are a thief. You make a living by sneaking into the
                            various camps of passersby, bandits and other forest-inhabitants,
                            stealing whatever you can and then hightailing it out of there hoping you weren't seen.
                            This tactic has worked out pretty well for you in the past.
                            For several years, you've managed to stay afloat, if just barely.
                            Some months you even had the money to buy yourself some actual food, 
                            instead of living off the scraps of others.
                        </div>
                        <div class="text">
                            Those golden days are over.
                            The Three-Tailed Rat Clan has recently raised its <span class="italic">fee</span>
                            to a level you can no longer afford. 
                            Fifty coins per month was a reasonable tariff, but one hundred is simply
                            <span class="bold">too</span> much for a simple kobold like yourself to steal on a monthly basis.
                            You have no choice <span class="bold">but</span> to pay, however. 
                            You have nowhere else to go, you've grown rather attached to your little shack in the woods,
                            and you'd rather <span class="bold">not</span> risk the ire of one of the deadliest gangs around.
                        </div>
                        <div class="text">
                            Over the past few months, you've been working your <span class="bold">ass</span> off trying to make ends meet.
                            You've been taking risks: stealing from caravans, stalking bandits to pick their prey clean of any leftover coin,
                            even making a trip to the city to try your hand at pickpocketing. 
                            Despite your best attempts, you've accumulated a significant amount of debt.
                            <span class="bold">Two hundred and fifty</span> coins worth, to be precise.
                        </div>
                        <div class="text">
                            The Three-Tailed Rat himself has issued you a warning.
                            Of course, not in person. 
                            Nobody talks to the Rat but his second-in-command.
                            A courier delivered a letter to your very doorstep.
                            Three days. 
                            That's all the time you had to repay your debt.
                            <span class="bold">Or else.</span>
                        </div>`;

    var contentMiddle2 = `  <div class="text">
                            Needless to say, the worries clouding your mind haven't been very conducive to your trade.
                            In fact, you've fumbled every single heist you've attempted since receiving the note.
                            Worse yet, last night, you were unfortunate enough to drop your meagre coin pouch while running from a pair of mounted caravan guards.
                            Today, the last day you had to pay up, you spent scouring every inch of the forest floor in a desperate attempt to recover your purse.
                        </div>
                        <div class="text">
                            The few coins within might have been enough to at least buy you some time, but alas.
                            The horsemen must have taken it as their just due. 
                            Or maybe another group stumbled upon it. 
                            Heck, the neck cord could have wrapped itself around some wild animal's leg for all you know.
                        </div>
                        <div class="text">
                            In the early afternoon, you were about ready to give up.
                            Not just on finding your lost pouch, but on <span class="bold">everything</span>.
                            You were done for.
                            A quarter of a thousand gold due tonight, and you hadn't a coin to your name.
                            All seemed lost.
                        </div>
                        <div class="text">
                            Still, a flicker of hope remained.
                            At least you didn't get caught by the horsemen. `;

    var contentEnd = `   Maybe, just maybe, you could meet your deadline still.
                            You just needed to score <span class="bold">one big hit.</span>
                            Driven by the will to live, you set out to do what you do best.
                        </div>`;


    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the first section of the third part.
     * @returns {string} The first section of the third part.
     */
    function buildName() {
        var gender = ctrl.getGenderChoice();
        var name = ctrl.getNameChoice();
        var part = contentStart;
        part += `Your name is <span class="bold">${name}</span>. `;
        part += contentMiddle;
        part += contentMiddle2;
        if (gender == GenderEnum.MALE) {
            part += `And the way you escaped them by <span class="bold">climbing up a tree</span> was pretty darn impressive.`;
        } else {
            part += `And the way you escaped them by <span class="bold">hiding in a bush</span> was pretty darn impressive.`;
        }

        part += contentEnd;
        return part;
    }

    /**
     * Builds and returns the continue button at the end of the scene.
     * @returns {string} The continue button.
     */
    function buildContinueButton() {
        var disabled = ctrl.hasContinued();
        var button = `<button class="button bigbutton continuebutton" id="${continueButtonId}" ${disabled ? "disabled" : ""}>`;
        button += `Continue &gt;&gt`;
        button += `</button>`;
        return button;
    }

    /**
     * Registers a listener for the continue button.
     */
    function registerContinueButtonListener() {
        $("#" + continueButtonId).click(function () {
            ctrl.pressContinue();
            global.ui.fadeOutContent(function () {
                global.sceneEnd();
                global.sceneLoader.loadScene(nextScene);
            });
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "part3";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {bool} invisible True if the entire section should be invisible.
     */
    this.build = function (invisible) {
        if (!ctrl.hasMadeNameChoice()) {
            return "";
        }
        
        var change = ctrl.thirdPartMentalEffects();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildName(change);
        part += change;
        part += buildContinueButton();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.hasContinued()) {
            return;
        }
        registerContinueButtonListener();
    };
}

// END scenes\Introduction\IntroductionThirdPart.js

// scenes\Introduction\IntroductionBuilder.js

/**
 * Builder for the Introduction scene.
 * @param {IntroductionController} controller The controller managing this scene.
 */
function IntroductionBuilder(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var thirdPart = new IntroductionThirdPart(controller);
    var secondPart = new IntroductionSecondPart(controller, thirdPart);
    var firstPart = new IntroductionFirstPart(controller, secondPart);
    
    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
    
    /**
     * Progressively builds the contents of the scene.
     * @returns {string} The innards of the scene, as a HTML string.
     */
    this.buildScene = function() {
        var scene = firstPart.build();
        scene += secondPart.build();
        scene += thirdPart.build();
        return scene;
    };
    
    /**
     * Registers all the required button listeners.
     */
    this.registerListeners = function() {
        firstPart.registerListeners();
        secondPart.registerListeners();
        thirdPart.registerListeners();
    };
}

// END scenes\Introduction\IntroductionBuilder.js

// scenes\Introduction\Introduction.js

/**
 * The Introduction scene. Shown after the big button in the title screen has been clicked.
 */
function Introduction() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "Introduction";
    var sceneName = "Introduction";
    var controller = new IntroductionController(sceneId);
    var builder = new IntroductionBuilder(controller);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += builder.buildScene();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        var audio = new AudioController(that.id, "res/audio/titlescreen/guitar_chord.mp3", 100);
        audio.playOnce();
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
        global.ui.applyTheme();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        builder.registerListeners();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new Introduction();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\Introduction\Introduction.js



// END SCENE Introduction



// SCENE Forest

// scenes\Forest\unknown\pry\bite\ForestUnknownInitialBiteContinue.js

/**
 * Bite section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownInitialBiteContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `initBiteLeftContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cock = ctrl.unknown.cock();
        var pants = ctrl.unknown.pants();
        var pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        var underwear = ctrl.unknown.underwear();
        var underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        var toReturn = `[p]
                            You lay down on the floor for several more minutes, until the sun has disappeared completely,
                            the darkness of night washing over the deepwoods, pseudo-dusk replaced by pitch-blackness.
                            You cry, you sob and you whimper, carefully rubbing your injured wrist.
                            You are shattered, both mentally as well as physically.
                            You feel used, pumped and [b]dumped[/b], taken advantage of by the first creature you encountered in this unhospitable part of the forest.
                        [/p]
                        [p]
                            The ground growing far too cold for comfort, you slowly rise to your shaky feet.
                            The steady stream of sweet-smelling, honey-like nectar flowing from your tailhole has slowed to a trickle, oozing down along your inner thighs.
                        `;

        if (pants) {
            toReturn += `   As you lift yourself off of the ground, the ripped remains of your pant legs slide off of your scaly frame,
                            leaving you naked from the waist-downwards.`;
        } else if (underwear) {
            toReturn += `   As you lift yourself off of the ground, you become aware of the fact that your ${underwearPopup} feels a lot heavier than before.
                            It must have soaked up a lot of the syrup oozing from your butt.`;
        }

        toReturn += `   [/p]
                        [p]
                            Then, you spot something that may offer a bit of consolation, or at the very least a bit of shielding from the steadily dropping temperature.
                            Your ${global.kobold.inventory.createItemWithTooltip(new WoolSweater())}, 
                            which you let go of during the struggle earlier, seems to have survived the fight mostly unscathed.
                            You pick it up off of the ground, and eagerly don it, 
                            focussing your mind on the warmth and comfort it provides to your upper body as opposed to the coldness afflicting your naked rear.
                        [/p]
                        [p]
                            At least you got [i]something[/i] out of this entire ordeal.
                            With your spirit slightly placated although still significantly battered,
                            you begin to ponder your current situation.
                            You doubt you can find your way home in the dark.
                            You are tired, but sleeping in a place this dangerous is [b]not[/b] an option.
                            Your best bet is to keep moving.
                            If you keep going in one direction, you'll reach the edge of this godforsaken forest at one point or another, 
                            and [i]then[/i] you can go home.
                        [/p]
                        [p]
                            Your plan formed, you pick a direction at random, and begin to march forwards,
                            your gait a little distorted by the slight amount of pain radiating from your bottom.
                            Instead of giving up, you double down.
                            You may have suffered a small setback, but you [b]are[/b] going to make it out of here [b]alive[/b],
                            and there's no amount of plants or tentacles, no matter how [b]thick[/b] or [b]slimy[/b],
                            that are you going stop you from doing so.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.continue();
            loadNextPart("RiversideMarsh", "Deepwoods");
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownInitialBiteContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.pry.initbiterape.continued()) {
            return "";
        }

        global.ui.applyTheme("forest_night");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.initbiterape.continueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.pry.initbiterape.continued()) {
            return;
        }

        if (ctrl.unknown.hasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\bite\ForestUnknownInitialBiteContinue.js

// scenes\Forest\unknown\pry\bite\ForestUnknownInitialBiteRape.js

/**
 * Initial bite section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownInitialBiteRape(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "unknownInitialBiteRapeContinueButton";

    var next = new ForestUnknownInitialBiteContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cock = ctrl.unknown.cock();
        var pants = ctrl.unknown.pants();
        var pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        var underwear = ctrl.unknown.underwear();
        var underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        var rearExposed = ctrl.unknown.rearExposed();
        var toReturn = `[p]
                            Panic building inside of you once more, you powerlessly wriggle about before limply hanging your head in defeat.
                            Unable to escape, you are at the monster's mercy.
                            The smooth tentacle slithers in between your feet, raising itself off of the ground behind you.
                            It gradually climbs upwards until reaching about the height of your waist, where it hovers ominously a few inches away from your bottom.
                        [/p]
                        [p]
                            One of the thorny ones, meanwhile, takes hold of your final unbound appendage: your tail, which is painfully yanked upwards.
                            Your last means of defence -- of somewhat shielding your rear and your groin from the creature -- taken from you, you burst into tears.
                            Without your tail to cover your backside, you feel [b]vulnerable[/b].
                            Entirely [b]defenceless[/b].
                            [b]Exposed[/b], ${rearExposed ? `your bare ass visible for all to see` : `despite your ${pantsPopup} still offering some modicum of modesty`}.
                        [/p]`;

        if (!rearExposed) {
            toReturn += `[p]
                            And even that final ounce of decency is taken from you.
                            The thornless tentacle surges forwards, probing your clothed rear, attempting to find an opening, only to discover fabric blocking its advances.
                            You can [b]feel[/b] the monster growing agitated, its grip on your limbs tightening,
                            squeezing down, threatening to puncture you, to hurt you, perhaps even [b]kill[/b] you out of sheer frustration.
                            Luckily for you, the prodding tendril soon finds its way to the waistband of your ${pantsPopup},
                            slipping underneath only to grip the textile and harshly [b]tear[/b] it from your body, 
                            your pants completely [b]ruined[/b] as the entire rear midsection gets ripped to pieces by the annoyed monstrosity,
                            exposing your bottom for the entire deepwoods to see.
                        [/p]`;
        }

        toReturn += `   [p]
                            The tentacle slowly approaches your naked rump, immediately finding its way in between your buttcheeks.
                            Its tapered tip rubs against your virginal tailhole, coating your sensitive rear-end with a slightly warm,
                            viscous yet slippery slime, the same stuff you reckon the tendril is soaked with.
                            You briefly wonder 
                            ${cock ? "how the monster knows you're a boy," : "why the monster is interested in your butt instead of your lady parts,"}
                            but you decide [i]not[/i] to question how nature works.
                        [/p]
                        [p]
                            After circling your hole for a few moments, which seem like aeons to you, the tentacle finally lines itself up proper.
                            In a single, smooth thrust, the narrow tip slides into your rear, the slick goop aiding its entrance.
                            But it doesn't stop there.
                            Continuing to force itself deeper,
                            the tentacle slides inch after inch of its gradually thickening [i]shaft[/i] into your body,
                            stretching your hole to just below your natural limits, yet still far beyond what you thought you could ever fit up your butt.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/unknown/pry/bite/buttsex-${cock ? "m" : "f"}-full.png" target="_blank">
                                    <img class="nsfw wideimage" src="scenes/Forest/unknown/pry/bite/buttsex-${cock ? "m" : "f"}.png" title="Illustration by Feldspar">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            You wince, not so much in pain, but out of pure discomfort, the insertion far too well lubricated to really [i]hurt[/i] all that much.
                            The tendril immediately begins to piston in and out of your behind, thoroughly [b]fucking[/b] your warm, tight, never-before-used ass.
                            A few thrusts in,
                            the tip of the tendril squirts a small amount of heated liquid, 
                            the same stuff it slathered across the exterior of your tailhole, deep inside of you,
                            slickening your insides to smoothen its motions.
                        [/p]
                        [p]
                            The added internal lubricant allows the tentacle to pick up its pace,
                            your head being forcibly tugged just a little closer to the ground to raise up your ass a bit more,
                            giving it slightly better access to your tender rear.
                            The plant-appendage lifts more of its own length into the air, allowing it to thrust more [i]horizontally[/i] and less [i]vertically[/i],
                            which, in turn, enables it to [b]shove[/b] just a little more of itself into you.
                            It's less than an extra inch, but you can [b]feel[/b] the difference in both girth and depth,
                            your body split open to its very limit.
                        [/p]
                        [p]
                            Luckily for you, the torment and indignation don't last long.
                            Mere minutes after having slid its slimy tip into your rear, the tentacle's rapid pounding becomes far more [i]erratic[/i],
                            a clear sign of its impending [i]orgasm[/i], if you can call it that.
                            Instead of the earlier rhytmic thrusting, it devolves into a series of raw, brutal [b]shoves[/b],
                            the monster trying, but failing to [b]ram[/b] just a little bit more of itself into you.
                        [/p]
                        [p]
                            Then, with one final, intense [b]push[/b], the creature lodges its tentacle up your rear as far as it possibly can,
                            the tendril's entire length visibly quivering and pulsing as wave after wave of thick, 
                            lukewarm [b]sap[/b] squirts from the tip, filling up your insides.
                            After pumping its hefty floral load into your body, the creature retracts from your thoroughly abused tailhole,
                            which is left slightly agape as the final few inches of the tentacle withdraw, syrupy goop immediately beginning to leak out.
                        [/p]
                        [p]
                            After the slick tendril has made its way back to the monstrosity's central mass, disappearing beneath the myriad of the thorns,
                            the bush slowly releases its grip on your limbs, your throat and your tail.
                            No longer supported by the threat of injury, you weakly fall forwards to the ground, your bare, stretched ass oozing honey-like tentacle-cum.
                            Hurting and defiled down to your very core, you sobbingly curl up into a ball, shivering both in fear and out of coldness, 
                            the sun going down as the organic terror makes its way out of the small clearing.
                        [/p]`;

        toReturn += `   [p]
                            [c]
                                [b]You have lost your anal virginity.[/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.pry.initbiterape.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.initbiterape.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownInitialBiteRape";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.pry.initbite.continued()) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.initbiterape.change(ctrl.unknown.pants());
        part += buildChoice();
        part += next.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.pry.initbite.continued()) {
            return;
        }

        next.registerListeners();

        if (ctrl.unknown.pry.initbiterape.continued()) {
            return;
        }

        registerContinueListener();
    };
}

var ForestUnknownInitialBiteChoice = {
    RELENT: 1,
    STRUGGLE: 2
};

// END scenes\Forest\unknown\pry\bite\ForestUnknownInitialBiteRape.js

// scenes\Forest\unknown\pry\bite\ForestUnknownInitialBite.js

/**
 * Initial bite section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownInitialBite(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "unknownInitialBiteContinueButton";

    var next = new ForestUnknownInitialBiteRape(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Your mind races to come up with a [i]way out[/i].
                            You try to tug against the plant's grasp once more, but it simply won't budge, and you'll end up hurting yourself if you continue to resist it.
                            With the other spiked tentacles inching ever closer, you figure that if the tendril won't let go, you'll just have to [i]sever it[/i] instead.
                        [/p]
                        [p]
                            The organic terror may have restrained one of your wrists, but [b]not[/b] your head!
                            Steeling your resolve, you firmly yank your grasped arm back one final time.
                            The spikes around your wrist sink into your flesh, drawing a small amount of blood,
                            but the pain is a small price to pay for fully stretching the tendril,
                            preparing the rope-like plant part for a well-aimed [b]bite[/b].
                        [/p]
                        [p]
                            You lean forwards, bending over slightly just to [i]reach[/i] the tentacle.
                            Then, tilting your head to align your muzzle along the length of the organic barbed wire, you [b]chomp[/b] down on it,
                            your pointed teeth easily cutting through the plant matter, [b]tearing out[/b] a piece of the tendril.
                            While the now-amputated segment around your wrist instantly loses its grip and falls to the ground, your victory is short-lived.
                        [/p]
                        [p]
                            Before you even manage to retract your head once more, another of the plant's appendages takes advantage of your temporary vulnerable state.
                            It lashes out, leaping from the very ground all the way to your neck, tightly wrapping around your tender throat,
                            preventing you from pulling back, lest you rake open your own windpipe.
                            Panicking, you wildly flail your limbs in an attempt to sever your new [i]leash[/i],
                            yet instead of freedom, you only grow more constricted as your already bleeding wrist gets taken hold of anew,
                            and your one free arm, too, gets entangled by thorny vines.
                        [/p]
                        [p]
                            You kick your legs and squirm about as much as your bondage allows, in an attempt to free yourself still,
                            or at least to prevent the monster from grabbing your ankles.
                            Your writhing, while undoubtedly a hindrance, only serves to delay the inevitable.
                            After a few unsuccessful tries, the creature finally manages to grab your left leg by the calf,
                            and with your remaining mobility halved, it isn't long until your right leg, too, gets snagged.
                        [/p]
                        [p]
                            You try to struggle, yet no matter which part of your body you move, you end up getting hurt by the many thorns surrounding your every limb.
                            You growl loudly, voicing your impotent frustration.
                            Fear builds up in the back of your mind.
                            The thornbush calmly reinforces its grip on your arms, almost fully [b]encasing[/b] them in tendrils.
                            Your legs, too, become wrapped in prickly plant matter.
                            You are completely immobilized.
                            While you may have hurt the plant when you bit off one of its tendrils, it has seemingly won this conflict.
                        [/p]
                        [p]
                            Despite the pain, you continue to struggle for a good minute more, [b]desperate[/b] to escape the organic terror's clutches.
                            Using the grip on your neck, it [b]yanks[/b] your head further downwards, properly bending you over.
                            You don't stop testing your restraints until you notice a new development,
                            one which makes your heart skip a beat when you realize its implications.
                        [/p]
                        [p]
                            With your head forcibly pointed downwards, you get a good, 
                            up-close look at a different kind of tendril which snakes out from underneath the central mass of the animated bush.
                            A smooth, thornless, brightly colored, oddly phallic-shaped one which seems to be covered from tip to base in 
                            some kind of sickly sweet-smelling [i]goop[/i], slick enough to allow it to [i]slide[/i] over the forest floor towards your legs.
                            [i]Oh no.[/i]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.pry.initbite.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.initbite.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownInitialBite";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.pry.base.get() !== ForestUnknownPryChoice.BITE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.initbite.change();
        part += buildChoice();
        part += next.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.pry.base.get() !== ForestUnknownPryChoice.BITE) {
            return;
        }

        next.registerListeners();

        if (ctrl.unknown.pry.initbite.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\bite\ForestUnknownInitialBite.js

// scenes\Forest\unknown\pry\claw\ForestUnknownClaw.js

/**
 * Claw section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownClaw(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `clawContinueButton`;
    
    var next = new ForestUnknownClawContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Your mind races to come up with a [i]way out[/i].
                            You try to tug against the plant's grasp once more, but it simply won't budge, and you'll end up hurting yourself if you continue to resist it.
                            With the other spiked tentacles inching ever closer, you figure that if the tendril won't let go, you'll just have to [i]slice it off[/i] instead.
                        [/p]
                        [p]
                            The organic terror may have restrained [i]one[/i] of your wrists, but [b]not[/b] both of them!
                            Gritting your teeth and steeling your resolve, you firmly yank your grasped arm back one final time.
                            The spikes around your wrist sink into your flesh, drawing a small amount of blood,
                            but the pain is a small price to pay for fully stretching the tendril,
                            preparing the rope-like plant part for a well-aimed slicing swipe from your razor-sharp claws.
                        [/p]
                        [p]
                            The strain from your tug pulling the tendril taut, your claws cut through it with ease,
                            your arm freed as the severed appendage limply falls to the ground.
                            The tentacle monster seems to recoil for a moment, as though hurt.
                            It was probably expecting a defenceless human maiden to walk into its trap instead of a crafty kobold such as yourself.
                        [/p]
                        [p]
                            With the thornbush receding, you briefly entertain the thought that you [i]won[/i] this battle.
                            Within moments, however, the plant already seems to have recovered from the loss of one of its graspers, and it begins its assault anew.
                            You barely dodge a second lunging tentacle trying to snag your non-bleeding wrist, 
                            and it is at that point you decide to [b]get the hell out of there[/b] while you still can.                           
                        [/p]
                        [p]
                            Reaching downwards, you quickly grab your ${global.kobold.inventory.createItemWithTooltip(new WoolSweater())} off the ground.
                            You're not leaving here without it, else this entire ordeal would have been for nothing!
                            You hop backwards to evade yet another springing tendril, yet accidentally step on one of them, small thorns digging into your rather soft sole.
                            Biting through the stinging pain, you turn around and [b]sprint[/b] away as fast as you can, 
                            a final tentacle lashing your back, yet failing to grab you as you bolt away from the animated thornbush.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.pry.claw.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.claw.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownClaw";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.pry.base.get() !== ForestUnknownPryChoice.CLAW) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.claw.change();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.pry.base.get() !== ForestUnknownPryChoice.CLAW) {
            return;
        }
        
        next.registerListeners();
        
        if (ctrl.unknown.pry.claw.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\claw\ForestUnknownClaw.js

// scenes\Forest\unknown\pry\claw\ForestUnknownClawContinue.js

/**
 * Claw section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownClawContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `runContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            For the second time today, you find yourself running for your life.
                            You don't know whether or not the plant is giving chase, but fright and adrenaline combined make you [b]keep running[/b] without looking back.
                            It isn't until the sun has gone down, that you come to a [b]completely[/b] exhausted stop.
                        [/p]
                        [p]
                            You slump down against a nearby tree, 
                            tightly clutching your ${global.kobold.inventory.createItemWithTooltip(new WoolSweater())} while loudly panting.
                            You missed this thing.
                            You really did.
                            You've had it for [b]years[/b].
                            It's served as a blanket, a garment and even a towel on rainy days.
                            It's the closest thing to a stuffed toy you've ever owned.
                            And with the temperature rapidly dropping, you are certain that you'll make good use of it once more, starting tonight.
                        [/p]
                        [p]
                            You check your wounds.
                            Your wrist has stopped bleeding, and while your sole still hurts a little, 
                            you reckon your foot is fit for walking on, even when [b]not[/b] fueled by sheer adrenaline.
                            Your back wasn't hurt all that badly.
                            The monster merely [i]grazed[/i] it while you were at the very edge of its reach.
                        [/p]
                        [p]
                            After taking a few more minutes to catch your breath and cuddle your sweater, you slowly rise to your feet once more.
                            You are tired.
                            [b]Very[/b] tired.
                            But you can't rest.
                            Not yet.
                            Not here.
                            It's too dangerous.
                            You take a look around.
                            You have no clue where you are.
                            All directions you gaze in look exactly the same.
                            Trees and darkness all around you.
                            You are officially [b]lost[/b].
                        [/p]
                        [p]
                            You slip on your ${global.kobold.inventory.createItemWithTooltip(new WoolSweater())}, which is slightly heated from the tender hug you gave it.
                            It fits snugly as ever, and it is bound to keep you warm throughout the night.
                            At least you won't have to worry about dying from hypothermia.
                            That just leaves about nine hundred and ninety nine other things that can kill you.
                        [/p]
                        [p]
                            Looks like you're going to have to pull an all nighter,
                            unless you can find a relatively safe place to catch some sleep in.
                            You're bound to run out of steam at some point, but that's a problem you'll face when it emerges.      
                            For now, you need to keep going.                      
                            As long as you keep heading in one direction, you'll reach the edge of the forest at one point or another, right?
                            You hope so, at least.
                        [/p]
                        [p]
                            With your body in shambles, but your spirit renewed, you begin to stride even deeper into the forest,
                            in hopes of ultimately finding your way back home, lest you [i]become[/i] one of the savages that live here, 
                            completely isolated from civilization.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.continue();
            loadNextPart("RiversideMarsh", "Deepwoods");
        });
    }


    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownClawContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.pry.claw.continued()) {
            return "";
        }
        
        global.ui.applyTheme("forest_night");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.claw.changeContinue();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.pry.claw.continued()) {
            return;
        }

        if (ctrl.unknown.hasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\claw\ForestUnknownClawContinue.js

// scenes\Forest\unknown\pry\still\bite\ForestUnknownBite.js

/**
 * Bite section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownBite(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `biteContinueButton`;
    var next = new ForestUnknownBiteContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            No.
                            While the urge to simply open your mouth and experience the taste of whatever sap the plant is spewing forth is great, you [b]don't[/b] give in.
                            You may not be the smartest of kobolds, but you know enough to realize that randomly tasting unknown substances is [b]bad[/b] for your health.
                            At best, it may simply be an alluring means to get you to suck the tentacle off.
                            At worst, it's [b]acid[/b] intended to [b]melt your insides[/b] so the monster can [b]eat[/b] you.
                        [/p]
                        [p]
                            The tendril's incessant attempts to get past your lips intensify.
                            It pushes and it prods and it presses against your sealed teeth, yet you [b]refuse[/b] to let it into your mouth.
                            Your brain, despite being addled by the strong, 
                            intoxicating scent which continues to build up as the tentacle slathers your lips with more of its syrup,
                            is working overtime to try and find a way out of this situation.
                        [/p]
                        [p]
                            You tentatively wriggle your limbs about, but any which way you move just ends up with your scales pressing against prickly thorns.
                            Your tail is the only part of yours that [i]isn't[/i] restrained, but unlike your claws, it isn't capable of slicing through vines.
                            You need something sharp.
                            Something that can cut through the plant's appendages.
                            Then it hits you.
                            Your teeth.
                        [/p]
                        [p]
                            You can [b]bite[/b] the plant, and sever its tendrils that way.
                            But then again, you can't reach any of them, barring the one pressing against your mouth.
                            In a way, the smooth tentacle seems far more [i]delicate[/i] than the thorny ones.
                            You wonder, could it perhaps be akin to the monster's genitals?
                            If so, [b]biting[/b] it [i]might[/i] prove painful enough to stun the creature, or perhaps even to knock it out cold.
                            If it feels pain in the first place, that is.
                            If it doesn't, it might just squeeze you [b]to death[/b].
                            But you'd rather die than get forced to suck off a plant anyway.
                        [/p]
                        [p]
                            Slowly, you open up your mouth, spreading your teeth and your lips both.
                            The impatient tendril immmediately finds its way into your mouth.
                            Then, before it has a chance to dull your senses with its potent honey, you brutally [b]chomp[/b] down,
                            your teeth sinking into the surprisingly flesh-like innards of the tentacle.
                            You pull your head to the side, practically [b]ripping[/b] off the small bit of the tender plant-part that had made it past your lips,
                            before immediately spitting out the chewed-off piece once more, fearing the effects of the sap that spews from its innards.
                        [/p]
                        [p]
                            The very instant you bite down, the entire organic terror recoils, 
                            its tendrils rapidly unfurling from your limbs, slinking back to the central mass.
                            The tentacle you bit falls lifelessly to the ground.
                            It can [b]definitely[/b] feel [b]pain[/b].
                        [/p]
                        [p]
                            With your hands freed, you quickly wipe your face of the monster's mind-boggling goop.
                            Still, this isn't quite over yet.
                            Hell-bent on revenge, or so you assume, the monstrosity assails you, flinging multiple thorn-covered appendages your way.
                            You duck, bob and weave to avoid most of the initial onslaught, but as one of the tendrils just [i]barely[/i] grazes your arm,
                            the resulting scratch acts like a signal, telling you to [b]get the hell out of there[/b]. 
                        [/p]
                        [p]
                            You quickly snag your ${global.kobold.inventory.createItemWithTooltip(new WoolSweater())} off of the ground,
                            before turning around and [b]bolting[/b] out of the small forest clearing.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.pry.bite.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.bite.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownBite";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.pry.stay.get() !== ForestUnknownStayStillChoice.BITE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.pry.stay.get() !== ForestUnknownStayStillChoice.BITE) {
            return;
        }
        
        next.registerListeners();

        if (ctrl.unknown.pry.bite.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\still\bite\ForestUnknownBite.js

// scenes\Forest\unknown\pry\still\bite\ForestUnknownBiteContinue.js

/**
 * Bite section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownBiteContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `runContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            For the second time today, you find yourself running for your life.
                            You don't know whether or not the plant is giving chase, but fright and adrenaline combined make you [b]keep running[/b] without looking back.
                            It isn't until the sun has gone down, that you come to a [b]completely[/b] exhausted stop.
                        [/p]
                        [p]
                            You slump down against a nearby tree, 
                            tightly clutching your ${global.kobold.inventory.createItemWithTooltip(new WoolSweater())} while loudly panting.
                            You missed this thing.
                            You really did.
                            You've had it for [b]years[/b].
                            It's served as a blanket, a garment and even a towel on rainy days.
                            It's the closest thing to a stuffed toy you've ever owned.
                            And with the temperature rapidly dropping, you are certain that you'll make good use of it once more, starting tonight.
                        [/p]
                        [p]
                            After taking a few more minutes to catch your breath and cuddle your sweater, you slowly rise to your feet once more.
                            You are tired.
                            [b]Very[/b] tired.
                            But you can't rest.
                            Not yet.
                            Not here.
                            It's too dangerous.
                            You take a look around.
                            You have no clue where you are.
                            All directions you gaze in look exactly the same.
                            Trees and darkness all around you.
                            You are officially [b]lost[/b].
                        [/p]
                        [p]
                            You slip on your ${global.kobold.inventory.createItemWithTooltip(new WoolSweater())}, which is slightly heated from the tender hug you gave it.
                            It fits snugly as ever, and it is bound to keep you warm throughout the night.
                            At least you won't have to worry about dying from hypothermia.
                            That just leaves about nine hundred and ninety nine other things that can kill you.
                        [/p]
                        [p]
                            Looks like you're going to have to pull an all nighter,
                            unless you can find a relatively safe place to catch some sleep in.
                            You're bound to run out of steam at some point, but that's a problem you'll face when it emerges.      
                            For now, you need to keep going.                      
                            As long as you keep heading in one direction, you'll reach the edge of the forest at one point or another, right?
                            You hope so, at least.
                        [/p]
                        [p]
                            With your body in shambles, but your spirit renewed, you begin to stride even deeper into the forest,
                            in hopes of ultimately finding your way back home, lest you [i]become[/i] one of the savages that live here, 
                            completely isolated from civilization.
                            If these woods think a little bit of sap is all it takes for you to let your guard down, they can think again.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.continue();
            loadNextPart("RiversideMarsh", "Deepwoods");
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownBiteContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.pry.bite.continued()) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.bite.change();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.pry.bite.continued()) {
            return;
        }

        if (ctrl.unknown.hasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\still\bite\ForestUnknownBiteContinue.js

// scenes\Forest\unknown\pry\still\suck\expose\ForestUnknownExpose.js

/**
 * Expose section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownExpose(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `exposeContinueButton`;
    var next = new ForestUnknownExposeContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cock = ctrl.unknown.cock();
        var pants = ctrl.unknown.pants();
        var pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        var underwear = ctrl.unknown.underwear();
        var underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        var toReturn = `[p]
                            You know fully well what it intends to do, and you have no plans of stopping it.
                            Your ${cock ? "member is fully erect" : "slit is dripping wet"}, practically [b]begging[/b] for attention.
                            Perhaps the tendril can sate your growing lust, your intensifying [b]need[/b], your [b]ache[/b] to be [b]touched[/b].
                        [/p]`;

        if (cock) {
            toReturn += `[p]
                            While the tentacle in your mouth continues to deftly pump in and out of your muzzle,
                            oozing a constant stream of aphrodisiac which you greedily gulp down,
                            the one in between your legs inches in between your buttcheeks, towards your tender tailhole.
                            While you would have preferred it to wrap around your length instead,
                            you don't complain as it begins to prod around your backside, circling around your hole,
                            spreading more of its slick, golden goop, undoubtedly lubing you up, preparing you for what is to come.
                        [/p]
                        [p]
                            While you've never quite touched yourself [i]in that way[/i],
                            even the teasing slathering of sap feels [b]heavenly[/b],
                            every slight stroke magnified tenfold due to the large amount of syrup you continue to ingest,
                            your body far more sensitive than usual.
                            Precum oozes from the tip of your member.
                            Still, it's not enough.
                            You need [b]more[/b].
                            You stick out your rear as best you can.
                            Raising your tail up high, you enticingly begin to shake your butt from side to side,
                            hoping to allure the monster into [b]fucking[/b] you.
                        [/p]
                        [p]
                            Your efforts are not in vain.
                            After observing your lust-fueled self-degradation for a few moments,
                            the monster decides to indulge your need.
                            In a single, smooth thrust, the smaller tentacle slides into your well-lubed rear,
                            the sensation [b]instantly[/b] driving you over the edge.
                            You let out a loud moan around the tendril in your mouth as you reach your end, 
                            instinctively bucking your hips forwards which incidentally also makes you [i]ride[/i] the tentacle up your butt as you shoot your thick,
                            sticky load all over the plant's thorny exterior, coating its brownish green tendrils a pearly white.
                        [/p]
                        [p]
                            Your climax seems to inspire the monster.
                            The tendril in your mouth ups its pace, outright [b]humping[/b] your muzzle,
                            whilst the one lodged up your ass immediately matches the other's pace, pounding your rear.
                            Your aphrodisiac-addled brain is barely able to [b]process[/b] the sheer amount of [b]pleasure[/b] you experience.
                            Your tongue lolls out of your mouth obscenely, and you loudly [b]scream[/b] in ecstasy as you are thoroughly taken from both ends.
                            Is this what [b]heaven[/b] feels like?
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/unknown/pry/still/suck/expose/consentacle-m-full.png" target="_blank">
                                    <img class="nsfw wideimage" src="scenes/Forest/unknown/pry/still/suck/expose/consentacle-m.png" title="Illustration by Feldspar">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            Sadly, the experience does not last long.
                            Within less than a minute, you can feel both of the tentacles beginning to quiver, pulse and spasm,
                            tangibly [b]pumping[/b] more of the plant's divine fluid into your body.
                            Your insides get filled to the very brim with the golden goo, as does your mouth, several times over,
                            but you absent-mindedly swallow every fresh wave of the delicious floral cum 
                            until your belly bulges ever so slightly from the huge amount you've ingested.
                        [/p]`;
        } else {
            toReturn += `[p]
                            While the tentacle in your mouth continues to deftly pump in and out of your muzzle,
                            oozing a constant stream of aphrodisiac which you greedily gulp down,
                            the one in between your legs inches towards your moist cunt.
                            As though it realizes your need, it begins to [b]toy[/b] with you,
                            tapered tip sliding along the entire length of your slit, parting your folds ever so slightly.
                        [/p]
                        [p]
                            Even these tiniest of touches are enough to send [b]spikes[/b] of pleasure down your spine,
                            your body made [b]far[/b] more sensitive by the large amount of syrup pumped into your mouth.
                            Still, it's not enough.
                            You wildly buck your hips against the prodding tendril as much as your thorny restraints allow,
                            in an attempt to spear yourself on its length, to fill yourself with its girth.
                            It is only after observing your frantic, lust-fueled attempts at riding its tentacle for a few moments,
                            that the monster decides to indulge your need.
                        [/p]
                        [p]
                            In a single, smooth thrust, the tentacle slides into your sopping wet cunt, spreading your folds and tearing your hymen.
                            While it isn't quite as [b]thick[/b] as you hoped it would be, 
                            the subtle mix between overwhelming pleasure and just a slight tinge of [i]pain[/i] immediately sends you over the edge.
                            Your entire body trembles, your back arches and your knees grow weak.
                            You let out a loud, pleasured moan around the tendril in your mouth as you reach your end,
                            a large amount of feminine fluids gushing out of your slit, running down the entire length of the thin tentacle,
                            all the way to the ground.
                        [/p]
                        [p]
                            Your climax seems to inspire the monster.
                            The tendril in your mouth ups its pace, outright [b]fucking[/b] your muzzle,
                            whilst the one lodged up your tight snatch immediately matches the other's pace, pounding your cunt.
                            Your aphrodisiac-addled brain is barely able to [b]process[/b] the sheer amount of [b]pleasure[/b] you experience.
                            Your tongue lolls out of your mouth obscenely, and you loudly [b]scream[/b] in ecstasy as you are thoroughly taken from both ends.
                            Is this what [b]heaven[/b] feels like?
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/unknown/pry/still/suck/expose/consentacle-f-full.png" target="_blank">
                                    <img class="nsfw wideimage" src="scenes/Forest/unknown/pry/still/suck/expose/consentacle-f.png" title="Illustration by Feldspar">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            Sadly, the experience does not last long.
                            Within less than a minute, you can feel both of the tentacles beginning to quiver, pulse and spasm,
                            tangibly [b]pumping[/b] more of the plant's divine fluid into your body.
                            Your cunt gets filled to the very brim with the golden goo, as does your mouth, several times over,
                            but you absent-mindedly swallow every fresh wave of the delicious floral cum 
                            until your belly bulges ever so slightly from the huge amount you've ingested.
                        [/p]`;
        }

        toReturn += `   [p]
                            Sated, the monstrosity removes its tentacles from your orifices, a few final squirts of sap splattering across your face and your loins.
                            As soon as both of the smooth tendrils have retreated back behind the shield of brambles, 
                            the animated thornbush slowly unwraps your restrained limbs, and begins to inch away from the small forest clearing.
                        [/p]
                        [p]
                            The moment your arms are released, you immediately begin to pleasure yourself once more,
                            the large amount of aphrodisiac still lingering in your body arousing you to no end.
                            Your loud moans echo throughout the deepwoods, as you continue playing with yourself until long after the sun has set,
                            and the tentacle monster seems like a distant memory, having vanished into the forest.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    You are no longer an oral virgin.<br/>`;
        if (cock) {
            toReturn += `           You are no longer an anal virgin.`;
        } else {
            toReturn += `           You are no longer a virgin.`;
        }
        toReturn += `           [/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.pry.expose.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.expose.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownExpose";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.pry.suck.get() !== ForestUnknownSuckChoice.EXPOSE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.expose.change();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.pry.suck.get() !== ForestUnknownSuckChoice.EXPOSE) {
            return;
        }

        next.registerListeners();

        if (ctrl.unknown.pry.expose.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\still\suck\expose\ForestUnknownExpose.js

// scenes\Forest\unknown\pry\still\suck\expose\ForestUnknownExposeContinue.js

/**
 * Expose section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownExposeContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `woodsContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cock = ctrl.unknown.cock();
        var pants = ctrl.unknown.pants();
        var pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        var underwear = ctrl.unknown.underwear();
        var underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        var toReturn = `[p]
                            You awaken from your lust-induced ${cock ? "rut" : "heat"}-like masturbation [b]craze[/b] hours later,
                            both your hands, the ground and your tail all entirely
                            ${cock ? "[b]coated[/b] in your spunk" : "[b]soaked[/b] in your excitement"}.
                            ${cock ? "Two of your fingers are toying with your own behind, while your other hand is firmly wrapped around your length." : ""}
                            You find yourself panting and huffing loudly, exerted from intensively touching yourself over and over again.
                            Coming to your senses, you 
                            ${cock ? "slide your fingers out of your rear, " : ""} 
                            clean yourself of any remaining sap using your palm, 
                            ${pants ? `pull up your ${pantsPopup},` : ""} 
                            wipe your hands off on a nearby patch of moss,
                            and take a minute to catch your breath.
                        [/p]
                        [p]
                            You can't recall quite how many times you came, but you feel completely [b]spent[/b].
                            Whatever that monster fed you, it must have been [b]extremely[/b] potent.
                            But then again, it did pump [b]a lot[/b] of its honey into you.
                            Due to the sheer amount of it lingering in your stomach, you feel as though you've just had a hearty meal.
                        [/p]
                        [p]
                            Slowly pulling your mind out of the gutter, you spot something out of the corner of your eye: 
                            your ${global.kobold.inventory.createItemWithTooltip(new WoolSweater)}, the entire reason you interacted with the thornbush in the first place!
                            You quickly don it, to protect yourself from the creeping cold of night.
                            With your lust sated, and your body at least somewhat protected from the freezing temperature, you begin to ponder about your current situation.
                        [/p]
                        [p]
                            You doubt you can find your way home in the dark.
                            You are tired, but sleeping in a place this dangerous is [b]not[/b] an option.
                            Your best bet is to keep moving.
                            If you keep going in one direction, you'll reach the edge of this godforsaken forest at one point or another, 
                            and [i]then[/i] you can go home.
                        [/p]
                        [p]
                            Your plan formed, you pick a direction at random, and begin to march forwards,
                            continuing to wipe your loins clean as you stride into the deepwoods.
                            While that was [i]fun[/i], you hope you won't encounter any more tentacle monsters, at least not in the nearby future.
                            You doubt you can weather another self-pleasuring session like that without some time to recover.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.continue();
            loadNextPart("RiversideMarsh", "Deepwoods");
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownExposeContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.pry.expose.continued()) {
            return "";
        }

        global.ui.applyTheme("forest_night");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.expose.continueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.pry.expose.continued()) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.unknown.hasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\still\suck\expose\ForestUnknownExposeContinue.js

// scenes\Forest\unknown\pry\still\suck\tail\ForestUnknownTail.js

/**
 * Tail section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownTail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `unknownTailContinueButton`;
    var next = new ForestUnknownTailContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cock = ctrl.unknown.cock();
        var pants = ctrl.unknown.pants();
        var pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        var underwear = ctrl.unknown.underwear();
        var underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        var toReturn = `[p]
                            Not wanting to lose your virginity to a random plant, no matter how [b]absolutely delectable[/b] its sap,
                            you press your unrestrained tail in between your legs, covering up your loins.
                            The smaller tentacle, upon reaching your inner thighs, begins to probe your private parts, intently looking for a hole to fuck.
                            Instead of finding what it seeks, it only ends up bumping against your smooth, scaly tail over and over again,
                            until ultimately resigning, slinking back down the ground, and crawling back to the thornbush's core.
                        [/p]
                        [p]
                            The monster seems aggravated about being unable to find a hole to fuck.
                            The tentacle in your mouth grows more [b]violent[/b], more [b]fierce[/b], its thrusts intensifying and deepening,
                            the tip bumping against the back of your throat more than once,
                            sometimes even extending far enough to squirt a globule of honey directly down into your stomach.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/unknown/pry/still/suck/tail/tail-${cock ? "m" : "f"}-full.png" target="_blank">
                                    <img class="nsfw wideimage" src="scenes/Forest/unknown/pry/still/suck/tail/tail-${cock ? "m" : "f"}.png" title="Illustration by Feldspar">
                                </a>
                            [/c]
                        [/p]`;

        if (cock) {
            toReturn += `[p]
                            During one of the tentacle's deeper thrusts, you accidentally shift your tail around,
                            bumping it into your needy, fully erect member, sending shivers of pleasure down your spine.
                            As the tendril pulls back, you loudly moan around it, your own precum staining the scaly underside of your only unbound appendage.
                            Then, an idea forms in your lust-clouded mind.
                            Your arms are tightly restrained, so you can't quite [i]touch[/i] yourself, no matter how much you want, no [b]need[/b] the pleasure.
                            Your tail, however, is not.
                        [/p]
                        [p]
                            You wrap the tip of the prehensile appendage around your shaft, and clumsily begin to ${cock ? "stroke" : "rub"} yourself,
                            your pleasure multiplied tenfold due to the constant influx of aphrodisiac.
                            While you aren't able to [i]properly[/i] jerk off, 
                            even the slightest bit of pumping you manage to achieve using nothing but your tail is still more than plenty to rapidly bring you over the edge.
                            Bucking your hips as much as your thorny restraints allow you to,
                            you tauntingly squirt your hefty load of thick kobold spunk all over the thornbush's exterior,
                            coating it a pearly white as it coated your muzzle a golden yellow.
                        [/p]`;
        } else {
            toReturn += `[p]
                            During one of the tentacle's deeper thrusts, you accidentally shift your tail around,
                            bumping it into your needy, sopping wet slit, sending shivers of pleasure down your spine.
                            As the tendril pulls back, you loudly moan around it, your own feminine fluids staining the scaly underside of your only unbound appendage.
                            Then, an idea forms in your lust-clouded mind.
                            Your arms are tightly restrained, so you can't quite [i]touch[/i] yourself, no matter how much you want, no [b]need[/b] the pleasure.
                            Your tail, however, is not.
                        [/p]
                        [p]
                            You deftly press the prehensile appendage in between your folds, and begin to stroke yourself,
                            your pleasure multiplied tenfold due to the constant influx of aphrodisiac.
                            Still, it isn't enough.
                            Then, you slide the very tip of your tail into your lust-craving cunt,
                            the feeling of being [b]filled[/b] instantly sending you over the edge.
                            You arch your back as your excitement [b]gushes[/b] out of your slit,
                            leaking out onto your tail before dripping down to the ground.
                        [/p]`;
        }

        toReturn += `   [p]
                            Smelling, seeing or otherwise [b]sensing[/b] your orgasm, the creature's tentacle shoots into overdrive.
                            It rapidly pumps itself in and out of your muzzle, pressing into your throat with every firm thrust.
                            A steady stream of sap oozes from the tip, as if to pacify you while the monster grows ever more rough,
                            its tentacle digging deeper, holding itself down your throat for longer stretches at a time,
                            preferring the tightness of your oesophagus over the warm wetness of your mouth.
                        [/p]
                        [p]
                            Finally, it forces itself just a [i]little[/i] deeper into your muzzle than ever before.
                            Holding itself in place deep down your throat, you can feel the tentacle begin to quiver, pulse and spasm inside of your mouth,
                            tangibly [b]pumping[/b] wave after wave of thick, sticky, warm sap directly into your stomach,
                            filling you up to the very [b]brim[/b] before retracting,
                            your belly bulging ever so slightly as a few more shallow squirts in your mouth fill your need for more of the heavenly taste.
                        [/p]
                        [p]
                            Fully spent, the creature withdraws from your mouth.
                            Its grip on your limbs loosening, it contently begins to slither away.
                            The moment your arms are released, you immediately begin to stroke yourself,
                            the large amount of aphrodisiac still lingering in your body arousing you to no end.
                            Your loud moans echo throughout the deepwoods, as you continue playing with yourself until long after the sun has set,
                            and the tentacle monster seems like a distant memory, having vanished into the forest.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    You are no longer an oral virgin.
                                [/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.pry.tail.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.tail.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownTail";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.pry.suck.get() !== ForestUnknownSuckChoice.TAIL) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.tail.change();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.pry.suck.get() !== ForestUnknownSuckChoice.TAIL) {
            return;
        }

        next.registerListeners();

        if (ctrl.unknown.pry.tail.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\still\suck\tail\ForestUnknownTail.js

// scenes\Forest\unknown\pry\still\suck\tail\ForestUnknownTailContinue.js

/**
 * Tail section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownTailContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `woodsContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cock = ctrl.unknown.cock();
        var pants = ctrl.unknown.pants();
        var pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        var underwear = ctrl.unknown.underwear();
        var underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        var toReturn = `[p]
                            You awaken from your lust-induced ${cock ? "rut" : "heat"}-like masturbation [b]craze[/b] hours later,
                            both your hands, the ground and your tail all entirely
                            ${cock ? "[b]coated[/b] in your spunk" : "[b]soaked[/b] in your excitement"}.
                            You find yourself panting and huffing loudly, exerted from intensively touching yourself over and over again.
                            Coming to your senses, you ${pants ? `pull up your ${pantsPopup},` : ""} wipe your hands off on a nearby patch of moss,
                            and take a minute to catch your breath.
                        [/p]
                        [p]
                            You can't recall quite how many times you came, but you feel completely [b]spent[/b].
                            Whatever that monster fed you, it must have been [b]extremely[/b] potent.
                            But then again, it did pump [b]a lot[/b] of its honey into you.
                            Due to the sheer amount of it lingering in your stomach, you feel as though you've just had a hearty meal.
                        [/p]
                        [p]
                            Slowly pulling your mind out of the gutter, you spot something out of the corner of your eye: 
                            your ${global.kobold.inventory.createItemWithTooltip(new WoolSweater)}, the entire reason you interacted with the thornbush in the first place!
                            You quickly don it, to protect yourself from the creeping cold of night.
                            With your lust sated, and your body at least somewhat protected from the freezing temperature, you begin to ponder about your current situation.
                        [/p]
                        [p]
                            You doubt you can find your way home in the dark.
                            You are tired, but sleeping in a place this dangerous is [b]not[/b] an option.
                            Your best bet is to keep moving.
                            If you keep going in one direction, you'll reach the edge of this godforsaken forest at one point or another, 
                            and [i]then[/i] you can go home.
                        [/p]
                        [p]
                            Your plan formed, you pick a direction at random, and begin to march forwards,
                            wiping your loins clean as you stride into the deepwoods.
                            You hope you won't encounter any more tentacle monsters, at least not in the nearby future.
                            You doubt you can weather another self-pleasuring session like that without some time to recover.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.continue();
            loadNextPart("RiversideMarsh", "Deepwoods");
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownTailContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.pry.tail.continued()) {
            return "";
        }

        global.ui.applyTheme("forest_night");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.tail.continueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.pry.tail.continued()) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.unknown.hasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\pry\still\suck\tail\ForestUnknownTailContinue.js

// scenes\Forest\unknown\pry\still\suck\ForestUnknownSuck.js

/**
 * Suck section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownSuck(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var exposeId = "unknownExposeButton";
    var tailId = "unknownTailButton";

    var toExpose = new ForestUnknownExpose(ctrl);
    var toTail = new ForestUnknownTail(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cock = ctrl.unknown.cock();
        var pants = ctrl.unknown.pants();
        var pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        var underwear = ctrl.unknown.underwear();
        var underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        var toReturn = `[p]
                            You slowly part your lips, just to lick off the honey slathered across the front-end of your muzzle,
                            but you hardly protest as the tentacle uses the opportunity to slip inside of the warm, wet confines of your mouth.
                            For being a good ${cock ? "boy" : "girl"}, you are instantly rewarded with a hefty [b]squirt[/b] of the heavenly goop,
                            which splatters along the entire length of your tongue, completely [b]saturating[/b] your tastebuds with its intense, sweet flavor.
                        [/p]
                        [p]
                            Your eyes roll into the back of your head as you discover that not only is this sap the [b]best[/b] thing you've [b]ever[/b] tasted,
                            it also acts as a potent aphrodisiac, 
                            your ${cock ? "member instantly springing to life, its tip emerging from your cockslit" : "slit instantly dampening, the urge to touch yourself rising with every passing second"}.
                            You moan out around the tentacle in your mouth, which slowly begins to thrust in and out of your muzzle,
                            outright rubbing its magnificent produce onto your tongue.
                            You eagerly close your mouth around it, hoping to [b]milk[/b] it of as much of the absolutely [b]divine[/b] substance as you can.
                        [/p]
                        [p]
                            While you continue to suckle the occasional addictive squirt of syrup from the tendril in your mouth,
                            another thornless tentacle inches out from underneath the central mass of the thornbush,
                            this one significantly smaller than the one you're suckling.
                            It makes its way to your legs, before slithering upwards towards your waist.
                            Even while completely befuddled by the monster's [b]delicious[/b], brain-numbing sexual stimulant,
                            your mind is still clear enough to realize its intentions.
                            ${pants ? `And as if to reinforce your suspicions, the tendril deftly [b]yanks[/b] down your ${pantsPopup}, 
                                    exposing your loins for the entire forest to see.` : ""}
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildExposeChoice() {
        var disabled = ctrl.unknown.pry.suck.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.pry.suck.get() === ForestUnknownSuckChoice.EXPOSE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${exposeId}" ${disabled} style="font-size: 18px;">Expose yourself</button>`;

        var tooltip = `Let the monster have its way with you.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 150px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildTailChoice() {
        var disabled = ctrl.unknown.pry.suck.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.pry.suck.get() === ForestUnknownSuckChoice.TAIL;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${tailId}" ${disabled}>Cover yourself</button>`;

        var tooltip = `If you use your tail to cover your loins, you might be able to fool the plant.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildExposeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildTailChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + exposeId), $("#" + tailId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerExposeListener() {
        $("#" + exposeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.suck.set(ForestUnknownSuckChoice.EXPOSE);
            loadNextPart(toExpose);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerTailListener() {
        $("#" + tailId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.suck.set(ForestUnknownSuckChoice.TAIL);
            loadNextPart(toTail);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownSuck";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.pry.stay.get() !== ForestUnknownStayStillChoice.SUCK) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.suck.change();
        part += buildChoice();
        part += toExpose.build();
        part += toTail.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.pry.stay.get() !== ForestUnknownStayStillChoice.SUCK) {
            return;
        }

        toExpose.registerListeners();
        toTail.registerListeners();

        if (ctrl.unknown.pry.suck.has()) {
            return;
        }

        registerExposeListener();
        registerTailListener();
    };
}

var ForestUnknownSuckChoice = {
    EXPOSE: 1,
    TAIL: 2
};

// END scenes\Forest\unknown\pry\still\suck\ForestUnknownSuck.js

// scenes\Forest\unknown\pry\still\ForestUnknownStayStill.js

/**
 * Stay still section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownStayStill(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var suckId = "unknownSuckButton";
    var biteId = "unknownStillBiteButton";

    var toSuck = new ForestUnknownSuck(ctrl);
    var toBite = new ForestUnknownBite(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You freeze up.
                            You don't move, not even an inch.
                            Maybe the bush thinks you're a threat.
                            Or maybe it can't see you unless you're moving.
                            Perhaps, if you stay perfectly still, the plant will just let you go.
                        [/p]
                        [p]
                            Sadly, you are mistaken.
                            Instead of releasing the grip on your wrist, more thorny vines wrap around your other limbs.
                            Two of them simultaneously snake onto your ankles, keeping your legs in place by the threat of puncturing your scales with thick thorns.
                            Another tendril wraps around your already-restricted arm, reinforcing the creature's hold on it.
                            Your free arm, too, gets snagged by one of the thorny appendages, leaving you completely immobilized, save for your head and your tail.
                        [/p]
                        [p]
                            Your resolve is wavering, yet still, you convince yourself that if you just remain calm, 
                            and not start pointlessly struggling against your spiky restraints, everything will work out [i]just fine[/i].
                            Plants don't eat meat, [i]you hope[/i], so as long as you don't [i]fight[/i] the monstrosity before you, 
                            surely it will have no reason whatsoever to [i]hurt you[/i], right?
                        [/p]
                        [p]
                            The central mass of the hulking monster slowly creeps closer, until it is barely a foot away from your fully restricted body.
                            As it stops, a single tentacle slips out from underneath the thorny exterior, one very much so [i]unlike[/i] the others.
                            While the other tendrils are covered in spikes and brambles, this one is perfectly [i]smooth[/i].
                            In fact, it appears [i]slick[/i], almost [i]shiny[/i], as though covered in a viscous liquid of some sort.
                            In color, too, it differs from the rest.
                            While the normal vines are a brownish green, this one is far more [b]vibrant[/b], a bright shade of green, borderline unnatural.
                        [/p]
                        [p]
                            It slowly slithers over the ground, towards your feet, before rising upwards, once again acting like a snake, this time one being charmed.
                            The long, slick tentacle lifts itself all the way up to your face, hovering directly in front of you.
                            It smells sickly sweet, like an exotic kind of honey you once stole a jar of.
                            Getting a closer look, you notice a small opening at the very tip of the oddly shaped plant-part, like a plant bud that hasn't bloomed yet.
                            Oozing from it at a steady pace is the same thick liquid you assume covers the rest the appendage's length.
                        [/p]
                        [p]
                            The tendril nears your face, hovering directly in front of your muzzle, mere inches away from your snout.
                            Then, it squirts a hefty rope of the sticky, golden-brown goo across your muzzle, 
                            splattering atop your scales all the way from the tip of your nose to in between your eyes.
                            A mouth-watering, overly-dominant fragrance of sheer honeyed [b]sweetness[/b] wafts from the syrupy substance,
                            completely [b]clogging[/b] your nostrils, filling up your lungs and clouding your mind.
                            It smells [b]so good[/b], it makes you want to [b]taste[/b] it.
                        [/p]
                        [p]
                            Just as you were about to open your mouth to try and snake your tongue to the top of your muzzle in an attempt to sample the goop,
                            the tentacle presses its tip against your lips.
                            It seductively drags itself from one side of your muzzle to the other, leaving a trail of mind-altering honey in its wake.
                            If you want to taste it, all you have to do is open your mouth and lick your lips.
                            Heck, why stop there?
                            Why not [b]suck[/b] the syrup directly from the tentacle,
                            so you can relish in as much of the undoubtedly [b]heavenly[/b] flavour as you could ever want?
                        [/p]
                        [p]
                            As though the monstrosity can read your mind,
                            the tendril positions itself at the front-end of your muzzle,
                            and slowly begins to press into your mouth.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSuckChoice() {
        var disabled = ctrl.unknown.pry.stay.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.pry.stay.get() === ForestUnknownStayStillChoice.SUCK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${suckId}" ${disabled}>Suck it</button>`;

        var tooltip = global.parser.parse(`You can't resist. You [b]need[/b] to know how it tastes.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildBiteChoice() {
        var disabled = ctrl.unknown.pry.stay.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.pry.stay.get() === ForestUnknownStayStillChoice.BITE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${biteId}" ${disabled}>Bite it</button>`;

        var tooltip = global.parser.parse(`Don't give in. No matter how enticing it may be, there's no telling what that goop might do to you!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildSuckChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildBiteChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + suckId), $("#" + biteId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerSuckListener() {
        $("#" + suckId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.stay.set(ForestUnknownStayStillChoice.SUCK);
            loadNextPart(toSuck);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerBiteListener() {
        $("#" + biteId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.stay.set(ForestUnknownStayStillChoice.BITE);
            loadNextPart(toBite);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownStayStill";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.pry.base.get() !== ForestUnknownPryChoice.STILL) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.stay.change();
        part += buildChoice();
        part += toBite.build();
        part += toSuck.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.pry.base.get() !== ForestUnknownPryChoice.STILL) {
            return;
        }

        toBite.registerListeners();
        toSuck.registerListeners();

        if (ctrl.unknown.pry.stay.has()) {
            return;
        }

        registerSuckListener();
        registerBiteListener();
    };
}

var ForestUnknownStayStillChoice = {
    SUCK: 1,
    BITE: 2
};

// END scenes\Forest\unknown\pry\still\ForestUnknownStayStill.js

// scenes\Forest\unknown\pry\ForestUnknownPry.js

/**
 * Unknown section of the forest scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownPry(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var clawId = "unknownPryClawButton";
    var biteId = "unknownPryInitialBiteButton";
    var stillId = "unknownPryStillButton";

    var toClaw = new ForestUnknownClaw(ctrl);
    var toBite = new ForestUnknownInitialBite(ctrl);
    var toStill = new ForestUnknownStayStill(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.unknown.cock();
        var toReturn = `[p]
                            You approach the spiky plant, taking care [i]not[/i] to step on any low-hanging tendrils with your bare feet.
                            Your shirt, hanging off the side of the decently-sized shrub, is impaled by several of its organic barbs, 
                            all the way from the end of one sleeve to the very collar.
                            Ignoring the fact that you don't quite recall getting entangled in the thornbush [i]that[/i] badly,
                            you deftly get to work freeing your garment from nature's fabric-ruining grasp.
                        [/p]
                        [p]
                            You painstakingly unhook your sweater from the myriad of thorns it is caught on, one at a time.
                            Luckily, each spike individually isn't all that hard to remove.
                            Due to the lack of jagged edges, you can simply pluck the thick woolen cloth away from the brambles without damaging it all too much.
                        [/p]
                        [p]
                            After a solid few minutes of concentrated effort, you've managed to completely remove your shirt from the bush.
                            Satisfied with your work, you hold the garment up to the sky with both your hands, 
                            one of the few rays of sunlight beaming through the canopy illuminating the sweater just enough for you to inspect it a bit more closely.
                            Apart from a few, hardly noticeable pinpoint pricks marking where the thorns used to be, it seems to be in perfect condition.
                        [/p]
                        [p]
                            With your gaze directed at your white woolen sweater, and with glee washing over your mind,
                            you fail to notice the many tendrils surrounding you slowly coming to life.
                            Just as you were about to proclaim this your [i]luckiest day [b]ever[/b][/i],
                            one of the thorny tentacles stretches out towards you, uncoiling like a snake leaping at its prey.
                            Caught completely off guard, your heart skips a beat as the surprisingly flexible, 
                            spike-covered plant part wraps itself around one of your wrists in the flash of an eye.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/unknown/pry/pry-${dick ? "m" : "f"}-full.png" target="_blank">
                                    <img class="nsfw wideimage" src="scenes/Forest/unknown/pry/pry-${dick ? "m" : "f"}.png" title="Illustration by Feldspar">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            You let out an audible, frightened gasp, the shirt falling from your hands onto the ground out of sheer astonishment.
                            Your first reflex is to yank your arm back, but this causes the thorns covering the tendril to press painfully against your scales.
                            You only barely manage to suppress your natural bodily response before drawing your own blood,
                            the tentacle completely unwavering in its determination to hold your arm in place.
                            A mixture of pain and fear causes adrenaline to rapidly flood your system, and time itself seems to slow down.
                            Even then, it takes your brain a few seconds to [b]process[/b] what exactly just happened.
                        [/p]
                        [p]
                            The thornbush seems to be [i]alive[/i], and intent on holding you in place.
                            Besides the single tendril already tightly wrapped around your wrist,
                            the ones on the ground surrounding you are creeping towards your feet, undoubtedly to further restrain you.
                            You've no clue as to what the plant's intentions are, but judging by the painful grip it has on your arm, 
                            you'd guess that it either wants to [b]eat you[/b] or...[i]do something else[/i].
                            You're not sure if you like either of those options.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the ant choice button.
     */
    function buildClawChoice() {
        var disabled = ctrl.unknown.pry.base.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.pry.base.get() === ForestUnknownPryChoice.CLAW;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${clawId}" ${disabled}>Swipe</button>`;

        var tooltip = `Use the sharp claws of your free hand in an attempt to cut the tendril clutching your arm.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildBiteChoice() {
        var disabled = ctrl.unknown.pry.base.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.pry.base.get() === ForestUnknownPryChoice.BITE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${biteId}" ${disabled}>Bite</button>`;

        var tooltip = `Attempt to sever the thorny tentacle using your pointed teeth.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }
    
    /**
     * Builds and returns the shake choice button.
     */
    function buildStillChoice() {
        var disabled = ctrl.unknown.pry.base.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.pry.base.get() === ForestUnknownPryChoice.STILL;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${stillId}" ${disabled}>Stay still</button>`;

        var tooltip = `Stay perfectly still.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 140px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildClawChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildBiteChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildStillChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + clawId), $("#" + biteId), $("#" + stillId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerClawListener() {
        $("#" + clawId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.base.set(ForestUnknownPryChoice.CLAW);
            loadNextPart(toClaw);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerBiteListener() {
        $("#" + biteId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.base.set(ForestUnknownPryChoice.BITE);
            loadNextPart(toBite);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerStillListener() {
        $("#" + stillId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.pry.base.set(ForestUnknownPryChoice.STILL);
            loadNextPart(toStill);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownPry";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.base.get() !== ForestUnknownChoice.PRY) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.pry.base.change();
        part += buildChoice();
        part += toClaw.build();
        part += toBite.build();
        part += toStill.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.base.get() !== ForestUnknownChoice.PRY) {
            return;
        }

        toClaw.registerListeners();
        toBite.registerListeners();
        toStill.registerListeners();

        if (ctrl.unknown.pry.base.has()) {
            return;
        }

        registerClawListener();
        registerBiteListener();
        registerStillListener();
    };
}

var ForestUnknownPryChoice = {
    CLAW: 1,
    BITE: 2,
    STILL: 3
};

// END scenes\Forest\unknown\pry\ForestUnknownPry.js

// scenes\Forest\unknown\pry\ForestUnknownPryController.js

/**
 * Controller for the unknown part of the forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestUnknownPryController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;

    var baseId = "unknownPryBase";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    this.base = new BaseController(baseId);
    this.initbite = new InitialBiteController("unknownPryBite");
    this.initbiterape = new InitialBiteRapeController("unknownPryBiteRape");
    this.stay = new StayStillController("unknownStayStill");
    this.suck = new SuckController("unknownSuck");
    this.tail = new TailController("unknownTail");
    this.expose = new ExposeController("unknownExpose");
    this.claw = new ClawController("unknownClaw");
    this.bite = new BiteController("unknownBite");

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    function BaseController(key) {
        var changeId = key + "Change";
        this.get = function () {
            return global.kobold.history.getKey(id, key);
        };
        this.set = function (choice) {
            global.kobold.history.setKey(id, key, choice);
        };
        this.has = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.physical.decrease(5, "Painful thorns");
            });
            return global.ui.change.buildNotification([
                new Change("Painful thorns", "Health", -5)
            ]);
        };
    }

    function InitialBiteController(key) {
        var changeId = key + "Change";
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.physical.decrease(15, "Struggling against thorns");
                global.kobold.status.mental.decrease(15, "Impending rape");
            });
            return global.ui.change.buildNotification([
                new Change("Body encased in thorns", "Health", -5),
                new Change("Struggling against thorns", "Health", -10),
                new Change("Lost fight", "Sanity", -10),
                new Change("Impending rape", "Sanity", -5)
            ]);
        };
    }
    
    function InitialBiteRapeController(key) {
        var changeId = key + "Change";
        var continueChangeId = key + "ContinueChange";
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function (pants) {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.body.analVirgin = false;
                global.kobold.status.physical.decrease(5, "Stretched rear");
                global.kobold.status.mental.decrease(25, "Defiled by monster");
                global.kobold.status.exertion.increase(25, "Tentacle buttsex");
                global.kobold.inventory.equipment.strip(SlotEnum.PANTS);
                if(global.kobold.inventory.equipment.underwear) {
                    global.kobold.inventory.equipment.underwear.stain("Honey");
                }
            });
            
            var changes = [
                new Change("Stretched rear", "Health", -5),
                new Change("Defiled by monster", "Sanity", -25),
                new Change("Tentacle buttsex", "Sexual exhaustion", +25, true)
            ];
            
            if(pants) {
                changes.push(new Change("Pants torn", global.kobold.inventory.createItemWithTooltip(pants, true), -1));
            }
            
            return global.ui.change.buildNotification(changes);
        };
        
        this.continueChange = function () {
            global.kobold.history.doIfNotSet(id, continueChangeId, true, function () {
                global.kobold.status.mental.increase(5, "Sheer determination");
                global.kobold.inventory.equipment.wear(new WoolSweater());
            });
            return global.ui.change.buildNotification([
                new Change("Recovered sweater", global.kobold.inventory.createItemWithTooltip(new WoolSweater()), +1),
                new Change("Doubling down", "Sanity", +5)
            ]);
        };
    }

    function StayStillController(key) {
        var changeId = key + "Change";
        this.get = function () {
            return global.kobold.history.getKey(id, key);
        };
        this.set = function (choice) {
            global.kobold.history.setKey(id, key, choice);
        };
        this.has = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.decrease(5, "Mind-boggling aroma");
            });
            return global.ui.change.buildNotification([
                new Change("Mind-altering fragrance", "Sanity", -5)
            ]);
        };
    }

    function SuckController(key) {
        var changeId = key + "Change";
        this.get = function () {
            return global.kobold.history.getKey(id, key);
        };
        this.set = function (choice) {
            global.kobold.history.setKey(id, key, choice);
        };
        this.has = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.decrease(10, "Mind-numbing sap");
            });
            return global.ui.change.buildNotification([
                new Change("Mind-numbing sap", "Sanity", -10)
            ]);
        };
    }
    
    function TailController(key) {
        var changeId = key + "Change";
        var continueChangeId = key + "ContinueChange";
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.decrease(15, "Mind-clouding honey");
                global.kobold.status.exertion.increase(15, "Tentacle facefuck");
                global.kobold.body.oralVirgin = false;
            });
            return global.ui.change.buildNotification([
                new Change("Mind-clouding honey", "Sanity", -15),
                new Change("Tentacle facefuck", "Sexual exhaustion", +15, true)
            ]);
        };
        this.continueChange = function () {
            global.kobold.history.doIfNotSet(id, continueChangeId, true, function () {
                global.kobold.status.physical.increase(5, "Honey-meal");
                global.kobold.status.mental.increase(5, "Honey-meal");
                global.kobold.status.mental.increase(10, "Recovered sweater");
                global.kobold.status.stamina.decrease(15, "Rampant masturbation");
                global.kobold.status.exertion.increase(45, "Rampant masturbation");
                global.kobold.inventory.equipment.wear(new WoolSweater());
            });
            return global.ui.change.buildNotification([
                new Change("Recovered sweater", global.kobold.inventory.createItemWithTooltip(new WoolSweater()), +1),
                new Change("Honey-meal", "Health", +5),
                new Change("Honey-meal", "Sanity", +5),
                new Change("Recovered sweater", "Sanity", +10),
                new Change("Rampant masturbation", "Stamina", -15),
                new Change("Rampant masturbation", "Sexual exhaustion", +45, true)
            ]);
        };
    }
    
    function ExposeController(key) {
        var changeId = key + "Change";
        var continueChangeId = key + "ContinueChange";
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.decrease(25, "Fucked silly");
                global.kobold.status.exertion.increase(25, "Tentacle threesome");
                global.kobold.body.oralVirgin = false;
                if(global.kobold.body.dick) {
                    global.kobold.body.analVirgin = false;
                } else {
                    global.kobold.body.vaginalVirgin = false;
                }
                if(global.kobold.inventory.equipment.underwear) {
                    global.kobold.inventory.equipment.underwear.stain("Honey");
                }
                if(global.kobold.inventory.equipment.pants) {
                    global.kobold.inventory.equipment.pants.stain("Honey");
                }
            });
            return global.ui.change.buildNotification([
                new Change("Fucked silly", "Sanity", -25),
                new Change("Tentacle threesome", "Sexual exhaustion", +25, true)
            ]);
        };
        this.continueChange = function () {
            global.kobold.history.doIfNotSet(id, continueChangeId, true, function () {
                global.kobold.status.physical.increase(5, "Honey-meal");
                global.kobold.status.mental.increase(5, "Honey-meal");
                global.kobold.status.mental.increase(10, "Recovered sweater");
                global.kobold.status.stamina.decrease(15, "Rampant masturbation");
                global.kobold.status.exertion.increase(45, "Rampant masturbation");
                global.kobold.inventory.equipment.wear(new WoolSweater());
            });
            return global.ui.change.buildNotification([
                new Change("Recovered sweater", global.kobold.inventory.createItemWithTooltip(new WoolSweater()), +1),
                new Change("Honey-meal", "Health", +5),
                new Change("Honey-meal", "Sanity", +5),
                new Change("Recovered sweater", "Sanity", +10),
                new Change("Rampant masturbation", "Stamina", -15),
                new Change("Rampant masturbation", "Sexual exhaustion", +45, true)
            ]);
        };
    }

    function ClawController(key) {
        var changeId = key + "Change";
        var changeContinueId = key + "ChangeContinue";
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.physical.decrease(15, "Minor puncture wounds");
            });
            return global.ui.change.buildNotification([
                new Change("Wrist bleeding", "Health", -10),
                new Change("Stepped on a thorny tentacle", "Health", -3),
                new Change("Back thorn-whipped", "Health", -2)
            ]);
        };
        this.changeContinue = function () {
            global.kobold.history.doIfNotSet(id, changeContinueId, true, function () {
                global.kobold.status.mental.increase(10, "Recovered sweater");
                global.kobold.status.stamina.decrease(30, "Running from monster");
                global.kobold.inventory.equipment.wear(new WoolSweater());
            });
            return global.ui.change.buildNotification([
                new Change("Recovered sweater", global.kobold.inventory.createItemWithTooltip(new WoolSweater()), +1),
                new Change("Recovered sweater", "Sanity", +10),
                new Change("Running on fumes", "Stamina", -30)
            ]);
        };
    }

    function BiteController(key) {
        var changeId = key + "Change";
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.increase(10, "Recovered sweater");
                global.kobold.status.stamina.decrease(30, "Running from monster");
                global.kobold.inventory.equipment.wear(new WoolSweater());
            });
            return global.ui.change.buildNotification([
                new Change("Recovered sweater", global.kobold.inventory.createItemWithTooltip(new WoolSweater()), +1),
                new Change("Recovered sweater", "Sanity", +10),
                new Change("Running on fumes", "Stamina", -30)
            ]);
        };
    }
}

// END scenes\Forest\unknown\pry\ForestUnknownPryController.js

// scenes\Forest\unknown\yank\dodge\ForestUnknownDodge.js

/**
 * Dodge section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownDodge(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `yankDodgeContinueButton`;
    var next = new ForestUnknownDodgeContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Showcasing an impressive display of kobold flexibility, you gracefully sidestep the rapidly approaching tendril,
                            hopping back ever so slightly to ensure it can't catch you on its way back.
                            Missing its first strike seems to aggrevate the thornbush.
                            A second tentacle is hurled your way, rapidly followed by a third, a fourth and even a fifth.
                            Yet sheer numbers alone are no match for a tiny frame and a lifetime of experience in [i]not getting caught[/i].
                        [/p]
                        [p]
                            You jump over one lunging for your feet, you sidestep the third, you duck below the fourth, 
                            and by the time the plant launches its fifth attack, you have already managed to back away out of its reach,
                            the last thrusting appendage reaching its apex mere inches away from striking you.
                        [/p]
                        [p]
                            Still, you're not out of danger quite yet.
                            The moment its range proves to be inadequate, the entirety of the bush begins to [b]creep[/b] forwards,
                            like a hulking [b]mass[/b] of tentacles snaking over the ground in an attempt to catch you still.
                            Something about this display unsettles you, disturbs you to your very core, inspires dread into every fiber of your existence.
                            For a moment, you stand there, completely terrified, but as the monstrosity hurls another one of its limbs your way, you snap back to reality.
                        [/p]
                        [p]
                            Dodging the final would-be blow, your fearful mind can think of only one course of action: to get the [b]hell[/b] away from that thing!
                            You turn around, and [b]sprint[/b] away as fast your legs can carry you, tattered remains of your sweater in hand.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.yank.dodge.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.yank.dodge.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownDodge";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.yank.base.get() !== ForestUnknownYankChoice.DODGE) {
            return "";
        }
        
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.yank.dodge.change();
        part += buildChoice();
        part += next.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.yank.base.get() !== ForestUnknownYankChoice.DODGE) {
            return;
        }
        
        next.registerListeners();

        if (ctrl.unknown.yank.dodge.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\yank\dodge\ForestUnknownDodge.js

// scenes\Forest\unknown\yank\dodge\ForestUnknownDodgeContinue.js

/**
 * Dodge section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownDodgeContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `runIntoWoodsContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            For the second time today, you find yourself running for your life.
                            You don't know whether or not the plant is still giving chase, but fright and adrenaline combined make you [b]keep running[/b] without looking back.
                            It isn't until the sun has gone down, that you come to a [b]completely[/b] exhausted stop.
                        [/p]
                        [p]
                            You slump down against a nearby tree, 
                            tightly clutching the tattered remains of what was once your shirt while loudly panting.
                            You [b]ruined[/b] it.
                            You've had that sweater for [b]years[/b].
                            It's served as a blanket, a garment and even a towel on rainy days.
                            It was the closest thing to a stuffed toy you've ever owned.
                            And now it's gone.
                        [/p]
                        [p]
                            After taking a few more minutes to catch your breath and choke back tears for your lost sweater, you slowly rise to your feet once more.
                            You are tired.
                            [b]Very[/b] tired.
                            But you can't rest.
                            Not yet.
                            Not here.
                            It's too dangerous.
                            You take a look around.
                            You have no clue where you are.
                            All directions you gaze in look exactly the same.
                            Trees and darkness all around you.
                            You are officially [b]lost[/b].
                        [/p]
                        [p]
                            You carefully put on the remnants of your ${global.kobold.inventory.createItemWithTooltip(new RippedWoolSweater())}.
                            It no longer keeps you warm, nor will it provide much defence from whatever the world hurls at you.
                            But you don't wear it for that.
                            You wear it as a memento.
                            To the best shirt that ever was, and the best shirt that ever will be.
                        [/p]
                        [p]
                            Looks like you're going to have to pull an all nighter,
                            unless you can find a relatively safe place to catch some sleep in.
                            You're bound to run out of steam at some point, but that's a problem you'll face when it emerges.      
                            For now, you need to keep going.                      
                            As long as you keep heading in one direction, you'll reach the edge of the forest at one point or another, right?
                            You hope so, at least.
                        [/p]
                        [p]
                            With your body in shambles, and your spirit demoralized, you begin to stride even deeper into the forest,
                            in hopes of ultimately finding your way back home, lest you [i]become[/i] one of the savages that live here, 
                            completely isolated from civilization.
                            At least you learned a thing or two by dodging those tentacles.
                            If you keep being as alert and uncatchable as you were while fighting that monster, you're bound to make it out of here alive.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.continue();
            loadNextPart("RiversideMarsh", "Deepwoods");
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownDodgeContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.yank.dodge.continued()) {
            return "";
        }

        global.ui.applyTheme("forest_night");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.yank.dodge.continueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.yank.dodge.continued()) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.unknown.hasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\yank\dodge\ForestUnknownDodgeContinue.js

// scenes\Forest\unknown\yank\enveloped\ForestUnknownEnveloped.js

/**
 * Enveloped section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownEnveloped(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `envelopedContinueButton`;
    var next = new ForestUnknownEnvelopedContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cock = ctrl.unknown.cock();
        var pants = ctrl.unknown.pants();
        var pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        var underwear = ctrl.unknown.underwear();
        var underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        var toReturn = `[p]
                            As your senses slowly return to you, you are greeted by the sight of the tentacle monster, standing right above you.
                            A multitude of thorny tendrils are wrapped tightly around your arms, pricklers threatening to puncture you at the slightest hint of resistance.
                            Your legs are completely enveloped by the central mass of the beast.
                            Its insides are warm, and you can feel something [i]wet[/i] and [i]slimey[/i] dripping onto your feet.
                            Is it...is it [b]eating[/b] you?!
                        [/p]
                        [p]
                            Panicking, you begin to struggle against your restraints, but only end up hurting yourself on the myriad of thorns.
                            You fearfully gaze up at the monstrosity.
                            It is slowly crawling upwards along your body, towards your midsection.
                            Then, something [i]thick[/i] and [i]slick[/i] start rubbing against your legs.
                            Feeling, touching, poking, [i]prodding[/i]?
                            The creature's tongue, perhaps?
                            Is it [b]tasting[/b] you${pants ? " through your pants" : ""}?!
                        [/p]
                        [p]
                            No, it doesn't feel like a [i]tongue[/i].
                            It's too [b]thick[/b] for that.
                            Another tendril, maybe?
                            One without spikes?
                            One coated in some kind of...[i]lubricant[/i]?
                            The terror continues to move upwards, your feet slowly emerging from underneath its warm innards.
                            As a cold breeze washes over your now-wet scaly hindpaws, the true intentions of the monster suddenly dawn on you.
                            It's not eating you.
                            It's trying to [b]fuck[/b] you!
                        [/p]
                        [p]
                            You don't know whether you're relieved or horrified by this revelation.
                            The creature, having reached your midsection, finally comes to a halt, 
                            its hulking mass of writhing tendrils covering your body all the way from your knees to your midriff.
                            The smooth, prodding tentacle continues to feel around.
                            At first, it finds your belly button, but after wriggling about the small crevice in an attempt to penetrate it for a few moments,
                            it seems to realize that's [i]not[/i] something it can slip into.
                        [/p]
                        [p]
                            Then, it ventures downwards once more.
                            Coming across the waistband of your ${pants ? pantsPopup : underwearPopup}, it slickly slides underneath the fabric,
                            leaving a trail of the thick, sticky, yet surprisingly slippery [i]goop[/i] it seems to be [b]covered[/b] in on your scales.
                            You squeeze your legs shut in an attempt to bar the [i]thing[/i] from accessing your most private parts, yet despite your best efforts, 
                            its lubrication combined with a little bit of [b]force[/b] allows the tendril to easily slide in between your thighs.
                        [/p]
                        [p]
                            You squirm your lower body in a futile attempt to dislodge the tentacle, 
                            but your movements come to an abrupt halt as the slithering intruder presses its smooth, 
                            tapered tip against your virginal ${cock ? "cock" : ""}slit.
                            You pleadingly shake your head no, but you doubt the monster understands, let alone [i]cares[/i] about your feelings.
                            Slowly, but surely, it begins to slip inside of your body, spreading your folds to make way for its ever-thickening girth.
                        [/p]
                        [p]`;

        if (cock) {
            toReturn += `   The tentacle lodges itself deep inside of your loins, spearing open a hole that was [b]never[/b] meant to be used this way.
                            Your mouth falls agape, and you try to beg, to scream for help, or even to pathetically [b]whimper[/b], yet not a single tone escapes your lips.
                            The slick plant-appendage's well-lubed penetration doesn't [i]hurt[/i] as much as it simply [b]discomforts[/b] you.
                            You feel like your very [b]masculinity[/b] is being threatened as the tip of the tendril comes to rest against your flaccid member.
                        [/p]
                        [p]
                            After a moment of inactivity, the tentacle pulls back ever so slightly, 
                            before shallowly beginning to thrust in and out of your tight cockslit,
                            the tip bumping against your tender, soft cock every time it bottoms out.
                            Realizing that squeezing your thighs together isn't doing much to dissuade the monster,
                            you hang your head in impotent shame and cease your efforts to resist the tendril,
                            spreading your legs a tiny bit, 
                            hoping this will all come to an end a little bit sooner if you just let it [i]use you[/i].
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/unknown/yank/enveloped/rape-m-full.png" target="_blank">
                                    <img class="nsfw wideimage" src="scenes/Forest/unknown/yank/enveloped/rape-m.png" title="Illustration by Feldspar">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            A warm, viscous liquid squirts from the tip of the tendril, coating your insides,
                            acting as a lubricant to further smoothen the monster's thrusts, its pace picking up.
                            Some sort of precum?
                            Or maybe more of the stuff the tentacle is covered in?
                            You don't know.
                            You don't care.
                            You try to look away, to mentally escape your current situation,
                            yet the thorns imprisoning your arms and pinning you to the ground serve as a constant reminder of the terror's presence.
                            You just want this to be over with, for it to become a memory for you to repress instead of a reality you can't face away from.
                        [/p]
                        [p]
                            Luckily, your wish is soon granted.
                            The animated thornbush's thrusts intensify, 
                            spiky tendrils wrapping around your lower body to painfully keep you in place as every firm tentacle-hump rocks you back and forth,
                            your cockslit thoroughly [b]fucked[/b] as though it was a regular cunt. 
                            The last remnants of your self-image as a proud, masculine kobold fade away as the creature reaches its end, 
                            the tentacle buried deep inside of you pulsing and quivering, before squirting wave after wave of thick [b]sap[/b] into the depths of your slit,
                            drowning your flaccid member in floral spunk.`;
        } else {
            toReturn += `   The tentacle slides deeper inside of your loins, spearing open your tender twat, 
                            your hymen first penetrated by the slender tip of the plant-like appendage,
                            and then gradually [b]torn[/b] by the rest of the tendril's far more plump [i]body[/i].
                            A small bit of blood seeps from your slit, signifying the loss of your purity.
                        [/p]
                        [p]
                            The unwanted insertion hurts, but not all that much, the slippery slime coating the entire length of the tentacle acting as a superb lubricant.
                            It isn't long until the tendril bottoms out, its tapered tip pressing against your very cervix,
                            whilst the [i]shaft[/i] fills the rest of your twat to the point of near-stretching,
                            staying [b]barely[/b] within the limits of what you can comfortably take.
                        [/p]
                        [p]
                            After a moment of inactivity, the tentacle pulls back ever so slightly, 
                            before shallowly beginning to thrust in and out of your tight snatch.
                            Realizing that squeezing your thighs together isn't doing much to dissuade the monster,
                            you hang your head in impotent shame and cease your efforts to resist the tendril,
                            spreading your legs a tiny bit,
                            hoping this will all come to an end a little bit sooner if you just let it [i]use you[/i].
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/unknown/yank/enveloped/rape-f-full.png" target="_blank">
                                    <img class="nsfw wideimage" src="scenes/Forest/unknown/yank/enveloped/rape-f.png" title="Illustration by Feldspar">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            A warm, viscous liquid squirts from the tip of the tendril, 
                            lubricating your insides to further smoothen the monster's thrusts, its pace picking up.
                            Some sort of precum?
                            Or maybe more of the stuff the tentacle is covered in?
                            You don't know.
                            You don't care.
                            You try to look away, to mentally escape your current situation,
                            yet the thorns imprisoning your arms and pinning you to the ground serve as a constant reminder of the terror's presence.
                            You just want this to be over with, for it to become a memory for you to repress instead of a reality you can't face away from.
                        [/p]
                        [p]
                            Luckily, your wish is soon granted.
                            The animated thornbush's thrusts intensify, 
                            spiky tendrils wrapping around your lower body to painfully keep you in place as every firm tentacle-hump rocks you back and forth,
                            your cunt thoroughly [b]fucked[/b] by the pistoning plant-appendage.
                            Then, suddenly, it stops.
                            The tentacle buried deep inside of you begins to quiver and pulse, 
                            before squirting wave after wave of thick [b]sap[/b] into the depths of your slit,
                            the prehensile tip teasingly testing your cervix, attempting, but failing to shoot its load directly into your uterus.`;
        }

        toReturn += `   [/p]
                        [p]
                            Once fully spent, the plant retracts itself from your well-fucked hole, 
                            a veritable [b]torrent[/b] of its honey-like ejaculate oozing out of your body.
                            Spiked tentacles unravel themselves from your arms, and the central mass of the plant-like horror crawls off of you.
                            You lie there, motionless, absent-mindedly staring as the sun goes down, and the beast slowly makes its way out of the small clearing.
                        [/p]`;
        toReturn += `   [p]
                            [c]
                                [b]
                                    You are no longer a virgin.
                                [/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.yank.envelop.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.yank.envelop.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownEnveloped";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.yank.grab.continued() && !ctrl.unknown.yank.hit.toEnvelop()) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.yank.envelop.change();
        part += buildChoice();
        part += next.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.yank.grab.continued() && !ctrl.unknown.yank.hit.toEnvelop()) {
            return;
        }

        next.registerListeners();

        if (ctrl.unknown.yank.envelop.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\yank\enveloped\ForestUnknownEnveloped.js

// scenes\Forest\unknown\yank\enveloped\ForestUnknownEnvelopedContinue.js

/**
 * Enveloped section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownEnvelopedContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `leftForDeadContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cock = ctrl.unknown.cock();
        var pants = ctrl.unknown.pants();
        var pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        var underwear = ctrl.unknown.underwear();
        var underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        var toReturn = `[p]
                            Darkness envelops the land, the small bit of sunlight that illuminated this part of the deepwoods fading away as day turns to night.
                            While you had kept brave so far, a few tears finally flow from your eyes.
                            You feel used and abused.
                            Robbed of your ${cock ? "masculinity" : "virginity"}.
                        [/p]
                        [p]
                            You reach down, sliding your hand underneath the waistband of your ${pants ? "pants" : "underwear"}.
                            The warm goop seeping from your loins, much of which was spilled on your meagre clothing, feels [b]sticky[/b] to the touch.
                            Like someone emptied a jar of honey inside of you.
                            Without bothering to get up off of the ground, you clean your slit with the flat of your hand to the best of your ability.
                        [/p]
                        [p]
                            As you wipe the golden syrup off on a nearby patch of moss, its sweet smell reaches your nose.
                            It sickens you enough to motivate you to rise to your shaky feet.
                            Out of the corner of your eye, you spot a tiny consolation prize: the tattered remains of your shirt.
                            You pick it up off of the ground, and slowly don it.
                            At least this wasn't all for nothing, or so you claim, to try to cheer yourself up.
                        [/p]
                        [p]
                            It works, to some effect.
                            Your tears stop flowing, and your mind clears just enough for you to start thinking about survival, rather than about [i]wanting[/i] to die.
                            You don't think you can find your way home through the darkness.
                            But you can't stay here, either.
                            That monster might be back, or maybe something [i]even worse[/i] than it.
                        [/p]
                        [p]
                            Disheartened and broken, both mentally and physically, you set off in a random direction, 
                            hoping to make it out of the deepwoods [i]without[/i] any more of your orifices being [b]defiled[/b] by plants,
                            beasts, or whatever else may be waiting for you in the uncharted parts of the forest.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.continue();
            loadNextPart("RiversideMarsh", "Deepwoods");
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownEnvelopedContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.yank.envelop.continued()) {
            return "";
        }

        global.ui.applyTheme("forest_night");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.yank.envelop.continueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.yank.envelop.continued()) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.unknown.hasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\yank\enveloped\ForestUnknownEnvelopedContinue.js

// scenes\Forest\unknown\yank\grab\ForestUnknownGrab.js

/**
 * Grab section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownGrab(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `grabContinueButton`;
    var toEnveloped = new ForestUnknownEnveloped(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Time appears to slow down as adrenaline floods your system.
                            You can see the spiky tentacle lunging towards you in slow motion.
                            It undoubtedly seeks to either hurt you or grab you.
                            But you bet it [i]isn't[/i] counting on [i]you[/i] grabbing [i]it[/i] first.
                        [/p]
                        [p]
                            You extend your arm towards the surging tentacle, your fingers spread, yet arched, ready to clamp down onto the plant's appendage.
                            As the tendril rushes past your prepared hand, you spring your trap, 
                            snagging it mid-flight, squeezing it with all of your might to halt its advance, and to trap it within your grasp.
                            While at first, you are proud of your achievement, 
                            it isn't until you feel a sharp, stinging pain emanating from your hand, that you realize your folly.
                        [/p]
                        [p]
                            You just [b]squeezed down[/b] on nature's equivalent of [b]barbed wire[/b].
                            The tendril's thorns have punctured your soft palm and the underside of your fingers, sinking deeply into your flesh.
                            You howl out in pain.
                            Tossing the tentacle to the ground, you clutch your bleeding hand, taking a moment to inspect the injury.
                            While some of the wounds are pretty deep, you don't think anything [i]crucial[/i] was damaged.
                            At least you can still move all of your fingers.
                        [/p]
                        [p]
                            Sensing an opportunity, the plant immediately takes advantage of your distracted state.
                            While you're still attempting to judge just [i]how badly[/i] your hand was hurt,
                            two tentacles sneakily inch along the forest floor, encircling you.
                            After crossing one another and then some, they swoop back towards the central mass of the monstrosity as a unified [i]rope[/i],
                            catching the back of your feet, yanking your legs out from underneath you.
                        [/p]
                        [p]
                            Surprised, you let out a loud yelp as you are yanked off your feet, falling backwards onto the ground,
                            your head smacking against an unfortunately placed rock while the rest of your body comes to rest onto a bed of leaves with a soft thud.
                            Dazed, you phase out for a few moments.
                            Your vision is blurry, and the entire world seems to be rocking back and forth.
                            You notice something moving, but you can't quite make out what.
                            You can feel something crawling onto your arms, wrapping around them tightly.
                            [i]Oh no.[/i]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.yank.grab.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.yank.grab.continue();
            loadNextPart(toEnveloped);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownGrab";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.yank.base.get() !== ForestUnknownYankChoice.GRAB) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.yank.hit.change(); // Hit and grab have the same change.
        part += buildChoice();
        part += toEnveloped.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.yank.base.get() !== ForestUnknownYankChoice.GRAB) {
            return;
        }
        
        toEnveloped.registerListeners();

        if (ctrl.unknown.yank.grab.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\yank\grab\ForestUnknownGrab.js

// scenes\Forest\unknown\yank\hit\ForestUnknownHit.js

/**
 * Hit section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownHit(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `unknownHitContinueButton`;
    var toEnveloped = new ForestUnknownEnveloped(ctrl);
    var toSuccess = new ForestUnknownHitSuccess(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function buildSwipe() {
        var toReturn = `[p]
                            Time appears to slow down as adrenaline floods your system.
                            You can see the spiky tentacle lunging towards you in slow motion.
                            It undoubtedly seeks to either hurt you or grab you.
                            But you bet it [i]isn't[/i] counting on you fighting back.
                        [/p]
                        [p]
                            You ball your hand up into a fist, and strike at the surging tentacle,
                            throwing out a widely-arced hook which [b]crashes[/b] into the side of the tendril,
                            throwing it off-course, causing it to miss your body by a large margin.
                            While at first, you are proud of your achievement, 
                            it isn't until you feel a sharp, stinging pain emanating from your hand, that you realize your folly.
                        [/p]
                        ${global.ui.condFail(`No weapon`)}
                        [p]
                            You just outright [b]punched[/b] nature's equivalent of [b]barbed wire[/b].
                            The brief moment during which you struck it, the tendril's thorns punctured your scales, sinking deeply into your fingers,
                            and into the tender webbing in between your knuckles.
                            You howl out in pain.
                            Clutching your bleeding hand, you take a moment to inspect the injury.
                            While some of the wounds are pretty deep, you don't think anything [i]crucial[/i] was damaged.
                            At least you can still move all of your fingers.
                        [/p]
                        [p]
                            The plant, meanwhile, has recovered from the minor blow you dealt to it.
                            Sensing an opportunity, it immediately takes advantage of your distracted state.
                            While you're still attempting to judge just [i]how badly[/i] your hand was hurt,
                            two tentacles sneakily inch along the forest floor, encircling you.
                            After crossing one another and then some, they swoop back towards the central mass of the monstrosity as a unified [i]rope[/i],
                            catching the back of your feet, yanking your legs out from underneath you.
                        [/p]
                        [p]
                            Surprised, you let out a loud yelp as you are yanked off your feet, falling backwards onto the ground,
                            your head smacking against an unfortunately placed rock while the rest of your body comes to rest onto a bed of leaves with a soft thud.
                            Dazed, you phase out for a few moments.
                            Your vision is blurry, and the entire world seems to be rocking back and forth.
                            You notice something moving, but you can't quite make out what.
                            You can feel something crawling onto your arms, wrapping around them tightly.
                            [i]Oh no.[/i]
                        [/p]`;
        toReturn += ctrl.unknown.yank.hit.change();
        return toReturn;
    }

    function buildFight() {
        var weapon = ctrl.unknown.weapon();
        var weaponPopup = global.kobold.inventory.createItemWithTooltip(weapon);
        var toReturn = `[p]
                            Time appears to slow down as adrenaline floods your system.
                            You can see the spiky tentacle lunging towards you in slow motion.
                            It undoubtedly seeks to either hurt you or grab you.
                            But you bet it [i]isn't[/i] counting on you being armed.
                        [/p]
                        [p]
                            You deftly swing your ${weaponPopup} at the surging tendril in a wide arc, 
                            and due to a combination of sheer [i]luck[/i] and adrenaline-fueled expertise,
                            you skillfully manage to [b]smack[/b] the very tip of the floral appendage with the [b]full[/b] force of your weapon.
                            You strike it [b]so[/b] hard that some of the thorns outright [b]break off[/b],
                            flying off of the tentacle like teeth out of a bar brawler's mouth, leaving sap-oozing holes in their wake.
                        [/p]
                        [p]
                            The moment your strike connects, the very passage of time seems to speed up again,
                            your temporary enhanced perception fading just as quickly as it emerged.
                            The tendril, along with the entire animated thornbush, recoils, almost visibly [i]hurt[/i].
                        [/p]
                        ${global.ui.condSucc(global.kobold.inventory.createItemWithTooltip(weapon, true))}
                        [p]
                            A small, cruel smirk creeps across your face.
                            While your sweater getting stuck on the bush was undeniably [i]your[/i] fault, 
                            it still feels good to [b]relieve some frustration[/b] on this monstrosity.
                            If it hadn't been [b]in your way[/b], none of this would have happened in the first place!
                            The tentacle monster recollects itself, before advancing upon you once more.
                            You brace your weapon and prepare yourself.
                            If it's a fight it wants, a fight it'll get.
                        [/p]`;
        return toReturn;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = ``;

        if (ctrl.unknown.weapon()) {
            toReturn += buildFight();
        } else {
            toReturn += buildSwipe();
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.yank.hit.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.yank.hit.continue();
            if (ctrl.unknown.weapon()) {
                ctrl.unknown.yank.hit.success();
                loadNextPart(toSuccess);
            } else {
                ctrl.unknown.yank.hit.envelop();
                loadNextPart(toEnveloped);
            }
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownHit";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.yank.base.get() !== ForestUnknownYankChoice.HIT) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toSuccess.build();
        part += toEnveloped.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.yank.base.get() !== ForestUnknownYankChoice.HIT) {
            return;
        }

        toSuccess.registerListeners();
        toEnveloped.registerListeners();

        if (ctrl.unknown.yank.hit.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\yank\hit\ForestUnknownHit.js

// scenes\Forest\unknown\yank\hit\ForestUnknownHitSuccess.js

/**
 * Hit success section of the unknown scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownHitSuccess(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `woodsContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.unknown.cock();
        var weapon = ctrl.unknown.weapon();
        var weaponImage = weapon.id === ItemEnum.ROCK ? "rock" : "stick";
        var weaponPopup = global.kobold.inventory.createItemWithTooltip(weapon);
        var toReturn = `[p]
                            The horrific creature sends another one of its tentacles towards you,
                            but just like before, you manage to strike it down before it can reach you.
                            Aggravated, the bush's next assault consists of two separate tendrils which both attempt to hit you at once.
                            You dodge one, bash the other and then decide to go on the offensive.
                        [/p]
                        [p]
                            With an overhead swing, you [b]plow[/b] your weapon into the monster's center mass.
                            While the smack was exceedingly effective, the biological terror instantly retalliates with a sharp, 
                            barbed [b]whip[/b] to your back, strong enough to draw blood.
                            You grunt in pain and take a step backwards.
                        [/p]
                        [p]
                            You've managed to hurt it, but it, in turn, hurt you,
                            and you doubt that you can outlive this monstrosity in a trading blows contest.
                            Instead of resuming your reckless assault, you fall back into the pattern of deflecting its attacks.
                            Surely, if you beat enough of these damned tentacles, the entire thing'll just [i]have to[/i] give up at some point, right?
                        [/p]
                        [p]
                            You defend yourself from hit upon hit, using your ${weaponPopup} as both a shield and an offensive tool.
                            A few of the more tricky swipes get through your parrying stance, yet they never manage to do more than [i]scrape[/i] you.
                            As the sun, which had already begun its descent when you first came upon the bush, makes way for nighttime,
                            the two of you are [i]still[/i] fighting, both you and the monster driven by sheer [b]determination[/b] and [b]stubbornness[/b].
                        [/p]
                        [p]
                            Neither of you wants to back down at this point.
                            You're panting loudly, covered in a myriad of small cuts and scrapes, a few of which are bleeding, but nothing too severe.
                            The plant is leaking sap from most of its appendages, a few of which lie limply on the forest floor.
                            Perhaps it had hoped the darkness would give it an advantage, but as a kobold, your nightvision is superb.
                            The both of you realize this [b]is[/b] going to end shortly, one way or the other.
                        [/p]
                        [p]
                            A short, unnerving stillness ensues.
                            Then, the plant launches its final attack, its two remaining unscathed tendrils lunging at you.
                            You bash one to the side, but then the creature manages to wrap the last of its tentacles around your wrist,
                            painfully driving the thorns that cover it into your flesh.
                        [/p]
                        [p]
                            You scream out in pain, dropping your ${weaponPopup}.
                            A few of the less wounded tendrils seem to spring alive anew, as though the bush is encouraged by its own success.
                            But you are not yet beaten.
                            Far from it.
                        [/p]
                        [p]
                            You lean down, and harshly [b]bite[/b] the tendril holding your wrist, your sharp,
                            pointed teeth sinking through the plant matter with ease, completely [b]severing[/b] a large part of the beast's appendage,
                            the amputated segment instantly losing its grip, limply falling off of your forearm, dropping to the ground.
                            Sap squirts liberally from the now-blunt remainder of the tip of the tentacle.
                            The beast writhes, squirms and slowly backs away from you before slumping down to the ground, curling up into a messy ball of sap-leaking thorns.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    You have beaten the tentacle monster!
                                [/b]
                            [/c]
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/unknown/yank/hit/victory-${dick ? "m" : "f"}-${weaponImage}-full.png" target="_blank">
                                    <img class="nsfw wideimage" src="scenes/Forest/unknown/yank/hit/victory-${dick ? "m" : "f"}-${weaponImage}.png" title="Illustration by Feldspar">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            You crudely spit out the small piece of tendril you bit off, warily rising from your combat stance as you notice the creature's defeat.
                            You don't know whether you knocked it out, killed it, or whether it's [i]just pretending[/i] to be downed.
                            While keeping an eye on the vanquished monstrosity's body, you pick your ${weaponPopup} up off the ground, and begin to inspect your wounds.
                        [/p]
                        [p]
                            While nearly every inch of your entire form was scraped, cut or punctured during the fight, none of the individual wounds are [i]all that bad[/i].
                            Even your wrist, which, admittedly, the creature [i]did[/i] get a good, solid grip on, isn't bleeding too much.
                            The initial blow to your back was the worst, but [i]that[/i] pain has luckily already subsided.
                            You'll surely be sore in the morning, but right now, you're feeling just [i]fine[/i].
                        [/p]
                        [p]
                            You pick up the tattered remains of your shirt, which you dropped to the ground during the fight.
                            You don the ${global.kobold.inventory.createItemWithTooltip(new RippedWoolSweater())}, not because it looks good, 
                            or because it'll protect you or keep you warm or anything, but as keepsake of you [b]beating the shit[/b] out of a scary deepwoods monster.
                        [/p]
                        [p]
                            The fight was rather disorienting, and with darkness having settled across the land, you doubt you can find your way back home.
                            While you are weary, you can't rest yet.
                            Not here, anyway.
                            Not near that [i]thing[/i].
                            You decide to press deeper into the woods.
                            Ultimately, you hope you'll be able to find your way back home,
                            but for now, you'd settle for a place to sleep.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.unknown.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.continue();
            loadNextPart("RiversideMarsh", "Deepwoods");
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownHitSuccess";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.unknown.yank.hit.toSuccess()) {
            return "";
        }

        global.ui.applyTheme("forest_night");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.yank.hit.successChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.unknown.yank.hit.toSuccess()) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.unknown.hasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\unknown\yank\hit\ForestUnknownHitSuccess.js

// scenes\Forest\unknown\yank\ForestUnknownYank.js

/**
 * Unknown section of the forest scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknownYank(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var grabId = "unknownYankGrabButton";
    var hitId = "unknownYankHitButton";
    var dodgeId = "unknownYankDodgeButton";

    var toGrab = new ForestUnknownGrab(ctrl);
    var toHit = new ForestUnknownHit(ctrl);
    var toDodge = new ForestUnknownDodge(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.unknown.cock();
        var toReturn = `[p]
                            You move a bit closer towards the spiky shrub, making sure not to step on any lowhanging, barb-covered tendrils.
                            Your shirt is thoroughly [b]impaled[/b] on a multitude of thorns.
                            Far too many for you to carefully remove.
                            You [i]could[/i] try to gently pry the fabric apart the plant, but the situation seems rather hopeless.
                            Besides, it'd take precious time, time that could be better spent evading whoever may be chasing you.
                        [/p]
                        [p]
                            Instead of setting yourself up for failure, you decide to go for the quick and easy route.
                            You take hold of the stuck garment with both your hands, and then fiercely [b]pull[/b] it away from the plant's exterior.
                            A loud tearing noise echoes throughout the small clearing, as one particularly well-punctured part simply [b]refuses[/b] to budge.
                            While you were hoping to retrieve the sweater with perhaps a few small holes in it,
                            you accidentally completely [b]shred[/b] the old, worn fabric to pieces in your attempt to yank it off of the thornbush.
                        [/p]
                        [p]
                            When you realize what you just did, it is already too late.
                            The deed has been done.
                            Plucks of what used to be your favorite piece of clothing stick to the thorny bush, 
                            while a large strip of now-useless fabric flutters to the ground.
                            You have inadvertently [b]destroyed[/b] one of your most precious possessions, worth ten times the gold around your neck in sentimental value.
                        [/p]
                        [p]
                            Stupid. Stupid! [b]Stupid![/b]
                            How could you have done this?!
                            You should have [b]thought[/b] before you [b]acted[/b]!
                            You nearly start to cry, but manage to withhold your tears, for now.
                        [/p]
                        [p]
                            You hold the ${global.kobold.inventory.createItemWithTooltip(new RippedWoolSweater())} up to the sky,
                            one of the sparse rays of light beaming through the canopy illuminating it, allowing you to properly inspect the results of your actions.
                            The backside is practically torn in half.
                            Large pieces of cloth were ripped from the front.
                            One of the sleeves is hanging on by a thread, and the other has a few large holes in it.
                        [/p]
                        [p]
                            The result of your rash move is a bunch of rags that are barely strung together,
                            although you guess you could still wear what [i]used to be[/i] your [i]shirt[/i], 
                            if just until the fabric gives way completely.
                            As a memento.
                            A keepsake.
                            A reminder to always [b]think[/b] before you [b]do[/b].
                        [/p]
                        [p]
                            As you continue to lament the loss of your most prized belonging, 
                            you suddenly see something [i]moving[/i] through one of the many holes lining your now-ruined shirt.
                            Lowering the fabric covering your view, you bare witness to the plant's many spiked tendrils starting to move about.
                        [/p]
                        [p]
                            At first, you can't believe your eyes.
                            You think you're hallucinating, that your mind is unable to cope with the loss of your shirt,
                            and is thus imagining some sort of [i]boogeyman[/i], but the longer you stare at the slithering mass of thorned tentacles in front of you, 
                            the more you realize that you [i]aren't[/i] dreaming.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/unknown/yank/yank-${dick ? "m" : "f"}-full.png" target="_blank">
                                    <img class="nsfw wideimage" src="scenes/Forest/unknown/yank/yank-${dick ? "m" : "f"}.png" title="Illustration by Feldspar">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            You decide to keep your guard up.
                            Most of the tentacles are inching towards you.
                            You don't know the intentions of this [i]thing[/i] you seem to have awoken,
                            but if whatever it has planned involves those prickly tendrils, then you doubt you want to be a part of it.
                            Your instincts tell you to flee, but your rational mind reasons that, if you [i]do[/i] turn tail and run,
                            you might simply be presenting the monster with an easier target.
                        [/p]
                        [p]
                            Instead of outright bolting, you slowly begin to back away from the monstrosity, its tentacles menacingly advancing upon you.
                            One of them seemingly coils up, like a snake preparing to strike.
                            Then, just like the predator it's mimicking, the thorny appendage [b]leaps[/b] from the ground, lunging towards you at high speed.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the ant choice button.
     */
    function buildGrabChoice() {
        var disabled = ctrl.unknown.yank.base.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.yank.base.get() === ForestUnknownYankChoice.GRAB;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${grabId}" ${disabled}>Grab it</button>`;

        var tooltip = `Grab the tendril before it can grab you.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 170px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildHitChoice() {
        var disabled = ctrl.unknown.yank.base.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.yank.base.get() === ForestUnknownYankChoice.HIT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${hitId}" ${disabled}>Hit it</button>`;

        var weapon = ctrl.unknown.weapon();
        var tooltip;
        
        if (weapon) {
            tooltip = `Bash the tentacle with your ${weapon.name()}.`;
        } else {
            tooltip = `Punch the tentacle.`;
        }
        
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildDodgeChoice() {
        var disabled = ctrl.unknown.yank.base.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.yank.base.get() === ForestUnknownYankChoice.DODGE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${dodgeId}" ${disabled}>Dodge it</button>`;

        var tooltip = `Dodge the lunging tendril.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 110px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildGrabChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildHitChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildDodgeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + grabId), $("#" + hitId), $("#" + dodgeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerGrabListener() {
        $("#" + grabId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.yank.base.set(ForestUnknownYankChoice.GRAB);
            loadNextPart(toGrab);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerHitListener() {
        $("#" + hitId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.yank.base.set(ForestUnknownYankChoice.HIT);
            loadNextPart(toHit);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerDodgeListener() {
        $("#" + dodgeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.yank.base.set(ForestUnknownYankChoice.DODGE);
            loadNextPart(toDodge);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknownYank";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unknown.base.get() !== ForestUnknownChoice.YANK) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unknown.yank.base.change();
        part += buildChoice();
        part += toGrab.build();
        part += toHit.build();
        part += toDodge.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unknown.base.get() !== ForestUnknownChoice.YANK) {
            return;
        }

        toGrab.registerListeners();
        toHit.registerListeners();
        toDodge.registerListeners();

        if (ctrl.unknown.yank.base.has()) {
            return;
        }

        registerGrabListener();
        registerHitListener();
        registerDodgeListener();
    };
}

var ForestUnknownYankChoice = {
    GRAB: 1,
    HIT: 2,
    DODGE: 3
};

// END scenes\Forest\unknown\yank\ForestUnknownYank.js

// scenes\Forest\unknown\yank\ForestUnknownYankController.js

/**
 * Controller for the unknown part of the forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestUnknownYankController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;

    var baseId = "unknownYankBase";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    this.base = new BaseController(baseId);
    this.grab = new GrabController("unknownYankGrab");
    this.hit = new HitController("unknownYankHit");
    this.dodge = new DodgeController("unknownYankDodge");
    this.envelop = new EnvelopController("unknownEnvelop");

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    function BaseController(key) {
        var changeId = key + "Change";
        this.get = function () {
            return global.kobold.history.getKey(id, key);
        };
        this.set = function (choice) {
            global.kobold.history.setKey(id, key, choice);
        };
        this.has = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.stats.decreaseStat(StatsEnum.INTELLIGENCE, 1);
                global.kobold.status.mental.decrease(20, "Ruined wool sweater");
            });
            return global.ui.change.buildNotification([
                new Change("Ruined favorite shirt", "Intelligence", -1),
                new Change("Ruined favorite shirt", "Sanity", -20)
            ]);
        };
    }

    function GrabController(key) {
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
    }

    function HitController(key) {
        var changeId = key + "Change";
        var envelopId = key + "Envelop";
        var successId = key + "Success";
        var successChangeId = key + "SuccessChange";
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        this.success = function () {
            global.kobold.history.setKey(id, successId, true);
        };
        this.toSuccess = function () {
            return global.kobold.history.has(id, successId);
        };
        this.envelop = function () {
            global.kobold.history.setKey(id, envelopId, true);
        };
        this.toEnvelop = function () {
            return global.kobold.history.has(id, envelopId);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.stats.decreaseStat(StatsEnum.INTELLIGENCE, 1);
                global.kobold.status.physical.decrease(25, "Hit your head");
            });
            return global.ui.change.buildNotification([
                new Change("Hit your head", "Intelligence", -1),
                new Change("Hurt your hand", "Health", -10),
                new Change("Hit your head", "Health", -15),
            ]);
        };
        this.successChange = function () {
            global.kobold.history.doIfNotSet(id, successChangeId, true, function () {
                global.kobold.stats.increaseStat(StatsEnum.STRENGTH, 1);
                global.kobold.status.physical.decrease(20, "Various minor wounds");
                global.kobold.status.mental.increase(10, "Tentacle monster beatdown");
                global.kobold.status.stamina.decrease(25, "Tentacle monster fight");
                global.kobold.inventory.equipment.wear(new RippedWoolSweater());
            });
            return global.ui.change.buildNotification([
                new Change("Combat trophy", global.kobold.inventory.createItemWithTooltip(new RippedWoolSweater()), +1),
                new Change("Display of power", "Strength", +1),
                new Change("Victory", "Sanity", +10),
                new Change("Lots of minor wounds", "Health", -20),
                new Change("Long fight", "Stamina", -25)
            ]);
        };
    }

    function DodgeController(key) {
        var changeId = key + "Change";
        var continueChangeId = key + "ContinueChange";
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.stats.increaseStat(StatsEnum.AGILITY, 1);
                global.kobold.status.stamina.decrease(10, "Dodging tentacles");
            });
            return global.ui.change.buildNotification([
                new Change("Skillful dodging", "Agility", +1),
                new Change("Skillful dodging", "Stamina", -10)
            ]);
        };
        this.continueChange = function () {
            global.kobold.history.doIfNotSet(id, continueChangeId, true, function () {
                global.kobold.status.mental.decrease(5, "Lamenting sweater");
                global.kobold.status.stamina.decrease(30, "Running from monster");
                global.kobold.inventory.equipment.wear(new RippedWoolSweater());
            });
            return global.ui.change.buildNotification([
                new Change("Recovered sweater", global.kobold.inventory.createItemWithTooltip(new RippedWoolSweater()), +1),
                new Change("Lamenting sweater", "Sanity", -5),
                new Change("Running on fumes", "Stamina", -30)
            ]);
        };
    }
    
    function EnvelopController(key) {
        var changeId = key + "Change";
        var continueChangeId = key + "ContinueChange";
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                if(global.kobold.body.dick) {
                    global.kobold.body.cockslitVirgin = false;
                } else {
                    global.kobold.body.vaginalVirgin = false;
                }
                if(global.kobold.inventory.equipment.pants) {
                    global.kobold.inventory.equipment.pants.stain("Honey");
                }
                if(global.kobold.inventory.equipment.underwear) {
                    global.kobold.inventory.equipment.underwear.stain("Honey");
                }
                global.kobold.status.exertion.increase(20, "Tentacle-fucked");
            });
            return global.ui.change.buildNotification([
                new Change("Tentacle-fucked", "Sexual exhaustion", +20, true)
            ]);
        };
        this.continueChange = function () {
            global.kobold.history.doIfNotSet(id, continueChangeId, true, function () {
                global.kobold.status.mental.decrease(20, "Abused by tentacle monster");
                global.kobold.inventory.equipment.wear(new RippedWoolSweater());
            });
            return global.ui.change.buildNotification([
                new Change("Abused", "Sanity", -20),
                new Change("Recovered sweater", global.kobold.inventory.createItemWithTooltip(new RippedWoolSweater()), +1)
            ]);
        };
    }
}

// END scenes\Forest\unknown\yank\ForestUnknownYankController.js

// scenes\Forest\unknown\ForestUnknown.js

/**
 * Unknown section of the forest scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestUnknown(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var pryId = "unknownPryButton";
    var yankId = "unknownYankButton";

    var toPry = new ForestUnknownPry(ctrl);
    var toYank = new ForestUnknownYank(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[st]Deepwoods exploration[/st]
                        [p]
                            Your quick breather over, you decide to journey into parts of the forest you've never been to before.
                            You tend to avoid these deeper stretches of woodland, as you've heard plenty of [i]stories[/i] of the [i]things[/i] that dwell there.
                            Roaming packs of feral predators, human-sized spiders, uncivilized tribes of various species,
                            and even a few green dragons all call the savage, hardly ever explored deepwoods their home.
                        [/p]
                        [p]
                            You're certain that whoever may be pursuing you, too, is aware of the dangers of the uncharted regions of the forest.
                            You hope they won't have the guts or the will to follow you on your reckless venture into the unknown.
                            And if they do, at least you won't be the [i]only[/i] morsel around for the various woodland beasts to hunt.
                        [/p]
                        [p]
                            Neither uncivilized tribes nor feral animals would have much interest in robbing you of your stolen hoard of gold,
                            save for perhaps fashioning the individual coins into trinkets. 
                            In that sense, you suppose the deepwoods are just a tiny hint [i]safer[/i] than the bandit-infested surroundings of Varanar, at least for you.
                            Still, the lack of highwaymen doesn't nearly make up for the myriad of other dangers.
                        [/p]
                        [p]
                            The moment you step off the beaten path, you instantly notice the forest growing more [b]dense[/b].
                            The underbrush is thicker and more prevalent, and the trees are packed closely together.
                            The canopy blocks most of the sunlight, making this area far less illuminated than the rest of the woods, as though enveloped in a perpetual dusk.
                            A darkness which you fear will only deepen after the sun has set.
                        [/p]
                        [p]
                            About half an hour into your journey, you come across a large, thorny bush sitting in the middle of a small clearing.
                            At first, you think nothing of it, but as you walk past, you notice something familiar hanging off the side of the bush.
                            It's your old ${global.kobold.inventory.createItemWithTooltip(new WoolSweater())}, which you lost about half a week ago!
                        [/p]
                        [p]
                            How did that end up [i]here[/i]?
                            Admittedly, when you lost it, you weren't paying all too much attention.
                            You were too busy running away from a bunch of angry humans to mind your surroundings.
                            But still, you're fairly certain you didn't lose it [i]here[/i].
                            In fact, you don't think you've [b]ever[/b] been to this part of the woods before,
                            not even [i]while[/i] absent-mindedly fleeing farmers with pitchforks.
                        [/p]
                        [p]
                            To make matters even more confusing, you lost your shirt to a thornbush that looks [b]just[/b] like this one.
                            Did someone find the garment, bring it here and then lose it again to a very similar plant?
                            Unlikely.
                            Did someone move the bush?
                            Why would they have gone through the trouble of doing that, without so much as removing your shirt from its thorny exterior?
                            Did the plant itself [i]move[/i]?
                            Nonsense.
                            Plants can't move.
                        [/p]
                        [p]
                            You're probably just misremembering exactly where you lost your sweater.
                            Come to think of it, that entire day is a bit [i]hazy[/i] in your memory.
                            Besides, the important thing isn't how or where you lost the shirt.
                            You found it once again, and that's what matters.
                            Now all you need to do is retrieve it without getting yourself stung.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the ant choice button.
     */
    function buildPryChoice() {
        var disabled = ctrl.unknown.base.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.base.get() === ForestUnknownChoice.PRY;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${pryId}" ${disabled}>Pry it off</button>`;

        var tooltip = `Carefully pry off your sweater, to prevent it from getting damaged.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 165px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildYankChoice() {
        var disabled = ctrl.unknown.base.has() ? "disabled" : "";
        var isSelected = ctrl.unknown.base.get() === ForestUnknownChoice.YANK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${yankId}" ${disabled}>Yank it off</button>`;

        var tooltip = `Crudely yank your shirt off of the bush.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildPryChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildYankChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + pryId), $("#" + yankId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerPryListener() {
        $("#" + pryId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.base.set(ForestUnknownChoice.PRY);
            loadNextPart(toPry);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerYankListener() {
        $("#" + yankId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unknown.base.set(ForestUnknownChoice.YANK);
            loadNextPart(toYank);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "forestUnknown";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.woods.getArrivalChoice() !== ForestWoodsArrivalChoiceEnum.UNKNOWN) {
            return "";
        }
        ctrl.bgm.fadeLoop();
        ctrl.unknown.introSound.playSettingsDependent();
        global.ui.applyTheme("deepwoods");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toPry.build();
        part += toYank.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.woods.getArrivalChoice() !== ForestWoodsArrivalChoiceEnum.UNKNOWN) {
            return;
        }

        toPry.registerListeners();
        toYank.registerListeners();

        if (ctrl.unknown.base.has()) {
            return;
        }

        registerPryListener();
        registerYankListener();
    };
}

var ForestUnknownChoice = {
    PRY: 1,
    YANK: 2
};

// END scenes\Forest\unknown\ForestUnknown.js

// scenes\Forest\unknown\ForestUnknownController.js

/**
 * Controller for the unknown part of the forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestUnknownController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;

    var continueId = "unknownContinue";
    var baseId = "unknownBase";
    var cockId = "unknownCock";
    var pantsId = "unknownPants";
    var underwearId = "unknownUnderwear";
    var weaponId = "unknownWeapon";
    var rearExposedId = "unknownRearExposed";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.pry = new ForestUnknownPryController(id);
    this.yank = new ForestUnknownYankController(id);
    this.base = new BaseController(baseId);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    this.continue = function () {
        global.kobold.history.setKey(id, continueId, true);
    };

    this.hasContinued = function () {
        return global.kobold.history.has(id, continueId);
    };


    function BaseController(key) {
        this.get = function () {
            return global.kobold.history.getKey(id, key);
        };
        this.set = function (choice) {
            global.kobold.history.setKey(id, key, choice);
        };
        this.has = function () {
            return global.kobold.history.has(id, key);
        };
    }


    this.cock = function () {
        global.kobold.history.setIfNotSet(id, cockId, global.kobold.body.dick);
        return global.kobold.history.getKey(id, cockId);
    };

    this.pants = function () {
        global.kobold.history.setIfNotSet(id, pantsId, global.kobold.inventory.equipment.pants);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, pantsId));
    };

    this.underwear = function () {
        global.kobold.history.setIfNotSet(id, underwearId, global.kobold.inventory.equipment.underwear);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, underwearId));
    };
    this.weapon = function () {
        global.kobold.history.setIfNotSet(id, weaponId, global.kobold.inventory.equipment.weapon);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, weaponId));
    };
    this.rearExposed = function () {
        global.kobold.history.setIfNotSet(id, rearExposedId, global.kobold.inventory.equipment.rearExposed());
        return global.kobold.history.getKey(id, rearExposedId);
    };

    this.introSound = new AudioController(id, "res/audio/forest/dark_ambience.mp3", 300);
}

// END scenes\Forest\unknown\ForestUnknownController.js

// scenes\Forest\bandit\reward\ForestBanditReward.js

/**
 * Reward section of the forest bandit section.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditReward(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var continueId = "banditRewardContinueButton";
    
    var scaryNoise = `  <div class="text">
                            You take a deep breath, before letting out a loud, terrifying, soul-piercing howl to the best of your abilities.
                            While you start off mimicking a wolf, your <span class="italic">outburst</span> turns into something different altogether nearing the end,
                            more resembling a low, unsteady growl than the high-pitched whine you started off with.
                        </div>
                        <div class="text">
                            You can <span class="italic">hear</span> the kobolds getting more anxious with every passing moment,
                            their leader rapidly losing control over his men.
                            You follow up your initial screech with a series of disconcerting noises, each one more unsettling than the last.
                        </div>
                        <div class="text center">
                            <span class="bold blue">
                                Run!
                            </span>
                        </div>
                        <div class="text">
                            Comes the insubordinate decision of one of the underlings, 
                            the kobold darting out of the clearing and into the woods as fast as his legs can take him.
                            He is quickly followed by two of his peers, with only the leader tenaciously staying behind for a little while longer, before relenting himself.
                            With a dissatisfied grunt, he hops off of the woman's face, threatening her to stay put or face the consequences one final time.
                        </div>
                        <div class="text">
                            She wisely abides to his demand, the dragoness lying still whilst the ringleader drags his remaining, 
                            unconscious henchman into the cover of the forest.
                            The final two members of the gang soon disappear from sight, as the other three had long before them.
                        </div>
                        <div class="text center">
                            <span class="bold">
                                You've managed to save the dragoness!
                            </span>
                        </div>
                        <div class="text">
                            You keep up your threatening sounds for a few moments more, to ensure that the gang has left for sure.
                            The woman removes her impromptu gag, wiping her muzzle as she rises to her feet.
                            She, too, seems a little disturbed by your audial show.
                            You figure you must be good at faking noises.
                        </div>
                        <div class="text">
                            Ceasing your spooky act, you slowly step out from behind your cover, and approach the dragoness.
                            She gasps as you come into view.
                            The fearful lady visibly goes through a myriad of different emotions.
                        </div>
                        <div class="text">
                            At first, your appearance startles her. 
                            She gasps in fright as her mind races to reason that you are <span class="italic">not</span> a ghost 
                            or some scary forest monster coming to eat her whole.
                            Then, an angry, hateful glare takes over her visage, thinking you're part of another, or perhaps even the same gang.
                            Her hands curl up into fists, yet her aggressive posture quickly fades to make room for a slightly more hospitable pose.
                            She gives you a questioning, doutbting, thinking look.
                            Then, finally, the dragoness manages to piece things together.
                        </div>
                        <div class="text">
                            She tilts her head, quizically, gracefully and somewhat gratefully at the same time.
                            Truly a diplomatic figure, capable of portraying so many emotions with such a simple nod.
                            She eyes you up and down for a few moments.
                            Then, she speaks, slowly, her words like honey to your ears, each of them ringing in your mind long after being uttered by the woman:
                        </div>
                        <div class="center">
                            <span class="bold pink">
                                Was it <span class="italic">you</span> who rescued me?
                            </span>
                        </div>
                        <div class="text">
                            Your turn to be dumbstruck.
                            You manage to respond to the woman's query with only a simple nod.
                            A small, courteous, genuine smile forms on her face.
                            It warms your very heart.
                            Dragons really are <span class="italic">magical</span> beings.
                        </div>
                        <div class="center">
                            <span class="bold pink">
                                I am lady Saphira, princess of the United Dragon Empire. And you, my savior, are...?
                            </span>
                        </div>
                        <div class="text">
                            Slightly awestruck, you barely manage to stammer out your name.
                            Along with it, you ask what the lady is doing out here, all alone in the forest.
                            You can hardly peel your eyes away from hers, as though you're stuck in a trance.
                        </div>`;
    
    var payment = `     [p]
                            You loudly and nervously state that you would like to offer the gang a deal.
                            Their leader instantly responds, a quizzical, yet sceptical look forming on his face.
                            "<span class="bold brightred">Deal? What deal?! Bad timing! Can't you see I...WE are busy?! Go away!</span>"
                            He frantically waves his one free arm in an attempt to dismiss you.
                            You hesitantly assure him that this is a deal he -- [b]they[/b] wouldn't want to miss out on.
                            He sighs annoyedly in response, but you [i]have[/i] seemingly caught the group's attention, for better or for worse.
                        [/p]
                        [p]
                            You cup your coin pouch.
                            Second thoughts race through your mind, but it is too late for you to back down now.
                            You slowly remove the hefty bundle from around your neck.
                            Staring at it longingly, you close your eyes and let out a sigh of your own.
                            You have to do this. 
                            For the safety of the dragoness. 
                            You [b]need[/b] to do this.
                        [/p]
                        [p]
                            Opening your eyes once more, you jiggle your purse, the golden coins within clattering about loudly.
                            You open up the small bag to show its contents are indeed authentic, to prove you're not trying to play a trick on the kobolds.
                            Then, you toss the pouch to the ground, right in front of you.
                            Their collective eyes remain greedily glued to it until you open your mouth to speak once more.
                        [/p]
                        [p]
                            You make the offer.
                            They leave the dragoness alone, they get the coins.
                            The leader immediately questions why he shouldn't just take [b]both[/b].
                            You hadn't given it much thought, so you are dumbstruck for a moment.
                            After a brief period of silence, you suddenly think of a sly and appropriate answer.
                            Your hesitation melts like snow on the first, sunny day of spring, as you begin to reason out loud.
                        [/p]
                        [p]
                            Right now, they have a choice.
                            Option one is for them to resume their business with the dragoness, and let you go on your merry way.
                            Option two is that they take up your offer, you step aside and let them take your gold, and they leave a lot richer than before.
                            Option three entails them trying to get both.
                            They lunge for you, you snag your pouch off the ground, and you bolt into the forest.
                            The dragoness escapes, and they get [i]neither[/i] the gold [i]nor[/i] the girl.
                            Heck, maybe they'll manage to catch [b]one[/i] of you, [i]if[/i] they're lucky.
                        [/p]
                        [p]
                            Your astute observation prompts a discussion amongst the kobolds.
                            While it's hard to follow all five of them chattering at the same time,
                            the underlings generally seem to agree with what you're saying.
                            The leader, meanwhile, continues to believe he can snag your gold [i]without[/i] letting the dragoness go,
                            but after a solid minute of heated arguing between him and his gang, he finally relents to the will of the majority.
                        [/p]
                        [p]
                            His face turns towards you once more, an unhappy look on his face.
                            "<span class="bold brightred">Fine.</span>" he practically spits out. "<span class="bold brightred">We take the gold.</span>"
                            The both of you exchange an uneasy nod.
                            The deal is done.
                        [/p]
                        [p]
                            The bandits let go of the dragoness' limbs.
                            She instantly removes the loincloth-gag from her mouth, and rises to her feet.
                            You carefully circle around them towards her, and they do their best to avoid you whilst beelining for your gold.
                            After verifying the contents of the pouch that was formerly yours, the ringleader gives both you and the dragoness a courteous bow,
                            resentment and reluctance still clearly plastered across his face.
                        [/p]
                        [p]
                            While unpleased with the deal his underlings are so keen on taking, he seems like a man of his word.
                            Bedgrudgingly, he speaks: "<span class="bold brightred">Pleasure doing business. We'll buy ten whores like her with this.</span>"
                            Then, with a wave of his hand, all five of the kobolds swiftly venture out of the clearing, disappearing into the forest.
                        [/p]
                        <div class="text center">
                            <span class="bold">
                                You've managed to save the dragoness!
                            </span>
                        </div>`;
    
    var paymentRock = ` [p]
                            You loudly and nervously state that you would like to offer the gang a deal.
                            Their leader instantly responds, a quizzical, yet sceptical look forming on his face.
                            "<span class="bold brightred">Deal? You throw rock, then offer deal?! Get lost before I gut you!</span>"
                            He threateningly waves his knife about in your direction.
                            You hesitantly assure him that this is a deal he wouldn't want to miss out on,
                            one that'll [b]more[/b] than make up for knocking out one of his comrades.
                            He sighs annoyedly in response, but you [i]have[/i] seemingly caught the group's attention, for better or for worse.
                        [/p]
                        [p]
                            You cup your coin pouch.
                            Second thoughts race through your mind, but it is too late for you to back down now.
                            You slowly remove the hefty bundle from around your neck.
                            Staring at it longingly, you close your eyes and let out a sigh of your own.
                            You have to do this. 
                            For the safety of the dragoness. 
                            You [b]need[/b] to do this.
                        [/p]
                        [p]
                            Opening your eyes once more, you jiggle your purse, the golden coins within clattering about loudly.
                            You open up the small bag to show its contents are indeed authentic, to prove you're not trying to play a trick on the kobolds.
                            Then, you toss the pouch to the ground, right in front of you.
                            Their collective eyes remain greedily glued to it until you open your mouth to speak once more.
                        [/p]
                        [p]
                            You make the offer.
                            They leave the dragoness alone, they get the coins.
                            The leader immediately questions why he shouldn't just take [b]both[/b].
                            You hadn't given it much thought, so you are dumbstruck for a moment.
                            After a brief period of silence, you suddenly think of a sly and appropriate answer.
                            Your hesitation melts like snow on the first, sunny day of spring, as you begin to reason out loud.
                        [/p]
                        [p]
                            Right now, they have a choice.
                            Option one is for them to resume their business with the dragoness, and let you go on your merry way.
                            Option two is that they take up your offer, you step aside and let them take your gold, and they leave a lot richer than before.
                            Option three entails them trying to get both.
                            They lunge for you, you snag your pouch off the ground, and you bolt into the forest.
                            Already a man down, 
                            there'll be nothing to prevent the dragoness from escaping, 
                            and they'll get [i]neither[/i] the gold [i]nor[/i] the girl.
                            Heck, maybe they'll manage to catch [b]one[/i] of you, [i]if[/i] they're lucky.
                        [/p]
                        [p]
                            Your astute observation prompts a discussion amongst the kobolds.
                            While it's hard to follow all four of them chattering at the same time,
                            the underlings generally seem to agree with what you're saying.
                            The leader, meanwhile, continues to believe he can snag your gold [i]without[/i] letting the dragoness go,
                            but after a solid minute of heated arguing between him and his gang, he finally relents to the will of the majority.
                        [/p]
                        [p]
                            His face turns towards you once more, an unhappy look on his face.
                            "<span class="bold brightred">Fine.</span>" he practically spits out. "<span class="bold brightred">We take the gold.</span>"
                            The both of you exchange an uneasy nod.
                            The deal is done.
                        [/p]
                        [p]
                            The bandits let go of the dragoness' limbs.
                            She instantly removes the loincloth-gag from her mouth, and rises to her feet.
                            You carefully circle around them towards her, and they do their best to avoid you whilst beelining for your gold,
                            only one of them concerned with strenuously dragging their unconscious friend along with.
                            After verifying the contents of the pouch that was formerly yours, the ringleader gives both you and the dragoness a courteous bow,
                            resentment and reluctance still clearly plastered across his face.
                        [/p]
                        [p]
                            While unpleased with the deal his underlings are so keen on taking, he seems like a man of his word.
                            Bedgrudgingly, he speaks: "<span class="bold brightred">Pleasure doing business. We'll buy ten whores like her with this.</span>"
                            Then, with a wave of his hand, three of the kobolds venture out of the clearing, 
                            disappearing into the forest, followed by the fourth tugging along the fifth.
                        [/p]
                        <div class="text center">
                            <span class="bold">
                                You've managed to save the dragoness!
                            </span>
                        </div>`;
    
    
    var postPay = `     [p]
                            As the bandits scurry off, the dragoness turns to face you.
                            A small, genuine smile forms on her face.
                            It warms your very heart.
                            Dragons truly are <span class="italic">magical</span> beings.
                            Then, she speaks, her words like honey to your ears, each of them ringing in your mind long after being uttered by the woman:
                        [/p]
                        [p]
                            "<span class="bold pink">I am lady Saphira, princess of the United Dragon Empire.
                                I am indebted to you, for your kind and most generous of actions.
                                That must have been quite a bit of money, for those vile creatures to release someone of my standing from their grasp.
                                May I know your name, my most beneficent of heroes?</span>"
                        [/p]
                        [p]
                            Slightly awestruck, you barely manage to stammer out your name.
                            Along with it, you ask what the lady is doing out here, all alone in the forest.
                            You can hardly peel your eyes away from hers, as though you're stuck in a trance.
                        [/p]`;
    
    var genReward = `   [p]
                            Noticing your sadness, the dragoness quickly changes the topic at hand.
                            A reward.
                            She wants to give you something for saving her, yet her own purse was plundered by the kobold bandits.
                            Picking up her deflated coin pouch off the leafy forest floor, she laments the entire ordeal for a couple of moments.
                        [/p]
                        [p]
                            While she has nothing of actual value left,
                            apart from her somewhat intact pride,
                            there is still [i]one[/i] thing she can give you.
                            Reaching into her bosom, the lady begins to fiddle with the fabric on the inside of her dress.
                            After a small bit of tangible frustration,
                            she pulls out a single, sharp ${global.kobold.inventory.createItemWithTooltip(new GoldenNeedle())}, 
                            which was properly concealed, hidden in her apparel.
                        [/p]
                        [p]
                            She hands the sharp object to you, as a keepsake.
                            It's the [i]least[/i] she can do to repay you.
                            Seeing the setting sun, she decides it's time to say goodbye.
                            And then, after condescendingly patting your head, 
                            the dragoness slowly walks out of the clearing, in the direction of the southern road.
                        [/p]`;
    
    var payReward = `   [p]
                            Noticing your sadness, the dragoness quickly changes the topic at hand.
                            A reward.
                            She wants to give you something for saving her, yet her own purse was plundered by the kobold bandits.
                            Picking up her deflated coin pouch off the leafy forest floor, she passes it to you.
                            You wear the empty bag around your neck, much like your previous money-container,
                            only this one is empty instead of being filled to the brim with much-needed money.
                        [/p]
                        [p]
                            While she has nothing of actual value left,
                            apart from her somewhat intact pride,
                            there is still [i]one[/i] thing she can give you.
                            Reaching into her bosom, the lady begins to fiddle with the fabric on the inside of her dress.
                            After a small bit of tangible frustration,
                            she pulls out a single, sharp ${global.kobold.inventory.createItemWithTooltip(new GoldenNeedle())}, 
                            which was properly concealed, hidden in her apparel.
                        [/p]
                        [p]
                            She hands the sharp object to you, as a keepsake.
                            It's the [i]least[/i] she can do to repay you.
                            Seeing the setting sun, she decides it's time to say goodbye.
                            And then, after condescendingly patting your head, 
                            the dragoness slowly walks out of the clearing, in the direction of the southern road.
                        [/p]`;

    var cRock = `       [p]
                            With the lady gone, you slowly drift out of your enamored state.
                            She had an undeniably powerful effect on you, which seems to fade along with her presence, if a bit delayed.
                            You wonder why she didn't use this [i]gift[/i] to stave off the bandits.
                            If she's even in control of it, that is.
                        [/p]
                        [p]
                            Oh well, some things aren't for simple minds like your own to understand.
                            You briefly shake your head to clear any lingering thoughts of Saphira, and after picking up your  you begin to make your way back home.
                        [/p]`;


    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    
    function buildName() {
        var name = ctrl.bandit.rewardName();
        return global.parser.parse(`
                        [p]
                            "<span class="bold pink">Ah, ${name}. 
                            Am I pronouncing that right?
                            A very fitting name for someone of your...[i]stature[/i].
                            Well, my little saviour, until very recently, I was a guest in the keep of Varanar.
                            However, the atmosphere there has turned rather [i]grim[/i] ever since the human king ventured out to war.</span>"
                        [/p]
                        [p]
                            A troubled look grows on the princess' face.
                            It makes you feel absolutely [b]terrible[/b], as though you're going through the exact same emotions she must have felt.
                        [/p]
                        [p]
                            "<span class="bold pink">Without divulging into details, let us just say that the current governor and I do not see eye to eye.
                            In fact, he imprisoned my retinue, cast me from the keep, and outright banished me from the city.</span>"
                        [/p]
                        [p]
                            You attentively continue to listen, enthralled by her every word, [b]craving[/b] for more of the nigh-orgasmic melody of her voice.
                            She seems to be happy to be able to tell her tale to someone who [i]cares[/i], seemingly unaware of her own enchanting aura.
                        [/p]
                        [p]
                            "<span class="bold pink">So I was making my lonely way southwards, to the nearby town of Salamar,
                            when I spotted a band of misfits headed my way, their caravan taking up most of the road.
                            Knowing how [i]uncivilized[/i] soldiers can get, I sought refuge in the forest, only to be assaulted by a band of [i]your[/i] kind.
                            And the rest, I believe you can piece together for yourself.</span>"
                        [/p]
                        [p]
                            You give the lady a slight nod. 
                            You can feel her [i]influence[/i] on you lessen, 
                            enough for you to [b]think[/b] for yourself instead of your mind simply echoing [i]her[/i] words.
                            You ask the woman if she needs an [i]escort[/i] to get to...Saramar, was it?
                            You mean, you've always dreamt of being a knight, and this is the closest you've ever gotten to kind of not really fulfilling that dream!
                        [/p]
                        [p]
                            Sadly, the lady turns you down, stating she can't [i]afford[/i] being seen with one of [i]your[/i] kind.
                            Her rejection makes you feel physically ill, devestated to the very core,
                            but the feeling rapidly vanishes, leaving only the disappointment of your broken childhood dreams in its wake.
                        [/p]`);
    }
    
    function buildContinue(){
        var didRock = (ctrl.bandit.getArrivalChoice() === ForestBanditArrivalChoiceEnum.ROCK);
        return global.parser.parse(`
                        [p]
                            With the lady gone, you slowly drift out of your enamored state.
                            She had an undeniably powerful effect on you, which seems to fade along with her presence, if a bit delayed.
                            You wonder why she didn't use this [i]gift[/i] to stave off the bandits.
                            If she's even in control of it, that is.
                        [/p]
                        [p]
                            Oh well, some things aren't for simple minds like your own to understand.
                            You briefly shake your head to clear any lingering thoughts of Saphira, 
                            and ${didRock ? "after picking up your rock once more, " : ""} you begin to make your way back home.
                        [/p]`);
    }
    
    function buildScaryNoise() {
        var toReturn = scaryNoise;
        toReturn += buildName();
        toReturn += global.parser.parse(genReward);
        toReturn += buildContinue();
        return toReturn;
    }
    
    function buildPay() {
        var didRock = (ctrl.bandit.getArrivalChoice() === ForestBanditArrivalChoiceEnum.ROCK);
        var toReturn;
        if(didRock){
            toReturn = global.parser.parse(paymentRock);
        } else {
            toReturn = global.parser.parse(payment);
        }
        toReturn += global.parser.parse(postPay);
        toReturn += buildName();
        toReturn += global.parser.parse(payReward);
        toReturn += buildContinue();
        return toReturn;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var source = ctrl.bandit.getRewardSource();
        var toReturn = `<div class=text>reward - no source</div>`;
        if (source === ForestBanditRewardSourceEnum.SCARYNOISE) {
            toReturn = buildScaryNoise();
            toReturn += ctrl.bandit.rewardScary();
        }
        if (source === ForestBanditRewardSourceEnum.PAY) {
            toReturn = buildPay();
            toReturn += ctrl.bandit.rewardPay();
        }
        return toReturn;
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.bandit.hasRewardContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }
    
    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.rewardContinue();
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditReward";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.bandit.hasRewardSource()) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.bandit.hasRewardSource()) {
            return;
        }

        // Register children listeners here.

        if (ctrl.bandit.hasRewardContinued()) {
            return;
        }

        registerContinueListener();
    };
}

/**
 * Sources for the reward of the bandit section.
 */
var ForestBanditRewardSourceEnum = {
    SCARYNOISE: 1,
    PAY: 2,
    PERSUASION: 3
};

// END scenes\Forest\bandit\reward\ForestBanditReward.js

// scenes\Forest\bandit\intervene\bargain\replace\ForestBanditReplace.js

/**
 * Replace part of the forest bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditReplace(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var acceptId = "banditReplaceAcceptButton";
    var protestId = "banditReplaceProtestButton";

    var toAccept = new ForestBanditAccept(ctrl);
    var toProtest = new ForestBanditProtest(ctrl);

    var didRock;
    var gender;
    var dick;
    var pants;
    var underwear;
    var bra;


    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You compliment the leader of the gang for his fine catch.
                            You imagine it's not [i]every[/i] day they manage to [i]capture[/i] one of the most regal creatures around.
                            You remark that you can't help but notice their group's intentions go beyond just robbing the poor lady.
                        [/p]
                        [p]
                            "<span class="bold brightred">We're not sharing! Leave!</span>" comes the reply.
                            You quickly assure him that you aren't looking for a handout of dragoness sex.
                            Instead, you say that you've come here to offer him something [b]even better[/b].
                            Seemingly having piqued his curiosity, you immediately play your hand, and say that instead of the woman,
                            him and his gang could instead feast themselves on [b]your[/b] body, which is far more height-appropriate than a dragoness twice their size.
                        [/p]`;
        if (didRock) {
            toReturn += `[p]
                            You immediately get the feeling your idea wasn't all that well accepted by the group.
                            Just as their leader was about to scold you, the dragoness suddenly makes a move.
                            She swings her free arm at the bandit leader, punching him off of her.
                            With his knife no longer on her neck, the feisty lady kicks off the remaining distracted and surprised gangmembers.
                        [/p]
                        [p]
                            She hops to her feet and bolts into the forest, one of the bandits giving chase before the leader has recovered enough to tell him not to.
                            Once recovered from their light injuries, the three remaining bandits focus their attention on [b]you[/b].
                            They swiftly advance upon you, and before you even properly realize what's going on, they have you surrounded.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You immediately get the feeling your idea wasn't all that well accepted by the group.
                            Just as their leader was about to scold you, the dragoness suddenly makes a move.
                            She wildly flails her right arm about, knocking off the kobold pinning her down.
                            She then swings her freed arm at the bandit leader, punching him off of her.
                            With his knife no longer on her neck, the feisty lady kicks off the remaining distracted and surprised gangmembers.
                        [/p]
                        [p]
                            She hops to her feet and bolts into the forest, two of the bandits giving chase before the leader has recovered enough to tell them not to.
                            Once recovered from their light injuries, the three remaining bandits focus their attention on [b]you[/b].
                            They swiftly advance upon you, and before you even properly realize what's going on, they have you surrounded.
                        [/p]`;
        }
        toReturn += `   [p]
                            "<span class="bold brightred">Offer accepted.</span>" the leader, right in front of you, speaks.
                            You let out a nervous giggle and say you were [i]just kidding[/i].
                            They didn't actually believe that someone would stumble across a bunch of rowdy bandits like themselves and simply offer to have sex with them, right?
                        [/p]
                        [p]
                            Disregarding your excuses, the boss promptly unties his belt, his pants sliding smoothly over his scaled legs down to the leafy forest floor, 
                            leaving him naked from the waist-down, his semi-hard, human-sized member on full display.
                            You struggle to come up with a way out of this.
                            Looking around, you notice the other two kobolds have stripped as well.
                            A mixture of fear and regret begins to spread through your mind.
                        [/p]
                        [p]
                            As you begin to stammer out another witty excuse, the leader of the gang takes a step towards you, 
                            threateningly cupping your jaw and squeezing your cheeks to silence you halfway through your first word.
                            He speaks, in a dead-serious tone, which hugely contrasts with his earlier near-incoherent shouting:
                            "<span class="bold brightred">You cost us a dragoness.
                            You offered us your body.
                            Pray that we don't take your life as well.</span>"
                        [/p]
                        [p]`;

        if (bra && pants) {
            toReturn += `   With his free hand, the kobold reaches for your ${global.kobold.inventory.createItemWithTooltip(bra)},
                            ripping the fabric from your body and tossing it to the side, your meagre chest exposed to the group.
                            From behind, one of the other two kobolds grips your ${global.kobold.inventory.createItemWithTooltip(pants)} by the waistband.
                            Attempting to pull them down, the ruffian tugs on the garment just a little too roughly.
                            A loud ripping noise echoes throughout the forest clearing, and within moments the torn, 
                            shredded remains of your pants join whatever is left of your bra on the leafy forest floor.`;
        } else if (underwear) {
            toReturn += `   With his free hand, the kobold reaches for your ${global.kobold.inventory.createItemWithTooltip(underwear)},
                            ripping the fabric from your body and tossing it to the side, your groin exposed to the group.
                            Your nakedness only adds to your building fear.`;
        }

        toReturn += `   [/p]
                        [p]
                            You move your hands to cover your nudity,
                            but your arms are grabbed and held in place before you even manage to reach your private parts.
                            Your tail, your last bastion of modesty, crudely gets yanked upwards by one of the bandits behind you.
                            The other swats your taut rear with the flat of his hand, letting his palm linger to squeeze the lower end of one of your buttcheeks, 
                            his fingers slipping in between your thighs.
                        [/p]`;
        return  global.parser.parse(toReturn);
    }

    /**
     * Initializes all private variables.
     */
    function initVars() {
        didRock = (ctrl.bandit.getArrivalChoice() === ForestBanditArrivalChoiceEnum.ROCK);
        gender = ctrl.bandit.intervene.replaceGender();
        dick = ctrl.bandit.intervene.replaceDick();
        pants = ctrl.bandit.intervene.replacePants();
        underwear = ctrl.bandit.intervene.replaceUnderwear();
        bra = ctrl.bandit.intervene.replaceBra();
        ctrl.bandit.intervene.replaceRemoveClothing();
    }

    /**
     * Builds and returns the accept choice button.
     */
    function buildAcceptChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeReplaceChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getReplaceChoice() == ForestBanditReplaceChoice.ACCEPT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${acceptId}" ${disabled} style="font-size: 17px;">Accept your fate</button>`;

        var tooltip = `You can't stop them. Surrender your body to the kobolds.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the protest choice button.
     */
    function buildProtestChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeReplaceChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getReplaceChoice() == ForestBanditReplaceChoice.PROTEST;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${protestId}" ${disabled}>Protest</button>`;

        var tooltip = `You can't let them do this to you. You don't want to lose your virginity to some gang in the middle of nowhere!`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildAcceptChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildProtestChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables both buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + acceptId), $("#" + protestId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the bargain button.
     */
    function registerAcceptListener() {
        $("#" + acceptId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeReplaceChoice(ForestBanditReplaceChoice.ACCEPT);
            loadNextPart(toAccept);
        });
    }

    /**
     * Adds a listener to the relieve button.
     */
    function registerProtestListener() {
        $("#" + protestId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeReplaceChoice(ForestBanditReplaceChoice.PROTEST);
            loadNextPart(toProtest);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditReplace";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.intervene.getBargainChoice() !== ForestBanditBargainChoice.REPLACE) {
            return "";
        }
        initVars();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        // Build children here.
        part += toAccept.build();
        part += toProtest.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.intervene.getBargainChoice() !== ForestBanditBargainChoice.REPLACE) {
            return;
        }

        toAccept.registerListeners();
        toProtest.registerListeners();

        if (ctrl.bandit.intervene.hasMadeReplaceChoice()) {
            return;
        }

        registerAcceptListener();
        registerProtestListener();
    };
}

var ForestBanditReplaceChoice = {
    ACCEPT: 1,
    PROTEST: 2
};

// END scenes\Forest\bandit\intervene\bargain\replace\ForestBanditReplace.js

// scenes\Forest\bandit\intervene\bargain\replace\virgin\ForestBanditProtest.js

/**
 * Protest part of the forest bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditProtest(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "banditProtestContinueButton";

    var didRock;
    var gender;
    var male;
    var dick;
    var pants;
    var underwear;
    var bra;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You begin to struggle against the kobolds' grasp, fighting against their hold while their leader's hand finds its way atop your chest.
                        `;

        if (gender === GenderEnum.MALE) {
            toReturn += `   He runs his palm along your smooth, scaled torso.`;
        } else {
            toReturn += `   He grabs one of your breasts, groping and rubbing it to his heart's content.`;
        }

        toReturn += `       You try, but fail to kick him, your foot barely grazing his leg.
                            "<span class="bold brightred">Feisty little bitch.</span>" he comments, a broad, ominous grin spreading across his face.
                            He finally lets go of your jaw, sharply slapping your cheek, drawing a few tears from your eyes.
                        [/p]
                        [p]
                            With your mouth freed, you quickly utter a loud plea, [b]begging[/b] them to stop.
                            With a shaky voice, you stammer out that they can't do this to you! 
                            You're still a virgin!
                            Upon hearing that final fact, the ringleader's eyes open wide.
                            His grin deepens, and he lustily licks his lips.
                            "<span class="bold brightred">Virgin? Good. Might make a decent replacement for the dragoness after all.</span>"
                        [/p]
                        [p]
                            Your jaw falls agape.
                            They simply don't care.
                            With a small nod from the boss, his underlings roughly yank you backwards, pulling you off your feet and tossing you to the ground.
                            Your back hits the ground with a loud thud, the impact luckily softened by a small patch of moss and a bunch of leaves.
                        [/p]
                        [p]
                            You panic.
                            You wildly flail your legs about, hitting nothing but air.
                            You wriggle your body around to try and escape the bandits' grip, but they prove simply [i]too strong[/i] for you.
                            Then, a harsh blow to your gut knocks the air out of your lungs.
                            "<span class="bold blue">Hold STILL!</span>" one of them commands.
                            Fearing for your very life, you comply, your struggles slowly dying down as you gasp for air.
                        [/p]
                        [p]
                            The leader of the gang drops to his knees.
                            Your legs are forcibly spread, your most private parts revealed.
                            He takes a moment to inspect your groin, his cock visibly hardening, a single droplet of precum forming at the very tip of it.
                            The kobold crawls on top of you, his slightly larger frame completely covering your own, his bare body pressing against yours.
                        [/p]
                        [p]
                            As the male comes face-to-face with you, 
                            you can feel his member ${male ? `pressing in between your butcheeks` : `pressing against your slit`}.
                            He leans in to loudly sniff your neck, his nostrils pressing directly against your scales.
                            Out of sheer impotent rage, you snap your muzzle at him, threatening to bite his face with your razor-sharp teeth.
                            He balls his hand up into a fist, raising it menacingly, as if to punch you directly in the face,
                            but decides [b]not[/b] to bloody you quite yet.
                        [/p]
                        [p]
                            "<span class="bold brightred">Just for that, no foreplay.</span>" he calmly states, anger dominating his voice.
                            One of his comrades passes the leader his knife, which he crudely stabs into the ground, right next to your face, 
                            only just barely missing your head.
                            Fear and fright race through your mind anew, your entire body freezing up.
                            "<span class="bold brightred">Try to bite again, and I'll fuck you with my knife instead of my dick.</span>"
                        [/p]
                        [p]
                            As if to test you, the bandit leans in close once more, this time extending his wet and slimy tongue,
                            giving you a single long, gut-wrenching, spit-trailing lick from the base of your neck all the way to the underside of your muzzle.
                            A smug, self-satisfied smile creeps across his face, revelling in your dread-induced passivity.
                            "<span class="bold brightred">You taste as good as you smell.</span>" he mocks, once again licking his lips as if to empower his words.
                        [/p]
                        [p]
                            The kobold reaches for his own groin, `;
        if (male) {
            toReturn += `   guiding his member in between your scaly buttcheeks, 
                            its warm tapered tip pressing against your virginal tailhole,
                            spreading around just a single droplet of precum to serve as lubricant. `;
        } else {
            toReturn += `   lining up his member with your virginal hole, 
                            the tapered tip of his member slightly parting your folds.`;
        }

        toReturn += `       He lowers his head, his muzzle mere inches away from yours, close enough for you to feel his warm breath on your face.
                            "<span class="bold brightred">Let's see how tight you really are.</span>"
                        [/p]
                        [p]`;

        if (male) {
            toReturn += `   Then, with a single, firm thrust, the ringleader forces himself inside of you,
                            tears flowing liberally from your eyes as your tailhole is stretched wide to accomodate his large, throbbing member.
                            Taken by another man, whatever remained of your fighting spirit quickly fades.`;
        } else {
            toReturn += `   Then, with a single, firm thrust, the ringleader forces himself inside of you,
                            tears flowing liberally from your eyes as your tender insides are spread wide by his large, throbbing member.
                            A small trickle of blood oozes from your groin, dripping down onto the leafy forest floor.`;
        }

        toReturn += `   [/p]
                        [p]
                            He barely gets halfway in, lack of lubrication making his insertion [i]far[/i] more painful than it should have been.
                            Both your body and your mind feel like they're about to break.
                            "<span class="bold brightred">You weren't lying.</span>" he taunts, his words burning into your mind, 
                            along with the faint smell of alcohol lingering on his breath.
                            He pulls his member back, easing your pain for just a few moments, before ramming forwards once more,
                            harshly shoving the remainder of his length inside of you all the way to the very base of it, 
                            spearing you on the full extent of his cock, his scaly groin coming to rest ${male ? "against your taut ass" : "against yours"}.
                        [/p]
                        [p]
                            You scream out in pain, yet your shout quickly turns into anguished sobbing.
                            Your misery seems to fuel the bandit.
                            He grabs onto your shoulders, and begins to mercilessly [b]fuck[/b] you with a series of fierce, body-wracking thrusts,
                            his hips loudly slapping against your ${male ? "ass" : "inner thighs"} every time he bottoms out inside of the overly tight, 
                            warm confines of your ${male ? "tailhole" : "cunt"}.
                        [/p]
                        [p]
                            While your mind [b]hates[/b] just about [b]everything[/b] about this situation, 
                            your body slowly seems to be responding to the unwanted attention.
                            ${male ? "The tip of your member emerges from your cockslit" : "Your slit grows just a little more damp"}, 
                            the ringleader immediately noticing your body's betrayal.
                            "<span class="bold brightred">Liking this, little whore? 
                            I [i]knew[/i] you were dying for dick. 
                            You wanted us to fuck you instead of the dragoness, after all.</span>"
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/bandit/intervene/bargain/replace/virgin/rape-${male ? "male" : "female"}-full.png" target="_blank">
                                    <img class="nsfw" src="scenes/Forest/bandit/intervene/bargain/replace/virgin/rape-${male ? "male" : "female"}.png" title="Made by Sneaky">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            You turn your head to the side, facing away from him as you face away from the truth.
                            In the back of your head, you realize that he's right.
                            You [i]did[/i] ask for this.
                            Your tears spill down to the ground.
                        [/p]
                        [p]
                            The both of you begin to breathe more heavily, heated huffing and panting reverberating across the forest clearing.
                            He's getting close.
                            You can feel it.
                            His slow, but powerful rhythm has dissolved into a fast and brutal pace, his shaft pistoning in and out of you, 
                            thrusts slickened only by the small amount of precum constantly drooping from the tip of his cock,
                            doing little to properly lubricate your ${male ? "tailhole" : "cunt"}.
                        [/p]
                        [p]
                            The kobold's grip on your shoulders suddenly tightens tenfold.
                            He lets out a loud grunt, practically [b]pulling[/b] you just a little further onto his length.
                            He rams himself forwards with all of his might, burying his length deep inside of you as he reaches his end, 
                            a first wave of thick, warm spunk flooding your insides.
                        [/p]
                        [p]
                            The ringleader keeps himself hilted for the entire duration of his orgasm.
                            He grips your head and forces you to look at him, a dominating, possessive,
                            smug yet hate-filled stare on his face which burns into your retinas as squirt after squirt of ${male ? "" : "fertile"} semen
                            droops from his quivering length deep inside of you.
                        [/p]
                        [p]
                        `;

        if (male) {
            toReturn += `   Finally, he pulls out, a dribble of his spunk immediately flooding from your thoroughly abused tailhole.
                            He moves his member upwards, letting the final few bits of cum leaking from his member drip right on top of your own exposed length.
                            "<span class="bold brightred">Marked with superior seed.</span>" he jests, chuckling loudly as he rises to his feet.`;
        } else {
            toReturn += `   Finally, he pulls out, a dribble of his spunk immediately flooding from your thoroughly abused cunt, 
                            washing away the small bit of blood still clinging to your scales.
                            He moves his member upwards, shooting one final rope of sperm across your lower body, 
                            as though to signify how deep he planted his seed inside of you.
                            "<span class="bold brightred">You'll make a good mother.</span>" he jests, chuckling loudly as he rises to his feet.`;
        }

        toReturn += `   [/p]
                        [p]
                            "<span class="bold brightred">Who's next?</span>" the ringleader continues, looking at his peers.
                            You shudder and groan, quietly shaking your head no. 
                            They [b]all[/b] want a go?!
                            You can't take any more of this!
                            Then, just as the argument of who would go second was settled, a distant, yet very audible shout rings throughout the clearing:
                            "<span class="bold green">SOLDIERS!!</span>"
                        [/p]
                        [p]
                            The three kobolds surrounding you glance at one another, then towards you.
                            "<span class="bold brightred">You're lucky. We'll be seeing you later, [i]${male ? "bitch" : "mommy"}[/i].</span>" the leader remarks,
                            before bolting out into the forest with remarkable speed.
                            He's quickly followed by the bandit slated to go third, the one who was about to get his turn lingering just a little while longer,
                            staring at you longingly before making his escape.
                        [/p]
                        [p]
                            You lie there for a few seconds more, your brain slow to process what just happened.
                            Then, you realize you have to move.
                            Those [i]soldiers[/i] might just be the ones hunting you.
                            You wearily get onto all fours and crawl out of the clearing, hiding in a nearby bush, a clear trail of spunk easily giving away your position.
                        [/p]
                        [p]
                            You wait for several minutes, recovering from the harsh, traumatic experience you just went through.
                            Part of you [b]wishes[/b] those soldiers would run along and cut you down, so you wouldn't have to live with what just happened.
                            Yet as usual, you don't get what you wish for.
                            Nobody comes.
                        [/p]
                        [p]
                            You slowly rise to your feet.
                            ${(bra || underwear) ? "You leer longingly at your torn clothing, yet you don't dare to even try to recover any of it. Who knows, there may actually be soldiers about." : ""}
                            At least the bandits didn't take your coin pouch.
                            They were too obsessed with [b]fucking[/b] you to even notice it.
                            Wiping some of the spunk off your ${male ? "ass" : "groin"} with your hand, you decide to limp home naked,
                            clinging to the shadows to avoid any attention.
                        [/p]
                        [p]
                            [c]
                                [b]You are no longer a virgin.[/b]
                            [/c]
                        [/p]`;

        return  global.parser.parse(toReturn);
    }

    /**
     * Initializes all private variables.
     */
    function initVars() {
        didRock = (ctrl.bandit.getArrivalChoice() === ForestBanditArrivalChoiceEnum.ROCK);
        gender = ctrl.bandit.intervene.replaceGender();
        dick = ctrl.bandit.intervene.replaceDick();
        pants = ctrl.bandit.intervene.replacePants();
        underwear = ctrl.bandit.intervene.replaceUnderwear();
        bra = ctrl.bandit.intervene.replaceBra();
        male = (gender === GenderEnum.MALE);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.bandit.intervene.protestHasContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }
    
    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.protestContinue();
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditReplaceProtest";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.intervene.getReplaceChoice() !== ForestBanditReplaceChoice.PROTEST) {
            return "";
        }
        initVars();
        var change = ctrl.bandit.intervene.protestChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `</div>`;
        // Build children here.
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.intervene.getReplaceChoice() !== ForestBanditReplaceChoice.PROTEST) {
            return;
        }

        if (ctrl.bandit.intervene.protestHasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\bandit\intervene\bargain\replace\virgin\ForestBanditProtest.js

// scenes\Forest\bandit\intervene\bargain\replace\accept\ForestBanditAccept.js

/**
 * Accept part of the forest bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditAccept(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var enjoyId = "banditAcceptEnjoyButton";
    var hateId = "banditAcceptHateButton";

    var toSex = new ForestBanditAcceptSex(ctrl);

    var didRock;
    var gender;
    var dick;
    var pants;
    var underwear;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide not to struggle against the kobolds' grasp, instead staying still as their leader's hand finds its way atop your chest.
                        `;

        if (gender === GenderEnum.MALE) {
            toReturn += `   He runs his palm along your smooth, scaled torso, studying your face as if to gauge your reaction.`;
        } else {
            toReturn += `   He roughly grabs one of your breasts, groping it whilst studying your face as if to gauge your reaction.`;
        }

        toReturn += `       "<span class="bold brightred">Meek little bitch.</span>" the leader comments, after about a solid minute of the treatment.
                            A wide, ominous grin spreads across the kobold's face.
                            He finally lets go of your jaw.
                            "<span class="bold brightred">We're going to have a lot of [i]fun[/i] with you.</span>"
                        [/p]
                        [p]
                            With a simple nod from the boss, his underlings force you to your knees.
                            Your face is pushed forwards, directly into his groin, your snout ending up rubbing against his bare scales, 
                            his member resting against the side of your muzzle, the tip of it pressing into your cheek.
                        [/p]
                        [p]
                            While up until now, your mind was racing to come up with a new plan to escape the kobolds' clutches still,
                            the scent of cock slowly permeating your nostrils makes you realize that there [i]is[/i] no escape.
                            They [b]are[/b] going to have their way with you, and there's nothing left you can do to stop them.
                            You could, however, make this easier on yourself by [i]playing along[/i].
                            Heck, you might even end up enjoying this entire ordeal.
                        [/p]`;
        return  global.parser.parse(toReturn);
    }

    /**
     * Initializes all private variables.
     */
    function initVars() {
        didRock = (ctrl.bandit.getArrivalChoice() === ForestBanditArrivalChoiceEnum.ROCK);
        gender = ctrl.bandit.intervene.replaceGender();
        dick = ctrl.bandit.intervene.replaceDick();
        pants = ctrl.bandit.intervene.replacePants();
        underwear = ctrl.bandit.intervene.replaceUnderwear();
    }

    /**
     * Builds and returns the enjoy choice button.
     */
    function buildEnjoyChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeAcceptChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getAcceptChoice() == ForestBanditAcceptChoice.ENJOY;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${enjoyId}" ${disabled}>Enjoy it</button>`;

        var tooltip = `You might as well enjoy yourself and pleasure the bandits to the best of your abilities.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the hate choice button.
     */
    function buildHateChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeAcceptChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getAcceptChoice() == ForestBanditAcceptChoice.HATE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${hateId}" ${disabled}>Hate it</button>`;

        var tooltip = `You hate what they're doing to you. Lie there and take it. Hope they'll be done soon.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildEnjoyChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildHateChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables both buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + enjoyId), $("#" + hateId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the enjoy button.
     */
    function registerEnjoyListener() {
        $("#" + enjoyId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeAcceptChoice(ForestBanditAcceptChoice.ENJOY);
            loadNextPart(toSex);
        });
    }

    /**
     * Adds a listener to the hate button.
     */
    function registerHateListener() {
        $("#" + hateId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeAcceptChoice(ForestBanditAcceptChoice.HATE);
            loadNextPart(toSex);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditReplaceAccept";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.intervene.getReplaceChoice() !== ForestBanditReplaceChoice.ACCEPT) {
            return "";
        }
        initVars();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        // Build children here.
        part += toSex.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.intervene.getReplaceChoice() !== ForestBanditReplaceChoice.ACCEPT) {
            return;
        }

        toSex.registerListeners();

        if (ctrl.bandit.intervene.hasMadeAcceptChoice()) {
            return;
        }

        registerEnjoyListener();
        registerHateListener();
    };
}

var ForestBanditAcceptChoice = {
    ENJOY: 1,
    HATE: 2
};

// END scenes\Forest\bandit\intervene\bargain\replace\accept\ForestBanditAccept.js

// scenes\Forest\bandit\intervene\bargain\replace\accept\ForestBanditAcceptSex.js

/**
 * Sex part of the forest bandit encounter's accept route.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditAcceptSex(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "banditAcceptContinueButton";
    
    var didRock;
    var gender;
    var dick;
    var pants;
    var underwear;
    var male;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function buildEnjoy() {
        var toReturn = `[p]
                            You loudly sniff the ringleader's groin, taking in as much of his wonderful, manly scent as you possibly can.
                            You endearingly nuzzle his scales, before pursing your lips and giving his groin a single, small kiss.
                            Your enthusiasm seems to surprise him, and the group as a whole.
                            "<span class="bold brightred">Eager whore. If you like the smell, why not give it a taste?</span>"
                        [/p]
                        [p]
                            You instantly oblige, your tongue snaking out of your mouth, lapping at the edge of his cockslit, 
                            your tongue occasionally brushing against his protruding member.
                            He's right.
                            It [b]does[/b] taste as [b]potent[/b] as it smells.
                            As you continue to wetly tongue-bathe his scales, your tongue flicks closer and closer to his length.
                            Before long, you're outright licking his dick, no longer keeping up the pretense of simply doing as you are told.
                        [/p]
                        [p]
                            One of your hands makes its way to your own groin. You begin to play with yourself, 
                            ${male ? "jerking your emerging member" : "rubbing your moistening slit"} to the scent and the taste of the ringleader's tool.
                            Your show arouses the other men, who begin to touch themselves in turn.
                        [/p]
                        [p]
                            "<span class="bold brightred">Enough, you cock-crazed slut.</span>" comes the interrupting command of the ringleader.
                            You instantly pull your head back, suggestively placing your muzzle at the tip of his manhood.
                            He takes hold of his own cock, giving himself a few slow strokes, spreading your lingering spit along the length of his shaft.
                            "<span class="bold brightred">I'm not the only one here.</span>" he muses, his underlings stressing his words by taking just a little step closer,
                            their dicks on either side of your face.
                        [/p]
                        [p]
                            A sly grin grows on your visage.
                            You turn to the bandit to your right.
                            Winking at him, you tell him to get down on the ground.
                            He looks over to his boss, who gives him a small, affirmative nod.
                            The kobold then slowly lowers himself, first to a seated position, and then, encouraged by your hand signals, 
                            he lies down flat on his back, his erection jutting upwards proudly.
                        [/p]
                        [p]
                            The ringleader seems to understand what you're trying to pull off.
                            He doesn't halt you as you drop backwards onto all fours, skillfully crawling on top of the prone gangmember,
                            before lowering your body once more, directly on top of his, his dick ending up directly in between your thighs,
                            mere inches away from your ${male ? "own shaft" : "dripping wet slit"}.
                        [/p]
                        [p]`;
        
        if(male) {
            toReturn += `   You get comfortable on top of the kobold for a moment, 
                            enticingly grinding your taut behind against his groin whilst encapsulating his member with your scaly thighs,
                            wriggling your legs about ever so slightly, a single droplet precum dripping from the tip onto your leg.
                            After a bit more teasing, you lift your upper body off of his, coming to a seated position.`;
        } else {
            toReturn += `   The eager kobold reaches out from underneath you, groping your chest.
                            You let him play with your breasts for a few moments, before lifting your upper body off of his, coming to a seated position.`;
        }
        
        toReturn += `       He isn't the only one around, after all.
                        [/p]
                        [p]
                            What you did wasn't quite what the ringleader had in mind, but he seems to be more than okay with it,
                            idly stroking himself to the tempting display you're putting on.
                            As you sit upright, he approaches you from the side, rather crudely shoving his dick in your face,
                            wiping a small build-up of pre onto your cheek.
                        [/p]
                        [p]
                            "<span class="bold brightred">You know what to do.</span>" he speaks, in a low, rumbling voice,
                            as if he's trying to hide his arousal behind a [i]tough guy[/i] attitude.
                            You find it endearing, almost cute even.
                            Turning your head to face his member, you part your lips and wrap them around the tapered tip,
                            your tastebuds instantly rewarded with a fresh squirt of precum along the entire length of your tongue,
                            your ${male ? "cock quivering" : "slit dampening"} in response to the absolutely [b]divine[/b] taste.
                        [/p]
                        [p]
                            You eagerly begin to bob up and down the ringleader's shaft,
                            swirling your tongue around it to soak in as much of its taste as you can,
                            revelling in the flavor of sheer [b]cock[/b].
                            Simultaneously, you move one of your hands to the dick of the kobold below you, grabbing it 
                        `;
        
        if(male) {
            toReturn += `   and pressing it against your own shaft, relishing in the warmth radiating from his length.
                            Frotting the two of you together, you take hold of both of your dicks, 
                            and you deftly begin to stroke, jerking off the bandit and yourself, whilst rubbing your shafts against one another.`;
        } else {
            toReturn += `   and pressing it against your sopping wet slit, in between your lower lips, parting them slightly, relishing in the warmth radiating from his shaft.
                            Moving your hips, you begin to rub your cunt against the kobold's shaft, smearing your feminine fluids along its entire length,
                            your thighs slowly milking his cock, your hand rubbing along his tip, which protrudes from in between your legs.`;
        }
        
        toReturn += `   [/p]
                        [p]
                            You pull off of the leader's dick for a moment, thick strands of spit connecting your muzzle to his length still.
                            Looking around, you notice one of the kobolds isn't really partaking in this entire affair.
                            He stands to the side, masturbating lonely whilst watching the other two have fun with you.
                            Using your free hand, you beckon him over, implying you'll do the jerking [i]for him[/i] with a few obscene gestures.
                        [/p]
                        [p]
                            He simply shakes his head no, continuing to mind his own business.
                            Still, you insist.
                            "<span class="bold brightred">Stop distracting the bitch. Get over here. Do your [i]thing[/i].</span>" his boss finally rules.
                            He rolls his eyes and sighs audibly, slowly approaching the three of you.
                        [/p]
                        [p]
                            As opposed to taking up your handjob-invitation, however, the odd kobold circumvents you, instead approaching the foot-end of the situation.
                            He drops to his knees, grabs your ankles and lifts up your legs, 
                            inadvertently giving his prone partner's dick a single stroke using your thighs,
                            before [b]burying[/b] his snout against your soles, sniffing loudly, 
                            his cock visibly twitching between his legs as his eyes roll into the back of his head.
                        [/p]
                        [p]
                            [i]Weirdo[/i].
                            Still, who are you to judge?
                            You're slutting yourself out to a bunch of bandits you've never met before.
                            After satisfying his unusual craving, the kobold lowers your hindpaws, guiding them directly onto his groin, 
                            your feet on either side of his throbbing length.
                            As the ringleader said before: [i]you know what to do[/i].
                        [/p]
                        [p]
                            You wrap your soles around the odd bandit's shaft, and begin to give him a footjob.
                            Slow at first, but your pace rapidly picks up as you grow more comfortable with the weird sexual act.
                            Sensing the small intermission has ended, the kobold beneath you begins to thrust his hips,
                            fucking your thighs while grinding his member 
                        `;
        
        if(male) {
            toReturn += `   against your own, your hand keeping the two together tightly,
                            a considerable amount of your combined precum slickening your palm and your both your shafts,
                            smoothing his frantic humping.`;
        } else {
            toReturn += `   against your cunt, a considerable amount of your feminine fluids mixed with his precum slickening both his shaft and your warm, scaly thighs,
                            heated up through sheer friction to the [b]perfect[/b] cock-warming temperature.`;
        }
        
        toReturn += `   [/p]
                        [p]
                            The leader, too, feels like it's about darn time you got back to sucking his dick.
                            He grabs your head and physically [b]pulls[/b] your attention away from the foot-enamored bandit, tugging your muzzle onto his member once more.
                            You eagerly resume your sloppy blowjob,
                            spit liberally dribbling down to the leafy forest floor as your lips slide along the length of his considerable cock,
                            filling your mouth to the brim, kind or perhaps [i]pleased[/i] enough not to make you [b]choke[/b] on it.
                        [/p]
                        [p]
                            Besides being a weirdo, the bandit toying with your feet also appears to have somewhat of a hair trigger, 
                            or maybe he's just so [i]happy[/i] that he [b]finally[/b] gets to act out something he's [b]always[/b] wanted to do.
                            Either way, he blows his load mere minutes after joining the fun, coating your soles and the sides of your feet in his warm, 
                            sticky seed, the heady scent of which reaches all the way to your nose.
                        [/p]
                        [p]
                            The others seems to smell it too.
                            It appears to drive them into some sort of [i]mating frenzy[/i], the kobold below you upping the pace of his thrusts,
                            rampantly ${male ? "frotting against your shaft" : "massaging your cunt with his length"} while the leader of the group steadies your head, 
                            and begins to hump against your lips, outright [b]fucking[/b] your face.
                        [/p]
                        [p]
                            The leader lets out a loud, domineering grunt as he slides most of his shaft from your maw, 
                            keeping only the tapered tip inside of your mouth.
                            Giving his spit-slick member a few pumps with his hand, a veritable [b]torrent[/b] of cum erupts from his cock,
                            splattering across the entire length of your tongue, rapidly filling up your mouth.
                            You greedily swallow every single drop of his hefty, salty, undeniably [b]masculine[/b] spunk,
                            the taste of which spurs a sexual [b]need[/b] inside of you.
                        [/p]
                        [p]
                            You begin to roll your hips back against the kobold below you,
                        `;
        
        if(male) {
            toReturn += `   frotting your dick against his, as he grinds his against yours.
                            You can feel your end rapidly approaching.
                            With a loud, rather feminine moan, you wildly buck your hips against his one final time, an intense orgasm rocking your body hard.
                            Your own thick cum oozes from the tip of your member, drooping down along the other kobold's shaft.
                            This seems to drive him to the edge in turn, a mixture between a satisfied sigh and a groan coming from behind you as he squirts
                            his first rope of cum, which lands on top of your belly.`;
        } else {
            toReturn += `   frantically rubbing your cunt against his cock, sliding along his length as he thrusts against your thighs.
                            You can feel your end rapidly approaching.
                            With a loud, feminine moan, you wildly buck your hips against his one final time, an intense peak of pleasure rocking your body hard.
                            Feminine fluids leak liberally from your quivering cunt, their warmth washing over the other kobold's cock, driving him to the edge in turn.
                            You hear a mixture between a satisfied sigh and a groan coming from behind you as he squirts the first wave of his climax, 
                            the thick rope of cum splattering down onto the top of your leg.`;
        }
        
        toReturn += `
                        [/p]
                        [p]
                            Both your orgasms last for several more moments, both your legs and your crotch ending up covered in kobold spunk.
                            After revelling in post-orgasmic bliss for a little while longer, you slowly slide off of the bandit, 
                            his groin almost as messy and sticky as yours is.
                        [/p]
                        [p]
                            You take a bit to properly recover from the intensive foursome.
                            After giving the cum splattered on your legs and your feet just a little taste, you slowly get up from the ground.
                            All three of the men look just as spent as you are.
                            But then you realize something.
                        [/p]
                        [p]
                            You take a few steps away from the gang, before bending over and lifting your tail,
                            immediately drawing the attention of all three bandits.
                            You merrily muse that you haven't actually been fucked yet, 
                            and going off of the lusty looks on the kobolds' faces, they are [b]definitely[/b] up for a round number two.
                        [/p]
                        [p]
                            Then, just as the prone kobold begins to rise to his feet, and the others start to move towards you, 
                            a distant, yet very audible shout rings throughout the clearing:
                            "<span class="bold green">SOLDIERS!!</span>"
                        [/p]
                        [p]
                            "<span class="bold brightred">Your ass will have to wait.</span>" the leader speaks, collecting his clothing in a hurry.
                            The other kobolds seem to join him, and as soon as they're half-dressed, they bolt into the forest.
                            Your own clothing torn to shreds, you decide to run after them, just in case those [i]soldiers[/i] are actually the men chasing you.
                            Sadly, the kobolds are already out of sight by the time you leave the clearing.
                            Instead of aimlessly running into the forest, you decide to head back home post-haste.
                        [/p]
                        [p]
                            [c]
                                [b]You have lost your oral virginity![/b]
                            [/c]
                        [/p]`;
        toReturn += ctrl.bandit.intervene.replaceEagerChange();
        return toReturn;
    }

    function buildHate() {
        var toReturn = `[p]
                            You immediately try to pull your head back, yet you are being firmly kept in place by one of the kobolds to your rear,
                            forced to inhale the stench of the boss kobold's groin, which is filling your lungs at an alarming rate, corrupting your body from the inside-out.
                            At least that's what you imagine is happening, as you are kept idly in place for an entire minute more,
                            until finally the leader himself removes your head from his groin, your muzzle ending up directly in front of his imposing member.
                        [/p]
                        [p]
                            "<span class="bold brightred">Now that you've smelled it, how about a taste?</span>" he taunts, removing his hand from your head,
                            instead taking hold of his dick and condescendingly tapping your snout with it, as though ordering you to [i]open up[/i].
                            You refuse, turning your head to the side, scoffing at the very notion of sucking that [b]thing[/b].
                        [/p]
                        [p]
                            Luckily for you, he doesn't expect you to [b]suck[/b] it just yet.
                            The leader harshly slaps you across the cheek, before forcibly turning your head towards his member once more.
                            "<span class="bold brightred">LICK!</span>" he commands loudly and authoratively.
                            Glaring up at him with tears in your eyes, you slowly open your maw.
                        [/p]
                        [p]
                            Your tongue reluctantly snakes out of your mouth, and you tentatively lap at the tapered tip of his member just [b]once[/b],
                            collecting a thick globule of precum which rolls over the entire length of your tongue, directly into your mouth, covering
                            nigh all of your tastebuds along the way.
                            The flavor of it almost makes you [b]retch[/b].
                        [/p]
                        [p]
                            Grumbling loudly as he realizes that's the best he'll get out of you, the boss takes a step back, nodding at his associates,
                            who promptly push you forwards, shoving you onto all fours.
                            Then, things speed up.
                        [/p]
                        [p]
                            Your tail gets harshly yanked upwards.
                            You try to yelp out in pain, but the moment you open your mouth, the ringleader in front of you thrusts forward,
                            ramming his member past your lips and into your mouth.
                            You consider biting down, but while using one hand to hold your head in place, he brandishes his knife in the other, 
                            threateningly pressing it against your cheek, doing all but drawing blood.
                            "<span class="bold brightred">Bite and die.</span>" he states the obvious, slowly pushing just a little bit more of his length into your mouth.
                        [/p]
                        [p]
                            The taste of it almost makes you sick, his shaft resting atop of your tongue, filling your mouth to the brim,
                            saturating your sense of taste with its intense, purely [b]masculine[/b] flavor.
                            To make matters worse, you feel a pair of groping hands at your rear, inching ever closer to your tender, virginal ${male ? "tailhole" : "cunt"}.
                            Even before the leader begins to rock back and forth, slowly thrusting in and out of your mouth,
                            one of the bandits' fingers has already found its way inside of your ${male ? "ass" : "slit"}, 
                            making you gasp around the dick in your muzzle.
                        [/p]
                        [p]
                            The kobold toying with your ${male ? "rear" : "twat"} immediately begins to pump his finger in and out of your hole,
                            and before long, a second one is added, loosening you for what is to come.
                            The leader picks up his pace, removing his knife from your head to avoid [i]accidentally[/i] cutting you, 
                            as he outright begins to [b]fuck[/b] your face.
                            Every few thrusts, a thick squirt of precum is splayed across your tongue, adding to the heady, unsavory taste of unwashed cock.
                        [/p]
                        [p]
                            The fingers pull out completely, and one of the kobolds takes up position behind you.
                            Out of pure sadism, he cruelly twists your tail, making you howl out in pain, your shout muffled by the leader's groin.
                            "<span class="bold brightred">Noisy bitch.</span>" he observes, tossing his weapon to the side so he can grab your head with both of his hands.
                            Then, as if to silence you, the boss of the gang pulls your head further into his lap, slamming his dick down your throat,
                            your snout coming to rest against his groin as he bottoms out.
                        [/p]
                        [p]
                            You sputter and gag, barely managing to contain the contents of your stomach as he harshly holds you in place, your air supply completely cut off.
                            Just then, as if to [b]break[/b] you completely, the kobold behind you lines up his member, and with a single, smooth thrust, 
                        `;
        
        if(male) {
            toReturn += `   he rams himself inside of your slightly-loosened tailhole, 
                            forcing over half of his member inside of your virginally tight hole in a single thrust.`;
        } else {
            toReturn += `   he rams himself inside of your finger-prepped slit, 
                            spreading your folds and forcing over half of his member inside of your overly tight hole.
                            A bit of blood immediately oozes from your cock-filled cunt, embodying the loss of your virginity.`;
        }
        
        toReturn += `   [/p]
                        [p]
                            You try to scream around the cock buried down your throat, yet fail to do so.
                            Even if you [i]could[/i] shout, you doubt any sensible words would leave your lips.
                            You want to struggle, to bite and fight, to shake them off and [b]kill[/b] them for what they're doing to you,
                            yet your body feels [b]weak[/b] to the point where your arms and legs can barely hold the weight of your own body.
                        [/p]
                        [p]
                            The bandit leader pulls back slightly, allowing you to breathe in some much-needed air through your nose, 
                            which inadvertently also makes you inhale the scent of his groin, 
                            the heavy concentration of male pheromones immediately clouding your mind,
                            hampering any kind of logical thought or reasoning.
                        [/p]
                        [p]
                            The leader exchanges an ominous nod with the bandit standing behind you,
                            and then, both of them begin to outright [b]fuck[/b] you at the same time.
                            The boss humps your face, his cock travelling down your throat every time he bottoms out, 
                            your snout wetly slapping against his groin-scales,
                            thick strands of your own spit bridging the gap between your muzzle and his nether regions whenever he pulls back.
                            The kobold behind you retracts his member from your ${male ? "rear" : "cunt"}, before plunging forwards once more,
                            spearing your body on the entire length of his cock, the girthy base only [b]barely[/b] managing to fit inside of you.
                        [/p]
                        [p]
                            After forcing the entirety of his shaft into your body, the bandit behind you begins to wildly hump away,
                            [b]pounding[/b] you hard enough to rock your entire frame back and forth, 
                            pushing you just a little deeper onto his boss' cock with every fierce thrust.
                            The only kobold not actively participating in the brutal gangrape is standing off to the side,
                            lustily jerking off to the scene unfolding before his very eyes.
                        [/p]
                        [p]
                            Luckily for you, the overly tight confines of your throat and your ${male ? "ass" : "slit"} quickly prove too much 
                            for both the leader and his subordinate.
                            The less important of the two is first to blow his load.
                            You can feel his entire length quiver and tremble inside of you,
                            yet just before he reaches his climax, he pulls out, gives himself a few more strokes,
                            and then heatedly squirts rope after rope of warm, thick, sticky cum all across your backside.
                            ${male ? "" : `"<span class="bold brightred">Only [i]I[/i] get to [i]breed[/i].</span>" the boss of the gang mumbles, as if to remind his partner, who was clearly already aware of that fact.`}
                        [/p]
                        [p]
                            Seeing his comrade finish, the leader's thrusts increase in ferocity.
                            He [b]crams[/b] his cock down your maw every time he pistons forwards, your snout [b]battering[/b] against his groin,
                            until he finally hilts himself inside of your mouth one final time.
                            With a loud grunt, the brutish kobold [b]tugs[/b] your head onto his cock, 
                            forcing just a [i]little bit more[/i] of his length down your throat before reaching his end,
                            wave after wave of his hefty spunk pumping directly into your stomach.
                        [/p]
                        [p]
                            Running out of air, you begin to struggle against the boss' grasp, 
                            trying your hardest, but utterly failing to pull your head back from his iron grip.
                            He intends to shoot [b]every single last drop[/b] of his cum down your throat, a fact he loudly proclaims,
                            but his words are lost on you in your panicking state.
                            Still, as your brain struggles to continue to function, your protests slowly die down.
                            As the corners of your eyes begin to go black, on the verge of passing out, the ruffian finally pulls back.
                        [/p]
                        [p]
                            Your arms give out and you fall to the ground, coughing out bits and pieces of kobold cum while loudly gasping and wheezing for air.
                            A couple of mixed cum and spit droplets droop down from above, yet you don't care about the goop right now, 
                            focussing on simply inhaling as much and as steadily as you can.
                            Your senses slowly flood back to you, your jaw aching and your ${male ? "ass" : "cunt"} feeling stretched out to [b]way[/b] beyond its limits.
                        [/p]
                        [p]
                            "<span class="bold brightred">Go fuck her so we can leave.</span>" is the first thing you hear as your mind recovers enough 
                            to be able to understand words once again.
                            However, as the third kobold slowly approaches you, another distant, yet very audible voice rings throughout the forest clearing:
                            "<span class="bold green">SOLDIERS!!</span>"
                        [/p]
                        [p]
                            The leader grumbles loudly.
                            "<span class="bold brightred">Your turn will have to wait.</span>" he speaks, halting his approaching peer.
                            His decision is met with little resistance.
                            The three kobolds pick up their clothing, leaving you in a puddle of semen and spit as they scurry out of the clearing and into the forest.
                        [/p]
                        [p]
                            You slowly process what just happened.
                            Soldiers...those could be the ones chasing after you.
                            You cough up another sizable amount of spunk, before raising your body onto your hands and feet once more,
                            crawling out of the clearing, and into a nearby bush, a clear trail of jizz easily giving away your position.
                        [/p]
                        [p]
                            You wait for several minutes, recovering from the harsh, traumatic experience you just went through, almost throwing up more than once.
                            Part of you [b]wishes[/b] those soldiers would run along and cut you down, so you wouldn't have to live with what just happened.
                            Yet as usual, you don't get what you wish for.
                            Nobody comes.
                        [/p]
                        [p]
                            You slowly rise to your unsteady feet.
                            You leer longingly at your torn clothing, yet you don't dare to even try to recover any of it. 
                            Who knows, there may actually be soldiers about.
                            At least the bandits didn't take your coin pouch.
                            They were too obsessed with [b]fucking[/b] you to even notice it.
                            Wiping some of the spunk off your ${male ? "ass" : "groin"} and your muzzle with your hand, you decide to limp home naked,
                            clinging to the shadows to avoid any attention.
                        [/p]
                        [p]
                            [c]
                                [b]You have lost your oral virginity.[/b]
                            [/c]
                        [/p]
                        [p]
                            [c]
                                [b]You are no longer a virgin.[/b]
                            [/c]
                        [/p]`;
        toReturn += ctrl.bandit.intervene.replaceHateChange();
        return toReturn;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = ``;
        if (ctrl.bandit.intervene.getAcceptChoice() === ForestBanditAcceptChoice.ENJOY) {
            toReturn += buildEnjoy();
        } else {
            toReturn += buildHate();
        }
        return  global.parser.parse(toReturn);
    }
    
    /**
     * Initializes all private variables.
     */
    function initVars() {
        didRock = (ctrl.bandit.getArrivalChoice() === ForestBanditArrivalChoiceEnum.ROCK);
        gender = ctrl.bandit.intervene.replaceGender();
        dick = ctrl.bandit.intervene.replaceDick();
        pants = ctrl.bandit.intervene.replacePants();
        underwear = ctrl.bandit.intervene.replaceUnderwear();
        male = (gender === GenderEnum.MALE);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.bandit.intervene.acceptHasContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }
    
    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.acceptContinue();
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditReplaceAcceptSex";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.bandit.intervene.hasMadeAcceptChoice()) {
            return "";
        }
        initVars();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        // Build children here.
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.bandit.intervene.hasMadeAcceptChoice()) {
            return;
        }

        if (ctrl.bandit.intervene.acceptHasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\bandit\intervene\bargain\replace\accept\ForestBanditAcceptSex.js

// scenes\Forest\bandit\intervene\bargain\relieve\ForestBanditRelieve.js

/**
 * Relieve choice of the bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditRelieve(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var eagerId = "banditRelieveEagerButton";
    var reluctantId = "banditRelieveReluctantButton";

    var toSex = new ForestBanditRelieveSex(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var didRock = (ctrl.bandit.getArrivalChoice() === ForestBanditArrivalChoiceEnum.ROCK);
        var toReturn = `[p]
                            You realize that you can't [i]stop[/i] the bandits.
                            They outnumber you, they have [i]actual[/i] weapons, and there's not a doubt in your mind that they know how to use them.
                            You are reminded of an old saying: [i]if you can't beat'em, join'em[/i].
                            That is to say, if you can't prevent the dragoness from getting [i]raped[/i],
                            you might as well [i]ease[/i] her burden by getting some of the kobolds off yourself.
                        [/p]
                        [p]
                            You still need to figure out a way to convince the gang to [i]give up[/i] a dragoness, 
                            one of the most sought after sexual partners, in favor of plain old [b]you[/b].
                            "<span class="bold brightred">Well?! What do you want?!</span>" the ringleader impatiently shouts,
                            yet you don't let him interrupt your train of thoughts.
                            Wait a minute.
                            The leader.
                            Yeah, that might work.
                        [/p]
                        [p]
                            "<span class="bold brightred">What, you mute?! Speak, or I'll gut you!</span>" the knife-wielding kobold threatens.
                            With a devious and bold plan hatched, you finally acknowledge him,
                            opening up with a flattering comment about how [i]great[/i] a job [b]he[/b] did in tackling the dragoness.
                            Hatred quickly melts from his face, and in its place comes an appreciative blush combined with a devilish grin.
                        [/p]
                        [p]
                            Seemingly on the right track, you continue to push your luck.
                            You explain that you witnessed [b]his[/b] [i]grand triumph[/i] from behind the cover of a nearby tree.
                            While [b]he[/b]'s doing an excellent job at keeping the dragoness in check, you mention that there's a small [i]problem[/i] with his current plan.
                            A little oversight, nothing more, you assure the kobold.
                        [/p]
                        [p]
                            Your scheme appears to be working, thus far.
                            You've smooth talked the leader of the gang to the point where he doesn't even ask you [i]what[/i] is wrong with his scheme,
                            he simply waits for you to explain yourself out of your own volition, a marked improvement from the angry 
                            ${didRock ? "knife waving" : "sneering"} you received about a minute ago.
                        [/p]
                        [p]
                            You explain that there's only [i]one[/i] dragoness to split amongst all ${didRock ? "four" : "five"} of them.
                            Hardly a problem, [b]but[/b] [i]you[/i], as a mere spectator to the entire ordeal, believe that wouldn't be very [i]fair[/i].
                            After all, [i]the great leader[/i] did all of the work, assisted by the ${didRock ? "proud man" : "two proud men"} holding down the woman's arms.
                        [/p]
                        [p]
                            While you completely agree that he and his ${didRock ? "finest comrade" : "finest couple of henchmen"} [b]deserve[/b] to reap the spoils of their labour,
                            you personally think that the two kobolds pinning down the dragoness' legs didn't really do all that much.
                            As such, you feel that they don't [i]deserve[/i] a go at one of the [i]hottest[/i], 
                            [i]tightest[/i] [b]most fuckable[/b] things in these entire woods.
                            They [i]do[/i], however, deserve [i]some[/i] compensation for their effort.
                            What a conundrum!
                        [/p]
                        [p]
                            The two bandits you spoke ill of glare at you much like their leader used to.
                            They obviously disagree with just about [b]everything[/b] you're saying.
                            The well-dressed and well-armed kobold, on the contrary, 
                            is eating out of the palm of your hand, nodding agreeably in response to your every word.
                            You've got him hooked.
                            Now to reel him in.
                        [/p]
                        [p]
                            In order to [i]solve[/i] this dilemma, you suggest the following:
                            the [i]great[/i] leader and his right-hand ${didRock ? "man" : "men"} get to [i]abuse[/i] the [b]dragoness whore[/b] for as long as they'd like,
                            while the two [i]inept[/i] members of the gang sate their lust on something -- some[i]one[/i] -- of lesser value.
                            Someone like [b]you[/b].
                        [/p]`;
        if (didRock) {
            toReturn += `[p]
                            Right as it seems you've [b]completely[/b] convinced the ringleader,
                            a sudden thought strikes his otherwise compliment-addled mind.
                            His face scrunches up inquisitively, and his eyes narrow.
                            In a questioning voice, he speaks:
                            "<span class="bold brightred">Why did you throw a rock at us?</span>"
                        [/p]
                        [p]
                            Luckily, you manage to keep your cool.
                            With a smug, toothy smirk you point towards their unconscious comrade, 
                            and reply that [i]he[/i] didn't deserve neither the dragoness [i]nor[/i] you.
                            Your apparent cruelness seems to have struck the leader's fancy.
                            He burst out into hearty laughter, amused sufficiently by your reasoning to [i]overlook[/i] your earlier transgression.
                        [/p]`;
        }
        toReturn += `   [p]
                            Complaints emerge from the two belittled gangmembers, yet they are quickly silenced by their boss.
                            He has made his decision.
                            He'll follow your [i]advice[/i].
                            Reluctantly, the scorned felons let go of the dragoness' legs.
                            Turning away from the scene, their anger-filled eyes fall upon [b]you[/b].
                            As the remaining bandits begin to abuse the woman, the pair of upset outcasts advance in your direction.
                        [/p]
                        [p]
                            To make sure they don't get any [i]funny ideas[/i], you preemptively drop to your knees.
                            You don't intend to lose your virginity to these disgruntled muggers.
                            Mistaking your [i]precaution[/i] for eagerness, they strip out of their meagre clothing mid-march,
                            a pair of fully erect kobold dicks of considerable size coming into view.
                        [/p]`;
        
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the eager choice button.
     */
    function buildEagerChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeRelieveChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getRelieveChoice() == ForestBanditRelieveChoice.EAGER;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${eagerId}" ${disabled}>Eagerly</button>`;

        var tooltip = `Fervently jerk and suck the two bandits off. 
                       Maybe you can even touch yourself a little, if you can spare one of your hands for a few minutes.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 205px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the reluctant choice button.
     */
    function buildReluctantChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeRelieveChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getRelieveChoice() == ForestBanditRelieveChoice.RELUCTANT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${reluctantId}" ${disabled}>Reluctantly</button>`;

        var tooltip = `Hesitantly stroke the two kobolds to orgasm.
                       They aren't expecting you to use your muzzle, are they?`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">How do you service them?</div>`;
        choice += buildEagerChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildReluctantChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all four buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + eagerId), $("#" + reluctantId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the bargain button.
     */
    function registerEagerListener() {
        $("#" + eagerId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeRelieveChoice(ForestBanditRelieveChoice.EAGER);
            loadNextPart(toSex);
        });
    }

    /**
     * Adds a listener to the bargain button.
     */
    function registerReluctantListener() {
        $("#" + reluctantId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeRelieveChoice(ForestBanditRelieveChoice.RELUCTANT);
            loadNextPart(toSex);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditRelieve";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.intervene.getBargainChoice() !== ForestBanditBargainChoice.RELIEVE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        // Build children here.
        part += toSex.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.intervene.getInterveneChoice() !== ForestBanditInterveneChoiceEnum.BARGAIN) {
            return;
        }

        toSex.registerListeners();

        if (ctrl.bandit.intervene.hasMadeRelieveChoice()) {
            return;
        }

        registerEagerListener();
        registerReluctantListener();
    };
}

var ForestBanditRelieveChoice = {
    EAGER: 1,
    RELUCTANT: 2
};

// END scenes\Forest\bandit\intervene\bargain\relieve\ForestBanditRelieve.js

// scenes\Forest\bandit\intervene\bargain\relieve\ForestBanditRelieveSex.js

/**
 * Sex part of the relieve choice of the bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditRelieveSex(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "banditRelieveContinueButton";

    var didRock;
    var gender;
    var dick;
    var pants;
    var underwear;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the eager sex part.
     * @returns {string} The eager sex part.
     */
    function buildEager() {
        var toReturn = `[p]
                            A mischievous smile creeps across your face upon seeing that the bandits were definitely 
                            [i]not[/i] overcompensating for anything with their rowdy behavior.
                            As they bridge the final few paces to where you're kneeling,
                            you assure the well-hung highwaymen that they can have [b]a lot[/b] more fun with a willing 
                            kobold than they can have with an [b]unwilling[/b] dragon.
                            To stress your words, you open your maw wide, and let your long tongue loll out to its full extent.
                        [/p]
                        [p]
                            While at first less than pleased to be torn away from their dragoness fucktoy,
                            the two bandits' attitude quickly changes at the sight of your eagerness [i]and your tongue[/i].
                            They stand in front of you, side by side, members at full attention, pointing directly at your face.
                        [/p]
                        [p]
                            The masculine smell of sheer [b]cock[/b] reaches your nose, slightly clouding your mind and 
                            ${dick ? "spurring your own length to emerge from its slit" : "dampening your groin"}.
                            You decide not to keep them, or yourself for that matter, waiting any longer.
                        [/p]
                        [p]
                            Out of the corner of your eye, you spot a glistening bauble of precum hanging precariously from the tapered tip of the shaft to your left.
                            You lean in to collect the salty fluid on your tongue, licking it from the tip of the bandit's cock, 
                            letting it linger on your tastebuds whilst revelling in the flavor.
                        [/p]
                        [p]
                            In the meanwhile, you firmly grip the other outlaw's dick.
                            After taking a moment to teasingly toy with the head of his member using your thumb,
                            you begin to stroke him properly, jerking him off with the warm, soft palm of your hand,
                            your fingers clutched tightly around his throbbing length.
                        [/p]
                        [p]
                            As you retract your tongue once more, the taste of precum spreads throughout your mouth.
                            [b]You like it.[/b]
                            You want [b]more[/b].
                            Using your free hand, you tilt the shaft to your left upwards ever so slightly, 
                            just enough for you to slide your muzzle underneath it, pressing your snout into his groin.
                            With your nostrils rubbing against the very base of his member, you take a loud [b]sniff[/b],
                            practically [b]snorting[/b] the thick miasma of pure male pheromones wafting from his cockslit.
                        [/p]
                        [p]
                            Your senses overload.
                            Your mind goes blank. `;

        if (dick) {
            toReturn += `Your dick rapidly grows to full mast`;
            if (underwear) {
                toReturn += `, a dark stain of arousal forming on your ${global.kobold.inventory.createItemWithTooltip(underwear)}.`;
            } else if (pants) {
                toReturn += `, a dark stain of arousal forming on your ${global.kobold.inventory.createItemWithTooltip(pants)}.`;
            } else {
                toReturn += `, a droplet of your own arousal dripping to the leafy forest floor.`;
            }
        } else {
            toReturn += `Your slit moistens`;
            if (underwear) {
                toReturn += `, a dark stain of arousal forming on your ${global.kobold.inventory.createItemWithTooltip(underwear)}.`;
            } else if (pants) {
                toReturn += `, a dark stain of arousal forming on your ${global.kobold.inventory.createItemWithTooltip(pants)}.`;
            } else {
                toReturn += `, a droplet of your arousal running down your naked thigh.`;
            }
        }


        toReturn += `       You let go of the kobold's length, letting it flop down onto the top of your muzzle,
                            a well-timed squirt of precum splattering against your scales.
                            You move your hand towards your own groin`;

        if (pants) {
            toReturn += `, sliding it underneath the waistband of your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (underwear) {
                toReturn += `, and into your ${global.kobold.inventory.createItemWithTooltip(underwear)}`;
            }
            toReturn += `.`;
        } else if (underwear) {
            toReturn += `, sliding it underneath your ${global.kobold.inventory.createItemWithTooltip(underwear)}.`;
        } else {
            toReturn += `.`;
        }

        if (dick) {
            toReturn += ` You take hold of your own member, and deftly start to stroke yourself.`;
        } else {
            toReturn += ` You start feverishly rubbing your [b]dripping wet[/b] slit with the flat of your hand.`;
        }

        toReturn += `   [/p]
                        [p]
                            After huffing the bandit's groin for a few more seconds,
                            you purse your lips, and plant a small kiss atop the base of his dick,
                            before letting your tongue slither out of your mouth.
                            Pressing the warm, wet appendage against his cock,
                            you slowly tilt your head backwards, 
                            lifting his shaft upwards as your tongue slides along its underside,
                            leaving a small, glistening trail of saliva in its wake. 
                        [/p]
                        [p]
                            After you reach the peak of your unorthodox, full-length lick,
                            the kobold's member slips off of the top of your snout, 
                            your tongue flicking along its tapered tip on its way down.
                            You glance upwards at the mugger.
                            He [b]definitely[/b] seems to be enjoying the treatment.
                        [/p]
                        [p]
                            Lowering your head in pursuit of cock, you decide it's time to stop fooling around.
                            Your lips part, and you finally take the first few inches of the bandit's length into your mouth,
                            swirling your tongue around his shaft.
                        [/p]
                        [p]
                            You haven't forgotten about your other suitor, either.
                            While still dutifully jacking him off, you slowly tug his member towards your face, pressing the very tip against your cheek.
                            Letting his dick frot against your scales, his precum quickly gets slathered all across the side of your face,
                            especially as you begin to bob up and down the leftmost bandit's length, taking more and more of his meat into your mouth.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/bandit/intervene/bargain/relieve/eager-threesome-${dick ? "male" : "female"}-full.png" target="_blank">
                                    <img class="nsfw" src="scenes/Forest/bandit/intervene/bargain/relieve/eager-threesome-${dick ? "male" : "female"}.png" title="Made by Sneaky">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            Luckily, the thickly-scented liquid eases the frottage, 
                            the dick of the kobold on the right rubbing along the side of your cheek every time you go down on his friend's shaft.
                            It isn't long until he begins to softly buck against your grasp,
                            thrusting his member into your hand and against your face.
                        [/p]
                        [p]
                            The dominating taste of [b]cock[/b] and the constant dribble of salty pre leaking out on top of the entire length of your tongue
                            spur your own masturbation, fueling you
                        `;

        if (dick) {
            toReturn += ` to up the pace of your rampant masturbation, fervently jerking yourself off. `;
        } else {
            toReturn += ` to slip a single finger inside of your tight, moist folds, pumping it in and out of yourself at a high pace. `;
        }

        if (pants) {
            toReturn += `You interrupt yourself only briefly to lower your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (underwear) {
                toReturn += `, and your ${global.kobold.inventory.createItemWithTooltip(underwear)}`;
            }
            toReturn += `, fully exposing your arousal to the two bandits.`;
        } else if (underwear) {
            toReturn += `You interrupt yourself only briefly to lower your ${global.kobold.inventory.createItemWithTooltip(underwear)}, 
                         fully exposing your arousal to the two bandits.`;
        } else {
            toReturn += `Your arousal is fully exposed to the two bandits.`;
        }

        toReturn += `   [/p]
                        [p]
                            The sight seems to spur them on in turn.
                            The lizard on the right lustily licks his lips.
                            He takes hold of your wrist, keeping your hand in place as he begins to outright [b]fuck[/b] your hand,
                            humping your precum-slick palm, the tip of his member pressing into the side of your face with every thrust.
                        [/p]
                        [p]
                            The kobold on the left, meanwhile, places one of his hands on the back of your head,
                            guiding you up and down his shaft, spurring you to up your pace, pushing you to take more and [b]more[/b] of his shaft.
                            You more than gladly oblige him, taking over half his size into your suckling maw,
                            your tongue slithering out once more to please the remainder of his length,
                            lapping at the underside of his cock while sliding across his meat along with the rest of your muzzle.
                        [/p]
                        [p]
                            You can practically [b]feel[/b] the sexual tension rising, all three of you nearing your ends.
                            The first to blow their load is the bandit thrusting against your cheek.
                            A pleasured grunt escapes his lips as he reaches his end, a first rope of sticky, 
                            heavily fragrant spunk squirting across your face, landing horizontally across the length of your muzzle.
                            He continues to frantically hump against your grasp for the entire duration of his climax, 
                            frotting the semen-spouting tip of his member against the side of your face every time his hips roll forwards, 
                            grinding his jizz against your scales, spreading the goopy mess all over the right side of your visage.
                        [/p]
                        [p]
                            The kobold in your mouth is next to go.
                            He practically [b]pulls[/b] you onto his cock, before letting out a loud, satisfied sigh, 
                            his head leaning backwards, his mouth slightly agape.
                            Instead of squirting, his thick sperm [b]oozes[/b] from his quivering member in a steady, 
                            pulsing stream, directly onto your tongue, spilling out into the rest of your mouth through sheer volume alone.
                            You greedily gulp down all of the heady nectar dribbling from his dick, gently suckling in an attempt to coax out [b]more[/b].
                        [/p]
                        [p]
                            The scent and the taste of spunk combined are enough to send you over the edge.`;
        if (dick) {
            toReturn += ` You continue to jack yourself off, the movements of your hand slickened by your own precum, spread across your shaft.
                          As you swallow another half-mouthful of cum, you finally reach your end, blowing your load all over the leafy forest floor,
                          a few of your squirts making it as far as one of the kobolds' feet, splattering on top of his foot-scales, and leaking down in between his toes.
                          He doesn't seem to mind.`;
        } else {
            toReturn += ` You moan around the bandit's cock as you slip a second finger inside of yourself, just in time for your intense climax.
                          You buck your hips wildly against your own hand, practically riding your deeply-inserted fingers.
                          You arch your back and curl your toes as heavenly pleasure rolls over your entire body, 
                          feminine fluids squirting liberally down atop the leafy forest floor.`;
        }
        toReturn += `   [/p]
                        [p]
                            The three of you are slow to wind down from your orgasms, 
                            preferring to revel in post-orgasmic bliss rather than having to pull away and face the harsh reality of the world once again.
                            After a solid minute of mutual motionless heavy breathing, the kobold to the left finally pulls his semi-erect member out of your mouth,
                            and his friend soon follows suit in removing his dick from your face.
                        [/p]
                        [p]
                            You wipe some of the cum off your face`;
        if (underwear) {
            toReturn += `, and pull up your ${global.kobold.inventory.createItemWithTooltip(underwear)}`;
            if (pants) {
                toReturn += ` followed by your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            }
            toReturn += `, covering your groin once more.`;
        } else if (pants) {
            toReturn += `, and pull up your ${global.kobold.inventory.createItemWithTooltip(pants)}, covering your groin once more. `;
        } else {
            toReturn += `.`;
        }
        toReturn += `       The two kobolds gratefully thank you.
                            One of them remarks that you were right when you said you could show them a better time than the dragoness ever could.
                            The thought of outperforming a literal [i]sex goddess[/i] puts a smile on your face.
                        [/p]
                        [p]
                            The bandits return to their partners, which have just about wrapped things up on their end.
                            Together, ${didRock ? `they depart from the clearing, their unconscious peer hoisted and carried by two of them` : `all five of them depart from the clearing`}.
                            You give your face a more thorough cleaning, before rising to your feet once more.
                            That was [i]fun[/i].
                        [/p]
                        [p]
                            You check up on the dragoness, who seems to be in complete and utter [i]shock[/i].
                            She doesn't appear to be responding to, well, [i]anything[/i].
                            You try to get her to snap out of it, 
                            yet with the sun setting and post-orgasmic weariness starting to set in,
                            you all-too-soon decide to leave her to her mind-broken fate.
                            With the taste of cum lingering in your mouth, and its scent still clinging to your face, you begin the short trip towards your home.
                        [/p]
                        [p]
                            [c]
                                [b]You have lost your oral virginity![/b]
                            [/c]
                        [/p]`;

        return toReturn;
    }

    /**
     * Builds and returns the reluctant sex part.
     * @returns {string} The reluctant sex part.
     */
    function buildReluctant() {
        var toReturn = `[p]
                            You sigh loudly as the two males approach you,
                            the stench of their arousal reaching your nose long before they manage to bridge the couple of paces to where you're kneeling,
                            and it only intensifies with every step they take towards you.
                        [/p]
                        [p]
                            They stand on either side of you, their human-sized dicks mere inches away from your face.
                            Both of them are glaring angrily down at you.
                            You [i]did[/i] cost them a romp with a dragoness, after all.
                            A dragoness, who, from the sounds of it, their comrades are [b]thoroughly[/b] enjoying.
                        [/p]
                        [p]
                            "<span class="bold blue">Wake up, stupid whore!</span>" one of the kobolds in front of you crudely shouts,
                            obviously irked that you sunk away into your mind for a moment to try and escape this [b]nightmare[/b].
                            "<span class="bold green">Get to work!</span>" his friend interjects, 
                            giving himself a few much-needed strokes while pointing his member directly towards your face,
                            a droplet of precum dripping from the very tip of his cock down onto the leafy forest floor, 
                            its fragrance adding to the already intense aroma hanging in the air.
                        [/p]
                        [p]
                            You let out another discontent sigh as you slowly reach for the groin of the bandit to your right.
                            You tentatively grip his shaft, weakly wrapping it with your fingers.
                            Turning your head to the side, looking away from the entire ordeal and catching some fresh, non-heady air in the process, 
                            you begin to half-heartedly jerk him off at an agonizingly sluggish pace.
                        [/p]
                        [p]
                            "<span class="bold blue">Me too!</span>" the kobold to your left exclaims, as though you'd forgotten about him.
                            You blindly reach for his dick with your free hand, 
                            your palm first ending up on his bare thigh before travelling towards the base of his length.
                            Your inaccuracy seems to upset the mugger.
                            Out of the corner of your eye, you see him raising his hand, 
                            threatening to slap you, yet luckily for you, his companion dissuades him from doing so.
                        [/p]
                        [p]
                            Still, the mere gesture suffices to unsettle your mind, making you feel a little more fearful of the two males in front of you.
                            The disgruntled bandit to your left mumbles to himself as he lowers his hand once more.
                            Instead of slapping you, he reaches for your head, and forcibly rotates it towards him,
                            [b]forcing[/b] you to stare at his groin as you meekly service him,
                            and making you inhale the cock-reeking smell lingering in between them with every breath you take.
                        [/p]
                        [p]
                            With fright slowly building at the back of your mind, 
                            and the intoxicating scent of kobold arousal permeating your surroundings,
                            you subconsciously begin to try a [i]little[/i] harder to get them off,
                            if just to get this all over with sooner.
                            You up the pace of your dual jerking, and you grip their members with a bit more [i]conviction[/i].
                        [/p]
                        [p]
                            The thug to your right seems pleased with your renewed efforts.
                            The one on the left, however, is still not satisfied.
                            He adds a second hand to the back of your head, ending up with a solid grip on your skull.
                            Then, with a sudden yank, he fiercely pulls your head into his groin, burying your snout underneath his shaft.
                        [/p]
                        [p]
                            You let out a small surprised shriek, followed by a low grumbling as you realize he's simply trying to demean you.
                            With your nose pressing directly against his meat, you try to breathe in through your mouth as much as you can.
                            Still, the hefty, masculine, pheromone-laden stench of the kobold's length seeps into your nostrils, 
                            strong enough to addle your very mind.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/Forest/bandit/intervene/bargain/relieve/reluctant-threesome-${dick ? "male" : "female"}-full.png" target="_blank">
                                    <img class="nsfw" src="scenes/Forest/bandit/intervene/bargain/relieve/reluctant-threesome-${dick ? "male" : "female"}.png" title="Made by Sneaky">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            Your hands grind to a halt as your cockscent-clouded brain is straining to process what [i]exactly[/i] just happened,
                            a fact the two bandits don't appreciate in the slightest.
                            The one to your right shuffles a bit closer, and begins to rub his member against your cheek,
                            grinding the tapered tip against your warm scales to provide him with at least [i]some[/i] pleasure.
                        [/p]
                        [p]
                            The bastard to the left isn't keen on doing [i]any[/i] work himself.
                            "<span class="bold blue">Stop huffing my dick and get back to work, stupid slut.</span>"
                            he speaks, with a smug grin plastered across his face.
                            You glare angrily upwards at him, your vision partly blocked by his shaft resting atop your muzzle.
                            You try to pull your head back, yet he firmly keeps you in place.
                            You think of baring your teeth, but you'd rather not part your lips this close to his member.
                        [/p]
                        [p]
                            Grudgingly, you give in to the kobold's command.
                            You resume your pleasuring of both of them, jacking off the one to the right as he frots against the side of your face,
                            whilst stroking the dick on top of your maw, which you reluctantly press deeper against your scales, 
                            sandwiching it between your palm and your muzzle, rubbing it with the flat of your hand.
                        [/p]
                        [p]
                            Before long, your cheek ends up slathered in the bandit to the right's precum, slickening his incessant grinding.
                            The kobold right in front of you slowly begins to thrust his hips back and forward, sliding his length along the smooth topside of your muzzle,
                            the tapered tip poking the scales in between your eyes whenever he bottoms out, with your snout pressed directly against his cockslit.
                            You keep your hand firmly pressed against his shaft, as though creating a [i]hole[/i] for him to fuck,
                            hoping the sensation will prove too great for him to handle, 
                            so you can remove your head from his groin and finally take a breath that isn't completely [b]saturated[/b] with pure masculinity. 
                        [/p]
                        [p]
                            You can feel their ends approaching, each of the males showing distinct signs of their nearing climaxes.
                            The one fucking the top of your muzzle begins to move more erraticly, his otherwise steady pace turning into a haphazard series of thrusts.
                            Meanwhile, you can feel the other kobold's dick starting to twitch and quiver in your grasp.
                        [/p]
                        [p]
                            Then, nearly at the same time, they reach their ends.
                            The bandit to your right presses the tapered tip of his member against your scales as the first wave of his orgasm oozes from his pulsing, 
                            throbbing length, most of the viscous fluid sticking to your cheek, some of it running down your neck.
                            He moves his cock along the side of your face, painting much of your visage a pearly white.
                            Simultaneously, the ruffian on top of your muzzle pulls back, gives himself a few more strokes, 
                            and then squirts his load on top of you from a distance, covering most of your muzzle in his thick, sticky cum.
                        [/p]
                        [p]
                            You close your eyes to prevent any of the spunk from getting in it.
                            Just in time, or so it seems, as you can feel a warm rope of semen splattering atop your right eyelid mere moments after shutting it.
                            With your sight gone, the relative perceptiveness of your other senses grows.
                            You can [b]smell[/b] their spent cocks and their jizz, 
                            you can [b]hear[/b] their ragged, heavy breathing and the dripping of the final bits of cum on the leafy forest floor,
                            and you can [b]feel[/b] the liquid covering most of your face slowly beginning to grow cold.
                        [/p]
                        [p]
                            "<span class="bold green">Next time we meet, just a handjob won't do.</span>" one of them states.
                            His comrade seems to agree, adding his own parting insults: 
                            "<span class="bold blue">You're lucky the boss-man likes you, stupid whore.
                             If I was the leader, I'd have compared your slutty ass to the dragoness'.</span>"
                            Then, both of the males pull back from you, the warmth radiating from their lengths fading away.
                            You hear the leaves crinkle under their feet as they turn around, and walk away.
                        [/p]
                        [p]
                            You slowly wipe the cum from your face, ensuring your eyelids especially are clean of any ejaculate.
                            Opening your eyes once more, you are greeted with the sight of a pair of naked kobold asses,
                            the bandits bending down to pick up their discarded clothing, dressing themselves as they return to their gang.
                            The others seemed to have finished as well.
                            You watch them leave together, ${didRock ? "knocked out member in tow," : ""} as you continue cleaning your face of most of the spunk coating it.
                        [/p]
                        [p]
                            You take a few deep breaths of fresh air to clear the scent of dick from your lungs,
                            yet sadly the stench of kobold jizz still heavily clinging to your scales prevents you from [i]completely[/i] purging your innards 
                            of the taint of masculine arousal.
                        [/p]
                        [p]
                            You rise to your feet, and waddle over to the dragoness.
                            She seems completely [i]in shock[/i], unresponsive to whatever attempts you make at communicating.
                            Dumb [i]bitch[/i] could at least be grateful after what you did for her, 
                            instead of being nigh-catatonic just from getting raped by a couple of bandits.
                        [/p]
                        [p]
                            Your mood is thoroughly soured from the unpleasant encounter, 
                            and from the fact that the woman isn't even [i]awake enough[/i] to appreciate your [i]help[/i].
                            With the sun slowly starting to set, you decide to cut your losses and head back home.
                        [/p]`;
        return toReturn;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = ``;
        if (ctrl.bandit.intervene.getRelieveChoice() == ForestBanditRelieveChoice.EAGER) {
            toReturn += buildEager();
        } else {
            toReturn += buildReluctant();
        }
        return global.parser.parse(toReturn);
    }

    /**
     * Initializes all private variables.
     */
    function initVars() {
        didRock = (ctrl.bandit.getArrivalChoice() === ForestBanditArrivalChoiceEnum.ROCK);
        gender = ctrl.bandit.intervene.relieveGender();
        dick = ctrl.bandit.intervene.relieveDick();
        pants = ctrl.bandit.intervene.relievePants();
        underwear = ctrl.bandit.intervene.relieveUnderwear();
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.bandit.intervene.relieveHasContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }
    
    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.relieveContinue();
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditRelieveSex";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.bandit.intervene.hasMadeRelieveChoice()) {
            return "";
        }
        initVars();
        var change = ctrl.bandit.intervene.relieveChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.bandit.intervene.hasMadeRelieveChoice()) {
            return;
        }

        if (ctrl.bandit.intervene.relieveHasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\bandit\intervene\bargain\relieve\ForestBanditRelieveSex.js

// scenes\Forest\bandit\intervene\bargain\ForestBanditBargain.js

/**
 * Bargain part of the forest bandit encounter.
 * @param {ForestController} controller The forest controller.
 * @param {ForestBanditReward} reward The reward section.
 */
function ForestBanditBargain(controller, reward) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var payId = "banditBargainPayButton";
    var persuadeId = "banditBargainPersuadeButton";
    var relieveId = "banditBargainRelieveButton";
    var replaceId = "banditBargainReplaceButton";

    var toReward = reward;
    var toRelieve = new ForestBanditRelieve(ctrl);
    var toReplace = new ForestBanditReplace(ctrl);

    var content = ` [p]
                        You decide to bargain with the kobolds.
                        While still bandits, they can't be [i]completely[/i] unreasonable, right?
                        You [b]pray[/b] that they're at least willing to [i]listen[/i] to an agreeable bargain.
                    [/p]
                    [p]
                        Putting on your best faked smile, you take another step forwards, inching a teensy bit closer to the group.
                        Nobody's gouged you yet, which is a good thing. 
                        In fact, they're all still clutching the dragoness' limbs, fearing she'll make a move with you there to distract them.
                        The woman, however, doesn't seem to be willing to risk her life just because a [i]sixth[/i] kobold has appeared on the scene.
                        For all she knows, you're just as [i]wicked[/i] as they are.
                    [/p]`;

    var rock = `    [p]
                        You decide to bargain with the kobolds.
                        While still bandits, they can't be [i]completely[/i] unreasonable, right?
                        You [b]pray[/b] that they're at least willing to [i]listen[/i] to an agreeable bargain, after you knocked one of them out with a rock.
                        Then again, maybe your show of strength will end up working out to your advantage.
                        Only time can tell.
                    [/p]
                    [p]
                        Putting on your best faked smile, you take another step forwards, inching a teensy bit closer to the group.
                        Nobody's gouged you yet, which is a good thing. 
                        In fact, they're all still clutching the dragoness' limbs, fearing she'll make a move now that one of them has been knocked out.
                        With one of her arms freed, she could very well pose a serious threat to the bandits.
                        For now, however, she remains perfectly still, fearful of the blade pressed against her throat.
                    [/p]`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function buildRock() {
        var toReturn = rock;
        toReturn = global.parser.parse(toReturn);
        return toReturn;
    }

    function buildRegular() {
        var toReturn = content;
        toReturn = global.parser.parse(toReturn);
        return toReturn;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var src = ctrl.bandit.intervene.getInterveneSource();
        if (src === ForestBanditInterveneSourceEnum.ROCK) {
            return buildRock();
        }
        return  buildRegular();
    }

    /**
     * Builds and returns the pay choice button.
     */
    function buildPayChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeBargainChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getBargainChoice() == ForestBanditBargainChoice.PAY;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${payId}" ${disabled}>Pay them</button>`;

        var tooltip = `Use your bundle of hard-earned gold to persuade the gang to leave the dragoness alone.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the relieve choice button.
     */
    function buildRelieveChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeBargainChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getBargainChoice() == ForestBanditBargainChoice.RELIEVE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${relieveId}" ${disabled}>Get them off</button>`;

        var tooltip = `Offer an extra mouth and a pair of hands for them to relieve themselves with.
                       The fewer bandits the dragoness has to please, the better, right?`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the replace choice button.
     */
    function buildReplaceChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeBargainChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getBargainChoice() == ForestBanditBargainChoice.REPLACE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${replaceId}" ${disabled}>Replace her</button>`;

        var tooltip = `Offer to replace the dragoness completely.
                       Surely someone half her size is bound to be tighter than the oversized woman herself, right?`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 205px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You attempt to: </div>`;
        choice += buildPayChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildRelieveChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildReplaceChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all four buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + payId), $("#" + persuadeId), $("#" + relieveId), $("#" + replaceId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the bargain button.
     */
    function registerPayListener() {
        $("#" + payId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeBargainChoice(ForestBanditBargainChoice.PAY);
            ctrl.bandit.setRewardSource(ForestBanditRewardSourceEnum.PAY);
            loadNextPart(toReward);
        });
    }
    
    /**
     * Adds a listener to the relieve button.
     */
    function registerRelieveListener() {
        $("#" + relieveId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeBargainChoice(ForestBanditBargainChoice.RELIEVE);
            loadNextPart(toRelieve);
        });
    }
    
    /**
     * Adds a listener to the replace button.
     */
    function registerReplaceListener() {
        $("#" + replaceId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeBargainChoice(ForestBanditBargainChoice.REPLACE);
            loadNextPart(toReplace);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditBargain";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.intervene.getInterveneChoice() !== ForestBanditInterveneChoiceEnum.BARGAIN) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        // Build children here.
        part += toRelieve.build();
        part += toReplace.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.intervene.getInterveneChoice() !== ForestBanditInterveneChoiceEnum.BARGAIN) {
            return;
        }

        // toReward.registerListeners(); // Already registered in arrival.
        toRelieve.registerListeners();
        toReplace.registerListeners();

        if (ctrl.bandit.intervene.hasMadeBargainChoice()) {
            return;
        }

        registerPayListener();
        registerRelieveListener();
        registerReplaceListener();
    };
}

var ForestBanditBargainChoice = {
    PAY: 1,
    RELIEVE: 2,
    REPLACE: 3
};

// END scenes\Forest\bandit\intervene\bargain\ForestBanditBargain.js

// scenes\Forest\bandit\intervene\ForestBanditInterveneController.js

/**
 * Information about the class.
 * @param {string} sceneId The id of the scene.  
 */
function ForestBanditInterveneController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;

    var interveneSourceId = "banditInterveneSource";
    var interveneId = "banditInterveneChoice";

    var bargainId = "banditBargainChoice";

    var relieveId = "banditRelieveChoice";
    var relieveGenderId = "banditRelieveGender";
    var relieveDickId = "banditRelieveDick";
    var relievePantsId = "banditRelievePants";
    var relieveUnderwearId = "banditRelieveUnderwear";
    var relieveContinueId = "banditRelieveContinue";
    var relieveChangeId = "banditRelieveChange";

    var replaceId = "banditReplaceChoice";
    var replaceGenderId = "banditReplaceGender";
    var replaceDickId = "banditReplaceDick";
    var replacePantsId = "banditReplacePants";
    var replaceUnderwearId = "banditReplaceUnderwear";
    var replaceBraId = "banditReplaceBra";
    var replaceRemoveId = "banditReplaceRemove";
    var replaceChangeId = "banditReplaceChange";

    var acceptId = "banditAcceptChoice";
    var acceptContinueId = "banditAcceptContinue";

    var protestContinueId = "banditProtestContinue";
    var protestChangeId = "banditProtestChange";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Sets the source for the intervene.
     * @param {ForestBanditInterveneSourceEnum} source The source scene.
     */
    this.setInterveneSource = function (source) {
        var history = global.kobold.history.get(id);
        history[interveneSourceId] = source;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the intervene source from history and returns it.
     * May return undefined if the source has not yet been set.
     * @returns {ForestBanditInterveneSourceEnum} The source scene.
     */
    this.getInterveneSource = function () {
        var history = global.kobold.history.get(id);
        return history[interveneSourceId];
    };

    /**
     * Returns true if the intervene source has been set.
     */
    this.hasInterveneSource = function () {
        return global.kobold.history.has(id, interveneSourceId);
    };

    /**
     * Makes the intervene choice.
     * @param {ForestBanditInterveneChoiceEnum} choice The choice.
     */
    this.makeInterveneChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[interveneId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the intervene choice from history and returns it.
     * May return undefined if the intervene choice has not yet been set.
     */
    this.getInterveneChoice = function () {
        var history = global.kobold.history.get(id);
        return history[interveneId];
    };

    /**
     * Returns true if the intervene choice has been made.
     */
    this.hasMadeInterveneChoice = function () {
        return global.kobold.history.has(id, interveneId);
    };


    /**
     * Makes the bargain choice.
     * @param {ForestBanditBargainChoice} choice The choice.
     */
    this.makeBargainChoice = function (choice) {
        global.kobold.history.setKey(id, bargainId, choice);
    };

    /**
     * Fetches the intervene choice from history and returns it.
     * May return undefined if the intervene choice has not yet been set.
     */
    this.getBargainChoice = function () {
        return global.kobold.history.getKey(id, bargainId);
    };

    /**
     * Returns true if the intervene choice has been made.
     */
    this.hasMadeBargainChoice = function () {
        return global.kobold.history.has(id, bargainId);
    };


    /**
     * Makes the relieve choice.
     * @param {ForestBanditRelieveChoice} choice The choice.
     */
    this.makeRelieveChoice = function (choice) {
        global.kobold.history.setKey(id, relieveId, choice);
        if (choice === ForestBanditRelieveChoice.EAGER) {
            global.kobold.body.oralVirgin = false;
        }
    };

    /**
     * Fetches the relieve choice from history and returns it.
     * May return undefined if the relieve choice has not yet been set.
     */
    this.getRelieveChoice = function () {
        return global.kobold.history.getKey(id, relieveId);
    };

    /**
     * Returns true if the relieve choice has been made.
     */
    this.hasMadeRelieveChoice = function () {
        return global.kobold.history.has(id, relieveId);
    };


    /**
     * Relieve gender.
     */
    this.relieveGender = function () {
        var gender = global.kobold.body.getBaseGender();
        global.kobold.history.setIfNotSet(id, relieveGenderId, gender);
        return global.kobold.history.getKey(id, relieveGenderId);
    };

    /**
     * Relieve dick.
     */
    this.relieveDick = function () {
        var dick = global.kobold.body.hasDick();
        global.kobold.history.setIfNotSet(id, relieveDickId, dick);
        return global.kobold.history.getKey(id, relieveDickId);
    };

    /**
     * Relieve pants.
     */
    this.relievePants = function () {
        var pants = global.kobold.inventory.equipment.pants;
        global.kobold.history.setIfNotSet(id, relievePantsId, pants);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, relievePantsId));
    };

    /**
     * Relieve underwear.
     */
    this.relieveUnderwear = function () {
        var underwear = global.kobold.inventory.equipment.underwear;
        global.kobold.history.setIfNotSet(id, relieveUnderwearId, underwear);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, relieveUnderwearId));
    };

    function relieveChangeEager() {
        global.kobold.history.doIfNotSet(id, relieveChangeId, true, function () {
            global.kobold.status.exertion.increase(25, "Eager bandit threesome");
            global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.UNHELPED;
            global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.EAGER_THREESOME;
        });
        return global.ui.change.buildNotification([
            new Change("Threesome", "Sexual exhaustion", +25, true)
        ]);
    }

    function relieveChangeReluctant() {
        global.kobold.history.doIfNotSet(id, relieveChangeId, true, function () {
            global.kobold.status.exertion.increase(10, "Reluctant bandit handjobs");
            global.kobold.status.mental.decrease(10, "Reluctant bandit handjobs");
            global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.UNHELPED;
            global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.RELUCTANT_THREESOME;
        });
        return global.ui.change.buildNotification([
            new Change("Handjobs", "Sexual exhaustion", +10, true),
            new Change("Rough treatment", "Sanity", -10)
        ]);
    }

    /**
     * Relieve change.
     */
    this.relieveChange = function () {
        if (that.getRelieveChoice() == ForestBanditRelieveChoice.EAGER) {
            return relieveChangeEager();
        } else {
            return relieveChangeReluctant();
        }
    };

    /**
     * Relieve continue.
     */
    this.relieveContinue = function () {
        global.kobold.history.setKey(id, relieveContinueId, true);
    };

    /**
     * Returns true if the relieve screen has continued.
     */
    this.relieveHasContinued = function () {
        return global.kobold.history.has(id, relieveContinueId);
    };


    /**
     * Makes the replace choice.
     * @param {ForestBanditReplaceChoice} choice The choice.
     */
    this.makeReplaceChoice = function (choice) {
        global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.RAN_AWAY;
        global.kobold.history.setKey(id, replaceId, choice);
        if (choice === ForestBanditReplaceChoice.PROTEST) {
            global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.GANGRAPE;
        }
    };

    /**
     * Fetches the replace choice from history and returns it.
     * May return undefined if the replace choice has not yet been set.
     */
    this.getReplaceChoice = function () {
        return global.kobold.history.getKey(id, replaceId);
    };

    /**
     * Returns true if the replace choice has been made.
     */
    this.hasMadeReplaceChoice = function () {
        return global.kobold.history.has(id, replaceId);
    };


    function acceptEnjoy() {
        global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.EAGER_GANGBANG;
        global.kobold.history.shared.tenPawGang.footsie = TenPawGangFootsieEnum.OBSERVED;
    }

    function acceptHate() {
		if (that.replaceGender() === GenderEnum.MALE) {
            global.kobold.body.analVirgin = false;
        } else {
            global.kobold.body.vaginalVirgin = false;
        }
        global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.RELUCTANT_GANGBANG;
    }

    /**
     * Makes the accept choice.
     * @param {ForestBanditAcceptChoice} choice The choice.
     */
    this.makeAcceptChoice = function (choice) {
        global.kobold.history.setKey(id, acceptId, choice);

        global.kobold.body.oralVirgin = false;

        if (choice === ForestBanditAcceptChoice.ENJOY) {
            acceptEnjoy();
        } else {
            acceptHate();
        }
    };

    /**
     * Fetches the accept choice from history and returns it.
     * May return undefined if the accept choice has not yet been set.
     */
    this.getAcceptChoice = function () {
        return global.kobold.history.getKey(id, acceptId);
    };

    /**
     * Returns true if the accept choice has been made.
     */
    this.hasMadeAcceptChoice = function () {
        return global.kobold.history.has(id, acceptId);
    };


    /**
     * Accept continue.
     */
    this.acceptContinue = function () {
        global.kobold.history.setKey(id, acceptContinueId, true);
    };

    /**
     * Returns true if the accept screen has continued.
     */
    this.acceptHasContinued = function () {
        return global.kobold.history.has(id, acceptContinueId);
    };


    /**
     * Protest continue.
     */
    this.protestContinue = function () {
        global.kobold.history.setKey(id, protestContinueId, true);
    };

    /**
     * Returns true if the protest screen has continued.
     */
    this.protestHasContinued = function () {
        return global.kobold.history.has(id, protestContinueId);
    };


    this.protestChange = function () {
        global.kobold.history.doIfNotSet(id, protestChangeId, true, function () {
            global.kobold.status.exertion.increase(50, "Raped");
            global.kobold.status.mental.decrease(50, "Raped");
            global.kobold.status.physical.decrease(20, "Punched in the gut");
            if(global.kobold.body.dick) {
                global.kobold.body.analVirgin = false;
            } else {
                global.kobold.body.vaginalVirgin = false;
            }
        });

        return global.ui.change.buildNotification([
            new Change("Raped", "Sexual exhaustion", +50, true),
            new Change("Raped", "Sanity", -50),
            new Change("Punched in the gut", "Health", -20)
        ]);
    };


    /**
     * Replace gender.
     */
    this.replaceGender = function () {
        var gender = global.kobold.body.getBaseGender();
        global.kobold.history.setIfNotSet(id, replaceGenderId, gender);
        return global.kobold.history.getKey(id, replaceGenderId);
    };

    /**
     * Replace dick.
     */
    this.replaceDick = function () {
        var dick = global.kobold.body.hasDick();
        global.kobold.history.setIfNotSet(id, replaceDickId, dick);
        return global.kobold.history.getKey(id, replaceDickId);
    };

    /**
     * Replace pants.
     */
    this.replacePants = function () {
        var pants = global.kobold.inventory.equipment.pants;
        global.kobold.history.setIfNotSet(id, replacePantsId, pants);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, replacePantsId));
    };

    /**
     * Replace bra.
     */
    this.replaceBra = function () {
        var bra = global.kobold.inventory.equipment.bra;
        global.kobold.history.setIfNotSet(id, replaceBraId, bra);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, replaceBraId));
    };

    /**
     * Strips the kobold.
     */
    this.replaceRemoveClothing = function () {
        global.kobold.history.doIfNotSet(id, replaceRemoveId, true, function () {
            global.kobold.inventory.equipment.stripNaked();
        });
    };

    /**
     * Replace underwear.
     */
    this.replaceUnderwear = function () {
        var underwear = global.kobold.inventory.equipment.underwear;
        global.kobold.history.setIfNotSet(id, replaceUnderwearId, underwear);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, replaceUnderwearId));
    };

    this.replaceEagerChange = function () {
        global.kobold.history.doIfNotSet(id, replaceChangeId, true, function () {
            global.kobold.status.exertion.increase(50, "Bandit foursome");
            global.kobold.status.stamina.decrease(20, "Bandit foursome");
        });

        return global.ui.change.buildNotification([
            new Change("Bandit foursome", "Sexual exhaustion", +50, true),
            new Change("Bandit foursome", "Stamina", -20)
        ]);
    };

    this.replaceHateChange = function () {
        global.kobold.history.doIfNotSet(id, replaceChangeId, true, function () {
            global.kobold.status.exertion.increase(60, "Cruel bandit threesome");
            global.kobold.status.stamina.decrease(20, "Nearly choked to death");
            global.kobold.status.physical.decrease(20, "Nearly choked to death");
            global.kobold.status.mental.decrease(30, "Nearly choked to death");
        });

        return global.ui.change.buildNotification([
            new Change("Cruel threesome", "Sexual exhaustion", +60, true),
            new Change("Nearly choked to death", "Health", -20),
            new Change("Nearly choked to death", "Sanity", -30),
            new Change("Nearly choked to death", "Stamina", -20)
        ]);
    };
}

// END scenes\Forest\bandit\intervene\ForestBanditInterveneController.js

// scenes\Forest\bandit\intervene\ForestBanditIntervene.js

/**
 * Information about the class.
 * @param {ForestController} controller The forest controller.
 * @param {ForestBanditReward} reward The reward section.
 */
function ForestBanditIntervene(controller, reward) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var bargainId = "banditInterveneBargainButton";
    var fightId = "banditInterveneFightButton";
    
    var toReward = reward;
    var toBargain = new ForestBanditBargain(ctrl, toReward);

    var content = ` <div class=text>
                        You step out from behind your cover, and announce your presence by loudly clearing your throat.
                        The bandits seem startled.
                        One of them audibly gasps in surprise.
                        Their combined stupefaction rapidly fades, however, 
                        as they realize you're not some knight in shining armor, but instead a lowly kobold like themselves.
                    </div>
                    <div class="text">
                        The leader addresses you in the deepest, most authorative voice he can muster:
                        "<span class="bold brightred">You! Kobold! What are you doing here?! Go away! This is MY...</span>"
                        He quickly corrects himself.
                        "<span class="bold brightred">...OUR dragon toy! Not yours! Get your own!</span>"
                    </div>
                    <div class=text>
                        You admittedly haven't thought this plan through very well. 
                        There's five of [i]them[/i] and only one of [i]you[/i].
                        Things might go [b]very[/b] wrong for you if you don't pick your next moves carefully.
                    </div>`;

    var rock = `    <div class=text>
                        With one of the kobolds knocked out, you figure your odds of striking a deal or manhandling the situation have improved dramatically.
                        At least you [i]hope[/i] so.
                        You step out from behind your cover, and slowly, non-threateningly approach the situation.
                    </div>
                    <div class=text>
                        The ringleader's visage says it all.
                        His face goes from slightly scared, to confused, to downright [b]angry[/b].
                        At least his knife seems to give the dragoness' throat just a little more leeway.
                        The three remaining underlings breathe a collective sigh of relief, before joining their leader in glaring you down.
                    </div>
                    <div class="text">
                        "<span class="bold brightred">You! Kobold!</span>" 
                        the ringleader finally speaks. 
                        "<span class="bold brightred">What is your problem?! Why you throw rock?! Is MY...</span>"
                        He quickly corrects himself.
                        "<span class="bold brightred">...OUR dragon toy! Not yours! Go away! Get your own!</span>"
                    </div>
                    <div class=text>
                        You admittedly haven't thought this plan through very well. 
                        There's still four of [i]them[/i] and only one of [i]you[/i].
                        Not to mention the fact that you're now completely disarmed.
                        Things might go [b]very[/b] wrong for you if you don't pick your next moves carefully.
                    </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function buildRock() {
        var toReturn = rock;
        toReturn = global.parser.parse(toReturn);
        return toReturn;
    }

    function buildRegular() {
        var toReturn = content;
        toReturn = global.parser.parse(toReturn);
        return toReturn;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var src = ctrl.bandit.intervene.getInterveneSource();
        if (src === ForestBanditInterveneSourceEnum.ROCK) {
            return buildRock();
        }
        return  buildRegular();
    }

    /**
     * Builds and returns the bargain choice button.
     */
    function buildBargainChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeInterveneChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getInterveneChoice() == ForestBanditInterveneChoiceEnum.BARGAIN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${bargainId}" ${disabled}>Bargain</button>`;

        var tooltip = `Perhaps you can reason with the bandits in one way or another.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the intervene choice button.
     */
    function buildFightChoice() {
        var disabled = ctrl.bandit.intervene.hasMadeInterveneChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.intervene.getInterveneChoice() == ForestBanditInterveneChoiceEnum.FIGHT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fightId}" ${disabled}>Fight</button>`;

        var tooltip = `Attempt to take on the entire gang in combat. <span class="todo">UNFINISHED</span>`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You attempt to: </div>`;
        choice += buildBargainChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFightChoice();
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables both buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + bargainId), $("#" + fightId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }
    
    /**
     * Adds a listener to the bargain button.
     */
    function registerBargainListener() {
        $("#" + bargainId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeInterveneChoice(ForestBanditInterveneChoiceEnum.BARGAIN);
            loadNextPart(toBargain);
        });
    }

    /**
     * Adds a listener to the fight button.
     */
    function registerFightListener() {
        $("#" + fightId).click(function () {
            /*$(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.makeInterveneChoice(ForestBanditInterveneChoiceEnum.FIGHT);
            loadNextPart(toFight);*/
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditIntervene";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.bandit.intervene.hasInterveneSource()) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toBargain.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.bandit.intervene.hasInterveneSource()) {
            return;
        }

        toBargain.registerListeners();

        if (ctrl.bandit.intervene.hasMadeInterveneChoice()) {
            return;
        }

        registerBargainListener();
        registerFightListener();
    };
}

var ForestBanditInterveneChoiceEnum = {
    BARGAIN: 1,
    FIGHT: 2
};

/**
 * Source for the intervention part of the bandit part.
 */
var ForestBanditInterveneSourceEnum = {
    APPROACH: 1,
    ROCK: 2
};

// END scenes\Forest\bandit\intervene\ForestBanditIntervene.js

// scenes\Forest\bandit\rock\ForestBanditRock.js

/**
 * Information about the class.
 * @param {ForestController} controller The forest controller.
 * @param {ForestBanditIntervene} intervene The intervene section.
 * @param {ForestBanditReward} reward The reward section.
 */
function ForestBanditRock(controller, intervene, reward) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toIntervene = intervene;
    var toReward = reward;

    var scaryId = "banditRockScaryButton";
    var interveneId = "banditRockInterveneButton";

    var content = ` <div class=text>
                        Luckily, you picked up a ${global.kobold.inventory.createItemWithTooltip(new Rock())} earlier,
                        The perfect weapon to draw first blood without even having to leave your concealed location.
                        You step out from behind the tree, your body concealed still by the bush right next to it.
                        You clutch the ${global.kobold.inventory.createItemWithTooltip(new Rock())} in your right hand, 
                        and set your sights on one of the kobolds pinning the dragoness' arms.
                    </div>
                    <div class=text>
                        Uncertainty and panic slowly settle in your mind as you take aim.
                        You've never really <span class="italic">thrown</span>, well, <span class="italic">anything</span> with the intent of hurting someone.
                        And if you miss, you could just as easily hit the woman.
                    </div>
                    <div class=text>
                        As your arm is already swinging forwards to propel the ${global.kobold.inventory.createItemWithTooltip(new Rock())} towards the gangmembers,
                        you close your eyes out of sheer self-doubt, your mind flipping sides from <span class="italic">do it</span> to <span class="italic">don't</span>.
                        But at that point, it was already too late.
                    </div>
                    <div class=text>
                        At the very apex of your pitch, your scaled fingers untangle themselves from the cold, hard surface of the heavy stone,
                        which is sent flying through the air.
                        With your eyes closed, you can only hope you didn't <span class="italic">screw up</span>.
                    </div>
                    <div class="text">
                        A loud, somewhat masculine scream rings throughout the forest, 
                        followed by the soft thud of a lightweight body falling atop the leafy forest floor.
                        Instinctively, you press yourself against the tree-trunk once more, 
                        re-entering cover before opening your fearfully shut eyes and peeking out from behind it.
                    </div>
                    <div class="text">
                        You managed to score a direct hit on your target, the kobold knocked out cold from the harsh impact to his cranium.
                        He lies motionlessly on the forest floor, one of the dragoness' arms freed once more, 
                        kept only in check by the threatening blade pressed to her throat.
                    </div>
                    <div class="text">
                        The knocked out bandit's companions gaze around the clearing fearfully.
                        A few panicked lines of dialogue come to pass before their leader brings his men in line,
                        the knife-wielding kobold clearly more than a little startled himself.
                        He gazes around warily, seemingly unable to pinpoint your position.
                        He then speaks out loudly, on the verge of shouting, in the deepest voice he can seemingly muster:
                    </div>
                    <div class="text center">
                        <span class="bold brightred">
                            Who's there?! 
                            Come out and fight!
                            Quit being a coward and face us! 
                            Come out or I'll kill the woman!!
                        </span>
                    </div>
                    <div class="text">
                        The other three, meanwhile, are practically cowering in fear.
                        They whisper amongst themselves about stories of ghosts and spectres,
                        about how they've <span class="bold">angered</span> the forest.
                    </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        return content;
    }

    /**
     * Builds and returns the scary noise choice button.
     */
    function buildScaryNoiseChoice() {
        var disabled = ctrl.bandit.hasMadeRockChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.getRockChoice() == ForestBanditRockChoiceEnum.SCARYNOISE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${scaryId}" ${disabled}>Scare them</button>`;

        var tooltip = `Making a loud noise might scare the kobolds enough to send them running.
                       Then again, you'd be giving away your position at the same time.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 215px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the intervene choice button.
     */
    function buildInterveneChoice() {
        var disabled = ctrl.bandit.hasMadeRockChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.getRockChoice() == ForestBanditRockChoiceEnum.INTERVENE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${interveneId}" ${disabled}>Intervene</button>`;

        var tooltip = `With one of the kobolds knocked out, you might stand a better chance at arguing with them, or perhaps even fighting them.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 225px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildScaryNoiseChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildInterveneChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables both buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + scaryId), $("#" + interveneId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the rock button.
     */
    function registerScaryListener() {
        $("#" + scaryId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.setRewardSource(ForestBanditRewardSourceEnum.SCARYNOISE);
            ctrl.bandit.makeRockChoice(ForestBanditRockChoiceEnum.SCARYNOISE);
            loadNextPart(toReward);
        });
    }

    /**
     * Adds a listener to the rock button.
     */
    function registerInterveneListener() {
        $("#" + interveneId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.setInterveneSource(ForestBanditInterveneSourceEnum.ROCK);
            ctrl.bandit.makeRockChoice(ForestBanditRockChoiceEnum.INTERVENE);
            loadNextPart(toIntervene);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditRock";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.getArrivalChoice() !== ForestBanditArrivalChoiceEnum.ROCK) {
            return "";
        }

        var change = global.ui.change.buildNotification([
            new Change("Thrown rock", new Rock().name(true), -1)
        ]);
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `</div>`;
        if (ctrl.bandit.getRockChoice() === ForestBanditRockChoiceEnum.INTERVENE) {
            part += toIntervene.build();
        }
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.getArrivalChoice() !== ForestBanditArrivalChoiceEnum.ROCK) {
            return;
        }

        // toIntervene.registerListeners(); // Already registered in arrival.
        // toReward.registerListeners(); // Already registered in arrival.

        if (ctrl.bandit.hasMadeRockChoice()) {
            return;
        }

        registerScaryListener();
        registerInterveneListener();
    };
}

/**
 * Choices after the rock has been throwin in the bandit encounter.
 */
var ForestBanditRockChoiceEnum = {
    SCARYNOISE: 1,
    INTERVENE: 2
};

// END scenes\Forest\bandit\rock\ForestBanditRock.js

// scenes\Forest\bandit\watch\help\ForestBanditHelp.js

/**
 * Help section of the bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditHelp(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var continueId = "banditHelpContinueButton";
    
    var content = ` <div class="text">
                        After waiting for a few more moments to ensure the gang of kobolds has left completely,
                        you emerge from your hiding spot, and slowly approach the dragoness.
                        The thick smell of 
                        ${(ctrl.bandit.watch.getHelpSource() === ForestBanditHelpSourceEnum.INDULGE) ? "four" : "five"} combined loads of hefty kobold spunk hits your sensitive nose like a truck.
                        You can only imagine how she must be feeling.
                    </div>
                    <div class="text">
                        You kneel down next to the woman, and gently put your hand on the top of her head.
                        You calmly pat her head in an effort to comfort her, making sure to avoid the large amount of cum sticking to her scales.
                        She doesn't acknowledge you, continuing to stare blankly up at the sky, her jaw slightly agape.
                        You utter some comforting words, yet much like your patting, you doubt they have much effect.
                    </div>
                    <div class="text">
                        You try to keep a smile on your face to appear more amicable than the other kobolds,
                        but you can't help but frown at the the mentally broken state the dragoness is in.
                        You decide to go about things a different way.
                        You pick up the woman's discarded, used underwear.
                        It's disgusting and sticky, but you bear with it.
                        Bringing the rag to her still-spread legs, you press it against her cum-leaking slit, 
                        and carefully clean the slightly gaping orifice, the silk fabric soaking up much of the jizz spilling from her folds.
                    </div>
                    <div class="text">
                        With her groin at least <span class="italic">somewhat</span> devoid of spunk, you set yourself to the task of closing her legs.
                        You strain yourself to lift one of the large, blue appendages, and bring it just a little bit closer to the other.
                        Then, you repeat the process for the second leg, improving her posture from spread-eagle whore to naked, catatonic damsel.
                    </div>
                    <div class="text">
                        She can't stay here.
                        It's dangerous out in the woods.
                        There's other bandits, woodland creatures and even some <span class="italic">slavers</span> that might find her.
                        You call out for her, try to get her to snap back to reality, yet nothing seems to work.
                        Slightly frustrated, you take hold of her arm, and pull it upwards, in an attempt to get her to sit upright.
                        Still no response.
                    </div>
                    <div class="text">
                        You sigh loudly.
                        There isn't much you can do for her without her cooperation.
                        Still, you can't just leave her out here.
                        Then, an idea crosses your mind.
                        The leaves.
                        You can use the leaves!
                        Picking up a few handfuls of them, you strew them about the unresponsive dragoness' body, 
                        camouflaging her blue scales with far more natural brown and yellow colors.
                    </div>
                    <div class="text">
                        You rise to your feet.
                        While not really satisfied with your work, you did the best you could.
                        You restored some of the dragoness' dignity, and you gave her some camouflage to prevent her from getting found by people who are <span class="italic">less kind</span> than you are.
                    </div>
                    <div class="text">
                        You try one final time to <span class="italic">get through</span> to her, without success.
                        While you'd love to stay and watch over the woman, you'd just end up attracting more attention, especially with the pouch of money strapped around your neck.
                        Besides, a single lone kobold like yourself wouldn't be able to offer her much protection anyway.
                        Despite your childhood dreams, you <span class="bold">aren't</span> a knight in shining armor.
                    </div>
                    <div class="text">
                        With a sigh, you wave the dragoness goodbye, before walking out of the clearing.
                        You head home, hoping she'll be alright. 
                        In the end, you did the right thing. 
                        But you can't help but feel a little guilty for not even <span class="bold">trying</span> to stop the gang.
                    </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.bandit.watch.helpHasContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }
    
    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.helpContinue();
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditHelp";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.bandit.watch.hasHelpSource()) {
            return "";
        }
        ctrl.bandit.watch.helpShared();
        var status = ctrl.bandit.watch.helpStatus();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += content;
        part += status;
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.bandit.watch.hasHelpSource()) {
            return;
        }

        // Register child listeners here

        if (ctrl.bandit.watch.helpHasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

/**
 * Source for the help part of the bandit encounter.
 */
var ForestBanditHelpSourceEnum = {
    MASTURBATE: 1,
    KEEPWATCHING: 2,
    INDULGE: 3
};

// END scenes\Forest\bandit\watch\help\ForestBanditHelp.js

// scenes\Forest\bandit\watch\keepwatching\sloppyseconds\ForestBanditSloppySeconds.js

/**
 * Sloppy seconds section of the bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditSloppySeconds(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "banditSloppySecondsContinueButton";

    var underwear;
    var underwearPopup;
    var pants;
    var pantsPopup;
    var bra;
    var braPopup;
    var shirt;
    var shirtPopup;
    var boobs;
    var cock;
    var cunt;
    var cuntVirgin;
    var cockVirgin;
    var gender;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds the male part of this section.
     */
    function buildMale() {
        var onWhat;
        if (pants) {
            onWhat = pantsPopup;
        } else if (underwear) {
            onWhat = underwearPopup;
        } else {
            onWhat = "bare scales";
        }

        var toReturn = `<div class="text">
                            You step in between the dragoness' still spread legs, and drop to your knees. 
                            Her well-fucked slit is mere inches away from your groin, 
                            so close you can feel the warmth radiating from it on your ${onWhat}.
                            Cum is still oozing from the woman's cunt, a small puddle of the stuff forming atop the leafy forest floor.
                            You're too aroused to let such a tiny <span class="italic">detail</span> dissuade you.
                        </div>
                        <div class="text">
                            This is, after all, a dragoness you're dealing with.
                            Besides being one of the more <span class="italic">regal</span> species around, dragons are also highly coveted sexual partners.
                            Whereas vixens are natural born whores, dragonesses are more akin to courtesans, if not queens or princesses themselves.
                            You've always <span class="italic">dreamed</span> of fucking one, 
                            and you're not going to let this opportunity go to waste just because someone else beat you to it.
                        </div>
                        <div class="text">
                            You slip a single finger into the lady's slightly gaping loins, which slides in with a wet squelching sound.
                            While fucked only just now, her innards are still <span class="italic">exceedingly</span> tight.
                            Your finger almost feels like it's being sucked in deeper into the messy hole.
                            No wonder her kind is lusted after.
                        </div>
                        <div class="text">
                            You remove your digit from her cunt, and wipe off any lingering semen on her belly.
                            With her complete lack of reaction, you assume you're good to go.`;

        if (pants) {
            toReturn += `   You lower your ${pantsPopup} to the ground, making sure they're safe from the growing pool of cum. `;
            if (underwear) {
                toReturn += `Your ${underwearPopup} soon follows suit.`;
            }
            toReturn += `   Your semi-erect member comes into view.`;
        } else if (underwear) {
            toReturn += `   You lower your ${underwearPopup} to the ground, making sure it's safe from the growing pool of cum.`;
            toReturn += `   Your semi-erect member comes into view.`;
        }

        toReturn += `       You quickly give your cock a few much-needed strokes with one of your hands.
                            Your length near-instantly hardens to full length, 
                            the mere <span class="italic">thought</span> of fucking a <span class="bold">dragoness</span> driving you wild with lust.
                        </div>
                        <div class="text">
                            With her loins already having been <span class="italic">warmed up</span> by your <span class="italic">predecessor</span>, 
                            there's no need for much foreplay.
                            You guide the tapered tip of your member to the woman's folds, parting them slightly, without pressing in just yet.
                            You can feel her <span class="bold">heat</span> on your shaft.
                            You take a few seconds to <span class="bold">relish</span> in it.
                            This is going to be a moment you remember for the rest of your life.
                        </div>
                        <div class="text">
                            Very slowly, you press your hips forwards, your cock sinking smoothly into the dragoness' sloppy cunt, 
                            which tightens around your shaft like a <span class="bold">vice</span>, enticing you to go deeper and deeper.
                            Using the plentiful spunk as lube, you manage to bottom out in a single thrust, your entire length engulfed in the dragoness' overly-warm innards,
                            which squeeze your cock tightly, almost as though <span class="bold">milking</span> it for your seed.
                        </div>
                        <div class="text">
                            <a href="scenes/Forest/bandit/watch/keepwatching/sloppyseconds/seconds-male-full.png" target="_blank">
                                <img class="nsfw" src="scenes/Forest/bandit/watch/keepwatching/sloppyseconds/seconds-male.png" title="Made by Sneaky">
                            </a>
                        </div>
                        <div class="text">
                            Overwhelmed with satisfaction and lust, you almost blow your load right then and there.
                            <span class="bold">Almost</span>.
                            You aren't about to waste a once-in-a-lifetime opportunity like this on just a single insertion.
                            Collecting yourself, you pull back from the woman's hole, keeping only the very tip of your cock nestled snugly inside of her,
                            your shaft coated in a mixture of semen and excessive amounts of her own fluids.
                        </div>
                        <div class="text">
                            Taking a deep breath, you begin to thrust in and out of her tight snatch.
                            Slowly, at first, but with the sheer <span class="bold">need</span> to plant your seed 
                            <span class="bold">deep</span> inside of her growing with every passing second,
                            so too does your pace rapidly increase.
                            It isn't long until you're wildly <span class="bold">bucking</span> away at the dragoness' cunt, your hands digging under her ass,
                            grabbing on to her plump behind to allow you to <span class="bold">fuck</span> her just a little bit harder.
                            Your hips slap against her pelvis with every rowdy motion forwards, 
                            the sound of which reverberates across the clearing, and echoes throughout the forest.
                        </div>
                        <div class="text">
                            You pant loudly, physically exerting yourself to 
                            <span class="bold">pound</span> this dragoness <span class="bold">whore</span> as hard as you can.
                            She may be twice your size, but you doubt she's ever been fucked by someone with the 
                            sheer driven determination to experience the best fuck of
                            ${(gender === GenderEnum.MALE) ? "his" : "her"} life.
                        </div>
                        <div class="text">
                            Still, despite your best efforts, the woman's overly tight cunt quickly brings you to the edge of orgasm.
                            You can feel your cock quiver, <span class="bold">begging</span> for release.
                            While you'd have loved to <span class="bold">fuck</span> her brains out for hours on end, 
                            you simply lack the sexual endurance for such a gargantuan task.
                            Tilting your head backwards, arching your spine and ramming yourself into the lady's slit as deep as you can,
                            you finally allow yourself to reach your end.
                        </div>
                        <div class="text">
                            You keep yourself hilted inside of the dragoness' snatch, 
                            as wave after wave of viscous spunk squirts from the tip of your member,
                            eventually settling down into a small, continuous <span class="italic">dribble</span>.
                            Your cum mixes with the jizz left over from the previous kobold to have fucked her, 
                            and as you slowly retract your member from her cunt, the combined stream begins to leak out 
                            of the lady's twice-abused folds.
                        </div>
                        <div class="text">
                            You let out a deep, content sigh.
                            That was <span class="bold">incredible</span>.
                            You wipe your cock off onto the dragoness' thigh.
                            You rise to your feet, still panting, and make a mental note to accrue a harem of draconic sluts, 
                            should you ever come into possession of a <span class="bold">huge</span> amount of coins.
                        </div>
                        <div class="text">`;

        if (underwear) {
            toReturn += `   You pull up your ${underwearPopup}`;
            if (pants) {
                toReturn += ` and your ${pantsPopup}`;
            }
            toReturn += `.`;
        } else if (pants) {
            toReturn += `   You pull up your ${pantsPopup}.`;
        }


        toReturn += `       While that was fun, you don't have time to linger. 
                            Night is slowly starting to fall.
                            You head out of the clearing, and back into the woods.
                            Destination: home.
                        </div>
                        <div class="text center bold">
                            You are no longer a virgin!
                        </div>`;
        return toReturn;
    }

    /**
     * Builds the female part of this section.
     */
    function buildFemale() {
        var toReturn = `<div class="text">
                            As you approach the dragoness, the heady smell of five hefty loads of kobold spunk assails your sensitive nostrils.
                            It excites you to no end, your legs growing weaker and your groin growing damper with every step and every breath you take.
                            Finally, you drop to your knees, right next to the woman's frame.
                            You intently huff the heavily scented air, tainted to the very core with a multitude of masculine odours.
                            You can't get enough of it.
                        </div>
                        <div class="text">`;

        if (pants && underwear) {
            toReturn += `One of your hands slides underneath the waistband of your ${pantsPopup}, down into your ${underwearPopup}.`;
        } else if (pants) {
            toReturn += `One of your hands finds its way into your ${pantsPopup}.`;
        } else if (underwear) {
            toReturn += `You slip a hand into your ${underwearPopup}.`;
        } else {
            toReturn += `You move one of your hands to your bare groin.`;
        }

        toReturn += `       You begin to slide two fingers along the length of your slit, already <span class="bold">dripping</span> with arousal.
                            Your feminine fragrance is added to the mixture of sexual smells, tinting it just a little, to sheer <span class="bold">perfection</span>.
                        </div>
                        <div class="text">`;

        if (pants && underwear) {
            toReturn += `You take off your ${pantsPopup}, along with your ${underwearPopup}, leaving you completely naked from the waist-down.`;
        } else if (pants) {
            toReturn += `You take off your ${pantsPopup}, leaving you completely naked from the waist-down.`;
        } else if (underwear) {
            toReturn += `You take off your ${underwearPopup}, leaving you completely naked from the waist-down.`;
        }

        toReturn += `       Your wet groin glistens beautifully in the final rays of the setting sun.
                            You could masturbate just to the ambient scent lingering in the clearing.
                            Yet you want more.
                            No, you <span class="bold">need</span> more.
                        </div>
                        <div class="text">
                            You crawl towards the dragoness' feet.
                            You press your snout against her sole, the same place one of the bandits' dicks was not ten minutes ago.
                            You sniff deeply.
                            Besides the distinctive scent of female dragon feet, you can still clearly smell the kobold's cock.
                        </div>
                        <div class="text">
                            Your two fingers turn into a full palm rubbing against your cunt, 
                            spreading your own excitement along your groin, slathering the viscous goo liberally across your nether scales.
                            The remnants of the spunk coating the woman's foot grab your attention.
                            You hover your head right in front of a single rope of jizz still sticking to her scales.
                            In the heat of the moment, you extend your tongue, and you lap the semen from her foot, 
                            along with a small bit of foot-grime, leaving behind a trail of your saliva.
                        </div>
                        <div class="text">
                            It tastes every bit as you thought it would.
                            Salty, tangy, <span class="bold">heavenly</span>.
                            In your aroused state, you barely even <span class="italic">taste</span> the dirt on your tongue.
                            Your mind focuses on one thing: getting <span class="bold">more</span>.
                            You crawl in between the dragoness' spread legs, on your hands and knees,
                            and there you are greeted with the motherload.
                        </div>
                        <div class="text">
                            A veritable <span class="bold">torrent</span> of kobold semen, mixed with the dragoness' own fluids,
                            is oozing down from her slightly gaping snatch, being <span class="bold">wasted</span> on the forest floor.
                            You lower your face in between her thighs, and are greeted with an aroma of pure <span class="bold">sex</span>.
                        </div>
                        <div class="text">
                            Dipping one of your digits into your needy slit, 
                            you begin to finger yourself as you press your muzzle directly against the woman's freshly-fucked cunt.
                            You greedily lap at her nether lips, collecting as much of the sickly-sweet mixture as you possibly can, 
                            swallowing a good mouthful of kobold spunk and dragoness arousal.
                            Your snout is covered in the stuff, your nostrils clogged to the brim with combined male and female pheromones,
                            almost driving you <span class="bold">crazy</span> with lust.
                        </div>
                        <div class="text">
                            <a href="scenes/Forest/bandit/watch/keepwatching/sloppyseconds/seconds-female-full.png" target="_blank">
                                <img class="nsfw" src="scenes/Forest/bandit/watch/keepwatching/sloppyseconds/seconds-female.png" title="Made by Sneaky">
                            </a>
                        </div>
                        <div class="text">
                            You press your tongue past her folds, into the lady's cunt, 
                            lapping at her innards to collect even <span class="bold">more</span> of the mind-numbing concoction.
                            Finally, with a good bit of the stuff in your mouth, you pull your head back, without swallowing just yet.
                        </div>
                        <div class="text">
                            The thick strands of mixed saliva, feminine arousal and jizz connecting your muzzle to her nethers are broken, 
                            as you crawl on top of and over the woman's body, all the way up to her chest, ending up face-to-face with the fair blue-scaled lady.
                            The blank, absent-minded stare plastered across her visage makes her look like an air-headed bimbo.
                            A broken dragoness fucktoy.
                            Right now, in the middle of getting yourself off, that doesn't seem like such a bad state to be in.
                        </div>
                        <div class="text">
                            With one hand, you resume your interrupted masturbation, adding a second digit to your intense fingering.
                            With the other, you take hold of the dragoness' muzzle.
                            You lower your head, and press your lips against hers.
                            With your fingers, you pry open her jaws, just so you can slip your tongue into her mouth,
                            the slop collected in your muzzle seeping downwards, giving her a taste of the fruits of her labour.
                        </div>
                        <div class="text">
                            The sheer <span class="build">deviancy</span> of the act alone almost brings you close to your climax.
                            You turn the snowballing into a full-blown ${(gender === GenderEnum.FEMALE) ? "lesbian" : ""} kiss,
                            your tongue exploring every inch of the insides of her mouth.
                            You remove your hand from her muzzle, and instead use it to grope one of her tits.
                            You can feel your orgasm rising within you, bubbling up from your groin.
                        </div>
                        <div class="text">
                            You shove your fingers inside of yourself to the very knuckle.
                            Your mind goes blank.
                            Your eyes roll to the back of your head.
                            You let out a loud moan, without breaking the sloppy kiss you find yourself engrossed in.
                            Your entire body shudders and trembles, as you experience the most intense climax you've ever felt.
                        </div>
                        <div class="text">
                            Feminine fluids squirt from your slit, splattering along the entire length of the dragoness' lower body.
                            Your snatch contracts and convulses, tightening sporadically around your fingers.
                            You collapse down on top of the blue-scaled woman, your fall dampened by her hefty rack.
                            You break free from the lengthy kiss, which leaves the both of you huffing and panting, 
                            sadly enough the most reaction you've managed to coax from her during this entire ordeal.
                        </div>
                        <div class="text">
                            For a few moments, you lie on your catatonic lover, revelling in post-orgasmic bliss.
                            Still breathing heavily, you lazily roll off of her, coming to rest next to the dragoness, atop the bed of leaves she, too, is lying on.
                            You could get used to this sort of stuff, perhaps with a somewhat more responsive partner.
                            Maybe even in a nicer place than the midst of the forest.
                            Line in an inn, or at least in the comfort of a bed.
                            ${(gender === GenderEnum.FEMALE) ? "This doesn't make you gay, does it?" : ""}
                        </div>
                        <div class="text">
                            After you catch your breath, you wipe your muzzle clean of any remaining goop.
                            This was fun, but with the sun rapidly disappearing behind the treetops, it's time for you to move on.
                            You use a couple of leaves to clean your groin of sticky arousal${(pants || underwear) ? ", before donning your garments once more" : ""}. 
                        </div>
                        <div class="text">
                            As you exit the clearing, you cast one final glance at the dragoness.
                            She hasn't so much as closed her mouth, her jaw still as agape as you left.
                            You're sure she'll be fine.
                            You set off into the woods, and begin the journey home.
                        </div>`;

        return toReturn;
    }

    /**
     * Builds and returns the content.
     */
    function buildContent() {
        var toReturn = `<div class="text">
                            Watching the other kobolds have their way with the dragoness has aroused you a great deal.
                            You wait a few more moments for them to disappear from sight.
                            Then, you emerge from your hiding spot, and step towards the woman.
                            She seems completely <span class="italic">out of it</span>, in shock, catatonic, 
                            in no shape to defend herself, or even to object to anything anyone might do to her.
                            The perfect condition for you to sate your lust on her body.
                        </div>`;

        if (cock) {
            toReturn += buildMale();
        } else {
            toReturn += buildFemale();
        }

        return toReturn;
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.bandit.watch.sloppySecondsHasContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }

    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.sloppySecondsContinue();
            loadNextPart();
        });
    }

    /**
     * Initializes all required variables.
     */
    function initVars() {
        pants = ctrl.bandit.watch.sloppySecondsPants();
        pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        underwear = ctrl.bandit.watch.sloppySecondsUnderwear();
        underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        bra = ctrl.bandit.watch.sloppySecondsBra();
        braPopup = global.kobold.inventory.createItemWithTooltip(bra);
        shirt = ctrl.bandit.watch.sloppySecondsShirt();
        shirtPopup = global.kobold.inventory.createItemWithTooltip(shirt);
        boobs = ctrl.bandit.watch.sloppySecondsBoobs();
        cock = ctrl.bandit.watch.sloppySecondsCock();
        cockVirgin = ctrl.bandit.watch.sloppySecondsCockVirgin();
        cunt = ctrl.bandit.watch.sloppySecondsCunt();
        cuntVirgin = ctrl.bandit.watch.sloppySecondsCuntVirgin();
        gender = ctrl.bandit.watch.sloppySecondsBaseGender();
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditSloppySeconds";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.watch.getKeepWatchingChoice() !== ForestBanditKeepWatchingChoiceEnum.SLOPPYSECONDS) {
            return "";
        }
        initVars();
        ctrl.bandit.watch.sloppySecondsShared();
        var status = ctrl.bandit.watch.sloppySecondsStatus();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += status;
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.watch.getKeepWatchingChoice() !== ForestBanditKeepWatchingChoiceEnum.SLOPPYSECONDS) {
            return;
        }

        // Register child listeners here

        if (ctrl.bandit.watch.sloppySecondsHasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\bandit\watch\keepwatching\sloppyseconds\ForestBanditSloppySeconds.js

// scenes\Forest\bandit\watch\keepwatching\ForestBanditKeepWatching.js

/**
 * Keep watching section of the bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditKeepWatching(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toHelp = new ForestBanditHelp(controller);
    var toScavenge = new ForestBanditScavenge(controller);
    var toSloppySeconds = new ForestBanditSloppySeconds(controller);

    var helpId = "banditHelpButton";
    var scavengeId = "banditScavengeButton";
    var sloppyId = "banditSloppyButton";

    var content = ` <div class="text">
                        The kobolds are too numerous.
                        If you tried to intervene, you'd just end up in the same position the dragoness is in.
                        And you don't want to risk your neck for a chance at some second-hand stolen goods.
                        While right now there's nothing you <span class="italic">can</span> do besides continue to watch,
                        there may yet be an opportunity for you to somehow profit off of this situation.
                    </div>
                    <div class="text">
                        The lady looks to be in some kind of shock, 
                        paralyzed by a mixture of fear, humiliation and self-hatred.
                        The dagger-wielding kobold has finished completely <span class="bold">destroying</span> her dress,
                        the entirety of her magnificent blue-scaled body coming into full view.
                        Lacking any signifant muscle tone,
                        her entire frame seems rather <span class="italic">plump</span>.
                        While not by any means fat, she's far too well fed to be a commoner.
                        Definitely a noble or a merchant.
                    </div>
                    <div class="text">
                        While the woman's dress and bra have been reduced to useless scraps,
                        the small male between her legs displays an uncharacteristic level of sophistication in properly taking off her panties,
                        instead of tearing them off like the rest of her clothing.
                        He slides the delicate fabric down the smooth scales of her legs, 
                        and gently tugs the underwear out from underneath her feet.
                        The reason why is quickly revealed,
                        as he passes the panties to his foot-obssessed friend, who presses them to his snout,
                        huffs them deeply, and then wraps the garment around his growing length,
                        jerking himself off with the silken cloth.
                        What a weirdo.
                    </div>
                    <div class="text">
                        With the dragoness' underwear removed,
                        the only thing left concealing her most intimate parts are her thick thighs, 
                        tightly pressed together as a last line of defence.
                        But not for long.
                        Two kobolds take hold of her legs.
                        She struggles, she squirms, and a few muffled, panicked pleas escape her lips,
                        yet despite her best efforts, her legs are spread open wide, draconic slit fully exposed for all to see.
                    </div>
                    <div class="text">
                        The woman renews her struggles,
                        trying her hardest to thrash about,
                        to no avail.
                        A third kobold crawls in between her legs.
                        He reaches down to give her cunt just a few quick pity-rubs with the flat of his hand,
                        both to <span class="italic">test the waters</span> and to taunt the lady.
                        Then, without much ado, he grabs her hips, and with a single, smooth thrust, he lodges himself inside of the dragoness' tight, warm snatch.
                    </div>
                    <div class="text">
                        Her eyes open wide,
                        and her resistance instantly dies down the moment the kobold spears apart her folds.
                        It was as though she suddenly realized that there <span class="italic">was</span> no escape.
                        A fresh wave of tears roll down the dragoness' face.
                        Her body goes limp, completely surrendered to the whims of her assailants.
                        Her eyes close once more, and she simply lies there as the bandit begins to fuck her properly,
                        her thighs jiggling with his every thrust, the slapping of scales against scales reverberating across the clearing.
                    </div>
                    <div class="text">
                        The other kobolds begin to look for their own ways to get off.
                        The one with the dagger moves towards the lady's side.
                        Taking hold of her hand, he places it atop his erect member.
                        While at first, she silently refused to cooperate, a few angry threats got her to half-heartedly wrap her fingers around his length.
                        Only when the mugger pressed his sharp, pointy blade against her bare scales, threatening to stab her, did she begin to jerk him off properly.
                    </div>
                    <div class="text">
                        The kobold that got naked first, tired of stroking himself off while fiddling with just one measly breast, climbs atop of the woman's body.
                        After idly caressing her boobs for a little while longer, he presses his cock in between them.
                        Then, the highwayman squishes the dragoness' tits together, enveloping his dick in warm, supple scales,
                        forming a hole to fuck which he instantly begins to thrust against.
                    </div>
                    <div class="text">
                        The ringleader harshly slaps the woman across the face, before using both his hands to pry her jaws apart, just wide open enough for him to ram his member into her mouth. 
                        Luckily for her, he wasn't <span class="italic">hung</span> enough to reach all the way to her throat, his dick barely large enough to fill her entire mouth.
                        After a brief reminder of what would happen should she choose to bite down, the leader lets go of her muzzle, the dragoness reluctantly wrapping her lips around his length.
                        Taking hold of her horns once more, to keep her head in place, the cruel kobold begins to mercilessly fuck the woman's face, 
                        the bottom of his shaft rubbing against her near-serpentine tongue with every thrust, ensuring the flavour of his cock would linger on her tastebuds long after he was through with her.
                    </div>
                    <div class="text">
                        The one with the pair of panties still hanging off his dick, finally, returns to the lady's feet,
                        yet instead of rubbing his muzzle against them like he did before, he begins to hump against her soles.
                        It isn't long before the underwear wrapped around the tip of his member becomes so 
                        <span class="italic">saturated</span> with precum that some of the slick fluid begins to seep out,
                        coating the scales at the bottom of the larger lizard's feet, slickening them, 
                        allowing the foot-obsessed bandit to up his pace.
                    </div>
                    <div class="text">
                        As each of the males finds his own spot, their combined cacophony of laughs, jests, jeers and insults slowly dies down.
                        Instead, it is replaced by the sheer sounds of <span class="bold">sex</span>:
                        a myriad of grunts, moans and wet, sloppy, messy sounds as well as the ever-intensifying panting of the overexcited kobolds.
                    </div>
                    <div class="text">
                        Only a few minutes of the start of the, admittedly, somewhat arousing spectacle, the bandit in between the dragoness' legs reaches his end.
                        He grunts loudly as he hilts himself inside of the dragoness' tight cunt, tugging her hips towards him with all of his might, in an effort to slide in just teensy bit more of his member.
                        After holding himself inside of her for a few moments more, he pulls his visibly twitching length out of her thoroughly fucked hole, from which his spunk instantly begins to seep.
                    </div>
                    <div class="text">
                        The paw-enamored kobold is next to go.
                        He removes the precum-drenched and footgrime-stained panties from his member, pressing them to his snout once more.
                        Sniffing the underwear loudly, he jerks himself to climax, squirting rope after rope of thick, sticky cum across the dragoness' sole,
                        painting the blue canvas of the bottom of her foot a pearly white.
                        The kobold getting jacked off by the woman and the one enjoying her breasts both reach their ends at the same time, 
                        covering the lady's hand, her arm, her side and the top of her chest with their hefty loads. 
                        A small puddle of semen forms directly above her tits.
                    </div>
                    <div class="text">
                        As becoming of a leader, the kobold fucking the dragoness' muzzle is the last to finish.
                        He continues to intensively <span class="bold">pound</span> the woman's face for several more minutes after his compatriots have already re-dressed.
                        As his friends are starting to get a little impatient, the warm, wet confines of the woman's mouth finally prove to be too much for the ringleader. 
                        He pulls out of her muzzle, gives his spit-slick shaft a few more strokes, and squirts out wave after wave of heady cum across her face.
                        Her eyes get covered, her forehead ends up dripping jizz, her cheeks gets a few ropes of spunk splattered atop them, and the entire length of her muzzle ends up more white than blue.
                    </div>
                    <div class="text">
                        After the last of the ringleader's intense orgasm droops from the tip of his member onto the lady's snout, he lets go of her horns.
                        Too shocked to keep herself upright, the dragoness mindlessly lets her head fall to the ground, the impact luckily cushioned by a blanket of leaves.
                        As the last of the kobolds gets off of her, she remains there, lying completely motionless, semen from all five of her assailants drying atop and dripping from her body.
                        Her eyes are open, but her mind is clearly absent.
                    </div>
                    <div class="text">
                        The ringleader puts his pants back on.
                        All five of the kobolds leave, heading towards the deeper parts of the woods.
                    </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the help button.
     */
    function buildHelpChoice() {
        var disabled = ctrl.bandit.watch.hasMadeKeepWatchingChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getKeepWatchingChoice() == ForestBanditKeepWatchingChoiceEnum.HELP;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${helpId}" ${disabled}>Help</button>`;

        var tooltip = `You feel kind of bad for the dragoness. Check if she's alright, and try to help her.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the scavenge button.
     */
    function buildScavengeChoice() {
        var disabled = ctrl.bandit.watch.hasMadeKeepWatchingChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getKeepWatchingChoice() == ForestBanditKeepWatchingChoiceEnum.SCAVENGE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${scavengeId}" ${disabled}>Scavenge</button>`;

        var tooltip = `With the gang gone, you can check the area and the woman for anything there may be left for you to pocket.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the sloppy seconds button.
     */
    function buildSloppyChoice() {
        var disabled = ctrl.bandit.watch.hasMadeKeepWatchingChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getKeepWatchingChoice() == ForestBanditKeepWatchingChoiceEnum.SLOPPYSECONDS;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${sloppyId}" ${disabled} style="font-size: 18px;">Get sloppy seconds</button>`;

        var tooltip = `This entire ordeal has left you more than a little aroused. Quench your lust on the dragoness' body.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildHelpChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildScavengeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSloppyChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + helpId), $("#" + scavengeId), $("#" + sloppyId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the help button.
     */
    function registerHelpListener() {
        $("#" + helpId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeKeepWatchingChoice(ForestBanditKeepWatchingChoiceEnum.HELP);
            ctrl.bandit.watch.setHelpSource(ForestBanditHelpSourceEnum.KEEPWATCHING);
            loadNextPart(toHelp);
        });
    }

    /**
     * Adds a listener to the scavenge button.
     */
    function registerScavengeListener() {
        $("#" + scavengeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeKeepWatchingChoice(ForestBanditKeepWatchingChoiceEnum.SCAVENGE);
            ctrl.bandit.watch.setScavengeSource(ForestBanditScavengeSourceEnum.KEEPWATCHING);
            loadNextPart(toScavenge);
        });
    }

    /**
     * Adds a listener to the scavenge button.
     */
    function registerSloppyListener() {
        $("#" + sloppyId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeKeepWatchingChoice(ForestBanditKeepWatchingChoiceEnum.SLOPPYSECONDS);
            loadNextPart(toSloppySeconds);
        });
    }


    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditKeepWatching";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.watch.getWatchChoice() !== ForestBanditWatchChoiceEnum.KEEPWATCHING) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += content;
        part += buildChoice();
        part += `</div>`;
        part += toHelp.build();
        part += toScavenge.build();
        part += toSloppySeconds.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.watch.getWatchChoice() !== ForestBanditWatchChoiceEnum.KEEPWATCHING) {
            return;
        }

        toHelp.registerListeners();
        toScavenge.registerListeners();
        toSloppySeconds.registerListeners();

        if (ctrl.bandit.watch.hasMadeKeepWatchingChoice()) {
            return;
        }

        registerHelpListener();
        registerScavengeListener();
        registerSloppyListener();
    };
}

/**
 * Choices for the keep watching part of the bandit encounter.
 */
var ForestBanditKeepWatchingChoiceEnum = {
    HELP: 1,
    SCAVENGE: 2,
    SLOPPYSECONDS: 3
};

// END scenes\Forest\bandit\watch\keepwatching\ForestBanditKeepWatching.js

// scenes\Forest\bandit\watch\scavenge\ForestBanditScavenge.js

/**
 * Scavenge section of the bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditScavenge(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toContinue = new ForestBanditScavengeContinue(controller);

    var doId = "banditScavengeDoButton";
    var dontId = "banditScavengeDontButton";

    var contentStart = `<div class="text">
                            You wait a few more moments, until the gang has left the area.
                            The dragoness lies motionless on the ground, seemingly completely <span class="italic">out of it</span>.
                            The perfect moment for you to grab whatever the bandits didn't.
                            If there's anything left to take, that is.
                        </div>
                        <div class="text">
                            ${(ctrl.bandit.watch.getScavengeSource() === ForestBanditScavengeSourceEnum.KEEPWATCHING) ? "You step out from behind your cover, and walk into the clearing." : "You step out of the bush, and walk into the clearing."}
                            The cum-coated dragoness doesn't even turn her head to face you.
                            She seems completely catatonic, blankly staring up at the sky.
                            Good, that means she won't object to you scouring about for stuff to steal.
                            Still, you keep an eye on her. 
                            You never know.
                            You remember an old rhyme you heard during one of your trips to Varanar:
                            <span class="italic">Around dragons, watch your wagons.</span>
                        </div>
                        <div class="text">
                            The first thing you check is the tattered remains of the woman's robe.
                            You look through the cloth scraps for pockets, a hidden compartiment or any concealed valuables.
                            Sadly, your search comes up empty.
                            There is nothing but torn textile left, and you doubt a merchant would pay any kind of coin for bits of fabric.
                        </div>
                        <div class="text">
                            You move on to inspecting the dragoness' body.
                            From a distance, of course.
                            You don't want to get your scales sticky.
                            Besides the large amount of spunk covering her body, nothing in particular seems to be sticking out.
                            Whatever riches she had were probably taken by the bandits.
                            A small piece of cut rope lies next to her neck, a remnant of her coin pouch, or so you assume.
                            No sign of the purse itself, though.
                        </div>
                        <div class="text">
                            You sigh audibly.
                            You didn't really <span class="italic">expect</span> to find anything, but you still feel kind of let down.
                            With the main points of interest searched, you begin to scrounge the rest of the clearing, 
                            in hopes of finding at least <span class="italic">something</span> to take home.
                        </div>
                        <div class="text">
                            You comb through the leaves and you brush through a few patches of grass,
                            looking for anything that may have been lost during the struggle.
                            A stray coin, a lost ring, maybe even a necklace.
                            Heck, you'd settle for a gold tooth.
                        </div>
                        <div class="text">
                            Again, you find nothing.
                            Just as you are about to completely abandon your search, 
                            you feel something <span class="italic">other</span> than the usual muddy dirt underneath your foot.
                            Looking down, you find that you are standing on top of the dragoness' thoroughly used and then discarded `;

    var contentEnd = `  </div>
                        <div class="text">
                            At first, you are revolted.
                            While made out of a fine material, the underwear is soaked with spunk and precum, dirty with mud, and filthy 
                            ${(ctrl.bandit.watch.getScavengeSource() === ForestBanditScavengeSourceEnum.KEEPWATCHING) ? "from it being rubbed against the dragoness' sole and" : ""}
                            from being stepped on by you, <span class="italic">just now</span>.
                            Then, you begin to remember the way they looked <span class="italic">before</span> the bandits, particularly 
                            ${(ctrl.bandit.watch.getScavengeSource() === ForestBanditScavengeSourceEnum.INDULGE) ? "your new friend" : "the foot-obsessed one"}, 
                            got their hands on it.
                            They covered the dragoness' loins rather nicely.
                            And they fit her frame <span class="bold">so</span> well.
                        </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content of this section.
     */
    function buildContent() {
        var content = contentStart;
        var panties = new SilkPanties();
        panties.stain("Cum and Mud");
        content += global.kobold.inventory.createItemWithTooltip(panties) + ".";
        content += contentEnd;
        return content;
    }

    /**
     * Builds and returns the do button.
     */
    function buildDoChoice() {
        var disabled = ctrl.bandit.watch.hasMadeScavengeChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getScavengeChoice() == ForestBanditScavengeChoiceEnum.DO;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${doId}" ${disabled}>Put them on</button>`;

        var tooltip = `They <span class="bold">are</span> made out of silk. It'd be a shame to let them go to waste.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the don't button.
     */
    function buildDontChoice() {
        var disabled = ctrl.bandit.watch.hasMadeScavengeChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getScavengeChoice() == ForestBanditScavengeChoiceEnum.DONT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${dontId}" ${disabled}>Do not</button>`;

        var tooltip = `Gross, you're not going to wear those!`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildDoChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildDontChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables both buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + doId), $("#" + dontId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the help button.
     */
    function registerDoListener() {
        $("#" + doId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeScavengeChoice(ForestBanditScavengeChoiceEnum.DO);
            loadNextPart(toContinue);
        });
    }

    /**
     * Adds a listener to the scavenge button.
     */
    function registerDontListener() {
        $("#" + dontId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeScavengeChoice(ForestBanditScavengeChoiceEnum.DONT);
            loadNextPart(toContinue);
        });
    }


    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditScavenge";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.bandit.watch.hasScavengeSource()) {
            return "";
        }
        ctrl.bandit.watch.scavengeShared();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toContinue.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.bandit.watch.hasScavengeSource()) {
            return;
        }

        toContinue.registerListeners();

        if (ctrl.bandit.watch.hasMadeScavengeChoice()) {
            return;
        }

        registerDoListener();
        registerDontListener();
    };
}

/**
 * Choices for the scavenge part of the bandit encounter.
 */
var ForestBanditScavengeChoiceEnum = {
    DO: 1,
    DONT: 2
};

/**
 * Source for the scavenge part of the bandit encounter.
 */
var ForestBanditScavengeSourceEnum = {
    MASTURBATE: 1,
    KEEPWATCHING: 2,
    INDULGE: 3
};

// END scenes\Forest\bandit\watch\scavenge\ForestBanditScavenge.js

// scenes\Forest\bandit\watch\scavenge\ForestBanditScavengeContinue.js

/**
 * Continue part of the scavenge section of the bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditScavengeContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var continueId = "banditScavengeContinueButton";
    
    var doNot = `   <div class="text">
                        <span class="bold">Hell no.</span>
                        The panties are far too filthy for you to even <span class="italic">consider</span> picking them up, let alone to wear them.
                        Using your foot, you grind the dirty fabric deeper into the mud, where it belongs.
                        With some luck, they will never again be found, let alone by some perverted, foot-huffing kobold.
                    </div>
                    <div class="text">
                        It's time to cut your losses.
                        You aren't going to find anything of value here.
                        You leave the traumatized dragoness behind, and venture back into the woods.
                        Night is starting to fall.
                        You decide to head home, to avoid any of the <span class="italic">unsavory elements</span>  that come out after the sun goes down.
                    </div>`;
    
    var doStart = ` <div class="text">
                        You lift your foot off of the filthy garment.
                        You bend over and pick the panties up off the ground.
                        Shaking the garment, you manage to get a lot of the loose dirt to fall off.
                        The mud, however, is too deeply engrained for you to be able to simply shake it away, not to mention the jizz.
                    </div>
                    <div class="text">
                        You hold the panties in front of your face.
                        They smell like the dragoness does: of drying semen.
                        Still, they seem about your waist size, and the few small, relatively unsoiled bits feel oh-so smooth to the touch.
                        You've never held underwear of this high quality, regardless of its current condition.
                    </div>
                    <div class="text">
                        After a moment of doubt, you finally make up your mind.
                        You're going to put on these filthy, cum-stained panties, you're going to wear them proudly, 
                        and hopefully, you're going to find a way to clean them before long, to restore them to their former glory.
                    </div>`;
    
    var warily = `  <div class="text">
                        You look around warily, scouting for any prying eyes.
                        You're reasonably sure nobody's looking.
                        Well, nobody apart from the dragoness, that is.
                        But somehow, you don't think she'll mind seeing just one more naked kobold today.
                    </div>`;
    
    var pantyDescription = `<div class="text">
                                You feel somewhat disgusted, but they fit <span class="bold">so</span> well!
                                The texture of the silk remains smooth, even when soaked with a mixture of spunk and mud.
                                While you doubt the restorative capabilities of semen, you've heard of some fancy nobles taking baths in mud,
                                so you're sure that won't be detrimental, at least.
                            </div>`;
    
    var pantyEnd = `<div class="text">
                        With your new panties on, you march out of the clearing, back into the woods.
                        You aren't quite sure whether you should feel proud or ashamed.
                        With mixed feelings, you begin the short trip back to your home.
                    </div>`;
    

    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    
    
    function buildPantsAndUnderwear() {
        var pants = global.kobold.inventory.createItemWithTooltip(ctrl.bandit.watch.scavengeContinuePants());
        var underwear = global.kobold.inventory.createItemWithTooltip(ctrl.bandit.watch.scavengeContinueUnderwear());
        var toReturn = warily;
        toReturn = `<div class="text">
                        You lower your ${pants} to the ground, and step out of them.
                        Checking your surroundings one more time, your ${underwear} follows suit.
                        Now completely naked, you quickly slip on the filthy panties.
                        They leave a trail of filth and dirt as they travel upwards along your scaly legs,
                        before wetly settling against your groin.
                    </div>`;
        toReturn += pantyDescription;
        toReturn += `<div class="text">
                        You put your ${pants} back on, hiding your new disgusting underwear beneath them.
                        Your previous pair, you toss towards the dragoness.
                        She'll be grateful to have at least <span class="italic">something</span> to wear.
                        When she snaps out of her shocked condition, that is.
                        <span class="italic">If</span> she snaps out of her shocked condition.
                    </div>`;
        toReturn += pantyEnd;
        return toReturn;
    }
    
    function buildPants() {
        var pants = global.kobold.inventory.createItemWithTooltip(ctrl.bandit.watch.scavengeContinuePants());
        var toReturn = warily;
        toReturn = `<div class="text">
                        Checking your surroundings one more time, you lower your ${pants} to the ground, and step out of them.
                        Now completely naked, you quickly slip on the filthy panties.
                        They leave a trail of filth and dirt as they travel upwards along your scaly legs,
                        before wetly settling against your groin.
                    </div>`;
        toReturn += pantyDescription;
        toReturn += `<div class="text">
                        You put your ${pants} back on, hiding your new disgusting underwear beneath them.
                        Well, at least you <span class="bold">have</span> underwear now.
                        The same can not be said about the dragoness, whom you leave behind without so much as her own filthy panties.
                    </div>`;
        toReturn += pantyEnd;
        return toReturn;
    }
    
    function buildUnderwear() {
        var underwear = global.kobold.inventory.createItemWithTooltip(ctrl.bandit.watch.scavengeContinueUnderwear());
        var toReturn = warily;
        toReturn = `<div class="text">
                        Doublechecking your surroundings one more time, you lower your ${underwear} to the ground, and step out of it.
                        Now completely naked, you quickly slip on the filthy panties.
                        They leave a trail of filth and dirt as they travel upwards along your scaly legs,
                        before wetly settling against your groin.
                    </div>`;
        toReturn += pantyDescription;
        toReturn += `<div class="text">
                        You toss your ${underwear} towards the dragoness, for her to wear when she snaps out of whatever trance she's in.
                        You won't be needing it anymore, 
                        and you imagine she'll be grateful to have at least <span class="italic">something</span> to wear.
                    </div>`;
        toReturn += pantyEnd;
        return toReturn;
    }
    
    function buildNeither() {
        var toReturn = `<div class="text">
                            Seeing as you don't really have anything to wear around your waist,
                            you'll take whatever you can get.
                            Even a mud-stained cumrag like this particular pair of panties is better than nothing.
                            You quickly slip on the filthy panties.
                            They leave a trail of filth and dirt as they travel upwards along your scaly legs,
                            before wetly settling against your groin.
                        </div>`;
        toReturn += pantyDescription;
        toReturn += pantyEnd;
        return toReturn;
    }
    
    /**
     * Builds and returns the wear section.
     */
    function buildWear() {
        var pants = ctrl.bandit.watch.scavengeContinuePants();
        var underwear = ctrl.bandit.watch.scavengeContinueUnderwear();
        var toReturn = doStart;
        
        if(pants && underwear) {
            toReturn += buildPantsAndUnderwear();
        } else if(pants) {
            toReturn += buildPants();
        } else if (underwear) {
            toReturn += buildUnderwear();
        } else {
            toReturn += buildNeither();
        }
        
        ctrl.bandit.watch.scavengeContinueChange();
        return toReturn;
    }
    
    /**
     * Builds and returns the do not wear section.
     */
    function buildDoNotWear() {
        return doNot;
    }
    
    /**
     * Builds and returns the content.
     */
    function buildContent() {
        var wear = (ctrl.bandit.watch.getScavengeChoice() == ForestBanditScavengeChoiceEnum.DO);
        if(wear) {
            return buildWear();
        } else {
            return buildDoNotWear();
        }
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.bandit.watch.scavengeHasContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }
    
    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.scavengeContinue();
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditScavengeContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.bandit.watch.hasMadeScavengeChoice()) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.bandit.watch.hasMadeScavengeChoice()) {
            return;
        }

        if (ctrl.bandit.watch.scavengeHasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\bandit\watch\scavenge\ForestBanditScavengeContinue.js

// scenes\Forest\bandit\watch\masturbate\indulge\ForestBanditMasturbateIndulge.js

/**
 * The indulge part of the masturbate section of the watch scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditMasturbateIndulge(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    var underwear;
    var underwearPopup;
    var pants;
    var pantsPopup;
    var bra;
    var braPopup;
    var shirt;
    var shirtPopup;
    var boobs;
    var cock;
    var cunt;
    var cuntVirgin;
    var gender;

    var scavengeId = "banditMasturbateIndulgeScavengeButton";
    var helpId = "banditMasturbateIndulgeHelpButton";

    var toHelp = new ForestBanditHelp(controller);
    var toScavenge = new ForestBanditScavenge(controller);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function buildIndulge() {
        var toReturn = `<div class="text">
                            You smile shyly at the kobold.
                            Does he really want to jerk off to <span class="italic">you</span> instead of to the sight of his friends fucking a dragoness?
                            In a weird way, this makes you feel some twisted kind of pride.
                            Like you're <span class="italic">better</span> than her.
                            Your heart races.
                            You decide to do it.
                        </div>
                        <div class="text">
                            You glance over to the other bandits, to make sure they're still occupied.
                            They sure as hell are, their collective attention focused on the dragoness.
                            Cautiously, you take a small step forwards, out of the bush you were concealed in, 
                            your naked body on full display for the panty-huffing kobold.
                            His grin deepens.
                            He seductively licks his lips, his member twitching visibly, 
                            the precum-stain on the panties at the tip of his dick darkening just a little.
                        </div>
                        <div class="text">
                            While you feel incredibly exposed, showing off your nude body to a bandit for him to jack off to <span class="bold">excites</span> you to no end.
                            The exhilarating rush of self-exposition causes `;
        if (cock) {
            toReturn += `   your own cock to beg for attention.`;
        } else if (cunt) {
            toReturn += `   your <span class="bold">dripping</span> wet slit to beg for attention.`;
        } else {
            toReturn += `   a fresh surge of arousal on your end.`;
        }

        toReturn += `       The other kobold takes the lead. 
                            He grips his dick by the base, and vigorously begins to stroke himself, pumping his length with slow, heavy motions.
                            It looks like he's trying to show off.
                            Wordlessly, he hints for you to continue your own self-pleasuring.
                            You gladly oblige.
                        </div>
                        <div class="text center">
                            <a href="scenes/Forest/bandit/watch/masturbate/indulge/footsie-full.png" target="_blank">
                                <img class="nsfw" src="scenes/Forest/bandit/watch/masturbate/indulge/footsie.png" style="width: 400px;" title="Made by Sneaky">
                            </a>
                        </div>
                        <div class="text">
                            With your mind less focused on the gangrape, and more on the rather attractive male in front of you, you resume your masturbation.`;
        if (cock) {
            toReturn += `   You take hold of your precum-slick shaft, and continue to jerk it.`;
        }
        if (cunt) {
            toReturn += `   You slide a second finger inside of your moist cunt.`;
        }
        if (boobs) {
            toReturn += `   You pinch one of your rock-hard nipples, and fail to stifle a soft moan.`;
        }

        toReturn += `       Your excitement is made clear through your hasty pace, your fingers moving up and down at an incredible rate.
                            The kobold removes the panties from his dick, clutching the fabric it in one hand, while stroking his considerable length with the other.
                            Prying your gaze away from his groin, for a moment, you notice his eyes roaming over your body, inspecting every inch of your bare scales.
                        </div>
                        <div class="text">
                            You are a little <span class="italic">too</span> eager, or so it seems.
                            It isn't long until you reach your end.`;

        if (cock) {
            toReturn += `   You frantically pump your shaft, practically <span class="bold">milking</span> it with your hand.
                            A small grunt escapes your lips as the first squirt of your spunk shoots from the tip of your dick,
                            the thick, white, sticky goo splattering down atop the leaves covering the ground.
                            You continue to stroke yourself, coaxing more and more ropes of cum from your length,
                            your jaw agape as you experience your most intense orgasm ever.`;
        } else if (cunt) {
            toReturn += `   You spread your legs to allow your fingers just a <span class="italic">little</span> better access to your folds.
                            You slide both of the digits in and out of your slit at a high pace, your climax rapidly building.
                            One final time, you hilt your fingers inside of you, all the way to the knuckle.
                            Your head rears backwards, and your tongue lolls out of your mouth obscenely, 
                            as the most intense orgasm of your entire life hits your body <span class="bold">hard</span>.
                            Feminine fluids liberally squirt from your loins, splattering down atop the leaves covering the ground.`;
        }

        toReturn += `   </div>
                        <div class="text">
                            Slowly, you begin to recover from the intense climax.
                            It had been a while. You <span class="bold">needed</span> that.
                            As you recover your senses, your attention drifts towards the other kobold once more.
                            He hasn't gotten off yet, although he's still jerking off.
                            He notices you staring, and his eyes move to meet yours, a pleased, yet lewd smile on his face.
                            With the hand still clutching the dragoness' panties, he motions for you to turn around.
                        </div>
                        <div class="text">
                            Lost in the heat of the moment, you comply with his unspoken order.
                            You turn around and lean forwards, bracing yourself against the tree you were hiding behind not too long ago.
                            You lift your tail, exposing your taut rear to the amorous bandit.
                            You hear him approach, leaves crinkling underneath his steps.
                        </div>
                        <div class="text">
                            Standing right behind you, he places a hand atop your back.
                            You can practically <span class="bold">feel</span> the heat radiating from his member, mere inches away from your ass.
                            Just as you brace yourself, as you think you're about to get fucked,
                            as your mind gets clogged by the thought of finally losing your virginity,
                            the bandit lets out a loud, relieved sigh.
                            Instead of the thorough pounding you were expecting, you feel his warm, sticky cum splattering across your backside,
                            some of his spunk rolling in between your buttcheeks to coat your tender, waiting ass.
                        </div>
                        <div class="text">
                            While you <span class="italic">are</span> slightly disappointed by the anticlimax, 
                            the reason-driven half of your brain argues that perhaps it is for the best that you did not lose your virginity
                            to a heat-of-the-moment fling with a random bandit you've never even seen before.
                        </div>
                        <div class="text">
                            His orgasm coming to an end, the kobold leans over you. 
                            Whispering and audibly out of breath, he speaks:<br/>
                            <div class="center">"<span class="bold blue">Good ${gender === GenderEnum.FEMALE ? "girl" : "boy"}.
                            Go hide, before the others see you.
                            We will meet again.</span>"</div>
                        </div>
                        <div class="text">
                            With those words uttered, the male rises off of you.
                            You come to an upright pose yourself.
                            The two of you exchange a solemn, mutual nod.
                            As the bandit finishes wiping his dick clean with the dragoness' panties,
                            he begins to walk back to his comrades.
                            You decide to follow his advice, and tuck yourself away in the bush once more.
                        </div>
                        <div class="text">
                            The other kobolds seem to have wrapped up their business with the dragoness,
                            leaving her a near-catatonic, cum-drenched mess.
                            As their friend re-joins them, they collectively get dressed.
                            The group prepares to leave.
                            Your newly acquired foot-loving friend casts a final look your way, 
                            giving you one last wink before heading out of the clearing along with his peers.
                        </div>
                        <div class="text">
                            You wipe the bandit's cum off your ass.
                            That sure was a heated encounter.
                            Your new friend seemed slightly <span class="italic">different</span> from the rest,
                            although he might just have been acting that way to get under your tail. 
                            ${(pants || bra || underwear) ? "After redressing, you think of what to do next." : "You think of what to do next."}
                        </div>`;
        toReturn += ctrl.bandit.watch.masturbateIndulgeTogetherChange();
        return toReturn;
    }

    function buildDoNot() {
        var toReturn = `<div class="text">
                            While completely apalled by the unspoken offer, you put on a fake, shy smile, not to anger the bandit.
                            You shake your head no, and remain in the cover of the bush.
                            He rolls his eyes at you, gives you a disapproving stare, and then turns back towards the dragoness.
                            Close call. 
                            You wonder why he doesn't just tell his friends about you.
                            Maybe he didn't see your coin pouch, or maybe he isn't all that much of a bad guy after all.
                        </div>
                        <div class="text">
                            With the amorous kobold dissuaded, you turn your attention back towards your own arousal,
                            and towards the scene unfolding in front of your very eyes.
                            The kobold fucking the dragoness has upped both his pace and the intensity of his thrusts,
                            the woman's thick thighs jiggling with every fierce motion of his hips.`;

        if (cock) {
            toReturn += `   You resume stroking yourself, imagining that you're the one in his position, 
                            pistoning away at a voluptuous <span class="bold">whore</span> with a tight, naturally lubricated cunt
                            that was practically <span class="bold">built</span> to <span class="bold">milk</span> the cum from hard, throbbing dicks like your own.`;
        } else if (cunt) {
            toReturn += `   You slide a second digit into your cunt and resume your fingering, 
                            imagining that you're the one getting <span class="bold">rutted</span> by an entire <span class="bold">gang</span> of kobolds,
                            each one of them taking turns in abusing your already spunk-filled slit.
                            Your mouth would of course be put to work cleaning up the ones that have already fucked you.
                            Your imagination is so vivid, that you can practically <span class="bold">taste</span> their members coated in the spent seed of their peers.`;
        }

        toReturn += `   </div>
                        <div class="text">
                            You close your eyes, and let your imagination take you, fueled by the myriad of grunts, 
                            groans and lewd, sloppy noises coming from the clearing.`;
        
        if (cock) {
            toReturn += `   You lean back slightly and pick up the pace of your jerking, the side of your hand bumping into your waist with every heated pump.`;
        } else if (cunt) {
            toReturn += `   Your fingers pick up their pace, pistoning in and out of your feminine fluid-lubricated hole.`;
        }
        
        toReturn += `       You are so close. So very, <span class="bold">very</span> close. You can feel your orgasm building up inside of you.
                        </div>
                        <div class="text">
                            As your need builds, your motions become more erratic.`;
        
        if (cock) {
            toReturn += `   Your tongue lolls out of your mouth obscenely as your cock begins to quiver and tremble with need,
                            your internal balls aching for relief.`;
        } else if (cunt) {
            toReturn += `   You hastily finger your arousal-dripping cunt. 
                            Bringing your second hand to your groin as well, you rub your tiny little clit with a single finger.
                            That last act sends you over the edge.`;
        }
        
        toReturn += `       Finally, you reach your end, a phenomenal climax washing over your body.`;
        
        if (cock) {
            toReturn += `   You squirt your load over the leafy forest floor, 
                            rope after rope of sticky white kobold cum splattering across the otherwise brown and yellow leaves.`;
        } else if (cunt) {
            toReturn += `   You pull out your fingers, yet continue to rub yourself as you squirt your excitement all over the forest floor,
                            your eyes rolling to the back of your head, your back arching and your tongue lolling out of your mouth obscenely.`;
        }
        
        toReturn += `   </div>
                        <div class="text">
                            Post-orgasmal bliss washes over your mind.
                            It had been a while.
                            You really <span class="bold">needed</span> that.
                            Slowly, you open your eyes, to find that the gang of kobolds has already left,
                            leaving behind a naked, well-fucked and cum covered dragoness.
                            You take a moment to clean your 
                            groin${(pants || bra || underwear) ? ", before redressing" : ""}.
                            While that was nice, it's time to decide what to do next.
                        </div>`;
        
        toReturn += ctrl.bandit.watch.masturbateIndulgeChange();
        return toReturn;
    }

    /**
     * Builds and returns the content.
     */
    function buildContent() {
        if (ctrl.bandit.watch.getMasturbateChoice() == ForestBanditMasturbateChoiceEnum.INDULGE) {
            return buildIndulge();
        } else {
            return buildDoNot();
        }
    }

    /**
     * Builds and returns the help button.
     */
    function buildHelpChoice() {
        var disabled = ctrl.bandit.watch.hasMadeMasturbateIndulgeChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getMasturbateIndulgeChoice() == ForestBanditMasturbateIndulgeChoiceEnum.HELP;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${helpId}" ${disabled}>Help</button>`;

        var tooltip = `You feel kind of bad for the dragoness. Check if she's alright, and try to help her.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the scavenge button.
     */
    function buildScavengeChoice() {
        var disabled = ctrl.bandit.watch.hasMadeMasturbateIndulgeChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getMasturbateIndulgeChoice() == ForestBanditMasturbateIndulgeChoiceEnum.SCAVENGE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${scavengeId}" ${disabled}>Scavenge</button>`;

        var tooltip = `With the gang gone, you can check the area and the woman for anything there may be left for you to pocket.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildHelpChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildScavengeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables both buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + scavengeId), $("#" + helpId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the help button.
     */
    function registerHelpListener() {
        $("#" + helpId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeMasturbateIndulgeChoice(ForestBanditMasturbateIndulgeChoiceEnum.HELP);
            if (ctrl.bandit.watch.getMasturbateChoice == ForestBanditMasturbateChoiceEnum.INDULGE) {
                ctrl.bandit.watch.setHelpSource(ForestBanditHelpSourceEnum.INDULGE);
            } else {
                ctrl.bandit.watch.setHelpSource(ForestBanditHelpSourceEnum.MASTURBATE);
            }
            loadNextPart(toHelp);
        });
    }

    /**
     * Adds a listener to the scavenge button.
     */
    function registerScavengeListener() {
        $("#" + scavengeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeMasturbateIndulgeChoice(ForestBanditMasturbateIndulgeChoiceEnum.SCAVENGE);
            if (ctrl.bandit.watch.getMasturbateChoice == ForestBanditMasturbateChoiceEnum.INDULGE) {
                ctrl.bandit.watch.setScavengeSource(ForestBanditScavengeSourceEnum.INDULGE);
            } else {
                ctrl.bandit.watch.setScavengeSource(ForestBanditScavengeSourceEnum.MASTURBATE);
            }
            loadNextPart(toScavenge);
        });
    }

    /**
     * Initializes all required variables.
     */
    function initVars() {
        pants = ctrl.bandit.watch.masturbatePants();
        pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        underwear = ctrl.bandit.watch.masturbateUnderwear();
        underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        bra = ctrl.bandit.watch.masturbateBra();
        braPopup = global.kobold.inventory.createItemWithTooltip(bra);
        shirt = ctrl.bandit.watch.masturbateShirt();
        shirtPopup = global.kobold.inventory.createItemWithTooltip(shirt);
        boobs = ctrl.bandit.watch.masturbateBoobs();
        cock = ctrl.bandit.watch.masturbateCock();
        cunt = ctrl.bandit.watch.masturbateCunt();
        cuntVirgin = ctrl.bandit.watch.masturbateCuntVirgin();
        gender = ctrl.bandit.watch.masturbateBaseGender();
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditMasturbateIndulge";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.bandit.watch.hasMadeMasturbateChoice()) {
            return "";
        }

        initVars();
        ctrl.bandit.watch.masturbateIndulgeShared();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toHelp.build();
        part += toScavenge.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.bandit.watch.hasMadeMasturbateChoice()) {
            return;
        }

        toHelp.registerListeners();
        toScavenge.registerListeners();

        if (ctrl.bandit.watch.hasMadeMasturbateIndulgeChoice()) {
            return;
        }

        registerHelpListener();
        registerScavengeListener();
    };

}

/**
 * Choices for the indulging part of the bandit encounter.
 */
var ForestBanditMasturbateIndulgeChoiceEnum = {
    HELP: 1,
    SCAVENGE: 2
};

// END scenes\Forest\bandit\watch\masturbate\indulge\ForestBanditMasturbateIndulge.js

// scenes\Forest\bandit\watch\masturbate\ForestBanditMasturbate.js

/**
 * The masturbate part of the watch scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditMasturbate(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    var underwear;
    var underwearPopup;
    var pants;
    var pantsPopup;
    var bra;
    var braPopup;
    var shirt;
    var shirtPopup;
    var boobs;
    var cock;
    var cunt;
    var cuntVirgin;
    
    var indulgeId = "banditMasturbateIndulgeButton";
    var dontId = "banditMasturbateDontIndulgeButton";
    
    var toIndulge = new ForestBanditMasturbateIndulge(ctrl);
    
    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the starting section.
     */
    function buildStart() {
        var toReturn = `<div class="text">
                            While morally questionable, if not outright unethical, 
                            the rare display of kobold dominance strikes your fancy in all the <span class="italic">wrong</span> ways.
                            Dragons are typically amongst the higher reaches of society.
                            Watching one of them getting abused by your own lowly kind is proving to be rather arousing, 
                            and somewhat mentally fulfilling, in a weird and twisted way. `;
        if (pants && underwear) {
            toReturn += `You slide a hand underneath the waistband of your ${pantsPopup}, down into your ${underwearPopup}.`;
        } else if (pants) {
            toReturn += `You slide a hand underneath the waistband of your ${pantsPopup}.`;
        } else if (underwear) {
            toReturn += `You slip a hand into your ${underwearPopup}.`;
        } else {
            toReturn += `You move one of your hands to your bare groin.`;
        }

        toReturn += ` Slowly, you begin to touch yourself.
                        </div>`;
        return toReturn;
    }

    /**
     * Builds and returns the middle section.
     */
    function buildMid() {
        var toReturn = `<div class="text">
                            The bandits waste little time.
                            One of them removes the dragoness' underwear, passing it to the foot-obsessed kobold, who seems overjoyed with the gift.
                            He eagerly huffs the ${global.kobold.inventory.createItemWithTooltip(new SilkPanties())}, stroking his member to their scent.
                            You can only imagine what the insides of a dragoness' underwear smells like.
                        </div>
                        <div class="text">
                            Besides being one of the more <span class="italic">regal</span> species around, dragons are also highly coveted sexual partners.
                            Whereas vixens are natural born whores, dragonesses are more akin to courtesans, if not queens or princesses themselves.
                        </div>
                        <div class="text">
                            With her panties removed, the woman's loins are covered only by her thick thighs.
                            Not for long.
                            Two of the muggers forcibly part her legs, despite her renewed resistance.
                            She squirms and struggles, yet her resistance is futile.
                            Her legs are pulled away from one another, and her most private parts are revealed to not only the kobolds, but to you as well.
                         </div>   
                        <div class="text">
                            You focus on the sight of the dragoness' folds. She looks tight. Barely used. 
                            This one must be a princess instead of a high-class prostitute. Or maybe she's been without customers for a long-ass time.
                            Still, a teensy bit of natural lubrication lingers on her nether lips, making them glisten invitingly.
                            You wonder whether this is a natural phenomenon amongst dragons, or whether the <span class="bold">whore</span> is actually getting off to this.
                            <span class="bold">Both</span> of those thoughts fuel your rising arousal.
                        </div>   
                        <div class="text">`;
        toReturn += buildMidMasturbate();
        toReturn += `       You slip into the bush next to the tree you're hiding behind to get a better look at the entire ordeal.
                        </div>
                        <div class="text">
                            A bandit crawls in between the dragoness' spread blue haunches and gives her loins a few courtesy rubs, 
                            spreading her slick fluids around before promptly lining up his member.
                            With a single, smooth rolling motion of his hips, the male hilts himself inside of the woman's cunt, 
                            her eyes going wide and her protests dying down the moment she gets penetrated.
                            Her body goes limp as she realizes there <span class="bold">is</span> no escape.
                        </div>
                        <div class="text">
                            Mistaking her <span class="italic">giving up</span> for actual submission, the gang is driven into a lust-frenzy.
                            The kobold in between the lady's legs grabs onto her hips, and begins to <span class="bold">fuck</span> her with reckless abandon.
                            The ringleader forces his member past her lips, and while keeping her head in place by her horns, 
                            he starts to hump the woman's face, thrusting his length in and out of her warm, wet mouth.
                        </div>
                        <div class="text">
                            While you <span class="italic">do</span> imagine yourself being the one riding the dragoness' muzzle,
                            you also find yourself wondering what it'd be like to be on the <span class="italic">receiving</span> end of the intensive bout of group sex.
                            For just a moment, you fantasize about the taste of the ringleader's cock, before pulling your mind out of the gutter.
                        </div>
                        <div class="text">
                            Two thugs, one on either side of the dragoness, take hold of her hands, and place her scaly palms atop their groins.
                            She outright refuses to grip onto their shafts, until threatened with the dagger once more.
                            Reminded that her life is at stake, she reluctantly wraps her fingers around the kobolds' cocks, 
                            and at an agonizingly slow pace, she begins to give them half-hearted handjobs. `;
        toReturn += buildMidWish();
        return toReturn;
    }

    /**
     * Builds and returns the masturbation part of the middle section.
     */
    function buildMidMasturbate() {
        var toReturn = `Your masturbation intensifies. `;
        if (cunt) {
            toReturn += `You eagerly rub two fingers along the length of your ${cuntVirgin ? "virginal " : ""} slit.
                         While the dragoness may be somewhat moist, you are completely <span class="bold">drenched</span> with excitement,
                         ${underwear ? "making a complete mess of your underwear" : "running stickily down your thighs"}. `;
        }
        if (boobs) {
            toReturn += `One of your hands ${shirt ? "slides underneath your shirt" : "slithers upwards over your body"}, towards your chest.
                         You cup one of your breasts, and start to rhythmically squeeze it${shirt ? ` through your ${shirtPopup}` : ``}. `;
        }
        if (cock) {
            toReturn += `The tip of your member has begun to emerge from your cockslit. You grip onto it with two of your fingers and stroke it,
                         causing more of your length to appear from the confines of your slit. Before long, `;
            if (pants) {
                toReturn += `you are clearly pitching a tent in your ${pantsPopup}. `;
            } else if (underwear) {
                toReturn += `your member visibly protrudes from your ${underwearPopup}. `;
            } else {
                `you are fully erect. `;
            }
        }

        return toReturn;
    }

    /**
     * Builds and returns the wishing part of the middle section.
     */
    function buildMidWish() {
        var toReturn = ``;

        if (cock) {
            toReturn += `You wish she was jerking <span class="bold">you</span> off, instead of those guys.`;
        } else {
            toReturn += `You wish you were the one jerking them off, with their dicks mere inches away from your face.`;
        }

        toReturn += `   The mere thought sends shivers down your spine.
                    </div>
                    <div class="text">
                        It's been <span class="bold">so</span> long since you've last properly rubbed one out. `;
        
        if(pants) {
            toReturn += `You lower your pants to your knees. `;
            if(underwear) {
                toReturn += `Your underwear quickly follows suit. `;
            }
        } else if(underwear) {
            toReturn += `You lower your underwear to your knees. `;
        }
        
        if(boobs && bra) {
            toReturn += `You lift your bra over your chest, exposing your tits, allowing you to caress them directly. 
                         You begin to play with your nipples, adding to your building lust. `;
        } else if(boobs) {
            toReturn += `You begin to toy with your nipples, adding to your building lust. `;
        }
        
        if(cunt) {
            toReturn += `You slip a single finger inside of your tight snatch.
                         <span class="bold">God</span>, that feels <span class="bold">good</span>.
                         You immediately begin to pump the small digit in and out of your slit, sending waves of pleasure coursing throughout your body. `;
        }
        if(cock) {
            toReturn += `You take hold of your shaft, and begin to stroke it.
                         A thick bauble of precum forms at the tip. 
                         Instead of letting it drip down to the ground, 
                         you collect the droplet using your palm, 
                         and rub it across the length of your member, 
                         slickening it slightly, allowing you to speed up your frantic jerking. `;
        }

        toReturn += `You have trouble controlling your breath, beginning to huff and pant loudly.
                     Luckily, the bandits are doing the same.
                    </div>
                    <div class="text">
                        <a href="scenes/Forest/bandit/watch/masturbate/masturbation-${cock ? "male" : "female"}-full.png" target="_blank">
                            <img class="nsfw" src="scenes/Forest/bandit/watch/masturbate/masturbation-${cock ? "male" : "female"}.png" title="Made by Sneaky">
                        </a>
                    </div>`;

        return toReturn;
    }
    
    /**
     * Builds and returns the end section.
     */
    function buildEnd() {
        var toReturn = `<div class="text">
                            Four of them seem to be thoroughly enjoying the dragoness' body.
                            The ones getting handjobs have taken to wrapping their own fingers around the woman's,
                            forcibly tightening her grasp, guiding her hand to increase her pace.
                            At this point, they might as well be jacking themselves off, 
                            but you imagine the dragoness' palms are <span class="bold">far</span> softer than their own rugged graspers.
                            At least it seems that way, with how they buck their hips against every stroke, practically <span class="bold">fucking</span> the woman's hands,
                            her scales slickened by their arousal, a droplet of which occasionally slips through her grasp and falls to the ground.
                        </div>
                        <div class="text">
                            The only bandit not actively partaking in the escalating gangrape is the one huffing the dragoness' panties.
                            Standing off to the side, he lowers the smooth fabric to his groin, and wraps it around his member, using the feminine garment to stroke himself with.
                            Then, suddenly, he turns around. 
                            He sniffs the air for a moment, before turning towards...you.
                            His eyes fixate upon yours, staring at you through the shrubbery.
                            A wicked smile forms atop his face.
                            You freeze up mid-masturbation.
                        </div>
                        <div class="text">
                            The thought of running crosses your mind, but the odd kobold doesn't seem to be alerting his comrades just yet.
                            He winks slyly at you, bucking his hips in your direction whilst giving you a sultry look.
                            He suggestively looks down at his panty-covered member, and then back up to you.
                            Does he want you to give him a show?
                        </div>`;
        return toReturn;
    }

    /**
     * Builds and returns the content.
     */
    function buildContent() {
        var toReturn = buildStart();
        toReturn += buildMid();
        toReturn += buildEnd();
        return toReturn;
    }
    
    /**
     * Builds and returns the do button.
     */
    function buildIndulgeChoice() {
        var disabled = ctrl.bandit.watch.hasMadeMasturbateChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getMasturbateChoice() == ForestBanditMasturbateChoiceEnum.INDULGE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${indulgeId}" ${disabled}>Indulge him</button>`;

        var tooltip = `He <span class="italic">is</span> asking nicely, and you <span class="italic">have</span> been spying
                       on his entire gang fucking a dragoness. It'd be rude not to give him a show in return.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the don't button.
     */
    function buildDontChoice() {
        var disabled = ctrl.bandit.watch.hasMadeMasturbateChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getMasturbateChoice() == ForestBanditMasturbateChoiceEnum.DONT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${dontId}" ${disabled}>Do not</button>`;

        var tooltip = `What a freak. You aren't going to expose yourself for his amusement!`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildIndulgeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildDontChoice();
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables both buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + indulgeId), $("#" + dontId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the help button.
     */
    function registerIndulgeListener() {
        $("#" + indulgeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeMasturbateChoice(ForestBanditMasturbateChoiceEnum.INDULGE);
            loadNextPart(toIndulge);
        });
    }

    /**
     * Adds a listener to the scavenge button.
     */
    function registerDontListener() {
        $("#" + dontId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeMasturbateChoice(ForestBanditMasturbateChoiceEnum.DONT);
            loadNextPart(toIndulge);
        });
    }

    /**
     * Initializes all clothing and clothing popup variables.
     */
    function initClothing() {
        pants = ctrl.bandit.watch.masturbatePants();
        pantsPopup = global.kobold.inventory.createItemWithTooltip(pants);
        underwear = ctrl.bandit.watch.masturbateUnderwear();
        underwearPopup = global.kobold.inventory.createItemWithTooltip(underwear);
        bra = ctrl.bandit.watch.masturbateBra();
        braPopup = global.kobold.inventory.createItemWithTooltip(bra);
        shirt = ctrl.bandit.watch.masturbateShirt();
        shirtPopup = global.kobold.inventory.createItemWithTooltip(shirt);
        boobs = ctrl.bandit.watch.masturbateBoobs();
        cock = ctrl.bandit.watch.masturbateCock();
        cunt = ctrl.bandit.watch.masturbateCunt();
        cuntVirgin = ctrl.bandit.watch.masturbateCuntVirgin();
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditMasturbate";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.watch.getWatchChoice() !== ForestBanditWatchChoiceEnum.MASTURBATE) {
            return "";
        }

        initClothing();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toIndulge.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.watch.getWatchChoice() !== ForestBanditWatchChoiceEnum.MASTURBATE) {
            return;
        }

        toIndulge.registerListeners();

        if (ctrl.bandit.watch.hasMadeMasturbateChoice()) {
            return;
        }

        registerIndulgeListener();
        registerDontListener();
    };

}

/**
 * Choices for the masturbate part of the bandit encounter.
 */
var ForestBanditMasturbateChoiceEnum = {
    INDULGE: 1,
    DONT: 2
};

// END scenes\Forest\bandit\watch\masturbate\ForestBanditMasturbate.js

// scenes\Forest\bandit\watch\steal\ForestBanditSteal.js

/**
 * Steal section of the bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditSteal(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var continueId = "banditStealContinueButton";
    
    var content = ` <div class="text">
                        Physically, you are no match for the group of kobolds.
                        If you got into a fight with them, they'd probably beat you up, 
                        stab you a couple of times and leave you for dead.
                        You <span class="bold">are</span>, however, 
                        just as cunning as they are, if not more so.
                        With all ten of their eyes focused on the aquamarine beauty,
                        you have no doubt that you'll be able to snatch at least <span class="italic">some</span> of their personal effects without getting noticed.
                    </div>
                    <div class="text">
                        You slip out from behind the tree trunk into the adjacent bush.
                        Luckily, the ruckus of the group's escalating gangrape covers up the rustling of leaves.
                        Still, you bide your time.
                        You don't want to attract <span class="bold">any</span> attention.
                        You play it cool.
                        You wait patiently, for any suspicion to fade.
                    </div>
                    <div class="text">
                        After a couple of minutes of remaining quiet and completely unnoticed,
                        you finally take a small step forwards, leaving your green concealment.
                        You are now in full view of the group of aroused males,
                        yet just as you suspected, just as you'd hoped, 
                        they are far too absorbed by their lust to notice you standing off to the side.
                    </div>
                    <div class="text">
                        You slowly inch forwards, 
                        step after step, 
                        until you reach the point where you're close enough to 
                        <span class="bold">smell</span> your fellow kobolds' excitement.
                        They've begun to properly <span class="bold">fuck</span> the dragoness,
                        their attention turned towards her waist, 
                        four of the males eagerly onlooking,
                        idly stroking themselves while waiting for the fifth to finish.
                    </div>
                    <div class="text">
                        Highly doubting the sexual stamina of a kobold bandit,
                        you realize you have to be quick about this.
                        In a bold and daring move,
                        you take a few quick large steps,
                        bringing you within grasping distance of the ringleader's discarded ${global.kobold.inventory.createItemWithTooltip(new LinenPants())}.
                        Taking a knee, you begin to quietly rifle through the pockets, finding them full of ill-gotten coins.
                    </div>
                    <div class="text">
                        Various moans, grunts, jests and jeers muffle the jangling noise of coins being stolen.
                        After taking two big handfuls and adding them to your own purse, you come across an odd find.
                        Apparently, the bandit saw fit to keep a key in his pocket, amidst all of the gold.
                        While you have no clue as to what lock this key may fit, you take it anyway.
                    </div>
                    <div class="text">
                        Judging by his increasingly rabid thrusts, 
                        you don't think the kobold fucking the dragoness'll last much longer.
                        While the urge to grab just a few more coins is tempting,
                        you decide not to push your luck any further than you already have.
                        You pussyfoot your way out of the clearing.
                        As you slip back into the treeline, 
                    </div>
                    <div class="text">
                        You pussyfoot your way out of the clearing.
                        As you slip back into the treeline, 
                        you cast one last glance at the dragoness,
                        just in time to witness the first of her rapists pull out,
                        a meagre dribble of spunk immediately oozing from her innards.
                        The spent kobold, 
                        in the middle of enjoying the afterglow of a good fuck,
                        gets pushed out of the way by one of his friends,
                        only to be replaced by the second in line.
                    </div>
                    <div class="text">
                        You head further into the woods, out of sight of the gang.
                        You wonder how long it'll take for them to realize they've just been robbed.
                        They probably won't even find out until after they've all finished having their turn with the dragoness.
                        By then, you'll be long gone.
                    </div>
                    <div class="text">
                        The sun is slowly setting.
                        Day is turning into night.
                        After counting your loot,
                        you decide to head home.
                    </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.bandit.watch.stealHasContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }
    
    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.stealContinue();
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditSteal";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.watch.getWatchChoice() !== ForestBanditWatchChoiceEnum.STEAL) {
            return "";
        }
        ctrl.bandit.watch.stealShared();
        var change = ctrl.bandit.watch.stealItems();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += content;
        part += change;
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.watch.getWatchChoice() !== ForestBanditWatchChoiceEnum.STEAL) {
            return;
        }

        // Register child listeners here

        if (ctrl.bandit.watch.stealHasContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\bandit\watch\steal\ForestBanditSteal.js

// scenes\Forest\bandit\watch\ForestBanditWatch.js

/**
 * Watch section of the bandit encounter.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditWatch(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toMasturbate = new ForestBanditMasturbate(controller);
    var toKeepWatching = new ForestBanditKeepWatching(controller);
    var toSteal = new ForestBanditSteal(controller);

    var masturbateId = "banditMasturbateButton";
    var watchId = "banditKeepWatchingButton";
    var stealId = "banditStealStuffButton";

    var content = ` <div class="text">
                        While you <span class="bold">would</span> like to help the poor dragoness, 
                        there is simply nothing you can do against a gang of five of your peers,
                        at least one of which is armed.
                        Besides, you don't want to risk losing your bag of freshly-acquired gold.
                        Still, you decide to stick around for just a little while longer, just to see the situation play out further.
                        There might still be something for you to gain out of this situation.
                    </div>
                    <div class="text">
                        As the dragoness' reluctant submission sinks in, the kobolds immediately begin to abuse her.
                        Tears flow down the woman's cheeks as her ${global.kobold.inventory.createItemWithTooltip(new RedDress())} gets lifted, 
                        one of her assailants sticking his head underneath the fine fabric,
                        the full extent of her long, scaly blue legs as well as
                        her fancy ${global.kobold.inventory.createItemWithTooltip(new SilkPanties())} now fully exposed to him.
                    </div>
                    <div class="text">
                        The nude kobold, meanwhile, has begun to grope one of the lady's breasts through her clothing.
                        He rubs, squeezes and caresses the sensitive mound to his heart's content, 
                        visibly discomforting the dragoness whilst openly stroking himself.
                        A single droplet of precum drips from the very tip of his cock down on top of her arm, 
                        the liquid instantly seeping into her dress, matting the once-pristine garment.
                    </div>
                    <div class="text">
                        The male in charge of her left leg seems to have taken an interest in the woman's foot, 
                        lowering his ${global.kobold.inventory.createItemWithTooltip(new LoinCloth())} whilst pressing his muzzle against her sole,
                        seemingly inhaling deeply as his eyes roll to the back of his head.
                        He intently begins to rub the slit between his legs with two of his fingers, 
                        the tapered tip of his member already protruding from it, 
                        growing more erect with every passing second.
                    </div>
                    <div class="text">
                        The ringleader decides it's time to move things along.
                        Out of the sleeve of his ${global.kobold.inventory.createItemWithTooltip(new PaddedLeatherShirt())},
                        he produces a second weapon: a sharp ${global.kobold.inventory.createItemWithTooltip(new Dagger())}.
                        He passes the pointy blade to the only member of the gang that hasn't gotten naked yet.
                        After the two exchange a mutual nod, the now-armed kobold climbs atop the woman's body,
                        straddling her chest, pressing the dagger against her bosom.
                    </div>
                    <div class="text">
                        For a moment, you feared he'd sunk the blade directly into her very heart.
                        A wave of relief washes over you as you realize he was simply cutting apart her clothing, 
                        ripping it with one clawed hand whilst slicing it with the dagger, 
                        the ${global.kobold.inventory.createItemWithTooltip(new RedDress())} rapidly reduced to mere shreds,
                        more and more of dragoness' enticing, curvy body coming into view with every scrap of fabric removed.
                    </div>
                    <div class="text">
                        Her ${global.kobold.inventory.createItemWithTooltip(new LinenBra())} gets torn along with her dress,
                        the woman's blue-scaled tits with dark black nipples now in plain sight.
                        The kobold playing with one of her breasts redoubles his efforts,
                        digging his fingers into her supple flesh, 
                        harshly pinching her bare nipple and rubbing his member against the side of her boob,
                        prodding into it ever so slightly whilst wiping off a fresh buildup of pre onto her scales, 
                        slickening them for further frottage.
                    </div>
                    <div class="text">
                        His underlings occupied, and the lady sufficiently pacified for the moment,
                        the leader removes his knife from her throat, sticking it in the ground right besides her head,
                        to remind her that things can still turn violent at his behest.
                        After warning her not to scream, he removes the loincloth from her mouth.
                        To your surprise, the dragoness complies, 
                        too intimidated and overwhelmed by the group molestation to utter even a single word.
                    </div>
                    <div class="text">
                        The ringleader removes his pants, tosses them aside and mounts the woman's face, 
                        taking hold of her small, nubby horns and using them as handlebars to lift her head upwards, 
                        forcing her muzzle into his groin. 
                        Grinding his cockslit against her lips and her nose,
                        he pre-emptively silences whatever plea she was preparing to utter.
                    </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the main content of this section.
     * @returns {string} The main content of this section.
     */
    function buildContent() {
        return content;
    }

    /**
     * Builds and returns the masturbate button.
     */
    function buildMasturbateChoice() {
        var disabled = ctrl.bandit.watch.hasMadeWatchChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getWatchChoice() == ForestBanditWatchChoiceEnum.MASTURBATE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${masturbateId}" ${disabled}>Masturbate</button>`;

        var tooltip = `Watching a group of kobolds having their way with a dragoness is rather arousing. Quietly let off some steam while watching from the shadows.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 250px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the keep watching button.
     */
    function buildKeepWatchingChoice() {
        var disabled = ctrl.bandit.watch.hasMadeWatchChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getWatchChoice() == ForestBanditWatchChoiceEnum.KEEPWATCHING;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${watchId}" ${disabled}>Keep watching</button>`;

        var tooltip = `Wait until the gang has finished up their business before making a move.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the steal stuff button.
     */
    function buildStealStuffChoice() {
        var disabled = ctrl.bandit.watch.hasMadeWatchChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.watch.getWatchChoice() == ForestBanditWatchChoiceEnum.STEAL;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${stealId}" ${disabled}>Steal stuff</button>`;

        var tooltip = `While the kobolds are occupied, you may be able to sneak up on them and steal some of their stuff. 
                       You <span class="italic">are</span> a thief, after all.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 220px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildMasturbateChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildKeepWatchingChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildStealStuffChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + masturbateId), $("#" + watchId), $("#" + stealId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the masturbate button.
     */
    function registerMasturbateListener() {
        $("#" + masturbateId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeWatchChoice(ForestBanditWatchChoiceEnum.MASTURBATE);
            loadNextPart(toMasturbate);
        });

    }

    /**
     * Adds a listener to the keep watching button.
     */
    function registerKeepWatchingListener() {
        $("#" + watchId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeWatchChoice(ForestBanditWatchChoiceEnum.KEEPWATCHING);
            loadNextPart(toKeepWatching);
        });
    }

    /**
     * Adds a listener to the steal button.
     */
    function registerStealStuffListener() {
        $("#" + stealId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.watch.makeWatchChoice(ForestBanditWatchChoiceEnum.STEAL);
            loadNextPart(toSteal);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditWatch";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bandit.getArrivalChoice() !== ForestBanditArrivalChoiceEnum.WATCH) {
            return "";
        }
        ctrl.bandit.watch.watchShared();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toMasturbate.build();
        part += toKeepWatching.build();
        part += toSteal.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bandit.getArrivalChoice() !== ForestBanditArrivalChoiceEnum.WATCH) {
            return;
        }

        toMasturbate.registerListeners();
        toKeepWatching.registerListeners();
        toSteal.registerListeners();

        if (ctrl.bandit.watch.hasMadeWatchChoice()) {
            return;
        }

        registerMasturbateListener();
        registerKeepWatchingListener();
        registerStealStuffListener();
    };
}

/**
 * Choices for the watching part of the bandit encounter.
 */
var ForestBanditWatchChoiceEnum = {
    MASTURBATE: 1,
    KEEPWATCHING: 2,
    STEAL: 3
};

// END scenes\Forest\bandit\watch\ForestBanditWatch.js

// scenes\Forest\bandit\watch\ForestBanditWatchController.js

/**
 * This class acts as a dedicated controller for the watch section of the bandit part of the Forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestBanditWatchController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;

    var watchId = "banditWatchChoice";
    var watchSharedId = "banditWatchShared";

    var keepWatchingId = "banditKeepWatchingChoice";

    var masturbateId = "banditMasturbateChoice";
    var masturbateBraId = "banditMasturbateBra";
    var masturbatePantsId = "banditMasturbatePants";
    var masturbateUnderwearId = "banditMasturbateUnderwear";
    var masturbateShirtId = "banditMasturbateShirt";
    var masturbateBoobsId = "banditMasturbateBoobs";
    var masturbateCuntId = "banditMasturbateCunt";
    var masturbateCuntVirginId = "banditMasturbateCuntVirgin";
    var masturbateCockId = "banditMasturbateCock";
    var masturbateGenderId = "banditMasturbateGender";

    var masturbateIndulgeId = "banditMasturbateIndulge";
    var masturbateIndulgeSharedId = "banditMasturbateIndulgeShared";
    var masturbateIndulgeChangeId = "banditMasturbateIndulgeChange";
    var masturbateIndulgeChangeTogetherId = "banditMasturbateIndulgeChangeTogether";

    var stealId = "banditStealChoice";
    var stealSharedId = "banditStealShared";
    var stealItemsId = "banditStealItems";

    var helpId = "banditHelpChoice";
    var helpSharedId = "banditHelpShared";
    var helpSource = "banditHelpSource";
    var helpStatusId = "banditHelpStatus";

    var scavengeId = "banditScavengeChoice";
    var scavengeSharedId = "banditScavengeShared";
    var scavengeSource = "banditScavengeSource";

    var scavengeContinueId = "banditScavengeContinueChoice";
    var scavengeContinueUnderwearId = "banditScavengeContinueUndies";
    var scavengeContinuePantsId = "banditScavengeContinuePants";
    var scavengeContinueChangeId = "banditScavengeContinueChange";

    var sloppySecondsId = "banditSloppySecondsChoice";
    var sloppySecondsSharedId = "banditSloppySecondsShared";
    var sloppySecondsStatusId = "banditSloppySecondsStatus";
    var sloppySecondsBraId = "banditSloppySecondsBra";
    var sloppySecondsPantsId = "banditSloppySecondsPants";
    var sloppySecondsUnderwearId = "banditSloppySecondsUnderwear";
    var sloppySecondsShirtId = "banditSloppySecondsShirt";
    var sloppySecondsBoobsId = "banditSloppySecondsBoobs";
    var sloppySecondsCuntId = "banditSloppySecondsCunt";
    var sloppySecondsCuntVirginId = "banditSloppySecondsCuntVirgin";
    var sloppySecondsCockId = "banditSloppySecondsCock";
    var sloppySecondsCockVirginId = "banditSloppySecondsCockVirgin";
    var sloppySecondsGenderId = "banditSloppySecondsGender";


    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Makes the watch choice.
     * @param {ForestBanditWatchChoiceEnum} choice The choice.
     */
    this.makeWatchChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[watchId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the watch choice from history and returns it.
     * May return undefined if the watch choice has not yet been set.
     * @returns {ForestBanditWatchChoiceEnum} The choice.
     */
    this.getWatchChoice = function () {
        var history = global.kobold.history.get(id);
        return history[watchId];
    };

    /**
     * Returns true if the watch choice has been made.
     */
    this.hasMadeWatchChoice = function () {
        return global.kobold.history.has(id, watchId);
    };

    /**
     * Makes the watch shared changes.
     */
    this.watchShared = function () {
        global.kobold.history.doIfNotSet(id, watchSharedId, true, function () {
            global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.DEEPLY_OBSERVED;
            global.kobold.history.shared.tenPawGang.footsie = TenPawGangFootsieEnum.OBSERVED;
        });
    };


    /**
     * Makes the keep watching choice.
     * @param {ForestBanditKeepWatchingChoiceEnum} choice The choice.
     */
    this.makeKeepWatchingChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[keepWatchingId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the keep watching choice from history and returns it.
     * May return undefined if the keep watching choice has not yet been set.
     * @returns {ForestBanditKeepWatchingChoiceEnum} The choice.
     */
    this.getKeepWatchingChoice = function () {
        var history = global.kobold.history.get(id);
        return history[keepWatchingId];
    };

    /**
     * Returns true if the keep watching choice has been made.
     */
    this.hasMadeKeepWatchingChoice = function () {
        return global.kobold.history.has(id, keepWatchingId);
    };


    /**
     * Makes the masturbate choice.
     * @param {ForestBanditMasturbateChoiceEnum} choice The choice.
     */
    this.makeMasturbateChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[masturbateId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the masturbate choice from history and returns it.
     * May return undefined if the masturbate choice has not yet been set.
     * @returns {ForestBanditMasturbateChoiceEnum} The choice.
     */
    this.getMasturbateChoice = function () {
        var history = global.kobold.history.get(id);
        return history[masturbateId];
    };

    /**
     * Returns true if the masturbate choice has been made.
     */
    this.hasMadeMasturbateChoice = function () {
        return global.kobold.history.has(id, masturbateId);
    };

    /**
     * Stores (if required) and retrieves the kobold's pants at masturbate-time.
     * @returns {Gear} The pants.
     */
    this.masturbatePants = function () {
        if (!global.kobold.history.has(id, masturbatePantsId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.PANTS)) {
                history[masturbatePantsId] = global.kobold.inventory.equipment.pants;
            } else {
                history[masturbatePantsId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[masturbatePantsId]);
    };

    /**
     * Stores (if required) and retrieves the kobold's underwear at masturbate-time.
     * @returns {Gear} The underwear.
     */
    this.masturbateUnderwear = function () {
        if (!global.kobold.history.has(id, masturbateUnderwearId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.UNDERWEAR)) {
                history[masturbateUnderwearId] = global.kobold.inventory.equipment.underwear;
            } else {
                history[masturbateUnderwearId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[masturbateUnderwearId]);
    };

    /**
     * Stores (if required) and retrieves the kobold's underwear at masturbate-time.
     * @returns {Gear} The shirt.
     */
    this.masturbateShirt = function () {
        if (!global.kobold.history.has(id, masturbateShirtId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.SHIRT)) {
                history[masturbateShirtId] = global.kobold.inventory.equipment.shirt;
            } else {
                history[masturbateShirtId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[masturbateShirtId]);
    };

    /**
     * Stores (if required) and retrieves the kobold's underwear at masturbate-time.
     * @returns {Gear} The bra.
     */
    this.masturbateBra = function () {
        if (!global.kobold.history.has(id, masturbateBraId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.BRA)) {
                history[masturbateBraId] = global.kobold.inventory.equipment.bra;
            } else {
                history[masturbateBraId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[masturbateBraId]);
    };

    /**
     * Stores (if required) and retrieves whether or not the kobold has boobs at masturbate-time.
     * @returns {boolean} Whether or not the kobold has boobs at masturbate-time.
     */
    this.masturbateBoobs = function () {
        if (!global.kobold.history.has(id, masturbateBoobsId)) {
            var history = global.kobold.history.get(id);
            history[masturbateBoobsId] = global.kobold.body.hasBreasts();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[masturbateBoobsId];
    };

    /**
     * Stores (if required) and retrieves whether or not the kobold has a cunt at masturbate-time.
     * @returns {boolean} Whether or not the kobold has a cunt at masturbate-time.
     */
    this.masturbateCunt = function () {
        if (!global.kobold.history.has(id, masturbateCuntId)) {
            var history = global.kobold.history.get(id);
            history[masturbateCuntId] = global.kobold.body.hasCunt();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[masturbateCuntId];
    };

    /**
     * Stores (if required) and retrieves whether or not the kobold's cunt is virginal at masturbate-time.
     * @returns {boolean} Whether or not the kobold's cunt is virginal at masturbate-time.
     */
    this.masturbateCuntVirgin = function () {
        if (!global.kobold.history.has(id, masturbateCuntVirginId)) {
            var history = global.kobold.history.get(id);
            history[masturbateCuntVirginId] = global.kobold.body.vaginalVirgin;
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[masturbateCuntVirginId];
    };

    /**
     * Stores (if required) and retrieves whether or not the kobold has a dick at masturbate-time.
     * @returns {boolean} Whether or not the kobold has a dick at masturbate-time.
     */
    this.masturbateCock = function () {
        if (!global.kobold.history.has(id, masturbateCockId)) {
            var history = global.kobold.history.get(id);
            history[masturbateCockId] = global.kobold.body.hasDick();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[masturbateCockId];
    };
    /**
     * Stores (if required) and retrieves the kobold's base gender at masturbate-time.
     * @returns {GenderEnum} The kobold's base gender at masturbate-time.
     */
    this.masturbateBaseGender = function () {
        if (!global.kobold.history.has(id, masturbateGenderId)) {
            var history = global.kobold.history.get(id);
            history[masturbateGenderId] = global.kobold.body.getBaseGender();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[masturbateGenderId];
    };


    /**
     * Makes the masturbate indulging choice.
     * @param {ForestBanditMasturbateIndulgeChoiceEnum} choice The choice.
     */
    this.makeMasturbateIndulgeChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[masturbateIndulgeId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the masturbate indulging choice from history and returns it.
     * May return undefined if the masturbate indulge choice has not yet been set.
     * @returns {ForestBanditMasturbateIndulgeChoiceEnum} The choice.
     */
    this.getMasturbateIndulgeChoice = function () {
        var history = global.kobold.history.get(id);
        return history[masturbateIndulgeId];
    };

    /**
     * Returns true if the masturbate indulging choice has been made.
     */
    this.hasMadeMasturbateIndulgeChoice = function () {
        return global.kobold.history.has(id, masturbateIndulgeId);
    };

    /**
     * Makes the masturbate indulge shared changes.
     */
    this.masturbateIndulgeShared = function () {
        global.kobold.history.doIfNotSet(id, masturbateIndulgeSharedId, true, function () {
            global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.DEEPLY_OBSERVED;
            if (that.getMasturbateChoice() === ForestBanditMasturbateChoiceEnum.INDULGE) {
                global.kobold.history.shared.tenPawGang.footsie = TenPawGangFootsieEnum.MASTURBATION_INDULGED;
            } else {
                global.kobold.history.shared.tenPawGang.footsie = TenPawGangFootsieEnum.MASTURBATION_DECLINED;
            }
        });
    };

    /**
     * Alters the kobold's stats if required.
     * @returns {string} HTML describing the change.
     */
    this.masturbateIndulgeTogetherChange = function () {
        if (!global.kobold.history.has(id, masturbateIndulgeChangeTogetherId)) {
            global.kobold.status.mental.increase(5, "Made a friend");
            global.kobold.status.exertion.increase(20, "Masturbation");
            var history = global.kobold.history.get(id);
            history[masturbateIndulgeChangeTogetherId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Made a friend", "Sanity", +5),
            new Change("Masturbation", "Sexual exhaustion", +20, true)
        ]);
    };

    /**
     * Alters the kobold's stats if required.
     * @returns {string} HTML describing the change.
     */
    this.masturbateIndulgeChange = function () {
        if (!global.kobold.history.has(id, masturbateIndulgeChangeId)) {
            global.kobold.status.exertion.increase(20, "Masturbation");
            var history = global.kobold.history.get(id);
            history[masturbateIndulgeChangeId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Masturbation", "Sexual exhaustion", +20, true)
        ]);
    };


    /**
     * Announces the steal continue button has been pressed.
     */
    this.stealContinue = function () {
        var history = global.kobold.history.get(id);
        history[stealId] = true;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns true if the steal continue button has been pressed.
     */
    this.stealHasContinued = function () {
        return global.kobold.history.has(id, stealId);
    };
    
    /**
     * Makes the steal shared changes.
     */
    this.stealShared = function () {
        global.kobold.history.doIfNotSet(id, stealSharedId, true, function () {
            global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.UNHELPED;
            global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.STOLEN_FROM;
        });
    };

    /**
     * Gives the kobold items if required.
     * @returns {string} HTML describing the change.
     */
    this.stealItems = function () {
        if (!global.kobold.history.has(id, stealItemsId)) {
            global.kobold.inventory.coins += 30;
            global.kobold.inventory.bag.addMisc(new BanditKey());
            var history = global.kobold.history.get(id);
            history[stealItemsId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Loot", "Gold", +30),
            new Change("Loot", "Bandit key", +1)
        ]);
    };


    /**
     * Sets the help section's source.
     * @param {ForestBanditHelpSourceEnum} choice The choice.
     */
    this.setHelpSource = function (source) {
        var history = global.kobold.history.get(id);
        history[helpSource] = source;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the source of the help section from history and returns it.
     * May return undefined if the source has not yet been set.
     * @returns {ForestBanditHelpSourceEnum} The source.
     */
    this.getHelpSource = function () {
        var history = global.kobold.history.get(id);
        return history[helpSource];
    };

    /**
     * Returns true if the help section has a source.
     */
    this.hasHelpSource = function () {
        return global.kobold.history.has(id, helpSource);
    };
    
    /**
     * Makes the help shared changes.
     */
    this.helpShared = function () {
        global.kobold.history.doIfNotSet(id, helpSharedId, true, function () {
            global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.HELPED;
        });
    };

    /**
     * Alters the kobold's stats if required.
     * @returns {string} HTML describing the change.
     */
    this.helpStatus = function () {
        if (!global.kobold.history.has(id, helpStatusId)) {
            global.kobold.status.stamina.decrease(5, "Helping the dragoness");
            var history = global.kobold.history.get(id);
            history[helpStatusId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Did your best", "Sanity", +5),
            new Change("Still feel bad", "Sanity", -5),
            new Change("Heavy lifting", "Stamina", -5)
        ]);
    };

    /**
     * Announces the help continue button has been pressed.
     */
    this.helpContinue = function () {
        var history = global.kobold.history.get(id);
        history[helpId] = true;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns true if the help continue button has been pressed.
     */
    this.helpHasContinued = function () {
        return global.kobold.history.has(id, helpId);
    };


    /**
     * Sets the scavenge section's source.
     * @param {ForestBanditScavengeSourceEnum} source The source.
     */
    this.setScavengeSource = function (source) {
        var history = global.kobold.history.get(id);
        history[scavengeSource] = source;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the source of the scavenge section from history and returns it.
     * May return undefined if the source has not yet been set.
     * @returns {ForestBanditScavengeSourceEnum} The source.
     */
    this.getScavengeSource = function () {
        var history = global.kobold.history.get(id);
        return history[scavengeSource];
    };

    /**
     * Returns true if the scavenge section has a source.
     */
    this.hasScavengeSource = function () {
        return global.kobold.history.has(id, scavengeSource);
    };

    /**
     * Makes the scavenge choice.
     * @param {ForestBanditScavengeChoiceEnum} choice The choice.
     */
    this.makeScavengeChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[scavengeId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the scavenge choice from history and returns it.
     * May return undefined if the scavenge choice has not yet been set.
     * @returns {ForestBanditScavengeChoiceEnum} The choice.
     */
    this.getScavengeChoice = function () {
        var history = global.kobold.history.get(id);
        return history[scavengeId];
    };

    /**
     * Returns true if the scavenge choice has been made.
     */
    this.hasMadeScavengeChoice = function () {
        return global.kobold.history.has(id, scavengeId);
    };
    
    /**
     * Makes the scavenge shared changes.
     */
    this.scavengeShared = function () {
        global.kobold.history.doIfNotSet(id, scavengeSharedId, true, function () {
            global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.UNHELPED;
        });
    };


    /**
     * Announces the scavenge continue button has been pressed.
     */
    this.scavengeContinue = function () {
        var history = global.kobold.history.get(id);
        history[scavengeContinueId] = true;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns true if the scavenge continue button has been pressed.
     */
    this.scavengeHasContinued = function () {
        return global.kobold.history.has(id, scavengeContinueId);
    };

    /**
     * Stores (if required) and retrieves the kobold's pants at the end of scavenging.
     * @returns {Gear} The pants.
     */
    this.scavengeContinuePants = function () {
        if (!global.kobold.history.has(id, scavengeContinuePantsId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.PANTS)) {
                history[scavengeContinuePantsId] = global.kobold.inventory.equipment.pants;
            } else {
                history[scavengeContinuePantsId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[scavengeContinuePantsId]);
    };

    /**
     * Stores (if required) and retrieves the kobold's underwear at the end of scavenging.
     * @returns {Gear} The underwear.
     */
    this.scavengeContinueUnderwear = function () {
        if (!global.kobold.history.has(id, scavengeContinueUnderwearId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.UNDERWEAR)) {
                history[scavengeContinueUnderwearId] = global.kobold.inventory.equipment.underwear;
            } else {
                history[scavengeContinueUnderwearId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[scavengeContinueUnderwearId]);
    };

    /**
     * Changes the kobold's underwear to panties if not done yet.
     */
    this.scavengeContinueChange = function () {
        if (!global.kobold.history.has(id, scavengeContinueChangeId)) {
            var panties = new SilkPanties();
            panties.stain("Cum and Mud");
            global.kobold.inventory.equipment.wear(panties);
            var history = global.kobold.history.get(id);
            history[scavengeContinueChangeId] = true;
            global.kobold.history.set(id, history);
        }
    };



    /**
     * Signals the continue button of the sloppy seconds scene has been pressed.
     */
    this.sloppySecondsContinue = function (choice) {
        var history = global.kobold.history.get(id);
        history[sloppySecondsId] = true;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns true if the sloppy seconds continue button has been pressed.
     */
    this.sloppySecondsHasContinued = function () {
        return global.kobold.history.has(id, sloppySecondsId);
    };
    
    /**
     * Makes the sloppy seconds shared changes.
     */
    this.sloppySecondsShared = function () {
        global.kobold.history.doIfNotSet(id, sloppySecondsSharedId, true, function () {
            global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.SLOPPY_SECONDS;
        });
    };

    /**
     * Stores (if required) and retrieves the kobold's pants at sloppy seconds-time.
     * @returns {Gear} The pants.
     */
    this.sloppySecondsPants = function () {
        if (!global.kobold.history.has(id, sloppySecondsPantsId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.PANTS)) {
                history[sloppySecondsPantsId] = global.kobold.inventory.equipment.pants;
            } else {
                history[sloppySecondsPantsId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[sloppySecondsPantsId]);
    };

    /**
     * Stores (if required) and retrieves the kobold's underwear at sloppy seconds-time.
     * @returns {Gear} The underwear.
     */
    this.sloppySecondsUnderwear = function () {
        if (!global.kobold.history.has(id, sloppySecondsUnderwearId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.UNDERWEAR)) {
                history[sloppySecondsUnderwearId] = global.kobold.inventory.equipment.underwear;
            } else {
                history[sloppySecondsUnderwearId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[sloppySecondsUnderwearId]);
    };

    /**
     * Stores (if required) and retrieves the kobold's underwear at sloppy seconds-time.
     * @returns {Gear} The shirt.
     */
    this.sloppySecondsShirt = function () {
        if (!global.kobold.history.has(id, sloppySecondsShirtId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.SHIRT)) {
                history[sloppySecondsShirtId] = global.kobold.inventory.equipment.shirt;
            } else {
                history[sloppySecondsShirtId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[sloppySecondsShirtId]);
    };

    /**
     * Stores (if required) and retrieves the kobold's underwear at sloppy seconds-time.
     * @returns {Gear} The bra.
     */
    this.sloppySecondsBra = function () {
        if (!global.kobold.history.has(id, sloppySecondsBraId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.BRA)) {
                history[sloppySecondsBraId] = global.kobold.inventory.equipment.bra;
            } else {
                history[sloppySecondsBraId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[sloppySecondsBraId]);
    };

    /**
     * Stores (if required) and retrieves whether or not the kobold has boobs at sloppy seconds-time.
     * @returns {boolean} Whether or not the kobold has boobs at sloppy seconds-time.
     */
    this.sloppySecondsBoobs = function () {
        if (!global.kobold.history.has(id, sloppySecondsBoobsId)) {
            var history = global.kobold.history.get(id);
            history[sloppySecondsBoobsId] = global.kobold.body.hasBreasts();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[sloppySecondsBoobsId];
    };

    /**
     * Stores (if required) and retrieves whether or not the kobold has a cunt at sloppy seconds-time.
     * @returns {boolean} Whether or not the kobold has a cunt at sloppy seconds-time.
     */
    this.sloppySecondsCunt = function () {
        if (!global.kobold.history.has(id, sloppySecondsCuntId)) {
            var history = global.kobold.history.get(id);
            history[sloppySecondsCuntId] = global.kobold.body.hasCunt();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[sloppySecondsCuntId];
    };

    /**
     * Stores (if required) and retrieves whether or not the kobold's cunt is virginal at sloppy seconds-time.
     * @returns {boolean} Whether or not the kobold's cunt is virginal at sloppy seconds-time.
     */
    this.sloppySecondsCuntVirgin = function () {
        if (!global.kobold.history.has(id, sloppySecondsCuntVirginId)) {
            var history = global.kobold.history.get(id);
            history[sloppySecondsCuntVirginId] = global.kobold.body.vaginalVirgin;
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[sloppySecondsCuntVirginId];
    };

    /**
     * Stores (if required) and retrieves whether or not the kobold has a dick at sloppy seconds-time.
     * @returns {boolean} Whether or not the kobold has a dick at sloppy seconds-time.
     */
    this.sloppySecondsCock = function () {
        if (!global.kobold.history.has(id, sloppySecondsCockId)) {
            var history = global.kobold.history.get(id);
            history[sloppySecondsCockId] = global.kobold.body.hasDick();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[sloppySecondsCockId];
    };

    /**
     * Stores (if required) and retrieves whether or not the kobold's cock (giving) is virginal at sloppy seconds-time.
     * @returns {boolean} Whether or not the kobold's cock is virginal at sloppy seconds-time.
     */
    this.sloppySecondsCockVirgin = function () {
        if (!global.kobold.history.has(id, sloppySecondsCockVirginId)) {
            var history = global.kobold.history.get(id);
            history[sloppySecondsCockVirginId] = global.kobold.body.dickVirgin;
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[sloppySecondsCockVirginId];
    };

    /**
     * Stores (if required) and retrieves the kobold's base gender at sloppy seconds-time.
     * @returns {GenderEnum} The kobold's base gender at sloppy seconds-time.
     */
    this.sloppySecondsBaseGender = function () {
        if (!global.kobold.history.has(id, sloppySecondsGenderId)) {
            var history = global.kobold.history.get(id);
            history[sloppySecondsGenderId] = global.kobold.body.getBaseGender();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[sloppySecondsGenderId];
    };

    /**
     * Alters the kobold's stats if required.
     * @returns {string} HTML describing the change.
     */
    this.sloppySecondsStatus = function () {
        if (that.sloppySecondsCock()) {
            if (!global.kobold.history.has(id, sloppySecondsStatusId)) {
                global.kobold.status.exertion.increase(25, "Dragoness sloppy seconds");
                global.kobold.status.stamina.decrease(15, "Fucking a dragoness");
                global.kobold.body.dickVirgin = false;
                var history = global.kobold.history.get(id);
                history[sloppySecondsStatusId] = true;
                global.kobold.history.set(id, history);
            }
            return global.ui.change.buildNotification([
                new Change("Sloppy seconds", "Sexual exhaustion", +25, true),
                new Change("Intense fucking", "Stamina", -15)
            ]);
        } else {
            if (!global.kobold.history.has(id, sloppySecondsStatusId)) {
                global.kobold.status.exertion.increase(25, "Bisexual dragoness cleaning");
                var history = global.kobold.history.get(id);
                history[sloppySecondsStatusId] = true;
                global.kobold.history.set(id, history);
            }
            return global.ui.change.buildNotification([
                new Change("Sloppy seconds", "Sexual exhaustion", +25, true)
            ]);
        }


    };
}

// END scenes\Forest\bandit\watch\ForestBanditWatchController.js

// scenes\Forest\bandit\arrival\ForestBanditArrival.js

/**
 * The first section of the bandit part of the Forest scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestBanditArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toReward = new ForestBanditReward(ctrl);
    var toWatch = new ForestBanditWatch(ctrl);
    var toIntervene = new ForestBanditIntervene(ctrl, toReward);
    var toRock = new ForestBanditRock(ctrl, toIntervene, toReward);

    var rockId = "banditRockButton";
    var interveneId = "banditInterveneButton";
    var watchId = "banditWatchButton";

    var contentFromWoods = `<div class="text">
                                You keep moving, yet stick to the woodland regions you're familiar with,
                                which inadvertently leads you just a little bit closer to your own home.
                                If anyone was following you, they've now most likely been thrown off the trail,
                                seeing as your decision lead you to take a sharp turn back towards the eastern road,
                                as opposed to continuing on in a straight line deeper into the woods.
                            </div>
                            <div class="text">
                                You go at a slower pace than before, beginning to grow more than a little tired from all the constant running and walking.
                                You figure the soldiers <span class="bold">must</span> have given up by now.
                                If they even noticed you to begin with, that is.
                                Which you're reasonably sure that they didn't.
                            </div>
                            <div class="text">
                                Overall, you end up enjoying a nice little walk through the forest, 
                                which gives you plenty of time to think a few things through.
                                Your legs are very grateful for the lower tempo you've attained.
                                Like winding down after a stressful workout, you start to feel oddly <span class="italic">satisfied</span>.
                                You've successfully pulled off one of the most exhilarating heists of your life,
                                and you may or may not have put yourself on the path of a miraculous recovery from crippling debt.
                            </div>
                            <div class="text">
                                Just as you were starting to feel wholesomely relaxed,
                                you hear a commotion coming from a small clearing up ahead.
                                At first it sounds like arguing, some kind of heated debate between a woman and a couple of guys.
                                Then, a single, loud, feminine scream for help echoes throughout the forest, 
                                although it appears to be forcibly silenced about halfway through.
                                In place of the debate comes the racket of a struggle.
                            </div>`;

    var contentFromHome = `<div class="text">
                                Your decision to investigate the source of the noise that interrupted your trip home
                                leads you back across the road, and a little ways into the forest.
                                The initial scream is followed by a plethora of panicked shouts for help.
                                Spurred on by the thought of rescuing a damsel in distress, you hastily make your way 
                                through the woods.
                            </div>
                            <div class="text">
                                The shrieks get louder as you approach, until finally the very last one gets cut off about halfway through.
                                Muffled, although you know not by <span class="italic">what</span>.
                                You double your pace, practically sprinting forwards as fast as you can, only to
                                stop dead in your tracks as you become aware of the sounds of a struggle coming from up ahead.
                            </div>`;

    var contentEnd = `      <div class="text">
                                Alarmed, you lower your posture and slowly sneak forwards.
                                Silent like a mouse, you cautiously approach the clearing.
                                Getting closer, you hide behind a young tree, which conveniently has a dense bush right besides it.
                                The shrubbery concealing your body, 
                                you can freely peek out from behind the trunk of the sapling to observe the situation unfolding before you.
                            </div>
                            <div class="text">
                                <a href="scenes/Forest/bandit/arrival/intro-full.png" target="_blank">
                                    <img class="nsfw" src="scenes/Forest/bandit/arrival/intro.png" title="Made by Sneaky">
                                </a>
                            </div>
                            <div class="text">
                                An all-male band of five scantily-dressed kobolds is wrestling an anthromorphic dragoness nearly twice their size, 
                                in what is no doubt a stickup gone wrong.
                                You're not sure how, but they've managed to knock the woman off her legs.
                                She's on her back on the leafy forest floor, visibly struggling to pry the smaller scalies off of her.
                                Her mouth is gagged with what appears to be the ${global.kobold.inventory.createItemWithTooltip(new LoinCloth())} of one of the males, 
                                which explains her sudden silence, as well as the nudity of one of them.
                            </div>
                            <div class="text">
                                Two of the kobolds are trying their hardest to subdue the dragoness' 
                                ${global.kobold.inventory.createItemWithTooltip(new RedDress())}-constrained legs, 
                                the powerful woman's tail wildly flailing about, lashing at them from underneath her frame.
                                The naked one is, together with one of the clothed kobolds, currently straining to keep the woman's arms in check.
                                The fifth and final kobold, who seems to be just a little bit better dressed than his peers, 
                                is pinning the lady's head to the ground with a single hand, 
                                whilst simultaneously keeping the loincloth firmly secured in her muzzle with the other.
                                You figure this dapper one is the ringleader of the bunch.
                            </div>
                            <div class="text">
                                While you don't immediately recognize them, you imagine the group is one of the many gangs that have made this forest their home.
                                That means the woman is probably a traveller or a local out for a stroll in the woods.
                                Maybe she was picking flowers or having a picnic or something when they confronted her and demanded who knows what.
                                She refused, a struggle ensued, and then you arrived to spy on the ordeal.
                                That's what you make of this situation, at least.
                            </div>
                            <div class="text">
                                Just as it seems like the dragoness is about to turn the tide against her assailants, 
                                the most fashionable of the kobolds decides it's time to stop playing around.
                                He pulls a ${global.kobold.inventory.createItemWithTooltip(new SmallKnife())} 
                                out of his ${global.kobold.inventory.createItemWithTooltip(new LinenPants())},
                                and threateningly presses the blade against her exposed, blue-scaled throat.
                            </div>
                            <div class="text">
                                The woman instantly freezes, her protests dying out in the blink of an eye.
                                The hateful glare on her face slowly turns into a shocked, defeated, almost sad expression.
                                The kobolds grin wickedly as the dragoness' limbs stop struggling against their collective grasp, 
                                one of them already letting go of his assigned body part 
                                to slide his hand underneath his own ${global.kobold.inventory.createItemWithTooltip(new KoboldUnderwear())},
                                a clear indication of the group's intentions.
                            </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the main content of this section.
     * @returns {string} The main content of this section.
     */
    function buildContent() {
        var content;
        var src = ctrl.bandit.getSource();
        if (src === ForestBanditSourceEnum.WOODS) {
            content = contentFromWoods;
        } else {
            content = contentFromHome;
        }
        content += contentEnd;
        return content;
    }

    /**
     * Builds and returns the throw rock choice button.
     */
    function buildRockChoice() {
        var hasRock = ctrl.bandit.arrivalHasRock();
        var disabled = (ctrl.bandit.hasMadeArrivalChoice() || !hasRock) ? "disabled" : "";
        var isSelected = ctrl.bandit.getArrivalChoice() == ForestBanditArrivalChoiceEnum.ROCK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${rockId}" ${disabled}>Throw rock</button>`;

        var tooltip;
        if (hasRock) {
            tooltip = `You're outnumbered. Throw your rock from the shadows to knock at least one of them out.`;
        } else {
            tooltip = `If you had a rock, you could throw it at them.`;
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 165px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the intervene choice button.
     */
    function buildInterveneChoice() {
        var disabled = ctrl.bandit.hasMadeArrivalChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.getArrivalChoice() == ForestBanditArrivalChoiceEnum.INTERVENE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${interveneId}" ${disabled}>Intervene</button>`;

        var tooltip = `You can't let them do this. Step out of the shadows and approach the gang.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the watch choice button.
     */
    function buildWatchChoice() {
        var disabled = ctrl.bandit.hasMadeArrivalChoice() ? "disabled" : "";
        var isSelected = ctrl.bandit.getArrivalChoice() == ForestBanditArrivalChoiceEnum.WATCH;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${watchId}" ${disabled}>Watch</button>`;

        var tooltip = `It's best not to get involved. Stay hidden and watch.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildRockChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildInterveneChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildWatchChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + rockId), $("#" + interveneId), $("#" + watchId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the rock button.
     */
    function registerRockListener() {
        $("#" + rockId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.makeArrivalChoice(ForestBanditArrivalChoiceEnum.ROCK);
            loadNextPart(toRock);
        });
    }

    /**
     * Adds a listener to the intervene button.
     */
    function registerInterveneListener() {
        $("#" + interveneId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.intervene.setInterveneSource(ForestBanditInterveneSourceEnum.APPROACH);
            ctrl.bandit.makeArrivalChoice(ForestBanditArrivalChoiceEnum.INTERVENE);
            loadNextPart(toIntervene);
        });
    }

    /**
     * Adds a listener to the watch button.
     */
    function registerWatchListener() {
        $("#" + watchId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bandit.makeArrivalChoice(ForestBanditArrivalChoiceEnum.WATCH);
            loadNextPart(toWatch);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "banditArrival";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var change = "";
        var src = ctrl.bandit.getSource();
        if (src === ForestBanditSourceEnum.WOODS) {
            change = ctrl.bandit.arrivalStatusChange();
        }

        ctrl.bgm.fadeLoop();
        ctrl.bandit.introSound.playSettingsDependent();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `</div>`;
        part += toWatch.build();
        if (ctrl.bandit.getArrivalChoice() === ForestBanditArrivalChoiceEnum.INTERVENE) { // Prevent building twice.
            part += toIntervene.build();
        }
        part += toRock.build();
        part += toReward.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        toWatch.registerListeners();
        toIntervene.registerListeners();
        toRock.registerListeners();
        toReward.registerListeners();

        if (ctrl.bandit.hasMadeArrivalChoice()) {
            return;
        }

        registerRockListener();
        registerInterveneListener();
        registerWatchListener();

    };
}

/**
 * Choices for the arrival of the bandit encounter.
 */
var ForestBanditArrivalChoiceEnum = {
    ROCK: 1,
    INTERVENE: 2,
    WATCH: 3
};

// END scenes\Forest\bandit\arrival\ForestBanditArrival.js

// scenes\Forest\bandit\ForestBanditController.js

/**
 * This class acts as a dedicated controller for the bandit part of the Forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestBanditController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    var sourceId = "banditSource";

    var arrivalId = "banditArrivalChoice";
    var arrivalWeapon = "banditArrivalWeapon";
    var arrivalStatus = "banditArrivalStatus";

    var rockId = "banditRockChoice";

    var rewardSource = "banditRewardSource";
    var rewardGoldId = "banditRewardGold";
    var rewardId = "banditRewardChoice";
    var rewardNameId = "banditRewardName";
    var rewardChangeId = "banditRewardChange";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.watch = new ForestBanditWatchController(sceneId);
    this.intervene = new ForestBanditInterveneController(sceneId);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Sets the source scene.
     * @param {ForestBanditSourceEnum} source The source scene.
     */
    this.setSource = function (source) {
        var history = global.kobold.history.get(id);
        history[sourceId] = source;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the source scene from history and returns it.
     * May return undefined if the source scene has not yet been set.
     * @returns {ForestBanditSourceEnum} The source scene.
     */
    this.getSource = function () {
        var history = global.kobold.history.get(id);
        return history[sourceId];
    };

    /**
     * Returns true if the source scene has been set.
     */
    this.hasSource = function () {
        return global.kobold.history.has(id, sourceId);
    };


    /**
     * Makes the arrival choice.
     * @param {ForestBanditArrivalChoiceEnum} choice The choice.
     */
    this.makeArrivalChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[arrivalId] = choice;
        global.kobold.history.set(id, history);
        if (choice == ForestBanditArrivalChoiceEnum.ROCK) {
            global.kobold.inventory.equipment.weapon = null;
        }
    };

    /**
     * Fetches the arrival choice from history and returns it.
     * May return undefined if the arrival choice has not yet been set.
     */
    this.getArrivalChoice = function () {
        var history = global.kobold.history.get(id);
        return history[arrivalId];
    };

    /**
     * Returns true if the arrival choice has been made.
     */
    this.hasMadeArrivalChoice = function () {
        return global.kobold.history.has(id, arrivalId);
    };

    /**
     * Increases the kobold's sanity and stamina if required.
     * @returns {string} HTML describing the change.
     */
    this.arrivalStatusChange = function () {
        if (!global.kobold.history.has(id, arrivalStatus)) {
            global.kobold.status.mental.increase(5, "Thought things over");
            global.kobold.status.stamina.increase(5, "Relaxing walk");
            var history = global.kobold.history.get(id);
            history[arrivalStatus] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Time to think", "Sanity", +5),
            new Change("Relaxing walk", "Stamina", +5)
        ]);
    };

    /**
     * Stores (if required) and retrieves the kobold's weapon at the time of traversing the path,
     * compares it to ItemEnum.ROCK and returns the result.
     * @returns {Gear} The weapon.
     */
    this.arrivalHasRock = function () {
        if (!global.kobold.history.has(id, arrivalWeapon)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.WEAPON)) {
                history[arrivalWeapon] = global.kobold.inventory.equipment.weapon.id;
            } else {
                history[arrivalWeapon] = 0;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[arrivalWeapon] === ItemEnum.ROCK;
    };


    /**
     * Makes the rock choice.
     * @param {ForestBanditRockChoiceEnum} choice The choice.
     */
    this.makeRockChoice = function (choice) {
        global.kobold.history.shared.tenPawGang.usedRock = true;
        if (choice === ForestBanditRockChoiceEnum.SCARYNOISE) {
            global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.SPOOKED;
        }
        var history = global.kobold.history.get(id);
        history[rockId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the arrival choice from history and returns it.
     * May return undefined if the arrival choice has not yet been set.
     */
    this.getRockChoice = function () {
        var history = global.kobold.history.get(id);
        return history[rockId];
    };

    /**
     * Returns true if the arrival choice has been made.
     */
    this.hasMadeRockChoice = function () {
        return global.kobold.history.has(id, rockId);
    };


    /**
     * Sets the source for the reward.
     * @param {ForestBanditRewardSourceEnum} source The source scene.
     */
    this.setRewardSource = function (source) {
        var history = global.kobold.history.get(id);
        history[rewardSource] = source;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the reward source from history and returns it.
     * May return undefined if the source has not yet been set.
     * @returns {ForestBanditRewardSourceEnum} The source scene.
     */
    this.getRewardSource = function () {
        var history = global.kobold.history.get(id);
        return history[rewardSource];
    };

    /**
     * Returns true if the reward source has been set.
     */
    this.hasRewardSource = function () {
        return global.kobold.history.has(id, rewardSource);
    };

    /**
     * Reward continue.
     */
    this.rewardContinue = function () {
        var history = global.kobold.history.get(id);
        history[rewardId] = true;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns true if the reward screen has continued.
     */
    this.hasRewardContinued = function () {
        return global.kobold.history.has(id, rewardId);
    };

    /**
     * The kobold's name at reward-time.
     * @returns {string} The kobold's name at reward-time.
     */
    this.rewardName = function () {
        var name = global.kobold.name;
        global.kobold.history.setIfNotSet(id, rewardNameId, name);
        return global.kobold.history.getKey(id, rewardNameId);
    };

    /**
     * Alters the kobold's inventory and stats and returns HTML describing the change.
     * @returns {string} HTML describing the change.
     */
    this.rewardScary = function () {
        global.kobold.history.doIfNotSet(id, rewardChangeId, true, function () {
            global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.SAVED;
            global.kobold.history.shared.dragonessPrincess.savedBy = DragonessPrincessSavedBy.SCARY_NOISE;
            global.kobold.status.mental.increase(50, "Saved the dragoness");
            global.kobold.inventory.bag.addMisc(new GoldenNeedle());
            global.kobold.inventory.equipment.weapon = new Rock();
        });
        return global.ui.change.buildNotification([
            new Change("Saved a princess", "Sanity", +50),
            new Change("Reward", global.kobold.inventory.createItemWithTooltip(new GoldenNeedle(), true), +1),
            new Change("Recovered rock", global.kobold.inventory.createItemWithTooltip(new Rock(), true), +1)
        ]);
    };
    
    /**
     * Alters the kobold's inventory and stats and returns HTML describing the change.
     * @returns {string} HTML describing the change.
     */
    function rewardPayRock() {
        var gold = rewardCoins();
        global.kobold.history.doIfNotSet(id, rewardChangeId, true, function () {
            global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.SAVED;
            global.kobold.history.shared.dragonessPrincess.savedBy = DragonessPrincessSavedBy.PAYING_GANG;
            global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.PAID;
            global.kobold.status.mental.increase(50, "Saved the dragoness");
            global.kobold.inventory.bag.addMisc(new GoldenNeedle());
            global.kobold.inventory.equipment.weapon = new Rock();
            global.kobold.inventory.coins = 0;
        });
        return global.ui.change.buildNotification([
            new Change("Saved a princess", "Sanity", +50),
            new Change("Reward", global.kobold.inventory.createItemWithTooltip(new GoldenNeedle(), true), +1),
            new Change("Recovered rock", global.kobold.inventory.createItemWithTooltip(new Rock(), true), +1),
            new Change("Paid", "Gold", -gold)
        ]);
    };
    
    /**
     * Alters the kobold's inventory and stats and returns HTML describing the change.
     * @returns {string} HTML describing the change.
     */
    function rewardPayNoRock() {
        var gold = rewardCoins();
        global.kobold.history.doIfNotSet(id, rewardChangeId, true, function () {
            global.kobold.history.shared.dragonessPrincess.relation = DragonessPrincessEnum.SAVED;
            global.kobold.history.shared.dragonessPrincess.savedBy = DragonessPrincessSavedBy.PAYING_GANG;
            global.kobold.history.shared.tenPawGang.relation = TenPawGangEnum.PAID;
            global.kobold.status.mental.increase(50, "Saved the dragoness");
            global.kobold.inventory.bag.addMisc(new GoldenNeedle());
            global.kobold.inventory.coins = 0;
        });
        return global.ui.change.buildNotification([
            new Change("Saved a princess", "Sanity", +50),
            new Change("Reward", global.kobold.inventory.createItemWithTooltip(new GoldenNeedle(), true), +1),
            new Change("Paid", "Gold", -gold)
        ]);
    };
    
    function rewardCoins() {
        var gold = global.kobold.inventory.coins;
        global.kobold.history.setIfNotSet(id, rewardGoldId, gold);
        return global.kobold.history.getKey(id, rewardGoldId);
    }
    
    /**
     * Alters the kobold's inventory and stats and returns HTML describing the change.
     * @returns {string} HTML describing the change.
     */
    this.rewardPay = function () {
        var didRock = (that.getArrivalChoice() === ForestBanditArrivalChoiceEnum.ROCK);
        if(didRock) {
            return rewardPayRock();
        } else {
            return rewardPayNoRock();
        }
    };
    
    this.introSound = new AudioController(id, "res/audio/forest/bandit.mp3", 160);
}

// END scenes\Forest\bandit\ForestBanditController.js

// scenes\Forest\bandit\ForestBandit.js

/**
 * The bandit section of the Forest scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestBandit(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var arrival = new ForestBanditArrival(ctrl);
    
    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    
    // PUBLIC MEMBERS \\        this.public = "test";

    // The id of this segment.
    this.id = "forestBandit";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
    
    /**
     * Builds and returns this part.
     * @param {boolean} invisible If true, will build this part invisible.
     */
    this.build = function(invisible) {
        if (!ctrl.bandit.hasSource()) {
            return "";
        }
        var src = ctrl.bandit.getSource();
        var bandit = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if(src === ForestBanditSourceEnum.HOME) {
            bandit += `<div class="subtitle">Investigation</div>`;
        } else {
            bandit += `<div class="subtitle">Safe exploration</div>`;
        }
        bandit += arrival.build();
        bandit += `</div>`;
        return bandit;
    };
    
    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function() {
        if (!ctrl.bandit.hasSource()) {
            return;
        }
        arrival.registerListeners();
    };
}

/**
 * Sources for the bandit encounter.
 */
var ForestBanditSourceEnum = {
    WOODS: 1,
    HOME: 2
};

// END scenes\Forest\bandit\ForestBandit.js

// scenes\Forest\bridge\ForestBridgeController.js

/**
 * This class acts as a dedicated controller for the bridge part of the Forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestBridgeController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

}

// END scenes\Forest\bridge\ForestBridgeController.js

// scenes\Forest\bridge\ForestBridge.js

/**
 * The bridge section of the Forest scene.
 * @param {ForestController} controller The forest controller.
 */
function ForestBridge(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    //var pathSelection = new ForestIntroductionPathSelection(ctrl);
    //var itemSelection = new ForestIntroductionItemSelection(ctrl, pathSelection);
    //var recovering = new ForestIntroductionRecovering(ctrl, itemSelection);
    
    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    
    
    // PUBLIC MEMBERS \\        this.public = "test";

    // The id of this segment.
    this.id = "forestBridge";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
    
    /**
     * Builds and returns this part.
     */
    this.build = function() {
        if (!ctrl.intro.hasMadePathChoice()) {
            return "";
        }
        if (ctrl.intro.getPathChoice() !== ForestIntroductionPathChoiceEnum.BRIDGE) {
            return "";
        }
        var introduction = `<div id="${that.id}">`;
        introduction += `<div class="text">bridge</div>`
        introduction += `</div>`;
        return introduction;
    };
    
    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function() {
        if (!ctrl.intro.hasMadePathChoice()) {
            return;
        }
        if (ctrl.intro.getPathChoice() !== ForestIntroductionPathChoiceEnum.BRIDGE) {
            return;
        }
    };
}

// END scenes\Forest\bridge\ForestBridge.js

// scenes\Forest\home\ForestHomeController.js

/**
 * This class acts as a dedicated controller for the home part of the Forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestHomeController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;
    var pathChoiceId = "homePathChoice";
    var pathStatusChangeId = "homePathStatus";
    var pathWeaponId = "homePathWeapon";
    var hideContinueId = "homeHideContinue";
    var hideStatusChangeId = "homeHideStatus";
    var hideWeaponId = "homeHideWeapon";
    var ignoreContinueId = "homeIgnoreContinue";
    var ignoreStatusChangeId = "homeIgnoreStatus";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Makes the path choice.
     * @param {ForestHomePathChoiceEnum} choice The choice that was made.
     */
    this.makePathChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[pathChoiceId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns whether or not the path choice has been made.
     * @returns {boolean} Whether or not the path choice has been made.
     */
    this.hasMadePathChoice = function () {
        return global.kobold.history.has(id, pathChoiceId);
    };

    /**
     * Fetches the path choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     * @returns {ForestHomePathChoiceEnum} The choice that was made.
     */
    this.getPathChoice = function () {
        var history = global.kobold.history.get(id);
        return history[pathChoiceId];
    };

    /**
     * Reduces the kobold's sanity if required.
     * @returns {string} HTML describing the change.
     */
    this.pathStatusChange = function () {
        if (!global.kobold.history.has(id, pathStatusChangeId)) {
            global.kobold.status.mental.decrease(5, "Disconcerting scream");
            var history = global.kobold.history.get(id);
            history[pathStatusChangeId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Disconcerted", "Sanity", -5)
        ]);
    };

    /**
     * Stores (if required) and retrieves the kobold's weapon at the time of traversing the path.
     * @returns {Gear} The weapon.
     */
    this.pathWeapon = function () {
        if (!global.kobold.history.has(id, pathWeaponId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.WEAPON)) {
                history[pathWeaponId] = global.kobold.inventory.equipment.weapon;
            } else {
                history[pathWeaponId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[hideWeaponId]);
    };

    /**
     * Signals that the hide section of the scene has continued.
     */
    this.hideContinue = function () {
        var history = global.kobold.history.get(id);
        history[hideContinueId] = true;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns whether or not the hide section of this part of the scene has continued.
     * @returns {boolean} Whether or not the hide section of this part of the scene has continued.
     */
    this.hideHasContinued = function () {
        return global.kobold.history.has(id, hideContinueId);
    };

    /**
     * Stores (if required) and retrieves the kobold's weapon at the time of hiding in the forest.
     * @returns {Gear} The weapon.
     */
    this.hideWeapon = function () {
        if (!global.kobold.history.has(id, hideWeaponId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.WEAPON)) {
                history[hideWeaponId] = global.kobold.inventory.equipment.weapon;
            } else {
                history[hideWeaponId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[hideWeaponId]);
    };

    /**
     * Reduces the kobold's sanity if required.
     * @returns {string} HTML describing the change.
     */
    this.hideStatusChange = function () {
        if (!global.kobold.history.has(id, hideStatusChangeId)) {
            global.kobold.status.mental.decrease(10, "Spooky noises");
            var history = global.kobold.history.get(id);
            history[hideStatusChangeId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Spooked", "Sanity", -10)
        ]);
    };

    /**
     * Signals that the ignore section of the scene has continued.
     */
    this.ignoreContinue = function () {
        global.kobold.history.doIfNotSet(id, ignoreContinueId, true, function () {
            global.kobold.history.shared.home.status = HomeStatusEnum.SAVED;
        });
    };

    /**
     * Returns whether or not the ignore section of this part of the scene has continued.
     * @returns {boolean} Whether or not the ignore section of this part of the scene has continued.
     */
    this.ignoreHasContinued = function () {
        return global.kobold.history.has(id, ignoreContinueId);
    };

    /**
     * Reduces the kobold's sanity if required.
     * @returns {string} HTML describing the change.
     */
    this.ignoreStatusChange = function () {
        if (!global.kobold.history.has(id, ignoreStatusChangeId)) {
            global.kobold.status.mental.decrease(20, "Crippling guilt");
            var history = global.kobold.history.get(id);
            history[ignoreStatusChangeId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Guilt", "Sanity", -20)
        ]);
    };
}

// END scenes\Forest\home\ForestHomeController.js

// scenes\Forest\home\ForestHomeIgnore.js

/**
 * Ignore the scream on the path home.
 * @param {ForestController} controller The forest controller.
 */
function ForestHomeIgnore(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "homeIgnoreContinueButton";

    var content = ` <div class="text">
                        You decide to ignore the screams echoing throughout the forest.
                        This isn't the first time you've heard someone getting mugged, and you doubt it'll be your last.
                        It simply isn't any of your concern.
                        Even if you did decide to intervene, the entire thing could be an elaborate trap or a setup of some kind.
                        Angering a local gang by getting between them and their prey wouldn't be in your best interests, 
                        and neither would getting caught with two hundred gold dangling from your neck.
                    </div>
                    <div class="text">
                        You continue the journey towards your home.
                        The panicked shrieks turn to wailing before slowly fading from hearing, diminished by the increasing distance between the source and your ears.
                        You just got a second chance at life, and you have no reason to risk losing it all in favor of some stranger.
                        That's what you keep telling yourself, as self-inflicted guilt continues to pile up at the back of your head.
                        But what if it was someone caught in a bear trap?
                        Or a poor, innocent child lost in the woods, frightened and alone?
                    </div>
                    <div class="text">
                        You feel kind of sorry for not even checking if you <span class="bold">could</span> have helped.
                        Regardless, it's too late now.
                        You try, but fail, to get the entire ordeal out of your head.
                        Following the concealed pathway, you begin to near the decrepit old shack that you call home.
                    </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the main content of this section.
     * @returns {string} The main content of this section.
     */
    function buildContent() {
        return content;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var disabled = ctrl.home.ignoreHasContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }
    
    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.home.ignoreContinue();
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeIgnore";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.home.hasMadePathChoice()) {
            return "";
        }
        if(ctrl.home.getPathChoice() !== ForestHomePathChoiceEnum.CONTINUE) {
            return "";
        }
        
        var change = ctrl.home.ignoreStatusChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.home.getPathChoice() !== ForestHomePathChoiceEnum.CONTINUE) {
            return;
        }
        if(ctrl.home.ignoreHasContinued()) {
            return;
        }
        
        registerContinueListener();
    };
}

// END scenes\Forest\home\ForestHomeIgnore.js

// scenes\Forest\home\ForestHomeHide.js

/**
 * Hide from the scream on the path home.
 * @param {ForestController} controller The forest controller.
 */
function ForestHomeHide(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "continueButton";

    var content = ` <div class="text">
                        The shrill shriek sends shivers down your spine.
                        You jump into a nearby bush, cower in fear and try to stifle your panicked breathing.
                        You've heard terrible tales of phantoms, spectres and plenty more otherwordly creatures that are said to haunt these very woods.
                        Of course, you <span class="sneakytext bold">no longer</span> believe in ghosts, but that shout you just heard was <span class="bold">very</span> real.
                    </div>
                    <div class="text">
                        You remain concealed as the scream is followed by a bunch of wails and moans of a disconcerting nature.
                        Luckily enough, their source seems to be stationary. 
                        Or at least, they don't seem to be growing noticably louder or softer.
                        With ghosts, you can never be certain.
                        You aren't even sure whether these <span class="italic">shrieks</span> are physical, or just a mental manifestation of whatever phantom you've encountered.
                        You notice the sound appears to be echoing off the trees surrounding you. 
                        That poltergeist sure is good at creating lifelike audio-illusions!
                    </div>
                    <div class="text">
                        Suddenly, the latest high-pitched whine to ring through your ears abruptly ends, as though silenced.
                        Oddly enough, the silence proves to be <span class="bold">more</span> unsettling than the wailing ever was.
                        Did the ghost leave?
                        Did someone <span class="span">kill</span> it?
                        Is it even possible to <span class="span">kill</span> a ghost?
                        Maybe <span class="span">banish</span> would be a better word.
                    </div>
                    <div class="text">
                        You remain in the bush for nearly an entire hour more, 
                        <span class="italic">contemplating the mortality of ghosts</span>, 
                        and totally not being scared shitless by the thought of a phantasm taking control of your body and making you`;
    
    var contentEnd = `  What grown-up kobold would be afraid of such silly things? 
                        Not you, that's for sure!
                      </div>
                      <div class="text">
                        As the sun is slowly starting to set, you decide to emerge from the bush once more.
                        You <span class="bold">don't</span> want to be around here come nightfall.
                        It's about time to resume your journey home.
                      </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the weapon dependent part of this section.
     * @returns {string} The weapon dependent part of this section.
     */
    function buildWeaponDependentPart() {
        var weapon = ctrl.home.hideWeapon();
        if(weapon) {
            return ` hit yourself on the head with your ${global.kobold.inventory.createItemWithTooltip(weapon)} until your brains spill out.`;
        } else {
            return ` claw your own eyes out.`;
        }
    }

    /**
     * Builds and returns the main content of this section.
     * @returns {string} The main content of this section.
     */
    function buildContent() {
        return content + buildWeaponDependentPart() + contentEnd;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var disabled = ctrl.home.hideHasContinued();
        var choice = `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(ctrl.homeSceneName);
        });
    }

    /**
     * Adds a listener to the continue button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.home.hideContinue();
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeHide";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.home.hasMadePathChoice()) {
            return "";
        }
        if(ctrl.home.getPathChoice() !== ForestHomePathChoiceEnum.HIDE) {
            return "";
        }
        
        var change = ctrl.home.hideStatusChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.home.getPathChoice() !== ForestHomePathChoiceEnum.HIDE) {
            return;
        }
        if(ctrl.home.hideHasContinued()) {
            return;
        }
        
        registerContinueListener();
    };
}

// END scenes\Forest\home\ForestHomeHide.js

// scenes\Forest\home\ForestHomePath.js

/**
 * The path to home of the Forest scene.
 * @param {ForestController} controller The forest controller.
 * @param {ForestBandit} toBandit A part the kobold may go to next.
 * @param {ForestHomeHide} toHide A part the kobold may go to next.
 * @param {ForestHomeIgnore} toIgnore A part the kobold may go to next.
 */
function ForestHomePath(controller, toBandit, toHide, toIgnore) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    var bandit = toBandit;
    var hide = toHide;
    var ignore = toIgnore;

    var investigateId = "homeInvestigateButton";
    var hideId = "homeHideButton";
    var ignoreId = "homeIgnoreButton";

    var content = ` <div class="text">
                        You're not too far away from your humble homestead.
                        You decide to go there, since there's bound to be some kind of debt collector already waiting on your porch.
                        And if there isn't, at least you can rest your sore legs on your shabby bed for a bit.
                        In fact, your entire body feels kind of stiff, burdened by overexertion and the rollercoaster of emotions you've been through today.
                        It'd be nice to be able to wind down in a somewhat comfortable, familiar place, if just for long enough to collect yourself.
                    </div>
                    <div class="text">
                        You head for the nearby <span class="bold">eastern Varanar road</span>, which differs from the southern one, 
                        which is where you robbed the money transport.
                        Before crossing, you cautiously look from side to side, scouting for anyone that may cause trouble.
                        Luckily for you, this is one of the least travelled roads to the capital, and hence, it is relatively unplagued by highwaymen.
                        Still, you leave little up to chance.
                        You bolt across the path, moving from one side to the other in the flash of an eye.
                        Once again you check your surroundings.
                        Nobody. 
                        Phew.
                    </div>
                    <div class="text">
                        Walking alongside the road to maintain some modicum of stealth, you quickly come upon the small, well-hidden path that leads to your shack.
                        All of a sudden, a loud scream echoes across the forest.
                        It sounds like a woman in distress.
                    `;

    var contentEnd = `  fully expecting an ambush. 
                        It is only after you've fully prepared yourself for battle, both physically and mentally, 
                        that you come to the conclusion it was more of a distant wail instead of a nearby rallying cry.
                        You realize you're more than a little overstressed, teetering <span class="italic">on the edge</span>.
                        You slip out of your fighting pose.
                        That shout scared the <span class="bold">hell</span> out of you.
                    </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the weapon dependent part of this section.
     * @returns {string} The weapon dependent part of this section.
     */
    function buildWeaponDependentPart() {
        var weapon = ctrl.home.pathWeapon();
        if (weapon) {
            return ` You ready your weapon, `;
        } else {
            return ` You drop into a combat stance, `;
        }
    }

    /**
     * Builds and returns the main content of this section.
     * @returns {string} The main content of this section.
     */
    function buildContent() {
        return content + buildWeaponDependentPart() + contentEnd;
    }

    /**
     * Builds and returns the investigate choice button.
     */
    function buildInvestigateChoice() {
        var disabled = ctrl.home.hasMadePathChoice() ? "disabled" : "";
        var isSelected = ctrl.home.getPathChoice() == ForestHomePathChoiceEnum.INVESTIGATE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${investigateId}" ${disabled}>Investigate</button>`;

        var tooltip = `Investigate the shriek. Someone may be in trouble!`;
        var popupContainer = global.ui.makeToolTip(button, tooltip);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the hide choice button.
     */
    function buildHideChoice() {
        var disabled = ctrl.home.hasMadePathChoice() ? "disabled" : "";
        var isSelected = ctrl.home.getPathChoice() == ForestHomePathChoiceEnum.HIDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${hideId}" ${disabled}>Hide</button>`;

        var tooltip = `That wail was terrifying! Hide from whatever may have uttered it.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the ignore choice button.
     */
    function buildIgnoreChoice() {
        var disabled = ctrl.home.hasMadePathChoice() ? "disabled" : "";
        var isSelected = ctrl.home.getPathChoice() == ForestHomePathChoiceEnum.CONTINUE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${ignoreId}" ${disabled}>Continue</button>`;

        var tooltip = `Ignore the scream and continue on your way home.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildInvestigateChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildHideChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildIgnoreChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + investigateId), $("#" + hideId), $("#" + ignoreId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {ForestHomePathChoiceEnum} next Which part to load.
     */
    function loadNextPart(next) {
        var part;
        if (next === ForestHomePathChoiceEnum.INVESTIGATE) {
            part = bandit;
        }
        if (next === ForestHomePathChoiceEnum.HIDE) {
            part = hide;
        }
        if (next === ForestHomePathChoiceEnum.CONTINUE) {
            part = ignore;
        }
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the investigate button.
     */
    function registerInvestigateListener() {
        $("#" + investigateId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.home.makePathChoice(ForestHomePathChoiceEnum.INVESTIGATE);
            ctrl.bandit.setSource(ForestBanditSourceEnum.HOME);
            loadNextPart(ForestHomePathChoiceEnum.INVESTIGATE);
        });
    }

    /**
     * Adds a listener to the hide button.
     */
    function registerHideListener() {
        $("#" + hideId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.home.makePathChoice(ForestHomePathChoiceEnum.HIDE);
            loadNextPart(ForestHomePathChoiceEnum.HIDE);
        });
    }

    /**
     * Adds a listener to the ignore button.
     */
    function registerIgnoreListener() {
        $("#" + ignoreId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.home.makePathChoice(ForestHomePathChoiceEnum.CONTINUE);
            loadNextPart(ForestHomePathChoiceEnum.CONTINUE);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homePath";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var change = ctrl.home.pathStatusChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `</div>`;
        if (ctrl.home.getPathChoice() == ForestHomePathChoiceEnum.INVESTIGATE) {
            part += bandit.build();
        }
        //part += hide.build();   // Build by ForestHome
        //part += ignore.build(); // Build by ForestHome
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.home.getPathChoice() == ForestHomePathChoiceEnum.INVESTIGATE) {
            bandit.registerListeners();
        }

        if (ctrl.home.hasMadePathChoice()) {
            return;
        }
        registerInvestigateListener();
        registerHideListener();
        registerIgnoreListener();
    };
}

/**
 * Path choices to be made.
 */
var ForestHomePathChoiceEnum = {
    INVESTIGATE: 1,
    HIDE: 2,
    CONTINUE: 3
};

// END scenes\Forest\home\ForestHomePath.js

// scenes\Forest\home\ForestHome.js

/**
 * The home section of the Forest scene.
 * @param {ForestController} controller The forest controller.
 * @param {ForestBandit} bandit One of the parts the kobold can go next.
 */
function ForestHome(controller, bandit) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var hideInBush = new ForestHomeHide(ctrl);
    var ignoreScream = new ForestHomeIgnore(ctrl);
    var pathSelection = new ForestHomePath(ctrl, bandit, hideInBush, ignoreScream);
    
    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    
    
    // PUBLIC MEMBERS \\        this.public = "test";

    // The id of this segment.
    this.id = "forestHome";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
    
    /**
     * Builds and returns this part.
     * @param {boolean} invisible If true, will build this part invisible.
     */
    this.build = function(invisible) {
        if (!ctrl.intro.hasMadePathChoice()) {
            return "";
        }
        if (ctrl.intro.getPathChoice() !== ForestIntroductionPathChoiceEnum.HOME) {
            return "";
        }
        var home = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        home += `<div class="subtitle">Going home</div>`;
        home += pathSelection.build();
        home += hideInBush.build();
        home += ignoreScream.build();
        home += `</div>`;
        return home;
    };
    
    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function() {
        if (!ctrl.intro.hasMadePathChoice()) {
            return;
        }
        if (ctrl.intro.getPathChoice() !== ForestIntroductionPathChoiceEnum.HOME) {
            return;
        }
        pathSelection.registerListeners();
        hideInBush.registerListeners();
        ignoreScream.registerListeners();
    };
}

// END scenes\Forest\home\ForestHome.js

// scenes\Forest\hide\run\fail\beg\ForestWoodsRunFailBeg.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRunFailBeg(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            The wolf sets you back down on the ground, to give you a few seconds to catch your breath.
                            You [b]beg[/b] and you [b]plead[/b] and you [b]cry[/b] for mercy.
                            You didn't know you were stealing from the [i]mighty[/i] [b]Furskin Mountain-Crew[/b]!
                            If you'd seen a [b]strong[/b] and [b]imposing[/b] black wolf like him guarding the caravan,
                            you'd have thought twice before attempting to rob it!
                            Surely, a man so big has room for a bit of [b]kindness[/b] in his heart, right?!
                        [/p]
                        [p]
                            Your flattery seems to fall on deaf ears.
                            The wolf undoes his pants, his swollen sheath coming into view, the tapered tip of his member already protruding from it.
                            Panicking, your brain works overtime to think of [i]something[/i], [b]anything[/b] that could save you now.
                            If you can't get him to let you go, surely you can convince the canine to keep his hands off of you, for now, right?!
                            Then it hits you.
                            Yes.
                            This might work.
                        [/p]
                        [p]
                            You loudly cry for the wolf to [b]stop[/b]!
                            You're a [b]virgin[/b]!
                            Virgin slaves sell for [b]a lot[/b] more!!
                            If he takes you back to his leader [i]without[/i] touching you, surely he'll be rewarded with enough coin to buy at least [b]two[/b] whores.
                            He knows of the legendary Varanarian brothel district, right?!
                        [/p]
                        [p]
                            Your new ploy seems to be working.
                            The lupine stops to consider his actions.
                            With a loud sigh, he buttons up  his pants once more.
                            He stares into your eyes, hatred clearly visible on his face.
                            Baring his teeth, he growls loudly.
                            "[color=gblue]Better not be lying.[/color]" he speaks in an angry tone.
                            "[color=gblue]About the whores [b]or[/b] your virginity.[/color]"
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            The upset canine takes one last lusting glance at your naked form.
                            Letting go of your throat, he grabs you by your new collar.
                            And thus, you are forcibly dragged throughout the forest by your new [i]captor[/i],
                            back towards the caravan where today's adventure began.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.enslaveSceneName, ctrl.enslaveFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.runFail.get() !== ForestWoodsRunFailChoice.BEG) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.beg.change();
        part += buildChoice();
        if (!ctrl.enslaveReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.runFail.get() !== ForestWoodsRunFailChoice.BEG) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.enslaveReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Forest\hide\run\fail\beg\ForestWoodsRunFailBeg.js

// scenes\Forest\hide\run\fail\struggle\hate\ForestWoodsRunFailHate.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRunFailHate(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cunt = ctrl.misc.cunt();
        var toReturn = `[p]
                            Rapidly regaining your senses, you kick, squirm and [b]shout[/b] for help, all to no avail.
                            Tears flow liberally from the corners of your eyes as you pathetically [b]beg[/b] the wolf [b]not[/b] to do this.
                            All you get in response to your miserable pleading is a harsh [b]slap[/b] to your cheek.
                            "[color=gblue]Want me to [b]kill you[/b] instead?!
                            Be a good [b]bitch[/b] and [b]shut the fuck up[/b] before I [b]make[/b] you.[/color]"
                            Reduced to pitiful snivelling, you turn your face to the side,
                            staring off into the distance while the caravan guard begins to press his tip inside of you.
                        [/p]`;

        if (cunt) {
            toReturn += `[p]
                            In one firm [b]shove[/b], the lupine [b]rams[/b] himself inside of you.
                            You [b]scream[/b] out in pain, your virginal folds parted by ever-thickening wolf cock,
                            your innards filled to the very [b]brim[/b] with throbbing canine meat.
                            A small trickle of blood oozes from your slit, signifying the loss of your virginity.
                            You grab onto the bandit's broad shoulders, in an attempt to [b]pull[/b] yourself off of his member,
                            but, after a second slap, he forcibly [b]pushes[/b] you down onto his length,
                            sliding in the rest of his considerable girth, his fully formed knot coming to rest against your blood-trickling cunt.
                        [/p]
                        [p]
                            "[color=gblue]Tight little [b]slut[/b].[/color]" the wolf comments.
                            He adjusts your pose ever so slightly, keeping your upper body pressed against the tree whilst tugging your lower body towards him,
                            positioning you perfectly for his cock to piston in and out of your unlubed slit.
                            Then, after a second of idly relishing in the tightness of your velvety insides,
                            he starts to [b]fuck you properly[/b], [b]slamming[/b] his member in and out of your warm snatch,
                            his tapered tip [b]bashing[/b] against your cervix with every thrust.
                        [/p]
                        [p]
                            There is no [i]buildup[/i], no time to [i]get used[/i] to his rhythm, no [i]grace period[/i] for [b]inexperienced lizards[/b] such as yourself.
                            One moment, the wolf is sitting still, firmly lodged inside of you.
                            The other, he's [b]pounding[/b] away at your cunt, his hefty knot smacking against your folds,
                            the very tree you're leaning against beginning to tremble and shake.
                        [/p]
                        [p]
                            It hurts.
                            It hurts like [b]hell[/b].
                            Every hump feels like he's [b]tearing[/b] you in half with his oversized schlong,
                            stretching your slit far beyond what it was ever meant to take.
                            While not wanting to get painfully slapped [i]again[/i],
                            you resume pleading, whimpering and whining in a soft voice, trying but failing to arouse sympathy from the canine.
                        [/p]
                        [p]
                            It seems as though the wolf wasn't lying about caravan guard duty being [i]lonely[/i],
                            or maybe he just wasn't prepared for the tightness of a virginal kobold.
                            Mere [b]seconds[/b] after you [b]beg[/b] him not to put the knot in, he pulls out of your slit, as though finally heading your requests.
                            After a few more sloppy humps against your crotch-scales, the canine reaches his end,
                            squirting a considerable amount of thin, yet [b]extremely[/b] potent and heady spunk across the entire length of your body,
                            all the way from your belly button to the underside of your chin.
                        [/p]`;
        } else {
            toReturn += `[p]
                            In one firm [b]shove[/b], the lupine [b]rams[/b] himself inside of you.
                            You [b]scream[/b] out in pain, your tailhole stretched to its very limit by ever-thickening wolf cock.
                            You grab onto the bandit's broad shoulders, in an attempt to [b]pull[/b] yourself off of his member,
                            but, after a second slap, he forcibly [b]pushes[/b] you down onto his length,
                            sliding in the rest of his considerable girth, his fully formed knot coming to rest against your clenching hole.
                        [/p]
                        [p]
                            "[color=gblue]Tight little [b]slut[/b].[/color]" the wolf comments.
                            He adjusts your pose ever so slightly, keeping your upper body pressed against the tree whilst pulling your lower body towards him,
                            positioning you perfectly for his cock to piston in and out of your rear end.
                            Then, after a second of idly relishing the warmth of your velvety insides,
                            he starts to [b]fuck you properly[/b], [b]slamming[/b] his member in and out of your tight tailhole,
                            his tapered tip [b]oozing[/b] precum, slickening your insides to ease his rabid humping.
                        [/p]
                        [p]
                            There is no [i]buildup[/i], no time to [i]get used[/i] to his rhythm, no [i]grace period[/i] for [b]inexperienced lizards[/b] such as yourself.
                            One moment, the wolf is sitting still, firmly lodged inside of you.
                            The other, he's [b]pounding[/b] away at your ass, his hefty knot smacking against your hole,
                            the very tree you're leaning against beginning to tremble and shake.
                        [/p]
                        [p]
                            It hurts.
                            It hurts like [b]hell[/b].
                            Every hump feels like he's [b]tearing[/b] you in half with his oversized schlong,
                            stretching your rear far beyond what it was ever meant to take.
                            While not wanting to get painfully slapped [i]again[/i],
                            you resume pleading, whimpering and whining in a soft voice, trying but failing to arouse sympathy from the canine.
                        [/p]
                        [p]
                            It seems as though the wolf wasn't lying about caravan guard duty being [i]lonely[/i],
                            or maybe he just wasn't prepared for the tightness of a virginal kobold's butt.
                            Mere [b]seconds[/b] after you [b]beg[/b] him not to put the knot in, he pulls out of your ass, as though finally heading your requests.
                            After a few more sloppy thrusts against your crotch-scales,
                            his [b]far[/b] larger length grinding against your cockslit,
                            the canine reaches his end,
                            squirting a considerable amount of thin, yet [b]extremely[/b] potent and heady spunk across the entire length of your body,
                            all the way from your belly button to the underside of your chin.
                        [/p]`;
        }

        toReturn += `   [p]
                            Without even [b]bothering[/b] to wait for his orgasm to finish fully,
                            the lupine casually drops you to the ground,
                            another squirt of wolf cum splattering across your face as your ${cunt ? `` : `sore`} butt lands on the soft forest floor.
                            While he wrestles his semi-hard, spunk-leaking member into his underwear, you curl up into a ball and cry your eyes out.
                            "[color=gblue]Don't act like a little baby.[/color]" he speaks.
                            "[color=gblue]I didn't even [b]knot[/b] you.
                            Sun's going down.
                            That would have taken too long.[/color]"
                            The bandit pulls up his pants and rebuckles his belt.
                            "[color=gblue]But I'll have plenty of time once we get back to the caravan.[/color]"
                        [/p]
                        [p]
                            Chuckling cruelly at his own statement, the wolf pulls you to your feet by your iron collar.
                            With his new kobold slave in tow, he begins to proudly march back to the place where today's adventure began for you,
                            his scent clinging to your scales letting any canine you pass by know to whom you now belong.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    You are no longer a virgin.
                                [/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.enslaveSceneName, ctrl.enslaveFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.struggle.get() !== ForestWoodsRunFailStruggleChoice.HATE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.hate.change();
        part += buildChoice();
        if (!ctrl.enslaveReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.struggle.get() !== ForestWoodsRunFailStruggleChoice.HATE) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.enslaveReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Forest\hide\run\fail\struggle\hate\ForestWoodsRunFailHate.js

// scenes\Forest\hide\run\fail\struggle\enjoy\ForestWoodsRunFailEnjoy.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRunFailEnjoy(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.captureEnjoyed = true;
        var cunt = ctrl.misc.cunt();
        var toReturn = `[p]
                            While not quite exotic, you admittedly have had [i]thoughts[/i] of [i]sleeping[/i] with a canine before.
                            You can't say you're left completely unaroused by the hot, furred body pinning you against a tree,
                            not to mention the precum-leaking tapered tip pressing directly against your ${cunt ? `cunt` : `rear`}.
                            It all feels so [b]carnal[/b], so [b]bestial[/b], so [b]primal[/b].
                            ${cunt ? `Your slit slightly dampens` : `Your member begins to harden`} in anticipation of your wet dreams coming true.
                        [/p]`;

        if (cunt) {
            toReturn += `[p]
                            Then, in one firm [b]shove[/b], the lupine [b]rams[/b] himself inside of you.
                            Your mind goes [b]blank[/b] with pleasure, your folds parted by ever-thickening wolf cock,
                            your innards filled to the very [b]brim[/b] with throbbing canine meat.
                            A small trickle of blood oozes from your slit, signifying the loss of your virginity, but you don't care.
                            Moaning out like a cheap [b]whore[/b], you grab onto the bandit's broad shoulders, and [b]pull[/b] yourself further down onto his member,
                            sliding in the rest of his considerable girth, his fully formed knot coming to rest against your moist cunt.
                        [/p]
                        [p]
                            "[color=gblue]Eager little [b]slut[/b].[/color]" the wolf comments.
                            He adjusts your pose ever so slightly, keeping your upper body pressed against the tree whilst pulling your lower body towards him,
                            positioning you perfectly for his cock to piston in and out of your receptive slit.
                            Then, after a second of idly relishing in the tightness of your velvety insides,
                            he starts to [b]fuck you properly[/b], [b]slamming[/b] his member in and out of your warm snatch,
                            his tapered tip [b]bashing[/b] against your cervix with every thrust.
                        [/p]
                        [p]
                            There is no [i]buildup[/i], no time to [i]get used[/i] to his rhythm, no [i]grace period[/i] for [b]inexperienced lizards[/b] such as yourself.
                            One moment, the wolf is sitting still, firmly lodged inside of you.
                            The other, he's [b]pounding[/b] away at your cunt, his hefty knot smacking against your wet folds,
                            the very tree you're leaning against beginning to tremble and shake.
                        [/p]
                        [p]
                            You [b]scream[/b] out in pleasure.
                            Being used to nothing but your own fingers,
                            the lupine's incredible pace combined with the sheer [b]size[/b] of his member rapidly bring you to orgasm,
                            your cunt twitching and quivering, [b]massaging[/b] his shaft while your feminine fluids [b]ooze[/b] out onto the base of his length,
                            your body practically [b]begging[/b] him to [b]knot you[/b].
                        [/p]
                        [p]
                            It seems as though the wolf wasn't lying about caravan guard duty being [i]lonely[/i],
                            or maybe he just wasn't prepared for the tightness of a virginal kobold mid-climax.
                            Mere [b]seconds[/b] after you've come, he pulls out of your slit.
                            After a few more wet and sloppy humps against your crotch-scales, the canine, too, reaches his end,
                            squirting a considerable amount of thin, yet [b]extremely[/b] potent and heady spunk across the entire length of your body,
                            all the way from your belly button to the underside of your chin.
                        [/p]`;
        } else {
            toReturn += `[p]
                            Then, in one firm [b]shove[/b], the lupine [b]rams[/b] himself inside of you.
                            Your mind goes [b]blank[/b] with pleasure, your tailhole stretched to its very limit by ever-thickening wolf cock.
                            You just lost your virginity to a bandit in the middle of the forest, but you don't care.
                            Moaning out like a cheap [b]whore[/b], you grab onto the bandit's broad shoulders, and [b]pull[/b] yourself further down onto his member,
                            sliding in the rest of his considerable girth, his fully formed knot coming to rest against your clenching hole,
                            your erect member a clear indicator of just how much you [b]wanted[/b] this.
                        [/p]
                        [p]
                            "[color=gblue]Eager little [b]slut[/b].[/color]" the wolf comments.
                            He adjusts your pose ever so slightly, keeping your upper body pressed against the tree whilst pulling your lower body towards him,
                            positioning you perfectly for his cock to piston in and out of your rear end.
                            Then, after a second of idly relishing the warmth of your velvety insides,
                            he starts to [b]fuck you properly[/b], [b]slamming[/b] his member in and out of your tight tailhole,
                            his tapered tip [b]oozing[/b] precum, slickening your insides to ease his rabid humping.
                        [/p]
                        [p]
                            There is no [i]buildup[/i], no time to [i]get used[/i] to his rhythm, no [i]grace period[/i] for [b]inexperienced lizards[/b] such as yourself.
                            One moment, the wolf is sitting still, firmly lodged inside of you.
                            The other, he's [b]pounding[/b] away at your ass, his hefty knot smacking against your hole,
                            the very tree you're leaning against beginning to tremble and shake.
                        [/p]
                        [p]
                            You [b]scream[/b] out in pleasure.
                            Being used to nothing but, admittedly, your own fingers,
                            the lupine's incredible pace combined with the sheer [b]size[/b] of his member rapidly bring you to orgasm,
                            your cock twitching and quivering, shooting a meagre load all over the canine's clothed chest,
                            the scent of which only serves to [b]spur him on further[/b], your cum inadvertently [b]enticing[/b] him to [b]knot you[/b].
                        [/p]
                        [p]
                            It seems as though the wolf wasn't lying about caravan guard duty being [i]lonely[/i],
                            or maybe he just wasn't prepared for the tightness of a virginal kobold's butt.
                            Mere [b]seconds[/b] after you've come, he pulls out of your ass.
                            After a few more sloppy thrusts against your crotch-scales,
                            his [b]far[/b] larger length frotting against your receding cock,
                            the canine, too, reaches his end,
                            squirting a considerable amount of thin, yet [b]extremely[/b] potent and heady spunk across the entire length of your body,
                            all the way from your belly button to the underside of your chin.
                        [/p]`;
        }

        toReturn += `   [p]
                            Without even [b]bothering[/b] to wait for his orgasm to finish fully,
                            the lupine casually drops you to the ground,
                            another squirt of wolf cum splattering across your face as your ${cunt ? `` : `sore`} butt lands on the soft forest floor.
                            While he wrestles his semi-hard, spunk-leaking member into his underwear, you ask your captor why he didn't put his [i]thing[/i] in [i]fully[/i].
                            "[color=gblue]Sun's going down.
                            That would have taken too long.
                            I didn't feel like carrying your ass all the way back to the caravan.[/color]"
                            The bandit pulls up his pants and rebuckles his belt.
                            "[color=gblue]Besides, there'll be plenty of time for me to knot that tight little ${cunt ? `cunt` : `ass`} of yours once we get back.[/color]"
                        [/p]
                        [p]
                            That being said, the wolf pulls you to your feet by your iron collar.
                            With his new kobold slave in tow, he begins to proudly march back to the place where today's adventure began for you,
                            his scent clinging to your scales letting any canine you pass by know to whom you now belong.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    You are no longer a virgin.
                                [/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.enslaveSceneName, ctrl.enslaveFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.struggle.get() !== ForestWoodsRunFailStruggleChoice.ENJOY) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.enjoy.change();
        part += buildChoice();
        if (!ctrl.enslaveReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.struggle.get() !== ForestWoodsRunFailStruggleChoice.ENJOY) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.enslaveReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Forest\hide\run\fail\struggle\enjoy\ForestWoodsRunFailEnjoy.js

// scenes\Forest\hide\run\fail\struggle\ForestWoodsRunFailStruggle.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRunFailStruggle(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstChoice = new ForestWoodsRunFailHate(ctrl);
    var secondChoice = new ForestWoodsRunFailEnjoy(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var cunt = ctrl.misc.cunt();
        global.kobold.history.shared.enslavement.captureFucked = true;
        var toReturn = `[p]
                            You smack the wolf's strong forearm, slapping and batting at it with both of your hands.
                            You begin to writhe your body around once more, kicking and failing wildly,
                            yet all of your protests remain unnoticed.
                            The canine simply undoes his pants, his swollen sheath coming into view, the tapered tip of his cock already protruding from it.
                            Panicking, your brain works overtime to think of [i]something[/i], [b]anything[/b] that could save you now, yet it draws a complete blank.
                        [/p]
                        [p]
                            While you continue to struggle and squirm, the lupine gives his emerging length a few strokes, his shaft rapidly growing to its full size.
                            Only after preparing himself to [b]fuck your brains out[/b], does the wolf pay your incessant floundering any mind.
                            He harshly [b]bashes[/b] you against the tree a second time, strong enough to make the entire wooden giant shake,
                            and to knock the resistance out of you.
                            With his prey pacified, the canine lines up his member with your virginal ${cunt ? `slit` : `tailhole`}.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = ctrl.struggle.has() ? "disabled" : "";
        var isSelected = ctrl.struggle.get() === ForestWoodsRunFailStruggleChoice.HATE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Hate it</button>`;

        var tooltip = global.parser.parse(`You'd rather dip your hand into a pile of rusty metal than get fucked by this [b]brute[/b].`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = ctrl.struggle.has() ? "disabled" : "";
        var isSelected = ctrl.struggle.get() === ForestWoodsRunFailStruggleChoice.ENJOY;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Love it</button>`;

        var tooltip = global.parser.parse(`If this [i]is[/i] happning, you might as well enjoy it.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.struggle.set(ForestWoodsRunFailStruggleChoice.HATE);
            loadNextPart(firstChoice);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.struggle.set(ForestWoodsRunFailStruggleChoice.ENJOY);
            loadNextPart(secondChoice);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.runFail.get() !== ForestWoodsRunFailChoice.STRUGGLE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstChoice.build();
        part += secondChoice.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.runFail.get() !== ForestWoodsRunFailChoice.STRUGGLE) {
            return;
        }

        firstChoice.registerListeners();
        secondChoice.registerListeners();

        if (ctrl.struggle.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var ForestWoodsRunFailStruggleChoice = {
    HATE: 1,
    ENJOY: 2
};

// END scenes\Forest\hide\run\fail\struggle\ForestWoodsRunFailStruggle.js

// scenes\Forest\hide\run\fail\ForestWoodsRunFail.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRunFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstChoice = new ForestWoodsRunFailStruggle(ctrl);
    var secondChoice = new ForestWoodsRunFailBeg(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var bra = ctrl.misc.bra();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        global.kobold.history.shared.enslavement.captured = true;
        var toReturn = `[p]
                            Your very life at stake, you line yourself up perfectly with the piece of wood blocking your path.
                            You're only going to get [b]one[/b] shot at this.
                            Succeed, and you can escape through the copse, and hopefully make it home by nightfall.
                            If you fail, the wolf'll catch up to you, and who knows what he'll do?
                        [/p]`;

        if (ctrl.run.get() === ForestWoodsRunChoice.BREAKTHROUGH) {
            toReturn += `[p]
                            You [b]storm[/b] towards the barring branch,
                            digging your claws into the dirt below to add an extra ounce of sheer [b]power[/b] to your barreling form.
                            Gritting your teeth, you [b]ram[/b] into the wooden barricade with [b]all[/b] of your might.
                            The blockade creaks loudly, but it does not break.
                            Instead of ramming through it like you intended,
                            the branch [b]smacks[/b] into your gut like a wooden mace,
                            your own momentum turned back on you with a force strong enough to knock you off your feet and onto your back.
                        [/p]
                        ${global.ui.condFail(`Strength below 6`)}`;
        } else {
            toReturn += `[p]
                            You [b]rush[/b] towards the barring branch,
                            eyeing the small clearance below it, envisioning yourself sliding underneath it and emerging safely on the other side. 
                            Waiting until the very last moment, you let your upper body drop to the ground, gliding feet-first underneath the barricade.
                            Alas, you tilt your head up just a [i]little[/i] too high, and end up smacking face-first into the wooden barrier,
                            your entire sliding body coming to a halt.
                        [/p]
                        ${global.ui.condFail(`Agility below 6`)}`;
        }

        toReturn += `   [p]
                            The heavy blow is enough to daze you for a while.
                            As you lay motionlessly on your back, trying to collect your senses,
                            the wolf pursuing you approaches your downed form.
                            As the worst of the blurriness fades from your eyes,
                            his broad figure looming over you comes into view.
                        [/p]
                        [p]
                            Panic clouds your mind.
                            You attempt to crawl underneath the branch, to escape still,
                            but all the big man has to do to stop you is plant his hefty, black-furred footpaw directly on top of your chest,
                            like a hunter standing over freshly felled prey.
                            You wildly flail about, writhing around your entire body like a [b]child[/b] trying,
                            but failing to get away from righteous parental punishment.
                            All the seemingly calm lupine has to do to get you to stop, is pressing his clawed toes against your scaly abdomen,
                            the simple, minimalistic act wordlessly conveying a threat to [b]tear your guts out[/b] unless you stop squirming.
                        [/p]
                        [p]
                            Once you've calmed a bit, the black wolf reaches down.
                            Grabbing you by your throat, he removes his foot from your chest,
                            picking you up and lifting you off the ground by your very neck instead of pinning you to the forest floor.
                            ${weapon ? `You try to swing your ${global.kobold.inventory.createItemWithTooltip(weapon)}, but accidentally drop it to the floor.` : ``}
                            With his free hand, he deftly snags and pockets your ill-gotten coinpouch.
                            Then, he produces an odd, iron object, the likes of which you've only ever seen worn by slaves.
                            [i]Oh no.[/i]
                        [/p]
                        [p]
                            He brings the ferrous collar to your throat, and with a loud, metallic [i]click[/i], it wraps around your neck.
                            "[color=gblue]For the crime of thievery...[/color]" the wolf speaks in a bass voice.
                            "[color=gblue]...and for making me run halfway through the fucking forest...[/color]" he continues, sounding a lot more disgruntled than before.
                            "[color=gblue]...you are now the property of the Furskin Mountain-Crew.[/color]"
                            The Furskin Mountain-Crew.
                            But those are bandits, from [i]way[/i] up north.
                            What are they doing guarding caravans coming from the south?!
                            Lack of air from the furred hand squeezing your throat shut getting to you, your train of thoughts goes amiss.
                        [/p]
                        [p]
                            The canine clears his throat, seemingly only now having fully recovered from the chase.
                            "[color=gblue]Normally, this is where I'd tell you you've just lost all of your rights.[/color]"
                            He gives you a sly wink.
                            "[color=gblue]But you never had any to begin with, now did you?[/color]"
                            The wolf wickedly chuckles at his own sadistic joke,
                            his face returning to a dead-serious glare as he realizes you [i]aren't[/i] sharing his jolly sentiment.
                            You're stuck between barely being able to breathe,
                            and trying to process the fact that your entire future is now in the hands of the bandit holding you,
                            or whatever misfits he may call his [i]friends[/i].
                        [/p]
                        [p]
                            With a loud growl, the hulking brute [b]slams[/b] your back against a nearby tree,
                            lifting you up high enough to come face-to-face with him.
                            His free hand moves to your chest, where he ${bra ? `uses one of his claws to slowly shred
                            your ${global.kobold.inventory.createItemWithTooltip(bra)} into tiny little bits` : `slowly runs one of his claws in circles atop your scales`}.
                            The lupine grins wickedly.
                            "[color=gblue]You know, I've always wanted a kobold [b]slave[/b].[/color]"
                            His hand roams ever-lower, claw trailing over your soft belly, down towards your groin.
                            "[color=gblue]Guarding these [b]stupid[/b] transports gets lonely at times, especially [i]at night[/i].[/color]"
                            With a single, firm tug, he [b]tears[/b] apart your
                            ${pants ? `${global.kobold.inventory.createItemWithTooltip(pants)}` : `${global.kobold.inventory.createItemWithTooltip(underwear)}`},
                            your last bastion of modesty ripped away from you.
                            "[color=gblue]Caravan master may want to sell you, but that won't stop me from [i]testing the merchandise[/i] first.[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = ctrl.runFail.has() ? "disabled" : "";
        var isSelected = ctrl.runFail.get() === ForestWoodsRunFailChoice.STRUGGLE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Struggle</button>`;

        var tooltip = global.parser.parse(`The wolf's intentions are clear. That doesn't mean you're going to make it easy for him.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = ctrl.runFail.has() ? "disabled" : "";
        var isSelected = ctrl.runFail.get() === ForestWoodsRunFailChoice.BEG;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Beg</button>`;

        var tooltip = global.parser.parse(`Use your mind instead of your body to resist. Beg, lie and panhandle your way out of this.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.runFail.set(ForestWoodsRunFailChoice.STRUGGLE);
            loadNextPart(firstChoice);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.runFail.set(ForestWoodsRunFailChoice.BEG);
            loadNextPart(secondChoice);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.runFail.change();
        part += buildChoice();
        part += `</div>`;
        part += firstChoice.build();
        part += secondChoice.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        firstChoice.registerListeners();
        secondChoice.registerListeners();

        if (ctrl.runFail.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var ForestWoodsRunFailChoice = {
    STRUGGLE: 1,
    BEG: 2
};

// END scenes\Forest\hide\run\fail\ForestWoodsRunFail.js

// scenes\Forest\hide\run\succeed\ForestWoodsRunSucceedContinue.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRunSucceedContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            After making your way through the crevasse-like path, you luckily end up back in the forest proper.
                            Having shaken both the human [i]and[/i] the wolf, and with the sun slowly starting to go down,
                            you feel safe enough to relax for a moment.
                            That was a [b]strenuous[/b] encounter.
                            You almost got caught [b]several times[/b].
                            But then again, you suppose a bit of [i]excitement[/i] is a meagre price to pay for a haul this big.
                        [/p]
                        [p]
                            High risk, high reward.
                            With your thumping, adrenaline-fueled heart and your rapid, exhausted breathing slowly returning to normal,
                            you decide it's time to head back home.
                            You doubt the caravan guards are brave or [i]foolish[/i] enough to venture into the forest at night.
                            You're in the clear, [i]for now[/i].
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.homeSceneName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.preContinue.continued()) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.preContinue.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\hide\run\succeed\ForestWoodsRunSucceedContinue.js

// scenes\Forest\hide\run\succeed\ForestWoodsRunSucceed.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRunSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new ForestWoodsRunSucceedContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Your very life at stake, you line yourself up perfectly with the piece of wood blocking your path.
                            You're only going to get [b]one[/b] shot at this.
                            Succeed, and you can escape through the copse, and hopefully make it home by nightfall.
                            If you fail, the wolf'll catch up to you, and who knows what he'll do?
                        [/p]`;

        if (ctrl.run.get() === ForestWoodsRunChoice.BREAKTHROUGH) {
            toReturn += `[p]
                            You [b]storm[/b] towards the barring branch,
                            digging your claws into the dirt below to add an extra ounce of sheer [b]power[/b] to your barreling form.
                            Gritting your teeth, you [b]ram[/b] into the wooden barricade with [b]all[/b] of your might.
                            With a loud [b]crack[/b], the rotten branch breaks in half, clearing a path into the heart of the thicket.
                            You disappear into the maze of trees with not a moment to spare, leaving the wolf to curse his broad build.
                        [/p]
                        ${global.ui.condSucc(`Strength greater than 5`)}`;
        } else {
            toReturn += `[p]
                            You [b]rush[/b] towards the barring branch,
                            eyeing the small clearance below it, envisioning yourself sliding underneath it and emerging safely on the other side. 
                            Waiting until the very last moment, you let your upper body drop to the ground, gliding feet-first underneath the barricade.
                            To your own surprise, you make it to the other side with a scratch!
                            After getting back up on your feet, you disappear into the maze of trees with not a moment to spare, leaving the wolf to wonder where you went.
                        [/p]
                        ${global.ui.condSucc(`Agility greater than 5`)}`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.preContinue.continued() ? "disabled" : "";
        var isSelected = ctrl.preContinue.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.preContinue.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.runSucceed.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.preContinue.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Forest\hide\run\succeed\ForestWoodsRunSucceed.js

// scenes\Forest\hide\run\ForestWoodsRunSlide.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRunSlide(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new ForestWoodsRunFail(ctrl);
    var toSuccess = new ForestWoodsRunSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.agility() >= 6;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.run.get() !== ForestWoodsRunChoice.SLIDE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.run.get() !== ForestWoodsRunChoice.SLIDE) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Forest\hide\run\ForestWoodsRunSlide.js

// scenes\Forest\hide\run\ForestWoodsRunBreakthrough.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRunBreakthrough(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new ForestWoodsRunFail(ctrl);
    var toSuccess = new ForestWoodsRunSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.strength() >= 6;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.run.get() !== ForestWoodsRunChoice.BREAKTHROUGH) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.run.get() !== ForestWoodsRunChoice.BREAKTHROUGH) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Forest\hide\run\ForestWoodsRunBreakthrough.js

// scenes\Forest\hide\run\ForestWoodsRun.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsRun(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstChoice = new ForestWoodsRunSlide(ctrl);
    var secondChoice = new ForestWoodsRunBreakthrough(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You won't be getting caught today.
                            No matter how bad things may look, this is still [b]your[/b] forest.
                            They may have tracked you this far, but having lived here your entire life, you know a thing or two that they don't.
                            This isn't your first time getting chased around the woods.
                            And you [b]refuse[/b] to let it be your last.
                        [/p]
                        [p]
                            There are certain places where [i]larger[/i] species simply [i]can't reach[/i].
                            Parts of the forest where trees are packed so closely together, they form a [b]barrier[/b],
                            impassable to anyone bigger than a child...[i]or a kobold[/i].
                            One of these sections is pretty close to where you are.
                            If you can make it there before your pursuers catch up to you,
                            you [i]might[/i] have a [i]chance[/i] at escaping their grasping clutches.
                        [/p]
                        [p]
                            You begin to [b]run[/b] towards what you believe to be your one shot at salvation.
                            Looking behind you, you can see the clear outlines of a pair of large men in the distance.
                            They're rapidly gaining on you, their sheer size allowing them to easily outrun you.
                            You pick up the pace, but it's still not enough.
                            In less than a minute since the chase began, they manage to get close enough for you to identify them.
                            An anthromorphic wolf and a human, with a leashed feral dog leading the latter in the chase.
                            The rabid spittle fluttering from its toothy maw leaves behind a disgusting trail, undoubtedly born from mindless [b]anger[/b].
                        [/p]
                        [p]
                            It's just a little further up ahead.
                            Your pursuers are getting drawing near.
                            You can't let them get any closer.
                            Pushing your body to its very limits, you burst into a sprint.
                            Still it's not enough.
                            The human and his dog are [b]right[/b] behind you, the feral's teeth on the [b]verge[/b] of sinking into your scaly tail.
                        [/p]
                        [p]
                            Suddenly, out of nowhere, a startled deer [b]hops[/b] in front of you, scaring you to the point of nearly costing you your balance.
                            You fear the small slowdown may have cost you your very life, but casting your gaze backwards,
                            you realize the animal was a hint of [b]good fortune[/b] instead of spelling your doom.
                            The dog chasing after you loses interest in catching your scaly tail, instead bolting after the fleeing cervine.
                            Its master, unwilling to let go of the canine's leash, gets tugged to the ground by the sudden change of direction.
                        [/p]
                        [p]
                            The human loudly curses at his pet, which disappears into the woods.
                            With him and the dog both out of the picture, you might make it to safety after all!
                            The wolf, being just a little [i]too[/i] big to nimbly manoeuvre around the woods, is still quite a bit off. 
                            Unlike his human companion, though, he doesn't simply [i]give up[/i] pursuit.
                        [/p]
                        [p]
                            After half a minute more of the wearying chase, the dense formation of trees you were heading for comes into view.
                            The wooden giants together form a living wall, separated by gaps too tiny for even [i]your kind[/i] to fit through.
                            There is, however, a single small passageway leading throughout the entirety of the thicket.
                            Big enough for you, not [b]nearly[/b] large enough for [i]him[/i].
                            That is your goal.
                        [/p]
                        [p]
                            Fear fills your heart as you near the tight corridor.
                            A large, rotten branch stuck between two trees blocks the entrance to the path.
                            With the wolf gaining on you, you don't have time to try to dislodge it.
                            You're going to have to deal with this without stopping.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = ctrl.run.has() ? "disabled" : "";
        var isSelected = ctrl.run.get() === ForestWoodsRunChoice.SLIDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled} style="width:180px;">Slide underneath</button>`;

        var tooltip = global.parser.parse(`A fancy slide through the dirt'll allow you to avoid the obstacle.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:180px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = ctrl.run.has() ? "disabled" : "";
        var isSelected = ctrl.run.get() === ForestWoodsRunChoice.BREAKTHROUGH;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Break through</button>`;

        var tooltip = global.parser.parse(`The barring branch looks old and weak. Bash through it with blunt force and momentum.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.run.set(ForestWoodsRunChoice.SLIDE);
            loadNextPart(firstChoice);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.run.set(ForestWoodsRunChoice.BREAKTHROUGH);
            loadNextPart(secondChoice);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.treeFail.get() !== ForestWoodsTreeFailChoice.RUN && ctrl.bush.get() !== ForestWoodsHideBushChoice.RUN) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.run.change();
        part += buildChoice();
        part += `</div>`;
        part += firstChoice.build();
        part += secondChoice.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.treeFail.get() !== ForestWoodsTreeFailChoice.RUN && ctrl.bush.get() !== ForestWoodsHideBushChoice.RUN) {
            return;
        }

        firstChoice.registerListeners();
        secondChoice.registerListeners();

        if (ctrl.run.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var ForestWoodsRunChoice = {
    SLIDE: 1,
    BREAKTHROUGH: 2
};

// END scenes\Forest\hide\run\ForestWoodsRun.js

// scenes\Forest\hide\surrender\spank\ForestWoodsHideSurrenderSpankContinue.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHideSurrenderSpankContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var toReturn = `[p]
                            After a few minutes of pouting and crying, you slowly rise to your feet.
                            You sadly don the empty coin pouch, and ${pants ? `pull up your pants after rubbing` : `rub`} your sore behind for good measure.
                            The wolf may have disciplined you like a child, but it still [b]really[/b] hurt, especially due to the size difference between the two of you.
                            Still, you're kind of glad you got out of this [i]mostly[/i] unharmed.
                            Those men could have [b]seriously[/b] injured you if they wanted to.
                            And in a way, you [b]deserved[/b] far more than a simple spanking.
                        [/p]
                        [p]
                            Saddened by the loss of gold and the impact this'll undoubtedly have on your relationship with the Three-Tailed Rat Clan,
                            you begin the shameful walk home empty-handed.
                            At least, with nothing of value to steal, you don't have to worry about bandits on the way there.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.homeSceneName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.preContinue.continued()) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.preContinue.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\hide\surrender\spank\ForestWoodsHideSurrenderSpankContinue.js

// scenes\Forest\hide\surrender\spank\ForestWoodsHideSurrenderSpank.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHideSurrenderSpank(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new ForestWoodsHideSurrenderSpankContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        // CAN NOT reach this part via treeFail, because treeFail implies rock or stick.
        var toReturn = `[p]
                            Game over.
                            They've got you trapped.
                            You [i]could[/i] run, but you doubt you'd be able to get away.
                            If you hadn't squandered away your time by trying to hide, perhaps you'd have stood a chance. 
                            As things are now, your best bet would be to [i]surrender[/i] yourself to them, and [b]hope[/b] they go easy on you.
                        [/p]
                        [p]
                            You emerge from the bush, and pre-emptively remove the bag of gold from around your neck.
                            Staring at it longingly, you let out a loud sigh.
                            If only you'd been a little [i]smarter[/i], you wouldn't have had to part with your one chance at paying your debts.
                            But alas, what's done is done.
                            In some weird way, this is the [b]right[/b] thing to do.
                            You swallow deeply, and mentally prepare yourself for whatever judgement your pursuers may render unto you.
                        [/p]
                        [p]
                            The party comes into view.
                            A big, buff anthromorphic wolf and a human, with a leashed feral dog that barks its lungs out at you.
                            They seem surprised to see you standing here, coinpouch in hand.
                            The human calms his loud pet, threatening to [b]hit it[/b] to make it [b]shut up[/b].
                            Then, the wolf steps towards you.
                        [/p]
                        [p]
                            The broad lupine leans down to bring his head to your level, glaring at you with hate-filled eyes, his large fangs bared.
                            You shiver in fear, frozen in place.
                            Without a word, the man takes the pouch from your hand, bouncing it up and down, as if [i]checking[/i] the weight.
                            Then, finally, he breaks the tense silence.
                            "[color=gblue]So you're our [i]little[/i] [b]thief[/b], huh?[/color]"
                            You barely manage to [b]nod[/b] in response, too scared to speak.
                            "[color=gblue]Did you really think you could [b]steal[/b] from the [b]baron[/b] and get away with it?[/color]"
                            You shake your head no, and stammer out a meek apology.
                        [/p]
                        [p]
                            "[color=gblue]I've [b]eaten[/b] lizards bigger than you.[/color]"
                            The wolf licks his lips, as though salivating at the thought.
                            "[color=gblue]But then again, I've also raised four pups of my own,
                            and one of them is a [b]thieving little shit[/b] just like you.[/color]"
                            The canine tosses the gold pouch to his human accomplice, hitting his mark without even needing to look back.
                        [/p]
                        [p]
                            "[color=gblue]Normally, me and Kajar would cut you up into [b]minced meat[/b] and feed you to Rex for what you did.
                            Rex is the dog, mind you.[/color]"
                            The pooch barks upon hearing its name.
                            "[color=gblue]But since you [b]presented[/b] yourself to us on a [b]silver platter[/b],
                            like a little [b]cub[/b] that [b]knows[/b] it [b]fucked up bigtime[/b], I'll go easy on you.[/color]"
                            A broad grin forms on the wolf's face.
                            "[color=gblue]After all, I know how to handle [b]brats[/b] like none other.[/color]"
                        [/p]
                        [p]
                            The wolf straightens his back once more, towering over you like a mountain over a valley.
                            Grabbing you by the scruff with his powerful hand, he outright [b]drags[/b] you with him as he walks towards a nearby felled tree.
                            You do little to resist, knowing fully well that the lupine could [b]snap your neck[/b] with ease if he wanted to.
                            You begin to fear the worst, however, as the strong canine forcibly [b]bends[/b] you over the horizontal
                            trunk${pants ? `. Your fear deepens as your ${global.kobold.inventory.createItemWithTooltip(pants)} get crudely lowered` : ``}, your
                            bare behind on display for the entire forest to see.
                        [/p]
                        [p]
                            Panic settles in as you think the wolf is about to [b]rape[/b] you.
                            What about all that talk about his children?!
                            Surely he doesn't do [i]that[/i] to them, right?!
                            Luckily, the lupine doesn't step behind you, instead ominously standing right next to you, a sadistic smirk plastered across his face.
                            What, is he going to let the [b]human[/b] have a go?
                            Or maybe the [b]dog[/b]?!
                        [/p]
                        [p]
                            You begin to struggle against the canine's unrelenting grasp, your mind coming up with [b]worse[/b] and [b]worse[/b] scenarios.
                            Ignoring your feeble resistance, the strong guard lifts one of his big, buff arms high up in the air,
                            before swinging it downwards at a high velocity, the flat palm of his furred hand [b]smacking[/b] heavily against your scaly rear.
                            You yelp out in surprise, stinging pain spreading forth from your slightly jiggling buttcheeks.
                            He's...[b]spanking[/b] you?!
                        [/p]
                        [p]
                            True to his word, the wolf treats you like nothing but a [b]child[/b].
                            He raises his hand once more, striking your rear a second time, and a third, and a fourth, pain mounting with every hit.
                            You [b]beg[/b] for him to stop, tears forming at the corners of your eyes, but there is no mercy to be found here.
                            The sound of the human laughing out loud at the display before him,
                            the rhythmic slapping of flesh against scales,
                            and your miserable cries for mercy reverberate across the woods,
                            your utter [b]humiliation[/b] broadcasted far beyond the eyes of your abusers.
                        [/p]
                        [p]
                            A few minutes of the demeaning treatment later, you are reduced to a pathetic, crying [b]mess[/b] of a would-be-thief.
                            One final time, the wolf's firm hand smacks against your fiery-red bottom, drawing one more pitiful whimper from your lips.
                            Finally, he is done.
                            Grabbing you by the scruff once more, he pulls you from the log, and tosses you to the ground,
                            forcing you to [b]sit[/b] on your tenderized behind, knowing fully well you have nobody but [b]yourself[/b] to blame for all of this.
                        [/p]
                        [p]
                            "[color=gblue]Think about what you've done. There [b]won't[/b] be a next time, got it?[/color]"
                            You wipe away your still-flowing tears, and meekly nod your head.
                            "[color=gblue]Good. Here's something to remember us by.[/color]"
                            The wolf signals to his comrade.
                            The human, in turn, tosses the emptied coin pouch back at you, as if to serve as a reminder of what happened here today.
                            After exchanging glances and nodding at one another, the two men and their pet leave without a word.
                            They got what they came for, and you got what you deserved.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.preContinue.continued() ? "disabled" : "";
        var isSelected = ctrl.preContinue.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.preContinue.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.surrenderSpank.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.preContinue.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Forest\hide\surrender\spank\ForestWoodsHideSurrenderSpank.js

// scenes\Forest\hide\surrender\rock\ForestWoodsHideSurrenderRockContinue.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHideSurrenderRockContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var rock = ctrl.intro.getItemChoice() === ItemEnum.ROCK;
        var weapon = ctrl.misc.weapon();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            After nearly half an hour of bawling your eyes out, you scrape yourself off of the ground, and onto your feet.
                            Your slit has stopped bleeding,
                            and the slight gape that was left in the wake of your ${global.kobold.inventory.createItemWithTooltip(weapon)} has all but vanished.
                            It still hurts, but you hope the pain, too, will fade in due time.
                        [/p]
                        [p]
                            You ${pants ? `pull up your ${global.kobold.inventory.createItemWithTooltip(pants)}` : `put
                            your ${global.kobold.inventory.createItemWithTooltip(underwear)} back on`}.
                            The sun is slowly starting to set.
                            You can't stay here.
                            Walking with a slight limp, you set off for home.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.homeSceneName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.preContinue.continued()) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.preContinue.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\hide\surrender\rock\ForestWoodsHideSurrenderRockContinue.js

// scenes\Forest\hide\surrender\rock\ForestWoodsHideSurrenderRock.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHideSurrenderRock(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new ForestWoodsHideSurrenderRockContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var treeFail = ctrl.treeFail.has();
        var rock = ctrl.intro.getItemChoice() === ItemEnum.ROCK;
        var weapon = ctrl.misc.weapon();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            Game over.
                            They've got you trapped.
                            You [i]could[/i] run, but you doubt you'd be able to get away.
                            If you hadn't squandered away your time by trying to hide, perhaps you'd have stood a chance. 
                            As things are now, your best bet would be to [i]surrender[/i] yourself to them, and [b]hope[/b] they go easy on you.
                        [/p]
                        [p]
                            ${treeFail ? `You` : `You step out of the bush, and`} pre-emptively remove the bag of gold from around your neck.
                            Staring at it longingly, you let out a loud sigh.
                            If only you'd been ${treeFail ? `able to jump a little higher` : `a little [i]smarter[/i]`},
                            you wouldn't have had to part with your one chance at paying your debts.
                            But alas, what's done is done.
                            In some weird way, this is the [b]right[/b] thing to do.
                            You swallow deeply, and mentally prepare yourself for whatever judgement your pursuers may render unto you.
                        [/p]
                        [p]
                            The party comes into view.
                            A big, buff anthromorphic wolf and a human, with a leashed feral dog that barks its lungs out at you.
                            They seem surprised to see you standing here, coinpouch in hand.
                            The human calms his loud pet, threatening to [b]hit it[/b] to make it [b]shut up[/b].
                            Then, the wolf steps towards you.
                        [/p]
                        [p]
                            The broad lupine leans down to bring his head to your level, glaring at you with hate-filled eyes, his large fangs bared.
                            You shiver in fear, frozen in place.
                            Without a word, the man takes the pouch from your hand, bouncing it up and down, as if [i]checking[/i] the weight.
                            Then, finally, he breaks the tense silence.
                            "[color=gblue]So you're our [i]little[/i] [b]thief[/b], huh?[/color]"
                            You barely manage to [b]nod[/b] in response, too scared to speak.
                            "[color=gblue]Did you really think you could [b]steal[/b] from the [b]baron[/b] and get away with it?[/color]"
                            You shake your head no, and stammer out a meek apology.
                        [/p]
                        [p]
                            "[color=gblue]I've [b]eaten[/b] lizards bigger than you.[/color]"
                            The wolf licks his lips, as though salivating at the thought.
                            "[color=gblue]But then again, I've also raised four pups of my own,
                            and one of them is a [b]thieving little shit[/b] just like you.[/color]"
                            The canine tosses the gold pouch to his human accomplice, hitting his mark without even needing to look back.
                        [/p]
                        [p]
                            "[color=gblue]Normally, me and Kajar would cut you up into [b]minced meat[/b] and feed you to Rex for what you did.
                            Rex is the dog, mind you.[/color]"
                            The pooch barks upon hearing its name.
                            "[color=gblue]But since you [b]presented[/b] yourself to us on a [b]silver platter[/b],
                            like a little [b]cub[/b] that [b]knows[/b] it [b]fucked up bigtime[/b], I'll go easy on you.[/color]"
                            A broad grin forms on the wolf's face.
                            "[color=gblue]After all, I know how to handle [b]brats[/b] like none other.[/color]"
                        [/p]
                        [p]
                            "[color=gblue]What do we have here?[/color]" the lupine continues,
                            snatching your ${global.kobold.inventory.createItemWithTooltip(weapon)} from your hand.
                            "[color=gblue]Planning to [b]hurt[/b] someone with this, now were you?[/color]"
                            The wolf grabs your throat, not quite [b]squeezing[/b] it shut, but still making sure you aren't going anywhere.
                            "[color=gblue]I know exactly what to do with you now, you [b]failure[/b] of a [b]bandit[/b].[/color]"
                        [/p]
                        [p]
                            The wolf straightens his back once more, towering over you like a mountain over a valley.
                            Dragging you along by your neck, he brings you to the very same tree you considered climbing earlier.
                            The canine roughly slams your back against the wooden behemoth.
                            With his free hand, he crudely tugs down your
                            ${pants ? `${global.kobold.inventory.createItemWithTooltip(pants)}` : `${global.kobold.inventory.createItemWithTooltip(underwear)}`},
                            your smooth groin on full display.
                        [/p]
                        [p]
                            You try to cover yourself using your tail, but a stern, disapproving, almost [i]fatherly[/i] finger dissuades you.
                            Using his feet, he kicks apart your legs, spreading your thighs and exposing your loins further.
                            While embarrassed, you don't fight the lupine.
                            He could snap you in half without even trying.
                            Then, the wolf presses your very own ${global.kobold.inventory.createItemWithTooltip(weapon)} against your
                            ${dick ? `cock` : `virginal `}slit, and before you even know what's going on,
                            the ${rock ? `blunt object` : `pointy end`} is [b]rammed[/b] inside of you,
                            your velvety insides parting to make room for the unnaturally-shaped intruder.
                        [/p]
                        [p]
                            Your eyes go wide.
                            You [b]scream[/b] in pain, the coarse surface of the ${rock ? `rock` : `stick`} [b]scraping[/b] against your sensitive interior.
                            The canine lets go of your throat.
                            You instantly fall to the ground, clutching your abused groin,
                            your former weapon sticking out of your stretched hole.
                            The human bursts out into laughter.
                            The wolf walks back over to him.
                            They toss the coinpouch, emptied of course, onto your curled up frame.
                            With their gold reclaimed, and punishment dealt out, they start heading back towards the caravan, their feral pet in tow.
                        [/p]
                        [p]
                            You loudly groan in agony, a small amount of blood trickling from your crotch, running over your inner thigh before dripping to the ground.
                            Despite your weapon not being inserted [b]fully[/b], you're still pretty certain
                            ${rock ? `it stretched you to well beyond what you were built to take` : `a few woodchips splintered off and got lodged in your flesh`}.
                            Even worse: this ordeal isn't over yet.
                            You still need to [b]get it out[/b].
                        [/p]
                        [p]
                            Breathing shakily, you reach in between your legs and tentatively tug on the foreign object embedded in your body.
                            It painfully shifts around a bit, sending fresh spikes of torment down your spine.
                            Gritting your teeth, you get a better grip on the ${rock ? `stone` : `branch`}.
                            There's no easy way to do this, so you might as well [b]get it over with[/b] all in one go.
                            With a single, firm [b]tug[/b], you [b]yank[/b] your former weapon from your ${dick ? `cockslit` : `cunt`},
                            letting out a defeated, anguished cry as it chafingly slides from your groin.
                        [/p]
                        [p]
                            The considerable pain is [b]nothing[/b] compared to the [b]shame[/b] and the [b]defilement[/b].
                            You spend several minutes crying on the ground,
                            your most [b]sensitive[/b] spot reduced to a [b]sheath[/b] for  ${rock ? `dirty rocks` : `splintery sticks`}.
                        [/p]
                        [p]
                            You can't believe that cruel wolf actually [b]did[/b] something like that.
                            Even worse: part of your mind believes you [b]deserved[/b] that punishment.
                            All you wanted to do was repay your debt.
                            Instead, you are left broken, bleeding and crying on the soft forest floor, with not a coin to your name.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    You are no longer a ${dick ? `cockslit` : ``} virgin.
                                [/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.preContinue.continued() ? "disabled" : "";
        var isSelected = ctrl.preContinue.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.preContinue.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.surrenderRock.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.preContinue.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Forest\hide\surrender\rock\ForestWoodsHideSurrenderRock.js

// scenes\Forest\hide\surrender\ForestWoodsHideSurrender.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHideSurrender(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new ForestWoodsHideSurrenderRock(ctrl);
    var toSuccess = new ForestWoodsHideSurrenderSpank(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.intro.getItemChoice() === 0;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.treeFail.get() !== ForestWoodsTreeFailChoice.SURRENDER && ctrl.bush.get() !== ForestWoodsHideBushChoice.SURRENDER) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.treeFail.get() !== ForestWoodsTreeFailChoice.SURRENDER && ctrl.bush.get() !== ForestWoodsHideBushChoice.SURRENDER) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Forest\hide\surrender\ForestWoodsHideSurrender.js

// scenes\Forest\hide\bush\stick\ForestWoodsHideStickContinue.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHideStickContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You did it!
                            By manipulating their pet, you managed to outsmart both the human, [b]and[/b] the wolf!
                            Still, for your own safety, you remain concealed inside of the bush for a few minutes more.
                            Y'know, so you can be sure that they [i]aren't[/i] trying to [i]fake you out[/i] or anything.
                        [/p]
                        [p]
                            Once you think the coast is clear, you slowly creep out of the bush, like a shark emerging from a water-surface.
                            After warily taking a look around, you begin to head back home.
                            The sun is slowly starting to set, and you don't want to be out here in the dark, especially not with two hundred coins dangling off your neck.
                        [/p]
                        [p]
                            In retrospect, you are glad things worked out this way.
                            You didn't get caught, nobody got hurt, and the dog seemed pretty happy about its new stick.
                            Sure, the men'll probably get [i]reprimanded[/i] or something for failing to find you, but that's none of your concern.
                            If they were actually [b]paying attention[/b], they would have probably spotted you when you did the whole stick-thing,
                            so they kind of [b]deserve[/b] to get punished, in a way.
                        [/p]
                        [p]
                            That's not to say you [b]want[/b] them to get fired, or whatever.
                            It's just the way things go.
                            Do a bad job, lose your job.
                            But then again, you've never really been employed yourself, and if they [i]had[/i] found you,
                            then things would have probably gone [b]horribly wrong[/b] for you.
                        [/p]
                        [p]
                            As you begin to near your home, you're left with slightly mixed feelings.
                            You [i]think[/i] you did the right thing, but as a thief,
                            the [i]right thing[/i] is not always the [b]best[/b] course of action.
                            Suffice to say, at least [b]you[/b] ended up better off today.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.homeSceneName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.preContinue.continued()) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.preContinue.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\hide\bush\stick\ForestWoodsHideStickContinue.js

// scenes\Forest\hide\bush\stick\ForestWoodsHideStick.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHideStick(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new ForestWoodsHideStickContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            Okay.
                            Calm down.
                            They have a dog.
                            What do dogs like?
                            Dogs like [b]sticks[/b].
                            Luckily for you, you just so happened to pick up a ${global.kobold.inventory.createItemWithTooltip(weapon)} earlier.
                            [i]Maybe[/i], just [b]maybe[/b], you [i]might[/i] be able to distract the beast for long enough for its masters to think it lead them astray.
                        [/p]
                        [p]
                            Your pursuers draw close, close enough for you to identify them.
                            A human, a broadly-built, buff anthromorphic wolf and a feral guard dog doubling as a tracking hound, held on a leash.
                            While the wolf and the human chatter amongst one another, the dog instantly [b]bolts[/b] towards the bush you're hiding in.
                            Luckily, its masters fail to notice its interest.
                        [/p]
                        [p]
                            You grab your ${global.kobold.inventory.createItemWithTooltip(weapon)}, and poke it out of the bush you're in.
                            As you wave it around, the dog stops its aggressive barking, instead seemingly enthralled by the piece of wood.
                            After a bit of teasing, you [b]throw[/b] your stick as far as you can to the side, the dog immediately [b]chasing[/b] after it,
                            nearly pulling the human holding its leash off of his feet.
                            "[color=red]He's got something! Get'em, Rex![/color]"
                        [/p]
                        [p]
                            The dog's handlers follow it to the stick you just threw,
                            completely disillusioned and disappointed with their pet,
                            once they realize it lead them to a [b]piece of wood[/b] instead of to the [b]lizard[/b] they were hoping to catch.
                            With a loud, communal sigh, the human and the wolf decide to call their chase a [i]bust[/i].
                            The dog clutching its new, wooden toy in between its sharp teeth, the party of two and a half head back the way they came.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.preContinue.continued() ? "disabled" : "";
        var isSelected = ctrl.preContinue.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.preContinue.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.bush.get() !== ForestWoodsHideBushChoice.STICK) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.hideStick.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.bush.get() !== ForestWoodsHideBushChoice.STICK) {
            return;
        }

        next.registerListeners();

        if (ctrl.preContinue.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Forest\hide\bush\stick\ForestWoodsHideStick.js

// scenes\Forest\hide\bush\ForestWoodsHideBush.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHideBush(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstChoice = new ForestWoodsHideSurrender(ctrl);
    var secondChoice = new ForestWoodsRun(ctrl);
    var thirdChoice = new ForestWoodsHideStick(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            The simplest option would be to hide in a bush.
                            Easy to get into, easy to get out of.
                            If you get spotted while hiding way up in a tree, you won't have anywhere to run.
                            You'll be trapped, stuck ten feet up without any way out.
                            If your pursuers somehow manage to [i]find[/i] you hiding in a bush, then you can simply [b]bolt[/b] out of it and run away.
                            A win-win situation, really.
                        [/p]
                        [p]
                            You pick a nice, dense shrub and quickly stride into it, your entire body,
                            from the claws of your feet to the scaly tip of your head covered and camouflaged with foliage without even [b]needing[/b] to duck down.
                            The perfect hiding spot!
                            They'll never find you here.
                            Not unless they have some kind of...
                            [b]Oh no![/b]
                        [/p]
                        [p]
                            You [b]completely[/b] forgot about the feral dog they had with them!
                            It'll sniff you out for sure!
                            You [i]should[/i] have gone for the tree, but it's [b]far[/b] too late to change places now.
                            You already can see your pursuers approaching in the distance, feral canine leading the party, barks echoing throughout the entire forest.
                        [/p]
                        [p]
                            You panic.
                            A million thoughts race through your mind.
                            You can't stay here.
                            Not without a way to distract the dog.
                            Even then, it'll be risky.
                            It might be better to just [i]come clean[/i].
                            But then again, you [b]really[/b] need this gold.
                            Maybe you can make a run for it?
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = ctrl.bush.has() ? "disabled" : "";
        var isSelected = ctrl.bush.get() === ForestWoodsHideBushChoice.SURRENDER;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Surrender</button>`;

        var tooltip = global.parser.parse(`The jig is up. If you surrender, they might spare your life.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = ctrl.bush.has() ? "disabled" : "";
        var isSelected = ctrl.bush.get() === ForestWoodsHideBushChoice.RUN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Run</button>`;

        var tooltip = global.parser.parse(`They haven't caught you [i]yet[/i]. Leg it!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 150px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var stick = ctrl.intro.getItemChoice() === ItemEnum.BRANCH;
        var disabled = ctrl.bush.has() || !stick ? "disabled" : "";
        var isSelected = ctrl.bush.get() === ForestWoodsHideBushChoice.STICK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Throw stick</button>`;

        var tooltip = global.parser.parse(`If you throw your branch, that might distract the dog for long enough for you to escape.`);
        if (!stick) {
            tooltip = global.parser.parse(`If only you had one.`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 220px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bush.set(ForestWoodsHideBushChoice.SURRENDER);
            loadNextPart(firstChoice);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bush.set(ForestWoodsHideBushChoice.RUN);
            loadNextPart(secondChoice);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.bush.set(ForestWoodsHideBushChoice.STICK);
            loadNextPart(thirdChoice);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.hide.get() !== ForestWoodsHideChoice.BUSH) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.bush.change();
        part += buildChoice();
        part += `</div>`;
        part += firstChoice.build();
        part += secondChoice.build();
        part += thirdChoice.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.hide.get() !== ForestWoodsHideChoice.BUSH) {
            return;
        }

        firstChoice.registerListeners();
        secondChoice.registerListeners();
        thirdChoice.registerListeners();

        if (ctrl.bush.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var ForestWoodsHideBushChoice = {
    SURRENDER: 1,
    RUN: 2,
    STICK: 3
};

// END scenes\Forest\hide\bush\ForestWoodsHideBush.js

// scenes\Forest\hide\tree\fail\ForestWoodsTreeFail.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsTreeFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstChoice = new ForestWoodsHideSurrender(ctrl);
    var secondChoice = new ForestWoodsRun(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            You frantically look around for the best tree to climb, yet sadly, [b]none[/b] of them seem to have [b]any[/b] low-hanging branches.
                            Not low enough for you to grab onto, at least.
                            Your best bet, or so you figure, is a large, sturdy looking oak tree with a medium-sized, moss-covered rock right next to it.
                            If you can use the outcrop to give you a boost, you [i]might[/i] just be able to reach high enough to grab an offshoot.
                        [/p]
                        [p]
                            After taking a few steps back, you [b]sprint[/b] towards the oak,
                            accelerating rapidly from standing still to [b]bolting[/b] as fast as you possibly can,
                            gaining momentum with every pace, your sharp claws [b]digging[/b] into the soft dirt below to propel you forwards.
                            Nearing the rock, you leap into the air as high as you can,
                            lifting off with just one foot so you can quickly place the other atop the mossy rock surface,
                            using it to launch yourself even higher.
                        [/p]
                        ${global.ui.condFail(`Carrying weapon`)}
                        [p]
                            Alas, perhaps in part due to the added weight brought on by your ${global.kobold.inventory.createItemWithTooltip(weapon)},
                            you miss the branch you were aiming for by just an inch, your fingers scraping against the underside of it,
                            failing to find anything to hold on to.
                            You plummet to the ground, 
                            only barely managing to refrain from hurting yourself by gracefully turning your accumulated kinetic energy into an adept roll,
                            a move with which you surprise even yourself.
                        [/p]
                        [p]
                            Having landed on the ground safe and sound may have prevented you from painfully face-planting the dirt,
                            but you're not out of the woods yet.
                            The barking dog is speedily approaching, and it's far too late for you to find a new hiding spot.
                            You'll either have to come clean, surrender and hope they [b]won't[/b] carve you in half for what you've done,
                            or you can collect your wits and make them [b]work[/b] for their gold by hightailing it out of here as fast as you can.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = ctrl.treeFail.has() ? "disabled" : "";
        var isSelected = ctrl.treeFail.get() === ForestWoodsTreeFailChoice.SURRENDER;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Surrender</button>`;

        var tooltip = global.parser.parse(`The jig is up. If you surrender, they might spare your life.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = ctrl.treeFail.has() ? "disabled" : "";
        var isSelected = ctrl.treeFail.get() === ForestWoodsTreeFailChoice.RUN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Run</button>`;

        var tooltip = global.parser.parse(`They haven't caught you [i]yet[/i]. Leg it!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 150px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.treeFail.set(ForestWoodsTreeFailChoice.SURRENDER);
            loadNextPart(firstChoice);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.treeFail.set(ForestWoodsTreeFailChoice.RUN);
            loadNextPart(secondChoice);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.treeFail.change();
        part += buildChoice();
        part += `</div>`;
        part += firstChoice.build();
        part += secondChoice.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        firstChoice.registerListeners();
        secondChoice.registerListeners();

        if (ctrl.treeFail.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var ForestWoodsTreeFailChoice = {
    SURRENDER: 1,
    RUN: 2
};

// END scenes\Forest\hide\tree\fail\ForestWoodsTreeFail.js

// scenes\Forest\hide\tree\succeed\throw\ForestWoodsThrowContinue.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsThrowContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You continue to run for a solid few minutes, ensuring you've shaken off both the guards [i]and[/i] the bees before coming to a brief halt.
                            All things considered, that went pretty well!
                            While they might not even [i]know[/i] it was [b]you[/b] who did it, you gave those guards a run for their money,
                            using nothing but your wits, your athleticism, and your forest knowledge.
                            Heck, even without the bees, they probably would have given up before finding your excellent hiding spot!
                        [/p]
                        [p]
                            Sure, you kind of feel bad for those guys.
                            They were probably just trying to do their job.
                            You [i]did[/i] steal from them, after all.
                            But still, who's going to miss a single two hundred gold bag from a transport carrying tens of thousands of coins?
                            Nobody, that's who.
                            You did what you had to.
                            And more importantly: you got away with it, at least for now.
                        [/p]
                        [p]
                            After catching your breath, you realize the direction in which you ran from the bees conveniently brought you closer to the path leading home.
                            Given that the sun is slowly starting to go down, you figure it's best if you return to your own little shack in the woods.
                            You don't want to stay out here at night, after all.
                            That's when the [i]worst[/i] of the bandits come out, and you [i]don't[/i] want to mess with [i]those[/i] guys,
                            especially not with two hundred gold dangling around your neck.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.homeSceneName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.preContinue.continued()) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.preContinue.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\hide\tree\succeed\throw\ForestWoodsThrowContinue.js

// scenes\Forest\hide\tree\succeed\throw\ForestWoodsThrow.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsThrow(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new ForestWoodsThrowContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Fearing for your life in more than one way, you hop to your feet atop the tree branch,
                            balancing your way over to the beehive as fast as you can.
                            At first, you intend to grab it and throw it down,
                            yet you really don't want to put that swarming ball of sting-happy buzzers anywhere [b]near[/b] your face.
                            Thinking swiftly and acting even quicker, you prompty [b]kick[/b] the dense ball of honeycomb and wax,
                            its binding to the tree loudly cracking and breaking.
                        [/p]
                        [p]
                            No longer attached to the tree, the beehive plummets to the ground,
                            your powerful kick sending it flying in the direction of the two guards and their [i]pet[/i].
                            Once the nest hits the ground, it cracks open, an entire [b]swarm[/b] of angry bees flying out,
                            assaulting the wolf, the human and the dog.
                            After the initial surprise, the group bolts away as fast as they can to escape the [b]barrage[/b] of painful stings,
                            the feral canine leading the way.
                        [/p]
                        [p]
                            Some of the bees, however, managed to escape their nest [i]before[/i] you kicked it.
                            Despite your best efforts, you yourself get stung more than a few times,
                            although you manage to protect your face and other [i]more sensitive[/i] parts by wildly flailing your arms about to swat the insects away.
                            As your pursuers run into the depths of the forest, concern for their own well-being trumping the hunt for a runaway kobold,
                            you decide it's time to leave [b]before[/b] the rest of the lingering swarm takes an interest in you.
                        [/p]
                        [p]
                            You skillfully hop down a few branches before leaping to the ground below,
                            dropping into an athletic roll to lessen the impact of the heighty jump,
                            the soft forest ground acting as a cushion to soften the blow even more.
                            Scrambling to your feet, you sprint away in the opposite direction the party chasing after you went,
                            running away from the bees while avoiding incidentally bumping into your pursuers still.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.preContinue.continued() ? "disabled" : "";
        var isSelected = ctrl.preContinue.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.preContinue.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.tree.get() !== ForestWoodsTreeChoice.THROW) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.treeThrow.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.tree.get() !== ForestWoodsTreeChoice.THROW) {
            return;
        }

        next.registerListeners();

        if (ctrl.preContinue.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Forest\hide\tree\succeed\throw\ForestWoodsThrow.js

// scenes\Forest\hide\tree\succeed\weather\ForestWoodsWeatherContinue.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsWeatherContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You are stuck in purgatory, between being conscious enough to [b]feel pain[/b],
                            and being too far [b]out of it[/b] to do anything but stare up at the forest canopy.
                            Slowly, but steadily, the bee-venom begins to fade from your system, your hazy mind growing sharper with every passing minute.
                            Your limbs, too, begin to regain their strength, although you doubt they'll [i]ever[/i] [b]fully[/b] return to normal.
                        [/p]
                        [p]
                            About half an hour after falling from the tree, you wearily roll off your aching back onto your relatively unscathed stomach,
                            a breath of cold air running over your body somewhat soothing the [b]burning pain[/b].
                            You begin to pluck stingers from your arms, a great deal of them luckily having fallen off on your way down from the branch.
                            Once your arms are mostly cleared, you toilsomely rise to your feet, your weakened legs barely able to carry your weight.
                        [/p]
                        [p]
                            You cast a final fearful gaze up at the beehive, almost losing your balance in the process.
                            You may not have fully recovered yet, but you [b]really[/b] need to get going.
                            Night will fall soon, and you don't want to be out here when the [b]real[/b] bandits show up.
                            Especially not like this.
                            Let alone while carrying a bundle of gold around your neck.
                            You need to go home.
                        [/p]
                        [p]
                            Step by step, you move your numb legs forwards.
                            It takes a massive amount of concentration just for you to be able to stand upright, let alone walk.
                            With the apitoxin dissipating, your gait swiftly improves from stumbling to limping.
                            It seems as though movement aids your recovery.
                            Mere minutes after you set off, you regain full control of your limbs.
                            While still severely injured, at least you [i]think[/i] you can make it home without collapsing.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.homeSceneName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.preContinue.continued()) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.preContinue.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Forest\hide\tree\succeed\weather\ForestWoodsWeatherContinue.js

// scenes\Forest\hide\tree\succeed\weather\ForestWoodsWeather.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsWeather(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new ForestWoodsWeatherContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.beesting = true;
        var toReturn = `[p]
                            No.
                            You can't move.
                            If you move, they'll see you.
                            Your best bet would be to stay put.
                            Surely, the bees won't sting you unless you look like a threat to them, right?
                        [/p]
                        [p]
                            Wrong.
                            While your pursuers take their time to comb through every single bush in the area,
                            a trickle of bees begins to cover your body, most of them concentrating around the initial prick,
                            some of them landing on your back, and a few even flying out as far as your legs and your feet.
                            You protect your face with your arms,
                            preventing the drones from getting to your snout, or your eyes,
                            the bottom half of your muzzle pressing against the bark of the branch you're laying atop.
                        [/p]
                        [p]
                            At first, the bees seem content in roaming your form, scuttling and scampering across your scaly skin,
                            seemingly judging your motionless frame as [i]harmless[/i].
                            Then, a sudden sting at the back of your head marks the beginning of a veritable [b]onslaught[/b] of stabbing punctures.
                            A single bee driving his barbed stinger into your scales spurs on his comrades,
                            who follow their fellow worker's lead, sacrificing their lives in defence of the hive and the queen from a motionless kobold invader.
                        [/p]
                        [p]
                            All over, you are repeatedly stung, from your forearm to your shoulder, to your back.
                            Even the parts of your head that you fail to cover get viciously punctured, over and over again.
                            The pain is [b]excruciating[/b], small dosages of apitoxin adding up to make your entire body feel like it's [b]on fire[/b].
                            You grit your teeth and silently cry into your arms as you weather the horrible assault,
                            the need to [b]scream[/b] out in [b]agony[/b] rapidly mounting, yet if you do, your pursuers'll find you for sure.
                        [/p]
                        [p]
                            Nearly an entire minute of pure [b]torture[/b] passes before you finally notice the party 
                            and their dog leaving out of the tear-filled corner of your eye.
                            Still, you remain on top of the branch for a little while longer, making sure they've left the area completely.
                            The bees may hurt,
                            but you are [b]far[/b] more scared of the [b]horrible[/b] things [i]they[/i] can inflict upon you than you are of a swarm of insects.
                        [/p]
                        [p]
                            Once you feel like you've hit the absolute [b]limit[/b] of how much pain you can take, you immediately attempt to climb down from the tree.
                            Your muscles weakened by the onslaught of apitoxin-injecting beestings, you find yourself almost [b]unable[/b] to move,
                            your limbs simply [b]failing[/b] to budge, panic mounting rapidly as you fear you might actually [b]die[/b] up here.
                            Scraping together every single last bit of [b]determination[/b] you have left, you [b]barely[/b] manage to roll to the side,
                            your form [b]sliding[/b] off of the thick branch, plummeting downwards, beckoned by the unrelenting pull of gravity.
                        [/p]
                        [p]
                            [b]Tumbling[/b] down the tree, your numb frame painfully hits several branches,
                            adding a new layer of [b]blunt trauma[/b] to your already near-unbearably bee-stung anatomy.
                            At least the multitude of drones still clinging to your scales are scared off by the heavy impacts,
                            the entire swarm, save for a few stragglers, vacating your form and returning to the hive.
                        [/p]
                        [p]
                            After smacking face-first into one final offshoot, you ultimately land on your back on the surprisingly soft ground,
                            in the middle of a blow-dampening patch of grass.
                            You are teetering on the [b]edge[/b] of consciousness, every moment bringing with it fresh [b]anguish[/b].
                            You can't move.
                            You can't think.
                            You [b]wish[/b] those bees had stung you [b]to death[/b] instead of condemning you to this [b]hell[/b].
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.preContinue.continued() ? "disabled" : "";
        var isSelected = ctrl.preContinue.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.preContinue.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.tree.get() !== ForestWoodsTreeChoice.WEATHER) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.treeWeather.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.tree.get() !== ForestWoodsTreeChoice.WEATHER) {
            return;
        }

        next.registerListeners();

        if (ctrl.preContinue.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Forest\hide\tree\succeed\weather\ForestWoodsWeather.js

// scenes\Forest\hide\tree\succeed\ForestWoodsTree.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsTree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstChoice = new ForestWoodsThrow(ctrl);
    var secondChoice = new ForestWoodsWeather(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You frantically look around for the best tree to climb, yet sadly, [b]none[/b] of them seem to have [b]any[/b] low-hanging branches.
                            Not low enough for you to grab onto, at least.
                            Your best bet, or so you figure, is a large, firm looking oak tree with a medium-sized, moss-covered rock right next to it.
                            If you can use the outcrop to give you a boost, you [i]might[/i] just be able to reach high enough to grab an offshoot.
                        [/p]
                        [p]
                            After taking a few steps back, you [b]sprint[/b] towards the oak,
                            accelerating rapidly from standing still to [b]bolting[/b] as fast as you possibly can,
                            gaining momentum with every pace, your sharp claws [b]digging[/b] into the soft dirt below to propel you forwards.
                            Nearing the rock, you leap into the air as high as you can,
                            lifting off with just one foot so you can quickly place the other atop the mossy rock surface,
                            using it to launch yourself even higher.
                        [/p]
                        ${global.ui.condSucc(`No weapon`)}
                        [p]
                            With the additional height gained from boosting off the rock,
                            you [b]barely[/b] manage to grab on to the top of the branch, inertia swinging your body forwards.
                            Once you've stabilized, you quickly pull yourself upwards, lifting your luckily not-all-too-heavy frame up onto the sturdy branch.
                            In retrospect, you're surprised you managed to jump high enough to reach, but you don't have much time to reflect upon your remarkable feat.
                        [/p]
                        [p]
                            You may be off the ground, but you're not in the clear yet.
                            It may have been a big leap for a kobold like you, but larger species can probably reach the branch you're on without even [b]trying[/b].
                            If you stay this low, you'll surely get noticed and caught.
                        [/p]
                        [p]
                            The barking dog nearing, you frantically begin to climb upwards, navigating branches and offshoots, scaling ever-higher.
                            By the time your pursuers come into view, you're high up above, several meters off the ground.
                            You press your body against the thick branch you end up on, concealing yourself as best as you can.
                            As the footsteps below come to an end, and the dog's barking turns to ominous growling,
                            you peek your head out to get a better idea of the situation.
                        [/p]
                        [p]
                            The party chasing you appears to consist of two men.
                            A human and a big, broad anthromorphic wolf.
                            As you suspected ages ago, they are accompanied by a feral guard dog doubling as a tracking hound.
                            The dog is scratching the tree you're in.
                            It clearly knows where you are, yet the ruffians accompanying it remain clueless.
                            You overhear their conversation.
                            "[color=red]Stupid dog.[/color]" "[color=gblue]Lizard can't have gone far. Let's look around.[/color]"
                        [/p]
                        [p]
                            Suddenly, you become aware of a slight [i]buzzing[/i] noise, soon followed by a tingling feeling on your forearm.
                            Tearing your attention away from your pursuers, you notice a large, fuzzy, black-and-yellow [b]bee[/b] sitting directly atop your exposed scales.
                            The motion of your head turning to face it must have scared the insect, because the moment it comes into view,
                            it [b]plunges[/b] down its hind-end, stabbing you with its stinger, which tears loose from the bee's abdomen,
                            sticking out from your scales like a poisoned dart would protrude from an assassin's victim.
                        [/p]
                        [p]
                            You barely manage to refrain from yelping in agony, instead letting out a small, hand-muffled grunt as pain begins to course through your arm.
                            The bee crawls along your scales for a few more inches, before dropping off the side, falling to the ground.
                            Dead.
                            You remove the barbed stinger, and discard it.
                            As the search party below begins to comb through the area, you realize where the annoying insect came from.
                            A honey-filled beehive hanging off the far-end of the very same branch you are on, more drones rapidly buzzing forth from it,
                            undoubtedly attracted by the alarm pheromone released by its singular vanguard.
                        [/p]
                        [p]
                            In other words: the bees are coming for you.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = ctrl.tree.has() ? "disabled" : "";
        var isSelected = ctrl.tree.get() === ForestWoodsTreeChoice.THROW;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled} style="width:160px;">Throw the nest</button>`;

        var tooltip = global.parser.parse(`Grab the hive and throw it at your pursuers. You might just be able to vanish in the ensuing panic.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 160px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = ctrl.tree.has() ? "disabled" : "";
        var isSelected = ctrl.tree.get() === ForestWoodsTreeChoice.WEATHER;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled} style="width:190px;">Weather the stings</button>`;

        var tooltip = global.parser.parse(`Hang on for dear life and hope the men go away before the bees sting you [b]to death[/b].`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 210px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tree.set(ForestWoodsTreeChoice.THROW);
            loadNextPart(firstChoice);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tree.set(ForestWoodsTreeChoice.WEATHER);
            loadNextPart(secondChoice);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.tree.change();
        part += buildChoice();
        part += `</div>`;
        part += firstChoice.build();
        part += secondChoice.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        firstChoice.registerListeners();
        secondChoice.registerListeners();

        if (ctrl.tree.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var ForestWoodsTreeChoice = {
    THROW: 1,
    WEATHER: 2
};

// END scenes\Forest\hide\tree\succeed\ForestWoodsTree.js

// scenes\Forest\hide\tree\ForestWoodsHideTree.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHideTree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new ForestWoodsTreeFail(ctrl);
    var toSuccess = new ForestWoodsTree(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.intro.getItemChoice() === 0;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.hide.get() !== ForestWoodsHideChoice.TREE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.hide.get() !== ForestWoodsHideChoice.TREE) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Forest\hide\tree\ForestWoodsHideTree.js

// scenes\Forest\hide\ForestWoodsHide.js

/**
 * Hide section of the woods.
 * @param {ForestController} controller The Forest controller.
 */
function ForestWoodsHide(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstChoice = new ForestWoodsHideTree(ctrl);
    var secondChoice = new ForestWoodsHideBush(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You make the wise decision to hide from your pursuers.
                            You stole [b]a lot[/b] of money, and you don't think your thievery went entirely unnoticed.
                            If they catch you, who knows what those guards will do?
                            They could execute you on the spot if they so please,
                            or torture you until you [b]wish[/b] they [b]had[/b].
                            There are fates worse than death, which are, sadly enough, often reserved for [i]your kind[/i].
                            You don't want to find out first-hand what those fates entail.
                        [/p]
                        [p]
                            The forest is a big place, with plenty of hidden hidey-holes.
                            It is your home, and you know it inside and out.
                            Well, [i]most[/i] of it anyway.
                            This area in particular, you don't really have any hideaways in.
                            Still, there are plenty of apt places for you to hide.
                            Within a few minutes of exploration,
                            you're bound to find a hollow tree stump or some overgrown roots for you to conceal yourself in.
                        [/p]
                        [p]
                            Right as you were about to [i]start[/i] looking for a hiding spot, you become aware of a faint noise in the distance.
                            Is that...barking?
                            It's getting louder.
                            Definitely a feral dog or a wolf of some kind.
                            You don't think wolves actually bark, though, so it [b]must[/b] be a dog.
                            A tracking dog, perhaps?
                            Surely the caravan guards didn't bring one of those, right?
                            To hunt down sneaky kobold thieves like yourself with?
                            Another bark.
                            It's getting closer.
                            [i]Oh no.[/i]
                        [/p]
                        [p]
                            All you needed was five minutes.
                            Five minutes, and you'd have been able to conceal yourself [b]so well[/b] they would [b]never[/b] have found you.
                            Your mind races a million miles an hour, trying to come up with a [b]backup plan[/b], a place to hide in [b]right now[/b].
                            Alas, there's nothing around but [b]trees[/b] and [b]bushes[/b].
                            Wait a minute.
                            Trees.
                            And bushes.
                            You could hide in those!
                            Better hurry, you doubt you have long before the guards arrive.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = ctrl.hide.has() ? "disabled" : "";
        var isSelected = ctrl.hide.get() === ForestWoodsHideChoice.TREE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>A tree</button>`;

        var tooltip = global.parser.parse(`Try to climb a tree in order to stay hidden.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = ctrl.hide.has() ? "disabled" : "";
        var isSelected = ctrl.hide.get() === ForestWoodsHideChoice.BUSH;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>A bush</button>`;

        var tooltip = global.parser.parse(`Hide in a nearby bush.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You hide in: </div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.hide.set(ForestWoodsHideChoice.TREE);
            loadNextPart(firstChoice);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.hide.set(ForestWoodsHideChoice.BUSH);
            loadNextPart(secondChoice);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.woods.getArrivalChoice() !== ForestWoodsArrivalChoiceEnum.HIDE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstChoice.build();
        part += secondChoice.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.woods.getArrivalChoice() !== ForestWoodsArrivalChoiceEnum.HIDE) {
            return;
        }

        firstChoice.registerListeners();
        secondChoice.registerListeners();

        if (ctrl.hide.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var ForestWoodsHideChoice = {
    TREE: 1,
    BUSH: 2
};

// END scenes\Forest\hide\ForestWoodsHide.js

// scenes\Forest\woods\ForestWoodsController.js

/**
 * This class acts as a dedicated controller for the woods part of the Forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestWoodsController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;
    var arrivalChoiceId = "woodsArrivalChoice";
    var arrivalStatusId = "woodsArrivalStatus";
    var unknownChoiceId = "woodsUnknownChoice";
    var hideChoiceId = "woodsHideChoice";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Makes the arrival choice.
     * @param {ForestWoodsArrivalChoiceEnum} choice The choice that was made.
     */
    this.makeArrivalChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[arrivalChoiceId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns whether or not the arrival choice has been made.
     * @returns {boolean} Whether or not the path choice has been made.
     */
    this.hasMadeArrivalChoice = function () {
        return global.kobold.history.has(id, arrivalChoiceId);
    };

    /**
     * Fetches the arrival choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     * @returns {ForestWoodsArrivalChoiceEnum} The choice that was made.
     */
    this.getArrivalChoice = function () {
        var history = global.kobold.history.get(id);
        return history[arrivalChoiceId];
    };
    
    /**
     * Reduces the kobold's stamina if required.
     * @returns {string} HTML describing the change.
     */
    this.arrivalStatusChange = function () {
        if (!global.kobold.history.has(id, arrivalStatusId)) {
            global.kobold.status.stamina.decrease(10, "Long hike");
            var history = global.kobold.history.get(id);
            history[arrivalStatusId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Hiking", "Stamina", -10)
        ]);
    };
    
    /**
     * Makes the unknown choice.
     * @param {} choice The choice that was made.
     */
    this.makeUnknownChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[unknownChoiceId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns whether or not the unknown choice has been made.
     * @returns {boolean} Whether or not the path choice has been made.
     */
    this.hasMadeUnknownChoice = function () {
        return global.kobold.history.has(id, unknownChoiceId);
    };

    /**
     * Fetches the unknown choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     * @returns {} The choice that was made.
     */
    this.getUnknownChoice = function () {
        var history = global.kobold.history.get(id);
        return history[unknownChoiceId];
    };
    
    /**
     * Makes the hide choice.
     * @param {} choice The choice that was made.
     */
    this.makeHideChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[hideChoiceId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns whether or not the hide choice has been made.
     * @returns {boolean} Whether or not the path choice has been made.
     */
    this.hasMadeHideChoice = function () {
        return global.kobold.history.has(id, hideChoiceId);
    };

    /**
     * Fetches the hide choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     * @returns {} The choice that was made.
     */
    this.getHideChoice = function () {
        var history = global.kobold.history.get(id);
        return history[hideChoiceId];
    };

}

// END scenes\Forest\woods\ForestWoodsController.js

// scenes\Forest\woods\arrival\ForestWoodsArrival.js

/**
 * The path to home of the Forest scene.
 * @param {ForestController} controller The forest controller.
 * @param {ForestWoodsUnknown} toUnknown A part the kobold may go to next.
 * @param {ForestHomeHide} toHide A part the kobold may go to next.
 * @param {ForestBandit} toBandit A part the kobold may go to next.
 */
function ForestWoodsArrival(controller, toUnknown, toHide, toBandit) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    var unknown = toUnknown;
    var hide = toHide;
    var bandit = toBandit;

    var unknownId = "woodsUnknownButton";
    var hideId = "woodsHideButton";
    var knownId = "woodsKnownButton";

    var content = ` <div class="text">
                        You decide to venture deeper into the woods, to try and shake off any would-be pursuers.
                        While you lack a map, you know exactly where you are: the large stretch of woodland between the
                        eastern and the southern roads to Varanar. The southern path is where you ambushed the money transport,
                        and your home lies a little bit off the eastern route. You form a mental image of your surroundings.
                    </div>
                    <div class="text center">
                        <a href="scenes/Forest/woods/arrival/woodsmap.png" target="_blank">
                            <img class="sfw" src="scenes/Forest/woods/arrival/woodsmap.png" title="Made by TinkeringTurian">
                        </a>
                    </div>
                    <div class="text">
                        Since you want to avoid the roads, and you'll just end up getting blocked off by the river if you head northwest, 
                        you wisely head off towards the southeast, into the denser parts of the forest.
                        You walk at a steady pace, sheer determination driving your weary legs to carry your lithe frame throughout the maze
                        of trees.
                    </div>
                    <div class="text">
                        After nearly half an hour of hiking, you take a little break.
                        You've made considerable progress into the forest, and you're nearing parts you've only rarely ventured into.
                        There's no telling what you might find should you press on, but do you really want to go exploring with a large sum of money dangling from your neck?
                        Alternatively, you could keep wandering about, yet restrict yourself to familiar territory.
                        That way, you can stay ahead of whoever may or may not be chasing you, without risking being unable to find your way home once the sun sets.
                        Or you could just stop walking altogether, and try to conceal yourself as best you can in the area you are in right now.
                    </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the main content of this section.
     * @returns {string} The main content of this section.
     */
    function buildContent() {
        return content;
    }

    /**
     * Builds and returns the investigate choice button.
     */
    function buildUnknownChoice() {
        var disabled = ctrl.woods.hasMadeArrivalChoice() ? "disabled" : "";
        var isSelected = ctrl.woods.getArrivalChoice() == ForestWoodsArrivalChoiceEnum.UNKNOWN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${unknownId}" ${disabled} style="font-size: 18px;">Venture into the unknown</button>`;

        var tooltip = `Head into parts of the forest you've never really been to before.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the hide choice button.
     */
    function buildHideChoice() {
        var disabled = ctrl.woods.hasMadeArrivalChoice() ? "disabled" : "";
        var isSelected = ctrl.woods.getArrivalChoice() == ForestWoodsArrivalChoiceEnum.HIDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${hideId}" ${disabled}>Hide</button>`;

        var tooltip = `Hide from the men who may or may not be pursuing you.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the ignore choice button.
     */
    function buildKnownChoice() {
        var disabled = ctrl.woods.hasMadeArrivalChoice() ? "disabled" : "";
        var isSelected = ctrl.woods.getArrivalChoice() == ForestWoodsArrivalChoiceEnum.KNOWN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${knownId}" ${disabled} style="font-size: 18px;">Venture into the known</button>`;

        var tooltip = `Explore relatively known parts of the forest and attempt to find something new.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildUnknownChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildHideChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildKnownChoice();
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + unknownId), $("#" + hideId), $("#" + knownId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {ForestWoodsArrivalChoiceEnum} next Which part to load.
     */
    function loadNextPart(next) {
        var part;
        if (next === ForestWoodsArrivalChoiceEnum.UNKNOWN) {
            part = unknown;
        }
        if (next === ForestWoodsArrivalChoiceEnum.HIDE) {
            part = hide;
        }
        if (next === ForestWoodsArrivalChoiceEnum.KNOWN) {
            part = bandit;
        }
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the unknown button.
     */
    function registerUnknownListener() {
        $("#" + unknownId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.woods.makeArrivalChoice(ForestWoodsArrivalChoiceEnum.UNKNOWN);
            loadNextPart(ForestWoodsArrivalChoiceEnum.UNKNOWN);
        });
    }

    /**
     * Adds a listener to the known button.
     */
    function registerHideListener() {
        $("#" + hideId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.woods.makeArrivalChoice(ForestWoodsArrivalChoiceEnum.HIDE);
            loadNextPart(ForestWoodsArrivalChoiceEnum.HIDE);
        });
    }

    /**
     * Adds a listener to the known button.
     */
    function registerKnownListener() {
        $("#" + knownId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.woods.makeArrivalChoice(ForestWoodsArrivalChoiceEnum.KNOWN);
            ctrl.bandit.setSource(ForestBanditSourceEnum.WOODS);
            loadNextPart(ForestWoodsArrivalChoiceEnum.KNOWN);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "woodsArrival";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var change = ctrl.woods.arrivalStatusChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `</div>`;
        part += bandit.build();
        part += unknown.build();
        part += hide.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        unknown.registerListeners();
        bandit.registerListeners();
        hide.registerListeners();

        if (ctrl.woods.hasMadeArrivalChoice()) {
            return;
        }

        registerUnknownListener();
        registerKnownListener();
        registerHideListener();
    };
}

/**
 * Path choices to be made.
 */
var ForestWoodsArrivalChoiceEnum = {
    UNKNOWN: 1,
    HIDE: 2,
    KNOWN: 3
};

// END scenes\Forest\woods\arrival\ForestWoodsArrival.js

// scenes\Forest\woods\ForestWoods.js

/**
 * The woods section of the Forest scene.
 * @param {ForestController} controller The forest controller.
 * @param {ForestBandit} bandit One of the parts the kobold can go next.
 */
function ForestWoods(controller, bandit) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var hide = new ForestWoodsHide(ctrl);
    var unknown = new ForestUnknown(ctrl);
    var arrival = new ForestWoodsArrival(ctrl, unknown, hide, bandit);
    
    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    
    
    // PUBLIC MEMBERS \\        this.public = "test";

    // The id of this segment.
    this.id = "forestWoods";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
    
    /**
     * Builds and returns this part.
     */
    this.build = function() {
        if (!ctrl.intro.hasMadePathChoice()) {
            return "";
        }
        if (ctrl.intro.getPathChoice() !== ForestIntroductionPathChoiceEnum.WOODS) {
            return "";
        }
        var woods = `<div id="${that.id}">`;
        woods += `<div class="subtitle">Into the woods</div>`;
        woods += arrival.build();
        woods += `</div>`;
        return woods;
    };
    
    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function() {
        if (!ctrl.intro.hasMadePathChoice()) {
            return;
        }
        if (ctrl.intro.getPathChoice() !== ForestIntroductionPathChoiceEnum.WOODS) {
            return;
        }
        arrival.registerListeners();
    };
}

// END scenes\Forest\woods\ForestWoods.js

// scenes\Forest\introduction\ForestIntroductionController.js

/**
 * This class acts as a dedicated controller for the introduction of the Forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestIntroductionController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;

    var thinkingChoice = "introRec";
    var recoveringStamina = "introRecStamina";
    var recoveringChestCoveredId = "introChest";
    var itemChoice = "introItem";
    var itemSharedId = "introItemShared";
    var itemChoiceStamina = "itemChoiceStamina";
    var itemChoiceBaseGender = "itemChoiceBaseGender";
    var pathChoice = "introPath";
    var pathChoiceStamina = "pathChoiceStamina";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Adds equipment to the kobold depending on the chosen item.
     * @param {GenderEnum} choice The selected item.
     */
    function itemChoiceEquipment(choice) {
        if(choice === 0) {
            return;
        }
        global.kobold.inventory.equipment.wear(choice);
    }

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Makes the thinking choice.
     * @param {ForestIntroductionThinkingChoiceEnum} choice The selected item.
     */
    this.makeThinkingChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[thinkingChoice] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the thinking choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     */
    this.getThinkingChoice = function () {
        var history = global.kobold.history.get(id);
        return history[thinkingChoice];
    };

    /**
     * Returns true if the thinking choice has already been made. False otherwise.
     */
    this.hasMadeThinkingChoice = function () {
        return global.kobold.history.has(id, thinkingChoice);
    };

    /**
     * Makes the item choice, giving the kobold a weapon.
     * @param {Item} choice The selected item.
     */
    this.makeItemChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[itemChoice] = choice;
        global.kobold.history.set(id, history);
        itemChoiceEquipment(choice);
    };

    /**
     * Fetches the item choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     */
    this.getItemChoice = function () {
        var history = global.kobold.history.get(id);
        return history[itemChoice];
    };

    /**
     * Returns true if the item choice has already been made. False otherwise.
     */
    this.hasMadeItemChoice = function () {
        return global.kobold.history.has(id, itemChoice);
    };
    
    /**
     * Makes the item shared changes.
     */
    this.itemShared = function () {
        global.kobold.history.doIfNotSet(id, itemSharedId, true, function () {
            if(that.getThinkingChoice() === ForestIntroductionThinkingChoiceEnum.PRESENT) {
                global.kobold.history.shared.caravanBandits.knowledge = CaravanBanditsKnowledgeEnum.SUSPICIOUS;
            }
        });
    };

    /**
     * Makes the path choice.
     * @param {ForestIntroductionPathChoiceEnum} choice The selected item.
     */
    this.makePathChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[pathChoice] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Fetches the path choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     */
    this.getPathChoice = function () {
        var history = global.kobold.history.get(id);
        return history[pathChoice];
    };

    /**
     * Returns true if the path choice has already been made. False otherwise.
     */
    this.hasMadePathChoice = function () {
        return global.kobold.history.has(id, pathChoice);
    };

    /**
     * Reduces the kobold's stamina and increases its sanity if required.
     * @returns {string} HTML describing the change.
     */
    this.recoveringStatusChange = function () {
        if (!global.kobold.history.has(id, recoveringStamina)) {
            global.kobold.status.stamina.decrease(60, "Running for your life");
            global.kobold.status.mental.decrease(10, "Close to passing out");
            global.kobold.status.physical.decrease(10, "Overexerted");
            global.kobold.inventory.coins += 203;
            var history = global.kobold.history.get(id);
            history[recoveringStamina] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Overexerted", "Health", -10),
            new Change("Close to passing out", "Sanity", -10),
            new Change("Out of breath", "Stamina", -60)
        ]);
    };

    /**
     * Returns whether or not the kobold's chest was covered at the recovery-phase.
     * @returns {string} Whether or not the kobold's chest was covered at the recovery-phase.
     */
    this.recoveringChestCovered = function () {
        if (!global.kobold.history.has(id, recoveringChestCoveredId)) {
            var history = global.kobold.history.get(id);
            history[recoveringChestCoveredId] = !global.kobold.inventory.equipment.chestExposed();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[recoveringChestCoveredId];
    };
    
    /**
     * Returns the kobold's base gender at the item choice phase.
     * @returns {GenderEnum} The kobold's base gender at the item choice phase.
     */
    this.itemChoiceBaseGEnder = function () {
        if (!global.kobold.history.has(id, itemChoiceBaseGender)) {
            var history = global.kobold.history.get(id);
            history[itemChoiceBaseGender] = global.kobold.body.baseGender();
            global.kobold.history.set(id, history);
        }
        return global.kobold.history.get(id)[itemChoiceBaseGender];
    };

    /**
     * Increase the kobold's stamina if required.
     * @returns {string} HTML describing the change.
     */
    this.itemSelectionStatusChange = function () {
        if (!global.kobold.history.has(id, itemChoiceStamina)) {
            global.kobold.status.stamina.increase(10, "Recovering");
            global.kobold.status.mental.increase(10, "Calmed down");
            global.kobold.status.physical.increase(5, "Recovering");
            var history = global.kobold.history.get(id);
            history[itemChoiceStamina] = true;
            global.kobold.history.set(id, history);
        }
        ``
        return global.ui.change.buildNotification([
            new Change("Recovering", "Health", +5), 
            new Change("Calmed down", "Sanity", +10), 
            new Change("Recovering", "Stamina", +10)
        ]);
    };

    /**
     * Increase the kobold's stamina if required.
     * @returns {string} HTML describing the change.
     */
    this.pathSelectionStatusChange = function () {
        if (!global.kobold.history.has(id, pathChoiceStamina)) {
            global.kobold.status.stamina.increase(10, "Recovering");
            global.kobold.status.mental.increase(5, "Coin-laden optimism");
            global.kobold.status.physical.increase(5, "Recovering");
            var history = global.kobold.history.get(id);
            history[pathChoiceStamina] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Recovering", "Health", +5), 
            new Change("Cautious optimism", "Sanity", +5), 
            new Change("Recovering", "Stamina", +10)
        ]);
    };

}

// END scenes\Forest\introduction\ForestIntroductionController.js

// scenes\Forest\introduction\ForestIntroductionPathSelection.js

/**
 * The third part of the introduction of the Forest scene.
 * @param {ForestController} controller The forest controller.
 * @param {ForestHome} home One of the parts the kobold can go next.
 * @param {ForestBridge} bridge One of the parts the kobold can go next.
 * @param {ForestWoods} woods One of the parts the kobold can go next.
 */
function ForestIntroductionPathSelection(controller, home, bridge, woods) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var bridgeId = "bridgeButton";
    var woodsId = "woodsButton";
    var homeId = "homeButton";
    
    var toHome = home;
    var toBridge = bridge;
    var toWoods = woods;

    var contentStart = `<div class="text">`;
    var rock = `You choose the ${global.kobold.inventory.createItemWithTooltip(new Rock())}.
                A hardy, dependable weapon your kind has been using since the dawn of ages.
                It might not be fancy and it may not be sharp, yet a good rock is all you need to bash someone's head in.
                Should the situation call for it, you could even throw it instead of trying to get up close and personal with someone twice your size.
                Which is mostly everyone, apart from other kobolds.`;
    var branch = `You choose the ${global.kobold.inventory.createItemWithTooltip(new Branch())}.
                  While a rather unorthodox weapon, this heavy stick can be used as a makeshift club.
                  You've heard of elves in a distant land fashioning sharp <span class="bold">swords</span> from wood like this.
                  If you had an entire workshop's worth of tools at hand, and a couple of days to fuck around with them, 
                  you're certain you could at the very least add something sharp and pointy to the hitting end.`;
    var neither = `You don't need a weapon to defend yourself.
                   The added weight would just slow you down.
                   Even if you'd picked either the rock or the branch, you wouldn't have been able to utilize either of them properly.
                   When you were little, you liked to fool around with the other young kobolds, smacking eachother with sticks and pretending you were brave, gallant knights.
                   Nowadays, you're old enough to realize that fighting of any kind is not, and never will be, your strong suit.
                   Unless some sort of miracle happens, that is.`;
    var rockOrBranchEnd = ` </div>
                            <div class="text">
                            You swing your new weapon around a bit, to get a feel for its weight. 
                            While you're certain you can cause some serious damage if you actually manage to strike someone with it, 
                            you heavily doubt your ability to land a hit in the heat of combat.`;
    var neitherEnd = `  </div>
                        <div class="text">
                            Still, that doesn't mean you're <span class="bold">completely</span> defenceless.
                            Your keen claws and razor-sharp teeth can rip through flesh as easily as they can shred clothing.
                            While you lack a weapon, the one thing you have plenty of is bravery.`;
    var contentEnd = `  </div>
                        <div class="text">
                            The issue of your personal safety resolved, you figure it's about time you found out <span class="bold">exactly</span> how much money you stole.
                            Your heart races as you lift the coin pouch's neck-strap over your head, and loosen the cord tying the top-end shut.
                            The moment the small bag opens, you are greeted with the warm glow of sunrays reflecting off the sparkling coins within, 
                            shining directly onto your face.
                        </div>
                        <div class="text">
                            You breathe an audible sigh of relief. 
                            The bag is filled to the brim with gold, alright.
                            Slightly awestruck, you waddle over to a nearby moss-covered rock.
                            Emptying the contents of the pouch atop the green-cushioned ancient,
                            you set about trying to count it all, one piece at a time.
                        </div>
                        <div class="text nomargin">
                            Two hundred and one...
                        </div>
                        <div class="text nomargin">
                            Two hundred and two...
                        </div>
                        <div class="text nomargin">
                            Two hundred and three!
                        </div>
                        <div class="text">
                            <span class="bold">You've managed to steal two hundred and three coins!</span>
                        </div>
                        <div class="text">
                            At first you get excited. 
                            <span class="bold">Really</span> excited.
                            You let out a rather cute, happy squeal.
                            Luckily enough, nobody was around to hear it.
                            For the first time in weeks, a content, genuinely happy smile forms at the end of your muzzle.
                            Then, the initial pleasant surprise fades, and more worries begin to flood your mind.
                            Two hundred and three coins is a lot of money, but it's still not enough to <span class="bold">completely</span> bail you out of debt.
                            But hey, at least it's more than nothing, right?
                        </div>
                        <div class="text">
                            You put the coins back into the pouch, which you secure around your neck.
                            Looking over your surroundings, you realize exactly where you are: 
                            at the edge of the forest, close to both Varanar as well as your own home.
                            It's time to decide where to go next.
                        </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds the rock part of the item selection.
     * @returns {string} The past part of the item selection.
     */
    function buildRock() {
        return rock + rockOrBranchEnd;
    }

    /**
     * Builds the branch part of the item selection.
     * @returns {string} The present part of the item selection.
     */
    function buildBranch() {
        return branch + rockOrBranchEnd;
    }

    /**
     * Builds the neither part of the item selection.
     * @returns {string} The future part of the item selection.
     */
    function buildNeither() {
        return neither + neitherEnd;
    }

    /**
     * Builds and returns the main content of this section.
     * @returns {string} The main content of this section.
     */
    function buildContent() {
        var content = contentStart;
        var previousChoice = ctrl.intro.getItemChoice();
        if (previousChoice === ItemEnum.ROCK) {
            content += buildRock();
        }
        if (previousChoice === ItemEnum.BRANCH) {
            content += buildBranch();
        }
        if (previousChoice === 0) {
            content += buildNeither();
        }
        content += contentEnd;
        return content;
    }

    /**
     * Builds and returns the Varanar choice button.
     * @returns {string} The Varanar choice button.
     */
    function buildVaranarChoice() {
        var disabled = ctrl.intro.hasMadePathChoice() ? "disabled" : "";
        var isSelected = ctrl.intro.getPathChoice() == ForestIntroductionPathChoiceEnum.BRIDGE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${bridgeId}" ${disabled}>Varanar</button>`;

        var popupMessage = `You know a small, hidden path that leads towards the capital. You should go pay off your debt immediately. <span class="todo">UNFINISHED</span>`;
        var popupContainer = global.ui.makeToolTip(button, popupMessage, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the woods choice button.
     * @returns {string} The woods choice button.
     */
    function buildWoodsChoice() {
        var disabled = ctrl.intro.hasMadePathChoice() ? "disabled" : "";
        var isSelected = ctrl.intro.getPathChoice() == ForestIntroductionPathChoiceEnum.WOODS;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${woodsId}" ${disabled}>The woods</button>`;

        var popupMessage = `Those soldiers could still be on your tail. You should hide from them somewhere in the woods.`;
        var popupContainer = global.ui.makeToolTip(button, popupMessage, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the home choice button.
     * @returns {string} The home choice button.
     */
    function buildHomeChoice() {
        var disabled = ctrl.intro.hasMadePathChoice() ? "disabled" : "";
        var isSelected = ctrl.intro.getPathChoice() == ForestIntroductionPathChoiceEnum.HOME;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${homeId}" ${disabled}>Home</button>`;

        var popupMessage = "You've had a long, emotion-filled day. It's time to go home. If a debt collector comes by, you can pay him all you have.";
        var popupContainer = global.ui.makeToolTip(button, popupMessage, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You set off towards:</div>`;
        choice += buildVaranarChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildWoodsChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildHomeChoice();
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + homeId), $("#" + bridgeId), $("#" + woodsId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     * @param {ForestHomePathChoiceEnum} next Which part to load.
     */
    function loadNextPart(next) {
        var part;
        if(next === ForestIntroductionPathChoiceEnum.HOME) {
            part = toHome;
        }
        if(next === ForestIntroductionPathChoiceEnum.BRIDGE) {
            part = toBridge;
        }
        if(next === ForestIntroductionPathChoiceEnum.WOODS) {
            part = toWoods;
        }
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }
    
    /*
     * Registers a listener for the woods button.
     */
    function registerWoodsListener() {
        $("#" + woodsId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.intro.makePathChoice(ForestIntroductionPathChoiceEnum.WOODS);
            loadNextPart(ForestIntroductionPathChoiceEnum.WOODS);
        });
    }
    
    /*
     * Registers a listener for the home button.
     */
    function registerHomeListener() {
        $("#" + homeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.intro.makePathChoice(ForestIntroductionPathChoiceEnum.HOME);
            loadNextPart(ForestIntroductionPathChoiceEnum.HOME);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "introductionPath";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.intro.hasMadeItemChoice()) {
            return "";
        }
        var statChange = ctrl.intro.pathSelectionStatusChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += statChange;
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.intro.hasMadeItemChoice()) {
            return;
        }
        if (ctrl.intro.hasMadePathChoice()) {
            return;
        }
        registerWoodsListener();
        registerHomeListener();
    };
}

/**
 * Path choices to be made.
 */
var ForestIntroductionPathChoiceEnum = {
    HOME: 1,
    BRIDGE: 2,
    WOODS: 3
};

// END scenes\Forest\introduction\ForestIntroductionPathSelection.js

// scenes\Forest\introduction\ForestIntroductionItemSelection.js

/**
 * The second part of the introduction of the Forest scene.
 * @param {ForestController} controller The forest controller.
 * @param {ForestIntroductionPathSelection} nextPart The next part to load once finished.
 */
function ForestIntroductionItemSelection(controller, nextPart) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    var next = nextPart;

    var rockId = "introRock";
    var branchId = "introBranch";
    var neitherId = "introNeither";

    var contentStart = `<div class="text">
                            You take some deep breaths.
                            You calm down a little.
                            You close your eyes. `;
    var past = `            Your mind wanders to your past.
                            A couple of years ago, you were in a similar situation as the one you find yourself in today.
                            Not quite as severe, but somewhat of the same nature.
                        </div>
                        <div class="text">
                            You couldn't pay your monthly fee.
                            You were terrified. 
                            Scared shitless of what might happen without the <span class="italic">protection</span> of the <span class="bold">Three-Tailed Rat Clan</span>.
                            The day the money was due, you marched into town with your tail between your legs.
                            You were granted an audience with the Rat's second-in-command.
                            You grovelled, you begged and you reluctantly humiliated yourself in front of him.
                        </div>
                        <div class="text">
                            Eventually, after humoring the kobold with your <span class="italic">patheticness</span> for the better half of an hour, 
                            he tentatively accepted <span class="bold">your failure</span> to provide the Rat with his <span class="bold">just due</span>.
                            You were let off the hook with a slap on the wrists, a thoroughly dented sense of self-esteem, and a twenty percent interest rate.
                        </div>
                        <div class="text">
                            You owe far more than fifty coins this time around, 
                            the Rat himself is personally involved in the matter, 
                            <span class="bold">and</span> you won't be able to make it to the city before your deadline expires,
                            but perhaps it might still be in your best interest to visit the gang's headquarters as soon as possible.
                            After all, you've just acquired an entire pouch filled with gold.
                            Surely that will be enough to at least buy you some time, right?
                        </div>`;
    var present = `         Your mind wanders to the men you stole from.
                            You had a chance to inspect them before making your move.
                            There was definitely something odd about them, but you can't quite put your finger on it.
                        </div>
                        <div class="text">
                            It was clear the caravan was an imperial money transport, all the animals branded with the seal of the king himself.
                            Probably taxes or something.
                            The footmen, however, looked <span class="bold">nothing</span> like <span class="bold">any</span> of the soldiers or guards you'd ever seen before.
                            Instead of a proper, well-maintained combat cuirass, many of them wore leather armor.
                            Only the overseer wore a distinctive imperial tunic.
                            Heck, some of them were clothed in not much more than the scraggy rags you're wearing.
                        </div>
                        <div class="text">
                            Their attitude, too, was very unbecoming of the king's finest.
                            A few of them were openly guzzling booze from animal-leather drinking pouches.
                            At least you assume it was some kind of alcoholic beverage, because no healthy man should wobble on his feet as he lazily walks alongside a slow-moving cart.
                        </div>
                        <div class="text">
                            While the drunkards were somewhat pacified from the inebriating brews they drank, the rest seemed like a rowdy bunch.
                            They were constantly either laughing at some crude joke, punching eachother's arms or measuring weapon sizes against one another.
                            When one of them spotted a feral brown bear scurrying off into the forest just to <span class="bold">avoid</span> the unpleasant group, 
                            the burly man instantly let out a rallying cry, and in the flash of an eye they all began to chase after it like a bunch of dimwits, their weapons drawn.
                        </div>
                        <div class="text">
                            If you didn't know any better, you'd almost come to think these men were brigands themselves.
                            But why would a bunch of bandits be escorting the king's gold?
                            It doesn't make any sense to you.
                            They couldn't have been keeping the caravan master hostage.
                            He'd have ran away the moment they turned their backs.
                            Nor could they have been guiding the gold elsewhere. The road they were on led to Varanar, and nowhere <span class="bold">but</span> Varanar.
                        </div>
                        <div class="text">
                            You decide not to break your head over this conundrum.
                            The one thing you can make from all of this, is that you should be extra careful around those guys.
                            While you're unsure whether or not they even <span class="bold">noticed</span> the missing bag, one of them <span class="italic">might</span> have seen you slip away from the caravan.
                            They might decide to <span class="italic">hunt</span> you, much like they did the bear.
                            And should these men manage to catch you, you sincerely doubt they'll punish your crime of theft according to law.
                        </div>`;
    var future = `          You turn your mind to your future, which looks a lot brighter now than it did just an hour ago.
                            Depending on the exact contents of the pouch around your neck, you may have bailed yourself out of trouble completely.
                        </div>
                        <div class="text">
                            You've never actually <span class="italic">had</span> more than a hundred coins on you at any given time, 
                            so you can't quite make out how much money this is by its weight alone.
                            Still, unless the bag is weighed down with rocks or something, it's bound to be <span class="bold">a lot</span>.
                            You'll count it properly once you've regained your composure.
                        </div>
                        <div class="text">
                            You need a few more moments to calm down further.
                            That's what you tell yourself, at least.
                            The truth is that you want to dream a bit more, to forget all about the harsh reality of things for a minute longer.
                            You imagine yourself proudly marching home, awaiting the debt collector the Rat has undoubtedly sent your way and
                            handing him whatever you owe with a shit-eating grin on your face, plus a little extra for himself, of course.
                        </div>
                        <div class="text">
                            Then, you'll take the rest of the coin to Varanar, 
                            treat yourself to the <span class="bold">biggest</span>, <span class="bold">juiciest</span> piece of meat you can find
                            and sleep at the most luxurious inn in town.
                            You can practically feel the warm blankets surrounding you, and taste the succulent steak served by a couple of scantily-clad vixens.
                            That is, until all too soon, the doubt and concern growing at the back of your head put a rapid end to your daydream.
                            You snap back to reality.
                        </div>`;

    var contentEnd = `  <div class="text">
                            Your breath stabilizes.
                            You slowly open your eyes once more.
                            You wipe the sweat from your brow.
                            You have calmed down from your hyperventilation, hysteria and near-panic.
                            Your legs are feeling a little better.
                            You wearily pull yourself to your feet.
                        </div>
                        <div class="text">
                            As you take a few moments to compose yourself further, a second wave of worry washes over you.
                            You are weak, almost to the point of being defenceless, and you have a large amount of money visibly strapped around your neck.
                            You present the perfect opportunity for any would-be bandit you cross to get a fat payday without having to put in all that much effort.
                            You need a weapon. 
                            Something to protect yourself with. 
                            You scrounge through some nearby bushes, and come up with two suitable candidates.
                        </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds the past part of the item selection.
     * @returns {string} The past part of the item selection.
     */
    function buildPast() {
        return past;
    }

    /**
     * Builds the present part of the item selection.
     * @returns {string} The present part of the item selection.
     */
    function buildPresent() {
        return present;
    }

    /**
     * Builds the future part of the item selection.
     * @returns {string} The future part of the item selection.
     */
    function buildFuture() {
        return future;
    }

    /**
     * Builds and returns the main content of this section.
     * @returns {string} The main content of this section.
     */
    function buildContent() {
        var content = contentStart;
        var previousChoice = ctrl.intro.getThinkingChoice();
        if (previousChoice == ForestIntroductionThinkingChoiceEnum.PAST) {
            content += buildPast();
        }
        if (previousChoice == ForestIntroductionThinkingChoiceEnum.PRESENT) {
            content += buildPresent();
        }
        if (previousChoice == ForestIntroductionThinkingChoiceEnum.FUTURE) {
            content += buildFuture();
        }
        content += contentEnd;
        return content;
    }

    /**
     * Builds and returns the neither choice button.
     */
    function buildRockChoice() {
        var disabled = ctrl.intro.hasMadeItemChoice() ? "disabled" : "";
        var isSelected = ctrl.intro.getItemChoice() == ItemEnum.ROCK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${rockId}" ${disabled}>A rock</button>`;

        var popupContainer = global.kobold.inventory.createItemTooltip(button, new Rock());
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;
        
        return button;
    }

    /**
     * Builds and returns the branch choice button.
     */
    function buildBranchChoice() {
        var disabled = ctrl.intro.hasMadeItemChoice() ? "disabled" : "";
        var isSelected = ctrl.intro.getItemChoice() == ItemEnum.BRANCH;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${branchId}" ${disabled}>A branch</button>`;

        var popupContainer = global.kobold.inventory.createItemTooltip(button, new Branch());
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the neither choice button.
     */
    function buildNeitherChoice() {
        var disabled = ctrl.intro.hasMadeItemChoice() ? "disabled" : "";        
        var isSelected = ctrl.intro.getItemChoice() === 0;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${neitherId}" ${disabled}>Neither</button>`;
        
        var popupContainer = global.ui.makeToolTip(button, `You don't need any weapons. They'd just slow you down. You're dead meat if you get into a fight either way.`);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;
        
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You take:</div>`;
        choice += buildRockChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildBranchChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildNeitherChoice();
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + rockId), $("#" + branchId), $("#" + neitherId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        var part = next.build(true);
        global.ui.appendBeforeFinisher(part, global.ui.genericHeaderId);
        $("#" + next.id).fadeIn(400);
        next.registerListeners();
    }
    
    /**
     * Adds a listener to the rock button.
     */
    function registerRockListener() {
        $("#" + rockId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.intro.makeItemChoice(ItemEnum.ROCK);
            loadNextPart();
        });
    }
    
    /**
     * Adds a listener to the branch button.
     */
    function registerBranchListener() {
        $("#" + branchId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.intro.makeItemChoice(ItemEnum.BRANCH);
            loadNextPart();
        });
    }
    
    /**
     * Adds a listener to the neither button.
     */
    function registerNeitherListener() {
        $("#" + neitherId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.intro.makeItemChoice(0);
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "introductionItem";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.intro.hasMadeThinkingChoice()) {
            return "";
        }
        ctrl.intro.itemShared();
        var change = ctrl.intro.itemSelectionStatusChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.intro.hasMadeThinkingChoice()) {
            return;
        }
        if (ctrl.intro.hasMadeItemChoice()) {
            return;
        }
        registerRockListener();
        registerBranchListener();
        registerNeitherListener();
    };
}

// END scenes\Forest\introduction\ForestIntroductionItemSelection.js

// scenes\Forest\introduction\ForestIntroductionRecovering.js

/**
 * The first part of the Forest scene.
 * @param {ForestController} controller The forest controller.
 * @param {ForestIntroductionItemSelection} nextPart The next part to load once finished.
 */
function ForestIntroductionRecovering(controller, nextPart) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    var next = nextPart;

    var pastId = "introPast";
    var presentId = "introPresent";
    var futureId = "introFuture";

    var contentStart = `<div class="subtitle">
                            Introduction
                        </div>
                        <div class="text">
                            You ran. 
                            You ran as fast as you could, for as long as you could.
                            You ran until your lungs began to burn, and your legs felt like they were about to explode.
                            You sprinted through the woods, ill-gotten gains in hand.
                            You don't think anyone saw you.
                            Oh god, you <span class="bold">hope</span> nobody saw you.
                        </div>
                        <div class="text">
                            Coming to a rest against a nearby tree, you pant, huff and wheeze loudly, practically gasping for air.
                            With one hand you clutch your `;

    var contentMid = ` With the other, you grope the hefty coin pouch hanging around your neck, 
                            the myriad of shiny gold pieces within clattering against one another in a soul-mending harmony.
                            Yet your body aches just as much as your spirit is recovering.
                        </div>
                        <div class="text">
                            You finally regain enough breath to <span class="bold">think</span>.
                            Your heart is racing, pounding so hard it makes your entire body shiver.
                            You can barely keep your hands from shaking.
                            A last tidbit of adrenaline courses through your system.
                            You look to the left. You look to the right. Nobody.
                            You did it. 
                            You actually pulled it off. 
                            You managed to steal from one of the king's highly secure money transports.
                        </div>
                        <hr class="smallseparator">
                        <div class="text">
                            In your desperation to meet your deadline, to stave off your impending doom,
                            the rewards seemed to far outweigh the risk.
                            The small convoy of gold-filled-coffer-carrying riding horses was momentarily
                            abandoned by the footmen guarding it, the mixture of anthros and humans rushing off 
                            to test their mettle against a big, brown forest bear.
                            With only the caravan master, perched atop the largest carriage at the very front of the 
                            transport left, you decided to make your move.
                        </div>
                        <div class="text">
                            You began at the very rear end of the convoy, a myriad of beasts of burden between you and the overseer.
                            At first, you went for the coffers, small chests filled to the brim with coin. 
                            They were locked, and far too heavy for you to carry.
                            To your despair, nearly everything of value was tucked away in those crates, where greedy hands like yours couldn't reach.
                            Moving forward along the length of the transport, you tested chest after chest, hoping they might have forgotten to lock at least <span class="bold">one</span> of them.
                            No success.
                        </div>
                        <div class="text">
                            You were about to give up hope once more, 
                            to give in to your fears and run away with nothing but your undoubtedly short-lived life to show for your daring venture.
                            Then, out of the corner of your eye, you witnessed your salvation.
                            While the rest of the caravan was indeed too secure for you to tinker with, 
                            several small bags of gold were tied to the side of the foreman's carriage, 
                            kept safe only by the watchful eye of the burly human himself.
                        </div>
                        <div class="text">
                            You snuck in between the legs of the <span class="sneakytext bold">visibly aroused</span> riding horse directly behind the overseer's wagon.
                            You inched closer and closer, crawling flat on your belly through the slightly-trampled dirt and mud that lined the small road.
                            Just as you were steeling yourself, bracing your mind for actually <span class="bold">going through</span> with this, distant voices
                            rang out from the rear of the convoy.
                            The footmen had returned.
                        </div>
                        <div class="text">
                            Panic surged through your mind.
                            If you'd attempted to stay hidden, they would have found you.
                            And even if they wouldn't have, the caravan would surely have trampled you.
                            A primal <span class="italic">fight or flight</span> reaction triggering in your brain, you hopped to your feet.
                            It felt as though your body was acting out of its own, against your better judgement.
                            In a spur of the moment, you reached for the carriage, yanking a single money-laden bag off the side.
                            With the pouch in hand, you bolted off the path, into the heart of the forest.
                            And the rest is history.
                        </div>
                        <hr class="smallseparator">
                        <div class="text">
                            As adrenaline slowly starts to fade from your body, you begin to feel incredibly lightheaded. 
                            You drop to your knees, your heavily overexerted legs far too weary to carry your meagre frame.
                            Your tongue lolls out of your mouth obscenely.
                            Your entire body feels like it's on fire.
                            You squeeze the coin pouch once more.
                            You still can't believe it's there.
                        </div>
                        <div class="text">
                            You pushed yourself <span class="bold">too</span> hard.
                            You feel like you're about to pass out.
                            You tilt your head downwards, staring blankly at the root of the tree in front of you.
                            You can't even <span class="bold">see</span> clearly anymore.
                            Blackness seeps into the corners of your eyes.
                            The rest of your vision is blurry.
                            You can't afford to pass out now.
                            You need something to think of, something to focus on, something to keep you from <span class="bold">falling apart</span>.
                        </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the main content of this section.
     * @returns {string} The main content of this section.
     */
    function buildContent() {
        var content = contentStart;
        content += ctrl.intro.recoveringChestCovered() ? "sweaty chest rag." : "bare chest.";
        content += contentMid;
        return content;
    }

    /**
     * Builds and returns the past choice button.
     */
    function buildPastChoice() {
        var disabled = ctrl.intro.hasMadeThinkingChoice() ? "disabled" : "";
        var isSelected = ctrl.intro.getThinkingChoice() == ForestIntroductionThinkingChoiceEnum.PAST;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${pastId}" ${disabled}>The past</button>`;
        return button;
    }

    /**
     * Builds and returns the present choice button.
     */
    function buildPresentChoice() {
        var disabled = ctrl.intro.hasMadeThinkingChoice() ? "disabled" : "";
        var isSelected = ctrl.intro.getThinkingChoice() == ForestIntroductionThinkingChoiceEnum.PRESENT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${presentId}" ${disabled}>The present</button>`;
        return button;
    }

    /**
     * Builds and returns the future choice button.
     */
    function buildFutureChoice() {
        var disabled = ctrl.intro.hasMadeThinkingChoice() ? "disabled" : "";
        var isSelected = ctrl.intro.getThinkingChoice() == ForestIntroductionThinkingChoiceEnum.FUTURE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${futureId}" ${disabled}>The future</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You think of:</div>`;
        choice += buildPastChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildPresentChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFutureChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Loads and displays the next part.
     */
    function loadNextPart() {
        var part = next.build(true);
        global.ui.appendBeforeFinisher(part, global.ui.genericHeaderId);
        $("#" + next.id).fadeIn(400);
        next.registerListeners();
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + pastId), $("#" + presentId), $("#" + futureId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Registers a listener to the past button.
     */
    function registerPastListener() {
        $("#" + pastId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.intro.makeThinkingChoice(ForestIntroductionThinkingChoiceEnum.PAST);
            loadNextPart();
        });
    }

    /**
     * Registers a listener to the present button.
     */
    function registerPresentListener() {
        $("#" + presentId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.intro.makeThinkingChoice(ForestIntroductionThinkingChoiceEnum.PRESENT);
            loadNextPart();
        });
    }

    /**
     * Registers a listener to the future button.
     */
    function registerFutureListener() {
        $("#" + futureId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.intro.makeThinkingChoice(ForestIntroductionThinkingChoiceEnum.FUTURE);
            loadNextPart();
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "introductionRecovering";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.woods.getArrivalChoice() !== ForestWoodsArrivalChoiceEnum.UNKNOWN && !ctrl.bandit.hasSource()) {
            ctrl.bgm.loop();
        }
        var loss = ctrl.intro.recoveringStatusChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += loss;
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.intro.hasMadeThinkingChoice()) {
            return;
        }
        registerPastListener();
        registerPresentListener();
        registerFutureListener();
    };
}

/**
 * Thinking choices to be made.
 */
var ForestIntroductionThinkingChoiceEnum = {
    PAST: 1,
    PRESENT: 2,
    FUTURE: 3
};

// END scenes\Forest\introduction\ForestIntroductionRecovering.js

// scenes\Forest\introduction\ForestIntroduction.js

/**
 * The introduction of the Forest scene.
 * @param {ForestController} controller The forest controller.
 * @param {ForestHome} home One of the parts the kobold can go next.
 * @param {ForestBridge} bridge One of the parts the kobold can go next.
 * @param {ForestWoods} woods One of the parts the kobold can go next.
 */
function ForestIntroduction(controller, home, bridge, woods) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var pathSelection = new ForestIntroductionPathSelection(ctrl, home, bridge, woods);
    var itemSelection = new ForestIntroductionItemSelection(ctrl, pathSelection);
    var recovering = new ForestIntroductionRecovering(ctrl, itemSelection);
    
    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    
    
    // PUBLIC MEMBERS \\        this.public = "test";

    // The id of this segment.
    this.id = "forestIntroduction";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
    
    /**
     * Builds and returns this part.
     */
    this.build = function() {
        var introduction = `<div id="${that.id}">`;
        introduction += recovering.build();
        introduction += itemSelection.build();
        introduction += pathSelection.build();
        introduction += `</div>`;
        return introduction;
    };
    
    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function() {
        recovering.registerListeners();
        itemSelection.registerListeners();
        pathSelection.registerListeners();
    };
}

// END scenes\Forest\introduction\ForestIntroduction.js

// scenes\Forest\ForestController.js

/**
 * Controller for the Forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.intro = new ForestIntroductionController(id);
    this.home = new ForestHomeController(id);
    this.woods = new ForestWoodsController(id);
    this.bandit = new ForestBanditController(id);
    this.unknown = new ForestUnknownController(id);

    this.homeSceneName = "Home";
    this.enslaveSceneName = "CaravanConfinement";
    this.enslaveFolder = "Enslavement";
    this.enslaveReady = true;

    this.misc = new MiscController(id);

    this.hide = new ChoiceController(id, "hide");
    this.tree = new ChangeController(id, "tree", function change() {
        global.kobold.status.physical.decrease(5, "Bee sting");
        global.kobold.status.stamina.decrease(10, "Tree climbing");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Got stung", "Health", -5),
            new Change("Tree climbing", "Stamina", -10)
        ]);
    });
    this.treeFail = new ChangeController(id, "treeFail", function change() {
        global.kobold.status.mental.decrease(10, "Panicking");
        global.kobold.status.stamina.decrease(10, "Attempted climbing");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Panic", "Sanity", -10),
            new Change("Attempted climbing", "Stamina", -10)
        ]);
    });
    this.bush = new ChangeController(id, "bush", function change() {
        global.kobold.status.mental.decrease(10, "Panicking");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Panic", "Sanity", -10)
        ]);
    });
    this.run = new ChangeController(id, "run", function change() {
        global.kobold.status.mental.increase(10, "Fortune");
        global.kobold.status.mental.decrease(5, "Misfortune");
        global.kobold.status.stamina.decrease(25, "Intensive running");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Fortune", "Sanity", +10),
            new Change("Misfortune", "Sanity", -5),
            new Change("Intensive running", "Stamina", -25)
        ]);
    });

    this.runFail = new ChangeController(id, "runFail", function change() {
        global.kobold.status.physical.decrease(15, "Getting choked");
        global.kobold.status.mental.decrease(20, "Imminent rape");
        global.kobold.inventory.equipment.strip(SlotEnum.BRA);
        global.kobold.inventory.equipment.strip(SlotEnum.PANTS);
        global.kobold.inventory.equipment.strip(SlotEnum.UNDERWEAR);
        global.kobold.inventory.equipment.strip(SlotEnum.WEAPON);
        global.kobold.inventory.coins -= that.misc.gold();
    }, function changeNote() {
        var weapon = that.misc.weapon();
        var bra = that.misc.bra();
        var pants = that.misc.pants();
        var underwear = that.misc.underwear();
        var changeArray =
                [new Change("Hit branch", "Health", -10),
                    new Change("Being choked", "Health", -5),
                    new Change("Caught", "Sanity", -10),
                    new Change("Imminent rape", "Sanity", -10),
                    new Change("Money taken", "Gold", -that.misc.gold())];
        if (weapon) {
            changeArray.push(new Change("Lost weapon", global.kobold.inventory.createItemWithTooltip(weapon, true), -1));
        }
        if (bra) {
            changeArray.push(new Change("Torn bra", global.kobold.inventory.createItemWithTooltip(bra, true), -1));
        }
        if (pants) {
            changeArray.push(new Change("Ripped pants", global.kobold.inventory.createItemWithTooltip(pants, true), -1));
        }
        if (underwear) {
            changeArray.push(new Change("Ripped underwear", global.kobold.inventory.createItemWithTooltip(underwear, true), -1));
        }
        return global.ui.change.buildNotification(changeArray);
    });

    this.struggle = new ChoiceController(id, "runFailStruggle");

    this.preContinue = new ContinueController(id, "preContinue");

    this.treeThrow = new ChangeController(id, "treeThrow", function change() {
        global.kobold.status.physical.decrease(10, "Bee stings");
        global.kobold.status.mental.increase(10, "Outsmarting pursuers");
        global.kobold.status.stamina.decrease(15, "Running away");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Multiple stings", "Health", -10),
            new Change("Outsmarted pursuers", "Sanity", +10),
            new Change("Jumped down", "Stamina", -5),
            new Change("Ran away", "Stamina", -10)
        ]);
    });

    this.treeWeather = new ChangeController(id, "treeWeather", function change() {
        global.kobold.status.physical.decrease(50, "Hellish pain");
        global.kobold.status.mental.decrease(20, "Near death experience");
        global.kobold.stats.decreaseStat(StatsEnum.STRENGTH, 1);
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Massive amounts of beestings", "Health", -30),
            new Change("Fell down tree", "Health", -20),
            new Change("Near death experience", "Sanity", -10),
            new Change("Traumatized", "Sanity", -10),
            new Change("Lasting effects", "Strength", -1)
        ]);
    });

    this.surrenderRock = new ChangeController(id, "surrenderRock", function change() {
        var rock = that.intro.getItemChoice() === ItemEnum.ROCK;
        global.kobold.status.physical.decrease(20, `${rock ? "Rock" : "Stick"} insertion`);
        global.kobold.status.mental.decrease(35, `${rock ? "Rock" : "Stick"} insertion`);
        global.kobold.status.exertion.increase(15, `${rock ? "Rock" : "Stick"} insertion`);
        global.kobold.inventory.coins -= that.misc.gold();
        if (that.misc.cunt()) {
            global.kobold.body.vaginalVirgin = false;
        } else {
            global.kobold.body.cockslitVirgin = false;
        }
    }, function changeNote() {
        var rock = that.intro.getItemChoice() === ItemEnum.ROCK;
        return global.ui.change.buildNotification([
            new Change(`${rock ? "Rock" : "Stick"} insertion`, "Health", -20),
            new Change(`${rock ? "Rock" : "Stick"} insertion`, "Sanity", -35),
            new Change(`${rock ? "Rock" : "Stick"} insertion`, "Sexual exhaustion", +15, true),
            new Change("Money taken", "Gold", -that.misc.gold())
        ]);
    });

    this.surrenderSpank = new ChangeController(id, "surrenderSpank", function change() {
        global.kobold.status.physical.decrease(10, "Firmly spanked");
        global.kobold.status.mental.decrease(30, "Extreme humiliation");
        global.kobold.status.exertion.increase(5, "Hurting rear");
        global.kobold.inventory.coins -= that.misc.gold();
        global.kobold.stats.decreaseStat(StatsEnum.ENDURANCE, 1);
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Spanked", "Health", -10),
            new Change("Humiliated", "Sanity", -30),
            new Change("Rear hurts", "Sexual exhaustion", +5, true),
            new Change("Money taken", "Gold", -that.misc.gold()),
            new Change("Sensitive behind", "Endurance", -1)
        ]);
    });

    this.hideStick = new ChangeController(id, "hideStick", function change() {
        global.kobold.status.mental.increase(10, "Outsmarting pursuers");
        global.kobold.inventory.equipment.strip(SlotEnum.WEAPON);
        global.kobold.stats.increaseStat(StatsEnum.INTELLIGENCE, 1);
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Outsmarted pursuers", "Sanity", +10),
            new Change("Lost stick", global.kobold.inventory.createItemWithTooltip(that.misc.weapon(), true), -1),
            new Change("Smart move", "Intelligence", +1)
        ]);
    });

    this.runSucceed = new ChangeController(id, "runSucceed", function change() {
        global.kobold.status.mental.increase(10, "Escaped pursuers");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Escaped pursuers", "Sanity", +10)
        ]);
    });

    this.beg = new ChangeController(id, "beg", function change() {
        global.kobold.status.mental.increase(10, "Rape avoided");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Rape avoided", "Sanity", +25),
            new Change("Captured", "Sanity", -15)
        ]);
    });

    this.hate = new ChangeController(id, "hate", function change() {
        if (that.misc.cunt()) {
            global.kobold.body.vaginalVirgin = false;
        } else {
            global.kobold.body.analVirgin = false;
        }
        global.kobold.status.physical.decrease(5, "Thoroughly stretched");
        global.kobold.status.mental.decrease(20, "Raped");
        global.kobold.status.exertion.increase(40, "Raped by an anthro wolf");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Thoroughly stretched", "Health", -5),
            new Change("Raped", "Sanity", -20),
            new Change("Rape", "Sexual exhaustion", +40, true)
        ]);
    });

    this.enjoy = new ChangeController(id, "enjoy", function change() {
        if (that.misc.cunt()) {
            global.kobold.body.vaginalVirgin = false;
        } else {
            global.kobold.body.analVirgin = false;
        }
        global.kobold.status.physical.decrease(5, "Thoroughly stretched");
        global.kobold.status.mental.decrease(10, "Didn't get the knot");
        global.kobold.status.stamina.decrease(10, "Intense sex");
        global.kobold.status.exertion.increase(30, "Sex with an anthro wolf");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Thoroughly stretched", "Health", -5),
            new Change("Reached climax", "Sanity", +5),
            new Change("Got off to rape", "Sanity", -10),
            new Change("Didn't get the knot", "Sanity", -5),
            new Change("Intense sex", "Stamina", -10),
            new Change("Rough wolf fucking", "Sexual exhaustion", +30, true)
        ]);
    });

    this.bgm = new AudioController(id, "res/audio/forest/forest.mp3", 65);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Forest\ForestController.js

// scenes\Forest\ForestBuilder.js

/**
 * Builder for the Forest scene.
 * @param {ForestController} controller The controller managing this scene.
 */
function ForestBuilder(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var ctrl = controller;
    var bandit = new ForestBandit(ctrl);
    var home = new ForestHome(ctrl, bandit);
    var woods = new ForestWoods(ctrl, bandit);
    var bridge = new ForestBridge(ctrl);
    var intro = new ForestIntroduction(ctrl, home, bridge, woods);
    var title = `<div class="title fancy">
                    the forest
                </div>`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Progressively builds the contents of the scene.
     * @returns {string} The innards of the scene, as a HTML string.
     */
    this.buildScene = function () {
        var scene = title;
        scene += intro.build();
        scene += bridge.build();
        scene += woods.build();
        scene += home.build();
        return scene;
    };

    /**
     * Registers all the required button listeners.
     */
    this.registerListeners = function () {
        intro.registerListeners();
        bridge.registerListeners();
        woods.registerListeners();
        home.registerListeners();
    };
}

// END scenes\Forest\ForestBuilder.js

// scenes\Forest\Forest.js

/**
 * The Forest scene. Shown after the Introduction.
 */
function Forest() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "Forest";
    var sceneName = "Forest - Daytime";
    var theme = "forest";
    var controller = new ForestController(sceneId);
    var builder = new ForestBuilder(controller);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
        global.ui.applyTheme(theme);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += builder.buildScene();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        builder.registerListeners();
        controller.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new Forest();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\Forest\Forest.js



// END SCENE Forest



// SCENE Home

// scenes\Home\saved\ambush\HomeAmbush.js

/**
 * Ambush section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeAmbush(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `ambushContinueButton`;
    var toBed = new HomeAmbushToBed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            While the odds of you taking on the oversized kobold-crossbreed in a fair fight are slim to none,
                            you figure that if you manage to sneak up behind him, you can get a critical first hit in.
                            If you can bonk his unprotected head hard enough, you might just knock him out before he knows what hit'em.
                        [/p]
                        [p]
                            You lower your posture and begin to prowl about, using the dense underbrush as concealment.
                            Sneaking around your shack in a big half-circle, you end up behind it, and more importantly: behind the debt collector.
                            You silently approach the pig-like kobold from the rear, every soft footstep bringing you closer to him.
                        [/p]
                        [p]
                            You stop just a couple of feet short of the treestump the pigbold is sitting on, 
                            intending to perform a charging, leaping bash which is sure to knock him out cold.
                            You close your eyes, and steel your nerves.`;

        var weapon = ctrl.saved.weapon();
        if (weapon) {
            toReturn += `   You ready your ${global.kobold.inventory.createItemWithTooltip(weapon)}.`;
        } else {
            toReturn += `   You ball up your hands, preparing to strike.`;
        }

        toReturn += `       Suddenly, a deep and rumbling voice snaps you out of your concentration.
                            You open your eyes, and are greeted by the sight of a scaly muscular belly protruding from below a leather chestpiece, 
                            mere inches away from your face. 
                        [/p]
                        [p]
                            [c]
                                "[color=brown]
                                    [b]Thought I didn't see you there, kid?[/b]
                                [/color]"
                            [/c]
                        [/p]
                        [p]
                            Before you have a chance to respond, the surprisingly agile and swift pig-hybrid [b]rams[/b] his powerful fist into your forehead,
                            with an immense force strong enough to send you [b]flying[/b].
                            The back of your head hits a nearby tree, nearly cracking open your skull from the rear.
                            Your brain, which was only [b]barely[/b] able to withstand the first blow, fails to weather the second impact.
                            You pass out, your lifeless body slumping to the ground.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.saved.ambush.knockContinued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.ambush.knockContinue();
            loadNextPart(toBed);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeAmbush";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.saved.getApproachChoice() !== HomeSavedChoice.AMBUSH) {
            return "";
        }
        var change = ctrl.saved.ambush.knockChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += toBed.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.saved.getApproachChoice() !== HomeSavedChoice.AMBUSH) {
            return;
        }
        toBed.registerListeners();

        if (ctrl.saved.ambush.knockContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Home\saved\ambush\HomeAmbush.js

// scenes\Home\saved\ambush\HomeAmbushToBed.js

/**
 * To bed section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeAmbushToBed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `ambushWakeContinueButton`;
    var toWake = new HomeAmbushWakeUp(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You begin to wake up a [b]long[/b] time later, your head absolutely [b]killing[/b] you.
                            Slowly regaining consciousness, you clutch your forehead, which feels like it was hit by a [b]catapult[/b].
                            The rear of your cranium, too, feels as though it's been through [b]hell[/b].
                            Carefully feeling around, 
                            your thoroughly rocked brain narrowly registers the small amount of cold, clotted
                            [b][color=red]blood[/color][/b] your fingers collect from around the gash at the back of your head.
                        [/p]
                        [p]
                            You wearily open your heavy eyelids, which are luckily greeted by night-time instead of the brightness of day.
                            You don't think you'd be able to process any vibrant colors right now.
                            Your brain struggles to cope with even the small amount of moonlight dropping in through the treetops.
                            At least you're still on the ground near your house.
                            Not being kidnapped is a good thing.
                        [/p]
                        [p]
                            Becoming aware of the blood on your fingers, you let out a pained groan.
                            You poke around your skull a bit more intently.
                            The aching wound seems to have closed on its own already, although you can't say for sure how much blood you lost.
                        [/p]
                        [p]
                            You examine the rest of your body.
                            Everything but your head seems to be intact.
                            Everything apart from your coin pouch, which you find on the ground next to you, completely emptied.
                        [/p]
                        [p]
                            You wince in pain as you slowly rise to your feet, steadying yourself against the tree you slammed into.
                            You feel like you're about to pass out again, but manage to collect yourself enough to stagger into your house.
                            Closing the wooden door behind you, you practically [b]drop[/b] onto the bale of hay that you call your [i]bed[/i], 
                            before losing consciousness once more.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.saved.ambush.sleepContinued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.ambush.sleepContinue();
            loadNextPart(toWake);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeAmbushBed";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.saved.ambush.knockContinued()) {
            return "";
        }
        global.ui.applyTheme("home_night");
        var change = ctrl.saved.ambush.sleepChange(ctrl.gold());
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += toWake.build();
        part += `</div>`;

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.saved.ambush.knockContinued()) {
            return;
        }

        toWake.registerListeners();

        if (ctrl.saved.ambush.sleepContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Home\saved\ambush\HomeAmbushToBed.js

// scenes\Home\saved\ambush\HomeAmbushWakeUp.js

/**
 * Wake up section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeAmbushWakeUp(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueHamletId = `toHamletContinueButton`;
    var continueVaranarId = `toVaranarContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Once again, you rise after a long period of slumber.
                            Your head still feels as though it was trampled on by the mounts of an entire cavalry unit,
                            but your mind is just a little more [i]clear[/i] than before.
                            Your brain has had time to recover from the two heavy blows it was forced to endure,
                            to the point where you can tolerate keeping your eyes open despite 
                            the early rays of sunlight shining through the cracks in between the wooden planks making up the walls of your shack.
                        [/p]
                        [p]
                            Sitting up straight on your [i]bed[/i], you try to recall exactly what happened.
                            One moment, you were sneaking up to a debt collector.
                            The next, he'd punched you [b]so[/b] hard he knocked you [b]the fuck[/b] out.
                            You can't quite remember [i]how[/i] you made it to your bed in between then and now, but you're glad that you did.
                        [/p]
                        [p]
                            The injury at the back of your head seems to have healed up nicely while you were unconscious.
                            It's not bleeding, not anymore, that is, and the pain has dulled to the point where you can gently touch it [b]without[/b] wincing.
                            You clean the blood off the back of your skull using a bit of straw, before getting up off of your bale of hay.
                        [/p]
                        [p]
                            With concerns over your current health slowly fading, the bigger picture of things slowly creeps to mind.
                            You're pretty sure that guy was a debt collector.
                            If he wasn't, you're fucked.
                            But you don't want to imagine that scenario right now.
                            Right now, you want to believe that despite getting seriously hurt, you managed to buy yourself a couple of days worth of time, at the very least.
                        [/p]
                        [p]
                            You stretch your limbs, and test your shaky balance.
                            It takes a few moments, but you manage to steady your posture and clear your mind.
                            You're fine.
                            Just fine.
                            Your brain wasn't hurt [i]that[/i] badly.
                            Or so you tell yourself to inspire some much-needed conviction, some [b]drive[/b] to [b]keep going[/b].
                            At least you [i]hope[/i] you didn't suffer any [i]permanent[/i] damage from this encounter.
                        [/p]
                        [p]
                            You sigh loudly.
                            The debt collector may have taken over two hundred coins, but you [b]still[/b] owe the clan another fifty orso.
                            While you [b]could[/b] do with [b]a lot[/b] more rest, you really don't want mister half-pig to pay you another visit.
                            Next time, he might outright [b]kill[/b] you instead of just knocking you out.
                        [/p]
                        [p]
                            You exit your hovel.
                            The sun is rising, and the forest is slowly coming to life, a myriad of animal noises reverberating across the woodland.
                            Realizing that you'll never be lucky enough to rob a money transport and live to tell the tale [b]a second time[/b], 
                            you decide to seek your fortune outside of the woods.
                            You can either go to Varanar directly, or you can first take a small detour to the rural hamlet just outside of town.
                            Whichever route you choose, you hope you'll end up better off than you are now.
                        [/p]`;
        
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueHamletId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `To the hamlet &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${continueVaranarId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `To Varanar &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueHamletId), $("#" + continueVaranarId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }
    
    /**
     * Adds a listener to the hamlet button.
     */
    function registerHamletListener() {
        $("#" + continueHamletId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.hamletSceneName, ctrl.hamletFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeAmbushWakeUp";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.saved.ambush.sleepContinued()) {
            return "";
        }
        global.ui.applyTheme("home_dawn");
        var change = ctrl.saved.ambush.wakeChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `<span class="todo">You can only go to the Hamlet.</span>`;
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.saved.ambush.sleepContinued()) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.hasContinued()) {
            return;
        }

        registerHamletListener();
    };
}

// END scenes\Home\saved\ambush\HomeAmbushWakeUp.js

// scenes\Home\saved\hide\HomeHide.js

/**
 * Hide section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeHide(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toBed = new HomeHideToBed(ctrl);
    
    var continueId = `homeHideContinue`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You're [i]smart[/i].
                            If you approach the debt collector with a bag of gold strapped around your neck,
                            he'll most certainly take it from you, to [i]settle the score[/i].
                            But, if you [i]don't have[/i] any gold to give, then he can't [i]take[/i] any.
                            Screw smart.
                            You're a [b]genius[/b].
                        [/p]
                        [p]
                            You take off your pouch of gold, dig a small hole into the ground, and bury it right next to a tree,
                            the loose dirt concealed by a small bush.
                            With your purse hidden, you approach the kobold-pig hybrid with a smug smirk plastered across your face.
                        [/p]
                        [p]
                            "[color=brown][b]You're late.[/b][/color]" the brute speaks, in a deep and bass voice.
                            He rises from his treestump-seat as you approach, towering a good foot above you,
                            undoubtedly weighing multiple times your own mass.
                            "[color=brown][b]Rat wants his money. I'm here to collect.[/b][/color]" the fat lizard continues, scratching his bulging belly.
                        [/p]
                        [p]
                            You calmly and clearly explain that you don't really [i]have[/i] any coins for the half-pig to collect.
                            The big kobold snorts loudly, staring down at you with a hateful glare on his face.
                            He loudly cracks his knuckles, hands balling up into fists.
                            You suddenly realize that you haven't really thought this plan all the way through.
                            Perhaps you aren't as much of a genius as you hoped you were.
                        [/p]
                        [p]
                            Out of the corner of your eye, you spot the motion of a burly arm swinging upwards,
                            but the bandit's scaly fist plows into the underside of your muzzle before you manage to react.
                            The hefty, unexpected uppercut knocks you off of your feet.
                            You fall backwards, onto the leafy forest floor, a bit of blood spilling from your maw onto the ground.
                        [/p]
                        [p]
                            Even as your body comes to a rest, your brain still can't fully comprehend what exactly just happened.
                            Only when the kobold hybrid straddles your chest, pinning you down with the full weight of his relatively [b]enormous[/b] body,
                            does fear begin to emerge at the back of your mind.
                            For a moment, you consider telling him about your buried gold, yet before you can so much as open your mouth,
                            the savage mercilessly begins to [b]wail[/b] on you, brutally bashing your face, 
                            alternating between both of his hands to keep the steady stream of violence coming.
                        [/p]
                        [p]
                            Your mind spaces out, seemingly [i]leaving[/i] your body as your head gets violently [b]thrashed[/b].
                            You wake up from your preservation instinct-induced blackout as the burly ruffian rises from your body,
                            huffing and panting from the vicious beatdown.
                        [/p]
                        [p]
                            Your face is in agonizing pain.
                            Your eyes are [b]punched[/b] shut.
                            You spit out the chipped remains of what was once one of your teeth.
                            You can barely breathe through the blood oozing from your snout, a lot of which is piling up at the back of your throat.
                            You roll onto your side, just to let some of it seep from your mouth more easily.
                        [/p]
                        [p]
                            He's still not quite yet done.
                            With a loud, angry growl, the enforcer ruthlessly [b]kicks[/b] you in the gut, his foot [b]burying[/b] into your unprotected belly,
                            knocking the air out of your meagre frame.
                            You cough up a considerable amount of blood, as your constant flow of tears intensifies.
                            Breathing raggedly, panting for air and whimpering for mercy, you curl up into a ball as the half-kobold retracts his foot.
                        [/p]
                        [p]
                            The debt collector snorts yet again, before hocking down a disgusting wad of phlegm, directly onto of your face.
                            "[color=brown][b]You're lucky I don't get paid enough to kill.[/b][/color]" he says, flexing his fingers and wiping your blood from his knuckles.
                            "[color=brown][b]Others do. Watch your back.[/b][/color]" 
                            With those words uttered, the sadistic pig walks away, leaving you a bloodied and broken mess, idly sobbing on the leafy forest floor.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.saved.hide.ballContinued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.hide.ballContinue();
            loadNextPart(toBed);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeHide";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.saved.getApproachChoice() !== HomeSavedChoice.HIDE) {
            return "";
        }
        var change = ctrl.saved.hide.ballChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += toBed.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.saved.getApproachChoice() !== HomeSavedChoice.HIDE) {
            return;
        }
        
        toBed.registerListeners();

        if (ctrl.saved.hide.ballContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Home\saved\hide\HomeHide.js

// scenes\Home\saved\hide\HomeHideToBed.js

/**
 * To bed section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeHideToBed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toWake = new HomeHideWakeUp(ctrl);

    var continueId = `homeHideBedContinue`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You lie there for a few more hours.
                            You are conscious, yet your body is crippled with pain.
                            You slowly rock back and forth in your fetal position, more than a little traumatized.
                            You've never quite been brutalized like that before.
                            The entire ordeal has left you mentally [b]demolished[/b].
                        [/p]
                        [p]
                            Your face hurts.
                            Your stomach hurts.
                            Your [b]everything[/b] hurts.
                            You bring your hands to your visage.
                            The swelling seems to have gone down a bit, enough for you to open your eyes.
                            The sun has set, and night-time is upon you.
                        [/p]
                        [p]
                            You slowly unfurl from your self-protecting posture.
                            You are shivering, not just from your poor mental condition, but from being cold.
                            You rise to your unsteady feet, wiping some of the blood off of your muzzle.
                        [/p]
                        [p]
                            At least...at least there's one thing that can offer you [i]some[/i] solace.
                            You limp away from your shed, stopping at a familiar tree and dropping to your knees.
                            You dig rampantly, with both your hands, uncovering your well-hidden stash of gold.
                            You press the dirt-covered coinpouch against your chest, cradling it.
                            Your near-[b]broken[/b] mind [b]needs[/b] something to cling to.
                            Your plan [b]worked[/b], or so you claim, to keep yourself from falling apart.
                        [/p]
                        [p]
                            A small smile forms upon your disfigured face, 
                            but it quickly fades to tears once more as you fail to hold up the pretense for more than a few moments.
                            The horrible pain you're in is [b]not[/b] worth [b]any[/b] amount of gold,
                            but you wish you didn't have to realize that cold, hard fact.
                        [/p]
                        [p]
                            You stagger back towards your home.
                            You push open your shed's wooden door.
                            After closing it behind you, you fall down on top of the bale of hay that you call your [i]bed[/i].
                            Still clutching your filthy purse, you slowly cry your battered form to sleep, hoping you'll wake up in the morning.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.saved.hide.sleepContinued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.hide.sleepContinue();
            loadNextPart(toWake);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeHideBed";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.saved.hide.ballContinued()) {
            return "";
        }
        global.ui.applyTheme("home_night");
        var change = ctrl.saved.hide.sleepChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += toWake.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.saved.hide.ballContinued()) {
            return;
        }
        
        toWake.registerListeners();

        if (ctrl.saved.hide.sleepContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Home\saved\hide\HomeHideToBed.js

// scenes\Home\saved\hide\HomeHideWakeUp.js

/**
 * Wake up section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeHideWakeUp(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueHamletId = `toHamletContinueButton`;
    var continueVaranarId = `toVaranarContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You wake up many hours later, after a pained and troubled slumber,
                            drifting in and out of the realm of consciousness,
                            on the brink of being too awake to sleep and too asleep to think.
                        [/p]
                        [p]
                            At the cusp of dawn, you are roused from your restless drowse by a ray of blinding light shining in through a crack in the wooden wall of your shed.
                            You groaningly open your still slightly swollen eyelids.
                            Your entire body hurts, yet not nearly as much as it did yesterday.
                            While your face took most of the intense beating, it has seemingly gone [i]numb[/i],
                            the pain radiating from your visage dull and distant, intensifying only when you touch your wounded scales.
                        [/p]
                        [p]
                            Your mouth has stopped bleeding.
                            With your tongue, which is one of the few parts of your body that isn't aching, you feel around your gums.
                            One of your teeth outright [b]broke[/b], and fragments of it tore open the rest of your mouth, save for your tongue.
                            While you've managed to spit out most of the debris, the entirety of the inside of your mouth is lined with small cuts,
                            which you assume you'll be feeling for a long time still, despite them having stopped bleeding already.
                        [/p]
                        [p]
                            Your jaw, which you believed to be broken last night, seems to be in slightly better condition than you thought it was.
                            While it definitely hurts to open and close, it doesn't feel like anything is severely [b]fractured[/b] or [i]in the wrong place[/i].
                            Your snout, similarly, isn't in quite the poor shape you figured it was.
                            Your muzzle isn't deformed in any way, and apart from the tooth your lost, you don't think any [i]major[/i] damage was done.
                        [/p]
                        [p]
                            Maybe you were panicking just a little [i]too[/i] much yesterday.
                            While the kobold hybrid undeniably [i]fucked you up[/i], he was kind enough not to [b]permanently[/b] disfigure you.
                            He probably figured you'd be dead in a few days anyway.
                        [/p]
                        [p]
                            You shift to a seated position on your [i]bed[/i].
                            Your stomach still hurts quite a lot.
                            You hope you have no lasting [i]internal[/i] injuries, but you have no real way to check.
                            As you rise to your unsteady feet, a sudden wave of inexplicable euphoria washes over you.
                            As though your brain realizes [b]most[/b] of the pain is over, [b]for now[/b].
                        [/p]
                        [p]
                            Your thoughts wander towards the future.
                            While you may have kept your considerable [i]hoard[/i] of gold for now,
                            you doubt the Three-Tailed Rat will simply [b]stop[/b] coming after you when his extorter comes home empty-handed.
                            At best, you figure you'll be kidnapped, and get a chance to pay off your debts still.
                            At worst, you'll be poisoned, stabbed in the back or otherwise assassinated [b]before[/b] 
                            you can even explain that you actually [b]do[/b] have the rat's money.
                        [/p]
                        [p]
                            You wipe any lingering blood off of your still-sensitive face, before leaving the comfortable, familiar snugness of your small, wooden shack.
                            You take a deep breath of fresh morning air, or at least as deep as you can before your battered body begins to ache.
                        [/p]
                        [p]
                            The cool air clears your mind.
                            You take a moment to think, realizing that you've been surviving on [i]pure luck[/i] for a long time now.
                            You decide to seek a more assured fortune outside of the woods.
                            You can either go to Varanar directly, or you can first make a small stop at the rural hamlet just outside of town.
                            Whichever route you choose, you hope you'll end up better off than you are now.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueHamletId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `To the hamlet &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${continueVaranarId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `To Varanar &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueHamletId), $("#" + continueVaranarId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }
    
    /**
     * Adds a listener to the hamlet button.
     */
    function registerHamletListener() {
        $("#" + continueHamletId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.hamletSceneName, ctrl.hamletFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeHideWakeUp";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.saved.hide.sleepContinued()) {
            return "";
        }
        global.ui.applyTheme("home_dawn");
        var change = ctrl.saved.hide.wakeChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `<span class="todo">You can only go to the hamlet.</span>`;
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.saved.hide.sleepContinued()) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.hasContinued()) {
            return;
        }

        registerHamletListener();
    };
}

// END scenes\Home\saved\hide\HomeHideWakeUp.js

// scenes\Home\saved\approach\give\HomeGive.js

/**
 * Give section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeGive(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `giveContinueButton`;
    var toNext = new HomeGiveWakeUp(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You tilt your head upwards, staring at the pigbold's face with a nervous smile on your face.
                            You are more than a little fearful of the broad and muscle-gutted thug.
                            He looks like he could outright [b]tear you in half[/b] should you refuse his request.
                            If anything, you should be thankful that he's asking you [i]somewhat[/i] nicely instead of just [b]taking[/b] your gold.
                        [/p]
                        [p]
                            You take hold of the small rope tying your purse around your neck, and you deftly lift it over your head.
                            After taking one last glance at the hefty bundle of money, you slowly hand it over to the half-kobold in front of you,
                            his arm not so much as [i]flinching[/i] as you place the heavy pouch in his open palm, staying perfectly still despite the added weight.
                        [/p]
                        [p]
                            The bandit inspects the pouch, opening it up to confirm its contents.
                            Being greeted by the sight of over two hundred coins staring back at him, he figures you [b]aren't[/b] trying to deceive him.
                            The ruffian moves your pouch to the side of his leather pants, tilting the small bag to let the coins within [i]pour[/i] into his pocket.
                        [/p]
                        [p]
                            As the clattering of coins comes to an end, the pigbold's attention turns towards you once more.
                            He tosses you your emptied purse, before speaking, his voice slightly less ominous than before, yet still rather threatening.
                            "[color=brown][b]Good. 
                            If this ain't enough, the Rat'll be expecting you to cough up the rest in a few days, in which case we'll be meetin' again.
                            If it's too much, consider the excess a fine for bein' late.[/b][/color]"
                        [/p]
                        [p]
                            After a small, mutual nod to [b]ensure[/b] that his threatening words had reached the depths of your fear-addled mind,
                            the burly hybrid turns away from you, and begins his journey back towards the capital, without looking back even once.
                        [/p]
                        [p]
                            You let out a deep sigh.
                            That was [b]stressful[/b], but you're glad you got it over with.
                            You feel like you did the right thing here.
                            You [i]manned up[/i], acted like an adult, and everything worked out fine in the end.
                        [/p]
                        [p]
                            As the slight euphoric rush you find yourself in quickly fades, you yawn loudly.
                            So far, you hadn't realized how [i]tired[/i] this strenuous, emotion and exertion-filled day had made you,
                            adrenaline and necessity driving you onwards up until now.
                            The hard part is finally over.
                            You can relax, if just for one night.
                        [/p]
                        [p]
                            You enter your small, homely shack, and plop yourself down on top of the comfortable, warm bale of hay that you call your [i]bed[/i].
                            You cover yourself in a few handfuls of straw which serve as a blanket of sorts.
                            After snugly nestling yourself just a little deeper into your resting spot, you slowly doze off into a well-deserved nap.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.saved.debt.firstSegmentContinued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.debt.firstSegmentContinue();
            loadNextPart(toNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeGive";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.saved.debt.get() !== HomeApproachChoice.ACCEPT) {
            return "";
        }
        var change = ctrl.saved.debt.acceptChange(ctrl.gold());
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += toNext.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.saved.debt.get() !== HomeApproachChoice.ACCEPT) {
            return;
        }

        toNext.registerListeners();

        if (ctrl.saved.debt.firstSegmentContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Home\saved\approach\give\HomeGive.js

// scenes\Home\saved\approach\give\HomeGiveWakeUp.js

/**
 * Wake up section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeGiveWakeUp(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueHamletId = `toHamletContinueButton`;
    var continueVaranarId = `toVaranarContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You wake up nearly half a day of deep, dreamless sleep later, still tucked in to your hay-bed, relishing in the warmth it provides.
                            As you slowly open your eyes, 
                            you are greeted with the first few rays of morning sunlight shining brightly through the cracks in the walls of your shack.
                            You stay in bed for just a few more minutes.
                        [/p]
                        [p]
                            Finally, you lazily rise from your comfortable position, wiping a few straws of hay off of the top of your head.
                            You yawn loudly, this time not from exhaustion, but out of the satisfaction of a solid night's sleep. 
                            You contently stretch your limbs.
                            You have a feeling today is going to be a good day.
                        [/p]
                        [p]
                            You step out of your small hovel, and are greeted with a gust of fresh, cool morning air, which you breathe in deeply.
                            It reinvigorates you, expells some of the sleepiness from your brain.
                            You rub your eyes as your mind wakes up at its own pace, lagging behind your body as usual.
                        [/p]
                        [p]
                            You've paid off most of your debts.
                            Great.
                            Now you only need to find a way to scrounge up about fifty more coins in a couple of days, and you'll be golden.
                            You doubt you'll be able to gather that money in the forest, though.
                        [/p]
                        [p]
                            You figure it's time to set out,
                            to venture into the city and test your luck there.
                            Maybe you should drop by the small hamlet outside of town first,
                            just to get a better feel for what humans and other species twice your size are like.
                            Whichever route you choose, you hope you'll end up feeling just as [b]good[/b] and [b]fulfilled[/b] as you do now.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueHamletId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `To the hamlet &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${continueVaranarId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `To Varanar &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueHamletId), $("#" + continueVaranarId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }
    
    /**
     * Adds a listener to the hamlet button.
     */
    function registerHamletListener() {
        $("#" + continueHamletId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.hamletSceneName, ctrl.hamletFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeGiveWakeUp";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.saved.debt.firstSegmentContinued()) {
            return "";
        }
        global.ui.applyTheme("home_dawn");
        var change = ctrl.saved.debt.acceptWakeChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `<span class="todo">You can only go to the hamlet.</span>`;
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.saved.debt.firstSegmentContinued()) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.hasContinued()) {
            return;
        }

        registerHamletListener();
    };
}

// END scenes\Home\saved\approach\give\HomeGiveWakeUp.js

// scenes\Home\saved\approach\refuse\HomeRefuse.js

/**
 * Refuse section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeRefuse(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = `refuseContinueButton`;

    var toNext = new HomeRefuseWakeUp(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            So this guy thinks he can [b]bully[/b] you out of [b]your[/b] hard-earned gold just because he's bigger, stronger and [b]heavier[/b] than you are?!
                            No way.
                            You glare up at him with a defiant look on your face as you shake your head no.
                            You aren't going to give him what he wants.
                        [/p]
                        [p]
                            "[color=brown][b]Suit yourself.[/b][/color]"
                            the burly half-kobold replies, looming over you as though he was going to let himself fall forwards and [b]squash[/b] you underneath his weight.
                            Then, out of nowhere, in the flash of an eye, the brute's fist strikes the lower half of your muzzle,
                            with a force strong enough to outright [b]knock[/b] you off your feet.
                        [/p]
                        [p]
                            You scream out in pain as you fall backwards onto the leafy forest floor.
                            The moment you hit the ground, you immediately try to scramble to your feet, or at least to scurry backwards,
                            your fight or flight and self-preservation instincts kicking in.
                            Before you manage to move even an inch, the heavy-set hybrid's foot comes crashing down upon your chest, pinning you against the ground.
                        [/p]
                        [p]
                            Panicking, you struggle and you squirm, yet you fail to escape from the pigbold's clutches.
                            As though he's done this a thousand times before, he ignores your wriggling, 
                            calmly leans down, reaches for the purse now lying atop your chest and harshly [b]tugs[/b] it away from your neck,
                            the flimsy cord keeping it in place snapping with ease.
                        [/p]
                        [p]
                            Having gotten what he was after, the enforcer removes his foot from your chest, and takes a few steps backwards.
                            You instantly hop to your feet, and drop into a combat stance, yet the thug seems uninterested in fighting.
                            Instead, he casually transfers the contents of your gold-pouch into his own pocket.
                        [/p]
                        [p]
                            You slowly slip out of your fighting posture.
                            Your jaw begins to ache as adrenaline fades from your system, 
                            a painful reminder of the sheer [b]strength[/b] of the man before you.
                            The rational part of your brain reasons that even [b]if[/b] he considered you a worthy foe,
                            you'd just end up getting the [b]shit[/b] beaten out of you regardless.
                        [/p]
                        [p]
                            The harsh fact that the half-kobold has [b]won[/b] this [i]not-quite-fight[/i] only truly sinks in 
                            as he tosses the now-empty pouch that used to contain your money in your face.
                            "[color=brown][b]Next time you act uppity, I'll beat you to a pulp.[/b][/color]" the brute warns.
                            You don't doubt his words for even a second.
                            "[color=brown][b]And if this ain't enough to cover your debt, we'll be seeing eachother again real soon.[/b][/color]"
                        [/p]
                        [p]
                            With your coin pocketed and his final threat uttered, the debt collector turns around, and casually strolls away from your humble abode.
                            Disheartened and hurting, you decide not to give chase.
                            Your clutch your jaw, opening and closing your mouth a few times to make sure nothing's [i]broken[/i].
                            It hurts, but you think you'll be fine.
                            Although he only hit you once, he hit you [b]good[/b].
                            And even then, you had the feeling he was still [b]holding back[/b].
                        [/p]
                        [p]
                            You sigh loudly, tying the neck-cord of your once-filled coin pouch back together.
                            You [b]really[/b] wanted to keep at least [b]some[/b] of your money.
                            At least you showed him, and yourself, that you have [b]guts[/b].
                            If you had ${ctrl.gender() === GenderEnum.MALE ? "external" : ""} balls, they'd be as big as a feral horse's.
                            You give yourself a figurative pat on the back for not backing down at first.
                        [/p]
                        [p]
                            After further lamenting your monetary loss, you decide to call it an early night.
                            You step into your small, homely shack, and plop yourself down on top of the bale of hay that you call your [i]bed[/i].
                            After such a long and tiring day, you easily drift to sleep in a matter of mere minutes.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.saved.debt.firstSegmentContinued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.debt.firstSegmentContinue();
            loadNextPart(toNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeRefuse";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.saved.debt.get() !== HomeApproachChoice.REFUSE) {
            return "";
        }
        var change = ctrl.saved.debt.refuseChange(ctrl.gold());
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += toNext.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.saved.debt.get() !== HomeApproachChoice.REFUSE) {
            return;
        }

        toNext.registerListeners();

        if (ctrl.saved.debt.firstSegmentContinued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Home\saved\approach\refuse\HomeRefuse.js

// scenes\Home\saved\approach\refuse\HomeRefuseWakeUp.js

/**
 * Wake up section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeRefuseWakeUp(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueHamletId = `toHamletContinueButton`;
    var continueVaranarId = `toVaranarContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Your sleep is troubled by [i]nightmares[/i], 
                            your mind subconsciously re-living the awful, scary moments of the passed day over and over again.
                            You toss and turn in your [i]bed[/i], until finally, at the break of dawn, you awake from your slumber,
                            finding yourself resting on the cold, dirt floor instead of in the warm embrace of hay.
                        [/p]
                        [p]
                            You let out an annoyed moan as you rise to your feet.
                            While you slept like shit, at least your jaw is feeling a bit better.
                            On the other hand, your body is feeling a bit [i]stiff[/i] from sleeping on the ground.
                            But you don't let that stop you from going out the door.
                            As you leave your home, you are immediately greeted by a gust of fresh morning air,
                            which does wonders to wipe away your budding morning-grumpiness.
                        [/p]
                        [p]
                            With your mind cleared, you think about the future for a moment.
                            You still owe the Three-Tailed Rat about fifty gold, a sum which you're certain he won't just forget about.
                            You take in a deep breath, and decide that the best course of action for you to take would be to head for Varanar,
                            although you could take a small detour and visit the tiny hamlet right outside the city limits first.
                            Whichever route you choose, you hope you'll end [i]this[/i] day with a better nap than the one you just woke up from.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueHamletId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `To the hamlet &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${continueVaranarId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `To Varanar &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueHamletId), $("#" + continueVaranarId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }
    
    /**
     * Adds a listener to the hamlet button.
     */
    function registerHamletListener() {
        $("#" + continueHamletId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.hamletSceneName, ctrl.hamletFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeRefuseWakeUp";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.saved.debt.firstSegmentContinued()) {
            return "";
        }
        global.ui.applyTheme("home_dawn");
        var change = ctrl.saved.debt.refuseWakeChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `<span class="todo">You can only go to the hamlet.</span>`;
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.saved.debt.firstSegmentContinued()) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.hasContinued()) {
            return;
        }

        registerHamletListener();
    };
}

// END scenes\Home\saved\approach\refuse\HomeRefuseWakeUp.js

// scenes\Home\saved\approach\HomeApproach.js

/**
 * Approachj section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeApproach(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var refuseId = "savedRefuseButton";
    var giveId = "savedGiveButton";
    
    var toRefuse = new HomeRefuse(ctrl);
    var toGive = new HomeGive(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide [b]not[/b] to fuck around and to simply approach the man and have a reasonable, civil conversation with him.
                            He takes note of you as you draw near, rising from his treestump-seat, towering a good foot above you, 
                            undoubtedly weighing multiple times your own mass.
                        [/p]
                        [p]
                            "[color=brown][b]You're late.[/b][/color]" the brute speaks, before crudely scratching his bulging belly.
                            He eyes your meagre frame up and down.
                            Noticing the coin-filled pouch hanging off your neck, the fat lizard rudely points one of his chubby fingers at it.
                            "[color=brown][b]That's what I'm here for.[/b][/color]"  he continues, turning his pointing hand into a receiving cup.
                            "[color=brown][b]Rat wants his money. Hand it over.[/b][/color]" 
                            The enforcer's other hand balls up into a menacingly raised fist.
                            "[color=brown][b]Or else.[/b][/color]" 
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the refuse choice button.
     */
    function buildRefuseChoice() {
        var disabled = ctrl.saved.debt.has() ? "disabled" : "";
        var isSelected = ctrl.saved.debt.get() === HomeApproachChoice.REFUSE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${giveId}" ${disabled}>Hand it over</button>`;

        var tooltip = `Give him your purse.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the give choice button.
     */
    function buildGiveChoice() {
        var disabled = ctrl.saved.debt.has() ? "disabled" : "";
        var isSelected = ctrl.saved.debt.get() === HomeApproachChoice.ACCEPT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${refuseId}" ${disabled}>Refuse</button>`;

        var tooltip = `Shake your head no.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildRefuseChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildGiveChoice();
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + refuseId), $("#" + giveId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerRefuseListener() {
        $("#" + refuseId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.debt.set(HomeApproachChoice.REFUSE);
            loadNextPart(toRefuse);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerGiveListener() {
        $("#" + giveId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.debt.set(HomeApproachChoice.ACCEPT);
            loadNextPart(toGive);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeApproach";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.saved.getApproachChoice() !== HomeSavedChoice.TALK) {
            return "";
        }
        global.ui.applyTheme("home");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toRefuse.build();
        part += toGive.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.saved.getApproachChoice() !== HomeSavedChoice.TALK) {
            return;
        }

        toRefuse.registerListeners();
        toGive.registerListeners();

        if (ctrl.saved.debt.has()) {
            return;
        }

        registerRefuseListener();
        registerGiveListener();
    };
}

var HomeApproachChoice = {
    REFUSE: 1,
    ACCEPT: 2
};

// END scenes\Home\saved\approach\HomeApproach.js

// scenes\Home\saved\HomeSaved.js

/**
 * Saved section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeSaved(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var hideId = "savedHideButton";
    var talkId = "savedTalkButton";
    var ambushId = "savedAmbushButton";

    var toHide = new HomeHide(ctrl);
    var toTalk = new HomeApproach(ctrl);
    var toAmbush = new HomeAmbush(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[st]Early arrival[/st]`;
        toReturn += `   [p]
                            After a short, relaxing trek through the forest,
                            you begin to near your small hovel,
                            located near the northern edge of the woods.
                            You've been living there for just about as long as you can remember.
                            You think it used to serve as a storage room for a larger building,
                            seeing as the somewhat concealed dirt track leading towards your home abruptly ends a good bit away from the actual shack.
                        [/p]
                        [p]
                            As your shed comes into view, you spot a shady figure sitting alongside it on the tree stump that has been serving as your outside-stool.
                            The person in question is a kobold like you, but even from a distance you can see that he's [b]far[/b] larger than you are,
                            in both size and in [b]girth[/b].
                            A crossbreed between your race and something [i]bigger[/i], or so you figure.
                            Probably a pig, judging by his [i]size[/i].
                        [/p]
                        [p]
                            The half-kobold is wearing a ${global.kobold.inventory.createItemWithTooltip(new StuddedLeatherChestPiece())}
                            with matching, equally combat-ready ${global.kobold.inventory.createItemWithTooltip(new LeatherPants())}.
                            His attire makes you think that he was sent to collect your [i]debt[/i], by any means necessary.
                            He isn't visibly armed, but you doubt someone of his [b]bulk[/b] needs anything other than his oversized fists to beat the shit out of you.
                        [/p]
                        [p]
                            The pigbold seemingly hasn't noticed you yet.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the sneak choice button.
     */
    function buildHideChoice() {
        var disabled = ctrl.saved.hasMadeApproachChoice() ? "disabled" : "";
        var isSelected = ctrl.saved.getApproachChoice() === HomeSavedChoice.HIDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${hideId}" ${disabled}>Hide your gold</button>`;

        var tooltip = `Stash your gold in a nearby bush before approaching.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the talk choice button.
     */
    function buildTalkChoice() {
        var disabled = ctrl.saved.hasMadeApproachChoice() ? "disabled" : "";
        var isSelected = ctrl.saved.getApproachChoice() === HomeSavedChoice.TALK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${talkId}" ${disabled}>Talk to him</button>`;

        var tooltip = `Approach the debt collector and talk to him.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the ambush choice button.
     */
    function buildAmbushChoice() {
        var disabled = ctrl.saved.hasMadeApproachChoice() ? "disabled" : "";
        var isSelected = ctrl.saved.getApproachChoice() === HomeSavedChoice.AMBUSH;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${ambushId}" ${disabled}>Ambush him</button>`;

        var tooltip = `Sneak up from behind and try to ambush the kobold.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 185px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildHideChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildTalkChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAmbushChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + hideId), $("#" + talkId), $("#" + ambushId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerHideListener() {
        $("#" + hideId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.makeApproachChoice(HomeSavedChoice.HIDE);
            loadNextPart(toHide);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerTalkListener() {
        $("#" + talkId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.makeApproachChoice(HomeSavedChoice.TALK);
            loadNextPart(toTalk);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerAmbushListener() {
        $("#" + ambushId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.saved.makeApproachChoice(HomeSavedChoice.AMBUSH);
            loadNextPart(toAmbush);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeSaved";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.status() !== HomeStatusEnum.SAVED) {
            return "";
        }
        if (!ctrl.saved.hasMadeApproachChoice()) {
            ctrl.introAudio.playSettingsDependent();
        }
        global.ui.applyTheme("home");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toHide.build();
        part += toTalk.build();
        part += toAmbush.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.status() !== HomeStatusEnum.SAVED) {
            return;
        }

        toHide.registerListeners();
        toTalk.registerListeners();
        toAmbush.registerListeners();

        if (ctrl.saved.hasMadeApproachChoice()) {
            return;
        }

        registerHideListener();
        registerTalkListener();
        registerAmbushListener();
    };
}

var HomeSavedChoice = {
    HIDE: 1,
    TALK: 2,
    AMBUSH: 3
};

// END scenes\Home\saved\HomeSaved.js

// scenes\Home\saved\HomeSavedController.js

/**
 * Controller for the saved part of the home scene.
 * @param {string} sceneId The id of the scene.
 */
function HomeSavedController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;
    var approachId = "savedApproach";
    var debtId = "savedDebt";
    var weaponId = "savedWeapon";
    var ambushId = "savedAmbush";
    var hideId = "savedHide";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Makes the approach choice.
     * @param {HomeSavedChoice} choice The choice.
     */
    this.makeApproachChoice = function (choice) {
        if (choice === HomeSavedChoice.AMBUSH) {
            global.kobold.history.shared.debtCollector.status = DebtCollectorEnum.AMBUSHED;
        }
        if (choice === HomeSavedChoice.HIDE) {
            global.kobold.history.shared.debtCollector.status = DebtCollectorEnum.UNPAID;
        }
        global.kobold.history.setKey(id, approachId, choice);
    };

    /**
     * Fetches the approach choice from history and returns it.
     * May return undefined if the approach choice has not yet been set.
     */
    this.getApproachChoice = function () {
        return global.kobold.history.getKey(id, approachId);
    };

    /**
     * Returns true if the approach choice has been made.
     */
    this.hasMadeApproachChoice = function () {
        return global.kobold.history.has(id, approachId);
    };


    function DebtController(key) {
        var segment = key + "first";
        var acceptChangeId = key + "acceptChange";
        var refuseChangeId = key + "refuseChange";
        var acceptWakeChangeId = key + "acceptWakeChange";
        var refuseWakeChangeId = key + "refuseWakeChange";
        this.get = function () {
            return global.kobold.history.getKey(id, key);
        };
        this.set = function (choice) {
            if (choice === HomeApproachChoice.ACCEPT) {
                global.kobold.history.shared.debtCollector.status = DebtCollectorEnum.PAID;
            }
            if (choice === HomeApproachChoice.REFUSE) {
                global.kobold.history.shared.debtCollector.status = DebtCollectorEnum.REFUSED;
            }
            global.kobold.history.setKey(id, key, choice);
        };
        this.has = function () {
            return global.kobold.history.has(id, key);
        };
        this.firstSegmentContinue = function () {
            global.kobold.history.setKey(id, segment, true);
        };
        this.firstSegmentContinued = function () {
            return global.kobold.history.has(id, segment);
        };
        this.acceptChange = function (coins) {
            global.kobold.history.doIfNotSet(id, acceptChangeId, true, function () {
                global.kobold.history.shared.debtCollector.paid += coins;
                global.kobold.inventory.coins = 0;
                global.kobold.status.mental.increase(20, "Immense relief");
                global.kobold.stats.increaseStat(StatsEnum.INTELLIGENCE, 1);
            });
            return global.ui.change.buildNotification([
                new Change("Debt paid", "Coins", -coins),
                new Change("Immense relief", "Sanity", +20),
                new Change("Mature conduct", "Intelligence", +1)
            ]);
        };
        this.refuseChange = function (coins) {
            global.kobold.history.doIfNotSet(id, refuseChangeId, true, function () {
                global.kobold.history.shared.debtCollector.paid += coins;
                global.kobold.inventory.coins = 0;
                global.kobold.status.physical.decrease(20, "Punched");
                global.kobold.status.mental.decrease(5, "Bravery");
            });
            return global.ui.change.buildNotification([
                new Change("Money taken", "Coins", -coins),
                new Change("Punched", "Health", -20),
                new Change("Money taken", "Sanity", -10),
                new Change("Bravery", "Sanity", +5)
            ]);
        };
        this.acceptWakeChange = function () {
            global.kobold.history.doIfNotSet(id, acceptWakeChangeId, true, function () {
                global.kobold.status.physical.increase(40, "Well-earned nap");
                global.kobold.status.stamina.increase(40, "Well-earned nap");
            });
            return global.ui.change.buildNotification([
                new Change("Well-earned nap", "Health", +40),
                new Change("Well-earned nap", "Stamina", +40)
            ]);
        };
        this.refuseWakeChange = function () {
            global.kobold.history.doIfNotSet(id, refuseWakeChangeId, true, function () {
                global.kobold.status.physical.increase(10, "Jaw-pain subsided");
                global.kobold.status.mental.decrease(10, "Troubled sleep");
                global.kobold.status.stamina.increase(20, "Mediocre rest");
            });
            return global.ui.change.buildNotification([
                new Change("Pain subsided", "Health", +10),
                new Change("Troubled sleep", "Sanity", -10),
                new Change("Mediocre rest", "Stamina", +20)
            ]);
        };
    }

    this.debt = new DebtController(debtId);

    function AmbushController(key) {
        var knockChangeId = key + "KnockChange";
        var knockId = key + "Knocked";
        var sleepChangeId = key + "SleepChange";
        var sleepId = key + "Sleep";
        var wakeChangeId = key + "WakeChange";
        this.knockChange = function () {
            global.kobold.history.doIfNotSet(id, knockChangeId, true, function () {
                global.kobold.status.physical.decrease(40, "Knocked out");
            });
            return global.ui.change.buildNotification([
                new Change("Knocked out", "Health", -40)
            ]);
        };
        this.knockContinue = function () {
            global.kobold.history.setKey(id, knockId, true);
        };
        this.knockContinued = function () {
            return global.kobold.history.has(id, knockId);
        };
        this.sleepChange = function (coins) {
            global.kobold.history.doIfNotSet(id, sleepChangeId, true, function () {
                global.kobold.history.shared.debtCollector.paid += coins;
                global.kobold.inventory.coins = 0;
                global.kobold.status.mental.decrease(20, "Brain damage");
            });
            return global.ui.change.buildNotification([
                new Change("Debt collected", "Coins", -coins),
                new Change("Brain damage", "Sanity", -20)
            ]);
        };
        this.sleepContinue = function () {
            global.kobold.history.setKey(id, sleepId, true);
        };
        this.sleepContinued = function () {
            return global.kobold.history.has(id, sleepId);
        };
        this.wakeChange = function () {
            global.kobold.history.doIfNotSet(id, wakeChangeId, true, function () {
                global.kobold.status.physical.increase(10, "Wounds healing");
                global.kobold.status.stamina.increase(20, "Pained slumber");
                global.kobold.stats.decreaseStat(StatsEnum.AGILITY, 1);
                global.kobold.stats.decreaseStat(StatsEnum.INTELLIGENCE, 1);
            });
            return global.ui.change.buildNotification([
                new Change("Wounds healing", "Health", +10),
                new Change("Pained slumber", "Stamina", +20),
                new Change("Permanent damage", "Agility", -1),
                new Change("Permanent damage", "Intelligence", -1)
            ]);
        };
    }

    this.ambush = new AmbushController(ambushId);

    function HideController(key) {
        var ballId = key + "Ball";
        var ballChangeId = key + "BallChange";
        var sleepChangeId = key + "SleepChange";
        var sleepId = key + "Sleep";
        var wakeChangeId = key + "WakeChange";
        this.ballChange = function () {
            global.kobold.history.doIfNotSet(id, ballChangeId, true, function () {
                global.kobold.status.physical.decrease(70, "Brutal beatdown");
            });
            return global.ui.change.buildNotification([
                new Change("Brutal beatdown", "Health", -70)
            ]);
        };
        this.ballContinue = function () {
            global.kobold.history.setKey(id, ballId, true);
        };
        this.ballContinued = function () {
            return global.kobold.history.has(id, ballId);
        };
        this.sleepChange = function () {
            global.kobold.history.doIfNotSet(id, sleepChangeId, true, function () {
                global.kobold.status.mental.decrease(10, "Brutal beatdown");
            });
            return global.ui.change.buildNotification([
                new Change("Kept coins", "Sanity", +10),
                new Change("But at what cost?", "Sanity", -20)
            ]);
        };
        this.sleepContinue = function () {
            global.kobold.history.setKey(id, sleepId, true);
        };
        this.sleepContinued = function () {
            return global.kobold.history.has(id, sleepId);
        };
        this.wakeChange = function () {
            global.kobold.history.doIfNotSet(id, wakeChangeId, true, function () {
                global.kobold.status.physical.increase(20, "Wounds healing nicely");
                global.kobold.status.mental.increase(10, "Cautious optimism");
                global.kobold.status.stamina.increase(20, "Pained slumber");
                global.kobold.stats.increaseStat(StatsEnum.ENDURANCE, 1);
            });
            return global.ui.change.buildNotification([
                new Change("Wounds healing", "Health", +20),
                new Change("Cautious optimism", "Sanity", +10),
                new Change("Pained slumber", "Stamina", +20),
                new Change("What doesn't kill you", "Endurance", +1)
            ]);
        };
    }

    this.hide = new HideController(hideId);

    /**
     * Stored and returns the kobold's weapon.
     */
    this.weapon = function () {
        global.kobold.history.setIfNotSet(id, weaponId, global.kobold.inventory.equipment.weapon);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, weaponId));
    };
}

// END scenes\Home\saved\HomeSavedController.js

// scenes\Home\unsaved\ground\ant\HomeAnt.js

/**
 * Ant section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeAnt(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueHamletId = `toHamletContinueButton`;
    var continueVaranarId = `toVaranarContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.gender() === GenderEnum.MALE;
        var pants = ctrl.pants();
        var bra = ctrl.bra();
        var shirt = ctrl.shirt();
        var underwear = ctrl.underwear();
        var toReturn = `[p]
                            In many parts of the world, these ants are [b]highly[/b] sought after.
                            Some harvest their venom to make powerful lust-inducing potions,
                            others keep entire colonies to [i]indulge[/i] themselves every now and again.
                            You've even heard of some brothels containing walled-off gardens filled with anthills full of this particular breed,
                            visited by the most exclusive of clientele who pay a bunch of gold for just a single encounter with the Venusian ants.
                        [/p]
                        [p]
                            You should consider yourself [b]lucky[/b] for having the rare chance to have some fun with these marvellous little insects.
                            A second one bites down, more aphrodisiac rushing into your system, 
                            ${male ? `your member rapidly hardening, emerging from your cockslit` : `your groin growing more damp with every passing second`}.
                        [/p]
                        [p]
                        `;

        if (pants) {
            toReturn += `   You quickly lower your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (underwear) {
                toReturn += ` and your ${global.kobold.inventory.createItemWithTooltip(underwear)}`;
            }
            toReturn += `, exposing your ${male ? `now fully erect length` : `heated folds`} to the coldness of your surroundings.`;
        } else if (underwear) {
            toReturn += `   You quickly lower your ${global.kobold.inventory.createItemWithTooltip(underwear)},
                            exposing your ${male ? `now fully erect length` : `heated folds`} to the coldness of your surroundings.`;
        }

        if (male) {
            toReturn += `   You grip your needy shaft, and immediately start to stroke it, a third bite eliciting a small,
                            feminine moan from your lips as every motion of your hand causes immense pleasure,
                            your body made temporarily oversensitive by the ants' venom.`;
        } else {
            if(shirt) {
                toReturn += `   You quickly take off your ${global.kobold.inventory.createItemWithTooltip(shirt)}.`;
            }
            if (bra) {
                toReturn += `   You remove your ${global.kobold.inventory.createItemWithTooltip(bra)}, and start to play with your breasts.`;
            } else {
                toReturn += `   You casually start to play with your breasts.`;
            }
            toReturn += `   One of your hands finds its way between your legs, and you wantonly begin to rub your slick slit,
                            a third bite eliciting a small moan from your lips. Every touch causes immense pleasure,
                            your body made temporarily oversensitive by the ants' venom.`;
        }

        toReturn += `   [/p]
                        [p]
                            But still, it's not enough. You need [b]more[/b].
                            You rise from the ground, causing a few more ants to bite, sending spikes of arousal up your spine, clouding your mind.
                            The rational part of your brain knows this is a [b]bad[/b] idea, yet that part is being drowned out by sheer [b]lust[/b].
                            As though your body is acting on its own, you walk over to the anthill, 
                            ${male ? `drop down onto all fours, and thrust your quivering cock directly into the dirt mound` : `squat down, and press your needy loins directly against the dirt mound`}.
                        [/p]
                        [p]
                            Their lair touched, the ants are driven into a frenzy.
                        `;

        if (male) {
            toReturn += `   They swarm over your length, your eyes rolling back into your head as an untold amount of aphrodisiac is injected directly into your loins.
                            You pull your hips back only so you can begin to thrust your shaft in and out of the anthill, outright [b]fucking[/b] the ground,
                            getting off not to any sort of [i]tightness[/i], 
                            but rather to the sheer amount of venom and the super-enhanced feeling of the ants moving about your most sensitive body part.
                        [/p]
                        [p]
                            Needless to say, you don't last long.
                            After just a few humps, you can quickly feel your end approaching.
                            You [b]pound[/b] the mound of dirt into [b]ground[/b], almost to the point of collapse,
                            until finally slamming your member all the way inside one final time, 
                            your hips pressing directly against the tip of the anthill as you reach your climax,
                            squirting wave after wave of thick, heady spunk directly into the lair of the ants.
                        [/p]
                        [p]
                            The moment you reach your orgasm, the tiny creatures instantly seem to recoil from your scales,
                            crawling off of you and back into their hive, like they got what they were after.
                            Your body is already filled to the brim with as much aphrodisiac as it can handle, so you don't really mind.
                        [/p]
                        [p]
                            It takes quite a while for your balls to empty, your climax fueled by the ants' venom.
                            When you finally do pull out of the anthill, you feel completely [b]drained[/b].
                            You roll off to the side, and take a moment to recover.
                            Panting loudly, you cast a gaze downwards, at your member.
                            It seems...[i]bigger[/i] than before.
                            A side effect of the aphrodisiac?
                        [/p]
                        [p]
                            As you slowly rise to your feet, you notice dawn is breaking.
                            No point in going back to sleep now.`;
        } else {
            toReturn += `   They crawl up your waist, some even finding their way into your snatch, 
                            your eyes rolling into the back of your head as their potent aphrodisiac is injected into your very innards.
                            You grab a few more off the ground and place them on your chest, one of them ending up directly on your left nipple,
                            a loud, pleasured grunt echoing through the woods as the little insect bites down.
                        [/p]
                        [p]
                            Needless to say, you don't last long.
                            As need builds up inside of your body, you begin to rub your sopping-wet cunt against the anthill, 
                            its coarse texture ultimately sending you over the edge.
                            You scream out in pure ecstasy as you press your soaking slit down onto the dirt mound as hard as you can,
                            your feminine fluids squirting liberally deep into the lair of the ants.
                        [/p]
                        [p]
                            The moment you reach your orgasm, the tiny creatures instantly seem to recoil from your scales,
                            crawling off of you and back into their hive, like they got what they were after.
                            Your body is already filled to the brim with as much aphrodisiac as it can handle, so you don't really mind.
                        [/p]
                        [p]
                            It takes quite a while for your climax to come to an end, your pleasure fueled by the ants' venom.
                            When you finally do pull away from the anthill, you feel completely [b]drained[/b].
                            You step to the side, and take a moment to recover.
                            Panting loudly, you cast a gaze downwards at your body, but find your view blocked by your breasts.
                            They seem...[i]bigger[/i] than before.
                            A side effect of the aphrodisiac?
                        [/p]
                        [p]
                            As you wipe some of the lingering fluids off your groin, you notice dawn is breaking.
                            No point in going back to sleep now.`;
        }
        
        if (pants || bra || underwear) {
            toReturn += `   You slip back into your attire.`;
        }


        toReturn += `       Despite the large amount of aphrodisiac still coursing through your veins,
                            your arousal seems to have faded along with your orgasm.
                            Your mind is quickly clearing, too.
                            With your lust sated, you turn your attention towards the future.
                        [/p]
                        [p]`;
        var gold = ctrl.gold();
        if (!gold) {
            toReturn += `   You are completely [b]broke[/b]. 
                            You have no money whatsoever to pay your overdue debts with.
                            Desperate times call for desperate measures.`;
        } else if (gold > 150) {
            toReturn += `   You have quite a bit of money, but sadly it's still not enough to pay your overdue debt.`;
        } else if (gold > 203) {
            toReturn += `   You have [b]a lot[/b] of money, more than you gained from just robbing the caravan, even.
                            You're [b]very[/b] close to being able to pay off your debt.
                            You just need a [i]teensy[/i] bit more.
                            But you aren't going to get it from these woods.`;
        }

        toReturn += `       You decide to head out, to seek your fortune in the great city of Varanar.
                            Maybe you'll stop at the small hamlet outside of the city walls along the way.
                            While you'd love to bring some of the ants along with you, they have all disappeared back into the hole they once came from.
                            A shame, they could have fetched a handsome price.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueHamletId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `To the hamlet &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${continueVaranarId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `To Varanar &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueHamletId), $("#" + continueVaranarId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }
    
    /**
     * Adds a listener to the hamlet button.
     */
    function registerHamletListener() {
        $("#" + continueHamletId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.hamletSceneName, ctrl.hamletFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeAnt";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unsaved.getGroundChoice() !== HomeGroundChoice.ANT) {
            return "";
        }
        global.ui.applyTheme("home_dawn");
        var change = ctrl.unsaved.antChange(ctrl.gender());
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `<span class="todo">You can only go to the hamlet.</span>`;
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unsaved.getGroundChoice() !== HomeGroundChoice.ANT) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.hasContinued()) {
            return;
        }

        registerHamletListener();
    };
}

// END scenes\Home\unsaved\ground\ant\HomeAnt.js

// scenes\Home\unsaved\ground\shake\HomeShake.js

/**
 * Shake section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeShake(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var masturbateId = "unsavedMasturbateButton";
    var sleepId = "unsavedSleepButton";

    var toMasturbate = new HomeMasturbate(ctrl);
    var toSleep = new HomeWetDream(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.gender() === GenderEnum.MALE;
        var toReturn = `[p]
                            Startled, you hop to your feet and frantically begin to swat the ants off of you.
                            The ones you don't manage to wipe off with the first pass of your hand immediately bite down,
                            your system flooded by enough venom to drive a dragon wild.
                        [/p]
                        [p]
                            You take a few steps backwards, away from the ant-pouring mound, as you continue to remove more bugs from your body.
                            You swat one on your side with the flat of your hand, you pluck one off of your tail, and crush one of the insects making its way up your chest.
                            Just as you thought you were completely ant-free, one final specimen digs its mandibles into your back, 
                            before being smitten by your rightful wrath.
                        [/p]
                        [p]
                            The rest of the bugs retreat back into their anthill, dissuaded by the distance you've created between you and it.
                            While you've managed to overcome the ants themselves, the potent aphrodisiac inflicted by their bites courses through your veins.
                            In between your legs, 
                            your ${male ? `fully erect member twitches and throbs,` : `sopping wet slit is`} practically [b]begging[/b] for attention.
                            You are [b]so[/b] aroused that your knees are growing weak, the urge, no, the [b]need[/b] 
                            to ${male ? `jerk off` : `touch yourself`} intensifying with every passing moment.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the ant choice button.
     */
    function buildMasturbateChoice() {
        var disabled = ctrl.unsaved.shake.has() ? "disabled" : "";
        var isSelected = ctrl.unsaved.shake.get() === HomeShakeChoice.MASTURBATE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${masturbateId}" ${disabled}>Masturbate</button>`;

        var tooltip = `Relieve yourself. You wouldn't be able to sleep like this anyway.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildSleepChoice() {
        var disabled = ctrl.unsaved.shake.has() ? "disabled" : "";
        var isSelected = ctrl.unsaved.shake.get() === HomeShakeChoice.SLEEP;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${sleepId}" ${disabled}>Sleep</button>`;

        var tooltip = `Repress the dirty thoughts flooding your mind and try to get some rest.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildMasturbateChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSleepChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + masturbateId), $("#" + sleepId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerMasturbateListener() {
        $("#" + masturbateId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unsaved.shake.set(HomeShakeChoice.MASTURBATE);
            loadNextPart(toMasturbate);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerSleepListener() {
        $("#" + sleepId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unsaved.shake.set(HomeShakeChoice.SLEEP);
            loadNextPart(toSleep);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeShake";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unsaved.getGroundChoice() !== HomeGroundChoice.SHAKE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toMasturbate.build();
        part += toSleep.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unsaved.getGroundChoice() !== HomeGroundChoice.SHAKE) {
            return;
        }

        toMasturbate.registerListeners();
        toSleep.registerListeners();

        if (ctrl.unsaved.shake.has()) {
            return;
        }

        registerMasturbateListener();
        registerSleepListener();
    };
}

var HomeShakeChoice = {
    MASTURBATE: 1,
    SLEEP: 2
};

// END scenes\Home\unsaved\ground\shake\HomeShake.js

// scenes\Home\unsaved\ground\shake\masturbate\HomeMasturbate.js

/**
 * Masturbate section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeMasturbate(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueHamletId = `toHamletContinueButton`;
    var continueVaranarId = `toVaranarContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.gender() === GenderEnum.MALE;
        var pants = ctrl.pants();
        var bra = ctrl.bra();
        var underwear = ctrl.underwear();
        var toReturn = `[p]
                            You let out a small, needy grunt as you lustily lick your lips, your hands drifting towards your groin.`;

        if (pants) {
            toReturn += `   You quickly lower your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (underwear) {
                toReturn += ` and your ${global.kobold.inventory.createItemWithTooltip(underwear)}`;
            }
            toReturn += `, exposing your ${male ? `rock-hard member` : `heated folds`} to the coldness of your surroundings.`;
        } else if (underwear) {
            toReturn += `   You quickly lower your ${global.kobold.inventory.createItemWithTooltip(underwear)},
                            exposing your ${male ? `rock-hard member` : `heated folds`} to the coldness of your surroundings.`;
        }

        if (male) {
            toReturn += `   You grip your quivering shaft, and deftly begin to stroke it with your hand,
                            every touch feeling [b]ten times[/b] more intense, your flesh overly sensitive due to the ants' venom.
                        [/p]
                        [p]
                            Leaning against a nearby tree, you pick up your pace, 
                            an abundance of warm precum oozing from the tapered tip of your needy cock lubing up your shaft and your palm,
                            slickening your jerking.
                        [/p]
                        [p]
                            In your aphrodisiac-addled state, you fail to last very long.
                            All too soon, you can feel your end rapidly approaching.
                            Arching your back and letting out a small, rather feminine moan, you shoot your load onto the bark of the tree you were leaning against,
                            your internal balls working overtime to pump out wave after wave of thick, 
                            heady spunk which glistens beautifully in the slowly rising sun's first rays of light.`;
        } else {
            toReturn += `   You slide two of your fingers in between your messy thighs, and press them against your slick slit.
                            Even this slight touch is enough to draw a soft moan from your lips, your body overly sensitive due to the ants' venom.
                        [/p]
                        [p]
                            You spread your legs, and without further ado, you slide both digits inside of your completely [b]drenched[/b] folds, to the very knuckle.
                            You can barely refrain from howling in pleasure as you begin to pump your fingers in and out of your sopping wet cunt,
                            the magnified feeling unlike anything you've ever felt before.
                            Sadly, it is [i]due to[/i] this enhanced satisfaction that you only manage to last for just about a minute of masturbation 
                            before you feel your climax building up inside of you.
                        [/p]
                        [p]
                            With the sun slowly starting to rise, and with your lust greater than ever before, 
                            you decide [i]not[/i] to [b]deprive[/b] yourself from a much-needed orgasm any longer.
                            One final time, you [b]ram[/b] your fingers deep inside of you, before reaching your end,
                            your entire body quivering in pleasure, an abundance of your feminine fluid squirting out onto the leafy forest floor,
                            glistening beautifully in the very first rays of sunlight.
                            `;
        }

        toReturn += `   [/p]
                        [p]
                            Panting loudly, you revel in post-orgasmic bliss for a few moments, 
                            before wiping your groin${pants || underwear ? ` and redressing` : ``}.
                            The ants' venom is slowly being filtered from your system, and with it, your spontaneous arousal vanishes as well.
                        [/p]
                        [p]
                            After licking your fingers clean, you decide there's no point in going back to sleep.
                            Those couple of hours that you rested [i]before[/i] the ants got to you'll have to suffice.
                            While this was fun, it's time to think about more important things.
                            You turn your mind to the future.
                        [/p]
                        [p]`;

        var gold = ctrl.gold();
        if (!gold) {
            toReturn += `   You are completely [b]broke[/b]. 
                            You have no money whatsoever to pay your overdue debts with.
                            Desperate times call for desperate measures.`;
        } else if (gold > 150) {
            toReturn += `   You have quite a bit of money, but sadly it's still not enough to pay your overdue debt.`;
        } else if (gold > 203) {
            toReturn += `   You have [b]a lot[/b] of money, more than you gained from just robbing the caravan, even.
                            You're [b]very[/b] close to being able to pay off your debt.
                            You just need a [i]teensy[/i] bit more.
                            But you aren't going to get it from these woods.`;
        }

        toReturn += `       You decide to head out, to seek your fortune in the great city of Varanar.
                            Maybe you'll stop at the small hamlet outside of the city walls along the way.
                            If anything, you hope you at least won't end up sleeping next to another anthill tonight.
                        [/p]`;


        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueHamletId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `To the hamlet &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${continueVaranarId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `To Varanar &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueHamletId), $("#" + continueVaranarId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }
    
    /**
     * Adds a listener to the hamlet button.
     */
    function registerHamletListener() {
        $("#" + continueHamletId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.hamletSceneName, ctrl.hamletFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeMasturbate";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unsaved.shake.get() !== HomeShakeChoice.MASTURBATE) {
            return "";
        }
        global.ui.applyTheme("home_dawn");
        var change = ctrl.unsaved.masturbateChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `<span class="todo">You can only go to the hamlet.</span>`;
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unsaved.shake.get() !== HomeShakeChoice.MASTURBATE) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.hasContinued()) {
            return;
        }

        registerHamletListener();
    };
}

// END scenes\Home\unsaved\ground\shake\masturbate\HomeMasturbate.js

// scenes\Home\unsaved\ground\shake\wetdream\HomeWetDream.js

/**
 * Wet dream section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeWetDream(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueHamletId = `toHamletContinueButton`;
    var continueVaranarId = `toVaranarContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.gender() === GenderEnum.MALE;
        var sawDragoness = global.kobold.history.shared.dragonessPrincess.relation !== DragonessPrincessEnum.UNMET;
        var toReturn = `[p]
                            No!
                            You [b]aren't[/b] giving in to these darned [b]pests[/b]' venom.
                            While they're just simple insects doing whatever their instinct tells them to, you are [b]not[/b].
                            You may be at the bottom of the civilized world, but if there's anything you [i]are[/i] above, it is simple-minded creatures such as ants.
                        [/p]
                        [p]
                            This is a matter of principle.
                            You will [b]not[/b] be brought to your knees by a bunch of tiny six-legged freaks, 
                            no matter [b]how[/b] much your groin is [b]aching[/b] for attention.
                            You take a deep breath and steel your resolve.
                            ${male ? `With your member proudly protruding from your groin,` : `With your arousal beginning to seep down your inner thighs,`}
                            you begin to look for a new place to sleep.
                        [/p]
                        [p]
                            It isn't long before you come across a tiny blanket of moss, big enough for at least [i]some[/i] of your small frame to rest on.
                            Tiredness beginning to set in once more, you decide not to look any further.
                            You quickly check for anthills, and are glad to see that this time, there are none around.
                            You drop down to the ground, rest your head atop of the surprisingly soft moss, and within seconds manage to drift to sleep once more.
                        [/p]
                        [p]
                            Being as [i]excited[/i] as you are, your slumber is plagued by erotic dreams of various kinds.
                            One moment, you are tied up in a dungeon, getting abused by a sadistic, whip-armed fox.
                            The next, you find yourself getting caught by the soldiers that you robbed, 
                            a particularly buff looking wolf amongst them ramming his oversized knot inside of you.
                            ${sawDragoness ? `The blue-scaled dragoness also plays a role in your wet dreams. You envision yourself lapping at her folds while she rides your face, moaning in pleasure.` : ``}
                        [/p]
                        [p]
                            You slowly drift away from your pleasured dreams, waking up to find your hand having subconsciously drifted towards your groin.
                            During the night, you may have touched yourself a bit, but you proudly come to the conclusion that you [b]didn't[/b] masturbate.
                            You can practically [b]feel[/b] the ants' venom fading from your body, your arousal vanishing along with it.
                        [/p]
                        [p]
                            You let out a sigh of relief.
                            You've managed to retain your pride in the face of adversity, 
                            and you've won a figurative mental battle against a bunch of tiny, mindless bugs.
                            Regardless of the mediocrity of this achievement,
                            you feel content to know that you can remain in control of your body even when aroused.
                        [/p]
                        [p]
                            As the sun rises once again, so do you.
                            Getting up off of the ground, the coldness of the morning grips your body.
                            It washes away the heat of your arousal, like a dip in the river on a frigid winter night.
                            The last hint of your excitement vanishing, your mind turns to the future.
                        [/p]
                        [p]`;
        var gold = ctrl.gold();
        if (!gold) {
            toReturn += `   You are completely [b]broke[/b]. 
                            You have no money whatsoever to pay your overdue debts with.
                            Desperate times call for desperate measures.`;
        } else if (gold > 150) {
            toReturn += `   You have quite a bit of money, but sadly it's still not enough to pay your overdue debt.`;
        } else if (gold > 203) {
            toReturn += `   You have [b]a lot[/b] of money, more than you gained from just robbing the caravan, even.
                            You're [b]very[/b] close to being able to pay off your debt.
                            You just need a [i]teensy[/i] bit more.
                            But you aren't going to get it from these woods.`;
        }

        toReturn += `       You decide to head out, to seek your fortune in the great city of Varanar.
                            Maybe you'll stop at the small hamlet outside of the city walls along the way.
                            If anything, you hope you at least won't end up sleeping next to another anthill tonight.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueHamletId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `To the hamlet &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${continueVaranarId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `To Varanar &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueHamletId), $("#" + continueVaranarId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the hamlet button.
     */
    function registerHamletListener() {
        $("#" + continueHamletId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.hamletSceneName, ctrl.hamletFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeWetDream";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unsaved.shake.get() !== HomeShakeChoice.SLEEP) {
            return "";
        }
        global.ui.applyTheme("home_dawn");
        var change = ctrl.unsaved.wetChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `<span class="todo">You can only go to the hamlet.</span>`;
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unsaved.shake.get() !== HomeShakeChoice.SLEEP) {
            return;
        }
        // REGISTER CHILDREN HERE

        if (ctrl.hasContinued()) {
            return;
        }

        registerHamletListener();
    };
}

// END scenes\Home\unsaved\ground\shake\wetdream\HomeWetDream.js

// scenes\Home\unsaved\ground\HomeGround.js

/**
 * Ground sleeping section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeGround(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var antId = "unsavedAntButton";
    var shakeId = "unsavedShakeButton";

    var toAnt = new HomeAnt(ctrl);
    var toShake = new HomeShake(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.gender() === GenderEnum.MALE;
        var toReturn = `[p]
                            You owe the clan [b]a lot[/b] of money.
                            You wouldn't be surprised if the Three-Tailed Rat were to send some kind of night-time 
                            assassination squad to your home to try and [i]teach you[/i] a [b]lethal late-night lesson[/b].
                        [/p]
                        [p]
                            You can't risk sleeping in your home. That's the first place they'd come looking for you.
                            While using the hammock is a nice idea, you'd be [b]far[/b] too exposed.
                            You decide to choose [b]safety[/b] over [i]comfort[/i].
                            The ground is where you'll sleep tonight.
                        [/p]
                        [p]
                            You wander a little further into the woods, away from your shed.
                            After a solid minute of walking, you stumble upon a bush large enough to conceal your entire body.
                            You sigh loudly, getting flashbacks from your younger days as you reluctantly lower yourself into the foliage,
                            your small frame coming to rest on the cold ground at the very root of the shrub.
                        [/p]
                        [p]
                            While by no means warm, at least the plant's leaves somewhat protect you from the freezing cold washing over the land.
                            This isn't your first time sleeping on the forest floor, but your mind is too weary to reminisce right now.
                            Curling up into a ball, you slowly drift to sleep.
                        [/p]
                        [p]
                            You wake up far earlier than you'd have liked, to the feeling of something small [i]crawling[/i] over your scales.
                            At first you think you're dreaming, but the sensation persist.
                            It moves along your exposed belly, creeping every which way.
                        [/p]
                        [p]
                            As you open your heavy eyelids and stare down at your body, you are greeted with the cause of the odd sensation.
                            A bunch of ants are crawling about your lower body, many more swarming out of a nearby anthill that you failed to spot earlier.
                            Worse yet, these aren't just regular ants.
                            Even in the near pitch-blackness of night, you can make out their distinctly [color=pink]pink[/color] hue.
                            These are the dreaded [b]Venusian ants[/b], 
                            whose bite is said to inject a potent aphrodisiac strong enough for ten of them to make a male horse erect.
                        [/p]
                        [p]
                            Your mind instantly snaps awake as one of the small bugs digs its pincers into your scales.
                            It doesn't [i]hurt[/i] more than a small pinch, yet your groin instantly responds to the venom,
                            ${male ? "your member stirring to life inside of your cockslit" : "your slit growing just a little more moist"}.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the ant choice button.
     */
    function buildAntChoice() {
        var disabled = ctrl.unsaved.hasMadeGroundChoice() ? "disabled" : "";
        var isSelected = ctrl.unsaved.getGroundChoice() === HomeGroundChoice.ANT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${antId}" ${disabled}>Want more</button>`;

        var tooltip = global.parser.parse(`The aphrodisiac feels good. [b]So[/b] good. You want [b]more[/b] of their venom. You [b]need[/b] more of it.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 210px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildShakeChoice() {
        var disabled = ctrl.unsaved.hasMadeGroundChoice() ? "disabled" : "";
        var isSelected = ctrl.unsaved.getGroundChoice() === HomeGroundChoice.SHAKE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${shakeId}" ${disabled} style="font-size: 18px;">Shake them off</button>`;

        var tooltip = `Get the little things off of you before they can do more harm!`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildAntChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildShakeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + antId), $("#" + shakeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerAntListener() {
        $("#" + antId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unsaved.makeGroundChoice(HomeGroundChoice.ANT);
            loadNextPart(toAnt);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerShakeListener() {
        $("#" + shakeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unsaved.makeGroundChoice(HomeGroundChoice.SHAKE);
            loadNextPart(toShake);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeGround";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unsaved.getSleepChoice() !== HomeUnsavedChoice.GROUND) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toAnt.build();
        part += toShake.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unsaved.getSleepChoice() !== HomeUnsavedChoice.GROUND) {
            return;
        }

        toAnt.registerListeners();
        toShake.registerListeners();

        if (ctrl.unsaved.hasMadeGroundChoice()) {
            return;
        }

        registerAntListener();
        registerShakeListener();
    };
}

var HomeGroundChoice = {
    ANT: 1,
    SHAKE: 2
};

// END scenes\Home\unsaved\ground\HomeGround.js

// scenes\Home\unsaved\hammock\HomeHammock.js

/**
 * Hammock section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeHammock(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueHamletId = `toHamletContinueButton`;
    var continueVaranarId = `toVaranarContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide you'd rather not risk sleeping inside of your home in case the Rat's men return.
                            Seeing as you don't think you'd get much sleep at all on the cold, hard ground, 
                            you choose to spend your night on the hammock, 
                            an excellent compromise between comfort and the risk of getting caught by [i]whoever[/i] is looking for you.
                        [/p]
                        [p]
                            Even [i]if[/i] someone were to come [i]back[/i] to your shack during the night, 
                            you reason there's no way they'd come looking for you in the forest itself.
                            While you'll be suspended above any sort of concealing shrubbery, 
                            your slightly more remote location as well as the darkness of night are sure to hide you from any prying eyes.
                        [/p]
                        [p]
                            You march into the woods beyond your hovel.
                            Even you, having a good idea of where your hammock [i]should[/i] be, have a hard time actually [b]finding[/b] it in the middle of night.
                            Only after a good five minutes of determined searching do you finally come across your chosen nighttime solution.
                            Of course, with someone sleeping in it, it's bound to be a little more visible than when empty, but oh well.
                            You've made your bed, and now you're going to lie in it.
                        [/p]
                        [p]
                            You crawl into the hammock, and after a bit of fumbling, manage to reach a comfortable resting position.
                            The only thing you didn't consider was the temperature.
                            It's pretty chilly outside, and you have nothing to cover yourself with.
                            In an attempt to conserve precious body heat, you curl up into a ball.
                            Shuddering from the cold, you slowly drift to sleep.
                        [/p]
                        [p]
                            While the near-freezing temperature rouses you from your slumber more than once,
                            it does so for only a few minutes at a time before you drift back into unconsciousness.
                            You manage to get a decent amount of oft-interrupted rest in.
                        [/p]
                        [p]
                            As the sun begins to rise, you wake up properly, chilled to the very bone.
                            Shivering uncontrollably, you slip out of your hammock.
                            You breathe some warm air onto your ice-cold scaly hands and rub them together to create friction.
                        [/p]
                        [p]
                            While you wish you'd be able to wake up nice and slow,
                            your frigid surroundings spur you to immediately begin to move about in order to generate some much-needed body heat.
                            You rub your eyes, stretch your legs, and walk back to your shed.
                        [/p]
                        [p]
                            It looks to be in exactly the same condition you left it yesterday.
                            You're starting to regret not sleeping indoors.
                            But alas, what is done, is done.
                        [/p]
                        [p]
                            You continue to warm up your body by rubbing your legs and your arms with your hands.
                            With the sun gradually warming your surroundings, 
                            you ultimately manage to reach the point where you're no longer [b]too[/b] cold to think.
                        [/p]
                        [p]`;

        var gold = ctrl.gold();
        if (!gold) {
            toReturn += `   You are completely [b]broke[/b]. 
                            You have no money whatsoever to pay your overdue debts with.
                            Desperate times call for desperate measures.`;
        } else if (gold > 150) {
            toReturn += `   You have quite a bit of money, but sadly it's still not enough to pay your overdue debt.`;
        } else if (gold > 203) {
            toReturn += `   You have [b]a lot[/b] of money, more than you gained from just robbing the caravan, even.
                            You're [b]very[/b] close to being able to pay off your debt.
                            You just need a [i]teensy[/i] bit more.
                            But you aren't going to get it from these woods.`;
        }

        toReturn += `       You decide to head out, to seek your fortune in the great city of Varanar.
                            Maybe you'll stop at the small hamlet outside of the city walls along the way.
                            If anything, you hope you'll at least sleep somewhere a little warmer tonight.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueHamletId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `To the hamlet &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${continueVaranarId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `To Varanar &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueHamletId), $("#" + continueVaranarId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }
    
    /**
     * Adds a listener to the hamlet button.
     */
    function registerHamletListener() {
        $("#" + continueHamletId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.hamletSceneName, ctrl.hamletFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeHammock";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unsaved.getSleepChoice() !== HomeUnsavedChoice.HAMMOCK) {
            return "";
        }
        global.ui.applyTheme("home_dawn");
        var change = ctrl.unsaved.hammockChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `<span class="todo">You can only go to the hamlet.</span>`;
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unsaved.getSleepChoice() !== HomeUnsavedChoice.HAMMOCK) {
            return;
        }

        // REGISTER CHILDREN HERE

        if (ctrl.hasContinued()) {
            return;
        }

        registerHamletListener();
    };
}

// END scenes\Home\unsaved\hammock\HomeHammock.js

// scenes\Home\unsaved\inside\HomePillaged.js

/**
 * Pillaged section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomePillaged(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueHamletId = `toHamletContinueButton`;
    var continueVaranarId = `toVaranarContinueButton`;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Your home may be pillaged, but that doesn't mean you can't still sleep in it.
                            Whoever raided your place is probably long-gone.
                            Sticking around a seemingly abandoned hovel would be meaningless for them,
                            and nobody's going to come looking for you while it's dark out.
                        [/p]
                        [p]
                            Having mentally reassured yourself that everything will be just fine, you walk up to the front-end of your shed.
                            You try to open the busted door, but to your annoyance, this causes it to fall off of its hinges completely, 
                            tumbling down against the wooden back-wall of your home, nearly taking the entire ramshackle thing down with it.
                        [/p]
                        [p]`;

        var male = ctrl.gender() === GenderEnum.MALE;
        if (male) {
            toReturn += `   You let out a loud sigh.
                            Stepping inside, over the remains of the entryway, you use your masculine strength to lift up the door.
                            It's human-sized, but that doesn't stop you from putting it back onto its hinges.
                        [/p]
                        [p]
                            Slightly inspired by your own prowess, `;
        } else {
            toReturn += `   You let out a loud sigh.
                            Stepping inside, over the remains of the entryway, you try, but fail to lift up your door.
                            It's human-sized, making it far too heavy for someone of your posture to properly handle.
                        [/p]
                        [p]
                            Slightly disheartened, `;
        }

        toReturn += `       you begin to collect the scattered remains of the bale of hay that you call your [i]bed[/i].
                            Those [b]savages[/b] completely [b]tore[/b] it apart, probably thinking you had some valuables concealed within.
                            You almost feel like a bird making a nest.
                            It takes a couple of minutes, but you eventually manage to form a small pile of straw, big enough for you to sleep on.
                        [/p]
                        [p]
                            You lie down on top of the warm, comfortable former-fodder.`;

        if (male) {
            toReturn += `   The hay, together with your fixed door, provide for a warm and snug environment.`;
        } else {
            toReturn += `   The hay shields you from the cold seeping in through the busted door.`;
        }

        toReturn += `       Due to sheer exhaustion, you manage to drift to sleep rather quickly.
                        [/p]
                        [p]
                            ${male ? "Thanks to your fixed door," : "Despite your door being broken,"} you end up having a nice, if brief, little nap.
                            You slowly wake up hours later, completely [b]buried[/b] in the pile of straw, your scaly body being nurtured by the lingering warmth it provides.
                            While you'd love to stay and rest for just a bit longer, dawn is breaking.
                            You were right about people [i]not[/i] looking for you during the night, but you don't want to risk sticking around during daytime.
                        [/p]
                        [p]
                            You rise from your nest feeling quite rejuvinated.
                            Your short slumber did you good.
                            It allowed your brain and your body both to recover significantly.
                            You are ready to put the trials and tribulations of yesterday aside, and do whatever you need to do to make it through another day.
                        [/p]
                        [p]
                            You step out of the ${male ? "fixed" : "busted"} door, and are instantly greeted with a gust of cool morning air.
                            You breathe in deeply, clearing your mind before thinking about the future.`;

        var gold = ctrl.gold();
        if (!gold) {
            toReturn += `   You are completely [b]broke[/b]. 
                            You have no money whatsoever to pay your overdue debts with.
                            Desperate times call for desperate measures.`;
        } else if (gold > 150) {
            toReturn += `   You have quite a bit of money, but sadly it's still not enough to pay your overdue debt.`;
        } else if (gold > 203) {
            toReturn += `   You have [b]a lot[/b] of money, more than you gained from just robbing the caravan, even.
                            You're [b]very[/b] close to being able to pay off your debt.
                            You just need a [i]teensy[/i] bit more.
                            But you aren't going to get it from these woods.`;
        }

        toReturn += `       You decide to head out, to seek your fortune in the great city of Varanar,
                            and maybe take a stop at the small hamlet outside of the city walls along the way.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.hasContinued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueHamletId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `To the hamlet &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${continueVaranarId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `To Varanar &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueHamletId), $("#" + continueVaranarId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }
    
    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }
    
    /**
     * Adds a listener to the hamlet button.
     */
    function registerHamletListener() {
        $("#" + continueHamletId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.hamletSceneName, ctrl.hamletFolder);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homePillaged";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unsaved.getSleepChoice() !== HomeUnsavedChoice.INSIDE) {
            return "";
        }
        global.ui.applyTheme("home_dawn");
        var change = ctrl.unsaved.pillageChange();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += change;
        part += buildChoice();
        part += `<span class="todo">You can only go to the hamlet.</span>`;
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unsaved.getSleepChoice() !== HomeUnsavedChoice.INSIDE) {
            return;
        }

        // REGISTER CHILDREN HERE

        if (ctrl.hasContinued()) {
            return;
        }

        registerHamletListener();
    };
}

// END scenes\Home\unsaved\inside\HomePillaged.js

// scenes\Home\unsaved\HomeUnsaved.js

/**
 * Unsaved section of the home scene.
 * @param {HomeController} controller The home controller.
 */
function HomeUnsaved(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var insideId = "unsavedInsideButton";
    var hammockId = "unsavedHammockButton";
    var groundId = "unsavedGroundButton";

    var toInside = new HomePillaged(ctrl);
    var toHammock = new HomeHammock(ctrl);
    var toGround = new HomeGround(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[st]Late arrival[/st]`;
        var deepwoods = global.kobold.history.sceneStackHas("RiversideMarsh");
        toReturn += `   [p]
                            After a short trek through the forest, you begin to near your small hovel, located near the northern edge of the woods.
                            Thank goodness you're not far off, because ${deepwoods ? "it's getting [b]very[/b] late" : "night is upon you"}, 
                            and you know how [i]dangerous[/i] things can get around the woodland after the sun goes down.
                        [/p]
                        [p]
                            Your wooden [i]shed[/i], illuminated dimly by moonlight shining through the treetops, soon comes into view.
                            As you approach, however, you notice something's [i]wrong[/i].
                            The door to your humble abode hangs crooked on its sturdy hinges, obviously [b]forced[/b] open.
                            You've no doubt that this was the work of the Three-Tailed Rat's men.
                            The only question that lingers in your mind is whether or not they're still in the area.
                        [/p]
                        [p]
                            You skulk around your shoddy shack, keeping your eyes peeled and your ears open,
                            yet you can't quite find anyone at first glance.
                            You reckon they must have left after discovering that your home contains nothing but the hay you sleep on.
                            Too bad for them, you carry about all you own, which isn't really all that much.
                        [/p]
                        [p]
                            You let out a loud yawn, your mind weary from the [i]active[/i] day you've had, worn out both physically as well as emotionally.
                            While you'd love to pull an all-nighter just to see if any of those no-good [b]looters[/b] would have the [b]guts[/b] to return 
                            to your shack while you're present, both your body and your brain are [b]far[/b] too tired to stay up for much longer, let alone the entire night.
                        [/p]
                        [p]
                            However sleepy you may be, you're still alert enough to realize that [i]should[/i] whoever raided your home return to find you sleeping,
                            that would [b]not[/b] end well for you.
                            Then again, why [i]would[/i] they return?
                            They didn't find anything of value, and you doubt anyone would come looking for you in the middle of night.
                            But with the Three-Tailed Rat Clan, you never know.
                        [/p]
                        [p]
                            Besides sleeping in what remains of your shed, there's also the possibility of resting outdoors,
                            which might prove to be a little more safe.
                        [/p]
                        [p]
                            A few months ago, you found a fisherman's abandoned net washed up on the bank of the river.
                            Instead of letting it go to waste, you tied up the intertwined rope-webbing in between two trees,
                            fashioning yourself a hammock not too far off from your home.
                            While undoubtedly more [i]comfortable[/i] than sleeping on the ground,
                            you would also be quite a bit more [i]visible[/i] and [i]vulnerable[/i] as opposed to sleeping in a bush.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the inside choice button.
     */
    function buildInsideChoice() {
        var disabled = ctrl.unsaved.hasMadeSleepChoice() ? "disabled" : "";
        var isSelected = ctrl.unsaved.getSleepChoice() === HomeUnsavedChoice.INSIDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${insideId}" ${disabled}>Inside</button>`;

        var tooltip = `Your home may be plundered, but you doubt they took <span class="italic">all</span> of your hay.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the hammock choice button.
     */
    function buildHammockChoice() {
        var disabled = ctrl.unsaved.hasMadeSleepChoice() ? "disabled" : "";
        var isSelected = ctrl.unsaved.getSleepChoice() === HomeUnsavedChoice.HAMMOCK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${hammockId}" ${disabled} style="font-size: 18px;">On the hammock</button>`;

        var tooltip = `The hammock is comfortable enough to get some decent rest in, but you'd be pretty exposed during the night.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the ground choice button.
     */
    function buildGroundChoice() {
        var disabled = ctrl.unsaved.hasMadeSleepChoice() ? "disabled" : "";
        var isSelected = ctrl.unsaved.getSleepChoice() === HomeUnsavedChoice.GROUND;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${groundId}" ${disabled}>On the ground</button>`;

        var tooltip = `Your home is no longer safe, and you'd rather not sleep in a strung-up net with a bunch of people potentially hunting you.
                       Rest on the cold ground, in a nearby bush.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 300px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You sleep: </div>`;
        choice += buildInsideChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildHammockChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildGroundChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + insideId), $("#" + hammockId), $("#" + groundId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerInsideListener() {
        $("#" + insideId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unsaved.makeSleepChoice(HomeUnsavedChoice.INSIDE);
            loadNextPart(toInside);
        });

    }

    /**
     * Adds a listener to the button.
     */
    function registerHammockListener() {
        $("#" + hammockId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unsaved.makeSleepChoice(HomeUnsavedChoice.HAMMOCK);
            loadNextPart(toHammock);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerGroundListener() {
        $("#" + groundId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unsaved.makeSleepChoice(HomeUnsavedChoice.GROUND);
            loadNextPart(toGround);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "homeUnsaved";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.status() !== HomeStatusEnum.UNSAVED) {
            return "";
        }
        if (!ctrl.unsaved.hasMadeSleepChoice()) {
            ctrl.introAudio.playSettingsDependent();
        }
        global.ui.applyTheme("home_night");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toInside.build();
        part += toHammock.build();
        part += toGround.build();
        part += `</div>`;

        // BUILD CHILDREN HERE

        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.status() !== HomeStatusEnum.UNSAVED) {
            return;
        }

        toInside.registerListeners();
        toHammock.registerListeners();
        toGround.registerListeners();

        if (ctrl.unsaved.hasMadeSleepChoice()) {
            return;
        }

        registerInsideListener();
        registerHammockListener();
        registerGroundListener();
    };
}

var HomeUnsavedChoice = {
    INSIDE: 1,
    HAMMOCK: 2,
    GROUND: 3
};

// END scenes\Home\unsaved\HomeUnsaved.js

// scenes\Home\unsaved\HomeUnsavedController.js

/**
 * Controller for the unsaved part of the home scene.
 * @param {string} sceneId The id of the scene.
 */
function HomeUnsavedController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;

    var sexExId = "unsavedSexEx";
    var sleepId = "unsavedSleep";
    var groundId = "unsavedGround";
    var shakeId = "unsavedShake";
    var hammockChangeId = "unsavedHammockChange";
    var pillageChangeId = "unsavedPillageChange";
    var wetChangeId = "unsavedWetChange";
    var masturbateChangeId = "unsavedMasturbateChange";
    var antChangeId = "unsavedAntChange";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Makes the sleep choice.
     * @param {HomeUnsavedChoice} choice The choice.
     */
    this.makeSleepChoice = function (choice) {
        global.kobold.history.setKey(id, sleepId, choice);
    };

    /**
     * Fetches the sleep choice from history and returns it.
     * May return undefined if the sleep choice has not yet been set.
     */
    this.getSleepChoice = function () {
        return global.kobold.history.getKey(id, sleepId);
    };

    /**
     * Returns true if the sleep choice has been made.
     */
    this.hasMadeSleepChoice = function () {
        return global.kobold.history.has(id, sleepId);
    };


    /**
     * Makes the ground choice.
     * @param {HomeGroundChoice} choice The choice.
     */
    this.makeGroundChoice = function (choice) {
        global.kobold.history.setKey(id, groundId, choice);
    };

    /**
     * Fetches the ground choice from history and returns it.
     * May return undefined if the ground choice has not yet been set.
     */
    this.getGroundChoice = function () {
        return global.kobold.history.getKey(id, groundId);
    };

    /**
     * Returns true if the ground choice has been made.
     */
    this.hasMadeGroundChoice = function () {
        return global.kobold.history.has(id, groundId);
    };

    this.hammockChange = function () {
        global.kobold.history.doIfNotSet(id, hammockChangeId, true, function () {
            global.kobold.status.physical.decrease(10, "Near-hypothermia");
            global.kobold.status.stamina.increase(15, "Poor sleep");
            if (global.kobold.status.exertion.get()) {
                global.kobold.history.setKey(id, sexExId, true);
                global.kobold.status.exertion.decrease(20, "Slept");
            }
        });

        var changes = [
            new Change("Near-hypothermia", "Health", -10),
            new Change("Poor sleep", "Stamina", +15)
        ];
        if (global.kobold.history.has(id, sexExId)) {
            changes.push(new Change("Slept", "Sexual exhaustion", -20, true));
        }
        return global.ui.change.buildNotification(changes);
    };

    this.pillageChange = function () {
        global.kobold.history.doIfNotSet(id, pillageChangeId, true, function () {
            global.kobold.status.physical.increase(25, "Short nap");
            global.kobold.status.mental.increase(25, "Short nap");
            global.kobold.status.stamina.increase(25, "Short nap");
            if (global.kobold.status.exertion.get()) {
                global.kobold.history.setKey(id, sexExId, true);
                global.kobold.status.exertion.decrease(25, "Short nap");
            }
        });

        var changes = [
            new Change("Short nap", "Health", +25),
            new Change("Short nap", "Sanity", +25),
            new Change("Short nap", "Stamina", +25)
        ];
        if (global.kobold.history.has(id, sexExId)) {
            changes.push(new Change("Short nap", "Sexual exhaustion", -25, true));
        }
        return global.ui.change.buildNotification(changes);
    };

    this.wetChange = function () {
        global.kobold.history.doIfNotSet(id, wetChangeId, true, function () {
            global.kobold.status.physical.increase(15, "Brief nap");
            global.kobold.status.mental.increase(10, "Mental victory");
            global.kobold.status.stamina.increase(15, "Brief nap");
            global.kobold.status.exertion.increase(10, "Unsated lust");
        });
        return global.ui.change.buildNotification([
            new Change("Brief nap", "Health", +15),
            new Change("Mental victory", "Sanity", +10),
            new Change("Brief nap", "Stamina", +15),
            new Change("Unsated lust", "Sexual exhaustion", +10, true)
        ]);
    };

    this.masturbateChange = function () {
        global.kobold.history.doIfNotSet(id, masturbateChangeId, true, function () {
            global.kobold.status.physical.decrease(5, "Sleep deprivation");
            global.kobold.status.mental.decrease(5, "Sleep deprivation");
            global.kobold.status.stamina.decrease(5, "Sleep deprivation");
            global.kobold.status.exertion.increase(20, "Masturbation");
        });
        return global.ui.change.buildNotification([
            new Change("Sleep deprivation", "Health", -5),
            new Change("Sleep deprivation", "Sanity", -5),
            new Change("Sleep deprivation", "Stamina", -5),
            new Change("Masturbation", "Sexual exhaustion", +20, true)
        ]);
    };

    this.antChange = function (gender) {
        global.kobold.history.doIfNotSet(id, antChangeId, true, function () {
            global.kobold.status.physical.decrease(5, "Sleep deprivation");
            global.kobold.status.mental.decrease(10, "Fucked an anthill");
            global.kobold.status.stamina.decrease(10, "Fucked an anthill");
            global.kobold.status.exertion.increase(25, "Fucked an anthill");
            if (gender === GenderEnum.MALE) {
                global.kobold.body.growDick(global.kobold.body.dick + 1);
            } else {
                global.kobold.body.growBreasts(global.kobold.body.breasts + 1);
            }
        });

        var changes = [
            new Change("Sleep deprivation", "Health", -5),
            new Change("Fucked an anthill", "Sanity", -10),
            new Change("Fucked an anthill", "Stamina", -10),
            new Change("Fucked an anthill", "Sexual exhaustion", +25, true)
        ];

        if (gender === GenderEnum.MALE) {
            changes.push(new Change("Lasting effects", "Dick size", +1));
        } else {
            changes.push(new Change("Lasting effects", "Breast size", +1));
        }

        return global.ui.change.buildNotification(changes);
    };

    function ShakeController(key) {
        this.get = function () {
            return global.kobold.history.getKey(id, key);
        };
        this.set = function (choice) {
            global.kobold.history.setKey(id, key, choice);
        };
        this.has = function () {
            return global.kobold.history.has(id, key);
        };
    }

    this.shake = new ShakeController(shakeId);
}

// END scenes\Home\unsaved\HomeUnsavedController.js

// scenes\Home\HomeController.js

/**
 * Controller for the Home scene.
 * @param {string} sceneId The id of the scene.
 */
function HomeController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;
    var statusId = "status";
    var genderId = "gender";
    var autoSaveId = "autoSave";
    var goldId = "gold";
    var continueId = "continue";
    var underwearId = "underwear";
    var pantsId = "pants";
    var braId = "bra";
    var shirtId = "shirt";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.hamletSceneName = "SekiWorkshop";
    this.hamletFolder = "HamletDay2";
    this.saved = new HomeSavedController(id);
    this.unsaved = new HomeUnsavedController(id);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /*
     * Stores the status of the home at scene-entry time.
     */
    this.status = function () {
        global.kobold.history.setIfNotSet(id, statusId, global.kobold.history.shared.home.status);
        return global.kobold.history.getKey(id, statusId);
    };

    /*
     * Stores the kobold's base gender.
     */
    this.gender = function () {
        global.kobold.history.setIfNotSet(id, genderId, global.kobold.body.getBaseGender());
        return global.kobold.history.getKey(id, genderId);
    };

    /*
     * Autosaves the game if necessary.
     */
    this.autoSave = function () {
        global.kobold.history.doIfNotSet(id, autoSaveId, true, function () {
            global.saveManager.autoSave();
        });
    };


    /**
     * Logs and returns the amount of gold the kobold had at the start of the scene.
     */
    this.gold = function () {
        global.kobold.history.setIfNotSet(id, goldId, global.kobold.inventory.coins);
        return global.kobold.history.getKey(id, goldId);
    };

    /**
     * Signals the scene has been continued.
     */
    this.continue = function () {
        global.kobold.history.setKey(id, continueId, true);
    };

    /**
     * Returns whether or not the scene has continued.
     * @returns {boolean} True if the scene has continued, false otherwise.
     */
    this.hasContinued = function () {
        return global.kobold.history.has(id, continueId);
    };

    this.pants = function () {
        global.kobold.history.setIfNotSet(id, pantsId, global.kobold.inventory.equipment.pants);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, pantsId));
    };

    this.bra = function () {
        global.kobold.history.setIfNotSet(id, braId, global.kobold.inventory.equipment.bra);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, braId));
    };
    
    this.shirt = function () {
        global.kobold.history.setIfNotSet(id, shirtId, global.kobold.inventory.equipment.shirt);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, shirtId));
    };

    this.underwear = function () {
        global.kobold.history.setIfNotSet(id, underwearId, global.kobold.inventory.equipment.underwear);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, underwearId));
    };
    
    this.introAudio =  new AudioController(id, "res/audio/home/tension.mp3", 100);
}

// END scenes\Home\HomeController.js

// scenes\Home\Home.js

/**
 * The Home scene. One of the options following the Forest.
 */
function Home() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "Home";
    var sceneNameUnsaved = "Home - Late";
    var sceneNameSaved = "Home - Early";
    var ctrl = new HomeController(sceneId);
    var saved = new HomeSaved(ctrl);
    var unsaved = new HomeUnsaved(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Returns true if the home was saved, and false otherwise.
     * @returns {boolean} True if the home was saved, and false otherwise.
     */
    function hasSaved() {
        var status = ctrl.status();
        return (status === HomeStatusEnum.SAVED);
    }

    /**
     * Sets this scene's name.
     */
    function setName() {
        var sceneName = hasSaved() ? sceneNameSaved : sceneNameUnsaved;
        that.name = sceneName;
    }

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]home[/t]`);
        tab += hasSaved() ? saved.build() : unsaved.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = "Error";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setName();
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        hasSaved() ? saved.registerListeners() : unsaved.registerListeners();
        ctrl.autoSave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new Home();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\Home\Home.js



// END SCENE Home



// SCENE SekiWorkshop

// scenes\HamletDay2\SekiWorkshop\center\SekiWorkshopCenter.js

/**
 * Buy section of the Workshop scene.
 * @param {SekiWorkshopController} controller The Seki controller.
 */
function SekiWorkshopCenter(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var breedingId = "centerBreedingButton";
    var farmHomeId = "centerFarmHomeButton";
    var wellId = "centerWellButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Disheartened, you leave Seki's store and walk towards the small well at the very center of the hamlet.
                            Leaning against it, you take a moment to collect your thoughts.
                            ${global.kobold.history.shared.enslavement.noonTwo.boughtBySeki ? `The stage you were put on display on has already been torn down.` : ``}
                        [/p]
                        [p]
                            Seki is a brilliant alchemist.
                            Perhaps even one of the best ones to have ever lived.
                            But to you, he's more than that.
                            He's your kind of, not really, somewhat of a [i]father figure[/i].
                            You wouldn't know what you'd do if he was gone.
                            You're not ready to lose him.
                        [/p]
                        [p]
                            You can't.
                            You [b]won't[/b].
                            You're [b]not[/b] going to.
                            Even if it means you need to travel through [b]hell[/b] itself to drag his old,
                            feathery ass out of the afterlife,
                            you are [b]not[/b] losing Seki, [b]no matter what[/b].
                            ${global.kobold.history.shared.enslavement.noonTwo.boughtBySeki ? `He saved you, now you're going to save him.` : ``}
                            If you need to punch the Grim Reaper himself in the face, then so be it.
                            That bony bastard is [b]not[/b] taking your feathery father from you.
                            [b]Over your dead body.[/b]
                        [/p]
                        [p]
                            You kick a small rock in frustration.
                            It skips along the dirt for a few feet, then comes to rest.
                            You wipe a few tears from your eyes.
                            Luckily, nobody is around to see you.
                            This place looks deserted.
                            Everyone is probably indoors or out working the field or in Varanar or wherever.
                        [/p]
                        [p]
                            You can't sit around here moping all day long.
                            Well, you could, but Seki wouldn't want that.
                            He'd probably chase you from the hamlet himself if he could run for more than five feet without being out of breath.
                            You're going to miss him.
                            A lot.
                            You almost burst into tears.
                            It's just [i]not fair[/i].
                            ${global.kobold.history.shared.enslavement.noonTwo.boughtBySeki ? `First you become a [b]slave[/b], now this.` : ``}
                        [/p]
                        [p]
                            Taking a look around, you spot three places of interest that might prove worthy of exploration.
                            The breeding farm across the road from Seki's workshop is an obvious choice.
                            Both the fancy-looking house as well as the attached barn could be full of stuff for you to steal.
                            You don't know [i]who[/i] owns that place, though.
                            You've heard Seki talk about some woman constantly buying up all of his ${global.kobold.inventory.createItemWithTooltip(new FemPlus())}s.
                            Could be related, but you don't really know for certain.
                        [/p]
                        [p]
                            Alternatively, you could break into a random farmer's house.
                            Should be empty, save for a stay-at-home spouse or perhaps some offspring.
                            ${global.kobold.history.shared.enslavement.noonTwo.boughtBySeki ? `Farmers have returned to their fields after the auction.` : ``}
                            There probably won't be [i]that much[/i] to steal, but it's a safer bet than the breeding farm.
                        [/p]
                        [p]
                            The final point of interest is the well you're leaning against.
                            It might prove valuable to give it a more [i]thorough[/i] inspection.
                            People around these rural parts are pretty superstitious.
                            You wouldn't be surprised if they hurled an entire fortune worth of individual coins down the waterhole,
                            praying for good luck or wishing for a better life.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${breedingId}" ${disabled ? "disabled" : ""} style="margin: auto; float: left;">`;
        choice += `Breeding farm &gt;&gt`;
        choice += `</button>`;
        choice += `<button class="button bigbutton continuebutton" id="${farmHomeId}" ${disabled ? "disabled" : ""} style="margin: auto; float: right;">`;
        choice += `Farmer's home &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        choice += `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${wellId}" ${disabled ? "disabled" : ""} style="margin: auto;">`;
        choice += `Well investigation &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + breedingId), $("#" + farmHomeId), $("#" + wellId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerBreedingListener() {
        $("#" + breedingId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.breedingSceneName, ctrl.folderName);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFarmHouseListener() {
        $("#" + farmHomeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.farmHomeSceneName, ctrl.folderName);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerWellListener() {
        $("#" + wellId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.wellSceneName, ctrl.folderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "sekiCenter";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.free.continued() && !ctrl.buyContinue.continued()) {
            return "";
        }
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.free.continued() && !ctrl.buyContinue.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerBreedingListener();
        registerFarmHouseListener();
        registerWellListener();
    };
}

// END scenes\HamletDay2\SekiWorkshop\center\SekiWorkshopCenter.js

// scenes\HamletDay2\SekiWorkshop\free\SekiWorkshopFree.js

/**
 * Free section of the Workshop scene.
 * @param {SekiWorkshopController} controller The Seki controller.
 */
function SekiWorkshopFree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "freeContinueButton";

    var next = new SekiWorkshopCenter(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            After about a minute of waiting, you begin to get a bit impatient.
                            You ring the bell a few more times, until you finally get a response in the form of a raspy,
                            bass shout coming from the darkness on the far end of the counter:
                            "[color=brown]Hold your horses!
                            Settle down!
                            I'm on my way!
                            Can't an old man get a minute to finish grinding his herbs?!
                            You youngsters are always so [b]restive[/b]![/color]"
                        [/p]
                        [p]
                            Loudly grumbling, the elderly owl emerges from the shadows, stepping into the dim candlelight,
                            which illuminates his once-brown, but now pale grey feathers.
                            Readjusting the heavy, thick spectacles atop his beak,
                            the angry frown on the bird's face changes to a happy,
                            heartfelt smile as he realizes the tiny little thing that interrupted his machinations was none other but, well, [b]you[/b].
                        [/p]
                        [p]
                            "[color=brown]Ahhh!
                            My favorite customer!
                            Uhhh...[/color]"
                            The avian brings his feathery hand towards his forehead, wiping his brow as he visibly tries his very hardest to recall your name.
                            You know he didn't forget you.
                            His old mind just can't remember things quite as well as it used to.
                            You understand.
                            Returning a smile, you coyly hint that your name is ${ctrl.name()}.
                        [/p]
                        [p]
                            "[color=brown]Yes, yes!
                            ${ctrl.name()}!
                            I knew that!
                            I was just [b]catching my breath[/b] from walking all the way over here.[/color]"
                            Silly old geezer.
                            "[color=brown]Let me take a good look at you.[/color]"
                            He continues, leaning forwards over the counter, gazing down at you with his large owl eyes.
                        [/p]
                        `;
        if (ctrl.health() < 80) {
            toReturn += `[p] 
                            "[color=brown]You look like you've seen better days!
                            Who'd you get in a tussle with this time, eh?
                            It's probably best if you [b]don't[/b] tell me.
                            If they can beat up a fierce youngling like yourself, an old geezer like me wouldn't stand much of a chance.
                            Twenty years ago, I'd have shown those ruffians what-for, let me tell you![/color]"
                        [/p]`;
        } else {
            toReturn += `[p] 
                            "[color=brown]You look fit as ever!
                            Taking good care of yourself, eh?
                            'Atta ${ctrl.gender() === GenderEnum.MALE ? "boy" : "girl"}.
                            It'd pain an old man's heart to see a pretty little snout like yours get hurt.
                            Do me a favor and keep yourself out of harm's way, won't you?
                            Not for your sake, but for mine.[/color]"
                        [/p]`;
        }

        toReturn += `   [p] 
                            Then, Seki begins to reminisce, as he always does when you come to visit him.
                            "[color=brown]When I was your age, I was out explorin' the world.
                            I fought trolls, I climbed mountains, I even found what I believed to be the fabled well of youth![/color]"
                            A small pause.
                            A smirk forms on Seki's face.
                            He winks at you.
                            "[color=brown]Turns out it was just an ordinary river spring, but you probably could have guessed that by taking a look at me, eh?[/color]"
                        [/p]
                        [p]
                            The old man bursts into laughter, which quickly devolves into a coughing fit.
                            The smile is wiped off his face.
                            Hacking and wheezing, he spits a distinctly red-colored wad of phlegm from his beak onto the ground.
                            You fear for his well-being, the bird visibly hurting, but the moment he catches his breath,
                            he turns towards you, smile plastered across his face once more.
                            You don't know whether or not he's [i]acting[/i] tough.
                        [/p]
                        [p]
                            "[color=brown]Don't worry about me.[/color]"
                            Seki says, as if that's supposed to comfort you.
                            "[color=brown]Just a common cold.
                            Nothin' more.[/color]"
                            You mention that he has the [b]cure[/b] for that.
                            He once made you feel [b]much[/b] better with just a small potion.
                            The owl looks away.
                            All playfulness vanishes from his face, a disconcerting seriousness taking its place.
                            Tears form at the corners of his eyes.
                            He swallows deeply.
                            "[color=brown]It's not the kind of cold that can be cured.[/color]"
                        [/p]
                        [p]
                            The old man pokes a feathery finger underneath his glasses to wipe his eyes, in order to pretend he wasn't crying.
                            He turns towards you once more, a shaky, obviously faked grin on his face.
                            With the same finger he used to collect his tears, he points towards the empty coinpurse hanging from your neck.
                            "[color=brown]Well, you won't be buyin' much with that.
                            And you know the rule: no buyin' means no sellin'.[/color]"
                        [/p]
                        [p]
                            You're still stuck trying to process what he just said.
                            Is Seki going to...die?
                            The old man notices your confusion, and in an attempt to break the grim mood that has fallen upon the conversation,
                            he opens up a drawer, and fetches a familiarly shaped piece of metal.
                        [/p]
                        [p]
                            "[color=brown]Can't send you off without a means to fill up that empty pouch of yours, now can I?[/color]"
                            He asks, rhetorically, whilst handing you a single ${global.kobold.inventory.createItemWithTooltip(new Lockpick())}.
                            "[color=brown]Don't go using it all in one place now, eh?[/color]"
                            The bird jokes, seemingly [i]wanting[/i] to laugh, but not wanting to cough up another piece of his lungs.
                        [/p]
                        [p]
                            "[color=brown]Anyway, off you go.[/color]"
                            He continues, turning around and beginning to walk away at a slow pace.
                            "[color=brown]I'll be in the city for the next few days, but be sure to stop by here sometime before next week, you hear?
                            I want to see my little [b]${ctrl.gender() === GenderEnum.MALE ? "man" : "woman"}[/b] one last time before...[/color]"
                            Stopping mid-sentence, Seki lets out a loud, depressed sigh before silently disappearing into the darkness,
                            either to tend to his wounds, or to lament in solitude.
                        [/p]
                        [p]
                            Heartbroken, you hang your head and barely refrain from crying as you walk out of the old owl's shop.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the alternate content of this section.
     * @returns {string} The content.
     */
    function buildAltContent() {
        var toReturn = `[p]
                            Suddenly, Seki bursts into a coughing fit.
                            The heart-warming smile is wiped off his face.
                            Hacking and wheezing, he spits a distinctly red-colored wad of phlegm from his beak onto the ground.
                            You fear for his well-being, the bird visibly hurting, but the moment he catches his breath,
                            and notices you staring, a small smile creeps onto his face once more.
                            You don't know whether or not he's [i]acting[/i] tough.
                        [/p]
                        [p]
                            "[color=brown]Don't worry about me.[/color]"
                            The old bird says, as if that's supposed to comfort you.
                            "[color=brown]Just a common cold.
                            Nothin' more.[/color]"
                            You mention that he has the [b]cure[/b] for that.
                            He once made you feel [b]much[/b] better with just a small potion.
                            The owl looks away.
                            All playfulness vanishes from his face, a disconcerting seriousness taking its place.
                            Tears form at the corners of his eyes.
                            He swallows deeply.
                            "[color=brown]It's not the kind of cold that can be cured.[/color]"
                        [/p]
                        [p]
                            The avian pokes a feathery finger underneath his glasses to wipe his eyes, in order to pretend he wasn't crying.
                            He gives you a shaky, obviously faked grin.
                            "[color=brown]You worry about yourself first and foremost.
                            I...might not be around to save you next time.[/color]"
                            You're still stuck trying to process what he just said.
                            Is Seki going to...die?
                            The old man notices your confusion, and in an attempt to break the grim mood that has fallen upon the conversation,
                            he quickly changes the topic.
                        [/p]
                        [p]
                            "[color=brown]Anyway, you caught me right as I was about to head for Varanar.[/color]"
                            He begins to walk away to the back of his store, at a slow pace.
                            You're not sure whether you should follow him or not.
                            "[color=brown]I'll be in the city for the next few days, but be sure to stop by here sometime before next week, you hear?
                            I want to see my little [b]${ctrl.gender() === GenderEnum.MALE ? "man" : "woman"}[/b] one last time before...[/color]"
                            Stopping mid-sentence, Seki lets out a loud, depressed sigh before silently disappearing into the darkness,
                            either to tend to his wounds, to lament in solitude, or to leave via the back door.
                        [/p]
                        [p]
                            Heartbroken, you hang your head and barely refrain from bursting into tears once more as you walk out of the old owl's shop.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.free.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.free.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "sekiFree";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.entrance.continued() || ctrl.gold() >= ctrl.minBuyGold) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (global.kobold.history.shared.enslavement.noonTwo.boughtBySeki) {
            part += buildAltContent();
            part += ctrl.free.altChange();
        } else {
            part += buildContent();
            part += ctrl.free.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.entrance.continued() || ctrl.gold() >= ctrl.minBuyGold) {
            return;
        }

        next.registerListeners();

        if (ctrl.free.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\SekiWorkshop\free\SekiWorkshopFree.js

// scenes\HamletDay2\SekiWorkshop\shop\SekiWorkshopBuy.js

/**
 * Buy section of the Workshop scene.
 * @param {SekiWorkshopController} controller The Seki controller.
 */
function SekiWorkshopBuy(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "buyContinueButton";
    var next = new SekiWorkshopBuyContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            After about a minute of waiting, you begin to get a bit impatient.
                            You ring the bell a few more times, until you finally get a response in the form of a raspy,
                            bass shout coming from the darkness on the far end of the counter:
                            "[color=brown]Hold your horses!
                            Settle down!
                            I'm on my way!
                            Can't an old man get a minute to finish grinding his herbs?!
                            You youngsters are always so [b]restive[/b]![/color]"
                        [/p]
                        [p]
                            Loudly grumbling, the elderly owl emerges from the shadows, stepping into the dim candlelight,
                            which illuminates his once-brown, but now pale grey feathers.
                            Readjusting the heavy, thick spectacles atop his beak,
                            the angry frown on the bird's face changes to a happy,
                            heartfelt smile as he realizes the tiny little thing that interrupted his machinations was none other but, well, [b]you[/b].
                        [/p]
                        [p]
                            "[color=brown]Ahhh!
                            My favorite customer!
                            Uhhh...[/color]"
                            The avian brings his feathery hand towards his forehead, wiping his brow as he visibly tries his very hardest to recall your name.
                            You know he didn't forget you.
                            His old mind just can't remember things quite as well as it used to.
                            You understand.
                            Returning a smile, you coyly hint that your name is ${ctrl.name()}.
                        [/p]
                        [p]
                            "[color=brown]Yes, yes!
                            ${ctrl.name()}!
                            I knew that!
                            I was just [b]catching my breath[/b] from walking all the way over here.[/color]"
                            Silly old geezer.
                            "[color=brown]Let me take a good look at you.[/color]"
                            He continues, leaning forwards over the counter, gazing down at you with his large owl eyes.
                        [/p]
                        `;
        if (ctrl.health() < 80) {
            toReturn += `[p] 
                            "[color=brown]You look like you've seen better days!
                            Who'd you get in a tussle with this time, eh?
                            It's probably best if you [b]don't[/b] tell me.
                            If they can beat up a fierce youngling like yourself, an old geezer like me wouldn't stand much of a chance.
                            Twenty years ago, I'd have shown those ruffians what-for, let me tell you![/color]"
                        [/p]`;
        } else {
            toReturn += `[p] 
                            "[color=brown]You look fit as ever!
                            Taking good care of yourself, eh?
                            'Atta ${ctrl.gender() === GenderEnum.MALE ? "boy" : "girl"}.
                            It'd pain an old man's heart to see a pretty little snout like yours get hurt.
                            Do me a favor and keep yourself out of harm's way, won't you?
                            Not for your sake, but for mine.[/color]"
                        [/p]`;
        }

        toReturn += `   [p]
                            Noticing your well-filled coin purse, the bird's smirk deepens.
                            "[color=brown]Come across some gold, now have you?
                            While I sometimes forget, I [i]am[/i] running a store here.
                            Feel free to browse, just don't sample anything without asking me first.[/color]"
                        [/p]
                        [p][sneak]
                            You are in a safe zone.
                            You can use items from your inventory here.
                            You are in a store.
                            You can sell items from your inventory here.
                            You may buy items from the list below.
                            Keep in mind that all purchases are final.
                            Worn equipment can not be sold.
                            Buying equipment will discard your previous attire.
                        [/sneak][/p]`;
        toReturn += ctrl.buy.getGenerator().build();
        toReturn += `[p][sneak]
                            Remember that you can not drink potions or eat food outside of safe zones.
                            You might want to use some of the items you've purchased before continuing.
                        [/sneak][/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.buy.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.buy.continue();
            ctrl.buy.getGenerator().disable();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "sekiBuy";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.entrance.continued() || ctrl.gold() < ctrl.minBuyGold) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.entrance.continued() || ctrl.gold() < ctrl.minBuyGold) {
            return;
        }

        next.registerListeners();

        if (ctrl.buy.continued()) {
            return;
        }

        registerContinueListener();
        ctrl.buy.getGenerator().registerListeners();
    };
}

// END scenes\HamletDay2\SekiWorkshop\shop\SekiWorkshopBuy.js

// scenes\HamletDay2\SekiWorkshop\shop\SekiWorkshopBuyContinue.js

/**
 * Buy section of the Workshop scene.
 * @param {SekiWorkshopController} controller The Seki controller.
 */
function SekiWorkshopBuyContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "buyContinueReduxButton";
    var next = new SekiWorkshopCenter(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p] 
                            With your business concluded, Seki begins to reminisce, as he always does when you come to visit him.
                            "[color=brown]When I was your age, I was out explorin' the world.
                            I fought trolls, I climbed mountains, I even found what I believed to be the fabled well of youth![/color]"
                            A small pause.
                            A smirk forms on Seki's face.
                            He winks at you.
                            "[color=brown]Turns out it was just an ordinary river spring, but you probably could have guessed that by taking a look at me, eh?[/color]"
                        [/p]
                        [p]
                            The old man bursts into laughter, which quickly devolves into a coughing fit.
                            The smile is wiped off his face.
                            Hacking and wheezing, he spits a distinctly red-colored wad of phlegm from his beak onto the ground.
                            You fear for his well-being, the bird visibly hurting, but the moment he catches his breath,
                            he turns towards you, smile plastered across his face once more.
                            You don't know whether or not he's [i]acting[/i] tough.
                        [/p]
                        [p]
                            "[color=brown]Don't worry about me.[/color]"
                            Seki says, as if that's supposed to comfort you.
                            "[color=brown]Just a common cold.
                            Nothin' more.[/color]"
                            You mention that he has the [b]cure[/b] for that.
                            He once made you feel [b]much[/b] better with just a small potion.
                            The owl looks away.
                            All playfulness vanishes from his face, a disconcerting seriousness taking its place.
                            Tears form at the corners of his eyes.
                            He swallows deeply.
                            "[color=brown]It's not the kind of cold that can be cured.[/color]"
                        [/p]
                        [p]
                            The old man pokes a feathery finger underneath his glasses to wipe his eyes, in order to pretend he wasn't crying.
                            He turns towards you once more, a shaky, obviously faked grin on his face.
                            You're still stuck trying to process what he just said.
                            Is Seki going to...die?
                            The owl notices your confusion.
                            He tries to come up with a way to break the grim mood that has fallen upon the conversation, but you can see him visibly struggling for words.
                        [/p]
                        [p]
                            The bird utters a loud, depressing sigh.
                            "[color=brown]Anyway, off you go.[/color]"
                            He speaks, dismissing you with a wave of his hand, before turning around and beginning to walk away at a slow pace.
                            "[color=brown]I'll be in the city for the next few days, but be sure to stop by here sometime before next week, you hear?
                            I want to see my little [b]${ctrl.gender() === GenderEnum.MALE ? "man" : "woman"}[/b] one last time before...[/color]"
                            Stopping mid-sentence, Seki halts his slow walking for a moment.
                            He sadly snorts his nose, before silently disappearing into the darkness,
                            either to tend to his wounds, or to lament in solitude.
                        [/p]
                        [p]
                            Heartbroken, you hang your head and barely refrain from crying as you walk out of the old owl's shop.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.buyContinue.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.buyContinue.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "sekiBuyContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.buy.continued()) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.buyContinue.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.buy.continued()) {
            return;
        }

        next.registerListeners();

        if (ctrl.buyContinue.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\SekiWorkshop\shop\SekiWorkshopBuyContinue.js

// scenes\HamletDay2\SekiWorkshop\entrance\SekiWorkshopEntrance.js

/**
 * Entrance section of the Workshop scene.
 * @param {SekiWorkshopController} controller The Seki controller.
 */
function SekiWorkshopEntrance(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "entranceContinueButton";

    var toFree = new SekiWorkshopFree(ctrl);
    var toBuy = new SekiWorkshopBuy(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[st]Seki's Workshop[/st]
                        [p]
                            You enter the workshop of Seki, the owl.
                            This kind old man has helped you through some tough times in the past.
                            In some ways, you consider him the father you never had.
                            On the other hand, he's more of a [i]fence[/i] that you sell stuff you've stolen to.
                            In that regard, you suppose he's somewhat of a [i]bad[/i] influence.
                        [/p]
                        [p]
                            The workshop itself is [i]dark[/i], barely illuminated.
                            Being an owl, Seki prefers pitch-blackness above all,
                            keeping just a few candles around to prevent his customers from knocking over any of the potions he's brewed.
                            As a kobold, darkness doesn't phase you.
                            Your nightvision is superb, a fact which has significantly aided your criminal career thus far.
                            You deftly manoeuvre around the various tables and displays, towards the wooden counter,
                            where you ring a small bell to summon your elderly comrade from the depths of his shop.
                        [/p]
                        [p]
                            The old man is quite the accomplished alchemist.
                            There isn't a potion he [i]doesn't[/i] know how to brew.
                            Rumor has it that the owl used to be an advisor to the single most influential person in the entire world, the Dragon Emperor himself.
                            You've never quite asked him about it, and he hasn't told you, so you figure it's one of those things that are best left unmentioned.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the alternate content of this section.
     * @returns {string} The content.
     */
    function buildAltContent() {
        var name = ctrl.name();
        var toReturn = `[st]Seki's Workshop[/st]
                        [p]
                            As soon as the door to the barely illuminated workshop closes behind you,
                            a comforting feeling of intense [b]safety[/b] washes over you.
                            It touches you, to know that [b]someone[/b] actually [b]cares[/b] about you.
                            Your eyes water up.
                            You grab Seki's feathery leg, cuddle up against it and burst out into tears,
                            letting go all of the emotions you've been bottling up since the start of your capture.
                        [/p]
                        [p]
                            The big bird struggles to take a knee,
                            but once he manages to lower himself to your eye-level,
                            he wraps his winged arms around you, embracing you in a tender hug,
                            letting you rest your weeping head on his age-old shoulder while you bawl your eyes out.
                            "[color=brown]Shhhh...there, there.[/color]" he consolingly whispers to you, gently rubbing your bare back to calm you down.
                            "[color=brown]The bad men can't hurt you anymore.
                            You're safe here, with me.
                            Everything's going to be okay.[/color]"
                        [/p]
                        [p]
                            The avian lets you cry for as long as you need to.
                            For several minutes, you find yourself relaxing in Seki's clutches, your bountiful tears slowly coming to a halt.
                            Once again, the elderly owl proves that he's the closest thing to a father you'll ever have.
                        [/p]
                        [p]
                            As you begin to recover from the traumatic night and a half you've been through,
                            Seki reaches into the pouch hanging off his neck.
                            After fiddling around for a bit, he produces a recurring present, something he's given you [b]dozens[/b] of times before.
                            A single, shiny ${global.kobold.inventory.createItemWithTooltip(new Lockpick())}.
                            "[color=brown]Look here, what I have for you.
                            Remember this?[/color]"
                            Lifting your head from his shoulder, you give the bird a meek nod.
                            Breaking his caring embrace, he places the trinket into your hand, using his feathery talon to wrap your fingers around it tightly.
                            "[color=brown]It's all yours, my precious little ${name}.
                            Just like when you were young.[/color]"
                        [/p]
                        [p]
                            A small smile forms on your teary-eyed face.
                            Seki always knows exactly how to cheer you up.
                            With great difficulty, and a little help from you, he rises to his feet once more.
                            He grabs an empty coin pouch off one of his shelves, and hands it to you.
                            You gratefully wear it around your neck, as he wears his.
                        [/p]
                        [p]
                            "[color=brown]I'm sorry for yelling at you earlier.
                            Seeing you in the hands of [b]slaver scum[/b] ruffled my old feathers quite a bit.
                            But it's all over now.[/color]"
                            You ask about the money he used to [i]buy[/i] you.
                            Wasn't that a lot?
                            "[color=brown]Don't worry, little lizard.
                            I...won't be needing it anymore.[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.entrance.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.entrance.continue();
            if (ctrl.gold() > ctrl.minBuyGold) {
                loadNextPart(toBuy);
            } else {
                loadNextPart(toFree);
            }
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "sekiEntrance";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.arrival.continued()) {
            return "";
        }

        ctrl.bgm.loop();
        global.ui.applyTheme("hamlet_shop");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (global.kobold.history.shared.enslavement.noonTwo.boughtBySeki) {
            part += buildAltContent();
            part += ctrl.entrance.altChange();
        } else {
            part += buildContent();
        }
        part += buildChoice();
        part += `</div>`;
        part += toFree.build();
        part += toBuy.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.arrival.continued()) {
            return;
        }

        toFree.registerListeners();
        toBuy.registerListeners();

        if (ctrl.entrance.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\SekiWorkshop\entrance\SekiWorkshopEntrance.js

// scenes\HamletDay2\SekiWorkshop\arrival\SekiWorkshopArrival.js

/**
 * Arrival section of the Workshop scene.
 * @param {SekiWorkshopController} controller The Seki controller.
 */
function SekiWorkshopArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "arrivalContinueButton";

    var next = new SekiWorkshopEntrance(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You follow the eastern road towards Varanar,
                            making sure to keep your eyes peeled for any would-be assailants.
                            Luckily, your stroll is quite uneventful.
                            Bandits usually operate closer to nighttime.
                            Most of them must still be asleep.
                            It doesn't take you long to reach the small,
                            branching dirt path that'll lead you directly towards the hamlet.
                        [/p]
                        [p]
                            Approaching the small village, you cross paths with a couple of farmers.
                            You figure they're either making their way to their respective fields, or heading to Varanar for whatever reason.
                            They don't seem to be carrying any goods for sale, not that it's any of your business.
                            The humans are too busy speaking to one another to pay you much mind.
                            Shrugging, you continue along your journey.
                        [/p]
                        [p]
                            Around noon, you arrive at the hamlet.
                            The sun is high up in the sky, although the temperature remains somewhat frigid.
                            The settlement you have entered considers of nothing but a few meagre houses, a [i]breeding farm[/i], 
                            and a single shop, ran by an old friend of yours.
                            That last place is where you decide to head towards first.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.arrival.continued();
        var choice = `<div class="choicecontainer">
                        <button class="button inlinechoice ${disabled ? "selected" : ""}" id="${continueId}" ${disabled ? "disabled" : ""}>
                            Continue
                        </button>
                      </div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.arrival.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "sekiArrival";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (global.kobold.history.shared.enslavement.noonTwo.boughtBySeki) {
            ctrl.arrival.continue();
            return next.build();
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.arrival.continued() || global.kobold.history.shared.enslavement.noonTwo.boughtBySeki) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\SekiWorkshop\arrival\SekiWorkshopArrival.js

// scenes\HamletDay2\SekiWorkshop\SekiWorkshopController.js

/**
 * Controller for the Home scene.
 * @param {string} sceneId The id of the scene.
 */
function SekiWorkshopController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;
    var genderId = "gender";
    var autoSaveId = "autoSave";
    var goldId = "gold";
    var continueId = "continue";
    var underwearId = "underwear";
    var pantsId = "pants";
    var braId = "bra";
    var shirtId = "shirt";
    var nameId = "name";
    var healthId = "health";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}
    function EntranceController(key) {
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
            global.kobold.history.shared.seki.knowsDying = true;
            global.kobold.history.shared.seki.hamletHadMoney = global.kobold.inventory.coins !== 0;
        };
        this.continued = function () {
            return global.kobold.history.has(id, key);
        };
        var altChangeKey = key + "change";
        this.altChange = function () {
            global.kobold.history.doIfNotSet(id, altChangeKey, true, function () {
                global.kobold.status.mental.increase(15, "Hugging Seki");
                global.kobold.inventory.bag.addMisc(new Lockpick());
            });
            return global.ui.change.buildNotification([
                new Change("Cheered up", "Sanity", +15),
                new Change("Free lockpick", global.kobold.inventory.createItemWithTooltip(new Lockpick(), true), +1)
            ]);
        };
    }

    function FreeController(key) {
        ContinueController.call(this, id, key);
        var changeKey = key + "change";
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeKey, true, function () {
                global.kobold.status.mental.decrease(20, "Bad news has wings");
                global.kobold.inventory.bag.addMisc(new Lockpick());
            });
            return global.ui.change.buildNotification([
                new Change("Bad news", "Sanity", -20),
                new Change("Free lockpick", global.kobold.inventory.createItemWithTooltip(new Lockpick(), true), +1)
            ]);
        };
        var altChangeKey = key + "change";
        this.altChange = function () {
            global.kobold.history.doIfNotSet(id, altChangeKey, true, function () {
                global.kobold.status.mental.decrease(20, "Bad news has wings");
            });
            return global.ui.change.buildNotification([
                new Change("Bad news", "Sanity", -20)
            ]);
        };
    }

    function BuyController(key) {
        var that = this;
        ContinueController.call(this, id, key);

        var shopItemKey = key + "Items";
        var shopGoldKey = key + "Gold";
        var initialShopItems = [
            new ItemAmount(1, new Dagger()),
            new ItemAmount(1, new LinenBra()),
            new ItemAmount(1, new KoboldUnderwear()),
            new ItemAmount(1, new LeatherPants()),
            new ItemAmount(1, new PaddedLeatherShirt()),
            new ItemAmount(1, new Rope()),
            new ItemAmount(1, new SmallMirror()),
            new ItemAmount(4, new Lockpick()),
            new ItemAmount(1, new AgilityPotion()),
            new ItemAmount(1, new CharismaPotion()),
            new ItemAmount(1, new EndurancePotion()),
            new ItemAmount(1, new IntelligencePotion()),
            new ItemAmount(1, new LuckPotion()),
            new ItemAmount(1, new StrengthPotion()),
            new ItemAmount(1, new RejuvenationPotion()),
            new ItemAmount(2, new mPlus()),
            new ItemAmount(2, new Apple()),
            new ItemAmount(1, new HeartyMeal())
        ];
        var sellmod = 0.7;
        var buymod = 1.0;

        function sellListener(item) {
            var itemAmounts = that.shopItems();
            var added = false;
            for (var i = 0; i < itemAmounts.length && !added; i++) {
                if (itemAmounts[i].item.id === item.id) {
                    itemAmounts[i].amount++;
                    added = true;
                }
            }
            if (!added) {
                itemAmounts.push(new ItemAmount(1, item));
            }
            that.updateCoins();
            that.updateShopItems(itemAmounts);
        }

        function buyListener(location) {
            var oldCoins = that.coins();
            that.updateCoins();
            var newCoins = that.coins();
            var itemAmounts = that.shopItems();
            itemAmounts[location].amount--;
            that.updateShopItems(itemAmounts);
            global.kobold.history.shared.seki.moneySpent += oldCoins - newCoins;
        }

        this.getGenerator = function () {
            return new MerchantInventoryGenerator("SekiShop", that.shopItems(), that.coins(), sellmod, buymod, that.continued(), sellListener, buyListener);
        };
        this.coins = function () {
            global.kobold.history.setIfNotSet(id, shopGoldKey, global.kobold.inventory.coins);
            return global.kobold.history.getKey(id, shopGoldKey);
        };
        this.updateCoins = function () {
            global.kobold.history.setKey(id, shopGoldKey, global.kobold.inventory.coins);
        };
        this.shopItems = function () {
            global.kobold.history.setIfNotSet(id, shopItemKey, initialShopItems);
            var jsonItems = global.kobold.history.getKey(id, shopItemKey);
            var items = global.kobold.inventory.constructItemAmountListFromMemory(jsonItems);
            return items;
        };
        this.updateShopItems = function (newItems) {
            global.kobold.history.setKey(id, shopItemKey, newItems);
        };
    }

    function BuyContinueController(id, key) {
        this.continue = function () {
            global.kobold.history.setKey(id, key, true);
        };
        this.continued = function () {
            // Version 2.11 (Pitchfork) introduced a bug where the key used for the BuyContinueController was the id of the scene.
            // This fix ensures saves will not retroactively break.
            return global.kobold.history.has(id, key) || global.kobold.history.has(id, id);
        };
        var changeKey = key + "change";
        this.change = function () {
            global.kobold.history.doIfNotSet(id, changeKey, true, function () {
                global.kobold.status.mental.decrease(20, "Bad news has wings");
            });
            return global.ui.change.buildNotification([
                new Change("Bad news", "Sanity", -20)
            ]);
        };
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.folderName = "HamletDay2";
    this.breedingSceneName = "BreedingFarm";
    this.farmHomeSceneName = "FarmerHouse";
    this.wellSceneName = "HamletWell";
    this.minBuyGold = 50;
    this.arrival = new ContinueController(id, "arrival");
    this.entrance = new EntranceController("entrance");
    this.free = new FreeController("free");
    this.buy = new BuyController("buy");
    this.buyContinue = new BuyContinueController(id, "buyContinue");

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /*
     * Autosaves the game if necessary.
     */
    this.autosave = function () {
        global.kobold.history.doIfNotSet(id, autoSaveId, true, function () {
            global.saveManager.autoSave();
        });
    };

    /**
     * Signals the scene has been continued.
     */
    this.continue = function () {
        global.kobold.history.setKey(id, continueId, true);
    };

    /**
     * Returns whether or not the scene has continued.
     * @returns {boolean} True if the scene has continued, false otherwise.
     */
    this.continued = function () {
        return global.kobold.history.has(id, continueId);
    };

    this.gold = function () {
        global.kobold.history.setIfNotSet(id, goldId, global.kobold.inventory.coins);
        return global.kobold.history.getKey(id, goldId);
    };

    this.pants = function () {
        global.kobold.history.setIfNotSet(id, pantsId, global.kobold.inventory.equipment.pants);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, pantsId));
    };

    this.bra = function () {
        global.kobold.history.setIfNotSet(id, braId, global.kobold.inventory.equipment.bra);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, braId));
    };

    this.shirt = function () {
        global.kobold.history.setIfNotSet(id, shirtId, global.kobold.inventory.equipment.shirt);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, shirtId));
    };

    this.underwear = function () {
        global.kobold.history.setIfNotSet(id, underwearId, global.kobold.inventory.equipment.underwear);
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.getKey(id, underwearId));
    };

    this.gender = function () {
        global.kobold.history.setIfNotSet(id, genderId, global.kobold.body.getBaseGender());
        return global.kobold.history.getKey(id, genderId);
    };

    this.name = function () {
        global.kobold.history.setIfNotSet(id, nameId, global.kobold.name);
        return global.kobold.history.getKey(id, nameId);
    };

    this.health = function () {
        global.kobold.history.setIfNotSet(id, healthId, global.kobold.status.physical.get());
        return global.kobold.history.getKey(id, healthId);
    };

    this.bgm = new AudioController(id, "res/audio/seki/sadness_cosmo.mp3", 100);
}

// END scenes\HamletDay2\SekiWorkshop\SekiWorkshopController.js

// scenes\HamletDay2\SekiWorkshop\SekiWorkshop.js

/**
 * Seki's Workshop. Starting scene of the Hamlet.
 */
function SekiWorkshop() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "SekiWorkshop";
    var sceneName = "Seki's Workshop";
    var sceneSubfolder = "HamletDay2";
    var ctrl = new SekiWorkshopController(sceneId);
    var arrival = new SekiWorkshopArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]hamlet[/t]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new SekiWorkshop();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\HamletDay2\SekiWorkshop\SekiWorkshop.js



// END SCENE SekiWorkshop



// SCENE BreedingFarm

// scenes\HamletDay2\BreedingFarm\barn\attack\free\NotBarnFree.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnFree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "NotBarnFreeContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.cassidy.koboldsfreed = true;
        var dead = global.kobold.history.shared.cassidy.dead;
        var toReturn = `[p]
                            With the vixen ${dead ? "" : "temporarily"} out of the equation, you decide to save your fellow kobolds from a lifetime of mindless breeding.
                            You run back towards the barn, and enter it once more.
                            You walk up to the nearest enclosure, and deftly lift the heavy latch holding the large wooden door shut.
                            Swinging open the animal pen, you are greeted with a terrible sight.
                        [/p]
                        [p]
                            A group of seven kobolds inhabits this particular [i]cell[/i], some of which are chained to the wooden walls.
                            All of them are naked.
                            In one corner lies a female, with tits so bloated you doubt she can move on her own.
                            A single tiny whelp, who can't be more than a couple of months old, is suckling on one of her lactating teats.
                        [/p]
                        [p]
                            Next to the nursing mother, an unchained male is idly humping a restrained, gagged and blindfolded kobold.
                            You can't quite discern the gender of the one on the receiving end,
                            but regardless of whether the hole being fucked is a cockslit or a cunt,
                            it's completely [b]overflowing[/b] with spunk.
                            Judging by the veritable [b]pool[/b] of jizz that has accumulated on the soggy ground below them,
                            you reckon those two must have been at it for [b]days[/b] on end.
                        [/p]
                        [p]
                            Further to the right, another [i]stallion[/i] is standing up, leaning against the wall, fervently stroking his oversized erection.
                            His very [b]existence[/b] seems to taunt a sixth of your kin, one of the females.
                            This one is a bit more [i]normally-proportioned[/i], apart from her swollen belly.
                            You presume she's pregnant, as opposed to just being [i]fat[/i].
                            Her eyes are glued to the masturbating male's groin, and the reason why is very clear.
                            Her wrists and her tail are hogtied, preventing her from satisfying the [i]painfully obvious[/i] need between her legs.
                            She squirms her lower body about, desperately trying but failing to pleasure herself.
                            She must have gotten injected with some kind of aphrodisiac, and a strong one at that.
                        [/p]
                        [p]
                            To your left, you spot what you believe to be the most recent [i]acquisition[/i] of the bunch.
                            A young woman, who you judge to be about your age, sits chained up to the wall, mindlessly staring off to the side.
                            Her loins have clearly been [b]violated[/b], but apart from a few cum-spatters staining her scales, she appears to be in rather good shape.
                            Slowly, the kobold shifts her gaze towards you, which is the most [i]recognition[/i] you've gotten from any of the captives thus far.
                        [/p]
                        [p]
                            At first, she looks uninterested, probably thinking you're yet [b]another[/b] poor sap that got enticed by Cassidy.
                            Then, you notice a [i]spark[/i], a certain [i]twinkle[/i] of consciousness, of sheer self-awareness in her otherwise jaded eyes.
                            Something the other kobolds in the pen lack.
                        [/p]
                        [p]
                            The realization that the vixen isn't around seems to inspire hope in the otherwise completely demoralized girl.
                            She tries to speak, to address you, to [b]beg[/b] you for help, yet her mouth is gagged by an old, filthy rag,
                            causing her desperate pleas to come out as pitiful whimpers and groaning moans.
                        [/p]
                        [p]
                            You quickly approach the distressed captive.
                            Kneeling beside her, you pull the disgusting piece of clothing serving as a silencing muzzle from her mouth.
                            She immediately begins to [b]beg[/b] you [i]not[/i] to fuck her.
                            Poor thing must have been through a lot.
                            You reassure her that you're not there to hurt her.
                            You just want to help.
                        [/p]
                        [p]
                            To showcase your good intentions, you unfasten the cuffs tying her to the wall.
                            Same mechanism as the table.
                            Easy to remove from the outside.
                            The moment her arms are freed from their restraints, she curls up into a fetal position on the ground,
                            and slowly begins to rock back and forth, softly muttering to herself before openly bursting out into tears.
                        [/p]
                        [p]
                            You decide to give the kobold a few minutes to herself.
                            You step out of the enclosure, to ponder the entire ordeal.
                            Had you let Cassidy have her way, you could have very well ended up in one of these pens yourself,
                            drugged up to the point where you can no longer [b]think[/b] of anything but getting fucked,
                            pumping out litter after litter of younglings to be sold into slavery, or perhaps even [b]butchered[/b] for their meat.
                        [/p]
                        [p]
                            As the heart-wrenching sobbing coming from the open pen fades to nothing, you head back inside.
                            Once more, you take a knee next to the pink-scaled captive.
                            You place your hand atop her head, and comfortingly pat it, ever so gently.
                            Her crying has stopped, although even the girl's heavy, unsteady breathing is permeated with pure [b]sadness[/b].
                        [/p]
                        [p]
                            You cautiously coax the traumatized kobold into sitting up straight.
                            You don't dare to let her to rest her back against the wall, lest it trigger some kind of flashback of being held captive.
                            After she's calmed down a bit more, you ask the woman for her name.
                            No response.
                            She continues to stare blankly ahead of her.
                            You ask if she'd like to leave this place.
                            A single nod, weak but solemn.
                        [/p]
                        [p]
                            You say there's a nice house just next door that she can go to.
                            But first, she needs to help you break out the rest of the kobolds, so they can all get out of this hell-hole together.
                            She tilts her visage your way.
                            She visibly tries to respond to your request, yet no words leave her mouth.
                            Instead, another slight, affirmative shaking of her head.
                        [/p]
                        [p]
                            You help the girl to her feet.
                            At first, she can barely stand, let alone walk, yet she quickly regains her leg-strength.
                            It takes just a few minutes of you supporting her before she can stand on her own once more.
                            Once you lead her outside of the containment pen, the sight of the downed vixen coming into view,
                            her spirits seem to be lifted further.
                            The two of you share a mutual nod, and you jointly set to work opening the remainder of the enclosures,
                            liberating the stables from vulpine oppression.
                        [/p]
                        [p]
                            In each pen, the situation is roughly the same.
                            There's a bunch of drugged up males, some thoroughly violated,
                            often pregnant, mostly mind-broken kobold women, a couple of which used to be [i]men[/i],
                            and finally there's one or two more relatively unscathed, yet completely immobilized [i]newcomers[/i], who retain much of their [i]sanity[/i].
                            You focus on this final group, setting them free like you did your new,
                            pink-scaled friend, and getting them to help in the ongoing effort to [i]get through[/i] to the [i]less fortunate[/i] members of the group.
                        [/p]
                        [p]
                            Eventually, you manage to get a somewhat functional form of [i]coordination[/i] going on.
                            The more mentally capable kobolds herd the drug-addled and physically debilitated ones out of the pens.
                            Most of them can't stop touching themselves, so [b]afflicted[/b] by various potions and drugs that they'll likely take days,
                            if not weeks to physically recover from this entire endeavour, let alone mentally.
                        [/p]
                        [p]
                            The entire group gets corralled through the barn's front door, into the adjacent house,
                            and hopefully towards a better future than what they'd have had in here.
                            ${dead ? "The vixen's corpse gets dragged indoors, probably to dispose of later. Or for other [i]purposes[/i]." : "The vixen, who [b]still[/b] has not regained her consciousness, gets dragged indoors."}
                            One of the last few stragglers to remain, diligently ensuring that each and every one of the once-prisoners finds his or her way inside,
                            is the pink-scaled girl you rescued first.
                        [/p]
                        [p]
                            Ultimately, you and her end up being the only two left on the porch of the house.
                            No words are spoken, but you can [i]sense[/i] her gratitude,
                            in the same way that she realizes that you can't help but make this [b]a lot[/b] more awkward than it should have been.
                            A small, heart-warming smile forms on her face.
                            With a loud, relieved sigh and a deepfully grateful nod, she silently turns around, and walks away to join the others inside.
                        [/p]
                        [p]
                            You wish you could stay, but alas, your presence here would only cause trouble for the kobolds.
                            Besides, you have business to attend to in Varanar, and you'd only get in the way of their healing process.
                            You're certain they'll be fine.
                            They're in at least one set of good hands.
                            With the knowledge that you did the right thing here today,
                            you quietly close the front door,
                            leaving the breeding farm not as the thief you arrived,
                            but as an unsung hero of your kind.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnAttackFree";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.postnotbarn.get() !== NotBarnAttackChoice.FREE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        //part += ctrl.unawareFree.change();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.postnotbarn.get() !== NotBarnAttackChoice.FREE) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\attack\free\NotBarnFree.js

// scenes\HamletDay2\BreedingFarm\barn\attack\loot\downstairs\NotBarnLootDownstairs.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnLootDownstairs(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnLootDownstairsCButton";
    
    var next = new NotBarnLootDownstairsContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dead = global.kobold.history.shared.cassidy.dead;
        var toReturn = `[p]
                            You don't have time to check out the second floor.
                            You need to grab whatever you can and get the hell out of there.
                            You bolt around the house, looking for some [i]obvious[/i] things to steal.
                            Sadly, many drawers are locked, with no sign of any keys laying around.
                            The ones that [i]are[/i] unlocked are filled with worthless stuff.
                            Letters, utensils, a few cooking ingredients, nothing worth taking.
                        [/p]
                        [p]
                            On the verge of giving up hope on finding anything of value whatsoever, your avarice resurges as you enter the dining room.
                            The large rectangular table at the center of the room is set for dinner, a veritable [b]hoard[/b] of silver tableware ripe for the taking.
                            You begin collecting forks, spoons and breadknives, but before you manage to empty the entirety of the wooden surface, 
                            you hear a noise at the front door.
                            A voice calling out to Cassidy, trying to get her to wake up.
                        [/p]
                        [p]
                            Deciding not to take any chances, you cut your gathering short.
                            You sprint towards the back door, which is luckily unlocked,
                            and you silently slip outside, escaping the scene of the crime through the back yard.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildContinueChoice() {
        var disabled = ctrl.postnotbarnloot.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarnloot.has();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildContinueChoice();
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarnloot.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnLootDownstairs";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.notbarnloot.get() !== NotBarnLootChoice.DOWNSTAIRS) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.notbarnDownstairsLootChange();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.notbarnloot.get() !== NotBarnLootChoice.DOWNSTAIRS) {
            return;
        }

        next.registerListeners();

        if (ctrl.postnotbarnloot.has()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\attack\loot\downstairs\NotBarnLootDownstairs.js

// scenes\HamletDay2\BreedingFarm\barn\attack\loot\downstairs\NotBarnLootDownstairsContinue.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnLootDownstairsContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnLootDownstairsContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dead = global.kobold.history.shared.cassidy.dead;
        var toReturn = `[p]
                            After about half an hour of running through fields, farms and small patches of woodland,
                            making sure to stay out of sight, you think you're finally in the clear.
                            By traversing less-well travelled paths, you've ensured nobody will be able to link you to 
                            ${dead ? "the murder" : "the robbing"} of Cassidy${dead ? "" : ", apart from the vixen herself"}.
                            ${dead ? "" : "You're certain she won't tell anyone, out of fear that you'll spread the word of her deviant sexual tastes, thus ruining her reputation."}
                            You have a feeling the city guards'll be pestering you enough [i]without[/i] allegations of ${dead ? "murder" : "robbery"}.
                        [/p]
                        [p]
                            You find yourself on the eastern road to Varanar once again.
                            After taking a moment to catch your breath, you begin to walk down the path towards the city.
                            At least you won't be meeting any more crazy, kobold-breeding bitches there.
                            Or so you hope.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnLootDownstairsContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.postnotbarnloot.continued()) {
            return "";
        }
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.notbarnLootContinueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.postnotbarnloot.continued()) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\attack\loot\downstairs\NotBarnLootDownstairsContinue.js

// scenes\HamletDay2\BreedingFarm\barn\attack\loot\upstairs\NotBarnLootUpstairs.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnLootUpstairs(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnLootUpstairsCButton";
    
    var next = new NotBarnLootUpstairsContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dead = global.kobold.history.shared.cassidy.dead;
        var toReturn = `[p]
                            You sprint towards the stairs, and ascend them with haste.
                            The second floor hallway contains even more pictures,
                            these ones far more risqu&eacute; than the ones downstairs.
                            Naked kobolds, fornicating foxes, even some of your kind mating with vulpines.
                            While the paintings are distracting and disturbing, to say the least,
                            you force yourself to face away from them, and set your mind to the task at hand.
                        [/p]
                        [p]
                            All of the doors are closed.
                            You'll have to search each room at a time.
                            The first one you enter is a small, dusty storageroom, with nothing of value in it for you to pocket.
                            The second seems to pass as some sort of bathroom, with a fancy mirror on the wall a large wooden tub at the center.
                            Again: nothing to steal.
                        [/p]
                        [p]
                            You enter a third room.
                            Bullseye.
                            The vixen's bedroom.
                            Surely there'll be some goodies in here, right?
                            You search through both her nightstands, finding nothing but various pairs of infrequently worn panties, none of which are your size.
                            On top of the underwear, you find a pair of fake, fur-lined handcuffs and a small, smooth, dog-phallic-shaped piece of wood,
                            the use of which you can only [i]imagine[/i].
                        [/p]
                        [p]
                            Just as you were about to give up hope on finding anything of value whatsoever,
                            you accidentally trip over a pair of panties, landing face-first on the hard, wooden floor.
                            Recovering from the fall, your gaze wanders to the underside of the vulpine's bed,
                            where you spot a tiny coinpouch, seemingly filled to the very brim.
                            You rub your head to ease your pain, and pinch yourself to make sure you're not hallucinating.
                            Nope.
                            It's real!
                        [/p]
                        [p]
                            Just as you reach out and grab the bag, you hear a noise at the front door.
                            A voice calling out to Cassidy, trying to get her to wake up.
                            Deciding not to take any chances, you swipe the pouch and hop to your feet.
                            You consider bolting for the back door, but with an unwelcome [i]guest[/i] around,
                            the chance of being seen is too great.
                            Instead, you open one of the bedroom windows, and sneakily climb down a handily-placed vine plant encroaching upon the building.
                            Silently dropping down to the back yard, you escape the scene of the crime unnoticed.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildContinueChoice() {
        var disabled = ctrl.postnotbarnloot.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarnloot.has();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildContinueChoice();
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarnloot.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnLootUpstairs";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.notbarnloot.get() !== NotBarnLootChoice.UPSTAIRS) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.notbarnUpstairsLootChange();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.notbarnloot.get() !== NotBarnLootChoice.UPSTAIRS) {
            return;
        }

        next.registerListeners();

        if (ctrl.postnotbarnloot.has()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\attack\loot\upstairs\NotBarnLootUpstairs.js

// scenes\HamletDay2\BreedingFarm\barn\attack\loot\upstairs\NotBarnLootUpstairsContinue.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnLootUpstairsContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnLootUpstairsContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dead = global.kobold.history.shared.cassidy.dead;
        var toReturn = `[p]
                            After about half an hour of running through fields, farms and small patches of woodland,
                            making sure to stay out of sight, you think you're finally in the clear.
                            By traversing less-well travelled paths, you've ensured nobody will be able to link you to 
                            ${dead ? "the murder" : "the robbing"} of Cassidy${dead ? "" : ", apart from the vixen herself"}.
                            ${dead ? "" : "You're certain she won't tell anyone, out of fear that you'll spread the word of her deviant sexual tastes, thus ruining her reputation."}
                            You have a feeling the city guards'll be pestering you enough [i]without[/i] allegations of ${dead ? "murder" : "robbery"}.
                        [/p]
                        [p]
                            You find yourself on the eastern road to Varanar once again.
                            You move the gold you stole from the small pouch it is in to your own moneybag.
                            After taking a moment to catch your breath, you begin to walk down the path towards the city.
                            At least you won't be meeting any more crazy, kobold-breeding bitches there.
                            Or so you hope.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnLootUpstairsContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.postnotbarnloot.continued()) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.notbarnLootContinueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.postnotbarnloot.continued()) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\attack\loot\upstairs\NotBarnLootUpstairsContinue.js

// scenes\HamletDay2\BreedingFarm\barn\attack\loot\NotBarnLoot.js

/**
 * Not barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnLoot(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var downstairsId = "notBarnLootDownstairsButton";
    var upstairsId = "notBarnLootUpstairsButton";

    var toDownstairs = new NotBarnLootDownstairs(ctrl);
    var toUpstairs = new NotBarnLootUpstairs(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dead = global.kobold.history.shared.cassidy.dead;
        var toReturn = `[p]
                            With the vixen downed, you bolt into the mansion through the open door.
                            It looks as fancy on the inside as it does on the outside.
                            The entire place is well-lit by candles in gilded candelabras.
                            Various paintings of kobolds and canines alike line the walls.
                            A myriad of luxurious cabinets and drawers clutter the broad hallway you find yourself in,
                            undoubtedly filled to the brim with [b]riches[/b].
                            Or so you imagine.
                        [/p]
                        [p]
                            Sadly, you don't have all the time in the world to strip this place of everything of value.
                            You doubt it'll be long before someone discovers Cassidy's ${dead? "corpse" : "passed out body"}.
                            The hamlet, while small, is a tight-knit community.
                            And the last thing you want to do right now is get chased down by farmers.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildDownstairsChoice() {
        var disabled = ctrl.notbarnloot.has() ? "disabled" : "";
        var isSelected = ctrl.notbarnloot.get() === NotBarnLootChoice.DOWNSTAIRS;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${downstairsId}" ${disabled}>Downstairs</button>`;
        var tooltip = global.parser.parse(`The ground floor's where the good stuff's at.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildUpstairsChoice() {
        var disabled = ctrl.notbarnloot.has() ? "disabled" : "";
        var isSelected = ctrl.notbarnloot.get() === NotBarnLootChoice.UPSTAIRS;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${upstairsId}" ${disabled}>Upstairs</button>`;
        var tooltip = global.parser.parse(`Go up to the second floor instead. There's bound to be some interesting finds there!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You loot: </div>`;
        choice += buildDownstairsChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildUpstairsChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + downstairsId), $("#" + upstairsId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerDownstairsListener() {
        $("#" + downstairsId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.notbarnloot.set(NotBarnLootChoice.DOWNSTAIRS);
            loadNextPart(toDownstairs);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerUpstairsListener() {
        $("#" + upstairsId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.notbarnloot.set(NotBarnLootChoice.UPSTAIRS);
            loadNextPart(toUpstairs);
        });
    }


    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnLoot";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.postnotbarn.get() !== NotBarnAttackChoice.LOOT) {
            return "";
        }
        global.ui.applyTheme("hamlet_mansion");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toDownstairs.build();
        part += toUpstairs.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.postnotbarn.get() !== NotBarnAttackChoice.LOOT) {
            return;
        }

        toDownstairs.registerListeners();
        toUpstairs.registerListeners();

        if (ctrl.notbarnloot.has()) {
            return;
        }

        registerDownstairsListener();
        registerUpstairsListener();
    };
}

var NotBarnLootChoice = {
    DOWNSTAIRS: 1,
    UPSTAIRS: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\attack\loot\NotBarnLoot.js

// scenes\HamletDay2\BreedingFarm\barn\attack\NotBarnAttack.js

/**
 * Not barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnAttack(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var lootId = "notBarnAttackLootButton";
    var freeId = "notBarnAttackFreeButton";

    var toLoot = new NotBarnLoot(ctrl);
    var toFree = new NotBarnFree(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon(); // TODO
        var toReturn = `[p]
                            No.
                            [b]No![/b]
                            What Cassidy is doing is [b]wrong[/b], no matter how you look at it.
                            She's [b]evil[/b].
                            Anger wells up inside of you.
                            You wield your ${global.kobold.inventory.createItemWithTooltip(weapon)} and approach the vulpine from behind.
                            She's too busy fumbling with the door to notice you drawing close.
                            You're not going to let this horrible [b]bitch[/b] get away with [b]enslaving[/b] your kind.
                        [/p]`;
        
        if(weapon.stab()) {
            global.kobold.history.shared.cassidy.dead = true;
            toReturn += `[p]
                            In a single smooth motion, you grab the vixen's waist with one arm, whilst driving your ${global.kobold.inventory.createItemWithTooltip(weapon)}
                            into her side with the other.
                            Like embracing her in a hug from behind, one almost as sick and twisted as she herself is.
                            The blade of your weapon slides in between her ribs, sinking into her delicate flesh all the way to the hilt.
                            Your rage surging, you stab her, again and again and again.
                        [/p]
                        [p]
                            She doesn't scream.
                            She doesn't struggle.
                            Instead, she solemnly closes her eyes, and lowers her head.
                            It's like she knew this was coming.
                            Like she'd long made peace with the fact that she's a horrible person.
                            Like she accepts, or perhaps even [i]welcomes[/i] this outcome to her wicked life.
                            Or maybe her mind just can't process what is happening.
                        [/p]
                        [p]
                            As your series of stabs comes to an end, blood [b]gushing[/b] from her side,
                            Cassidy takes one last deep breath, before her still-standing figure lifelessly topples forwards.
                            The vulpine's naked frame tumbles against the door, her arm impacting the door handle, causing it to swing open.
                            Her body ends up halfway inside of her own mansion, legs and rump laying on the porch, 
                            a large pool of blood forming on the indoors floor, ending up leaking from the doorway.
                            [b]You've killed Cassidy.[/b]
                        [/p]
                        [p]
                            You take a few deep breaths, excitement rushing from your system.
                            You acted on impulse, there's no denying that.
                            But she said it herself.
                            She breeds kobolds to be [b]slaughtered[/b].
                            To be [b]eaten[/b].
                            You couldn't let her get away with that.
                            Kobolds don't have any rights.
                            The guards wouldn't have stopped her.
                            You did what you had to.
                            Or so you tell yourself, to dull your aching conscience.
                        [/p]
                        [p]
                            In the back of your mind, you realize you've crossed a terrible threshold,
                            but the adrenaline coursing through your veins makes you ignore the crushing guilt building up inside of you, at least for now.
                            You glance around.
                            You don't think anyone saw you, but it's only a matter of time before someone notices the open door and the corpse in the doorway.
                            Whatever you decide to do, you'll have to act fast.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You brace your weapon, holding it high above your head, preparing your entire body to strike down with all of your might.
                            You fail to contain a loud, angry growl, alerting the vixen to your presence directly behind her, yet before she even manages to react,
                            you bring your ${global.kobold.inventory.createItemWithTooltip(weapon)} down upon her head,
                            the hard surface impacting against the woman's skull with a twistingly satisfying [b]crack[/b].
                        [/p]
                        [p]
                            The vulpine's naked frame tumbles against the door, her arm impacting the door handle, causing it to swing open.
                            Her body ends up halfway inside of her own mansion, legs and rump laying on the porch, 
                            For a moment, you fear you might have [b]killed[/b] her, but then you notice the vixen's bare, 
                            prone torso slowly, yet rhythmically inflating and deflating.
                            Still breathing.
                            You didn't kill her, you've just managed to knock her unconscious in a single blow.
                        [/p]
                        [p]
                            Content with your work, you take a few deep breaths to calm yourself down.
                            That'll teach her.
                            You acted on impulse, there's no denying that.
                            But she said it herself.
                            She breeds kobolds to be [b]slaughtered[/b].
                            To be [b]eaten[/b].
                            You couldn't let her get away with that.
                        [/p]
                        [p]
                            You glance around.
                            You don't think anyone saw you, but it's only a matter of time before someone notices the open door and the passed out vixen in the doorway.
                            Or worse yet: she might get up herself.
                            Whatever you decide to do, you'll have to act fast.
                        [/p]`;
        }
        
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildLootChoice() {
        var disabled = ctrl.postnotbarn.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarn.get() === NotBarnAttackChoice.LOOT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${lootId}" ${disabled}>Steal</button>`;
        var tooltip = global.parser.parse(`Enter the mansion and take whatever you can.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;
        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFreeChoice() {
        var disabled = ctrl.postnotbarn.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarn.get() === NotBarnAttackChoice.FREE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${freeId}" ${disabled} style="width:200px;">Free the kobolds</button>`;
        var tooltip = global.parser.parse(`Go back to the barn and set the vixen's captives free.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice" style="width:200px;">`;
        button += popupContainer;
        button += `</div>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildLootChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFreeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + lootId), $("#" + freeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerLootListener() {
        $("#" + lootId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarn.set(NotBarnAttackChoice.LOOT);
            loadNextPart(toLoot);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFreeListener() {
        $("#" + freeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarn.set(NotBarnAttackChoice.FREE);
            loadNextPart(toFree);
        });
    }


    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnAttack";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.notbarn.get() !== NotBarnTeaDoorChoice.ATTACK && ctrl.notbarn.get() !== NotBarnChoice.ATTACK) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toLoot.build();
        part += toFree.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.notbarn.get() !== NotBarnTeaDoorChoice.ATTACK && ctrl.notbarn.get() !== NotBarnChoice.ATTACK) {
            return;
        }

        toLoot.registerListeners();
        toFree.registerListeners();

        if (ctrl.postnotbarn.has()) {
            return;
        }

        registerLootListener();
        registerFreeListener();
    };
}

var NotBarnAttackChoice = {
    LOOT: 1,
    FREE: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\attack\NotBarnAttack.js

// scenes\HamletDay2\BreedingFarm\barn\notbarn\agree\NotBarnAgreeContinue.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnAgreeContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnAgreeContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Your lust sated, you wander back to the center of the hamlet.
                            It's already late in the afternoon, so you don't have much time to linger if you want to make it to Varanar before nightfall.
                            You cast one last look at the vixen's farm-home.
                            That sure was an [b]experience[/b] you'll never forget.
                            With a soft sigh, you set off towards the human capital.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnAgreeContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.postnotbarnsex.continued()) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.postnotbarnsex.continued()) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\notbarn\agree\NotBarnAgreeContinue.js

// scenes\HamletDay2\BreedingFarm\barn\notbarn\agree\NotBarnAgreeSex.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnAgreeSex(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnAgreeSexButton";

    var next = new NotBarnAgreeContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function buildMaleContent() {
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var dick = ctrl.misc.dick();
        var name = ctrl.misc.name();
        var virgin = ctrl.misc.dickVirgin();
        var toReturn = `[p]
                            The canine takes a few steps forwards, towards a small, low, decorative table which stands against the leftmost wall.
                            While continuing to touch herself, she uses her free arm to brush the couple of unopened letters atop its surface to the floor. `;

        if (pants || underwear) {
            toReturn += `Meanwhile, you begin to undress yourself. `;
        } else {
            toReturn += `Cassidy's constant teasing and the sight of her pristine, naked body have thoroughly aroused you,
                         your member fully erect, throbbing and ready for action.`;
        }

        if (pants) {
            toReturn += `You take off your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (underwear) {
                toReturn += ` and your ${global.kobold.inventory.createItemWithTooltip(underwear)}`;
            }
            toReturn += `, your already fully erect member coming into view,
                        Cassidy's constant teasing along with the sight of her naked body having thoroughly aroused you.`;
        } else if (underwear) {
            toReturn += `You take off your ${global.kobold.inventory.createItemWithTooltip(underwear)},
                        your already fully erect member coming into view,
                        Cassidy's constant teasing along with the sight of her naked body having thoroughly aroused you.`;
        }

        toReturn += `   [/p]
                        [p]
                            The table cleared, the woman looks behind her, directly at you.
                            Giving you a small wink, she slowly sinks to her knees,
                            her gaze drifting down to your groin as she presses her upper body flat against the low, wooden surface,
                            squishing her formidable rack against it,
                            whilst sticking her rump out at [b]exactly[/b] the right height for you to be able to reach it properly,
                            her tail enticingly raised up high.
                            She's done this before, and it shows.
                        [/p]`;

        toReturn += `   [p]
                            After eyeing your length for a couple of seconds, the vixen speaks.`;

        if (dick < SizeEnum.AVERAGE) {
            toReturn += `"[color=orange]Not much of a shower, huh? Don't worry, I don't mind. Don't keep me waiting, though.[/color]"`;
        } else if (dick < SizeEnum.LARGER) {
            toReturn += `"[color=orange]A decent sized dick for a decent-mannered kobold. What are you waiting for, hot stuff?[/color]"`;
        } else {
            toReturn += `"[color=orange]Oh [b]my[/b]. You're a big boy, aren't you? Get over here and [b]fuck me silly[/b].[/color]"`;
        }

        toReturn += `       To stress her words, and to further allure you, she temptingly begins to sway her rear from side to side.
                            As if you needed any more persuasion.
                        [/p]`;

        if (dick < SizeEnum.AVERAGE) {
            toReturn += `[p]
                            You eagerly bridge the small gap between the two of you with a couple of big paces.
                            Gripping the woman's furred haunches with one of your hands, you use the other to line up your undersized member.
                            With a single rolling of your hips, you easily manage to slip inside of Cassidy's moist cunt,
                            which is barely tight enough to offer any resistance at all.
                            You can see her fighting back a demeaning giggle.
                            "[color=orange]It's okay, darling. Try thrusting.[/color]"
                        [/p]
                        [p]
                            While you dislike the vulpine practically [b]laughing[/b] at the size of your length,
                            and you don't enjoy being patronized,
                            you still decide to take her advice.
                            You begin to hump against her rear, vigorously pumping your shaft in and out of her less-than-tight confines,
                            easily bottoming out each and every time, your scaly hips smacking against her wet slit with every thrust.
                            She tries to act like she's [i]into it[/i], but you can clearly tell that you're simply not cut out to satisfy a woman of Cassidy's size.
                        [/p]
                        [p]
                            [c]
                                ${global.ui.inlineImage.create(`scenes/HamletDay2/BreedingFarm/img/con-small-m`, "png", ImageWidthEnum.SMALLER, false, "Sneaky")}
                            [/c]
                        [/p]
                        [p]
                            Despite not properly clenching your member like a smaller species' loins would,
                            the female fox's warm confines still prove exceedingly pleasurable.
                            You wish you could say this entire ordeal was equally enjoyable for her,
                            but in between streaks of [i]pretending[/i] to like it and simply smiling back at you while you do your thing,
                            you can sense that she's merely letting you do this as some twisted form of a [i]favor[/i].
                            You try a few fiercer, more domineering thrusts,
                            you even tug on her tail a little to make up for your size,
                            but it's all for naught.
                        [/p]
                        [p]
                            Luckily for Cassidy, it only takes a few minutes of rabid humping for you to reach your end.
                            She barely even flinches as you blow your meagre load inside of her, painting her insides with rope after rope of thick kobold spunk.
                            "[color=orange]Finished?[/color]" she asks, as though she didn't feel a thing.
                            You're pretty sure she's joking.
                            At least you hope she is.
                        [/p]
                        [p]
                            After you pull back from her folds, the vixen rises to her feet.
                            She wipes her loins with the flat of her hand, collecting the smidgen of cum oozing from her cunt,
                            just to playfully wipe it off on your face.
                            "[color=orange]Well that was...[/color]" she stops mid-phrase,
                            preventing herself from [i]accidentally[/i] speaking her mind,
                            coming up with an apt description before finishing her sentence.
                            "[color=orange]...[i]fun[/i].[/color]"
                        [/p]
                        [p]
                            You and her exchange nervous smiles.
                            You almost feel like you should [i]apologize[/i],
                            but then again, that might just make this even [b]more[/b] awkward than it already is.
                            An uncomfortable silence falls upon the conversation, but it is broken as your [i]partner[/i] remembers something.
                            "[color=orange]Oh, right![/color]" she exclaims, walking over to a nearby cabinet.
                            ${pants || underwear ? `While her back is turned, you put your ${pants && underwear ? `clothes` : `${pants ? `pants` : `underwear`}`} back on to hide your shame.` : ""}
                        [/p]
                        [p]
                            Opening up a drawer, the canine produces an oddly-shaped bottle from within.
                            Returning to you, she hands you an ${global.kobold.inventory.createItemWithTooltip(new mPlus())}, which you begrudgingly pocket.
                            "[color=orange]You might want to drink that before, uh, you...[i]you know[/i].
                            With someone bigger than a kobold.[/color]"
                            You don't know whether you should feel thankful or [i]insulted[/i].
                        [/p]
                        [p]
                            Your silence, however, isn't helping the situation along.
                            Once again, Cassidy is forced to take the ropes.
                            "[color=orange]Don't worry about it.
                            Consider it a gift.
                            But anyway, I have matters to attend to.
                            [small]Matters such as rubbing one out[/small].
                            I'm sure you can find your own way to the door?[/color]"
                        [/p]
                        [p]
                            Without waiting for your response,
                            probably eager to get [i]away[/i] from this [b]mess[/b] of a hookup,
                            the vulpine turns around and walks away without so much as a goodbye.
                            You let out a loud sigh, before doing the same.
                            Feeling almost [i]rejected[/i], you step back towards the front door.
                            As you swing it open, you can hear the vixen ascending stairs,
                            probably heading for her bedroom to do what you were unable to.
                            With a heavy heart, you leave the mansion.
                        [/p]`;
        } else if (dick < SizeEnum.LARGER) {
            toReturn += `[p]
                            You eagerly bridge the small gap between the two of you with a couple of big paces.
                            Gripping the woman's furred haunches with one of your hands, you use the other to line up your adequate length.
                            With a single thrust, you spread Cassidy's folds, slamming your member inside of her,
                            not quite managing to hilt yourself fully on the first go, but getting close to it.
                            A pleasured grunt escapes her lips.
                            "[color=orange]Keep going.[/color]"
                        [/p]
                        [p]
                            [c]
                                ${global.ui.inlineImage.create(`scenes/HamletDay2/BreedingFarm/img/con-small-m`, "png", ImageWidthEnum.SMALLER, false, "Sneaky")}
                            [/c]
                        [/p]
                        [p]
                            You happily oblige her request.
                            Reinforcing your grip on her thighs,
                            you begin to rock your hips back and forth,
                            slowly inching the remainder of your shaft inside of the vulpine.
                            Smooth and gentle, right until you bottom out,
                            the tapered tip of your member barely prodding into her cervix as the scales covering your hips press into her loins.
                            A perfect fit.
                        [/p]
                        [p]
                            You hold yourself there for a few moments, relishing the warmth surrounding your length,
                            savoring this very moment, imprinting it into your memory so you can recall it for years to come.
                            Just as the canine was about to open her mouth once more,
                            as she was on the verge of telling you to get things going,
                            you take the initiative, and begin to fuck her [b]properly[/b],
                            outright [b]humping[/b] Cassidy, bottoming out with every quick thrust.
                        [/p]
                        [p]
                            Instead of a command or a request, the only thing that escapes her lips is an unexpected moan.
                            She begins to rock back against your motions, as though bouncing off your lap in between humps,
                            riding your cock as you give her not what she wants, but what she [b]needs[/b].
                            The sounds of your scales slapping against her moist cunt reveberate throughout the hallway,
                            accompanied by an occasional pleasured groan on her end and joint, ever-intensifying panting.
                        [/p]
                        [p]
                            Alas, the tight confines of the female fox's loins prove far too pleasurable for a kobold like yourself,
                            your mutual enjoyment cut short by your rapidly approaching orgasm.
                            Your high-paced fucking turns into a series of slower, rougher, more intensive thrusts,
                            something the vixen doesn't seem to mind in the slightest.
                            She takes hold of the far-end of the table to brace herself as you pound away,
                            your motions becoming a lot more erratic, your climax nigh-imminent. 
                        [/p]
                        [p]
                            You try to hold yourself back for as long as you can, but the pleasure is simply [b]too much[/b] to endure.
                            After a few more frantic pumps of your hips, you hilt yourself inside of Cassidy one final time before blowing your load,
                            your thrusts reduced to the slow-paced back-and-forth-rocking you started off with
                            as wave after wave of your spunk squirts out of the tip of your member,
                            filling the fox up to the brim.
                        [/p]
                        [p]
                            Panting, you gradually pull yourself from the woman's well-fucked hole.
                            While she didn't quite reach her own end, you can clearly tell that she, too, enjoyed this.
                            Once the tip of your spent length exits her cunt, her slightly gaping slit begins to ooze a steady stream of cum.
                            It pools up on the wooden floor, and its heady scent quickly overpowers the lingering feminine odour of vulpine need.
                        [/p]
                        [p]
                            After catching her breath, Cassidy rises to her feet.
                            She rubs her loins with the palm of her hand, scooping up some of your jizz and then bringing it to her muzzle,
                            loudly sniffing the stuff, rubbing some of it onto her snout before licking off the rest.
                            You roll your eyes.
                            [i]Canines[/i].
                            "[color=orange]That was...[b]good[/b].[/color]" she speaks, panting in between words.
                        [/p]
                        [p]
                            ${pants || underwear ? `While the vixen regains her composure, you casually put your ${pants && underwear ? `clothes` : `${pants ? `pants` : `underwear`}`} back on.` : "It takes a minute for the vixen to regain her composure."}
                            Collecting her thoughts, she addresses you once more.
                            "[color=orange]Thank you for that, ${name}.
                            I'd appreciate it if you could stop by again sometime.[/color]"
                            You nod in agreement.
                            This is [b]definitely[/b] something you'd be willing to repeat.
                            You'll be sure to visit again if you find yourself in the neighbourhood.
                        [/p]
                        [p]
                            "[color=orange]Anyway, I [i]do[/i] have some matters to attend to.
                            After all, I've a farm to run.[/color]"
                            That being said, Cassidy escorts you to the door.
                            After giving you one last kiss on your cheek,
                            and a reminder that she'd [b]really[/b] like to see you again some day,
                            she sends you on your way back out into the cold world outside of her fancy home.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You eagerly bridge the small gap between the two of you with a couple of big paces.
                            Gripping the woman's furred haunches with one of your hands, you use the other to line up your oversized cock.
                            Realizing this'll be a tight fit,
                            you give yourself a couple of strokes whilst rubbing the tapered tip of your member up and down the entire length of Cassidy's slit,
                            smearing an ample amount of precum which mixes with her own moistness to serve as much-needed lubricant.
                        [/p]
                        [p]
                            Once Cassidy's entrance is somewhat slickened, you [b]ram[/b] your hips forwards, spearing open her folds with your girthy shaft,
                            outright [b]forcing[/b] over half of your member inside of her.
                            The vixen lets out a loud, surprised, yet pleasured moan.
                            You think she might have [i]underestimated[/i] you.
                            Her obvious arousal only spurs you on more.
                            You pull back ever so slightly, just so you can plunge forwards once again,
                            sinking another few inches of your hefty tool into the vulpine's overly tight passage,
                            spreading her to her very limits, your tapered tip already pressing into her cervix, despite there still being several more inches to go. 
                        [/p]
                        [p]
                            [c]
                                ${global.ui.inlineImage.create(`scenes/HamletDay2/BreedingFarm/img/con-big-m`, "png", ImageWidthEnum.SMALLER, false, "Sneaky")}
                            [/c]
                        [/p]
                        [p]
                            Cassidy seems dumbstruck, completely [b]enthralled[/b] by the sheer size of your cock.
                            You can feel her trying to [b]relax[/b], doing her best, but [b]failing[/b] to accomodate [b]more[/b] of you,
                            her body simply not meant to take someone of your size.
                            Realizing you won't be able to force yourself in [i]all the way[/i],
                            no matter how much either of you try, you decide to roll with what you've got.
                        [/p]
                        [p]
                            You slowly begin to rock back and forth, fucking the canine at a tauntingly slow pace,
                            allowing her to get used to being pushed to her limit and beyond.
                            A small moan accompanies your every motion, the female fox already [b]dripping wet[/b], her excitement easing your smooth rocking.
                            While initially gripping you like a vice, her cunt gradually loosens to the point where you can comfortable slide in and out,
                            a perfect preparation for what is to come.
                        [/p]
                        [p]
                            Cassidy, too, knows that this is just the start.
                            She takes hold of the far-end of the table, right on time to brace herself as you up the pace.
                            Crudely [b]tugging[/b] on her tail, as if to signal that the [b]real[/b] fun is about to begin,
                            you begin to [b]pound[/b] away,
                            rocking the entire table back and forth with every fierce thrust as you thoroughly [b]plow[/b] the vixen's twat,
                            the tip of your member slamming against her cervix every time you roll your hips forwards.
                        [/p]
                        [p]
                            Instead of moaning, the cowgirl loudly [b]screams[/b] out in pleasure,
                            her entire mind going blank as she experiences pleasure unlike anything she's never felt before.
                            Only a few thrusts in, she rises up from the table's surface, her back arching, her cunt pulsing and contracting around your length,
                            so much of her excitement spilling forth that an entire [b]puddle[/b] forms on the floor below, 
                            its hefty feminine fragrance inspiring you to fuck her [b]harder[/b].
                        [/p]
                        [p]
                            She may have reached her end, but you haven't reached yours yet.
                            In an incredible display of breeding-need-fueled dominance, you [b]force[/b] the vixen back down onto the table,
                            yanking up her ass by her tail to raise it ever so slightly, giving you a better angle from which to thrust into her,
                            allowing you to slide a teensy bit [b]more[/b] of your girth into her body.
                            Overtaken by lust, you wildly [b]fuck[/b] the ever-loving [b]shit[/b] out of the bitch in heat.
                        [/p]
                        [p]
                            Cassidy's tongue obscenely lolls out of her mouth over the edge of the table, saliva leaking from her open mouth.
                            Having gone hoarse from screaming her lungs out in pleasure, she has resorted to meekly moaning to voice her ecstasy,
                            a sound so soft you can barely hear it over your own loud panting.
                            While growing exerted, you can sense your orgasm slowly building up inside of you.
                        [/p]
                        [p]
                            You slap the vulpine's furred ass,
                            waking her up from her lust-induced haze just so she can [b]feel[/b] you [b]ramming[/b] your member inside of her one final time,
                            with [b]all[/b] of your might, the tapered tip ever-so-slightly piercing her cervix,
                            the first load of your incredible climax squirting straight into her womb.
                            Her eyes open wide, before rolling into the back of her head.
                            For the first time since you started fucking her, she actively begins to hump back at you,
                            riding your length, milking you of every single droplet contained within your hefty internal balls,
                            your fragrant seed all-too-soon spilling out from her well-fucked hole, oozing out around your girth,
                            her cunt completely filled to the very [b]brim[/b].
                        [/p]
                        [p]
                            As your orgasm fades, you begin to pull out from the lusty vixen, a [b]torrent[/b] of your spunk flowing from her near-abused slit,
                            mixing with the puddle of her own excitement.
                            As your shaft leaves Cassidy's body completely, she lazily rolls off of low table,
                            coming to a rest on her back on the floor, fully spread eagle, cunt visibly gaping while large amounts of your jizz continues to seep from it.
                        [/p]
                        [p]
                            She seems [b]completely[/b] [i]out of it[/i].
                            Like you've [b]broken[/b] her.
                            You catch yourself staring concernedly.
                            And she must have caught you as well.
                            "[color=orange]I'm...fine.[/color]" she speaks, panting in between words.
                            "[color=orange]Just...give me a few minutes.[/color]"
                        [/p]
                        [p]
                            You honor her request, taking the time to admire your handiwork.
                            ${pants || underwear ? `While you wait, you casually put your ${pants && underwear ? `clothes` : `${pants ? `pants` : `underwear`}`} back on.` : ""}
                            Finally, after a good minute and a half, the vixen lets her hoarse voice be heard once more.
                            "[color=orange]That. Was. [b]Incredible[/b].[/color]"
                            She doesn't even bother to lift her head up off the ground to look at you.
                            "[color=orange]I don't think I can [b]ever[/b] go back to my own breeders after [b]that[/b].
                            [b]Please[/b] promise me you'll be back.[/color]"
                        [/p]
                        [p]
                            You say you can't make any promises, but if you find yourself in the neighbourhood, you might decide to pay her a visit.
                            On one condition, that is.
                            Next time, you do it somewhere more comfortable than the hallway.
                            You [b]are[/b] a kobold with [b]standards[/b], after all.
                        [/p]
                        [p]
                            Amused by your own joke, you find your way to the door.
                            Taking one last look at your thoroughly bred [b]bitch[/b], you leave her home, almost as satisfied as she is.
                            [i]Almost[/i].
                        [/p]`;
        }

        if(virgin) {
            toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
        }

        return toReturn;
    }

    function buildFemaleContent() {
        var shirt = ctrl.misc.shirt();
        var bra = ctrl.misc.bra();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var name = ctrl.misc.name();
        var virgin = ctrl.misc.cuntVirgin();
        
        var toReturn = `[p]
                            The canine takes a few steps forwards, towards a small, low, decorative table which stands against the leftmost wall.
                            While continuing to touch herself, she uses her free arm to brush the couple of unopened letters atop its surface to the floor. `;

        if (pants || underwear) {
            toReturn += `Meanwhile, you begin to undress yourself. `;
        } else {
            toReturn += `Cassidy's constant teasing and the sight of her pristine, naked body have thoroughly aroused you,
                         your loins already moist, fully ready for action.`;
        }

        if (pants) {
            toReturn += `You take off your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (underwear) {
                toReturn += ` and your ${global.kobold.inventory.createItemWithTooltip(underwear)}`;
            }
            toReturn += `, your already moist loins coming into view,
                        Cassidy's constant teasing along with the sight of her naked body having thoroughly aroused you.`;
        } else if (underwear) {
            toReturn += `You take off your ${global.kobold.inventory.createItemWithTooltip(underwear)},
                        your already moist loins coming into view,
                        Cassidy's constant teasing along with the sight of her naked body having thoroughly aroused you.`;
        }

        toReturn += `   [/p]
                        [p]
                            The table cleared, the woman looks behind her, directly at you.
                            She lustily licks her lips, inspecting your ${shirt || bra ? "half-naked" : "naked"} body.
                            Then, suddenly, she reaches out, grabbing you by the arm and tugging you close to her.
                            The vulpine pulls you into a tender embrace, your head lodged firmly in between her hefty, furred tits.
                            She hugs and squeezes you tight enough to lift you off the ground, setting you down atop the table.
                        [/p]
                        [p]
                            Cassidy drops to her knees in front of you.
                            Forcing your legs open wide, she presses her face in between your scaly thighs, her snout rubbing against your slit.
                            The vixen loudly and deeply [b]sniffs[/b] your groin, visibly [b]getting off[/b] to your scent,
                            beginning to outright [b]masturbate[/b] to the fragrance of female kobold excitement.
                            You lean backwards, against the wall, further exposing your most intimate parts, giving her full access to your groin,
                            anticipation and the the warm breath being exhaled onto your loins dampening them further.
                        [/p]
                        [p]
                            Unable to restrain herself for much longer, the fox decides to [b]dig in[/b].
                            She begins to lap at your folds with her warm, wet canine tongue, licking your cunt like a dehydrated [b]dog[/b],
                            savoring every single droplet of your excitement, fingering herself to the taste.
                            You spread your legs a little bit more, opening them as far as they can go,
                            a loud moan escaping your lips as pleasure surges throughout your body.
                        [/p]
                        [p]
                            [c]
                                ${global.ui.inlineImage.create(`scenes/HamletDay2/BreedingFarm/img/con-f`, "png", ImageWidthEnum.SMALLER, false, "Sneaky")}
                            [/c]
                        [/p]
                        [p]
                            You already feel like you're in [b]heaven[/b], but the vixen has only just [b]begun[/b] to show you what she's capable of.
                            Her licks gradually get more fierce, more [b]intense[/b], burying [b]deeper[/b] into your slit, spreading your folds more and more,
                            her face inching ever-closer to your loins until her furred muzzle is pressing directly against your heated, arousal-stained scales.
                            And that, is when the [b]real[/b] fun begins.
                        [/p]
                        [p]
                            Cassidy [b]rams[/b] her tongue forwards, slipping it inside of your spit-lubed cunt,
                            sliding most of its elongated length inside of your body, reaching parts so deep you didn't even know they [b]existed[/b].
                            Your mind goes blank with pleasure as she starts to wriggle her tongue about, pleasuring every square inch of your sensitive innards.
                            You wrap your legs around her head, forcing her to stay in place, pulling her in even [b]closer[/b],
                            her long tongue, extended to its very limits, sliding in just a teensy bit further,
                            enough to reach that one special spot that drives you [b]crazy[/b] with lust.
                        [/p]
                        [p]
                            Nearing your climax, you begin to [b]hump[/b] against the vulpine's face,
                            rubbing your cunt along her muzzle as her tongue pistons in and out of your tight snatch.
                            The canine, too, begins to bounce up and down her own fingers,
                            the addded motion of her snout bobbing up and down ever so gently,
                            warm, wet fur brushing against your parted folds, is enough to send you over the edge.
                        [/p]
                        [p]
                            You curl your toes and arch your back, an intense orgasm hitting your body hard.
                            You squeeze down on Cassidy's head using your thighs, wildly rocking your hips back and forth,
                            grinding against her face as she tongues your snatch,
                            squirting a considerable amount of your feminine fluids directly into her open maw,
                            the vixen eagerly swallowing as much of your excitement as she can.
                        [/p]
                        [p]
                            Her tongue slips out of your spasming hole as your climax begins to fade,
                            continuing to lap up even more of your sweet arousal, cleaning you up while sating her own need.
                            You slowly release your grip on her head,
                            limply slinking down onto the table, completely spent,
                            panting loudly while your tender loins continue to be licked.
                        [/p]
                        [p]
                            Finally, after a solid minute more of contant lapping,
                            the vixen finally pulls back from your groin,
                            a few thick strands of mixed spit and feminine fluids connecting her snout to your most delicate scales.
                            With her still-lolling tongue, she licks her muzzle clean, collecting the large amount of fluids lingering on her fur,
                            before swallowing it all, looking into your eyes to ensure you'd witness the small bulge in her throat as she downs the fruits of her labour.
                        [/p]
                        [p]
                            Cassidy rises to her feet, her own [b]extremely messy[/b] groin coming into view.
                            Regaining some of your strength, your take a glimpse downwards,
                            noticing a veritable [b]puddle[/b] of canine excitement spilled on the floor,
                            its heady scent overpowering even the distinct smell of your freshly-sated groin.
                        [/p]
                        [p]
                            "[color=orange]That was [b]fun[/b].[/color]" she states, lending you a hand to pull you off the table, and back onto your shaky feet.
                            ${pants || underwear ? "You wobble towards your discarded clothing and redress." : ""}
                            "[color=orange][b]Do[/b] come visit me again sometimes, won't you?[/color]" the vixen asks, almost pleadingly.
                            You say you'll consider stopping by again if you find yourself in the neighbourhood, but only on one condition.
                            Next time, you get to eat her out as well.
                        [/p]
                        [p]
                            The two of you exchange sultry looks.
                            Not bothering to reply, Cassidy simply gives you one final peck on your cheek, and then she escorts you to the door.
                            After wishing you good luck on your journey, the kobold-breeding vulpine sends you on your way.
                        [/p]`;
        return toReturn;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var toReturn = ``;

        if (dick) {
            toReturn += buildMaleContent();
        } else {
            toReturn += buildFemaleContent();
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildContinueChoice() {
        var disabled = ctrl.postnotbarnsex.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarnsex.has();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildContinueChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarnsex.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnAgreeSex";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.postnotbarn.continued()) {
            return "";
        }
        global.ui.applyTheme("hamlet_mansion");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.notbarnAgreeSexChange(ctrl.misc.dick());
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.postnotbarn.continued()) {
            return;
        }

        next.registerListeners();

        if (ctrl.postnotbarnsex.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\notbarn\agree\NotBarnAgreeSex.js

// scenes\HamletDay2\BreedingFarm\barn\notbarn\agree\NotBarnAgree.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnAgree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnAgreeButton";
    
    var next = new NotBarnAgreeSex(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            While you internally hold some doubts,
                            you tell Cassidy that you don't see how a life without worries, full of sex,
                            with a constant supply of free food and a [b]good-looking[/b] caretaker such as herself could be a bad thing.
                            Your flattering reply seems to please the woman, a broad smirk forming on her face, visible even though she doesn't look back.
                            A loud [i]click[/i] resonates from the door lock, and with a simple push, the vixen opens up the way into her home.
                        [/p]
                        [p]
                            The two of you step inside.
                            You first, the vulpine second.
                            She closes the door behind her, using nothing but her tail.
                            The inside of her home looks every bit as luxurious as the outside.
                            A myriad of paintings line the walls, interspersed with gilded candelabras containing lit candles.
                            The entrance hallway you find yourself in is at least three times the size of your little shack in the woods.
                            You can't say you're left unamazed.
                        [/p]
                        [p]
                            After taking in the splendour on display for a few moments, you turn around to face your hostess.
                            She stares at you with a lusty, predatory look.
                            Now that the two of you are alone, she no longer needs to hold back.
                            One of her hands has already slinked between her own legs,
                            subconsciously rubbing away at her inner thigh.
                            You can tell that she's just about ready to [b]pounce[/b] at you,
                            to [b]leap[/b] on top of you and [b]fuck[/b] until you're both sated.
                            The two of you wordlessly exchange sultry smiles.
                            Your inaction only seems to entice her further,
                            until finally, she can hold herself back no longer.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildContinueChoice() {
        var disabled = ctrl.postnotbarn.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarn.has();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildContinueChoice();
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarn.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnAgree";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.notbarn.get() !== NotBarnChoice.AGREE) {
            return "";
        }
        global.ui.applyTheme("hamlet_mansion");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.notbarnAgreeChange();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.notbarn.get() !== NotBarnChoice.AGREE) {
            return;
        }

        next.registerListeners();

        if (ctrl.postnotbarn.has()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\notbarn\agree\NotBarnAgree.js

// scenes\HamletDay2\BreedingFarm\barn\notbarn\disagree\NotBarnDisagreeContinue.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnDisagreeContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnDisagreeContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            After the both of you recover your respective breaths, Cassidy walks over to your lower end.
                            You get your hopes up, thinking she's about to [i]return the favor[/i],
                            but instead of bringing her muzzle to your groin as you'd hoped,
                            the woman deftly takes hold of one of your ankles,
                            and begins to [b]drag[/b] you towards the door, over the smooth, wooden floor.
                            You stay silent.
                            After all, she hasn't [b]told you[/b] to speak.
                        [/p]
                        [p]
                            Reaching the doorway, the canine opens up the door once more, and promptly kicks you out onto the porch.
                            You look back at her, her loins almost as messy as your face is.
                            She speaks, in a stern tone, glaring at you as though you've done something wrong.
                            "[color=orange]Come back when you realize that your kind amounts to nothing good but pleasuring others.
                            I'll save a spot in the barn, just for you.
                            Until then, you stay the [b]hell[/b] away from my farm.
                            And if you [b]ever[/b] dare to insinuate that I don't have what's best in mind for the kobolds I breed,
                            I'll make you [b]wish[/b] I sent you to the [b]slaughterhouse[/b].[/color]"
                        [/p]
                        [p]
                            After that last threat, Cassidy harshly slams the door shut, leaving you cold and alone on the porch of her home,
                            her arousal sticking to your face, her taste lingering in your mouth,
                            and her scent more than likely having seeped into your scales to the point where you'll be [b]reeking[/b] of fox femcum for days to come.
                        [/p]
                        [p]
                            Part of you is glad that this entire ordeal is finally over with,
                            even if you didn't get anything in return for pleasuring the vulpine, apart from getting yelled at.
                            But then again, she [i]didn't[/i] end up forcing you into a lifetime of sexual servitude, which is a good thing, come to think of it.
                            You pick yourself up off of the ground, wipe your face clean of most of the goop staining it,
                            and swiftly leave the vicinity of the farm, returning to the center of the hamlet.
                        [/p]
                        [p]
                            It's late in the afternoon.
                            Too late to go explore another part of the hamlet.
                            In fact, if you want to reach Varanar before nightfall,
                            you have to leave right away.
                            You take one final glance at the barn and the mansion beside it.
                            You've got mixed feelings about that entire place.
                            You don't think you'll ever go there again.
                            You [i]hope[/i] you'll never go there again.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnDisagreeContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.postnotbarnsex.continued()) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.notbarnDisagreeContinueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.postnotbarnsex.continued()) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\notbarn\disagree\NotBarnDisagreeContinue.js

// scenes\HamletDay2\BreedingFarm\barn\notbarn\disagree\NotBarnDisagreeSex.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnDisagreeSex(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnDisagreeSexButton";

    var next = new NotBarnDisagreeContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Your admission of defeat seems to stroke the cowgirl's ego, the smirk plastered across her face only deepening.
                            Still, she's graceful enough [b]not[/b] to rub it in your face, metaphorically speaking.
                            Without saying a word, the vixen presses one of her soft palms against your forehead,
                            pinning you against the floor with something that [i]isn't[/i] her foot.
                        [/p]
                        [p]
                            "[color=orange][b]Stay[/b].[/color]" comes the unnecessary, simplistic, almost taunting order.
                            And like the [i]tamed[/i], [i]broken[/i] [b]animal[/b] that you are, you oblige Cassidy's command,
                            moving not even a muscle while she lifts her heavy boot off your chest,
                            ending up standing over you with a foot on either side of your lithe frame.
                        [/p]
                        [p]
                            Then, the woman begins to lower her own body, squatting down over your head,
                            her thighs widely spread, her visibly needy cunt rapidly approaching your face.
                            If it weren't for her hand atop your cranium [b]forcing[/b] you to stay in place, you'd have faced away once more.
                            Right before her moist lips would have pressed against your muzzle, the canine suddenly halts her motions,
                            hovering her loins directly in front of your lips, perhaps to keep you from biting.
                        [/p]
                        [p]
                            "[color=orange][b]Lick[/b].[/color]" she demands.
                            A part of your mind would rather [b]die[/b] than submit, but the consequences of disobeying the vixen could be [b]far worse[/b] than mere death.
                            After a moment of hesitation, fear triumphs over what little pride you have left.
                            You slowly part your lips, and from your open muzzle snakes your long tongue.
                            You lightly press the tip of the prehensile appendage against the vulpine's folds,
                            giving her a single tentative lick,
                            a droplet of her feminine fluids collecting atop your tongue,
                            rolling down across its entire length into your mouth,
                            trailing the concentrated taste of a [b]bitch[/b] in heat along your tastebuds.
                        [/p]
                        [p]
                            Her scent, too, reaches your nose, every breath you take growing more and more [b]saturated[/b] with the fragrance of pure, feminine fertility.
                            "[color=orange][b]Keep going[/b].[/color]" comes your next order.
                            You begrudgingly comply, licking her again and again and again,
                            the tip of your tongue trailing back and forth along the entire length of her slit,
                            parting her folds ever so slightly with every flick of your tongue.
                        [/p]
                        [p]
                            You don't know whether she's inching closer,
                            or if you subconsciously [b]enjoy[/b] the intense taste of the female fox's cunt,
                            but you find your tongue digging deeper and deeper into her snatch,
                            a mixture of saliva and excitement dripping down atop your face as you vigorously [b]lap[/b] at her folds.
                            Cassidy has taken to playing with her chest using her one free hand, her loud moans echoing throughout the entire house.
                        [/p]
                        [p]
                            Still, she needs [b]more[/b].
                            A small growl emanates from the back of the canine's throat, her bestial side somewhat taking hold of her.
                            She removes her hand from your head, instead pressing her loins directly up to your muzzle,
                            [b]burying[/b] you in between her furred thighs, pinning you in place by outright [b]sitting[/b] on your face.
                            Your snout ends up bumping against the vixen's clit,
                            your nostrils and your lungs both filling with a thick concentration of pure female pheromones,
                            while your entire muzzle gets [b]soaked[/b] by her arousal, which liberally leaks from her loins,
                            rolling down the side of your face all the way to the floor.
                        [/p]
                        [p]
                            [c]
                                ${global.ui.inlineImage.create(`scenes/HamletDay2/BreedingFarm/img/sit`, "png", ImageWidthEnum.SMALLER, false, "Sneaky")}
                            [/c]
                        [/p]
                        [p]
                            "[color=orange][b]Deeper[/b]![/color]" she shouts,
                            far more forceful than any of the previous commands, sheer lust clouding the woman's thoughts.
                            Partly out of intimidation, partly out of your own volition, you stop licking Cassidy for a moment.
                            Instead, you force your tongue deep inside of her cunt, and begin to properly [b]eat[/b] the bitch out.
                            You open your muzzle as wide as it'll go, enveloping most of her folds with the warm wetness of your mouth,
                            slightly suckling on her slit, collecting mouthful after mouthful of sweet,
                            feminine fluids while your tongue twists and turns inside of her, stroking every part of her innards, as far as you can reach.
                        [/p]
                        [p]
                            Lodging the prehensile appendage into the vulpine's body triggers a primal urge, a [b]need[/b] to [b]mate[/b] inside of her.
                            Losing all inhibitions, she wildly begins to ride your face, outright [b]humping[/b] your muzzle,
                            rubbing her twat along your scales, stopping only to bounce up and down your pistoning tongue.
                            Suddenly, her movements come to an end.
                            She arches her back, pressing down on you one final time with all of her might,
                            nearly [b]crushing[/b] your head against the floor just to lodge an extra inch of your tongue inside of her,
                            a hefty climax rocking her body hard.
                        [/p]
                        [p]
                            Lustful excretions [b]gush[/b] from Cassidy's contracting cunt, spilling into your wide open mouth.
                            You greedily gulp them down, your senses completely [b]overloaded[/b] with the taste and the scent of the female fox,
                            your own mind going blank for a few moments, your body relaxing and even [b]enjoying[/b] the feeling of bringing 
                            a${ctrl.misc.baseGender() === GenderEnum.MALE ? "" : "nother"} woman to climax.
                        [/p]
                        [p]
                            After a few seconds of idly sitting atop your face, [b]revelling[/b] in post-orgasmic bliss,
                            the vixen slowly rises to her feet, your tongue sloppily sliding out of her thoroughly-sated slit.
                            You let it loll out over your chin, to show off its length, your pheromone-addled brain taking a twisted form of pride in your actions.
                            "[color=orange]Good ${ctrl.misc.baseGender() === GenderEnum.MALE ? "boy" : "girl"}.[/color]" she cooes, in between huffs and puffs.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildContinueChoice() {
        var disabled = ctrl.postnotbarnsex.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarnsex.has();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildContinueChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarnsex.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnDisagreeSex";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.postnotbarn.continued()) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.notbarnDisagreeSexChange();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.postnotbarn.continued()) {
            return;
        }

        next.registerListeners();

        if (ctrl.postnotbarnsex.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\notbarn\disagree\NotBarnDisagreeSex.js

// scenes\HamletDay2\BreedingFarm\barn\notbarn\disagree\NotBarnDisagree.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnDisagree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnDisagreeButton";

    var next = new NotBarnDisagreeSex(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var shirt = ctrl.misc.shirt();
        var toReturn = `[p]
                            No!
                            Keeping kobolds as [b]slaves[/b] is [b]wrong[/b]!
                            You passionately express your concerns about what the vixen is doing,
                            but she seemingly [i]ignores[/i] you, instead concentrating on getting the door open.
                            You raise your voice, you even yell at her a little to grab her attention,
                            or at least to elicit [b]any[/b] kind of response, without success.
                        [/p]
                        [p]
                            To an outsider, the entire scene probably resembled a child throwing a tantrum,
                            with its mother not being in the mood to humor it.
                            Right as you're getting fed up with being ignored, as you've reached your limit,
                            as you've had about [b]enough[/b] of Cassidy not so much as acknowledging your [b]existence[/b],
                            a loud [i]click[/i] resonates from the lock she's trying to open, interrupting your outburst of anger.
                        [/p]
                        [p]
                            She pushes open the front door to her house.
                            Then, in the flash of an eye, before you even [i]know[/i] what's going on,
                            the canine reaches behind her, crudely grabbing your arm with pinpoint precision.
                            She hurls you forwards, practically [b]tossing[/b] you into her home.
                            You fail to keep your balance, landing face-first on the smooth,
                            wooden floor, sliding along it for a good foot or two before coming to a halt.
                        [/p]
                        [p]
                            You flip yourself onto your back,
                            right on time to witness Cassidy entering her own mansion as well, closing the door once more,
                            the luxurious hallway you've been cast into suddenly growing a lot more claustrophobic 
                            as the last rays of sunlight beaming in through the doorway get blocked by the wooden surface.
                            Now it's just you, her, and the multitude of lit candles and paintings lining the walls.
                        [/p]
                        [p]
                            The sultry smile on the vixen's face has been replaced by an angry glare.
                            "[color=orange][b]Uppity[/b] little [b]sluts[/b] like [b]you[/b] need to be [b]tamed[/b].[/color]" she says,
                            her tone seemingly [b]dead-serious[/b].
                            Right after those words have left her mouth, the near-naked vulpine begins to advance upon you, slowly walking towards your downed form.
                        [/p]
                        [p]
                            You panic, wildly flailing your limbs about in a desperate attempt to get away,
                            your mind torn between either crawling backwards or rising to your feet, your body doing neither.
                            Your smooth-scaled hands and feet find little to no grip on the surface of the polished floor,
                            and with nothing nearby to grab onto, you are left flailing like a fish out of water,
                            watching in horror as the woman leisurely approaches.
                        [/p]
                        [p]
                            Realizing you aren't getting away, at least not like this, you cease your frantic squirming.
                            You try to think of another way out, of a plan of action, of a way to escape, yet it's already too late.
                            A heavy leather boot plops down onto your chest, [i]almost[/i] [b]crushing[/b] the air out of you,
                            the vixen grateful enough [b]not[/b] to stomp down with all of her might, or even to lean her entire weight onto you,
                            applying [i]just[/i] enough pressure to keep you pinned down, and nothing more.
                        [/p]
                        [p]
                            You consider fighting against her foot-hold, yet the sharp spur at the back of her boot, which is mere inches away from
                            ${shirt ? `tearing into your ${global.kobold.inventory.createItemWithTooltip(shirt)}` : `carving into your flesh`},
                            dissuades you from struggling.
                            Instead, you lie limply, staring up at Cassidy with fear in your eyes.
                            She leans over you like a hunter would over felled prey, her toothy, predatory grin only serving to make you even [b]more[/b] afraid.
                        [/p]
                        [p]
                            Staring down at you, the woman speaks:
                            "[color=orange]You promised me a good time, and that's [b]exactly[/b] what I intend on having.
                            I don't care how much you [b]hate[/b] my guts.
                            You're going to do [b]exactly[/b] as I say,
                            and then maybe I'll consider letting you [b]leave[/b] instead of hogtying you,
                            throwing you into the barn, and showing you exactly how [b]depraved[/b] your kind can get.[/color]"
                        [/p]
                        [p]
                            You swallow deeply, the true severity of the situation only now dawning upon you.
                            You continue to look for an exit plan, for a way to wriggle out from underneath her, yet you fail to come up with anything.
                            You are trapped.
                            Stuck between a boot and a hard place.
                            There [b]is[/b] no way out.
                        [/p]
                        [p]
                            "[color=orange]So, [b]lizard[/b].
                            What'll it be?
                            Are you going to be a good ${ctrl.misc.baseGender() === GenderEnum.MALE ? "boy" : "girl"}?
                            Or am I going to have to [b]punish[/b] you?[/color]"
                        [/p]
                        [p]
                            Panic turning to unwillingness, you look away to the side for a few moments,
                            simply staring at the bottom of the wall while your brain struggles to decide.
                            A single tear spills from your eyes, rolling down over your scales, dropping to the ground.
                            Your teary gaze wanders back towards Cassidy, who is patiently awaiting your reply.
                            She knows she's fully in control of the situation.
                            And you, too, are painfully well aware of that fact.
                            Your fighting spirit broken, you give her but a single, defeated nod.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildContinueChoice() {
        var disabled = ctrl.postnotbarn.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarn.has();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildContinueChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarn.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnDisagree";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.notbarn.get() !== NotBarnChoice.DISAGREE) {
            return "";
        }
        global.ui.applyTheme("hamlet_mansion");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.notbarnDisagreeChange();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.notbarn.get() !== NotBarnChoice.DISAGREE) {
            return;
        }

        next.registerListeners();

        if (ctrl.postnotbarn.has()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\notbarn\disagree\NotBarnDisagree.js

// scenes\HamletDay2\BreedingFarm\barn\notbarn\NotBarn.js

/**
 * Not barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarn(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var disagreeId = "notBarnDisagreeB";
    var agreeId = "notBarnAgreeB";
    var attackId = "notBarnAttackB";

    var toDisagree = new NotBarnDisagree(ctrl);
    var toAgree = new NotBarnAgree(ctrl);
    var toAttack = new NotBarnAttack(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            You slowly nod, but at the same time you stammer out that you don't want to do it [i]here[/i].
                            While that table looks inviting, you mention you'd rather [i]do it[/i] inside of the house.
                            Unless she has a problem with that?
                            A confused look forms on the vixen's face.
                            It takes a few seconds for her to process what you just said.
                            It's like she doesn't know how to feel about your response.
                        [/p]
                        [p]
                            "[color=orange]In the house? Not here, on the table?[/color]" the vulpine finally asks.
                            You reply with a single, swift nod, full of conviction.
                            You would [b]really[/b] prefer a [i]cleaner[/i] environment than this dusty old barn.
                            The woman removes her hand from your shoulder, and straightens her back once more, her face no longer mere inches away from yours.
                            "[color=orange]You're not the first kobold I make that offer, you know.
                            You [i]are[/i], however, the first to...[/color]" She considers her words. "[color=orange]...[i]dislike[/i] the barn.[/color]"
                        [/p]
                        [p]
                            Your turn to feel conflicted.
                            You aren't sure whether the canine is complimenting or berating you.
                            Noticing your confusion, she shoots you a small grin and a wink.
                            "[color=orange]It's nice to meet a kobold with standards, for once.[/color]"
                            A mountain of doubt rolls off your shoulders, Cassidy's compliment striking every right note,
                            pushing all the buttons to make a prideful smile form on your face.
                            She's right.
                            You're not like the others.
                            You have [b]standards[/b].
                        [/p]
                        [p]
                            The female fox extends her arm towards you, offering her furred hand.
                            "[color=orange]To the house, then, you [i]demanding little lizard[/i]?[/color]"
                            Ignoring her jest, you eagerly grab her hand, and the two of you exit the barn together.
                        [/p]
                        [p]
                            You walk hand in hand towards the fancy mansion adjacent to the much older barn.
                            While you don't accidentally want to break the mood, you do think a little bit of smalltalk is in order.
                            Your lust may be sated soon, but that still leaves your burning curiosity unchecked.
                            You ask the vixen what [i]exactly[/i] it is she breeds here.
                            You never really [i]did[/i] get a good look at the inside of those pens.
                        [/p]
                        [p]
                            The canine takes a moment to think.
                            She speaks slowly, carefully selecting her words.
                            "[color=orange]Oh, you know.
                            Small, cute things.
                            Some people like keeping them as pets.
                            Others like eating them.
                            Me?
                            I like [b]sleeping[/b] with them.[/color]"
                        [/p]
                        [p]
                            You fail to pick up on the blatant hint contained within the fox's reply.
                            Still clueless, you begin to think.
                            A small, cute pet.
                            Rabbits?
                            People like eating rabbits.
                            And earlier on, you even thought that that might have been it.
                            But Cassidy wouldn't [i]fuck[/i] a bunny, [i]would she[/i]?
                            And if she [i]did[/i] sleep with animals, why would she tell you about it?
                        [/p]
                        [p]
                            As the two of you near the front door, you are just as confused as you were [i]before[/i] her suggestive comment.
                            With an inquisitive look on your face, you repeat your question: what [b]does[/b] she breed?
                            She rolls her eyes, clearly unamused by your lack of comprehension.
                            Still, you can [i]feel[/i] that she doesn't feel comfortable outright [i]saying[/i] it.
                        [/p]
                        [p]
                            Her smile turns into a nervous grin.
                            The vixen lets go of your hand, and takes a step forward, towards the door, her bare back facing you.
                            "[color=orange]Well...[/color]"
                            she begins, whilst deeply bending over, lifting her tail up high to give you a clear look at her furred rump,
                            the fertile loins placed in between her thick thighs distracting you more than a little, 
                            almost as if she's trying to trick you, to prevent you from focussing on her reply.
                            While your eyes are glued to to the vulpine's groin, sheer inquisitiveness wins out from arousal, your attention remaining on-point.
                            Finally, the answer is revealed to you.
                            "[color=orange]...[small]kobolds[/small]. I breed kobolds.[/color]"
                        [/p]
                        [p]
                            Oh.
                            [i]Oh[/i].
                            Wait.
                            What?
                            Kobolds?
                            Your kind?
                            She said she bred things for fucking, for selling them as pets, and for...[b]eating[/b]?!
                            Reaching into her boot, Cassidy produces a small metal object from her footwear.
                            The key to her house.
                            Looking back at you, she notices how upset you are.
                            She realizes her [i]distraction[/i] must not have worked.
                        [/p]
                        [p]
                            Stepping towards the door, the canine begins to ratify her actions as she fiddles with the lock, her tone changing considerably.
                            "[color=orange]I thought you knew.
                            They live good lives.
                            Without me, they'd probably be off robbing someone or dying in a ditch somewhere.
                            In the barn, they get to spend all day fucking eachother, and sometimes even [i]me[/i].
                            They get plenty of food, straw to keep them warm at night, heck, I even read them a bedtime story once every week.
                            Doesn't that sound like a good time, ${name}?[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildDisagreeChoice() {
        var disabled = ctrl.notbarn.has() ? "disabled" : "";
        var isSelected = ctrl.notbarn.get() === NotBarnChoice.DISAGREE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${disagreeId}" ${disabled}>Disagree</button>`;

        var tooltip = global.parser.parse(`No! Your kind is [b]more[/b] than just [b]cattle[/b]!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 155px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildAgreeChoice() {
        var disabled = ctrl.notbarn.has() ? "disabled" : "";
        var isSelected = ctrl.notbarn.get() === NotBarnChoice.AGREE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${agreeId}" ${disabled}>Agree</button>`;
        var tooltip = global.parser.parse(`You don't want to ruin your chances of [i]getting some[/i]. Agree with the vixen.`);
        
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildAttackChoice() {
        var weapon = ctrl.misc.weapon();
        var disabled = ctrl.notbarn.has() || !weapon ? "disabled" : "";
        var isSelected = ctrl.notbarn.get() === NotBarnChoice.ATTACK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${attackId}" ${disabled}>Attack her</button>`;

        var tooltip;
        if (weapon) {
            tooltip = global.parser.parse(`How [b]dare[/b] she keep your kind as [b]sex slaves[/b]?! Time to teach this bitch a lesson she won't soon forget.`);
        } else {
            tooltip = global.parser.parse(`You don't have any weapons to attack the woman with.`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildDisagreeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAgreeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAttackChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + disagreeId), $("#" + agreeId), $("#" + attackId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerDisagreeListener() {
        $("#" + disagreeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.notbarn.set(NotBarnChoice.DISAGREE);
            loadNextPart(toDisagree);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerAgreeListener() {
        $("#" + agreeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.notbarn.set(NotBarnChoice.AGREE);
            loadNextPart(toAgree);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerAttackListener() {
        $("#" + attackId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.notbarn.set(NotBarnChoice.ATTACK);
            loadNextPart(toAttack);
        });
    }
    

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarn";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if(ctrl.barn.get() !== BreedingBarnChoice.NOTBARN) {
            return "";
        }
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toDisagree.build();
        part += toAgree.build();
        part += toAttack.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.barn.get() !== BreedingBarnChoice.NOTBARN) {
            return;
        }
        
        toDisagree.registerListeners();
        toAgree.registerListeners();
        toAttack.registerListeners();

        if (ctrl.notbarn.has()) {
            return;
        }

        registerDisagreeListener();
        registerAgreeListener();
        registerAttackListener();
    };
}

var NotBarnChoice = {
    DISAGREE: 1,
    AGREE: 2,
    ATTACK: 3
};

// END scenes\HamletDay2\BreedingFarm\barn\notbarn\NotBarn.js

// scenes\HamletDay2\BreedingFarm\barn\tea\agree\NotBarnTeaAgreeContinue.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnTeaAgreeContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnTeaAgreeContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Your belly filled with tea and treats, you make your way back to the center of the hamlet.
                            The vixen may have an odd profession, but she isn't the worst of people.
                            In the end, you're glad you met her.
                            It was nice to speak to someone that [i]didn't[/i] want your money or your life, for once.
                            And, for the first time in your life, you got to see what it's like inside one of those big houses rich people live in.
                            Not to mention the free food and the well-brewed tea.
                        [/p]
                        [p]
                            Overall, you had a nice time.
                            With your spirits lifted, and your body warmed up at least a little, you set off for Varanar.
                            Hopefully, you can make even more friends there, although preferably ones with less morally ambiguous occupations.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnTeaAgreeContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.postnotbarn.continued()) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.teaAgreeContinueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.postnotbarn.continued()) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\tea\agree\NotBarnTeaAgreeContinue.js

// scenes\HamletDay2\BreedingFarm\barn\tea\agree\NotBarnTeaAgree.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnTeaAgree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnTeaAgreeButton";
    
    var next = new NotBarnTeaAgreeContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Yeah.
                            You can kind of see where she's coming from.
                            While you don't [i]hate[/i] your own kind, you realize that most kobolds are [i]not[/i] very good people.
                            Most of your peers are bandits, beggars, lowlives and scam artists.
                            The ones that make a living for themsevles often do so via clandestine ways.
                            And the ones that [i]don't[/i] would do [b]anything[/b] for a meal a day, let alone a roof above their heads.
                        [/p]
                        [p]
                            Cassidy seems glad you understand her position.
                            Finally managing to open the front door, she invites you into her house.
                            An invitation which you don't decline, the warmth rushing forth from her hallway seeming ever-so enticing.
                        [/p]
                        [p]
                            The vixen's mansion is as fancy on the inside as it is on the outside.
                            The entire place is well-lit by candles.
                            Various paintings of kobolds and foxes alike line the walls.
                            The canine leads you to the den, where you get to take a seat in a comfortable padded chair.
                            After a bit of smalltalk, the woman wanders into the kitchen to prepare the tea.
                            You take a moment to admire the sheer luxury you are surrounded with.
                            You've never been in a place this lavish.
                        [/p]
                        [p]
                            It is only when Cassidy returns with a pair of filled teacups and a platter full of biscuits,
                            that you stop staring at the various gold-plated items on display, you attention returning to her naked frame.
                            The vulpine hands you your tea, and then sits down ever so daintily.
                            She sips her own drink, pinky finger fully extended.
                            She'd almost be lady-like if she weren't half-naked wearing cowboy attire.
                            [i]Almost[/i].
                        [/p]
                        [p]
                            As you mutually partake in delicious, body-warming tea and crunchy, vanilla-flavored biscuits, the two of you get to chatting.
                            Topic at hand: how she came to be a [i]kobold breeder[/i].
                            Apparently, this place used to be a regular farm, until the vixen's fox-husband one day decided [i]not[/i] to come home from working the field.
                            While she doesn't know what happened to him, Cassidy assumes her other half [i]ran off[/i] to go live a better life elsewhere.
                        [/p]
                        [p]
                            Sick and tired of her own kind, unwilling to become a prostitute,
                            and unable to cultivate any meaningful amount of grain by herself,
                            the female fox turned to her most secretive of kinks.
                            The one thing that turned her on the most.
                            Something most people would ridicule her for, should they find out her deviant tastes.
                            Kobolds.
                        [/p]
                        [p]
                            She renovated the storage barn, bought her first pair of your kind from a slave trader, and the rest is history.
                            Nowadays, on top of making a pretty penny, she has access to all the kobold sex she could ever want.
                            But she's always on the lookout for new faces.
                            That final statement is accompanied by a wink.
                            You say you're flattered, but you really don't have time to stick around.
                        [/p]
                        [p]
                            As the both of you finish your drinks, the pleasant chat winds down.
                            The canine rises from her seat, and escorts you to the door,
                            giving you one final gift before letting you go:
                            the remainder of the ${global.kobold.inventory.createItemWithTooltip(new TeaBiscuit)}s,
                            wrapped up in a neat little package.
                            You gratefully accept the present,
                            and with a smile on your face, you bid Cassidy farewell,
                            stepping back out into the cold, harsh, unforgiving world that lies outside of your new friend's luxurious home.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildContinueChoice() {
        var disabled = ctrl.postnotbarn.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarn.has();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildContinueChoice();
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarn.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnTeaAgree";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.notbarn.get() !== NotBarnTeaDoorChoice.AGREE) {
            return "";
        }
        global.ui.applyTheme("hamlet_mansion");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.teaAgreeChange();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.notbarn.get() !== NotBarnTeaDoorChoice.AGREE) {
            return;
        }

        next.registerListeners();

        if (ctrl.postnotbarn.has()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\tea\agree\NotBarnTeaAgree.js

// scenes\HamletDay2\BreedingFarm\barn\tea\disagree\NotBarnTeaDisagreeContinue.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnTeaDisagreeContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnTeaDisagreeContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var int = ctrl.misc.intelligence();
        var cha = ctrl.misc.charisma();
        var toReturn = `[p]
                            Cassidy wishes you a good day,
                            the last thing you see of the vixen being the wide, smug grin plastered across her face as she closes the door behind you.
                            You are left standing in the cold, feeling utterly distraught,
                            doubting the very [i]right[/i] of your entire [b]race[/b] to exist outside of the confines of animal enclosures and cages.
                            Perhaps it [i]would[/i] be a good thing for kobolds to be [i]domesticated[/i] by a more [i]civil[/i] species.
                        [/p]
                        [p]
                            Even [b]you[/b] are, after all, nothing but a [b]thief[/b].
                            A negative influence on society as a whole.
                            A leech.
                            Perhaps you'd be better off being some rich nobleman's [i]pet[/i].
                            At least then you wouldn't have to worry about getting [b]hunted down[/b] by people aiming to collect the money you owe.
                        [/p]
                        [p]
                            With a resigning sigh, you walk back to the center of the hamlet.
                            After pondering the debate for a few more minutes, questioning the merits of your own freedom,
                            you decide to put the entire ordeal behind you, at least for now.
                            If you [b]truly[/b] still feel like you shouldn't be in charge of your own life, for your own sake,
                            you can always start looking for a family in need of a [b]pet[/b] [i]after[/i] you settle your debt.
                            Heck, maybe you'll even find your way back to Cassidy.
                            [i]Maybe[/i].
                        [/p]
                        [p]
                            You decide it's time to head for Varanar.
                            It's already late in the afternoon, and the sun tends to set rather early during the winter.
                            Even this close to the capital of the humans, the roads [b]aren't[/b] safe at night.
                            Especially not for an ownerless pet like yourself.
                        [/p]`;

        if (int >= 6 && cha >= 6) {
            toReturn = `[p]
                            Cassidy wishes you a good day,
                            the last thing you see of the vixen being a confused, pensive look plastered across her face as she closes the door behind you.
                            You are left standing in the cold, feeling equally confused,
                            doubting the very [i]right[/i] of your entire [b]race[/b] to exist outside of the confines of animal enclosures and cages.
                            Still, you are convinced it would [b]not[/b] be a good thing for kobolds to be [i]domesticated[/i] by another species.
                        [/p]
                        [p]
                            With a soft, self-inspiring grunt, you walk back to the center of the hamlet.
                            You ponder the debate for a few more minutes, questioning the merits of your entire kind.
                            The vixen thinks kobolds are nothing more than thieves, bandits and other lowlife scum, but you [b]can[/b] and you [b]will[/b] prove her wrong.
                            Just because kobolds have never achieved greatness before doesn't mean that they –- that [b]you[/b] -- never will!
                        [/p]
                        [p]
                            If you want to become the best kobold you can be, if just to prove Cassidy wrong, then you need to get started as soon as you can.
                            You decide it's time to head for Varanar.
                            It's already late in the afternoon, and the sun tends to set rather early during the winter.
                            Even this close to the capital of the humans, the roads [b]aren't[/b] safe at night.
                            And you doubt bandits care whether or not you're trying to get your life back on track.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnTeaDisagreeContinue";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.postnotbarn.continued()) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.teaDisagreeContinueChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.postnotbarn.continued()) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\tea\disagree\NotBarnTeaDisagreeContinue.js

// scenes\HamletDay2\BreedingFarm\barn\tea\disagree\NotBarnTeaDisagree.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnTeaDisagree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnTeaDisagreeButton";

    var next = new NotBarnTeaDisagreeContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var int = ctrl.misc.intelligence();
        var cha = ctrl.misc.charisma();
        var toReturn = `[p]
                            No!
                            What Cassidy is doing is [b]wrong[/b], no matter how you look at it.
                            Take yourself, for example.
                            If she had her way, would you be in one of those pens right now?!
                            As you struggle to formulate a well-worded answer, the vixen finally manages to open the door to her house.
                            Without replying, you silently follow her inside, if just to give her a piece of your mind.
                        [/p]
                        [p]
                            Cassidy's house is as fancy on the inside as it is on the outside.
                            The entire place is well-lit by candles.
                            Various paintings of kobolds and canines alike line the walls.
                            You stay your tongue as she leads you to the den, where you get to take a seat in a luxurious padded chair.
                            The lack of words between the two of you continues as the vixen speechlessly wanders into the kitchen.
                            She knows you don't agree with her ways.
                            She'd have to be blind [i]not[/i] to be able to read your expression.
                            It is only when she returns with a pair of filled teacups, that you voice your concerns.
                        [/p]
                        [p]
                            The vulpine hands you your tea, and then sits down ever so daintily.
                            She sips her own drink, pinky finger fully extended.
                            She'd almost be lady-like if she weren't half-naked wearing cowboy attire.
                            [i]Almost[/i].
                        [/p]`;

        if (int < 6 && cha < 6) {
            toReturn += `[p]
                            After a long sip on her end, the two of you enter the most [i]philosophical[/i] debate you've [b]ever[/b] been a part of.
                            Well, at least it [i]seems[/i] a bit more intricate than your usual dialogue.
                            But, as a kobold, even that isn't saying much.
                            You have to say, she makes a pretty darn decent case for herself,
                            whereas your only arguments are that what she's doing is [i]unethical[/i], and [i]wrong[/i].
                            But, as she argues, you may just [i]view[/i] it that way because you [b]are[/b] a kobold.
                        [/p]
                        [p]
                            As the both of you finish your drinks, the discussion winds down.
                            The canine rises from her seat, and escorts you to the door.
                            You try to [i]act[/i] as though you still condemn Cassidy's trade,
                            but deep down inside, you know you don't.
                            Not anymore.
                            Or at least [i]not as much[/i].
                            And it [b]shows[/b].
                            Besides, she very clearly [i]won[/i] the argument.
                            You can hate what she does all you'd like,
                            but that won't change the fact that she's undeniably in the right here.
                        [/p]`;
        } else if (int >= 6 && cha < 6) {
            toReturn += `[p]
                            After a long sip on her end, the two of you enter the most [i]philosophical[/i] debate you've [b]ever[/b] been a part of.
                            Well, at least it [i]seems[/i] a bit more intricate than your usual dialogue.
                            But, as a kobold, even that isn't saying much.
                            The vixen makes a pretty compelling case for herself.
                            She brings up many points that you'd never even considered up until now.
                            It seems that she's thought about the ethics of her trade far more than you'd have guessed.
                        [/p]
                        [p]
                            Still, you don't [b]fully[/b] agree with what she's saying.
                            Kobolds may be a [i]detrimental[/i] part of society,
                            but that doesn't make your entire kind any lesser than the outlaws of other species,
                            who [b]still[/b] have basic rights, despite [b]also[/b] not contributing to society as a whole.
                            Sadly, you fail to convey this sentiment properly.
                            In between stuttering, forgetting your train of thoughts halfway through and the vixen constantly [b]twisting[/b] your words,
                            you end up concluding the very opposite of what you had in mind.
                        [/p]
                        [p]
                            As the both of you finish your drinks, the discussion winds down.
                            The canine rises from her seat, and escorts you to the door.
                            You try to [i]act[/i] as though you still condemn Cassidy's trade,
                            but deep down inside, you know you don't.
                            Not anymore.
                            Or at least [i]not as much[/i].
                            And it [b]shows[/b].
                            Besides, she very clearly [i]won[/i] the argument.
                            You can hate what she does all you'd like,
                            but that won't change the fact that she's undeniably in the right here.
                        [/p]`;
        } else if(cha >= 6 && int < 6) {
            toReturn += `[p]
                            After a long sip on her end, the two of you enter the most [i]philosophical[/i] debate you've ever been a part of.
                            Well, at least it [i]seems[/i] a bit more intricate than your usual dialogue.
                            But, as a kobold, even that isn't saying much.
                            You have to say, she makes a pretty darn decent case for herself.
                            Still, you [b]refuse[/b] to let her gain any ground.
                        [/p]
                        [p]
                            You argue that you are the same as her.
                            You can think, you can dream and you can drink tea just as well as the vixen can.
                            Viewing [b]any[/b] sentient being as inherent [b]less[/b] than others is plain [b]wrong[/b].
                            You [b]deserve[/b] the same rights as anyone else!
                        [/p]
                        [p]
                            Cassidy rebutes that, just because two species are capable of reasonable thought and speech, that does [b]not[/b] make them the same.
                            She asks you to consider the dragons, founders and rulers of multiple, successive, continent-spanning empires.
                            "[color=orange]Has any kobold ever done [b]anything[/b] of even remotely equal merit?[/color]"
                        [/p]
                        [p]
                            Of course not.
                            The biggest organized group of kobolds you know is the gang you owe money to.
                            Instead of palaces, your kind prefers to spend their lives in caves, huts and temporary encampments, if they have a place to call home at all.
                            Kobolds would rather skulk through the night like thieves and vagabonds as opposed to holding civil discussions during the day.
                        [/p]
                        [p]
                            The vulpine has made her point, yet still she does not relent.
                            "[color=orange]How many of your kind even know how to read, let alone write?
                            Kobolds are not merely [i]smaller[/i] than other species, they are strictly [b]inferior[/b] in every way imaginable.
                            [b]That[/b] is why [i]they[/i] -- that is why [b]you[/b] -- have no rights.[/color]"
                        [/p]
                        [p]
                            As the both of you finish your drinks, the discussion winds down.
                            The canine rises from her seat, and escorts you to the door.
                            You try to [i]act[/i] as though you still condemn Cassidy's trade,
                            but deep down inside, you know you don't.
                            Not anymore.
                            Or at least [i]not as much[/i].
                            And it [b]shows[/b].
                            Besides, she very clearly [i]won[/i] the argument.
                            You can hate what she does all you'd like,
                            but that won't change the fact that she's undeniably in the right here.
                            Kobolds are inferior to other sentient beings, and you just need to learn to live with that fact.
                        [/p]`;
        } else {
            toReturn += `[p]
                            After a long sip on her end, the two of you enter the most [i]philosophical[/i] debate you've [b]ever[/b] been a part of.
                            Well, at least it [i]seems[/i] a bit more intricate than your usual dialogue.
                            But, as a kobold, even that isn't saying much.
                            The vixen makes a pretty compelling case for herself, a fact which you readily and openly admit.
                            Still, you refuse to give her even an [b]inch[/b].
                            Kobolds may not be as valuable to society as other species, but that doesn't mean they should be treated any worse!
                        [/p]
                        [p]
                            The argument between the two of you goes back and forth.
                            The vixen comes up with a good point why your kind should [b]not[/b] have any rights,
                            you rebuttle with a cleverly constructed comment relating kobolds to criminals of other species,
                            the woman replies that criminals may reform, yet kobolds will always be kobolds,
                            and so the discussion continues for the better part of an hour.
                        [/p]
                        [p]
                            As the both of you finish your drinks, the discussion winds down.
                            The canine rises from her seat, and escorts you to the door.
                            Ultimately, neither of you managed to convince the other.
                            She had a few points for which you had no rebuttal, and you in turn left her speechless more than once.
                            In the end, the two of you agree to disagree.
                            Cassidy gave you a lot to think about, and you gave her a few good reasons to stop doing what she's doing.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildContinueChoice() {
        var disabled = ctrl.postnotbarn.has() ? "disabled" : "";
        var isSelected = ctrl.postnotbarn.has();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildContinueChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postnotbarn.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnTeaDisagree";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.notbarn.get() !== NotBarnTeaDoorChoice.DISAGREE) {
            return "";
        }
        global.ui.applyTheme("hamlet_mansion");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.teaDisagreeChange();
        part += buildChoice();
        part += next.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.notbarn.get() !== NotBarnTeaDoorChoice.DISAGREE) {
            return;
        }

        next.registerListeners();

        if (ctrl.postnotbarn.has()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\tea\disagree\NotBarnTeaDisagree.js

// scenes\HamletDay2\BreedingFarm\barn\tea\NotBarnTeaDecline.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnTeaDecline(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "notBarnTeaDeclineContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You politely decline the invitation, claiming that you have business to attend to.
                            Urgent business.
                            You're obviously lying,
                            but for the sake of [i]not[/i] making this entire ordeal [b]even more[/b] uncomfortable for the both of you,
                            the vixen decides not to ask what sort of a business a [i]kobold[/i] would have that can't stand to wait for an hour more,
                            let alone inquiring about why you stepped foot in her barn in the first place, if you have somewhere to be.
                        [/p]
                        [p]
                            Instead, a somewhat sad frown forms on her face.
                            Like you've not only rejected her as a lover, but as a [b]friend[/b].
                            She closes her eyes, presumably to stop herself from crying.
                            "[color=orange]Very well. Run along.[/color]"
                            With that said, she dismissively waves her hand, turns around and begins to walk away in silence, presumably to hide her sadness.
                        [/p]
                        [p]
                            You feel somewhat [i]guilty[/i] for making Cassidy upset, but at the same time, you never asked for any of this.
                            Internally conflicted, you decide not to stick around for much longer.
                            You return to the front door of the barn, and leave the ancient structure.
                        [/p]
                        [p]
                            A gust of fresh air makes you realize how [i]stale[/i] and [i]musty[/i] the air lingering in the barn was.
                            Like awakening from a drunken haze, your senses sharpen slightly.
                            As you begin to walk back to the center of the hamlet, you ponder the events that transpired.
                            While you [i]were[/i] a bit rude in declining the canine's tea-invitation,
                            surely that was no reason for the woman to burst out into tears, was it?
                            Maybe she was trying to [b]lure you in[/b], to guilt-trip you into humoring her.
                            Then again, she might just have been a nice lady trying to show you some hospitality.
                        [/p]
                        [p]
                            Alas, you will never know.
                            Unsure if you even [b]should[/b] be upset at yourself, you start to make your way towards Varanar,
                            hoping that not [i]everyone[/i] you meet will be as [i]thin-skinned[/i] as the vixen.
                            Your heart wouldn't be able to take upsetting half the city.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnTeaDecline";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.tea.get() !== NotBarnTeaChoice.DECLINE) {
            return "";
        }
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.teaDeclineChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.tea.get() !== NotBarnTeaChoice.DECLINE) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\tea\NotBarnTeaDecline.js

// scenes\HamletDay2\BreedingFarm\barn\tea\NotBarnTeaDoor.js

/**
 * Not barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnTeaDoor(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var disagreeId = "notBarnDisagreeButton";
    var agreeId = "notBarnAgreeButton";
    var attackId = "notBarnAttackButton";

    var toDisagree = new NotBarnTeaDisagree(ctrl);
    var toAgree = new NotBarnTeaAgree(ctrl);
    var toAttack = new NotBarnAttack(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            Well, you [b]have[/b] spent most of your day thus far walking through the blistering cold.
                            You could use a hot beverage, if just to warm yourself up a bit.
                            You nod in agreement.
                            A smile forms on the vulpine's face.
                            But she better not try to seduce you or anything!
                            Her time to roll her eyes.
                        [/p]
                        [p]
                            Cassidy walks past you, beckoning you to follow her.
                            The both of you step outside.
                            You traverse the small dirt path, heading back towards the fancy mansion adjacent to the much older barn.
                            As you walk, you decide to strike up a small conversation.
                            You ask the vixen what [i]exactly[/i] it is she breeds here.
                            You never really [i]did[/i] get a good look at the inside of those pens.
                        [/p]
                        [p]
                            The canine takes a moment to think.
                            She speaks slowly, carefully selecting her words.
                            "[color=orange]Oh, you know.
                            Small, cute things.
                            Some people like keeping them as pets.
                            Others like eating them.
                            A select group of individuals, [small]myself included[/small], even enjoys [i]sleeping[/i] with them.[/color]"
                        [/p]
                        [p]
                            It's as though she doesn't [b]want[/b] to tell you what is hidden in that barn of hers.
                            But that makes you even [b]more[/b] curious.
                            You begin to think.
                            A small, cute pet.
                            Rabbits?
                            People like eating rabbits.
                            And earlier on, you even thought that that might have been it.
                            But people don't have [b]sex[/b] with rabbits, [i]or do they[/i]?
                            You're pretty sure that they don't.
                            And even if they did, you doubt Cassidy would outright tell you she's a [i]bunnyfucker[/i].
                        [/p]
                        [p]
                            As the two of you near the front door, you are just as clueless as you were [i]before[/i] her hints.
                            With an inquisitive look on your face, you repeat your question: what [b]does[/b] she breed?
                            Her smile turns into a nervous grin.
                            You can [i]feel[/i] that she's uncomfortable discussing this topic, but you really want to know.
                            You insist.
                        [/p]
                        [p]
                            The vixen takes a step forward, towards the door, her bare back facing you.
                            "[color=orange]Well...[/color]"
                            she begins, whilst deeply bending over, lifting her tail up high to give you a clear look at her furred rump,
                            the fertile loins placed in between her thick thighs distracting you more than a little, 
                            almost as if she's trying to trick you, to prevent you from focussing on her reply.
                            While your eyes are glued to to the vulpine's groin, sheer inquisitiveness causes your attention to remain on-point.
                            Finally, the answer is revealed to you.
                            "[color=orange]...[small]kobolds[/small]. I breed kobolds.[/color]"
                        [/p]
                        [p]
                            What?
                            Kobolds?
                            Your kind?
                            Wait.
                            She said she bred things for fucking, for selling them as pets, and for...[b]eating[/b]?!
                            Reaching into her boot, Cassidy produces a small metal object from her footwear.
                            The key to her house.
                            Looking back at you, she notices how upset you are.
                            She realizes her [i]distraction[/i] must not have worked.
                        [/p]
                        [p]
                            The canine sighs loudly.
                            Stepping towards the door, she begins to ratify her actions as she fiddles with the lock, her tone changing considerably.
                            "[color=orange]It's a better life for [b]most[/b] of them.
                            Without me, they'd end up on the streets, or as bandits in the forest.
                            Sure, some of the young ones are marked for death, but the rest of the kobolds under my care are treated fairly.
                            They haven't a care in the world, and some of them may even end up being [i]adopted[/i].
                            Surely you can see the logic in my actions, right, ${name}?[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildDisagreeChoice() {
        var disabled = ctrl.notbarn.has() ? "disabled" : "";
        var isSelected = ctrl.notbarn.get() === NotBarnTeaDoorChoice.DISAGREE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${disagreeId}" ${disabled}>Disagree</button>`;

        var tooltip = global.parser.parse(`What?! No! Your kind is [b]more[/b] than just [b]cattle[/b]!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildAgreeChoice() {
        var disabled = ctrl.notbarn.has() ? "disabled" : "";
        var isSelected = ctrl.notbarn.get() === NotBarnTeaDoorChoice.AGREE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${agreeId}" ${disabled}>Agree</button>`;
        var tooltip = global.parser.parse(`Constant sex, free food and nothing to worry about? Sounds like she's doing them a favor!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildAttackChoice() {
        var weapon = ctrl.misc.weapon();
        var disabled = ctrl.notbarn.has() || !weapon ? "disabled" : "";
        var isSelected = ctrl.notbarn.get() === NotBarnTeaDoorChoice.ATTACK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${attackId}" ${disabled}>Attack her</button>`;

        var tooltip;
        if (weapon) {
            tooltip = global.parser.parse(`How [b]dare[/b] she keep your kind as [b]sex slaves[/b]?! Time to teach this bitch a lesson she won't soon forget.`);
        } else {
            tooltip = global.parser.parse(`You don't have any weapons to attack the woman with.`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildDisagreeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAgreeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAttackChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + disagreeId), $("#" + agreeId), $("#" + attackId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerDisagreeListener() {
        $("#" + disagreeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.notbarn.set(NotBarnTeaDoorChoice.DISAGREE);
            loadNextPart(toDisagree);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerAgreeListener() {
        $("#" + agreeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.notbarn.set(NotBarnTeaDoorChoice.AGREE);
            loadNextPart(toAgree);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerAttackListener() {
        $("#" + attackId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.notbarn.set(NotBarnTeaDoorChoice.ATTACK);
            loadNextPart(toAttack);
        });
    }
    

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnTeaDoor";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if(ctrl.tea.get() !== NotBarnTeaChoice.ACCEPT) {
            return "";
        }
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toDisagree.build();
        part += toAgree.build();
        part += toAttack.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.tea.get() !== NotBarnTeaChoice.ACCEPT) {
            return;
        }
        
        toDisagree.registerListeners();
        toAgree.registerListeners();
        toAttack.registerListeners();

        if (ctrl.notbarn.has()) {
            return;
        }

        registerDisagreeListener();
        registerAgreeListener();
        registerAttackListener();
    };
}

var NotBarnTeaDoorChoice = {
    DISAGREE: 4,
    AGREE: 5,
    ATTACK: 6
};

// END scenes\HamletDay2\BreedingFarm\barn\tea\NotBarnTeaDoor.js

// scenes\HamletDay2\BreedingFarm\barn\tea\NotBarnTea.js

/**
 * Not barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function NotBarnTea(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var declineId = "notBarnTeaDeclineButton";
    var acceptId = "notBarnTeaAcceptButton";

    var toDecline = new NotBarnTeaDecline(ctrl);
    var toAccept = new NotBarnTeaDoor(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You slowly shake your head no, softly stammering out that you aren't really interested in Cassidy's [i]proposal[/i].
                            You didn't come here to have sex, let alone with someone you've never met before.
                            A confused look forms on the vixen's face.
                            As though she doesn't quite understand what you just said.
                            She visibly attempts to form a coherent reply, yet no words escape her lips.
                            You guess she's not used to being rejected.
                        [/p]
                        [p]
                            "[color=orange]Are you sure you understood me correctly?[/color]" the vulpine asks, after a moment of painfully awkward silence.
                            You reply with a single, swift nod, full of conviction.
                            "[color=orange]So you're saying you [b]don't[/b] want to have sex with me? Right now? With no strings attached?[/color]"
                            Another nod.
                            It feels like you're finally starting to get through to her.
                            "[color=orange]Are you [b]not[/b] [i]into[/i] women?[/color]"
                            You roll your eyes and face away from her, refusing to even acknowledge, let alone [i]answer[/i] her question.
                        [/p]
                        [p]
                            The vixen lets out a loud sigh.
                            She removes her hand from your shoulder, and straightens her back once more, her face no longer mere inches away from yours.
                            Instead of her sultry look, Cassidy puts on a more [i]dignified[/i] visage, and clears her throat, before speaking once more:
                            "[color=orange]I [i]do[/i] apologize.
                            Usually, when one of your kind wanders in here due to sheer curiosity, they [b]jump[/b] at the opportunity to [i]have some fun[/i].
                            But I now see that [i]some[/i] of you aren't [b]completely[/b] driven by desire.[/color]"
                        [/p]
                        [p]
                            Her odd explanation only serves to raise more questions than it provides answers.
                            She does this thing all the time?
                            Why?
                            Does she [i]love[/i] kobolds [i]that much[/i]?
                            For a moment, you begin to doubt yourself.
                            Perhaps you shouldn't have been so suspicious.
                            Maybe she really [i]did[/i] just want to have sex with you.
                            Alas, it seems that ship has sailed.
                            Judging by her posture, you've thoroughly [b]soured[/b] the canine's excited mood.
                        [/p]
                        [p]
                            "[color=orange]Well then, regardless of whether or not you're in the mood to [i]indulge[/i] my need, you [i]are[/i] still my guest.
                            So, little one, how about we leave this dusty old barn and head indoors?
                            I'll make you a nice cup of tea, and we can chat for a bit.
                            If I can't use my body to warm you up, the least I can do is offer you some shelter from the coldness of the world, if only for a while.[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildDeclineChoice() {
        var disabled = ctrl.tea.has() ? "disabled" : "";
        var isSelected = ctrl.tea.get() === NotBarnTeaChoice.DECLINE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${declineId}" ${disabled}>Decline</button>`;
        var tooltip = global.parser.parse(`You'd rather not take the risk of entering a nymphomaniac's home,
                                           especially after shooting down her initial request.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 220px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildAcceptChoice() {
        var disabled = ctrl.tea.has() ? "disabled" : "";
        var isSelected = ctrl.tea.get() === NotBarnTeaChoice.ACCEPT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${acceptId}" ${disabled}>Accept</button>`;
        var tooltip = global.parser.parse(`Sure. You could go for a cup of tea. But [b]no[/b] funny business!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildDeclineChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAcceptChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + declineId), $("#" + acceptId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerDeclineListener() {
        $("#" + declineId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tea.set(NotBarnTeaChoice.DECLINE);
            loadNextPart(toDecline);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerAcceptListener() {
        $("#" + acceptId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tea.set(NotBarnTeaChoice.ACCEPT);
            loadNextPart(toAccept);
        });
    }


    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "notBarnTea";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.barn.get() !== BreedingBarnChoice.DECLINE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toDecline.build();
        part += toAccept.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.barn.get() !== BreedingBarnChoice.DECLINE) {
            return;
        }

        toDecline.registerListeners();
        toAccept.registerListeners();

        if (ctrl.tea.has()) {
            return;
        }

        registerDeclineListener();
        registerAcceptListener();
    };
}

var NotBarnTeaChoice = {
    DECLINE: 1,
    ACCEPT: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\tea\NotBarnTea.js

// scenes\HamletDay2\BreedingFarm\barn\eager\unaware\run\BarnRun.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnRun(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "barnRunContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide not to push your luck any further.
                            You'd rather not risk wasting away in one of these animal pens, or perhaps even getting flayed alive, just so you can get some revenge.
                            Looking around for avenues of escape, you notice the barn has a small back door, the large iron lock of which is clearly open.
                            Without alerting Cassidy, you creep towards it.
                        [/p]
                        [p]
                            The moment you reach the back exit, the vixen rises from her crouched searching pose,
                            ${global.kobold.inventory.createItemWithTooltip(new FemPlus())} in hand.
                            She turns around, just in time to witness you [b]slamming[/b] the door shut behind you.
                            You take in a deep breath of fresh air, the sight of the afternoon sky inspiring a sense of safety inside of you.
                            Like you've made it out.
                            Like you're a free kobold once more.
                            But you're not out of trouble yet.
                            For all you know, that crazy [b]bitch[/b] could be rushing after you this very instant!
                        [/p]
                        [p]
                            While Cassidy may be twice your size, you're confident you can at least [b]outrun[/b] the canine.
                            Especially if she's still wearing those cowboy boots of hers.
                            Knocking over a nearby barrel to somewhat [i]block[/i] the door you just came from,
                            you begin to mindlessly sprint away, in the general direction of Varanar.
                            You aren't sure whether or not she's in pursuit, but you'd rather not take any chances.
                        [/p]
                        [p]
                            After about half an hour of intensive running through fields, farms and small patches of woodland,
                            you think you've finally escaped the wicked woman fully, if she was even on your tail to begin with.
                            Better safe than sorry.
                            You find yourself back on the eastern road to Varanar, just a few miles away from the city limits.
                            You are exhausted.
                            [b]Very[/b] exhausted.
                            You lean against a lone nearby tree to catch your breath.
                        [/p]
                        [p]
                            You keep replaying the barn-scenario in your head.
                            It was a close call.
                            [b]Too[/b] close.
                            If you hadn't figured out a way to escape that table, who knows what Cassidy would have done to you?!
                            You might have ended up spending your entire life getting fucked nonstop by a bunch of your drug-infused peers.
                        [/p]
                        [p]
                            You take a moment to lament the other kobolds imprisoned by the vixen.
                            Sadly, there isn't much you can do for them.
                            What Cassidy is doing is completely [b]legal[/b], despite being abhorrent, undeniably [b]evil[/b] and totally sickening.
                            Disheartened, you let out a loud sigh.
                            This world is unfair, especially for kobolds.
                            At least [b]you[/b] got out of there alive, and that's what really matters.
                        [/p]
                        [p]
                            Having recuperated enough to continue your journey, you begin to walk down the road towards Varanar.
                            At least there aren't any kobold breeding farms there.
                            Or so you hope.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnRun";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unaware.get() !== BarnUnawareChoice.RUN) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unawareRun.change();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unaware.get() !== BarnUnawareChoice.RUN) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\unaware\run\BarnRun.js

// scenes\HamletDay2\BreedingFarm\barn\eager\unaware\force\free\BarnForceFree.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnForceFree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "barnForceFreeContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.cassidy.koboldsfreed = true;
        var toReturn = `[p]
                            With the vixen restrained, you decide to save your fellow kobolds from a lifetime of mindless breeding.
                            You walk up to the nearest enclosure, and deftly lift the heavy latch holding the large wooden door shut.
                            Swinging open the animal pen, you are greeted with a terrible sight.
                        [/p]
                        [p]
                            A group of seven kobolds inhabits this particular [i]cell[/i], some of which are chained to the wooden walls.
                            All of them are naked.
                            In one corner lies a female, with tits so bloated you doubt she can move on her own.
                            A single tiny whelp, who can't be more than a couple of months old, is suckling on one of her lactating teats.
                        [/p]
                        [p]
                            Next to the nursing mother, an unchained male is idly humping a restrained, gagged and blindfolded kobold.
                            You can't quite discern the gender of the one on the receiving end,
                            but regardless of whether the hole being fucked is a cockslit or a cunt,
                            it's completely [b]overflowing[/b] with spunk.
                            Judging by the veritable [b]pool[/b] of jizz that has accumulated on the soggy ground below them,
                            you reckon those two must have been at it for [b]days[/b] on end.
                        [/p]
                        [p]
                            Further to the right, another [i]stallion[/i] is standing up, leaning against the wall, fervently stroking his oversized erection.
                            His very [b]existence[/b] seems to taunt a sixth of your kin, one of the females.
                            This one is a bit more [i]normally-proportioned[/i], apart from her swollen belly.
                            You presume she's pregnant, as opposed to just being [i]fat[/i].
                            Her eyes are glued to the masturbating male's groin, and the reason why is very clear.
                            Her wrists and her tail are hogtied, preventing her from satisfying the [i]painfully obvious[/i] need between her legs.
                            She squirms her lower body about, desperately trying but failing to pleasure herself.
                            She must have gotten injected with some kind of aphrodisiac, and a strong one at that.
                        [/p]
                        [p]
                            To your left, you spot what you believe to be the most recent [i]acquisition[/i] of the bunch.
                            A young woman, who you judge to be about your age, sits chained up to the wall, mindlessly staring off to the side.
                            Her loins have clearly been [b]violated[/b], but apart from a few cum-spatters staining her scales, she appears to be in rather good shape.
                            Slowly, the kobold shifts her gaze towards you, which is the most [i]recognition[/i] you've gotten from any of the captives thus far.
                        [/p]
                        [p]
                            At first, she looks uninterested, probably thinking you're yet [b]another[/b] poor sap that got tricked by Cassidy.
                            Then, you notice a [i]spark[/i], a certain [i]twinkle[/i] of consciousness, of sheer self-awareness in her otherwise jaded eyes.
                            Something the other kobolds in the pen lack.
                        [/p]
                        [p]
                            The realization that the vixen isn't around seems to inspire hope in the otherwise completely demoralized girl.
                            She tries to speak, to address you, to [b]beg[/b] you for help, yet her mouth is gagged by an old, filthy rag,
                            causing her desperate pleas to come out as pitiful whimpers and groaning moans.
                        [/p]
                        [p]
                            You quickly approach the distressed captive.
                            Kneeling beside her, you pull the disgusting piece of clothing serving as a silencing muzzle from her mouth.
                            She immediately begins to [b]beg[/b] you [i]not[/i] to fuck her.
                            Poor thing must have been through a lot.
                            You reassure her that you're not there to hurt her.
                            You just want to help.
                        [/p]
                        [p]
                            To showcase your good intentions, you unfasten the cuffs tying her to the wall.
                            Same mechanism as the table.
                            Easy to remove from the outside.
                            The moment her arms are freed from their restraints, she curls up into a fetal position on the ground,
                            and slowly begins to rock back and forth, softly muttering to herself before openly bursting out into tears.
                        [/p]
                        [p]
                            You decide to give the kobold a few minutes to herself.
                            You step out of the enclosure, to check up on Cassidy.
                            She hurls a myriad of insults your way, and threatens to [b]kill[/b] you if you don't stop toucing her [b]pets[/b].
                            Growing sick of her loud mouth, you crudely [b]shove[/b] the filthy piece of cloth you found earlier into her muzzle.
                            She tries, but fails to bite you, ending up only slightly tearing the rag instead, dust and bits of fabric clinging to her gums.
                        [/p]
                        [p]
                            With the bitch silenced, you return to the enclosure.
                            The heart-wrenching sobbing coming from the open pen has seemingly come to a halt
                            You head back inside, and once more, you take a knee next to the pink-scaled captive.
                            You place your hand atop her head, and comfortingly pat it, ever so gently.
                            Her crying has stopped, although even the girl's heavy, unsteady breathing is permeated with pure [b]sadness[/b].
                        [/p]
                        [p]
                            You cautiously coax the traumatized kobold into sitting up straight.
                            You don't dare to let her to rest her back against the wall, lest it trigger some kind of flashback of being held captive.
                            After she's calmed down a bit more, you ask the woman for her name.
                            No response.
                            She continues to stare blankly ahead of her.
                            You ask if she'd like to leave this place.
                            A single nod, weak but solemn.
                        [/p]
                        [p]
                            You say there's a nice house just next door that she can go to.
                            But first, she needs to help you break out the rest of the kobolds, so they can all get out of this hell-hole together.
                            She tilts her visage your way.
                            She visibly tries to respond to your request, yet no words leave her mouth.
                            Instead, another slight, affirmative shaking of her head.
                        [/p]
                        [p]
                            You help the girl to her feet.
                            At first, she can barely stand, let alone walk, yet she quickly regains her leg-strength.
                            It takes just a few minutes of you supporting her before she can stand on her own once more.
                            Once you lead her outside of the containment pen, the sight of the restrained vixen coming into view,
                            her spirits seem to be lifted further.
                            The two of you share a mutual nod, and you jointly set to work opening the remainder of the enclosures,
                            liberating the stables from vulpine oppression.
                        [/p]
                        [p]
                            In each pen, the situation is roughly the same.
                            There's a bunch of drugged up males, some thoroughly violated,
                            often pregnant, mostly mind-broken kobold women, a couple of which used to be [i]men[/i],
                            and finally there's one or two more relatively unscathed, yet completely immobilized [i]newcomers[/i], who retain much of their [i]sanity[/i].
                            You focus on this final group, setting them free like you did your new,
                            pink-scaled friend, and getting them to help in the ongoing effort to [i]get through[/i] to the [i]less fortunate[/i] members of the group.
                        [/p]
                        [p]
                            Eventually, you manage to get a somewhat functional form of [i]coordination[/i] going on.
                            The more mentally capable kobolds herd the drug-addled and physically debilitated ones out of the pens.
                            Most of them can't stop touching themselves, so [b]afflicted[/b] by various potions and drugs that they'll likely take days,
                            if not weeks to physically recover from this entire endeavour, let alone mentally.
                        [/p]
                        [p]
                            The entire group gets corralled through the barn's single side-door, leading into the adjacent house,
                            and hopefully towards a better future than what they'd have had in here.
                            One of the last few stragglers to remain, diligently ensuring that each and every one of the once-prisoners finds his or her way inside,
                            is the pink-scaled girl you rescued first.
                        [/p]
                        [p]
                            Ultimately, you and her end up being the only two left in the barn,
                            apart from a small amount of males which have taken to abusing their former mistress.
                            No words are spoken between the two of you, but you can [i]sense[/i] her gratitude,
                            in the same way that she realizes that you can't help but make this [b]a lot[/b] more awkward than it should have been.
                            A small, heart-warming smile forms on her face.
                            With a loud, relieved sigh and a deepfully grateful nod, she silently turns around, and walks away to join the others inside of the house.
                        [/p]
                        [p]
                            You wish you could stay, but alas, your presence here would only cause trouble for the kobolds.
                            Besides, you have business to attend to in Varanar, and you'd only get in the way of their healing process.
                            You're certain they'll be fine.
                            They're in at least one set of good hands.
                            With the knowledge that you did the right thing here today,
                            you quietly leave the barn through the back door,
                            not as the thief you entered as,
                            but as an unsung hero of your kind.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnForceFree";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.barnForce.get() !== BarnForceChoice.FREE) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.barnForceFree.change();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.barnForce.get() !== BarnForceChoice.FREE) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\unaware\force\free\BarnForceFree.js

// scenes\HamletDay2\BreedingFarm\barn\eager\unaware\force\fuck\BarnForceFuck.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnForceFuck(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "barnForceFuckContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You walk over to the bag Cassidy was searching through.
                            You find the ${global.kobold.inventory.createItemWithTooltip(new FemPlus())} she dropped as you scooped her off the ground.
                            Luckily, it didn't get damaged by the fall.
                            While this isn't what you were looking for, you pocket it regardless.
                        [/p]
                        [p]
                            The vulpine has already cleared out most of the [i]empty[/i] flasks.
                            It takes you a mere three blind grabs to find something other than FemPlus.
                            Not a potion, but [b]lotion[/b].
                            Something to sprinkle atop skin so you don't have to bother with getting her to swallow it,
                            or finding something to inject her with.
                            A full vial of ${global.kobold.inventory.createItemWithTooltip(new aPlusLotion())}.
                        [/p]
                        [p]
                            With the bottle in your hand and a mischievous smile on your face, you return to the breeding table.
                            Cassidy snaps her jaws at you, still attempting to act tough.
                            It's actually somewhat [i]endearing[/i] to see her like this,
                            not as a crazy [b]bitch[/b] that wants to turn you into some kind of baby factory,
                            but as a meek little vixen trying to keep up appearances despite being completely at your mercy.
                        [/p]
                        [p]
                            You pop the lid of the vial, drawing the woman's attention to it.
                            "[color=orange]Don't touch that, you stupid [b]lizard[/b]!
                            Do you even know what that is?![/color]"
                            With your free hand, you pick up the rag you spotted before off the ground,
                            threateningly waving it around, which suffices to shut her up.
                        [/p]
                        [p]
                            Holding the opened bottle as high above the canine's bare, restrained body as your arm can reach,
                            you slightly tip it over, the viscous oil within oozing out, slowly seeping down onto Cassidy's body in a single, elongated strand.
                            You spill the liquid all over her body.
                            Her firm breasts, her soft belly, her upper thighs, and an extra dosage in between her legs,
                            ensuring her groin is [b]bathing[/b] in potent aphrodisiac.
                            The final few droplets, you sprinkle onto her neck and her face, an act which she doesn't seem to appreciate.
                        [/p]
                        [p]
                            Normally, this oil is supposed to be [b]rubbed[/b] into someone's skin, but you don't really want to get any of it on your hands.
                            Besides, even without you applying it more thoroughly, the lotion already seems to be taking effect.
                            You watch as Cassidy's visage shifts from an angry glare to a flustered, needy look.
                            Her nipples visibly harden, and the intense scent of a bitch in heat wafts from her loins.
                            While she does her best to resist the effects, to remain stoic and unyielding, it takes less than a minute for her masquerade to [b]break[/b].
                        [/p]
                        [p]
                            The vixen struggles against her restraints,
                            desperately wriggling her thighs about as much as her shackles will allow,
                            in an attempt to cool the fire rising from the stoked furnace in between her legs.
                            She swings her chest from side to side, rolling around as best she can,
                            her tits slightly rubbing against one another, providing not [b]nearly[/b] enough pleasure to get her off,
                            but still more than [b]nothing[/b].
                        [/p]
                        [p]
                            You stare at her, your cruel smirk only adding to her agony.
                            But you've only just [b]begun[/b] toying with her.
                            You rest your hand atop the woman's thigh, her writhing instantly coming to an end as you grab the attention of every fiber of her being.
                            She pleadingly looks at you, panting, her tongue lolling out of her mouth,
                            her cheeks reddened to the point where you can see her blush through her dense fur.
                        [/p]
                        [p]
                            You give the soft flesh of her thick upper leg a single, firm squeeze,
                            a needy moan escaping Cassidy's lips, showcasing the extent of how utterly [b]sensitive[/b] the lotion has made her.
                            You move your hand upwards, trailing it along her fur, ruffling it all the way from her hips to her chest.
                            With two of your fingers, you sharply [b]pinch[/b] one of her rock-hard nipples,
                            eliciting a far louder moan this time, the vulpine pushing her chest outwards in a futile attempt to rub her tit against your hand.
                        [/p]
                        [p]
                            You deny her that pleasure.
                            She hasn't [b]earned[/b] it.
                            Instead, you trace your hand back down her belly, as though patting her supersensitive skin.
                            As you reach her groin, you remove your hand from her body once more,
                            temptingly hovering the appendage mere inches away from her loins, teasing [i]without[/i] touching.
                        [/p]
                        [p]
                            "[color=orange]P-please...[/color]" the vixen stammers, the most sense she's made since you poured the aphrodisiac onto her.
                            You tell her to hush.
                            You ask her if what she did was wrong.
                            "[color=orange]Yes! I shouldn't have forced you onto the [b]goddamn[/b] table! Now [b]please[/b] just [b]fucking[/b]...[/color]" she replies,
                            anger, frustration and sheer [b]need[/b] heavy on her voice.
                            Exactly what you wanted to hear, or well, at least the first part.
                        [/p]
                        [p]
                            You hush her again, threatening to leave her like this unless she fixes that tone of hers.
                            Her menacing growling reaches a new level of volume,
                            yet it ceases altogether the moment you press those same two fingers you used to pinch her nipple against her labia,
                            spreading her folds, exposing the very core of the heat overtaking her body to the chilly air lingering in the barn.
                        [/p]
                        [p]
                            You decide to toy with your vulpine prey a bit.
                            You ask her who a good puppy is.
                            She glares at you angrily.
                            Brushing your thumb over her clit, sending a [b]surge[/b] of pleasure coursing throughout her body, seems to [i]loosen her lips[/i].
                            You re-state your question.
                            Succumbing to your will, she begrudgingly replies: "[color=orange][b]Me![/b][/color]"
                            She, what?
                            "[color=orange]I'm a good puppy.[/color]"
                        [/p]
                        [p]
                            You burst out into sadistic laughter.
                            Cassidy fails to see the humor in your actions.
                            She shifts her hips around uncomfortably, trying to grind her groin against your fingers.
                            You plan on giving her what she needs.
                            But not yet.
                            No, she still has one thing left to do before you inguldge her.
                        [/p]
                        [p]
                            The order you give her is simple: bark.
                            Bark like the [b]bitch[/b] she is.
                            You can see her hesitation.
                            Her reluctance to humiliate herself.
                            It fuels you, tempering your resolve [b]not[/b] to pleasure her until she relents.
                            You stare at her, dead-serious.
                            She glares back at you, barely able to keep a straight face.
                            "[color=orange][small]Woof.[/small][/color]"
                            You say you didn't hear that.
                            What did she say?
                            "[color=orange]Woof.[/color]" she repeats, louder this time.
                            Again, you command.
                            "[color=orange]Woof! [b]Do it already!![/b][/color]"
                        [/p]
                        [p]
                            Dogs can't talk, you mention.
                            You tell her to bark one more time.
                            As loud as she can.
                            "[color=orange][b]WOO---[/b][/color]"
                            You interrupt her shout by plunging both of your fingers into her [b]sopping-wet[/b] cunt,
                            plugging her leaking hole, scratching a burning itch deep inside of her.
                        [/p]
                        [p]
                            Her pretend-bark turns into a [b]scream[/b] of pleasure.
                            Cassidy wildly bucks her hips back against your hand, riding your two digits as best as her restraints allow,
                            a mind-wracking orgasm rocking her body hard.
                            A [b]flood[/b] of femine fluids squirt from her loins, completely [b]soaking[/b] the palm of your hand.
                        [/p]
                        [p]
                            So intense is the woman's climax, that her mind goes [b]completely[/b] blank.
                            Her tongue lolls obscenely from her gaping muzzle, her body limply laying atop the table, her groin still twitching.
                            By the time you remove your fingers from her warm snatch, the vixen has resorted to softly repeating the same word over and over again,
                            in hopes you'd see to further pleasuring her:
                            "[color=orange][small]Woof...woof...woof...[/small][/color]"
                        [/p]
                        [p]
                            You sloppily wipe your fingers off on Cassidy's snout, her tongue meekly lapping at your palm, cleaning up the rest of her mess.
                            With the canine thoroughly [b]broken[/b], you decide to take your leave.
                            You're certain she can find her own way out of her predicament after the lotions wears off.
                            After all, [b]you[/b] managed to escape the table, so why wouldn't she be able to?
                        [/p]
                        [p]
                            You exit the barn through the back door, and continue on your way to Varanar.
                            It is late in the afternoon, and you'll have to walk at a brisk pace if you want to make it there come nightfall.
                            Maybe, once there, you can pay some prostitute to let you do to her what you did to Cassidy.
                            And if not, you might just have to return to the breeding farm sometimes, to give your little puppy a treat.
                            Whether she likes it, or not.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnForceFuck";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.barnForce.get() !== BarnForceChoice.FUCK) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.barnForceFuck.change();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.barnForce.get() !== BarnForceChoice.FUCK) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\unaware\force\fuck\BarnForceFuck.js

// scenes\HamletDay2\BreedingFarm\barn\eager\unaware\force\BarnForce.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnForce(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var fuckId = "barnForceFuckButton";
    var freeId = "barnForceFreeButton";

    var toFree = new BarnForceFree(ctrl);
    var toFuck = new BarnForceFuck(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You stealthily approach the vulpine.
                            Being a fox, she may have incredible hearing and a sense of smell well beyond your own,
                            but you are a [b]kobold[/b], one of the sneakiest races around.
                            That, and she's far too occupied with searching through an entire bag full of noisy rubbish to notice you drawing close.
                        [/p]
                        [p]
                            "[color=orange]Aha![/color]" the vixen exclaims, rising to her feet with a single filled vial.
                            Spotting the perfect opportunity to strike, you rush towards her.
                            Embracing her in a powerful [b]bear-hug[/b] from behind,
                            you showcase your incredible strength and [b]lift[/b] Cassidy off the ground.
                        [/p]
                        [p]
                            She yelps in surprise, immediately starting to struggle, kick and squirm, yet your grip is unrelenting.
                            Stepping backwards, you slowly carry the much larger woman all the way to the breeding table.
                            She manages to scratch you a few times using the spurs on the back of her boots,
                            but she fails to land any significant blows.
                            Once close enough to the table, you [b]throw[/b] her on top of it,
                            rapidly flipping the switch to mechanically pin her in place before she even realizes what's going on.
                        [/p]
                        [p]
                            "[color=orange]What are you doing?! Let me [b]go[/b]![/color]" she demands, screaming so loud it almost gives you a headache.
                            You tell her to shut up, threatening to shove an old, discarded rag you spot on the floor into her muzzle, lest she be quiet.
                            Her angry shouts fade into bestial growling,
                            the canine baring her teeth, attempting to act [b]tough[/b] and [b]dangerous[/b] while being completely restrained.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFuckChoice() {
        var exertion = ctrl.misc.exertion();
        var disabled = ctrl.barnForce.has() ? "disabled" : "";
        var isSelected = ctrl.barnForce.get() === BarnForceChoice.FUCK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fuckId}" ${disabled} style="width: 200px;">Torment the bitch</button>`;
        var tooltip = global.parser.parse(`She thought she could have [b]you[/b] raped?
                                           Time to show Cassidy who's the boss around these parts.
                                           Pump her full of aphrodisiac, toy with her for a bit, then leave her on the table to writhe in the mess she created.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 200px;">`;
        button += popupContainer;
        button += `</div>`;
        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFreeChoice() {
        var disabled = ctrl.barnForce.has() ? "disabled" : "";
        var isSelected = ctrl.barnForce.get() === BarnForceChoice.FREE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${freeId}" ${disabled} style="width:200px;">Free your peers</button>`;
        var tooltip = global.parser.parse(`Open the cubicles and free the other kobolds. Surely some of them must still be able-minded, right?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildFuckChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFreeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + fuckId), $("#" + freeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFuckListener() {
        $("#" + fuckId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.barnForce.set(BarnForceChoice.FUCK);
            loadNextPart(toFuck);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFreeListener() {
        $("#" + freeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.barnForce.set(BarnForceChoice.FREE);
            loadNextPart(toFree);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnForce";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unaware.get() !== BarnUnawareChoice.FORCE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.barnForce.change();
        part += buildChoice();
        part += toFree.build();
        part += toFuck.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unaware.get() !== BarnUnawareChoice.FORCE) {
            return;
        }

        toFree.registerListeners();
        toFuck.registerListeners();

        if (ctrl.barnForce.has()) {
            return;
        }

        registerFuckListener();
        registerFreeListener();
    };
}

var BarnForceChoice = {
    FUCK: 1,
    FREE: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\eager\unaware\force\BarnForce.js

// scenes\HamletDay2\BreedingFarm\barn\eager\unaware\attack\free\BarnAttackFree.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnAttackFree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "barnAttackFreeContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.cassidy.koboldsfreed = true;
        var dead = global.kobold.history.shared.cassidy.dead;
        var toReturn = `[p]
                            With the vixen ${dead ? "" : "temporarily"} out of the equation, you decide to save your fellow kobolds from a lifetime of mindless breeding.
                            You walk up to the nearest enclosure, and deftly lift the heavy latch holding the large wooden door shut.
                            Swinging open the animal pen, you are greeted with a terrible sight.
                        [/p]
                        [p]
                            A group of seven kobolds inhabits this particular [i]cell[/i], some of which are chained to the wooden walls.
                            All of them are naked.
                            In one corner lies a female, with tits so bloated you doubt she can move on her own.
                            A single tiny whelp, who can't be more than a couple of months old, is suckling on one of her lactating teats.
                        [/p]
                        [p]
                            Next to the nursing mother, an unchained male is idly humping a restrained, gagged and blindfolded kobold.
                            You can't quite discern the gender of the one on the receiving end,
                            but regardless of whether the hole being fucked is a cockslit or a cunt,
                            it's completely [b]overflowing[/b] with spunk.
                            Judging by the veritable [b]pool[/b] of jizz that has accumulated on the soggy ground below them,
                            you reckon those two must have been at it for [b]days[/b] on end.
                        [/p]
                        [p]
                            Further to the right, another [i]stallion[/i] is standing up, leaning against the wall, fervently stroking his oversized erection.
                            His very [b]existence[/b] seems to taunt a sixth of your kin, one of the females.
                            This one is a bit more [i]normally-proportioned[/i], apart from her swollen belly.
                            You presume she's pregnant, as opposed to just being [i]fat[/i].
                            Her eyes are glued to the masturbating male's groin, and the reason why is very clear.
                            Her wrists and her tail are hogtied, preventing her from satisfying the [i]painfully obvious[/i] need between her legs.
                            She squirms her lower body about, desperately trying but failing to pleasure herself.
                            She must have gotten injected with some kind of aphrodisiac, and a strong one at that.
                        [/p]
                        [p]
                            To your left, you spot what you believe to be the most recent [i]acquisition[/i] of the bunch.
                            A young woman, who you judge to be about your age, sits chained up to the wall, mindlessly staring off to the side.
                            Her loins have clearly been [b]violated[/b], but apart from a few cum-spatters staining her scales, she appears to be in rather good shape.
                            Slowly, the kobold shifts her gaze towards you, which is the most [i]recognition[/i] you've gotten from any of the captives thus far.
                        [/p]
                        [p]
                            At first, she looks uninterested, probably thinking you're yet [b]another[/b] poor sap that got tricked by Cassidy.
                            Then, you notice a [i]spark[/i], a certain [i]twinkle[/i] of consciousness, of sheer self-awareness in her otherwise jaded eyes.
                            Something the other kobolds in the pen lack.
                        [/p]
                        [p]
                            The realization that the vixen isn't around seems to inspire hope in the otherwise completely demoralized girl.
                            She tries to speak, to address you, to [b]beg[/b] you for help, yet her mouth is gagged by an old, filthy rag,
                            causing her desperate pleas to come out as pitiful whimpers and groaning moans.
                        [/p]
                        [p]
                            You quickly approach the distressed captive.
                            Kneeling beside her, you pull the disgusting piece of clothing serving as a silencing muzzle from her mouth.
                            She immediately begins to [b]beg[/b] you [i]not[/i] to fuck her.
                            Poor thing must have been through a lot.
                            You reassure her that you're not there to hurt her.
                            You just want to help.
                        [/p]
                        [p]
                            To showcase your good intentions, you unfasten the cuffs tying her to the wall.
                            Same mechanism as the table.
                            Easy to remove from the outside.
                            The moment her arms are freed from their restraints, she curls up into a fetal position on the ground,
                            and slowly begins to rock back and forth, softly muttering to herself before openly bursting out into tears.
                        [/p]`;
        if (dead) {
            toReturn += `[p]
                            You decide to give the kobold a few minutes to herself.
                            You step out of the enclosure, to ponder the entire ordeal.
                            Had you not managed to free yourself from the breeding table, you could have very well ended up in one of these pens yourself,
                            drugged up to the point where you can no longer [b]think[/b] of anything but getting fucked,
                            pumping out litter after litter of younglings to be sold into slavery, or perhaps even [b]butchered[/b] for their meat.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You decide to give the kobold a few minutes to herself.
                            You step out of the enclosure, to check up on the knocked out bitch.
                            She's still down, and out for the count.
                            Using a handily placed length of rope, you tie her limbs together to ensure she won't be causing any trouble when she finally [i]does[/i] wake up.
                        [/p]`;
        }

        toReturn += `   [p]
                            As the heart-wrenching sobbing coming from the open pen fades to nothing, you head back inside.
                            Once more, you take a knee next to the pink-scaled captive.
                            You place your hand atop her head, and comfortingly pat it, ever so gently.
                            Her crying has stopped, although even the girl's heavy, unsteady breathing is permeated with pure [b]sadness[/b].
                        [/p]
                        [p]
                            You cautiously coax the traumatized kobold into sitting up straight.
                            You don't dare to let her to rest her back against the wall, lest it trigger some kind of flashback of being held captive.
                            After she's calmed down a bit more, you ask the woman for her name.
                            No response.
                            She continues to stare blankly ahead of her.
                            You ask if she'd like to leave this place.
                            A single nod, weak but solemn.
                        [/p]
                        [p]
                            You say there's a nice house just next door that she can go to.
                            But first, she needs to help you break out the rest of the kobolds, so they can all get out of this hell-hole together.
                            She tilts her visage your way.
                            She visibly tries to respond to your request, yet no words leave her mouth.
                            Instead, another slight, affirmative shaking of her head.
                        [/p]
                        [p]
                            You help the girl to her feet.
                            At first, she can barely stand, let alone walk, yet she quickly regains her leg-strength.
                            It takes just a few minutes of you supporting her before she can stand on her own once more.
                            Once you lead her outside of the containment pen, the sight of the downed vixen coming into view,
                            her spirits seem to be lifted further.
                            The two of you share a mutual nod, and you jointly set to work opening the remainder of the enclosures,
                            liberating the stables from vulpine oppression.
                        [/p]
                        [p]
                            In each pen, the situation is roughly the same.
                            There's a bunch of drugged up males, some thoroughly violated,
                            often pregnant, mostly mind-broken kobold women, a couple of which used to be [i]men[/i],
                            and finally there's one or two more relatively unscathed, yet completely immobilized [i]newcomers[/i], who retain much of their [i]sanity[/i].
                            You focus on this final group, setting them free like you did your new,
                            pink-scaled friend, and getting them to help in the ongoing effort to [i]get through[/i] to the [i]less fortunate[/i] members of the group.
                        [/p]
                        [p]
                            Eventually, you manage to get a somewhat functional form of [i]coordination[/i] going on.
                            The more mentally capable kobolds herd the drug-addled and physically debilitated ones out of the pens.
                            Most of them can't stop touching themselves, so [b]afflicted[/b] by various potions and drugs that they'll likely take days,
                            if not weeks to physically recover from this entire endeavour, let alone mentally.
                        [/p]
                        [p]
                            The entire group gets corralled through the barn's single side-door, leading into the adjacent house,
                            and hopefully towards a better future than what they'd have had in here.
                            ${dead ? "" : "The vixen, who [b]still[/b] has not regained her consciousness, gets dragged along with."}
                            One of the last few stragglers to remain, diligently ensuring that each and every one of the once-prisoners finds his or her way inside,
                            is the pink-scaled girl you rescued first.
                        [/p]
                        [p]
                            Ultimately, you and her end up being the only two left in the barn.
                            No words are spoken, but you can [i]sense[/i] her gratitude,
                            in the same way that she realizes that you can't help but make this [b]a lot[/b] more awkward than it should have been.
                            A small, heart-warming smile forms on her face.
                            With a loud, relieved sigh and a deepfully grateful nod, she silently turns around, and walks away to join the others inside of the house.
                        [/p]
                        [p]
                            You wish you could stay, but alas, your presence here would only cause trouble for the kobolds.
                            Besides, you have business to attend to in Varanar, and you'd only get in the way of their healing process.
                            You're certain they'll be fine.
                            They're in at least one set of good hands.
                            With the knowledge that you did the right thing here today,
                            you quietly leave the barn through the back door,
                            not as the thief you entered as,
                            but as an unsung hero of your kind.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnAttackFree";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unawareAttack.get() !== BarnAttackChoice.FREE) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unawareFree.change();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unawareAttack.get() !== BarnAttackChoice.FREE) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\unaware\attack\free\BarnAttackFree.js

// scenes\HamletDay2\BreedingFarm\barn\eager\unaware\attack\loot\BarnAttackLoot.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnAttackLoot(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "barnAttackLootContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dead = global.kobold.history.shared.cassidy.dead;
        var toReturn = `[p]
                            With the vixen out of the equation, you decide to help yourself to whatever valuables you can find.
                            Sadly, judging by Cassidy's prior frantic searching through the bag of potions she ${dead ? "died" : "passed out"} near,
                            you somehow doubt there's a lot left for you to plunder.
                        [/p]`;
        if (dead) {
            toReturn += `[p]
                            Your eyes drifting towards the canine's half-nude corpse for a moment, you notice one of her hands being firmly clenched shut,
                            a very distinctive shade of pink visible through her furred fingers.
                            Reaching down, you roughly [b]pry[/b] her hand open, stealing the potion she once intended to use on you from her dead grasp.
                        [/p]
                        [p]
                            At least you won't be leaving completely empty-handed.
                            For a moment, you consider walking out the front door of the barn,
                            but it's best you [b]don't[/b] risk being seen after having committed a murder.
                            Instead, you decide to leave via the barn's shoddy back door.
                        [/p]
                        [p]
                            As you cross the table you were strapped to not a few minutes ago, you notice something small and shiny resting atop the small stool next to it.
                            Inspecting the small trinket, it appears to be the key to the veritable [b]mansion[/b] the barn is attached to.
                            You decide to pocket the key.
                            It's not like Cassidy will be needing her home anymore.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You take to searching through the sack yourself.
                            It seems the canine had already removed most of the useless things from it, a myriad of empty potions cluttering the ground.
                            Still, there's plenty garbage left.
                            This must have been an entire year's supply of various nefarious concotions.
                            After a few minutes of rummaging about, you finally come across the thing she intended to addle you with:
                            a single, filled ${global.kobold.inventory.createItemWithTooltip(new FemPlus())}, which you take as your just due.
                        [/p]
                        [p]
                            At least you won't be leaving completely empty-handed.
                            For a moment, you consider walking out the front door of the barn,
                            but it's best you [b]don't[/b] risk being seen after having struck down someone who [i]is[/i] protected by law.
                            Instead, you decide to leave via the barn's shoddy back door, before she wakes up.
                        [/p]`;
        }

        toReturn += `   [p]
                            Exiting the barn through the back, you continue on your way to Varanar.
                            To make sure that nobody ${dead ? "" : "but the vixen herself"} can link [b]you[/b] to the events that transpired at the breeding farm,
                            you decide to take a few less-well travelled paths through various fields and a few small patches of woodland.
                            Before long, you arrive back at the eastern road to Varanar, just a few miles away from the city limits.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnAttackLoot";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.unawareAttack.get() !== BarnAttackChoice.LOOT) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unawareLoot.change();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.unawareAttack.get() !== BarnAttackChoice.LOOT) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\unaware\attack\loot\BarnAttackLoot.js

// scenes\HamletDay2\BreedingFarm\barn\eager\unaware\attack\BarnAttack.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnAttack(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var lootId = "unawareAttackLootButton";
    var freeId = "unawareAttackFreeButton";

    var toLoot = new BarnAttackLoot(ctrl);
    var toFree = new BarnAttackFree(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            You wield your ${global.kobold.inventory.createItemWithTooltip(weapon)} and slowly, stealthily approach the vulpine from behind.
                            Being a fox, she may have incredible hearing and a sense of smell well beyond your own,
                            but you are a [b]kobold[/b], one of the sneakiest races around.
                            That, and she's far too occupied with searching through an entire bag full of noisy rubbish to notice you drawing close.
                        [/p]`;
        
        if(weapon.stab()) {
            global.kobold.history.shared.cassidy.dead = true;
            toReturn += `[p]
                            "[color=orange]Aha![/color]" the vixen exclaims, rising to her feet with a single filled vial.
                            Just as you were about to strike, she turns around, audibly gasping at the sight of her captive brandishing a weapon.
                            Adrenaline courses through your veins.
                            Your ambush foiled, your fight or flight instinct kicks in.
                            With a loud, menacing growl of sheer rage, you leap towards Cassidy, weapon in hand.
                        [/p]
                        [p]
                            Your body collides with the fox's slender frame, knocking the stunned vixen off her feet.
                            You fall on top of the backwards-tumbling woman as she lands on the ground with a heavy thud.
                            You come face to face with the [b]bitch[/b] that tried to [b]ruin[/b] your life.
                            You stare deeply into her shocked, wide-open eyes.
                            Her jaw falls agape.
                            A single [i]gurgle[/i] escapes her lips.
                            Her eyes slowly close, and her body slumps down lifelessly.
                        [/p]
                        [p]
                            Looking downwards, you find your blade having sunk into Cassidy's very heart, felling her with a single lethal [b]stab[/b].
                            You must have managed to get off a lucky strike during your initial lunge, or perhaps when the both of you fell to the ground.
                            Regardless, the deed is done.
                            You pull your weapon from the canine's corpse, and wipe the blood it is stained with off onto the ground.
                            You take a few deep breaths, excitement rushing from your system.
                            [b]You've just killed Cassidy.[/b]
                        [/p]
                        [p]
                            Calming down a little, you rise to your feet.
                            You convince yourself that you did what [b]had[/b] to be done, no matter how horrible of an experience it was.
                            You had no choice.
                            It was her, or you.
                            Besides, she was a [b]terrible[/b] person.
                            She [b]deserved[/b] this.
                            Or so you tell yourself, to dull your aching conscience.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You brace your weapon, holding it high above your head, preparing your entire body to strike down with all of your might.
                            You fail to contain a loud, angry growl, alerting the vixen to your presence directly behind her, yet before she even manages to react,
                            you bring your ${global.kobold.inventory.createItemWithTooltip(weapon)} down upon her head,
                            the hard surface impacting against the woman's skull with a twistingly satisfying [b]crack[/b].
                        [/p]
                        [p]
                            The canine's body slumps down to the ground, completely lifeless.
                            For a moment, you fear you might have [b]killed[/b] her, but then you notice the vixen's bare, 
                            prone torso slowly, yet rhythmically inflating and deflating.
                            Still breathing.
                            You didn't kill her, you've just managed to knock her unconscious in a single blow.
                            Righteous retribution.
                        [/p]
                        [p]
                            Satisfied with your work, you take a few deep breaths to calm yourself down.
                            You did it.
                            You [b]won[/b].
                            That [b]bitch[/b] won't be turning you into a breeding kobold, or worse yet some rich guy's side dish.
                            Not today.
                            Not [b]ever[/b].
                        [/p]`;
        }
        
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildLootChoice() {
        var disabled = ctrl.unawareAttack.has() ? "disabled" : "";
        var isSelected = ctrl.unawareAttack.get() === BarnAttackChoice.LOOT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${lootId}" ${disabled}>Loot</button>`;
        var tooltip = `Steal whatever you can, then run away from the scene of the crime.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFreeChoice() {
        var disabled = ctrl.unawareAttack.has() ? "disabled" : "";
        var isSelected = ctrl.unawareAttack.get() === BarnAttackChoice.FREE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${freeId}" ${disabled} style="width:160px;">Free your peers</button>`;
        var tooltip = `Open the cubicles and free the other kobolds. Surely some of them must still be able-minded, right?`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice" style="width:160px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildLootChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFreeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + lootId), $("#" + freeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerLootListener() {
        $("#" + lootId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unawareAttack.set(BarnAttackChoice.LOOT);
            loadNextPart(toLoot);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerFreeListener() {
        $("#" + freeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unawareAttack.set(BarnAttackChoice.FREE);
            loadNextPart(toFree);
        });
    }
    

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnAttack";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if(ctrl.unaware.get() !== BarnUnawareChoice.ATTACK) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unawareAttack.change(ctrl.misc.weapon());
        part += buildChoice();
        part += toLoot.build();
        part += toFree.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.unaware.get() !== BarnUnawareChoice.ATTACK) {
            return;
        }
        
        toLoot.registerListeners();
        toFree.registerListeners();

        if (ctrl.unawareAttack.has()) {
            return;
        }

        registerLootListener();
        registerFreeListener();
    };
}

var BarnAttackChoice = {
    LOOT: 1,
    FREE: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\eager\unaware\attack\BarnAttack.js

// scenes\HamletDay2\BreedingFarm\barn\eager\unaware\BarnUnaware.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnUnaware(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var runId = "unawareRun";
    var attackId = "unawareAttack";
    var forceId = "unawareForce";

    var toRun = new BarnRun(ctrl);
    var toAttack = new BarnAttack(ctrl);
    var toForce = new BarnForce(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var shirt = ctrl.misc.shirt();
        var bra = ctrl.misc.bra();
        var dick = ctrl.misc.dick();
        var cunt = ctrl.misc.cunt();
        var cuntVirgin = ctrl.misc.cuntVirgin();
        var gender = ctrl.misc.gender();
        var clothes = shirt || pants || underwear;
        ctrl.misc.wearShirt(shirt);
        ctrl.misc.wearPants(pants);
        ctrl.misc.wearUnderwear(underwear);
        var toReturn = `[p]
                            ${clothes ? "You sneakily don your garments once more, the modesty they provide bolstering your courage." : ""}
                            Cassidy is still searching through a stash of various bottles and trinkets, most of which seem to be empty and worthless.
                            She hasn't noticed your escape.
                            The tables have turned.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRunChoice() {
        var disabled = ctrl.unaware.has() ? "disabled" : "";
        var isSelected = ctrl.unaware.get() === BarnUnawareChoice.RUN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${runId}" ${disabled}>Run</button>`;

        var tooltip = `Get the hell out of here, as fast as you can.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildAttackChoice() {
        var weapon = ctrl.misc.weapon();
        var disabled = ctrl.unaware.has() || !weapon ? "disabled" : "";
        var isSelected = ctrl.unaware.get() === BarnUnawareChoice.ATTACK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${attackId}" ${disabled} style="width:160px;">Attack Cassidy</button>`;

        var tooltip;
        if (weapon) {
            tooltip = global.parser.parse(`That [b]bitch[/b] thought she could keep you here against your will? Time for some payback.`);
        } else {
            tooltip = global.parser.parse(`You don't have any weapons to attack the woman with.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice" style="width:160px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildForceChoice() {
        var str = ctrl.misc.strength();
        var strongEnough = str >= 6;
        var disabled = ctrl.unaware.has() || !strongEnough ? "disabled" : "";
        var isSelected = ctrl.unaware.get() === BarnUnawareChoice.FORCE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${forceId}" ${disabled} style="width: 250px">Force her onto the table</button>`;

        var tooltip;
        if (strongEnough) {
            tooltip = global.parser.parse(`Approach from behind, pick her up and [b]wrestle[/b] the woman onto the table herself.
                                            She may be twice your size, but she doesn't have muscles like you do.`);
        } else {
            tooltip = global.parser.parse(`Requires [b]6 strength[/b].`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildRunChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAttackChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildForceChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + runId), $("#" + attackId), $("#" + forceId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerRunListener() {
        $("#" + runId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unaware.set(BarnUnawareChoice.RUN);
            loadNextPart(toRun);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerAttackListener() {
        $("#" + attackId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unaware.set(BarnUnawareChoice.ATTACK);
            loadNextPart(toAttack);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerForceListener() {
        $("#" + forceId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.unaware.set(BarnUnawareChoice.FORCE);
            loadNextPart(toForce);
        });
    }
    

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnUnaware";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.unaware.change();
        part += buildChoice();
        part += toRun.build();
        part += toAttack.build();
        part += toForce.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        toRun.registerListeners();
        toAttack.registerListeners();
        toForce.registerListeners();

        if (ctrl.unaware.has()) {
            return;
        }

        registerRunListener();
        registerAttackListener();
        registerForceListener();
    };
}

var BarnUnawareChoice = {
    RUN: 1,
    ATTACK: 2,
    FORCE: 3
};

// END scenes\HamletDay2\BreedingFarm\barn\eager\unaware\BarnUnaware.js

// scenes\HamletDay2\BreedingFarm\barn\eager\unhinged\BarnUnhinged.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnUnhinged(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toUnaware = new BarnUnaware(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Frustrated, you writhe around, fighting against your restraints,
                            testing both the limits of the iron cuffs holding you in place,
                            and the patience of your [i]hostess[/i].
                            While your efforts prove to be quite fruitless, you [b]do[/b] notice something.
                            Three of the cuffs are firmly locked in place, yet the one holding your right arm seems a bit...[i]loose[/i].
                            It rattles whenever you tug your arm to the side, moving ever so slightly.
                        [/p]
                        [p]
                            As Cassidy continues to look for a single vial that [i]isn't[/i] empty,
                            you quietly begin to [i]experiment[/i] with the wobbly mechanism.
                            While the other cuffs have fallen into fastening latches, this one seems to have [i]missed[/i] its locking apparatus.
                            If you can apply [i]just[/i] the right amount of tension in the correct manner, you [b]might[/b] be able to free yourself, or at least your arm.
                        [/p]
                        [p]
                            As though your life depends on it -- which it does -- you start to twiddle the janky piece of metal around.
                            At first you try simply tugging it open, to no avail.
                            Then, you test some [i]smoother[/i], more delicate motions.
                            You come to the conclusion that shaking your arm from side to side seems to loosen the mechanism ever so slightly,
                            although you're not quite sure why.
                        [/p]
                        [p]
                            Realizing your window of opportunity is rapidly closing, an unsettling kind of nervousness springs up at the back of your mind.
                            You attempt to yank the cuff to the left, yet it doesn't budge beyond a certain point.
                            To the right, neither, proves to be the right answer.
                            Your brain is working overtime to comprehend how this thing works.
                            Then it dawns on you.
                            The solution!
                        [/p]
                        [p]
                            Instead of moving your arm one way or the other,
                            you [b]rotate[/b] the scaly appendage as best as you are able to,
                            the metal cuff being [b]lifted[/b] out of its improperly-fastened clasp.
                            It's just a little, but it's enough to allow you to [b]yank[/b] the mechanism open the rest of the way,
                            freeing your right arm through sheer brute force!
                        [/p]
                        [p]
                            With your now-free hand, you begin to pry at the other cuffs.
                            While nearly impossible to remove by whoever's trapped on the table,
                            barring any sort of malfunction as you were lucky enough to experience,
                            it seems that these restraints were made to be easily undone from the outside.
                            You hastily free the rest of your limbs, before hopping off of the table.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnUnhinged";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.table.get() !== BarnTableChoice.UNHINGED) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += toUnaware.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.table.get() !== BarnTableChoice.UNHINGED) {
            return;
        }

        toUnaware.registerListeners();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\unhinged\BarnUnhinged.js

// scenes\HamletDay2\BreedingFarm\barn\eager\recognize\BarnRecognize.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnRecognize(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toUnaware = new BarnUnaware(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You try to come up with a suitable insult to hurl Cassidy's way,
                            yet your mind is far too busy trying to come up with a way to escape this darn table to think of anything.
                            Then you remember.
                            You've seen this table before.
                            Well, not exactly [i]this[/i] one, but the very same model.
                        [/p]
                        [p]
                            A few years ago,
                            when you were summoned to the headquarters of the Three-Tailed Rat Clan to [i]explain[/i] your lack of [i]contribution[/i] to [i]the cause[/i],
                            you were [i]introduced[/i] to some of their [i]torture[/i] practices to [b]scare[/b] you into coughing up the coins you owed at the time.
                            Luckily, you were spared any sort of pain yourself,
                            but you were quite well informed what would happen to you should you [b]fail[/b] to pay them back.
                            One of the tools used by the gang to keep whoever they're trying to torment firmly in place,
                            just so happened to be a table like the one you're strapped to now.
                        [/p]
                        [p]
                            You calm down a little, useful memories and a sliver of hope coursing through your mind.
                            "[color=orange]Good. I knew you wouldn't put up too much of a fight.[/color]"
                            Cassidy speaks, mistaking your slight relaxation for some form of acceptance.
                            Like you've already come to terms with what she plans on doing to you.
                            Like you'd [b]ever[/b] willingly agree to this [b]horrible[/b] life she has in store for you.
                        [/p]
                        [p]
                            The canine turns around, facing away from you, her tail wagging excitedly,
                            her rear tauntingly swaying from side to side, as if to add insult to injury.
                            "[color=orange]Anyway, time for your first shot. I'll be [b]right[/b] back. Don't go anywhere![/color]"
                            She giggles at her own joke in a way that only a true [b]psychopath[/b] could,
                            as she begins to walk towards a nearby stash of what you can only presume to be Seki's
                            sold out ${global.kobold.inventory.createItemWithTooltip(new FemPlus())} supply.
                        [/p]
                        [p]
                            With the woman's back turned, you spot the perfect opportunity to make a move.
                            If you recall, the mechanism strapping you to the table is operated via a small lever on the underside of the flat, wooden surface.
                            While your hands and legs may be bound, [i]one[/i] of your appendages is not: your tail!
                            You slither the prehensile body-part underneath the table, blindly searching for the lever.
                        [/p]
                        [p]
                            You swing your tail to the left.
                            Nothing.
                            To the right.
                            Again, nothing!
                            Are you mistaken?
                            Is this a different model?
                            Can you simply not reach far enough?
                            Frustrated and scared, you check the left hand side once more.
                            [b]Nothing![/b]
                            Almost giving up hope, you flail your tail about in random directions, like a wounded animal writhing in pain.
                            Then, suddenly, the scaly appendage hits something!
                            With a soft [b]clink[/b], the cuffs keeping you bound to the table unlock.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnRecognize";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.eager.get() !== BarnEagerChoice.RECOGNIZE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += toUnaware.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.eager.get() !== BarnEagerChoice.RECOGNIZE) {
            return;
        }

        toUnaware.registerListeners();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\recognize\BarnRecognize.js

// scenes\HamletDay2\BreedingFarm\barn\eager\tied\BarnEscape.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnEscape(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "barnEscapeContinueButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var shirt = ctrl.misc.shirt();
        var bra = ctrl.misc.bra();
        var weapon = ctrl.misc.weapon();
        var gear = pants || underwear || shirt || bra || weapon;
        var toReturn = `[p]
                            You lay there, motionless, a million thoughts racing through your mind while you listen to the vixen's muffled footsteps fading.
                            You hear a door opening, closing again, and then locking.
                            With Cassidy having left the barn, it feels as though a great burden is lifted off your shoulders.
                            You let out a loud, deeply depressed sigh.
                            For a few more minutes, you audibly cry, no longer needing to hold yourself back to save face.
                            You slowly rock back and forth, wishing you were [i]anywhere but here[/i].
                        [/p]
                        [p]
                            During your soothing motion, you notice something.
                            The rope around your wrists is kind of...[i]slack[/i].
                            In fact, you aren't properly tied together at all.
                            The bitch must have been relying on the paralyzing agent, or perhaps on your weariness.
                            While you [b]are[/b] tired, oh so [b]very[/b] tired, you aren't about to pass out.
                            Not yet, anyway.
                        [/p]
                        [p]
                            The bindings holding your wrists together were made poorly enough for you to be able to reach the rope itself with your fingers.
                            Using your sharp claws, you manage to free yourself in a matter of minutes, the thick rope somewhat tough to cut through.
                            Before you free your legs, you first wipe your tears and somewhat tidy your loins, wiping several palmfuls of sticky spunk off onto the ground.
                            After removing your bonds completely, you rise to your shaky legs, leaning against the wall to prevent yourself from falling over.
                        [/p]
                        [p]
                            Luckily, some of the longer lasting aftereffects of the transformation-mix are starting to fade.
                            While, at first, your legs are exceedingly unsteady, you are able to walk anew, without any kind of support, in just a few minutes.
                            Your loins still aching, you [i]do[/i] end up walking with a slight limp, however.
                            And your limbs, in general, feel just a little less [i]powerful[/i] than before.
                            Having recovered most of your motor skills, you begin trying to break out of the enclosure you are trapped in.
                        [/p]
                        [p]
                            Your first thought is to [b]dig[/b] your way out, but the ground the barn was built on feels pretty solid.
                            Too hard to dig through.
                            Besides, it might take too long.
                            Then, you recall a plan you had when you first entered this place.
                            You wanted to crawl on top of one of the walls, to see what was inside of the enclosures.
                            And that's exactly how you're going to get out of here.
                        [/p]
                        [p]
                            With a small jump, you skillfully manage to grab onto the top of the wooden wall.
                            While your arms are still feeling rather weak, you still manage to lift your slim body upwards.
                            Swinging a single leg over the enclosure, you almost lose your balance and fall down to the ground,
                            yet you manage to keep yourself stable due to years of experience in thievery and burglary.
                            Swinging your second leg over the wall, you hop down to the ground,
                            your weakened legs only barely refraining from buckling, a second near-faceplant narrowly avoided.
                            You hope this enfeebling effect isn't permanent.
                        [/p]
                        [p]
                            Looking around, you notice three ways to leave the barn.
                            A door leading to the house, which you're reasonably certain was the one Cassidy locked on her way out,
                            the front door, through which you entered this horrible place,
                            and the back door, the one you are closest to.
                            ${gear ? "You try, but fail to find your stripped gear. Bitch must have taken it indoors." : ""}
                        [/p]
                        [p]
                            With a loud sigh, you decide to leave through the back door.
                            ${gear ? "You'd rather be naked and free than clothed and at risk of being imprisoned as a breeding slave [b]for life[/b]." : ""}
                            ${gear ? "Besides, if" : "If"} Cassidy catches you out of your pen,
                            she'd probably send you straight to the slaughterhouse instead of trying to [i]tame[/i] you.
                            Opening the large, rickety wooden door, you slip outside, your naked body assaulted by a cold winter breeze.
                            Shivering, you sneak out of the wicked bitch's back yard unnoticed.
                        [/p]
                        [p]
                            Still, you'd rather not take any chances.
                            Travelling via back roads, skipping through fields and traversing small patches of woodland, you make your way towards Varanar.
                            Close to the evening, you arrive at the eastern road, just a few miles away from the city limits.
                            Your adventure at the barn has left you scarred for life,
                            but at least you're left with a brighter future than becoming a sentient [b]cow[/b].
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnEscape";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.tiedUp.get() !== BarnTiedChoice.ESCAPE) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.barnEscape.change();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.tiedUp.get() !== BarnTiedChoice.ESCAPE) {
            return;
        }

        //toEager.registerListeners();

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\tied\BarnEscape.js

// scenes\HamletDay2\BreedingFarm\barn\eager\tied\BarnPassOut.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnPassOut(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "barnPassOutRevertButton";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            You lay there, motionless, a million thoughts racing through your mind while you listen to the vixen's muffled footsteps fading.
                            You hear a door opening, closing again, and then locking.
                            With Cassidy having left the barn, it feels as though a great burden is lifted off your shoulders.
                            You let out a loud, deeply depressed sigh.
                            For a few more minutes, you audibly cry, no longer needing to hold yourself back to save face.
                            You slowly rock back and forth, wishing you were [i]anywhere but here[/i].
                            The soothing motion slowly makes your exhausted body drift off into a deep slumber.
                        [/p]
                        [p]
                            When you wake up, you are already bound to the breeding table once more.
                            The bitch must have moved you while you were fast asleep.
                            She introduces you to another member of her [i]herd[/i],
                            and just like the first, he thoroughly [b]breeds[/b] you.
                            You get another shot of femplus, and from that point onwards, events grow kind of blurry.
                        [/p]
                        [p]
                            You get fucked over and over again, for hours on end.
                            Finally removed from the table, you get to drink from a trough to wet your parched throat,
                            only to end up getting pounded from behind, and nearly drowned in the process.
                            Another injection, but an aphrodisiac this time around.
                            You get placed inside of an enclosure with other kobolds,
                            chained to a wall with your ${dick ? "newly formed," : ""} sopping wet cunt practically [b]begging[/b] for attention.
                        [/p]
                        [p]
                            You desperately try to please yourself, but with both your hands and your tail bound,
                            there isn't much you can do besides wriggling your lower body around.
                            You [b]beg[/b] for one of the others to touch you, to [b]fuck you[/b], to [b]sate[/b] your [b]need[/b], yet none of them heed your call.
                            One of the more mentally coherent members of the group begins to taunt you,
                            rubbing his dick against your snout, jerking off onto your face,
                            making you perform various depraved acts in return for the promise of [b]pleasure[/b],
                            without ever living up to his side of the deal.
                        [/p]
                        [p]
                            Come morning of the third -- or was it the fourth? -- day, you are finally removed from the pen,
                            placed back onto the table, your anticipating excitement deeply staining the hard wooden surface even before your mistress reveals her plans.
                            Instead of the [b]cock[/b] you so desperately desire, you end up having to please [b]her[/b],
                            the act of serving another female only resulting in arousing you [b]more[/b],
                            lust [b]gushing[/b] from your pleasure-deprived slit, pooling up on the table before liberally dripping to the ground.
                        [/p]
                        [p]
                            Then, as you finish eating her out for the umpteenth time, Cassidy finally shows you some mercy.
                            She brings the exact same two fingers she'd been using to fondle your groin for the entire duration of your stay,
                            ever since the very first day, to your absolutely [b]soaked[/b] cunt.
                            After teasing you for what seems like an eternity, gently stroking without pleasing,
                            she finally [b]shoves[/b] both of her fingers into your needy slit,
                            sinking her furred digits inside of you, down to the very knuckle.
                        [/p]
                        [p]
                            Your mind goes blank.
                            You wildly buck your hips against her fingers,
                            the simple insertion enough to rock your stimulation-deprived body with the most intense orgasm you've ever experienced,
                            a veritable [b]torrent[/b] of your feminine fluids squirting onto the ground below.
                            Your life becomes even more of a haze of lust, sex, arousal and denial.
                        [/p]
                        [p]
                            Days, perhaps even weeks pass.
                            You find yourself chained, probably to the wall of one of the pens.
                            Your mouth is gagged.
                            A heavy blindfold prevents you from seeing.
                            A malicious fluid injected into your arm prevents you from hearing.
                            All you can do is feel.
                        [/p]
                        [p]
                            Feel, as you get [b]fucked[/b], again and again and again.
                            A different suitor every time.
                            Some big, some a little smaller, a few exceptionally [b]huge[/b], almost to the point of [b]breaking[/b] you.
                            The entire barn is having a go at you.
                            Or perhaps the people fucking you aren't kobolds at all.
                            You can't tell.
                            All you know is that it feels [b]good[/b].
                            Not just physically, but mentally.
                            Like this is how things were meant to be.
                            Like this is where you belong.
                        [/p]
                        [p]
                            You don't know how long it's been.
                            Months?
                            Years?
                            You can barely even remember your old life.
                            You sit in a corner, idly rubbing your bloated, pregnant belly with one hand,
                            while jerking off one of the males in your pen with the other.
                            You wish he'd fuck you instead, but he's too concerned about the baby growing inside of you.
                            The large door to your enclosure swings open.
                            Cassidy walks in, bucket in hand.
                            It's time for your daily milking.
                            Just like she promised.
                        [/p]
                        [p]
                            [c]
                                ${global.ui.inlineImage.create(`scenes/HamletDay2/BreedingFarm/img/cow`, "png", ImageWidthEnum.SMALLER, false, "Sneaky")}
                            [/c]
                        [/p]
                        [p]
                            [c]
                                [b]
                                    Cow end. Moo.
                                [/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" style="margin-left: auto; position: relative; float: right;" id="${continueId}">`;
        choice += `&lt;&lt Revert`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400, function () {
            var main = $("#main");
            main.scrollTop(main[0].scrollHeight - main.height() - 5); // Fix scrolling
        });
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tiedUp.unset();
            var parent = new BarnTied(ctrl);
            $("#" + that.id).remove();
            $("#" + parent.id).remove();
            loadNextPart(parent);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnPassOut";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.tiedUp.get() !== BarnTiedChoice.PASSOUT) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.tiedUp.get() !== BarnTiedChoice.PASSOUT) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\tied\BarnPassOut.js

// scenes\HamletDay2\BreedingFarm\barn\eager\tied\BarnTied.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnTied(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var passOutId = "barnTiedPassoutButton";
    var escapeId = "barnTiedEscapeButton";

    var toBadEnd = new BarnPassOut(ctrl);
    var toEscape = new BarnEscape(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var shirt = ctrl.misc.shirt();
        var bra = ctrl.misc.bra();
        var dick = ctrl.misc.dick();
        var cunt = ctrl.misc.cunt();
        var cuntVirgin = ctrl.misc.cuntVirgin();
        var gender = ctrl.misc.gender();
        var toReturn = `[p]
                            While you have been crying ever since getting injected,
                            it is only when Cassidy and her pet disappear into the animal pen that you truly burst out into tears,
                            no longer feeling like you need to [i]hold strong[/i] as a matter of pride.
                            Your loud snivelling and sad sobbing echo throughout the barn.
                            Your loins ache.
                            You feel betrayed, abused, violated and defiled.
                            You wish you'd never trusted that mean [b]bitch[/b].
                        [/p]
                        [p]
                            The vixen emerges from the cubicle, closing it behind her.
                            Her pet is no longer with her.
                            You assume she restrained him in there,
                            to bottle up his precious reproductive fluid until the next time he gets to fuck a [i]bitch[/i] [sneak]like you[/sneak].
                            You try to regain your composure, to look [b]strong[/b] and [b]unbroken[/b] before your captor,
                            but you simply can't stop crying.
                        [/p]
                        [p]
                            You glare at the bitch through your tear-filled eyes, with the most hate-filled look you can muster.
                            She seems to [b]enjoy[/b] your [b]torment[/b].
                            "[color=orange]Now that was [b]fun[/b], wasn't it?[/color]" she mocks while approaching.
                            If you weren't bound, you'd leap at her neck and [b]rip out her throat[/b] with your bare teeth.
                            Alas, the iron cuffs pinning you down hold fast, keeping your impotent rage in check.
                        [/p]
                        [p]
                            Approaching the table, the woman presses two fingers 
                            -- the same two fingers she's been using to touch your groin during this entire ordeal --
                            against your gaping slit once more.
                            Slipping them inside of your well-fucked hole, she parts your folds, 
                            a [b]rush[/b] of thick, heady kobold spunk flooding from your genitals,
                            as though she's [b]inspecting[/b] her pet's work.
                            Satisfied, the half-naked vixen retracts her digits once more, before licking her fingertips clean, clearly savoring the flavor of fresh semen.
                        [/p]
                        [p]
                            "[color=orange]You should be happy.
                            Not every newcomer gets to [b]enjoy[/b] my most prized kobold as their [b]first[/b].[/color]"
                            the canine speaks.
                            You can barely stand listening to her voice.
                            "[color=orange]Anyway, let's get you ready to spend the night.[/color]"
                        [/p]
                        [p]
                            Reaching underneath the table, the vixen flips a switch, and with a loud [b]clink[/b],
                            the cuffs holding you in place retract into the table, freeing your limbs.
                            An opportunity?
                            You try to move your limbs, to sit upstraight, to hop off the table and [b]bolt[/b] out the door,
                            to lunge at the woman and [b]tear her heart out[/b], but no matter how much your mind feels like moving,
                            your body simply refuses to budge.
                        [/p]
                        [p]
                            It seems that you are near-paralyzed from the neck down, able to turn your head still, but not much more.
                            Panicking, you fearfully gaze up at Cassidy, the smug smile on her face only deepening.
                            "[color=orange]Don't worry.
                            It's just a temporary muscle relaxant.
                            Wouldn't want a certain uppity little [b]bitch[/b] fighting back, now would we?[/color]"
                            Of course.
                            She spiked the potion, the stuff she injected you with.
                            Will this [b]hell[/b] never end?
                        [/p]
                        [p]
                            Bending down, the female fox grabs you by the ankles, and harshly pulls you off of the table,
                            your back falling to the ground with a soft thud, your head almost hitting the flat, wooden surface on the way down. `;

        if (shirt || bra) {
            toReturn += `Without the restraints blocking your limbs, Cassidy can [b]properly[/b] strip you, removing the last bits of clothing from your body completely.
                        "[color=orange]Won't be needing those, now will we?[/color]"`;
            if (ctrl.misc.gold()) {
                toReturn += `The only thing she doesn't take is the gold pouch wrapped around your neck, probably mistaking it for a keepsake.
                            You doubt she figures you for the kind of kobold that keeps any form of currency.`;
            }
        } else {
            if (ctrl.misc.gold()) {
                toReturn += `Cassidy notices your coin pouch.
                            "[color=orange]That's a cute bag of pebbles you have around your neck.
                            I'll let you keep it, for now.
                            That way, I'll have something to take from you, should you misbehave.[/color]"
                            Luckily, she mistakes it for a keepsake.`;
            }
        }


        toReturn += `       After taking one final glance at your naked, cum-dripping form, the vulpine lazily begins to drag you towards the single open cubicle,
                            which so happens to also be the nearest one, sparing you a lengthy trip through the dirt.
                        [/p]
                        [p]
                            Luckily, this particular pen is devoid of any drugged up, sex-craving, relief-deprived breeding kobolds.
                            In fact, it is completely empty, apart from the woman and yourself.
                            Tossing your light frame against one of the wooden walls of the enclosure,
                            the vixen uses a nearby piece of rope to tie your arms and legs together.
                            "[color=orange]There we go. Don't go anywhere, now. I'll be back in the morning, and we'll continue your [b]training[/b].[/color]"
                        [/p]
                        [p]
                            Bending downwards yet again, Cassidy plants a single, sickly-sweet kiss atop your forehead.
                            "[color=orange]Sleep tight, my little [b]pet[/b]. See you tomorrow.[/color]"
                            And with those words, she leaves you lying hog-tied on the cold ground.
                            As she swings close and shuts the large door to your new [i]cell[/i],
                            you can feel your extremities regaining the slightest bit of functionality.
                        [/p]`;

        ctrl.misc.strip();
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildBadChoice() {
        var disabled = ctrl.tiedUp.has() ? "disabled" : "";
        var isSelected = ctrl.tiedUp.get() === BarnTiedChoice.PASSOUT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${passOutId}" ${disabled}>Pass out</button>`;
        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildGoodChoice() {
        var disabled = ctrl.tiedUp.has() ? "disabled" : "";
        var isSelected = ctrl.tiedUp.get() === BarnTiedChoice.ESCAPE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${escapeId}" ${disabled} style="width: 200px;">Escape</button>`;
        var tooltip = global.parser.parse(`Cassidy did a poor job at tying you up. Wriggle out of your restraints and try to leave the barn unnoticed.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildBadChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildGoodChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + passOutId), $("#" + escapeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerPassOutListener() {
        $("#" + passOutId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tiedUp.set(BarnTiedChoice.PASSOUT);
            loadNextPart(toBadEnd);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerEscapeListener() {
        $("#" + escapeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tiedUp.set(BarnTiedChoice.ESCAPE);
            loadNextPart(toEscape);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnTied";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.tableRape.continued()) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toBadEnd.build();
        part += toEscape.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.tableRape.continued()) {
            return;
        }

        toBadEnd.registerListeners();
        toEscape.registerListeners();

        if (ctrl.tiedUp.has()) {
            return;
        }

        registerPassOutListener();
        registerEscapeListener();
    };
}

var BarnTiedChoice = {
    PASSOUT: 1,
    ESCAPE: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\eager\tied\BarnTied.js

// scenes\HamletDay2\BreedingFarm\barn\eager\BarnTableRape.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnTableRape(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = "barnTableRapeContinueButton";

    var toTied = new BarnTied(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function chestDescriptor() {
        var resisted = ctrl.barnResist.has();
        var boobs = ctrl.misc.boobs();
        var dick = ctrl.misc.dick();

        if (resisted) {
            if (boobs) {
                return "tits";
            } else {
                return "flat chest";
            }
        } else {
            if (dick) {
                return "flat, but still sensitive chest";
            } else if (boobs) {
                return "enlarged, still sensitive breasts";
            } else {
                return "small, newly grown pair of tits";
            }
        }
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var resisted = ctrl.barnResist.has();
        var cunt = ctrl.misc.cunt();
        var virgin = (dick && ctrl.misc.cockslitVirgin()) || (cunt && ctrl.misc.cuntVirgin());
        var toReturn = `[p]
                            As Cassidy and her [i]pet[/i] begin to near you, the male starts to incessantly sniff the air surrounding him.
                            You don't know whether it's the faint scent of the vixen he's indulging in,
                            or if he can somehow [b]smell[/b] the fresh meat spread-eagle on the breeding table for him.
                            Maybe the potion ${dick ? "made your fragrance a bit more [i]feminine[/i]" : "enhanced your own feminine fragrance"},
                            or perhaps he's simply enjoying the relatively fresh air outside of his [i]cell[/i].
                        [/p]
                        [p]
                            Stepping in between your legs, the woman presses two of her fingers against your slit,
                            giving you a single, almost [i]taunting[/i] full-length swipe.
                            Bringing those same fingers to the male's face, she presses them against his snout.
                            He eagerly huffs her furred digits, relishing in the traces of your scent, his cock twitching in anticipation.
                        [/p]
                        [p]
                            "[color=orange]You know, I think I'm going to enjoy this almost as much as he is. [i]Almost[/i].[/color]"
                            That being said, Cassidy steps aside, and harshly tugs the kobold forwards by his leash.
                            He bumps against the table, almost falling on top of you, his hands ending up on your ${chestDescriptor()} as he attempts to catch himself,
                            his member mere inches away from your loins.
                        [/p]
                        [p]
                            Feeling your scales underneath his palms, the pent-up male immediately begins to grope your ${!dick && !resisted ? "" : "flat"} chest.
                            He wildly thrusts his hips, humping nothing but air in a desperate attempt to find something to [b]fuck[/b].
                            The vixen, meanwhile, takes a seat on the small stool next to the table.
                            After a minute of observing her sightless pet's failed attempts at doing what he was trained to do, she decides to intervene.
                        [/p]
                        [p]
                            Gripping him by the collar, the vulpine forcibly [b]yanks[/b] the desperate kobold closer to you, pressing his body down against yours completely,
                            almost laying him flat on top of you, were it not for his legs still anchoring him to the ground.
                            Getting a better feel for the contours of your frame, the sex-crazed male's humps become more precise,
                            yet he still doesn't manage to find his target right away.
                            His member prods around your inner thighs, poking against your soft scales with every motion of his hips.
                        [/p]
                        [p]
                            While deeply concerned for your own safety,
                            and practically on the verge of having a mental breakdown,
                            for a moment, a feeling of smugness washes over you.
                            She can't even get her pet to fuck you right.
                            You tilt your head to face Cassidy, and despite the tears rapidly flowing from your eyes, a weak sneer forms on your face.
                            A small smile that is quickly washed away as the male kobold finds his mark,
                            spearing your tender${virgin ? ", virginal" : ""} ${dick ? "cock" : ""}slit wide open,
                            ramming over half of his member inside of your overly tight passage in one fierce thrust.
                            ${!dick && virgin ? "[b]A few droplets of blood spill from your loins, signifying the loss of your virginity.[/b]" : ""}
                        [/p]
                        [p]
                            [c]
                                ${global.ui.inlineImage.create(`scenes/HamletDay2/BreedingFarm/img/rape-${dick ? "m" : "f"}`, "png", ImageWidthEnum.SMALLER, false, "Sneaky")}
                            [/c]
                        [/p]
                        [p]
                            Cassidy seems to delight in seeing the smile wiped off of your face, the final few sparks of resistance fading along with it.
                            A large, self-satisfied grin forms atop her face, as she wordlessly claims her victory over you, her place as your [i]superior[/i].
                            You want to scream, not just in pain and discomfort, but at [b]her[/b], yet the only sound that escapes your lips is a small whimper.
                            A whine, like a scolded dog would utter when struck by its master,
                            whatever dignity you had left taken away by a single pelvic thrust${virgin ? ", along with your virginity" : ""}.
                        [/p]`;
        
        if(dick) {
            toReturn += `[p]
                            The kobold, driven by lust, instantly begins to pump his hips,
                            [b]ramming[/b] his cock deeper and deeper inside of your body with every rocking motion,
                            [b]forcing[/b] inch and after inch of his considerable girth into your sensitive cockslit,
                            until he finally bottoms out, the tapered tip of his member harshly pressing against your own flaccid length,
                            as though prodding into your very masculinity, a thing which is clearly at stake here.
                        [/p]
                        [p]
                            His pelvis coming to rest against your forcibly-spread inner thighs,
                            the well-hung male pauses for a moment, savoring the warmth and tightness of your body.
                            You can feel his oversized length [b]pulsing[/b] and [b]quivering[/b] inside of you,
                            as though he's already on the verge of climax, holding back just so he can take his time to truly [b]enjoy[/b] this.
                        [/p]
                        [p]
                            Pressing his snout against your neck, the kobold loudly sniffs your scales,
                            the injection you received earlier making you smell more like a girl than a 
                            boy${resisted ? ", despite you having resisted most of its other effects" : ""}.
                            Your scent seems to fuel his lust, his [b]need[/b].
                            The [i]stallion[/i] clasps the far end of the table, steadying himself,
                            rocking the entire thing back and forth as he begins to outright [b]pound[/b] you with a series of powerful, savage, near-bestial thrusts.
                            The sounds of your quiet sobbing, the slapping of his scales against yours,
                            the creaking of the wooden table and his loud panting and huffing all reverberate throughout the barn,
                            culminating in a cacophony of debauchery that breaks the eerie silence otherwise permeating the decrepit building.
                        [/p]
                        [p]
                            Eventually, he settles into a more steady, but still high-paced rhythm,
                            thoroughly [b]fucking[/b] a hole that was never meant to be used this way,
                            his tapered tip poking your soft cock every time he bottoms out,
                            each hump a sore reminder that, in his eyes, you are nothing more than a fertile breeding bitch to be.
                            A newcomer that hasn't had enough femplus yet.
                            Another hole for him to [b]fuck[/b] and [b]impregnate[/b], in due time.
                        [/p]
                        [p]
                            Luckily, despite being a well-trained breeder, the kobold is [b]so[/b] pent up, that he can't last for more than a few minutes.
                            All too soon, his thrusts increase in intensity once more, just a [i]little bit[/i] more of his member pressing into your cockslit,
                            enough to start somewhat [b]crushing[/b] your own dick, battering into it to the point of extreme discomfort.
                            If Cassidy wanted to neuter you, all she'd have to do is give this guy a bottle of lubricant.
                        [/p]
                        [p]
                            Mere seconds after breaking his rhythm in favor of more ferocious thrusts, the male reaches his end.
                            He lets out a loud, domineering growl before [b]slamming[/b] himself inside of you one last time,
                            ensuring his member is pressed directly up against your own,
                            before a first thick splattering of cum squirts from the tapered tip,
                            covering your flaccid length, marking your very groin as [b]his[/b].
                        [/p]
                        [p]
                            His intense orgasm continuing, the breeder gives you a few, final, shallow humps,
                            rocking back and forth to ensure he coats every inch of your insides with his potent seed,
                            enough of which is pumped directly from his relief-depraved internal balls to completely [b]fill[/b] your genital slit up to the very [b]brim[/b],
                            before idly spilling out around the base of his shaft.
                        [/p]`;
        } else {
            toReturn += `[p]
                            The kobold, driven by lust, instantly begins to pump his hips,
                            [b]ramming[/b] his cock deeper and deeper inside of your body with every rocking motion,
                            [b]forcing[/b] inch and after inch of his considerable girth into your sensitive twat,
                            until he finally bottoms out, the tapered tip of his member sharply pressing against your cervix,
                            as though prodding into your very womb, the clear risk of pregnancy stressed once again.
                        [/p]
                        [p]
                            His pelvis coming to rest against your forcibly-spread inner thighs,
                            the well-hung male pauses for a moment, savoring the warmth and tightness of your body.
                            You can feel his oversized length [b]pulsing[/b] and [b]quivering[/b] inside of you,
                            as though he's already on the verge of climax, holding back just so he can take his time to truly [b]enjoy[/b] this.
                        [/p]
                        [p]
                            Pressing his snout against your neck, the kobold loudly sniffs your scales,
                            the injection you received earlier enhancing your girly smell${resisted ? ", despite you having resisted most of its other effects" : ""}.
                            Your scent seems to fuel his lust, his [b]need[/b].
                            The [i]stallion[/i] clasps the far end of the table, steadying himself,
                            rocking the entire thing back and forth as he begins to outright [b]pound[/b] you with a series of powerful, savage, near-bestial thrusts.
                            The sounds of your quiet sobbing, the slapping of his scales against yours,
                            the creaking of the wooden table and his loud panting and huffing all reverberate throughout the barn,
                            culminating in a cacophony of debauchery that breaks the eerie silence otherwise permeating the decrepit building.
                        [/p]
                        [p]
                            Eventually, he settles into a more steady, but still high-paced rhythm,
                            thoroughly [b]fucking[/b] you, his tapered tip poking the passageway to your womb every time he bottoms out,
                            each hump a sore reminder that, in his eyes, you are nothing more than a fertile breeding bitch.
                            A new hole for him to [b]break in[/b].
                            Another cunt for him to [b]fuck[/b] and [b]impregnate[/b].
                        [/p]
                        [p]
                            Luckily, despite being a well-trained breeder, the kobold is [b]so[/b] pent up, that he can't last for more than a few minutes.
                            All too soon, his thrusts increase in intensity once more, just a [i]little bit[/i] more of his member shoved into your slit,
                            enough for the tapered tip to start somewhat [b]pressing into[/b] your cervix,
                            slightly spreading open the tiny hole leading to the innermost cradle of your femininity.
                            It doesn't hurt [i]that[/i] much, but the sheer discomfort and defilement weigh heavily on your mind.
                        [/p]
                        [p]
                            Mere seconds after breaking his rhythm in favor of more ferocious thrusts, the male reaches his end.
                            He lets out a loud, domineering growl before [b]slamming[/b] himself inside of you one last time,
                            ensuring his member is deeply prodding into your cervix before a first thick, fertile rope of cum squirts from the tapered tip,
                            almost directly into your uterus.
                        [/p]
                        [p]
                            His intense orgasm continuing, the breeder gives you a few, final, shallow humps,
                            rocking back and forth to ensure he [b]floods[/b] every inch of your insides with his potent seed,
                            enough of which is pumped directly from his relief-depraved internal balls to completely [b]fill[/b] your twat up to the very [b]brim[/b],
                            before idly spilling out around the base of his shaft, your pregnancy all but assured.
                        [/p]`;
        }
        
        toReturn += `   [p]
                            Cassidy, who had been silently watching you getting [b]broken in[/b] from the side, slowly rises from her wooden stool.
                            She steps behind the kobold, and with a rough, cruel yank on the iron chain serving as his leash, she pulls the male off of you,
                            his still cum-squirting cock ripped from your loins.
                            The breeder falls backwards, to the ground.
                            For a few moments, he fights against the vixen's pull, clamoring to return to the table,
                            to continue pumping you full of spunk, to feel your warmth around his length once more, perhaps even wanting to go for a [b]second[/b] round.
                        [/p]
                        [p]
                            But the canine'll have none of that.
                            She unrelentingly tugs him backwards, his resistance only serving to choke him on his own collar.
                            Realizing his struggle is futile, the kobold succumbs, limply allowing his [i]mistress[/i] to drag him through the dirt,
                            back towards his enclosure, his trail marked by the remarkable amount of jizz still oozing from his cock,
                            the same stuff that liberally flows from your slightly gaping ${dick ? "cock" : ""}slit, dripping down to the ground below.
                        [/p]`;
        
        if(virgin) {
            toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildContinueChoice() {
        var disabled = ctrl.tableRape.continued() ? "disabled" : "";
        var isSelected = ctrl.tableRape.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildContinueChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tableRape.continue();
            loadNextPart(toTied);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnTableRape";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if ((ctrl.barnResist.get() !== BarnResistChoice.RAPE) && (ctrl.transformation.get() !== BarnTransformationChoice.RAPE)) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.tableRape.change();
        part += buildChoice();
        part += toTied.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if ((ctrl.barnResist.get() !== BarnResistChoice.RAPE) && (ctrl.transformation.get() !== BarnTransformationChoice.RAPE)) {
            return;
        }

        toTied.registerListeners();

        if (ctrl.tableRape.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\BreedingFarm\barn\eager\BarnTableRape.js

// scenes\HamletDay2\BreedingFarm\barn\eager\BarnTransformation.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnTransformation(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var rapeId = "barnTableRapeButton";
    var persuadeId = "barnTablePersuadeButton";

    var toRape = new BarnTableRape(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function getBoobSize(boobs) {
        var size = "flat chest";
        if (boobs > SizeEnum.NONE) {
            size = "small tits";
        }
        if (boobs > SizeEnum.SMALL) {
            size = "average breasts";
        }
        if (boobs > SizeEnum.AVERAGE) {
            size = "large tits";
        }
        return size;
    }

    function getDickSize(dick) {
        var size = "small";
        if (dick > SizeEnum.SMALL) {
            size = "average";
        }
        if (dick > SizeEnum.AVERAGE) {
            size = "sizeable";
        }
        return size;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var shirt = ctrl.misc.shirt();
        var bra = ctrl.misc.bra();
        var dick = ctrl.misc.dick();
        var dickSize = getDickSize(dick);
        var boobs = ctrl.misc.boobs();
        var boobSize = getBoobSize(boobs);
        var cunt = ctrl.misc.cunt();
        var cuntVirgin = ctrl.misc.cuntVirgin();
        var gender = ctrl.misc.gender();
        var toReturn = `[p]
                            You can already feel the potion starting to kick in.`;

        if (dick) {
            toReturn += `   Your groin tingles.
                            In fact, your entire body starts to feel...[i]odd[/i].
                            You can sense something about yourself [b]changing[/b].
                            An odd feeling courses through your body.
                            Your face contorts as your ${dickSize} prick begins to [b]shrink[/b],
                            your cockslit growing just a little more roomy in the process.`;
        } else if (!boobs) {
            toReturn += `   Your chest begin to tingle.
                            In fact, your entire body starts to feel...[i]odd[/i].
                            You gaze downwards, just in time to witness your bosom rapidly expanding,
                            a pair of small tits forming atop your previously flat chest.`;
        } else {
            toReturn += `   Your nipples begin to tingle.
                            In fact, your entire bosom starts to feel...[i]odd[/i].
                            You gaze downwards at your body, just in time to witness your ${boobSize} rapidly expanding,
                            growing ${boobs > SizeEnum.AVERAGE ? "even" : ""} bigger than they were before!`;
        }
        
        toReturn += `   [/p]
                        [p]
                            [c]
                                [b]You have grown more feminine.[/b]
                            [/c]
                        [/p]
                        [p]
                            While the entire ordeal lasts no longer than half a minute,
                            the transformation seems to sap you of your strength, of your stamina, of your very [b]will[/b] to resist.
                            You're not certain whether that's a part of the potion, or if the [b]bitch[/b] mixed in something of her own.
                            The changes -- both physically and mentally -- are subtle, but [b]very[/b] apparent to you.
                        [/p]
                        [p]
                            You don't get long to get used to the [i]new you[/i].
                            Cassidy emerges from the animal pen, iron chain in hand.
                            With a firm tug on the chain, she pulls a male kobold from the enclosure,
                            the spiky collar around his neck linked to the metal she holds, which acts like a leash of sorts.
                            The male is naked and visibly aroused, his member fully erect, a drop of precum already formed at the very tip.
                            His head is covered by what appears to be some kind of falconry hood, tailor-made for kobolds, blocking his sight without restricting his muzzle.
                        [/p]
                        [p]
                            The vixen closes up the cubicle once more.
                            Her [i]pet[/i] in tow, she begins to march towards you.
                            "[color=orange]Surprise.[/color]" the woman states plainly, too upset with you to so much as [i]feign[/i] enthusiasm.
                            "[color=orange]I thought I'd bring you a little [i]friend[/i].
                            Aren't you glad to see another of your kind?
                            I think he's happy to meet you too.[/color]"
                            The kobold reaches for his groin, but a cruel [b]slap[/b] makes him recoil, his hand begrudgingly returning to resting idly at his side.
                        [/p]
                        [p]
                            "[color=orange]Poor little thing hasn't had any [b]relief[/b] for [b]weeks[/b] now.
                            Given the amount of homebrewed aphrodisiac I've been administering, I doubt there's anything but [b]sex[/b] on his mind right now.
                            While wasting a pent-up load like this on a newcomer that
                            ${cunt ? "hasn't even had any fertility shots yet" : "hasn't even grown the right [i]equipment[/i] yet"}
                            might be a bit wasteful, I think I'll make an exception, [b]just for you[/b].[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRapeChoice() {
        var disabled = ctrl.transformation.has() ? "disabled" : "";
        var isSelected = ctrl.transformation.get() === BarnTransformationChoice.RAPE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${rapeId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildPersuadeChoice() {
        var charisma = ctrl.misc.charisma();
        var likeableEnough = charisma >= 6;
        var disabled = ctrl.transformation.has() || !likeableEnough ? "disabled" : "";
        var isSelected = ctrl.transformation.get() === BarnTransformationChoice.PERSUADE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${persuadeId}" ${disabled} style="width: 200px;">Persuade</button>`;

        var tooltip;
        if (likeableEnough) {
            tooltip = global.parser.parse(`Try to talk the kobold into helping you. <span class="todo">Unavailable</span>`);
        } else {
            tooltip = global.parser.parse(`Requires [b]6 charisma[/b]. <span class="todo">Unavailable</span>`);
        }


        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildRapeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildPersuadeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + rapeId), $("#" + persuadeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + rapeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.transformation.set(BarnTransformationChoice.RAPE);
            loadNextPart(toRape);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerPersuadeListener() {
        $("#" + persuadeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.transformation.set(BarnTransformationChoice.PERSUADE);
            //loadNextPart(toUnhinged);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnTransformation";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.injection.get() !== BarnInjectionChoice.TRANSFORM) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.transformation.change();
        part += buildChoice();
        part += toRape.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.injection.get() !== BarnInjectionChoice.TRANSFORM) {
            return;
        }

        toRape.registerListeners();

        if (ctrl.transformation.has()) {
            return;
        }

        registerContinueListener();
        //registerPersuadeListener();
    };
}

var BarnTransformationChoice = {
    RAPE: 1,
    PERSUADE: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\eager\BarnTransformation.js

// scenes\HamletDay2\BreedingFarm\barn\eager\BarnResist.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnResist(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var rapeId = "barnTableRapeButton";
    var persuadeId = "barnTablePersuadeButton";

    var toRape = new BarnTableRape(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var shirt = ctrl.misc.shirt();
        var bra = ctrl.misc.bra();
        var dick = ctrl.misc.dick();
        var cunt = ctrl.misc.cunt();
        var cuntVirgin = ctrl.misc.cuntVirgin();
        var gender = ctrl.misc.gender();
        var toReturn = `[p]
                           No.
                           You [b]refuse[/b] to give in.
                           She may have managed to inject you, but that just means the battlefield has shifted from outside of your body, to inside of it.
                           You grit your teeth and focus your mind.
                        [/p]
                        [p]
                            You can feel your body starting to tingle, as though attempting to change, but you [b]won't[/b] let it.
                            You won't let [b]her[/b] win.
                            Losing focus for a moment, you can instantly [b]feel[/b] a few minor alterations occurring,
                            but within seconds you put your mind back to the task of resisting the potion's influence.
                            You are [b]stronger[/b] than it.
                            You can outlast it.
                            It will [b]not[/b] break you.
                            [b]She[/b] will not break you.
                        [/p]
                        [p]
                            After a solid minute of mentally hardening every fiber of your being,
                            you can feel the concoction's influence over your body slowly fading.
                            You hold your metaphorical ground for a bit longer, before dropping out of your mentally focussed state.
                            You have won, this time.
                            But you aren't sure how much more of that stuff you can weather before it'll start to alter your body, your mind or both.
                        [/p]
                        [p]
                            Your inward struggle coming to an end, you turn your attention back to your surroundings,
                            just in time to witness Cassidy emerging from the animal pen, iron chain in hand.
                            With a firm tug on the chain, she pulls a male kobold from the enclosure,
                            the spiky collar around his neck linked to the metal she holds, which acts like a leash of sorts.
                            The male is naked and visibly aroused, his member fully erect, a drop of precum already formed at the very tip.
                            His head is covered by what appears to be some kind of falconry hood, tailor-made for kobolds, blocking his sight without restricting his muzzle.
                        [/p]
                        [p]
                            The vixen closes up the cubicle once more.
                            Her [i]pet[/i] in tow, she begins to march towards you.
                            "[color=orange]Surprise.[/color]" the woman states plainly, too upset with you to so much as [i]feign[/i] enthusiasm.
                            "[color=orange]I thought I'd bring you a little [i]friend[/i].
                            Aren't you glad to see another of your kind?
                            I think he's happy to meet you too.[/color]"
                            The kobold reaches for his groin, but a cruel [b]slap[/b] makes him recoil, his hand begrudgingly returning to resting idly at his side.
                        [/p]
                        [p]
                            "[color=orange]Poor little thing hasn't had any [b]relief[/b] for [b]weeks[/b] now.
                            Given the amount of homebrewed aphrodisiac I've been administering, I doubt there's anything but [b]sex[/b] on his mind right now.
                            While wasting a pent-up load like this on a newcomer that
                            ${cunt ? "hasn't even had any fertility shots yet" : "hasn't even grown the right [i]equipment[/i] yet"}
                            might be a bit wasteful, I think I'll make an exception, [b]just for you[/b].[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRapeChoice() {
        var disabled = ctrl.barnResist.has() ? "disabled" : "";
        var isSelected = ctrl.barnResist.get() === BarnResistChoice.RAPE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${rapeId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildPersuadeChoice() {
        var charisma = ctrl.misc.charisma();
        var likeableEnough = charisma >= 6;
        var disabled = ctrl.barnResist.has() || !likeableEnough ? "disabled" : "";
        var isSelected = ctrl.barnResist.get() === BarnResistChoice.PERSUADE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${persuadeId}" ${disabled} style="width: 200px;">Persuade</button>`;

        var tooltip;
        if (likeableEnough) {
            tooltip = global.parser.parse(`Try to talk the kobold into helping you. <span class="todo">Unavailable</span>`);
        } else {
            tooltip = global.parser.parse(`Requires [b]6 charisma[/b]. <span class="todo">Unavailable</span>`);
        }


        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildRapeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildPersuadeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + rapeId), $("#" + persuadeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + rapeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.barnResist.set(BarnResistChoice.RAPE);
            loadNextPart(toRape);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerPersuadeListener() {
        $("#" + persuadeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.barnResist.set(BarnResistChoice.PERSUADE);
            //loadNextPart(toUnhinged);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnResist";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.injection.get() !== BarnInjectionChoice.RESIST) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.barnResist.change();
        part += buildChoice();
        part += toRape.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.injection.get() !== BarnInjectionChoice.RESIST) {
            return;
        }

        toRape.registerListeners();

        if (ctrl.barnResist.has()) {
            return;
        }

        registerContinueListener();
        //registerPersuadeListener();
    };
}

var BarnResistChoice = {
    RAPE: 1,
    PERSUADE: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\eager\BarnResist.js

// scenes\HamletDay2\BreedingFarm\barn\eager\BarnInjection.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnInjection(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var transformId = "barnTableTransformButton";
    var resistId = "barnTableResistButton";

    var toTransform = new BarnTransformation(ctrl);
    var toResist = new BarnResist(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function getBoobSize(boobs) {
        var size = "flat chest";
        if (boobs > SizeEnum.NONE) {
            size = "small tits";
        }
        if (boobs > SizeEnum.SMALL) {
            size = "average breasts";
        }
        if (boobs > SizeEnum.AVERAGE) {
            size = "large tits";
        }
        return size;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var shirt = ctrl.misc.shirt();
        var bra = ctrl.misc.bra();
        var dick = ctrl.misc.dick();
        var boobs = ctrl.misc.boobs();
        var boobSize = getBoobSize(boobs);
        var cunt = ctrl.misc.cunt();
        var cuntVirgin = ctrl.misc.cuntVirgin();
        var gender = ctrl.misc.gender();
        var baseGender = ctrl.misc.baseGender();
        var toReturn = `[p]
                            Frustrated, you struggle against your restraints, 
                            testing both the limits of the iron cuffs holding you in place, and the patience of your hostess.
                            All four of the mechanisms pinning you to the table appear to be firmly fastened, not so much as budging an inch.
                            While you writhe around fruitlessly, Cassidy searches through a large bag filled with countless empty vials and flasks,
                            which she tosses aside, digging through the abundant trash in search of 
                            a ${global.kobold.inventory.createItemWithTooltip(new FemPlus())} that isn't drained of its contents.
                        [/p]
                        [p]
                            After about a solid minute of frantically trying to escape your bonds, you relent.
                            It's no use.
                            You can't get out.
                            Mere moments after your struggling comes to an end, the vixen finally finds what she was looking for.
                            "[color=orange]Aha![/color]" she exclaims, rising to her feet, holding a single vial filled with bright pink liquid in one hand,
                            and a large syringe made out of glass, tipped with a long, sharp needle in the other.
                        [/p]
                        [p]
                            Cassidy turns to face you once more.
                            Using the syringe, she draws fluid from the vial, the translucent barrel filling up to about the halfway point.
                            Discarding the now-empty bottle, the vulpine advances upon you,
                            sadistic intent clearly and proudly on display through a toothy grin plastered across her face.
                            "[color=orange]Scared of needles?[/color]" she mockingly asks,
                            applying pressure to the syringe's plunger to squirt a bit of the pink contents out of the needle.
                            "[color=orange]${gender === GenderEnum.MALE ? "Being a girl" : "Having a nice pair of breasts"} isn't all that bad, you know.[/color]"
                        [/p]
                        [p]
                            Reaching the table, the vixen places one of her furred hands atop your restrained right arm.
                            You snap your jaws in an attempt to frighten her, but she doesn't even seem [i]phased[/i] by the act.
                            Either she places great faith in her equipment,
                            or she's done this [b]so[/b] many times before your kind no longer seems threatening in the slightest.`;
        if (shirt) {
            toReturn += `   Cassidy deftly rolls up the sleeve of your ${global.kobold.inventory.createItemWithTooltip(shirt)}, exposing your scaly forearm.`;
        }
        if (bra) {
            ctrl.misc.removeBra();
            toReturn += `   After giving your arm a firm squeeze, the bitch moves her hand to your near-naked torso,
                            your ${boobs ? "bosom" : "lack of breasts"} covered only by your ${global.kobold.inventory.createItemWithTooltip(bra)}.
                            In a single, smooth motion, she tugs the fabric upwards, fully exposing your ${boobSize}.`;
        }

        toReturn += `   [/p]
                        [p]`;

        if (shirt || bra) {
            toReturn += `   Your body made bare, the woman places her hand firmly atop your arm once more,
                            pinning it to the table, holding it in place.`;
        } else {
            toReturn += `   The woman pins your arm to the table, firmly holding it in place.`;
        }

        toReturn += `       You struggle against her grasp, until you feel something [b]sharp[/b] and [b]pointy[/b] pressing against your scales.
                            Not the needle.
                            Her sharpened nails.
                            Cassidy speaks, anger audible in her voice:
                            "[color=orange]Stay [b]still[/b], you little [b]runt[/b].
                            Do you want me to [b]ram[/b] this needle into your fucking [b]forehead[/b] instead?![/color]"
                            She threateningly hovers the medical tool in front of your face, a single droplet of transformative liquid dripping onto your muzzle.
                            At this point, you wouldn't be surprised if she actually [b]meant[/b] that threat.
                            "[color=orange]I'm [b]trying[/b] to be [b]nice[/b] here, but if you keep struggling, I'll make you [b]wish[/b]
                            ${gender === GenderEnum.MALE ? "I used this potion to turn you into a girl" : "I chose to send you to the slaughterhouse instead."}.[/color]"
                        [/p]
                        [p]
                            Her stern words and harsh threats carving through your fighting spirit like a hot knife through butter, you cease your fidgeting.
                            A single tear escapes your eyes, sheer [b]fear[/b] and mental exertion slowly getting the best of you.
                            "[color=orange]That's better.[/color]" comments the vixen, who moves the syringe towards your steadied arm.
                            Without so much as a warning, she [b]plunges[/b] the needle past your scales, into your flesh.
                            You let out a surprised yelp, but decide not to attempt to fight back, lest she decide to stick the needle [i]elsewhere[/i].
                        [/p]
                        [p]
                            What, in reality, takes only a few seconds, seems to take [b]ages[/b] for you,
                            the canine slowly forcing the syringe's contents into your bloodstream.
                            Only after every single last droplet of pink liquid has disappeared into your body, does she remove the needle from your arm,
                            a tiny bit of blood spilling from the small puncture wound.
                        [/p]
                        [p]
                            "[color=orange]There. That wasn't so bad, now was it?
                            While we wait for this stuff to take effect, how about I go fetch you a little [b]surprise[/b], huh?[/color]"
                            Without waiting for your response, the obviously annoyed woman puts down her syringe, and walks off towards one of the enclosures.
                            Opening the large, wooden door keeping the pen shut, she slips inside for a moment, leaving you to your fate.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildTransformChoice() {
        var disabled = ctrl.injection.has() ? "disabled" : "";
        var isSelected = ctrl.injection.get() === BarnInjectionChoice.TRANSFORM;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${transformId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildResistChoice() {
        var endurance = ctrl.misc.endurance();
        var strongEnough = endurance >= 6;
        var disabled = ctrl.injection.has() || !strongEnough ? "disabled" : "";
        var isSelected = ctrl.injection.get() === BarnInjectionChoice.RESIST;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${resistId}" ${disabled} style="width: 200px;">Resist the potion</button>`;

        var tooltip;
        if (strongEnough) {
            tooltip = global.parser.parse(`Weather the foul liquid. Focus your mind in an attempt to block it from altering your body.`);
        } else {
            tooltip = global.parser.parse(`Requires [b]6 endurance[/b].`);
        }


        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildTransformChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildResistChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + transformId), $("#" + resistId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + transformId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.injection.set(BarnInjectionChoice.TRANSFORM);
            loadNextPart(toTransform);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerResistListener() {
        $("#" + resistId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.injection.set(BarnInjectionChoice.RESIST);
            loadNextPart(toResist);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnInjection";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.table.get() !== BarnTableChoice.INJECT) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.injection.change();
        part += buildChoice();
        part += toTransform.build();
        part += toResist.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.table.get() !== BarnTableChoice.INJECT) {
            return;
        }

        toTransform.registerListeners();
        toResist.registerListeners();

        if (ctrl.injection.has()) {
            return;
        }

        registerContinueListener();
        registerResistListener();
    };
}

var BarnInjectionChoice = {
    TRANSFORM: 1,
    RESIST: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\eager\BarnInjection.js

// scenes\HamletDay2\BreedingFarm\barn\eager\BarnTable.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnTable(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var injectId = "barnTableInjectButton";
    var unhingedId = "barnTableUnhingedButton";

    var toInject = new BarnInjection(ctrl);
    var toUnhinged = new BarnUnhinged(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var shirt = ctrl.misc.shirt();
        var bra = ctrl.misc.bra();
        var dick = ctrl.misc.dick();
        var cunt = ctrl.misc.cunt();
        var cuntVirgin = ctrl.misc.cuntVirgin();
        var gender = ctrl.misc.gender();
        var toReturn = `[p]
                            You try to come up with a suitable insult to hurl Cassidy's way,
                            yet your mind is far too busy trying to come up with a way to escape this darn table to think of anything.
                            All that escapes your lips is a deep, low, hate-filled growling.
                            She betrayed you.
                            She played you like a fool.
                            And now she's got you caught in her trap like a fly in a spider's web.
                        [/p]
                        [p]
                            "[color=orange]Now, now, my little [b]pet[/b].[/color]" the evil bitch says, noticing your furious glare.
                            While her voice seemed oh-so [b]alluring[/b] not a few minutes ago, at this very moment, you can't [b]stand[/b] her condescending tone.
                            Every word she says only fuels the raging fire that is your [b]hatred[/b] for the vixen.
                            "[color=orange]Settle down.[/color]" she resumes, the grin on her face turning souring into a scolding glance, annoyance audible in her voice.
                            "[color=orange]You wouldn't want me to revise my plans and send you to the [b]butcher[/b] instead of keeping you around as a breeder,
                            now would you?${dick ? " If anything, you should be grateful I'm willing to turn you into a girl just to keep you here." : ""}[/color]"
                        [/p]
                        [p]
                            T-the butcher?
                            Is that why she's breeding your kind?
                            To sell kobold offspring for their [b]meat[/b]?!
                            Mixed feelings of being sickened and terrified at the same time wash over you.
                            You feel an intense urge to [b]vomit[/b], yet your stomach's contents remain in place, which is probably for the best.
                        [/p]
                        [p]
                            As your glare fades to a pleading, scared, awestruck stare, a fresh smirk creeps onto Cassidy's face, as though she relishes in your fright.
                            "[color=orange]If you act like a good little slut, it'll be [b]your kids[/b] getting sent to the slaughterhouse instead of you.
                            And even then, not [b]all[/b] of them will end up on some nobleman's plate.
                            Some'll become slaves, others will get turned into expensive sex toys, and of course I'll keep a few for myself.[/color]"
                        [/p]
                        [p]
                            You are at a loss for words.
                            This bitch is [b]insane[/b].
                            Sure, your kind has no [i]official[/i] rights, but to be treated as a commodity, or worse, as cattle?!
                            To get [b]harvested[/b] for meat like scaly cows?!
                            Your mind can not comprehend how [b]anyone[/b] could treat a fellow sapient race like this.
                            Do kobolds -- and by extension: you -- matter [b]that[/b] little to the world?
                        [/p]
                        [p]
                            No.
                            This can't be happening.
                            It [b]has[/b] to be a bad dream.
                            You close your eyes and imagine you're still at home, resting in your bed, dreaming about some sadistic fox [b]whore[/b].
                            While the thought helps to calm you down a little, when you open your eyes,
                            you are sadly greeted with the sight of the barn and the vixen both.
                        [/p]
                        [p]
                            "[color=orange]Good. Now don't start growling again, or I'll have to gag you prematurely.[/color]"
                            Cassidy speaks, mistaking your slight calming for some form of acceptance.
                            Like you've already come to terms with what she plans on doing to you.
                            Like you'd [b]ever[/b] willingly agree to this [b]horrible[/b] life she has in store for you.
                            There's a million things you want to call her, yet you stay your tongue,
                            if just so you can [b]bite[/b] her later instead of getting your mouth plugged with whatever filthy rag she can produce.
                        [/p]
                        [p]
                            The canine turns around, facing away from you, her tail wagging excitedly,
                            her rear tauntingly swaying from side to side, as if to add insult to injury.
                            "[color=orange]Anyway, time for your first shot. I'll be [b]right[/b] back. Don't go anywhere![/color]"
                            She giggles at her own joke in a way that only a true [b]psychopath[/b] could,
                            as she begins to walk towards a nearby stash of what you can only presume to be Seki's
                            sold out ${global.kobold.inventory.createItemWithTooltip(new FemPlus())} supply.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildInjectChoice() {
        var disabled = ctrl.table.has() ? "disabled" : "";
        var isSelected = ctrl.table.get() === BarnTableChoice.INJECT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${injectId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildUnhingedChoice() {
        var luck = ctrl.misc.luck();
        var luckyEnough = luck >= 6;
        var disabled = ctrl.table.has() || !luckyEnough ? "disabled" : "";
        var isSelected = ctrl.table.get() === BarnTableChoice.UNHINGED;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${unhingedId}" ${disabled} style="width: 200px;">Unhinged cuff</button>`;

        var tooltip;
        if (luckyEnough) {
            tooltip = global.parser.parse(`One of the cuffs restraining your arms seems a bit loose. You might be able to pry it open.`);
        } else {
            tooltip = global.parser.parse(`Requires [b]6 luck[/b].`);
        }


        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildInjectChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildUnhingedChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + injectId), $("#" + unhingedId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + injectId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.table.set(BarnTableChoice.INJECT);
            loadNextPart(toInject);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerUnhingedListener() {
        $("#" + unhingedId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.table.set(BarnTableChoice.UNHINGED);
            loadNextPart(toUnhinged);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnTable";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.eager.get() !== BarnEagerChoice.TABLE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.table.change();
        part += buildChoice();
        part += toInject.build();
        part += toUnhinged.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.eager.get() !== BarnEagerChoice.TABLE) {
            return;
        }

        toInject.registerListeners();
        toUnhinged.registerListeners();

        if (ctrl.table.has()) {
            return;
        }

        registerContinueListener();
        registerUnhingedListener();
    };
}

var BarnTableChoice = {
    INJECT: 1,
    UNHINGED: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\eager\BarnTable.js

// scenes\HamletDay2\BreedingFarm\barn\eager\BarnEager.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BarnEager(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var tableId = "barnTableStrapButton";
    var recognizeId = "barnRecognizeButton";

    var toTable = new BarnTable(ctrl);
    var toRecognize = new BarnRecognize(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var shirt = ctrl.misc.shirt();
        var bra = ctrl.misc.bra();
        var dick = ctrl.misc.dick();
        var cuntVirgin = ctrl.misc.cuntVirgin();
        var gender = ctrl.misc.gender();
        var toReturn = `[p]
                            A coy smirk forms on your face.
                            It's not every day a pretty, good looking woman asks you to have sex with her.
                            Well, unless you visit the brothel district in Varanar, that is.
                            The thought that Cassidy may be a [i]prostitute[/i] trying to trick you into some kind of extortionist racket,
                            where she demands payment for sleeping with you, crosses your mind, yet you quickly dismiss it.
                            Vixens may be natural-born whores, but you doubt a hooker would operate out of a [i]barn[/i] in the middle of nowhere.
                        [/p]
                        [p]
                            Mimicking her lip-licking, you give the fox an eager, determined nod.
                            The flirtatious smile on her face deepens.
                            After another small kiss, she grabs your hand, and the both of you begin to walk further into the barn.
                        [/p]
                        [p]
                            You try to come up with a question or something meaningful to say, yet nothing springs to mind.
                            When you finally [i]do[/i] figure out a decent conversational topic, the moment you open your mouth,
                            Cassidy silences you before you manage to utter even a single word, by pressing a finger against your lips.
                            After hushing you, that same finger travels in between her thighs, and she casually begins to play with herself.
                        [/p]
                        [p]
                            For some reason, that [i]small detail[/i] turns you on far more than it should.
                            You attempt to touch yourself mid-stride as well, yet due to your considerably shorter legs,
                            you fail to both masturbate and keep up with Cassidy's walking pace at the same time.
                            After almost tripping and falling down,
                            you decide to wait until the both of you have reached the table before enjoying yourself.
                        [/p]
                        [p]
                            Luckily, your clumsiness doesn't seem to phase the vixen.
                            If anything, the closer you get to the table, the more enthusiastic she gets.
                            She must [b]really[/b] want this.
                            As you near the end of the rows of cubicles, the canine roughly tugs you forwards,
                            swinging you by your arm, propelling you to her front.
                            She then hugs you tightly, pressing your tiny frame into her furred body, your face rubbing against her soft belly.
                        [/p]
                        [p]
                            As abruptly as the hug began, so too does it come to a quick end.
                            Cassidy pushes you away from her, making you take a few steps backwards.
                            You end up bumping your rear against the edge of the wooden table.
                            The woman rapidly advances upon you, seemingly driven by wanton lust.
                        [/p]
                        [p]
                            Gripping both of your wrists, she lifts your arms, and forcibly bends you backwards over the rather low table,
                            pressing your back against the hard, wooden surface, your legs bending at the knees.
                            Leaning over you, the vixen pins you against the table by pressing her rack against your chest.
                            Her face dips downwards, her muzzle pressing against yours once more.
                        [/p]
                        [p]
                            This time, however, a small peck isn't enough to sate the vulpine's need.
                            Her long, broad tongue snakes out of her mouth, slipping in between your lips and past your teeth.
                            Taken by surprise, you barely reciprocate as Cassidy begins to outright 
                            [b]explore[/b] every square inch of the insides of your mouth with her tongue,
                            giving you the deepest, most [b]passionate[/b] kiss you've [b]ever[/b] experienced, if a bit one-sided.
                        [/p]`;

        if (pants || underwear || shirt) {
            ctrl.misc.removePants();
            ctrl.misc.removeUnderwear();
            ctrl.misc.removeShirt();
            toReturn += `[p]
                            Without breaking the kiss, Cassidy releases one of your arms, and uses her free hand to take off your clothing.
                            ${shirt ? `Your ${global.kobold.inventory.createItemWithTooltip(shirt)} gets pulled up towards your neck, your belly made bare.` : ""}
                            ${pants ? `Your ${global.kobold.inventory.createItemWithTooltip(pants)} get tugged down in a single, firm yank, all the way to the ground.` : ""}
                            ${underwear ? `Your ${global.kobold.inventory.createItemWithTooltip(underwear)} gets removed, exposing your loins.` : ""}
                        [/p]
                        [p]
                            With ${bra ? "most of" : ""} your garments removed, `;
        } else {
            toReturn += `[p]
                            Seeing as you don't have any garments to remove, ${bra ? `save for your ${global.kobold.inventory.createItemWithTooltip(bra)},` : ""}`;
        }

        toReturn += `       the female fox breaks the intense kiss, which leaves the both of you huffing and panting.
                            With her long tongue still lolling from her mouth, a wicked grin forms on Cassidy's face.
                            While continuing to press her body down upon yours, she gently kicks apart your legs, spreading them wide,
                            your slit on full display.
                            Then, she reaches underneath the table, and with a loud, metallic [b]clink[/b],
                            a series of intricate, hidden metal cuffs spring forth from the flat surface and its legs,
                            wrapping around both your wrists and your ankles,
                            constraining you to the heavy, wooden construct.
                        [/p]
                        [p]
                            [c]
                                ${global.ui.inlineImage.create(`scenes/HamletDay2/BreedingFarm/img/pin-${dick ? "m" : "f"}`, "png", ImageWidthEnum.SMALLER, false, "Sneaky")}
                            [/c]
                        [/p]
                        [p]
                            At first, confusion surges through your mind.
                            You try to move your arms and your legs, but find them tightly bound.
                            With a questioning smile, you stare at the vixen, inquisitively interrogating her without uttering so much as a single word.
                            Surely, this isn't necessary, right?
                            It's undeniably kinky, but you're not sure if you feel [i]comfortable[/i] being [i]tied down[/i] like this.
                        [/p]
                        [p]
                            Tongue retracting back into her mouth, Cassidy lifts her body off of yours, the devilish smirk on her face only deepening.
                            Without offering an explanation, she drops to her knees in between your legs, her face mere inches away from your tender loins.`;

        if (dick) {
            toReturn += `   You feel extremely conflicted.
                            One part of you is growing concerned about the abrupt, sudden bondage, and the lack of an explanation for it.
                            Another fragment of your mind is barely able to contain its excitement, your dick on the verge of being sucked by a lusty vixen.
                        [/p]
                        [p]
                            You can [b]feel[/b] her warm breath on your cockslit, your member hardening internally in anticipation.
                            But, instead of the lapping tongue you expected, Cassidly presses two of her furred fingers against your slit.
                            Before you have time to utter any sort of protest, she deftly slips the tips of her furred digits inside of you,
                            spreading your most private parts open as wide as possible, seemingly inspecting your very innards.
                        [/p]
                        [p]
                            After another few moments of uncomfortable silence, 
                            her fingers testing your very limits before being removed from your body,
                            the woman finally speaks: `;

            if (dick < SizeEnum.AVERAGE) {
                toReturn += `"[color=orange]Cute.
                              A tiny little member for a tiny little kobold.
                              I'd hardly even call you a [b]boy[/b].
                              But don't worry, that just means it'll take less drugs to turn you into the perfect little breeder girl.[/color]"`;
            } else if (dick < SizeEnum.LARGER) {
                toReturn += `"[color=orange]Decent size.
                              Better than most of your kind, at least.
                              But don't worry, I've plenty of drugs to turn you into the perfect little breeder girl.[/color]"`;
            } else {
                toReturn += `"[color=orange]Oh my!
                              You're a big one, aren't you?
                              Too bad I already have plenty of studs.
                              But don't worry, I'll just have to give you a few extra injections to turn you into the perfect little breeder girl.[/color]"`;
            }

            toReturn += `[/p]
                        [p]
                            As Cassidy's words sink into your mind, panic slowly begins to settle in.
                            She wants to [b]turn you into a girl[/b]?!
                            Not just a girl, but a [b]breeder[/b]?!`;

        } else if (gender === GenderEnum.CUNTBOY) {
            toReturn += `   You feel extremely conflicted.
                            One part of you is growing concerned about the abrupt, sudden bondage, and the lack of an explanation for it.
                            Another fragment of your mind is barely able to contain its excitement.
                            You're about to get eaten out by a lusty vixen!
                        [/p]
                        [p]
                            You can [b]feel[/b] her warm breath on your ${cuntVirgin ? "virginal" : ""} slit, your folds dampening in anticipation.
                            Cassidy presses her face further in between your thighs, her cold nose rubbing against your clit,
                            while the furred tip of her muzzle ends up resting against your twat.
                            She sniffs you loudly, not once, but twice, her strong, canine nose taking in every minute detail about your scent.
                            Must be a canine thing.
                            Or maybe she just gets off to smelling kobold slits.
                        [/p]
                        [p]
                            Just as your lust for her mouth reaches its peak, the woman suddenly pulls her head back.
                            Instead of the lapping tongue you expected, Cassidly presses two of her furred fingers against your folds.
                            She deftly slips the tips of her furred digits inside of you,
                            spreading your most private parts open as wide as possible, seemingly inspecting your very innards.
                        [/p]
                        [p]
                            After another few moments of uncomfortable silence, 
                            her fingers testing your very limits before being removed from your body,
                            the woman finally speaks: `;

            if (cuntVirgin) {
                toReturn += `"[color=orange]A virgin kobold?
                              Such a rarity nowadays.
                              But then again, people probably just assumed you were a regular boy, didn't they?
                              I could just sell you outright for a [b]lot[/b] of gold, but I [b]do[/b] need more breeding bitches.
                              No, I think I'll keep you instead.[/color]"`;
            } else {
                toReturn += `"[color=orange]Not a virgin, huh?
                              Don't worry, sweetie.
                              That just saves my breeding studs the trouble of breaking you in.[/color]"`;
            }

            toReturn += `[/p]
                        [p]
                            As Cassidy's words sink into your mind, panic slowly begins to settle in.
                            She wants to [b]turn you into a breeder[/b]?!`;
        } else {
            toReturn += `   You feel extremely conflicted.
                            One part of you is growing concerned about the abrupt, sudden bondage, and the lack of an explanation for it.
                            Another fragment of your mind is barely able to contain its excitement.
                            You're about to get eaten out by a lusty vixen!
                        [/p]
                        [p]
                            You can [b]feel[/b] her warm breath on your ${cuntVirgin ? "virginal" : ""} slit, your folds dampening in anticipation.
                            Cassidy presses her face further in between your thighs, her cold nose rubbing against your clit,
                            while the furred tip of her muzzle ends up resting against your twat.
                            She sniffs you loudly, not once, but twice, her strong, canine nose taking in every minute detail about your scent.
                            Must be a canine thing.
                            Or maybe she just gets off to smelling kobold slits.
                        [/p]
                        [p]
                            Just as your lust for her mouth reaches its peak, the woman suddenly pulls her head back.
                            Instead of the lapping tongue you expected, Cassidly presses two of her furred fingers against your folds.
                            She deftly slips the tips of her furred digits inside of you,
                            spreading your most private parts open as wide as possible, seemingly inspecting your very innards.
                        [/p]
                        [p]
                            After another few moments of uncomfortable silence, 
                            her fingers testing your very limits before being removed from your body,
                            the woman finally speaks: `;

            if (cuntVirgin) {
                toReturn += `"[color=orange]A virgin kobold?
                              Such a rarity nowadays.
                              I could just sell you outright for a [b]lot[/b] of gold, but I [b]do[/b] need more breeding bitches.
                              No, I think I'll keep you instead.[/color]"`;
            } else {
                toReturn += `"[color=orange]Not a virgin, huh?
                              Don't worry, sweetie.
                              That just saves my breeding studs the trouble of breaking you in.[/color]"`;
            }

            toReturn += `[/p]
                        [p]
                            As Cassidy's words sink into your mind, panic slowly begins to settle in.
                            She wants to [b]turn you into a breeder[/b]?!`;
        }

        toReturn += `       You begin to struggle against your restraints, to no avail.
                            The vixen slowly rises to her feet, a bittersweet, cruel yet twistingly caring smile on her face.
                            "[color=orange]Relax, darling.
                            After the first few weeks, you'll learn to love it here.
                            And I'll personally come milk your udders [i]every single day[/i].[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildTableChoice() {
        var disabled = ctrl.eager.has() ? "disabled" : "";
        var isSelected = ctrl.eager.get() === BarnEagerChoice.TABLE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${tableId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRecognizeChoice() {
        var name = ctrl.misc.name();
        var intelligence = ctrl.misc.intelligence();
        var smartEnough = intelligence >= 6;
        var disabled = ctrl.eager.has() || !smartEnough ? "disabled" : "";
        var isSelected = ctrl.eager.get() === BarnEagerChoice.RECOGNIZE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${recognizeId}" ${disabled} style="width: 200px;">Recognize table</button>`;

        var tooltip;
        if (smartEnough) {
            tooltip = global.parser.parse(`You recall seeing a very similar table once upon a time, but you can't quite remember where. Think, ${name}, think!`);
        } else {
            tooltip = global.parser.parse(`Requires [b]6 intelligence[/b].`);
        }


        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildTableChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildRecognizeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + tableId), $("#" + recognizeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + tableId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.eager.set(BarnEagerChoice.TABLE);
            loadNextPart(toTable);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerRecognizeListener() {
        $("#" + recognizeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.eager.set(BarnEagerChoice.RECOGNIZE);
            loadNextPart(toRecognize);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "barnEager";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.barn.get() !== BreedingBarnChoice.EAGER) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.eager.change();
        part += buildChoice();
        part += toTable.build();
        part += toRecognize.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.barn.get() !== BreedingBarnChoice.EAGER) {
            return;
        }

        toTable.registerListeners();
        toRecognize.registerListeners();

        if (ctrl.eager.has()) {
            return;
        }

        registerContinueListener();
        registerRecognizeListener();
    };
}

var BarnEagerChoice = {
    TABLE: 1,
    RECOGNIZE: 2
};

// END scenes\HamletDay2\BreedingFarm\barn\eager\BarnEager.js

// scenes\HamletDay2\BreedingFarm\barn\BreedingBarn.js

/**
 * Barn section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BreedingBarn(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var declineId = "barnOfferDeclineButton";
    var notBarnId = "barnOfferNotBarnButton";
    var eagerId = "barnOfferEagerButton";

    var toDecline = new NotBarnTea(ctrl);
    var toNotBarn = new NotBarn(ctrl);
    var toEager = new BarnEager(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var name = ctrl.misc.name();
        var toReturn = `[st]The barn[/st]
                        [p]
                            You head back up the trail a bit, and this time follow the [i]other[/i] path, the one leading towards the barn.
                            Even before you enter, the familiar smell of straw mixed with a less common fragrance,
                            one you can't quite put your finger on, creeps into your nose.
                            You walk up to the sizeable wooden door, and after another threefold-series of firm knocks,
                            you invite yourself into the barn,
                            the sturdy, yet ill-maintained door nearly falling off its hinges as you open it and close it behind you once more.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/HamletDay2/BreedingFarm/barn/barn-full.png" target="_blank">
                                    <img class="sfw" src="scenes/HamletDay2/BreedingFarm/barn/barn.png" title="Made by Sneaky">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            While the outside looks decrepit, the inside of the barn looks rather tidy and neat,
                            besides the hay strewn about the flattened ground in a disorderly manner, but you suppose that's to be expected on a farm of sorts.
                            The vast main area is divided into two rows of large cubicles, one on either side of the barn.
                            In between these animal pens is a broad walkway filled with bales of hay and boxes of varying sizes.
                            On the far end of the barn, there stands a lone table, with a single stool next to it.
                        [/p]
                        [p]
                            The cubicles themselves are relatively new, judging by the far brighter tone of wood they are made of.
                            They must be a recent addition to this age-old structure.
                            All of them are closed, apart from a single one, all the way at the end of the walkway.
                            The cracks in between the planks are, sadly, not nearly big enough to see anything through.
                            While high enough for humans to gaze over the top of,
                            the exact species of animals contained within the enclosures remain a mystery to someone as [i]small[/i] as you are.
                        [/p]
                        [p]
                            You'd imagine such pens to contain feral horses or other big barnyard animals, yet you don't hear any animal noises.
                            You wonder what's in them.
                            Must be something small and silent, like rabbits maybe?
                            A rabbit breeding farm.
                            That [i]would[/i] make sense.
                            But why build a barn this big to keep rabbits in it?
                            Wouldn't they tunnel out?
                        [/p]
                        [p]
                            Just as you were about to climb up the side of one of the wooden walls to sate your curiosity,
                            a loud voice with a heavy western accent disturbs your train of thoughts:
                            [c][b]<span style="font-size: 30px;">[color=orange]Howdy, partner![/color]</span>[/b][/c]
                        [/p]
                        [p]
                            Out of the single open cubicle at the other end of the barn strides a half-naked anthromorphic vixen,
                            most of her lithe frame covered by nothing but pristine fur.
                            The only clothing she's wearing is a pair of ${global.kobold.inventory.createItemWithTooltip(new SpurredBoots())} 
                            and a large ${global.kobold.inventory.createItemWithTooltip(new CowboyHat())} that cutely flattens her ears.
                            ${global.kobold.history.shared.enslavement.captured ? `You think you saw her at the auction earlier,
                            although she's ditched her black shirt since.` : ``}
                            With a hearty smile on her face, she gives you a cheerful wave, before jogging towards you, her hefty,
                            furred breasts jiggling substantially with every graceful step she takes.
                        [/p]
                        [p]
                            As the woman nears you, her brisk running pace turns into a slow, elegant walk, 
                            bare hips swaying seductively back and forth, along with her tail.
                            Your eyes practically moving by themselves, you take a closer look at the female fox's body.
                            Her fur is a mixture of orange and white, with her extremities colored black.
                            Most of the front of her body, including her substantial bosom, her belly and her groin are all colored an immaculate white,
                            which you imagine runs all the way to the underside of her tail.
                        [/p]
                        [p]
                            [c]
                                ${global.ui.inlineImage.create("scenes/HamletDay2/BreedingFarm/img/intro", "png", ImageWidthEnum.CUSTOM, false, "Sneaky")}
                            [/c]
                        [/p]
                        [p]
                            You snap out of the dirty, daydreaming thoughts invading your mind as the woman extends her arm towards you,
                            grabbing your hand and shaking it eagerly.
                            You awkwardly fail to reciprocate her enthusiasm, but she doesn't seem to mind.
                            "[color=orange]I'm Cassidy! What do they call you?[/color]"
                            Still slightly enamored with her naked body, and awestruck by her lack of clothing, you barely manage to stammer out your name.
                            "[color=orange]${name}?
                            How cute!${global.kobold.history.shared.enslavement.captured ? `You're the one that was being sold earlier, if I'm not mistaken?` : ``}[/color]"
                            ${global.kobold.history.shared.enslavement.captured ? `You give the vixen a slight nod.
                            "[color=orange]I wish I was the one who took you home.
                            You look adorable enough to [b]eat[/b]![/color]"` : ``}
                        [/p]
                        [p]
                            The vixen lets go of your hand, instead placing her palm firmly atop your shoulder.
                            She bends down promiscuously, her cleavage almost pressing against your face.
                            She's so close you can [b]smell[/b] her, a faintly feminine fragrance mixing with the domineering scent of straw and hay.
                            Paralyzed with a mixture of anxiety and budding arousal, you stand perfectly still as she leans in,
                            pressing her furred muzzle against the scaly side of your face.
                            Then, she whispers, softly, ever so softly, yet her words pierce your mind like a crossbow bolt would your soft flesh.
                            "[color=orange][small]So, ${name}, have you ever [b]fucked[/b] a vixen before?[/small][/color]"
                        [/p]
                        [p]
                            Your jaw falls agape.
                            Your eyes open wide.
                            Closing your mouth, you swallow audibly.
                            Did she say what you [i]think[/i] she said?
                            A million thoughts race through your mind,
                            yet your mental panic is interrupted as the woman slides her tongue across your scales,
                            trailing a lick over the side of your face before aligning her muzzle with your own.
                            Pressing her soft, velvety lips against yours, she gives you a single, tender kiss.
                        [/p]
                        [p]
                            The vixen pulls away from the kiss, but keeps her face close to yours.
                            After a short moment of her lustfully licking her lips while staring into your eyes, she speaks yet again.
                            "[color=orange]Well then, what'll it be,
                            ${ctrl.misc.baseGender() === GenderEnum.MALE ? "my handsome little devil" : "you pretty little girl"}?[/color]"
                            She gestures behind her, as though inviting you deeper into the barn.
                            "[color=orange]There's a table in the back, and I've an [b]itch[/b] only something [b]small[/b] and [b]cute[/b] like you can scratch.[/color]"
                            ${global.kobold.history.shared.enslavement.captured ? `You're starting to regret not being bought by her.` : ``}
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the ant choice button.
     */
    function buildDeclineChoice() {
        var disabled = ctrl.barn.has() ? "disabled" : "";
        var isSelected = ctrl.barn.get() === BreedingBarnChoice.DECLINE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${declineId}" ${disabled}>Decline</button>`;

        var tooltip = `What? You've only just met, and she already wants to sleep with you? Something smells fishy.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildNotBarnChoice() {
        var disabled = ctrl.barn.has() ? "disabled" : "";
        var isSelected = ctrl.barn.get() === BreedingBarnChoice.NOTBARN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${notBarnId}" ${disabled}>Not here</button>`;

        var tooltip = global.parser.parse(`What a freaky [i]bitch[/i].
                                           You can't say her proposal has left you uninterested, but you don't want to have sex in a filthy barn like this.
                                           Ask to go back to her place instead.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the shake choice button.
     */
    function buildEagerChoice() {
        var disabled = ctrl.barn.has() ? "disabled" : "";
        var isSelected = ctrl.barn.get() === BreedingBarnChoice.EAGER;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${eagerId}" ${disabled}>Accept</button>`;

        var tooltip = `Hot damn! Does she greet everyone this way?
                       Time to ${ctrl.misc.baseGender() === GenderEnum.MALE ? "take a ride on the town bicycle" : "show her exactly what kobold tongues are capable of"}.`;
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildDeclineChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildNotBarnChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildEagerChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + declineId), $("#" + notBarnId), $("#" + eagerId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerDeclineListener() {
        $("#" + declineId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.barn.set(BreedingBarnChoice.DECLINE);
            loadNextPart(toDecline);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerNotBarnListener() {
        $("#" + notBarnId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.barn.set(BreedingBarnChoice.NOTBARN);
            loadNextPart(toNotBarn);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerEagerListener() {
        $("#" + eagerId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            global.kobold.history.shared.cassidy.capturedby = true;
            ctrl.barn.set(BreedingBarnChoice.EAGER);
            loadNextPart(toEager);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "breedingBarn";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.arrival.get() !== BreedingFarmArrivalChoice.BARN) {
            return "";
        }
        global.ui.applyTheme("hamlet_barn");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += toDecline.build();
        part += toNotBarn.build();
        part += toEager.build();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.arrival.get() !== BreedingFarmArrivalChoice.BARN) {
            return;
        }

        toDecline.registerListeners();
        toNotBarn.registerListeners();
        toEager.registerListeners();

        if (ctrl.barn.has()) {
            return;
        }

        registerDeclineListener();
        registerNotBarnListener();
        registerEagerListener();
    };
}

var BreedingBarnChoice = {
    DECLINE: 1,
    NOTBARN: 2,
    EAGER: 3
};

// END scenes\HamletDay2\BreedingFarm\barn\BreedingBarn.js

// scenes\HamletDay2\BreedingFarm\house\BreedingHouse.js

/**
 * Breakin section of the farm scene.
 * @param {BreedingFarmController} controller The breeding farm controller.
 */
function BreedingHouse(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        ctrl.createMinigame();
        ctrl.minigame.preload();
        return ctrl.minigame.build();
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        return "";
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "breedingBarn";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.arrival.get() !== BreedingFarmArrivalChoice.HOUSE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.arrival.get() !== BreedingFarmArrivalChoice.HOUSE) {
            return;
        }

        ctrl.minigame.registerListeners();
    };
}

var BreedingBarnChoice = {
    DECLINE: 1,
    NOTBARN: 2,
    EAGER: 3
};

// END scenes\HamletDay2\BreedingFarm\house\BreedingHouse.js

// scenes\HamletDay2\BreedingFarm\arrival\BreedingFarmArrival.js

/**
 * Arrival section of the Breeding Farm scene.
 * @param {BreedingFarmController} controller The Breeding Farm controller.
 */
function BreedingFarmArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var barnId = "arrivalBarnButton";
    var houseId = "arrivalHouseButton";

    var toBarn = new BreedingBarn(ctrl);
    var toHouse = new BreedingHouse(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide to pay a visit to the breeding farm, right across the rather broad dirt road from Seki's workshop.
                            Despite having visited Seki dozens of times, you've never really paid much mind to his [i]closest[/i] neighbour.
                            Mostly because you didn't want the old bird to get associated with a [i]thief[/i] such as yourself.
                            That would have hurt his business.
                            But now, you figure none of that really matters anymore.
                        [/p]
                        [p]
                            You walk down the small trail cutting through the extensive [i]front lawn[/i] of the breeding farm.
                            At a certain point, it splits into two.
                            One way goes towards the large, fancy house, which you presume whoever runs the farm lives in.
                            The other heads straight to the attached, dilapidated [i]barn[/i].
                            You reckon that's where the actual [i]work[/i] takes place at.
                            Separating living from your livelihood.
                            You wish you had the opportunity to do that.
                        [/p]
                        [p]
                            Alas, not everyone can be a natural born [i]whatever[/i]-breeder, or a farmer or a guard or something.
                            Some people are just [i]rotten apples[/i].
                            People like [b]you[/b].
                            You're a [b]thief[/b], and that's all you'll [b]ever[/b] be.
                            A no-good, lousy deadbeat mooching off of the success and the stupidity of others.
                            You'll [b]never[/b] become a knight.
                            And with your only [i]fence[/i] on the verge of dying,
                            and your debt to one of the major bosses of the shady underworld not yet repaid,
                            even your criminal career is rapidly reaching a dead end.
                            The future, for you, is looking grim.
                        [/p]
                        [p]
                            With a loud sigh, you set your gloomy thoughts aside, and approach the baroque house.
                            Stepping onto the porch, you reach upwards as high as you can, and firmly knock on the door.
                            Not once, not twice, but thrice.
                            As is customary in this part of the world.
                            Your courteous knocking, however, goes unresponded to.
                            You wait for several minutes, and even try a second time, but nobody answers your beckoning call.
                        [/p]
                        [p]
                            You come to the conclusion that the house must be empty.
                            You try opening the door, to see for yourself, but find it locked.
                            Peering through one of the windows, you notice a bunch of lit candles lining the walls.
                            Definitely not [i]abandoned[/i].
                            Whoever lives here must be in the barn, or out back somewhere.
                            Probably the former.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the barn choice button.
     */
    function buildBarnChoice() {
        var disabled = ctrl.arrival.has() ? "disabled" : "";
        var isSelected = ctrl.arrival.get() === BreedingFarmArrivalChoice.BARN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${barnId}" ${disabled} style="width: 180px;">Head to the barn</button>`;

        var tooltip = global.parser.parse(`It would be rude to prance about their property without making yourself known first.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the house choice button.
     */
    function buildHouseChoice() {
        var disabled = ctrl.arrival.has() ? "disabled" : "";
        var isSelected = ctrl.arrival.get() === BreedingFarmArrivalChoice.HOUSE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${houseId}" ${disabled}>Break in</button>`;

        var tooltip = global.parser.parse(`Break into the house. You [i]are[/i] a thief, after all. <span class="todo">Unavailable.</span>`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildHouseChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildBarnChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + barnId), $("#" + houseId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerBarnListener() {
        $("#" + barnId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.arrival.set(BreedingFarmArrivalChoice.BARN);
            loadNextPart(toBarn);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerHouseListener() {
        $("#" + houseId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.arrival.set(BreedingFarmArrivalChoice.HOUSE);
            loadNextPart(toHouse);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = "BFArrival";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toBarn.build();
        part += toHouse.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        toBarn.registerListeners();
        toHouse.registerListeners();

        if (ctrl.arrival.has()) {
            return;
        }

        registerBarnListener();
        //registerHouseListener();
    };
}

var BreedingFarmArrivalChoice = {
    BARN: 1,
    HOUSE: 2
};

// END scenes\HamletDay2\BreedingFarm\arrival\BreedingFarmArrival.js

// scenes\HamletDay2\BreedingFarm\BreedingFarmController.js

/**
 * Controller for the Breeding Farm scene.
 * @param {string} sceneId The id of the scene.
 */
function BreedingFarmController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;

    var autoSaveId = "autoSave";
    var continueId = "continue";


    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function EagerController(key) {
        ChoiceController.call(this, id, key);

        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.decrease(10, "Caught by Cassidy");
                global.kobold.status.exertion.increase(5, "Loins fondled");
            });
            return global.ui.change.buildNotification([
                new Change("Betrayed", "Sanity", -5),
                new Change("Trapped", "Sanity", -5),
                new Change("Loins fondled", "Sexual exhaustion", +5, true)
            ]);
        };
    }

    function TableController(key) {
        ChoiceController.call(this, id, key);

        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.decrease(15, "Sickened by kobold breeding");
            });
            return global.ui.change.buildNotification([
                new Change("Terrified", "Sanity", -10),
                new Change("Sickened", "Sanity", -5)
            ]);
        };
    }

    function InjectionController(key) {
        ChoiceController.call(this, id, key);

        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.physical.decrease(5, "Injection");
                global.kobold.status.mental.decrease(5, "Injection");
            });
            return global.ui.change.buildNotification([
                new Change("Injection", "Health", -5),
                new Change("Injection", "Sanity", -5)
            ]);
        };
    }

    function TransformationController(key) {
        ChoiceController.call(this, id, key);

        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.decrease(10, "Transformation");
                if (global.kobold.body.dick) {
                    global.kobold.body.growDick(global.kobold.body.dick - 1);
                } else {
                    global.kobold.body.growBreasts(global.kobold.body.breasts + 1);
                }
            });
            return global.ui.change.buildNotification([
                new Change("Transformation", "Sanity", -10)
            ]);
        };
    }

    function BarnResistController(key) {
        ChoiceController.call(this, id, key);

        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.increase(15, "Resisted transformation");
            });
            return global.ui.change.buildNotification([
                new Change("Resisted transformation", "Sanity", +15)
            ]);
        };
    }

    function BarnTableRapeController(key) {
        ChoiceController.call(this, id, key);

        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                if (global.kobold.body.dick) {
                    global.kobold.body.cockslitVirgin = false;
                } else {
                    global.kobold.body.vaginalVirgin = false;
                }
                global.kobold.status.physical.decrease(5, "Rough sex");
                global.kobold.status.mental.decrease(20, "Raped");
                global.kobold.status.exertion.increase(20, "Raped");
                global.kobold.history.shared.cassidy.tableraped = true;
            });
            return global.ui.change.buildNotification([
                new Change("Rough sex", "Health", -5),
                new Change("Raped", "Sanity", -20),
                new Change("Raped", "Sexual exhaustion", +20, true)
            ]);
        };
    }

    function BarnEscapeController(key) {
        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.increase(10, "Escaped from the barn");
                global.kobold.status.stamina.decrease(25, "Escaped from the barn");
                global.kobold.stats.decreaseStat(StatsEnum.STRENGTH, 1);
            });
            return global.ui.change.buildNotification([
                new Change("Escaped", "Sanity", +10),
                new Change("Escaped", "Stamina", -25),
                new Change("Lasting effects", "Strength", -1),
            ]);
        };
    }

    function UnawareController(key) {
        ChoiceController.call(this, id, key);

        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.increase(10, "Escaped from table");
            });
            return global.ui.change.buildNotification([
                new Change("Escaped table", "Sanity", +10)
            ]);
        };
    }

    function UnawareRunController(key) {
        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.stamina.decrease(25, "Ran from Cassidy");
            });
            return global.ui.change.buildNotification([
                new Change("Ran away", "Stamina", -25)
            ]);
        };
    }

    function UnawareAttackController(key) {
        ChoiceController.call(this, id, key);

        this.change = function (weapon) {
            var changeId = key + "Change";
            if (weapon.stab()) {
                global.kobold.history.doIfNotSet(id, changeId, true, function () {
                    global.kobold.status.mental.decrease(10, "Killed Cassidy");
                });
                return global.ui.change.buildNotification([
                    new Change("Revenge", "Sanity", +10),
                    new Change("Murder", "Sanity", -20)
                ]);
            } else {
                global.kobold.history.doIfNotSet(id, changeId, true, function () {
                    global.kobold.status.mental.increase(10, "Got revenge");
                });
                return global.ui.change.buildNotification([
                    new Change("Revenge", "Sanity", +10)
                ]);
            }
        };
    }

    function UnawareFreeController(key) {
        this.change = function (weapon) {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.increase(30, "Kobold saviour");
            });
            return global.ui.change.buildNotification([
                new Change("Unsung hero", "Sanity", +30)
            ]);
        };
    }

    function UnawareLootController(key) {
        ChoiceController.call(this, id, key);

        this.change = function () {
            var changeId = key + "Change";
            if (global.kobold.history.shared.cassidy.dead) {
                global.kobold.history.doIfNotSet(id, changeId, true, function () {
                    global.kobold.inventory.bag.addMisc(new CassidyHouseKey());
                    global.kobold.inventory.bag.addConsumable(new FemPlus());
                });
                return global.ui.change.buildNotification([
                    new Change("Loot", global.kobold.inventory.createItemWithTooltip(new CassidyHouseKey(), true), +1),
                    new Change("Loot", global.kobold.inventory.createItemWithTooltip(new FemPlus(), true), +1)
                ]);
            } else {
                global.kobold.history.doIfNotSet(id, changeId, true, function () {
                    global.kobold.inventory.bag.addConsumable(new FemPlus());
                });
                return global.ui.change.buildNotification([
                    new Change("Loot", global.kobold.inventory.createItemWithTooltip(new FemPlus(), true), +1)
                ]);
            }
        };
    }

    function BarnForceController(key) {
        ChoiceController.call(this, id, key);
        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.physical.decrease(5, "Light scratches");
                global.kobold.status.stamina.decrease(10, "Bear-hugging a vixen");
            });
            return global.ui.change.buildNotification([
                new Change("Scratched by spurs", "Health", -5),
                new Change("Physical exertion", "Stamina", -10)
            ]);
        };
    }

    function BarnForceFreeController(key) {
        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.increase(30, "Kobold saviour");
            });
            return global.ui.change.buildNotification([
                new Change("Unsung hero", "Sanity", +30)
            ]);
        };
    }

    function BarnForceFuckController(key) {
        this.change = function () {
            var changeId = key + "Change";
            global.kobold.history.doIfNotSet(id, changeId, true, function () {
                global.kobold.status.mental.increase(20, "Righteous retribution");
                global.kobold.inventory.bag.addConsumable(new FemPlus());
            });
            return global.ui.change.buildNotification([
                new Change("Righteous retribution", "Sanity", +20),
                new Change("Loot", global.kobold.inventory.createItemWithTooltip(new FemPlus(), true), +1)
            ]);
        };
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.varanarSceneName = "NightTwo";
    this.varanarFolderName = "Varanar";

    this.misc = new MiscController(id, "");
    this.arrival = new ChoiceController(id, "arrival");
    this.barn = new ChoiceController(id, "barn");
    this.eager = new EagerController("eager");
    this.table = new TableController("table");
    this.injection = new InjectionController("injection");
    this.transformation = new TransformationController("transformation");
    this.barnResist = new BarnResistController("barnResist");
    this.tableRape = new BarnTableRapeController("barnTableRape");
    this.tiedUp = new ChoiceController(id, "barnTied");
    this.barnEscape = new BarnEscapeController("barnEscape");
    this.unaware = new UnawareController("unaware");
    this.unawareRun = new UnawareRunController("unawareRun");
    this.unawareAttack = new UnawareAttackController("unawareAttack");
    this.unawareLoot = new UnawareLootController("unawareLoot");
    this.unawareFree = new UnawareFreeController("unawareFree");
    this.barnForce = new BarnForceController("barnForce");
    this.barnForceFree = new BarnForceFreeController("barnForceFree");
    this.barnForceFuck = new BarnForceFuckController("barnForceFuck");

    this.tea = new ChoiceController(id, "tea");
    this.notbarn = new ChoiceController(id, "notbarn");
    this.postnotbarn = new ChoiceController(id, "postnotbarn");
    this.postnotbarnsex = new ChoiceController(id, "postnotbarnsex");
    this.notbarnloot = new ChoiceController(id, "notbarnloot");
    this.postnotbarnloot = new ChoiceController(id, "postnotbarnloot");

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    this.teaDeclineChange = function () {
        var changeId = "teaDeclineChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.status.mental.decrease(10, "Self-doubt");
        });
        return global.ui.change.buildNotification([
            new Change("Self-doubt", "Sanity", -10)
        ]);
    };

    this.teaDisagreeChange = function () {
        var changeId = "teaDisagreeChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.status.physical.increase(5, "Drank tea");
        });
        return global.ui.change.buildNotification([
            new Change("Tea", "Health", 5)
        ]);
    };

    this.teaDisagreeContinueChange = function () {
        var changeId = "teaDisagreeContinueChange";
        var int = that.misc.intelligence();
        var cha = that.misc.charisma();
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            if (int < 6 && cha < 6) {
                global.kobold.status.mental.decrease(20, "Kobolds are pets");
            } else if (int >= 6 && cha < 6) {
                global.kobold.status.mental.decrease(15, "Lost argument");
            } else if (cha >= 6 && int < 6) {
                global.kobold.status.mental.decrease(15, "Kobolds are inferior");
            } else {
                global.kobold.status.mental.increase(10, "Good debate");
            }
        });

        if (int < 6 && cha < 6) {
            return global.ui.change.buildNotification([
                new Change("Realization", "Sanity", -20)
            ]);
        } else if (int >= 6 && cha < 6) {
            return global.ui.change.buildNotification([
                new Change("Realization", "Sanity", -15)
            ]);
        } else if (cha >= 6 && int < 6) {
            return global.ui.change.buildNotification([
                new Change("Realization", "Sanity", -15)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Held your own", "Sanity", +10)
            ]);
        }
    };

    this.teaAgreeChange = function () {
        var changeId = "teaAgreeChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.history.shared.cassidy.agree = true;
            global.kobold.status.physical.increase(10, "Tea and biscuits");
            global.kobold.inventory.bag.addConsumable(new TeaBiscuit(), 5);
        });
        return global.ui.change.buildNotification([
            new Change("Tea and biscuits", "Health", 10),
            new Change("Parting gift", global.kobold.inventory.createItemWithTooltip(new TeaBiscuit(), true), 5)
        ]);
    };

    this.teaAgreeContinueChange = function () {
        var changeId = "teaAgreeContinueChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.status.mental.increase(10, "Made a friend");
        });
        return global.ui.change.buildNotification([
            new Change("Made a friend", "Sanity", 10)
        ]);
    };

    this.notbarnDownstairsLootChange = function () {
        var changeId = "notbarnDownstairsLootChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.inventory.bag.addMisc(new Silverware());
        });
        return global.ui.change.buildNotification([
            new Change("Stolen goods", global.kobold.inventory.createItemWithTooltip(new Silverware(), true), +1)
        ]);
    };

    this.notbarnUpstairsLootChange = function () {
        var changeId = "notbarnUpstairsLootChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.status.physical.decrease(5, "Fell down");
            global.kobold.inventory.coins += 43;
        });
        return global.ui.change.buildNotification([
            new Change("Tripped", "Health", -5),
            new Change("Loot", "Coins", +43)
        ]);
    };

    this.notbarnLootContinueChange = function () {
        var changeId = "notbarnLootContinueChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.status.stamina.decrease(10, "Post-break-in escape");
        });
        return global.ui.change.buildNotification([
            new Change("Running away", "Stamina", -10)
        ]);
    };

    this.notbarnDisagreeChange = function () {
        var changeId = "notbarnDisagreeChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.history.shared.cassidy.sex = true;
            global.kobold.status.physical.decrease(5, "Thrown to the floor");
        });
        return global.ui.change.buildNotification([
            new Change("Fell down", "Health", -5)
        ]);
    };

    this.notbarnDisagreeSexChange = function () {
        var changeId = "notbarnDisagreeSexChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.body.oralVirgin = false;
            global.kobold.status.exertion.increase(10, "Vixen facesitting");
        });
        return global.ui.change.buildNotification([
            new Change("Intense oral", "Sexual exhaustion", +10)
        ]);
    };

    this.notbarnDisagreeContinueChange = function () {
        var changeId = "notbarnDisagreeContinueChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.status.mental.decrease(15, "Forced femdom");
        });
        return global.ui.change.buildNotification([
            new Change("Used for pleasure", "Sanity", -15)
        ]);
    };

    this.notbarnAgreeChange = function () {
        var changeId = "notbarnAgreeChange";
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.history.shared.cassidy.agree = true;
            global.kobold.history.shared.cassidy.sex = true;
        });
        return "";
    };

    function notbarnAgreeSmallDickSexChange(changeId) {
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.body.dickVirgin = false;
            global.kobold.status.mental.decrease(15, "Inadequate lover");
            global.kobold.status.stamina.decrease(10, "Fucked Cassidy");
            global.kobold.status.exertion.increase(15, "Fucked Cassidy");
            global.kobold.inventory.bag.addConsumable(new mPlus());
        });
        return global.ui.change.buildNotification([
            new Change("Failed to please", "Sanity", -15),
            new Change("One-sided sex", "Stamina", -10),
            new Change("One-sided sex", "Sexual exhaustion", +15, true),
            new Change("Degrading gift", global.kobold.inventory.createItemWithTooltip(new mPlus(), true), +1)
        ]);
    }
    ;

    function notbarnAgreeMediumDickSexChange(changeId) {
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.body.dickVirgin = false;
            global.kobold.status.mental.increase(10, "Pleased the bitch");
            global.kobold.status.stamina.decrease(15, "Fucked Cassidy");
            global.kobold.status.exertion.increase(20, "Fucked Cassidy");
        });
        return global.ui.change.buildNotification([
            new Change("Pleased Cassidy", "Sanity", +10),
            new Change("Pleasant fuck", "Stamina", -15),
            new Change("Pleasant fuck", "Sexual exhaustion", +20, true)
        ]);
    }
    ;

    function notbarnAgreeBigDickSexChange(changeId) {
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.body.dickVirgin = false;
            global.kobold.status.mental.increase(15, "Broke the bitch");
            global.kobold.status.stamina.decrease(20, "Demolished Cassidy");
            global.kobold.status.exertion.increase(25, "Demolished Cassidy");
        });
        return global.ui.change.buildNotification([
            new Change("Broke the bitch", "Sanity", +15),
            new Change("Intense sex", "Stamina", -20),
            new Change("Intense sex", "Sexual exhaustion", +25, true)
        ]);
    }
    ;

    function notbarnAgreeFemaleSexChange(changeId) {
        global.kobold.history.doIfNotSet(id, changeId, true, function () {
            global.kobold.status.exertion.increase(20, "Eaten out");
        });
        return global.ui.change.buildNotification([
            new Change("Eaten out", "Sexual exhaustion", +20, true)
        ]);
    }
    ;

    this.notbarnAgreeSexChange = function (dick) {
        var changeId = "notbarnAgreeSexChange";
        if (!dick) {
            return notbarnAgreeFemaleSexChange(changeId);
        }
        if (dick < SizeEnum.AVERAGE) {
            return notbarnAgreeSmallDickSexChange(changeId);
        }
        if (dick < SizeEnum.LARGER) {
            return notbarnAgreeMediumDickSexChange(changeId);
        }
        return notbarnAgreeBigDickSexChange(changeId);
    };

    this.minigame;

    this.createMinigame = function () {
        that.minigame = new LockPickingMinigame(id, "breakinMinigame", 10, 2);
    };

    /*
     * Autosaves the game if necessary.
     */
    this.autosave = function () {
        global.kobold.history.doIfNotSet(id, autoSaveId, true, function () {
            global.saveManager.autoSave();
        });
    };

    /**
     * Signals the scene has been continued.
     */
    this.continue = function () {
        global.kobold.history.setKey(id, continueId, true);
    };

    /**
     * Returns whether or not the scene has continued.
     * @returns {boolean} True if the scene has continued, false otherwise.
     */
    this.continued = function () {
        return global.kobold.history.has(id, continueId);
    };
}

// END scenes\HamletDay2\BreedingFarm\BreedingFarmController.js

// scenes\HamletDay2\BreedingFarm\BreedingFarm.js

/**
 * Breeding farm. Followup scene of the Hamlet.
 */
function BreedingFarm() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "BreedingFarm";
    var sceneName = "Breeding Farm";
    var sceneSubfolder = "HamletDay2";
    var ctrl = new BreedingFarmController(sceneId);
    var arrival = new BreedingFarmArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]breeding farm[/t]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new BreedingFarm();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\HamletDay2\BreedingFarm\BreedingFarm.js



// END SCENE BreedingFarm



// SCENE FarmerHouse

// scenes\HamletDay2\FarmerHouse\confront\intimidate\FarmerHouseIntimidate.js

/**
 * Confront section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseIntimidate(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toSuccess = new FarmerHouseScared(ctrl);
    var toFail = new FarmerHouseIntimidateFail(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}


    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.confront.get() !== FarmerHouseConfrontChoice.INTIMIDATE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (ctrl.misc.charisma() < 6) {
            part += toFail.build();
        } else {
            part += toSuccess.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.confront.get() !== FarmerHouseConfrontChoice.INTIMIDATE) {
            return;
        }

        if (ctrl.misc.charisma() < 6) {
            toFail.registerListeners();
        } else {
            toSuccess.registerListeners();
        }
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\intimidate\FarmerHouseIntimidate.js

// scenes\HamletDay2\FarmerHouse\confront\intimidate\FarmerHouseIntimidateFail.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHouseIntimidateFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You rise to your tip-toes to seem as large and imposing as you possibly can,
                            before loudly and threateningly [b]shouting[/b] at the woman to [b]stay put[/b] and to [b]keep quiet[/b].
                            Despite your best efforts, you fail to appear very menacing.
                            The human seems unimpressed.
                            She reaches behind her, grabbing a rather sharp-looking ${global.kobold.inventory.createItemWithTooltip(new KitchenKnife)} from a wooden block.
                            [i]Oh no.[/i]
                        [/p]
                        ${global.ui.condFail(`Charisma below 6`)}
                        [p]
                            The housewife menacingly advances upon you, knife in hand.
                            Every step she takes towards you, you take one backwards in fear.
                            Before long, you've been pushed back into the hallway you came from.
                            Unarmed, you could have easily taken her on, but with a knife [i]that big[/i],
                            she could [b]seriously[/b] hurt you, if not [b]kill you[/b] outright.
                        [/p]
                        [p]
                            Your fight or flight instincts kick in, and fighting a crazy lady armed with a sharp cutting tool is [b]not[/b] on your mind.
                            You spot the perfect opportunity to flee when the woman nearly trips over a crooked, protruding floorboard.
                            While she's busy recovering her balance, you turn tail and [b]bolt[/b] towards the rear exit.
                            As you open the door, the enraged human hurls her knife towards you,
                            the steel blade missing you by a hair, lodging itself into the wooden doorframe.
                        [/p]
                        [p]
                            For a moment, you consider retrieving the weapon and turning the tables on the housewife,
                            but the blade is stuck fast, and the woman lets out a shriek for help.
                            You decide to keep running, sprinting into the adjacent potato field.
                            A bunch of farmers emerge from their homes, rushing to the aid of the damsel in distress.
                            Luckily, by the time she's pointed out to them that you're escaping through the acreage,
                            you've already vanished into the small stretch of woodland on the opposite end.
                        [/p]
                        [p]
                            Despite having ditched your pursuers, you continue running all the way back to the eastern Varanar road.
                            The villagers might call the guards to investigate, and should anyone ask,
                            at least here you can [b]pretend[/b] to be a traveller instead of a [b]miserable failure[/b] of a thief.
                            Exhausted, you brace yourself against a nearby tree.
                        [/p]
                        [p]
                            You're glad you managed to get out of that one alive.
                            If you were a bit more intimidating and [b]persuasive[/b], perhaps your threatening appearance would have been taken a bit more seriously.
                            Alas, what's done is done.
                            You can't win [i]every[/i] confrontation you get into, and at least you didn't get [b]caught[/b] or [b]cut[/b].
                            You've gained nothing, but you haven't lost much either, besides an opportunity.
                            Your newfound fear of kitchen knives is sure to vanish in due time.
                        [/p]
                        [p]
                            Your brief period of respite over, you continue on the way to Varanar.
                            Hopefully, the humans there will be less [i]ballsy[/i].
                            That's what the hamlet's inhabitants constantly claim, anyway.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        global.kobold.history.shared.farmerhouse.failed = true;
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.intimidateFail();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\intimidate\FarmerHouseIntimidateFail.js

// scenes\HamletDay2\FarmerHouse\confront\loot\FarmerHouseFullLoot.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHouseFullLoot(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseLootContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With the woman out of the picture, your way into the rest of the house is cleared.
                            It'd be a shame to pass up an opportunity like this just because you can't control your libido.
                            You came here to steal stuff, and that's exactly what you are going to do.
                            Better make it fast, though.
                            While you tied her up good, she may still escape if given plenty of time.
                        [/p]
                        [p]
                            You step into the dining room.
                            Nothing worth a damn in here, at first glance.
                            You keep moving, entering the living room, or at least what's supposed to pass for it.
                            The room is dominated by a single padded chair, the most comfortable piece of furniture you've come across in the house, so far.
                            It stands before a small fireplace, probably the only source of [i]warmth[/i] apart from the oven in the kitchen.
                            Still, nothing interesting to take.
                        [/p]
                        [p]
                            There's only one room left, which you assume to be the bedroom.
                            Opening the shabby wooden door, your suspicions are confirmed.
                            Two beds.
                            A big one, and a small one.
                            A dresser, and a nightstand.
                            You decide to go for the latter.
                        [/p]
                        [p]
                            Opening the small drawer, you are surprised to find that it [b]isn't[/b] full of underwear.
                            Instead, the nightstand contains a pair of worn books,
                            atop which lay some out-of-place looking ${global.kobold.inventory.createItemWithTooltip(new FancyGlasses())}.
                            Your eyes open wide, avarice surging to a whole new level.
                            This find more than makes up for having to deal with the housewife.
                            If you knew these farmers had something [b]this[/b] expensive laying around,
                            you wouldn't have stopped attempting to steal from them after your first attempt.
                        [/p]
                        [p]
                            You figure they must be a family heirloom or something.
                            Regardless, they're [b]yours[/b] now.
                            You pocket the valuable item, and start thinking of a way out of this place.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.fullLoot.continued() ? "disabled" : "";
        var isSelected = ctrl.fullLoot.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.fullLoot.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.tied.get() !== FarmerHouseTiedUpChoice.LOOT) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.fullLoot.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.tied.get() !== FarmerHouseTiedUpChoice.LOOT) {
            return;
        }

        next.registerListeners();

        if (ctrl.fullLoot.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\loot\FarmerHouseFullLoot.js

// scenes\HamletDay2\FarmerHouse\confront\loot\FarmerHouseLootContinue.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHouseLootContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function buildFullLoot() {
        var toReturn = `[p]
                            You exit the bedroom, returning to the living room.
                            Just going out the front door would be risky.
                            [b]Too[/b] risky.
                            You make your way back through the dining room into the kitchen.
                            The woman has managed to free her arms, and she is currently in the middle of untying herself.
                            Looks like your time here is running out, and if you don't escape before she gets the rope out of her mouth,
                            you'll be [b]hunted down[/b] by whoever comes to her aid.
                            You bolt past her into the adjacent hallway, and exit the building the same way you entered it: quietly, and through the back.
                        [/p]
                        [p]
                            The moment you close the door behind you, a cry for help resonates from within the building.
                            You weren't planning on sticking around anyway.
                            You sprint into the adjacent potato field, and after crossing it with some difficulty,
                            you escape into a small patch of woodland with nobody but the housewife having seen you.
                        [/p]
                        [p]
                            Still, just because you [i]may[/i] not be getting chased, doesn't mean you don't expeditiously make your way back to the eastern Varanar road.
                            The villagers might call the guards to investigate, and should anyone ask,
                            at least while on the road you can [b]pretend[/b] to be a traveller instead of a [b]housewife-tying bandit[/b].
                            You're glad you didn't leave the farmer's house empty-handed.
                            Thank goodness you invested some of your coin in a coil of rope.
                        [/p]
                        [p]
                            Reaching the road, you continue on your way to Varanar.
                            Hopefully, your streak of good luck will continue in the city.
                            And maybe you won't even have to tie anyone up, the next time you attempt to break in somewhere.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    function buildPartialLoot() {
        var toReturn = `[p]
                            The moment you close the door behind you, a cry for help resonates from within the building.
                            You weren't planning on sticking around anyway.
                            You sprint into the adjacent potato field, and after crossing it with some difficulty,
                            you escape into a small patch of woodland with nobody but the housewife having seen you.
                        [/p]
                        [p]
                            Still, just because you [i]may[/i] not be getting chased, doesn't mean you don't expeditiously make your way back to the eastern Varanar road.
                            The villagers might call the guards to investigate, and should anyone ask,
                            at least while on the road you can [b]pretend[/b] to be a traveller instead of a [b]housewife-threatening bandit[/b].
                            You're glad you didn't leave the farmer's house empty-handed,
                            even if the loot you managed to score isn't as great as you'd hoped it would be.
                            You'd rather have some food than nothing at all.
                        [/p]
                        [p]
                            Reaching the road, you continue on your way to Varanar.
                            Hopefully, there will be some better pickings in the city.
                            And maybe you won't even have to intimidate anyone, the next time you attempt to break in somewhere.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        if (ctrl.tied.get() === FarmerHouseTiedUpChoice.LOOT) {
            return buildFullLoot();
        } else {
            return buildPartialLoot();
        }
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.partialLoot.continued() && !ctrl.fullLoot.continued()) {
            return "";
        }

        global.kobold.history.shared.farmerhouse.stole = true;
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.partialLoot.continued() && !ctrl.fullLoot.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\loot\FarmerHouseLootContinue.js

// scenes\HamletDay2\FarmerHouse\confront\loot\FarmerHousePartialLoot.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHousePartialLoot(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseLootContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With the woman in check, you decide it's time to do what you came here to do: loot whatever you can.
                            Reminding her once more to [b]stay down[/b], you step towards the kitchen counter.
                            Keeping an eye on the housewife, you begin to feel around for things to steal.
                            Sadly, the countertop contains pretty much [b]nothing[/b] of value.
                            The only thing your hand comes across is the ${global.kobold.inventory.createItemWithTooltip(new Bread())},
                            the ${global.kobold.inventory.createItemWithTooltip(new Butter())}
                            and the ${global.kobold.inventory.createItemWithTooltip(new Ham())}
                            the woman had brought with her.
                        [/p]
                        [p]
                            You can tell that the housewife, while still pacified for the moment being, is just [b]waiting[/b] for you to slip up,
                            so she can run, scream for help or perhaps even [b]attack[/b] you.
                            You don't dare go into the next room to see if the pickings there are better.
                            Not wanting to leave here empty-handed either, you decide to take the peasant family's food.
                        [/p]
                        [p]
                            You pocket the bread, the butter and the ham.
                            Frustrated that you weren't able to steal more, but afraid of what the woman might do, should you leave her out of your sight,
                            you slowly back away into the hallway from which you came.
                            After casting one last menacing glare her way, you face away from the housewife, and [b]bolt[/b] out of the backdoor as fast as you can.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.partialLoot.continued() ? "disabled" : "";
        var isSelected = ctrl.partialLoot.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.partialLoot.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.scared.get() !== FarmerHouseScaredChoice.LOOT) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.partialLoot.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.scared.get() !== FarmerHouseScaredChoice.LOOT) {
            return;
        }

        next.registerListeners();

        if (ctrl.partialLoot.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\loot\FarmerHousePartialLoot.js

// scenes\HamletDay2\FarmerHouse\confront\rape\FarmerHouseSexContinue.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseSexContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var toReturn = `[p]
                            ${dick ? `You give yourself a few final strokes,
                            dribbling a last few bits of spunk down on top of the woman.` : `You wipe your loin with the flat of your hand,
                            sprinkling a last bit of feminine arousal atop the woman.`}`;

        if (pants || underwear) {
            toReturn += ` You quickly redress, your garments soaking up the remnants of your excitement.`;
        }

        toReturn += `       With your lust sated, you decide to leave before the unconscious human wakes up.
                        [/p]
                        [p]
                            You return to the hallway from whence you came, and you slip out the back door, closing it behind you.
                            Assuming the housewife could be waking up any time now, you jog into the adjacent potato field, and after crossing it with some difficulty,
                            you escape into a small patch of woodland with nobody having seen you.
                        [/p]
                        [p]
                            Still, just because you [i]may[/i] not be getting chased, doesn't mean you don't expeditiously make your way back to the eastern Varanar road.
                            The villagers might call the guards to investigate when they find one of their own having been abused, and should anyone ask,
                            at least while on the road you can [b]pretend[/b] to be a traveller instead of a [b]rapist[/b].
                            You're glad you got to show that human who's boss.
                            You really needed something to lift your spirit, and turning the tables on society itself did just that.
                        [/p]
                        [p]
                            Reaching the road, you continue on your way to Varanar.
                            Hopefully, the city will be filled with equally meek human women for you to abuse.
                        [/p]`;

        if (!dick && ctrl.wrestle.get() === FarmerHouseWrestleSuccessChoice.RAPE) {
            toReturn = `[p]
                            Arousal fading and frustration rising in its stead, you pull yourself to your feet.
                            ${pants || underwear ? "You quickly redress and let out a loud sigh." : "You let out a loud sigh."}
                            You consider kicking the unconscious woman,
                            but waking her up and having her scream together the entire neighbourhood wouldn't really fix anything.
                            All she had to do was stay awake while you abused her butt, and she couldn't even do that.
                            Stupid human.
                        [/p]
                        [p]
                            Defeated and put off, you return to the hallway from whence you came.
                            You slip out the back door, closing it behind you.
                            Assuming the housewife could be waking up any time, you jog into the adjacent potato field, and after crossing it with some difficulty,
                            you escape into a small patch of woodland with nobody having seen you.
                        [/p]
                        [p]
                            Still, just because you [i]may[/i] not be getting chased, doesn't mean you don't expeditiously make your way back to the eastern Varanar road.
                            The villagers might call the guards to investigate when they find one of their own having been abused, and should anyone ask,
                            at least while on the road you can [b]pretend[/b] to be a traveller instead of a [b]butt-fingering molester[/b].
                            You're glad you [i]somewhat[/i] got to show that human who's boss,
                            but quite upset that she fainted before you had a chance to really [b]enjoy[/b] yourself.
                        [/p]
                        [p]
                            Reaching the road, you continue on your way to Varanar.
                            Hopefully, the city will have some human women that can take a few fingers up the ass without passing out.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.rape.continued()) {
            return "";
        }

        global.kobold.history.shared.farmerhouse.raped = true;
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        //part += ctrl.runChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.rape.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\rape\FarmerHouseSexContinue.js

// scenes\HamletDay2\FarmerHouse\confront\rape\FarmerHouseForcedOral.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHouseForcedOral(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseSexContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var toReturn = `[p]
                            The feeling of being in command inspires a sadistic urge inside of you.
                            A certain [b]lust[/b] is growing, enhanced by the sight of a woman on her knees; a human nonetheless.
                            ${dick ? `You can feel your member stirring within your cockslit, its tip already protruding.
                                    The very [b]thought[/b] of abusing the housewife is causing your loins to swell.
                                    ` : `You can feel your loins moisten at the very [b]thought[/b] of abusing the housewife.`}
                            For a brief moment, you remember the [i]actual[/i] reason you came here,
                            but a few meagre scraps of loot pale in comparison to the amount of [b]fun[/b] you can have with the woman.
                        [/p]
                        [p]
                            A wicked smile creeps across your face as you ogle the housewife.
                            Her linen shirt, while well-stitched, is not nearly thick enough to hide the distinct, flesh-like color of her bare body underneath.
                            Her sizable breasts, obviously unsupported, are large enough to slightly stretch the fabric of her somewhat undersized shirt, pulling it taut.
                            You fantasize about her topless, sewing together a new shirt,
                            ending up making it a [i]little bit[/i] too small because she was running out of yarn.
                            Or maybe her husband just likes her dressing like a [b]cheap whore[/b].
                        [/p]
                        [p]
                            You tell the woman once again to [b]stay down[/b], reminding her of what you'll do to her should she move.
                            She seems fully pacified and compliant.
                            You advance upon her, `;

        if (pants && underwear) {
            toReturn += `lowering your ${global.kobold.inventory.createItemWithTooltip(pants)} and your ${global.kobold.inventory.createItemWithTooltip(underwear)} mid-stride,
                         a horrified look forming on the woman's face as your clearly aroused loins are made bare to her.`;
        } else if (pants) {
            toReturn += `lowering your ${global.kobold.inventory.createItemWithTooltip(pants)} mid-stride,
                         a horrified look forming on the woman's face as your clearly aroused loins are made bare to her.`;
        } else if (underwear) {
            toReturn += `lowering your ${global.kobold.inventory.createItemWithTooltip(underwear)} mid-stride,
                         a horrified look forming on the woman's face as your clearly aroused loins are made bare to her.`;
        } else {
            toReturn += `a horrified look forming on the woman's face as she becomes aware of your clearly aroused loins.`;
        }

        toReturn += `       You can feel her hesitance and her urge to [b]resist[/b] rise, yet she's still far too intimidated by your approaching presence to react.
                        [/p]
                        [p]
                            After a few slow steps, you find yourself standing directly in front of the woman, her face mere inches away from your exposed groin.
                            You can nearly [b]feel[/b] her warm breath on your ${dick ? "semi-erect length" : "needy slit"}.
                            She tries to turn away to the side, yet you grab her head with both your hands and firmly keep it in place,
                            your sharp claws further incentivizing her compliance.
                        [/p]`;

        if (dick) {
            toReturn += `[p]
                            A small growl escapes your lips.
                            You move one of your hands to clutch the housewife's dark brown hair,
                            freeing up the other to grip your growing member with.
                            You give yourself a few much-needed strokes, coaxing the rest of your length from your slit,
                            the tip of your member ending up directly beneath the human's nose, pressing against her top lip.
                        [/p]
                        [p]
                            The housewife knows what's coming, yet still she tries to delay the inevitable.
                            You demand she open her mouth, but she stubbornly keeps her lips sealed.
                            Aggravated, you cruelly [b]twist[/b] the peasant's hair, her mouth opening to voice her anguish,
                            only for you to [b]ram[/b] your member past her lips, silencing her pained grunt.
                        [/p]
                        [p]
                            In one smooth rolling motion of your hips, you slide the entirety of your length into the woman's mouth.
                            Fully bottomed out, her face buried in your scaly groin,
                            you take a moment to savor the feeling of your shaft being enveloped by the warm wetness of a human's mouth.
                            You warn her that if she [b]bites[/b],
                            you'll [b]wring her neck[/b] and the last thing she'll experience before she [b]dies[/b] will be the taste of kobold cock mixed with blood.
                            She glares up at you, sheer [b]hatred[/b] clear in her eyes.
                            Still, you get the feeling that she got the message.
                        [/p]
                        [p]
                            Making sure her head isn't going anywhere, you slowly start to rock back and forth, thrusting in and out of the woman's mouth.
                            While her lips are simply [i]resting[/i] around your shaft, instead of actively suckling,
                            the feeling of them sliding up and down your length excites you to no end.
                            Her tongue remains idle, yet you thrust at just the right angle to frot your cock along its entirety,
                            [b]forcing[/b] her to taste every inch of your manhood
                        [/p]
                        [p]
                            The human's bountiful saliva slickens your member, and in turn her lips, allowing you to gradually increase your pace.
                            It isn't long until you're outright [b]fucking[/b] her face, deftly pumping your cock in and out of her mouth,
                            your spit-covered abdomen smacking against her face every time you bottom out, a few thick,
                            slobbery strings connecting the human's skin to your scales.
                        [/p]
                        [p]
                            While the facefucking itself is pleasurable,
                            the thing that [b]sates[/b] your lust the most is the fact that the person sucking your dick is a [b]human[/b].
                            Just [b]thinking[/b] about how [i]wrong[/i] this situation is excites you to the point of barely being able to contain yourself.
                            You wildly [b]hump[/b] away at the housewife's face.
                            Various demeaning and degrading insults escape your lips in between heated grunts and a constant, low-pitched growling.
                        [/p]
                        [p]
                            Finally, you can take no more.
                            You grab the woman's hair with both of your hands, and forcibly [b]pull[/b] her onto your length as you reach your end,
                            squirting rope after rope of thick, heady spunk onto the far-end of her tongue, some of your cum shooting directly down her throat.
                            The woman struggles against your hurtful grasp, but you hold her down for the entire duration of your orgasm,
                            wasting not a single drop of potent kobold semen on her face.
                        [/p]
                        [p]
                            Your climax fading, you slowly remove yourself from the housewife's mouth.
                            It seems that while you were lost in pleasure, her struggles must have ceased somewhere along the road.
                            In fact, as you let go of your grip on her hair, the woman falls limply to the floor.
                            Fainted.
                            This entire ordeal must have been a little too much for her to process.
                            Heh.
                            Human can't handle kobold cock.
                        [/p]`;
            if(ctrl.misc.dickVirgin()) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        } else {
            toReturn += `[p]
                            A small growl escapes your lips.
                            You move one of your hands to clutch the housewife's dark brown hair,
                            while grabbing the back of her head with the other, roughly pulling her face in between your thighs.
                            You let out a small, pleasured grunt as you can feel the human's nose press in between your folds.
                        [/p]
                        [p]
                            The housewife knows what's coming, yet still she tries to delay the inevitable.
                            You tell her to start [b]licking[/b], but she stubbornly keep her mouth closed.
                            Aggravated, you cruelly [b]twist[/b] the peasant's hair, her mouth opening to voice her anguish,
                            only for you to start [b]grinding[/b] your groin against her parted lips.
                        [/p]
                        [p]
                            The woman glares up at you, with tears in her eyes.
                            You cease your movements for a moment, to gauge her attitude.
                            Realizing resistance is futile, she slowly extends her tongue, and plants a single, tentative lick across the entire length of your slit.
                            The feeling of the warm, wet, slimy appendage pressing against your folds fuels your desire.
                        [/p]
                        [p]
                            You [b]demand[/b] she keep going, and the fearful woman complies to your whims.
                            She begins to lap at your cunt, shallowly scraping her tongue along your slit.
                            Letting out a soft moan, you slightly part your thighs, giving her better access to your loins.
                            Simultaneously, you force the human's head [b]deeper[/b] in between your legs, telling her to go [b]deeper[/b] with a single word.
                        [/p]
                        [p]
                            Once again, she complies, licking you with a bit more [b]conviction[/b],
                            her tongue now [b]parting[/b] your folds instead of simply rubbing against your cunt,
                            collecting more than a little of your arousal in the process.
                            It's not just the woman's tongue that's [b]sating[/b] your [b]need[/b].
                            The sheer fact that a [b]human[/b] is eating you out spurs your libido far more than her lapping ever could.
                        [/p]
                        [p]
                            You recommence your rocking back and forth, outright [b]humping[/b] the housewife's face,
                            slathering her pristine skin with feminine excitement.
                            You tighten your grip on her head, feeling your end rapidly approaching.
                            As her tongue accidentally slips inside of you, probably due to a miscalculation on her end, you are sent over the edge.
                            You let out a loud, satisfied sigh as you lower your entire body onto the human's face,
                            practically [b]sitting[/b] atop it as arousal [b]gushes[/b] from your slit, completely [b]soaking[/b] the woman's visage.
                        [/p]
                        [p]
                            The housewife struggles for a few moments, but you keep her pinned in place using both your scaly thighs and your hands.
                            As your orgasm fades, so too do her protests.
                            You let go of her face, and the woman instantly drops to the ground.
                            Fainted.
                            This entire ordeal must have been a little too much for her to process.
                            Heh.
                            Guess these rural humans aren't as tough as they claim to be.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.rape.continued() ? "disabled" : "";
        var isSelected = ctrl.rape.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.rape.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.scared.get() !== FarmerHouseScaredChoice.ORAL) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.rape.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.scared.get() !== FarmerHouseScaredChoice.ORAL) {
            return;
        }

        next.registerListeners();

        if (ctrl.rape.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\rape\FarmerHouseForcedOral.js

// scenes\HamletDay2\FarmerHouse\confront\rape\FarmerHouseTiedUpRape.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHouseTiedUpRape(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseSexContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();

        var toReturn = `[p]
                            For the first time in your life, you have a human completely at your mercy.
                            Their kind has pushed you around for as long as you can remember.
                            They treat you like garbage.
                            Like complete and utter [b]trash[/b].
                            For once, you are not subservient to humans, but domineering one of them.
                            And you intend to take full advantage of this opportunity.
                        [/p]
                        [p]
                            Lewd thoughts had crept into your mind from the moment you first laid eyes upon the housewife.
                            Her long, brown hair, her curvaceous frame,
                            her fair face and her considerable rack have been [b]taunting[/b] you ever since she entered the kitchen.
                            Watching the woman struggle and squirm in a feeble attempt to free herself from her bondage only serves to fuel your desire.
                            A wicked grin creeps onto your face.
                            You lustily lick your lips, `;

        if (pants && underwear) {
            toReturn += `lowering your ${global.kobold.inventory.createItemWithTooltip(pants)} and your ${global.kobold.inventory.createItemWithTooltip(underwear)}, `;
        } else if (pants) {
            toReturn += `lowering your ${global.kobold.inventory.createItemWithTooltip(pants)}, `;
        } else if (underwear) {
            toReturn += `lowering your ${global.kobold.inventory.createItemWithTooltip(underwear)}, `;
        }

        toReturn += `       your ${dick ? `fully erect cock` : `moist loins`} a clear indicator of your arousal.
                            This is going to be [b]fun[/b].
                        [/p]`;

        if (dick) {
            toReturn += `[p]
                            You kneel down next to the human, and crudely [b]grope[/b] her ass with one of your hands.
                            She freezes up for a moment, then begins to struggle, kick and scream as best as she can through the rope covering her mouth.
                            Luckily, even despite her frantic wriggling, your bindings hold.
                            She isn't going anywhere, which means you can take your sweet time feeling up her ass through her shabby leather pants.
                        [/p]
                        [p]
                            You lecherously squeeze the woman's soft, garment-clad posterior.
                            While she clearly hasn't had the easiest of lives so far,
                            her butt feels like it's been sitting idly on a comfortable throne for the most part of it.
                            Soft and squishy, truly an ass fit for an expensive [b]whore[/b] instead of a run-down housewife.
                            Your free hand finds its way to your groin, and you begin playing with yourself,
                            fantasizing about fucking her, your pelvis [b]smacking[/b] against her rear, making it [b]jiggle[/b] and [b]bounce[/b].
                        [/p]
                        [p]
                            Every touch and every thought entices you more and more.
                            You didn't tie her up [i]just[/i] to get a feel, though.
                            You [b]want[/b], no, you [b]need[/b] more.
                            Gripping the waistband of her pants, you tug them down as far as the rope will allow,
                            exposing the entirety of her behind, the waistband ending up nestling snugly just below her buttcheeks,
                            giving you an excellent rear-view of her loins.
                            Underwear must be too expensive for this family, for she is wearing none.
                            What a [b]slut[/b].
                        [/p]
                        [p]
                            You get down on the floor next to the woman, guiding your rock-hard member in between her legs,
                            pressing your tapered length into the small gap separating her thick thighs.
                            While you can't quite [i]fuck her[/i] with her legs bound this tightly together,
                            the warm, soft-skinned flesh of her thighs enveloping your manhood is [b]more[/b] than enough stimulation for you.
                            Gripping the housewife's waist to steady yourself, you slowly begin to rock back and forth,
                            grinding your cock along the length of her slit, parting her folds with the the top-side of your dick,
                            while her bare upper legs squeeze down on your shaft, milking it far better than your own hand ever could.
                        [/p]
                        [p]
                            Feeling your cock in between her thighs, the woman begins to squirm anew, desperate to free herself and throw you off of her.
                            Little does she know that the small motions of her legs only serve to [b]massage[/b] your shaft,
                            pleasuring you further instead of [b]dissuading[/b] you from toying with her body.
                            A steady stream of precum oozing from the tapered tip of your member serves to smoothen the peasant's already pristine flesh,
                            allowing you to pick up the pace, turning your slow grinding into [i]actual[/i] thighfucking,
                            the fantasies of your hips smacking against her ass becoming reality.
                        [/p]
                        [p]
                            Realizing she can't shake you off, the housewife ceases her protests,
                            resorting to laying perfectly still as you [b]use[/b] her body to get yourself off,
                            hoping her lack of resistance would speed things along.
                            Her assumption proves to be far from false.
                            Without the constant wriggling, you can up your pace one final time,
                            the loud slapping of scales against skin echoing through the decrepit house as you hump away to your heart's delight.
                        [/p]
                        [p]
                            You can feel your end rapidly approaching,
                            stimulated not only by the thighs squeezing around your length and the warm cunt your member rubs against with every thrust,
                            but also by the very situation itself.
                            You are fucking the thighs of a [b]human[/b], a human that isn't even [b]struggling[/b] anymore.
                            Like she realizes that [b]you[/b] are in command here.
                            Her unwillingness turning into emotionless reluctance excites you to no end,
                            stroking the sadistic side of your personality in [b]all[/b] the right ways.
                        [/p]
                        [p]
                            You reach up to grab one of the woman's clothed boobs, kneading it tightly as you reach your end,
                            your frantic fucking slowing down considerably, ending up in a series of firm,
                            full-length thrusts to [b]squeeze[/b] every bit of cum out of your internal balls.
                            The tip of your member protruding from the front-end of the housewife's groin,
                            you squirt rope after rope of thick, heady cum onto the wooden floor, a few of which get caught on the waistband of her pants.
                        [/p]
                        [p]
                            Your intense orgasm coming to an end,
                            you make sure to wipe a bit of oozing spunk off directly onto the human's cunt as you retract your length from in between her thighs.
                            It is only when you rise to your feet, that you notice the woman isn't even [i]conscious[/i] anymore.
                            She must have passed while you were having your way with her.
                            Fainted from emotional distress.
                            Guess these rural humans aren't as tough as they claim to be.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You kneel down next to the human, and crudely [b]grope[/b] her ass with one of your hands.
                            She freezes up for a moment, then begins to struggle, kick and scream as best as she can through the rope covering her mouth.
                            Luckily, even despite her frantic wriggling, your bindings hold.
                            She isn't going anywhere, which means you can take your sweet time feeling up her ass through her shabby leather pants.
                        [/p]
                        [p]
                            You lecherously squeeze the woman's soft, garment-clad posterior.
                            While she clearly hasn't had the easiest of lives so far,
                            her butt feels like it's been sitting idly on a comfortable throne for the most part of it.
                            Soft and squishy, truly an ass fit for an expensive [b]whore[/b] instead of a run-down housewife.
                            Your free hand finds its way to your groin, and you begin playing with yourself,
                            fantasizing about pressing your face in between her near-perfect buttcheeks,
                            and lapping at her twat until she cums all over your face.
                        [/p]
                        [p]
                            Every touch and every thought entices you more and more.
                            You didn't tie her up [i]just[/i] to get a feel, though.
                            You [b]want[/b], no, you [b]need[/b] more.
                            Gripping the waistband of her pants, you tug them down as far as the rope will allow,
                            exposing the entirety of her behind, the waistband ending up nestling snugly just below her buttcheeks,
                            giving you an excellent rear-view of her loins.
                            Underwear must be too expensive for this family, for she is wearing none.
                            What a [b]slut[/b].
                        [/p]
                        [p]
                            More importantly, lowering her pants has made bare a large stretch of rope-less leg,
                            just big enough for you to...[i]maybe[/i]...
                        [/p]
                        [p]
                            You step over the human, one leg on either side of her frame, before squatting down,
                            pressing your sopping wet slit against the side of her leg.
                            Using one hand to keep the woman pinned down, and the other to stabilize yourself with,
                            you slowly begin to rub your groin alongside the smooth surface of her skin.
                            The sensation is surprisingly pleasurable, more so than you had anticipated.
                            This was a good idea.
                        [/p]
                        [p]
                            Your absolutely [b]drenched[/b] cunt leaves a trail of arousal in its wake,
                            slickening the human's skin, allowing you to up your pace.
                            You begin to roll your hips back and forth, grinding against the side of her leg,
                            innate warmth enhanced by sheer [b]friction[/b] heating up your feminine excitement.
                        [/p]
                        [p]
                            Much more than the simple act of humping against the housewife's leg,
                            the feeling of having a [b]human[/b] underneath you spurs your libido.
                            To make things even better, she's not even [b]struggling[/b] against your hold.
                            She could very easily knock you off of her, if she tried, although that would only end bad for her.
                            Still, being in control of someone instead of being the meek person [b]being[/b] controlled feels [b]oh so[/b] good, for once.
                            It strokes the sadistic side of your personality in [b]all[/b] the right ways.
                        [/p]
                        [p]
                            Your soft moans and pleasured grunts echo throughout the decrepit house.
                            You lean forwards, pressing against the side of the woman's body to make your movements a bit more natural,
                            allowing you to [b]ride[/b] her skin like a professional kobold whore would a cock.
                            You grip one of her clothed tits, kneading it to your heart's content, toying with it without so much as an [b]objection[/b] from the housewife.
                        [/p]
                        [p]
                            Your breathing intensifies.
                            You can [b]feel[/b] your end rapidly approaching, your legs clutching tightly around hers.
                            You press your groin as tightly down onto the human's leg as you possibly can,
                            rubbing back and forth just a few more times before reaching your end.
                            You can [b]barely[/b] prevent yourself from [b]screaming[/b] in pleasure as feminine fluids [b]gush[/b] from your slit,
                            completely [b]soaking[/b] the woman's legs, a small puddle forming on the wooden floor below.
                        [/p]
                        [p]
                            Rising from the housewife's leg, you give yourself a few more strokes with your hand, basking in super-sensitive, post-orgasmic bliss.
                            With a loud, pleased sigh, you slowly rise to your feet.
                            It is only then, that you notice the woman isn't even [i]conscious[/i] anymore.
                            She must have passed while you were having your way with her.
                            Fainted from emotional distress.
                            So that's why she didn't struggle.
                            Guess these rural humans aren't as tough as they claim to be.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.rape.continued() ? "disabled" : "";
        var isSelected = ctrl.rape.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.rape.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.tied.get() !== FarmerHouseTiedUpChoice.RAPE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.rape.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.tied.get() !== FarmerHouseTiedUpChoice.RAPE) {
            return;
        }

        next.registerListeners();

        if (ctrl.rape.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\rape\FarmerHouseTiedUpRape.js

// scenes\HamletDay2\FarmerHouse\confront\rape\FarmerHouseWrestleRape.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseWrestleRape(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var continueId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseSexContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        
        var toReturn = `[p]
                            This [b]stupid human[/b] thought she could keep you from picking this place clean.
                            She's probably one of those types that thinks kobolds are [i]inferior[/i].
                            The kind of person that would [b]kill[/b] one of your kind without cause and without remorse, if given the tools to do so.
                            But now, the tables have turned.
                            [b]You[/b] aren't at [i]her[/i] mercy, but [b]she[/b] is at [i]yours[/i].
                            And you intend to make her feel every bit as [b]miserable[/b] as her species makes your kind feel [b]every single day[/b].
                        [/p]
                        [p]
                            While keeping the housewife firmly in place, you lean down, pressing your muzzle into the back of her head,
                            burying your snout into her long, brown hair.
                            You take a loud [b]sniff[/b], the scent of unprocessed [b]wheat[/b] and other farm produce filling your nostrils,
                            along with a distinctly [b]feminine[/b] scent buried underneath layers upon layers of a lifetime's worth of hard work.
                        [/p]
                        [p]
                            You brush her hair to the side with the tip of your nose.
                            With the woman's neck exposed, you extend your tongue, and plant a single long, wet lick across her smooth skin,
                            which leaves a trail of saliva in its wake.
                            The human tastes exactly like she smells: [i]rough around the edges[/i], but with a subtle, unmistaken core of femininity,
                            so subtle that species with less-well-defined tastebuds probably wouldn't even [b]notice[/b] it.
                        [/p]
                        [p]
                            In that way, your kind is blessed, just like most canines are.
                            You are more [i]cognitive[/i], able to pick up and separate distinct traces of scents and of tastes.
                            This is both a blessing and a curse, depending on [b]what[/b] you end up smelling or tasting.
                            Right now, it is a [b]godsend[/b], the subtle feminine fragrance and flavour of the woman spurring your libido to new heights.
                        [/p]
                        [p]
                            The housewife seems to be starting to understand what your intentions are.
                            Her struggling intensifies, to the point of nearly throwing you off her back.
                            Using your knee to keep both of her arms in position,
                            you use your now-freed hand to take hold of her hair.
                            Cruelly twisting it seems sufficient to pacify the woman, pain reminding that [b]you[/b] are in charge here.
                            Once her squirming stops, `;
        
        if (pants && underwear) {
            toReturn += `you take the time to lower your ${global.kobold.inventory.createItemWithTooltip(pants)} 
                        and your ${global.kobold.inventory.createItemWithTooltip(underwear)}, your fully aroused loins made bare.`;
        } else if (pants) {
            toReturn += `you take the time to lower your ${global.kobold.inventory.createItemWithTooltip(pants)}, your fully aroused loins made bare.`;
        } else if (underwear) {
            toReturn += `you take the time to lower your ${global.kobold.inventory.createItemWithTooltip(underwear)}, your fully aroused loins made bare.`;
        } else {
            toReturn += `you give your loins a few much-needed strokes.`;
        }
        
        toReturn += `     Time for the real [b]fun[/b] to begin.  
                        [/p]`;
        
        if(dick) {
            toReturn += `[p]
                            You shift around on the human's back, laying yourself down atop her, using your chest to keep her arms pinned.
                            Your groin aligns perfectly with her soft rear, your hard member pressing into her shabby leather pants from behind.
                            You remind the housewife that if she struggles, you'll be forced to [b]hurt[/b] her again.
                            Then, you forcefully [b]tug[/b] her pants down as far as they'll go, her ass fully exposed,
                            the tip of your cock directly in between her buttcheeks.
                            Slut isn't even wearing any underwear.
                            Like she [b]wanted[/b] this to happen all along.
                        [/p]
                        [p]
                            Gripping your member, you guide it deeper in between the human's buttocks, her warm flesh enveloping your length.
                            She writhes around like crazy, yet you manage to keep the lower half of her body somewhat steady,
                            simply by forcing it down against the floor with your all of your might, using your entire frame to keep her in place.
                            You press your tapered tip directly against her [i]other[/i] hole, the one place she's least likely to have ever been fucked.
                            She [b]begs[/b] you not to, her pleas muffled by your palm still covering her mouth.
                            Her miserable whining only serves to spur you on further.
                        [/p]
                        [p]
                            For being such a horrible, kobold-hating [b]human[/b], and for her constant, annoying [b]struggling[/b], she doesn't deserve your mercy.
                            You mercilessly press forwards, slowly forcing your rock-hard length into her tight behind,
                            powering inch after inch of ever-thickening kobold meat into the woman's ass.
                            Her writhing instantly comes to an end.
                            She [b]screams[/b] into your hand, before slumping down onto the ground, utterly [b]broken[/b] and [b]defeated[/b].
                            But her [b]punishment[/b] has only just begun.
                        [/p]
                        [p]
                            You manage to lodge over half of your member inside of her overly-tight hole in a single thrust.
                            You pull back, only to [b]ram[/b] yourself forwards once more, shoving another couple of inches inside.
                            A lusty grunt escapes your lips.
                            [i]Almost there.[/i]
                            A third time, you slip out everything but the tip.
                            With your free hand, you take hold of the woman's hips to steady yourself.
                            Then, you give her the most [b]fiercesome[/b] of thrusts so far, your scaly pelvis [b]smacking[/b] against her rear as you bottom out,
                            sinking the entirety of your length, up to the very base, into her far-too-tight behind.
                            [b]Fuck yes.[/b]
                        [/p]
                        [p]
                            The woman's insides [b]squeeze down[/b] on your shaft like a [b]vice[/b].
                            Still, you need more.
                            You begin to roll your hips back and forth, ever-so-slightly at first, giving her tender,
                            unlubed hole a few small thrusts while the precum oozing from your tip slickens her insides.
                            As the human's behind begins to loosen somewhat, you pick up both the pace and the ferocity of your motions,
                            fucking her ass [b]deeper[/b] and [b]harder[/b].
                            It isn't long until you settle into a steady rhythm, pumping your cock in and out of her,
                            your low growls and the sound of your hips smacking against her behind echoing across the room.
                        [/p]
                        [p]
                            The thing that turns you on the most about this situation isn't the actual [i]sex[/i].
                            Sure, fucking the woman's rear end is pleasurable, but if you really wanted to have a good time, you would have just fucked her cunt instead.
                            No, the thing that arouses you most, is the fact that you're having your way with a [b]human[/b].
                            You're making up for [b]years[/b] of your kind being [b]treaded on[/b], [b]used[/b] and [b]abused[/b].
                            You're punishing her for the misdeeds of her entire [b]species[/b].
                            For the wrongdoings of [b]all[/b] of the races that think they're so much [b]better[/b] than you.
                        [/p]
                        [p]
                            The feeling of finally [b]getting back[/b] at the humans for making your life [b]miserable[/b],
                            for making the lives of [b]all[/b] kobolds a living hell, [b]that[/b] is the thought that spurs your desire the most.
                            Combined with a hole so tight your member barely [b]fits[/b], you don't manage to last very long.
                            Mere [b]minutes[/b] after you first forced yourself into the housewife's ass, you can already feel your end approaching.
                        [/p]
                        [p]
                            Your fucking become more [b]bestial[/b], even more [b]violent[/b] than before.
                            You wildly [b]pound[/b] away at the woman's rear with a series of firm, stabbing thrusts, until finally you can hold yourself no longer.
                            One final time, you [b]slam[/b] yourself inside of her, hilting your member up to the base as you reach your climax,
                            squirting your hot, sticky load deep inside of her guts, filling her up with the very seed of the race she [b]despises[/b] so.
                        [/p]
                        [p]
                            You keep your member firmly in place for the entire duration of your orgasm.
                            Only once the multitude of individual waves of spunk fade into a continuous dribble of jizz,
                            do you decide to pull yourself out of the woman's rear.
                            Retracting your cock, the human's ass is left slightly agape.
                            Mere moments after you remove yourself from her behind, a trickle of spunk begins to ooze from it,
                            running over and alongside her cunt before spilling to the floor below.
                        [/p]
                        [p]
                            Breathing heavily, you pull your focus from the housewife's ass to her visage.
                            You wanted to see the look on her face, but alas, it seems she fainted somewhere between you inserting your member and you pulling out.
                            No longer needing to hold her down, you rise to your feet.
                            Stupid human can't even take a buttfucking without passing out.
                        [/p]`;
            if(ctrl.misc.dickVirgin()) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        } else {
            toReturn += `[p]
                            You shift around on the human's back, laying yourself down atop her, using your chest to keep her arms pinned.
                            You remind the housewife that if she struggles, you'll be forced to [b]hurt[/b] her again.
                            Then, you forcefully [b]tug[/b] her pants down as far as they'll go, her ass fully exposed.
                            Slut isn't even wearing any underwear.
                            Like she [b]wanted[/b] this to happen all along.
                        [/p]
                        [p]
                            You press your free hand in between the woman's buttcheeks, your scaly fingers rubbing against her sensitive rear hole.
                            You intend to [b]punish[/b] her for all the bad things she's done to your kind.
                            You figure that toying with her cunt would be more of a [b]reward[/b].
                            She begs you not to go through with this, yet her voice is muffled by your palm still plastered atop her mouth.
                        [/p]
                        [p]
                            Without mercy, you deftly plunge [b]two[/b] of your fingers into her overly-tight rear,
                            stretching her ass with a pair of unlubed digits.
                            The woman [b]screams[/b] into your hand, before slumping down onto the ground, utterly [b]broken[/b] and [b]defeated[/b].
                            But her [b]torment[/b] has only just begun.
                            Without delay, you begin to pump your fingers in and out of her, thoroughly [b]fucking[/b] her obviously inexperienced butt.
                        [/p]
                        [p]
                            But you aren't done just yet.
                            With a cruel smirk on your face, you force a [b]third[/b] thick finger up her rear, inserting all but your thumb,
                            testing and [b]exceeding[/b] the housewife's limits without ever halting your incessant fingerfucking.
                            You'd hoped to get a rise out of the woman, yet looking up at her face, you notice that she's completely [b]passed out[/b].
                            That's kind of a [b]buzzkill[/b].
                            You retract your fingers from her rear.
                            There's no point to fingerfucking an unconscious human.
                        [/p]`;
        }
        
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.rape.continued() ? "disabled" : "";
        var isSelected = ctrl.rape.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.rape.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if(ctrl.wrestle.get() !== FarmerHouseWrestleSuccessChoice.RAPE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.rape.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.wrestle.get() !== FarmerHouseWrestleSuccessChoice.RAPE) {
            return;
        }
        
        next.registerListeners();

        if (ctrl.rape.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\rape\FarmerHouseWrestleRape.js

// scenes\HamletDay2\FarmerHouse\confront\scared\FarmerHouseScared.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHouseScared(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var oralId = arguments.callee.name + "ButtonOne";
    var tieId = arguments.callee.name + "ButtonTwo";
    var lootId = arguments.callee.name + "ButtonThree";

    var toOral = new FarmerHouseForcedOral(ctrl);
    var toTie = new FarmerHouseTiedUp(ctrl);
    var toLoot = new FarmerHousePartialLoot(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function buildIntimidate() {
        var toReturn = `[p]
                            You rise to your tip-toes to seem as large and imposing as you possibly can,
                            before loudly [b]shouting[/b] at the woman to [b]stay put[/b] and to [b]keep quiet[/b].
                            Your booming voice and your lizard-like appearance seem to overcome your lacking [i]size[/i],
                            making you seem far more threatening than you actually are.
                            The human looks scared out of her mind, a terrified expression on her face as she cowers in place.
                        [/p]
                        ${global.ui.condSucc(`Charisma greater than 5`)}
                        [p]
                            Your intimidation successful, you command the woman to get down on her knees.
                            A twisted form of sadistic pride makes your chest swell as she actually [b]follows[/b] your order.
                            For the first time in your life, a human is doing as [b]you[/b] say, instead of it being the other way around.
                            You can't say you [b]dislike[/b] the feeling of being in charge.
                            It mentally empowers you, makes you feel like you're on top of the world.
                            Is this how humans feel when they boss your kind around?
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    function buildThreaten() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            You tell the woman to keep quiet and stay put.
                            You threaten to [b]hurt[/b] and even [b]kill[/b] her, should she so much as [b]dare[/b] to make a noise.
                            You brandish your ${global.kobold.inventory.createItemWithTooltip(weapon)} in an attempt to appear more dangerous, and it works.
                            The human seems scared out of her mind, a terrified expression on her face as she eyes your weapon.
                        [/p]
                        ${global.ui.condSucc(`${global.kobold.inventory.createItemWithTooltip(weapon, true)} offence greater than 3`)}
                        [p]
                            Your threatening successful, you command the woman to get down on her knees.
                            A twisted form of sadistic pride makes your chest swell as she actually [b]follows[/b] your order.
                            For the first time in your life, a human is doing as [b]you[/b] say, instead of it being the other way around.
                            You can't say you [b]dislike[/b] the feeling of being in charge.
                            It mentally empowers you, makes you feel like you're on top of the world.
                            Is this how humans feel when they boss your kind around?
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        if (ctrl.confront.get() === FarmerHouseConfrontChoice.INTIMIDATE) {
            return buildIntimidate();
        } else {
            return buildThreaten();
        }
    }

    /**
     * Builds and returns the choice button.
     */
    function buildOralChoice() {
        var disabled = ctrl.scared.has() ? "disabled" : "";
        var isSelected = ctrl.scared.get() === FarmerHouseScaredChoice.ORAL;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${oralId}" ${disabled}>Fuck her face</button>`;

        var tooltip = global.parser.parse(`Seeing the woman drop to her knees has awoken some unsavory thoughts. Act on them.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildTieChoice() {
        var hasRope = ctrl.hasRope();
        var disabled = ctrl.scared.has() || !hasRope ? "disabled" : "";
        var isSelected = ctrl.scared.get() === FarmerHouseScaredChoice.TIE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${tieId}" ${disabled}>Tie her up</button>`;

        var tooltip = global.parser.parse(`Requires a rope.`);
        if (hasRope) {
            tooltip = global.parser.parse(`You got this rope for a reason: to [s]hang yourself[/s] tie people up with.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 170px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildLootChoice() {
        var disabled = ctrl.scared.has() ? "disabled" : "";
        var isSelected = ctrl.scared.get() === FarmerHouseScaredChoice.LOOT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${lootId}" ${disabled}>Loot</button>`;

        var tooltip = global.parser.parse(`Loot as best as you can without taking your eyes off of the human.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildOralChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildTieChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildLootChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + oralId), $("#" + lootId), $("#" + tieId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerOralListener() {
        $("#" + oralId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.scared.set(FarmerHouseScaredChoice.ORAL);
            loadNextPart(toOral);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerTieListener() {
        $("#" + tieId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.scared.set(FarmerHouseScaredChoice.TIE);
            loadNextPart(toTie);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerLootListener() {
        $("#" + lootId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.scared.set(FarmerHouseScaredChoice.LOOT);
            loadNextPart(toLoot);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toOral.build();
        part += toTie.build();
        part += toLoot.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        toOral.registerListeners();
        toTie.registerListeners();
        toLoot.registerListeners();

        if (ctrl.scared.has()) {
            return;
        }

        registerOralListener();
        registerTieListener();
        registerLootListener();
    };
}

var FarmerHouseScaredChoice = {
    ORAL: 1,
    TIE: 2,
    LOOT: 3
};

// END scenes\HamletDay2\FarmerHouse\confront\scared\FarmerHouseScared.js

// scenes\HamletDay2\FarmerHouse\confront\scared\FarmerHouseTiedUp.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHouseTiedUp(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var lootId = arguments.callee.name + "ButtonOne";
    var rapeId = arguments.callee.name + "ButtonTwo";

    var toLoot = new FarmerHouseFullLoot(ctrl);
    var toRape = new FarmerHouseTiedUpRape(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            The woman may be pacified, for now, but you still don't trust her.
                            You can [i]sense[/i] that she's just [b]waiting[/b] for you to slip up,
                            for you to turn your back so she can [b]plunge a knife[/b] into it.
                            Luckily, you brought just the right tool to deal with these kinds of situations.
                        [/p]
                        [p]
                            You produce your rope.
                            After telling the human to [b]stay down[/b], you slowly advance upon her.
                            You crudely [b]grab[/b] both her arms, and loop your rope around her wrists.
                            Forcing the meek woman to the ground, you then firmly bind her ankles,
                            looping the cord around her legs multiple times.
                            As a finishing touch, you wind the rope upwards, and spin it around her head to cover her mouth,
                            preventing her from screaming for help while completely immobilizing the housewife.
                        [/p]
                        [p]
                            After tying a firm knot, you rise to your feet and admire your handiwork.
                            The human struggles fiercely against her restraints, wriggling her body back and forth in a vain attempt to escape the well-tied rope.
                            She tries pleading with you through her improvised gag,
                            but all that escapes her lips are a bunch of muffled, high-pitched tones
                            barely even resembling the ever more frustrated and desperate implorations she had in mind.
                        [/p]
                        [p]
                            It looks like you did an excellent job.
                            She can't move, she can't scream for help, and the house is all yours now.
                            But then again: she [b]can't move[/b] and she [b]can't scream for help[/b].
                            You could very easily do something [i]other[/i] than stealing things.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildLootChoice() {
        var disabled = ctrl.tied.has() ? "disabled" : "";
        var isSelected = ctrl.tied.get() === FarmerHouseTiedUpChoice.LOOT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${lootId}" ${disabled}>Loot</button>`;

        var tooltip = global.parser.parse(`With the human restrained, you can loot in peace.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 140px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRapeChoice() {
        var disabled = ctrl.tied.has() ? "disabled" : "";
        var isSelected = ctrl.tied.get() === FarmerHouseTiedUpChoice.RAPE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${rapeId}" ${disabled}>Fuck her</button>`;

        var tooltip = global.parser.parse(`Have your way with the woman. She's completely restrained, and thus unable to fight back.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildRapeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildLootChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + lootId), $("#" + rapeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerLootListener() {
        $("#" + lootId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tied.set(FarmerHouseTiedUpChoice.LOOT);
            loadNextPart(toLoot);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerRapeListener() {
        $("#" + rapeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.tied.set(FarmerHouseTiedUpChoice.RAPE);
            loadNextPart(toRape);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.scared.get() !== FarmerHouseScaredChoice.TIE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.removeRope();
        part += buildChoice();
        part += `</div>`;
        part += toLoot.build();
        part += toRape.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.scared.get() !== FarmerHouseScaredChoice.TIE) {
            return;
        }

        toLoot.registerListeners();
        toRape.registerListeners();

        if (ctrl.tied.has()) {
            return;
        }

        registerLootListener();
        registerRapeListener();
    };
}

var FarmerHouseTiedUpChoice = {
    LOOT: 1,
    RAPE: 2
};

// END scenes\HamletDay2\FarmerHouse\confront\scared\FarmerHouseTiedUp.js

// scenes\HamletDay2\FarmerHouse\confront\threaten\FarmerHouseThreatenFail.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHouseThreatenFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            You tell the woman to keep quiet and stay put.
                            You threaten to [b]hurt[/b] and even [b]kill[/b] her, should she so much as [b]dare[/b] to make a noise.
                            ${weapon ? `You brandish your ${global.kobold.inventory.createItemWithTooltip(weapon)} in an attempt to seem more dangerous,
                                        yet you seem to achieve the opposite effect.` : `Sadly, you don't have a weapon to back your threats up with.`}
                            The human seems unimpressed.
                            She reaches behind her, grabbing a rather sharp-looking ${global.kobold.inventory.createItemWithTooltip(new KitchenKnife)} from a wooden block.
                            [i]Oh no.[/i]
                        [/p]
                        ${global.ui.condFail(weapon ? `${global.kobold.inventory.createItemWithTooltip(weapon, true)} offence below 4` : `No weapon`)}
                        [p]
                            The housewife menacingly advances upon you, knife in hand.
                            Every step she takes towards you, you take one backwards in fear.
                            Before long, you've been pushed back into the hallway you came from.
                            Unarmed, you could have easily taken her on, but with a knife [i]that big[/i],
                            she could [b]seriously[/b] hurt you, if not [b]kill you[/b] outright.
                        [/p]
                        [p]
                            Your fight or flight instincts kick in, and fighting a crazy lady armed with a sharp cutting tool is [b]not[/b] on your mind.
                            You spot the perfect opportunity to flee when the woman nearly trips over a crooked, protruding floorboard.
                            While she's busy recovering her balance, you turn tail and [b]bolt[/b] towards the rear exit.
                            As you open the door, the enraged human hurls her knife towards you,
                            the steel blade missing you by a hair, lodging itself into the wooden doorframe.
                        [/p]
                        [p]
                            For a moment, you consider retrieving the weapon and turning the tables on the housewife,
                            but the blade is stuck fast, and the woman lets out a shriek for help.
                            You decide to keep running, sprinting into the adjacent potato field.
                            A bunch of farmers emerge from their homes, rushing to the aid of the damsel in distress.
                            Luckily, by the time she's pointed out to them that you're escaping through the acreage,
                            you've already vanished into the small stretch of woodland on the opposite end.
                        [/p]
                        [p]
                            Despite having ditched your pursuers, you continue running all the way back to the eastern Varanar road.
                            The villagers might call the guards to investigate, and should anyone ask,
                            at least here you can [b]pretend[/b] to be a traveller instead of a [b]miserable failure[/b] of a thief.
                            Exhausted, you brace yourself against a nearby tree.
                        [/p]
                        [p]
                            You're glad you managed to get out of that one alive.
                            If you had ${weapon ? "a more intimidating weapon" : "some kind of weapon"}, perhaps your threats would have been taken a bit more seriously.
                            Alas, what's done is done.
                            You can't win [i]every[/i] confrontation you get into, and at least you didn't get [b]caught[/b] or [b]cut[/b].
                            You've gained nothing, but you haven't lost much either, besides an opportunity.
                            Your newfound fear of kitchen knives is sure to vanish in due time.
                        [/p]
                        [p]
                            Your brief period of respite over, you continue on the way to Varanar.
                            Hopefully, the humans there will be less [i]ballsy[/i].
                            That's what the hamlet's inhabitants constantly claim, anyway.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        global.kobold.history.shared.farmerhouse.failed = true;
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.threatenFail();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\threaten\FarmerHouseThreatenFail.js

// scenes\HamletDay2\FarmerHouse\confront\threaten\FarmerHouseThreaten.js

/**
 * Confront section of the Farmer House scene.
 * @param {FarmerHouseController} controller The Farmer House controller.
 */
function FarmerHouseThreaten(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toSuccess = new FarmerHouseScared(ctrl);
    var toFail = new FarmerHouseThreatenFail(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}


    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.confront.get() !== FarmerHouseConfrontChoice.THREATEN) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (!ctrl.misc.weapon() || ctrl.misc.weapon().offence() < 4) {
            part += toFail.build();
        } else {
            part += toSuccess.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.confront.get() !== FarmerHouseConfrontChoice.THREATEN) {
            return;
        }

        if (!ctrl.misc.weapon() || ctrl.misc.weapon().offence() < 4) {
            toFail.registerListeners();
        } else {
            toSuccess.registerListeners();
        }
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\threaten\FarmerHouseThreaten.js

// scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestleSuccess.js

/**
 * Arrival section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseWrestleSuccess(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var chokeId = arguments.callee.name + "ButtonOne";
    var rapeId = arguments.callee.name + "ButtonTwo";

    var toChoke = new FarmerHouseWrestleChoke(ctrl);
    var toRape = new FarmerHouseWrestleRape(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You have to stop her from screaming for help.
                            The only way you can do that is by covering her mouth, which you can't reach.
                            You're going to have to [b]bring her down[/b] to your level somehow.
                            And that means you need to get [i]physical[/i].
                        [/p]
                        [p]
                            You wildly charge the scared housewife, roughly grabbing her by the hips, the very centre of her mass, in an attempt to throw her to the floor.
                            She harshly [b]slaps[/b] you in the face, but that doesn't prevent you from tugging her away from the kitchen counter,
                            and [b]hurling[/b] her to the ground, the human falling face-first on top of the rotten, wooden floor with a loud thud, and a soft pained grunt.
                        [/p]
                        ${global.ui.condSucc(`Strength greater than 5`)}
                        [p]
                            You leap on top of the woman, [b]slamming[/b] your body into hers.
                            While you don't weigh that much, your move still suffices to knock the air out of the housewife's lungs,
                            preventing her from screaming for just long enough for you take hold of both of her arms, restraining them behind her back.
                            The moment she recovers her breath enough to attempt to shriek, you firmly plant the palm of your free hand on the lower half of her face,
                            her shout muffled to the point of inaudibility.
                        [/p]
                        [p]
                            The feisty human tries, but fails, to bite your palm.
                            She attempts to free her arms, to no avail.
                            She may be larger than you, but she [i]isn't[/i] by any means [b]stronger[/b].
                            Giving up on overpowering you, the woman begins to wildly kick her legs and wriggle her body around, striving to throw you off of her.
                        [/p]
                        [p]
                            Growing tired of her resistance, you cruelly press your knee into her back,
                            causing her more than a little discomfort while pinning her to the floor.
                            Still, she does not relent, both her legs still wildly flailing about.
                            Stupid human just doesn't know when to give up.
                            You press your sharp claws against her cheeks, digging into her soft and tender flesh without drawing blood,
                            while at the same time painfully twisting one of her arms.
                        [/p]
                        [p]
                            Finally, her protests die down, at least to the point where you don't have to actively [b]struggle[/b] to keep her in place.
                            While the housewife clearly hasn't [i]given up[/i] yet, you've scared her into calming down, for now.
                            Still, you doubt she'll remain [i]pacified[/i] for long.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildChokeChoice() {
        var disabled = ctrl.wrestle.has() ? "disabled" : "";
        var isSelected = ctrl.wrestle.get() === FarmerHouseWrestleSuccessChoice.CHOKE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${chokeId}" ${disabled}>Choke her</button>`;

        var tooltip = global.parser.parse(`Knock her out so you can loot in peace.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRapeChoice() {
        var disabled = ctrl.wrestle.has() ? "disabled" : "";
        var isSelected = ctrl.wrestle.get() === FarmerHouseWrestleSuccessChoice.RAPE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${rapeId}" ${disabled}>Rape her</button>`;

        var tooltip = global.parser.parse(`You've got her pinned down. Time to have some [i]fun[/i] with the human.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildChokeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildRapeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + chokeId), $("#" + rapeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerChokeListener() {
        $("#" + chokeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.wrestle.set(FarmerHouseWrestleSuccessChoice.CHOKE);
            loadNextPart(toChoke);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerRapeListener() {
        $("#" + rapeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.wrestle.set(FarmerHouseWrestleSuccessChoice.RAPE);
            loadNextPart(toRape);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.wrestle.change();
        part += buildChoice();
        part += `</div>`;
        part += toChoke.build();
        part += toRape.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        toChoke.registerListeners();
        toRape.registerListeners();

        if (ctrl.wrestle.has()) {
            return;
        }

        registerChokeListener();
        registerRapeListener();
    };
}

var FarmerHouseWrestleSuccessChoice = {
    CHOKE: 1,
    RAPE: 2
};

// END scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestleSuccess.js

// scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestleChokeContinue.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseWrestleChokeContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You exit the bedroom, returning to the living room.
                            Just going out the front door would be risky.
                            [b]Too[/b] risky.
                            You make your way back through the dining room into the kitchen.
                            The woman is gradually starting to wake up, her groans inspiring a sense of fear in you.
                            Looks like your time here is running out, and if you don't escape before she regains her senses fully,
                            you'll be [b]hunted down[/b] by whoever comes to her aid.
                            You bolt past her downed frame into the adjacent hallway, and exit the building the same way you entered it: quietly, and through the back.
                        [/p]
                        [p]
                            The moment you close the door behind you, a cry for help resonates from within the building.
                            Looks like the human woke up.
                            You weren't planning on sticking around anyway.
                            You sprint into the adjacent potato field, and after crossing it with some difficulty,
                            you escape into a small patch of woodland with nobody but the housewife having seen you.
                        [/p]
                        [p]
                            Still, just because you [i]may[/i] not be getting chased, doesn't mean you don't expeditiously make your way back to the eastern Varanar road.
                            The villagers might call the guards to investigate, and should anyone ask,
                            at least while on the road you can [b]pretend[/b] to be a traveller instead of a [b]woman-choking bandit[/b].
                            You're glad you didn't leave the farmer's house empty-handed.
                            Getting your loot might have been a little harder to achieve than what you'd hoped for,
                            but regardless of the amount of extra effort put in, the results are there.
                        [/p]
                        [p]
                            Reaching the road, you continue on your way to Varanar.
                            Hopefully, your streak of good luck will continue in the city.
                            And maybe you won't even have to [b]strangle[/b] anyone, the next time you attempt to break in somewhere.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }
    
    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.choke.continued()) {
            return "";
        }

        global.kobold.history.shared.farmerhouse.stole = true;
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.choke.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestleChokeContinue.js

// scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestleChoke.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseWrestleChoke(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var continueId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseWrestleChokeContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You remove your hand from the housewife's mouth, instantly wrapping your arm around her neck before she manages to utter a shout for help,
                            her attempts at a desperate plea reduced to a bunch of surprised gurgling noises as you squeeze her throat shut.
                            With her air supply cut off, the woman resumes her frantic struggling, even more intense than before.
                            While you've no intention of [i]killing[/i] her, [b]she[/b] doesn't know that.
                            Fearing for her very life, the human wildly writhes about, kicking her legs,
                            fighting against your grasp on her arms and wriggling her body from side to side in an effort to throw you off of her.
                        [/p]
                        [p]
                            While threatening violence doesn't seem to dissuade her this time,
                            you manage to keep the housewife firmly pinned down and in place throughout the entirety of her protests.
                            Merely half a minute after you began to choke her out, her floundering slowly begins to die down,
                            until her body finally slumps down to the ground, unconsciousness setting in.
                            You could very well [b]kill[/b] her, but you're a [b]thief[/b], and not a murderer.
                            You'd have preferred if things didn't get this [i]physical[/i] to begin with.
                        [/p]
                        [p]
                            The woman knocked out, you rise to your feet and take a quick breather.
                            You doubt she'll be out for long, however.
                            You need to get to looting this place, and fast.
                        [/p]
                        [p]
                            You step into the dining room.
                            Nothing worth a damn in here, at first glance.
                            You keep moving, entering the living room, or at least what's supposed to pass for it.
                            The room is dominated by a single padded chair, the most comfortable piece of furniture you've come across in the house, so far.
                            It stands before a small fireplace, probably the only source of [i]warmth[/i] apart from the oven in the kitchen.
                            Still, nothing interesting to take.
                        [/p]
                        [p]
                            There's only one room left, which you assume to be the bedroom.
                            Opening the shabby wooden door, your suspicions are confirmed.
                            Two beds.
                            A big one, and a small one.
                            A dresser, and a nightstand.
                            You decide to go for the latter.
                        [/p]
                        [p]
                            Opening the small drawer, you are surprised to find that it [b]isn't[/b] full of underwear.
                            Instead, the nightstand contains a pair of worn books,
                            atop which lay some out-of-place looking ${global.kobold.inventory.createItemWithTooltip(new FancyGlasses())}.
                            Your eyes open wide, avarice surging to a whole new level.
                            This find more than makes up for having to deal with the housewife.
                            If you knew these farmers had something [b]this[/b] expensive laying around,
                            you wouldn't have stopped attempting to steal from them after your first attempt.
                        [/p]
                        [p]
                            You figure they must be a family heirloom or something.
                            Regardless, they're [b]yours[/b] now.
                            You pocket the valuable item, and start thinking of a way out of this place.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.choke.continued() ? "disabled" : "";
        var isSelected = ctrl.choke.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.choke.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if(ctrl.wrestle.get() !== FarmerHouseWrestleSuccessChoice.CHOKE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.choke.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.wrestle.get() !== FarmerHouseWrestleSuccessChoice.CHOKE) {
            return;
        }
        
        next.registerListeners();

        if (ctrl.choke.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestleChoke.js

// scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestleFail.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseWrestleFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You have to stop her from screaming for help.
                            The only way you can do that is by covering her mouth, which you can't reach.
                            You're going to have to [b]bring her down[/b] to your level somehow.
                            And that means you need to get [i]physical[/i].
                        [/p]
                        [p]
                            You wildly charge the scared housewife, roughly grabbing her by the hips, the very centre of her mass, in an attempt to throw her to the floor.
                            While you succeed in throwing off her balance, the woman manages to grab the countertop behind her,
                            preventing herself from falling to the floor despite the most [b]ferocious[/b] pulling, tugging and throwing you manage to muster.
                            Realizing your initial plan has failed, you reach for the human's mouth in an attempt to cover it,
                            but the housewife seems far more [i]feisty[/i] than you'd have thought.
                        [/p]
                        ${global.ui.condFail(`Strength below 6`)}
                        [p]
                            She crudely [b]bashes[/b] your face with her elbow, 
                            her linen shirt doing little to soften the heavy blow to your snout, which sends you recoiling backwards.
                            You let out a pained grunt, clutching your muzzle.
                            Blood flows liberally from your nostrils.
                            The woman, no longer being held by you, flees into adjacent dining room.
                            While you're still reeling from her blow, trying to figure out whether your nose is broken or merely bleeding,
                            the human starts [b]screaming[/b] for help, just as you feared would happen.
                        [/p]
                        [p]
                            Her cries resonate throughout the entire decrepit house, and are undoubtably heard all over the hamlet.
                            Dreading the [i]retribution[/i] of whoever may be coming to [i]help[/i],
                            you collect your senses enough to [b]bolt[/b] out of the back door from whence you came.
                            A small mob of pitchfork-wielding farmers has already formed at the front of the house,
                            being instructed by the lady-in-distress to go after the kobold that spooked her.
                            With nowhere to hide that [i]wouldn't[/i] lead to certain death, you sprint into the nearby potato field, which is luckily unmanned.
                        [/p]
                        [p]
                            As the peasants comb through the house and the back yard,
                            one of them notices you vanishing into a small gathering of trees at the far end of the farmland.
                            Luckily, the group decides [b]not[/b] to give chase.
                            Your lead is too great for them to have any hopes of catching up,
                            and the men consider having chased you off a victory on its own.
                        [/p]
                        [p]
                            Despite having ditched your pursuers, you continue running all the way back to the eastern Varanar road.
                            The villagers might call the guards to investigate, and should anyone ask,
                            at least here you can [b]pretend[/b] to be a traveller instead of a [b]miserable failure[/b] of a thief.
                            Exhausted, you brace yourself against a nearby tree.
                            Your nose still hurts like hell, but the bleeding must have stopped somewhere between the hamlet and the road.
                            You're fairly sure nothing is [i]broken[/i], although it might be wise to see a medician once you're in town.
                        [/p]
                        [p]
                            If you were just a [i]little bit[/i] [b]stronger[/b], you [i]might[/i] have been able to overpower the woman.
                            Alas, what's done is done.
                            You can't win [i]every[/i] fight you get into, and at least you didn't get [b]caught[/b].
                            You've gained nothing, but you haven't lost much either, besides an opportunity and the ability to breathe through your nose.
                            The latter is bound to return once the rest of the blood clears out.
                        [/p]
                        [p]
                            Your brief period of respite over, you continue on the way to Varanar.
                            Hopefully, the humans there will be less [i]hardy[/i].
                            That's what the hamlet's inhabitants constantly claim, anyway.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        global.kobold.history.shared.farmerhouse.raped = true;
        global.kobold.history.shared.farmerhouse.failed = true;
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.wrestleFail();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestleFail.js

// scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestle.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseWrestle(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toSuccess = new FarmerHouseWrestleSuccess(ctrl);
    var toFail = new FarmerHouseWrestleFail(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}


    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.confront.get() !== FarmerHouseConfrontChoice.WRESTLE) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (ctrl.misc.strength() < 6) {
            part += toFail.build();
        } else {
            part += toSuccess.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.confront.get() !== FarmerHouseConfrontChoice.WRESTLE) {
            return;
        }

        if (ctrl.misc.strength() < 6) {
            toFail.registerListeners();
        } else {
            toSuccess.registerListeners();
        }
    };
}

// END scenes\HamletDay2\FarmerHouse\confront\wrestle\FarmerHouseWrestle.js

// scenes\HamletDay2\FarmerHouse\confront\FarmerHouseConfront.js

/**
 * Confront section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseConfront(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var threatenId = arguments.callee.name + "ButtonOne";
    var intimidateId = arguments.callee.name + "ButtonTwo";
    var wrestleId = arguments.callee.name + "ButtonThree";

    var toThreaten = new FarmerHouseThreaten(ctrl);
    var toIntimidate = new FarmerHouseIntimidate(ctrl);
    var toWrestle = new FarmerHouseWrestle(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Your mind races to come up with a solution to this [i]problem[/i].
                            You can't just leave the human be.
                            She's blocking your way into the rest of the house,
                            and if she keeps wandering about, she's bound to stumble upon you sooner or later.
                            If that happens, you're toast.
                            You decide it's best to take matters into your own hands.
                            That way, you might have a chance to stifle the woman's startled shouts instead of [i]you[/i] being the one getting surprised by [i]her[/i].
                        [/p]
                        [p]
                            You step out into the kitchen, approaching the housewife from behind.
                            She seems to be making ham sandwiches, likely the family's dinner.
                            Most human women you encounter wear dresses, but this one seems to be clothed a bit more [i]practically[/i].
                            She wears a plain, ragged linen top with torn sleeves, combined with what appears to be a home-made pair of leather pants.
                            Despite visibly being stitched together from various different pieces of animal hide,
                            the peasant-lady's trousers accentuate her form rather nicely.
                            While by no means young, she's far from [i]old[/i].
                            You'd wager her to be in her late thirties, perhaps early fourties.
                        [/p]
                        [p]
                            Suddenly, a floorboard creaks underneath your foot, alerting the woman to your presence.
                            She turns around, at first looking [i]over[/i] you,
                            seemingly on the verge of returning to her duties before lowering her eyes,
                            her gaze falling directly on top of the kobold standing in the middle of her kitchen.
                        [/p]
                        [p]
                            The woman's mouth drops agape.
                            She rubs her eyes to make sure she isn't dreaming.
                            Time itself seems to slow down as adrenaline begins to flood your body.
                            Your brain shoots into overdrive to try and formulate a plan of action that [i]doesn't[/i] entail you [b]running away[/b],
                            a primal urge to [i]flee[/i] building at the back of your head.
                        [/p]
                        [p]
                            On one hand, she's a human.
                            She's [i]much[/i] taller than you, and probably smarter, too.
                            Plus she lives here, which means she knows the layout of the house.
                            You don't.
                            A disadvantage, should this end in a chase.
                            On the other hand, she seems almost as scared as you were when she first entered the kitchen.
                            You [i]could[/i] try to threaten violence to keep her in line, or you could intimidate her in some other way.
                            A final possibility would be to [i]tackle[/i] the woman.
                            She may be twice your size, but that won't matter if you can force her to the floor.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThreatenChoice() {
        var disabled = ctrl.confront.has() ? "disabled" : "";
        var isSelected = ctrl.confront.get() === FarmerHouseConfrontChoice.THREATEN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${threatenId}" ${disabled}>Threaten her</button>`;

        var tooltip = global.parser.parse(`Threaten the woman's life to keep her in check.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildIntimidateChoice() {
        var disabled = ctrl.confront.has() ? "disabled" : "";
        var isSelected = ctrl.confront.get() === FarmerHouseConfrontChoice.INTIMIDATE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${intimidateId}" ${disabled}>Intimidate her</button>`;

        var tooltip = global.parser.parse(`Shout loudly to inspire fear.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 130px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildWrestleChoice() {
        var disabled = ctrl.confront.has() ? "disabled" : "";
        var isSelected = ctrl.confront.get() === FarmerHouseConfrontChoice.WRESTLE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${wrestleId}" ${disabled}>Wrestle</button>`;

        var tooltip = global.parser.parse(`Tackle the human and wrestle her to the ground.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You loot: </div>`;
        choice += buildThreatenChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildIntimidateChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildWrestleChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + threatenId), $("#" + intimidateId), $("#" + wrestleId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerThreatenListener() {
        $("#" + threatenId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.confront.set(FarmerHouseConfrontChoice.THREATEN);
            loadNextPart(toThreaten);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerIntimidateListener() {
        $("#" + intimidateId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.confront.set(FarmerHouseConfrontChoice.INTIMIDATE);
            loadNextPart(toIntimidate);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerWrestleListener() {
        $("#" + wrestleId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.confront.set(FarmerHouseConfrontChoice.WRESTLE);
            loadNextPart(toWrestle);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.arrival.get() !== FarmerHouseArrivalChoice.CONFRONT) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toThreaten.build();
        part += toIntimidate.build();
        part += toWrestle.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.arrival.get() !== FarmerHouseArrivalChoice.CONFRONT) {
            return;
        }

        toThreaten.registerListeners();
        toIntimidate.registerListeners();
        toWrestle.registerListeners();

        if (ctrl.confront.has()) {
            return;
        }

        registerThreatenListener();
        registerIntimidateListener();
        registerWrestleListener();
    };
}

var FarmerHouseConfrontChoice = {
    THREATEN: 1,
    INTIMIDATE: 2,
    WRESTLE: 3
};

// END scenes\HamletDay2\FarmerHouse\confront\FarmerHouseConfront.js

// scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHouseCabinetRun.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseCabinetRun(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You silently close the door behind you, and turn around to face what you hope will be an easy getaway.
                            Instead, you are greeted with the sight of a broad-postured, scythe-wielding human, presumably the farmer who lives here.
                            He glares down at you.
                            You stare back at him.
                            For a moment, neither of you move.
                            He grips his scythe with both hands.
                            You can almost [b]feel[/b] the tension in the air.
                        [/p]
                        [p]
                            [big]"[color=dred][b]Thief![/b][/color]"[/big]
                            the man bellows out in a deep, bass voice, breaking the small stalemate between the two of you, rallying comrades to his aid.
                            The human advances upon you.
                            He attacks you with an overhead scythe-swipe, which you narrowly manage to avoid.
                            The blade of his weapon ends up being stuck in the wooden backdoor, giving you just enough time to run past him, into the adjacent potato field.
                            The farmer struggles to recover his lethal tool, although not for long.
                            He gives chase, wildly swings his scythe at you, only [b]barely[/b] missing your tail with every cut.
                            To make matters worse, he's [b]gaining[/b] on you.
                        [/p]
                        [p]
                            Being almost twice your size, he can take far bigger steps, thus traversing the rough farmland with greater ease.
                            Halfway across the field, you can almost [b]feel[/b] his weapon cutting through the air, nearly slicing off the very tip of your hind appendage.
                            Right as he was about to land a definitive strike, the clumsy human trips over a pile of dirt, landing face-first on the hard, tilled ground.
                            You glance back at him just to gloat, but the sight of a veritable [b]mob[/b] of armed farmers rushing out of the hamlet towards you 
                            inspires you to keep running instead of rubbing the man's defeat in his face.
                        [/p]
                        [p]
                            For the second time in recent history, you find yourself running from a group of angry peasants.
                            This time, luckily, you have somewhat of a headstart.
                            That, and the farmers chasing you seem more interested in helping the man that originally chased you off to his feet,
                            rather than pursuing you as you escape through a sparse set of trees.
                            Still, you keep running for quite a while, all the way back to the eastern road to Varanar.
                        [/p]
                        [p]
                            Confident that you've lost your pursuers, you take a moment to catch your breath.
                            That was close.
                            If that stupid farmer hadn't tripped, you'd be minced meat by now.
                            Luckily, you made it out alive, with some goodies to show for it.
                            The cupboard was annoying to loot, to say the least, but in the end, you feel like you at at least got something out of this entire ordeal.
                            It may not be much, but it's more than nothing.
                            And more than nothing is all you need to survive in Varanar.
                            [small]Or so you hope.[/small]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.cabinetSteal.continued()) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.runChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.cabinetSteal.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHouseCabinetRun.js

// scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHouseDrawer.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseDrawer(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var nextId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseCabinetRun(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You can't open the cupboard, and you can't reach the top shelf.
                            That leaves the drawer.
                            You grab the metallic handle and tug on it, half-expecting the tray to be jammed much like the bottom compartiment is.
                            To your surprise, the drawer opens rather smoothly.
                            Overestimating the amount of force needed, you almost pull it out completely, your yank luckily stopping just short of disaster.
                        [/p]
                        [p]
                            The drawer seems to be mostly filled with worthless stuff.
                            Letters, a ball of yarn, an old, moldy cookie, some religious book you don't recognize and a spare broom head.
                            Disappointed, you begin to rummage through the stuff.
                            Surely there's something better underneath all of this [b]crap[/b], right?
                        [/p]
                        [p]
                            Nothing.
                            You start to get annoyed.
                            You toss the letters to the floor.
                            You put the well-worn book on the cabinet.
                            You throw the ball of yarn across the hallway in aggravation.
                            And then, hiding in the corner of the drawer, beyond the now-discarded thread, your prize is revealed to you:
                            a single ${global.kobold.inventory.createItemWithTooltip(new CopperRing())}.
                        [/p]
                        [p]
                            It probably slipped off of someone's finger, or maybe it was stored in there for safekeeping.
                            Regardless, you pocket it as your just due for having to stand through the harrowing sensation of almost being caught.
                            It isn't nearly as much as you'd hoped to gain from this venture, but it's more than nothing.
                            Honestly, you shouldn't have expected anything more valuable coming from a farmer's home.
                            You're glad you got at least [b]something[/b] out of this ordeal.
                        [/p]
                        [p]
                            You don't bother cleaning up after yourself.
                            Honestly, you're doing these dumb, disorganized humans a favor by emptying out their drawer full of trash.
                            Maybe when they pick up the stuff you threw to the floor, they'll store it in a proper fashion.
                            [i]And get rid of that disgusting cookie in the process.[/i]
                            Still somewhat frustrated, you silently slip out through the back door from whence you came.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.cabinetSteal.continued() ? "disabled" : "";
        var isSelected = ctrl.cabinetSteal.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${nextId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + nextId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + nextId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.cabinetSteal.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.cabinet.get() !== FarmerHouseCabinetChoice.DRAWER) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.cabinetSteal.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.cabinet.get() !== FarmerHouseCabinetChoice.DRAWER) {
            return;
        }

        next.registerListeners();

        if (ctrl.cabinetSteal.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHouseDrawer.js

// scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHousePry.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHousePry(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var nextId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseCabinetRun(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            The compartiment at the bottom of the cabinet probably contains the best goodies.
                            It makes sense, right?
                            Biggest storage, best stuff.
                            You take hold of the metallic handle attached to the wooden door, and firmly [b]yank[/b] it backwards,
                            but it absolutely [b]refuses[/b] to budge.
                            You don't know whether there's some kind of lock you're not seeing, if the hinges are rusted, or if the wood is somehow blocked.
                        [/p]
                        [p]
                            Annoyed, you begin to tug at the jammed door, briskly rattling it back and forth to little effect.
                            You stop once you become aware of the noise you're making.
                            Luckily, the woman doesn't seem to notice.
                            The walls of this house, while rotten, must be pretty thick.
                        [/p]
                        [p]
                            You go for a new approach.
                            Applying pressure to the winch-side of the door with one hand, you start to steadily pull on the handle with the other.
                            You begin to exert ever-intensifying force.
                            The wood begins to creak and crack.
                            Right as you think you've [b]nearly[/b] got it, the entire thing gives way!
                            While you had meant to force the door open, you [b]tear[/b] it from the cupboard as a whole,
                            the decaying wood no longer suffice to hold the winches in place.
                        [/p]
                        [p]
                            The cabinet breaks open surprisingly quietly.
                            You place the torn-off door on the floor.
                            The contents of the cupboard are revealed to you:
                            a single, tiny bag of gold, which you presume to be the farmer's life savings.
                            A small amount of coins stashed away for a rainy day.
                        [/p]
                        [p]
                            While you do feel kind of bad for the peasant family, you probably need this money more than they do.
                            They may be pretty low on the social ladder, but they're still [b]humans[/b].
                            They have rights, and laws protecting them and whatnot.
                            Compared to you, they'll have an easy life even without this coin purse.
                            Besides, you're a thief.
                            Stealing is just what you [i]do[/i].
                            It's in your nature.
                            In your blood.
                        [/p]
                        [p]
                            This isn't the time to grow remorseful.
                            You pocket the bag of coins.
                            There's no point in trying to [i]fix[/i] the cabinet.
                            You leave the broken part on the floor, and slip out the back door from whence you came.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.cabinetSteal.continued() ? "disabled" : "";
        var isSelected = ctrl.cabinetSteal.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${nextId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + nextId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + nextId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.cabinetSteal.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.cabinet.get() !== FarmerHouseCabinetChoice.PRY) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.cabinetSteal.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.cabinet.get() !== FarmerHouseCabinetChoice.PRY) {
            return;
        }

        next.registerListeners();

        if (ctrl.cabinetSteal.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHousePry.js

// scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHouseTopShelf.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseTopShelf(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var nextId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseCabinetRun(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            The box on top of the cabinet has piqued your attention.
                            You can't quite see what it contains,
                            but you're pretty certain a poor family like this would have scrapped the crate for firewood long ago if it weren't something [i]good[/i].
                            Why put it all the way [i]up there[/i], though?
                            These humans have to make [b]everything[/b] harder than it needs to be, don't they?
                        [/p]
                        [p]
                            With a soft sigh and a rolling of your eyes, you hoist yourself onto the bottom half of the cupboard.
                            The wooden piece of furniture creaks menacingly as you place your entire weight upon it,
                            almost as though it's [b]threatening[/b] to crack should you proceed to scale its rotten shelves.
                            Even standing upright on the lower half, you still can't reach the top.
                        [/p]
                        [p]
                            You consider your options.
                            You [i]could[/i] jump up and try to snag the crate, but that would probably cause the entire cabinet to break upon your landing.
                            Besides the noise attracting unwanted attention, you might fall,
                            and if there's something fragile in the box, all of your efforts may have been for nothing.
                        [/p]
                        [p]
                            Jumping ruled out, you begin to test the individual [i]racks[/i].
                            Most of them are rotten through and through, a few barely even withstanding your gentle, testing pats.
                            Only [b]one[/b] of the planks seems stable enough to hold any sort of weight, at least for a little while.
                            It's the bottom-most one, a few inches above the surface you're standing on,
                            perfect for giving you a small boost, which may be just enough for you to reach the top.
                        [/p]
                        [p]
                            You gingerly place one of your feet atop of the non-rotten plank, testing its limits.
                            You aren't too sure about this, but it's the only way you're going to get to the crate that [i]doesn't[/i] involve making a ruckus.
                            You decide to go for it.
                            Hoping for the best, you lift yourself upwards, hoisting your entire body up onto the tips of your toes.
                            The box coming into reach, you grab onto it, slide it off the top of the cabinet, and in a single, smooth motion,
                            you drop back down onto the lower half of the cupboard, your descent aided by the sheer [b]weight[/b] of the crate.
                        [/p]
                        [p]
                            The wooden piece of furniture creaks even loader than before, as though it's on the [b]verge[/b] of breaking.
                            You decide to hop off of it [i]before[/i] you ruin your slick move.
                            You quietly hop to the ground.
                            It is only when you land for a second time, that the heaviness of the open-topped crate begins to burden you.
                            You quickly set it down on the floor, careful not to break any of the glass bottles contained within.
                        [/p]
                        [p]
                            You let out a soft, relieved sigh.
                            Those were a couple of intense moments, but all things considered, you made it out just fine.
                            Time to inspect the loot.
                            You take a bottle out of the box.
                            Empty.
                            No worries, you probably just picked the wrong one to begin with.
                            You pull out another.
                            Also empty?!
                            Did you just risk [b]all of that[/b] for a crate filled with [b]empty bottles[/b]?!
                        [/p]
                        [p]
                            You frantically remove container after container from the box, the most filled one thus far containing naught but a few [b]droplets[/b].
                            This can't be happening.
                            Why would they put [b]empty bottles[/b] on top of the darned cupboard?!
                            Your hope of scoring a good haul having all but sunk, you pull out the very last of the glass flasks.
                            You can immediately feel that it's different.
                            Heavier.
                            A smile forms on your face.
                            It's full.
                        [/p]
                        [p]
                            "${global.kobold.inventory.createItemWithTooltip(new CouragePotion(), true)}", it says on the label.
                            You'll have to try it out later.
                            Or you could sell it at Varanar.
                            Either way, now is not the time to think about what you're going to do with the bottle.
                            You pocket it, and silently slip through the back door from which you came.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.cabinetSteal.continued() ? "disabled" : "";
        var isSelected = ctrl.cabinetSteal.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${nextId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + nextId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + nextId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.cabinetSteal.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.cabinet.get() !== FarmerHouseCabinetChoice.TOPSHELF) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.cabinetSteal.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.cabinet.get() !== FarmerHouseCabinetChoice.TOPSHELF) {
            return;
        }

        next.registerListeners();

        if (ctrl.cabinetSteal.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHouseTopShelf.js

// scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHouseCabinet.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseCabinet(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var drawerId = arguments.callee.name + "ButtonOne";
    var pryId = arguments.callee.name + "ButtonTwo";
    var topShelfId = arguments.callee.name + "ButtonThree";

    var toDrawer = new FarmerHouseDrawer(ctrl);
    var toPry = new FarmerHousePry(ctrl);
    var toTopShelf = new FarmerHouseTopShelf(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            The cupboard is your best bet.
                            You doubt the woman'll be returning to the kitchen any time soon,
                            and even if she does, she might just [i]overlook[/i] you like she did the first time.
                            Besides, what better place to store a bunch of valuables than a sole piece of furniture at the very rear end of the house?
                        [/p]
                        [p]
                            You step out from behind your cover, and turn to face the cabinet properly.
                            It is a decently-sized, rickety construct, parts of which are rotten to the point where you could probably punch right through the wood.
                            The bottom half consists of a single, rather large compartiment, which is topped off by a drawer.
                            Further upwards are a bunch of empty shelves, but you can see a crate of some sort on the very topmost one, far out of your reach.
                            You try to open the compartiment instead, but it is jammed shut.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildDrawerChoice() {
        var disabled = ctrl.cabinet.has() ? "disabled" : "";
        var isSelected = ctrl.cabinet.get() === FarmerHouseCabinetChoice.DRAWER;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${drawerId}" ${disabled} style="width:190px;">Check the drawer</button>`;

        var tooltip = global.parser.parse(`The easiest part to get to. Bound to contain some goodies.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildPryChoice() {
        var strongEnough = ctrl.misc.strength() >= 6;
        var disabled = ctrl.cabinet.has() || !strongEnough ? "disabled" : "";
        var isSelected = ctrl.cabinet.get() === FarmerHouseCabinetChoice.PRY;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${pryId}" ${disabled} style="width:190px;">Force it open</button>`;

        var tooltip = global.parser.parse(`Requires [b]6 strength[/b].`);
        if (strongEnough) {
            tooltip = global.parser.parse(`The small wooden door is jammed shut, but you can pry it open with some elbow grease.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildTopShelfChoice() {
        var statCheck = ctrl.misc.agility() >= 6;
        var disabled = ctrl.cabinet.has() || !statCheck ? "disabled" : "";
        var isSelected = ctrl.cabinet.get() === FarmerHouseCabinetChoice.TOPSHELF;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${topShelfId}" ${disabled} style="width:190px;">Climb up</button>`;

        var tooltip = global.parser.parse(`Requires [b]6 agility[/b].`);
        if (statCheck) {
            tooltip = global.parser.parse(`Climb on top of the cabinet to reach the top shelf.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildDrawerChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildPryChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildTopShelfChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + drawerId), $("#" + pryId), $("#" + topShelfId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerDrawerListener() {
        $("#" + drawerId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.cabinet.set(FarmerHouseCabinetChoice.DRAWER);
            loadNextPart(toDrawer);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerPryListener() {
        $("#" + pryId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.cabinet.set(FarmerHouseCabinetChoice.PRY);
            loadNextPart(toPry);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerTopShelfListener() {
        $("#" + topShelfId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.cabinet.set(FarmerHouseCabinetChoice.TOPSHELF);
            loadNextPart(toTopShelf);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.inside.get() !== FarmerHouseInsideChoice.CABINET) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toDrawer.build();
        part += toPry.build();
        part += toTopShelf.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.inside.get() !== FarmerHouseInsideChoice.CABINET) {
            return;
        }

        toDrawer.registerListeners();
        toPry.registerListeners();
        toTopShelf.registerListeners();

        if (ctrl.cabinet.has()) {
            return;
        }

        registerDrawerListener();
        registerPryListener();
        registerTopShelfListener();
    };
}

var FarmerHouseCabinetChoice = {
    DRAWER: 1,
    PRY: 2,
    TOPSHELF: 3
};

// END scenes\HamletDay2\FarmerHouse\steal\inside\cabinet\FarmerHouseCabinet.js

// scenes\HamletDay2\FarmerHouse\steal\inside\follow\FarmerHouseFollowRun.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseFollowRun(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You silently close the door behind you, and turn around to face what you hope will be an easy getaway.
                            Instead, you are greeted with the sight of a broad-postured, scythe-wielding human, presumably the farmer who lives here.
                            He glares down at you.
                            You stare back at him.
                            For a moment, neither of you move.
                            He grips his scythe with both hands.
                            You can almost [b]feel[/b] the tension in the air.
                        [/p]
                        [p]
                            [big]"[color=dred][b]Thief![/b][/color]"[/big]
                            the man bellows out in a deep, bass voice, breaking the small stalemate between the two of you, rallying comrades to his aid.
                            The human advances upon you.
                            He attacks you with an overhead scythe-swipe, which you narrowly manage to avoid.
                            The blade of his weapon ends up being stuck in the wooden backdoor, giving you just enough time to run past him, into the adjacent potato field.
                            The farmer struggles to recover his lethal tool, although not for long.
                            He gives chase, wildly swings his scythe at you, only [b]barely[/b] missing your tail with every cut.
                            To make matters worse, he's [b]gaining[/b] on you.
                        [/p]
                        [p]
                            Being almost twice your size, he can take far bigger steps, thus traversing the rough farmland with greater ease.
                            Halfway across the field, you can almost [b]feel[/b] his weapon cutting through the air, nearly slicing off the very tip of your hind appendage.
                            Right as he was about to land a definitive strike, the clumsy human trips over a pile of dirt, landing face-first on the hard, tilled ground.
                            You glance back at him just to gloat, but the sight of a veritable [b]mob[/b] of armed farmers rushing out of the hamlet towards you 
                            inspires you to keep running instead of rubbing the man's defeat in his face.
                        [/p]
                        [p]
                            For the second time in recent history, you find yourself running from a group of angry peasants.
                            This time, luckily, you have somewhat of a headstart.
                            That, and the farmers chasing you seem more interested in helping the man that originally chased you off to his feet,
                            rather than pursuing you as you escape through a sparse set of trees.
                            Still, you keep running for quite a while, all the way back to the eastern road to Varanar.
                        [/p]
                        [p]
                            Confident that you've lost your pursuers, you take a moment to catch your breath.
                            That was close.
                            If that stupid farmer hadn't tripped, you'd be minced meat by now.
                            Luckily, you made it out alive, with some yummy sandwiches to show for it.
                            The housewife was annoying to deal with, and you could have wrung her neck in frustration when she closed the door to the den behind her,
                            but in the end, you feel like you at at least got something out of this entire ordeal.
                            It may not be much, but it's more than nothing.
                            And more than nothing is all you need to survive in Varanar.
                            [small]Or so you hope.[/small]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.follow.continued()) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.runChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.follow.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\steal\inside\follow\FarmerHouseFollowRun.js

// scenes\HamletDay2\FarmerHouse\steal\inside\follow\FarmerHouseFollow.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseFollow(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var nextId = arguments.callee.name + "ButtonOne";

    var next = new FarmerHouseFollowRun(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You doubt the cupboard'll contain much of interest.
                            The best stuff is probably scattered throughout the house.
                            Who puts all their eggs in one basket, right?
                            Hopefully not these humans.
                        [/p]
                        [p]
                            You sneak into the kitchen and peek around the corner.
                            The housewife, as you expected, has just finished setting up the small, round table for dinner.
                            Three metallic plates and two sandwiches, one of which is cut in half and distributed over two plates.
                            The family's meagre meal set up, the human steps out into the living room, closing the loudly creaking,
                            half-crooked dining room door behind her.
                            Must not be dinnertime yet.
                        [/p]
                        [p]
                            Darnit.
                            While you [i]do[/i] like having a solid surface in between you and the housewife,
                            the wooden door she shut bars your way to the rest of the house.
                            Once again, your plans are foiled and your progress is hampered.
                            You figure it's too late to back out now.
                            Best to grab what you can and get the hell out of this place.
                        [/p]
                        [p]
                            You scour the kitchen, looking for anything of value to take.
                            The counter is barren, apart from some basic, well-worn tools.
                            The small coal-oven underneath looks like it hasn't been used in [b]months[/b].
                            The coal-prod is too rusted to be of any use, and there aren't even any spices like salt or pepper around.
                            The woman took the bread along with her when she left this part of the house, leaving absolutely [b]nothing[/b] for you to steal.
                        [/p]
                        [p]
                            Annoyed, you step into the dining room.
                            Through the cracks of the door, you can see the woman brushing the floor on the other side.
                            This room contains naught but a single, kobold-sized cupboard and the table set for dinner.
                            You quietly rummage through the cupboard.
                            Nothing but worthless metal dishes, only the fanciest of which are made of iron.
                            Not even a blacksmith would pay a single coin for low-grade copper and bronze.
                        [/p]
                        [p]
                            The cutlery on the table, too, is exceedingly supbar.
                            That leaves just one option: the food itself.
                            The full-sized sandwich and its two smaller siblings are smeared with some kind of butter, upon which rests a thick slice of ham.
                            Your mouth waters just [b]thinking[/b] about the taste of an actual [b]meal[/b].
                            While you might not be able to sell the food for much, at least you'll have gained [i]something[/i] out this adventure.
                        [/p]
                        [p]
                            The moral implications of stealing from an obviously poor family, 
                            who may or may not have a child, as indicated by the third plate, come to mind for a moment.
                            You figure that if they had enough food to prepare [b]one[/b] meal, you're certain they'll have plenty to make a second one.
                            Plus, you've gone for [b]way[/b] longer than a single day without eating in the past.
                            You're certain it won't kill them to miss out on a couple of slices of bread and ham.
                        [/p]
                        [p]
                            While the allure of digging in and having a feast right here and now is great, you know better than to linger when trespassing.
                            You'll have to save these for later.
                            You grab the large sandwich, and both of the smaller ones, and pocket them.
                            While you'd have liked to get a lot more loot out of this place, you'll take what you can get.
                            You silently slip out through the back door from whence you came.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.follow.continued() ? "disabled" : "";
        var isSelected = ctrl.follow.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${nextId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + nextId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + nextId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.follow.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if(ctrl.inside.get() !== FarmerHouseInsideChoice.FOLLOW) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.follow.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.inside.get() !== FarmerHouseInsideChoice.FOLLOW) {
            return;
        }
        
        next.registerListeners();

        if (ctrl.follow.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\steal\inside\follow\FarmerHouseFollow.js

// scenes\HamletDay2\FarmerHouse\steal\inside\FarmerHouseInside.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseInside(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var cabinetId = arguments.callee.name + "ButtonOne";
    var followId = arguments.callee.name + "ButtonTwo";

    var toCabinet = new FarmerHouseCabinet(ctrl);
    var toFollow = new FarmerHouseFollow(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Going to the shed would be a dumb idea.
                            There's probably [b]nothing[/b] of value in there.
                            At least in here, there might be [b]something[/b] worth stealing.
                            Still, it's not going to be easy with the human in the kitchen.
                        [/p]
                        [p]
                            Then, as though your prayers were heard, a miracle happens.
                            Footsteps, once more.
                            You peek out from behind the cabinet, just in time to spot the woman leaving the kitchen, a pair of sandwiches in hand.
                            You wait for a few moments to see if she might return, but she doesn't.
                            You hear her putting down what you presume to be a pair of metallic plates in an adjacent room, probably the dining room or the den.
                        [/p]
                        [p]
                            Your way further into the house is cleared.
                            You [b]knew[/b] staying indoors was the correct way to go about things!
                            But still, this isn't the time to get cocky.
                            The woman may have left the kitchen, but she's still around.
                            It'd be a shame if you [i]bumped[/i] into her after all this stressful waiting and careful planning.
                            If you want to play it safe, you could simply see what the cabinet you've been hiding behind contains.
                            Following the woman as she goes about her business is a more [i]risky[/i] strategy,
                            but it still beats randomly roaming about the house in both security and stealing potential.
                            The human can't see you if you're behind her back, after all.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildCabinetChoice() {
        var disabled = ctrl.inside.has() ? "disabled" : "";
        var isSelected = ctrl.inside.get() === FarmerHouseInsideChoice.CABINET;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${cabinetId}" ${disabled} style="width:180px;">Rob the cabinet</button>`;

        var tooltip = global.parser.parse(`Steal from the cabinet.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:180px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFollowChoice() {
        var disabled = ctrl.inside.has() ? "disabled" : "";
        var isSelected = ctrl.inside.get() === FarmerHouseInsideChoice.FOLLOW;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${followId}" ${disabled}>Follow her</button>`;

        var tooltip = global.parser.parse(`Follow the woman around and take what you can along the way.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildFollowChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildCabinetChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + cabinetId), $("#" + followId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerCabinetListener() {
        $("#" + cabinetId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.inside.set(FarmerHouseInsideChoice.CABINET);
            loadNextPart(toCabinet);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFollowListener() {
        $("#" + followId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.inside.set(FarmerHouseInsideChoice.FOLLOW);
            loadNextPart(toFollow);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.steal.get() !== FarmerHouseStealChoice.INSIDE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toCabinet.build();
        part += toFollow.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.steal.get() !== FarmerHouseStealChoice.INSIDE) {
            return;
        }

        toCabinet.registerListeners();
        toFollow.registerListeners();

        if (ctrl.inside.has()) {
            return;
        }

        registerCabinetListener();
        registerFollowListener();
    };
}

var FarmerHouseInsideChoice = {
    CABINET: 1,
    FOLLOW: 2
};

// END scenes\HamletDay2\FarmerHouse\steal\inside\FarmerHouseInside.js

// scenes\HamletDay2\FarmerHouse\steal\shed\FarmerHouseShedTrident.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseShedTrident(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var nextId = arguments.callee.name + "ButtonOne";

    var next = new FarmerHouseShedRun(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            You grab the pitchfork leaning against the wall.
                            It's [b]way[/b] too big to be of much use to you.
                            You place it atop the tiny table that holds most of the tools, the wooden handle sticking out over the edge.
                            You need to make this thing smaller somehow, so you can use it as a weapon.
                            An idea forms within your mind.
                        [/p]
                        [p]
                            You take the large, well-used, two-handed saw.
                            Another tool not cut for someone of your stature.
                            You can barely even [b]lift[/b] it due to the two heavy weights attached to either end.
                            You imagine this saw is used for cutting down trees with, but it's the best thing around to shorten the pitchfork with.
                        [/p]
                        [p]
                            Through some fidgeting and a lot of physical exertion to the point of annoyance,
                            you finally manage to get the saw to stand upright near the edge of the table, with the pitchfork's handle underneath its dull blade.
                            You gingerly begin to roll the heavy contraption back and forth as best you can, slowly carving it into the wooden grip below.
                            You need to stop your sawing ever so often, either to adjust the pitchfork to ensure it doesn't fall down, or just to take a breather.
                        [/p]
                        [p]
                            This tool was [b]not[/b] meant to be used  by a kobold.
                            You don't even think it was made for a [i]single[/i] human to employ.
                            The fact that it's more dull than your claws isn't helping.
                            Still, you perservere, deftly sawing away millimetre after millimetre, until finally, you reach the end.
                            As the saw carves slightly into the table, a large part of what was once the pitchfork's handle tumbles down to the ground.
                        [/p]
                        [p]
                            You did it!
                            With one final burst of effort, you quietly lay the saw back down onto table,
                            before taking hold of your brand new ${global.kobold.inventory.createItemWithTooltip(new KoboldTrident())}.
                            You raise the former tool up to inspect it.
                            Far smaller than before.
                            Lighter, and easier for you to use.
                            Perfect.
                            You stab the air a few times to get a feeling for your new weapon.
                            ${weapon ? `You ditch your old ${global.kobold.inventory.createItemWithTooltip(weapon)}.
                            You have found something better.` : `You feel like a spear-wielding knight.`}
                        [/p]
                        [p]
                            You [b]love[/b] it.
                            And you made it yourself!
                            Well, sort of, anyway.
                            The important part is that it's yours now.
                            With the pitchfork-turned-trident in hand, you excitedly leave the shed, ready to face whatever the world has to throw at you.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.postshed.continued() ? "disabled" : "";
        var isSelected = ctrl.postshed.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${nextId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + nextId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + nextId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postshed.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.shed.get() !== FarmerHouseShedChoice.TRIDENT) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.postshed.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.shed.get() !== FarmerHouseShedChoice.TRIDENT) {
            return;
        }

        next.registerListeners();

        if (ctrl.postshed.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\steal\shed\FarmerHouseShedTrident.js

// scenes\HamletDay2\FarmerHouse\steal\shed\FarmerHouseShedRun.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseShedRun(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var gotWeapon = ctrl.shed.get() === FarmerHouseShedChoice.TRIDENT;
        var toReturn = `[p]
                            The moment you exit the shed, you are greeted with the sight of a broad-postured, scythe-wielding human,
                            presumably the farmer who lives here.
                            He glares down at you.
                            You stare back at him.
                            For a moment, neither of you move.
                            He grips his scythe with both hands.
                            You can almost [b]feel[/b] the tension in the air.
                        [/p]
                        [p]
                            [big]"[color=dred][b]Thief![/b][/color]"[/big]
                            the man bellows out in a deep, bass voice, breaking the small stalemate between the two of you, rallying comrades to his aid.
                            ${gotWeapon ? `You consider attacking him with your new weapon, 
                            but if you stick around for much longer, the entire hamlet'll be up your ass. Instead, you` : `You`}
                            turn tail and run, bolting past the wooden shed into the potato field.
                            The farmer gives chase.
                            He wildly swings his scythe at you, only [b]barely[/b] missing your tail with every swipe.
                            To make matters worse, he's [b]gaining[/b] on you.
                        [/p]
                        [p]
                            Being almost twice your size, he can take far bigger steps, thus traversing the rough farmland with greater ease.
                            Halfway across the field, you can almost [b]feel[/b] his weapon cutting through the air, nearly slicing off the very tip of your hind appendage.
                            Right as he was about to land a definitive strike, the clumsy human trips over a pile of dirt, landing face-first on the hard, tilled ground.
                            You glance back at him just to gloat, but the sight of a veritable [b]mob[/b] of armed farmers rushing out of the hamlet towards you 
                            inspires you to keep running instead of rubbing the man's defeat in his face.
                        [/p]
                        [p]
                            For the second time in recent history, you find yourself running from a group of angry peasants.
                            This time, luckily, you have somewhat of a headstart.
                            That, and the farmers chasing you seem more interested in helping the man that originally chased you off to his feet,
                            rather than pursuing you as you escape through a sparse set of trees.
                            Still, you keep running for quite a while, all the way back to the eastern road to Varanar.
                        [/p]
                        [p]
                            Confident that you've lost your pursuers, you take a moment to catch your breath.
                            That was close.
                            If that stupid farmer hadn't tripped, you'd be minced meat by now.
                            ${gotWeapon ? `Luckily, you made it out alive, with a fancy new weapon to boot.` : `Luckily, the only thing that got sliced was your hand.`}
                            You consider your trip to the farmer's house a ${gotWeapon ? "relative success" : "horrible failure"}.
                            You can only hope your luck ${gotWeapon ? "won't change" : "will improve"} when you reach Varanar.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     *  @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.postshed.continued()) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.runChange();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.postshed.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\steal\shed\FarmerHouseShedRun.js

// scenes\HamletDay2\FarmerHouse\steal\shed\FarmerHouseShedNoWeapon.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseShedNoWeapon(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var continueId = arguments.callee.name + "ButtonOne";
    var next = new FarmerHouseShedRun(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            You decide against taking a pitchfork as your weapon.
                            ${weapon ? `You already have a perfectly fine ${global.kobold.inventory.createItemWithTooltip(weapon)} anyway.` : ``}
                            Still, you aren't going to walk away from this entire ordeal empty-handed.
                            You don't want to have risked trespassing with absolutely [b]nothing[/b] to show for it.
                            You need [i]something[i], [b]anything[/b] to make it seem like you [b]didn't[/b] waste your time coming here.
                        [/p]
                        [p]
                            While most of the tools are neatly arrayed on a small table or hanging off the wall on half-crooked nails,
                            there's a metallic pile of rusty [b]rubbish[/b] cluttering a corner of the shed that you had neglected at first.
                            Seeing as everything else is practically worthless, you may as well look through it and see if there's anything of any value.
                        [/p]
                        [p]
                            You drop to your knees, and begin to rummage through the pile of metallic scrap.
                            After removing a few rusted pieces, you see a small glint of something shiny at the very bottom of the heap.
                            Excited, you eagerly reach for what you [b]hope[/b] to be a pile of coins buried underneath the rest of the trash,
                            yet long before you manage to make it to the object in question,
                            your hand brushes alongside an exceptionally [b]sharp[/b] piece of metal, which [b]rakes[/b] open your tender palm.
                        [/p]
                        [p]
                            You barely refrain from [b]screaming[/b] in pain.
                            You use your free hand to cover your muzzle, silencing the few pathetic whimpers that escape your lips.
                            You slowly retract the bleeding appendage from the pile of metal refuse, careful not to bump into any other jagged edges.
                            Blood leaks liberally from it, dripping down on top of the rusted heap, obscuring the glint that had caught your attention,
                            burying it, along with your self-esteem.
                        [/p]
                        [p]
                            Once fully removed from the scrapheap, you inspect the damage done to your palm, clutching your wrist with your free hand.
                            The cut is luckily not very [b]deep[/b], but it [b]is[/b] pretty long, 
                            running from the top of your wrist along the entire length of your palm, stopping just short of your index finger.
                            Despite it being shallow, your wound [b]hurts[/b] like [b]hell[/b], and it's bleeding quite a lot.
                        [/p]
                        [p]
                            You sit yourself down on the dusty wooden floor.
                            A few tears spill from your eyes.
                            You try to remain stoic, yet the pain mixed with the humiliation of utter [b]failure[/b] are too much.
                            That shiny thing you saw probably wasn't even a coin.
                            You quietly begin to sob.
                        [/p]
                        [p]
                            You're a thief, yet that's [b]twice[/b] now that you've completely [b]fucked up[/b] stealing from these simple farmers,
                            some of which have a life worse than your own.
                            Although right now, you find that a bit hard to believe.
                            As the bleeding slowly comes to a halt, you painfully wriggle your fingers around to make sure nothing critical was damaged.
                            Through teary eyes, you observe that all four of them are still in working condition,
                            although it'll be a while still before you can move them without hurting yourself.
                        [/p]
                        [p]
                            This is all that stupid farmer's fault.
                            If he had some [b]decent[/b] tools, you wouldn't have had to go rummaging in his pile of [b]crap[/b].
                            Why does he keep rusted metal around anyway?!
                            Stupid human.
                        [/p]
                        [p]
                            You rise to your feet once more.
                            You wipe your eyes with your non-hurting hand.
                            Your tears have stopped, but your dignity remains dented.
                            You have been bested by a pile of scrap.
                            You [i]want[/i] to kick it, to get some [i]revenge[/i] on the inanimate objects that hurt you,
                            but you'd rather not draw attention, let alone cut your foot as well.
                            With a hurting hand and wounded pride, you decide to leave the shed.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.postshed.continued() ? "disabled" : "";
        var isSelected = ctrl.postshed.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.postshed.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if(ctrl.shed.get() !== FarmerHouseShedChoice.NOWEAPON) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.postshed.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.shed.get() !== FarmerHouseShedChoice.NOWEAPON) {
            return;
        }
        
        next.registerListeners();

        if (ctrl.postshed.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\FarmerHouse\steal\shed\FarmerHouseShedNoWeapon.js

// scenes\HamletDay2\FarmerHouse\steal\shed\FarmerHouseShed.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseShed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var tridentId = arguments.callee.name + "ButtonOne";
    var noWeaponId = arguments.callee.name + "ButtonTwo";

    var toTrident = new FarmerHouseShedTrident(ctrl);
    var toNoWeapon = new FarmerHouseShedNoWeapon(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            You make the wise decision of [b]not[/b] taking any unnecessary risks.
                            You creep towards the door, quietly open it once more, and slip out of the dilapidated homestead.
                            A second wave of relief is your body's way of thanking you for [b]not[/b] making it have to weather more uncertainty and stress.
                        [/p]
                        [p]
                            You walk towards the shed.
                            Like the back door, it is unlocked, and like the house, it is rickety to the point of barely being able to stand upright.
                            You reckon a firm gust of wind could break it down, but you're no engineer.
                            Like your home, it might be more stable than it appears.
                        [/p]
                        [p]
                            It looks like you were right.
                            Tools, but not just farming tools.
                            A near-blunt saw, a chipped hammer, a bunch of nails, a plow that's seen better days,
                            a spade made out of wood, a hoe, a rake and a pitchfork.
                            Those are the ones you recognize.
                            The rest, you imagine, are various farming and carpenting tools that a dishonest kobold like yourself has no use for.
                        [/p]
                        [p]
                            Darnit.
                            Where's all the good stuff at?!
                            You were hoping to find a scythe or a sickle or at least a pickaxe.
                            The farmer must have taken those.
                            Or maybe he just doesn't own them.
                            Seeing as he doesn't seem too keen on [b]maintaining[/b] what little he [i]does[/i] have,
                            you doubt the [i]quality[/i] of the tools would have been all that great anyway.
                        [/p]
                        [p]
                            The only thing serviceable in this entire shed would be the pitchfork.
                            It's far too big for someone of your [i]posture[/i] to use,
                            but maybe you can [i]trim[/i] the handle somehow.
                            ${weapon ? `You [b]would[/b] have to leave behind your ${global.kobold.inventory.createItemWithTooltip(weapon)} behind, though.
                            No point in carrying two weapons.` : ``}
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildTridentChoice() {
        var disabled = ctrl.shed.has() ? "disabled" : "";
        var isSelected = ctrl.shed.get() === FarmerHouseShedChoice.TRIDENT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${tridentId}" ${disabled}>Take it</button>`;

        var tooltip = global.parser.parse(`With some slight alterations, the pitchfork could make a formidable weapon.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNoWeaponChoice() {
        var disabled = ctrl.shed.has() ? "disabled" : "";
        var isSelected = ctrl.shed.get() === FarmerHouseShedChoice.NOWEAPON;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${noWeaponId}" ${disabled}>Do not</button>`;

        var tooltip = global.parser.parse(`A pitchfork? Hardly a weapon for a thief. What are you, a farmer?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildTridentChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildNoWeaponChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + tridentId), $("#" + noWeaponId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerTridentListener() {
        $("#" + tridentId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.shed.set(FarmerHouseShedChoice.TRIDENT);
            loadNextPart(toTrident);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerNoWeaponListener() {
        $("#" + noWeaponId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.shed.set(FarmerHouseShedChoice.NOWEAPON);
            loadNextPart(toNoWeapon);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.steal.get() !== FarmerHouseStealChoice.SHED) {
            return "";
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toTrident.build();
        part += toNoWeapon.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.steal.get() !== FarmerHouseStealChoice.SHED) {
            return;
        }

        toTrident.registerListeners();
        toNoWeapon.registerListeners();

        if (ctrl.shed.has()) {
            return;
        }

        registerTridentListener();
        registerNoWeaponListener();
    };
}

var FarmerHouseShedChoice = {
    TRIDENT: 1,
    NOWEAPON: 2
};

// END scenes\HamletDay2\FarmerHouse\steal\shed\FarmerHouseShed.js

// scenes\HamletDay2\FarmerHouse\steal\FarmerHouseSteal.js

/**
 * Steal section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseSteal(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var toolshedId = arguments.callee.name + "ButtonOne";
    var insideId = arguments.callee.name + "ButtonTwo";

    var toToolshed = new FarmerHouseShed(ctrl);
    var toInside = new FarmerHouseInside(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You press yourself against the wall, hiding behind a tall cabinet to ensure you won't be spotted while you think of a plan.
                            The woman softly hums to herself while she goes about her business, making sandwiches for dinner.
                            While the hallway you are in somewhat separates the two of you, the only way further into the house is via the kitchen,
                            and you doubt you can sneak by the human without her noticing.
                        [/p]
                        [p]
                            You struggle to come up with an alternate route that won't lead to your detection.
                            You [i]could[/i] leave out the back and go through the front door, or maybe try a different house,
                            but you'd essentially be trading one roll of the dice for another.
                            Who's to say there won't be a guard dog or another housewife or an inquisitive child next door?
                            Heck, you got [b]lucky[/b] this time around.
                            The door was unlocked, and you managed to enter without making much noise.
                            Besides, if the woman is making dinner, you doubt it'll be long before the men of the village return.
                        [/p]
                        [p]
                            A thought bubbles up in your brain.
                            The shed outside.
                            You [i]could[/i] go take a look in there.
                            It'd be [b]safer[/b] than staying in the house,
                            and you might find something useful, although probably nothing [i]too[/i] valuable.
                            But then again, this place doesn't look too promising either.
                            [b]If[/b] you can get anything at all from it without the woman in the kitchen spotting you.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildToolshedChoice() {
        var disabled = ctrl.steal.has() ? "disabled" : "";
        var isSelected = ctrl.steal.get() === FarmerHouseStealChoice.SHED;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${toolshedId}" ${disabled}>Go to the shed</button>`;

        var tooltip = global.parser.parse(`At least the toolshed outside won't have a housewife running around in it.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 195px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildInsideChoice() {
        var disabled = ctrl.steal.has() ? "disabled" : "";
        var isSelected = ctrl.steal.get() === FarmerHouseStealChoice.INSIDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${insideId}" ${disabled}>Stay indoors</button>`;

        var tooltip = global.parser.parse(`Profit trumps risk. Find a way to steal things without being noticed.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 150px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildToolshedChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildInsideChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + toolshedId), $("#" + insideId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerToolshedListener() {
        $("#" + toolshedId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.steal.set(FarmerHouseStealChoice.SHED);
            loadNextPart(toToolshed);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerInsideListener() {
        $("#" + insideId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.steal.set(FarmerHouseStealChoice.INSIDE);
            loadNextPart(toInside);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if(ctrl.arrival.get() !== FarmerHouseArrivalChoice.STEAL) {
            return "";
        }
        
        global.kobold.history.shared.farmerhouse.stole = true;
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toToolshed.build();
        part += toInside.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if(ctrl.arrival.get() !== FarmerHouseArrivalChoice.STEAL) {
            return;
        }
        
        toToolshed.registerListeners();
        toInside.registerListeners();

        if (ctrl.steal.has()) {
            return;
        }

        registerToolshedListener();
        registerInsideListener();
    };
}

var FarmerHouseStealChoice = {
    SHED: 1,
    INSIDE: 2
};

// END scenes\HamletDay2\FarmerHouse\steal\FarmerHouseSteal.js

// scenes\HamletDay2\FarmerHouse\arrival\FarmerHouseArrival.js

/**
 * Arrival section of the Farmer House scene.
 * @param {BreedingFarmController} controller The Farmer House controller.
 */
function FarmerHouseArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;
    
    var confrontId = arguments.callee.name + "ButtonOne";
    var stealId = arguments.callee.name + "ButtonTwo";

    var toConfront = new FarmerHouseConfront(ctrl);
    var toSteal = new FarmerHouseSteal(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide some good old thievery would be the best way to get over the loss of your favorite person in the entire world.
                            Your target: one of the dozen or so homes that make up most of the little village.
                            The farmers around these parts aren't [i]rich[/i] by any means, but the more well off amongst them aren't [i]poor[/i] either.
                            They eat what they grow, and with low rural taxes, their hard work manages to net them a decent profit.
                            Hopefully, some of that profit will soon be [i]yours[/i].
                        [/p]
                        [p]
                            You walk down the dirt road that runs through the middle of the hamlet.
                            You've stolen from here before,
                            although you hope this time [i]won't[/i] end in you being chased off by a mob of angry humans with pitchforks and torches.
                            It was late at night, so you don't think anyone'll [i]recognize[/i] you.
                            Not that there [b]is[/b] anyone present [b]to[/b] remember your face.
                            This place is a ghost town.
                            Everyone's probably off working the fields or selling their wares.
                            And that makes [b]your[/b] [i]job[/i] a whole lot easier.
                        [/p]
                        [p]
                            You take one final look around to make sure absolutely [b]no one[/b] is spying on you,
                            before casually wandering off the main road, strolling into a random farmer's back yard as though you own the place.
                            Reaching the rear-end, you turn the corner and lower your posture, dropping into a sneaking stance.
                            You take note of a small shed a bit further out back, one not unlike your [i]home[/i].
                            It doesn't look like there's anyone in it.
                            It's probably used as a storage place for farming tools.
                            Luckily enough, it blocks the view of anyone working the neighbouring potato field,
                            hiding you from the prying eyes of honest, hard-working men.
                        [/p]
                        [p]
                            The house you chose is not nearly the biggest, nor is it the most fancy.
                            In fact, it's one of the less [i]prestigious[/i] buildings in town.
                            You doubt whoever lives here is expecting to get [b]robbed[/b], which means security will be lax.
                            And indeed, testing the back door, you find the shabby, wooden thing completely unlocked!
                            The only well maintained parts of it seem to be the hinges, which function noiselessly and smoothly,
                            the door opening in complete [b]silence[/b], allowing you to slip inside in an equally quiet fashion.
                        [/p]
                        [p]
                            You find yourself in a small hallway leading into the humble household's kitchen.
                            The inside of the house is every bit as rundown as the exterior is.
                            The undecorated wooden walls look nigh-rotten.
                            You spot a [i]nasty[/i] discoloration in a particularly [i]damp[/i] ceiling-corner.
                            [b]Mold[/b].
                            Disgusting.
                            You briefly wonder how people can [i]live[/i] like this,
                            but then you remind yourself of the fact that this place is [b]several times[/b] the size of your own home.
                            Still, [i]your[/i] shack [b]doesn't[/b] have mold growing in it, despite it being in the middle of the forest.
                        [/p]
                        [p]
                            The moment you were about to head deeper into the house, [b]footsteps[/b] resonate from within.
                            You freeze in place.
                            Your heart skips a beat.
                            Someone's home.
                            Flashbacks of your last [i]heist[/i] in this very hamlet shoot through your brain.
                            Your mind is torn between hiding and running, but your body ends up doing neither.
                        [/p]
                        [p]
                            While you're stuck halfway between [i]concealing yourself[/i] and [i]getting the hell out of there[/i],
                            a human saunters into the kitchen.
                            A woman.
                            You mindlessly stare at her as she turns to the small countertop dominating the cramped room,
                            and leisurely begins to slice a loaf of bread using a rusty-looking butter knife.
                            She's too engrossed in her day-to-day activities to notice you.
                            For now, at least.
                        [/p]
                        [p]
                            A mixture of relief and anxiety washes over you.
                            You haven't been caught just yet, but the presence of a housewife [i]complicates[/i] things, to say the least.
                            You need to do [i]something[/i], and you need to do it fast.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildConfrontChoice() {
        var disabled = ctrl.arrival.has() ? "disabled" : "";
        var isSelected = ctrl.arrival.get() === FarmerHouseArrivalChoice.CONFRONT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${confrontId}" ${disabled}>Confront her</button>`;

        var tooltip = global.parser.parse(`Approach the woman. Maybe you can prevent her from screaming somehow.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 150px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildStealChoice() {
        var disabled = ctrl.arrival.has() ? "disabled" : "";
        var isSelected = ctrl.arrival.get() === FarmerHouseArrivalChoice.STEAL;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${stealId}" ${disabled}>Steal stuff</button>`;

        var tooltip = global.parser.parse(`Ignore the housewife. Do what you came here to do. You can run if she sees you. That's what you did last time.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 185px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildConfrontChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildStealChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + confrontId), $("#" + stealId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerConfrontListener() {
        $("#" + confrontId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.arrival.set(FarmerHouseArrivalChoice.CONFRONT);
            loadNextPart(toConfront);
        });
    }
    
    /**
     * Adds a listener to the button.
     */
    function registerStealListener() {
        $("#" + stealId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.arrival.set(FarmerHouseArrivalChoice.STEAL);
            loadNextPart(toSteal);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        global.kobold.history.shared.farmerhouse.went = true;
        global.ui.applyTheme("hamlet_barn");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toConfront.build();
        part += toSteal.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        toConfront.registerListeners();
        toSteal.registerListeners();

        if (ctrl.arrival.has()) {
            return;
        }

        registerConfrontListener();
        registerStealListener();
    };
}

var FarmerHouseArrivalChoice = {
    CONFRONT: 1,
    STEAL: 2
};

// END scenes\HamletDay2\FarmerHouse\arrival\FarmerHouseArrival.js

// scenes\HamletDay2\FarmerHouse\FarmerHouseController.js

/**
 * Controller for the Farmer's House scene.
 * @param {string} sceneId The id of the scene.
 */
function FarmerHouseController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function postShedChange() {
        if (that.shed.get() === FarmerHouseShedChoice.NOWEAPON) {
            global.kobold.status.physical.decrease(10, "Cut by rusty metal");
            global.kobold.status.mental.decrease(10, "Beaten by garbage");
        } else {
            global.kobold.inventory.equipment.wear(new KoboldTrident());
            global.kobold.status.mental.increase(5, "Crafted weapon");
            global.kobold.status.stamina.decrease(10, "Hard work");
        }
    }

    function postShedChangeNote() {
        if (that.shed.get() === FarmerHouseShedChoice.NOWEAPON) {
            return global.ui.change.buildNotification([
                new Change("Rust-cut", "Health", -10),
                new Change("Beaten by garbage", "Sanity", -10)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Crafted", global.kobold.inventory.createItemWithTooltip(new KoboldTrident(), true), +1),
                new Change("Made something", "Sanity", +5),
                new Change("Hard work", "Stamina", -10)
            ]);
        }
    }

    function followChange() {
        global.kobold.inventory.bag.addConsumable(new Sandwich(), 2);
    }

    function followChangeNote() {
        return global.ui.change.buildNotification([
            new Change("Loot", global.kobold.inventory.createItemWithTooltip(new Sandwich(), true), +2)
        ]);
    }

    function cabinetChange() {
        if (that.cabinet.get() === FarmerHouseCabinetChoice.DRAWER) {
            global.kobold.status.mental.decrease(5, "Frustration");
            global.kobold.inventory.bag.addMisc(new CopperRing());
        } else if (that.cabinet.get() === FarmerHouseCabinetChoice.PRY) {
            global.kobold.inventory.coins += 20;
        } else {
            global.kobold.inventory.bag.addConsumable(new CouragePotion());
        }
    }

    function cabinetChangeNote() {
        if (that.cabinet.get() === FarmerHouseCabinetChoice.DRAWER) {
            return global.ui.change.buildNotification([
                new Change("Frustration", "Sanity", -5),
                new Change("Loot", global.kobold.inventory.createItemWithTooltip(new CopperRing(), true), +1)
            ]);
        } else if (that.cabinet.get() === FarmerHouseCabinetChoice.PRY) {
            return global.ui.change.buildNotification([
                new Change("Loot", "Gold", +20)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Loot", global.kobold.inventory.createItemWithTooltip(new CouragePotion(), true), +1)
            ]);
        }
    }

    function wrestleChange() {
        global.kobold.status.physical.decrease(5, "Slapped");
        global.kobold.status.stamina.decrease(15, "Wrestling");
    }

    function wrestleChangeNote() {
        return global.ui.change.buildNotification([
            new Change("Slapped", "Health", -5),
            new Change("Wrestling", "Stamina", -15)
        ]);
    }

    function wrestleLootChange() {
        global.kobold.inventory.addItem(new FancyGlasses());
        global.kobold.status.stamina.decrease(10, "Choking a housewife");
    }

    function wrestleLootChangeNote() {
        return global.ui.change.buildNotification([
            new Change("Loot", global.kobold.inventory.createItemWithTooltip(new FancyGlasses(), true), +1),
            new Change("Further wrestling", "Stamina", -10)
        ]);
    }

    function rapeChange() {
        if (that.wrestle.has()) {
            // Wrestle rape change
            if (that.misc.dick()) {
                global.kobold.body.dickVirgin = false;
                global.kobold.status.mental.increase(10, "Sadistic pleasure");
                global.kobold.status.stamina.decrease(15, "Forceful buttsex");
                global.kobold.status.exertion.increase(25, "Forceful buttsex");
            } else {
                global.kobold.status.mental.decrease(5, "Sadism denied");
                global.kobold.status.exertion.increase(5, "Fingered human ass");
            }
        } else if (that.tied.has()) {
            // Tied rape change
            global.kobold.status.mental.increase(10, "Sadistic pleasure");
            global.kobold.status.stamina.decrease(10, "Non-penetrative sex");
            global.kobold.status.exertion.increase(20, "Non-penetrative sex");
        } else {
            // Forced oral change
            if (that.misc.dick()) {
                global.kobold.body.dickVirgin = false;
            }
            global.kobold.status.mental.increase(10, "Sadistic pleasure");
            global.kobold.status.stamina.decrease(10, "Facefucking housewife");
            global.kobold.status.exertion.increase(20, "Facefucking housewife");
        }
    }

    function rapeChangeNote() {
        if (that.wrestle.has()) {
            // Wrestle rape note
            if (that.misc.dick()) {
                return global.ui.change.buildNotification([
                    new Change("Dominated human", "Sanity", +10),
                    new Change("Buttsex", "Stamina", -15),
                    new Change("Buttsex", "Sexual exhaustion", +25, true)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Dominated human", "Sanity", +10),
                    new Change("Pleasure denied", "Sanity", -15),
                    new Change("Pleasure denied", "Sexual exhaustion", +5, true)
                ]);
            }
        } else if (that.tied.has()) {
            // Tied rape change note
            return global.ui.change.buildNotification([
                new Change("Dominated human", "Sanity", +10),
                new Change("Non-penetrative sex", "Stamina", -10),
                new Change("Non-penetrative sex", "Sexual exhaustion", +20, true)
            ]);
        } else {
            // Forced oral change note
            return global.ui.change.buildNotification([
                new Change("Dominated human", "Sanity", +10),
                new Change("Facefucking", "Stamina", -10),
                new Change("Facefucking", "Sexual exhaustion", +20, true)
            ]);
        }
    }

    function partialLootChange() {
        global.kobold.inventory.bag.addConsumable(new Bread());
        global.kobold.inventory.bag.addConsumable(new Butter());
        global.kobold.inventory.bag.addConsumable(new Ham());
    }

    function partialLootChangeNote() {
        return global.ui.change.buildNotification([
            new Change("Loot", global.kobold.inventory.createItemWithTooltip(new Bread(), true), +1),
            new Change("Loot", global.kobold.inventory.createItemWithTooltip(new Butter(), true), +1),
            new Change("Loot", global.kobold.inventory.createItemWithTooltip(new Ham(), true), +1)
        ]);
    }

    function fullLootChange() {
        global.kobold.inventory.addItem(new FancyGlasses());
    }

    function fullLootChangeNote() {
        return global.ui.change.buildNotification([
            new Change("Loot", global.kobold.inventory.createItemWithTooltip(new FancyGlasses(), true), +1)
        ]);
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.varanarSceneName = "NightTwo";
    this.varanarFolderName = "Varanar";
    
    this.misc = new MiscController(id);
    this.arrival = new ChoiceController(id, "arrival");

    this.steal = new ChoiceController(id, "steal");
    this.shed = new ChoiceController(id, "shed");
    this.postshed = new ChangeController(id, "postshed", postShedChange, postShedChangeNote);
    this.inside = new ChoiceController(id, "inside");
    this.follow = new ChangeController(id, "follow", followChange, followChangeNote);
    this.cabinet = new ChoiceController(id, "cabinet");
    this.cabinetSteal = new ChangeController(id, "cabinetSteal", cabinetChange, cabinetChangeNote);

    this.confront = new ChoiceController(id, "confront");
    this.wrestle = new ChangeController(id, "wrestle", wrestleChange, wrestleChangeNote);
    this.choke = new ChangeController(id, "choke", wrestleLootChange, wrestleLootChangeNote);
    this.rape = new ChangeController(id, "rape", rapeChange, rapeChangeNote);
    this.scared = new ChoiceController(id, "scared");
    this.tied = new ChoiceController(id, "tied");
    this.partialLoot = new ChangeController(id, "partialLoot", partialLootChange, partialLootChangeNote);
    this.fullLoot = new ChangeController(id, "fullLoot", fullLootChange, fullLootChangeNote);


    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    this.hasRope = function () {
        global.kobold.history.setIfNotSet(id, "hasRope", global.kobold.inventory.bag.hasMisc(ItemEnum.ROPE));
        return global.kobold.history.getKey(id, "hasRope");
    };

    this.removeRope = function () {
        global.kobold.history.doIfNotSet(id, "removeRope", true, function () {
            global.kobold.inventory.bag.removeMisc(ItemEnum.ROPE);
        });
        return "";
    };

    this.intimidateFail = function () {
        global.kobold.history.doIfNotSet(id, "wrestleFail", true, function () {
            global.kobold.status.stamina.decrease(15, "Ran from farmer");
        });
        return global.ui.change.buildNotification([
            new Change("Ran", "Stamina", -15)
        ]);
    };

    this.threatenFail = function () {
        global.kobold.history.doIfNotSet(id, "wrestleFail", true, function () {
            global.kobold.status.stamina.decrease(15, "Ran from farmer");
        });
        return global.ui.change.buildNotification([
            new Change("Ran", "Stamina", -15)
        ]);
    };

    this.wrestleFail = function () {
        global.kobold.history.doIfNotSet(id, "wrestleFail", true, function () {
            global.kobold.status.physical.decrease(15, "Elbow-bashed");
            global.kobold.status.mental.decrease(10, "Beaten by a woman");
            global.kobold.status.stamina.decrease(15, "Ran from angry mob");
        });
        return global.ui.change.buildNotification([
            new Change("Elbow-bashed", "Health", -15),
            new Change("Beaten by a woman", "Stamina", -15),
            new Change("Ran from angry mob", "Stamina", -15)
        ]);
    };

    this.runChange = function () {
        global.kobold.history.doIfNotSet(id, "shedRun", true, function () {
            global.kobold.status.stamina.decrease(15, "Ran from farmer");
        });
        return global.ui.change.buildNotification([
            new Change("Ran", "Stamina", -15)
        ]);
    };
}

// END scenes\HamletDay2\FarmerHouse\FarmerHouseController.js

// scenes\HamletDay2\FarmerHouse\FarmerHouse.js

/**
 * Farmer's house. Followup scene of the Hamlet.
 */
function FarmerHouse() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "FarmerHouse";
    var sceneName = "Farmer's House";
    var sceneSubfolder = "HamletDay2";
    var ctrl = new FarmerHouseController(sceneId);
    var arrival = new FarmerHouseArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]farmer's house[/t]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new FarmerHouse();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\HamletDay2\FarmerHouse\FarmerHouse.js



// END SCENE FarmerHouse



// SCENE HamletWell

// scenes\HamletDay2\HamletWell\bottom\stuck\HamletWellStuck.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellStuck(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = false;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var help = ctrl.down.get() === HamletWellFallChoice.HELP;
        var toReturn = `[p]
                            The shaking intensifies.
                            You narrowly dodge a few bigger chunks of cobblestone falling down from the near-collapsing wall.
                            Pebbles rain on your head and plummet into the water besides you like an intense, earthen [i]hailstorm[/i].
                            You need to [b]get out of here[/b] and you need to do so [b]fast[/b]!
                            You can't climb, not with this entire place coming down on you.
                            You could dive, but that would mean you'd get buried slowly instead of getting knocked out by a piece of debris.
                            You scream for help${help ? ` again` : ``}, but nobody comes to your aid.
                            Is this well to be your grave?
                        [/p]
                        [p]
                            Panicking, you look all around, for [b]any[/b] way out of this soon-to-be [b]coffin[/b].
                            Then, out of the corner of your eye, you spot what might be your salvation.
                            A small hole in the side of the well, directly underneath the surface of the water,
                            one barely big enough for you to squeeze through.
                            You aren't sure whether that was always there, or if the earthquake uncovered it just now, but you aren't going to question it.
                            You don't even know whether it leads [i]anywhere[/i], but you'd rather die drowning trying to save your own hide,
                            than simply standing here cowering and waiting for your own death.
                        [/p]
                        [p]
                            The well looks about ready to collapse, bigger and bigger pieces of cobblestone breaking off from the wall,
                            dirt starting to seep in from the holes left behind.
                            Taking a deep breath, one that may be your very last one [b]ever[/b],
                            you dive underwater, and you squeeze through the rough stone gap,
                            your belly and your back both chafing against the coarse rocks, your tail only [b]barely[/b] managing to fit.
                        [/p]
                        [p]
                            The moment your feet disappear into the crevice, the well collapses fully behind you, barring your way back,
                            debris falling so close it nearly scrapes your soles.
                            The underwater hole seems to lead to a tiny, submerged tunnel!
                            You only hope it will actually take you somewhere you can [b]breathe[/b],
                            instead of being blocked off at one point or another.
                            You may not have been buried alive, but you're not nearly out of danger yet.
                            Already starting to run out of air, you swim forwards, into the unknown.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\HamletWell\bottom\stuck\HamletWellStuck.js

// scenes\HamletDay2\HamletWell\bottom\swim\HamletWellPants.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellPants(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.pants;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new HamletWellStuck(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var drank = global.kobold.history.shared.well.drankWater;
        var drowned = global.kobold.history.shared.well.drowned;
        var rocky = global.kobold.history.shared.well.fell;
        var arrivalChoice = ctrl.arrival.get();
        var wanted = arrivalChoice === HamletWellArrivalChoice.CLIMB || arrivalChoice === HamletWellArrivalChoice.ROPE;
        var arrivalDescriptor = `You may not have wanted to go down this damp hole, but now that you're here, you might as well explore it fully.`;
        if (wanted) {
            arrivalDescriptor = `You wanted to explore this place, and you're not going to back down just because the air in here is a little stale.`;
            if (rocky) {
                arrivalDescriptor = `Your entrance might not have been very [i]graceful[/i], `;
                arrivalDescriptor += `but you're not going to let a single small setback dissuade you from properly exploring this damp hole.`;
            }
        }
        var toReturn = `[p]
                            You don't concern yourself with escaping the well just yet.
                            ${arrivalDescriptor}
                            You take a few deep breaths, filling your ${drowned ? `still slightly aching` : ``} lungs with as much air as they can possibly hold.
                            Steeling your nerves, you dip your head
                            underwater${rocky ? ` yet again` : ``}${drowned ? `, your mind working overtime to suppress flashbacks
                            from your near-death experience just a few minutes ago` : ``}.
                            Athletically lifting the lower half of your body upwards, while simultaneously sinking your torso deeper,
                            you dive down and submerge yourself completely.
                        [/p]
                        [p]
                            With your eyes having adapted to the dark, and all kobolds luckily being able to see underwater,
                            you immediately notice something dinstinctly [b]shiny[/b] at the bottom of the well.
                            Coins, perhaps?
                            Maybe ${wanted ? `coming down here wasn't such a bad idea after all` : `falling down here wasn't such a bad thing after all`}.
                        [/p]
                        [p]
                            You swim down further, until you suddenly feel a dull scraping at the side of your leg.
                            Looking upwards, you find out that your thigh accidentally rubbed along a sharp rock protruding
                            from the edge of the well${rocky ? `, one you were lucky enough to miss on your earlier way down, ` : ``} without you noticing.
                            Thankfully, your ${global.kobold.inventory.createItemWithTooltip(pants)} protected your scaly hide.
                        [/p]
                        [p]
                            You were so focussed on the promise of gold that you momentarily forgot you're swimming around in a decrepit, age-old well.
                            You don't think the people that made this thing were too concerned about kobolds swimming around in it thousands of years in the future.
                            Alas, your ${global.kobold.inventory.createItemWithTooltip(pants)} were completely [b]torn apart[/b] by the unfortunate scraping.
                            You steer clear of the sharp rock, tear off the remnants of your pants and toss them to the side, letting them float to the bottom of the well.
                            If that [i]is[/i] gold you spy, you'll be able to buy a new pair later.
                        [/p]
                        [p]
                            With your ruined pants discarded, you continue swimming downwards.
                            A few strokes later, you arrive at the muddy ground.
                            To your pleasant surprise, the glint you saw from afar was indeed a shiny coin!
                            And it's not alone.
                            Brushing off some of the mud, you find piece after piece of gold, probably tossed down here by farmers wishing for good fortune.
                            Their prayers were answered, maybe not for them, but at least for you.
                        [/p]
                        [p]
                            You pick through the watery dirt, scrounging as many coins as you can before needing to go up for air.
                            Making sure to avoid hitting the same sharp stone again, you flip yourself upright once more,
                            kick off from the muddy ground, and swim back up to the surface.
                        [/p]
                        [p]
                            Surfacing from your dive, you take a closer look at the armful of coins you salvaged.
                            Most of them are coated with years worth of dirt.
                            You take the time to individually rinse them off in the water, before inspecting them further.
                            They seem to be pretty old, engraved with a design you've never quite seen before.
                            Oh well, you're certain they're still good for use.
                        [/p]
                        [p]
                            You notice that none of the coins have rusted,
                            despite being underwater for what you can only imagine to be quite a long time.
                            Come to think of it, you've never actually seen a rusty piece of gold.
                            You think that perhaps gold [i]can't[/i] rust, but you're not quite certain.
                            It would make sense to make money out of something that can't fall prey to the elements, right?
                            You briefly wonder why people don't make [i]unrustable[/i] weapons out of gold,
                            but you know pretty much nothing of smithing [i]or[/i] weaponry,
                            so you quickly drop the topic and get to work putting the coins into your pouch.
                        [/p]
                        [p]
                            Right as you were about to dive down a second time to gather more shinies, a loud [i]rumbling[/i] noise echoes throughout the well.
                            Everything begins to shake.
                            Small pebbles and rocks break off the wall and plunge into the water besides you.
                            Earthquake?!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\HamletWell\bottom\swim\HamletWellPants.js

// scenes\HamletDay2\HamletWell\bottom\swim\HamletWellScrape.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellScrape(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.scrape;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new HamletWellStuck(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.scrapedThigh = true;
        var drank = global.kobold.history.shared.well.drankWater;
        var drowned = global.kobold.history.shared.well.drowned;
        var rocky = global.kobold.history.shared.well.fell;
        var arrivalChoice = ctrl.arrival.get();
        var wanted = arrivalChoice === HamletWellArrivalChoice.CLIMB || arrivalChoice === HamletWellArrivalChoice.ROPE;
        var arrivalDescriptor = `You may not have wanted to go down this damp hole, but now that you're here, you might as well explore it fully.`;
        if (wanted) {
            arrivalDescriptor = `You wanted to explore this place, and you're not going to back down just because the air in here is a little stale.`;
            if (rocky) {
                arrivalDescriptor = `Your entrance might not have been very [i]graceful[/i], `;
                arrivalDescriptor += `but you're not going to let a single small setback dissuade you from properly exploring this damp hole.`;
            }
        }
        var toReturn = `[p]
                            You don't concern yourself with escaping the well just yet.
                            ${arrivalDescriptor}
                            You take a few deep breaths, filling your ${drowned ? `still slightly aching` : ``} lungs with as much air as they can possibly hold.
                            Steeling your nerves, you dip your head
                            underwater${rocky ? ` yet again` : ``}${drowned ? `, your mind working overtime to suppress flashbacks
                            from your near-death experience just a few minutes ago` : ``}.
                            Athletically lifting the lower half of your body upwards, while simultaneously sinking your torso deeper,
                            you dive down and submerge yourself completely.
                        [/p]
                        [p]
                            With your eyes having adapted to the dark, and all kobolds luckily being able to see underwater,
                            you immediately notice something dinstinctly [b]shiny[/b] at the bottom of the well.
                            Coins, perhaps?
                            Maybe ${wanted ? `coming down here wasn't such a bad idea after all` : `falling down here wasn't such a bad thing after all`}.
                        [/p]
                        [p]
                            You swim down further, until you suddenly feel a sharp, stinging [b]pain[/b] at the side of your leg.
                            Looking upwards, you find out that your thigh accidentally rubbed along a sharp rock protruding
                            from the edge of the well${rocky ? `, one you were lucky enough to miss on your earlier way down, ` : ``} without you noticing.
                            Blood flows liberally from your upper leg, painting part of the water an ever-darkening shade of red.
                            It seems that the pain is only now starting to fully set in.
                            You clutch your thigh, wincing and letting out an underwater [b]scream[/b] out of sheer pain.
                        [/p]
                        [p]
                            You were so focussed on the promise of gold that you momentarily forgot you're swimming around in a decrepit, age-old well.
                            You don't think the people that made this thing were too concerned about kobolds swimming around in it thousands of years in the future.
                            Money be damned, your leg hurts [b]far[/b] too much to continue downwards.
                            Making sure to avoid hitting the same sharp stone again, you flip yourself upright once more, and swim back up to the surface.
                        [/p]
                        [p]
                            Surfacing from your dive, you take a closer look at the wound you sustained.
                            It's a pretty [i]lengthy[/i] gash, running along the upper half of your outer thigh towards your rear,
                            stopping inches before reaching your butt proper.
                            While you don't think it's a very [i]deep[/i] cut, you're still bleeding like a stuck pig,
                            tainting the clear, ${drank ? `sedating` : `fresh`} water with kobold blood.
                        [/p]
                        [p]
                            As if to make matters even worse, a loud [i]rumbling[/i] noise echoes throughout the well.
                            Everything begins to shake.
                            Small pebbles and rocks break off the wall and plunge into the water besides you.
                            Earthquake?!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\HamletWell\bottom\swim\HamletWellScrape.js

// scenes\HamletDay2\HamletWell\bottom\help\beg\HamletWellBegSucceed.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellBegSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var changeController = ctrl.helpBegSucceed;

    var continueId = arguments.callee.name + "ButtonOne"

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.escapedRope = true;
        var toReturn = `[p]
                            You raise your voice loud enough for them to hear, and you speak up.
                            You say it's a good thing the boys found you when they did.
                            You were about to give up hope before they showed up!
                            You, uh, fell down here and you can't seem to get out on your own.
                            You ask them if they could, pretty please, perhaps toss down a rope attached to something so you can climb out of here,
                            or maybe go get one of their parents to help you out or something?
                        [/p]
                        [p]
                            You get the feeling the boys weren't exactly paying attention to your desperate pleading,
                            or maybe they're just not taking you very serious.
                            Althroughout your [i]begging for help[/i], you could hear them utter snickers and sneers, giggles and chortles,
                            like they were joking around instead of listening to you.
                            "[color=red]Hey![/color]" finally comes a response, after a moment of silence.
                            "[color=red]When kobolds are left in the water for too long, do they turn into fish?[/color]"
                        [/p]
                        [p]
                            More laughter.
                            They sound like they're having the time of their lives picking on you,
                            maybe because nothing interesting [i]ever[/i] happens around these parts.
                            Once they're done teasing you, surely, they'll come to your aid, right?
                            They wouldn't just leave you here, would they?
                        [/p]
                        [p]
                            Desperation growing at the back of your head, you decide to play along with the boys' little [i]game[/i].
                            You reply that, no, you don't turn into a fish, but you might become a scary crocodile if they don't let you out soon!
                            "[color=blue]A crocodile?[/color]" answers the third child, in a satirical tone.
                            "[color=blue]A kobold like you is more likely to turn into an [i]alley-gay-whore[/i]![/color]"
                        [/p]
                        [p]
                            The audacity!
                            Where does a young boy even pick up such vulgar words?!
                            All three begin to cackle loudly, laughing their asses off at the third boy's joke.
                            Stupid inbred farmer hicks not raising their kids right.
                            The laughter goes on for the better half of a minute.
                            You're starting to lose your patience with these kids.
                        [/p]
                        [p]
                            The laughter slowly dies down.
                            All three of the kids' heads turn towards the same direction.
                            "[color=red]We're being careful, mom![/color]"
                            You overhear one of the boys replying to who you can only imagine to be his mother.
                            "[color=red]Fiiiine! Let's go.[/color]"
                        [/p]
                        [p]
                            Without even giving you a proper goodbye,
                            and more importantly: without telling this [i]mother[/i] figure about the [b]kobold trapped in the well[/b],
                            the boys hop off of the edge of the well, and disappear from sight.
                            You shout for them to come back, to [b]help you[/b], but your pleas fall on deaf ears, if they were heard at all.
                        [/p]
                        [p]
                            Disillusioned about the pure, good-hearted nature of children, you begin to think of another way out of the well.
                            Calling for help clearly won't get you anywhere.
                            If the children of the hamlet won't help you, then you have little hope for the adults.
                        [/p]
                        [p]
                            Suddenly, a rope falls into the water, right next to you.
                            Wait, what?
                            A rope?
                            Where did that come from?!
                            You look upwards, but you see nobody there.
                            How weird.
                            You give the cord a tug.
                            It seems to be secured to something.
                            That means you can use it to climb out!
                            Someone just [b]saved[/b] you!
                        [/p]
                        ${global.ui.condSucc(`Charisma greater than 5`)}
                        [p]
                            Maybe one of the kids [b]did[/b] care, or perhaps a random passerby decided to help you out.
                            Regardless, you don't have much time to think things over. 
                            If you wait too long, something [i]bad[/i] might happen to [b]ruin[/b] your one and only hope of getting out of here.
                            Something bad [b]always[/b] happens.
                        [/p]
                        [p]
                            But not this time.
                            You're not going to take any chances sitting here pondering just [i]who[/i] it was that came to your rescue.
                            You'll see for yourself when you get out of the well.
                            You immediately grab the rope, and begin to shimmy upwards, towards your freedom.
                        [/p]
                        [p]
                            While the climb is exhausting, at least the rope makes it somewhat bearable.
                            You press your feet against the wall, and vertically walk to safety, pulling yourself upwards one step at a time.
                            A few minutes later, you find yourself at the very top, and after climbing over the small above-ground portion of the well,
                            your feet plop down on steady ground once more.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    You have escaped the well!
                                [/b]
                            [/c]
                        [/p]
                        [p]
                            After dropping to your knees and only [i]barely[/i] preventing yourself from kissing the dirt out of sheer joy,
                            you stand up and look for your oh-so kind benefactor.
                            However, there is nobody to be found.
                            The hamlet looks as deserted as it did the moment you entered it.
                        [/p]
                        [p]
                            You are left with more questions than answers.
                            Who saved you?
                            Was it one of the kids?
                            A farmer?
                            A random, compassionate guard passing by the area?
                            You're glad you got out, but you'd at least like to [b]thank[/b] whoever helped you.
                            Alas, it seems your [i]sympathizer[/i] wishes to remain anonymous.
                            A shame.
                        [/p]
                        [p]
                            With mixed feelings, you decide you've stuck around the hamlet for too long already.
                            It's time to get moving.
                            If you hurry, you might still make it to Varanar before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\HamletWell\bottom\help\beg\HamletWellBegSucceed.js

// scenes\HamletDay2\HamletWell\bottom\help\beg\HamletWellBegFail.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellBegFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.helpBegFail;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new HamletWellStuck(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You raise your voice loud enough for them to hear, and you speak up.
                            You say it's a good thing the boys found you when they did.
                            You were about to give up hope before they showed up!
                            You, uh, fell down here and you can't seem to get out on your own.
                            You ask them if they could, pretty please, perhaps toss down a rope attached to something so you can climb out of here,
                            or maybe go get one of their parents to help you out or something?
                        [/p]
                        [p]
                            You get the feeling the boys weren't exactly paying attention to your desperate pleading,
                            or maybe they're just not taking you very serious.
                            Althroughout your [i]begging for help[/i], you could hear them utter snickers and sneers, giggles and chortles,
                            like they were joking around instead of listening to you.
                            "[color=red]Hey![/color]" finally comes a response, after a moment of silence.
                            "[color=red]When kobolds are left in the water for too long, do they turn into fish?[/color]"
                        [/p]
                        [p]
                            More laughter.
                            They sound like they're having the time of their lives picking on you,
                            maybe because nothing interesting [i]ever[/i] happens around these parts.
                            Once they're done teasing you, surely, they'll come to your aid, right?
                            They wouldn't just leave you here, would they?
                        [/p]
                        [p]
                            Desperation growing at the back of your head, you decide to play along with the boys' little [i]game[/i].
                            You reply that, no, you don't turn into a fish, but you might become a scary crocodile if they don't let you out soon!
                            "[color=blue]A crocodile?[/color]" answers the third child, in a satirical tone.
                            "[color=blue]A kobold like you is more likely to turn into an [i]alley-gay-whore[/i]![/color]"
                        [/p]
                        [p]
                            The audacity!
                            Where does a young boy even pick up such vulgar words?!
                            All three begin to cackle loudly, laughing their asses off at the third boy's joke.
                            Stupid inbred farmer hicks not raising their kids right.
                            The laughter goes on for the better half of a minute.
                            You're starting to lose your patience with these kids.
                        [/p]
                        [p]
                            The laughter slowly dies down.
                            All three of the kids' heads turn towards the same direction.
                            "[color=red]We're being careful, mom![/color]"
                            You overhear one of the boys replying to who you can only imagine to be his mother.
                            "[color=red]Fiiiine! Let's go.[/color]"
                        [/p]
                        [p]
                            Without even giving you a proper goodbye,
                            and more importantly: without telling this [i]mother[/i] figure about the [b]kobold trapped in the well[/b],
                            the boys hop off of the edge of the well, and disappear from sight.
                            You shout for them to come back, to [b]help you[/b], but your pleas fall on deaf ears, if they were heard at all.
                        [/p]
                        [p]
                            Disillusioned about the pure, good-hearted nature of children, you begin to think of another way out of the well.
                            Calling for help clearly won't get you anywhere.
                            If the children of the hamlet won't help you, then you have little hope for the adults.
                        [/p]
                        ${global.ui.condFail(`Charisma below 6`)}
                        [p]
                            Halfway through conceiving an intricate escape plan concerning your tail and a star-shaped rock of some kind,
                            A loud [i]rumbling[/i] noise echoes throughout the well.
                            Everything begins to shake.
                            Small pebbles and rocks break off the wall and plunge into the water besides you.
                            Earthquake?!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\HamletWell\bottom\help\beg\HamletWellBegFail.js

// scenes\HamletDay2\HamletWell\bottom\help\HamletWellBeg.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellBeg(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellHelpChoice.BEG;
    var reqController = ctrl.help;

    var toFail = new HamletWellBegFail(ctrl);
    var toSuccess = new HamletWellBegSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.charisma() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\HamletWell\bottom\help\HamletWellBeg.js

// scenes\HamletDay2\HamletWell\bottom\help\HamletWellScold.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellScold(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellHelpChoice.SCOLD;
    var reqController = ctrl.help;
    var subController = ctrl.preContinue;
    var changeController = ctrl.helpScold;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new HamletWellStuck(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You raise your voice loud enough for them to hear, and you speak up.
                            You tell the boys that they should mind their manners.
                            They can't just go around calling random people dumb!
                            Now, they should go get their parents, and if they're super nice to you,
                            you won't even tell the grown-ups what their children called you.
                        [/p]
                        [p]
                            For a moment, the trio goes silent.
                            While you can't quite tell from where you are [i]swimming[/i],
                            you can imagine them staring at one another in disbelief.
                            Any minute now, they'll apologize, and go fetch some farmers to get you out of here.
                            "[color=red]Hey![/color]" finally comes a response.
                            "[color=red]Since the kobold likes water so much, how about I give it some more?[/color]"
                        [/p]
                        [p]
                            [i]Give you some more?[/i]
                            What does that even mean?
                            Are they going to pour water down the well?
                            "[color=blue]No, Leif, your mom said you weren't allowed to pee in the well![/color]"
                            Wait, what?
                            [i]Pee in the well?[/i]
                            "[color=red]Yeah, but she never said anything about peeing on kobolds![/color]"
                            [i]Oh no.[/i]
                        [/p]
                        [p]
                            From all the way at the top, a thin stream of watery urine sprays down into the well.
                            Due to the remarkable distance between the two of you,
                            the boy mostly ends up doing nothing but splattering the cobblestone wall with pee.
                            A few droplets, manage to make into the water next to you, but you never get hit directly by the troublesome youth.
                        [/p]
                        [p]
                            The stream soon comes to an end, the echoing sound of the tiny flow of splattering piss replaced by your loud, angry hissing,
                            which, in turn, is overshadowed by the trio's building, booming laughter.
                            "[color=red]That enough for you, lizard?![/color]"
                            One of them shouts down at you.
                            You curse your lungs out at them, hurling a barrage of insults and threats their way.
                            "[color=blue]Sounds to me like it wants some more.[/color]" comments one of them.
                            "[color=green]Isn't your uncle a horse? I think I once saw him pee for a minute straight. Let's go get him![/color]"
                        [/p]
                        [p]
                            With that being said, the boys rush off, presumably to find some adult to help them pick on the [b]kobold trapped in the well[/b],
                            You scream for them to come back, to stop playing around and to [b]help you[/b], but your pleas fall on deaf ears, if they were heard at all.
                            You don't really know what to do.
                            If they [i]do[/i] end up getting that [i]horse[/i] they were talking about, the adult might be smart enough to get you out of here.
                            On the other hand, if he's as sadistic as his brother's son is, then you might not want to stick around to wait for his [i]assistance[/i].
                        [/p]
                        [p]
                            Disillusioned about the pure, good-hearted nature of children, you begin to think of another way out of the well.
                            Halfway through conceiving an intricate escape plan concerning your tail and a star-shaped rock of some kind,
                            A loud [i]rumbling[/i] noise echoes throughout the well.
                            Everything begins to shake.
                            Small pebbles and rocks break off the wall and plunge into the water besides you.
                            Earthquake?!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\HamletWell\bottom\help\HamletWellScold.js

// scenes\HamletDay2\HamletWell\bottom\climb\HamletWellClimbRelent.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellClimbRelent(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellClimbUpChoice.GIVE;
    var reqController = ctrl.climbUp;
    var subController = ctrl.preContinue;
    var changeController = ctrl.climbUpRelent;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new HamletWellStuck(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var drank = global.kobold.history.shared.well.drankWater;
        var drowned = global.kobold.history.shared.well.drowned;
        var rocky = global.kobold.history.shared.well.fell;
        var arrivalChoice = ctrl.arrival.get();
        var wanted = arrivalChoice === HamletWellArrivalChoice.CLIMB || arrivalChoice === HamletWellArrivalChoice.ROPE;
        var rope = arrivalChoice === HamletWellArrivalChoice.ROPE;
        var toReturn = `[p]
                            This is dangerous.
                            The further up you go, the more likely you'll be to [b]severely[/b] hurt yourself when you inevitably fuck up.
                            Growing ever more exhausted, you doubt it'll take long for you to [i]slip and fall[/i], or to pick another unstable rock.
                            You've already gotten lucky, and managed to prevent yourself from falling down [i]once[/i].
                            You don't want to keep testing fate.
                        [/p]
                        [p]
                            Letting out a loud, exacerbating sigh, you begin to climb back down the
                            well${wanted ? `, just like before${rope ? `, but this time without a rope` : ``}` : ``}.
                            ${!rocky ? `You've done this already, a fact that significantly boosts confidence in your climbing abilities.
                            Still, your journey is not without peril.` : `You hope you can go down the well [i]without[/i] falling this time around.
                            The shoddy cobblestone wall you're scaling isn't making this any easier for you, though.`}
                        [/p]
                        [p]
                            The earlier near-fall has made you somewhat paranoid.
                            You question every step you take, thoroughly testing every stone before grabbing it with your hand or placing your foot down upon it.
                            While being diligent makes the entire ordeal a bit [i]safer[/i], it also complicates things.
                            Entire paths which you would have taken without a doubt before are suddenly deemed too hazardous by your frightened mind.
                            You find yourself having to [b]stretch[/b] your body a lot more than usual to be able to [b]reach[/b] [i]safer[/i] points,
                            to put your aching sense of self-preservation at ease.
                        [/p]
                        [p]
                            Despite needing to test the very limits of how far you can extend your body, you manage to make it back down to the cold water without falling down.
                            You slip back into the cold liquid, to give your aching limbs a bit of much needed rest.
                            Even if you didn't manage to escape the well, sure was one hell of a workout.
                            You hope you'll be able to make it out of here alive, so you can put your newfound flexibility to use.
                        [/p]
                        [p]
                            Right as you were starting to consider other avenues of escape, a loud [i]rumbling[/i] noise echoes throughout the well.
                            Everything begins to shake.
                            Small pebbles and rocks break off the wall and plunge into the water besides you.
                            Earthquake?!
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\HamletWell\bottom\climb\HamletWellClimbRelent.js

// scenes\HamletDay2\HamletWell\bottom\climb\keep\HamletWellKeepClimbingSucceed.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellKeepClimbingSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var changeController = ctrl.climbUpSucceed;

    var continueId = arguments.callee.name + "ButtonOne"

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.escapedClimb = true;
        var toReturn = `[p]
                            You need to keep going.
                            This might be your only way out, and you can't give up just because of a single bad thing happening.
                            Besides, you're already halfway to the top.
                            It'd be a darn [b]shame[/b] to quit now!
                            The way down would be as much of a climb as the way up is.
                            Plus, you're far more motivated by the thought of being [b]freed[/b] from this [i]watery prison[/i]
                            than you are by the premise of being stuck at the bottom [i]again[/i].
                        [/p]
                        [p]
                            It's decided, then.
                            You have to reach the top, no matter how tired or worn out your muscles may be.
                            You raise your arm to a different rock on the same level as the one that broke off a minute ago.
                            Then, the unexpected happens.
                            Like the one before it, this stone, too, breaks from the wall!
                            You are thrown off balance a second time, feeling even more startled than before.
                            You would never have expected two [b]adjacent[/b] rocks to [b]both[/b] be loose!
                        [/p]
                        [p]
                            Time itself seems to slow down as adrenaline floods your system.
                            You can't keep your balance on the shaky rock both your feet are on.
                            You need something new to grab onto, and you need it quick.
                            Looking upwards, you see two stones in the perfect position to hang off of.
                            They're far too high up for you to be able to reach them, though.
                            There's only one answer.
                            Only one thing you can do.
                        [/p]
                        [p]
                            With a soft growl, you [b]leap[/b] upwards, mere [b]moments[/b] before losing your balance completely.
                            Impressing even yourself, you manage to grab onto both of the rocks you were oggling earlier,
                            suspending your entire weight by your arms, while your legs fail to find support.
                            Adrenaline fueling your actions,
                            you amaze yourself further by clambering upwards using nothing but sheer armstrength to hoist your body up onto the next set of grabbable stones,
                            allowing your feet to rest on the rocks you jumped to, and bringing you a whole lot closer to your goal of escaping the well.
                        [/p]
                        ${global.ui.condSucc(`Agility and strength greater than 5`)}
                        [p]
                            You decide to keep climbing, before your sudden burst of strength fades.
                            A few minutes later, you emerge from the well, lifting yourself up and over the stone wall.
                            As adrenaline fades from your system, you collapse on the ground, all four of your limbs completely spent,
                            even your [b]tail[/b] feeling exhausted from the intensive, heart-rattling climb.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    You have escaped the well!
                                [/b]
                            [/c]
                        [/p]
                        [p]
                            You roll onto your back, spread eagle, breathing heavily, staring at the clear sky which you are grateful to see fully once more.
                            That was [b]hard[/b].
                            You didn't even [b]know[/b] you could do some of the things you did.
                            You thought you were going to die.
                            You only [b]barely[/b] made it out alive by the skin of your teeth.
                            In a weird way, however, the rush of excitement you felt when that rock broke off was [b]exhilarating[/b].
                            Defying death was just what you needed to take your mind off of Seki.
                            Maybe you should try that again sometimes.
                            Worst case scenario, you'll meet the old bird in the afterlife.
                        [/p]
                        [p]
                            After a small recovery period, you rise to your spent feet, wisely refraining from leaning against the well.
                            While your limbs are still aching from the extreme strain you put them under, and your heart is still reeling from the intense stress,
                            you feel like you're fit enough to continue on your journey to Varanar.
                            A bit of walking might put both your legs and your mind at ease, and your arms will appreciate a bit of inactivity.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.varanarSceneName, ctrl.varanarFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\HamletWell\bottom\climb\keep\HamletWellKeepClimbingSucceed.js

// scenes\HamletDay2\HamletWell\bottom\climb\keep\HamletWellKeepClimbingFail.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellKeepClimbingFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.climbUpFail;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new HamletWellStuck(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You need to keep going.
                            This might be your only way out, and you can't give up just because of a single bad thing happening.
                            Besides, you're already halfway to the top.
                            It'd be a darn [b]shame[/b] to quit now!
                            The way down would be as much of a climb as the way up is.
                            Plus, you're far more motivated by the thought of being [b]freed[/b] from this [i]watery prison[/i]
                            than you are by the premise of being stuck at the bottom [i]again[/i].
                        [/p]
                        [p]
                            It's decided, then.
                            You have to reach the top, no matter how tired or worn out your muscles may be.
                            You raise your arm to a different rock on the same level as the one that broke off a minute ago.
                            Then, the unexpected happens.
                            Like the one before it, this stone, too, breaks from the wall!
                            You are thrown off balance a second time, feeling even more startled than before.
                            You would never have expected two [b]adjacent[/b] rocks to [b]both[/b] be loose!
                        [/p]
                        [p]
                            You try to reach for a new gripping point, much like you did before,
                            but this time you [b]aren't[/b] so lucky as to be able to simply grab a new rock.
                            You manage to stay put, balancing only on your feet, for just a few seconds more.
                            Time itself seems to slow down, as you tumble backwards, falling from the wall.
                        [/p]
                        ${global.ui.condFail(ctrl.misc.strength() < 6 ? `Strength below 6` : `Agility below 6`)}
                        [p]
                            Instinctively, to try to prevent yourself from [b]scraping[/b] against the well's rocky interior,
                            you kick yourself off from the ledge you were standing on at the very last moment.
                            Bad idea.
                            Due to the space you're in being so [b]confined[/b],
                            you end up harshly [b]bashing[/b] the soft back of your head against the other side of the well.
                            Your vision goes black.
                            You pass out, your limp body plummeting down into the cold water below.
                        [/p]
                        [hl]
                        [p]
                            You wearily open your eyes, what you can only imagine to be minutes, or maybe [b]hours[/b] after passing out and falling down.
                            You find yourself floating on your back on the surface of the frigid well water.
                            Your head hurts.
                            Your [b]everything[/b] hurts.
                            You slowly come to an upright swimming position, clutching the back of your skull.
                            Is that...[i]blood[/i] you feel?
                        [/p]
                        [p]
                            You don't get long to inspect your injuries, or even to contemplate what happened.
                            A loud [i]rumbling[/i] noise echoes throughout the well.
                            Everything begins to shake, clearing your foggy mind despite rousing a killer headache.
                            Small pebbles and rocks break off the wall and plunge into the water besides you.
                            Earthquake?!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\HamletWell\bottom\climb\keep\HamletWellKeepClimbingFail.js

// scenes\HamletDay2\HamletWell\bottom\climb\HamletWellKeepClimbing.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellKeepClimbing(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellClimbUpChoice.KEEP;
    var reqController = ctrl.climbUp;

    var toFail = new HamletWellKeepClimbingFail(ctrl);
    var toSuccess = new HamletWellKeepClimbingSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.strength() > 5 && ctrl.misc.agility() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\HamletWell\bottom\climb\HamletWellKeepClimbing.js

// scenes\HamletDay2\HamletWell\bottom\HamletWellClimbUp.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellClimbUp(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellFallChoice.CLIMB;
    var reqController = ctrl.down;
    var subController = ctrl.climbUp;
    var changeController = ctrl.climbUp;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new HamletWellClimbRelent(ctrl);
    var secondNext = new HamletWellKeepClimbing(ctrl);

    var firstChoice = HamletWellClimbUpChoice.GIVE;
    var secondChoice = HamletWellClimbUpChoice.KEEP;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var rocky = global.kobold.history.shared.well.fell;
        var arrivalChoice = ctrl.arrival.get();
        var wanted = arrivalChoice === HamletWellArrivalChoice.CLIMB || arrivalChoice === HamletWellArrivalChoice.ROPE;
        var arrivalDescriptor = `You didn't want to be in here to begin with. Falling down was an accident, and nothing more. `;
        arrivalDescriptor += `If you wanted to explore these depths, you'd have climbed down here on your own accord.`;
        if (wanted) {
            arrivalDescriptor = `You wanted to explore this place, but after the tiring climb downwards, all you really want to do is leave. `;
            arrivalDescriptor += `You've already proven yourself. There's no point to sticking around.`;
            if (rocky) {
                arrivalDescriptor = `You wanted to explore this place, but after your harrowing [i]entrance[/i], all you really want to do is leave.`;
            }
        }
        var toReturn = `[p]
                            Time to get out of this damp hole.
                            ${arrivalDescriptor}
                            You reach up as high as you can, grabbing onto a rock protruding from the wall,
                            you find a similar stone to place a foot on, and so you begin to climb from the water, and hopefully out of the well.
                        [/p]
                        [p]
                            Your wall-scaling, while exhausting, goes without a hitch.
                            That is, until you reach the approximate halfway point.
                            You are getting tired, weary, perhaps even [i]too[/i] weary to get to the top.
                            You grab one of the cobblestone walls' rocks, as you've done a dozen times before.
                            Then, you try to lift yourself up to the next set of grip-points, only for the lime holding the stone in place to fail to support your weight.
                            You accidentally tear it straight from the wall, and in surprise, you let it slip from your grasp.
                            It drops down into the water below with a siumple, yet fright-inspiring [i]plop[/i].
                        [/p]
                        [p]
                            The sudden lack of grip throws you off balance, and almost sends you tumbling back down the well${rocky ? ` again` : ``}.
                            It is only through sheer luck and well-honed reflexes that you manage to quickly take hold of a different anchor point to maintain your position.
                            While you didn't fall, you are thoroughly shaken.
                            The moment the rock dislodged, your heart skipped a beat.
                            Doubt floods your brain.
                            Maybe this wasn't the best idea after all.
                            Maybe you should go back down and find another way to escape, before you hurt yourself.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Give up</button>`;

        var tooltip = global.parser.parse(`You'd rather not hurt yourself trying to get out.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Keep trying</button>`;

        var tooltip = global.parser.parse(`You're halfway there. Just a little bit more!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var HamletWellClimbUpChoice = {
    GIVE: 1,
    KEEP: 2
};

// END scenes\HamletDay2\HamletWell\bottom\HamletWellClimbUp.js

// scenes\HamletDay2\HamletWell\bottom\HamletWellHelp.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellHelp(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellFallChoice.HELP;
    var reqController = ctrl.down;
    var subController = ctrl.help;
    var changeController = ctrl.help;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new HamletWellScold(ctrl);
    var secondNext = new HamletWellBeg(ctrl);

    var firstChoice = HamletWellHelpChoice.SCOLD;
    var secondChoice = HamletWellHelpChoice.BEG;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var drank = global.kobold.history.shared.well.drankWater;
        var drowned = global.kobold.history.shared.well.drowned;
        var rocky = global.kobold.history.shared.well.fell;
        var arrivalChoice = ctrl.arrival.get();
        var wanted = arrivalChoice === HamletWellArrivalChoice.CLIMB || arrivalChoice === HamletWellArrivalChoice.ROPE;
        var arrivalDescriptor = `You didn't want to be in here to begin with.`;
        if (wanted) {
            arrivalDescriptor = `You wanted to explore this place, but after the tiring climb downwards,
                                all you can do is dread the thought that you might not be able to escape.`;
            if (rocky) {
                arrivalDescriptor = `You wanted to explore this place, but you couldn't even climb [i]down[/i] the well properly.`;
            }
        }
        var toReturn = `[p]
                            You don't think you can get out of this damp hole on your own.
                            ${arrivalDescriptor}
                            Your only option left is to shout for help, and hope somebody hears you.
                        [/p]
                        [p]
                            You take a deep breath, and [b]scream[/b] for aid.
                            Your shriek echoes throughout the well, but otherwise goes unanswered.
                            Doubt growing in the back of your mind, you shout for assistance a second time.
                            Once the reverberation comes to an end, you are left with silence.
                            Completely disheartened, you gaze upwards at the small bit of sky you can see, and let out a small sigh.
                        [/p]
                        [p]
                            Suddenly, a trio of silhouettes appear at the top of the well.
                            Are you imagining things?
                            You rub your eyes, thinking you were hit by a rock and are now hallucinating.
                            The figures persist.
                            They seem to be talking to one another, although you can't quite make out what they're saying.
                            Right as you were about to call out for aid a third time,
                            one of the three shouts into the well, with a shrill and juvenile sounding voice:
                            "[color=green]That's not a swimming pool, stupid lizard![/color]"
                        [/p]
                        [p]
                            All three of them burst out into laughter, loud enough for you to hear all the way down the well.
                            Judging by what little you can see of them, the pitch of their voices and their general attitude,
                            you guess these aren't [i]grown men[/i] you're talking to, but rather a group of young boys.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Scold</button>`;

        var tooltip = global.parser.parse(`Berate the kids for being so rude.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Beg</button>`;

        var tooltip = global.parser.parse(`Beg for help. Your life might be at stake!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var HamletWellHelpChoice = {
    SCOLD: 1,
    BEG: 2
};

// END scenes\HamletDay2\HamletWell\bottom\HamletWellHelp.js

// scenes\HamletDay2\HamletWell\bottom\HamletWellSwim.js

/**
 * Bottom section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellSwim(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellFallChoice.SWIM;
    var reqController = ctrl.down;

    var toFail = new HamletWellScrape(ctrl);
    var toSuccess = new HamletWellPants(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.pants();
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }
        
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\HamletWell\bottom\HamletWellSwim.js

// scenes\HamletDay2\HamletWell\fall\climb\HamletWellClimbSucceed.js

/**
 * Arrival section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellClimbSucceed(controller, firstNext, secondNext, thirdNext) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.down;
    var changeController = ctrl.climbSucceed;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    //var firstNext = new HamletWellSwim(ctrl);
    //var secondNext = new HamletWellClimb(ctrl);
    //var thirdNext = new HamletWellHelp(ctrl);

    var firstChoice = HamletWellFallChoice.SWIM;
    var secondChoice = HamletWellFallChoice.CLIMB;
    var thirdChoice = HamletWellFallChoice.HELP;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            After your depressing train of thoughts, you decide it's time to inspect the well more thoroughly.
                            There might be some coins at the bottom, and who knows what else you might find?
                            Some good old adventuring is exactly what you need to lift your spirits.
                        [/p]
                        [p]
                            You figure you might be able to scale down the cobblestone wall.
                            There's plenty of rocks sticking out, and like you noticed earlier,
                            the lime in between the stones is soft enough for you to carve your claws into, should the need arise.
                            While you're no expert in climbing, you've scaled many trees and even a couple of buildings before.
                            You're certain this'll go fine, but still, you're more than a little scared of the dark depths that await you.
                        [/p]
                        [p]
                            Gathering your courage, you let out a determined grunt.
                            Knights are fearless, and if you want to be one, you can't shy away from adventure.
                            You need to plunge head-first into danger!
                            Well, maybe not head-first, but you still need to go down the well,
                            if just to prove to yourself that you [b]do[/b] have what it takes to stop being a thief one day.
                        [/p]
                        [p]
                            You hop on top of the well's edge.
                            Facing away from the watering hole, you squat down.
                            With both your hands, you take hold of the rim of the cobblestone wall.
                            You let out a deep, nerve-soothing sigh.
                            Here you go.
                        [/p]
                        [p]
                            You lower your body into the watering hole, hanging off your hands until you can find a good place to put your feet.
                            At first, you only find [b]one[/b] rock sticking out within your reach, and it doesn't seem to be a very stable one.
                            Stretching your agile body to its very limits, however, you discover a second, far more stable stone platform just a few inches lower.
                            You place one of your feet down onto the rock, and with the first step made, you begin your descent proper.
                        [/p]
                        [p]
                            You deftly climb down the entire height of the circular cobblestone wall.
                            More than once do you find yourself in a perilous situation, but despite the well itself falling apart at the seams,
                            you manage to stick to the wall all the way to the bottom, where you lower your lithe frame into the cool water.
                            The water's too deep for you to stand up right.
                            Your gently kicking legs and your arms slowly rocking back and forth along the surface of the water keep your light frame afloat.
                            The damp, stale air at the bottom of the well is of poor quality,
                            but at least keeping yourself from sinking takes less effort than climbing down did.
                        [/p]
                        ${global.ui.condSucc(`Agility greater than 5`)}
                        [p]
                            You take a moment to recover from the strenuous climb.
                            You're glad and proud that you made it all the way down without falling.
                            A good display of flexibility, that you're certain not everyone can match.
                            You let out a sigh of relief, quickly followed by an annoyed groan as you tilt your head upwards,
                            and realize that you have to climb all the way back [b]up[/b] to get out of this place.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Dive down</button>`;

        var tooltip = global.parser.parse(`You're not worried about getting out just yet. Dive down to see what's at the bottom of the well.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Climb out</button>`;

        var tooltip = global.parser.parse(`Scale the walls to get out of this claustrophobic hole.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Call for help</button>`;

        var tooltip = global.parser.parse(`You don't think you can climb up by yourself. Maybe someone's around to help you out.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

// END scenes\HamletDay2\HamletWell\fall\climb\HamletWellClimbSucceed.js

// scenes\HamletDay2\HamletWell\fall\climb\HamletWellClimbFail.js

/**
 * Arrival section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellClimbFail(controller, firstNext, secondNext, thirdNext) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.down;
    var changeController = ctrl.climbFail;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    //var firstNext = new HamletWellSwim(ctrl);
    //var secondNext = new HamletWellClimb(ctrl);
    //var thirdNext = new HamletWellHelp(ctrl);

    var firstChoice = HamletWellFallChoice.SWIM;
    var secondChoice = HamletWellFallChoice.CLIMB;
    var thirdChoice = HamletWellFallChoice.HELP;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.fell = true;
        var toReturn = `[p]
                            After your depressing train of thoughts, you decide it's time to inspect the well more thoroughly.
                            There might be some coins at the bottom, and who knows what else you might find?
                            Some good old adventuring is exactly what you need to lift your spirits.
                        [/p]
                        [p]
                            You figure you might be able to scale down the cobblestone wall.
                            There's plenty of rocks sticking out, and like you noticed earlier,
                            the lime in between the stones is soft enough for you to carve your claws into, should the need arise.
                            While you're no expert in climbing, you've scaled many trees and even a couple of buildings before.
                            You're certain this'll go fine, but still, you're more than a little scared of the dark depths that await you.
                        [/p]
                        [p]
                            Gathering your courage, you let out a determined grunt.
                            Knights are fearless, and if you want to be one, you can't shy away from adventure.
                            You need to plunge head-first into danger!
                            Well, maybe not head-first, but you still need to go down the well,
                            if just to prove to yourself that you [b]do[/b] have what it takes to stop being a thief one day.
                        [/p]
                        [p]
                            You hop on top of the well's edge.
                            Facing away from the watering hole, you squat down.
                            With both your hands, you take hold of the rim of the cobblestone wall.
                            You let out a deep, nerve-soothing sigh.
                            Here you go.
                        [/p]
                        [p]
                            You lower your body into the watering hole, hanging off your hands until you can find a good place to put your feet.
                            There's only [b]one[/b] rock sticking out within your reach, and it doesn't seem to be a very stable one.
                            Darnit, if only you were a little bit bigger!
                            Fearfully, you place one of your feet down onto the unsteady stone, which shifts around ever so slightly.
                            Still, it holds.
                            For now.
                        [/p]
                        [p]
                            As you move one of your hands from the rim of the well in search of a lower place to put it,
                            the rock under your foot suddenly gives way, breaking from the wall and falling down!
                            Your let out a fearful, surprised yelp.
                            Your foot no longer supported, you find yourself hanging from only a single hand, your claws digging into the rim of the well.
                            It's not enough.
                            You can feel your fingers slipping away.
                            You try to bring your other hand up again, to return to hanging off the edge with eight digits instead of four.
                            Before you manage to strengthen your hold, however, your fingers give way to gravity's relentless pull.
                            Your last bit of grip lost, you fall back-first into the pitch-black darkness of the well.
                        [/p]
                        ${global.ui.condFail(`Agility below 6`)}
                        [p]
                            You panic, [b]screaming[/b] in terror, thinking this might be the end of you.
                            You wantonly flail your limbs about, scratching and scraping the cobblestone wall surrounding you on all sides with your claws,
                            hoping to [i]latch on[/i] to something, or at least to somewhat slow your fall.
                            Your efforts are in vain.
                            In a matter of seconds, your terrified shout is cut off,
                            interrupted by a loud [b]splash[/b] as you plummet into the cold water at the bottom of the well.
                        [/p]
                        [p]
                            Frigid water rushes into your throat via your open mouth, filling your lungs long before you manage to close your maw.
                            Your descent is quickly halted by the cool liquid, doubly so as your back bumps against the ground at the very base of the well.
                            Luckily, you quickly manage to find your way back up to surface of the water.
                        [/p]
                        [p]
                            Emerging from the ice-cold liquid, you immediately begin coughing up the watery contents of your respiratory system.
                            It's like a drink accidentally shot down the wrong hole, only ten times worse.
                            After agonizingly [b]vomiting[/b] water from your trachea, you resume your coughing fit.
                            It is only after a couple of minutes of recovering, both mentally and physically, that you can breathe normally once more.
                            Your gently kicking legs and your arms slowly rocking along the surface of the water keep your light frame afloat.
                            The damp, stale air at the bottom of the well is of poor quality, but at least with the liquid expelled from your lungs, you can breathe again.
                        [/p]
                        [p]
                            The rock you put your foot on broke off, leading to you falling down the well.
                            An unfortunate spell of bad luck that nearly cost you your life.
                            If it weren't for the water cushioning your fall, you'd have been a [b]dead[/b] lizard by now.
                            You didn't die, though.
                            Not yet, anyway.
                            You let out a sigh of relief, quickly followed by an annoyed groan as you tilt your head upwards,
                            and realize that you're several meters underground with no easy way up.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Dive down</button>`;

        var tooltip = global.parser.parse(`You're not worried about getting out just yet. Dive down to see what's at the bottom of the well.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Climb out</button>`;

        var tooltip = global.parser.parse(`Scale the walls to get out of this claustrophobic hole. Surely you won't fall [b]twice[/b], right?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Call for help</button>`;

        var tooltip = global.parser.parse(`You don't think you can climb up by yourself. Maybe someone's around to help you out.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

// END scenes\HamletDay2\HamletWell\fall\climb\HamletWellClimbFail.js

// scenes\HamletDay2\HamletWell\fall\HamletWellClimb.js

/**
 * Arrival section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellClimb(controller, firstNext, secondNext, thirdNext) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellArrivalChoice.CLIMB;
    var reqController = ctrl.arrival;

    var toFail = new HamletWellClimbFail(ctrl, firstNext, secondNext, thirdNext);
    var toSuccess = new HamletWellClimbSucceed(ctrl, firstNext, secondNext, thirdNext);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.agility() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }
        
        ctrl.bgm.loop();
        global.ui.imageBackground.hide();
        global.ui.applyTheme("hamlet_well");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\HamletWell\fall\HamletWellClimb.js

// scenes\HamletDay2\HamletWell\fall\HamletWellRope.js

/**
 * Arrival section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellRope(controller, firstNext, secondNext, thirdNext) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellArrivalChoice.ROPE;
    var reqController = ctrl.arrival;
    var subController = ctrl.down;
    var changeController = ctrl.rope;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    //var firstNext = new HamletWellSwim(ctrl);
    //var secondNext = new HamletWellClimb(ctrl);
    //var thirdNext = new HamletWellHelp(ctrl);

    var firstChoice = HamletWellFallChoice.SWIM;
    var secondChoice = HamletWellFallChoice.CLIMB;
    var thirdChoice = HamletWellFallChoice.HELP;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.drowned = true;
        global.kobold.history.shared.well.fell = true;
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            After your depressing train of thoughts, you decide it's time to inspect the well more thoroughly.
                            There might be some coins at the bottom, and who knows what else you might find?
                            Some good old adventuring is exactly what you need to lift your spirits.
                        [/p]
                        [p]
                            You grab your ${global.kobold.inventory.createItemWithTooltip(new Rope())}, hoping to scale down the well.
                            There isn't anything around to tie it to, though.
                            The cable attached to the well's bucket is simply coiled off to the side,
                            but you don't think your own ${global.kobold.inventory.createItemWithTooltip(new Rope())} is long enough to reach the very bottom.
                            Still, you'd rather trust something Seki sold you than a random cord that's probably been out here for [b]years[/b].
                        [/p]
                        [p]
                            Pondering the logistics of this situation for a few seconds, you spot the solution lying on the ground.
                            A chunk of cobblestone has broken off the side of the well.
                            You figure it's perfect for tying your ${global.kobold.inventory.createItemWithTooltip(new Rope())} around!
                            It [i]seems[/i] and [i]feels[/i] heavy enough to support your weight, although if it actually [i]will[/i] remains to be seen.
                        [/p]
                        [p]
                            You attach one end of your cable to the piece of rock-wall, and toss the other end down the well.
                            You cast your gaze downwards, but it's too dark to see the bottom.
                            You're sure your eyes'll adapt to the darkness once you're down there.
                            And if they don't, you can climb back out, and it'll be like nothing happened! 
                        [/p]
                        [p]
                            Gathering your courage, you let out a determined grunt.
                            Knights are fearless, and if you want to be one, you can't shy away from adventure.
                            You need to plunge head-first into danger!
                            Well, maybe not head-first, but you still need to go down the well,
                            if just to prove to yourself that you [b]do[/b] have what it takes to stop being a thief one day.
                        [/p]
                        [p]
                            You hop on top of the well's edge.
                            Facing away from the watering hole, you squat down.
                            With one hand, you tightly grip your ${global.kobold.inventory.createItemWithTooltip(new Rope())},
                            and with the other you take hold of the rim of the cobblestone wall.
                            You let out a deep, nerve-soothing sigh.
                            Here you go.
                        [/p]
                        [p]
                            You lower your body into the watering hole, distributing your weight evenly between the rope and the well itself.
                            To your pleasant surprise, everything seems to be holding!
                            You place your feet against the rock wall, and begin your descent.
                            The moment you let go of the edge of the well, however, something happens.
                        [/p]
                        [p]
                            Shifting your entire weight -- however light that may be -- onto the rope causes the rock you tied it to to [b]jerk[/b] forwards,
                            instantly dropping you down a couple of feet more.
                            [b]Oh no[/b]!
                            You yelp out in surprise, your heart beating a million times a minute, fearing for your very life.
                            You [b]knew[/b] you should have tested the chunk of cobblestone more thoroughly!
                            It [b]stopped moving[/b], for now, so you think it might be [i]trapped[/i] against the side of the well.
                            As long as nothing [i]breaks[/i], you should be fine.
                        [/p]
                        [p]
                            Adrenaline coursing through your system, you slowly and carefully begin to make your way back upwards.
                            Right as confidence resurges inside of your mind, as you believe that you're going to make it back up to the edge of the well,
                            that you're going to survive this frightening trial with only a near-heart attack to show for it,
                            a loud [b]crack[/b] resonates across the watering hole.
                            The lime-based cement is failing to hold together the well under the added stress of supporting your lightweight frame.
                        [/p]
                        [p]
                            You freeze in place, hoping that hanging still will prevent the circular wall from breaking further, but alas, it is too late.
                            After a few tense moments, the rock-like chunk you tied your rope to [b]breaks apart[/b] the ancient structure.
                            A considerable piece of cobblestone tears away from the rest,
                            and with nothing left to keep your rope-holding rock in place,
                            it gets pulled downwards by your weight, sealing your fate.
                            With nothing left to support you, you fall back-first into the well, your descent followed by several menacing rocks of varying sizes.
                        [/p]
                        [p]
                            You panic, [b]screaming[/b] in terror, thinking this might be the end of you.
                            You wantonly flail your limbs about, scratching and scraping the cobblestone wall surrounding you on all sides with your claws,
                            hoping to [i]latch on[/i] to something, or at least to somewhat slow your fall.
                            Your efforts are in vain.
                            Even if you [b]did[/b] manage to slow yourself down, you'd just get hit by the debris.
                            In a matter of seconds, your terrified shout is cut off,
                            interrupted by a loud series of [b]splashes[/b] as you and the falling rocks both plummet into the cold water at the bottom of the well.
                        [/p]
                        [p]
                            Frigid water rushes into your throat via your open mouth, filling your lungs long before you manage to close your maw.
                            Your descent is quickly halted by the cool liquid, doubly so as your back bumps against the ground at the very base of the well.
                            A large chunk of cobblestone heavily and painfully lands on your stomach, squishing you against the well's muddy base, pinning you in place.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/HamletDay2/HamletWell/img/drown-debris-${dick ? "m" : "f"}-full.png" target="_blank">
                                    <img class="sfw wideimage" src="scenes/HamletDay2/HamletWell/img/drown-debris-${dick ? "m" : "f"}.png" title="Illustration by ByondRAGE, colored by Nyumii">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            After falling, you find yourself [b]drowning[/b].
                            Panicking, you put every ounce of effort into pushing the heavy rock off of you, to regain your freedom.
                            Your consciousness is rapidly fading due to lack of air.
                            The thing just won't budge!
                            As you can feel life itself starting to slip away from you, you finally manage to dislodge the stone from your abdomen,
                            shoving it to the side a tiny little bit, enough for you to slip out from underneath it.
                            You swim upright.
                            Your feet touch the muddy ground below.
                            In a last-ditch attempt at survival, with your very last shred of power, you kick off from the submerged dirt,
                            launching you towards the surface of the water.
                        [/p]
                        [p]
                            You emerge from the ice-cold liquid, and immediately begin coughing up the watery contents of your respiratory system.
                            Your barely active brain [b]begs[/b] for air, but you can't breathe, not until you've emptied your [b]burning[/b] lungs.
                            After agonizingly [b]vomiting[/b] water from your trachea, you can finally take a much-needed gasp of oxygen,
                            the blackness that had begun to envelop the edges of your vision slightly clearing.
                        [/p]
                        [p]
                            It's not enough.
                            You need [b]more[/b].
                            You need to [b]breathe[/b], lest this damned well become your watery grave still.
                            You hack up more water, taking deep breaths in between each coughing fit.
                            Slowly, but steadily, you recover from the harrowing ordeal you just went through,
                            your breathing stabilizing as liquid drains from your lungs.
                        [/p]
                        [p]
                            It is only after a couple of minutes of recovering, both mentally and physically, that you can breathe normally once more.
                            The damp, stale air at the bottom of the well is of poor quality, but it sure as hell beats drowning.
                            Your gently kicking legs and your arms slowly rocking along the surface of the water keep your light frame afloat.
                            The water isn't [i]that[/i] deep.
                            You figure a tall human could stand in here, but you can't.
                        [/p]
                        [p]
                            The rock you tied your ${global.kobold.inventory.createItemWithTooltip(new Rope())} to came loose.
                            It slid against the well's edge, and after a little while, it ended up breaking the old construct.
                            An unfortunate series of events that nearly cost you your life.
                            You didn't die, though.
                            Not yet, anyway.
                        [/p]
                        [p]
                            Your ${global.kobold.inventory.createItemWithTooltip(new Rope())} is floating on the water's surface next to you.
                            It must have gotten untangled from the chunk you tied it to somewhere during or after the fall.
                            You pocket it once more.
                            While there's a bad memory attached to it now, it might still be useful later on.
                            You let out a sigh of relief, quickly followed by an annoyed groan as you tilt your head upwards,
                            and realize that you're several meters underground with no easy way up.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Dive down</button>`;

        var tooltip = global.parser.parse(`You're not worried about getting out just yet. Dive down to see what's at the bottom of the well.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Climb out</button>`;

        var tooltip = global.parser.parse(`Scale the walls to get out of this claustrophobic hole.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Call for help</button>`;

        var tooltip = global.parser.parse(`You don't think you can climb up by yourself. Maybe someone's around to help you out.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.loop();
        global.ui.imageBackground.hide();
        global.ui.applyTheme("hamlet_well");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

// END scenes\HamletDay2\HamletWell\fall\HamletWellRope.js

// scenes\HamletDay2\HamletWell\fall\HamletWellWalk.js

/**
 * Arrival section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellWalk(controller, firstNext, secondNext, thirdNext) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellArrivalChoice.WALK;
    var reqController = ctrl.arrival;
    var subController = ctrl.down;
    var changeController = ctrl.walk;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    //var firstNext = new HamletWellSwim(ctrl);
    //var secondNext = new HamletWellClimb(ctrl);
    //var thirdNext = new HamletWellHelp(ctrl);

    var firstChoice = HamletWellFallChoice.SWIM;
    var secondChoice = HamletWellFallChoice.CLIMB;
    var thirdChoice = HamletWellFallChoice.HELP;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.drowned = true;
        global.kobold.history.shared.well.fell = true;
        var toReturn = `[p]
                            After your depressing train of thoughts, you'd rather not hang around this place for much longer.
                            You'd hoped the well could draw your mind away from your elderly bird friend, but alas,
                            it only ended up making you think of him even more.
                            In fact, the entire hamlet reminds you of Seki.
                            You remember how you used to refer to this place as [i]Seki's village[/i].
                            Good times.
                        [/p]
                        [p]
                            You let out a loud sigh.
                            With a heavy heart, you decide to leave the hamlet.
                            The first step you take, however, your foot lands atop something uneven and unsteady!
                            You accidentally step on top of a broken off chunk of the cobblestone well,
                            and before you manage to recover your balance, the rock slides out from underneath you along the slightly muddy ground.
                        [/p]
                        [p]
                            Thrown off balance further, you recoil to the rear.
                            Your butt bumping against the well, you attempt to grab onto it,
                            to stabilize yourself against its rocky, waist-high exterior, but it's too late.
                            You fall backwards, over the edge of the well, tumbling head-first into pitch-black darkness before you even know what's going on.
                        [/p]
                        [p]
                            You panic, [b]screaming[/b] in terror, fearing for your very life.
                            You wantonly flail your limbs about, scratching and scraping the cobblestone wall surrounding you on all sides with your claws,
                            hoping to [i]latch on[/i] to something, or at least to somewhat slow your fall.
                            Your efforts are in vain.
                            In a matter of seconds, your terrified shout is cut off,
                            interrupted by a loud [b]splash[/b] as you plummet into the cold water at the bottom of the well.
                        [/p]
                        [p]
                            Frigid water rushes into your throat via your open mouth, filling your lungs long before you manage to close your maw.
                            Your descent is quickly halted by the cool liquid, doubly so as your head bumps against the ground at the very base of the well,
                            hard enough to significantly daze you.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/HamletDay2/HamletWell/img/drown-disorient-${ctrl.misc.dick() ? "m" : "f"}-full.png" target="_blank">
                                    <img class="sfw wideimage" src="scenes/HamletDay2/HamletWell/img/drown-disorient-${ctrl.misc.dick() ? "m" : "f"}.png" title="Illustration by ByondRAGE, colored by Nyumii">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            After falling, you find yourself [b]drowning[/b].
                            In your disoriented panic, you can't tell top from bottom.
                            You twist and turn in an attempt to get yourself upright, your consciousness,
                            already diminished from hitting your head, rapidly fading due to lack of air.
                            Just as you can feel life itself starting to slip away from you, one of your feet touches the muddy ground below.
                            In a last-ditch attempt at survival, with your very last ounce of power, you kick off from the submerged dirt,
                            launching you towards the surface of the water.
                        [/p]
                        [p]
                            You emerge from the ice-cold liquid, and immediately begin coughing up the watery contents of your respiratory system.
                            Your barely active brain [b]begs[/b] for air, but you can't breathe, not until you've emptied your [b]burning[/b] lungs.
                            After painfully [b]vomiting[/b] water from your trachea, you can finally take a much-needed gasp of oxygen,
                            the blackness that had begun to envelop the edges of your vision slightly clearing.
                        [/p]
                        [p]
                            It's not enough.
                            You need [b]more[/b].
                            You need to [b]breathe[/b], lest this damned well become your watery grave still.
                            You hack up more water, taking deep breaths in between each coughing fit.
                            Slowly, but steadily, you recover from the harrowing ordeal you just went through,
                            your breathing stabilizing as liquid drains from your lungs.
                        [/p]
                        [p]
                            It is only after a couple of minutes of recovering, both mentally and physically, that you can breathe normally once more.
                            The damp, stale air at the bottom of the well is of poor quality, but it sure as hell beats drowning.
                            Your gently kicking legs and your arms slowly rocking along the surface of the water keep your light frame afloat.
                            The water isn't [i]that[/i] deep.
                            You figure a tall human could stand in here, but you can't.
                        [/p]
                        [p]
                            You tripped, and you fell.
                            Dumb mistake.
                            It nearly cost you your life, but you didn't die.
                            Not yet, anyway.
                            You let out a sigh of relief, quickly followed by an annoyed groan as you tilt your head upwards,
                            and realize that you're several meters underground with no easy way up.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Dive down</button>`;

        var tooltip = global.parser.parse(`You're not worried about getting out just yet. Dive down to see what's at the bottom of the well.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Climb out</button>`;

        var tooltip = global.parser.parse(`Scale the walls to get out of this claustrophobic hole.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Call for help</button>`;

        var tooltip = global.parser.parse(`You don't think you can climb up by yourself. Maybe someone's around to help you out.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.loop();
        global.ui.imageBackground.hide();
        global.ui.applyTheme("hamlet_well");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

// END scenes\HamletDay2\HamletWell\fall\HamletWellWalk.js

// scenes\HamletDay2\HamletWell\fall\water\HamletWellWaterSucceed.js

/**
 * Arrival section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellWaterSucceed(controller, firstNext, secondNext, thirdNext) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.down;
    var changeController = ctrl.waterSucceed;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    //var firstNext = new HamletWellSwim(ctrl);
    //var secondNext = new HamletWellClimb(ctrl);
    //var thirdNext = new HamletWellHelp(ctrl);

    var firstChoice = HamletWellFallChoice.SWIM;
    var secondChoice = HamletWellFallChoice.CLIMB;
    var thirdChoice = HamletWellFallChoice.HELP;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.fell = true;
        global.kobold.history.shared.well.drowned = true;
        global.kobold.history.shared.well.drankWater = true;
        var toReturn = `[p]
                            After your depressing train of thoughts, you need something, [b]anything[/b] to take your mind off of Seki.
                            Desperately looking for a way to forget about your friend for a while, you turn to the rope disappearing into the well.
                            Drawing water might be a good way to pass the time, and honestly, you could use a refreshing drink.
                        [/p]
                        [p]
                            You grab the rope, and give it a firm [b]yank[/b].
                            It doesn't budge.
                            The bucket must be stuck on something, or maybe it's just too heavy for you to lift.
                            Determined to get some water, you dig your claws into the ground and lean back,
                            tugging at the rope with not just your arms, but your entire weight, however little that may be.
                        [/p]
                        [p]
                            Your new tactic proves to be effective!
                            After a bit of pulling, the bucket finally shoots free, causing you to fall on your behind.
                            Luckily, you're strong enough to keep the filled bucket lifted through sheer arm-strength alone.
                            You scamper to your feet without letting go of the rope, and begin to painstakingly hoist the container upwards, one tug at a time.
                        [/p]
                        ${global.ui.condSucc(`Strength greater than 5`)}
                        [p]
                            Success!
                            After a long and arduous game of tug of war against gravity itself, a bucket of fresh well-water appears near the edge of the well.
                            You inch both your hands towards it, along the length of the rope,
                            until you can take hold of the bucket with both of your hands, to lift it out of the well proper.
                            You did it!
                            It was hard, but you've managed to draw water from the well.
                            Your reward is a bucket of fresh, cool water.
                            Now for the fun part: to drink it!
                        [/p]
                        [p]
                            You bring the rim of the bucket to your lips.
                            Opening your muzzle, you tilt both your head and the container upwards, letting the refreshing liquid [b]gush[/b] into your open maw,
                            greedily gulping it down like the exhausted lizard you are.
                            After emptying about half of the bucket, a lot of which was spilled over your chest instead of disappearing down your throat,
                            you set the considerably lighter wooden container down on the ground.
                        [/p]
                        [p]
                            Suddenly, your sight goes kind of blurry.
                            You feel woozy, and your body rocks from side to side as though you're drunk!
                            You stumble about, trying to find the edge of the well, so you can take a seat until the feeling passes.
                            While you do manage to find the rim of the watering hole, a fresh wave of drowsiness washes over you.
                            Your balance disturbed, you fall backwards, over the edge of the well,
                            tumbling head-first into pitch-black darkness before you even know what's going on.
                        [/p]
                        [p]
                            Adrenaline floods your system, seemingly counteracting the sudden onset of tiredness.
                            Panic quickly settles in.
                            You begin to [b]scream[/b] in terror, fearing for your very life.
                            You wantonly flail your limbs about, scratching and scraping the cobblestone wall surrounding you on all sides with your claws,
                            hoping to [i]latch on[/i] to something, or at least to somewhat slow your fall.
                            Your efforts are in vain.
                            In a matter of seconds, your terrified shout is cut off,
                            interrupted by a loud [b]splash[/b] as you plummet into the cold water at the bottom of the well.
                        [/p]
                        [p]
                            Frigid water rushes into your throat via your open mouth, filling your lungs long before you manage to close your maw.
                            Your descent is quickly halted by the cool liquid, doubly so as your head bumps against the ground at the very base of the well,
                            hard enough to significantly daze you.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/HamletDay2/HamletWell/img/drown-disorient-${ctrl.misc.dick() ? "m" : "f"}-full.png" target="_blank">
                                    <img class="sfw wideimage" src="scenes/HamletDay2/HamletWell/img/drown-disorient-${ctrl.misc.dick() ? "m" : "f"}.png" title="Illustration by ByondRAGE, colored by Nyumii">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            After falling, you find yourself [b]drowning[/b].
                            In your disoriented panic, you can't tell top from bottom.
                            You twist and turn in an attempt to get yourself upright, your consciousness,
                            already diminished from hitting your head, rapidly fading due to lack of air.
                            Just as you can feel life itself starting to slip away from you, one of your feet touches the muddy ground below.
                            In a last-ditch attempt at survival, with your very last ounce of power, you kick off from the submerged dirt,
                            launching you towards the surface of the water.
                        [/p]
                        [p]
                            You emerge from the ice-cold liquid, and immediately begin coughing up the watery contents of your respiratory system.
                            Your barely active brain [b]begs[/b] for air, but you can't breathe, not until you've emptied your [b]burning[/b] lungs.
                            After painfully [b]vomiting[/b] water from both your esophagus and your trachea, you can finally take a much-needed gasp of oxygen,
                            the blackness that had begun to envelop the edges of your vision slightly clearing.
                        [/p]
                        [p]
                            It's not enough.
                            You need [b]more[/b].
                            You need to [b]breathe[/b], lest this damned well become your watery grave still.
                            You hack up more water, taking deep breaths in between each coughing fit.
                            Slowly, but steadily, you recover from the harrowing ordeal you just went through,
                            your breathing stabilizing as liquid drains from your lungs.
                        [/p]
                        [p]
                            It is only after a couple of minutes of recovering, both mentally and physically, that you can breathe normally once more.
                            The damp, stale air at the bottom of the well is of poor quality, but it sure as hell beats drowning.
                            Your gently kicking legs and your arms slowly rocking along the surface of the water keep your light frame afloat.
                            The water isn't [i]that[/i] deep.
                            You figure a tall human could stand in here, but you can't.
                        [/p]
                        [p]
                            One moment you were happily enjoying the spoils of your labor, the other you felt so sick you could hardly [b]stand[/b].
                            Was it the water?
                            You're swimming in the stuff right now, and it's not affecting you at all.
                            But then again, you've already thrown up the liquid you drank.
                            Maybe it only does it when it's ingested.
                            Or perhaps it just a momentary lapse of strength.
                            Regardless, your clumsiness nearly cost you your life.
                            You didn't die, though.
                            Not yet, anyway.
                            You let out a sigh of relief, quickly followed by an annoyed groan as you tilt your head upwards,
                            and realize that you're several meters underground with no easy way up.
                        [/p]`;
        
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Dive down</button>`;

        var tooltip = global.parser.parse(`You're not worried about getting out just yet. Dive down to see what's at the bottom of the well.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Climb out</button>`;

        var tooltip = global.parser.parse(`Scale the walls to get out of this claustrophobic hole.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Call for help</button>`;

        var tooltip = global.parser.parse(`You don't think you can climb up by yourself. Maybe someone's around to help you out.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

// END scenes\HamletDay2\HamletWell\fall\water\HamletWellWaterSucceed.js

// scenes\HamletDay2\HamletWell\fall\water\HamletWellWaterFail.js

/**
 * Arrival section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellWaterFail(controller, firstNext, secondNext, thirdNext) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.down;
    var changeController = ctrl.waterFail;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    //var firstNext = new HamletWellSwim(ctrl);
    //var secondNext = new HamletWellClimb(ctrl);
    //var thirdNext = new HamletWellHelp(ctrl);

    var firstChoice = HamletWellFallChoice.SWIM;
    var secondChoice = HamletWellFallChoice.CLIMB;
    var thirdChoice = HamletWellFallChoice.HELP;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.drowned = true;
        global.kobold.history.shared.well.fell = true;
        var toReturn = `[p]
                            After your depressing train of thoughts, you need something, [b]anything[/b] to take your mind off of Seki.
                            Desperately looking for a way to forget about your friend for a while, you turn to the rope disappearing into the well.
                            Drawing water might be a good way to pass the time, and honestly, you could use a refreshing drink.
                        [/p]
                        [p]
                            You grab the rope, and give it a firm [b]yank[/b].
                            It doesn't budge.
                            The bucket must be stuck on something, or maybe it's just too heavy for you to lift.
                            Determined to get some water, you dig your claws into the ground and lean back,
                            tugging at the rope with not just your arms, but your entire weight, however little that may be.
                        [/p]
                        [p]
                            Your new tactic proves to be effective!
                            After a bit of pulling, the bucket finally shoots free, causing you to fall on your behind.
                            Then, something unexpected happens.
                            While your weight and all of the muscle power you could muster were sufficient to dislodge the water-container,
                            you aren't powerful enough to keep it lifted!
                        [/p]
                        ${global.ui.condFail(`Strength below 6`)}
                        [p]
                            Your hands still clinging to the rope, you are pulled forwards, towards the well.
                            Before you even realize what's going on, you find yourself yanked over the waist-height well-wall,
                            pulled into the depths of the watering hole.
                            You let go of the rope, but it's too late!
                            You're already falling, tumbling head-first into pitch-black darkness.
                        [/p]
                        [p]
                            You panic, [b]screaming[/b] in terror, fearing for your very life.
                            You wantonly flail your limbs about, scratching and scraping the cobblestone wall surrounding you on all sides with your claws,
                            hoping to [i]latch on[/i] to something, or at least to somewhat slow your fall.
                            Your efforts are in vain.
                            In a matter of seconds, your terrified shout is cut off,
                            interrupted by a loud [b]splash[/b] as you plummet into the cold water at the bottom of the well.
                        [/p]
                        [p]
                            Frigid water rushes into your throat via your open mouth, filling your lungs long before you manage to close your maw.
                            Your descent is quickly halted by the cool liquid, doubly so as your head bumps against the ground at the very base of the well,
                            hard enough to significantly daze you.
                        [/p]
                        [p]
                            [c]
                                <a href="scenes/HamletDay2/HamletWell/img/drown-disorient-${ctrl.misc.dick() ? "m" : "f"}-full.png" target="_blank">
                                    <img class="sfw wideimage" src="scenes/HamletDay2/HamletWell/img/drown-disorient-${ctrl.misc.dick() ? "m" : "f"}.png" title="Illustration by ByondRAGE, colored by Nyumii">
                                </a>
                            [/c]
                        [/p]
                        [p]
                            After falling, you find yourself [b]drowning[/b].
                            In your disoriented panic, you can't tell top from bottom.
                            You twist and turn in an attempt to get yourself upright, your consciousness,
                            already diminished from hitting your head, rapidly fading due to lack of air.
                            Just as you can feel life itself starting to slip away from you, one of your feet touches the muddy ground below.
                            In a last-ditch attempt at survival, with your very last ounce of power, you kick off from the submerged dirt,
                            launching you towards the surface of the water.
                        [/p]
                        [p]
                            You emerge from the ice-cold liquid, and immediately begin coughing up the watery contents of your respiratory system.
                            Your barely active brain [b]begs[/b] for air, but you can't breathe, not until you've emptied your [b]burning[/b] lungs.
                            After painfully [b]vomiting[/b] water from your trachea, you can finally take a much-needed gasp of oxygen,
                            the blackness that had begun to envelop the edges of your vision slightly clearing.
                        [/p]
                        [p]
                            It's not enough.
                            You need [b]more[/b].
                            You need to [b]breathe[/b], lest this damned well become your watery grave still.
                            You hack up more water, taking deep breaths in between each coughing fit.
                            Slowly, but steadily, you recover from the harrowing ordeal you just went through,
                            your breathing stabilizing as liquid drains from your lungs.
                        [/p]
                        [p]
                            It is only after a couple of minutes of recovering, both mentally and physically, that you can breathe normally once more.
                            The damp, stale air at the bottom of the well is of poor quality, but it sure as hell beats drowning.
                            Your gently kicking legs and your arms slowly rocking along the surface of the water keep your light frame afloat.
                            The water isn't [i]that[/i] deep.
                            You figure a tall human could stand in here, but you can't.
                        [/p]
                        [p]
                            You'd managed to lift the water bucket out of the water, only for it to fall back down, pulling you along with it.
                            You should have known you weren't strong enough to operate a man-made well.
                            Your mistake nearly cost you your life, but you didn't die.
                            Not yet, anyway.
                            You let out a sigh of relief, quickly followed by an annoyed groan as you tilt your head upwards,
                            and realize that you're several meters underground with no easy way up.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Dive down</button>`;

        var tooltip = global.parser.parse(`You're not worried about getting out just yet. Dive down to see what's at the bottom of the well.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Climb out</button>`;

        var tooltip = global.parser.parse(`Scale the walls to get out of this claustrophobic hole.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Call for help</button>`;

        var tooltip = global.parser.parse(`You don't think you can climb up by yourself. Maybe someone's around to help you out.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

// END scenes\HamletDay2\HamletWell\fall\water\HamletWellWaterFail.js

// scenes\HamletDay2\HamletWell\fall\HamletWellWater.js

/**
 * Arrival section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellWater(controller, firstNext, secondNext, thirdNext) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = HamletWellArrivalChoice.WATER;
    var reqController = ctrl.arrival;

    var toFail = new HamletWellWaterFail(ctrl, firstNext, secondNext, thirdNext);
    var toSuccess = new HamletWellWaterSucceed(ctrl, firstNext, secondNext, thirdNext);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.strength() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.loop();
        global.ui.imageBackground.hide();
        global.ui.applyTheme("hamlet_well");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\HamletWell\fall\HamletWellWater.js

// scenes\HamletDay2\HamletWell\arrival\HamletWellArrival.js

/**
 * Arrival section of Hamlet, the Well.
 * @param {HamletWellController} controller The Hamlet, Well controller.
 */
function HamletWellArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.arrival;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";
    var fourthId = arguments.callee.name + "ButtonFour";

    var childFirstNext = new HamletWellSwim(ctrl);
    var childSecondNext = new HamletWellClimbUp(ctrl);
    var childThirdNext = new HamletWellHelp(ctrl);

    var firstNext = new HamletWellWalk(ctrl, childFirstNext, childSecondNext, childThirdNext);
    var secondNext = new HamletWellWater(ctrl, childFirstNext, childSecondNext, childThirdNext);
    var thirdNext = new HamletWellClimb(ctrl, childFirstNext, childSecondNext, childThirdNext);
    var fourthNext = new HamletWellRope(ctrl, childFirstNext, childSecondNext, childThirdNext);

    var firstChoice = HamletWellArrivalChoice.WALK;
    var secondChoice = HamletWellArrivalChoice.WATER;
    var thirdChoice = HamletWellArrivalChoice.CLIMB;
    var fourthChoice = HamletWellArrivalChoice.ROPE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            In an attempt to take your mind off of Seki, you turn your attention to the well.
                            It's a small, circular structure that reaches slightly above your waist.
                            The cobblestone it's made of looks pretty shabby, like it's fit to collapse any minute now.
                            Then again, judging by the ancient-looking moss clinging to the damp dirt on the northern end,
                            you figure the well's been around for a [b]very[/b] long time.
                            You probably won't look that good when you reach its age.
                        [/p]
                        [p]
                            If there's one thing humans are good at, it's building things that last.
                            From gigantic stone walls surrounding their enormous cities, to intricately engraved cathedral-windows.
                            Compared to the slums and sheds kobolds live in, even the simplest of human homes could pass for a [b]palace[/b].
                            Fit for [b]royalty[/b], if your kind actually [i]had[/i] rulership beyond petty crime lords.
                            Maybe that's why your species never learned to construct things beyond small cabins.
                            There isn't any need for big homes when everyone is [i]equally unimportant[/i].
                            Or maybe it's the other way around?
                        [/p]
                        [p]
                            You, personally, have never really built anything.
                            You kind of [i]stumbled upon[/i] the shack that you live in,
                            and you don't have need for anything but a place to rest your weary head at night.
                            You imagine building things isn't all that easy, especially not for someone your size.
                            Even if you somehow made it through the whole [i]planning[/i] phase,
                            the sheer amount of physical work needed to build something is far too much for any one kobold to handle.
                            And good luck getting an entire team to work together for more than a couple of days.
                        [/p]
                        [p]
                            What were you thinking about again?
                            Before you started [i]daydreaming[/i]?
                            Oh right, the well.
                        [/p]
                        [p]
                            You leisurely pace a few laps around it, inspecting the old structure.
                            A rope lies coiled next to it, one end disappearing into the watering hole's pitch-blackness.
                            The irregular stones the well is made of vary heavily in size.
                            Some are as big as your fist, others are but mere pebbles.
                            They're tied together by a flimsy lime-mixture.
                            Heck, you could probably tear the entire thing apart using nothing but your claws.
                        [/p]
                        [p]
                            You somehow doubt the locals would appreciate you [i]demolishing[/i] their convenient supply of water, though.
                            Then again, the river isn't [i]that[/i] far off.
                            Those lazy farmers getting fat off their own produce could definitely use the extra exercise.
                            Maybe you'll come back here next week for some mindless, stress-relieving vandalism.
                            If what Seki says is true, you'll no longer have to worry about [i]inconveniencing[/i] him by then,
                            and you'll undoubtedly have plenty of frustration to work out on inanimate objects.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled} style="width: 170px;">Walk away</button>`;

        var tooltip = global.parser.parse(`The well doesn't interest you anymore. Now it just makes you sad.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 170px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled} style="width: 170px;">Draw water</button>`;

        var tooltip = global.parser.parse(`Attempt to fetch some water from the bottom of the pit.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 205px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 170px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled} style="width: 170px;">Climb down</button>`;

        var tooltip = global.parser.parse(`Scale down the rocky well wall.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 210px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 170px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }
    /**
     * Builds and returns the choice button.
     */
    function buildFourthChoice() {
        var check = ctrl.misc.hasMisc(new Rope());
        var disabled = subController.has() || !check ? "disabled" : "";
        var isSelected = subController.get() === fourthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fourthId}" ${disabled} style="width: 170px;">Use rope</button>`;

        var tooltip = global.parser.parse(`Use your rope to climb down into the well.`);
        if (!check) {
            tooltip = global.parser.parse(`You don't have a rope to climb down with.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 170px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += `<div class="choicerow" style="margin-bottom: 10px;">`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        choice += `<div class="choicerow">`;
        choice += buildThirdChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFourthChoice();
        choice += `</div>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId), $("#" + fourthId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFourthListener() {
        $("#" + fourthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fourthChoice);
            loadNextPart(fourthNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.imageBackground.show("scenes/HamletDay2/HamletWell/img/wellbackground.png");
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        if (changeController) {
            part += changeController.change();
        }
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        part += fourthNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();
        fourthNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
        registerFourthListener();
    };
}

var HamletWellArrivalChoice = {
    WALK: 1,
    WATER: 2,
    CLIMB: 3,
    ROPE: 4
};

// END scenes\HamletDay2\HamletWell\arrival\HamletWellArrival.js

// scenes\HamletDay2\HamletWell\HamletWellController.js

/**
 * Controller for the Well scene.
 * @param {string} sceneId The id of the scene.
 */
function HamletWellController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.nextSceneReady = true;
    this.nextSceneName = "UndergroundCave";
    this.nextFolderName = "HamletDay2";
    this.varanarSceneName = "NightTwo";
    this.varanarFolderName = "Varanar";

    this.misc = new MiscController(id, "misc");

    this.arrival = new ChoiceController(id, "arrival");

    this.rope = new ChangeController(id, "rope", function () {
        global.kobold.inventory.equipment.cleanAll();
        global.kobold.status.physical.decrease(15, "Nearly drowning");
        global.kobold.status.mental.decrease(8, "Stuck in the well");
        global.kobold.status.stamina.decrease(17, "Almost drowning");
        if (that.misc.exertion()) {
            global.kobold.status.exertion.decrease(5, "Cold water");
        }
    }, function () {
        var changes = [
            new Change("Fell into water", "Health", -3),
            new Change("Squished by debris", "Health", -5),
            new Change("Nearly drowned", "Health", -7),
            new Change("Fell in well", "Sanity", -6),
            new Change("Nearly drowned", "Sanity", -5),
            new Change("Survived", "Sanity", +3),
            new Change("Climbing", "Stamina", -3),
            new Change("Screaming", "Stamina", -2),
            new Change("Pushing rock off", "Stamina", -4),
            new Change("Drowning", "Stamina", -8)
        ];
        if (that.misc.exertion()) {
            changes.push(new Change("Cold water", "Sexual exhaustion", -5, true));
        }
        changes.concat(
                [new Change("Used rope", global.kobold.inventory.createItemWithTooltip(new Rope(), true), -1),
                    new Change("Recovered rope", global.kobold.inventory.createItemWithTooltip(new Rope(), true), +1)]);

        return global.ui.change.buildNotification(changes);
    });

    this.walk = new ChangeController(id, "walk", function () {
        global.kobold.inventory.equipment.cleanAll();
        global.kobold.status.physical.decrease(18, "Nearly drowning");
        global.kobold.status.mental.decrease(8, "Stuck in the well");
        global.kobold.status.stamina.decrease(10, "Almost drowning");
        if (that.misc.exertion()) {
            global.kobold.status.exertion.decrease(5, "Cold water");
        }
    }, function () {
        var changes = [
            new Change("Fell into water", "Health", -4),
            new Change("Bonked head", "Health", -7),
            new Change("Nearly drowned", "Health", -7),
            new Change("Fell in well", "Sanity", -6),
            new Change("Nearly drowned", "Sanity", -5),
            new Change("Survived", "Sanity", +3),
            new Change("Screaming", "Stamina", -2),
            new Change("Drowning", "Stamina", -8)
        ];
        if (that.misc.exertion()) {
            changes.push(new Change("Cold water", "Sexual exhaustion", -5, true));
        }
        return global.ui.change.buildNotification(changes);
    });

    this.waterFail = new ChangeController(id, "waterFail", function () {
        global.kobold.inventory.equipment.cleanAll();
        global.kobold.status.physical.decrease(18, "Nearly drowning");
        global.kobold.status.mental.decrease(8, "Stuck in the well");
        global.kobold.status.stamina.decrease(12, "Almost drowning");
        if (that.misc.exertion()) {
            global.kobold.status.exertion.decrease(5, "Cold water");
        }
    }, function () {
        var changes = [
            new Change("Fell into water", "Health", -4),
            new Change("Bonked head", "Health", -7),
            new Change("Nearly drowned", "Health", -7),
            new Change("Fell in well", "Sanity", -6),
            new Change("Nearly drowned", "Sanity", -5),
            new Change("Survived", "Sanity", +3),
            new Change("Tugging", "Stamina", -2),
            new Change("Screaming", "Stamina", -2),
            new Change("Drowning", "Stamina", -8)
        ];
        if (that.misc.exertion()) {
            changes.push(new Change("Cold water", "Sexual exhaustion", -5, true));
        }
        return global.ui.change.buildNotification(changes);
    });

    this.waterSucceed = new ChangeController(id, "waterSucceed", function () {
        global.kobold.inventory.equipment.cleanAll();
        global.kobold.status.physical.decrease(18, "Nearly drowning");
        global.kobold.status.mental.decrease(3, "Stuck in the well");
        global.kobold.status.stamina.decrease(12, "Almost drowning");
        if (that.misc.exertion()) {
            global.kobold.status.exertion.decrease(5, "Cold water");
        }
    }, function () {
        var changes = [
            new Change("Fell into water", "Health", -4),
            new Change("Bonked head", "Health", -7),
            new Change("Nearly drowned", "Health", -7),
            new Change("Drew water", "Sanity", +5),
            new Change("Fell in well", "Sanity", -6),
            new Change("Nearly drowned", "Sanity", -5),
            new Change("Survived", "Sanity", +3),
            new Change("Tugging", "Stamina", -2),
            new Change("Screaming", "Stamina", -2),
            new Change("Drowning", "Stamina", -8)
        ];
        if (that.misc.exertion()) {
            changes.push(new Change("Cold water", "Sexual exhaustion", -5, true));
        }
        return global.ui.change.buildNotification(changes);
    });

    this.climbFail = new ChangeController(id, "climbFail", function () {
        global.kobold.inventory.equipment.cleanAll();
        global.kobold.status.physical.decrease(7, "Falling into the well");
        global.kobold.status.mental.decrease(3, "Stuck in the well");
        global.kobold.status.stamina.decrease(6, "Falling into the well");
        if (that.misc.exertion()) {
            global.kobold.status.exertion.decrease(5, "Cold water");
        }
    }, function () {
        var changes = [
            new Change("Fell into water", "Health", -4),
            new Change("Water in lungs", "Health", -3),
            new Change("Fell in well", "Sanity", -6),
            new Change("Survived", "Sanity", +3),
            new Change("Climbing", "Stamina", -2),
            new Change("Screaming", "Stamina", -2),
            new Change("Coughing", "Stamina", -2)
        ];
        if (that.misc.exertion()) {
            changes.push(new Change("Cold water", "Sexual exhaustion", -5, true));
        }
        return global.ui.change.buildNotification(changes);
    });

    this.climbSucceed = new ChangeController(id, "climbSucceed", function () {
        global.kobold.inventory.equipment.cleanAll();
        global.kobold.status.mental.increase(7, "Climbing into the well");
        global.kobold.status.stamina.decrease(12, "Climbing into the well");
        if (that.misc.exertion()) {
            global.kobold.status.exertion.decrease(5, "Cold water");
        }
    }, function () {
        var changes = [
            new Change("Climbed down", "Sanity", +10),
            new Change("Stuck", "Sanity", -3),
            new Change("Climbing", "Stamina", -12)
        ];
        if (that.misc.exertion()) {
            changes.push(new Change("Cold water", "Sexual exhaustion", -5, true));
        }
        return global.ui.change.buildNotification(changes);
    });

    this.down = new ChoiceController(id, "down");

    this.pants = new ChangeController(id, "pants", function () {
        that.misc.pants();
        global.kobold.status.mental.increase(5, "Finding gold in the well");
        global.kobold.status.stamina.decrease(6, "Diving");
        global.kobold.inventory.coins = global.kobold.inventory.coins + 37;
        global.kobold.inventory.equipment.strip(SlotEnum.PANTS);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Lost pants", "Sanity", -2),
            new Change("Found gold", "Sanity", +7),
            new Change("Diving", "Stamina", -6),
            new Change("Found gold", "Coins", +37),
            new Change("Ripped pants", global.kobold.inventory.createItemWithTooltip(that.misc.pants(), true), -1)
        ]);
    });

    this.scrape = new ChangeController(id, "scrape", function () {
        global.kobold.status.physical.decrease(15, "Scraped leg");
        global.kobold.status.mental.decrease(10, "Pain");
        global.kobold.status.stamina.decrease(3, "Diving");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Scraped", "Health", -15),
            new Change("Pain", "Sanity", -10),
            new Change("Diving", "Stamina", -3)
        ]);
    });

    this.climbUp = new ChangeController(id, "climbUp", function () {
        global.kobold.status.mental.decrease(5, "Almost fell down");
        global.kobold.status.stamina.decrease(15, "Climbing up");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Almost fell", "Sanity", -5),
            new Change("Climbing up", "Stamina", -15)
        ]);
    });

    this.climbUpRelent = new ChangeController(id, "climbUpRelent", function () {
        global.kobold.status.mental.increase(4, "Still stuck");
        global.kobold.status.stamina.decrease(8, "Climbing down");
        global.kobold.stats.increaseStat(StatsEnum.AGILITY, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stopped before getting hurt", "Sanity", +6),
            new Change("Still stuck", "Sanity", -2),
            new Change("Climbing down", "Stamina", -8),
            new Change("Good workout", "Agility", +1)
        ]);
    });

    this.climbUpFail = new ChangeController(id, "climbUpFail", function () {
        global.kobold.status.physical.decrease(25, "Hit head against wall");
        global.kobold.status.mental.decrease(16, "Seriously hurt");
        global.kobold.stats.decreaseStat(StatsEnum.INTELLIGENCE, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Hit head", "Health", -15),
            new Change("Knocked unconscious", "Health", -10),
            new Change("Fell down", "Sanity", -6),
            new Change("Seriously hurt", "Sanity", -10),
            new Change("Hit head", "Intelligence", -1)
        ]);
    });

    this.climbUpSucceed = new ChangeController(id, "climbUpSucceed", function () {
        global.kobold.status.physical.decrease(5, "Overexertion");
        global.kobold.status.mental.increase(20, "Escaped the well");
        global.kobold.status.stamina.decrease(20, "Intensive climbing");
        global.kobold.stats.increaseStat(StatsEnum.ENDURANCE, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Overexertion", "Health", -5),
            new Change("Impressed self", "Sanity", +10),
            new Change("Escaped the well", "Sanity", +10),
            new Change("Intensive climbing", "Stamina", -20),
            new Change("Pushed your limits", "Endurance", +1)
        ]);
    });

    this.help = new ChangeController(id, "help", function () { // Do NOT change key, used in shared history.
        global.kobold.history.shared.well.metBoys = true;
        global.kobold.status.mental.increase(2, "Help?");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Potential help", "Sanity", +2)
        ]);
    });

    this.helpScold = new ChangeController(id, "helpScold", function () {
        global.kobold.status.mental.decrease(19, "Rude children");
        global.kobold.stats.decreaseStat(StatsEnum.CHARISMA, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Peed on (kind of)", "Sanity", -10),
            new Change("Laughed at", "Sanity", -2),
            new Change("Frustration", "Sanity", -7),
            new Change("Botched negotiations", "Charisma", -1)
        ]);
    });

    this.helpBegFail = new ChangeController(id, "helpBegFail", function () {
        global.kobold.status.mental.decrease(13, "Mean kids");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Insulted", "Sanity", -1),
            new Change("Laughed at", "Sanity", -2),
            new Change("Left to die", "Sanity", -10)
        ]);
    });

    this.helpBegSucceed = new ChangeController(id, "helpBegSucceed", function () {
        global.kobold.status.mental.increase(20, "Shown kindness by a stranger");
        global.kobold.status.stamina.decrease(10, "Climbed out of the well");
        global.kobold.stats.increaseStat(StatsEnum.LUCK, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Helped", "Sanity", +20),
            new Change("Climbing", "Stamina", +10),
            new Change("Saved by a stranger", "Luck", +1)
        ]);
    });

    this.preContinue = new ChoiceController(id, "preContinue");

    this.bgm = new AudioController(id, "res/audio/effects/waterdrip.mp3", 70);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

}

var HamletWellFallChoice = {
    SWIM: 1,
    CLIMB: 2,
    HELP: 3
};

// END scenes\HamletDay2\HamletWell\HamletWellController.js

// scenes\HamletDay2\HamletWell\HamletWell.js

/**
 * Well in the hamlet.
 */
function HamletWell() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "HamletWell";
    var sceneName = "Hamlet, the Well";
    var sceneSubfolder = "HamletDay2";
    var ctrl = new HamletWellController(sceneId);
    var arrival = new HamletWellArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]hamlet[/t]`);
        tab += global.parser.parse(`[st]The well[/st]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new HamletWell();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\HamletDay2\HamletWell\HamletWell.js



// END SCENE HamletWell



// SCENE UndergroundCave

// scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sub\UndergroundCaveReflectionGetFuckedContinue.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionGetFuckedContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.reflectSubContinue;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With the basilisk slowly drifting to sleep in a puddle of your combined fluids,
                            you decide it's time to find a way out of the cave.
                            You may return here, one day, but for now you must leave.
                            You begin to follow the river-branch, hoping it'll lead you to some kind of exit.
                        [/p]
                        [p]
                            You leisurely stroll alongside the water, until daylight comes into sight.
                            As you expected: the water [b]was[/b] a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that,
                            you managed to have the time of your life getting fucked by a basilisk.
                            You doubt a lot of people can claim to have had sex with a mythological being, and much less still lived to tell the tale.
                        [/p]
                        [p]
                            You've gained a few personal memories that you'll undoubtedly find yourself masturbating to later,
                            along with some awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sub\UndergroundCaveReflectionGetFuckedContinue.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sub\UndergroundCaveReflectionGetFuckedInside.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionGetFuckedInside(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveReflectionGetFuckedChoice.INSIDE;
    var reqController = ctrl.reflectSub;
    var setChoice = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.reflectInside;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveReflectionGetFuckedContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskRodeCameInside = true;
        var dick = ctrl.misc.dick();
        var shirt = ctrl.misc.shirt();
        var pants = ctrl.misc.pants();
        var undies = ctrl.misc.underwear();

        var toReturn = ``;

        if (dick) {
            var virgin = ctrl.misc.analVirgin();
            toReturn += `
                        [p]
                            You [b]need[/b] this to continue, to [b]want[/b] the basilisk to [b]fuck you silly[/b],
                            to [b]pound you[/b] until you can barely even [b]walk[/b] anymore!
                            Even though you've already reached your end, this is the best feeling you've [b]ever[/b] felt in your entire life.
                            You want it to last [b]forever[/b]!
                            And while you know it'll eventually [i]stop[/i],
                            you intend to ride out every single second of it as though it is your last!
                        [/p]
                        [p]
                            A split second later you can feel it starting to happen.
                            The basilisk's member [b]twitches[/b] and [b]convulses[/b] inside of your ass.
                            Its loud hissing turns into a low-pitched growl, and the first of many ropes of hot,
                            thick, sticky cum squirts from its beastly length,
                            your senses soothed with the feeling of being [b]filled up[/b],
                            your battered mind [b]overloading[/b] with the sheer [b]taboo[/b] of letting a feral finish inside of your rear.
                        [/p]
                        [p]
                            Your brain completely short-circuiting, you begin to wildly [b]ride[/b] the basilisk's shaft once more,
                            intent on squeezing every single [b]drop[/b] of spunk and pleasure that you possibly can from its exotic prick.
                            Your tongue lolls out of your mouth in the same way the beast's did,
                            its [b]mind-breaking[/b] climax seemingly lasting for [b]minutes[/b],
                            minutes of absolute [b]ecstacy[/b] for both you and it.
                        [/p]
                        [p]
                            You lose all track of time, direction, and being.
                            You black out, your mind and body utterly overwhelmed by the sheer amount of pleasure you're feeling.
                            When you come to your senses, you find yourself lying on top of the monster's soft scales,
                            your belly slightly bulging from the sheer amount of cum inside of you, despite a lot of it already having flooded out.
                            Even more heady spunk is squirting on top of your body, covering you from head to toe,
                            ${shirt ? `completely soaking your ${global.kobold.inventory.createItemWithTooltip(shirt)},` : ``}
                            the basilisk's unmatched orgasm only [b]now[/b] starting to fade.
                        [/p]
                        [p]
                            Completely exhausted, you find yourself resting for a while in the warm, comforting care of your new feral [b]lover[/b],
                            idly taking in massive amounts of animalistic male pheromones with every breath.
                            You can feel the basilisk's strong chest comfortingly heaving up and down below you.
                            It's contently [i]purring[/i], its needs thoroughly sated.
                            What a [b]good boy[/b].
                        [/p]
                        [p]
                            As much as you'd like to make this cave your new home, you're afraid you simply can't.
                            Not yet, anyway.
                            As the cum on your scales starts to cool,
                            and the rational part of your brain slowly starts to kick in,
                            you realize that you have places to be and things to do.
                            With a loud sigh, you reluctantly roll off of the big lizard, who lets out a distinctly sad whimper.
                        [/p]
                        [p]
                            Trying to rise to your feet, you discover just how sore the intense fucking has left your entire body.
                            You're certain it'll fade in due time, though.
                            Desspite having a rough time reaching the cave, you can say with a smile on your face that it was all worth it.
                            You thank the basilisk, though you doubt it understands, and give a small kiss on its snout.
                            Then, ${pants || undies ? `you redress, and` : `you`} prepare to leave.
                        [/p]`;
            if (virgin) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        } else {
            var virgin = ctrl.misc.cuntVirgin();
            toReturn += `
                        [p]
                            You don't care anymore.
                            You don't care if you get pregnant, if the beast [b]breeds[/b] you,
                            if it mates you with and plants its seed deep inside of your womb.
                            You [b]need[/b] this to continue, to [b]want[/b] the basilisk to [b]fuck you silly[/b],
                            to [b]pound you[/b] until you can barely even [b]walk[/b] anymore!
                            This is the best feeling you've [b]ever[/b] felt in your entire life.
                            You want it to last [b]forever[/b]!
                            And while you know it'll eventually [i]stop[/i],
                            you intend to ride out every single second of it as though it is your last!
                        [/p]
                        [p]
                            Thinking about it, the risk of getting knocked up makes this even [b]more[/b] exciting.
                            Even more [b]taboo[/b].
                            A split second later you can feel it starting to happen.
                            The basilisk's member [b]twitches[/b] and [b]convulses[/b] inside of your cunt.
                            The tapered tip prods against your cervix, positioned perfectly for its load to shoot straight into your very womb.
                            Its loud hissing turns into a low-pitched growl, and the first of many ropes of hot,
                            thick, fertile feral cum squirts from its beastly length,
                            a second orgasm wracking your battered mind as the feeling of being [b]filled up[/b] sends you over the edge once more.
                        [/p]
                        [p]
                            Your brain completely short-circuiting, you begin to wildly [b]ride[/b] the basilisk's shaft once more,
                            intent on squeezing every single [b]drop[/b] of spunk and pleasure that you possibly can from its exotic prick.
                            Your tongue lolls out of your mouth in the same way the beast's did,
                            its [b]mind-breaking[/b] climax seemingly lasting for [b]minutes[/b],
                            minutes of absolute [b]ecstacy[/b] for both you and it.
                        [/p]
                        [p]
                            You lose all track of time, direction, and being.
                            You black out, your mind and body utterly overwhelmed by the sheer amount of pleasure you're feeling.
                            When you come to your senses, you find yourself lying atop the monster's soft scales,
                            your belly slightly bulging from the sheer amount of cum inside of you, despite a lot of it already having flooded out.
                            Even more heady spunk is squirting on top of your body, covering you from head to toe,
                            ${shirt ? `completely soaking your ${global.kobold.inventory.createItemWithTooltip(shirt)},` : ``}
                            the basilisk's unmatched orgasm only [b]now[/b] starting to fade.
                        [/p]
                        [p]
                            Completely exhausted, you find yourself resting for a while in the warm, comforting care of your new feral [b]lover[/b],
                            idly taking in massive amounts of male pheromones with every breath.
                            You can feel the basilisk's strong chest comfortingly heaving up and down below you.
                            It's contently [i]purring[/i], its needs thoroughly sated.
                            What a [b]good boy[/b].
                        [/p]
                        [p]
                            As much as you'd like to make this cave your new home, you're afraid you simply can't.
                            Not yet, anyway.
                            As the cum on your scales starts to cool,
                            and the rational part of your brain slowly starts to kick in,
                            you realize that you have places to be and things to do.
                            With a loud sigh, you reluctantly roll off of the big lizard, who lets out a distinctly sad whimper.
                        [/p]
                        [p]
                            Trying to rise to your feet, you discover just how sore the intense fucking has left your entire body.
                            You're certain it'll fade in due time, though.
                            Desspite having a rough time reaching the cave, you can say with a smile on your face that it was all worth it.
                            You thank the basilisk, though you doubt it understands, and give a small kiss on its snout.
                            Then, ${pants || undies ? `you redress, and` : `you`} prepare to leave.
                        [/p]`;
            if (virgin) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            setChoice ? subController.set(setChoice) : subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }


        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sub\UndergroundCaveReflectionGetFuckedInside.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sub\UndergroundCaveReflectionGetFuckedOutside.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionGetFuckedOutside(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveReflectionGetFuckedChoice.OUTSIDE;
    var reqController = ctrl.reflectSub;
    var setChoice = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.reflectOutside;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveReflectionGetFuckedContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var shirt = ctrl.misc.shirt();
        var pants = ctrl.misc.pants();
        var undies = ctrl.misc.underwear();

        var toReturn = ``;

        if (dick) {
            var virgin = ctrl.misc.analVirgin();
            toReturn += `
                        [p]
                            In the throes of an explosive orgasm,
                            right as the basilisk is about to shoot its load,
                            you slide right off your bestial partner's dick!
                            Just in time, too.
                            The [b]moment[/b] the tip of its member leaves your ass,
                            a first rope of thick, hot, sticky feral cum squirts out, splattering against your own dick,
                            marking your length with the male's scent,
                            the heady, masculine smell inspiring you to pleasure yourself further.
                        [/p]
                        [p]
                            You lay yourself down on top of the monster's chest, your hand finds its way in between your legs,
                            and you begin to rapidly jerk your already spent member off whilst fingering your slightly gaping tailhole,
                            all in an effort to keep your rush of arousal going strong.
                            A fresh wave of [b]intoxicating[/b] basilisk spunk splatters across your chest.
                            You grind your dick against the base of its throbbing length,
                            coaxing even more semen from the lizard's cock,
                            although it would have obviously preferred to have finished inside of you instead of letting its massive load go to waste.
                        [/p]
                        [p]
                            Alas, despite every bit of stimulation available, your cock is spent,
                            and you fail to find your second wind.
                            Instead of focusing on masturbation, you find yourself lying still atop the monster's soft scales.
                            Even more heady spunk is squirting on top of your already soaked body, covering you from head to toe,
                            ${shirt ? `completely drenching your ${global.kobold.inventory.createItemWithTooltip(shirt)},` : ``}
                            the basilisk's unmatched orgasm only [b]now[/b] starting to fade.
                        [/p]
                        [p]
                            Completely exhausted, you find yourself resting for a while in the warm, comforting care of your new feral [b]lover[/b],
                            idly taking in massive amounts of animalistic male pheromones with every breath.
                            You can feel the basilisk's strong chest comfortingly heaving up and down below you.
                            It's contently [i]purring[/i], its needs thoroughly sated.
                            What a [b]good boy[/b].
                        [/p]
                        [p]
                            As much as you'd like to make this cave your new home, you're afraid you simply can't.
                            Not yet, anyway.
                            As the cum on your scales starts to cool,
                            and the rational part of your brain slowly starts to kick in,
                            you realize that you have places to be and things to do.
                            With a loud sigh, you reluctantly roll off of the big lizard, who lets out a distinctly sad whimper.
                        [/p]
                        [p]
                            Trying to rise to your feet, you discover just how sore the intense fucking has left your entire body.
                            You're certain it'll fade in due time, though.
                            Desspite having a rough time reaching the cave, you can say with a smile on your face that it was all worth it.
                            You thank the basilisk, though you doubt it understands, and give a small kiss on its snout.
                            Then, ${pants || undies ? `you redress, and` : `you`} prepare to leave.
                        [/p]`;
            if (virgin) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        } else {
            var virgin = ctrl.misc.cuntVirgin();
            toReturn += `
                        [p]
                            In the throes of an explosive orgasm,
                            right as the basilisk is about to shoot its load,
                            you slide right off your bestial partner's dick!
                            Just in time, too.
                            The [b]moment[/b] the tip of its member leaves your slit,
                            a first rope of thick, hot, fertile feral cum squirts out, splattering against your stomach,
                            its heady, masculine scent inspiring you to pleasure yourself further.
                        [/p]
                        [p]
                            You lay yourself down on top of the monster's chest, your hand finds its way in between your legs,
                            and you begin to rapidly rub and finger yourself to keep your rush of arousal going strong,
                            [b]screaming[/b] out in pleasure as a fresh wave of [b]intoxicating[/b] basilisk spunk splatters across your chest.
                            You grind your cunt against the base of its throbbing length,
                            coaxing even more semen from the lizard's cock,
                            although it would have obviously preferred to [b]breed[/b] you instead of letting its massive load go to waste.
                        [/p]
                        [p]
                            Everything is a blur, from that point onwards.
                            You might have reached a second climax, but you can't quite remember.
                            When you come to your senses, you find yourself still lying atop the monster's soft scales.
                            Even more heady spunk is squirting on top of your already soaked body, covering you from head to toe,
                            ${shirt ? `completely drenching your ${global.kobold.inventory.createItemWithTooltip(shirt)},` : ``}
                            the basilisk's unmatched orgasm only [b]now[/b] starting to fade.
                        [/p]
                        [p]
                            Completely exhausted, you find yourself resting for a while in the warm, comforting care of your new feral [b]lover[/b],
                            idly taking in massive amounts of male pheromones with every breath.
                            You can feel the basilisk's strong chest comfortingly heaving up and down below you.
                            It's contently [i]purring[/i], its needs thoroughly sated.
                            What a [b]good boy[/b].
                        [/p]
                        [p]
                            As much as you'd like to make this cave your new home, you're afraid you simply can't.
                            Not yet, anyway.
                            As the cum on your scales starts to cool,
                            and the rational part of your brain slowly starts to kick in,
                            you realize that you have places to be and things to do.
                            With a loud sigh, you reluctantly roll off of the big lizard, who lets out a distinctly sad whimper.
                        [/p]
                        [p]
                            Trying to rise to your feet, you discover just how sore the intense fucking has left your entire body.
                            You're certain it'll fade in due time, though.
                            Desspite having a rough time reaching the cave, you can say with a smile on your face that it was all worth it.
                            You thank the basilisk, though you doubt it understands, and give a small kiss on its snout.
                            Then, ${pants || undies ? `you redress, and` : `you`} prepare to leave.
                        [/p]`;
            if (virgin) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            setChoice ? subController.set(setChoice) : subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sub\UndergroundCaveReflectionGetFuckedOutside.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sub\UndergroundCaveReflectionGetFucked.js

/**
 * Stare section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionGetFucked(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveReflectionDominanceChoice.SUB;
    var reqController = ctrl.reflectDominance;
    var subController = ctrl.reflectSub;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new UndergroundCaveReflectionGetFuckedOutside(ctrl);
    var secondNext = new UndergroundCaveReflectionGetFuckedInside(ctrl);

    var firstChoice = UndergroundCaveReflectionGetFuckedChoice.OUTSIDE;
    var secondChoice = UndergroundCaveReflectionGetFuckedChoice.INSIDE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskRode = true;
        var dick = ctrl.misc.dick();
        var breasts = ctrl.misc.boobs();
        var pants = ctrl.misc.pants();
        var undies = ctrl.misc.underwear();

        var lowerPantsAndUndies = `and already naked`;

        if (pants) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (undies) {
                lowerPantsAndUndies += ` and your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
            }
        } else if (undies) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
        }

        var toReturn = `[p]
                            The basilisk is more of a [b]big wimp[/b] than a menacing predator.
                            You can't imagine you were ever scared of this subby [b]bitch[/b] to begin with.
                            It [i]did[/i] look pretty hostile at first, but you guess that's almost a [i]requirement[/i] for living in a place like this.
                            Now that the creature has realized you're not its prey, it has revealed its true, submissive nature.
                        [/p]
                        [p]
                            Thinking of this situation as a [i]reversal of power[/i] awakens lustful feelings deep inside of you.
                            You've had this recurring [i]wet dream[/i] a few times, where kobolds would stage an uprising,
                            and you'd find yourself in front of the [b]juiciest[/b] maiden in all of Varanar,
                            with ten of your kind holding her struggling, naked form down for you to [b]ravage[/b].
                            This is [b]exactly[/b] like that dream, only instead of a maiden, you have a [b]basilisk[/b] to quench your lust upon,
                            the uprising was more of a [b]staring contest[/b], and there aren't really any kobolds around to hold the monster still.
                        [/p]
                        [p]
                            Okay, in retrospect, this entire ordeal might not be very much like your dream after all.
                            Still, you were reminded of it, and that's enough to get your
                            ${dick ? `member to slowly start to harden within your cockslit` : `slit to slightly moisten`}.
                            You begin to imagine all the things you could do to the creature, now that you've subdued it.
                            While not so similar to your dream, having your way with a large, ferocious animal [b]does[/b] pique your interest.
                        [/p]`;

        if (dick) {
            var virgin = ctrl.misc.analVirgin();
            global.kobold.body.analVirgin = false;
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at fucking a basilisk, you decide to act on your impulses.
                            You begin to stroke the emerging tip of your member,
                            which rapidly grows erect, fueled by the dirty thoughts popping up in your mind.
                            Hearing the monster give another pathetic whine makes you realize that it is [b]desperately[/b] needy.
                            Must not have been able to find a mate all season long.
                            But you can fix that.
                        [/p]
                        [p]
                            Your shaft grows painfully hard at the sheer thought of the feral's large, throbbing member having been pent up for [b]months[/b].
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, ${lowerPantsAndUndies}, your heart pounding and your breathing intensifying as you approach it.
                            You idly touch yourself a little,
                            inadvertently ending up spray a squirt of precum onto the basilisk's chin.
                            You watch heatedly as its tongue flicks out of its mouth, and greedily laps up your arousal staining its scales.
                            Its cock twitches, a small bead of its own pre at the tip signalling its readiness and eagerness.
                            It wants this.
                            It wants [b]you[/b].
                        [/p]
                        [p]
                            You straddle the beast, sitting down on its belly, its supple under-scales surprisingly warm to the touch.
                            You spread your legs, pressing the base of your cock against its frame, letting it feel your [b]heat[/b] in turn.
                            You can sense its entire body heaving up and down with every lustful breath it takes, its very heart excitedly beating.
                            You [b]wish[/b] kobolds were this [i]big[/i].
                        [/p]
                        [p]
                            Your fingers caress its scales, rubbing teasingly at its belly and sides.
                            Caressing soon turns to stroking, as you take hold of its shaft,
                            feeling it [b]throb[/b] in your palm while you move your hand up and down at an agonizingly slow pace.
                            It grunts in [b]pleasure[/b].
                            But you can tell that it wants [b]more[/b].
                            It [b]needs[/b] more.
                            A glob of heady precum squirts from the tip of its member, splattering directly onto your face,
                            [b]drenching[/b] your snout in the arousing scent of raw, feral male [b]potency[/b].
                            You, too, need this.
                        [/p]
                        [p]
                            You stand up again, wobbling on shaky feet, your dick harder than it's ever been.
                            Dizzy with lust from the inebriating, aphrodisiac-like liquid coating your face,
                            it takes some effort to properly position yourself over the basilisk's cock.
                            You grind your rear against its member,
                            lubricating your rear with its bountiful pre to ease the coming insertion.
                            The big beast [b]growls[/b] in [b]desire[/b] and [b]impatience[/b], and you can't help but mirror its sentiments.
                            One of your hands dips to the base of its shaft, to help hold it in place, while you use the other to help balance yourself with.
                            Then, with a long deep, lusty sigh, you begin to lower yourself onto the monster's length.
                        [/p]
                        [p]
                            You gasp out in pleasure, and a tiny bit of pain,
                            as you impale yourself on the the basilisk's warm, girthy manhood.
                            Your mind gets [b]flooded[/b] with need, your cock twitching.
                            It feels like you're about to cum on the spot.
                            It takes [b]all[/b] of your dwindling inhibitions to prevent yourself from trying to take the feral to the base.
                            Luckily, even while drunk with desire,
                            you realize that attempting to take a beast this [b]big[/b] in one go would be a [b]bad[/b] idea.  
                        [/p]
                        [p]
                            You come to a halt, after having only managed to take about a third of the basilisk's length on your own.
                            You barely manage to refrain from blowing your load from just putting it in for the first time.
                            You want [b]more[/b].
                            The lizard, too, is not yet satisfied.
                            It [b]bucks[/b] upwards, [b]ramming[/b] a few extra inches inside of you, sending waves of [b]pleasure[/b] coursing down your spine.
                            Your [b]whorish[/b] moans [b]echo[/b] throughout the cavern, your ass stuffed full to the very brim.
                            And the fun has only just begun.
                        [/p]
                        [p]
                            While your limbs feel [b]tingly[/b] and [b]numb[/b],
                            after [b]relishing[/b] in the feeling of the basilisk's cock for a few moments,
                            you find the strength to pull your way back up its shaft,
                            shuddering in ecstacy as each of its ridges pops from your rear like a set of beads.
                            Failing to hold yourself back, you let your body sink down again long before reaching the tip.
                            [b]Needing[/b] to be filled, you lower yourself [b]far[/b] faster this time,
                            [b]forcing[/b] half an inch extra into your already overly-full tailhole, your cock quivering in absolute [b]delight[/b].
                        [/p]
                        [p]
                            No longer needing to stabilize yourself, you begin fervently jerking your overly-sensitive member,
                            slightly humping against your own palm, which simultaneously causes you to pleasurably ride the well-hung lizard's tool.
                            Your small rocking motions, while enjoyable, aren't [b]nearly[/b] enough to get you off.
                            You begin to lift yourself up again, and getting the hang of things, you manage to raise yourself far faster than the last time,
                            pulling back a little further only to immediately slam down once more,
                            slowly settling into a rhythm, [b]milking[/b] the beast's shaft with your body,
                            riding it like the depraved, feral-fucking [b]slut[/b] that you are.
                        [/p]
                        [p]
                            The basilisk's cock proves every bit as [b]amazing[/b] as you imagined it would,
                            but the sheer [b]debasement[/b] of riding a feral creature in rut fuels your excitement.
                            You, a small and weak kobold,
                            usually preyed on by almost [b]everyone[/b] and [b]everything[/b] in the entire world,
                            are [b]on top[/b] of a [b]deadly monster[/b], which is entirely at the [b]mercy[/b] of your desires!
                            [b]You[/b] are the dominant and driving force of this animalistic [b]mating[/b],
                            and that thought drives you [b]wild[/b] with lust.
                            While you managed to avert it before, you can once again feel your premature end rapidly approaching.
                        [/p]
                        [p]
                            You let out another loud moan,
                            returning to rubbing the base of the creature's ridged member for a while,
                            your efforts rewarded by a hefty squirt of precum splattering against your insdies, further slickening your rear,
                            allowing you to pick up the pace one final time.
                        [/p]
                        [p]
                            The big reptile below you has all but gone [b]limp[/b] out of pleasure.
                            Its mouth hangs wide open, its tongue lolls out obscenely,
                            its hips are [b]struggling[/b] to hump up against you,
                            and one of its hind legs is absent-mindedly kicking air like it's nothing but an oversized [b]dog[/b]!
                            The once [b]mighty[/b] and [b]proud[/b] beast
                            has been reduced to little more than a meek [b]fucktoy[/b] for you to ride at your own behest.
                            And the thought of [i]breaking[/i] the basilisk, of reducing it to its most [b]base[/b] desires, that is what sends you over the edge.
                        [/p]
                        [p]
                            You only need to slide yourself down a few more times,
                            by now fitting more than [b]half[/b] of the monster's exotic dick inside you,
                            before reaching your end.
                            Every muscle in your meagre frame locks and tenses,
                            your cock [b]spasms[/b] wildly,
                            your internal balls [b]pump[/b] their contents through your shaft,
                            thick strands of your spunk squirting through the air before splattering down on the creature's tail,
                            and your eyes roll into the back of your head out of sheer [b]ecstacy[/b],
                            as a world-rocking orgasm hits your body hard.
                        [/p]
                        [p]
                            Even as your mind goes blank with pleasure,
                            you still realize the basilisk has not finished yet.
                            It hisses loudly, recovering enough of its vigor to start [b]wildly[/b] humping away at you once more,
                            outright [b]rutting[/b] your ass, sending spikes of pure [b]bliss[/b] shooting throughout overly-sensitive body!
                            It's close.
                            You know it is.
                        [/p]`;
        } else {
            var virgin = ctrl.misc.cuntVirgin();
            global.kobold.body.vaginalVirgin = false;
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at fucking a basilisk, you decide to act on your impulses.
                            You begin to touch your own slit, which rapidly dampens, fueled by the dirty thoughts popping up in your mind.
                            Hearing the monster give another pathetic whine makes you realize that it is [b]desperately[/b] needy.
                            Must not have been able to find a mate all season long.
                            But you can fix that.
                        [/p]
                        [p]
                            Your thighs moisten at the sheer thought of the feral's large, throbbing member having been pent up for [b]months[/b].
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, ${lowerPantsAndUndies}, your heart pounding and your breathing intensifying as you approach it.
                            Lust drips from your bare loins.
                            The scent of your own excitement fills the cave, inevitably caught by the basilisk as well.
                            It greedily huffs the air as its own arousal builds,
                            its cock twitching, a small bead of precum at the tip signalling its readiness and eagerness.
                            It wants this.
                            It wants [b]you[/b].
                        [/p]
                        [p]
                            You straddle the beast, sitting down on its belly, its supple under-scales surprisingly warm to the touch.
                            You spread your legs, pressing your slit against its frame, letting it feel your [b]heat[/b] in turn.
                            You can sense its entire body heaving up and down with every lustful breath it takes, its very heart excitedly beating.
                            You [b]wish[/b] kobolds were this [i]big[/i].
                        [/p]
                        [p]
                            Your fingers caress its scales, rubbing teasingly at its belly and sides.
                            Caressing soon turns to stroking, as you take hold of its shaft,
                            feeling it [b]throb[/b] in your palm while you move your hand up and down at an agonizingly slow pace.
                            It grunts in [b]pleasure[/b].
                            But you can tell that it wants [b]more[/b].
                            It [b]needs[/b] more.
                            A glob of heady precum squirts from the tip of its member, splattering directly onto your face,
                            [b]drenching[/b] your snout in the arousing scent of raw, feral male [b]potency[/b].
                            You, too, need this.
                        [/p]
                        [p]
                            You stand up again, wobbling on shaky feet, moisture dripping liberally down your cunt.
                            Dizzy with lust from the inebriating, aphrodisiac-like liquid coating your face,
                            it takes some effort to properly position yourself over the basilisk's cock.
                            You grind your loins against its member,
                            mixing your bountiful fluids together to form a suitable lubricant,
                            which you liberally coat your slit in to ease the coming insertion.
                            The big beast [b]growls[/b] in [b]desire[/b] and [b]impatience[/b], and you can't help but mirror its sentiments.
                            One of your hands dips to the base of its shaft, to help hold it in place, while you use the other to help balance yourself with.
                            Then, with a long deep, lusty sigh, you begin to lower yourself onto the monster's length.
                        [/p]
                        [p]
                            You gasp out in pleasure, and a tiny bit of pain,
                            as you impale yourself on the the basilisk's warm, girthy manhood.
                            ${virgin ? `Your virginity is taken, as signified by a trickle of blood running down the creature's member.` : ``}
                            Your mind gets [b]flooded[/b] with need.
                            It takes [b]all[/b] of your dwindling inhibitions to prevent yourself from trying to take the feral to the base.
                            Luckily, even while drunk with desire,
                            you realize that attempting to take a beast this [b]big[/b] in one go would be a [b]bad[/b] idea.    
                        [/p]
                        [p]
                            Copious amounts of female arousal leak down the basilisk's shaft, helping to lubricate the way.
                            Still, you only manage to take about a third of the its length on your own.
                            You want [b]more[/b].
                            The lizard, too, is not yet satisfied.
                            It [b]bucks[/b] upwards, [b]ramming[/b] a few extra inches inside of you,
                            its tip ending up poking and [b]prodding[/b] at your cervix, sending waves of [b]pleasure[/b] coursing down your spine.
                            Your [b]whorish[/b] moans [b]echo[/b] throughout the cavern, your cunt stuffed full to the very brim.
                            And the fun has only just begun.
                        [/p]
                        [p]
                            While your limbs feel [b]tingly[/b] and [b]numb[/b],
                            after [b]relishing[/b] in the feeling of the basilisk's cock for a few moments,
                            you find the strength to pull your way back up its shaft,
                            shuddering in ecstacy as each of its ridges wetly pops from your cunt.
                            Failing to hold yourself back, you let your body sink down again long before reaching the tip.
                            [b]Needing[/b] to be filled, you lower yourself [b]far[/b] faster this time,
                            [b]forcing[/b] half an inch extra into your already overly-full cunt, which quivers in absolute [b]delight[/b].
                        [/p]
                        [p]
                            No longer needing to stabilize yourself, you begin fervently toying with your ${breasts ? `breasts` : `flat chest`},
                            ${breasts ? `rubbing and squeezing them` : `rubbing your hands all over your tender scales`},
                            while simultaneously working to lift yourself up again.
                            Getting the hang of things, you manage to raise yourself faster than the last time,
                            pulling back a little further only to immediately slam down once more,
                            slowly settling into a rhythm, [b]milking[/b] the beast's shaft with your cunt,
                            riding it like the depraved, feral-fucking [b]slut[/b] that you are.
                        [/p]
                        [p]
                            The basilisk's cock proves every bit as [b]amazing[/b] as you imagined it would,
                            but the sheer [b]debasement[/b] of riding a feral creature in rut fuels your excitement.
                            You, a small and weak kobold,
                            usually preyed on by almost [b]everyone[/b] and [b]everything[/b] in the entire world,
                            are [b]on top[/b] of a [b]deadly monster[/b], which is entirely at the [b]mercy[/b] of your desires!
                            [b]You[/b] are the dominant and driving force of this animalistic [b]mating[/b],
                            and that thought drives you [b]wild[/b] with lust.
                            You can already feel your premature end rapidly approaching.
                        [/p]
                        [p]
                            You let out another loud moan,
                            returning to rubbing the base of the creature's ridged member for a while,
                            your efforts rewarded by a hefty squirt of precum splattering against your cervix, further slickening your insides,
                            allowing you to pick up the pace one final time.
                        [/p]
                        [p]
                            The big reptile below you has all but gone [b]limp[/b] out of pleasure.
                            Its mouth hangs wide open, its tongue lolls out obscenely,
                            its hips are [b]struggling[/b] to hump up against you,
                            and one of its hind legs is absent-mindedly kicking air like it's nothing but an oversized [b]dog[/b]!
                            The once [b]mighty[/b] and [b]proud[/b] beast
                            has been reduced to little more than a meek [b]fucktoy[/b] for you to ride at your own behest.
                            And the thought of [i]breaking[/i] the basilisk, of reducing it to its most [b]base[/b] desires, that is what sends you over the edge.
                        [/p]
                        [p]
                            You only need to slide yourself down a few more times,
                            by now fitting more than [b]half[/b] of the monster's exotic dick inside you,
                            before reaching your end.
                            Every muscle in your meagre frame locks and tenses,
                            your cunt [b]clenches[/b] down around the feral's girth,
                            your juices liberally [b]flood[/b] from your slit,
                            [b]soaking[/b] the remainder of the basilisk's cock,
                            a pool of feminine fluids forming at the very base,
                            and your eyes roll into the back of your head out of sheer [b]ecstacy[/b],
                            as a world-rocking orgasm hits your body hard.
                        [/p]
                        [p]
                            Even as your mind goes blank with pleasure,
                            you still realize the basilisk has not finished yet.
                            It hisses loudly, recovering enough of its vigor to start [b]wildly[/b] humping away at you once more,
                            outright [b]rutting[/b] you, sending spikes of pure [b]bliss[/b] shooting throughout overly-sensitive body,
                            [b]forcing[/b] your orgasm to go [b]on and on[/b]!
                            You know it's about to cum, but you [b]really[/b] don't want to pull off of it yet,
                            despite the rational part of your mind telling you [b]not[/b] to let it finish inside of you.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Outside</button>`;

        var tooltip = global.parser.parse(`Finish it off some other way.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Inside</button>`;

        var tooltip = global.parser.parse(`Let it fill you up.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">Make it cum:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var UndergroundCaveReflectionGetFuckedChoice = {
    OUTSIDE: 1,
    INSIDE: 2
};

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sub\UndergroundCaveReflectionGetFucked.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sex\UndergroundCaveReflectionSexContinue.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionSexContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.reflectSexContinue;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var baseMale = ctrl.misc.baseGender() === GenderEnum.MALE;

        var toReturn = `[p]
                            With the basilisk shown its place,
                            and yourself asserted as the [b]alpha ${baseMale ? `male` : `female`}[/b],
                            you decide it's time to find a way out of the cave.
                            Leaving the feral to revel in post-orgasmic bliss, 
                            you begin to follow the river-branch, hoping it'll lead you to some kind of exit.
                        [/p]
                        [p]
                            You leisurely stroll alongside the water, until daylight comes into sight.
                            As you expected: the water [b]was[/b] a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that,
                            you managed to ${dick ? `butt` : `finger`}fuck a basilisk into [i]willing[/i] submission.
                            You doubt a lot of people can claim to have had sex with a mythological being, and much less still lived to tell the tale.
                        [/p]
                        [p]
                            You've gained a few personal memories that you'll undoubtedly find yourself masturbating to later,
                            along with some awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sex\UndergroundCaveReflectionSexContinue.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sex\UndergroundCaveReflectionSex.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionSex(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveReflectionDominanceChoice.SEX;
    var reqController = ctrl.reflectDominance;
    var setChoice = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.reflectSex;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveReflectionSexContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskDomFucked = true;
        var dick = ctrl.misc.dick();
        var dick = ctrl.misc.dick();
        var pants = ctrl.misc.pants();
        var undies = ctrl.misc.underwear();
        var oralVirgin = ctrl.misc.oralVirgin();

        var lowerPantsAndUndies = `and already naked`;

        if (pants) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (undies) {
                lowerPantsAndUndies += ` and your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
            }
        } else if (undies) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
        }

        var toReturn = `[p]
                            The basilisk is more of a [b]big wimp[/b] than a menacing predator.
                            You can't imagine you were ever scared of this subby [b]bitch[/b] to begin with.
                            It [i]did[/i] look pretty hostile at first, but you guess that's required for living in a place like this.
                            Now that the creature has realized you're not its prey, it has revealed its true, submissive nature.
                        [/p]
                        [p]
                            Thinking of this situation as a [i]reversal of power[/i] awakens lustful feelings deep inside of you.
                            You've had this recurring [i]wet dream[/i] a few times, where kobolds would stage an uprising,
                            and you'd find yourself in front of the [b]juiciest[/b] maiden in all of Varanar,
                            with ten of your kind holding her struggling, naked form down for you to [b]ravage[/b].
                            This is [b]exactly[/b] like that dream, only instead of a maiden, you have a [b]basilisk[/b] to quench your lust upon,
                            the uprising was more of a [b]staring contest[/b], and there aren't really any kobolds around to hold the monster still.
                        [/p]
                        [p]
                            Okay, in retrospect, this entire ordeal might not be very much like your dream after all.
                            Still, you were reminded of it, and that's enough to get your
                            ${dick ? `member to slowly start to harden within your cockslit` : `slit to slightly moisten`}.
                            You begin to imagine all the things you could do to the creature, now that you've subdued it.
                            While not so similar to your dream, having your way with a large, ferocious animal [b]does[/b] pique your interest.
                        [/p]`;

        if (dick) {
            var virgin = ctrl.misc.dickVirgin();
            global.kobold.body.dickVirgin = false;
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at molesting a basilisk, you decide to act on your impulses.
                            You begin to stroke the emerging tapered tip of your member,
                            which rapidly grows erect, fueled by the dirty thoughts popping up in your mind.
                            You want to fuck its stony ass, concealed underneath its tail.
                            You want to walk on the wild side, to tempt fate itself,
                            to see whether or not the beast still has enough [b]fight[/b] in it to resist, to [b]thrash around[/b] and [b]maim you[/b].
                        [/p]
                        [p]
                            This cute little, scale-covered, oversized [b]puppy[/b]?
                            Surely not.
                            After being half-petrified, it has become completely and utterly [b]tame[/b].
                            You bet the creature won't even [b]try[/b] to stop you from fucking its behind.
                            To prove your point, you reach down to scratch its chin.
                            Instead of fighting back in any form or shape,
                            it rumbles pleasantly underneath your touch, like a dog being pet.
                            Its maw opens wide, serpentine tongue lolling out.
                            The beast sways its tail from side to side, seductively wriggling its ass about.
                            While in reality it's simply [i]enjoying[/i] your scratching,
                            you prefer to interpret it like the crocodillian [b]offering[/b] itself to you.
                        [/p]
                        [p]
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, ${lowerPantsAndUndies}, you approach the basilisk's backside,
                            and straddle the base of its tail, dropping to your knees on top of it.
                            Its member mere inches from your face, the creeping scent of a feral male in rut [b]fills[/b] your lungs.
                            The smell acts like an aphrodisiac, arousing you further, [b]enticing[/b] you to fuck the beast.
                        [/p]
                        [p]
                            With one of your hands, you begin to [b]toy[/b] with the basilisk's member,
                            giving it a few tentative squeezes and touches.
                            In return, you are rewarded with a thick, heady squirt of precum splattering from the tapered tip,
                            along the entire length of the monster's underside,
                            its potent fragrance adding to the already intoxicating aroma of male [b]lust[/b].
                            With your other hand, you begin to [i]probe[/i] around its rear.
                            It doesn't seem to mind one bit.
                        [/p]
                        [p]
                            After giving yourself a few much-needed strokes as well,
                            you line up the tip of your member with the basilisk's tailhole.
                            You immediately begin to apply pressure, pressing your tapered length into its overly tight rear.
                            Once past its relaxed, but still exceedingly tight sphincter,
                            you easily manage to ram the rest of your shaft inside, bottoming out in a single, smooth thrust.
                            The monster hisses, but it remains calm and submissive.
                            It stays in place, accepting its position as your [b]bitch[/b].
                            Just as planned.
                        [/p]
                        [p]
                            While the creature's exterior is hard, tough and cold, its insides are the exact opposite.
                            Warm, fleshy, and exceedingly tight,
                            almost what you imagine a human's ass to feel like, minus a bubbly butt to rest your groin against.
                            A pleasant surprise, and not an unwelcome one at that.
                            You lean forwards ever so slightly, almost pressing your chest up against the monster's length,
                            your snout almost [b]rubbing[/b] in the bauble of precum collected at the tip of its exotic cock.
                            While outright [b]huffing[/b] its enticing smell, you slowly begin to thrust in and out of the crocodillian's ass.
                        [/p]
                        [p]
                            The basilisk's tight insides prove to be quite pleasurable,
                            but the sheer [b]debasement[/b] of what you're doing drives your excitement through the roof.
                            You're [b]fucking[/b] a feral.
                            The untamed [b]beast[/b] has been completely [b]broken[/b], turned into your [b]bitch[/b].
                            And on top of all of this, you're tempting fate itself by ramming your dick into a creature that could easily [b]tear you apart[/b]!
                            Not that it shows any signs of [i]disliking[/i] what is happening.
                            If anything, its quivering shaft shows it wants this every bit as much as you do, if not even [b]more[/b].
                        [/p]
                        [p]
                            Closing your eyes for a moment, your mind briefly wanders back to the dream that inspired you to fuck the basilisk.
                            Snapping back to reality, however, you realize that this situation is [b]even better[/b] than what you imagined.
                            Having the [b]pleasure[/b] of dominating a [b]lethal beast[/b],
                            one that could [b]rip you to shreds[/b] were it not such a submissive [b]bitch[/b],
                            stirs your loins far more than the thought of raping a helpless damsel.
                            While you've only been absent-mindedly humping away at the big lizard's rear for a few seconds short of a minute,
                            you can already feel your premature end rapidly approaching.
                        [/p]
                        [p]
                            Lost in the heat of the moment, you press your lips against the tapered tip of the basilisk's cock,
                            taking the very end of its length into your mouth, suckling it like a needy [b]whore[/b],
                            collecting as much of its precum onto your tongue as you possibly can.
                            The intense [b]taste[/b] of basilisk dick entices you further,
                            its pre coating your tastebuds and spreading throughout your mouth makes you feel [b]lightheaded[/b] with [b]lust[/b].
                            While upping your pace, you begin to outright suck the beast off,
                            its lower body slowly humping back against your muzzle, its tailhole riding your dick in the process.
                        [/p]
                        [p]
                            You let out a lustful groan, rubbing one of your hands along the creature's soft underside.
                            You briefly consider giving it a few [b]spanks[/b], but you decide against that idea.
                            Instead, you wrap both of your hands around its girthy length,
                            and dutifully begin to [b]jerk off[/b] the large part of its shaft that you fail to fit in your mouth.
                            It's been a [b]good boy[/b], after all.
                            The fact that it doesn't resist, that it seems [b]willing[/b] to cooperate in all of this,
                            [b]that[/b] sends you over the edge.
                        [/p]
                        [p]
                            To your surprise, while you were just about to finish, the big creature reaches its end before you do!
                            A first wave of [b]thick[/b], [b]potent[/b] feral semen floods your mouth, driving you [b]wild[/b] with desire.
                            You give the beast's rear a few more rapid humps, before reaching your own climax.
                            Your member [b]pulses[/b] and [b]spasms[/b], spunk pumping from deep within your internal balls,
                            through your shaft, squirting directly inside of the monster's body.
                            You eagerly gulp down several waves of the basilisk's intoxicating jizz,
                            before removing your muzzle from its cock, letting the rest of its intense orgasm squirt along its own underbelly,
                            painting its own body a pearly white, leaving it a [b]very[/b] messy lizard.
                        [/p]
                        [p]
                            Your lust sated, and the entire contents of your nuts emptied deep within the big beast's body,
                            you pull your length from its butt, your cum immediately beginning to ooze from its hole,
                            staining the base of its tail, although there's not nearly as much of the stuff as there is on its upper body,
                            which is completely [b]soaked[/b].
                            You rise to your feet${pants || undies ? ` and redress` : ``},
                            observing the basilisk as it simply lies there in a dazed, hazy afterglow,
                            basking in pleasure, completely relaxed, its tongue lolling out of its mouth.
                            It's almost kind of adorable.
                            Too bad you can't stay here forever.
                        [/p]`;
            if (virgin) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        } else {
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at molesting a basilisk, you decide to act on your impulses.
                            You begin to touch your own slit, which rapidly dampens, fueled by the dirty thoughts popping up in your mind.
                            You want to touch its stony ass, concealed underneath its tail.
                            You want to walk on the wild side, to tempt fate itself,
                            to see whether or not the beast still has enough [b]fight[/b] in it to resist, to [b]thrash around[/b] and [b]maim you[/b].
                        [/p]
                        [p]
                            This cute little, scale-covered, oversized [b]puppy[/b]?
                            Surely not.
                            After being half-petrified, it has become completely and utterly [b]tame[/b].
                            You bet the creature won't even [b]try[/b] to stop you from toying with its behind.
                            To prove your point, you reach down to scratch its chin.
                            Instead of fighting back in any form or shape,
                            it rumbles pleasantly underneath your touch, like a dog being pet.
                            Its maw opens wide, serpentine tongue lolling out.
                            The beast sways its tail from side to side, seductively wriggling its ass about.
                            While in reality it's simply [i]enjoying[/i] your scratching,
                            you prefer to interpret it like the crocodillian [b]offering[/b] itself to you.
                        [/p]
                        [p]
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, ${lowerPantsAndUndies}, you approach the basilisk's backside,
                            and straddle its tail, dropping to your knees before lying flat down on top of it.
                            The base of the monster's member mere inches from your face, the creeping scent of a feral male in rut [b]fills[/b] your lungs.
                            The smell acts like an aphrodisiac, arousing you further, [b]enticing[/b] you to [i]play[/i] with the beast.
                        [/p]
                        [p]
                            With one of your hands, you begin to [b]toy[/b] with the basilisk's member,
                            giving it a few tentative squeezes and touches.
                            In return, you are rewarded with a thick, heady squirt of precum splattering from the tapered tip,
                            along the entire length of the monster's underside,
                            its potent fragrance adding to the already intoxicating aroma of male [b]lust[/b].
                            With your other hand, you begin to [i]probe[/i] around its rear.
                            It doesn't seem to mind one bit.
                        [/p]
                        [p]
                            Enticed, you grab hold of the basilisk cock, and slowly rub your comparatively small palm up and down the its shaft,
                            stroking its length at an agonizingly slow pace.
                            Wouldn't want it to finish [i]too soon[/i], now would you?
                            You have plenty more in store for the big beast.
                            Leaning your head in even closer, you press your muzzle not against its member, but against its tender rear.
                            Your tongue snakes out of your mouth, and in a fit of deviancy, you tentatively lick the creature's tailhole,
                            elliciting another lewd grunt from its muzzle.
                            Sterile and tasteless, apart from a remote hint of [i]sweat[/i], but the basilisk [i]did[/i] seem to like it.
                            Its member throbs in your hand, and it slowly starts to move its hips on its own,
                            humping against your palm.
                        [/p]
                        [p]
                            Again and again, you lap at the monster's rear, your tongue settling into a rhythm,
                            your saliva wettening its scales and lubing its hole, preparing it for what is to come.
                            The feral is blissfully unaware of your devious plans.
                            You strengthen your grip on its member, and increase the speed of your hand, properly jerking it off.
                            You can feel its precum slickening your palm, reducing friction,
                            making the entire ordeal even [b]more[/b] pleasurable for the crocodillian.
                        [/p]
                        [p]
                            Finally, with the beast completely lost in pleasure, you decide to move on to the [i]second[/i] phase of your plan.
                            With the large lizard fully relaxed, you pull your head back a little,
                            a few thick strings of saliva connecting your muzzle to its ass still.
                            You move your other hand to the creature's rear.
                            This is where the real fun begins.
                        [/p]
                        [p]
                            You press one of your fingers against the monster's tailhole, teasingly circling it, spreading your spit around.
                            Then, while simultaneously upping the pace of your frantic jerking once more,
                            you slide the digit into the feral's spit-lubed rear, ramming it inside of the basilisk to the very knuckle.
                            The large lizard is obviously surprised, but it remains calm and submissive, accepting its position as your [b]bitch[/b].
                            Or maybe it [b]likes[/b] its ass being played with.
                        [/p]
                        [p]
                            While continuing to firmly stroke the beast's ridged member,
                            you slide a second finger inside of its thoroughly lubricated behind,
                            and begin to pump both of the digits in and out of the creature's body, properly [b]fingering[/b] it while jerking it off,
                            sending spikes of pleasure coursing down its spine.
                            You doubt it's [b]ever[/b] felt anything like [b]this[/b] before.
                            And you're just getting [b]started[/b].
                        [/p]
                        [p]
                            Your mouth freed, you shuffle upwards along the basilisk's body,
                            bringing your snout to the very tip of its member, outright [b]huffing[/b] its potent, male aroma for a few moments.
                            Through its own humping, it inadvertently presses its tapered tip against your mouth,
                            rubbing a considerable bauble of collected precum off on your muzzle.
                            No longer able to contain your excitement,
                            you eagerly part your lips,
                            taking the very end of its length into your mouth, and start [b]suckling[/b] it like a needy [b]whore[/b],
                            collecting as much of the monster's precum onto your tongue as you possibly can.
                        [/p]
                        [p]
                            The intense [b]taste[/b] of basilisk dick makes you feel [b]lightheaded[/b] with [b]desire[/b].
                            The pungent pre coating your tastebuds and spreading throughout your mouth awakens a [b]lust[/b],   
                            a [b]craving[/b], a [b]need[/b] for the contents of its internal balls.
                            If it weren't for your fingers up its ass, you'd say the [b]feral[/b] was in complete control of this situation, and [b]not[/b] you.
                            You begin to rapidly stroke the remaining part of its member that you can't fit into your mouth.
                        [/p]
                        [p]
                            Its humping intensifies, the feral outright [b]fucking[/b] your face.
                            You can feel the its end approaching.
                            You up the intensity of your tailhole-fingering one last time, vigorously pumping your fingers in and out of its tight, warm body.
                            Its shaft twitches in your grasp.
                            Precum shoots down the back of your mouth, into your throat.
                            Its end is getting [b]very[/b] near.
                        [/p]
                        [p]
                            The big creature lets out a pleasured hiss as it reaches its climax,
                            rabidly humping away at your muzzle while a first wave of [b]thick[/b],
                            [b]potent[/b] feral semen floods your mouth, your brain [b]blanking out[/b] with desire.
                            You grind your loins against its stony tail, not quite pleasuring yourself properly,
                            but still slightly scratching the itch sucking off this [b]beast[/b] of male has awoken.
                            You eagerly gulp down several waves of the basilisk's intoxicating jizz,
                            before pulling your muzzle from its cock,
                            letting the rest of its intense orgasm squirt along its own underbelly,
                            painting its own body a pearly white, leaving it a [b]very[/b] messy lizard.
                        [/p]
                        [p]
                            You slide your fingers from the creature's rear.
                            Your belly filled, and the entire contents of the big beast's nuts emptied,
                            your job is done here.
                            That was [b]fun[/b].
                            You rise to your feet${pants || undies ? ` and redress` : ``}.
                            Before leaving, you take a moment to observe the basilisk as it simply lies there in a dazed, hazy afterglow,
                            basking in pleasure, completely relaxed, its tongue lolling out of its mouth.
                            It's almost kind of adorable.
                        [/p]`;
        }

        global.kobold.body.oralVirgin = false;
        if (oralVirgin) {
            toReturn += `[p][c][b]You are no longer an oral virgin.[/b][/c][/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            setChoice ? subController.set(setChoice) : subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\sex\UndergroundCaveReflectionSex.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\oral\UndergroundCaveReflectionOralContinue.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionOralContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.reflectOralContinue;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var baseMale = ctrl.misc.baseGender() === GenderEnum.MALE;

        var toReturn = `[p]
                            With the basilisk shown its place,
                            and yourself asserted as the [b]alpha ${baseMale ? `male` : `female`}[/b],
                            you decide it's time to find a way out of the cave.
                            Leaving the feral to enjoy the aftertaste of your ${dick ? `cum` : `feminine fluids`},
                            you begin to follow the river-branch, hoping it'll lead you to some kind of exit.
                        [/p]
                        [p]
                            The beast whines as you start to walk away, hoping it'd get some action itself, but luckily it does not follow you.
                            You leisurely stroll alongside the water, until daylight comes into sight.
                            As you expected: the water [b]was[/b] a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that, you managed to facefuck a basilisk into [i]willing[/i] submission.
                            You doubt a lot of people can claim to have had sex with a mythological being, and much less still lived to tell the tale.
                        [/p]
                        [p]
                            You've gained a few personal memories that you'll undoubtedly find yourself masturbating to later,
                            along with some awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\oral\UndergroundCaveReflectionOralContinue.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\oral\UndergroundCaveReflectionOral.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionOral(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveReflectionDominanceChoice.ORAL;
    var reqController = ctrl.reflectDominance;
    var setChoice = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.reflectOral;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveReflectionOralContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskOral = true;
        var dick = ctrl.misc.dick();
        var pants = ctrl.misc.pants();
        var undies = ctrl.misc.underwear();

        var lowerPantsAndUndies = `and already naked`;

        if (pants) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (undies) {
                lowerPantsAndUndies += ` and your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
            }
        } else if (undies) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
        }

        var toReturn = `[p]
                            The basilisk is more of a [b]big wimp[/b] than a menacing predator.
                            You can't imagine you were ever scared of this subby [b]bitch[/b] to begin with.
                            It [i]did[/i] look pretty hostile at first, but you guess that's required for living in a place like this.
                            Now that the creature has realized you're not its prey, it has revealed its true, submissive nature.
                        [/p]
                        [p]
                            Thinking of this situation as a [i]reversal of power[/i] awakens lustful feelings deep inside of you.
                            You've had this recurring [i]wet dream[/i] a few times, where kobolds would stage an uprising,
                            and you'd find yourself in front of the [b]juiciest[/b] maiden in all of Varanar,
                            with ten of your kind holding her struggling, naked form down for you to [b]ravage[/b].
                            This is [b]exactly[/b] like that dream, only instead of a maiden, you have a [b]basilisk[/b] to quench your lust upon,
                            the uprising was more of a [b]staring contest[/b], and there aren't really any kobolds around to hold the monster still.
                        [/p]
                        [p]
                            Okay, in retrospect, this entire ordeal might not be very much like your dream after all.
                            Still, you were reminded of it, and that's enough to get your
                            ${dick ? `member to slowly start to harden within your cockslit` : `slit to slightly moisten`}.
                            You begin to imagine all the things you could do to the creature, now that you've subdued it.
                            While not so similar to your dream, having your way with a large, ferocious animal [b]does[/b] pique your interest.
                        [/p]`;

        if (dick) {
            var virgin = ctrl.misc.dickVirgin();
            global.kobold.body.dickVirgin = false;
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at molesting a basilisk, you decide to act on your impulses.
                            You begin to stroke the emerging tapered tip of your member, which rapidly grows erect, fueled by the dirty thoughts popping up in your mind.
                            Its muzzle.
                            You want to fuck its muzzle, filled with sharp, pointy teeth.
                            You want to walk on the wild side, to tempt fate itself,
                            to see whether or not the beast still has enough [b]fight[/b] in it to resist, to [b]bite down[/b].
                        [/p]
                        [p]
                            This cute little, scale-covered, oversized [b]puppy[/b]?
                            Surely not.
                            After being half-petrified, it has become completely and utterly [b]tame[/b].
                            You bet the creature won't even [b]try[/b] to stop you from fucking its maw.
                            To prove your point, you reach down to scratch its chin.
                            Instead of fighting back in any form or shape,
                            it rumbles pleasantly underneath your touch, like a dog being pet.
                            Its tail slowly sways back and forth, and its maw opens wide, serpentine tongue lolling out.
                            While in reality it's simply [i]enjoying[/i] your scratching,
                            you prefer to interpret it like the crocodillian [b]offering[/b] its mouth for your length.
                        [/p]
                        [p]
                            As you remove your hand from its chin, the dog-like creature rolls back onto its feet,
                            its movements seeming a little more [i]fluid[/i] than before.
                            The petrification must be wearing off.
                        [/p]
                        [p]
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, ${lowerPantsAndUndies}, you approach the basilisk's head, dropping to your knees directly in front of it,
                            your erect member mere inches away from the feral's snout.
                            Its mouth is still open, and the monster dares not look at you.
                            A perfect opportunity.
                        [/p]
                        [p]
                            You line up the tip of your member with the beast's mouth, and with a single, smooth roll of your hips,
                            you slide your length inside of its muzzle, the bottom of your shaft pleasantly rubbing alongside the basilisk's slender, forked tongue.
                            While the creature's exterior is hard, tough and cold, the insides of its mouth are the exact opposite.
                            Warm, fleshy, [b]wet[/b] with saliva, almost what you imagine a human's gums must feel like.
                            A pleasant surprise, and not an unwelcome one.
                        [/p]
                        [p]
                            The lizard doesn't seem to mind your dick being rammed into its gaping maw one bit.
                            Like it's [i]used[/i] to the the taste of kobold cock.
                            If you didn't know any better, you'd think it's even trying to [b]suckle[/b] at your member, and [b]probing[/b] your length with its tongue.
                            While you're certain these are all just natural reflexes,
                            and the beast has never done anything of the sort before, you can't help but imagine that it [b]has[/b].
                            You fantasize that it's completely and utterly submissive to you, a [b]slut[/b] eager to [b]please[/b].
                        [/p]
                        [p]
                            Lost in pleasure, you close your eyes and lean back ever so slightly,
                            while starting to thrust in and out of the monster's muzzle.
                            The sensation isn't [b]too[/b] pleasurable, but the sheer [b]debasement[/b] of what you're doing excites you to no end.
                            You're [b]fucking[/b] a feral.
                            The untamed [b]beast[/b] has been completely [b]broken[/b], turned into your [b]bitch[/b].
                            And on top of all of this, you're tempting fate itself by ramming your cock into a mouth full of sharp teeth!
                        [/p]
                        [p]
                            With your eyes closed, your mind briefly wanders back to the dream that inspired you to facefuck the basilisk.
                            Snapping back to reality, however, you realize that this situation is [b]even better[/b] than what you imagined.
                            Having the [b]pleasure[/b] of dominating a [b]lethal beast[/b],
                            one that could [b]rip you to shreds[/b] were it not so interested in [b]sucking your dick[/b],
                            stirs your loins far more than the thought of raping a helpless damsel.
                            While you've only been absent-mindedly humping away at the big lizard's maw for a few seconds short of a minute,
                            you can already feel your premature end rapidly approaching.
                        [/p]
                        [p]
                            You let out a lustful groan, rubbing one of your hands along the soft underside of the creature's muzzle,
                            poking at its sharp teeth with one of your fingers to remind yourself of the [b]arousing danger[/b] you're in.
                            The basilisk shows no signs of biting, though.
                            Either it doesn't know what's happening, or it's a [b]willing[/b] participant in all of this.
                            The thought of having reduced the once-proud crocodillian to an [b]eager[/b] cocksucker,
                            [b]that[/b] sends you over the edge.
                        [/p]
                        [p]
                            You give the beast's mouth a few more rapid thrusts, before reaching your end.
                            Your member [b]pulses[/b] and [b]spasms[/b], spunk pumping from deep within your internal balls,
                            through your shaft, squirting directly atop the monster's forked tongue.
                            You don't even need to hold it in place.
                            It deftly stays put, letting you empty the entire contents of your nuts into its muzzle.
                        [/p]
                        [p]
                            Your lust sated, you pull your cock from the big beast's maw.
                            It closes its jaws, and licks its lips, almost looking somewhat [i]pleased[/i], audibly gulping down your cum.
                            Maybe it [i]did[/i] want this.
                            Maybe basilisks [i]do[/i] suck eachother off during rut or something.
                            Whatever the case, you didn't get killed, and ended up getting your rocks off.
                            You'll chalk this one up as a win.
                            You rise to your feet${pants || undies ? ` and redress` : ``}.
                        [/p]`;

            if (virgin) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        } else {
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at molesting a basilisk, you decide to act on your impulses.
                            You begin to touch your own slit, which rapidly dampens, fueled by the dirty thoughts popping up in your mind.
                            Its muzzle.
                            You want to ride its muzzle, filled with sharp, pointy teeth.
                            You want to walk on the wild side, to tempt fate itself,
                            to see whether or not the beast still has enough [b]fight[/b] in it to resist, to [b]bite[/b].
                        [/p]
                        [p]
                            This cute little, scale-covered, oversized [b]puppy[/b]?
                            Surely not.
                            After being half-petrified, it has become completely and utterly [b]tame[/b].
                            You bet the creature won't even [b]try[/b] to stop you from riding its maw.
                            To prove your point, you reach down to scratch its chin.
                            Instead of fighting back in any form or shape,
                            it rumbles pleasantly underneath your touch, like a dog being pet.
                            Its tail slowly sways back and forth, and its maw opens wide, serpentine tongue lolling out.
                            While in reality it's simply [i]enjoying[/i] your scratching,
                            you prefer to interpret it like the crocodillian [b]offering[/b] its [i]services[/i].
                        [/p]
                        [p]
                            As you remove your hand from its chin, the dog-like creature rolls back onto its feet,
                            its movements seeming a little more [i]fluid[/i] than before.
                            The petrification must be wearing off.
                        [/p]
                        [p]
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, ${lowerPantsAndUndies}, you approach the basilisk's head, dropping to your knees directly in front of it,
                            your wet slit mere inches away from the feral's snout.
                            Its mouth is still open, and it dares not look at you.
                            A perfect opportunity.
                        [/p]
                        [p]
                            You gently pet the beast's head, and find it all too eager to comply,
                            easy to direct and willing to play along as you plant your lusty twat directly on top of its snout. 
                            The scent of a female lizard seems to rouse the creature from its passive, submissive stance.
                            Its nostrils flare, and its slender, forked tongue slips out to taste your arousal, 
                            lightly flicking along your most sensitive scales, before retracting into its massive, toothy maw once more.
                        [/p]
                        [p]
                            The feminine taste of an aroused kobold girl spreading throughout its mouth only encourages the feral further.
                            The basilisk's tongue flicks out again, pressing deeper into your folds this time, even slightly parting them,
                            collecting as much of your sweet feminine excitement as it possible can, before retracting the flexible organ once more.
                        [/p]
                        [p]
                            You let out a pleasured sigh as the basilisk begins to outright [b]lap[/b] at your slit,
                            licking you much like a common bear would steal honey from a beekeeping friar's storehouse:
                            greedily and speedily.
                            Like it's [b]eager[/b] to please its new [i]mate[/i].
                            Despite the creature's rock-like appearance, its tongue feels [b]warm[/b] and [b]wet[/b],
                            and the forked split at the end is just exotic enough to get your filthy mind [b]racing[/b].
                        [/p]
                        [p]
                            Lost in pleasure, you close your eyes and lean back ever so slightly,
                            slowly rubbing your crotch back and forth over the monster's muzzle while it continues to eat you out.
                            The sensation itself is pleasurable, but the sheer [b]debasement[/b] of what you're doing is what drives your excitement through the roof.
                            You're getting [b]licked[/b] by a [b]feral[/b].
                            The untamed [b]beast[/b] has been completely [b]broken[/b], turned into your [b]bitch[/b].
                            And on top of all of this, you're tempting fate itself by grinding your loins against a mouth full of sharp teeth!
                        [/p]
                        [p]
                            With your eyes closed, your mind briefly wanders back to the dream that inspired you to get the basilisk to lick you.
                            Snapping back to reality, however, you realize that this situation is [b]even better[/b] than what you imagined.
                            Having the [b]pleasure[/b] of dominating a [b]lethal beast[/b],
                            one that could [b]rip you to shreds[/b] were it not more interested in [i]serving you[/i],
                            stirs your loins far more than the thought of forcing a helpless damsel to eat you out.
                            While you've only been absent-mindedly humping away at the big lizard's lapping maw for a few seconds short of a minute,
                            you can already feel your premature end rapidly approaching.
                        [/p]
                        [p]
                            You let out a lustful groan, rubbing one of your hands along the soft underside of the creature's muzzle,
                            poking at its sharp teeth with one of your fingers to remind yourself of the [b]arousing danger[/b] you're in.
                            It [b]willingly[/b] digs its tongue into your cunt, scooping ever-deeper for more of your [b]delicious[/b] arousal.
                            It must be into this every bit as much as you are.
                        [/p]
                        [p]
                            You give the beast's mouth a few more rapid grinding humps, before reaching your end.
                            Your innards spasm around the monster's forked tongue, your excitement liberally dripping from your loins,
                            a small squirt coating the basilisk's stone face with feminine fluids.
                            The subservient crocodillian, instead of pulling back, envelops your slit with its muzzle,
                            collecting as much of the bittersweet, viscous, luke-warm arousal flooding from your cunt as it can,
                            longingly swallowing every single droplet, one mouthful at a time, all the while the creature's tongue continues to work its magic.
                        [/p]
                        [p]
                            Absolutely [b]loving[/b] the feeling of the pleasant aftercare, you stay in place and simply let the big beast lick away for a while.
                            When you finally [i]do[/i] pull back, it closes its jaws once more.
                            The monster lewdly licks its lips clean, as though [i]pleased[/i] about the job it did.
                            If you didn't know any better, you'd say the feral [b]wanted[/b] this almost as much as you did.
                            You don't know whether you should feel confused, insulted or [i]flattered[/i].
                            You rise to your feet${pants || undies ? ` and redress` : ``}.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            setChoice ? subController.set(setChoice) : subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\oral\UndergroundCaveReflectionOral.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\UndergroundCaveReflectionDominance.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionDominance(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveReflectionSucceedChoice.DOMINATE;
    var reqController = ctrl.reflectSucceed;
    var subController = ctrl.reflectDominance;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new UndergroundCaveReflectionOral(ctrl);
    var secondNext = new UndergroundCaveReflectionSex(ctrl);
    var thirdNext = new UndergroundCaveReflectionGetFucked(ctrl);

    var firstChoice = UndergroundCaveReflectionDominanceChoice.ORAL;
    var secondChoice = UndergroundCaveReflectionDominanceChoice.SEX;
    var thirdChoice = UndergroundCaveReflectionDominanceChoice.SUB;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskDominated = true;
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            The basilisk is already beaten.
                            There is no point to taking its life, or turning it into an ornament for that matter.
                            It has submitted to you.
                            You could just leave, but seeing the creature on its back brings some [i]unsavory[/i] thoughts to mind.
                        [/p]
                        [p]
                            You study the creature, going around it with the mirror-shard at the ready, in case it suddenly decides to fight back.
                            As your sight wanders over the crocodillian's underside, you become aware of its hard,
                            throbbing, fully erect draconic-looking member jutting out from its nethers, standing at attention like the mast of a large and scaly ship.
                            The beast isn't even trying to hide its arousal.
                            It must be in rut or something.
                            Maybe that's why it was so fiercely defending its lair.
                            Perhaps you've inadvertently taken part of some sort of a basilisk mating ritual.
                            Maybe you've even asserted yourself as the [b]dominant[/b] partner, by petrifying the monster.
                        [/p]
                        [p]
                            Lying there, panting slightly, with its cock brimming with lust, the once-terrifying feral almost looks [i]cute[/i].
                            Like a perfectly tamed and loyal pet, waiting for a [i]treat[/i].
                            The longer you look at it, the more often your eyes find their way back to its member.
                            While you were already planning on sating your lust on the basilisk,
                            its shameless display has cemented your desire.
                            ${dick ? `Your hardens within your slit, the tapered tip peeking out already.` : `Your slit moistens in anticipation.`}
                        [/p]
                        [p]
                            The basilisk wants to mate.
                            Who are you to deny it?
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Use its muzzle</button>`;

        var tooltip = global.parser.parse(`Time to facefuck a beast's mouth filled with sharp teeth. What could go wrong?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Fuck it</button>`;

        var male = ctrl.misc.dick();
        var tooltip = global.parser.parse(`Make the basilisk your [b]bitch[/b].`);
        if (!male) {
            tooltip = global.parser.parse(`Fingerfuck the basilisk's tailhole. Make it your [b]bitch[/b].`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Ride it</button>`;

        var tooltip = global.parser.parse(`The beast's erect member allures you. You wonder how it would feel inside of you.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var UndergroundCaveReflectionDominanceChoice = {
    ORAL: 1,
    SEX: 2,
    SUB: 3
};

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\dominance\UndergroundCaveReflectionDominance.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\leave\UndergroundCaveReflectionLeaveContinue.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionLeaveContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.reflectLeaveContinue;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You run alongside the water as best you can,
                            scared the basilisk might chase after you still.
                            You nearly fall down, only barely managing to retain your balance, and decide to lower your pace to a brisk march instead.
                            A few minutes of expeditious walking later, daylight comes into sight.
                            As you expected: the water was a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You find yourself close to the road to Varanar.
                            In fact, you believe you passed this very place earlier, when heading for the hamlet.
                            You never knew there was a cave here, but then again,
                            you've never really bothered to scour the edge of the river for points of interest.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that, you managed to subdue a basilisk.
                            Or at least you [i]prevented[/i] it from [b]killing you[/b].
                        [/p]
                        [p]
                            You've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\leave\UndergroundCaveReflectionLeaveContinue.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\leave\UndergroundCaveReflectionLeave.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionLeave(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveReflectionSucceedChoice.LEAVE;
    var reqController = ctrl.reflectSucceed;
    var setChoice = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.reflectLeave;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveReflectionLeaveContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With the beast subdued, you have little reason to stick around.
                            You could probably [i]kill it[/i] or something,
                            but killing a defenceless, surrendering animal seems pointlessly cruel.
                            And you really don't want to tempt fate by trying anything [i]dumb[/i].
                        [/p]
                        [p]
                            You slowly back away from the basilisk,
                            holding the mirror shard up like a preacher would a crucifix in front of a vampire.
                            It doesn't move, instead staying in place, lying on its back, its tongue partially lolling out of its gaping maw.
                            The monster looks like it's playing dead, which is almost funny, in a way.
                        [/p]
                        [p]
                            You put some distance in between yourself and the crocodillian, backing off until its green-glowing eyes start to fade into the dark.
                            Despite no longer being able to see them, you can still picture those two intriguing emerald orbs in your head.
                            While the big reptile is no longer a threat to you, you don't think you'll ever be able to forget its terrifying, soul-gazing glare.
                            With the monster out of sight, you discard the glass shard, turn tail and flee.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            setChoice ? subController.set(setChoice) : subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }


        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\leave\UndergroundCaveReflectionLeave.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\petrify\UndergroundCaveReflectionPetrifyContinue.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionPetrifyContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.reflectPetrifyContinue;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You did it!
                            You killed a deadly monster, all by yourself!
                            Just like a knight would do!
                            Somehow, you don't feel very [i]knightly[/i], though.
                            You should be proud of yourself, but for some reason, you aren't.
                            Maybe your actions weren't as [i]chivalrous[/i] as you would have liked,
                            but it's the outcome that matters, and not how you achieved it, right?
                            The ends justify the means.
                        [/p]
                        [p]
                            But then again, the basilisk had surrendered.
                            It was no longer intent on hurting you.
                            It even rolled belly-up to display its submissiveness.
                            And yet you still persisted in petrifying it to death.
                            Now you're confused.
                            Did you do the right thing?
                        [/p]
                        [p]
                            You honestly don't know anymore.
                            Does it even matter?
                            What's done is done.
                            If you hadn't killed the creature, some pompous knight in shining armor
                            -- with nothing left to prove, neither to himself, nor to the world --
                            would have strolled into this cave to add another tally to his already impressive list of slain monsters.
                            For someone properly armed and armored, this would have been a [b]chore[/b], a menial task barely even worthy of mention.
                            But for you, this is a feat you'll remember until the day you die.
                        [/p]
                        [p]
                            Your conviction bolstered, you wait for a few more minutes to make sure it's really [b]dead[/b],
                            before approaching the giant lizard once again.
                            Kneeling besides it, you figure it's time to take a trophy.
                            Something to remember this day by.
                            Something you can show to people, so they'll believe you when you tell them you singlehandedly killed a basilisk.
                            But what?
                        [/p]
                        [p]
                            While pondering, you run one of your hands over the deceased basilisk's exposed belly, now a lot [i]harder[/i] than what it used to be.
                            You briefly consider taking the entire [i]statue[/i] with you, but it's [b]far[/b] too heavy.
                            As your sight wanders over the monster's underbelly, you notice that its once-supple scales are not the only thing [i]hardened[/i].
                            The beast seems to have been [i]fully erect[/i] at the time of its death.
                            Its girthy, ridged and draconic-looking member stands fully erect, every bit as petrified as the rest of its body.
                            It must've been in rut or something.
                            Maybe that's why it was so fiercely defending its lair.
                        [/p]
                        [p]
                            You briefly consider [i]snapping off[/i] the dead monster's stone-turned dick,
                            but you figure it would be [b]very[/b] awkward to use [i]that[/i] as a token of your kill, let alone trying to [i]sell it[/i].
                        [/p]
                        [p]
                            Figuring there isn't anything else worthwhile to take, you decide it's time to find a way out of the cave.
                            Following the river-branch should lead you to some kind of exit.
                            Without the threat of the basilisk around, you don't even need the glass shard anymore.
                            You discard it, before leisurely strolling alongside the water, until daylight comes into sight.
                            As you expected: the water was a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that, you managed to kill a basilisk by making it petrify itself!
                        [/p]
                        [p]
                            You've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\petrify\UndergroundCaveReflectionPetrifyContinue.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\petrify\UndergroundCaveReflectionPetrify.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionPetrify(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveReflectionSucceedChoice.PETRIFY;
    var reqController = ctrl.reflectSucceed;
    var setChoice = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.reflectPetrify;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveReflectionPetrifyContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Seeing an opportunity for greatness, you press your advantage.
                            You hold the mirror shard directly in front of the basilisk's eyes, [b]forcing[/b] it to stare at its own enchanting gaze.
                            You could probably just [b]rake open[/b] its throat instead, but you kind of [i]like[/i] turning its own weapon against the beast.
                            The creature tries to look away, but after a few seconds, its own alluring reflection captures its sight.
                            It lets out another whine, audibly scared.
                            This is clearly not what it expected – or hoped – to happen.
                        [/p]
                        [p]
                            The big reptile is too weak to resist now.
                            Too far gone already.
                            Its breathing grows slower, weaker.
                            Its flicking tongue grows limp.
                            Its bestial snout locks in place.
                            Its rocky scales lose whatever color they had, and the muscles underneath them stop moving.
                            Finally, as its breathing comes to a complete halt, its very eyes turn to stone, and then you know it is over.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    You have killed the basilisk.
                                [/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            setChoice ? subController.set(setChoice) : subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\petrify\UndergroundCaveReflectionPetrify.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\strike\UndergroundCaveReflectionStrikeContinue.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionStrikeContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.reflectStrikeContinue;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You did it!
                            You killed a deadly monster, all by yourself!
                            Just like a knight would do!
                            Somehow, you don't feel very [i]knightly[/i], though.
                            You should be proud of yourself, but for some reason, you aren't.
                            Maybe your actions weren't as [i]chivalrous[/i] as you would have liked,
                            but it's the outcome that matters, and not how you achieved it, right?
                            The ends justify the means.
                        [/p]
                        [p]
                            But then again, the basilisk had surrendered.
                            It was no longer intent on hurting you.
                            It even rolled belly-up to display its submissiveness.
                            Now you're confused.
                            Did you do the right thing?
                        [/p]
                        [p]
                            You honestly don't know anymore.
                            Does it even matter?
                            What's done is done.
                            If you hadn't killed the creature, some pompous knight in shining armor
                            -- with nothing left to prove, neither to himself, nor to the world --
                            would have strolled into this cave to add another tally to his already impressive list of slain monsters.
                            For someone properly armed and armored, this would have been a [b]chore[/b], a menial task barely even worthy of mention.
                            But for you, this is a feat you'll remember until the day you die.
                        [/p]
                        [p]
                            Your conviction bolstered, you wait for a few more minutes to make sure it's really [b]dead[/b],
                            before approaching the giant lizard once again.
                            Kneeling besides it, you figure it's time to take a trophy.
                            Something to remember this day by.
                            Something you can show to people, so they'll believe you when you tell them you singlehandedly killed a basilisk.
                            But what?
                        [/p]
                        [p]
                            While pondering, you run one of your hands over the deceased basilisk's back.
                            Some of its scales seem to be rather [i]loose[/i].
                            You take hold of one with both of your hands, and with surprising ease, you manage to [b]tear[/b] the scale from the monster's backside.
                            Like plucking a chicken.
                            You repeat the process a couple of times,
                            until you have collected a suitable handful of ${global.kobold.inventory.createItemWithTooltip(new BasiliskScales())}
                            to take with as your prize for ridding the world of this horrible monstrosity.
                        [/p]
                        [p]
                            Your trophies collected, you decide it's time to find a way out of the cave.
                            Following the river-branch should lead you to some kind of exit.
                            Without the threat of the basilisk around, you don't even need the glass shard anymore.
                            You discard it, before leisurely strolling alongside the water, until daylight comes into sight.
                            As you expected: the water was a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that, you managed to kill a basilisk.
                            It might not have been the most [i]honorable[/i] way of slaying the beast, but it was slain nonetheless.
                        [/p]
                        [p]
                            Besides the ${global.kobold.inventory.createItemWithTooltip(new BasiliskScales())} you [i]harvested[/i],
                            you've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\strike\UndergroundCaveReflectionStrikeContinue.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\strike\UndergroundCaveReflectionStrike.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionStrike(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveReflectionSucceedChoice.STRIKE;
    var reqController = ctrl.reflectSucceed;
    var setChoice = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.reflectFail;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveReflectionStrikeContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Seeing an opportunity for greatness, you decide to act on it.
                            While still brandishing the mirror shard with one hand, you move the other onto the monster's throat.
                            Expecting you to show it some mercy after it surrendered,
                            the basilisk's knows not what to do as you instead sink your sharp claws into its tender throat,
                            and rake open its windpipe in a single vicious [b]tear[/b].
                        [/p]
                        [p]
                            The beast's eyes shoot open wide.
                            It flips back upright, and limply thrashes its limbs around, hoping to mangle you, to take you down along with it,
                            but you manage to avoid most of its slow-paced rage by taking a quick step back.
                            While it continues to blindly rampage about, despite you having long cleared its imminent vicinity,
                            the basilisk's semi-petrified struggles soon die down.
                            Running out of air, its body slumps lifelessly down on the ground, in a pool of its own blood.
                            After one last, bloody gurgle, the big crocodillian closes its eyes, and drifts away into the afterlife.
                        [/p]
                        [p]
                            [c]
                                [b]You have killed the basilisk![/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            setChoice ? subController.set(setChoice) : subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\strike\UndergroundCaveReflectionStrike.js

// scenes\HamletDay2\UndergroundCave\mirror\succeed\UndergroundCaveReflectionSucceed.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.reflectSucceed;
    var changeController = ctrl.reflectSucceed;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";
    var fourthId = arguments.callee.name + "ButtonFour";

    var firstNext = new UndergroundCaveReflectionStrike(ctrl);
    var secondNext = new UndergroundCaveReflectionLeave(ctrl);
    var thirdNext = new UndergroundCaveReflectionPetrify(ctrl);
    var fourthNext = new UndergroundCaveReflectionDominance(ctrl);

    var firstChoice = UndergroundCaveReflectionSucceedChoice.STRIKE;
    var secondChoice = UndergroundCaveReflectionSucceedChoice.LEAVE;
    var thirdChoice = UndergroundCaveReflectionSucceedChoice.PETRIFY;
    var fourthChoice = UndergroundCaveReflectionSucceedChoice.DOMINATE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskMirror = true;
        global.kobold.history.shared.well.basiliskMet = true;
        var toReturn = `[p]
                            Your mind races, even as your body freezes.
                            How are you supposed to respond?!
                            You can't look away.
                            If it's a predator, it might take that as a sign of weakness.
                            Still, you're too [b]afraid[/b] to boldly stare back.
                            Those eyes are pretty scary.
                            You're left in an impasse, your frailty showing regardless of whether or not you look at the eyes.
                        [/p]
                        [p]
                            Then a thought flashes through your head.
                            An idea.
                            You bought something, something that can help!
                            Something that'll allow you to combine the best of both worlds!
                            You're talking about the ${global.kobold.inventory.createItemWithTooltip(new SmallMirror())} in your backpack.
                            With it, you can give the monster something to stare at, without having to look at it yourself.
                            You're saved!
                        [/p]
                        [p]
                            A lot of animals get spooked by their own reflections.
                            Seki once told you that, anyway.
                            He mentioned [i]why[/i], but you can't really recall that part of his [i]lecture[/i].
                            Would the monster think you're its peer instead of prey, and thus leave you alone?
                            Or perhaps it'd find out that its got a real nasty look on its face.
                            Either way, the mirror is your best bet.
                        [/p]
                        [p]
                            You keep your vision focused on the cavern wall, right next to the set of eyes staring at you,
                            trying to make it look like you [i]aren't[/i] afraid, without challenging the monster too much.
                            At the same time, you begin to look for the ${global.kobold.inventory.createItemWithTooltip(new SmallMirror())} on your person.
                        [/p]
                        [p]
                            Suddenly, your whole body jolts as something sharp pierces your fingertip!
                            A pained grunt escapes your muzzle.
                            Did you accidentally bump against an edgy stone?
                            No, it wasn't a rock.
                            Thanks to your haphazard adventure in the well,
                            your ${global.kobold.inventory.createItemWithTooltip(new SmallMirror())} seems to have broken into a million little pieces.
                            Must have gotten hit by some of the debris when the entire place came down on top of your head.
                            Your heart sinks.
                            Terror settles in, your best plan shattered.
                        [/p]
                        [p]
                            The creature in the dark scuttles closer, redoubling its stare.
                            You can remotely make out the outline of something [b]big[/b] and [b]scaly[/b].
                            You can feel it growing [i]impatient[/i], like it's almost done [i]playing[/i] with you.
                            No, no, no!
                            You have no choice [b]but[/b] to gaze directly into its eyes, putting on your bravest face to hopefully keep the monster at bay.
                        [/p]
                        [p]
                            Fueled by desperation, you scoop the broken pieces to your front.
                            They loudly [b]scrape[/b] across the stone floor.
                            Once repositioned, your hands keep searching through the broken mirror's shards.
                            You get cut several times more, but you barely even notice, your mind completely [b]clogged[/b] with fright.
                            Your fingers clench around several fragments, trying, but failing to find one big enough to [i]wield[/i].
                            All the while, the creature's stare [b]burns[/b] itself into your mind.
                        [/p]
                        [p]
                            You stubbornly cling to your plan.
                            Both your hands continue to fumble about the shards.
                            If you can find a large enough mirror-piece, you'll be [b]saved[/b]!
                            But it's no use.
                            All of the fragments are far too tiny, completely useless.
                            You can feel your fingertips growing slightly more [i]stiff[/i] and [i]rigid[/i],
                            a fact you attribute to the multitude of small cuts you've sustained.
                        [/p]
                        [p]
                            Right as you were about to give up hope, due to a combination of sheer [b]luck[/b] and skillfull searching,
                            your digits stroke across a medium-sized smooth surface.
                            An intact piece!
                            Despite your hand hurting, you grip on to the intact mirror-piece,
                            the edges cutting into your scales as you arduously lift it up for the creature to see.
                        [/p]
                        ${global.ui.condSucc(`Agility and luck greater than 5`)}
                        [p]
                            The dim green glow radiated by the monster's eyes is reflected back onto itself, giving you a first look at what it [i]actually[/i] is.
                            At first, you can't even [i]identify[/i] it.
                            Before you stands a feral, quadrupedal lizard with large, protruding, rock-like scales and a sizeable maw filled with sharp, pointy teeth. 
                            A crocodilian with a stony hide.
                            You think Seki once told you a story about a beast like this.
                            Could it be... a basilisk?!
                        [/p]
                        [p]
                            Startled, you rise to your feet.
                            If Seki's tales are to be believed,
                            this creature has the ability to petrify anyone who dares to look at it!
                            That explains the sudden [i]stiffness[/i] you were feeling just now.
                            No wonder it has such enchanting eyes.
                        [/p]
                        [p]
                            The basilisk appears to be focusing all of its attention on the shard in your hands.
                            It seems perplexed by its own, reflected gaze, undoubtedly in the process of turning to stone itself.
                            You can't help but smirk a little.
                            You unintentionally outsmarted it even more than you thought you were!
                            You turned the beast's petrification against itself!
                            You feel like one of those heroes people write songs about,
                            matching horrifying monsters' evil tricks with your quick wit and bravery!
                        [/p]
                        [p]
                            Emboldened by the thought of being seen as a hero, you take a few steps forwards,
                            holding the shard in front of the feral the entire time, until you're naught but a few feet away from it.
                            This could still end horribly for you.
                            If you drop your guard for even a moment, or if the creature snaps from its enthrallment,
                            it could easily pounce on you and tear you to shreds.
                            Perhaps you advanced a little too [i]rashly[/i], but there's no backing out now.
                            You've committed to this [i]strategy[/i], and now you need to see it through.
                        [/p]
                        [p]
                            The crocodilian finally notices your approach.
                            It tries to attack you, visibly surprised that it can hardly move any of its own limbs.
                            The basilisk hisses loudly, but about halfway through, the hiss turns into a [i]whine[/i],
                            as the beast realizes that it can't fight like this.
                            Its harsh glare softens.
                            It's almost as though the monster looks kind of [i]sad[/i].
                            Does that mean you've [i]won[/i]?
                        [/p]
                        [p]
                            While you're still cautious and [i]on edge[/i],
                            the feral slowly moves its half-petrified limbs in an unexpected way.
                            Clumsily, and with great effort, it rolls over,
                            submissively exposing its vulnerable underside to you, whilst letting out a pitiful whimper.
                            There's no doubt about it, you've bested the basilisk into submission.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled} style="width:200px;">Strike throat</button>`;

        var tooltip = global.parser.parse(`Finish the beast.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:200px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled} style="width:200px;">Run</button>`;

        var tooltip = global.parser.parse(`Leave while you still can.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:200px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled} style="width:200px;">Petrify it</button>`;

        var tooltip = global.parser.parse(`It can't resist forever.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:200px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFourthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === fourthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fourthId}" ${disabled} style="width:200px;">Assert dominance</button>`;

        var tooltip = global.parser.parse(`The beast has submitted to you. Time to claim what is rightfully yours.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:200px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += `<div class="choicerow" style="margin-bottom: 10px;">`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        choice += `<div class="choicerow">`;
        choice += buildThirdChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFourthChoice();
        choice += `</div>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId), $("#" + fourthId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFourthListener() {
        $("#" + fourthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fourthChoice);
            loadNextPart(fourthNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.imageBackground.hide();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        part += fourthNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();
        fourthNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
        registerFourthListener();
    };
}

var UndergroundCaveReflectionSucceedChoice = {
    STRIKE: 1,
    LEAVE: 2,
    PETRIFY: 3,
    DOMINATE: 4
};

// END scenes\HamletDay2\UndergroundCave\mirror\succeed\UndergroundCaveReflectionSucceed.js

// scenes\HamletDay2\UndergroundCave\mirror\fail\UndergroundCaveReflectionFail.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflectionFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.stare;
    var changeController = ctrl.reflectFail;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveRelent(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskMet = true;
        global.kobold.history.shared.well.basiliskPetrified = true;
        var toReturn = `[p]
                            Your mind races, even as your body freezes.
                            How are you supposed to respond?!
                            You can't look away.
                            If it's a predator, it might take that as a sign of weakness.
                            Still, you're too [b]afraid[/b] to boldly stare back.
                            Those eyes are pretty scary.
                            You're left in an impasse, your frailty showing regardless of whether or not you look at the eyes.
                        [/p]
                        [p]
                            Then a thought flashes through your head.
                            An idea.
                            You bought something, something that can help!
                            Something that'll allow you to combine the best of both worlds!
                            You're talking about the ${global.kobold.inventory.createItemWithTooltip(new SmallMirror())} in your backpack.
                            With it, you can give the monster something to stare at, without having to look at it yourself.
                            You're saved!
                        [/p]
                        [p]
                            A lot of animals get spooked by their own reflections.
                            Seki once told you that, anyway.
                            He mentioned [i]why[/i], but you can't really recall that part of his [i]lecture[/i].
                            Would the monster think you're its peer instead of prey, and thus leave you alone?
                            Or perhaps it'd find out that its got a real nasty look on its face.
                            Either way, the mirror is your best bet.
                        [/p]
                        [p]
                            You keep your vision focused on the cavern wall, right next to the set of eyes staring at you,
                            trying to make it look like you [i]aren't[/i] afraid, without challenging the monster too much.
                            At the same time, you begin to look for the ${global.kobold.inventory.createItemWithTooltip(new SmallMirror())} on your person.
                        [/p]
                        [p]
                            Suddenly, your whole body jolts as something sharp pierces your fingertip!
                            A pained grunt escapes your muzzle.
                            Did you accidentally bump against an edgy stone?
                            No, it wasn't a rock.
                            Thanks to your haphazard adventure in the well,
                            your ${global.kobold.inventory.createItemWithTooltip(new SmallMirror())} seems to have broken into a million little pieces.
                            Must have gotten hit by some of the debris when the entire place came down on top of your head.
                            Your heart sinks.
                            Terror settles in, your best plan shattered.
                        [/p]
                        [p]
                            The creature in the dark scuttles closer, redoubling its stare.
                            You can remotely make out the outline of something [b]big[/b] and [b]scaly[/b].
                            You can feel it growing [i]impatient[/i], like it's almost done [i]playing[/i] with you.
                            No, no, no!
                            You have no choice [b]but[/b] to gaze directly into its eyes, putting on your bravest face to hopefully keep the monster at bay.
                        [/p]
                        [p]
                            Fueled by desperation, you scoop the broken pieces to your front.
                            They loudly [b]scrape[/b] across the stone floor.
                            Once repositioned, your hands keep searching through the broken mirror's shards.
                            You get cut several times more, but you barely even notice, your mind completely [b]clogged[/b] with fright.
                            Your fingers clench around several fragments, trying, but failing to find one big enough to [i]wield[/i].
                            All the while, the creature's stare [b]burns[/b] itself into your mind.
                        [/p]
                        [p]
                            You stubbornly cling to your plan.
                            Both your hands continue to fumble about the shards.
                            If you can find a large enough mirror-piece, you'll be [b]saved[/b]!
                            But it's no use.
                            All of the fragments are far too tiny, completely useless.
                            You can feel your fingertips growing slightly more [i]stiff[/i] and [i]rigid[/i],
                            a fact you attribute to the multitude of small cuts you've sustained.
                            You doubt you could even properly [b]hold[/b] a shard at this point, even [i]if[/i] you managed to find one of sufficient size.
                        [/p]
                        ${global.ui.condFail(ctrl.misc.agility() < 6 ? `Agility below 6` : `Luck below 6`)}
                        [p]
                            Then you realize it's not [i]just[/i] your hands becoming numb.
                            [b]All[/b] of your extremities seem to be [i]unresponsive[/i].
                            As though half your body is [i]asleep[/i]!
                            Panic wells up inside you.
                            You try to rise to your feet, to flee as you should have done in the first place, but your legs refuse to function.
                            Your arms, too, appear to be frozen in place.
                        [/p]
                        [p]
                            You break eye contact with the monster.
                            You can't even turn your head away, your neck every bit as [i]stiff[/i] as your arms and legs.
                            Casting your gaze downwards, you notice your arms are [i]discolored[/i].
                            Your scales have become a dull grey instead of the usual green.
                        [/p]
                        [p]
                            Are you asleep?
                            Dreaming, perhaps?
                            Is this all a nightmare?
                            Did you never make it out of the well?
                            You try to open your mouth to scream, but even your muzzle isn't budging.
                            Staring at your own snout, you finally figure out what happened.
                            Your scales have turned to stone.
                            You've been petrified!
                        [/p]
                        [p]
                            You try every muscle in your body, to see which parts you can still move.
                            Your tongue is still in working order, although your mouth won't open.
                            Your eyes can dart around all they please, but your head itself is unmoving.
                            You retain [i]some[/i] form of control over your tail, the largest of your extremities still in some manner of working order.
                            Fingers are a no-go.
                            Entirely petrified.
                            Your toes, finally.
                            You can wriggle those around slightly, despite both your legs and your feet being locked in place.
                        [/p]
                        [p]
                            While most of your smaller organs and limbs are still fully functional,
                            your body as a whole is completely immobilized.
                            Shocked by this revelation, you barely even notice the creature you were so afraid of leaving its hole,
                            stepping out into the cave proper.
                            Your cross-eyed stare at your own muzzle slowly fading, you become aware of the large,
                            quadrupedal, rock-hided crocodillian standing right in front of you.
                            Your heart nearly skips a beat as you realize what you're dealing with here.
                            A [b]basilisk[/b]!
                        [/p]
                        [p]
                            Seki once told you stories about creatures that could petrify
                            anyone or anything unlucky enough to stare at them for prolonged periods of time,
                            but you thought he was just making things up!
                            You should have listened more carefully to the old bird!
                            On the bright side, while your fingers aren't fully petrified,
                            the small cuts you sustained while looking through the mirror shards have stopped bleeding,
                            probably sealed shut by your hardening scales.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(UndergroundCaveStareChoice.RELENT);
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.imageBackground.hide();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\fail\UndergroundCaveReflectionFail.js

// scenes\HamletDay2\UndergroundCave\mirror\UndergroundCaveReflection.js

/**
 * Reflect section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveReflection(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveSurfaceChoice.MIRROR;
    var reqController = ctrl.surface;

    var toFail = new UndergroundCaveReflectionFail(ctrl);
    var toSuccess = new UndergroundCaveReflectionSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.agility() > 5 && ctrl.misc.luck() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\mirror\UndergroundCaveReflection.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\leave\UndergroundCaveDominanceLeave.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveDominanceLeave(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.dominanceLeave;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var baseMale = ctrl.misc.baseGender() === GenderEnum.MALE;

        var choice = ctrl.dominance.get();
        var dominanceResult = "covered in your piss";
        if (choice === UndergroundCaveDominanceChoice.ORAL) {
            if (dick) {
                dominanceResult = "with your cum in its mouth";
            } else {
                dominanceResult = "to clean its muzzle";
            }
        } else if (choice === UndergroundCaveDominanceChoice.SEX) {
            if (dick) {
                dominanceResult = "leaking your spunk from its rear";
            } else {
                dominanceResult = "to bask in post-orgasmic bliss";
            }
        }

        var toReturn = `[p]
                            With the basilisk properly [b]dominated[/b],
                            and yourself asserted as the [b]alpha ${baseMale ? `male` : `female`}[/b],
                            you decide it's time to find a way out of the cave.
                            Leaving the feral ${dominanceResult}, you begin to follow the river-branch, hoping it'll lead you to some kind of exit.
                            You leisurely stroll alongside the water, until daylight comes into sight.
                            As you expected: the water [b]was[/b] a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that, you managed to wrestle a basilisk into submission.
                            Toying with it afterwards might not have been absolutely [i]necessary[/i], but you feel like it's now learned its lesson.
                            The basilisk should be happy that you decided to [b]let it live[/b] instead of finishing it off for accosting you.
                        [/p]
                        [p]
                            You've gained a few personal memories that you'll undoubtedly find yourself masturbating to later,
                            along with some awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\leave\UndergroundCaveDominanceLeave.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\oral\UndergroundCaveOralSex.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveOralSex(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.oral;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveDominanceLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskOral = true;
        var dick = ctrl.misc.dick();
        var pants = ctrl.misc.pants();
        var undies = ctrl.misc.underwear();

        var lowerPantsAndUndies = `and already naked`;

        if (pants) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (undies) {
                lowerPantsAndUndies += ` and your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
            }
        } else if (undies) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
        }
        var toReturn = `[p]
                            At first, the basilisk seemed like a determined, lethal [b]killing machine[/b],
                            but now that you've wrestled the fight out of it, it seems to have become a lot more [b]passive[/b].
                            It knows it did something bad, and it realizes that you're about to punish it.
                            It's as though the monster was hiding its submissive nature under a veil of hostility.
                            Now the creature is more of a [b]big wimp[/b] than a menacing beast,
                            and you're the [b]bully[/b] about to [b]push it around[/b] in whatever way you please.
                        [/p]
                        [p]
                            Thinking of this situation as a [i]reversal of power[/i] awakens lustful feelings deep inside of you.
                            You've had this recurring [i]wet dream[/i] a few times, where kobolds would stage an uprising,
                            and you'd find yourself in front of the [b]juiciest[/b] maiden in all of Varanar,
                            with ten of your kind holding her struggling, naked form down for you to [b]ravage[/b].
                            This is [b]just[/b] like that dream, only instead of a maiden, you have a [b]basilisk[/b] to quench your lust upon,
                            the uprising was more of a [b]wrestling match[/b], and there aren't really any kobolds around to hold the monster still.
                        [/p]
                        [p]
                            Okay, in retrospect, this entire ordeal might not be very much like your dream after all.
                            Still, you were reminded of it, and that's enough to get your
                            ${dick ? `member to slowly start to harden within your cockslit` : `slit to slightly moisten`}.
                            You begin to imagine all the things you could do to the creature, now that you've subdued it.
                            While not so similar to your dream, having your way with a large, ferocious animal [b]does[/b] pique your interest.
                        [/p]`;

        if (dick) {
            var virgin = ctrl.misc.dickVirgin();
            global.kobold.body.dickVirgin = false;
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at molesting a basilisk, you decide to act on your impulses.
                            You begin to stroke the emerging tapered tip of your member, which rapidly grows erect, fueled by the dirty thoughts popping up in your mind.
                            Its muzzle.
                            You want to fuck its muzzle, filled with sharp, pointy teeth.
                            You want to walk on the wild side, to tempt fate itself,
                            to see whether or not the beast still has enough [b]fight[/b] in it to resist, to [b]bite down[/b].
                        [/p]
                        [p]
                            You bet it won't.
                            You bet you've [b]broken[/b] it.
                            You bet it won't even [b]try[/b] to stop you from fucking its maw.
                            You've shown it that it's [b]not[/b] the boss around here.
                            You've overpowered it, despite all odds.
                            It won't resist.
                            Not after the [b]ass-whooping[/b] you gave it.
                            Not after it attacked you, so [b]certain[/b] that you'd be an easy prey.
                        [/p]
                        [p]
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, ${lowerPantsAndUndies}, you approach the basilisk's head, dropping to your knees directly in front of it,
                            your erect member mere inches away from the feral's snout.
                            Its maw is still open, the monster panting heavily, not yet fully recovered from the intense tussle.
                            A perfect opportunity.
                        [/p]
                        [p]
                            You line up the tip of your member with the beast's mouth, and with a single, smooth roll of your hips,
                            you slide your length inside of its muzzle, the bottom of your shaft pleasantly rubbing alongside the basilisk's slender, forked tongue.
                            While the creature's exterior is hard, tough and cold, the insides of its mouth are the exact opposite.
                            Warm, fleshy, [b]wet[/b] with saliva, almost what you imagine a human's gums must feel like.
                            A pleasant surprise, and not an unwelcome one.
                        [/p]
                        [p]
                            Your dick being rammed into its gaping maw seems to rouse the lizard from its post-battle drowse.
                            Its eyes shoot wide open.
                            The taste of kobold cock spreading throughout its mouth, it quickly realizes what's going on.
                            The monster doesn't like it one bit.
                            It attempts to chomp down on your member, but your [b]girth[/b] is simply too large for it to be able to close its jaws properly!
                            Instead of biting you, the basilisk ends up pleasantly [b]nibbling[/b] on your shaft, giving the [i]illusion[/i] that it knows its place.
                        [/p]
                        ${global.ui.condSucc(`Big dick`)}
                        [p]
                            Luckily, the crocodillian is far too worn out to put up any other kind of resistance.
                            Lost in pleasure, you close your eyes and lean back ever so slightly,
                            while starting to thrust in and out of the monster's muzzle.
                            The sensation isn't [b]too[/b] pleasurable, but the sheer [b]debasement[/b] of what you're doing excites you to no end.
                            You're [b]fucking[/b] a feral.
                            The [b]beast[/b] that tried to murder you has been turned into your [b]bitch[/b].
                            You get to be the person on the [b]giving[/b] end of abuse, for once.
                            And on top of all of this, you're tempting fate itself by ramming your dick into a mouth full of sharp teeth!
                        [/p]
                        [p]
                            With your eyes closed, your mind briefly wanders back to the dream that inspired you to facefuck the basilisk.
                            Snapping back to reality, however, you realize that this situation is [b]even better[/b] than what you imagined.
                            Having the [b]pleasure[/b] of dominating a [b]lethal beast[/b],
                            one that could [b]rip you to shreds[/b] were it not too exhausted,
                            stirs your loins far more than the thought of raping a helpless damsel.
                            While you've only been absent-mindedly humping away at the big lizard's maw for a few seconds short of a minute,
                            you can already feel your premature end rapidly approaching.
                        [/p]
                        [p]
                            You let out a lustful groan, rubbing one of your hands along the soft underside of the creature's muzzle,
                            poking at its sharp teeth with one of your fingers to remind yourself of the [b]arousing danger[/b] you're in.
                            The basilisk shows no signs of biting, though.
                            While, in reality, its jaws are spread too wide for it chomp down,
                            you mistake its inability to bite you for [b]willingness[/b].
                            And the thought of having reduced the once-proud crocodillian to an [b]eager[/b] cocksucker,
                            [b]that[/b] sends you over the edge.
                        [/p]
                        [p]
                            You give the beast's mouth a few more rapid thrusts, before reaching your end.
                            Your member [b]pulses[/b] and [b]spasms[/b], spunk pumping from deep within your internal balls,
                            through your shaft, squirting directly atop the monster's forked tongue.
                            Clearly reluctant, or perhaps even outright [b]unwilling[/b], it tries to pull its head back,
                            but using both of your hands you firmly keep it in place for the entire duration of your orgasm,
                            forcing the feral to take your load in its muzzle.
                        [/p]
                        [p]
                            Your lust sated, you pull your cock from the big beast's maw.
                            It doesn't even bother trying to close its mouth, simply letting your jizz leak from its jaws,
                            the thick, creamy goop spilling out from in between the gaps in its teeth.
                            It looks almost kind of [b]sad[/b], but you both know that it got what it deserved.
                            It shouldn't have tried to murder you.
                            You rise to your feet${pants || undies ? ` and redress` : ``}.
                        [/p]`;
            if (virgin) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        } else {
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at molesting a basilisk, you decide to act on your impulses.
                            You begin to touch your own slit, which rapidly dampens, fueled by the dirty thoughts popping up in your mind.
                            Its muzzle.
                            You want to ride its muzzle, filled with sharp, pointy teeth.
                            You want to walk on the wild side, to tempt fate itself,
                            to see whether or not the beast still has enough [b]fight[/b] in it to resist, to [b]bite[/b].
                        [/p]
                        [p]
                            You bet it won't.
                            You bet you've [b]broken[/b] it.
                            You bet it won't even [b]try[/b] to stop you from rubbing your cunt all over its maw.
                            You've shown it that it's [b]not[/b] the boss around here.
                            You've overpowered it, despite all odds.
                            It won't resist.
                            Not after the [b]ass-whooping[/b] you gave it.
                            Not after it attacked you, so [b]certain[/b] that you'd be an easy prey.
                        [/p]
                        [p]
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, ${lowerPantsAndUndies}, you approach the basilisk's head, dropping to your knees directly in front of it,
                            your wet slit mere inches away from the feral's snout.
                            Its maw is still open, the monster panting heavily, not yet fully recovered from the intense tussle.
                            A perfect opportunity.
                        [/p]
                        [p]
                            You grab the beast's head, and tilt it up ever so slightly, allowing you to plant your eager twat directly on top of its snout.
                            The scent of a female lizard seems to rouse the creature from its post-battle drowse.
                            Its nostrils flare, and its slender, forked tongue slips out to [b]taste[/b] your arousal,
                            lightly flicking along your most sensitive scales, before retracting into its massive, toothy maw once more.
                        [/p]
                        [p]
                            The monster's eyes shoot wide open, staring up at you.
                            The feminine taste of an aroused kobold girl spreading throughout its mouth, it quickly realizes what's going on.
                            Furthermore, it seems to understand what is asked of it.
                            The basilisk's tongue flicks out again, pressing deeper into your folds this time, even slightly parting them,
                            collecting as much of your sweet feminine excitement as it possible can, before retracting the flexible organ once more.
                        [/p]
                        [p]
                            You let out a pleasured sigh as the basilisk begins to outright [b]lap[/b] at your slit,
                            licking you much like a common bear would steal honey from a beekeeping friar's storehouse:
                            greedily and speedily.
                            Despite the creature's rock-like appearance, its tongue feels [b]warm[/b] and [b]wet[/b],
                            and the forked split at the end is just exotic enough to get your filthy mind [b]racing[/b].
                        [/p]
                        [p]
                            Lost in pleasure, you close your eyes and lean back ever so slightly,
                            slowly rubbing your crotch back and forth over the monster's muzzle while it continues to eat you out.
                            The sensation itself is pleasurable, but the sheer [b]debasement[/b] of what you're doing is what drives your excitement through the roof.
                            You're getting [b]licked[/b] by a [b]feral[/b].
                            The [b]beast[/b] that tried to murder you has been turned into your [b]bitch[/b].
                            You get to be the person [b]getting licked[/b], instead of the submissive lizard being forced to service someone else.
                            And on top of all of this, you're tempting fate itself by grinding your loins against a mouth full of sharp teeth!
                        [/p]
                        [p]
                            With your eyes closed, your mind briefly wanders back to the dream that inspired you to get the basilisk to lick you.
                            Snapping back to reality, however, you realize that this situation is [b]even better[/b] than what you imagined.
                            Having the [b]pleasure[/b] of dominating a [b]lethal beast[/b],
                            one that could [b]rip you to shreds[/b] were it not too exhausted to fight,
                            stirs your loins far more than the thought of forcing a helpless damsel to eat you out.
                            While you've only been absent-mindedly humping away at the big lizard's lapping maw for a few seconds short of a minute,
                            you can already feel your premature end rapidly approaching.
                        [/p]
                        [p]
                            You let out a lustful groan, rubbing one of your hands along the soft underside of the creature's muzzle,
                            poking at its sharp teeth with one of your fingers to remind yourself of the [b]arousing danger[/b] you're in.
                            The basilisk shows no signs of biting, though.
                            It [b]willingly[/b] digs its tongue into your cunt, scooping ever-deeper for more of your [b]delicious[/b] arousal.
                            It must be into this every bit as much as you are.
                        [/p]
                        [p]
                            You give the beast's mouth a few more rapid grinding humps, before reaching your end.
                            Your innards spasm around the monster's forked tongue, your excitement liberally dripping from your loins,
                            a small squirt coating the basilisk's stone face with feminine fluids.
                            Surprised by the sudden torrent, the crocodillian tries to pull its head back,
                            but using both of your hands you firmly keep it in place for the entire duration of your orgasm.
                        [/p]
                        [p]
                            Getting over the initial [i]scare[/i], the feral begins to eagerly lap up as much of the viscous,
                            luke-warm liquid flooding from your slit as it can.
                            Absolutely [b]loving[/b] the feeling of the pleasant aftercare, you stay in place and simply let the big beast lick away for a while.
                            It lets out a loud whine when you finally [i]do[/i] pull back from its maw.
                            It looks almost kind of [b]sad[/b], but this [i]was[/i] supposed to be a [b]punishment[/b] instead of something you could both enjoy.
                            Oh well.
                            You rise to your feet${pants || undies ? ` and redress` : ``}.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\oral\UndergroundCaveOralSex.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\oral\UndergroundCaveBite.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveBite(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;

    var continueId = arguments.callee.name + "ButtonOne";

    var idsToRemove = [
        "UndergroundCaveBite",
        "UndergroundCaveOral",
        "UndergroundCaveDominance",
        "UndergroundCaveWrestlingWin",
        "UndergroundCaveWrestling",
        "UndergroundCaveOffence",
        "UndergroundCaveFight",
        "UndergroundCaveResistSucceed",
        "UndergroundCaveResist",
        "UndergroundCaveStare",
        "UndergroundCaveAvert",
        "UndergroundCaveSurface"
    ];

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            At first, the basilisk seemed like a determined, lethal [b]killing machine[/b],
                            but now that you've wrestled the fight out of it, it seems to have become a lot more [b]passive[/b].
                            It knows it did something bad, and it realizes that you're about to punish it.
                            It's as though the monster was hiding its submissive nature under a veil of hostility.
                            Now the creature is more of a [b]big wimp[/b] than a menacing beast,
                            and you're the [b]bully[/b] about to [b]push it around[/b] in whatever way you please.
                        [/p]
                        [p]
                            Thinking of this situation as a [i]reversal of power[/i] awakens lustful feelings deep inside of you.
                            You've had this recurring [i]wet dream[/i] a few times, where kobolds would stage an uprising,
                            and you'd find yourself in front of the [b]juiciest[/b] maiden in all of Varanar,
                            with ten of your kind holding her struggling, naked form down for you to [b]ravage[/b].
                            This is [b]just[/b] like that dream, only instead of a maiden, you have a [b]basilisk[/b] to quench your lust upon,
                            the uprising was more of a [b]wrestling match[/b], and there aren't really any kobolds around to hold the monster still.
                        [/p]
                        [p]
                            Okay, in retrospect, this entire ordeal might not be very much like your dream after all.
                            Still, you were reminded of it, and that's enough to get your member to slowly start to harden within your cockslit.
                            You begin to imagine all the things you could do to the creature, now that you've subdued it.
                            While not so similar to your dream, having your way with a large, ferocious animal [b]does[/b] pique your interest.
                        [/p]
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at molesting a basilisk, you decide to act on your impulses.
                            You begin to stroke the emerging tapered tip of your member, which rapidly grows erect, fueled by the dirty thoughts popping up in your mind.
                            Its muzzle.
                            You want to fuck its muzzle, filled with sharp, pointy teeth.
                            You want to walk on the wild side, to tempt fate itself,
                            to see whether or not the beast still has enough [b]fight[/b] in it to resist, to [b]bite down[/b].
                        [/p]
                        [p]
                            You bet it doesn't.
                            You bet you've [b]broken[/b] it.
                            You bet it won't even [b]try[/b] to stop you from fucking its maw.
                            You've shown it that it's [b]not[/b] the boss around here.
                            You've overpowered it, despite all odds.
                            It won't resist.
                            Not after the [b]ass-whooping[/b] you gave it.
                            Not after it attacked you, so [b]certain[/b] that you'd be an easy prey.
                        [/p]
                        [p]
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, you approach the basilisk's head, dropping to your knees directly in front of it,
                            your erect member mere inches away from the feral's snout.
                            Its maw is still open, the monster panting heavily, not yet fully recovered from the intense tussle.
                            A perfect opportunity.
                        [/p]
                        [p]
                            You line up the tip of your member with the beast's mouth, and with a single, smooth roll of your hips,
                            you slide your length inside of its muzzle, the bottom of your shaft pleasantly rubbing alongside the basilisk's slender, forked tongue.
                            While the creature's exterior is hard, tough and cold, the insides of its mouth are the exact opposite.
                            Warm, fleshy, [b]wet[/b] with saliva, almost what you imagine a human's gums must feel like.
                            A pleasant surprise, and not an unwelcome one.
                        [/p]
                        [p]
                            Your dick being rammed into its gaping maw seems to rouse the lizard from its post-battle drowse.
                            Its eyes shoot wide open.
                            The taste of kobold cock spreading throughout its mouth, it quickly realizes what's going on.
                            The monster doesn't like it one bit.
                            Lost in pleasure, you close your eyes and lean back ever so slightly.
                            Then, a loud [b]snap[/b], like the sound of teeth clattering together, echoes throughout the cave.
                            Opening your eyes, you stare down at your groin, to find the entire tip missing, bitten off by the basilisk's snapping jaws.
                        [/p]
                        ${global.ui.condFail(`Small dick`)}
                        [p]
                            The remaining fleshy stump protruding from your slit liberally squirts blood atop the crocodillian's muzzle.
                            You stare in horror at your mangled groin.
                            An immense pain, unlike anything you've ever felt before, shoots up your spine.
                            Your sight fades, and you tumble to the ground, unconscious.
                            Spurred by the taste of blood and fresh kobold meat, the basilisk rises to its feet,
                            [b]rips[/b] apart your stomach with its sharp claws, and feasts upon your entrails.
                            You don't wake up again.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    Bite end.
                                [/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" style="margin-left: auto; position: relative; float: right;" id="${continueId}">`;
        choice += `&lt;&lt Revert`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400, function () {
            var main = $("#main");
            main.scrollTop(main[0].scrollHeight - main.height() - 5); // Fix scrolling
        });
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.misc.unsetAll();
            ctrl.dominance.undo();
            ctrl.dominance.unset();
            ctrl.wrestleWin.undo();
            ctrl.wrestleWin.unset();
            ctrl.fight.unset();
            ctrl.resistSucceed.undo();
            ctrl.resistSucceed.unset();
            ctrl.stare.undo();
            ctrl.stare.unset();
            ctrl.avert.undo();
            ctrl.avert.unset()
            ctrl.surface.undo();
            ctrl.surface.unset();
            for (var i = 0; i < idsToRemove.length; i++) {
                $("#" + idsToRemove[i]).remove();
            }
            loadNextPart(ctrl.fragments.surface);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\oral\UndergroundCaveBite.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\oral\UndergroundCaveOral.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveOral(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveDominanceChoice.ORAL;
    var reqController = ctrl.dominance;

    var toFail = new UndergroundCaveOralSex(ctrl);
    var toSuccess = new UndergroundCaveBite(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        var dick = ctrl.misc.dick();
        return dick && dick < SizeEnum.LARGE;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\oral\UndergroundCaveOral.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\pee\UndergroundCavePee.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCavePee(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveDominanceChoice.PEE;
    var reqController = ctrl.dominance;
    var subController = ctrl.preContinue;
    var changeController = ctrl.pee;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveDominanceLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskPeed = true;
        var dick = ctrl.misc.dick();
        var pants = ctrl.misc.pants();
        var undies = ctrl.misc.underwear();
        var wellDrown = global.kobold.history.shared.well.drowned;

        var lowerPantsAndUndies = `Already naked`;

        if (pants) {
            lowerPantsAndUndies = `After lowering your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (undies) {
                lowerPantsAndUndies += ` and your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
            }
        } else if (undies) {
            lowerPantsAndUndies = `After lowering your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
        }

        var toReturn = `[p]
                            A big dog.
                            That's all it is.
                            A [b]puppy[/b] that intended to [b]bite[/b] its ${dick ? `master` : `mistress`}.
                            While you don't know much about canine interactions,
                            and most of what you [i]do[/i] know is based off of hearsay and [i]calculated guesses[/i],
                            you [i]think[/i] they show dominance by [i]pissing[/i] on one another.
                            The alpha male pees on the beta, the beta pees on the omega, and so on and so forth.
                            Or was that how they marked their territory?
                            You can't quite remember.
                        [/p]
                        [p]
                            It doesn't matter anyway.
                            You've gulped down a lot of water in the past hour orso,
                            in the process of nearly drowning${wellDrown ? ` more than once` : ``},
                            and you feel like you need to take a leak.
                            The basilisk has deserved a punishment, and you intend to deliver.
                        [/p]
                        [p]
                            ${lowerPantsAndUndies}, you take up position next to the monster's slumped down body.
                            You spread your legs, stick out your groin,
                            and slightly ${dick ? `pry your cockslit open` : `spread your folds`} using two of your fingers.
                            With a relieved sigh, you release your budding bladder,
                            a stream of watery piss squirting from your ${dick ? `slit` : `urethra`},
                            arcing through the air before splattering down onto the feral's stone back. 
                        [/p]
                        [p]
                            The basilisk groans in discontent as it feels the warm liquid soiling its body.
                            It knows fully well what you're doing, but it doesn't resist.
                            It [b]knows[/b] that it [b]deserves[/b] this punishment.
                            And you've only gotten [b]started[/b].
                        [/p]
                        [p]
                            You redirect your groin, wetting the upper half of the creature's body while the stream travels over its head, towards its muzzle.
                            The lizard hisses angrily, but it is far too exhausted to even attempt to stop you.
                            It simply lays there as you empty the entire contents of your bladder over its face,
                            its angry eyes closing in resignation, or perhaps to prevent any splatter from getting into them.
                        [/p]
                        [p]
                            You feel [b]powerful[/b], taking a piss on a mythological being several times your weight.
                            Like you [b]own it[/b].
                            Like it's [b]your bitch[/b].
                            And to be frank, you [b]did[/b] beat it in a straight-up fight.
                            It [b]is[/b] your [b]bitch[/b].
                            You have every right to feel accomplished.
                            And you have every right to reap the spoils of your labour, however you see fit.
                        [/p]
                        [p]
                            Sensing the end of your leak approaching, you wildly rock your hips from side to side,
                            splattering urine all over the basilisk's body, from the tip of its snout to the base of its tail.
                            Your stream of piss rapidly weakens, and soon fades into a series of droplets dripping from your loins.
                            Satisfied, and convinced the basilisk now knows its place, ${pants || undies ? `you redress, and` : `you`} prepare to leave the cave.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\pee\UndergroundCavePee.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\sex\UndergroundCaveFuck.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveFuck(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveDominanceChoice.SEX;
    var reqController = ctrl.dominance;
    var subController = ctrl.preContinue;
    var changeController = ctrl.sex;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveDominanceLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskDomFucked = true;
        var dick = ctrl.misc.dick();
        var dick = ctrl.misc.dick();
        var pants = ctrl.misc.pants();
        var undies = ctrl.misc.underwear();

        var lowerPantsAndUndies = `and already naked`;

        if (pants) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(pants)}`;
            if (undies) {
                lowerPantsAndUndies += ` and your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
            }
        } else if (undies) {
            lowerPantsAndUndies = `after lowering your ${global.kobold.inventory.createItemWithTooltip(undies)}`;
        }
        var toReturn = `[p]
                            At first, the basilisk seemed like a determined, lethal [b]killing machine[/b],
                            but now that you've wrestled the fight out of it, it seems to have become a lot more [b]passive[/b].
                            It knows it did something bad, and it realizes that you're about to punish it.
                            It's as though the monster was hiding its submissive nature under a veil of hostility.
                            Now the creature is more of a [b]big wimp[/b] than a menacing beast,
                            and you're the [b]bully[/b] about to [b]push it around[/b] in whatever way you please.
                        [/p]
                        [p]
                            Thinking of this situation as a [i]reversal of power[/i] awakens lustful feelings deep inside of you.
                            You've had this recurring [i]wet dream[/i] a few times, where kobolds would stage an uprising,
                            and you'd find yourself in front of the [b]juiciest[/b] maiden in all of Varanar,
                            with ten of your kind holding her struggling, naked form down for you to [b]ravage[/b].
                            This is [b]just[/b] like that dream, only instead of a maiden, you have a [b]basilisk[/b] to quench your lust upon,
                            the uprising was more of a [b]wrestling match[/b], and there aren't really any kobolds around to hold the monster still.
                        [/p]
                        [p]
                            Okay, in retrospect, this entire ordeal might not be very much like your dream after all.
                            Still, you were reminded of it, and that's enough to get your
                            ${dick ? `member to start to harden within your cockslit` : `slit to slightly moisten`}.
                            You begin to imagine all the things you could do to the creature, now that you've subdued it.
                            While not so similar to your dream, having your way with a large, ferocious animal [b]does[/b] pique your interest.
                        [/p]`;

        if (dick) {
            var virgin = ctrl.misc.dickVirgin();
            global.kobold.body.dickVirgin = false;
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at molesting a basilisk, you decide to act on your impulses.
                            You begin to stroke the emerging tapered tip of your member, which rapidly grows erect, fueled by the dirty thoughts popping up in your mind.
                            You want to fuck its stony ass, concealed underneath its tail.
                            You want to walk on the wild side, to tempt fate itself,
                            to see whether or not the beast still has enough [b]fight[/b] in it to resist, to [b]thrash around[/b] and [b]maim you[/b].
                        [/p]
                        [p]
                            You bet it won't.
                            You bet you've [b]broken[/b] it.
                            You bet it won't even [b]try[/b] to stop you from pounding its butt.
                            You've shown it that it's [b]not[/b] the boss around here.
                            You've overpowered it, despite all odds.
                            It won't resist.
                            Not after the [b]ass-whooping[/b] you gave it.
                            Not after it attacked you, so [b]certain[/b] that you'd be an easy prey.
                        [/p]
                        [p]
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, ${lowerPantsAndUndies}, you approach the basilisk's backside,
                            dropping to your knees directly behind it, at the base of its tail.
                            The monster is panting heavily, not yet fully recovered from the intense tussle.
                            A perfect opportunity.
                        [/p]
                        [p]
                            Using both your hands, you lift up the basilisk's heavy tail, revealing its smooth backside, along with its loins.
                            Oddly enough, the crocodillian seems to be aroused [i]already[/i].
                            Its girthy, ridged and draconic-looking member hangs fully erect below its soft underbelly.
                            It must be in rut or something.
                            Maybe that's why it was so fiercely defending its lair.
                            Or perhaps it's just a natural born [b]slut[/b].
                            That last option, while the least probable, is also the most exciting.
                            You think you'll go with that one.
                        [/p]
                        [p]
                            Once lifted, you manage to hold the beast's tail up using only one hand.
                            With the other, you begin to [i]probe[/i] around its rear.
                            At least the underside is [b]far[/b] more [i]soft[/i] than its back.
                            In fact, you think you might actually end up [b]enjoying[/b] this.
                        [/p]
                        [p]
                            After giving yourself a few more strokes, you line up the tip of your member with the basilisk's tailhole.
                            You immediately begin to apply pressure, pressing your tapered length into its overly tight rear.
                            Once past its clenching sphincter, you easily manage to ram the rest of your length inside, bottoming out in a single, smooth thrust.
                            The monster hisses, but it does not fight back.
                            It stays in place, accepting its position as your [b]bitch[/b].
                            Just as planned.
                        [/p]
                        [p]
                            While the creature's exterior is hard, tough and cold, its insides are the exact opposite.
                            Warm, fleshy, and exceedingly tight, almost what you imagine a human's ass to feel like, minus a bubbly butt to rest your groin against.
                            A pleasant surprise, and not an unwelcome one at that.
                            You lean back ever so slightly, and slowly begin to thrust in and out of the crocodillian's ass.
                            Luckily, it is far too worn out to put up any other kind of resistance.
                            Or maybe it really [b]does[/b] like this.
                        [/p]
                        [p]
                            The basilisk's tight insides prove to be quite pleasurable,
                            but the sheer [b]debasement[/b] of what you're doing drives your excitement through the roof.
                            You're [b]fucking[/b] a feral.
                            The [b]beast[/b] that tried to murder you has been turned into your [b]bitch[/b].
                            You get to be the person on the [b]giving[/b] end of abuse, for once.
                            And on top of all of this, you're tempting fate itself by ramming your dick into a creature that could easily [b]tear you apart[/b]!
                        [/p]
                        [p]
                            Closing your eyes for a moment, your mind briefly wanders back to the dream that inspired you to abuse the basilisk.
                            Snapping back to reality, however, you realize that this situation is [b]even better[/b] than what you imagined.
                            Having the [b]pleasure[/b] of dominating a [b]lethal beast[/b],
                            one that could [b]rip you to shreds[/b] were it not too exhausted,
                            stirs your loins far more than the thought of raping a helpless damsel.
                            While you've only been absent-mindedly humping away at the big lizard's rear for a few seconds short of a minute,
                            you can already feel your premature end rapidly approaching.
                        [/p]
                        [p]
                            You let out a lustful groan, rubbing your free hand along the creature's soft underside.
                            You consider jerking it off as you fuck it, but you decide against giving it pleasure.
                            This is supposed to be a [b]punishment[/b], after all.
                            Instead, you grab onto its thigh, upping the intensity of your thrusts.
                            The fact that it doesn't resist,
                            the [b]thought[/b] of having reduced the once-proud crocodillian to an [b]eager[/b] tail-lifter,
                            [b]that[/b] sends you over the edge.
                        [/p]
                        [p]
                            You give the beast's rear a few more rapid humps, before reaching your end.
                            Your member [b]pulses[/b] and [b]spasms[/b], spunk pumping from deep within your internal balls,
                            through your shaft, squirting directly inside of the monster's body.
                            Clearly reluctant, or perhaps even outright [b]unwilling[/b], it tries to shuffle forwards,
                            but using both of your hands you firmly keep it in place for the entire duration of your orgasm.
                        [/p]
                        [p]
                            Your lust sated, you pull your cock from the big beast's butt, and let go of its tail.
                            It's almost kind of [b]sad[/b] to leave the basilisk without getting it off, too,
                            but you both know that it hasn't [b]earned[/b] that right.
                            It shouldn't have tried to murder you.
                            If it hadn't, maybe you'd seen fit to give it a pity-handjob after finishing.
                            You rise to your feet${pants || undies ? ` and redress` : ``}.
                        [/p]`;
            if (virgin) {
                toReturn += `[p][c][b]You are no longer a virgin.[/b][/c][/p]`;
            }
        } else {
            toReturn += `
                        [p]
                            Doubting you'll [b]ever[/b] get a second chance at molesting a basilisk, you decide to act on your impulses.
                            You begin to touch your own slit, which rapidly dampens, fueled by the dirty thoughts popping up in your mind.
                            You want to touch its stony ass, concealed underneath its tail.
                            You want to walk on the wild side, to tempt fate itself,
                            to see whether or not the beast still has enough [b]fight[/b] in it to resist, to [b]thrash around[/b] and [b]maim you[/b].
                        [/p]
                        [p]
                            You bet it won't.
                            You bet you've [b]broken[/b] it.
                            You bet it won't even [b]try[/b] to stop you from ramming your fingers up its butt.
                            You've shown it that it's [b]not[/b] the boss around here.
                            You've overpowered it, despite all odds.
                            It won't resist.
                            Not after the [b]ass-whooping[/b] you gave it.
                            Not after it attacked you, so [b]certain[/b] that you'd be an easy prey.
                        [/p]
                        [p]
                            Lust clouds your mind, preventing you from thinking about the consequences should things go awry.
                            Driven by desire, you approach the basilisk's backside,
                            dropping to your knees directly behind it, at the base of its tail.
                            The monster is panting heavily, not yet fully recovered from the intense tussle.
                            A perfect opportunity.
                        [/p]
                        [p]
                            Using both your hands, you lift up the basilisk's heavy tail, revealing its smooth backside, along with its loins.
                            Oddly enough, the crocodillian seems to be aroused [i]already[/i].
                            Its girthy, ridged and draconic-looking member hangs fully erect below its soft underbelly.
                            It must be in rut or something.
                            Maybe that's why it was so fiercely defending its lair.
                            Or perhaps it's a natural born [b]slut[/b].
                            That last option, while the least probable, is also the most exciting.
                            You think you'll go with that one.
                        [/p]
                        [p]
                            Once lifted, you manage to hold the beast's tail up using only one hand.
                            You don't [i]solely[/i] intend to finger its butt, though.
                            You'll need [b]both[/b] your hands for what you have in store.
                            Intent on making the creature a bit more [i]cooperative[/i],
                            you gingerly grab hold of its cock, and slowly rub your comparatively small palm up and down the basilisk's shaft,
                            stroking its length at an agonizingly slow pace.
                        [/p]
                        [p]
                            The feral lets out a surprised, but remotely [i]pleased[/i] grunt.
                            A viscous, heady squirt of precum shoots from the tapered tip of its member.
                            The scent of the stuff in [i]inebriating[/i], like a potent aphrodisiac in its own right.
                            The monster doesn't seem quite comfortable enough to willingly keep its tail raised yet,
                            but you know exactly how to spur it on.
                        [/p]
                        [p]
                            Leaning your head in close, you press your muzzle against the beast's rear.
                            Your tongue snakes out of your mouth, and in a fit of deviancy, you tentatively lick the creature's tailhole,
                            elliciting another confused, but far from upset grunt from its muzzle.
                            Sterile and tasteless, apart from a remote hint of [i]sweat[/i], but the basilisk [i]did[/i] seem to like it.
                            Its member throbs in your hand, and it slowly starts to move its hips on its own,
                            humping against your palm.
                            Just a little bit more.
                        [/p]
                        [p]
                            Again and again, you lap at the monster's rear, your tongue settling into a rhythm,
                            your saliva wettening its scales and lubing its hole, preparing it for what is to come.
                            The feral is blissfully unaware of your devious plans.
                            You strengthen your grip on its member, and increase the speed of your hand, properly jerking it off.
                            You can feel its precum slickening your palm, reducing friction, making the entire ordeal even [b]more[/b] pleasurable for the crocodillian.
                        [/p]
                        [p]
                            Finally, with the beast completely lost in pleasure, you get what you were after:
                            a burden is lifted off your hand, the basilisk willingly keeping its tail lifted,
                            granting you access to its loins [i]without[/i] needing to work for it.
                            [b]Perfect[/b].
                            With the large lizard fully relaxed, you pull your head back a little,
                            a few thick strings of saliva connecting your muzzle to its ass still.
                            You move your now-free hand to the creature's rear.
                            This is where the real fun begins.
                        [/p]
                        [p]
                            You press one of your fingers against the monster's tailhole, teasingly circling it, spreading your spit around.
                            Then, while simultaneously upping the pace of your frantic jerking once more,
                            you slide the digit into the feral's spit-lubed rear, ramming it inside of the basilisk to the very knuckle.
                            The large lizard is obviously startled.
                            Its legs frantically shuffle about, but after the initial panic subsides, it decides to remain in place.
                            It [b]knows[/b] it deserves this.
                            Or perhaps it's [i]confused[/i] about what's going on.
                            Or maybe it [b]likes it[/b].
                        [/p]
                        [p]
                            While continuing to firmly stroke the beast's ridged member,
                            you slide a second finger inside of its thoroughly lubricated behind,
                            and begin to pump both of the digits in and out of the creature's body, properly [b]fingering[/b] it while jerking it off,
                            sending mixed messages of both [b]submission[/b] and [b]pleasure[/b] coursing down its spine.
                        [/p]
                        [p]
                            You can feel the feral's end approaching.
                            It begins to thrust against your hand once more,
                            at the same time unintentionally [b]riding[/b] your fingers,
                            even more mixed messages filling its mind.
                            Its shaft twitches in your grasp.
                            Its end is getting [b]very[/b] near.
                        [/p]
                        [p]
                            In order to get a rise out of the basilisk, you hold the hand wrapped around its member perfectly still,
                            while continuing to deftly molest its rear.
                            [b]Needing[/b] to get off, it [b]wildly[/b] humps against your slickened palm,
                            outright [b]fucking[/b] your hand.
                            With a pleasured hiss, the monster reaches its climax,
                            but feeling exceptionally cruel, you let go of its cock the [b]moment[/b] it finishes,
                            leaving the draconic-looking length to spasm and squirm unattended,
                            the beast's orgasm driven on solely by the stimulation of its rear end being fingered.
                        [/p]
                        [p]
                            Wave after wave of fragrant feral spunk splatters against the stone cavern floor.
                            The basilisk desperately humps against [b]nothing[/b], trying but failing to find any pleasure,
                            eventually resorting to grinding its cock against the ground, rubbing its length through its own puddle of cum.
                            Satisfied with this [i]punishment[/i], you pull your fingers from the lizard's ass,
                            and rise to your feet once more${pants || undies ? ` before redressing` : ``}.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\sex\UndergroundCaveFuck.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\UndergroundCaveDominance.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveDominance(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveWrestlingWinChoice.DOMINANCE;
    var reqController = ctrl.wrestleWin;
    var subController = ctrl.dominance;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new UndergroundCavePee(ctrl);
    var secondNext = new UndergroundCaveOral(ctrl);
    var thirdNext = new UndergroundCaveFuck(ctrl);

    var firstChoice = UndergroundCaveDominanceChoice.PEE;
    var secondChoice = UndergroundCaveDominanceChoice.ORAL;
    var thirdChoice = UndergroundCaveDominanceChoice.SEX;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskDominated = true;
        var toReturn = `[p]
                            The basilisk is already beaten.
                            There is no point to taking its life.
                            It has submitted to you.
                            Still, you feel like it hasn't learned its lesson yet.
                            It probably won't turn aggressive after you release it from your chokehold,
                            but you imagine that, in time, it might return to its kobold-attacking ways.
                            You think you know exactly what it'll take to prevent that.
                            You need to show the stony lizard who's boss.
                        [/p]
                        [p]
                            You slowly release your grip on the monster's throat.
                            It immediately begins to gasp for air, taking a few deep breaths.
                            You roll off of the beast, and rise to your feet.
                            It looks up at you, submissiveness clear on its face, like a dog that knows it wronged its master.
                            And like a dog, you need to rub its defeat in its face, to make sure it'll never forget this day.
                            You need to assert your dominance over it, in a way that it'll remember for [b]years[/b] to come.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Pee on it</button>`;

        var tooltip = global.parser.parse(`Display your dominance the way canines do.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Use its muzzle</button>`;

        var tooltip = global.parser.parse(`Time to facefuck a beast's mouth filled with sharp teeth. What could go wrong?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Fuck it</button>`;

        var male = ctrl.misc.dick();
        var tooltip = global.parser.parse(`Make the basilisk your [b]bitch[/b].`);
        if(!male) {
            tooltip = global.parser.parse(`Fingerfuck the basilisk's tailhole. Make it your [b]bitch[/b].`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var UndergroundCaveDominanceChoice = {
    PEE: 1,
    ORAL: 2,
    SEX: 3
};

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\dominance\UndergroundCaveDominance.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\go\UndergroundCaveLetGoLeave.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveLetGoLeave(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.letGoLeave;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You and the basilisk slake your thirst, communally lapping at the cool water for nearly an entire minute on end.
                            Finishing a little bit before it does, you rise to your feet.
                            Noticing you've stopped drinking, it follows suit, seemingly standing at attention, awaiting your next [i]command[/i].
                            But no order comes.
                            You have no interest in being the creature's master.
                            You just want to get out of here already.
                        [/p]
                        [p]
                            You start following the river-branch, hoping it'll lead you to some kind of exit.
                            The monster tags along, scuttling after you, like a trained dog that doesn't even need a leash to know its place.
                            Not feeling threatened by it, you take your time to leisurely stroll alongside the water, until daylight comes into sight.
                            As you expected: it was a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            The basilisk, which followed you until the cave-entrance, stares at your seated frame from the darkness of the cavern's interior.
                            It seems [b]scared[/b] of the sunlight, perhaps unable to weather it.
                            You give it a dismissive wave, wordlessly ordering it to go back home.
                            It utters a loud whine, hesitating to follow your command,
                            but after a bit of what you can only imagine to be [i]pouting[/i],
                            the creature rejectedly turns its tail, and saunters back into its lair. 
                        [/p]
                        [p]
                            You could have used a basilisk companion, especially one that knows its place.
                            But since it can't travel during the day,
                            and seeing as you'd have a hard time explaining to the guards of Varanar why you're trying to bring a [b]lethal beast[/b] into the city at night,
                            you figure it'd be best to leave the monster here.
                            It's better off without you anyway.
                            Around you, it'd just end up getting killed.
                            Besides, you can barely even scrounge up enough food to feed yourself,
                            let alone a feral several times your weight.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that, you managed to wrestle and tame a basilisk.
                            You doubt there's other kobolds that can claim they tussled with a mythological beast and [b]won[/b].
                        [/p]
                        [p]
                            You've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\go\UndergroundCaveLetGoLeave.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\go\UndergroundCaveLetGo.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveLetGo(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveWrestlingWinChoice.GO;
    var reqController = ctrl.wrestleWin;
    var subController = ctrl.preContinue;
    var changeController = ctrl.letGo;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveLetGoLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            The basilisk is already beaten.
                            There is no point to taking its life.
                            It has submitted to you.
                            You hope it won't turn aggressive once more after you release it,
                            but you don't really want to take its life either.
                        [/p]
                        [p]
                            Despite doubt growing in the back of your mind, you slowly release your grip on the monster's throat.
                            It immediately begins to gasp for air, taking a few deep breaths.
                            At least it doesn't immediately appear hostile.
                            If it wanted to continue fighting, it'd have thrown you off its back by now.
                            You roll off of the beast, and exhaustedly rise to your feet.
                            It looks up at you, submissiveness clear on its face, like a dog that knows it wronged its master.
                            This cave must be its lair or something.
                            It was just trying to protect its territory.
                        [/p]
                        [p]
                            You give the basilisk a short, understanding nod, before gesturing towards the water.
                            You could sure use a drink after a fight like that, and you've no doubt that it, too, is thirsty.
                            The both of you walk towards the water.
                            At its edge, you drop to your hands and knees.
                            Together with the monster you were fighting not a minute ago,
                            you lower your head to the surface of the water,
                            and the both of you start greedily lapping up the fresh, revitalizing liquid.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\go\UndergroundCaveLetGo.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\snap\UndergroundCaveSnapLeave.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveSnapLeave(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.snapLeave;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You did it.
                            You won.
                            You killed the monster.
                            Just like a knight would do!
                            Somehow, you don't feel very [i]knightly[/i], though.
                            You should be proud of yourself, but for some reason, you aren't.
                            Maybe your actions weren't as [i]chivalrous[/i] as you would have liked,
                            but it's the outcome that matters, and not how you achieved it, right?
                            The ends justify the means.
                        [/p]
                        [p]
                            But then again, the basilisk was just a simple beast minding its own business.
                            You were intruding on its territory.
                            You could have probably left and it wouldn't have harmed you at all.
                            Now you're confused.
                            Did you do the right thing?
                        [/p]
                        [p]
                            You honestly don't know anymore.
                            Does it even matter?
                            What's done is done.
                            If you hadn't killed the creature, some pompous knight in shining armor
                            -- with nothing left to prove, neither to himself, nor to the world --
                            would have strolled into this cave to add another tally to his already impressive list of slain monsters.
                            For someone properly armed and armored, this would have been a [b]chore[/b], a menial task barely even worthy of mention.
                            But for you, this is a feat you'll remember until the day you die.
                        [/p]
                        [p]
                            Your conviction bolstered, you exhaustedly roll off of the lizard.
                            After resting for a few minutes, you kneel besides it.
                            You figure it'd be best to take a trophy.
                            Something to remember this day by.
                            Something you can show to people, so they'll believe you when you tell them you singlehandedly killed a basilisk.
                            But what?
                        [/p]
                        [p]
                            While pondering, you run one of your hands over the deceased basilisk's back.
                            Some of its scales seem to be rather [i]loose[/i].
                            You take hold of one with both of your hands, and with surprising ease, you manage to [b]tear[/b] the scale from the monster's backside.
                            Like plucking a chicken.
                            You repeat the process a couple of times,
                            until you have collected a suitable handful of ${global.kobold.inventory.createItemWithTooltip(new BasiliskScales())}
                            to take with as your prize for ridding the world of this horrible monstrosity.
                        [/p]
                        [p]
                            Your trophies collected, you decide it's time to find a way out of the cave.
                            Following the river-branch should lead you to some kind of exit.
                            Without the threat of the basilisk around, you don't even need to [b]swim[/b].
                            You leisurely stroll alongside the water, until daylight comes into sight.
                            As you expected: the water was a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that, you managed to kill a basilisk.
                            It might not have been the most [i]honorable[/i] way of slaying the beast, but it was slain nonetheless.
                        [/p]
                        [p]
                            Besides the ${global.kobold.inventory.createItemWithTooltip(new BasiliskScales())} you [i]harvested[/i],
                            you've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\snap\UndergroundCaveSnapLeave.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\snap\UndergroundCaveSnap.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveSnap(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveWrestlingWinChoice.SNAP;
    var reqController = ctrl.wrestleWin;
    var subController = ctrl.preContinue;
    var changeController = ctrl.snap;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveSnapLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            While it is already beaten, you can [b]not[/b] allow the basilisk to live.
                            [b]It[/b] attacked [i]you[/i], not the other way around.
                            Unless you finish it off right here, and right now,
                            it'll continue to [b]terrorize[/b] whoever may wander into this cave.
                            It's a foul [b]beast[/b], and as an aspiring knight to be, it is your [b]duty[/b] to put an end to it.
                        [/p]
                        [p]
                            You don't stop squeezing down, denying the monster any fresh air at all.
                            With the last bit of breath remaining in its lungs, it manages to utter a whine, [b]begging[/b] you to spare its life.
                            But you can't do that.
                            It tried to [b]kill you[/b], and it failed to do the job.
                            If the tables were turned, it wouldn't show you mercy.
                            It'd [b]rip out your guts[/b] and eat them in front of you.
                        [/p]
                        [p]
                            Despite growing doubts inside of your mind, you continue to choke out the big lizard.
                            The creature is nearing its end.
                            You know it.
                            It knows it.
                            But still, it does not resist.
                            Like it realizes its time has come.
                            Like it understands that there is nothing left it can do to stop a kobold as determined as you are.
                            It's almost sad, in a way.
                        [/p]
                        [p]
                            The basilisk's body slowly grows limp underneath you.
                            Its big, enthralling eyes close.
                            Its head lifelessly slumps down to the ground.
                            It is over.
                        [/p]
                        [p]
                            [c]
                                [b]You have killed the basilisk.[/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\snap\UndergroundCaveSnap.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\UndergroundCaveWrestlingWin.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveWrestlingWin(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.wrestleWin;
    var changeController = ctrl.wrestleWin;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new UndergroundCaveSnap(ctrl);
    var secondNext = new UndergroundCaveLetGo(ctrl);
    var thirdNext = new UndergroundCaveDominance(ctrl);

    var firstChoice = UndergroundCaveWrestlingWinChoice.SNAP;
    var secondChoice = UndergroundCaveWrestlingWinChoice.GO;
    var thirdChoice = UndergroundCaveWrestlingWinChoice.DOMINANCE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            The basilisk thinks it's tough.
                            And, admittedly, being made out of mostly stone, it most likely [b]is[/b].
                            Heck, it probably weighs several times more than you do.
                            But you have something it doesn't.
                            You have evolved [b]beyond[/b] being a [b]petty lizard[/b].
                            It might be able to swing its brute limbs around with unmatched ferocity,
                            but what you lack in strength, you make up for in wits.
                        [/p]
                        ${global.ui.condFail(`No weapon`)}
                        [p]
                            Your brain kicks into overdrive, a million thoughts flashing by in the blink of an eye.
                            Its hide is made out of rock.
                            You can't [b]hit it[/b], you'd just end up breaking your hand.
                            Your claws won't penetrate stone.
                            You don't have a weapon to use, and there aren't any loose rocks around.
                            You could try to lure it into the water, but that's a longshot at best.
                            Besides, it [i]might[/i] be able to swim.
                            That leaves only one option.
                            And it's [b]risky[/b].
                        [/p]
                        [p]
                            After a tense moment of inactivity, the monster suddenly [b]charges[/b] at you, bursting into a sprint.
                            It tries to bite you, but you deftly step to the side, avoiding its barreling frame.
                            The creature's powerful jaws snap shut, catching nothing but air.
                            You spot the perfect opportunity.
                            Here it goes.
                            All or nothing.
                        [/p]
                        [p]
                            You [b]leap[/b] on top of the crocodillian's back, instantly grabbing its throat in a choke-hold.
                            The bottom of the beast feels a lot [b]squishier[/b] than the top.
                            Perhaps its soft belly is its weak spot?
                        [/p]
                        [p]
                            You are in the [b]perfect[/b] position.
                            While on top of the basilisk, it can't hit you with its claws, it can't smack you with its tail, nor can it bite you.
                            Instead of outright accosting you, the monster begins to thrash about, wildly shaking its body in an attempt to throw you off.
                            You hold on for dear life, wrapping a second arm around the creature's neck, squeezing its windpipe shut while clinging to its back.
                        [/p]
                        [p]
                            Realizing you aren't going to let go like this, the basilisk switches tactics.
                            For a few moments, it seemingly calms down, its muscles tangibly [b]tensing[/b].
                            Right as thought it was on the verge of [b]surrendering[/b],
                            the beast [b]rolls[/b] onto its side, and then onto its back,
                            [b]crushing[/b] you beneath its heavy, stone frame.
                        [/p]
                        [p]
                            You [b]scream[/b] out in pain, but you do [b]not[/b] let go.
                            The torment lasts for all but a second, the basilisk continuing to roll around,
                            returning to its feet before tumbling to its back yet again,
                            [b]squishing[/b] you between its rocky hide and the stone cavern floor a second time.
                            No wonder they call this the [b]death roll[/b].
                        [/p]
                        ${global.ui.condSucc(`Strength and endurance greater than 6`)}
                        [p]
                            The massive monster rolls around three, no, [b]four[/b] times before grinding to a halt.
                            Despite everything, you've managed to stay on top of it, but only [b]barely[/b].
                            [b]Every[/b] part of your body [b]hurts[/b].
                            You can't handle any more.
                            You feel like if the monster rolls onto its back one more time, you'll [b]break[/b].
                        [/p]
                        [p]
                            You are breathing heavily.
                            It, too, [b]tries[/b] to catch some breath, but your forearm is still deftly blocking the creature's windpipe.
                            While it's done rolling, for now, it still does [b]not[/b] yield.
                            You [b]need[/b] to end this [b]now[/b].
                            If it gathers enough energy to start rolling again, you won't survive.
                        [/p]
                        [p]
                            Keeping one arm around the basilisk's neck,
                            you begin to wantonly [b]punch[/b] its underside with the other,
                            repeatedly bashing the feral from the bottom of its maw to its soft and vulnerable belly.
                            After about half a minute more of choking it while hitting it with your fist,
                            the beast finally slumps down to the ground, letting out a submissive whimper.
                        [/p]
                        [p]
                            [c]
                                [b]You have beaten the basilisk![/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Kill it</button>`;

        var tooltip = global.parser.parse(`This has gone on for long enough. End the basilisk.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Let it go</button>`;

        var tooltip = global.parser.parse(`The fight's over. You won. There's no need to murder the creature.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled} style="width:200px;">Assert dominance</button>`;

        var male = ctrl.misc.dick() ? "male" : "female";
        var tooltip = global.parser.parse(`The fight is over. You won. Time to assert yourself as the alpha ${male}.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width:200px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var UndergroundCaveWrestlingWinChoice = {
    SNAP: 1,
    GO: 2,
    DOMINANCE: 3
};

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\succeed\UndergroundCaveWrestlingWin.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\fail\UndergroundCavePetrificationRape.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCavePetrificationRape(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;

    var continueId = arguments.callee.name + "ButtonOne";

    var idsToRemove = [
        "UndergroundCavePetrificationRape",
        "UndergroundCaveWrestling",
        "UndergroundCaveOffence",
        "UndergroundCaveFight",
        "UndergroundCaveResistSucceed",
        "UndergroundCaveResist",
        "UndergroundCaveStare",
        "UndergroundCaveAvert",
        "UndergroundCaveSurface"
    ];

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var undies = ctrl.misc.underwear();
        var dick = ctrl.misc.dick();
        var vaginalVirgin = ctrl.misc.cuntVirgin();
        var slitVirgin = ctrl.misc.cockslitVirgin();
        var virgin = (dick && slitVirgin) || (!dick && vaginalVirgin);
        var toReturn = `[p]
                            The basilisk thinks it's tough.
                            And, admittedly, being made out of mostly stone, it most likely [b]is[/b].
                            Heck, it probably weighs several times more than you do.
                            But you have something it doesn't.
                            You have evolved [b]beyond[/b] being a [b]petty lizard[/b].
                            It might be able to swing its brute limbs around with unmatched ferocity,
                            but what you lack in strength, you make up for in wits.
                        [/p]
                        ${global.ui.condFail(`No weapon`)}
                        [p]
                            Your brain kicks into overdrive, a million thoughts flashing by in the blink of an eye.
                            Its hide is made out of rock.
                            You can't [b]hit it[/b], you'd end up breaking your hand.
                            Your claws won't penetrate stone.
                            You don't have a weapon to use, and there aren't any loose rocks around.
                            You could try to lure it into the water, but that's a longshot at best.
                            Besides, it [i]might[/i] be able to swim.
                            That leaves only one option.
                            And it's [b]risky[/b].
                        [/p]
                        [p]
                            After a tense moment of inactivity, the monster suddenly [b]charges[/b] at you, bursting into a sprint.
                            It tries to bite you, but you deftly step to the side, avoiding its barreling frame.
                            The creature's powerful jaws snap shut, catching nothing but air.
                            You spot the perfect opportunity.
                            Here it goes.
                            All or nothing.
                        [/p]
                        [p]
                            You [b]leap[/b] on top of the crocodillian's back, instantly grabbing its throat in a choke-hold.
                            The bottom of the beast feels a lot [b]squishier[/b] than the top.
                            Perhaps its soft belly is its weak spot?
                        [/p]
                        [p]
                            You are in the [b]perfect[/b] position.
                            While on top of the basilisk, it can't hit you with its claws, it can't smack you with its tail, nor can it bite you.
                            Instead of outright accosting you, the monster begins to thrash about, wildly shaking its body in an attempt to throw you off.
                            You hold on for dear life, wrapping a second arm around the creature's neck, squeezing its windpipe shut while clinging to its back.
                        [/p]
                        [p]
                            Realizing you aren't going to let go like this, the basilisk switches tactics.
                            For a few moments, it seemingly calms down, its muscles tangibly [b]tensing[/b].
                            Right as thought it was on the verge of [b]surrendering[/b],
                            the beast [b]rolls[/b] onto its side, and then onto its back,
                            [b]crushing[/b] you beneath its heavy, stone frame.
                        [/p]
                        [p]
                            You [b]scream[/b] out in pain, but you do [b]not[/b] let go.
                            The torment lasts for all but a second, the basilisk continuing to roll around,
                            returning to its feet before tumbling to its back yet again,
                            [b]squishing[/b] you between its rocky hide and the stone cavern floor a second time.
                            No wonder they call this the [b]death roll[/b].
                        [/p]
                        ${global.ui.condFail(ctrl.misc.strength() < 6 ? `Strength below 6` : `Endurance below 6`)}
                        [p]
                            You can't hold on.
                            Not much longer.
                            Your entire body [b]hurts[/b].
                            Your arms are growing weak.
                            ${pants ? `Your ${global.kobold.inventory.createItemWithTooltip(pants)} 
                            ${undies ? `and ${global.kobold.inventory.createItemWithTooltip(undies)}` : ``} have been torn to shreds in the struggle.` : `
                            ${undies ? `Your ${global.kobold.inventory.createItemWithTooltip(undies)} have been torn to shreds in the struggle.` : ``}`}
                            You feel like if the monster rolls onto its back one more time, you'll [b]break[/b].
                        [/p]
                        [p]
                            And it does just that.
                            For a third time, the basilisk crushes you beneath its stony back and the cavern floor.
                            Your weary limbs can no longer hold on.
                            Your arms unwrap from its throat.
                            The crocodillian has successfully thrown you off of its back.
                            you lay limply on the ground, completely dazed, no longer in any condition to fight,
                            while the beast rolls off of you, and gets back on its feet. 
                        [/p]
                        [p]
                            The creature crawls on top of you, pinning you to the ground by placing its front paws atop your shoulders,
                            resting its immense weight atop your fragile frame.
                            Its razor-sharp claws mere inches away from your throat, you don't [b]dare[/b] to struggle.
                            For the first time, you get a good look at the feral's underside, and what you see inspires [b]dread[/b].
                            The lizard is visibly aroused, its girthy, ridged and draconic-looking member hanging fully erect below its soft underbelly.
                            It must be in rut or something.
                            [i]Oh no.[/i]
                        [/p]
                        [p]
                            ${pants || undies ? `Having lost your below-waist clothing during the intense struggle, nothing` : `Nothing`}
                            stands in between the monstrosity's throbbing length, and your tender${virgin ? `, virginal` : ``} ${dick ? `cockslit` : `slit`}.
                            Before you even have the [b]chance[/b] to protest, it [b]rams[/b] itself forwards, the tapered tip of its member instantly finding its mark,
                            agonizingly ${dick ? `penetrating a hole that was never meant to be fucked` : `spreading your folds`},
                            [b]forcing[/b] over half of its ridged member inside of your body in a single thrust,
                            stretching you well beyond what your body was made to take.
                        [/p]
                        [p]
                            You cry out in pain, meekly kicking your legs to no avail.
                            The basilisk wastes no time, immediately starting to hump away at a rapid pace,
                            pounding you in an animalistic fashion, its girthy cock practically [b]tearing[/b] you apart.
                            It isn't long until the beast manages to bottom out.
                            ${dick ? `The tip of its member prods against your own flaccid length, threatening to [b]crush[/b] it with every wild thrust.` : `
                            The tip of its member prods against your cervix, threatening to force itself even deeper with every wild thrust.`}
                        [/p]
                        [p]
                            Distressed, and in immense pain, you try to find something, [b]anything[/b] to take your mind off of what is happening,
                            to be able to [b]escape[/b] reality.
                            After frantically looking around for [b]any[/b] way out of the basilisk's mating grip,
                            your sight falls upon its enchanting green eyes once more.
                            They captivate your aching mind, ensnaring every fiber of your being.
                            The feral returns your stare, gazing deep into your very [b]soul[/b].
                        [/p]
                        [p]
                            Within seconds, you can feel your limbs swiftly turning to stone.
                            The transformation is far more than scale-deep, your very flesh itself hardening, even your [b]bones[/b] petrifying.
                            Snapping back to reality, you try to [b]scream[/b] for help,
                            but find your muzzle having turned rock solid, frozen in place, unable to open.
                            Your arms.
                            Your legs.
                            Your tail.
                            You can't move any of them.
                        [/p]
                        [p]
                            The petrification creeps from your extremities onto your torso, like a [b]blight[/b] crawling over your scales.
                            You can feel it overtake your groin, slowly turning inch after inch of your slit into stone, and not just externally.
                            The coarseness of the hard surface surrounding its member doesn't seem to bother the lizard one bit.
                            It continues to wildly thrust its lower body, pistoning its member in and out of your petrified ${dick ? `cockslit` : `cunt`},
                            copious amounts of its own precum slickening the otherwise extremely [b]rough[/b] fuck.
                        [/p]
                        [p]
                            The contour of your entire body has turned to rock.
                            You can sense your innards, your vital organs themselves, slowly starting to change.
                            Retaining the slightest bit of sensibility in your loins, you can feel the big beast's end rapidly approaching,
                            its member pulsing and twitching inside of your stone slit.
                            The last thing you experience before your very heart itself is petrified,
                            is a first squirt of thick, potent basilisk cum splattering against your ${dick ? `flaccid member` : `cervix`},
                            one final indignation before your conscience fades.
                        [/p]
                        [hl]
                        [p]
                            Months pass.
                            Years, maybe even?
                            Decades?
                            The cave remains undisturbed, its brand new kobold statue preserved throughout the ages, permanent crotch-gape and all.
                            A warning sign to all intruders.
                            A testament to the basilisk's might.
                            But while [i]well[/i]-hidden, nothing can remain unexplored forever.
                            Especially not something this close to Varanar.
                            On a day that was bound to come,
                            a trio of particularly rowdy fellows from the nearby hamlet stumbles upon the cavern while poaching near the forest's edge.
                        [/p]
                        [p]
                            "[color=green]Hey, look![/color]" a deep, masculine voice reverberates across the cavern.
                            "[color=green]A statue of a kobold! Looks pretty lifelike.[/color]"
                            The human's two companions, only now catching up to him, stare at the promiscuous sculpture.
                            "[color=red]That's not a statue, you dumbass.[/color]" the least naive of the three speaks.
                            "[color=red]It's [b]obviously[/b] a [b]urinal[/b].[/color]"
                        [/p]
                        [p]
                            The third member of the party steps forward, leaning down and audibly [b]sniffing[/b] the air around the effigy.
                            "[color=blue]It [b]smells[/b] like a urinal alright. When's the last time this thing was [b]cleaned[/b]?! Ten years ago?[/color]"
                            The group bursts out into a short fit of laughter.
                            When the hilarity dies down, an interest proposition is made:
                            "[color=green]Five coins says you won't pee in it.[/color]"
                            "[color=red]You're on![/color]"
                        [/p]
                        [p]
                            The challenged human takes position in between the statue's spread legs.
                            Lowering his trousers, he deftly begins to take a leak into the petrified kobold's gaping ${dick ? `cockslit` : `cunt`},
                            his urine washing out the absolutely [b]revolting[/b] goop formed by basilisk spunk fermenting for more than a [b]decade[/b].
                            The disgusting sludge is replaced by fresh, warm, dark-yellow human piss,
                            the far less viscous liquid filling the statue's hole up to the [b]brim[/b] and then some,
                            like a [b]goblet[/b] overflowing with foul ale.
                        [/p]
                        [p]
                            The unwanted intrusion and the marking of territory did not go unnoticed.
                            From various dark holes in the walls, not one but [b]three[/b] sets of deep-green eyes stare at the invaders.
                            The family had grown since the last time they had a visitor.
                            The young one was getting hungry.
                            Dinner was served, and not a month too soon.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    Art end.
                                [/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" style="margin-left: auto; position: relative; float: right;" id="${continueId}">`;
        choice += `&lt;&lt Revert`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400, function () {
            var main = $("#main");
            main.scrollTop(main[0].scrollHeight - main.height() - 5); // Fix scrolling
        });
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.misc.unsetAll();
            ctrl.fight.unset();
            ctrl.resistSucceed.undo();
            ctrl.resistSucceed.unset();
            ctrl.stare.undo();
            ctrl.stare.unset();
            ctrl.avert.undo();
            ctrl.avert.unset();
            ctrl.surface.undo();
            ctrl.surface.unset();
            for (var i = 0; i < idsToRemove.length; i++) {
                $("#" + idsToRemove[i]).remove();
            }
            loadNextPart(ctrl.fragments.surface);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\fail\UndergroundCavePetrificationRape.js

// scenes\HamletDay2\UndergroundCave\fight\offence\fail\UndergroundCaveWrestling.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveWrestling(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;

    var toFail = new UndergroundCavePetrificationRape(ctrl);
    var toSuccess = new UndergroundCaveWrestlingWin(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.strength() >= 6 && ctrl.misc.endurance() >= 6;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\fail\UndergroundCaveWrestling.js

// scenes\HamletDay2\UndergroundCave\fight\offence\succeed\UndergroundCaveOffenceLoot.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveOffenceLoot(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.offenceLoot;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            You did it.
                            You won.
                            You killed the monster.
                            Just like a knight would do!
                            Somehow, you don't feel very [i]knightly[/i], though.
                            You should be proud of yourself, but for some reason, you aren't.
                            Maybe your actions weren't as [i]chivalrous[/i] as you would have liked,
                            but it's the outcome that matters, and not how you achieved it, right?
                            The ends justify the means.
                        [/p]
                        [p]
                            But then again, the basilisk was just a simple beast minding its own business.
                            You were intruding on its territory, and not the other way around.
                            You could have probably left and it wouldn't have harmed you at all.
                            Now you're confused.
                            Did you do the right thing?
                        [/p]
                        [p]
                            You honestly don't know anymore.
                            Does it even matter?
                            What's done is done.
                            If you hadn't killed the creature, some pompous knight in shining armor
                            -- with nothing left to prove, neither to himself, nor to the world --
                            would have strolled into this cave to add another tally to his already impressive list of slain monsters.
                            For someone properly armed and armored, this would have been a [b]chore[/b], a menial task barely even worthy of mention.
                            But for you, this is a feat you'll remember until the day you die.
                        [/p]
                        [p]
                            Your conviction bolstered, you wipe the blood off your ${global.kobold.inventory.createItemWithTooltip(weapon)}.
                            You kneel besides the lizard.
                            You figure it'd be best to take a trophy.
                            Something to remember this day by.
                            Something you can show to people, so they'll believe you when you tell them you singlehandedly killed a basilisk.
                            But what?
                        [/p]
                        [p]
                            While pondering, you run one of your hands over the deceased basilisk's back.
                            Some of its scales seem to be rather [i]loose[/i].
                            You take hold of one with both of your hands, and with surprising ease, you manage to [b]tear[/b] the scale from the monster's backside.
                            Like plucking a chicken.
                            You repeat the process a couple of times,
                            until you have collected a suitable handful of ${global.kobold.inventory.createItemWithTooltip(new BasiliskScales())}
                            to take with as your prize for ridding the world of this horrible monstrosity.
                        [/p]
                        [p]
                            Your trophies collected, you decide it's time to find a way out of the cave.
                            Following the river-branch should lead you to some kind of exit.
                            Without the threat of the basilisk around, you don't even need to [b]swim[/b].
                            You leisurely stroll alongside the water, until daylight comes into sight.
                            As you expected: the water was a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that, you managed to kill a basilisk.
                            It might not have been the most [i]honorable[/i] way of slaying the beast, but it was slain nonetheless.
                        [/p]
                        [p]
                            Besides the ${global.kobold.inventory.createItemWithTooltip(new BasiliskScales())} you [i]harvested[/i],
                            you've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\succeed\UndergroundCaveOffenceLoot.js

// scenes\HamletDay2\UndergroundCave\fight\offence\succeed\UndergroundCaveOffenceWin.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveOffenceWin(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.offenceWin;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveOffenceLoot(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskKilled = true;
        var weapon = ctrl.misc.weapon();
        var blunt = weapon.blunt();
        var stab = weapon.stab();
        var slash = weapon.slash();
        var toReturn = `[p]
                            The basilisk thinks it's tough.
                            And, admittedly, being made out of mostly stone, it most likely [b]is[/b].
                            Heck, it probably weighs several times more than you do.
                            But you have something it doesn't.
                            You have evolved [b]beyond[/b] being a [b]petty lizard[/b].
                            You have [b]thumbs[/b] and [b]hands[/b] instead of [b]paws[/b] and forelegs.
                            Unlike the monster, you can manipulate tools, do cartwheels, and most importantly: you can wield weapons.
                        [/p]
                        [p]
                            You brandish your ${global.kobold.inventory.createItemWithTooltip(weapon)}, the beast seemingly [b]hesitating[/b], but not exactly [i]backing down[/i] upon seeing it.
                            While its claws may be sharper than yours, and its body might be better protected,
                            you still feel like [b]you[/b] have the upper hand in this engagement.
                            Whether or not your self-confidence is well-placed, you will soon find out.
                        [/p]
                        [p]
                            The crocodillian suddenly bursts into a sprint, barreling towards you with its maw wide open, clearly intent on [b]biting you[/b].
                            You dodge to the side.
                            It follows up with a swipe that you only [b]barely[/b] manage to avoid.
                            Another attempted bite, but this time, the creature makes a critical mistake!
                            It leaves its head wide open to a counterattack, its muzzle scraping past your leg, missing your tender scales by less than an inch!
                        [/p]
                        [p]
                            Seeing an opportunity, you decide to capitalize on it.
                            You thrust your ${global.kobold.inventory.createItemWithTooltip(weapon)} downwards,
                            ${stab ? `plunging` : `swinging`} it towards the basilisk's head.
                            Through some kind of [b]miracle[/b], your weapon ends up directly
                            ${blunt ? `impacting against` : `${slash ? `slashing through` : `piercing through`}`}
                            one of the monster's large eyeballs,
                            the delicate organ audibly [b]popping[/b] with a loud, wet [b]squish[/b].
                            Critical hit!
                            You quickly retract your weapon, and take a small step back.
                        [/p]
                        ${global.ui.condSucc(global.kobold.inventory.createItemWithTooltip(weapon, true))}
                        [p]
                            The lizard recoils, clearly in extreme [b]pain[/b], blood [b]gushing[/b] from its eyesocket, along with [i]eye goop[/i].
                            It [b]screeches[/b] out a mixture between a threatening hiss, a pathetic whine and a call for help.
                            But no help comes.
                            This is between it and you.
                            And you've just tasted [b]blood[/b] in the water.
                        [/p]
                        [p]
                            Growing desperate, the wounded feral lashes out once more.
                            Now without depth perception, it poses little threat to you.
                            It snaps its jaws long before its muzzle has reached your legs, it swipes at nothing but [b]air[/b],
                            and you can feel it growing weaker with every passing second, its already sluggish movement slowing down to a crawl.
                        [/p]
                        [p]
                            At this point, finishing the beast off would be a [b]mercy killing[/b].
                            It can barely even stands on its legs anymore, overtaken by sheer agony.
                            As you approach, it tries to meekly swipe at you, but instead falls down flat on its face, in a small puddle of its own blood.
                            It hisses softly, pretending to act menacingly, but you both know it's at the end of its rope.
                        [/p]
                        [p]
                            You ready your ${global.kobold.inventory.createItemWithTooltip(weapon)} directly above the monster's head.
                            It whines, [b]begging[/b] you to spare its life.
                            But you can't do that.
                            It attacked you.
                            It tried to [b]kill you[/b].
                            If you let it live, someone else might fall prey to it.
                            It's a foul [b]beast[/b], and as an aspiring knight-to-be, it is your duty to rid the world of it.
                            Besides, it wouldn't survive out here with just one eye.
                            You're not taking its life, you're putting it out of its misery.
                        [/p]
                        [p]
                            Despite growing doubts inside of your mind, you decide to go through with your actions.
                            You face your head the other way, and in one fluid stroke, you
                            ${blunt ? `bash the creature's head in, its skull fracturing open with a gut-wrenching [b]crack[/b]` : `${slash ? `
                            slash the back of the creature's neck, severing its spine` : `sink your weapon directly into the creature's brain`}`}.
                            Its whining slowly dies down.
                            Its one remaining eye closes.
                            Its ragged breathing stops.
                            The basilisk is dead.
                        [/p]
                        [p]
                            [c]
                                [b]You have killed the basilisk![/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\succeed\UndergroundCaveOffenceWin.js

// scenes\HamletDay2\UndergroundCave\fight\offence\UndergroundCaveOffence.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveOffence(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveFightChoice.OFFENCE;
    var reqController = ctrl.fight;

    var toFail = new UndergroundCaveWrestling(ctrl);
    var toSuccess = new UndergroundCaveOffenceWin(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.weapon();
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\offence\UndergroundCaveOffence.js

// scenes\HamletDay2\UndergroundCave\fight\defence\eyepatch\UndergroundCaveDefenceLeave.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveDefenceLeave(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.defenceLeave;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            With the basilisk no longer around, you take a closer look at your aching hand.
                            One of your fingers is all kinds of [b]fucked up[/b].
                            While you don't think it [b]broke[/b], it's definitely [b]dislocated[/b].
                            You've had this happen once before, a few years back when you fell out of a tree.
                            There's only one way to fix this, and it isn't pretty.
                        [/p]
                        [p]
                            Dreading what is to come, you gingerly take hold of the awkwardly-angled digit.
                            It hurts just to [b]touch[/b] it.
                            Come on, ${name}, you have to be [b]brave[/b].
                            You need to fix this [b]right now[/b], or it'll only get worse.
                            You close your eyes.
                            You grit your teeth.
                            You can do this.
                            Just think about something else.
                        [/p]
                        [p]
                            [big]
                                [color=red][b]SNAP[/b][/color]
                            [/big]
                        [/p]
                        [p]
                            With one firm yank, you [b]tug[/b] your dislocated finger back into place,
                            your bones agonizingly [b]grinding[/b] against one another before coming to rest in their normal position.
                            You [b]scream[/b] out in pain, your shriek echoing throughout the cavern as you collapse to your knees.
                            Tears drip down onto the stone floor below.
                            It hurt [b]so[/b] much.
                            You are hyperventilating.
                            The sickening snapping sound it made is [b]burned[/b] into your mind.
                            You feel like you're about to throw up.
                        [/p]
                        [p]
                            Holding your fixed hand by the wrist, you inspect it anew through your teary vision.
                            Your finger seems to be in its correct position once more.
                            You can painfully move it, but you'd prefer to keep it still right now.
                            You swallow deeply.
                            Your breathing stabilizes.
                            The worst of the pain is over.
                        [/p]
                        [p]
                            You slowly rise to your feet.
                            While still hurting, you need to get out of here, before the monster comes back for a second round.
                            You start walking alongside the water, hoping it'll lead you to an exit.
                            You go as fast as you can without falling down, the terrain uneven and cluttered with various stalagmites and rocks.
                            Ever so often, you stop to test your finger, but never for more than a couple of seconds.
                        [/p]
                        [p]
                            Before long, daylight comes into view.
                            As you expected: it was a stale branch of the river.
                            Climbing out of the cave, you sit your weary butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                            You can move your finger again, somewhat, but it'll need a bit more time before it stops aching completely.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and ended up punching a basilisk in the face.
                            How many kobolds can say they did that?
                            Probably not a lot, since even your kind isn't dumb enough to hit a mythological rock monster without putting on a pair of steel gloves first.
                        [/p]
                        [p]
                            You've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                            And hopefully your finger'll have healed a bit more by then.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\defence\eyepatch\UndergroundCaveDefenceLeave.js

// scenes\HamletDay2\UndergroundCave\fight\defence\eyepatch\UndergroundCaveDefenceDoNot.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveDefenceDoNot(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveDefencePatchChoice.DO_NOT;
    var reqController = ctrl.defencePatch;
    var subController = ctrl.preContinue;
    var changeController = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveDefenceLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Do not take the eyepatch.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\defence\eyepatch\UndergroundCaveDefenceDoNot.js

// scenes\HamletDay2\UndergroundCave\fight\defence\eyepatch\UndergroundCaveDefenceTake.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveDefenceTake(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveDefencePatchChoice.TAKE;
    var reqController = ctrl.defencePatch;
    var subController = ctrl.preContinue;
    var changeController = ctrl.defencePatchTake;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveDefenceLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Take the eyepatch. No basilisk around.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\defence\eyepatch\UndergroundCaveDefenceTake.js

// scenes\HamletDay2\UndergroundCave\fight\defence\eyepatch\UndergroundCaveDefencePatch.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveDefencePatch(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.defence;
    var subController = ctrl.defencePatch;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new UndergroundCaveBasiTake(ctrl);
    var secondNext = new UndergroundCaveBasiDoNot(ctrl);

    var firstChoice = UndergroundCaveDefencePatchChoice.TAKE;
    var secondChoice = UndergroundCaveDefencePatchChoice.DO_NOT;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Stumble upon eyepatch. Take it or leave it. No basilisk.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Take it</button>`;

        var tooltip = global.parser.parse(`Take the eyepatch.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Do not</button>`;

        var tooltip = global.parser.parse(`Leave it.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var UndergroundCaveDefencePatchChoice = {
    TAKE: 1,
    DO_NOT: 2
};

// END scenes\HamletDay2\UndergroundCave\fight\defence\eyepatch\UndergroundCaveDefencePatch.js

// scenes\HamletDay2\UndergroundCave\fight\defence\UndergroundCaveDefence.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveDefence(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveFightChoice.DEFENCE;
    var reqController = ctrl.fight;
    var subController = ctrl.preContinue;
    var changeController = ctrl.defence;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveDefenceLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskFingerDislocated = true;
        var toReturn = `[p]
                            It's [i]only[/i] a feral lizard.
                            If you can dodge its swipes and bites, then surely it will tire out soon, right?
                            All you need to do is [b]not get hit[/b].
                            A basilisk relies on its targets to be [b]petrified[/b], so it shouldn't be able to hit you if you keep moving.
                            At least you hope so.
                            You [b]really[/b] hope so.
                        [/p]
                        [p]
                            While you're still mid-thought, the beast suddenly bursts out into a [b]sprint[/b] towards you.
                            You only [b]barely[/b] manage to pull your left leg back before its teeth sink into your scales, avoiding the lethal bite with a speedy side-step.
                            Angrily, the monster swipes at you with its claws.
                            You take a step backwards to avoid getting hit, ending up at the edge of the pool of liquid you crawled out of.
                            The creature's barrage of attacks [b]keep coming[/b].
                        [/p]
                        [p]
                            Determined to drive you into the water, it follows up with another attempt at biting you.
                            No longer able to retreat, you instead roll to the side,
                            using your innate flexibility and the freedom of motion granted by bipedal movement to stay one step ahead of the rock-clad crocodillian.
                        [/p]
                        [p]
                            Coming out of your roll, you lower into a defensive pose, and hold still for a moment.
                            You and the basilisk are panting,
                            but your water-side duel is not over yet.
                            You both have fight left in you.
                            Although you [b]wish[/b] it would [b]give up[/b] and [b]walk away[/b] already.
                            Surely a skinny kobold like yourself, with barely any meat to pick at, isn't worth a lengthy brawl, right?
                        [/p]
                        [p]
                            Apparently you're wrong.
                            The brief respite is broken.
                            The furious feral lashes at you, turning to the side to swing its rock-hard tail, using it as a weighty whip.
                            Seeing an opening, you tumble not backwards, but [b]forwards[/b],
                            elegantly jumping over the lizard's tail, surprising both it and yourself as you end up directly in front of its drooling maw.
                        [/p]
                        [p]
                            The basilisk is unsure how to react.
                            It could bite you, but it hesitates.
                            You, too, hadn't thought this far ahead.
                            Your mind freezing up, you do the first thing that pops into your brain.
                            You curl your hand into a fist, and bash bash the crocodillian's snout, outright [b]punching[/b] the front of the animal's muzzle.
                        [/p]
                        [p]
                            [big]
                                [color=red][b]CRACK[/b][/color]
                            [/big]
                        [/p]
                        [p]
                            [b]Ow[/b]!!
                            You clutch your hand.
                            It [b]hurts[/b].
                            It feels like you just [b]hit[/b] a [b]rock[/b]!
                            You think something [b]broke[/b], and it wasn't the basilisk's nose.
                        [/p]
                        [p]
                            Still, despite having barely scratched it, the beast recoils as though it's been struck by a mortal wound.
                            You must have frightened it or something.
                            Shown it that you [b]can[/b] and you [b]will[/b] hurt it.
                            Even if you end up hurting yourself [b]even more[/b].
                            You aren't afraid of fighting a pyrrhic battle to the death.
                            A fight in which [b]both[/b] sides lose.
                            And that scares the monster.
                            Or maybe it's just smart enough to know when to run away.
                            It hisses loudly, before turning tail and scampering off back to the hole it came from.
                        [/p]
                        [p]
                            [c]
                                [b]You have scared off the basilisk![/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\defence\UndergroundCaveDefence.js

// scenes\HamletDay2\UndergroundCave\fight\observe\UndergroundCaveObserve.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveObserve(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveFightChoice.OBSERVE;
    var reqController = ctrl.fight;
    var changeController = false;

    var continueId = arguments.callee.name + "ButtonOne";

    var idsToRemove = [
        "UndergroundCaveObserve",
        "UndergroundCaveFight",
        "UndergroundCaveResistSucceed",
        "UndergroundCaveResist",
        "UndergroundCaveStare",
        "UndergroundCaveAvert",
        "UndergroundCaveSurface"
    ];

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var petri = ctrl.stare.get();
        var toReturn = `[p]
                            You need to be well-aware of the slippery lizard's actions.
                            One wrong move, and you could very well end up injured or even [b]dead[/b].
                            Before committing to actual combat, you decide to [i]study[/i] the basilisk first.
                            If you know how it moves, you'll be better able to avoid its blows.
                            At least in theory.
                        [/p]
                        [p]
                            You keep some distance between the creature and you, taking a step back whenever it tries to close the gap.
                            While slow in general, you note that it can burst into short, speedy sprints.
                            You don't allow it close enough to test out your theory, but you imagine its attacks must be somewhat similar:
                            slow in general, with some lightning-fast shots mixed in.
                            While the former will probably [b]not[/b] be a major threat, it is latter that you fear the most.
                            A simple flick of its tail, for instance, could push you off balance, after which you'll be easy prey for the beast.
                        [/p]
                        [p]
                            While observing the basilisk, you find your sight drifting towards its eyes over and over again.
                            The creature's bright irises dance and twirl, like a smidgen of milk in a bowl of soup that's being constantly stirred.
                            It's [b]intriguing[/b].
                            [b]Hypnotizing[/b].
                            The entire cave feels like a pitch-black [b]void[/b] with the eyes at its very center.
                            While you know you shouldn't, you can't help [b]but[/b] gaze into the abyss.
                            Before long, you find yourself [i]lost[/i] in the alluring, mind-emptying void that is the creature's stare${petri ? ` yet again` : ``}.
                        [/p]`;

        if (petri) {
            toReturn += `
                        [p]
                            This time, instead of taking several minutes, you find yourself starting to petrify within [b]seconds[/b].
                            Feeling threatened, the basilisk must have upped its game.
                            Your mind [b]screams[/b] for you to look away, but your body fails to heed the mental beckoning,
                            your panicking thoughts drowned out by [b]emptiness[/b], brought about by the mere sight of the emerald eyes in front of you.
                            Instead of averting your gaze, you keep blankly staring as a rock-like texture overtakes your extremities.
                            Your limbs petrify, and before you know it, you are frozen in place for a second time.
                        [/p]`;
        } else {
            toReturn += ` 
                        [p]
                            Within seconds, you can start to feel your limbs stiffen, your scales rapidly turning to stone.
                            Your mind [b]screams[/b] for you to look away, but your body fails to heed the mental beckoning,
                            your panicking thoughts drowned out by [b]emptiness[/b], brought about by the mere sight of the emerald eyes in front of you.
                            Instead of averting your gaze, you keep blankly staring as a rock-like texture overtakes your extremities.
                            Your limbs petrify, and before you know it, you are frozen in place.
                        [/p]`;
        }

        toReturn += `   [p]
                            But you aren't out of your trance yet.
                            Even though the entire exterior of your body has already turned to stone, you simply keep staring.
                            You feel the petrification dig [b]deeper[/b], your very flesh becoming cold, hard rock.
                            The basilisk's eyes prove [i]comforting[/i], dulling your aching mind while your very bones fossilize.
                        [/p]
                        [p]
                            Right before the intrusive blight reaches your vital organs, the basilisk suddenly breaks eyecontact, and surges forwards, towards you.
                            No longer having the sweet, hypnotizing, relieving comfort of staring into its eyes, you find yourself [b]panicking[/b].
                            You can barely [b]breathe[/b], your rock-hard chest keeping your lungs permanently compressed.
                            Your heart has a hard time pumping, almost completely encased in stone.
                            Many of your veins have slipped shut.
                            Whatever feeling you still had in your hands and your feet soon fades, blood circulation entirely cut off.
                        [/p]
                        [p]
                            Luckily, your agony is short-lived.
                            The monster viciously [b]rakes[/b] at one of your petrified legs,
                            its incredibly sharp and powerful claws [b]easily[/b] tearing through the hard, yet brittle rock,
                            ripping out an entire chunk of stone, completely [b]severing[/b] your foot from the rest of your body.
                        [/p]
                        [p]
                            You feel no pain, yet your mind is flooded with [b]terror[/b].
                            An intense [b]fear[/b] of death washes over you, which only intesifies as you realize you [b]aren't[/b] [i]stable[/i] anymore.
                            With one foot no longer supporting you, and no way to readjust the other, your body slowly starts to topple to the side.
                            You feel sick to your stomach, not even able to utter a scream as you tumble to the ground, your fate sealed.
                            The last thing you hear, before losing consciousness, 
                            is the sound of your own body [b]shattering[/b] against the cavern floor,
                            breaking into a million pieces.
                        [/p]
                        [p]
                            As the debris of your stone-turned body settles down,
                            the basilisk immediately begins picking through it for the tender,
                            juicy innards that were trapped within the stone shell.
                            It's not every day that the beast could have such a feast.
                            While the organs are succulent,
                            the few small morsels of unpetrified flesh that the monster can find, sate it the most.
                            No wonder humans butcher kobolds.
                            Their meat really [b]is[/b] a delicacy, to feral and anthro alike.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    Rock end.
                                [/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" style="margin-left: auto; position: relative; float: right;" id="${continueId}">`;
        choice += `&lt;&lt Revert`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400, function () {
            var main = $("#main");
            main.scrollTop(main[0].scrollHeight - main.height() - 5); // Fix scrolling
        });
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.misc.unsetAll();
            ctrl.fight.unset();
            ctrl.resistSucceed.undo();
            ctrl.resistSucceed.unset();
            ctrl.stare.undo();
            ctrl.stare.unset();
            ctrl.avert.undo();
            ctrl.avert.unset();
            ctrl.surface.undo();
            ctrl.surface.unset();
            for (var i = 0; i < idsToRemove.length; i++) {
                $("#" + idsToRemove[i]).remove();
            }
            loadNextPart(ctrl.fragments.surface);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\fight\observe\UndergroundCaveObserve.js

// scenes\HamletDay2\UndergroundCave\fight\UndergroundCaveFight.js

/**
 * Fight section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveFight(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveAvertChoice.FIGHT;
    var reqController = ctrl.avert;
    var subController = ctrl.fight;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new UndergroundCaveDefence(ctrl);
    var secondNext = new UndergroundCaveOffence(ctrl);
    var thirdNext = new UndergroundCaveObserve(ctrl);

    var firstChoice = UndergroundCaveFightChoice.DEFENCE;
    var secondChoice = UndergroundCaveFightChoice.OFFENCE;
    var thirdChoice = UndergroundCaveFightChoice.OBSERVE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskFought = true;
        var toReturn = ``;

        if (ctrl.avert.get()) {
            toReturn += `
                        [p]
                            Whatever the beast's intentions, you [b]aren't[/b] going to be an easy prey.
                            Whether it wants to [b]gorge[/b] on your [b]flesh[/b], simply [b]kill you[/b] for intruding on its territory,
                            or if it has a [b]hungry brood[/b] somewhere in the hole it crawled out of, you won't go down without a fight.
                            You drop into a combat stance, and mentally prepare yourself for battle.
                            Noticing your hostility, the crocodillian hisses loudly, and assumes a threatening pose.
                            [i]It's on.[/i]
                        [/p]`;
        } else {
            toReturn += `
                        [p]
                            Pushing your mind to its very limit, you manage to move your second hand as well.
                            The basilisk, while a mindless creature, is visibly in fightened awe as you slowly rise to your feet,
                            a wave of green scales starting from each of your extremities, rolling over the stony grey that overtaken you,
                            washing it away like it was never even present.
                        [/p]
                        [p]
                            Able to move unhindered once again, you drop into a combat stance.
                            The creature hisses loudly, assuming a threatening pose.
                            The two of you have a score to settle.
                            And neither of you are going down without a fight.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Defence</button>`;

        var tooltip = global.parser.parse(`Make sure that thing doesn't hit you. Retalliate whenever possible.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Offence</button>`;

        var tooltip = global.parser.parse(`Kill it before it can kill you.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Observing</button>`;

        var tooltip = global.parser.parse(`Keep a very watchful eye on the basilisk. See if there's a pattern to its movements.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You focus on:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var UndergroundCaveFightChoice = {
    DEFENCE: 1,
    OFFENCE: 2,
    OBSERVE: 3
};

// END scenes\HamletDay2\UndergroundCave\fight\UndergroundCaveFight.js

// scenes\HamletDay2\UndergroundCave\avert\still\eyepatch\UndergroundCaveBasiLeave.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveBasiLeave(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.basiLeave;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var doggoChoice = ctrl.doggo.get();
        var treats = doggoChoice === ctrl.choices.doggo.FEED;
        var act = "petting";
        if (doggoChoice === ctrl.choices.doggo.FEED) {
            act = "feeding";
        }
        if(doggoChoice === ctrl.choices.doggo.FETCH) {
            act = "playing fetch with";
        }
        var toReturn = `[p]
                            After a few more minutes of walking, the two of you reach the entrance to the cave.
                            As you expected: the water was a stale branch of the river!
                            Climbing out of the cavern, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            Uttering an exhausted sigh, you realize your basilisk buddy didn't follow you out here.
                            Looking back at the cave, you can see the creature staring at you from the shade.
                            It tentatively pokes one of its legs out into the sunlight, but with a loud hiss, quickly retracts the limb once more.
                            You figure it can't stand the sun, or something like that.
                            The beast stares at you longingly, before letting out a sad whine.
                            Hanging its head, it turns around and ventures back into the cave.
                        [p]
                            You consider going after the basilisk for a [b]proper[/b] goodbye, but you figure this way is for the best.
                            You can't stick around this cave forever, and even if you [b]did[/b] wait until nightfall,
                            the guards of Varanar would [b]never[/b] let such a monster into the city, no matter how well-behaved it is.
                            Maybe you'll return with one day, with some ${treats ? `more` : ``} tasty treats for it to eat.
                        [/p]
                        [p]
                            Who are you kidding?
                            You can't even properly feed yourself, let alone a monster twice your weight.
                            You hope it'll be okay.
                            If a knight finds this cave, you fear your new friend won't last long.
                            But it's managed to survive until now, as have you.
                            With some luck, the both of you'll be just fine.
                        [/p]
                        [p]
                            The trip into the well was a harrowing experience.
                            You defied death more than once, but meeting the basilisk was all worth it.
                            You've never really owned a [i]pet[/i].
                            You wonder if that [i]hearty[/i] and [i]wholesome[/i] feeling you felt when ${act} the basilisk is how pet-owners feel all the time.
                        [/p]
                        [p]
                            Besides a friend, you've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            You find yourself close to the road to Varanar.
                            In fact, you believe you passed this very place earlier, when heading for the hamlet.
                            You never knew there was a cave here, but then again,
                            you've never really bothered to scour the edge of the river for points of interest.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            The capital isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\eyepatch\UndergroundCaveBasiLeave.js

// scenes\HamletDay2\UndergroundCave\avert\still\eyepatch\UndergroundCaveBasiDoNot.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveBasiDoNot(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveBasiPatchChoice.DO_NOT;
    var reqController = ctrl.basiPatch;
    var subController = ctrl.preContinue;
    var changeController = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveBasiLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            An eyepatch is useless to a kobold like you.
                            Your night vision is already as good as it's ever going to get.
                            While you kind of like the way this thing looks,
                            you're not a fan of losing [i]depth perception[/i].
                            You toss it back down to the ground, where you found it, and the two of you continue on your way.
                            You've no idea whether intentional or not, but your crocodillian friend steps on top of the piece of leather you just threw away,
                            the basilisk's sharp claws completely shredding the eyepatch to oblivion.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\eyepatch\UndergroundCaveBasiDoNot.js

// scenes\HamletDay2\UndergroundCave\avert\still\eyepatch\UndergroundCaveBasiTake.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveBasiTake(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveBasiPatchChoice.TAKE;
    var reqController = ctrl.basiPatch;
    var subController = ctrl.preContinue;
    var changeController = ctrl.basiTake;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveBasiLeave(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            An eyepatch is useless to a kobold like you.
                            Your night vision is already as good as it's ever going to get.
                            Still, you kind of like the way this thing looks.
                            You're not a fan of losing [i]depth perception[/i], though.
                            But it's so [i]stylish[/i]!
                            It'll make you look like a dashing rogue!
                        [/p]
                        [p]
                            After a bit of thinking, you come up with the perfect way to make use of the eyepatch's good looks without hampering your ability to see.
                            Tugging the strap taut, you pull it over your head.
                            Instead of having the piece of leather cover one of your eyes, you let it rest on your forehead,
                            like the diamond of a tiara, or the gemstone of a circlet.
                            Perfect!
                            Your crocodillian friend seems to nod his head up and down in approval,
                            although you're pretty sure he was just scratching his chin against the ground instead of [i]actually[/i] telling you he likes your new look.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\eyepatch\UndergroundCaveBasiTake.js

// scenes\HamletDay2\UndergroundCave\avert\still\eyepatch\UndergroundCaveBasiPatch.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveBasiPatch(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.prePatch;
    var subController = ctrl.basiPatch;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new UndergroundCaveBasiTake(ctrl);
    var secondNext = new UndergroundCaveBasiDoNot(ctrl);

    var firstChoice = UndergroundCaveBasiPatchChoice.TAKE;
    var secondChoice = UndergroundCaveBasiPatchChoice.DO_NOT;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var headrubs = ctrl.doggo.get() === ctrl.choices.doggo.PET;
        var toReturn = `[p]
                            With the basilisk befriended, you decide it's time to leave this place.
                            Signalling for the beast to follow you via a simple gesture,
                            you begin to follow the water, to what you hope will be your freedom.
                            The creature seems to understand your intentions.
                            It happily crawls alongside you, like a pet dog being taken for a walk.
                        [/p]
                        [p]
                            You spend a few minutes walking,
                            your gaze shifting between the bumpy ground in front of you,
                            and your pet basilisk at your side.
                            Suddenly, you notice something on the ground.
                            A small, black patch, with a strap attached to it.
                            You reach down to pick it up, your crocodillian friend getting excited,
                            mistakenly thinking it's about to get its head rubbed${headrubs ? ` again` : ``}.
                            You can almost feel the beast's disappointment when you, instead of patting it,
                            pick up an old ${global.kobold.inventory.createItemWithTooltip(new Eyepatch)} off the ground.
                        [/p]
                        [p]
                            You inspect the item for a moment.
                            Odd.
                            This cave must have been used by river-smugglers in the past or something.
                            You'd search around for hidden treasures or stashed booties, but you feel like you've had enough adventure for now.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Take it</button>`;

        var tooltip = global.parser.parse(`Put on the eyepatch.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Do not</button>`;

        var tooltip = global.parser.parse(`Leave it.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var UndergroundCaveBasiPatchChoice = {
    TAKE: 1,
    DO_NOT: 2
};

// END scenes\HamletDay2\UndergroundCave\avert\still\eyepatch\UndergroundCaveBasiPatch.js

// scenes\HamletDay2\UndergroundCave\avert\still\pet\UndergroundCaveScratch.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveScratch(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCavePetChoice.SCRATCH;
    var reqController = ctrl.pet;
    var subController = ctrl.prePatch;
    var changeController = ctrl.scratch;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveBasiPatch(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            While you give the beast one more pat on its head, you move your other hand towards its chin.
                            Partly with your claws, partly with your fingers, and partly with the palm of your hand,
                            you begin to gently rub the bottom of the fearsome creature's maw.
                            It [b]definitely[/b] likes this, the crocodillian's incessant, cat-like purring intensifying.
                            It lifts its head just a little bit higher, its maw opening slightly out of sheer delight.
                        [/p]
                        [p]
                            You intensively pet the basilisk's head, while continually scratching underneath its muzzle,
                            giving it the most thorough rubbing you bet the beast has ever had!
                            After nearly an entire minute of working your hands along its rocky scales, you finally relent.
                            Pulling back, the creature keeps its head lifted, seemingly wanting more.
                            A few moments pass, and then its eyes open once more.
                            Realizing it's not being pet anymore, the basilisk submissively lies down at your feet.
                        [/p]
                        [p]
                            [c]
                                [b]You have befriended the basilisk![/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\pet\UndergroundCaveScratch.js

// scenes\HamletDay2\UndergroundCave\avert\still\pet\UndergroundCaveStrikeLoot.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveStrikeLoot(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.throatLoot;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You did it!
                            You killed a deadly monster, all by yourself!
                            Just like a knight would do!
                            Somehow, you don't feel very [i]knightly[/i], though.
                            You should be proud of yourself, but for some reason, you aren't.
                            Maybe your actions weren't as [i]chivalrous[/i] as you would have liked,
                            but it's the outcome that matters, and not how you achieved it, right?
                            The ends justify the means.
                        [/p]
                        [p]
                            But then again, the basilisk was just minding its own business.
                            It wasn't hurting anyone.
                            It rubbed up against you and even let you pat its head.
                            Now you're confused.
                            Did you do the right thing?
                        [/p]
                        [p]
                            You honestly don't know anymore.
                            Does it even matter?
                            What's done is done.
                            If you hadn't killed the creature, some pompous knight in shining armor
                            -- with nothing left to prove, neither to himself, nor to the world --
                            would have strolled into this cave to add another tally to his already impressive list of slain monsters.
                            For someone properly armed and armored, this would have been a [b]chore[/b], a menial task barely even worthy of mention.
                            But for you, this is a feat you'll remember until the day you die.
                        [/p]
                        [p]
                            Your conviction bolstered, you wait for a few more minutes to make sure it's really [b]dead[/b], before approaching the giant lizard once again.
                            Kneeling besides it, you figure it's time to take a trophy.
                            Something to remember this day by.
                            Something you can show to people, so they'll believe you when you tell them you singlehandedly killed a basilisk.
                            But what?
                        [/p]
                        [p]
                            While pondering, you run one of your hands over the deceased basilisk's back.
                            Some of its scales seem to be rather [i]loose[/i].
                            You take hold of one with both of your hands, and with surprising ease, you manage to [b]tear[/b] the scale from the monster's backside.
                            Like plucking a chicken.
                            You repeat the process a couple of times,
                            until you have collected a suitable handful of ${global.kobold.inventory.createItemWithTooltip(new BasiliskScales())}
                            to take with as your prize for ridding the world of this horrible monstrosity.
                        [/p]
                        [p]
                            Your trophies collected, you decide it's time to find a way out of the cave.
                            Following the river-branch should lead you to some kind of exit.
                            Without the threat of the basilisk around, you don't even need to [b]swim[/b].
                            You leisurely stroll alongside the water, until daylight comes into sight.
                            As you expected: the water was a stale branch of the river!
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and after all of that, you managed to kill a basilisk.
                            It might not have been the most [i]honorable[/i] way of slaying the beast, but it was slain nonetheless.
                        [/p]
                        [p]
                            Besides the ${global.kobold.inventory.createItemWithTooltip(new BasiliskScales())} you [i]harvested[/i],
                            you've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\pet\UndergroundCaveStrikeLoot.js

// scenes\HamletDay2\UndergroundCave\avert\still\pet\UndergroundCaveStrike.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveStrike(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCavePetChoice.THROAT;
    var reqController = ctrl.pet;
    var subController = ctrl.preContinue;
    var changeController = ctrl.throat;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveStrikeLoot(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskKilled = true;
        var toReturn = `[p]
                            Seeing an opportunity for greatness, you decide to act on it.
                            While you give the creature one more pat on its head, you move your other hand onto its throat.
                            Expecting more pleasant rubbing,
                            the basilisk's knows not what to do as you instead sink your sharp claws into its tender throat,
                            and rake open its windpipe in a single vicious [b]tear[/b].
                        [/p]
                        [p]
                            The monster's eyes shoot open wide.
                            It wildly thrashes its limbs around, hoping to mangle you, to take you down along with it,
                            but you manage to avoid most of its rage by taking a quick step back.
                            While it continues to blindly rampage about, despite you having long cleared its imminent vicinity,
                            the basilisk's struggles soon die down.
                            Running out of air, its body slumps lifelessly down on the ground, in a pool of its own blood.
                            After one last, bloody gurgle, the big crocodillian closes its eyes, and drifts away into the afterlife.
                        [/p]
                        [p]
                            [c]
                                [b]You have killed the basilisk![/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\pet\UndergroundCaveStrike.js

// scenes\HamletDay2\UndergroundCave\avert\still\pet\UndergroundCavePet.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCavePet(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = ctrl.choices.doggo.PET;
    var reqController = ctrl.doggo;
    var subController = ctrl.pet;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new UndergroundCaveStrike(ctrl);
    var secondNext = new UndergroundCaveScratch(ctrl);

    var firstChoice = UndergroundCavePetChoice.THROAT;
    var secondChoice = UndergroundCavePetChoice.SCRATCH;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Still wary of the beast, you decide to try something.
                            You wait for the basilisk to finish circling around you, then you slowly reach for its head,
                            calmly laying your palm down atop of the hard, stone-like surface that constitutes his forehead.
                            The creature stops dead in its tracks, closing its eyes and uttering a deep rumbling noise remotely reminiscent of a [i]purr[/i].
                        [/p]
                        [p]
                            Taking the throaty sound as a request for more, you rub your hand along the top of the basilisk's head.
                            It [b]definitely[/b] seems to be enjoying this.
                            One of its hind legs starts to scrape against the floor, like a thumping little rabbit.
                            You go in for a second stroke, and then a third.
                            Fully relaxed, the big stony lizard raises its head, exposing the most fragile part of its entire body for some affectionate attention.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Strike throat</button>`;

        var tooltip = global.parser.parse(`The beast has exposed a weak spot. If you strike it now, you might be able to fell it in one swift blow.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Scratch chin</button>`;

        var tooltip = global.parser.parse(`Rub the basilisk's chin.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var UndergroundCavePetChoice = {
    THROAT: 1,
    SCRATCH: 2
};

// END scenes\HamletDay2\UndergroundCave\avert\still\pet\UndergroundCavePet.js

// scenes\HamletDay2\UndergroundCave\avert\still\fetch\UndergroundCaveFetch.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveFetch(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = ctrl.choices.doggo.FETCH;
    var reqController = ctrl.doggo;
    var subController = ctrl.prePatch;
    var changeController = ctrl.fetch;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveBasiPatch(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var branch = ctrl.misc.weapon();
        var toReturn = `[p]
                            Still wary of the beast, you decide to try something.
                            You slowly reach for your ${global.kobold.inventory.createItemWithTooltip(branch)} while the basilisk circles around you.
                            As it reaches your front-end once more, you prudently wave the stick in front of its face.
                            The creature seems captivated, its eyes following the branch's every move.
                            Having captured its attention, you throw your ${global.kobold.inventory.createItemWithTooltip(branch)} as far as you can.
                        [/p]
                        [p]
                            The basilisk instantly gives chase, bursting out into a sprint,
                            moving far faster than you'd have imagined a beast of its [i]size[/i] to be capable of,
                            let alone one completely covered in weighty stone scales.
                            What an odd creature.
                            It [i]acts[/i] like a mixture between a dog and a cat,
                            but it [b]looks[/b] like an oversized alligator with a bunch of aesthetically-shaped stones attached to its back.
                        [/p]
                        [p]
                            Arriving at the stick, the massive lizard comes to a halt, its massive talons digging into the rock-like ground itself to slow it down.
                            Curious, you curl your toes in an attempt to mimic its motions,
                            yet your own claws can barely even [b]scratch[/b] the stone surface you're standing on.
                            It must be [b]really[/b] strong.
                            Or maybe it has razor-sharp talons.
                            Or both.
                        [/p]
                        [p]
                            The basilisk, having stopped directly in front of the branch you threw, is [i]staring[/i] at it.
                            At first, you wonder what the crocodilian is doing, but then you notice the stick slowly starting to change.
                            Its color changes from tree-brown to a deep grey, while simultaneously its texture visibly alters from vibrant wood to dull stone.
                            The creature is petrifying your weapon!
                        [/p]
                        [p]
                            The branch fully turned to stone, the lizard takes in its mouth, and happily carries it back over to you,
                            its tail wagging excitedly from side to side while dragging along the rock-floor.
                            You gingerly take your new, [i]enhanced[/i] club from the returning basilisk's mouth.
                            While a bit more [i]clumsy[/i] to wield,
                            you suppose the combined power of wood and stone make for a far more capable weapon than what you had before.
                            Once it realizes you don't intend on throwing the now-far-too-heavy stick a second time,
                            the basilisk submissively lies down at your feet.
                        [/p]
                        [p]
                            [c]
                                [b]You have befriended the basilisk![/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\fetch\UndergroundCaveFetch.js

// scenes\HamletDay2\UndergroundCave\avert\still\feed\UndergroundCaveFeed.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveFeed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = ctrl.choices.doggo.FEED;
    var reqController = ctrl.doggo;
    var subController = ctrl.prePatch;
    var changeController = ctrl.feed;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveBasiPatch(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Still wary of the beast, you decide to try something.
                            You slowly reach for your ${global.kobold.inventory.createItemWithTooltip(new HeartyMeal())} while the basilisk circles around you.
                            As it reaches your front-end once more, you prudently set the basket down on the floor.
                            The creature seems captivated, its tongue rapidly flicking in and out of its mouth.
                            It turns its head towards the basket.
                            It can [b]smell[/b] the food.
                        [/p]
                        [p]
                            Having captured the crocodilian's attention, you open the lid of the wooden container.
                            The beast immediately [b]pounces[/b] on the basket, [b]ramming[/b] its head inside and [b]feasting[/b] on the soggy contents within,
                            like it's been starved for [b]weeks[/b].
                            Then again, maybe this is just how basilisks eat.
                            From what you can gather, they are [b]fierce[/b] creatures.
                            If it really was hungry, you have no doubt it would be eating [b]you[/b] right now.
                        [/p]
                        [p]
                            The contents of the ${global.kobold.inventory.createItemWithTooltip(new HeartyMeal())} are devoured wholly in less than a minute,
                            the monster [b]gorging[/b] and [b]feasting[/b] on them like there's no tomorrow.
                            At one point, you even fear it's going to eat the basket itself, the loud crunching of wood reverberating across the cavern.
                            Luckily, the beast seems smart enough to realize that's [b]not[/b] part of its meal.
                            It must have [i]accidentally[/i] chomped it a little.
                            Well, as long as it doesn't [i]accidentally[/i] chomp [b]you[/b], you're fine with that.
                        [/p]
                        [p]
                            Once finished with its meal, the basilisk retracts its face from the basket.
                            It looks up at you, as though begging for more, its forked tongue picking off some of the scraps clinging to the gaps in between his teeth.
                            Realizing you're out of food, it shoves the slightly damaged wooden basket to the side.
                            Taking the food container's place, the beast submissively lays itself down at your feet.
                        [/p]
                        [p]
                            [c]
                                [b]You have befriended the basilisk![/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\feed\UndergroundCaveFeed.js

// scenes\HamletDay2\UndergroundCave\avert\still\UndergroundCaveLick.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveLick(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.doggo;
    var changeController = ctrl.lick;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new UndergroundCavePet(ctrl);
    var secondNext = new UndergroundCaveFetch(ctrl);
    var thirdNext = new UndergroundCaveFeed(ctrl);

    var firstChoice = ctrl.choices.doggo.PET;
    var secondChoice = ctrl.choices.doggo.FETCH;
    var thirdChoice = ctrl.choices.doggo.FEED;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskFriend = true;
        var shirt = ctrl.misc.shirt();
        var toReturn = `[p]
                            You stand perfectly still, hoping to make yourself invisible to the beast.
                            At first, your strategy appears to be working.
                            The basilisk looks off to the left, then to the right.
                            Its eyes return to you, but you imagine it's just gazing at where it [i]last saw you[/i] or something.
                            The lizard may be leering directly at you, but that doesn't mean it can [b]see you[/b], right?
                            Clinging to the hope that remaining motionless will prevent the monster from noticing you,
                            you refuse to move an inch even as it crawls closer.
                            You tremble in fright,
                            knowing fully well that the creature approaching you is more than capable of [b]ripping your throat out[/b] should it judge you a threat.
                        [/p]
                        [p]
                            The basilisk hisses, sending shivers of fear down your spine.
                            Part of your brain wants to [b]run[/b], to flee and hope it won't be able to catch up to you.
                            But it's already too late.
                            The beast is too near now.
                            If you turn tail, it'll [b]snap[/b] one of your legs in between its powerful jaws and
                            [b]drag[/b] you down to the ground before [b]tearing your guts out[/b].
                            You've chosen your strategy, and no matter how much you regret it, you can't back out now.
                        [/p]
                        [p]
                            The crocodillian creeps closer and closer, until it's directly in front of you.
                            It tilts its head upwards, its eyes staring into yours.
                            Still, you do not move.
                            The basilisk extends its front legs, lifting its upper body off the ground, raising its head to about your waist-height.
                            You close your eyes, fearing your end is near, thinking the monster'll bite down
                            and [b]mangle[/b] your delicate belly${shirt ? ` through your ${global.kobold.inventory.createItemWithTooltip(shirt)}` : ``}.
                        [/p]
                        [p]
                            Instead of the [b]horrible pain[/b] you were anticipating, suddenly, you feel a slight [i]bump[/i] against your right hand.
                            You open your eyes, and quizzically stare down to find the rock-scaled beast pressing its stony snout into your palm,
                            affectionately nuzzling your soft scales like a friendly, oversized cat.
                            Absolutely dumbfounded, you watch as the creature ceases in its attempts to get you to pet it,
                            and instead lowers its muzzle to your thigh, hovering in front of the painful cut you sustained while diving in the well.
                        [/p]
                        [p]
                            After a couple of loud sniffs, the basilisk's maw opens, and its forked, snake-like tongue rolls out.
                            You wince as the monster laps at your injury, seemingly [b]tasting[/b] your [b]blood[/b].
                            It stings a bit, at first, but the more the crocodillian licks, the less it hurts.
                            Its saliva is seemingly mending your wound, patching the gash by slightly petrifiying your flesh.
                            Once the laceration is patched, the gentle creature steps besides you to rub its coarse, scaled side against your leg,
                            with a force strong enough to nearly [b]topple[/b] you over.
                        [/p]
                        [p]
                            The basilisk lets out a pleased, rumbling growl while it uses your leg as a scratching post,
                            grinding its rock-hard body against your scales.
                            You're beginning to think this beast isn't as much of a threat as you initially believed it to be.
                            While this might all still be some kind of [i]trick[/i], you're starting to think the lizard is [i]friendly[/i].
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Pet it</button>`;

        var tooltip = global.parser.parse(`Rub your hand along the basilisk's stony carapace.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var weapon = ctrl.misc.weapon();
        var req = weapon && weapon.id === ItemEnum.BRANCH;
        var disabled = subController.has() || !req ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Play fetch</button>`;

        var tooltip = global.parser.parse(`Throw your branch and hope the basilisk will retrieve it for you.`);
        if (!req) {
            tooltip = global.parser.parse(`You don't have a stick to play fetch with.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var req = ctrl.misc.hasConsumable(ItemEnum.HEARTY_MEAL);
        var disabled = subController.has() || !req ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Feed it</button>`;

        var tooltip = global.parser.parse(`Give the basilisk some food.`);
        if (!req) {
            tooltip = global.parser.parse(`You don't have any food that could interest a basilisk.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\UndergroundCaveLick.js

// scenes\HamletDay2\UndergroundCave\avert\still\UndergroundCaveNuzzle.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveNuzzle(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.doggo;
    var changeController = ctrl.nuzzle;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new UndergroundCavePet(ctrl);
    var secondNext = new UndergroundCaveFetch(ctrl);
    var thirdNext = new UndergroundCaveFeed(ctrl);

    var firstChoice = ctrl.choices.doggo.PET;
    var secondChoice = ctrl.choices.doggo.FETCH;
    var thirdChoice = ctrl.choices.doggo.FEED;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskFriend = true;
        var shirt = ctrl.misc.shirt();
        var pants = ctrl.misc.pants();
        var toReturn = `[p]
                            You stand perfectly still, hoping to make yourself invisible to the beast.
                            At first, your strategy appears to be working.
                            The basilisk looks off to the left, then to the right.
                            Its eyes return to you, but you imagine it's just gazing at where it [i]last saw you[/i] or something.
                            The lizard may be leering directly at you, but that doesn't mean it can [b]see you[/b], right?
                            Clinging to the hope that remaining motionless will prevent the monster from noticing you,
                            you refuse to move an inch even as it crawls closer.
                            You tremble in fright,
                            knowing fully well that the creature approaching you is more than capable of [b]ripping your throat out[/b] should it judge you a threat.
                        [/p]
                        [p]
                            The basilisk hisses, sending shivers of fear down your spine.
                            Part of your brain wants to [b]run[/b], to flee and hope it won't be able to catch up to you.
                            But it's already too late.
                            The beast is too near now.
                            If you turn tail, it'll [b]snap[/b] one of your legs in between its powerful jaws and
                            [b]drag[/b] you down to the ground before [b]tearing your guts out[/b].
                            You've chosen your strategy, and no matter how much you regret it, you can't back out now.
                        [/p]
                        [p]
                            The crocodillian creeps closer and closer, until it's directly in front of you.
                            It tilts its head upwards, its eyes staring into yours.
                            Still, you do not move.
                            The basilisk extends its front legs, lifting its upper body off the ground, raising its head to about your waist-height.
                            You close your eyes, fearing your end is near, thinking the monster'll bite down
                            and [b]mangle[/b] your delicate belly${shirt ? ` through your ${global.kobold.inventory.createItemWithTooltip(shirt)}` : ``}.
                        [/p]
                        [p]
                            Instead of the [b]horrible pain[/b] you were anticipating, suddenly, you feel a slight [i]bump[/i] against your right hand.
                            You open your eyes, and quizzically stare down to find the rock-scaled beast pressing its stony snout into your palm,
                            affectionately nuzzling your soft scales like a friendly, oversized cat.
                            Absolutely dumbfounded, you watch as the creature ceases in its attempts to get you to pet it,
                            and instead steps besides you to rub its coarse, scaled side against your ${pants ? `exposed` : ``} leg,
                            with a force strong enough to nearly [b]topple[/b] you over.
                        [/p]
                        [p]
                            The basilisk lets out a pleased, rumbling growl while it uses your leg as a scratching post,
                            grinding its rock-like body against your ${pants ? `${global.kobold.inventory.createItemWithTooltip(pants)}` : `scales`}.
                            You're beginning to think this beast isn't as much of a threat as you initially believed it to be.
                            While this might all still be some kind of [i]trick[/i], you're starting to think the lizard is [i]friendly[/i].
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Pet it</button>`;

        var tooltip = global.parser.parse(`Rub your hand along the basilisk's stony carapace.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var weapon = ctrl.misc.weapon();
        var req = weapon && weapon.id === ItemEnum.BRANCH;
        var disabled = subController.has() || !req ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Play fetch</button>`;

        var tooltip = global.parser.parse(`Throw your branch and hope the basilisk will retrieve it for you.`);
        if (!req) {
            tooltip = global.parser.parse(`You don't have a stick to play fetch with.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var req = ctrl.misc.hasConsumable(ItemEnum.HEARTY_MEAL);
        var disabled = subController.has() || !req ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Feed it</button>`;

        var tooltip = global.parser.parse(`Give the basilisk some food.`);
        if (!req) {
            tooltip = global.parser.parse(`You don't have any food that could interest a basilisk.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\UndergroundCaveNuzzle.js

// scenes\HamletDay2\UndergroundCave\avert\still\UndergroundCaveStandStill.js

/**
 * Doggo section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveStandStill(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveAvertChoice.STAND;
    var reqController = ctrl.avert;

    var toFail = new UndergroundCaveNuzzle(ctrl);
    var toSuccess = new UndergroundCaveLick(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return global.kobold.history.shared.well.scrapedThigh;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\avert\still\UndergroundCaveStandStill.js

// scenes\HamletDay2\UndergroundCave\avert\UndergroundCaveAvert.js

/**
 * Avert section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveAvert(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveSurfaceChoice.AVERT;
    var reqController = ctrl.surface;
    var subController = ctrl.avert;
    var changeController = ctrl.avert;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new UndergroundCaveStandStill(ctrl);
    var secondNext = ctrl.fragments.fight;

    var firstChoice = UndergroundCaveAvertChoice.STAND;
    var secondChoice = UndergroundCaveAvertChoice.FIGHT;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskMet = true;
        var toReturn = `[p]
                            You may be a kobold, but even [i]you[/i] know that blankly staring into a monster's eyes is a bad idea.
                            It might take it as a sign of aggression, a challenge to its reign over this cave.
                            You wisely avert your gaze, occasionally glancing back out of the corners of your eyes to prevent any nasty [i]surprises[/i].
                            After about half a minute of you not looking at it, the creature begins to [i]move[/i].
                            Its bright green eyes shift closer towards you as it crawls out of the dark hole it was concealed in.
                            Stepping out into the cavern proper, the monster is revealed.
                        [/p]
                        [p]
                            Before you stands a feral, quadrupedal lizard with large, protruding, rock-like scales and a sizeable maw filled with sharp, pointy teeth.
                            A crocodilian with a stony hide.
                            You think Seki once told you a story about a beast like this.
                            Could it be...[b]a basilisk[/b]?!
                        [/p]
                        [p]
                            Startled, you rise to your feet.
                            Thank goodness you decided to look away from it!
                            If Seki's tales are to be trusted, these creatures have the ability to petrify all those who look at them!
                            No wonder it has such [i]enchanting[/i] eyes.
                            Mere trickery, to lure innocent saps like yourself to get lost in their unrivalled depth while slowly turning to stone.
                        [/p]
                        [p]
                            The basilisk flicks out its tongue.
                            You [i]think[/i] that's how they [i]smell[/i], but you aren't exactly sure.
                            You should have paid more attention when the old bird was recounting his tales!
                            Who knew that his life stories could one day come in handy?
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Stand still</button>`;

        var tooltip = global.parser.parse(`Maybe basilisks can't see things that aren't moving?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Fight</button>`;

        var tooltip = global.parser.parse(`If this thing wants a fight, it can bring it on.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.imageBackground.hide();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var UndergroundCaveAvertChoice = {
    STAND: 1,
    FIGHT: 2
};

// END scenes\HamletDay2\UndergroundCave\avert\UndergroundCaveAvert.js

// scenes\HamletDay2\UndergroundCave\stare\petrified\UndergroundCaveRegain.js

/**
 * Stare section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveRegain(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.regain;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    function buildEnjoyed() {
        var dick = ctrl.misc.dick();
        var shirt = ctrl.misc.shirt();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;
        var toReturn = ``;
        if (dick) {
            if (clothingMale) {
                toReturn += `
                        [p]
                            Your brain [b]flooded[/b] with the strong smell of basilisk cock,
                            your nostrils inhaling more with every breath you take,
                            you can't help [b]but[/b] imagine the various ways the beast could have taken you,
                            were you not turned to stone.
                            Your dirty thoughts keep your trapped member at attention, and your mind occupied, while your limbs slowly return to normal.
                        [/p]
                        [p]
                            About half a lust-filled hour later, you finally regain some form of control over your arms.
                            The stony texture that had overtaken your limbs has given way to smooth scales once more.
                            Alas, it's already too late to sate your desire.
                            Most of the once-pungent scent of the basilisk's arousal has faded, much like the beast itself.
                            Along with it, went your cock's pheromone-fueled need for attention.
                            The once steamy puddle of cum below you is now but a mere [i]stain[/i] on the cavern's stony floor.
                        [/p]
                        [p]
                            Your muzzle in working order again, you sigh loudly in sexual frustration.
                            You aren't even going to [b]try[/b] to rub one out.
                            That could have been [b]so[/b] much fun, 
                            if only you'd remembered to take off your ${global.kobold.inventory.createItemWithTooltip(clothingMale)} first.
                            But then again, you didn't even [b]know[/b] you were being petrified, let alone that the beast wanted to [b]fuck[/b].
                            While you can't really blame yourself for what happened, you are still left somewhat upset.
                        [/p]
                        [p]
                            Your extremities still stiff, you roll onto your side, away from the basilisk's [i]fluids[/i].
                            You mindlessly stare up at the stalactite-laden roof of the cave, while you stretch your limbs every which way.
                            Understandably, you feel somewhat more [i]sluggish[/i] than you did before this entire ordeal.
                            You think your scales might have permanently [i]hardened[/i] a bit.
                            Once you feel like you're fit enough to walk, you slowly rise to your feet, and start making your way [i]out[/i] of the damp cave.
                        [/p]
                        [p]
                            You follow the water, hoping it'll lead you to some kind of exit.
                            Your legs still somewhat unresponsive, the journey is slow, and you nearly fall down more than once.
                            As you go, your stride improves, going from a half-limp to a decent walk alongside the water, until daylight comes into sight.
                            As you expected: it was a stale branch of the river.
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and missed a once-in-a-lifetime opportunity for some steamy basilisk sex.
                            A shame, but at least you made it out alive.
                        [/p]
                        [p]
                            You've gained a couple of memories you'll undoubtedly be masturbating to later,
                            along with a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            Your brain [b]flooded[/b] with the strong smell of basilisk jizz,
                            your nostrils inhaling more with every breath you take,
                            you can't help [b]but[/b] imagine the various ways the beast could have taken you,
                            were you not turned to stone.
                            Your dirty thoughts keep your trapped member at attention, and your mind occupied, while your limbs slowly return to normal.
                        [/p]
                        [p]
                            About half a lust-filled hour later, you finally regain some form of control over your arms.
                            The stony texture that had overtaken your limbs has given way to smooth scales once more.
                            Alas, it's already too late to sate your desire.
                            Most of the once-pungent scent of the basilisk's arousal has faded, much like the beast itself.
                            Along with it, went your cock's pheromone-fueled need for attention.
                            Your tailhole gapes no longer, and your insides feel even more empty than [i]before[/i] the intense romp.
                        [/p]
                        [p]
                            Your muzzle in working order again, you sigh loudly in sexual frustration.
                            You aren't even going to [b]try[/b] to rub one out.
                            You'd have cummed your brains out during the basilisk's pounding, if your cockslit hadn't been petrified shut.
                            A missed opportunity, for sure, but despite not reaching your end, you feel like this was a memorable experience regardless.
                        [/p]
                        [p]
                            Your extremities still stiff, you roll onto your side, away from the basilisk's [i]fluids[/i].
                            You mindlessly stare up at the stalactite-laden roof of the cave, while you stretch your limbs every which way.
                            Understandably, you feel somewhat more [i]sluggish[/i] than you did before this entire ordeal.
                            You think your scales might have permanently [i]hardened[/i] a bit.
                            Once you feel like you're fit enough to walk, you slowly rise to your feet, and start making your way [i]out[/i] of the damp cave.
                        [/p]
                        [p]
                            You follow the water, hoping it'll lead you to some kind of exit.
                            Your legs still somewhat unresponsive, the journey is slow, and you nearly fall down more than once.
                            As you go, your stride improves, going from a half-limp to a decent walk alongside the water, until daylight comes into sight.
                            As you expected: it was a stale branch of the river.
                            Climbing out of the cave, you sit your sore butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and got some steamy basilisk sex, although you didn't quite manage to get off to it.
                            A shame, but at least you made it out alive.
                        [/p]
                        [p]
                            You've gained a couple of memories you'll undoubtedly be masturbating to later,
                            along with a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;
            }
        } else {
            if (clothingFemale) {
                toReturn += `
                        [p]
                            Your brain [b]flooded[/b] with the strong smell of basilisk cock,
                            your nostrils inhaling more with every breath you take,
                            you can't help [b]but[/b] imagine the various ways the beast could have taken you,
                            were you not turned to stone.
                            Your dirty thoughts keep your loins burning with desire, and your mind occupied, while your limbs slowly return to normal.
                        [/p]
                        [p]
                            About half a lust-filled hour later, you finally regain some form of control over your arms.
                            The stony texture that had overtaken your limbs has given way to smooth scales once more.
                            Alas, it's already too late to sate your desire.
                            Most of the once-pungent scent of the basilisk's arousal has faded, much like the beast itself.
                            Along with it, went your slit's pheromone-fueled need for attention.
                            The once steamy puddle of cum below you is now but a mere [i]stain[/i] on the cavern's stony floor.
                        [/p]
                        [p]
                            Your muzzle in working order again, you sigh loudly in sexual frustration.
                            You aren't even going to [b]try[/b] to rub one out.
                            That could have been [b]so[/b] much fun, 
                            if only you'd remembered to take off your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)} first.
                            But then again, you didn't even [b]know[/b] you were being petrified, let alone that the beast wanted to [b]fuck[/b].
                            While you can't really blame yourself for what happened, you are still left somewhat upset.
                        [/p]
                        [p]
                            Your extremities still stiff, you roll onto your side, away from the basilisk's [i]fluids[/i].
                            You mindlessly stare up at the stalactite-laden roof of the cave, while you stretch your limbs every which way.
                            Understandably, you feel somewhat more [i]sluggish[/i] than you did before this entire ordeal.
                            You think your scales might have permanently [i]hardened[/i] a bit.
                            Once you feel like you're fit enough to walk, you slowly rise to your feet, and start making your way [i]out[/i] of the damp cave.
                        [/p]
                        [p]
                            You follow the water, hoping it'll lead you to some kind of exit.
                            Your legs still somewhat unresponsive, the journey is slow, and you nearly fall down more than once.
                            As you go, your stride improves, going from a half-limp to a decent walk alongside the water, until daylight comes into sight.
                            As you expected: it was a stale branch of the river.
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and missed a once-in-a-lifetime opportunity for some steamy basilisk sex.
                            A shame, but at least you made it out alive.
                        [/p]
                        [p]
                            You've gained a couple of memories you'll undoubtedly be masturbating to later,
                            along with a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            Your brain [b]flooded[/b] with the strong smell of basilisk jizz,
                            your nostrils inhaling more with every breath you take,
                            you can't help [b]but[/b] imagine the various ways the beast could have taken you,
                            were you not turned to stone.
                            Your dirty thoughts keep your cunt wet, and your mind occupied, while your limbs slowly return to normal.
                        [/p]
                        [p]
                            About half a lust-filled hour later, you finally regain some form of control over your arms.
                            The stony texture that had overtaken your limbs has given way to smooth scales once more.
                            Alas, it's already too late to sate your desire.
                            Most of the once-pungent scent of the basilisk's arousal has faded, much like the beast itself.
                            Along with it, went your twat's pheromone-fueled need for attention.
                            Your slit gapes no longer, and your insides feel even more empty than [i]before[/i] the intense romp.
                        [/p]
                        [p]
                            Your muzzle in working order again, you sigh loudly in sexual frustration.
                            You came [i]once[/i], but had the beast had the stamina for just a minute more, you would have been able to reach a second climax.
                            Oh well.
                            One is more than none.
                            A missed opportunity, for sure, but you feel like this was a very memorable experience regardless.
                        [/p]
                        [p]
                            Your extremities still stiff, you roll onto your side, away from the basilisk's [i]fluids[/i].
                            You mindlessly stare up at the stalactite-laden roof of the cave, while you stretch your limbs every which way.
                            Understandably, you feel somewhat more [i]sluggish[/i] than you did before this entire ordeal.
                            You think your scales might have permanently [i]hardened[/i] a bit.
                            Once you feel like you're fit enough to walk, you slowly rise to your feet, and start making your way [i]out[/i] of the damp cave.
                        [/p]
                        [p]
                            You follow the water, hoping it'll lead you to some kind of exit.
                            Your legs still somewhat unresponsive, the journey is slow, and you nearly fall down more than once.
                            As you go, your stride improves, going from a half-limp to a decent walk alongside the water, until daylight comes into sight.
                            As you expected: it was a stale branch of the river.
                            Climbing out of the cave, you sit your tender butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and got some steamy basilisk sex, although you wish it would have lasted a bit longer.
                            A shame, but at least you managed to get off a single time.
                        [/p]
                        [p]
                            You've gained a couple of memories you'll undoubtedly be remembering for the rest of your life,
                            along with a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;
            }
        }
        return toReturn;
    }

    function buildHated() {
        var dick = ctrl.misc.dick();
        var shirt = ctrl.misc.shirt();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;
        var toReturn = ``;
        if (dick) {
            if (clothingMale) {
                toReturn += `
                        [p]
                            You hate this.
                            As though it wasn't enough for the beast to use your body like a cheap fucktoy,
                            now you're stuck above a pool of fresh basilisk cum, with no way out.
                            Your limbs show no sign of regaining motion any time soon.
                            The scent of the monster's cooling semen reminds you of what happened with every breath you take through your nose,
                            constantly making your mind re-live the indignation [com]and the pain[/com] over and over again.
                            You try to breathe through your mouth, but your muzzle is petrified shut, sealing your fate.
                        [/p]
                        [p]
                            About half an hour of wallowing in self-pity later,
                            you finally regain some form of control over your arms.
                            The stony texture that had overtaken your limbs has given way to smooth scales once more.
                            Your muzzle, too, can open once again.
                            Alas, there's no longer a need to breathe through your mouth.
                            Most of the once-pungent scent of the basilisk's arousal has faded, the puddle of cum now but a mere [i]stain[/i] on the cavern's stony floor.
                        [/p]
                        [p]
                            While your encounter with the monster didn't quite go as planned, it could have gone worse.
                            [b]Far[/b] worse.
                            Your pride is severely dented,
                            but you should be grateful that you didn't get [b]hurt[/b].
                            Plus, thanks to your trusty ${global.kobold.inventory.createItemWithTooltip(clothingMale)},
                            you even managed to refrain from getting fucked by the beast.
                        [/p]
                        [p]
                            Your extremities still stiff, you roll onto your side, away from the basilisk's [i]fluids[/i].
                            After wiping your tears, you mindlessly stare up at the stalactite-laden roof of the cave.
                            You sigh loudly.
                            At long last, you've hit rock bottom,
                            being used against your will as a stress relief toy by a feral, who came inches away from [i]actually[/i] fucking you.
                            On the bright side, you doubt there's much worse that can force itself upon you than a mindless beast.
                            At least you [b]hope[/b] so.
                        [/p]
                        [p]
                            You stretch your aching limbs every which way.
                            Understandably, you feel somewhat more [i]sluggish[/i] than you did before this entire ordeal.
                            You think your scales might have permanently [i]hardened[/i] a bit.
                            Once you feel like you're fit enough to walk, you slowly rise to your feet, and start making your way [i]out[/i] of the damp cave,
                            before the beast returns for sloppy seconds.
                        [/p]
                        [p]
                            You follow the water, hoping it'll lead you to some kind of exit.
                            Your legs still somewhat unresponsive, the journey is slow, and you nearly fall down more than once.
                            As you go, your stride improves, going from a half-limp to a decent walk alongside the water, until daylight comes into sight.
                            As you expected: it was a stale branch of the river.
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and got raped by a basilisk, even if it didn't manage to shove its dick inside of you.
                        [/p]
                        [p]
                            You've gained a couple of bad memories you wish you could forget.
                            Some parts of the experience, though, could make for a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            You hate this.
                            As though it wasn't enough for the beast to use your body like a cheap fucktoy,
                            now you're stuck while its cum oozes from your gaping hole.
                            Your limbs show no sign of regaining motion any time soon.
                            The scent of the monster's cooling semen reminds you of what happened with every breath you take,
                            constantly making your mind re-live the indignation and the pain over and over again.
                            You try to breathe through your mouth, but your muzzle is petrified shut, sealing your fate.
                        [/p]
                        [p]
                            About half an hour of wallowing in self-pity later,
                            you finally regain some form of control over your arms.
                            The stony texture that had overtaken your limbs has given way to smooth scales once more.
                            Your muzzle, too, can open once again.
                            Alas, there's no longer a need to breathe through your mouth.
                            Most of the once-pungent scent of the basilisk's arousal has faded, the puddle of cum now but a mere [i]stain[/i] on the cavern's stony floor.
                        [/p]
                        [p]
                            While your encounter with the monster didn't quite go as planned, it could have gone worse.
                            [b]Far[/b] worse.
                            Your pride is severely dented,
                            and your ass thoroughly [b]rutted[/b],
                            but you should be grateful that you didn't get [b]hurt[/b] any more than you did.
                            The creature could have easily killed you after it finished, if it wanted to.
                            Or even [b]before[/b] it got started.
                            You don't want to think about your corpse being defiled right now.
                        [/p]
                        [p]
                            Your extremities still stiff, you roll onto your side, away from the basilisk's [i]fluids[/i].
                            After wiping your tears, you mindlessly stare up at the stalactite-laden roof of the cave.
                            You sigh loudly.
                            At long last, you've hit rock bottom, being used against your will as a fucktoy by a feral.
                            On the bright side, you doubt there's much worse that can force itself upon you than a mindless beast.
                            At least you [b]hope[/b] so.
                        [/p]
                        [p]
                            You stretch your aching limbs every which way.
                            Understandably, you feel somewhat more [i]sluggish[/i] than you did before this entire ordeal.
                            You think your scales might have permanently [i]hardened[/i] a bit.
                            Once you feel like you're fit enough to walk, you slowly rise to your feet, and start making your way [i]out[/i] of the damp cave,
                            before the beast returns for sloppy seconds.
                        [/p]
                        [p]
                            You follow the water, hoping it'll lead you to some kind of exit.
                            Your legs still somewhat unresponsive, the journey is slow, and you nearly fall down more than once.
                            As you go, your stride improves, going from a half-limp to a decent walk alongside the water, until daylight comes into sight.
                            As you expected: it was a stale branch of the river.
                            Climbing out of the cave, you sit your sore butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and then got raped by a basilisk, even if it didn't manage to shove its [i]entire[/i] dick inside of you.
                        [/p]
                        [p]
                            You've gained a couple of bad memories you wish you could forget.
                            Some parts of the experience, though, could make for a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;
            }
        } else {
            if (clothingFemale) {
                toReturn += `
                        [p]
                            You hate this.
                            As though it wasn't enough for the beast to use your body like a cheap fucktoy,
                            now you're stuck above a pool of fresh basilisk cum, with no way out.
                            Your limbs show no sign of regaining motion any time soon.
                            The scent of the monster's cooling semen reminds you of what happened with every breath you take through your nose,
                            constantly making your mind re-live the indignation [com]and the pain[/com] over and over again.
                            You try to breathe through your mouth, but your muzzle is petrified shut, sealing your fate.
                        [/p]
                        [p]
                            About half an hour of wallowing in self-pity later,
                            you finally regain some form of control over your arms.
                            The stony texture that had overtaken your limbs has given way to smooth scales once more.
                            Your muzzle, too, can open once again.
                            Alas, there's no longer a need to breathe through your mouth.
                            Most of the once-pungent scent of the basilisk's arousal has faded, the puddle of cum now but a mere [i]stain[/i] on the cavern's stony floor.
                        [/p]
                        [p]
                            While your encounter with the monster didn't quite go as planned, it could have gone worse.
                            [b]Far[/b] worse.
                            Your pride is severely dented,
                            but you should be grateful that you didn't get [b]hurt[/b].
                            Plus, thanks to your trusty ${global.kobold.inventory.createItemWithTooltip(clothingFemale)},
                            you even managed to refrain from getting fucked by the beast.
                        [/p]
                        [p]
                            Your extremities still stiff, you roll onto your side, away from the basilisk's [i]fluids[/i].
                            After wiping your tears, you mindlessly stare up at the stalactite-laden roof of the cave.
                            You sigh loudly.
                            At long last, you've hit rock bottom,
                            being used against your will as a stress relief toy by a feral, who came inches away from [i]actually[/i] fucking you.
                            On the bright side, you doubt there's much worse that can force itself upon you than a mindless beast.
                            At least you [b]hope[/b] so.
                        [/p]
                        [p]
                            You stretch your aching limbs every which way.
                            Understandably, you feel somewhat more [i]sluggish[/i] than you did before this entire ordeal.
                            You think your scales might have permanently [i]hardened[/i] a bit.
                            Once you feel like you're fit enough to walk, you slowly rise to your feet, and start making your way [i]out[/i] of the damp cave,
                            before the beast returns for sloppy seconds.
                        [/p]
                        [p]
                            You follow the water, hoping it'll lead you to some kind of exit.
                            Your legs still somewhat unresponsive, the journey is slow, and you nearly fall down more than once.
                            As you go, your stride improves, going from a half-limp to a decent walk alongside the water, until daylight comes into sight.
                            As you expected: it was a stale branch of the river.
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and got raped by a basilisk, even if it didn't manage to shove its dick inside of you.
                        [/p]
                        [p]
                            You've gained a couple of bad memories you wish you could forget.
                            Some parts of the experience, though, could make for a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            You hate this.
                            As though it wasn't enough for the beast to use your body like a cheap fucktoy,
                            now you're stuck while its cum oozes from your gaping cunt.
                            Your limbs show no sign of regaining motion any time soon.
                            The scent of the monster's cooling semen reminds you of what happened with every breath you take,
                            constantly making your mind re-live the indignation and the pain over and over again.
                            You try to breathe through your mouth, but your muzzle is petrified shut, sealing your fate.
                        [/p]
                        [p]
                            About half an hour of wallowing in self-pity later,
                            you finally regain some form of control over your arms.
                            The stony texture that had overtaken your limbs has given way to smooth scales once more.
                            Your muzzle, too, can open once again.
                            Alas, there's no longer a need to breathe through your mouth.
                            Most of the once-pungent scent of the basilisk's arousal has faded, the puddle of cum now but a mere [i]stain[/i] on the cavern's stony floor.
                        [/p]
                        [p]
                            While your encounter with the monster didn't quite go as planned, it could have gone worse.
                            [b]Far[/b] worse.
                            Your pride is severely dented,
                            and your loins thoroughly [b]bred[/b],
                            but you should be grateful that you didn't get [b]hurt[/b] any more than you did.
                            The creature could have easily killed you after it finished, if it wanted to.
                            Or even [b]before[/b] it got started.
                            You don't want to think about your corpse being defiled right now.
                        [/p]
                        [p]
                            Your extremities still stiff, you roll onto your side, away from the basilisk's [i]fluids[/i].
                            After wiping your tears, you mindlessly stare up at the stalactite-laden roof of the cave.
                            You sigh loudly.
                            At long last, you've hit rock bottom, being used against your will as a fucktoy by a feral.
                            On the bright side, you doubt there's much worse that can force itself upon you than a mindless beast.
                            At least you [b]hope[/b] so.
                        [/p]
                        [p]
                            You stretch your aching limbs every which way.
                            Understandably, you feel somewhat more [i]sluggish[/i] than you did before this entire ordeal.
                            You think your scales might have permanently [i]hardened[/i] a bit.
                            Once you feel like you're fit enough to walk, you slowly rise to your feet, and start making your way [i]out[/i] of the damp cave,
                            before the beast returns for sloppy seconds.
                        [/p]
                        [p]
                            You follow the water, hoping it'll lead you to some kind of exit.
                            Your legs still somewhat unresponsive, the journey is slow, and you nearly fall down more than once.
                            As you go, your stride improves, going from a half-limp to a decent walk alongside the water, until daylight comes into sight.
                            As you expected: it was a stale branch of the river.
                            Climbing out of the cave, you sit your butt down on the riverside dirt, and take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You nearly got crushed, almost drowned, and then got raped by a basilisk, even if it didn't manage to shove its [i]entire[/i] dick inside of you.
                        [/p]
                        [p]
                            You've gained a couple of bad memories you wish you could forget.
                            Some parts of the experience, though, could make for a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;
            }
        }
        return toReturn;
    }

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {

        var toReturn = ``;

        if (ctrl.relent.get() === UndergroundCaveRelentChoice.ENJOY) {
            toReturn += buildEnjoyed();
        } else {
            toReturn += buildHated();
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\stare\petrified\UndergroundCaveRegain.js

// scenes\HamletDay2\UndergroundCave\stare\petrified\UndergroundCaveEnjoy.js

/**
 * Stare section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveEnjoy(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveRelentChoice.ENJOY;
    var reqController = ctrl.relent;
    var subController = ctrl.preContinue;
    var changeController = ctrl.enjoy;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveRegain(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskEnjoyed = true;
        var analVirgin = ctrl.misc.analVirgin();
        var cuntVirgin = ctrl.misc.cuntVirgin();
        var dick = ctrl.misc.dick();
        var shirt = ctrl.misc.shirt();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;
        var toReturn = `[p]
                            While you didn't exactly [i]ask[/i] for this, you can't deny that something about this entire situation [i]turns you on[/i].
                            The feeling of being [b]helpless[/b] while a large, feral,
                            predatory animal is about to have its way with you scratches a certain kind of itch you didn't even know you had.
                            You guess you're just submissive at heart.
                            Or maybe you have a [i]thing[/i] for petrification.
                            Either way, ${dick ? "you can already feel your length stirring within your cockslit" : "your slit is already dampening in anticipation"}.
                        [/p]
                        [p]
                            Despite the basilisk having moved outside of your field of view, the image of the its loins remains burned in your brain.
                            You've witnessed plenty of males of varying species naked, but you've never before seen a dick quite like [i]this[/i].
                            It's not the [i]biggest[/i] around, especially not when compared to [b]horses[/b] or other [i]large[/i] species,
                            but the shape is unlike anything you've ever seen.
                            The closest thing you have to compare it to is a painting of a dragon's genitals you once saw while breaking into a manor.
                            Perhaps basilisks and dragons are related somehow?
                        [/p]
                        [p]
                            You try to shift your tail out of the way, to ensure the beast has proper access to your groin,
                            but before you manage to move the stiff, near-petrified appendage, the basilisk has already reached your rear.
                            It lifts up its upper body, and deftly mounts your backside, resting most of its massive weight on top of your frail frame, 
                            its front two legs ending up directly behind your arms, its muzzle right next to your face.
                            If you weren't petrified, you're certain your legs and arms would [b]buckle[/b] under the pressure.
                        [/p]`;

        if (dick) {
            if (clothingMale) {
                toReturn += `
                        [p]
                            Despite most of your butt being completely numb, you can feel the creature's
                            member pressing against your still somewhat sensitive
                            slit${clothingFemale ? `, even through your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)}` : ``}.
                            It's so [b]warm[/b] and [b]thick[/b].
                            [b]Definitely[/b] dragon-like.
                            You [b]want[/b] it.
                        [/p]
                        [p]
                            After a few moments of resting idly, the reptile pulls its length back.
                            It begins to hump against your rear, prodding and poking your petrified butt, looking for a hole to [b]fuck[/b].
                            Alas, your ${global.kobold.inventory.createItemWithTooltip(clothingMale)},
                            which cover [com]Should be "covers" in case of "loincloth". Singular/multiplicative in items?[/com]
                            your stone-turned-scales, are in the creature's way.
                            Incapable of tearing through the fabric with its genitals alone,
                            the beast relents in its efforts to find a suitable hole to pound after a few more tries.
                        [/p]
                        [p]
                            Both it and you disappointed, the basilisk resorts to instead fucking the gap between your petrified thighs,
                            grinding its ridged length against your
                            ${clothingFemale ? `${global.kobold.inventory.createItemWithTooltip(clothingFemale)}-clad` : ` bare`} cockslit.
                            The crocodillian appears to be entirely unfazed by the coarse texture of your once-soft upper legs.
                            As their kind is partly composed out of stone-like material,
                            you imagine basilisk sex must [b]always[/b] feel like ramming your dick into a rock-hewn hole.
                        [/p]
                        [p]
                            The basilisk's considerable length, pumping deftly from the back of your legs all the way up to your belly button,
                            occasionally squirts a heady glob of precum, a small puddle of which is forming on the ground beneath you.
                            Its strong, pheromone-laden scent clouds your mind.
                            It's almost as [b]hypnotizing[/b] as the beast's eyes were.
                            You find yourself eagerly taking in deep breaths through your nose,
                            in an attempt to inhale as much of the [b]delectable[/b] fragrance as you possibly can.
                        [/p]
                        [p]
                            While far less pleasurable than the intense, bestial pounding you had hoped for,
                            you find yourself growing exceedingly aroused still,
                            the constant frotting against your cockslit,
                            combined with the undeniably [b]masculine[/b] smell permeating every breath you take,
                            causing your member to rapidly harden.
                            As your shaft grows erect, however, you experience something you've never felt before.
                        [/p]
                        [p]
                            Your slit feels [b]tight[/b].
                            Filled to the very brim.
                            Your loins petrified, your manhood is trapped inside of your cockslit,
                            stuck in a state between being flaccid and fully erect,
                            constrained like a bird in a far too small cage,
                            unable to fly free.
                        [/p]
                        [p]
                            It's a weird feeling, to be aroused to no end by oh-so-little stimulation,
                            yet being denied the full pleasure you hoped you would receive from this situation.
                            Despite desperately wanting to get off, some part of you actually [i]enjoys[/i] the pseudo-cock-cage formed by your petrified loins.
                            It reinforces the notion that this entire thing is [b]not[b] about [i]you[/i].
                            It's about a large, uncaring [b]monster[/b] sating its [b]animalistic lust[/b] on your body, in whatever way it can.
                            And [b]that[/b] turns you on more than anything else.
                        [/p]
                        [p]
                            The basilisk, uncaring about your pleasure (or the lack thereof),
                            continues to hump away at your thighs, its ridged shaft grinding along your bustling cockslit and your stone-turned belly both.
                            Being an animal, however, it doesn't last long.
                            No more than five minutes after crawling on your back, you can already [b]sense[/b] its approaching orgasm.
                            Its movements become more erratic, a small growl rumbles from the back of its throat, and its large maw opens ever so slightly,
                            the creature seemingly [b]huffing[/b] and [b]panting[/b] through its mouth the same way a dog would.
                        [/p]
                        [p]
                            A few moments later, the monster reaches its end, 
                            its incessant thigh-fucking giving way to a series of fierce thrusts,
                            while squirt after squirt of thick basilisk spunk covers the stone floor and
                            your ${shirt ? `${global.kobold.inventory.createItemWithTooltip(shirt)}` : `lower body`},
                            the heavy scent adding to the lingering odor of bestial masculinity hanging in the air, driving your mind [b]wild[/b] with unsated desire.
                        [/p]
                        [p]
                            Once its climax starts to fade, the large reptillian slowly shifts its weight off of your back.
                            Its member receding into its body, the creature crawls back towards the hole it came from,
                            and before you even have the [b]chance[/b] to [b]try[/b] to [b]beg[/b] it to continue,
                            the basilisk disappears back into the darkness, leaving you trapped and aroused,
                            unable to get off while constantly being tempted by the aroma of the puddle of feral cum directly below you.
                        [/p]`;
            } else {
                global.kobold.history.shared.well.basiliskPenetrated = true;
                toReturn += `
                        [p]
                            Despite most of your butt being completely numb, you can feel the creature's
                            member pressing against your still somewhat sensitive
                            slit${clothingFemale ? `, even through your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)}` : ``}.
                            It's so [b]warm[/b] and [b]thick[/b].
                            [b]Definitely[/b] dragon-like.
                            You [b]want[/b] it.
                        [/p]
                        [p]
                            After a few moments of resting idly, the reptile pulls its length back.
                            It begins to hump against your rear, prodding and poking your petrified butt, looking for a hole to [b]fuck[/b].
                            After the third or fourth try, it hits its mark,
                            its thick, tapered cock [b]ramming[/b] into your unlubed${analVirgin ? `, virginal` : ``} tailhole.
                            Once firmly lodged inside of you, the beast immediately begins to thrust in and out of your body,
                            its rapid pace only matched by its feral ferocity.
                        [/p]
                        [p]
                            While your buttcheeks are numb from petrification, your stretched insides are not.
                            Your ass [b]hurts[/b] from the slightly oversized insertion,
                            although the pain quickly gets drowned out by the pleasure of being [b]taken[/b] by a feral [b]beast[/b],
                            your own dick rapidly swelling in your slit.
                            The crocodillian appears to be entirely unfazed by the coarse texture of your buttcheeks.
                            As their kind is partly composed out of stone-like material,
                            you imagine basilisk sex must [b]always[/b] feel like ramming your dick into a rock-hewn hole.
                        [/p]
                        [p]
                            You aren't sure exactly [i]how much[/i] of the creature's length is being forced inside of you,
                            but you figure it must be at least more than half.
                            It [b]feels[/b] like a lot, anyway.
                            Each of the individual ridges lining its shaft pop into and out of your rear, sending waves of pure [b]pleasure[/b] coursing down your spine.
                            Every few thrusts, a squirt of thick, warm precum shoots from the tip of the monster's draconic member, slickening your insides.
                            The lubrication allows the beast to fuck you [b]harder[/b] and [b]faster[/b],
                            its cock going just [i]slightly[/i], but [b]noticeably[/b] deeper than before.
                        [/p]
                        [p]
                            The smell of pure, bestial [b]sex[/b] spreads throughout the cave,
                            the strong and masculine-pheromone laden scent of the self-exerting basilisk clouding your mind.
                            It's almost as [b]hypnotizing[/b] as the beast's eyes were.
                            You find yourself eagerly taking in deep breaths through your nose,
                            in an attempt to inhale as much of the [b]delectable[/b] fragrance as you possibly can.
                        [/p]
                        [p]
                            The intense, feral rutting is every bit as pleasurable as you imagined it to be.
                            Your member rapidly grows more and more erect, until suddenly, you experience something you've never felt before. 
                            Your slit feels [b]tight[/b].
                            Filled to the very brim.
                            Your loins petrified, your manhood is trapped inside of your cockslit,
                            stuck in a state between being flaccid and fully erect,
                            constrained like a bird in a far too small cage,
                            unable to fly free.
                        [/p]
                        [p]
                            It's a weird feeling, to be aroused to no end, yet still being denied [b]full[/b] pleasure.
                            Despite desperately wanting to get off, some part of you actually [i]enjoys[/i] the pseudo-cock-cage formed by your petrified loins.
                            It reinforces the notion that this entire thing is [b]not[b] about [i]you[/i].
                            It's about a large, uncaring [b]monster[/b] sating its [b]animalistic lust[/b] on your body.
                            And [b]that[/b] turns you on more than anything else.
                        [/p]
                        [p]
                            The basilisk, uncaring about your pleasure (or the lack thereof),
                            continues to hump away, its ridged shaft almost bottoming out inside of you.
                            Your belly would be bulging ever-so-slightly if not turned to stone.
                            While the rough rutting is one of the best things you've ever experienced,
                            your animal partner fails to last for more than a few minutes,
                            the tight confines of your ${analVirgin ? `virginal` : ``} tailhole proving far more pleasurable than the loins of its own kind.
                        [/p]
                        [p]
                            You can already [b]sense[/b] its approaching orgasm.
                            Its movements become more erratic, a small growl rumbles from the back of its throat, and its large maw opens ever so slightly,
                            the creature seemingly [b]huffing[/b] and [b]panting[/b] through its mouth the same way a dog would.
                            Its incessant buttfucking gives way to a series of fierce thrusts, a clear indicator of what is to come.
                            A few moments later, the monster reaches its end,
                            wave after wave of hot, thick basilisk spunk flooding your body, filling up your insides to the [b]brim[/b].
                        [/p]
                        [p]
                            Once its climax starts to fade, the large reptillian slowly shifts its weight off of your back,
                            and pulls its length from your rear.
                            A [b]flood[/b] of heady feral cum leaks from your slightly gaping ass,
                            the heavy scent of the stuff adding to the lingering odor of bestial masculinity hanging in the air,
                            driving your mind [b]wild[/b] with unsated desire.
                        [/p]
                        [p]
                            Its member receding into its body, the creature crawls back towards the hole it came from,
                            and before you even have the [b]chance[/b] to [b]try[/b] to [b]beg[/b] it for more,
                            the monster disappears back into the darkness, leaving you freshly fucked, trapped and aroused,
                            unable to get off while constantly being tempted by the aroma of the puddle of feral cum dripping from your tailhole.
                        [/p]`;
                if (analVirgin) {
                    toReturn += `
                        [p]
                            [c]
                                [b]You are no longer a virgin.[/b]
                            [/c]
                        [/p]`;
                }
            }
        } else {
            if (clothingFemale) {
                toReturn += `
                        [p]
                            Despite most of your rear end being completely numb, you can feel the creature's
                            member pressing against your still sensitive
                            slit, even through your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)}.
                            It's so [b]warm[/b] and [b]thick[/b].
                            [b]Definitely[/b] dragon-like.
                            You [b]want[/b] it.
                        [/p]
                        [p]
                            After a few moments of resting idly, the reptile pulls its length back.
                            It begins to hump against your groin, prodding and poking your petrified loins, looking for a hole to [b]fuck[/b].
                            Alas, your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)},
                            which cover [com]Should be "covers" in case of "loincloth". Singular/multiplicative in items?[/com]
                            your stone-turned-scales, are in the creature's way.
                            Incapable of tearing through the fabric with its genitals alone,
                            the beast relents in its efforts to find a suitable hole to pound after a few more tries.
                        [/p]
                        [p]
                            Both it and you disappointed, the basilisk resorts to instead fucking the gap between your petrified thighs,
                            grinding its ridged length against your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)}-clad cunt.
                            The crocodillian appears to be entirely unfazed by the coarse texture of your once-soft upper legs.
                            As their kind is partly composed out of stone-like material,
                            you imagine basilisk sex must [b]always[/b] be like ramming a dick into a rock-hewn hole.
                        [/p]
                        [p]
                            The basilisk's considerable length, pumping deftly from the back of your legs all the way up to your belly button,
                            occasionally squirts a heady glob of precum, a small puddle of which is forming on the ground beneath you.
                            Its strong, pheromone-laden scent clouds your mind.
                            It's almost as [b]hypnotizing[/b] as the beast's eyes were.
                            You find yourself eagerly taking in deep breaths through your nose,
                            in an attempt to inhale as much of the [b]delectable[/b] fragrance as you possibly can.
                        [/p]
                        [p]
                            While far less pleasurable than the intense, bestial pounding you had hoped for,
                            you find yourself growing exceedingly aroused still,
                            the constant frotting against your slit,
                            combined with the undeniably [b]masculine[/b] smell permeating every breath you take,
                            causes your cunt to rapidly moisten.
                            A large spot of feminine excitement forms on your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)},
                            the slick feeling of which spurs the creature on further.
                        [/p]
                        [p]
                            It's a weird feeling, to be aroused to no end by oh-so-little stimulation,
                            yet being denied the full pleasure you hoped you would receive from this situation.
                            Despite desperately wanting to get off, some part of you actually [i]enjoys[/i] the fact you aren't being fucked directly.
                            It reinforces the notion that this entire thing is [b]not[b] about [i]you[/i].
                            It's about a large, uncaring [b]monster[/b] sating its [b]animalistic lust[/b] on your body, in whatever way it can.
                            And [b]that[/b] turns you on more than anything else.
                        [/p]
                        [p]
                            The basilisk, uncaring about your pleasure,
                            continues to hump away at your thighs, its ridged shaft grinding along your twat and your stone-turned belly both.
                            Being an animal, however, it doesn't last long.
                            No more than five minutes after crawling on your back, you can already [b]sense[/b] its approaching orgasm.
                            Its movements become more erratic, a small growl rumbles from the back of its throat, and its large maw opens ever so slightly,
                            the creature seemingly [b]huffing[/b] and [b]panting[/b] through its mouth the same way a dog would.
                        [/p]
                        [p]
                            A few moments later, the monster reaches its end, 
                            its incessant thigh-fucking giving way to a series of fierce thrusts,
                            while squirt after squirt of thick basilisk spunk covers the stone floor and
                            your ${shirt ? `${global.kobold.inventory.createItemWithTooltip(shirt)}` : `lower body`},
                            the heavy scent adding to the lingering odor of bestial masculinity hanging in the air, driving your mind [b]wild[/b] with unsated desire.
                        [/p]
                        [p]
                            Once its climax starts to fade, the large reptillian slowly shifts its weight off of your back.
                            Its member receding into its body, the creature crawls back towards the hole it came from,
                            and before you even have the [b]chance[/b] to [b]try[/b] to [b]beg[/b] it to continue,
                            the basilisk disappears back into the darkness, leaving you trapped and aroused,
                            unable to get off while constantly being tempted by the aroma of the puddle of feral cum directly below you.
                        [/p]`;
            } else {
                global.kobold.history.shared.well.basiliskPenetrated = true;
                toReturn += `
                        [p]
                            Despite most of your rear end being completely numb, you can feel the creature's
                            member pressing against your still sensitive slit.
                            It's so [b]warm[/b] and [b]thick[/b].
                            [b]Definitely[/b] dragon-like.
                            You [b]want[/b] it.
                        [/p]
                        [p]
                            After a few moments of resting idly, the reptile pulls its length back.
                            It begins to hump against your groin, prodding and poking your petrified loins, looking for a hole to [b]fuck[/b].
                            After the third or fourth try, it hits its mark,
                            its thick, tapered cock [b]ramming[/b] into your cunt.
                            ${cuntVirgin ? `A small amount of blood leaks from your slit, signifying the loss of your virginity, to a monster in a cave no less.` : ``}
                            Once firmly lodged inside of you, the beast immediately begins to thrust in and out of your body,
                            its rapid pace only matched by its feral ferocity.
                        [/p]
                        [p]
                            While your folds are numb from petrification, your stretched insides are not.
                            Your twat [b]hurts[/b] from the slightly oversized insertion,
                            although the pain quickly gets drowned out by the pleasure of being [b]taken[/b] by a feral [b]beast[/b].
                            The crocodillian appears to be entirely unfazed by the coarse texture of your folds.
                            As their kind is partly composed out of stone-like material,
                            you imagine basilisk sex must [b]always[/b] feel like ramming a dick into a rock-hewn hole.
                        [/p]
                        [p]
                            You aren't sure exactly [i]how much[/i] of the creature's length is being forced inside of you,
                            but you figure it must be at least more than half.
                            It [b]feels[/b] like a lot, anyway.
                            Each of the individual ridges lining its shaft pop into and out of your cunt, sending waves of pure [b]pleasure[/b] coursing down your spine.
                            Every few thrusts, a squirt of thick, warm precum shoots from the tip of the monster's draconic member,
                            which combined with your own arousal, slickens your insides.
                            The extra lubrication allows the beast to fuck you [b]harder[/b] and [b]faster[/b],
                            its cock going just [i]slightly[/i], but [b]noticeably[/b] deeper than before.
                        [/p]
                        [p]
                            The smell of pure, bestial [b]sex[/b] spreads throughout the cave,
                            the strong and masculine-pheromone laden scent of the self-exerting basilisk clouding your mind.
                            It's almost as [b]hypnotizing[/b] as the beast's eyes were.
                            You find yourself eagerly taking in deep breaths through your nose,
                            in an attempt to inhale as much of the [b]delectable[/b] fragrance as you possibly can.
                        [/p]
                        [p]
                            The intense, feral [b]rutting[/b] is every bit as pleasurable as you imagined it to be.
                            Less than a minute in, you can already feel your end approaching.
                            Your insides quiver and spasm, [b]milking[/b] the beast for all he has.
                            If your mouth would open, you'd [b]scream[/b] in pleasure, but all you manage is a muffled moan.
                            Your toes, one of the few parts of your body that you can actually [i]move[/i],
                            curl up from sheer orgasmic bliss as a world-rocking climax hits your body hard.
                        [/p]
                        [p]
                            Your feminine fluids drip down to the ground, but the crocodillian isn't done yet.
                            You slowly curl your stiff tail around it, pulling it in just a little bit closer while it continues to hump away.
                            It is then that you realize that this entire thing is [b]not[b] about [i]you[/i].
                            It's about a large, uncaring [b]monster[/b] sating its [b]animalistic lust[/b] on your body.
                            And [b]that[/b] turns you on more than anything else.
                        [/p]
                        [p]
                            The basilisk's ridged shaft almost completely bottoms out inside of you.
                            Your belly would be bulging ever-so-slightly if not turned to stone.
                            While the rough [b]breeding[/b] is one of the best things you've ever experienced,
                            your animal partner fails to last long enough to rock your world more than once,
                            the tight confines of your ${cuntVirgin ? `virginal` : ``} cunt proving far more pleasurable than the loins of its own kind.
                        [/p]
                        [p]
                            You can already [b]sense[/b] its approaching orgasm.
                            Its movements become more erratic, a small growl rumbles from the back of its throat, and its large maw opens ever so slightly,
                            the creature seemingly [b]huffing[/b] and [b]panting[/b] through its mouth the same way a dog would.
                            Its incessant rutting giving way to a series of fierce thrusts, a clear indicator of what is to come.
                            A few moments later, the monster reaches its end,
                            wave after wave of hot, thick, [b]fertile[/b] basilisk spunk flooding your insides, filling you to the [b]brim[/b].
                        [/p]
                        [p]
                            Once its climax starts to fade, the large reptillian slowly shifts its weight off of your back,
                            and pulls its length from your slit.
                            A [b]flood[/b] of heady feral cum leaks from your slightly gaping twat,
                            the heavy scent of the stuff adding to the lingering odor of bestial masculinity hanging in the air,
                            driving your mind [b]wild[/b] with desire.
                        [/p]
                        [p]
                            Its member receding into its body, the creature crawls back towards the hole it came from,
                            and before you even have the [b]chance[/b] to [b]try[/b] to [b]beg[/b] it for more,
                            the monster disappears back into the darkness, leaving you freshly fucked, trapped and aroused,
                            unable to get off a second time, no matter how much you want to,
                            all the while being tempted by the aroma of the puddle of fragrant cum dripping from your well-bred slit.
                        [/p]`;
                if (cuntVirgin) {
                    toReturn += `
                        [p]
                            [c]
                                [b]You are no longer a virgin.[/b]
                            [/c]
                        [/p]`;
                }
            }
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\stare\petrified\UndergroundCaveEnjoy.js

// scenes\HamletDay2\UndergroundCave\stare\petrified\UndergroundCaveHate.js

/**
 * Stare section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveHate(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveRelentChoice.HATE;
    var reqController = ctrl.relent;
    var subController = ctrl.preContinue;
    var changeController = ctrl.hate;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveRegain(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var analVirgin = ctrl.misc.analVirgin();
        var cuntVirgin = ctrl.misc.cuntVirgin();
        var dick = ctrl.misc.dick();
        var shirt = ctrl.misc.shirt();
        var pants = ctrl.misc.pants();
        var underwear = ctrl.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;
        var toReturn = `[p]
                            Despite the basilisk having moved outside of your field of view, the image of the its loins remains burned in your brain.
                            You've witnessed plenty of males of varying species naked, but you've never before seen a dick quite like [i]this[/i].
                            It's not the [i]biggest[/i] around, especially not when compared to [b]horses[/b] or other [i]large[/i] species,
                            but the shape is unlike anything you've ever seen.
                            The closest thing you have to compare it to is a painting of a dragon's genitals you once saw while breaking into a manor.
                            Perhaps basilisks and dragons are related somehow?
                        [/p]
                        [p]
                            Panic wells up inside of you.
                            You can't let this happen!
                            Those ridges'll tear your insides apart!
                            You struggle and strain, desperately trying, but failing to move [b]any[/b] of your limbs.
                            They've all turned to hard, unmoving stone.
                            You use every ounce of concentration you can muster try to shift your tail 
                            -- one of the last few parts of your body you can actually still move -- in between your legs,
                            to at least [b]try[/b] to guard your ${dick ? `butt` : `loins`} from the basilisk's...[i]thing[/i].
                        [/p]
                        [p]
                            Even when trying your very hardest,
                            you still fail to move the near-petrified appendage more than a few inches before the basilisk reaches your rear.
                            It lifts up its upper body, and deftly mounts your backside, resting most of its massive weight on top of your frail frame, 
                            its front two legs ending up directly behind your arms, its muzzle right next to your face.
                            If you weren't petrified, you're certain your legs and arms would [b]buckle[/b] under the pressure.
                            Your tail gets trapped between its underbelly and your back, effectively nullifying your attempt to preserve your dignity.
                        [/p]`;

        if (dick) {
            if (clothingMale) {
                toReturn += `
                        [p]
                            Despite most of your butt being completely numb, you can feel the creature's
                            member pressing against your still somewhat sensitive
                            slit${clothingFemale ? `, even through your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)}` : ``}.
                            It's [b]disgustingly[/b] warm, and the [b]thickness[/b] of it [b]worries you[/b].
                            You can't take something of that size, not to mention the ridges!
                        [/p]
                        [p]
                            After a few seconds of the reptile resting idly, the moment you've feared arrives.
                            It pulls its length back, and begins to hump against your rear, prodding and poking your petrified butt, looking for a hole to [b]fuck[/b].
                            Thank goodness, your ${global.kobold.inventory.createItemWithTooltip(clothingMale)},
                            which cover [com]Should be "covers" in case of "loincloth". Singular/multiplicative in items?[/com]
                            your stone-turned-scales, are in the creature's way.
                            Incapable of tearing through the fabric with its genitals alone,
                            the beast relents in its efforts to find a suitable hole to pound after a few more tries.
                        [/p]
                        [p]
                            It disappointed, and you relieved, the basilisk resorts to instead fucking the gap between your petrified thighs,
                            grinding its ridged length against your
                            ${clothingFemale ? `${global.kobold.inventory.createItemWithTooltip(clothingFemale)}-clad` : ` bare`} cockslit.
                            The crocodillian appears to be entirely unfazed by the coarse texture of your once-soft upper legs.
                            As their kind is partly composed out of stone-like material,
                            you imagine basilisk sex must [b]always[/b] feel like ramming your dick into a rock-hewn hole.
                        [/p]
                        [p]
                            While the worst of your fears have been avoided,
                            getting your thighs fucked by a humping basilisk is still more than a little degrading.
                            You felt deeply ashamed for having fallen for its [i]trap[/i] in the first,
                            but now you feel like a piece of petrified [b]meat[/b],
                            being used by a large, uncaring [b]monster[/b] interested only in sating its [b]animalistic lust[/b] on your body,
                            in whatever way it can.
                            Your pride is severely dented, and that hurts not your body, but your soul.
                        [/p]
                        [p]
                            The basilisk's considerable length, pumping deftly from the back of your legs all the way up to your belly button,
                            occasionally squirts a heady glob of precum, a small puddle of which is forming on the ground beneath you.
                            The strong, pheromone-laden scent of the stuff sickens you to your very core.
                            It doesn't smell [i]too[/i] bad,
                            but the sheer concentration of [b]testosterone[/b] makes every breath you take feel like a [i]challenge[/i] to your masculinity.
                            Even worse is the realization that you'll be stuck, forced to inhale the unwanted smell of basilisk pre -- and ultimately actual -- cum,
                            until your limbs regain their mobility, which you fear might be a long time still.
                        [/p]
                        [p]
                            Despite hating [b]everything[/b] about the situation you find yourself in,
                            you can't say the constant frotting against your cockslit [i]doesn't[/i] feel somewhat [i]good[/i].
                            Still, you find the slight pleasant sensation to be more [b]embarrassing[/b] than it is arousing.
                            What kind of [b]freak[/b] would feel even the tiniest bit of [i]pleasure[/i] while getting [b]abused[/b] by a [b]feral beast?![/b]
                            Not you, that's who.
                            At least so you tell yourself, in an effort to retain [b]some[/b] of your remaining self-esteem.
                        [/p]
                        [p]
                            The basilisk, uncaring about your pleasure (or the lack thereof),
                            continues to hump away at your thighs, its ridged shaft grinding along your cockslit and your stone-turned belly both.
                            Being an animal, it luckily doesn't last very long.
                            No more than five minutes after crawling on your back, you can already [b]sense[/b] its approaching orgasm.
                            Its movements become more erratic, a small growl rumbles from the back of its throat, and its large maw opens ever so slightly,
                            the creature seemingly [b]huffing[/b] and [b]panting[/b] through its mouth the same way a dog would.
                            As if you needed any more reminders that you're being fucked by nothing more than a lowly, feral [b]beast[/b].
                        [/p]
                        [p]
                            A few moments later, the monster reaches its end, 
                            its incessant thigh-fucking giving way to a series of fierce thrusts,
                            while squirt after squirt of thick basilisk spunk covers the stone floor and
                            your ${shirt ? `${global.kobold.inventory.createItemWithTooltip(shirt)}` : `lower body`},
                            the heavy scent adding to the lingering odor of bestial masculinity hanging in the air,
                            which is almost strong enough to make you [b]retch[/b] and [b]gag[/b].
                            Your eyes water out of sheer disgust, a perfect way to mask the few tears that had been welling up since the start of this entire ordeal.
                        [/p]
                        [p]
                            Once its climax starts to fade, the large reptillian slowly shifts its weight off of your back.
                            Its member receding into its body, the creature crawls back towards the hole it came from,
                            and before you even have the [b]chance[/b] to [b]try[/b] to [b]shout obscenities[/b] at it,
                            the basilisk disappears back into the darkness, leaving you trapped, disgusted and crying,
                            unable to escape while forced to inhale the aroma of the puddle of feral cum directly below you.
                        [/p]`;
            } else {
                global.kobold.history.shared.well.basiliskPenetrated = true;
                toReturn += `
                        [p]
                            Despite most of your butt being completely numb, you can feel the creature's
                            member pressing against your still somewhat sensitive
                            slit${clothingFemale ? `, even through your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)}` : ``}.
                            It's [b]disgustingly[/b] warm, and the [b]thickness[/b] of it [b]worries you[/b].
                            You can't take something of that size, not to mention the ridges!
                        [/p]
                        [p]
                            After a few seconds of the reptile resting idly, the moment you've feared arrives.
                            It pulls its length back, and begins to hump against your rear, prodding and poking your petrified butt, looking for a hole to [b]fuck[/b].
                            After the third or fourth try, it hits its mark,
                            its thick, tapered cock crudely [b]ramming[/b] into your unlubed${analVirgin ? `, virginal` : ``} tailhole.
                            Spikes of pain shoot up your spine.
                            You want to [b]scream[/b] out in agony, yet all can manage to utter is a muffled grunt.
                            Once firmly lodged inside of you, the beast painfully begins to thrust in and out of your body,
                            its rapid pace only matched by its feral ferocity.
                        [/p]
                        [p]
                            While your buttcheeks are numb from petrification, your stretched insides are not.
                            Your ass [b]hurts[/b] from the oversized insertion,
                            the monster's cock far too [i]girthy[/i] for your tender, frail body.
                            The crocodillian appears to be entirely unfazed by the coarse texture of your buttcheeks.
                            As their kind is partly composed out of stone-like material,
                            you imagine basilisk sex must [b]always[/b] feel like ramming your dick into a rock-hewn hole.
                        [/p]
                        [p]
                            You aren't sure exactly [i]how much[/i] of the beast's length is being forced inside of you,
                            but you figure it must be at least more than half.
                            You don't know whether you're too tight for it to push in any deeper,
                            or if the creature is showing you some kind of [i]mercy[/i] by not making you take its full length.
                            The ridges, which you dreaded the most, are luckily not as bad as you'd imagined.
                            They don't feel [i]pleasant[/i] popping in and out of your rear, but neither do they [i]hurt[/i].
                        [p]
                        [/p]
                            Every few thrusts, a squirt of thick, warm precum shoots from the tip of the monster's draconic member, slickening your insides.
                            A double edged blade.
                            On one hand, the lubrication eases the beast's thrusts, making the entire ordeal slightly less [b]agonizing[/b].
                            Your tailhole slowly growing accustomed to being stretched to its near-limits,
                            also helps reducing what was once excruciating torment into the mere feeling of constant uncomfortableness,
                            with just a little bit of pain on the side.
                            The other end of the rope is that the crocodillian can now pick up its pace, thoroughly [b]rutting[/b] you like the [b]basilisk bitch[/b] you are.
                            The less your ass hurts, the more your pride does.
                        [/p]
                        [p]
                            The basilisk, uncaring about your pain,
                            continues to dutifully hump away, its ridged shaft uncomfortably forcing itself just a little bit deeper inside of you.
                            Your belly would be bulging ever-so-slightly if not turned to stone.
                            While the rough fucking is one of the most degrading things you've ever experienced,
                            the animal luckily fails to last for more than a few minutes,
                            the tight confines of your ${analVirgin ? `virginal` : ``} tailhole proving far more pleasurable than the loins of its own kind,
                            even if that pleasure is far from mutual.
                        [/p]
                        [p]
                            You can already [b]sense[/b] its approaching orgasm.
                            Its movements become more erratic, a small growl rumbles from the back of its throat, and its large maw opens ever so slightly,
                            the creature seemingly [b]huffing[/b] and [b]panting[/b] through its mouth the same way a dog would.
                            As if you needed any more reminders that you're being rutted by nothing more than a lowly, feral [b]beast[/b].
                            Its incessant buttfucking gives way to a series of fierce thrusts, a clear indicator of what is to come.
                            A few extra-rough pumps later, the monster reaches its end,
                            wave after wave of warm, thick basilisk spunk flooding your body, filling up your insides to the [b]brim[/b].
                        [/p]
                        [p]
                            Once its climax starts to fade, the large reptillian slowly shifts its weight off of your back,
                            and pulls its length from your rear.
                            A [b]flood[/b] of heady feral cum leaks from your slightly gaping ass,
                            the heavy scent of the stuff adding to the lingering odor of bestial masculinity hanging in the air,
                            which is almost strong enough to make you [b]retch[/b] and [b]gag[/b].
                            Your eyes water out of sheer disgust, a perfect way to mask the few tears that had been welling up since the start of this entire ordeal.
                        [/p]
                        [p]
                            Its member receding into its body, the creature crawls back towards the hole it came from,
                            and before you even have the [b]chance[/b] to [b]try[/b] to [b]shout obscenities[/b] at it,
                            the monster disappears back into the darkness, leaving you trapped, disgusted and crying,
                            unable to escape while forced to inhale the aroma of the puddle of feral cum dripping from your abused tailhole.
                        [/p]`;
                if (analVirgin) {
                    toReturn += `
                        [p]
                            [c]
                                [b]You are no longer a virgin.[/b]
                            [/c]
                        [/p]`;
                }
            }
        } else {
            if (clothingFemale) {
                toReturn += `
                        [p]
                            Despite most of your rear end being completely numb, you can feel the creature's
                            member pressing against your still somewhat sensitive
                            slit, even through your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)}.
                            It's [b]disgustingly[/b] warm, and the [b]thickness[/b] of it [b]worries you[/b].
                            You can't take something of that size, not to mention the ridges!
                        [/p]
                        [p]
                            After a few seconds of the reptile resting idly, the moment you've feared arrives.
                            It pulls its length back, and begins to hump against your groin, prodding and poking your petrified loins, looking for a hole to [b]fuck[/b].
                            Thank goodness, your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)},
                            which cover [com]Should be "covers" in case of "loincloth". Singular/multiplicative in items?[/com]
                            your stone-turned-scales, are in the creature's way.
                            Incapable of tearing through the fabric with its genitals alone,
                            the beast relents in its efforts to find a suitable hole to pound after a few more tries.
                        [/p]
                        [p]
                            It disappointed, and you relieved, the basilisk resorts to instead fucking the gap between your petrified thighs,
                            grinding its ridged length against your ${global.kobold.inventory.createItemWithTooltip(clothingFemale)}-clad cunt.
                            The crocodillian appears to be entirely unfazed by the coarse texture of your once-soft upper legs.
                            As their kind is partly composed out of stone-like material,
                            you imagine basilisk sex must [b]always[/b] be like ramming a dick into a rock-hewn hole.
                        [/p]
                        [p]
                            While the worst of your fears have been avoided,
                            getting your thighs fucked by a humping basilisk is still more than a little degrading.
                            You felt deeply ashamed for having fallen for its [i]trap[/i] in the first,
                            but now you feel like a piece of petrified [b]meat[/b],
                            being used by a large, uncaring [b]monster[/b] interested only in sating its [b]animalistic lust[/b] on your body,
                            in whatever way it can.
                            Your pride is severely dented, and that hurts not your body, but your soul.
                        [/p]
                        [p]
                            The basilisk's considerable length, pumping deftly from the back of your legs all the way up to your belly button,
                            occasionally squirts a heady glob of precum, a small puddle of which is forming on the ground beneath you.
                            The strong, pheromone-laden scent of the stuff sickens you to your very core.
                            It doesn't smell [i]too[/i] bad,
                            but the sheer concentration of [b]testosterone[/b] makes every breath you take feel like a
                            [i]challenge[/i] to your natural instinct of [b]submitting[/b] to the powerful, if feral, male on top of you.
                            Even worse is the realization that you'll be stuck, forced to inhale the unwanted smell of basilisk pre -- and ultimately actual -- cum,
                            until your limbs regain their mobility, which you fear might be a long time still.
                        [/p]
                        [p]
                            Despite hating [b]everything[/b] about the situation you find yourself in,
                            you can't say the constant frotting against your slit [i]doesn't[/i] feel somewhat [i]good[/i].
                            Still, you find the slight pleasant sensation to be more [b]embarrassing[/b] than it is arousing.
                            What kind of [b]freak[/b] would feel even the tiniest bit of [i]pleasure[/i] while getting [b]abused[/b] by a [b]feral beast?![/b]
                            Not you, that's who.
                            At least so you tell yourself, in an effort to retain [b]some[/b] of your remaining self-esteem.
                        [/p]
                        [p]
                            The basilisk, uncaring about your pleasure (or the lack thereof),
                            continues to hump away at your thighs, its ridged shaft grinding along your twat and your stone-turned belly both.
                            Being an animal, it luckily doesn't last very long.
                            No more than five minutes after crawling on your back, you can already [b]sense[/b] its approaching orgasm.
                            Its movements become more erratic, a small growl rumbles from the back of its throat, and its large maw opens ever so slightly,
                            the creature seemingly [b]huffing[/b] and [b]panting[/b] through its mouth the same way a dog would.
                            As if you needed any more reminders that you're being fucked by nothing more than a lowly, feral [b]beast[/b].
                        [/p]
                        [p]
                            A few moments later, the monster reaches its end, 
                            its incessant thigh-fucking giving way to a series of fierce thrusts,
                            while squirt after squirt of thick basilisk spunk covers the stone floor and
                            your ${shirt ? `${global.kobold.inventory.createItemWithTooltip(shirt)}` : `lower body`},
                            the heavy scent adding to the lingering odor of bestial masculinity hanging in the air,
                            which is almost strong enough to make you [b]retch[/b] and [b]gag[/b].
                            Your eyes water out of sheer disgust, a perfect way to mask the few tears that had been welling up since the start of this entire ordeal.
                        [/p]
                        [p]
                            Once its climax starts to fade, the large reptillian slowly shifts its weight off of your back.
                            Its member receding into its body, the creature crawls back towards the hole it came from,
                            and before you even have the [b]chance[/b] to [b]try[/b] to [b]shout obscenities[/b] at it,
                            the basilisk disappears back into the darkness, leaving you trapped and disgusted,
                            unable to escape while forced to inhale the aroma of the puddle of feral cum directly below you.
                        [/p]`;
            } else {
                global.kobold.history.shared.well.basiliskPenetrated = true;
                toReturn += `
                        [p]
                            Despite most of your rear end being completely numb, you can feel the creature's
                            member pressing against your still sensitive slit.
                            It's [b]disgustingly[/b] warm, and the [b]thickness[/b] of it [b]worries you[/b].
                            You can't take something of that size, not to mention the ridges!
                        [/p]
                        [p]
                            After a few seconds of the reptile resting idly, the moment you've feared arrives.
                            It begins to hump against your groin, prodding and poking your petrified loins, looking for a hole to [b]fuck[/b].
                            After the third or fourth try, it hits its mark,
                            its thick, tapered cock crudely [b]ramming[/b] into your unlubed cunt.
                            ${cuntVirgin ? `A small amount of blood leaks from your slit, signifying the loss of your virginity, to a monster in a cave no less.` : ``}
                            Once firmly lodged inside of you, the beast immediately begins to thrust in and out of your body,
                            its rapid pace only matched by its feral ferocity.
                        [/p]
                        [p]
                            While your folds are numb from petrification, your stretched insides are not.
                            Your twat [b]hurts[/b] from the oversized insertion,
                            the monster's cock far too [i]girthy[/i] for your tender, frail body.
                            The crocodillian appears to be entirely unfazed by the coarse texture of your folds.
                            As their kind is partly composed out of stone-like material,
                            you imagine basilisk sex must [b]always[/b] feel like ramming a dick into a rock-hewn hole.
                        [/p]
                        [p]
                            You aren't sure exactly [i]how much[/i] of the creature's length is being forced inside of you,
                            but you figure it must be at least more than half.
                            You don't know whether you're too tight for it to push in any deeper,
                            or if the creature is showing you some kind of [i]mercy[/i] by not making you take its full length.
                            The ridges, which you dreaded the most, are luckily not as bad as you'd imagined.
                            They don't feel [i]pleasant[/i] popping in and out of your slit, but neither do they [i]hurt[/i].
                        [/p]
                        [p]
                            Every few thrusts, a squirt of thick, warm precum shoots from the tip of the monster's draconic member, slickening your insides.
                            A double edged blade.
                            On one hand, the lubrication eases the beast's thrusts, making the entire ordeal slightly less [b]agonizing[/b].
                            Your cunt slowly growing accustomed to being stretched to its near-limits,
                            also helps reducing what was once excruciating torment into the mere feeling of constant uncomfortableness,
                            with just a little bit of pain on the side.
                            The other end of the rope is that the crocodillian can now pick up its pace, thoroughly [b]rutting[/b] you like the [b]basilisk bitch[/b] you are.
                            The less your groin hurts, the more your pride does.
                        [/p]
                        [p]
                            The basilisk, uncaring about your pain,
                            continues to dutifully hump away, its ridged shaft uncomfortably forcing itself just a little bit deeper inside of you.
                            Your belly would be bulging ever-so-slightly if not turned to stone.
                            While the rough fucking is one of the most degrading things you've ever experienced,
                            the animal luckily fails to last for more than a few minutes,
                            the tight confines of your ${cuntVirgin ? `virginal` : ``} slit proving far more pleasurable than the loins of its own kind,
                            even if that pleasure is far from mutual.
                        [/p]
                        [p]
                            You can already [b]sense[/b] its approaching orgasm.
                            Its movements become more erratic, a small growl rumbles from the back of its throat, and its large maw opens ever so slightly,
                            the creature seemingly [b]huffing[/b] and [b]panting[/b] through its mouth the same way a dog would.
                            As if you needed any more reminders that you're being rutted by nothing more than a lowly, feral [b]beast[/b].
                            Its incessant humping give way to a series of fierce thrusts, a clear indicator of what is to come.
                            A few moments later, the monster reaches its end,
                            wave after wave of warm, thick, [b]fertile[/b] basilisk spunk flooding your insides, filling you to the [b]brim[/b].
                        [/p]
                        [p]
                            Once its climax starts to fade, the large reptillian slowly shifts its weight off of your back,
                            and pulls its length from your slit.
                            A [b]flood[/b] of heady feral cum leaks from your slightly gaping twat,
                            the heavy scent of the stuff adding to the lingering odor of bestial masculinity hanging in the air,
                            which is almost strong enough to make you [b]retch[/b] and [b]gag[/b].
                            Your eyes water out of sheer disgust, a perfect way to mask the few tears that had been welling up since the start of this entire ordeal.
                        [/p]
                        [p]
                            Its member receding into its body, the creature crawls back towards the hole it came from,
                            and before you even have the [b]chance[/b] to [b]try[/b] to [b]shout obscenities[/b] at it,
                            the monster disappears back into the darkness, leaving you trapped, disgusted and crying,
                            unable to escape while forced to inhale the aroma of the puddle of feral cum dripping from your abused, well-bred slit.
                        [/p]`;
                if (cuntVirgin) {
                    toReturn += `
                        [p]
                            [c]
                                [b]You are no longer a virgin.[/b]
                            [/c]
                        [/p]`;
                }
            }
        }
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\stare\petrified\UndergroundCaveHate.js

// scenes\HamletDay2\UndergroundCave\stare\resist\UndergroundCaveResistFail.js

/**
 * Stare section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveResistFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.relent;
    var changeController = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveHate(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            [b]No![/b]
                            You can't let this happen!
                            You may have stared into the basilisk's eyes like an [b]idiot[/b], playing right into its cards, but this fight is [b]not[/b] over yet!
                            You begin to focus all of your willpower, all of your energy, [b]everything[/b] you've got on [b]moving[/b] your limbs.
                        [/p]
                        [p]
                            Your hand.
                            That first.
                            If you can move your hand, then you can move [b]anything[/b].
                            You close your eyes.
                            You take deep breaths through your nose.
                            You try to find some inner peace, not an easy task with a huge feral beast drawing near,
                            every scrape of its sharp claws over the stone cavern floor interrupting your meditation.
                            [b]Move, damnit![/b]
                        [/p]
                        ${global.ui.condFail(`Insufficient sanity, intelligence and endurance`)}
                        [p]
                            You try and you try, but your hand remains frozen in place.
                            You can move your fingers, but your palm budges not even an inch.
                            Losing faith, you open your eyes again, only to be greeted by the basilisk's toothy maw, mere inches from your petrified muzzle.
                            You'd be able to feel its warm breath on your scales, were they not turned to numb stone.
                            The creature flicks out its forked tongue, licking your rock-like snout.
                            Perhaps [i]testing[/i] if you're properly turned to stone or just [i]faking it[/i]?
                        [/p]
                        [p]
                            It's too late.
                            Even [b]if[/b] you managed to regain your motion somehow,
                            the beast would probably [b]get startled[/b] and [b]lash out[/b] if you suddenly rose to your feet.
                        [/p]
                        [p]
                            The monster stares deeply into your eyes, but you don't return its inquisitive, probing leer.
                            You've learned [b]not[/b] to peer into a basilisk's eyes, no matter how alluring.
                            Instead, you nervously look away, first to the side, then to the top of the cavern, and finally to the creature's underside.
                            Your sight wandering around the bottom end of the crocodillian's body, you realize that it's definitely [b]male[/b].
                            Its girthy, ridged and draconic-looking member hangs fully erect below its soft underbelly.
                            [i]Oh no.[/i]
                        [/p]
                        [p]
                            After the solemn tongue-flick, the beast lowers its head once more,
                            and begins to crawl around you, towards your backside.
                            Its intentions are [b]very[/b] clear, even to a dumb kobold like yourself.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(UndergroundCaveRelentChoice.HATE)
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\stare\resist\UndergroundCaveResistFail.js

// scenes\HamletDay2\UndergroundCave\stare\resist\UndergroundCaveResistSucceed.js

/**
 * Stare section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveResistSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.avert;
    var changeController = ctrl.resistSucceed;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.fight;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            [b]No![/b]
                            You can't let this happen!
                            You may have stared into the basilisk's eyes like an [b]idiot[/b], playing right into its cards, but this fight is [b]not[/b] over yet!
                            You begin to focus all of your willpower, all of your energy, [b]everything[/b] you've got on [b]moving[/b] your limbs.
                        [/p]
                        [p]
                            Your hand.
                            That first.
                            If you can move your hand, then you can move [b]anything[/b].
                            You close your eyes.
                            You take deep breaths through your nose.
                            You try to find some inner peace, not an easy task with a huge feral beast drawing near,
                            every scrape of its sharp claws over the stone cavern floor interrupting your meditation.
                            [b]Move, damnit![/b]
                        [/p]
                        ${global.ui.condSucc(`Sufficient sanity, intelligence and endurance`)}
                        [p]
                            Suddenly, something changes.
                            Your body feels [i]off balance[/i].
                            You open your eyes and look down, to find your right hand lifted ever-so slightly above the ground,
                            no longer supporting the rest of your petrified body.
                            You did it!
                            Through sheer [b]determination[/b] alone, you have managed to raise your hand off of the floor!
                        [/p]
                        [p]
                            Hope rising inside of your mind, you watch as the rocky exterior that had overtaken your limb gives way to your natural scales once more.
                            You realize that the petrification is [b]not[/b] a death sentence.
                            It's like a fast-spreading [b]disease[/b].
                            A [b]blight[/b] brought upon you by staring into the monster's oh-so-alluring eyes.
                            But not an uncurable one.
                            It can be [b]fought[/b].
                            It can be [b]beaten[/b].
                            You just need to [b]concentrate[/b]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(UndergroundCaveAvertChoice.FIGHT);
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\stare\resist\UndergroundCaveResistSucceed.js

// scenes\HamletDay2\UndergroundCave\stare\resist\UndergroundCaveResist.js

/**
 * Stare section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveResist(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveStareChoice.RESIST;
    var reqController = ctrl.stare;

    var toFail = new UndergroundCaveResistFail(ctrl);
    var toSuccess = new UndergroundCaveResistSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        var willpower = ctrl.misc.intelligence() + ctrl.misc.endurance();
        var sanityReq = 350 / willpower;
        return ctrl.misc.sanity() > sanityReq;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\stare\resist\UndergroundCaveResist.js

// scenes\HamletDay2\UndergroundCave\stare\relent\UndergroundCaveRelent.js

/**
 * Stare section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveRelent(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveStareChoice.RELENT;
    var reqController = ctrl.stare;
    var subController = ctrl.relent;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new UndergroundCaveHate(ctrl);
    var secondNext = new UndergroundCaveEnjoy(ctrl);

    var firstChoice = UndergroundCaveRelentChoice.HATE;
    var secondChoice = UndergroundCaveRelentChoice.ENJOY;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskFucked = true;
        var toReturn = `[p]
                            It's hopeless.
                            You gazed at the basilisk's eyes like an [b]idiot[/b], playing right into its cards.
                            Now most of your body has turned to stone, and you're left completely [b]defenceless[/b].
                            You are at the mercy of the hulking crocodillian, which slowly approaches you from the front.
                        [/p]
                        [p]
                            The creature draws near, crawling over the ground like the reptile it resembles.
                            Once close enough, it raises its upper body off of the ground,
                            positioning its toothy maw mere inches from your petrified muzzle.
                            You'd be able to feel its warm breath on your scales, were they not turned to numb stone.
                            The basilisk flicks out its forked tongue, licking your rock-like snout.
                            Perhaps [i]testing[/i] if you're properly turned to stone or just [i]faking it[/i]?
                        [/p]
                        [p]
                            The monster stares deeply into your eyes, but you don't return its inquisitive, probing leer.
                            While it's too late now, you've learned [b]not[/b] to peer into a basilisk's eyes, no matter how alluring.
                            Instead, you nervously look away, first to the side, then to the top of the cavern, and finally to the creature's underside.
                            Your sight wandering around the bottom end of the crocodillian's body, you realize that it's definitely [b]male[/b].
                            Its girthy, ridged and draconic-looking member hangs fully erect below its soft underbelly.
                            [i]Oh no.[/i]
                        [/p]
                        [p]
                            After the solemn tongue-flick, the beast lowers its head once more,
                            and begins to crawl around you, towards your backside.
                            Its intentions are [b]very[/b] clear, even to a dumb kobold like yourself.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Hate it</button>`;

        var tooltip = global.parser.parse(`Why the [b]hell[/b] would you enjoy getting fucked by a [b]wild beast[/b]?!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Love it</button>`;

        var tooltip = global.parser.parse(`Its member looked [i]exotic[/i]. This might end up being [b]fun[/b].`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var UndergroundCaveRelentChoice = {
    HATE: 1,
    ENJOY: 2
};

// END scenes\HamletDay2\UndergroundCave\stare\relent\UndergroundCaveRelent.js

// scenes\HamletDay2\UndergroundCave\stare\UndergroundCaveStare.js

/**
 * Stare section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveStare(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveSurfaceChoice.STARE;
    var reqController = ctrl.surface;
    var subController = ctrl.stare;
    var changeController = ctrl.stare;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new UndergroundCaveResist(ctrl);
    var secondNext = new UndergroundCaveRelent(ctrl);

    var firstChoice = UndergroundCaveStareChoice.RESIST;
    var secondChoice = UndergroundCaveStareChoice.RELENT;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.well.basiliskMet = true;
        global.kobold.history.shared.well.basiliskPetrified = true;
        var toReturn = `[p]
                            You can't look away.
                            If this is some kind of predator, then you [b]need[/b] to keep staring at it.
                            Doing anything else'll be seen as a sign of weakness, which might trigger it to attack.
                            Besides, if you stop looking, you might not be able to react in time if it charges you or something.
                        [/p]
                        [p]
                            So you stare.
                            You stare deep into the pair of reptillian eyes, hoping whatever creature is watching you'll just [i]go away[/i].
                            While you may look like it, you are [b]not[/b] a quick bite or an easy prey.
                            You'll fight to your very last breath, in every way you can.
                            Even if that means you have to glare at it for an entire [b]hour[/b], you will not back down, and you will not relent.
                        [/p]
                        [p]
                            Neither the creature, nor you are willing to back down.
                            Minutes pass.
                            You haven't moved an inch.
                            The sight of the alluring, enchanting bright green eyes,
                            combined with the constant sound of dripping of water is sort of [i]hypnotizing[/i].
                            [i]Relaxing[/i], even.
                            If you weren't in the middle of an intense standoff with a lethal monster,
                            you could easily fall asleep to the sound of the splashing droplets.
                        [/p]
                        [p]
                            More time passes.
                            How much more?
                            You don't know.
                            You sort of [i]lost yourself[/i] in the creature's eyes for a while there.
                            Its cold stare is [b]captivating[/b].
                            [b]Enthralling[/b].
                            Maybe this wasn't such a good idea after all.
                            You should, at the very least, change your posture to be better prepared for an eventual [i]attack[/i].
                            You're still on your hands and knees, as you've been ever since you woke up after passing out.
                            Hardly the best stance for [i]combat[/i].
                        [/p]
                        [p]
                            You try to rise to your feet, but for some reason, you [i]can't[/i].
                            Neither your legs, nor your arms seem to be [i]moving[/i].
                            Are your limbs asleep from being on your hands and knees for so long or something?
                            You try to tilt your head downwards to get a better look at your forearms, but even your neck refuses service!
                            Panicking, you break eye contact with the monster.
                            At least your eyeballs can still move.
                        [/p]
                        [p]
                            Looking down, you notice your arms are [i]discolored[/i].
                            Your scales are [i]grey[/i] instead of the usual [b]green[/b].
                            Are you [i]asleep[/i]?
                            Dreaming, perhaps?
                            Is this all a nightmare?
                            You try to open your mouth to scream, but even your muzzle isn't budging.
                            Staring at your own snout, you finally figure out what happened.
                            Stone.
                            Your scales have turned to stone.
                            You've been petrified!
                        [/p]
                        [p]
                            You try every muscle in your body, to see which parts you can still move.
                            Your tongue is still in working order, although your mouth won't open.
                            Your eyes can dart around all they please, but your head itself is unmoving.
                            You retain [i]some[/i] form of control over your tail, the largest of your extremities still in some manner of working order.
                            Fingers are a no-go.
                            Entirely petrified.
                            Your toes, finally.
                            You can wriggle those around slightly, despite both your legs and your feet being locked in place.
                        [/p]
                        [p]
                            While most of your smaller organs and limbs are still fully functional,
                            your body as a whole is completely immobilized.
                            Shocked by this revelation, you barely even notice the creature you were so afraid of leaving its hole,
                            stepping out into the cave proper.
                            Your cross-eyed stare at your own muzzle slowly fading, you become aware of the large,
                            quadrupedal, rock-hided crocodillian standing right in front of you.
                            Your heart nearly skips a beat as you realize what you're dealing with here.
                            A [b]basilisk[/b]!
                        [/p]
                        [p]
                            Seki once told you stories about creatures that could petrify
                            anyone or anything unlucky enough to stare at them for prolonged periods of time,
                            but you thought he was just making things up!
                            You should have listened more carefully to the old bird!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Resist</button>`;

        var tooltip = global.parser.parse(`You need to break loose from the monster's enchantment!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Relent</button>`;

        var tooltip = global.parser.parse(`It's already too late. There's no point to struggling. Let it happen.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.imageBackground.hide();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var UndergroundCaveStareChoice = {
    RESIST: 1,
    RELENT: 2
};

// END scenes\HamletDay2\UndergroundCave\stare\UndergroundCaveStare.js

// scenes\HamletDay2\UndergroundCave\run\UndergroundCaveRunContinue.js

/**
 * Run section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveRunContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.run;
    var changeController = ctrl.run;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Swimming in this condition is tough.
                            Your entire body feels [i]stiff[/i], probably due to all the stuff you've been through already.
                            Progress is slow, but the at least the beast isn't following you.
                            After a few minutes of tiring progress, your knackered limbs seem to [i]loosen[/i] a bit more,
                            slightly easing your journey to the entrance of the cave.
                        [/p]
                        [p]
                            A while later, you finally reach the grotto's ingress.
                            As you expected: the waterway led you straight to the river.
                            You swim to the riverbank, and clamber out of water one last time.
                            Sitting your butt down on the dirt, you take a breather.
                            You feel like you've earned one.
                        [/p]
                        [p]
                            You find yourself close to the road to Varanar.
                            In fact, you believe you passed this very place earlier, when heading for the hamlet.
                            You never knew there was a cave here, but then again,
                            you've never really bothered to scour the edge of the river for points of interest.
                        [/p]
                        [p]
                            You've defied death more than once today.
                            You should consider yourself lucky to be alive.
                            The trip into the well was a harrowing experience.
                            You aren't too sure what to take away from all of this.
                            All you know is that you never want to be stuck in another watering hole for the rest of your life.
                        [/p]
                        [p]
                            If nothing else, you've gained a few awesome stories to tell once you've grown old and wise.
                            Stories like the ones Seki used to tell you when you were little.
                            You went through all that trouble to get your mind off of him, and here you are reminiscing once more.
                            You just can't help yourself from clinging to times long gone.
                            Things were so much simpler back then.
                            You miss those days.
                        [/p]
                        [p]
                            The sun is slowly starting to set.
                            You drag yourself to your feet.
                            It's time to go.
                            Varanar isn't [i]that[/i] far off.
                            If your legs don't give out from underneath you, you might still make it there before nightfall.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\HamletDay2\UndergroundCave\run\UndergroundCaveRunContinue.js

// scenes\HamletDay2\UndergroundCave\run\UndergroundCaveRun.js

/**
 * Run section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveRun(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = UndergroundCaveSurfaceChoice.SWIM;
    var reqController = ctrl.surface;
    var subController = ctrl.run;
    var changeController = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new UndergroundCaveRunContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Nope.
                            You nearly got crushed, and then you almost drowned, but [b]mysterious cave creatures[/b] is where you draw the line.
                            You are [b]not[/b] ready for this, neither physically nor mentally.
                            It's time to do what kobolds do best: running the fuck away.
                            You may be tired, you may be worn out, you may have gone through [b]hell[/b] already,
                            but you figure you have a tiny [i]smidgen[/i] of adrenaline left, just enough for you to hightail it out of here.
                        [/p]
                        [p]
                            That creature, whatever it is, was here before you, so this must be its home or its nest or whatever you want to call it.
                            If you leave it be, it should have no business pursuing you.
                            Keeping your eyes on it to watch for any sudden movements, you slowly back away, towards the water.
                            You [i]could[/i] run alongside the river-branch, but you think you can move quicker while swimming,
                            plus you won't have to worry about falling down.
                            You crawl into the cold water once again, and breaking eye-contact, you begin to swim away.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.imageBackground.hide();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\run\UndergroundCaveRun.js

// scenes\HamletDay2\UndergroundCave\arrival\UndergroundCaveSurface.js

/**
 * Arrival section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveSurface(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.arrival;
    var subController = ctrl.surface;
    var changeController = ctrl.surface;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";
    var fourthId = arguments.callee.name + "ButtonFour";

    var firstNext = new UndergroundCaveRun(ctrl);
    var secondNext = new UndergroundCaveAvert(ctrl);
    var thirdNext = new UndergroundCaveStare(ctrl);
    var fourthNext = new UndergroundCaveReflection(ctrl);

    var firstChoice = UndergroundCaveSurfaceChoice.SWIM;
    var secondChoice = UndergroundCaveSurfaceChoice.AVERT;
    var thirdChoice = UndergroundCaveSurfaceChoice.STARE;
    var fourthChoice = UndergroundCaveSurfaceChoice.MIRROR;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With a throbbing head, and still-hurting lungs, you wearily awaken from your unpleasant slumber.
                            Your heavy eyelids are slow to open, the cave's dim lighting luckily not [i]too[/i] taxing on your semi-functional brain.
                            You don't know how long you were out for.
                            Heck, you can't even remember getting out of the flooded tunnel.
                            For a few minutes, you lie motionlessly, your mind and body both still recovering from the near-death experience.
                        [/p]
                        [p]
                            While your oxygen intake is steady, every breath you take causes serious chestpains.
                            You feel like you've ingested a box of rusty nails down your windpipe,
                            your respiratory system sore from the back of your throat to the depths of your lungs.
                            Your entire body aches, having been pushed to its very limits and beyond.
                            By all accounts, you should have been long dead,
                            yet through some miraculous combination of luck and determination,
                            or perhaps some kind of divine intervention, you have managed to make it out alive.
                        [/p]
                        [p]
                            You sigh loudly.
                            [b]Everything[/b] hurts.
                            All the oxygen deprivation has taken a hefty toll on your body, not to mention the physical strain.
                            You don't think you can stand up just yet.
                            But you can't keep lying here, either.
                            At the very least, you need to figure out where you are.
                        [/p]
                        [p]
                            You slowly get onto your unsteady hands and knees, and begin to look around.
                            You are definitely in a [i]cave[/i] of some sorts.
                            There's some faint traces of natural light, so there must be an exit somewhere.
                            The body of water you crawled out of stretches along the entire length of the cavern.
                            Must be an offshoot of the river, which means it [i]might[/i] lead back to the outside world.
                        [/p]
                        [p]
                            Besides the water, there isn't much to see.
                            A few holes line the cavern's walls, most of them too tiny -- or too [i]scary[/i] -- for you to squeeze into.
                            You figure they're tunnels, like the one you crawled in through, although yours was [i]submerged[/i].
                            Not that [i]adventure[/i] is on your mind after nearly [b]dying[/b].
                        [/p]
                        [p]
                            While tracing the walls of the cave looking for any [i]danger[/i], your eyes suddenly fall upon a pair of strange, twinkling, sparkling things.
                            Two green lights in the dark depths of one of the many holes.
                            Like wicked braziers in a sea of blackness.
                            They captivate you, their inherent beauty alluring you to stare mindlessly at them.
                            What could they be?
                            Wet, green stones?
                            Emeralds?
                            Jade crystals?
                        [/p]
                        [p]
                            Suddenly, the lights fade.
                            They flicker, if only for half a second.
                            Then you realize.
                            They're not gemstones or rocks.
                            They're [b]eyes[/b].
                            Something is staring at you from the darkness.
                            And it just blinked.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Swim away</button>`;

        var tooltip = global.parser.parse(`You can probably swim faster than you can run in the dark.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Look away</button>`;

        var tooltip = global.parser.parse(`Whatever this thing is, you don't trust its beady eyes.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Stare</button>`;

        var tooltip = global.parser.parse(`If a staring contest is what this thing wants, you're going to give it one!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFourthChoice() {
        var check = ctrl.misc.hasMisc(new SmallMirror());
        var disabled = subController.has() || !check ? "disabled" : "";
        var isSelected = subController.get() === fourthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fourthId}" ${disabled}>Reflect</button>`;

        var tooltip = global.parser.parse(`Use your mirror to reflect the glare.`);
        if (!check) {
            tooltip = global.parser.parse(`You don't have a mirror to reflect the glare with.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 150px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += `<div class="choicerow" style="margin-bottom: 10px;">`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        choice += `<div class="choicerow">`;
        choice += buildThirdChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFourthChoice();
        choice += `</div>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId), $("#" + fourthId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFourthListener() {
        $("#" + fourthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fourthChoice);
            loadNextPart(fourthNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.imageBackground.show("scenes/HamletDay2/UndergroundCave/img/eyes.gif");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        part += fourthNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();
        fourthNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
        registerFourthListener();
    };
}

var UndergroundCaveSurfaceChoice = {
    SWIM: 1,
    AVERT: 2,
    STARE: 3,
    MIRROR: 4
};

// END scenes\HamletDay2\UndergroundCave\arrival\UndergroundCaveSurface.js

// scenes\HamletDay2\UndergroundCave\arrival\UndergroundCaveArrival.js

/**
 * Arrival section of the Underground Cave.
 * @param {UndergroundCaveController} controller The Underground Cave controller.
 */
function UndergroundCaveArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.arrival;
    var changeController = ctrl.arrival;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.surface;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var drowned = global.kobold.history.shared.well.drowned;
        var toReturn = `[p]
                            You keep moving forwards through the tiny, cramped tunnel flooded with water,
                            your motions more akin to [b]crawling[/b] than swimming.
                            The shaft is far too small for you to practice any proper swimming strokes.
                            You end up digging the claws of both your hands and your feet into the submerged,
                            muddy dirt surrounding you on all sides, using it to propel you forwards, to clamber onwards.
                        [/p]
                        [p]
                            While the well was dark, there was still [i]some[/i] natural light shining in from the hole at the top.
                            No such luck here.
                            In the submerged pitch-blackness of these claustrophobic depths, not even your perfect night-sight can distinguish rock from mud.
                            Not that you [i]need[/i] to see much.
                            There's only one way to go.
                            And if your head bumps into something, you'll know you've reached the end, of not only the tunnel, but of your life.
                        [/p]
                        [p]
                            You crawl onwards for nearly an entire minute,
                            driven only by the will to live and the fading hope that there will be light at the end of the tunnel.
                            Your lungs ache.
                            They burn, desperately in [b]need[/b] of fresh air${drowned ? ", a feeling you are not entirely unfamiliar with" : ""}.
                            Moving is getting harder.
                            Your limbs begin to slow down.
                            Your extremities grow numb.
                            Your consciousness is fading.
                            The tunnel is widening.
                            Just a little further.
                        [/p]
                        [hl]
                        [p]
                            With a loud [b]splash[/b], you emerge from the water,
                            your half-gone mind barely even [b]registering[/b] the fact that you can breathe once more.
                            Through basic survival instinct, and nothing else, you take a first, small breath,
                            the tiniest sliver of oxygen breathing new life into your fading form.
                            Your eyes shoot open, greeted by the scarcely illuminated sight of a rock-hewn cave,
                            no doubt carved out by the elongated body of water you find yourself in.
                        [/p]
                        [p]
                            But where exactly you are is a concern for later.
                            For now, you greedily gulp down the precious gift of air, savoring it like a fine meal after being starved for [b]months[/b] on end.
                            It fans the flickering flame of your very existence, a fire which came [b]very[/b] close to being extinguished for good.
                            Feeling very lightheaded, like you might pass out still, you meekly swim to the water's stony edge,
                            and with your last ounce of energy, you pull yourself from the liquid, onto dry land.
                            Once secured on the stony cavern floor, your overexerted body collapses, and your weary mind slips into unconsciousness.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.bgm.loop();
        global.ui.applyTheme("hamlet_cave");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\HamletDay2\UndergroundCave\arrival\UndergroundCaveArrival.js

// scenes\HamletDay2\UndergroundCave\UndergroundCaveController.js

/**
 * Controller for the Underground Cave scene.
 * @param {string} sceneId The id of the scene.
 */
function UndergroundCaveController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.nextSceneReady = true;
    this.nextSceneName = "NightTwo";
    this.nextFolderName = "Varanar";

    this.misc = new MiscController(id, "misc");

    this.arrival = new ChangeController(id, "arrival", function () {
        global.kobold.status.physical.decrease(10, "Drowning");
        global.kobold.status.mental.decrease(9, "Near-death experience");
        global.kobold.status.stamina.decrease(10, "Drowning");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Drowning", "Health", -10),
            new Change("Near-death experience", "Sanity", -10),
            new Change("Still alive", "Sanity", +1),
            new Change("Underwater crawling", "Stamina", -5),
            new Change("Drowning", "Stamina", -5)
        ]);
    });

    // Actually a reset controller.
    this.surface = new UndoController(id, "surface", function change() {
        var healthBacklog = new ChoiceController(id, "surfaceHealthBacklog");
        var sanityBacklog = new ChoiceController(id, "surfaceSanityBacklog");
        var staminaBacklog = new ChoiceController(id, "surfaceStaminaBacklog");
        var exertionBacklog = new ChoiceController(id, "surfaceExertionBacklog");
        var healthLHBacklog = new ChoiceController(id, "surfaceLHHealthBacklog");
        var sanityLHBacklog = new ChoiceController(id, "surfaceLHSanityBacklog");
        var staminaLHBacklog = new ChoiceController(id, "surfaceLHStaminaBacklog");
        var exertionLHBacklog = new ChoiceController(id, "surfaceLHExertionBacklog");
        healthBacklog.set(global.kobold.status.physical.get());
        sanityBacklog.set(global.kobold.status.mental.get());
        staminaBacklog.set(global.kobold.status.stamina.get());
        exertionBacklog.set(global.kobold.status.exertion.get());
        healthLHBacklog.set(global.kobold.status.physical.lasthit);
        sanityLHBacklog.set(global.kobold.status.mental.lasthit);
        staminaLHBacklog.set(global.kobold.status.stamina.lasthit);
        exertionLHBacklog.set(global.kobold.status.exertion.lasthit);
    }, function changeNote() {
        return "";
    }, function undoChange() {
        var healthBacklog = new ChoiceController(id, "surfaceHealthBacklog");
        var sanityBacklog = new ChoiceController(id, "surfaceSanityBacklog");
        var staminaBacklog = new ChoiceController(id, "surfaceStaminaBacklog");
        var exertionBacklog = new ChoiceController(id, "surfaceExertionBacklog");
        var healthLHBacklog = new ChoiceController(id, "surfaceLHHealthBacklog");
        var sanityLHBacklog = new ChoiceController(id, "surfaceLHSanityBacklog");
        var staminaLHBacklog = new ChoiceController(id, "surfaceLHStaminaBacklog");
        var exertionLHBacklog = new ChoiceController(id, "surfaceLHExertionBacklog");
        global.kobold.status.physical.set(healthBacklog.get());
        global.kobold.status.physical.lasthit = healthLHBacklog.get();
        global.kobold.status.mental.set(sanityBacklog.get());
        global.kobold.status.mental.lasthit = sanityLHBacklog.get();
        global.kobold.status.stamina.set(staminaBacklog.get());
        global.kobold.status.stamina.lasthit = staminaLHBacklog.get();
        global.kobold.status.exertion.set(exertionBacklog.get());
        global.kobold.status.exertion.lasthit = exertionLHBacklog.get();
        global.kobold.history.shared.well.basiliskMet = false;
        global.kobold.history.shared.well.basiliskFought = false;
        global.kobold.history.shared.well.basiliskKilled = false;
        global.kobold.history.shared.well.basiliskFriend = false;
        global.kobold.history.shared.well.basiliskPetrified = false;
        global.kobold.history.shared.well.basiliskFucked = false;
        global.kobold.history.shared.well.basiliskPenetrated = false;
        global.kobold.history.shared.well.basiliskEnjoyed = false;
        global.kobold.history.shared.well.basiliskFingerDislocated = false;
        global.kobold.history.shared.well.basiliskDominated = false;
        global.kobold.history.shared.well.basiliskPeed = false;
        global.kobold.history.shared.well.basiliskOral = false;
        global.kobold.history.shared.well.basiliskDomFucked = false;
    });

    this.run = new ChangeController(id, "run", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.decrease(7, "Swimming while exerted");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Swimming", "Stamina", -10),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.stare = new UndoController(id, "stare", function () {
        global.kobold.status.mental.decrease(10, "Being petrified");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Petrification", "Sanity", -6),
            new Change("Panic", "Sanity", -4)
        ]);
    });

    this.relent = new ChoiceController(id, "relent");

    this.enjoy = new ChangeController(id, "enjoy", function () {
        var dick = that.misc.dick();
        var pants = that.misc.pants();
        var underwear = that.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;
        if (dick) {
            if (clothingMale) {
                that.misc.shirt() && global.kobold.inventory.equipment.shirt.stain("Basilisk Cum");
                global.kobold.status.mental.decrease(3, "Thighfucked by basilisk");
                global.kobold.status.exertion.increase(8, "Thighfucked by basilisk");
            } else {
                global.kobold.body.analVirgin = false;
                global.kobold.status.mental.increase(4, "Basilisk buttsex");
                global.kobold.status.exertion.increase(20, "Basilisk buttsex");
            }
        } else {
            if (clothingFemale) {
                that.misc.shirt() && global.kobold.inventory.equipment.shirt.stain("Basilisk Cum");
                global.kobold.status.mental.decrease(3, "Thighfucked by basilisk");
                global.kobold.status.exertion.increase(8, "Thighfucked by basilisk");
            } else {
                global.kobold.body.vaginalVirgin = false;
                global.kobold.status.mental.increase(10, "Basilisk sex");
                global.kobold.status.exertion.increase(25, "Basilisk sex");
            }
        }
    }, function () {
        var dick = that.misc.dick();
        var pants = that.misc.pants();
        var underwear = that.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;
        if (dick) {
            if (clothingMale) {
                return global.ui.change.buildNotification([
                    new Change("Excitement", "Sanity", +2),
                    new Change("Pleasure denied", "Sanity", -5),
                    new Change("Petrified-thigh-fucked", "Sexual exhaustion", +8, true)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Excitement", "Sanity", +2),
                    new Change("Thoroughly rutted", "Sanity", +7),
                    new Change("Pleasure denied", "Sanity", -5),
                    new Change("Ferally buttfucked", "Sexual exhaustion", +20, true)
                ]);
            }
        } else {
            if (clothingFemale) {
                return global.ui.change.buildNotification([
                    new Change("Excitement", "Sanity", +2),
                    new Change("Pleasure denied", "Sanity", -5),
                    new Change("Petrified-thigh-fucked", "Sexual exhaustion", +8, true)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Excitement", "Sanity", +2),
                    new Change("Thoroughly rutted", "Sanity", +5),
                    new Change("Reached climax", "Sanity", +5),
                    new Change("Only one orgasm", "Sanity", -2),
                    new Change("Bestially bred", "Sexual exhaustion", +20, true),
                    new Change("Reached orgasm", "Sexual exhaustion", +5, true)
                ]);
            }
        }
    });

    this.hate = new ChangeController(id, "hate", function () {
        var dick = that.misc.dick();
        var pants = that.misc.pants();
        var underwear = that.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;
        if (dick) {
            if (clothingMale) {
                that.misc.shirt() && global.kobold.inventory.equipment.shirt.stain("Basilisk Cum");
                global.kobold.status.mental.decrease(12, "Thighfucked by basilisk");
                global.kobold.status.exertion.increase(8, "Thighfucked by basilisk");
            } else {
                global.kobold.body.analVirgin = false;
                global.kobold.status.physical.decrease(5, "Raped by basilisk");
                global.kobold.status.mental.decrease(20, "Raped by basilisk");
                global.kobold.status.exertion.increase(25, "Raped by basilisk");
            }
        } else {
            if (clothingFemale) {
                that.misc.shirt() && global.kobold.inventory.equipment.shirt.stain("Basilisk Cum");
                global.kobold.status.mental.decrease(12, "Thighfucked by basilisk");
                global.kobold.status.exertion.increase(8, "Thighfucked by basilisk");
            } else {
                global.kobold.body.vaginalVirgin = false;
                global.kobold.status.physical.decrease(5, "Raped by basilisk");
                global.kobold.status.mental.decrease(20, "Raped by basilisk");
                global.kobold.status.exertion.increase(25, "Raped by basilisk");
            }
        }
    }, function () {
        var dick = that.misc.dick();
        var pants = that.misc.pants();
        var underwear = that.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;
        if (dick) {
            if (clothingMale) {
                return global.ui.change.buildNotification([
                    new Change("Defiled", "Sanity", -8),
                    new Change("Thigh-fucked by a feral", "Sanity", -8),
                    new Change("Didn't get penetrated", "Sanity", +4),
                    new Change("Petrified-thigh-raped", "Sexual exhaustion", +8, true)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Painful insertion", "Health", -5),
                    new Change("Defiled", "Sanity", -8),
                    new Change("Raped by a beast", "Sanity", -12),
                    new Change("Ferally buttfucked", "Sexual exhaustion", +20, true),
                    new Change("Raped", "Sexual exhaustion", +5, true)
                ]);
            }
        } else {
            if (clothingFemale) {
                return global.ui.change.buildNotification([
                    new Change("Defiled", "Sanity", -8),
                    new Change("Thigh-fucked by a feral", "Sanity", -8),
                    new Change("Didn't get penetrated", "Sanity", +4),
                    new Change("Petrified-thigh-raped", "Sexual exhaustion", +8, true)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Painful insertion", "Health", -5),
                    new Change("Defiled", "Sanity", -8),
                    new Change("Raped by a beast", "Sanity", -12),
                    new Change("Ferally bred", "Sexual exhaustion", +20, true),
                    new Change("Raped", "Sexual exhaustion", +5, true)
                ]);
            }
        }
    });

    function regainRelentChange() {
        var dick = that.misc.dick();
        var pants = that.misc.pants();
        var underwear = that.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;

        global.kobold.status.stamina.increase(3, "Minor reprieve");
        global.kobold.stats.decreaseStat(StatsEnum.AGILITY, 1);
        global.kobold.stats.increaseStat(StatsEnum.ENDURANCE, 1);

        if (dick) {
            if (clothingMale) {
                global.kobold.status.mental.decrease(3, "Sexual frustration");
                global.kobold.status.exertion.increase(2, "Sexual frustration");
            } else {
                global.kobold.status.mental.increase(1, "Escaped the cave");
                global.kobold.status.exertion.increase(2, "Sexual frustration");
            }
        } else {
            if (clothingFemale) {
                global.kobold.status.mental.decrease(3, "Sexual frustration");
                global.kobold.status.exertion.increase(2, "Sexual frustration");
            } else {
                global.kobold.status.mental.increase(1, "Escaped the cave");
            }
        }
    }

    function regainRelentNote() {
        var dick = that.misc.dick();
        var pants = that.misc.pants();
        var underwear = that.misc.underwear();
        var underwearMale = underwear && underwear.coversRear();
        var underwearFemale = underwear && underwear.coversGroin();
        var clothingMale = underwearMale ? underwear : false;
        clothingMale = pants ? pants : clothingMale;
        var clothingFemale = underwearFemale ? underwear : false;
        clothingFemale = pants ? pants : clothingFemale;
        if (dick) {
            if (clothingMale) {
                return global.ui.change.buildNotification([
                    new Change("Sexual frustration", "Sanity", -6),
                    new Change("Escaped the cave", "Sanity", +3),
                    new Change("Minor reprieve", "Stamina", +3),
                    new Change("Sexual frustration", "Sexual exhaustion", +2, true),
                    new Change("Petrification after effects", "Agility", -1),
                    new Change("Petrification after effects", "Endurance", +1)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Sexual frustration", "Sanity", -2),
                    new Change("Escaped the cave", "Sanity", +3),
                    new Change("Minor reprieve", "Stamina", +3),
                    new Change("Sexual frustration", "Sexual exhaustion", +2, true),
                    new Change("Petrification after effects", "Agility", -1),
                    new Change("Petrification after effects", "Endurance", +1)
                ]);
            }
        } else {
            if (clothingFemale) {
                return global.ui.change.buildNotification([
                    new Change("Sexual frustration", "Sanity", -6),
                    new Change("Escaped the cave", "Sanity", +3),
                    new Change("Minor reprieve", "Stamina", +3),
                    new Change("Sexual frustration", "Sexual exhaustion", +2, true),
                    new Change("Petrification after effects", "Agility", -1),
                    new Change("Petrification after effects", "Endurance", +1)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Minor sexual frustration", "Sanity", -2),
                    new Change("Escaped the cave", "Sanity", +3),
                    new Change("Minor reprieve", "Stamina", +3),
                    new Change("Petrification after effects", "Agility", -1),
                    new Change("Petrification after effects", "Endurance", +1)
                ]);
            }
        }
    }

    function regainHateChange() {
        global.kobold.status.mental.decrease(1, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
        global.kobold.stats.decreaseStat(StatsEnum.AGILITY, 1);
        global.kobold.stats.increaseStat(StatsEnum.ENDURANCE, 1);
    }

    function regainHateNote() {
        return global.ui.change.buildNotification([
            new Change("Re-living rape", "Sanity", -2),
            new Change("Self-pity", "Sanity", -2),
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3),
            new Change("Petrification after effects", "Agility", -1),
            new Change("Petrification after effects", "Endurance", +1)
        ]);
    }

    this.regain = new ChangeController(id, "regain", function () {
        if (that.relent.get() === UndergroundCaveRelentChoice.ENJOY) {
            regainRelentChange();
        } else {
            regainHateChange();
        }
    }, function () {
        if (that.relent.get() === UndergroundCaveRelentChoice.ENJOY) {
            return regainRelentNote();
        } else {
            return regainHateNote();
        }
    });

    this.resistFail = new ChoiceController(id, "resistFail");

    this.resistSucceed = new UndoController(id, "resistSucceed", function () {
        global.kobold.status.mental.increase(5, "Resisted petrification");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Resisted petrification", "Sanity", +5)
        ]);
    });

    this.avert = new UndoController(id, "avert", function () {
        global.kobold.status.mental.decrease(5, "Panicking");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Panic", "Sanity", -5)
        ]);
    });

    this.nuzzle = new ChangeController(id, "nuzzle", function () {
        global.kobold.status.mental.increase(6, "Cautious optimism");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Cautious optimism", "Sanity", +6)
        ]);
    });

    this.lick = new ChangeController(id, "lick", function () {
        global.kobold.status.physical.increase(8, "Wound healed");
        global.kobold.status.mental.increase(8, "Cautious optimism");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Wound healed", "Health", +8),
            new Change("Pain fading", "Sanity", +2),
            new Change("Cautious optimism", "Sanity", +6)
        ]);
    });

    this.doggo = new ChoiceController(id, "doggo");

    this.pet = new ChoiceController(id, "pet");

    this.scratch = new ChangeController(id, "scratch", function () {
        global.kobold.status.mental.increase(9, "Intensive petting");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Pet the basilisk", "Sanity", +5),
            new Change("Made a friend", "Sanity", +4)
        ]);
    });

    this.throat = new ChangeController(id, "throat", function () {
        global.kobold.status.mental.decrease(5, "Killing the basilisk");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Killed a vicious creature?", "Sanity", +3),
            new Change("Murdered an animal", "Sanity", -4),
            new Change("Betrayed trust", "Sanity", -4)
        ]);
    });

    this.throatLoot = new ChangeController(id, "throatLoot", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
        global.kobold.inventory.bag.addMisc(new BasiliskScales());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3),
            new Change("Trophies", global.kobold.inventory.createItemWithTooltip(new BasiliskScales(), true), +1)
        ]);
    });

    this.fetch = new ChangeController(id, "fetch", function () {
        that.misc.weapon();
        global.kobold.status.mental.increase(6, "Awkward game of fetch");
        global.kobold.inventory.equipment.strip(SlotEnum.WEAPON);
        global.kobold.inventory.equipment.wear(new PetrifiedBranch());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Played fetch", "Sanity", +2),
            new Change("Made a friend", "Sanity", +4),
            new Change("Branch petrified", global.kobold.inventory.createItemWithTooltip(that.misc.weapon(), true), -1),
            new Change("New weapon", global.kobold.inventory.createItemWithTooltip(new PetrifiedBranch(), true), +1)
        ]);
    });

    this.feed = new ChangeController(id, "feed", function () {
        global.kobold.status.mental.increase(6, "Fed animal");
        global.kobold.inventory.bag.removeConsumable(new HeartyMeal());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Fed animal", "Sanity", +2),
            new Change("Made a friend", "Sanity", +4),
            new Change("Lost a meal", global.kobold.inventory.createItemWithTooltip(new HeartyMeal(), true), -1)
        ]);
    });

    this.prePatch = new ChoiceController(id, "prePatch");

    this.basiPatch = new ChoiceController(id, "basiPatch");

    this.basiTake = new ChangeController(id, "basiTake", function () {
        global.kobold.inventory.equipment.wear(new Eyepatch());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Took eyepatch", global.kobold.inventory.createItemWithTooltip(new Eyepatch(), true), +1)
        ]);
    });

    this.basiLeave = new ChangeController(id, "basiLeave", function () {
        global.kobold.status.mental.decrease(1, "Left friend behind");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Left behind the basilisk", "Sanity", -4),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.fight = new ChoiceController(id, "fight");

    this.defence = new ChangeController(id, "defence", function () {
        global.kobold.status.physical.decrease(10, "Fighting the basilisk");
        global.kobold.status.mental.increase(5, "Scared off basilisk");
        global.kobold.status.stamina.decrease(10, "Defensive combat");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Broke hand", "Health", -10),
            new Change("Broke hand", "Sanity", -5),
            new Change("Scared off the basilisk", "Sanity", +10),
            new Change("Defensive combat", "Stamina", -10)
        ]);
    });

    /*this.defencePatch = new ChoiceController(id, "defencePatch");
     
     this.defencePatchTake = new ChangeController(id, "defencePatchTake", function () {
     
     }, function () {
     return global.ui.change.buildNotification([
     new Change("TODO", "CHANGE", +1)
     ]);
     });*/

    this.defenceLeave = new ChangeController(id, "defenceLeave", function () {
        global.kobold.status.physical.increase(2, "Fixed hand");
        global.kobold.status.mental.decrease(3, "Escaping the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Fixed hand", "Health", +2),
            new Change("Intense pain", "Sanity", -6),
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.offenceWin = new ChangeController(id, "offenceWin", function () {
        global.kobold.status.mental.decrease(6, "Killing the basilisk");
        global.kobold.status.stamina.decrease(8, "Combat");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Killed a monster", "Sanity", +4),
            new Change("Became the monster?", "Sanity", -10),
            new Change("Combat", "Stamina", -8)
        ]);
    });

    this.offenceLoot = new ChangeController(id, "offenceLoot", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
        global.kobold.inventory.bag.addMisc(new BasiliskScales());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3),
            new Change("Trophies", global.kobold.inventory.createItemWithTooltip(new BasiliskScales(), true), +1)
        ]);
    });

    this.wrestleWin = new UndoController(id, "wrestleWin", function () {
        global.kobold.status.physical.decrease(18, "Repeatedly crushed");
        global.kobold.status.mental.decrease(4, "Head crushed");
        global.kobold.status.stamina.decrease(15, "Intense struggle");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Repeatedly crushed", "Health", -18),
            new Change("Head crushed", "Sanity", -4),
            new Change("Intense struggle", "Stamina", -15)
        ]);
    });

    this.snap = new ChangeController(id, "snap", function () {
        global.kobold.status.mental.decrease(6, "Killing the basilisk");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Killed a monster", "Sanity", +4),
            new Change("Became the monster?", "Sanity", -10)
        ]);
    });

    this.snapLeave = new ChangeController(id, "snapLeave", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
        global.kobold.inventory.bag.addMisc(new BasiliskScales());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3),
            new Change("Trophies", global.kobold.inventory.createItemWithTooltip(new BasiliskScales(), true), +1)
        ]);
    });

    this.letGo = new ChangeController(id, "letGo", function () {
        global.kobold.status.physical.increase(3, "Drinking water");
        global.kobold.status.mental.increase(8, "Sparing the basilisk");
        global.kobold.status.stamina.increase(3, "Drinking water");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Fresh water", "Health", +3),
            new Change("Spared a life", "Sanity", +8),
            new Change("Fresh water", "Stamina", +3)
        ]);
    });

    this.letGoLeave = new ChangeController(id, "letGoLeave", function () {
        global.kobold.status.mental.increase(3, "Escaping the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.dominance = new UndoController(id, "dominance", function () {
        // Unused.
    }, function () {
        return global.ui.change.buildNotification([
            new Change("TODO", "CHANGE", +1)
        ]);
    });

    this.pee = new ChangeController(id, "pee", function () {
        global.kobold.status.mental.increase(6, "Peeing on a basilisk");
        global.kobold.status.stamina.increase(5, "Relaxing whizz");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Dominated foe", "Sanity", +10),
            new Change("Peeing on a feral", "Sanity", -4),
            new Change("Relaxing whizz", "Stamina", +5)
        ]);
    });

    this.oral = new ChangeController(id, "oral", function () {
        if (that.misc.dick()) {
            global.kobold.status.mental.increase(14, "Facefucked a basilisk");
            global.kobold.status.exertion.increase(16, "Facefucked a basilisk");

        } else {
            global.kobold.status.mental.increase(14, "Eaten out by a basilisk");
            global.kobold.status.exertion.increase(16, "Eaten out by a basilisk");
        }
    }, function () {
        if (that.misc.dick()) {
            return global.ui.change.buildNotification([
                new Change("Dominated foe", "Sanity", +12),
                new Change("Tempted fate", "Sanity", -2),
                new Change("Fucked a feral", "Sanity", -4),
                new Change("Reached climax", "Sanity", +8),
                new Change("Facefucked the basilisk", "Sexual exhaustion", +12, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Dominated foe", "Sanity", +12),
                new Change("Tempted fate", "Sanity", -2),
                new Change("Fucked a feral", "Sanity", -4),
                new Change("Reached climax", "Sanity", +8),
                new Change("Eaten out by the basilisk", "Sexual exhaustion", +12, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true)
            ]);
        }
    });

    this.sex = new ChangeController(id, "sex", function () {
        if (that.misc.dick()) {
            global.kobold.status.mental.increase(14, "Fucking a basilisk");
            global.kobold.status.stamina.decrease(7, "Fucking a basilisk");
            global.kobold.status.exertion.increase(24, "Fucking a basilisk");
        } else {
            global.kobold.status.mental.increase(6, "Abusing a basilisk");
            global.kobold.status.exertion.increase(5, "Abusing a basilisk");
        }
    }, function () {
        if (that.misc.dick()) {
            return global.ui.change.buildNotification([
                new Change("Dominated foe", "Sanity", +12),
                new Change("Tempted fate", "Sanity", -2),
                new Change("Fucked a feral", "Sanity", -4),
                new Change("Reached climax", "Sanity", +8),
                new Change("Intense sex", "Stamina", -7),
                new Change("Fucked the basilisk", "Sexual exhaustion", +20, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Dominated foe", "Sanity", +12),
                new Change("Tempted fate", "Sanity", -2),
                new Change("Fingered a feral", "Sanity", -2),
                new Change("Rimmed a feral", "Sanity", -2),
                new Change("Abused basilisk", "Sexual exhaustion", +3, true),
                new Change("Smell of cum", "Sexual exhaustion", +2, true)
            ]);
        }
    });

    this.dominanceLeave = new ChangeController(id, "dominanceLeave", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.preContinue = new ChoiceController(id, "preContinue");

    this.choices = {
        doggo: {
            PET: 1,
            FETCH: 2,
            FEED: 3
        }
    };

    this.reflectFail = new ChangeController(id, "reflectFail", function () {
        global.kobold.status.physical.decrease(2, "Cut hand");
        global.kobold.status.mental.decrease(10, "Being petrified");
        global.kobold.inventory.bag.removeMisc(ItemEnum.SMALL_MIRROR, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Cut hand", "Health", -6),
            new Change("Wound petrified", "Health", +4),
            new Change("Petrification", "Sanity", -6),
            new Change("Panic", "Sanity", -4),
            new Change("Mirror broke", global.kobold.inventory.createItemWithTooltip(new SmallMirror(), true), -1)
        ]);
    });

    this.reflectSucceed = new ChangeController(id, "reflectSucceed", function () {
        global.kobold.status.physical.decrease(6, "Cut hand");
        global.kobold.status.mental.increase(6, "Besting the basilisk");
        global.kobold.inventory.bag.removeMisc(ItemEnum.SMALL_MIRROR, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Cut hand", "Health", -6),
            new Change("Panic", "Sanity", -4),
            new Change("Heroism", "Sanity", +5),
            new Change("Bested basilisk", "Sanity", +5),
            new Change("Mirror broke", global.kobold.inventory.createItemWithTooltip(new SmallMirror(), true), -1)
        ]);
    });

    this.reflectStrike = new ChangeController(id, "reflectStrike", function () {
        global.kobold.status.mental.decrease(5, "Killing the basilisk");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Killed a vicious creature?", "Sanity", +3),
            new Change("Murdered an animal", "Sanity", -4),
            new Change("Did not accept surrender", "Sanity", -4)
        ]);
    });

    this.reflectStrikeContinue = new ChangeController(id, "reflectStrikeContinue", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
        global.kobold.inventory.bag.addMisc(new BasiliskScales());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3),
            new Change("Trophies", global.kobold.inventory.createItemWithTooltip(new BasiliskScales(), true), +1)
        ]);
    });

    this.reflectLeave = new ChangeController(id, "reflectLeave", function () {
        global.kobold.status.mental.increase(4, "Refrained from violence");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Refrained from violence", "Sanity", +4)
        ]);
    });

    this.reflectLeaveContinue = new ChangeController(id, "reflectLeaveContinue", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.reflectPetrify = new ChangeController(id, "reflectPetrify", function () {
        global.kobold.status.mental.decrease(2, "Petrifying the basilisk");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Killed a vicious creature?", "Sanity", +3),
            new Change("Murdered an animal", "Sanity", -4),
            new Change("Did not accept surrender", "Sanity", -4),
            new Change("Delicious irony", "Sanity", +3)
        ]);
    });

    this.reflectPetrifyContinue = new ChangeController(id, "reflectPetrifyContinue", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.reflectDominance = new ChoiceController(id, "reflectDominance");

    this.reflectOral = new ChangeController(id, "reflectOral", function () {
        if (that.misc.dick()) {
            global.kobold.status.mental.increase(14, "Facefucked a basilisk");
            global.kobold.status.exertion.increase(16, "Facefucked a basilisk");

        } else {
            global.kobold.status.mental.increase(14, "Eaten out by a basilisk");
            global.kobold.status.exertion.increase(16, "Eaten out by a basilisk");
        }
    }, function () {
        if (that.misc.dick()) {
            return global.ui.change.buildNotification([
                new Change("Served by basilisk", "Sanity", +12),
                new Change("Tempted fate", "Sanity", -2),
                new Change("Fucked a feral", "Sanity", -4),
                new Change("Reached climax", "Sanity", +8),
                new Change("Facefucked the basilisk", "Sexual exhaustion", +12, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Served by basilisk", "Sanity", +12),
                new Change("Tempted fate", "Sanity", -2),
                new Change("Fucked a feral", "Sanity", -4),
                new Change("Reached climax", "Sanity", +8),
                new Change("Eaten out by the basilisk", "Sexual exhaustion", +12, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true)
            ]);
        }
    });

    this.reflectOralContinue = new ChangeController(id, "reflectOralContinue", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.reflectSex = new ChangeController(id, "reflectSex", function () {
        if (that.misc.dick()) {
            global.kobold.status.mental.increase(14, "Fucking a basilisk");
            global.kobold.status.stamina.decrease(7, "Fucking a basilisk");
            global.kobold.status.exertion.increase(28, "Fucking a basilisk");
        } else {
            global.kobold.status.mental.increase(4, "Fingering a basilisk");
            global.kobold.status.exertion.increase(11, "Fingering a basilisk");
        }
    }, function () {
        if (that.misc.dick()) {
            return global.ui.change.buildNotification([
                new Change("Buttfucked basilisk", "Sanity", +12),
                new Change("Tempted fate", "Sanity", -2),
                new Change("Fucked a feral", "Sanity", -2),
                new Change("Sucked a feral", "Sanity", -2),
                new Change("Reached climax", "Sanity", +8),
                new Change("Intense sex", "Stamina", -7),
                new Change("Fucked the basilisk", "Sexual exhaustion", +20, true),
                new Change("Sucked the basilisk", "Sexual exhaustion", +4, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Basilisk blowjob", "Sanity", +12),
                new Change("Tempted fate", "Sanity", -2),
                new Change("Fingered a feral", "Sanity", -2),
                new Change("Rimmed a feral", "Sanity", -2),
                new Change("Sucked a feral", "Sanity", -2),
                new Change("Fingered basilisk", "Sexual exhaustion", +3, true),
                new Change("Sucked basilisk", "Sexual exhaustion", +4, true),
                new Change("Smell of cum", "Sexual exhaustion", +2, true),
                new Change("Taste of cum", "Sexual exhaustion", +2, true)
            ]);
        }
    });

    this.reflectSexContinue = new ChangeController(id, "reflectSexContinue", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.reflectSub = new ChoiceController(id, "reflectSub");

    this.reflectOutside = new ChangeController(id, "reflectOutside", function () {
        var shirt = global.kobold.inventory.equipment.shirt;
        shirt && shirt.stain("Basilisk Cum");
        if (that.misc.dick()) {
            global.kobold.status.physical.decrease(4, "Riding a basilisk");
            global.kobold.status.mental.increase(20, "Riding a basilisk");
            global.kobold.status.stamina.decrease(13, "Riding a basilisk");
            global.kobold.status.exertion.increase(30, "Riding a basilisk");
        } else {
            global.kobold.status.physical.decrease(4, "Riding a basilisk");
            global.kobold.status.mental.increase(20, "Riding a basilisk");
            global.kobold.status.stamina.decrease(13, "Riding a basilisk");
            global.kobold.status.exertion.increase(30, "Riding a basilisk");
        }
    }, function () {
        if (that.misc.dick()) {
            return global.ui.change.buildNotification([
                new Change("Very sore", "Health", -4),
                new Change("Rode basilisk", "Sanity", +10),
                new Change("Fucked a feral", "Sanity", -4),
                new Change("Kissed a feral", "Sanity", -2),
                new Change("Immense climax", "Sanity", +10),
                new Change("Scaly lover", "Sanity", +6),
                new Change("Overly intense sex", "Stamina", -13),
                new Change("Rode basilisk", "Sexual exhaustion", +26, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Very sore", "Health", -4),
                new Change("Rode basilisk", "Sanity", +10),
                new Change("Fucked a feral", "Sanity", -4),
                new Change("Kissed a feral", "Sanity", -2),
                new Change("Intense climax", "Sanity", +8),
                new Change("Scaly lover", "Sanity", +6),
                new Change("Overly intense sex", "Stamina", -13),
                new Change("Rode basilisk", "Sexual exhaustion", +26, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true)
            ]);
        }
    });

    this.reflectInside = new ChangeController(id, "reflectInside", function () {
        var shirt = global.kobold.inventory.equipment.shirt;
        shirt && shirt.stain("Basilisk Cum");
        if (that.misc.dick()) {
            global.kobold.status.physical.decrease(4, "Riding a basilisk");
            global.kobold.status.mental.increase(20, "Riding a basilisk");
            global.kobold.status.stamina.decrease(13, "Riding a basilisk");
            global.kobold.status.exertion.increase(35, "Riding a basilisk");
        } else {
            global.kobold.status.physical.decrease(4, "Riding a basilisk");
            global.kobold.status.mental.increase(16, "Riding a basilisk");
            global.kobold.status.stamina.decrease(13, "Riding a basilisk");
            global.kobold.status.exertion.increase(35, "Riding a basilisk");
        }
    }, function () {
        if (that.misc.dick()) {
            return global.ui.change.buildNotification([
                new Change("Very sore", "Health", -4),
                new Change("Rode basilisk", "Sanity", +10),
                new Change("Fucked a feral", "Sanity", -4),
                new Change("Kissed a feral", "Sanity", -2),
                new Change("Immense climax", "Sanity", +10),
                new Change("Scaly lover", "Sanity", +6),
                new Change("Overly intense sex", "Stamina", -13),
                new Change("Rode basilisk", "Sexual exhaustion", +26, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true),
                new Change("It came inside", "Sexual exhaustion", +5, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Very sore", "Health", -4),
                new Change("Rode basilisk", "Sanity", +10),
                new Change("Fucked a feral", "Sanity", -4),
                new Change("Kissed a feral", "Sanity", -2),
                new Change("Double climax", "Sanity", +10),
                new Change("Scaly lover", "Sanity", +6),
                new Change("Pregnant?", "Sanity", -4),
                new Change("Overly intense sex", "Stamina", -13),
                new Change("Rode basilisk", "Sexual exhaustion", +26, true),
                new Change("Reached climax", "Sexual exhaustion", +4, true),
                new Change("It came inside", "Sexual exhaustion", +5, true)
            ]);
        }
    });

    this.reflectSubContinue = new ChangeController(id, "reflectSubContinue", function () {
        global.kobold.status.mental.increase(3, "Escaped the cave");
        global.kobold.status.stamina.increase(3, "Minor reprieve");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Escaped the cave", "Sanity", +3),
            new Change("Minor reprieve", "Stamina", +3)
        ]);
    });

    this.fragments = {};
    this.fragments.fight = new UndergroundCaveFight(this);
    this.fragments.surface = new UndergroundCaveSurface(this); // Needs to be separate or else can't access the fragments variable itself.

    this.bgm = new AudioController(id, "res/audio/effects/cave.mp3", 70);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

}

// END scenes\HamletDay2\UndergroundCave\UndergroundCaveController.js

// scenes\HamletDay2\UndergroundCave\UndergroundCave.js

/**
 * Underground cave after the well.
 */
function UndergroundCave() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "UndergroundCave";
    var sceneName = "Underground Cave";
    var sceneSubfolder = "HamletDay2";
    var ctrl = new UndergroundCaveController(sceneId);
    var arrival = new UndergroundCaveArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]the CaVe[/t]`);
        tab += global.parser.parse(`[st]Underwater tunnel[/st]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new UndergroundCave();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\HamletDay2\UndergroundCave\UndergroundCave.js



// END SCENE UndergroundCave



// SCENE NightTwo

// scenes\Varanar\NightTwo\confront\crude\appease\eager\NightTwoCrudeEagerContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeEagerContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Crude eager continue.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.crudeEager.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.crudeEager.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\crude\appease\eager\NightTwoCrudeEagerContinue.js

// scenes\Varanar\NightTwo\confront\crude\appease\eager\NightTwoCrudeEager.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeEager(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoCrudeEagerContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Crude eager.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.crudeEager.continued() ? "disabled" : "";
        var isSelected = ctrl.crudeEager.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.crudeEager.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.appease.get() !== NightTwoCrudeAppeaseChoice.EAGER) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.crudeEager.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.appease.get() !== NightTwoCrudeAppeaseChoice.EAGER) {
            return;
        }

        next.registerListeners();

        if (ctrl.crudeEager.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\crude\appease\eager\NightTwoCrudeEager.js

// scenes\Varanar\NightTwo\confront\crude\appease\reluctant\NightTwoCrudeReluctantContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeReluctantContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You spend a good half hour stuck in between wanting to vomit, and not being able to.
                            While you wisely held your breath while down on your knees,
                            the small bit of sweat still staining your snout carries an unpleasant scent along with it.
                            The smell is probably not [i]that[/i] intense, but being smeared around your nostrils, it taints every breath you take.
                        [/p]
                        [p]
                            As you finally manage to compose yourself a little, the sun starts to rise above the rooftops.
                            You'd rather have spent the remainder of the night in bed, as opposed to on your knees and bent over trying to throw up, but alas.
                            You head back towards the town square, where your [i]urban adventure[/i] started.
                            Your first night in Varanar is over, and your first day is just about to begin.
                            If there's anything you've learned, it's that you need to keep your guard up at all times.
                            The next group of humans to corner you might not be sated with a small peck on their buttcheeks.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.crudeReluctant.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.crudeReluctant.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\crude\appease\reluctant\NightTwoCrudeReluctantContinue.js

// scenes\Varanar\NightTwo\confront\crude\appease\reluctant\NightTwoCrudeReluctant.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeReluctant(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoCrudeReluctantContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.nightTwo.buttKiss = true;
        var toReturn = `[p]
                            You try your hardest to look for a way out of this mess you've created, but your brain draws a blank.
                            It seems as though you've talked yourself into a corner.
                            You can simply [i]refuse[/i] to play along, but judging by the group's intense behavior,
                            you can make a pretty decent guess that things [b]won't[/b] turn out nicely if you back out now.
                            The best course of action, for now, would be to just [i]play along[/i].
                            If you do ask they ask of you, this'll all be over soon...you hope.
                        [/p]
                        [p]
                            "[color=bred]Well? Don't keep me waiting.[/color]" the half-naked human's voice snaps you back to reality.
                            He bends over slightly, his unsavory ass jutting out towards you.
                            You can feel your stomach churning.
                            It's just so [b]hairy[/b] and [b]disgusting[/b].
                            You'd rather kiss a [b]horse[/b]'s rear end.
                            You [b]really[/b] don't want to do this, but the men have left you with little choice.
                        [/p]
                        [p]
                            You sigh loudly and take a deep breath of cool, late-night air.
                            Your lungs filled with clean air, you drop to your knees, and hold your breath.
                            Closing your eyes, you lean forwards, and plant a single small kiss on the human's right buttcheek,
                            your snout pressing against his damp skin.
                            As you pull back from the simple peck, the group bursts out into laughter.
                        [/p]
                        [p]
                            You are completely disgusted.
                            You wipe off the sweat clinging to the front of your muzzle.
                            Rising to your feet once more, an intense feeling of [b]sickness[/b] washes over you.
                            You heave, retch and gag.
                            It's like you're on the verge of vomiting, without being able to hurl and get it over with already.
                            All the while, the humans are taunting you.
                            "[color=green]The lizard actually did it![/color]"
                            "[color=brown]You kiss your mother with that mouth?![/color]"
                            "[color=purple]Ass-kissing kobold![/color]"
                            "[color=dred]Bet it tastes better than what you're used to eating![/color]"
                        [/p]
                        [p]
                            The exposed man pulls up his pants.
                            While the crowd slowly simmers down, he addresses his comrade.
                            "[color=bred]Told you they were good at something, Jacoby.
                            You owe us a round at the tavern![/color]"
                            The group bursts into cheer.
                            Pushing you aside, they break their encircling formation, and wander off towards the brothel district you came from,
                            leaving you alone in the street, struggling to overcome your sheer revulsion.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.crudeReluctant.continued() ? "disabled" : "";
        var isSelected = ctrl.crudeReluctant.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.crudeReluctant.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.appease.get() !== NightTwoCrudeAppeaseChoice.RELUCTANT) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.crudeReluctant.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.appease.get() !== NightTwoCrudeAppeaseChoice.RELUCTANT) {
            return;
        }

        next.registerListeners();

        if (ctrl.crudeReluctant.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\crude\appease\reluctant\NightTwoCrudeReluctant.js

// scenes\Varanar\NightTwo\confront\crude\appease\unfit\NightTwoCrudeUnfitContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeUnfitContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With the humans leaving, you decide that you, too, should probably get moving, lest they change their minds.
                            You begin to make your way back towards the city square, where your nighttime adventure began.
                            While you walk, you ponder the events that just transpired.
                            Were it not for your clever way of handling the group of drunkards, things could have gone [b]very[/b] bad for you.
                            You're glad you made it out unharmed, with your dignity mostly intact.
                        [/p]
                        [p]
                            As you near the marketplace, the sun slowly rises above the rooftops.
                            You guess you won't be getting any sleep tonight.
                            Oh well.
                            Taking a deep breath of fresh morning air, you steel yourself for your first day in Varanar.
                            You'll need to be on your guard if it's to be anything like your first [i]night[/i].
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.crudeUnfit.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.crudeUnfit.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\crude\appease\unfit\NightTwoCrudeUnfitContinue.js

// scenes\Varanar\NightTwo\confront\crude\appease\unfit\NightTwoCrudeUnfit.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeUnfit(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoCrudeUnfitContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You say you are [b]flattered[/b] that the men would even [b]consider[/b] letting a [b]lowly[/b] kobold as yourself kiss their behinds,
                            but you are afraid that you simply [b]can't[/b].
                            You see, you have a [b]rare[/b] and [b]dangerous[/b] disease called [b]tongorrhea[/b],
                            which pretty much means that if your tongue so much as [b]comes close[/b] to their skin, they'll get [b]infected[/b],
                            and their [b]dicks[/b] will [b]shrivel[/b] and [b]fall off[/b] in a couple of months.
                            Not to mention the horrible [b]itching[/b].
                            Their wives won't be too pleased if they go home with a [b]nasty[/b] kobold disease.
                            Surely they can find a non-infected member of your kind to settle this [i]dispute[/i], right?
                        [/p]
                        [p]
                            The group bursts into disarray.
                            "[color=purple]Mangy kobold![/color]"
                            "[color=bred]Jacoby picked this one because he knew he was going to lose the bet![/color]"
                            "[color=green]Hey now, [b]you're[/b] the one with your bare ass close to the diseased vermin's snout![/color]"
                            Realizing his pants are still down, the exposed human in front of you quickly re-dresses, before turning around to face you once more.
                            "[color=bred]Lizard.[/color]" he speaks, drawing your attention.
                            "[color=bred]Thanks for telling us.
                            Wouldn't want to catch any infections from your kind.[/color]"
                            You wave away his thanks, and mention that you're glad things worked out for the better, for [b]both[/b] of you.
                        [/p]
                        [p]
                            Their interest in you vanishing, but their argument still going strong, the drunkards slowly break their encirclement.
                            "[color=brown]Let's go find another one![/color]"
                            "[color=purple]At this hour? Where?[/color]"
                            "[color=dred]To the lizard district![/color]"
                            Like they've completely forgotten you even exist,
                            the group begins to head for the kobold district, leaving you all alone in the middle of the street.
                            Just the way you like it.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.crudeUnfit.continued() ? "disabled" : "";
        var isSelected = ctrl.crudeUnfit.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.crudeUnfit.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.appease.get() !== NightTwoCrudeAppeaseChoice.UNFIT) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.crudeUnfit.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.appease.get() !== NightTwoCrudeAppeaseChoice.UNFIT) {
            return;
        }

        next.registerListeners();

        if (ctrl.crudeUnfit.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\crude\appease\unfit\NightTwoCrudeUnfit.js

// scenes\Varanar\NightTwo\confront\crude\appease\NightTwoCrudeAppease.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeAppease(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reluctantId = arguments.callee.name + "ButtonOne";
    var eagerId = arguments.callee.name + "ButtonTwo";
    var unfitId = arguments.callee.name + "ButtonThree";

    var toReluctant = new NightTwoCrudeReluctant(ctrl);
    var toEager = new NightTwoCrudeEager(ctrl);
    var toUnfit = new NightTwoCrudeUnfit(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            No matter what you say, they'll find a way to turn it against you,
                            but if you don't play along, you doubt things would end well.
                            If you just give them the laugh they're after, you're certain they'll leave you alone.
                            Besides, if you show the group that you can take a [i]joke[/i] just as well as any human can,
                            perhaps they'll get bored and find a different target to pick on.
                        [/p]
                        [p]
                            Forsaking your pride, you half-heartedly stammer out that yes, kobolds are [i]indeed[/i] good at [b]kissing ass[/b].
                            You can barely even get those final two words past your lips, your mind screaming at you [b]not[/b] to put yourself down like this.
                            But, in a way, the statement isn't even [i]wrong[/i].
                            While most kobolds are [i]scoundrels[/i] of some form or another, the ones that aren't tend to be [b]suck-ups[/b],
                            relying on innate [b]cuteness[/b] and [b]flattering words[/b] in the hopes of being tossed a morsel or a coin every now and again.
                            In that sense: yes. Your kind [i]is[/i] good at [b]kissing ass[/b].
                        [/p]
                        [p]
                            The entire group bursts out into laughter at the sight of a kobold humiliating itself.
                            The drunken men begin to taunt you anew, although the mood is slightly more cheerful now than it was before,
                            their insults clearly less meant to [b]hurt[/b] you this time around.
                            It seems as though your gamble has paid off.
                        [/p]
                        [p]
                            One of the men still seems displeased.
                            As the posse begins to quiet down, he raises his voice:
                            "[color=green]Kobolds are [b]liars[/b].
                            Who knows if this one is telling the truth?!
                            Make'em [b]prove[/b] it![/color]"
                            At first, the group pays the dissenter no mind.
                            Slowly, but steadily, however, cogs begin to turn inside of their drunken heads.
                            As they come to a communal conclusion, some of the men begin to voice their response in unison:
                            "[color=purple]Yeah. Make'em prove it![/color]"
                            "[color=brown]The kobold lies![/color]"
                            "[color=dred]Put on a show![/color]"
                        [/p]
                        [p]
                            The man in front of you realizes he's losing the crowd's favour.
                            A grin forms atop his visage.
                            "[color=bred]Jacoby's right.[/color]" he speaks, his hands travelling down to his waist, resting atop his belt buckle.
                            "[color=bred]You lizards are a bunch of liars.[/color]"
                            He unfastens the clasp sealing the leather cord, his pants visible loosening.
                            "[color=bred]And we don't take kindly to liars around these parts.[/color]"
                            That being said, the drunkard turns around, facing away from you before promptly lowering his pants and his underwear,
                            his sweaty and hairy, undoubtedly [b]masculine[/b] rear end coming into view.
                            "[color=bred]Let's see if you were telling the truth or not.[/color]"
                        [/p]
                        [p]
                            The entire group seems riled up.
                            They [b]really[/b] want this to happen, and judging by the exposed human's tone,
                            you're in too deep to back out now.
                            Unless you can present a strong case for yourself, you fear this [b]is[/b] happening.
                            That doesn't mean you need to like it, though.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildReluctantChoice() {
        var disabled = ctrl.appease.has() ? "disabled" : "";
        var isSelected = ctrl.appease.get() === NightTwoCrudeAppeaseChoice.RELUCTANT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${reluctantId}" ${disabled}>Reluctant</button>`;

        var tooltip = global.parser.parse(`You don't really want to humiliate yourself for these guys' entertainment, but you don't have much of a choice.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildEagerChoice() {
        var disabled = ctrl.appease.has() ? "disabled" : "";
        var isSelected = ctrl.appease.get() === NightTwoCrudeAppeaseChoice.EAGER;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${eagerId}" ${disabled}>Eager</button>`;

        var tooltip = global.parser.parse(`Give the men a show they'll never forget.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildUnfitChoice() {
        var test = ctrl.misc.intelligence() > 5 && ctrl.misc.charisma() > 5;
        var disabled = (ctrl.appease.has() || !test) ? "disabled" : "";
        var isSelected = ctrl.appease.get() === NightTwoCrudeAppeaseChoice.UNFIT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${unfitId}" ${disabled}>Unfit</button>`;

        var tooltip = global.parser.parse(`Surely these men can find a more fitting kobold to abuse, right?`);
        if (!test) {
            tooltip = `Requires 6 intelligence and 6 charisma.`;
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You are: </div>`;
        choice += buildReluctantChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildUnfitChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + reluctantId), $("#" + eagerId), $("#" + unfitId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerReluctantListener() {
        $("#" + reluctantId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.appease.set(NightTwoCrudeAppeaseChoice.RELUCTANT);
            loadNextPart(toReluctant);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerEagerListener() {
        $("#" + eagerId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.appease.set(NightTwoCrudeAppeaseChoice.EAGER);
            loadNextPart(toEager);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerUnfitListener() {
        $("#" + unfitId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.appease.set(NightTwoCrudeAppeaseChoice.UNFIT);
            loadNextPart(toUnfit);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.lewdCrude.get() !== NightTwoCrudeChoice.APPEASE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toReluctant.build();
        part += toEager.build();
        part += toUnfit.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.lewdCrude.get() !== NightTwoCrudeChoice.APPEASE) {
            return;
        }

        toReluctant.registerListeners();
        toEager.registerListeners();
        toUnfit.registerListeners();

        if (ctrl.appease.has()) {
            return;
        }

        registerReluctantListener();
        registerEagerListener();
        registerUnfitListener();
    };
}

var NightTwoCrudeAppeaseChoice = {
    RELUCTANT: 1,
    EAGER: 2,
    UNFIT: 3
};

// END scenes\Varanar\NightTwo\confront\crude\appease\NightTwoCrudeAppease.js

// scenes\Varanar\NightTwo\confront\crude\rude\fight\lose\NightTwoCrudeLose.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeLose(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    var idsToRemove = [
        "NightTwoCrudeLose",
        "NightTwoCrudeFight",
        "NightTwoCrudeRude",
        "NightTwoCrude",
        "NightTwoConfront",
        "NightTwoRunIn",
        "NightTwoStage"
    ];

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            No.
                            You aren't going to let them do this.
                            You can't and you [b]won't[/b] get [b]anywhere[/b] near that filthy human's disgusting rear, let alone pressing your lips against it!
                            [b]Over your dead body.[/b]
                            They may be stronger than you, they may outnumber you, they may have you in an arm-grip,
                            but all of that just means they [b]won't[/b] be [i]expecting[/i] you to go on the offensive.
                        [/p]
                        [p]
                            They think they have you scared shitless, that you're just another kobold for them to [b]bully[/b] and [b]abuse[/b],
                            that you won't put up a fight simply because this is [b]their[/b] town.
                            But you aren't any ordinary run-of-the-mill lizard.
                            The other members of your kind may be content being thieves, or bandits or prostitutes or beggars, but not [b]you[/b].
                            You may be a thief right now, but you [b]aspire[/b] to be a [b]knight[/b].
                            And knights don't back down in the face of adversity.
                            They charge head-forward into danger.
                            Just like you're about to do.
                        [/p]`;

        if (weapon) {
            toReturn += `[p]
                            While one of your arms is being restrained, you reach for your ${global.kobold.inventory.createItemWithTooltip(weapon)} with the other.
                            With your head being forced ever-closer to the drunkard's butt, you decide to strike before your snout brushes against his body hair.
                            Swiftly and silently, you [b]surge[/b] your weapon forwards, intent on striking the oaf's rear,
                            yet before your strike manages to connect, your free arm is grabbed and painfully twisted, much like the other.
                            With a pained yelp, you drop your ${global.kobold.inventory.createItemWithTooltip(weapon)} to the ground,
                            where it is quickly picked up by one of the humans, who promptly [b]bashes[/b] you on the head with it.
                        [/p]`;
        } else {
            toReturn += `[p]
                            While one of your arms is being restrained, you lash out with the other, intent on striking the oaf's rear,
                            yet before your strike manages to connect, your free arm is grabbed and painfully twisted, much like the other.
                            You let out a pained yelp.
                            Your attempted aggression is soon punished, one of the humans brutally [b]bashing[/b] the top of your head with his balled up hand.
                        [/p]`;
        }

        toReturn += `   ${global.ui.condFail(weapon ? `${global.kobold.inventory.createItemWithTooltip(weapon, true)} offence below 4` : `No weapon`)}
                        [p]
                            Your vision fades to black.
                            You lose consciousness.
                            You dream about a ceremony.
                            One in your honor.
                            You are in a castle.
                            There's princes and princesses, barons and dukes, royals from all corners of the world.
                            You are kneeling.
                            The Dragon King himself puts a gilded blade atop your shoulder.
                            Then suddenly, the dream turns into a nightmare, the castle replaced by a terrible, all-consuming void that sucks you into its core.
                            Once the void has consumed all, you are left with nothing, not even yourself.
                            The nightmare ends, and so do you.
                            In the real world, the men kicked you to death for passing out instead of doing their bidding.
                            Luckily, you were not awake to witness the horrible pain inflicted upon you.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    Nightmare ending.
                                [/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" style="margin-left: auto; position: relative; float: right;" id="${continueId}">`;
        choice += `&lt;&lt Revert`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400, function () {
            var main = $("#main");
            main.scrollTop(main[0].scrollHeight - main.height() - 5); // Fix scrolling
        });
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.misc.unsetAll();
            ctrl.rude.undo();
            ctrl.rude.unset();
            ctrl.lewdCrude.undo();
            ctrl.lewdCrude.unset();
            ctrl.runin.unset();
            ctrl.stage.unset();
            for (var i = 0; i < idsToRemove.length; i++) {
                $("#" + idsToRemove[i]).remove();
            }
            loadNextPart(new NightTwoStage(ctrl));
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        registerContinueListener();
    };
}

// END scenes\Varanar\NightTwo\confront\crude\rude\fight\lose\NightTwoCrudeLose.js

// scenes\Varanar\NightTwo\confront\crude\rude\fight\win\NightTwoCrudeWinContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeWinContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            As your adrenaline rush fades, the harsh reality of things comes crashing down on you.
                            You just [b]attacked[/b] and seriously [b]hurt[/b] a bunch of humans.
                            They might have been bothering you, but the guards won't care.
                            To them, you're a stupid, unruly lizard that crossed the line.
                            The men you assaulted had rights.
                            You don't.
                        [/p]
                        [p]
                            You make the smart decision to vacate the scene of the crime.
                            After jogging a few streets over, you lower your pace to a brisk walk.
                            You think you're in the clear.
                            If the famous guards of Varanar were called, you wouldn't have made it this far.
                            You'd have been captured or killed before you even left the street.
                            Still, you probably shouldn't linger in this particular area for too long.
                        [/p]
                        [p]
                            As you begin to make your way back to the town square, where your late-night adventure began, the sun slowly begins to rise above the rooftops.
                            It seems as though you've managed to retain your dignity throughout the night, but at what cost?
                            You don't want to think about it.
                            It's too late, or rather [i]too early[/i] to go to bed.
                            Who needs sleep?
                            You'd just have bad dreams about the things you did anyway.
                            You know you acted in self-defence, but your conscience doesn't agree. 
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.fightWin.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.fightWin.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\crude\rude\fight\win\NightTwoCrudeWinContinue.js

// scenes\Varanar\NightTwo\confront\crude\rude\fight\win\NightTwoCrudeWin.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeWin(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoCrudeWinContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        global.kobold.history.shared.varanar.nightTwo.fought = true;
        var toReturn = `[p]
                            No.
                            You aren't going to let them do this.
                            You can't and you [b]won't[/b] get [b]anywhere[/b] near that filthy human's disgusting rear, let alone pressing your lips against it!
                            [b]Over your dead body.[/b]
                            They may be stronger than you, they may outnumber you, they may have you in an arm-grip,
                            but all of that just means they [b]won't[/b] be [i]expecting[/i] you to go on the offensive.
                        [/p]
                        [p]
                            They think they have you scared shitless, that you're just another kobold for them to [b]bully[/b] and [b]abuse[/b],
                            that you won't put up a fight simply because this is [b]their[/b] town.
                            But you aren't any ordinary run-of-the-mill lizard.
                            The other members of your kind may be content being thieves, or bandits or prostitutes or beggars, but not [b]you[/b].
                            You may be a thief right now, but you [b]aspire[/b] to be a [b]knight[/b].
                            And knights don't back down in the face of adversity.
                            They charge head-forward into danger.
                            Just like you're about to do.
                        [/p]
                        [p]
                            While one of your arms is being restrained, you reach for your ${global.kobold.inventory.createItemWithTooltip(weapon)} with the other.
                            With your head being forced ever-closer to the drunkard's butt, you decide to strike before your snout brushes against his body hair.
                            Swiftly and silently, you [b]surge[/b] your weapon forwards,
                            shoving the pointy-end directly into the ruffian's exposed ass, stabbing him in his buttcheek.
                            The man screams in pain.
                            His companions laugh out loud, thinking you bit him.
                            Their laughter comes to a halt, however, as the group slowly realizes what has transpired.
                        [/p]
                        [p]
                            You pull your weapon from the pantless delinquent's rear, and in one swift move, you bring it to bear on the person holding your arm.
                            Blindly stabbing backwards, you land a seemingly solid strike on the human's abdomen.
                            He loses his hold on your arm nigh-instantly as your ${global.kobold.inventory.createItemWithTooltip(weapon)} sinks into his gut.
                            The man whose butt you perforated tumbles to the ground.
                            Blood oozes liberally from his mutilated ass.
                            The guy behind you stumbles backwards, clutching his wounded belly.
                            Communal panic settles in.
                            [i]Perfect[/i].
                        [/p]
                        ${global.ui.condSucc(`${global.kobold.inventory.createItemWithTooltip(weapon, true)} offence greater than 3`)}
                        [p]
                            Your freedom regained, you drop into a combat stance and begin to wildly swing your weapon about in all directions,
                            attempting and succeeding in pushing back the humans surrounding you.
                            Their encircling formation broken, one of their comrades downed, and the other vomiting blood, the frightened men turn tail and run away.
                            By growing hysteria, fostered by your ballsy outburst, you have managed to drive the overwhelming odds to flee,
                            causing an entire group of intoxicated humans to sprint away from a single kobold with a weapon,
                            not even bothering to pick up their wounded in the process.
                        [/p]
                        [p]
                            As the dust settles, the only people left in the street are you, and the guy you stabbed in the butt.
                            Only one of you is standing, and it isn't him.
                            You take a few deep breaths, slowly calming down from the adrenaline rush that took over you.
                            You did it.
                            You scared them off.
                            You showed them that they can't just push kobolds around without any repercussions,
                            that you and your kind alike [b]will[/b] fight back if cornered.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.fightWin.continued() ? "disabled" : "";
        var isSelected = ctrl.fightWin.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.fightWin.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.fightWin.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.fightWin.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\crude\rude\fight\win\NightTwoCrudeWin.js

// scenes\Varanar\NightTwo\confront\crude\rude\fight\NightTwoCrudeFight.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeFight(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new NightTwoCrudeLose(ctrl);
    var toSuccess = new NightTwoCrudeWin(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        var weapon = ctrl.misc.weapon();
        return weapon && weapon.offence() >= 3 && weapon.stab();
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.rude.get() !== NightTwoCrudeRudeChoice.FIGHT) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.rude.get() !== NightTwoCrudeRudeChoice.FIGHT) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\crude\rude\fight\NightTwoCrudeFight.js

// scenes\Varanar\NightTwo\confront\crude\rude\struggle\NightTwoCrudeStruggleContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeStruggleContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You are absolutely [b]disgusted[/b].
                            Sickened, not only by what you were made to do, but by yourself, and by society as a whole.
                            [b]Nobody[/b] should be forced to do something so [b]foul[/b] and [b]revolting[/b] just to settle some stupid drunken quarrel.
                            You cry for what seems like an eternity.
                            The ground is as cold as the hearts of the humans that humiliated you.
                        [/p]
                        [p]
                            The sun slowly rises above the building-rooftops, and as you feel its first warming rays shine atop your back,
                            you pick yourself off the ground, along with the broken pieces of your shattered self-image.
                            While staying on the ground until you [b]die[/b] sounds like a good plan right now,
                            you decide it'd be best if you made your way back to the city square.
                            You won't be getting any sleep tonight.
                            At least you learned a valuable lesson: humans are [b]not[/b] nice people.
                            They [b]hate[/b] your kind, and you shouldn't trust them any further than you can throw them,
                            which isn't very far at all.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.crudeStruggle.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.crudeStruggle.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\crude\rude\struggle\NightTwoCrudeStruggleContinue.js

// scenes\Varanar\NightTwo\confront\crude\rude\struggle\NightTwoCrudeStruggle.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeStruggle(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoCrudeStruggleContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.nightTwo.forced = true;
        global.kobold.history.shared.varanar.nightTwo.buttKiss = true;
        var toReturn = `[p]
                            You wriggle around, attempting to escape the human's grasp,
                            but all he has to do is apply some extra [b]force[/b] to halt your squirming.
                            Your arm feels like it's about to [b]break[/b].
                            The pain almost makes you burst into tears, yet you manage to save face for just a little while longer.
                            Your mind [b]screams[/b] for you to beg him to let go, but you don't want to humiliate yourself any further.
                        [/p]
                        [p]
                            Your head gets gradually pushed lower and lower, closer to the hulking brute's bare behind.
                            "[color=bred]Hurry up back there, I haven't got all night.[/color]" he complains,
                            slightly bending over to make his ass just out further, its unsavory fragrance reaching your nose.
                            "[color=purple]You heard him, lizard.
                            Here you go.[/color]" the man holding your arm speaks.
                            Then, the slow, constant lowering is replaced by a single fierce [b]shove[/b],
                            your face forcibly pressed against the exposed delinquent's rear.
                        [/p]
                        [p]
                            Whether by accident, or just as planned, your muzzle ends up lodged directly in between the drunkard's buttcheeks,
                            warm, sweaty, hairy flesh enveloping your scales from all sides.
                            Your lips end up pressing directly against his [i]least sanitary[/i] part.
                            You make the mistake of gasping in surprise,
                            inadvertently taking in a mouthful and a noseful of the foul, concentrated stench of the human's butthole,
                            an intense feeling of [b]sickness[/b] building inside of you.
                        [/p]
                        [p]
                            The group goes wild.
                            Half of them burst out into laughter, the other half start to ridicule you,
                            hurling every insult they know your way in an attempt to make your life even more [b]miserable[/b] than it already is.
                            Your free arm gets grabbed, and held in place much like the other.
                            Your tail gets harshly yanked as a prank.
                            You don't know what brings more tears to your eyes:
                            the foul act you're forced to perform, or the oh-so-casual cruelty and complete disregard of your well-being.
                            You're nothing but an [b]animal[/b] to these people.
                        [/p]
                        [p]
                            You struggle with all of your might, yet the grip on your arms and on your head is unyielding.
                            You thought they just wanted you to give a peck on his buttcheek, not [b]this[/b]!
                            If you knew they wanted you to kiss his [b]asshole[/b], you'd have [b]fought them[/b] with a bit more [b]conviction[/b]!
                            But it's too late for that now.
                            You're completely trapped.
                            And as the humans so kindly remind you:
                            "[color=brown]You ain't gettin' out 'till you kiss it![/color]"
                        [/p]
                        [p]
                            Crying out loud, you swallow what little remains of your pride.
                            Puckering your lips, you plant a single, small kiss atop the half-naked human's hole,
                            and begin to [b]retch[/b] and [b]gag[/b] immediately afterwards.
                            "[color=bred]The lizard did it![/color]" the drunkard exclaims, and immediately you are yanked from his rear,
                            tossed to the ground like a used piece of toilet paper.
                            The moment your body hits the cold, hard dirt, you curl up into a ball and begin to cry your eyes out.
                        [/p]
                        [p]
                            Some of the humans taunt and humiliate you, but luckily,
                            most of them are more interested in the fact that the [i]bet[/i] between Jacoby and Jehova has been settled.
                            You tune out for a little while, not interested in what they, or anyone, for that matter has to say.
                            After a bit of rowdy discussion, the group sets off towards the brothel district, leaving you to wallow in self-pity.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.crudeStruggle.continued() ? "disabled" : "";
        var isSelected = ctrl.crudeStruggle.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.crudeStruggle.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.rude.get() !== NightTwoCrudeRudeChoice.STRUGGLE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.crudeStruggle.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.rude.get() !== NightTwoCrudeRudeChoice.STRUGGLE) {
            return;
        }

        next.registerListeners();

        if (ctrl.crudeStruggle.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\crude\rude\struggle\NightTwoCrudeStruggle.js

// scenes\Varanar\NightTwo\confront\crude\rude\NightTwoCrudeRude.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrudeRude(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var fightId = arguments.callee.name + "ButtonOne";
    var struggleId = arguments.callee.name + "ButtonTwo";

    var toFight = new NightTwoCrudeFight(ctrl);
    var toStruggle = new NightTwoCrudeStruggle(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            No matter what you say, they'll find a way to turn it against you.
                            Might as well not even [b]bother[/b] to play along.
                            Besides, if you show the group that you don't take kindly to them messing with you,
                            perhaps they'll leave you alone and find a different target to pick on.
                            Heck, given their [i]rough[/i] appearance, they might even [i]appreciate[/i] you not being a [b]total pushover[/b].
                        [/p]
                        [p]
                            Raising your voice, you loudly tell the man to [b]fuck off[/b].
                            You're not solving any disputes for a bunch of rude, inbred [b]humans[/b]!
                            Kobolds are ass kissers, huh?!
                            That's funny, coming from a race of [b]bootlickers[/b] that'd [b]suck their lord off[/b] for an extra acre of land,
                            or some food to make it through the winter!
                            If anything, [b]humans[/b] are the ass kissers, and you can't [b]wait[/b] for your kind to replace theirs as the dominant party around here!
                        [/p]
                        [p]
                            Wew.
                            You, uhh, didn't mean to go on a [i]rant[/i] like that.
                            You just wanted to throw some cuss words around to impress the drunkards, so they'd leave you alone.
                            A lifetime of [b]taking shit[/b] from their kind left you with a [b]lot[/b] of pent up frustration,
                            and you couldn't really stop yourself after you started venting some of that budding resentment.
                            Alas, the deed is done.
                            There's no going back now.
                            All you can do is [b]pray[/b] that the humans'll take your little [i]tirade[/i] as a [i]joke[/i].
                        [/p]
                        [p]
                            Sadly, that doesn't seem to be the case.
                            The group goes eerily quiet, angry eyes glaring down at you from all directions.
                            "[color=bred]I don't think you understand where you are.[/color]" the man that approached you breaks the silence,
                            his tone far more [b]hostile[/b] than before.
                            His hands travel down to his waist, and he deftly unbuckles the heavy belt keeping his pants up.
                            "[color=bred]This is a [b]human[/b] city, [b]lizard[/b].[/color]"
                        [/p]
                        [p]
                            From behind, one of the men roughly grabs your arm, cruelly and painfully twisting it.
                            You barely refrain from voicing your anguish, still attempting to keep up the [i]tough[/i] facade.
                            "[color=bred]And in Varanar, when [b]humans[/b] ask you a question, you [b]answer them[/b].[/color]"
                            Your head is pushed down, [b]forcing[/b] you to bend over forwards.
                            "[color=bred]But since you don't want to [b]talk[/b] to us...[/color]"
                            The hulking brute turns around as he speaks, lowering his trousers and his underwear simultaneously,
                            his sweat-ridden, hairy behind coming into view, buttcheeks mere inches away from your face.
                            "[color=bred]...you're just going to have to [b]show[/b] us how good you are at [b]kissing ass[/b].[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFightChoice() {
        var disabled = ctrl.rude.has() ? "disabled" : "";
        var isSelected = ctrl.rude.get() === NightTwoCrudeRudeChoice.FIGHT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fightId}" ${disabled}>Fight</button>`;

        var tooltip = global.parser.parse(`You're not going to go down without a fight. Come out swinging.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildStruggleChoice() {
        var disabled = ctrl.rude.has() ? "disabled" : "";
        var isSelected = ctrl.rude.get() === NightTwoCrudeRudeChoice.STRUGGLE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${struggleId}" ${disabled}>Struggle</button>`;

        var tooltip = global.parser.parse(`Struggle against the humans' grasp, but don't outright attack them.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildFightChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildStruggleChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + fightId), $("#" + struggleId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFightListener() {
        $("#" + fightId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.rude.set(NightTwoCrudeRudeChoice.FIGHT);
            loadNextPart(toFight);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerStruggleListener() {
        $("#" + struggleId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.rude.set(NightTwoCrudeRudeChoice.STRUGGLE);
            loadNextPart(toStruggle);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.lewdCrude.get() !== NightTwoCrudeChoice.RUDE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.rude.change();
        part += buildChoice();
        part += `</div>`;
        part += toFight.build();
        part += toStruggle.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.lewdCrude.get() !== NightTwoCrudeChoice.RUDE) {
            return;
        }

        toFight.registerListeners();
        toStruggle.registerListeners();

        if (ctrl.rude.has()) {
            return;
        }

        registerFightListener();
        registerStruggleListener();
    };
}

var NightTwoCrudeRudeChoice = {
    FIGHT: 1,
    STRUGGLE: 2
};

// END scenes\Varanar\NightTwo\confront\crude\rude\NightTwoCrudeRude.js

// scenes\Varanar\NightTwo\confront\crude\NightTwoCrude.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoCrude(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var appeaseId = arguments.callee.name + "ButtonOne";
    var rudeId = arguments.callee.name + "ButtonTwo";

    var toAppease = new NightTwoCrudeAppease(ctrl);
    var toRude = new NightTwoCrudeRude(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide [b]not[/b] to empower the group's belligerent behavior.
                            Instead of giving them attention, or even the [b]satisfaction[/b] of chasing after you,
                            you're going to walk right past the hoodlums without so much as [b]looking[/b] at them.
                            Tilting your head the other way,
                            staring at the front-faces of the buildings you pass as though you're inspecting them,
                            you begin to stride down the road.
                            As you approach, the group appears to simmer down a bit, your tactic of not paying them any mind seemingly effective.
                        [/p]
                        [p]
                            Looking away to the side, however, quickly proves to have been a mistake.
                            The moment you cross the delinquents' party, you suddenly [b]bump[/b] into something,
                            the unexpected collision almost making you fall back onto your butt.
                            Your gaze snapping forwards, you become aware that the [i]thing[/i] you walked into is none other than one of the men.
                            Taking a few steps backwards to regain your balance, you find your rear end pressing against something as well.
                            Another human.
                            It looks like, while you were [i]pretending[/i] not to notice them, the drunkards somehow managed to surround you!
                        [/p]
                        [p]
                            The man you bumped into last pushes you forwards, into the middle of the little [i]circle[/i] they've formed.
                            While they were quiet enough to sneak up on you, now that they have you surrounded, the group soon returns to its rowdy nature.
                            Cruel jokes and a myriad of insults are hurled your way.
                            "[color=purple]Who left the cage open?![/color]"
                            "[color=dred]It lost its collar![/color]"
                            "[color=brown]My wife cooked two of you for dinner last night![/color]"
                            Growing concerned, you desperately look around for a way out, yet they have you [b]completely[/b] encircled.
                            Any time you take so much as a step towards them, you get roughly [b]shoved[/b] back into the very centre of the group.
                        [/p]
                        [p]
                            One of the humans breaks formation, stepping towards you with a swaying gait that [b]screams[/b] drunkenness.
                            You fear for the worst, but instead of the [b]brutal pummeling[/b] you half-expected,
                            the ruffian comes to a halt a few feet in front of you.
                            It seems he is sober enough to slur out a few coherent words:
                            "[color=bred]Me and...[/color]" *hic* "[color=bred]...my buddy Jacoby back there was arguin' 'bout you [i]lizards[/i] earlier.[/color]"
                            The brute takes a lengthy pause to consider his next words, his inebriated brain clearly working overtime.
                            "[color=bred]He says your kind is good for [b]nothing[/b].
                            I disagree.
                            You kobolds are good at [b]kissing ass[/b], especially when it comes to [b]civilized people[/b] like [b]us[/b].
                            I think you kiss ass better than [b]any[/b] other species!
                            So, which one of us is right?[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildAppeaseChoice() {
        var disabled = ctrl.lewdCrude.has() ? "disabled" : "";
        var isSelected = ctrl.lewdCrude.get() === NightTwoCrudeChoice.APPEASE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${appeaseId}" ${disabled}>Appeasing</button>`;

        var tooltip = global.parser.parse(`Attempt to appease the man. If you play along, the group'll have no reason to hurt you, right?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRudeChoice() {
        var disabled = ctrl.lewdCrude.has() ? "disabled" : "";
        var isSelected = ctrl.lewdCrude.get() === NightTwoCrudeChoice.RUDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${rudeId}" ${disabled}>Rude</button>`;

        var tooltip = global.parser.parse(`Don't give them an inch. Refuse to play along. Give them a piece of your mind, and send them on their way.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You are: </div>`;
        choice += buildRudeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAppeaseChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + appeaseId), $("#" + rudeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerAppeaseListener() {
        $("#" + appeaseId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.lewdCrude.set(NightTwoCrudeChoice.APPEASE);
            loadNextPart(toAppease);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerRudeListener() {
        $("#" + rudeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.lewdCrude.set(NightTwoCrudeChoice.RUDE);
            loadNextPart(toRude);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.lewdCrude.change();
        part += buildChoice();
        part += `</div>`;
        part += toAppease.build();
        part += toRude.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        toAppease.registerListeners();
        toRude.registerListeners();

        if (ctrl.lewdCrude.has()) {
            return;
        }

        registerAppeaseListener();
        registerRudeListener();
    };
}

var NightTwoCrudeChoice = {
    RUDE: 1,
    APPEASE: 2
};

// END scenes\Varanar\NightTwo\confront\crude\NightTwoCrude.js

// scenes\Varanar\NightTwo\confront\lewd\appease\eager\NightTwoLewdEagerContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdEagerContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Lewd eager continue.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.lewdEager.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.lewdEager.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\appease\eager\NightTwoLewdEagerContinue.js

// scenes\Varanar\NightTwo\confront\lewd\appease\eager\NightTwoLewdEager.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdEager(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoLewdEagerContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Lewd eager.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.lewdEager.continued() ? "disabled" : "";
        var isSelected = ctrl.lewdEager.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.lewdEager.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.appease.get() !== NightTwoLewdAppeaseChoice.EAGER) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.lewdEager.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.appease.get() !== NightTwoLewdAppeaseChoice.EAGER) {
            return;
        }

        next.registerListeners();

        if (ctrl.lewdEager.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\appease\eager\NightTwoLewdEager.js

// scenes\Varanar\NightTwo\confront\lewd\appease\reluctant\NightTwoLewdReluctantContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdReluctantContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You rise to your feet, before spending a good half hour stuck in between wanting to vomit, and not being able to.
                            You should have wiped your tongue before retracting it,
                            but your mind was too occupied with figuring out that the human you [i]licked[/i] just couldn't get it up,
                            which meant the other ruffian won their little [i]bet[/i], or whatever.
                        [/p]
                        [p]
                            As you finally manage to compose yourself a little, the sun starts to rise above the rooftops.
                            You'd rather have spent the remainder of the night in bed, as opposed to on your knees and bent over trying to throw up, but alas.
                            You head back towards the town square, where your [i]urban adventure[/i] started.
                            Your first night in Varanar is over, and your first day is just about to begin.
                            If there's anything you've learned, it's that you need to keep your guard up at all times.
                            The next group of humans to corner you might not be limp-dicked drunkards.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.lewdReluctant.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.lewdReluctant.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\appease\reluctant\NightTwoLewdReluctantContinue.js

// scenes\Varanar\NightTwo\confront\lewd\appease\reluctant\NightTwoLewdReluctant.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdReluctant(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoLewdReluctantContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.nightTwo.flaccidOral = true;
        var toReturn = `[p]
                            You try your hardest to look for a way out of this mess you've created, but your brain draws a blank.
                            It seems as though you've talked yourself into a corner.
                            You can simply [i]refuse[/i] to play along, but judging by the group's intense behavior,
                            you can make a pretty decent guess that things [b]won't[/b] turn out nicely if you back out now.
                            The best course of action, for now, would be to just [i]play along[/i].
                            If you do ask they ask of you, this'll all be over soon...you hope.
                        [/p]
                        [p]
                            "[color=bred]Well? Don't keep me waiting.[/color]" the half-naked human's voice snaps you back to reality.
                            He rocks his hips forwards, his unsavory groin jutting out towards you.
                            You can feel your stomach churning.
                            It's just so [b]hairy[/b] and [b]disgusting[/b].
                            You'd rather suck off a [b]horse[/b] than him.
                            You [b]really[/b] don't want to do this, but the men have left you with little choice.
                        [/p]
                        [p]
                            You sigh loudly and take a deep breath of cool, late-night air.
                            Your lungs filled with clean air, you drop to your knees, and hold your breath.
                            Closing your eyes, you lean forwards, and plant a single tiny kiss on the human's flaccid length,
                            your lips pressing against his foreskin, his pubic hair brushing against your snout.
                            "[color=bred]That's it. Keep going.[/color]"
                        [/p]
                        [p]
                            While utterly revolted by the unwashed hairs rubbing against your scales,
                            you deftly plant a second kiss atop his shrunken shaft, followed by a small lick,
                            the [b]horrid[/b] taste of the human's member only now revealed to you.
                            You barely refrain from gagging, but your wincing face clearly reveals your aversion.
                            Biting through your own reluctance, you give the man a second, a third and a fourth lick without so much as retracting your tongue,
                            your tastebuds luckily growing a little numb, sparing you from the worst of the appalling taste of the filth clinging to your tongue.
                        [/p]
                        [p]
                            Despite your best efforts, the human's member remains completely flaccid,
                            growing not even a single inch since the start of your [i]reluctant admiration[/i].
                            You pull back slightly, and cast a quizzical gaze upwards at the human.
                            "[color=bred]I...I...[/color]" he stammers, clearly lost for words.
                            "[color=bred]It's the booze, I swear![/color]"
                            The entire group bursts out into laughter.
                        [/p]
                        [p]
                            "[color=green]Jehova can't get his dick up![/color]"
                            "[color=brown]I bet your wife could use a [b]real[/b] man![/color]"
                            "[color=purple]What's the matter? Kobold not so good after all?[/color]"
                            "[color=dred]Bet he'd get hard if it were a wolfess instead![/color]"
                            Embarrassed, the exposed man quickly pulls up his pants.
                            It seems the group's attention has shifted from you to him.
                        [/p]
                        [p]
                            They bicker amongst eachother for a while, before reaching a conclusion:
                            "[color=purple]Audric wins the bet. Drinks are on Jehovah![/color]"
                            "[color=brown]To the tavern![/color]"
                            And with that, the group of drunken humans rushes off towards the brothel district,
                            leaving you on your knees in the middle of the street with your tongue still lolling out,
                            somewhat wondering what the hell just happened.
                        [/p]
                        [p]
                            You absent-mindedly retract your tongue, the dry layer of grime on top of it mixing with your saliva,
                            which seemingly rejuvinates the filth's nauseating taste, while simultaneously spreading all across your mouth.
                            You heave, retch and gag.
                            It's like you're on the verge of vomiting, without being able to hurl and get it over with already.
                            At least the humans aren't around to witness this.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.lewdReluctant.continued() ? "disabled" : "";
        var isSelected = ctrl.lewdReluctant.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.lewdReluctant.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.appease.get() !== NightTwoLewdAppeaseChoice.RELUCTANT) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.lewdReluctant.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.appease.get() !== NightTwoLewdAppeaseChoice.RELUCTANT) {
            return;
        }

        next.registerListeners();

        if (ctrl.lewdReluctant.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\appease\reluctant\NightTwoLewdReluctant.js

// scenes\Varanar\NightTwo\confront\lewd\appease\unfit\NightTwoLewdUnfitContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdUnfitContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With the humans leaving, you decide that you, too, should probably get moving, lest they change their minds.
                            You begin to make your way back towards the city square, where your nighttime adventure began.
                            While you walk, you ponder the events that just transpired.
                            Were it not for your clever way of handling the group of drunkards, things could have gone [b]very[/b] bad for you.
                            You're glad you made it out unharmed, with your dignity mostly intact.
                        [/p]
                        [p]
                            As you near the marketplace, the sun slowly rises above the rooftops.
                            You guess you won't be getting any sleep tonight.
                            Oh well.
                            Taking a deep breath of fresh morning air, you steel yourself for your first day in Varanar.
                            You'll need to be on your guard if it's to be anything like your first [i]night[/i].
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.lewdUnfit.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.lewdUnfit.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\appease\unfit\NightTwoLewdUnfitContinue.js

// scenes\Varanar\NightTwo\confront\lewd\appease\unfit\NightTwoLewdUnfit.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdUnfit(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoLewdUnfitContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var gender = ctrl.misc.baseGender();
        var genderName = GenderEnum.properties[gender].name.toLowerCase();
        var toReturn = `[p]
                            You say you are [b]flattered[/b] that the men would even [b]consider[/b] letting such an [b]ugly[/b] kobold as yourself near their loins,
                            but surely they can find a [b]prettier[/b] one than you to quell their need, right?
                            You mean, you've got some well-defined muscles, and some very visible scars.
                            Unless they [i]like[/i] that sort of stuff, that is.
                            But wouldn't that make them attracted to [b]masculine[/b] features?
                            You don't know much about humans and what they like, but that sounds kind of [i]gay[/i] to you.
                        [/p]
                        [p]
                            The group bursts into disarray.
                            "[color=purple]The lizard's right![/color]"
                            "[color=dred]Audric's a queer![/color]"
                            "[color=green]Hey now, it was [b]Jehova[/b] that pulled down his pants![/color]"
                            They begin to argue about the merits of having sex with butch-looking ${genderName} kobolds.
                            After a bit of back and forth bickering, the human in front of you pulls up his pants, and buckles his belt once more.
                            "[color=bred]Can't even get my dick sucked without a fight these days.[/color]" he mumbles.
                        [/p]
                        [p]
                            While you aren't keen on interrupting the men while they're mid-[b]discussion[/b],
                            you quietly suggest that they should perhaps go visit a whorehouse,
                            where there are plenty of more [b]feminine[/b] people eager to give them [b]all[/b] blowjobs.
                            Heck, perhaps the whores can give further insights on their little [i]disagreement[/i].
                            They [i]are[/i] experts at having sex, after all.
                            Surely they must know plenty about whether or not it's gay to face-fuck a somewhat masculine kobold.
                        [/p]
                        [p]
                            "[color=brown]Right again.
                            You're pretty smart, for a lizard![/color]"
                            One of the most attentive humans commends.
                            You quietly whisper that he has no idea just [i]how[/i] smart you are.
                            Luckily, he doesn't hear you.
                            "[color=green]Let's go settle this at the whorehouse, then![/color]"
                            "[color=purple]Yeah, let's go prove Audric's gay![/color]"
                            "[color=green]You wish, Cedric![/color]"
                            "[color=brown][b]To the brothel![/b][/color]"
                            "[color=dred]God damnit.[/color]"
                        [/p]
                        [p]
                            With that, the drunkards break their encirclement.
                            Like they've completely forgotten you even exist, they head for the brothel district, leaving you all alone in the middle of the street.
                            Just the way you like it.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.lewdUnfit.continued() ? "disabled" : "";
        var isSelected = ctrl.lewdUnfit.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.lewdUnfit.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.appease.get() !== NightTwoLewdAppeaseChoice.UNFIT) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.lewdUnfit.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.appease.get() !== NightTwoLewdAppeaseChoice.UNFIT) {
            return;
        }

        next.registerListeners();

        if (ctrl.lewdUnfit.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\appease\unfit\NightTwoLewdUnfit.js

// scenes\Varanar\NightTwo\confront\lewd\appease\NightTwoLewdAppease.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdAppease(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reluctantId = arguments.callee.name + "ButtonOne";
    var eagerId = arguments.callee.name + "ButtonTwo";
    var unfitId = arguments.callee.name + "ButtonThree";

    var toReluctant = new NightTwoLewdReluctant(ctrl);
    var toEager = new NightTwoLewdEager(ctrl);
    var toUnfit = new NightTwoLewdUnfit(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            No matter what you say, they'll find a way to turn it against you,
                            but if you don't play along, you doubt things would end well.
                            If you just give them the laugh they're after, you're certain they'll leave you alone.
                            Besides, if you show the group that you can take a [i]joke[/i] just as well as any human can,
                            perhaps they'll get bored and find a different target to pick on.
                        [/p]
                        [p]
                            Forsaking your pride, you half-heartedly stammer out that yes, kobolds are [i]indeed[/i] good at [b]sucking human cock[/b].
                            You can barely even get those final three words past your lips, your mind screaming at you [b]not[/b] to put yourself down like this.
                            But, in a way, the statement isn't even [i]wrong[/i].
                            While most kobolds are [i]scoundrels[/i] of some form or another, the ones that aren't tend to be [b]prostitutes[/b],
                            relying on innate [b]tightness[/b] due to size difference, and a knack for shoving things down their throats without gagging or retching.
                            In that sense: yes. Your kind [i]is[/i] good at [b]giving blowjobs[/b].
                        [/p]
                        [p]
                            The entire group bursts out into laughter at the sight of a kobold humiliating itself.
                            The drunken men begin to taunt you anew, although the mood is slightly more cheerful now than it was before,
                            their insults clearly less meant to [b]hurt[/b] you this time around.
                            It seems as though your gamble has paid off.
                        [/p]
                        [p]
                            One of the men still seems displeased.
                            As the posse begins to quiet down, he raises his voice:
                            "[color=green]Kobolds are [b]liars[/b].
                            Who knows if this one is telling the truth?!
                            Make'em [b]prove[/b] it![/color]"
                            At first, the group pays the dissenter no mind.
                            Slowly, but steadily, however, cogs begin to turn inside of their drunken heads.
                            As they come to a communal conclusion, some of the men begin to voice their response in unison:
                            "[color=purple]Yeah. Make'em prove it![/color]"
                            "[color=brown]The kobold lies![/color]"
                            "[color=dred]Put on a show![/color]"
                        [/p]
                        [p]
                            The man in front of you realizes he's losing the crowd's favour.
                            A grin forms atop his visage.
                            "[color=bred]Audric's right.[/color]" he speaks, his hands travelling down to his waist, resting atop his belt buckle.
                            "[color=bred]You lizards are a bunch of liars.[/color]"
                            He unfastens the clasp sealing the leather cord, his pants visible loosening.
                            "[color=bred]And we don't take kindly to liars around these parts.[/color]"
                            That being said, the drunkard promptly lowers his pants and his underwear,
                            his flaccid, uncut member surrounded with a hefty amount of pubic hair
                            "[color=bred]Let's see if you were telling the truth or not.[/color]"
                        [/p]
                        [p]
                            The entire group seems riled up.
                            They [b]really[/b] want this to happen, and judging by the exposed human's tone,
                            you're in too deep to back out now.
                            Unless you can present a strong case for yourself, you fear this [b]is[/b] happening.
                            That doesn't mean you need to like it, though.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildReluctantChoice() {
        var disabled = ctrl.appease.has() ? "disabled" : "";
        var isSelected = ctrl.appease.get() === NightTwoLewdAppeaseChoice.RELUCTANT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${reluctantId}" ${disabled}>Reluctant</button>`;

        var tooltip = global.parser.parse(`You don't really want to participate in these guys' debauchery, but you don't have much of a choice.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 185px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildEagerChoice() {
        var disabled = ctrl.appease.has() ? "disabled" : "";
        var isSelected = ctrl.appease.get() === NightTwoLewdAppeaseChoice.EAGER;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${eagerId}" ${disabled}>Eager</button>`;

        var tooltip = global.parser.parse(`Give the men a show they'll never forget.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildUnfitChoice() {
        var test = ctrl.misc.strength() > 5 && ctrl.misc.endurance() > 5;
        var disabled = (ctrl.appease.has() || !test) ? "disabled" : "";
        var isSelected = ctrl.appease.get() === NightTwoLewdAppeaseChoice.UNFIT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${unfitId}" ${disabled}>Ugly</button>`;

        var tooltip = global.parser.parse(`Surely these men can find a prettier kobold to abuse, right?`);
        if (!test) {
            tooltip = `Requires 6 strength and 6 endurance.`;
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You are: </div>`;
        choice += buildReluctantChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildUnfitChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + reluctantId), $("#" + eagerId), $("#" + unfitId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerReluctantListener() {
        $("#" + reluctantId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.appease.set(NightTwoLewdAppeaseChoice.RELUCTANT);
            loadNextPart(toReluctant);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerEagerListener() {
        $("#" + eagerId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.appease.set(NightTwoLewdAppeaseChoice.EAGER);
            loadNextPart(toEager);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerUnfitListener() {
        $("#" + unfitId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.appease.set(NightTwoLewdAppeaseChoice.UNFIT);
            loadNextPart(toUnfit);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.lewdCrude.get() !== NightTwoCrudeChoice.APPEASE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toReluctant.build();
        part += toEager.build();
        part += toUnfit.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.lewdCrude.get() !== NightTwoLewdChoice.APPEASE) {
            return;
        }

        toReluctant.registerListeners();
        toEager.registerListeners();
        toUnfit.registerListeners();

        if (ctrl.appease.has()) {
            return;
        }

        registerReluctantListener();
        registerEagerListener();
        registerUnfitListener();
    };
}

var NightTwoLewdAppeaseChoice = {
    RELUCTANT: 1,
    EAGER: 2,
    UNFIT: 3
};

// END scenes\Varanar\NightTwo\confront\lewd\appease\NightTwoLewdAppease.js

// scenes\Varanar\NightTwo\confront\lewd\rude\fight\lose\NightTwoLewdLose.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdLose(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    var idsToRemove = [
        "NightTwoLewdLose",
        "NightTwoLewdFight",
        "NightTwoLewdRude",
        "NightTwoLewd",
        "NightTwoConfront",
        "NightTwoRunIn",
        "NightTwoStage"
    ];

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            No.
                            You aren't going to let them do this.
                            You can't and you [b]won't[/b] get [b]anywhere[/b] near that filthy human's disgusting groin, let alone taking his [i]thing[/i] into your mouth!
                            [b]Over your dead body.[/b]
                            They may be stronger than you, they may outnumber you, they may have you in an arm-grip,
                            but all of that just means they [b]won't[/b] be [i]expecting[/i] you to go on the offensive.
                        [/p]
                        [p]
                            They think they have you scared shitless, that you're just another kobold for them to [b]bully[/b] and [b]abuse[/b],
                            that you won't put up a fight simply because this is [b]their[/b] town.
                            But you aren't any ordinary run-of-the-mill lizard.
                            The other members of your kind may be content being thieves, or bandits or prostitutes or beggars, but not [b]you[/b].
                            You may be a thief right now, but you [b]aspire[/b] to be a [b]knight[/b].
                            And knights don't back down in the face of adversity.
                            They charge head-forward into danger.
                            Just like you're about to do.
                        [/p]`;

        if (weapon) {
            toReturn += `[p]
                            While one of your arms is being restrained, you reach for your ${global.kobold.inventory.createItemWithTooltip(weapon)} with the other.
                            With your head being forced ever-closer to the drunkard's dick, you decide to strike before your snout is pushed into his unwashed pubic hair.
                            Swiftly and silently, you [b]surge[/b] your weapon forwards, intent on striking the ruffian's exposed groin,
                            yet before your strike manages to connect, your free arm is grabbed and painfully twisted, much like the other.
                            With a pained yelp, you drop your ${global.kobold.inventory.createItemWithTooltip(weapon)} to the ground,
                            where it is quickly picked up by one of the humans, who promptly [b]bashes[/b] you on the head with it.
                        [/p]`;
        } else {
            toReturn += `[p]
                            While one of your arms is being restrained, you lash out with the other, intent on striking the oaf's exposed groin,
                            yet before your strike manages to connect, your free arm is grabbed and painfully twisted, much like the other.
                            You let out a pained yelp.
                            Your attempted aggression is soon punished, one of the humans brutally [b]bashing[/b] the top of your head with his balled up hand.
                        [/p]`;
        }

        toReturn += `   ${global.ui.condFail(weapon ? `${global.kobold.inventory.createItemWithTooltip(weapon, true)} offence below 4` : `No weapon`)}
                        [p]
                            Your vision fades to black.
                            You lose consciousness.
                            You dream about a ceremony.
                            One in your honor.
                            You are in a castle.
                            There's princes and princesses, barons and dukes, royals from all corners of the world.
                            You are kneeling.
                            The Dragon King himself puts a gilded blade atop your shoulder.
                            Then suddenly, the dream turns into a nightmare, the castle replaced by a terrible, all-consuming void that sucks you into its core.
                            Once the void has consumed all, you are left with nothing, not even yourself.
                            The nightmare ends, and so do you.
                            In the real world, the men kicked you to death for passing out instead of doing their bidding.
                            Luckily, you were not awake to witness the horrible pain inflicted upon you.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    Nightmare ending.
                                [/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" style="margin-left: auto; position: relative; float: right;" id="${continueId}">`;
        choice += `&lt;&lt Revert`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400, function () {
            var main = $("#main");
            main.scrollTop(main[0].scrollHeight - main.height() - 5); // Fix scrolling
        });
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.misc.unsetAll();
            ctrl.rude.undo();
            ctrl.rude.unset();
            ctrl.lewdCrude.undo();
            ctrl.lewdCrude.unset();
            ctrl.runin.unset();
            ctrl.stage.unset();
            for (var i = 0; i < idsToRemove.length; i++) {
                $("#" + idsToRemove[i]).remove();
            }
            loadNextPart(new NightTwoStage(ctrl));
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        registerContinueListener();
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\rude\fight\lose\NightTwoLewdLose.js

// scenes\Varanar\NightTwo\confront\lewd\rude\fight\win\NightTwoLewdWinContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdWinContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            As your adrenaline rush fades, the harsh reality of things comes crashing down on you.
                            You just [b]attacked[/b] and seriously [b]hurt[/b] a bunch of humans.
                            They might have been bothering you, but the guards won't care.
                            To them, you're a stupid, unruly lizard that crossed the line.
                            The men you assaulted had rights.
                            You don't.
                        [/p]
                        [p]
                            You make the smart decision to vacate the scene of the crime.
                            After jogging a few streets over, you lower your pace to a brisk walk.
                            You think you're in the clear.
                            If the famous guards of Varanar were called, you wouldn't have made it this far.
                            You'd have been captured or killed before you even left the street.
                            Still, you probably shouldn't linger in this particular area for too long.
                        [/p]
                        [p]
                            As you begin to make your way back to the town square, where your late-night adventure began, the sun slowly begins to rise above the rooftops.
                            It seems as though you've managed to retain your dignity throughout the night, but at what cost?
                            You don't want to think about it.
                            It's too late, or rather [i]too early[/i] to go to bed.
                            Who needs sleep?
                            You'd just have bad dreams about the things you did anyway.
                            You know you acted in self-defence, but your conscience doesn't agree. 
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.fightWin.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.fightWin.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\rude\fight\win\NightTwoLewdWinContinue.js

// scenes\Varanar\NightTwo\confront\lewd\rude\fight\win\NightTwoLewdWin.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdWin(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoLewdWinContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        global.kobold.history.shared.varanar.nightTwo.fought = true;
        var toReturn = `[p]
                            No.
                            You aren't going to let them do this.
                            You can't and you [b]won't[/b] get [b]anywhere[/b] near that filthy human's disgusting groin, let alone put his [i]thing[/i] in your mouth!
                            [b]Over your dead body.[/b]
                            They may be stronger than you, they may outnumber you, they may have you in an arm-grip,
                            but all of that just means they [b]won't[/b] be [i]expecting[/i] you to go on the offensive.
                        [/p]
                        [p]
                            They think they have you scared shitless, that you're just another kobold for them to [b]bully[/b] and [b]abuse[/b],
                            that you won't put up a fight simply because this is [b]their[/b] town.
                            But you aren't any ordinary run-of-the-mill lizard.
                            The other members of your kind may be content being thieves, or bandits or prostitutes or beggars, but not [b]you[/b].
                            You may be a thief right now, but you [b]aspire[/b] to be a [b]knight[/b].
                            And knights don't back down in the face of adversity.
                            They charge head-forward into danger.
                            Just like you're about to do.
                        [/p]
                        [p]
                            While one of your arms is being restrained, you reach for your ${global.kobold.inventory.createItemWithTooltip(weapon)} with the other.
                            With your head being forced ever-closer to the drunkard's dick, you decide to strike before your snout is pushed into his unwashed pubic hair.
                            Swiftly and silently, you [b]surge[/b] your weapon forwards,
                            shoving the pointy-end directly into the ruffian's exposed groin, stabbing him in his flaccid length.
                            The man screams in pain.
                            His companions laugh out loud, thinking you bit him.
                            Their laughter comes to a halt, however, as the group slowly realizes what has transpired.
                        [/p]
                        [p]
                            You pull your weapon from the pantless delinquent's loins, and in one swift move, you bring it to bear on the person holding your arm.
                            Blindly stabbing backwards, you land a seemingly solid strike on the human's abdomen.
                            He loses his hold on your arm nigh-instantly as your ${global.kobold.inventory.createItemWithTooltip(weapon)} sinks into his gut.
                            The man whose cock you perforated tumbles backwards to the ground.
                            Blood oozes liberally from his mutilated groin.
                            The guy behind you stumbles backwards, clutching his wounded belly.
                            Communal panic settles in.
                            [i]Perfect[/i].
                        [/p]
                        ${global.ui.condSucc(`${global.kobold.inventory.createItemWithTooltip(weapon, true)} offence greater than 3`)}
                        [p]
                            Your freedom regained, you drop into a combat stance and begin to wildly swing your weapon about in all directions,
                            attempting and succeeding in pushing back the humans surrounding you.
                            Their encircling formation broken, one of their comrades downed, and the other vomiting blood, the frightened men turn tail and run away.
                            By growing hysteria, fostered by your ballsy outburst, you have managed to drive the overwhelming odds to flee,
                            causing an entire group of intoxicated humans to sprint away from a single kobold with a weapon,
                            not even bothering to pick up their wounded in the process.
                        [/p]
                        [p]
                            As the dust settles, the only people left in the street are you, and the guy you stabbed in the dick.
                            Only one of you is standing, and it isn't him.
                            You take a few deep breaths, slowly calming down from the adrenaline rush that took over you.
                            You did it.
                            You scared them off.
                            You showed them that they can't just push kobolds around without any repercussions,
                            that you and your kind alike [b]will[/b] fight back if cornered.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.fightWin.continued() ? "disabled" : "";
        var isSelected = ctrl.fightWin.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.fightWin.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.fightWin.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.fightWin.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\rude\fight\win\NightTwoLewdWin.js

// scenes\Varanar\NightTwo\confront\lewd\rude\fight\NightTwoLewdFight.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdFight(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new NightTwoLewdLose(ctrl);
    var toSuccess = new NightTwoLewdWin(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        var weapon = ctrl.misc.weapon();
        return weapon && weapon.offence() >= 3 && weapon.stab();
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.rude.get() !== NightTwoLewdRudeChoice.FIGHT) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.rude.get() !== NightTwoLewdRudeChoice.FIGHT) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\rude\fight\NightTwoLewdFight.js

// scenes\Varanar\NightTwo\confront\lewd\rude\struggle\NightTwoLewdStruggleContinue.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdStruggleContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You are absolutely [b]disgusted[/b].
                            Sickened, not only by what you were made to do, but by yourself, and by society as a whole.
                            [b]Nobody[/b] should be forced to do something so [b]foul[/b] and [b]revolting[/b] just to settle some stupid drunken quarrel.
                            You cry for what seems like an eternity, stuck in between wanting to throw up, and not being able to.
                            The ground is as cold as the hearts of the humans that humiliated you.
                        [/p]
                        [p]
                            The sun slowly rises above the building-rooftops, and as you feel its first warming rays shine atop your back,
                            you pick yourself off the ground, along with the broken pieces of your shattered self-image.
                            While staying on the ground until you [b]die[/b] sounds like a good plan right now,
                            you decide it'd be best if you made your way back to the city square.
                            You won't be getting any sleep tonight.
                            At least you learned a valuable lesson: humans are [b]not[/b] nice people.
                            They [b]hate[/b] your kind, and you shouldn't trust them any further than you can throw them,
                            which isn't very far at all.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.lewdStruggle.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.lewdStruggle.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\rude\struggle\NightTwoLewdStruggleContinue.js

// scenes\Varanar\NightTwo\confront\lewd\rude\struggle\NightTwoLewdStruggle.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdStruggle(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoLewdStruggleContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.nightTwo.forced = true;
        global.kobold.history.shared.varanar.nightTwo.flaccidOral = true;
        var toReturn = `[p]
                            You wriggle around, attempting to escape the human's grasp,
                            but all he has to do is apply some extra [b]force[/b] to halt your squirming.
                            Your arm feels like it's about to [b]break[/b].
                            The pain almost makes you burst into tears, yet you manage to save face for just a little while longer.
                            Your mind [b]screams[/b] for you to beg him to let go, but you don't want to humiliate yourself any further.
                        [/p]
                        [p]
                            Your head gets gradually pushed lower and lower, closer to the hulking brute's bare loins.
                            "[color=bred]Hurry up, I haven't got all night.[/color]" he complains,
                            slightly rocking his hips forwards to jut his groin out further, its unsavory fragrance reaching your nose.
                            "[color=purple]You heard him, lizard.
                            Here you go.[/color]" the man holding your arm speaks.
                            Then, the slow, constant lowering is replaced by a single fierce [b]shove[/b],
                            your face forcibly pressed against the exposed delinquent's hefty sack, his member rubbing against your snout.
                        [/p]
                        [p]
                            Your lips end up pressing directly against the drunkard's balls.
                            You make the mistake of gasping in surprise,
                            inadvertently taking in a mouthful and a noseful of the foul, concentrated stench of the man's unwashed malehood,
                            an intense feeling of [b]sickness[/b] building inside of you.
                            The thought of [b]biting down[/b], of [b]tearing[/b] apart his most [b]sensitive[/b] bodypart crosses your mind,
                            but the rest of the louts would undoubtedly [b]kill you[/b] for such an act of defiance.
                            You really don't want to die with human scrotum stuck to your teeth.
                        [/p]
                        [p]
                            The group goes wild.
                            Half of them burst out into laughter, the other half start to ridicule you,
                            hurling every insult they know your way in an attempt to make your life even more [b]miserable[/b] than it already is.
                            Your free arm gets grabbed, and held in place much like the other.
                            Your tail gets harshly yanked as a prank.
                            You don't know what brings more tears to your eyes:
                            the foul act you're forced to perform, or the oh-so-casual cruelty and complete disregard of your well-being.
                            You're nothing but an [b]animal[/b] to these people.
                        [/p]
                        [p]
                            You struggle with all of your might, yet the grip on your arms and on your head is unyielding.
                            You thought they just wanted you to play with his dick a little, not lap at his filthy nuts!
                            If you knew they wanted you to do [b]this[/b], you'd have [b]fought them[/b] with a bit more [b]conviction[/b]!
                            But it's too late for that now.
                            You're completely trapped.
                            "[color=brown]Get to work or we'll [b]bash your brains in[/b] and find a different lizard that [b]will[/b] cooperate![/color]"
                        [/p]
                        [p]
                            Crying out loud, you swallow what little remains of your pride.
                            Puckering your lips, you plant a single, small kiss atop the half-naked human's nuts.
                            "[color=bred]That's it. Keep going.[/color]"
                            Despite every fiber of your being begging you not to give in, your self-preservation instinct takes control.
                            You give the man's balls another kiss, and then a small lick, your tongue collecting more than a little ballsack-grime.
                            The taste is [b]abhorrent[/b].
                            Without retracting the flexible appendage, you give the drunkard's nuts a second, and then a third lick,
                            your tastebuds luckily going numb, sparing you from the worst of the flavour.
                        [/p]
                        [p]
                            Slowly, but steadily, the flow of insults directed at [b]you[/b] comes to an end.
                            Instead, the ruffians seem to be picking on the man whose balls are in your face.
                            "[color=green]Jehova can't get his dick up![/color]" one of the group's members comments.
                            "[color=bred]What're you lookin' at my dick for, Audric?![/color]" comes the angry reply.
                            "[color=brown]He's right! It's as soft as a virgin dragoness' ass![/color]"
                            "[color=purple]What's the matter? Kobold not so good after all?[/color]"
                            "[color=dred]Bet he'd get hard if it were a wolfess instead![/color]"
                        [/p]
                        [p]
                            "[color=bred]It's the booze, alright?![/color]"
                            Embarrassed, the exposed man pushes you to the ground, before quickly pulling up his pants.
                            It seems the group's attention has shifted from you to him.
                            The moment your body hits the cold, hard dirt, you curl up into a ball and begin to cry your eyes out.
                            You accidentally retract your lolling tongue, scrotum-filth spreading throughout your mouth,
                            your saliva seemingly reinvigorating its horrid taste.
                        [/p]
                        [p]
                            Some of the humans taunt and humiliate you, but luckily,
                            most of them are more interested in the fact that the [i]bet[/i] between Audric and Jehova has been settled.
                            You tune out for a little while, not interested in what they, or anyone, for that matter has to say.
                            After a bit of rowdy discussion, the group sets off towards the brothel district, leaving you to wallow in self-pity,
                            while retching from the taste of unwashed human balls.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.lewdStruggle.continued() ? "disabled" : "";
        var isSelected = ctrl.lewdStruggle.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.lewdStruggle.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.rude.get() !== NightTwoLewdRudeChoice.STRUGGLE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.lewdStruggle.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.rude.get() !== NightTwoLewdRudeChoice.STRUGGLE) {
            return;
        }

        next.registerListeners();

        if (ctrl.lewdStruggle.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\confront\lewd\rude\struggle\NightTwoLewdStruggle.js

// scenes\Varanar\NightTwo\confront\lewd\rude\NightTwoLewdRude.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewdRude(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var fightId = arguments.callee.name + "ButtonOne";
    var struggleId = arguments.callee.name + "ButtonTwo";

    var toFight = new NightTwoLewdFight(ctrl);
    var toStruggle = new NightTwoLewdStruggle(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            No matter what you say, they'll find a way to turn it against you.
                            Might as well not even [b]bother[/b] to play along.
                            Besides, if you show the group that you don't take kindly to them messing with you,
                            perhaps they'll leave you alone and find a different target to pick on.
                            Heck, given their [i]rough[/i] appearance, they might even [i]appreciate[/i] you not being a [b]total pushover[/b].
                        [/p]
                        [p]
                            Raising your voice, you loudly tell the man to [b]fuck off[/b].
                            You're not solving any disputes for a bunch of rude, inbred [b]humans[/b]!
                            Kobolds suck dick, huh?!
                            That's funny, coming from a race of [b]bootlickers[/b] that'd [b]jerk their lord off[/b] for an extra acre of land,
                            or some food to make it through the winter!
                            If anything, [b]humans[/b] are the cocksuckers,
                            and you can't [b]wait[/b] for your kind to replace theirs as the dominant party around these parts!
                        [/p]
                        [p]
                            Wew.
                            You, uhh, didn't mean to go on a [i]rant[/i] like that.
                            You just wanted to throw some cuss words around to impress the drunkards, so they'd leave you alone.
                            A lifetime of [b]taking shit[/b] from their kind left you with a [b]lot[/b] of pent up frustration,
                            and you couldn't really stop yourself after you started venting some of that budding resentment.
                            Alas, the deed is done.
                            There's no going back now.
                            All you can do is [b]pray[/b] that the humans'll take your little [i]tirade[/i] as a [i]joke[/i].
                        [/p]
                        [p]
                            Sadly, that doesn't seem to be the case.
                            The group goes eerily quiet, angry eyes glaring down at you from all directions.
                            "[color=bred]I don't think you understand where you are.[/color]" the man that approached you breaks the silence,
                            his tone far more [b]hostile[/b] than before.
                            His hands travel down to his waist, and he deftly unbuckles the heavy belt keeping his pants up.
                            "[color=bred]This is a [b]human[/b] city, [b]lizard[/b].[/color]"
                        [/p]
                        [p]
                            From behind, one of the men roughly grabs your arm, cruelly and painfully twisting it.
                            You barely refrain from voicing your anguish, still attempting to keep up the [i]tough[/i] facade.
                            "[color=bred]And in Varanar, when [b]humans[/b] ask you a question, you [b]answer them[/b].[/color]"
                            Your head is pushed down, [b]forcing[/b] you to bend over forwards.
                            "[color=bred]But since you don't want to [b]talk[/b] to us...[/color]"
                            The hulking brute lowers his trousers and his underwear as he speaks,
                            his pubic hair-engulfed, flaccid, uncut member mere inches away from your face.
                            "[color=bred]...you're just going to have to [b]show[/b] us how good you are at [b]sucking dick[/b].[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFightChoice() {
        var disabled = ctrl.rude.has() ? "disabled" : "";
        var isSelected = ctrl.rude.get() === NightTwoLewdRudeChoice.FIGHT;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fightId}" ${disabled}>Fight</button>`;

        var tooltip = global.parser.parse(`You're not going to go down without a fight. Come out swinging.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildStruggleChoice() {
        var disabled = ctrl.rude.has() ? "disabled" : "";
        var isSelected = ctrl.rude.get() === NightTwoLewdRudeChoice.STRUGGLE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${struggleId}" ${disabled}>Struggle</button>`;

        var tooltip = global.parser.parse(`Struggle against the humans' grasp, but don't outright attack them.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildFightChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildStruggleChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + fightId), $("#" + struggleId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFightListener() {
        $("#" + fightId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.rude.set(NightTwoLewdRudeChoice.FIGHT);
            loadNextPart(toFight);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerStruggleListener() {
        $("#" + struggleId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.rude.set(NightTwoLewdRudeChoice.STRUGGLE);
            loadNextPart(toStruggle);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.lewdCrude.get() !== NightTwoLewdChoice.RUDE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.rude.change();
        part += buildChoice();
        part += `</div>`;
        part += toFight.build();
        part += toStruggle.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.lewdCrude.get() !== NightTwoLewdChoice.RUDE) {
            return;
        }

        toFight.registerListeners();
        toStruggle.registerListeners();

        if (ctrl.rude.has()) {
            return;
        }

        registerFightListener();
        registerStruggleListener();
    };
}

var NightTwoLewdRudeChoice = {
    FIGHT: 1,
    STRUGGLE: 2
};

// END scenes\Varanar\NightTwo\confront\lewd\rude\NightTwoLewdRude.js

// scenes\Varanar\NightTwo\confront\lewd\NightTwoLewd.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoLewd(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var appeaseId = arguments.callee.name + "ButtonOne";
    var rudeId = arguments.callee.name + "ButtonTwo";

    var toAppease = new NightTwoLewdAppease(ctrl);
    var toRude = new NightTwoLewdRude(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var cd = ctrl.misc.crossdressing();
        var toReturn = `[p]
                            You decide [b]not[/b] to empower the group's belligerent behavior.
                            Instead of giving them attention, or even the [b]satisfaction[/b] of chasing after you,
                            you're going to walk right past the hoodlums without so much as [b]looking[/b] at them.
                            Tilting your head the other way,
                            staring at the front-faces of the buildings you pass as though you're inspecting them,
                            you begin to stride down the road.
                            As you approach, the group appears to simmer down a bit, your tactic of not paying them any mind seemingly effective.
                        [/p]
                        [p]
                            Looking away to the side, however, quickly proves to have been a mistake.
                            The moment you cross the delinquents' party, you suddenly [b]bump[/b] into something,
                            the unexpected collision almost making you fall back onto your butt.
                            Your gaze snapping forwards, you become aware that the [i]thing[/i] you walked into is none other than one of the men.
                            Taking a few steps backwards to regain your balance, you find your rear end pressing against something as well.
                            Another human.
                            It looks like, while you were [i]pretending[/i] not to notice them, the drunkards somehow managed to surround you!
                        [/p]
                        [p]
                            The man you bumped into last pushes you forwards, into the middle of the little [i]circle[/i] they've formed.
                            While they were quiet enough to sneak up on you, now that they have you surrounded, the group soon returns to its rowdy nature.
                            Cruel jokes and a myriad of lewd insults are hurled your way.
                            "[color=purple]${cd && male ? "It's dressed like a girl!" : "Nude whore!"}[/color]"
                            "[color=dred]What's your rate? Two coins?![/color]"
                            "[color=brown]Tighter than my wife, I bet![/color]"
                            Growing concerned, you desperately look around for a way out, yet they have you [b]completely[/b] encircled.
                            Any time you take so much as a step towards them, you get roughly [b]shoved[/b] back into the very centre of the group.
                        [/p]
                        [p]
                            One of the humans breaks formation, stepping towards you with a swaying gait that [b]screams[/b] drunkenness.
                            You fear for the worst, but instead of the [b]brutal pummeling[/b] you half-expected,
                            the ruffian comes to a halt a few feet in front of you.
                            It seems he is sober enough to slur out a few coherent words:
                            "[color=bred]Me and...[/color]" *hic* "[color=bred]...my buddy Audric back there was arguin' 'bout you [i]lizards[/i] earlier.[/color]"
                            The brute takes a lengthy pause to consider his next words, his inebriated brain clearly working overtime.
                            "[color=bred]He says your kind is good for [b]nothing[/b].
                            I disagree.
                            You kobolds are good at [b]suckin' dick[/b], especially when it comes to [b]bigger species[/b] like [b]us[/b].
                            I think you give better blowjobs than [b]any[/b] other race!
                            So, which one of us is right?[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildAppeaseChoice() {
        var disabled = ctrl.lewdCrude.has() ? "disabled" : "";
        var isSelected = ctrl.lewdCrude.get() === NightTwoLewdChoice.APPEASE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${appeaseId}" ${disabled}>Appeasing</button>`;

        var tooltip = global.parser.parse(`Attempt to appease the man. If you play along, the group'll have no reason to hurt you, right?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 230px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRudeChoice() {
        var disabled = ctrl.lewdCrude.has() ? "disabled" : "";
        var isSelected = ctrl.lewdCrude.get() === NightTwoLewdChoice.RUDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${rudeId}" ${disabled}>Rude</button>`;

        var tooltip = global.parser.parse(`Don't give them an inch. Refuse to play along. Give them a piece of your mind, and send them on their way.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You are: </div>`;
        choice += buildRudeChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAppeaseChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + appeaseId), $("#" + rudeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerAppeaseListener() {
        $("#" + appeaseId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.lewdCrude.set(NightTwoLewdChoice.APPEASE);
            loadNextPart(toAppease);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerRudeListener() {
        $("#" + rudeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.lewdCrude.set(NightTwoLewdChoice.RUDE);
            loadNextPart(toRude);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.lewdCrude.change();
        part += buildChoice();
        part += `</div>`;
        part += toAppease.build();
        part += toRude.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        toAppease.registerListeners();
        toRude.registerListeners();

        if (ctrl.lewdCrude.has()) {
            return;
        }

        registerAppeaseListener();
        registerRudeListener();
    };
}

var NightTwoLewdChoice = {
    RUDE: 1,
    APPEASE: 2
};

// END scenes\Varanar\NightTwo\confront\lewd\NightTwoLewd.js

// scenes\Varanar\NightTwo\confront\NightTwoConfront.js

/**
 * Confrontation section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoConfront(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new NightTwoLewd(ctrl);
    var toSuccess = new NightTwoCrude(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        if (ctrl.misc.baseGender() === GenderEnum.MALE) {
            return !ctrl.misc.crossdressing() && !ctrl.misc.groinExposed();
        } else {
            return !ctrl.misc.groinExposed();
        }
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.runin.get() !== NightTwoRunInChoice.IGNORE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.runin.get() !== NightTwoRunInChoice.IGNORE) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\NightTwo\confront\NightTwoConfront.js

// scenes\Varanar\NightTwo\flee\alley\hide\attack\NightTwoHideAttackContinue.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoHideAttackContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            After removing yourself from the [i]brothel district[/i] proper,
                            you start walking back towards the town square,
                            where your [i]urban adventure[/i] began.
                            As you begin to near it, you notice the sun slowly rising above the rooftops.
                            You must have lost track of time while exploring the city.
                            Or maybe you were left waiting in the alleyway for the group to scatter for a longer time than you care to remember.
                            Either way, you guess you won't be sleeping tonight.
                        [/p]
                        [p]
                            The twinkle of magic you felt at midnight has all but faded, giving way to a sense of [b]vulnerability[/b].
                            You are but a lone, weak, small, kobold in a big city filled with people who can and [b]want[/b] to maim,
                            hurt, kill and perhaps even [b]rape[/b] you.
                            If you let your guard down for even one moment, it could cost you your money, your life or both.
                            Still, you held your own tonight.
                            You waited for exactly the right moment to strike, and you felled a human twice your size with a single precise blow.
                            Maybe living in the city for a while won't be too bad after all.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.hideAttack.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.hideAttack.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\alley\hide\attack\NightTwoHideAttackContinue.js

// scenes\Varanar\NightTwo\flee\alley\hide\attack\NightTwoHideAttack.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoHideAttack(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoHideAttackContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.nightTwo.alleyAttack = true;
        var toReturn = `[p]
                            This guy thinks he can just [b]piss[/b] all over whatever the hell he wants?!
                            Has he no decency?
                            No way.
                            You aren't going to let some [b]filthy human[/b] take a leak in your face, on top of a bunch of barrels that aren't even his.
                            He might not [b]know[/b] you're down here, but that just means you have the advantage of surprise.
                            Time to teach this guy a lesson about keeping his dick in his pants until he gets home.
                        [/p]
                        [p]
                            While a puddle of urine steadily forms next to you, you slowly rise from your huddled-up position on the ground,
                            like a snake rearing its head before [b]striking[/b].
                            Your eyes set on the ruffian's considerable package, you make your move.
                            You leap forwards, bundling all of your [b]anger[/b] and [b]hatred[/b] into a single powerful [b]swipe[/b].
                            Your sharp claws [b]smack[/b] against the human's exposed balls, [b]tearing[/b] through his scrotum, 
                            [b]ripping[/b] open his sack, tearing his nuts directly from his groin, mangling his genitals to beyond recognition.
                        [/p]
                        [p]
                            Your fingers [b]bathe[/b] in the human's blood, which gushes liberally from the gaping hole left where his testicles used to be.
                            His senses dulled by alcohol, it takes the man a moment to [b]realize[/b] what the hell is going on.
                            Looking downwards, he stares at the remnants of what used to be his balls, laying limply on the ground,
                            blood starting to mix with the ever-expanding piss-puddle.
                            His jaw drops agape.
                            He lets out an uncharacteristic high-pitched squeak.
                            His eyes rolling into the back of his head, the scoundrel tumbles backwards, falling lifelessly to the cold, hard ground.
                            Unable to cope with the immense amount of pain, he loses his consciousness.
                            The ruffian faints, with his now neutered, blood-oozing flaccid cock still on full display.
                        [/p]
                        [p]
                            As soon as the hoodlum falls, you immediately rush past him towards the entry to the alleyway.
                            Luckily, none of his [i]friends[/i] seem to have noticed the sound of his body falling to the floor.
                            They have continued walking along the road, and are starting to disperse.
                            You wait for a few more minutes for them to leave the area completely.
                            After wiping your hands on the downed human's clothes, 
                            you quickly and quietly escape the scene,
                            swiftly making your way to a more [i]hospitable[/i] neighbourhood.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.hideAttack.continued() ? "disabled" : "";
        var isSelected = ctrl.hideAttack.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.hideAttack.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.hide.get() !== NightTwoHideChoice.ATTACK) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.hideAttack.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.hide.get() !== NightTwoHideChoice.ATTACK) {
            return;
        }

        next.registerListeners();

        if (ctrl.hideAttack.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\flee\alley\hide\attack\NightTwoHideAttack.js

// scenes\Varanar\NightTwo\flee\alley\hide\stay\NightTwoHideStayContinue.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoHideStayContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var shirt = ctrl.misc.shirt();
        var clothes = shirt || pants;
        var toReturn = `[p]
                            The moment the human leaves the alleyway, you instantly rise to your feet and clean your face off with your single dry hand.
                            You feel completely [b]disgusted[/b].
                            Unable to weather the stench for even a second longer, you step out from behind the barrels,
                            and wander deeper into the alleyway in search of some fresh air.
                            At this point, you don't even [b]care[/b] about getting caught anymore.
                            The group seems to have wander off, but you'd have broken your cover even if they hadn't.
                        [/p]
                        [p]
                            Leaning against the brick wall sealing off the alley, you bend over and retch for a few solid minutes straight,
                            stuck in between the intense urge to vomit and the desire [b]not[/b] to let your last meal go to waste.
                            You can still [b]reek[/b] the human's piss, splatters of which linger on your ${clothes ? `clothing` : `body`} and on your face.
                            You fear the scent may have [b]seeped[/b] into your very scales, even though you didn't get [i]that much[/i] of it on you.
                        [/p]
                        [p]
                            While sickened by the lasting urine-smell, you are grateful that at least the fragrance of refuse hasn't reached this end of the alley.
                            At least here, you don't have to suffer the mixture of sweaty human groin, reeking barrels and dehydrated pee all at once.
                            Slowly but surely, your retching comes to an end.
                            Holding your breath, you walk past the scene of the crime, 
                            looking everywhere [b]but[/b] at the stagnating puddle in between the barrels to check whether the hoodlums are still around or not.
                            The street seems empty.
                            You think they've left.
                            You [b]hope[/b] they've left.
                        [/p]
                        [p]
                            You sneak out of the alleyway, before quickly and quietly making your way back to a more [i]hospitable[/i] part of town.
                            While, at first, you curse the small, chilly morning wind that picks up,
                            you soon learn to appreciate it for washing the stench of piss off of you,
                            and for fully drying your ${clothes ? `clothes` : `body`} in a matter of minutes.
                        [/p]
                        [p]
                            After removing yourself from the [i]brothel district[/i] proper,
                            you start walking back towards the town square,
                            where your [i]urban adventure[/i] began.
                            As you begin to near it, you notice the sun slowly rising above the rooftops.
                            You must have lost track of time while exploring the city.
                            Or maybe you were left retching in the alleyway for a longer time than you care to remember.
                            Either way, you guess you won't be sleeping tonight.
                        [/p]
                        [p]
                            The twinkle of magic you felt at midnight has all but faded, giving way to a sense of [b]vulnerability[/b].
                            You are but a lone, weak, small, kobold in a big city filled with people who can and [b]want[/b] to maim,
                            hurt, kill and perhaps even [b]rape[/b] you.
                            If you let your guard down for even one moment, it could cost you your money, your life or both.
                            And next time, the drunk taking a leak against your hiding spot might actually [b]notice[/b] you.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.hideStay.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.hideStay.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\alley\hide\stay\NightTwoHideStayContinue.js

// scenes\Varanar\NightTwo\flee\alley\hide\stay\NightTwoHideStay.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoHideStay(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoHideStayContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var pants = ctrl.misc.pants();
        var shirt = ctrl.misc.shirt();
        var clothes = shirt || pants;
        var toReturn = `[p]
                            While you would [b]love[/b] to [b]rip his nuts off[/b] with your sharp claws,
                            you simply can't risk the human calling his buddies over before he goes down.
                            There's no way you can take on all of them, unless they each approach one of at atime with their pants down and their dicks out.
                            While you [b]hate[/b] being forced to do this, you're just going to have to sit here and be quiet while the dehydrated drunkard does his thing.
                        [/p]
                        [p]
                            The man lets out a loud sigh of relief as he continues to let out an entire night's worth of booze-turned-urine,
                            a considerable puddle of the stuff forming on the ground right next to you.
                            You try to scooch over, to scuttle to the side just a little bit more to avoid the absolutely [b]revolting[/b] liquid,
                            yet after managing to spare yourself for half an inch more, there is simply nowhere left to run.
                            You almost [b]vomit[/b] as you feel the lukewarm fluid flow around your scaly hand, placed on the ground to keep you balanced,
                            but knowing your very life is at stake, you remain staunch and stalwart, resolute and unwavering.
                        [/p]
                        [p]
                            However, there are limits to your determination, limits which are put to the test as the ruffian begins to wobble on his feet,
                            his steady stream of urine erratically flailing about, splatterings of piss ${clothes ? `matting your clothing` : `befouling your scales`}.
                            As the human nearly topples over fully, his torrent deviates the most, the flow splashing directly against your face for just a brief moment,
                            completely [b]sullying[/b] your once-clean visage, drenching you in foul-smelling secretions,
                            your stomach managing to retain its contents by your sheer will to live alone.
                        [/p]
                        [p]
                            While your [b]hatred[/b] for the hoodlum runs deep, and the urge to attack him, to [b]maim[/b] him,
                            to [b]scar[/b] him for [/b]life[/b] as punishment for outright taking a [b]piss[/b] on you,
                            you don't even [b]dare[/b] to wipe your face out of fear that he'll notice.
                            His warm urine nauseatingly runs down your forehead and alongside your muzzle, before dripping to the ground below.
                            Luckily, none of it runs down your neck.
                        [/p]
                        [p]
                            Not long after being outright [b]sprayed[/b], the flood of vile fluid begins to falter.
                            After squirting out a final few jets of piss, and shaking off any lingering droplets,
                            the man repackages his flaccid cock, sloppily buttons up his pants, and walks away,
                            leaving you huddled up in a puddle of his slowly cooling mess.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.hideStay.continued() ? "disabled" : "";
        var isSelected = ctrl.hideStay.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.hideStay.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.hide.get() !== NightTwoHideChoice.STAY) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.hideStay.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.hide.get() !== NightTwoHideChoice.STAY) {
            return;
        }

        next.registerListeners();

        if (ctrl.hideStay.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\flee\alley\hide\stay\NightTwoHideStay.js

// scenes\Varanar\NightTwo\flee\alley\hide\NightTwoHide.js

/**
 * Fleeing section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoHide(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var stayHiddenId = arguments.callee.name + "ButtonOne";
    var attackId = arguments.callee.name + "ButtonTwo";

    var toStayHidden = new NightTwoHideStay(ctrl);
    var toAttack = new NightTwoHideAttack(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            It's too late to run, and you can't get over a wall that high.
                            The barrels.
                            Those are your best option.
                            Combined with the inherent darkness of the alleyway, the cover provided by them is sure to keep you hidden from a gang of drunkards.
                            At least you hope it will.
                        [/p]
                        [p]
                            You crouch down, crawling in between two of the barrels.
                            They [b]reek[/b] of refuse, probably containing something rotten, or maybe a family of rats made their home inside.
                            You have no choice [b]but[/b] to tolerate the stench, for now.
                            Concealed between a pair of barrels and a stone wall, you slowly turn around in your hiding spot,
                            facing the street-end to monitor for any new developments.
                        [/p]
                        [p]
                            A few members of the group pass by without even casting a glance down the alley in which you are hidden.
                            Your plan is working [b]perfectly[/b].
                            Heck, since they're not even looking this way, you didn't even need to bother with squeezing in between these smelly containers.
                            Taking your safety for granted, you let down your guard for a few seconds,
                            covering your snout to prevent the worst of the smell from reaching your nostrils.
                        [/p]
                        [p]
                            Your gaze returning to the entrance of the alley, a small, inadvertent gasp escapes your lips.
                            One of the men is drowsily standing at the far-end of the corridor, peering into it with half-attentive eyes.
                            You lower the hand covering your nose to your mouth, to muffle the quiet noise of your audible surprise.
                            Scared shitless, you reduce your posture as much as you can, curling up into a tiny little ball in hopes that he hasn't seen you [i]yet[/i].
                        [/p]
                        [p]
                            Luckily, humans are [b]not[/b] as skilled as kobolds are when it comes to seeing in the dark.
                            And even if they [b]were[/b], this one looks [b]so[/b] drunk he can barely stand on his own two feet,
                            let alone spot a tiny little lizard like yourself concealed behind a set of barrels in a pitch-black alley.
                            Still, despite not having seen you, the man casually saunters forwards, approaching your position with an unsteady gait.
                        [/p]
                        [p]
                            In spite of needing to brace himself against one of the walls to stay upright,
                            the human walks up to the very set of barrels you are hiding behind.
                            [b]Like he knows you're there.[/b]
                            Your heart races a million miles an hour.
                            You can [b]smell[/b] the stench of [b]booze[/b] lingering about him even through the odour of rot.
                            You let out a small whimper, almost betraying your location.
                            Luckily, he doesn't seem to notice.
                            You're still safe.
                            He [b]doesn't[/b] know.
                        [/p]
                        [p]
                            The man steadies himself against one of the barrels you're hiding behind, his boot-clad feet mere inches from your face.
                            He moves one of his hands to his groin, undoing the buttons of his linen pants.
                            Wait, what?
                            Reaching into his trousers, the man deftly lowers his underwear, exposing his [b]huge[/b], heavily oversized flaccid member,
                            an unwashed, manly fragrance mixing with the rest of the unpleasant scents.
                            Taking hold of his length, he points it downwards, not directly at you, but in your general vicinity.
                            [i]Oh no.[/i]
                        [/p]
                        [p]
                            Like the unsavory drunk that he is, the man releases his bladder,
                            a dark-yellow stream of pungent piss splattering against the wall besides you,
                            its hefty, acrid fragrance immediately [b]overpowering[/b] all of the other scents,
                            clogging your nose to the point of nearly making you [b]retch[/b].
                            That filthy human wasn't even looking for [b]you[/b], just for a place to take a leak!
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildStayHiddenChoice() {
        var disabled = ctrl.hide.has() ? "disabled" : "";
        var isSelected = ctrl.hide.get() === NightTwoHideChoice.STAY;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${stayHiddenId}" ${disabled}>Stay hidden</button>`;

        var tooltip = global.parser.parse(`Weather through the disgusting ordeal. He'll leave when he's done, right?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildAttackChoice() {
        var disabled = ctrl.hide.has() ? "disabled" : "";
        var isSelected = ctrl.hide.get() === NightTwoHideChoice.ATTACK;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${attackId}" ${disabled} style="width: 160px;">Attack the man</button>`;

        var tooltip = global.parser.parse(`With his pants down, you have a clear shot at the human's most [i]delicate[/i] parts.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildStayHiddenChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildAttackChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + stayHiddenId), $("#" + attackId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerStayHiddenListener() {
        $("#" + stayHiddenId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.hide.set(NightTwoHideChoice.STAY);
            loadNextPart(toStayHidden);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerAttackListener() {
        $("#" + attackId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.hide.set(NightTwoHideChoice.ATTACK);
            loadNextPart(toAttack);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.alley.get() !== NightTwoAlleyChoice.HIDE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toStayHidden.build();
        part += toAttack.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.alley.get() !== NightTwoAlleyChoice.HIDE) {
            return;
        }

        toStayHidden.registerListeners();
        toAttack.registerListeners();

        if (ctrl.hide.has()) {
            return;
        }

        registerStayHiddenListener();
        registerAttackListener();
    };
}

var NightTwoHideChoice = {
    STAY: 1,
    ATTACK: 2
};

// END scenes\Varanar\NightTwo\flee\alley\hide\NightTwoHide.js

// scenes\Varanar\NightTwo\flee\alley\run\fail\stay\NightTwoAlleyStayContinue.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlleyStayContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var underwear = ctrl.misc.underwear();
        var pants = ctrl.misc.pants();
        var shirt = ctrl.misc.shirt();
        var clothing = underwear || pants || shirt;
        var toReturn = `[p]
                            As soon as the man has left the alleyway, you rise out of the puddle of pee.
                            Tears flow liberally from your eyes.
                            You feel [b]unclean[/b], [b]defiled[/b] from head to toe.
                            You wipe your face as best you can, but even dry, your scales seem [b]filthy[/b].
                            There's a horrid [b]stench[/b] lingering about, a revolting fragrance so intense you fear it may seep into your very skin.
                        [/p]
                        [p]
                            You clean off your arms, your legs and your tail as best you can.
                            Luckily, the digusting yellow fluid rolls easily off your scales, dripping down to the ground.
                            While warm at first, the liquid is quickly starting to cool down, leaving you chilly and cold on top of being absolutely [b]revolted[/b].
                            Despite cleaning your body as best as you can, the sleazy, unsanitary feeling remains.
                            ${clothing ? `Your clothes are tainted.
                            You don't want to take them off in the middle of town.
                            They'll just have to dry on their own.` : ""}
                        [/p]
                        [p]
                            While completely disheartened, reduced to a sobbing mess of a piss-stained lizard,
                            you are glad that you escaped that encounter relatively unharmed.
                            After a few solid minutes of bawling your eyes out, you wipe your tears, and stumble towards the alleyway's entrance.
                            The men are nowhere around to be seen.
                            Having traded your dignity for your life, you start walking back towards the town square.
                        [/p]
                        [p]
                            While, at first, you curse the cool morning wind that picks up,
                            you soon learn to appreciate it for washing the stench of piss off of you,
                            and for somewhat helping to dry your ${clothing ? `clothes` : `body`}.
                            As you begin to near your destination, you notice the sun slowly rising above the rooftops.
                            You must have lost track of time while exploring the city.
                            Or maybe you were left crying in the alleyway for a longer period than you care to remember.
                            Either way, you guess you won't be sleeping tonight.
                        [/p]
                        [p]
                            The twinkle of magic you felt at midnight has all but faded, giving way to a sense of [b]vulnerability[/b].
                            You are but a lone, weak, small, kobold in a big city filled with people who can and [b]want[/b] to maim,
                            hurt, kill and perhaps even [b]rape[/b] you.
                            If you let your guard down for even one moment, it could cost you your money, your life or both.
                            You'd worry about losing your pride, but you fear you have little of that left as it is.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.alleyStay.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.alleyStay.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\alley\run\fail\stay\NightTwoAlleyStayContinue.js

// scenes\Varanar\NightTwo\flee\alley\run\fail\stay\NightTwoAlleyStay.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlleyStay(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoAlleyStayContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.nightTwo.alleyPiss = true;
        var underwear = ctrl.misc.underwear();
        var pants = ctrl.misc.pants();
        var shirt = ctrl.misc.shirt();
        var clothing = underwear || pants || shirt;
        var toReturn = `[p]
                            You don't have a choice.
                            The hoodlum is twice your size,
                            and he has a bunch of equally menacing friends waiting just outside of the alleyway to come [b]fuck you up[/b] unless you do as he says.
                            You glare up at the human.
                            He smirks down at you.
                            Taking your silence for acceptance, he grabs his humongous shaft, and points it directly at you.
                            Before you have a chance to protest, the drunkard  releases his bladder, and a thick stream of dark-yellow, acrid,
                            foul-smelling urine hits you straight on the forehead, splattering across the entirety of your face.
                        [/p]
                        [p]
                            You immediately close your eyes and your mouth, your tooth-baring scowl replaced by a pathetic, submissive frown.
                            The man lets out a domineering chuckle, as he begins to shift his member about, directing the stream of disgusting, lukewarm liquid.
                            First, he ensures your entire face has been coated, from the top of your head to the underside of your muzzle.
                            You gag and retch as a few droplets find their way inside of you nostrils,
                            an absolutely [b]vile[/b] taste building up at the back of your throat, as though you can [b]taste[/b] it.
                        [/p]
                        [p]
                            You lift up your hands and your arms in an attempt to cover your face, trying to [b]block[/b] the flow, but the human simply lowers his aim,
                            ${clothing ? `completely [b]drenching[/b] your clothing` : `soaking your body`} in piss instead.
                            A small puddle of the stuff forms underneath you, which makes you feel even [b]more[/b] dirty and uncomfortable,
                            as though you're [b]wallowing[/b] in it.
                            On the very [b]verge[/b] of throwing up, you finally feel the flow starting to lessen,
                            the once-powerful stream blasting across your chest reduced to a minor trickle spilling down onto your legs.
                        [/p]
                        [p]
                            "[color=dred]You look good in yellow.[/color]" the man taunts, pulling up and re-buttoning his pants.
                            Snickering at his own joke, he realizes you aren't sharing his jolly sentiment.
                            He crudely spits on the ground next to you, his saliva quickly mixing with the expanding pool of urine.
                            "[color=dred]Next time, I'll let my buddies take a leak on you as well, you ungrateful [b]bitch[/b].
                            That's all you [b]filthy lizards[/b] are good for.[/color]"
                            That last part being more of a mumbling to himself, the man turns around and begins to walk away,
                            leaving you crying in the middle of a puddle of his piss.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.alleyStay.continued() ? "disabled" : "";
        var isSelected = ctrl.alleyStay.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.alleyStay.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.alleyRunFail.get() !== NightTwoAlleyRunFailChoice.STAY) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.alleyStay.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.alleyRunFail.get() !== NightTwoAlleyRunFailChoice.STAY) {
            return;
        }

        next.registerListeners();

        if (ctrl.alleyStay.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\flee\alley\run\fail\stay\NightTwoAlleyStay.js

// scenes\Varanar\NightTwo\flee\alley\run\fail\eager\NightTwoAlleyEagerContinue.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlleyEagerContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Dawn breaks.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.alleyEager.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.alleyEager.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\alley\run\fail\eager\NightTwoAlleyEagerContinue.js

// scenes\Varanar\NightTwo\flee\alley\run\fail\eager\NightTwoAlleyEager.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlleyEager(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoAlleyEagerContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Eagerly gulp down piss. Slut.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.alleyEager.continued() ? "disabled" : "";
        var isSelected = ctrl.alleyEager.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.hideAttack.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.alleyRunFail.get() !== NightTwoAlleyRunFailChoice.EAGER) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.alleyEager.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.alleyRunFail.get() !== NightTwoAlleyRunFailChoice.EAGER) {
            return;
        }

        next.registerListeners();

        if (ctrl.alleyEager.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\flee\alley\run\fail\eager\NightTwoAlleyEager.js

// scenes\Varanar\NightTwo\flee\alley\run\fail\bite\NightTwoAlleyBite.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlleyBite(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    var idsToRemove = [
        "NightTwoAlleyBite",
        "NightTwoAlleyRunFail",
        "NightTwoAlleyRun",
        "NightTwoAlley",
        "NightTwoFlee",
        "NightTwoRunIn",
        "NightTwoStage"
    ];

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Who the [b]hell[/b] does this guy think he is?!
                            Does he really expect you to [b]stay put[/b] and let him [b]take a piss on you[/b]?!
                            He might have met some meek kobolds in his lifetime, but you're [b]not[/b] one of those.
                            You're a fighter.
                            You don't spit [b]or[/b] swallow.
                            You [b]bite[/b].
                        [/p]
                        [p]
                            Glaring upwards at the man, you take a mental snapshot of the smug, coy look on his face.
                            You're going to [b]enjoy[/b] his look of anguish after you [b]bite his dick off[/b].
                            You shift around a little, positioning yourself in a better way to [b]lash out[/b].
                            "[color=dred]Quit fidgetin'.
                            It's comin', just you wait.[/color]"
                            the hoodlum comments, mistaking your assault-preparations for anxiousness.
                            He probably thinks you're going to attempt to run away.
                            Poor him.
                        [/p]
                        [p]
                            You bide your time for a few more moments,
                            trying to find a good balance between getting the ruffian to [i]relax[/i] so you can catch him off guard,
                            without waiting [i]too[/i] long, lest he start [i]doing his thing[/i].
                            You have no intentions of getting peed on, at least not today.
                            Mere moments before the scoundrel was about to release his bladder, you make your move.
                        [/p]
                        [p]
                            You lift yourself off the ground, and [b]leap[/b] forwards in a single, smooth move.
                            Your muzzle is wide open, and before he even knows what's going on, you've taken the man's disgusting, uncut and unwashed member into your mouth.
                            His thick shaft barely even [b]fits[/b] in your mouth.
                            Ere he has a chance to piss down your throat, you [b]bite down[/b] as [b]hard[/b] as you can,
                            your teeth sinking into the soft flesh of the human's cock, severing off more than a few inches.
                        [/p]
                        [p]
                            The hoodlum [b]cries[/b] out in [b]pain[/b], screaming loud enough to wake up the entire neighbourhood as he slowly steps backwards,
                            tearing the final few bits of flesh holding the tip of his member in place himself.
                            Blood gushes from the remainder of his length, splattering your face.
                            For a few moments, the human clutches his groin in disbelief.
                            Then, unable to process the pain, he falls down to the ground, knocked unconscious through sheer pain alone.
                        [/p]
                        [p]
                            Alas, it seems like the other delinquents [b]heard[/b] their comrade's cry for help.
                            The entire group [b]rushes[/b] into the alleyway, to find a kobold crouching over their downed friend's body.
                            You throw a wicked grin their way, baring your teeth and the fleshy organ in between them.
                            To make your point even more clear, you spit the severed cock-length out onto the unconscious ruffian's body.
                        [/p]
                        [p]
                            "[color=bred]That kobold bit off Carl's [b]dick[/b]! [b]Let's get'em![/b][/color]" the ringleader of the pack proclaims.
                            It is at that point you realize you might have [i]bitten off a bit more than you can chew[/i].
                            Stuck in self-blaming thoughts, you remain frozen in place as the gang rushes towards you.
                            The first human to arrive crudely [b]kicks[/b] you in the face, sending you flying backwards,
                            the rear-end of your head painfully [b]crashing[/b] into the brick wall, almost knocking you out from the get-go.
                        [/p]
                        [p]
                            Alas, you are spared the mercy of unconsciousness.
                            Left reeling from the initial blow, you clutch your hurting face, your nose bleeding profusely.
                            The men give you no respite.
                            Mere moments after kicking you into the wall, a second human jumps on top of you, and begins to [b]pound[/b] your body with blow after blow,
                            using both of his fists to [b]brutalize[/b] your upper body, while pinning you down using nothing but his own weight,
                            preventing you from so much as covering yourself with your arms.
                        [/p]
                        [p]
                            The brutal torment goes on for what feels like [b]ages[/b].
                            You can feel your bones [b]cracking[/b], breaking under the incredible strain of repeated, full-strength blows.
                            Blood begins to pile up at the back of your throat as one of your ribs painfully [b]punctures[/b] your lungs.
                            You try to [b]beg[/b] the man for mercy, yet nothing but a bloody gurgle escapes your lips.
                        [/p]
                        [p]
                            Growing tired, the man rises from your battered frame.
                            The rest of the group is [b]far[/b] from done with you, however.
                            Your next tormentor steps forwards.
                            Looking to get a rise out of you, and wanting to look cool in front of his friends,
                            he cruelly plants his foot in between your slightly parted legs, brutally [b]kicking[/b] you in the groin,
                            harsh enough for the tip of his boot to [b]sink[/b] into your slit.
                            Horrible pain like nothing you've [b]ever[/b] felt before shoots up your spine,
                            but there's barely enough air in your blood-filled lungs to keep you [b]conscious[/b], let alone to voice your agony with.
                        [/p]
                        [p]
                            Upset that he didn't get to hear you [b]squeal[/b], the cruel hoodlum starts a series of [b]kicks[/b],
                            focusing on your lower half, bruising and severely [b]injuring[/b] everything from your stomach to your tail,
                            on which he [b]stomps[/b] on repeatedly.
                            At first, you try your hardest to move your tail out of the way, but after the second time his heavy boot comes down onto it,
                            something [b]snaps[/b], and a new wave of [b]anguish[/b] floods your body.
                            You find yourself unable to move your tail from there on out.
                        [/p]
                        [p]
                            The third delinquent focuses on your limbs.
                            He grabs your arm, and pins it.
                            He demands you [b]beg[/b] him for mercy, yet you manage to utter nothing but a painful, blood-filled cough.
                            Without warning, he cruelly [b]twists[/b] the appendage in his grasp, bending your arm in a way it was [b]never[/b] meant to go.
                            With a loud, sickening [b]crack[/b], your bones break in half.
                            A fresh wave of tears flow from your abused eyesockets as your arm is snapped like a twig.
                            The pain is [b]horrendous[/b], but for whatever reason, your mind [b]refuses[/b] to let you pass out.
                        [/p]
                        [p]
                            Every member of the group takes a turn.
                            Some simply batter your face, others break your fingers, knock out your teeth or torture your feet.
                            One of the humans outright [b]cuts off[/b] the tip of your tail, maiming you like you did their friend.
                            Then, for a moment, the constant beatings come to an end.
                            Barely conscious, you blurrily see the ringleader of the group approach through your one still-functioning eye.
                            He presses his shoe-clad foot onto your throat, cutting off what little air intake you had left.
                            Then, he twists his heel.
                            You hear something crack.
                            Your torment is finally over.
                            You welcome death, and death welcomes you.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    Dead end. Just like the alleyway.
                                [/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" style="margin-left: auto; position: relative; float: right;" id="${continueId}">`;
        choice += `&lt;&lt Revert`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400, function () {
            var main = $("#main");
            main.scrollTop(main[0].scrollHeight - main.height() - 5); // Fix scrolling
        });
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.misc.unsetAll();
            ctrl.alleyRunFail.undo();
            ctrl.alleyRunFail.unset();
            ctrl.alley.unset();
            ctrl.flee.undo();
            ctrl.flee.unset();
            ctrl.runin.unset();
            ctrl.stage.unset();
            for (var i = 0; i < idsToRemove.length; i++) {
                $("#" + idsToRemove[i]).remove();
            }
            loadNextPart(new NightTwoStage(ctrl));
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.alleyRunFail.get() !== NightTwoAlleyRunFailChoice.BITE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.alleyRunFail.get() !== NightTwoAlleyRunFailChoice.BITE) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Varanar\NightTwo\flee\alley\run\fail\bite\NightTwoAlleyBite.js

// scenes\Varanar\NightTwo\flee\alley\run\fail\NightTwoAlleyRunFail.js

/**
 * Fleeing section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlleyRunFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var biteId = arguments.callee.name + "ButtonOne";
    var stayId = arguments.callee.name + "ButtonTwo";
    var eagerId = arguments.callee.name + "ButtonThree";

    var toBite = new NightTwoAlleyBite(ctrl);
    var toStay = new NightTwoAlleyStay(ctrl);
    var toEager = new NightTwoAlleyEager(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var agiFail = ctrl.misc.agility() < 6;
        var toReturn = `[p]
                            You envision your way up along the side of the wall on the far end of the alley.
                            First you'll run towards it, then you intend to [b]jump[/b], and using the momentum gained from your dash,
                            you'll kick yourself off of one of the walls, ending up high enough to latch on to the top of the brick wall.
                            From then, it's a matter of pulling yourself up, and after that you'll be in the clear.
                        [/p]
                        [p]
                            With your plan visualized, it's time to put it into motion.
                            You [b]sprint[/b] forwards, accelerating at a rapid pace.
                            Running past the last of the barrels, you [b]leap[/b] from the ground as high up as you can.
                            Mid-air, you press your right foot against the alley-wall, and expertly push yourself up even higher,
                            converting horizontal momentum into vertical lift.
                        [/p]`;

        if (agiFail) {
            toReturn += `[p]
                            Alas, you seem to have misjudged the actual height of the wall.
                            You reach for the top, yet come just a few inches short of being able to grab onto it.
                            Without anything to grip onto, your entire body smacks into the solid barrier, and you fall to the ground.
                            Slightly dazed, you slump down with your back against the wall, just in time to witness one of the humans entering the alleyway.
                        [/p]
                        ${global.ui.condFail(`Agility below 6`)}`;
        } else {
            toReturn += `[p]
                            Exactly as planned, you narrowly reach the height required to grab the top of the wall,
                            your claws digging in to the mortar in between the bricks ever so slightly.
                            Despite trying with all your might, however, you utterly [b]fail[/b] to pull your meagre bodyweight up onto the barrier.
                            You struggle in vain for nearly half a minute to try and lift yourself up higher, before your arms can carry your weight no longer.
                            You fall to the ground, landing painfully on your rear.
                            Slightly dazed, you slump down with your back against the wall, just in time to witness one of the humans entering the alleyway.
                        [/p]
                        ${global.ui.condFail(`Strength below 6`)}`;
        }

        toReturn += `   [p]
                            Paralyzed with a mixture of fear and stinging pain, you are frozen in place as the ruffian slowly approaches.
                            He clearly knows you're there, staring directly at your downed form, yet he isn't calling for his friends.
                            Perhaps he wants the pleasure of [i]beating you to death[/i] all for himself.
                            You [i]should[/i] ready yourself for combat,
                            yet your body and your mind both feel [b]far[/b] too exhausted and utterly [b]defeated[/b] to even [b]attempt[/b] to put up a fight.
                        [/p]
                        [p]
                            The scoundrel stops just a few feet in front of you.
                            Glaring down at you with angry, hate-filled eyes, he speaks.
                            "[color=dred]You're that lizard we're chasin'.[/color]"
                            His voice is bass, and some of his words come out slurred due to his heavy intoxication.
                            Still, not unfamiliar with being drunk yourself, you understand every word he's saying.
                            "[color=dred]My friends want to [b]kill[/b] you.
                            I think that'd be a [b]waste[/b].
                            You'll make a fine [b]kobold whore[/b] one day,
                            and Varanar could use more cheap pieces of [b]ass[/b].[/color]"
                        [/p]
                        [p]
                            A sadistic smile forms atop of the bulky hoodlum's face.
                            He moves one of his hands towards his groin, and casually begins to unbutton his pants.
                            "[color=dred]I came into this alley to take a leak,
                            and that's exactly what I'm gonna do.
                            You stay put like the lil' [b]slut[/b] you are,
                            and I won't have a reason to call my buddies over here, you got that?[/color]"
                            Without even bothering to wait for a response, the man promptly lowers his underwear,
                            the stench of unwashed human groin creeping into your nostrils as his [b]huge[/b],
                            heavily oversized flaccid member is made bare to you.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildBiteChoice() {
        var disabled = ctrl.alleyRunFail.has() ? "disabled" : "";
        var isSelected = ctrl.alleyRunFail.get() === NightTwoAlleyRunFailChoice.BITE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${biteId}" ${disabled}>Bite</button>`;

        var tooltip = global.parser.parse(`What the hell does this guy think he's doing?! Bite his dick off. That'll show him.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildStayChoice() {
        var disabled = ctrl.alleyRunFail.has() ? "disabled" : "";
        var isSelected = ctrl.alleyRunFail.get() === NightTwoAlleyRunFailChoice.STAY;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${stayId}" ${disabled}>Stay still</button>`;

        var tooltip = global.parser.parse(`You don't want to get bludgeoned to death by a bunch of angry, drunken humans.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildEagerChoice() {
        var disabled = ctrl.alleyRunFail.has() ? "disabled" : "";
        var isSelected = ctrl.alleyRunFail.get() === NightTwoAlleyRunFailChoice.EAGER;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${eagerId}" ${disabled}>Open your mouth</button>`;

        var tooltip = global.parser.parse(`You thought this guy was going to [b]kill[/b] you. Now that you know he isn't, you might as well have some fun.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildBiteChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildStayChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + biteId), $("#" + stayId), $("#" + eagerId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerBiteListener() {
        $("#" + biteId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.alleyRunFail.set(NightTwoAlleyRunFailChoice.BITE);
            loadNextPart(toBite);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerStayListener() {
        $("#" + stayId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.alleyRunFail.set(NightTwoAlleyRunFailChoice.STAY);
            loadNextPart(toStay);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerEagerListener() {
        $("#" + eagerId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.alleyRunFail.set(NightTwoAlleyRunFailChoice.EAGER);
            loadNextPart(toEager);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.alleyRunFail.change();
        part += buildChoice();
        part += `</div>`;
        part += toBite.build();
        part += toStay.build();
        part += toEager.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        toBite.registerListeners();
        toStay.registerListeners();
        toEager.registerListeners();

        if (ctrl.alleyRunFail.has()) {
            return;
        }

        registerBiteListener();
        registerStayListener();
        registerEagerListener();
    };
}

var NightTwoAlleyRunFailChoice = {
    BITE: 1,
    STAY: 2,
    EAGER: 3
};

// END scenes\Varanar\NightTwo\flee\alley\run\fail\NightTwoAlleyRunFail.js

// scenes\Varanar\NightTwo\flee\alley\run\succeed\NightTwoAlleyRunSuccessContinue.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlleyRunSuccessContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            After removing yourself from the [i]brothel district[/i] proper,
                            you start walking back towards the town square,
                            where your [i]urban adventure[/i] began.
                            As you begin to near it, you notice the sun slowly rising above the rooftops.
                            You must have lost track of time while exploring the city.
                            Or maybe you were being chased by those men for a longer time than you remember.
                            Either way, you guess you won't be sleeping tonight.
                        [/p]
                        [p]
                            The twinkle of magic you felt at midnight has all but faded, giving way to a sense of [b]vulnerability[/b].
                            You are but a lone, weak, small, kobold in a big city filled with people who can and [b]want[/b] to maim,
                            hurt, kill and perhaps even [b]rape[/b] you.
                            If you let your guard down for even one moment, it could cost you your money, your life or both.
                            And next time you might not be so lucky as to be able to freerun your way up a wall without screwing up.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.alleyRunSuccess.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.alleyRunSuccess.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\alley\run\succeed\NightTwoAlleyRunSuccessContinue.js

// scenes\Varanar\NightTwo\flee\alley\run\succeed\NightTwoAlleyRunSuccess.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlleyRunSuccess(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoAlleyRunSuccessContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You envision your way up along the side of the wall on the far end of the alley.
                            First you'll run towards it, then you intend to [b]jump[/b], and using the momentum gained from your dash,
                            you'll kick yourself off of one of the walls, ending up high enough to latch on to the top of the brick wall.
                            From then, it's a matter of pulling yourself up, and after that you'll be in the clear.
                        [/p]
                        [p]
                            With your plan visualized, it's time to put it into motion.
                            You [b]sprint[/b] forwards, accelerating at a rapid pace.
                            Running past the last of the barrels, you [b]leap[/b] from the ground as high up as you can.
                            Mid-air, you press your right foot against the alley-wall, and expertly push yourself up even higher,
                            converting horizontal momentum into vertical lift.
                        [/p]
                        [p]
                            Exactly as planned, you narrowly reach the height required to grab the top of the wall,
                            your claws digging in to the mortar in between the bricks ever so slightly.
                            Using all of your strength, you deftly begin to pull yourself upwards,
                            lifting your light frame up onto the barrier, coming to a seated rest on top of it.
                        [/p]
                        ${global.ui.condSucc(`Strength and agility greater than 5`)}
                        [p]
                            Just in time, or so it seems.
                            One of the men chasing you wanders into the alleyway.
                            Whether he didn't see you perched up high, or if he simply had no intention of calling for his friends, you will never know.
                            You lower yourself on the opposite end of the brick wall, and with a feline-resembling grace, you hop down to the ground once more.
                            With the group shaken off, for now, at least, you quickly and quietly make your way to [i]less hostile[/i] parts of Varanar.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.alleyRunSuccess.continued() ? "disabled" : "";
        var isSelected = ctrl.alleyRunSuccess.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.alleyRunSuccess.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.alleyRunSuccess.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.alleyRunSuccess.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\flee\alley\run\succeed\NightTwoAlleyRunSuccess.js

// scenes\Varanar\NightTwo\flee\alley\run\NightTwoAlleyRun.js

/**
 * Fleeing section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlleyRun(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new NightTwoAlleyRunFail(ctrl);
    var toSuccess = new NightTwoAlleyRunSuccess(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.agility() > 5 && ctrl.misc.strength() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.alley.get() !== NightTwoAlleyChoice.RUN) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.alley.get() !== NightTwoAlleyChoice.RUN) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\alley\run\NightTwoAlleyRun.js

// scenes\Varanar\NightTwo\flee\alley\NightTwoAlley.js

/**
 * Fleeing section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoAlley(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var hideId = arguments.callee.name + "ButtonOne";
    var runId = arguments.callee.name + "ButtonTwo";

    var toHide = new NightTwoHide(ctrl);
    var toRun = new NightTwoAlleyRun(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Desperate to get away from the men, your mind forms a fickle escape plan.
                            You turn a corner, and immediately [b]bolt[/b] for the nearest alleyway.
                            Stepping into the tight, unlit corridor, you let your eyes adjust to the complete darkness for a few moments, your superb nightvision kicking in.
                            The alley you've entered is a dead end, the other side blocked off by a pretty high brick barrier.
                            A collection of barrels are unorderly arrayed against the leftmost wall, leaving barely enough room for a full-grown human to squeeze on by.
                        [/p]
                        [p]
                            This place [b]stinks[/b] like rotten fish.
                            You consider backing out and either picking a different alleyway, or just ditching the idea of hiding completely, but alas, it is already too late.
                            Judging by the noise coming from the street, the men have followed in your footsteps, and they are currently out looking for you.
                            If you remain standing in the entrance, they'll undoubtedly find you.
                            You need to make a decision, and quick.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildHideChoice() {
        var disabled = ctrl.alley.has() ? "disabled" : "";
        var isSelected = ctrl.alley.get() === NightTwoAlleyChoice.HIDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${hideId}" ${disabled}>Hide</button>`;

        var tooltip = global.parser.parse(`Conceal yourself behind the barrels.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRunChoice() {
        var disabled = ctrl.alley.has() ? "disabled" : "";
        var isSelected = ctrl.alley.get() === NightTwoAlleyChoice.RUN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${runId}" ${disabled}>Run</button>`;

        var tooltip = global.parser.parse(`Try to parkour up and over the wall.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildRunChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildHideChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + hideId), $("#" + runId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerRunListener() {
        $("#" + runId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.alley.set(NightTwoAlleyChoice.RUN);
            loadNextPart(toRun);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerHideListener() {
        $("#" + hideId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.alley.set(NightTwoAlleyChoice.HIDE);
            loadNextPart(toHide);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.flee.get() !== NightTwoFleeChoice.HIDE) {
            return "";
        }

        global.ui.applyTheme("varanar_alley");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toRun.build();
        part += toHide.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.flee.get() !== NightTwoFleeChoice.HIDE) {
            return;
        }

        toRun.registerListeners();
        toHide.registerListeners();

        if (ctrl.alley.has()) {
            return;
        }

        registerRunListener();
        registerHideListener();
    };
}

var NightTwoAlleyChoice = {
    HIDE: 1,
    RUN: 2
};

// END scenes\Varanar\NightTwo\flee\alley\NightTwoAlley.js

// scenes\Varanar\NightTwo\flee\run\fail\die\NightTwoRunDeath.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoRunDeath(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    var idsToRemove = [
        "NightTwoRunDeath",
        "NightTwoRunBeatdown",
        "NightTwoKeepRunning",
        "NightTwoFlee",
        "NightTwoRunIn",
        "NightTwoStage"
    ];

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Desperate to get away from the men, you burst into a sprint the moment you turn a corner,
                            hoping to be able to outrun them before they even [i]know[/i] you're escaping.
                            While your gambit does pay off in the form of a head start, sadly, you don't manage to shake the group from the get-go.
                            Noticing their lizard-prey running away,
                            the men drop their tight-packed group-cohesion and begin to outright [b]chase[/b] after you.
                        [/p]
                        [p]
                            You run from street to street, periodically checking behind you to make sure they haven't given up yet.
                            While far bigger than you, and probably able to outrun you whilst sober, their heavy inebriation hinders the belligerent group significantly.
                            A couple of pursuers tumble over their own feet, falling to the ground, being left behind by their peers.
                            A few of the men get too sick to keep pursuing you, retreating into alleyways to empty their bellies of food and drink.
                            Some of them are simply too weary to keep up pursuit.
                            Initially numbering high in the single-digits,
                            it takes only a few solid minutes of running for their [i]pack[/i] to be reduced to three hard-boiled men.
                        [/p]
                        [p]
                            But still, even [b]one[/b] of them can [b]cave your head in[/b] if they manage to catch you.
                            You aren't out of the woods until you can shake off [b]all three[/b] of them.
                            And you get the feeling that all this running is making them [b]even more frustrated[/b].
                            There's no doubt in your mind: your very life is at stake here.
                            And the humans chasing you are [b]determined[/b] to rid their city of another [b]pest[/b] infesting it.
                        [/p]
                        [p]
                            You push your body to its very limits, sprinting away as fast as you can.
                            You can't hold out for much longer.
                            You can feel the end of your rope approaching.
                            You hope they're every bit as tired as you are.
                            That their bodies are as much on the verge of collapsing as your tiny, scaled frame is.
                        [/p]
                        [p]
                            Crossing into a more dignified area of Varanar, one of the few districts with cobblestone roads,
                            you feel the last of your energy leaving your body.
                            You try to keep running, you [b]really[/b] do, but your legs simply can no longer carry you.
                            As the flat dirt under your feet suddenly makes way for a more uneven, rock-like surface,
                            you lose your balance and plummet to the ground.
                            Time itself seems to slow down as your body falls limply onto the cold, hard street, completely exhausted.
                        [/p]
                        ${global.ui.condFail(`Insufficient stamina`)}
                        [p]
                            You roll onto your side right on time to witness one of your pursuers rapidly approaching.
                            Using the hefty momentum from his drunken running as a catalyst, he cruelly plants the tip of his heavy boot into your stomach,
                            kicking you [b]so[/b] hard your leathery body is lifted off the ground as a whole.
                            You scream in agony as you are hurled through the air, your panicked shout coming to an end as you fall down once more,
                            your frame heavily impacting with the stone street a second time, painfully skimming across it for a few more inches before finally coming to rest.
                        [/p]
                        [p]
                            Seriously hurt, you clutch your abdomen and slowly curl up into a ball.
                            You're pretty certain the [b]immense[/b] force behind the kick [b]broke[/b] something inside of you.
                            As you whimper and cry, the other two men that were chasing you catch up with their peer.
                            They advance upon you, and waste no time in resuming their brutal clobbering.
                        [/p]
                        [p]
                            One of them leans down, and mercilessly begins to [b]pound[/b] your side, battering your ribs with blow after blow of his powerful, clenched fist.
                            A second human begins to repeatedly kick your back with all of his might, your spine on the verge of [b]snapping[/b] in half.
                            The third and final abuser concerns himself with your face,
                            alternating between using his hands and his shoe-clad feet to batter every part of your visage from your forehead to your snout.
                        [/p]
                        [p]
                            You [b]beg[/b] for mercy, but the humans seem determined not to stop until you're [b]dead[/b].
                            Your eyes are beaten shut to the point where you can no longer [b]see[/b],
                            but you [b]hear[/b] someone walking by over the gut-wrenching sounds of your bones breaking and your scales bruising.
                            You cry out for help, hoping it's a kind-hearted guard, yet the clacking of spurred boots on stone don't cease, not even for an instant.
                            Help a kobold?
                            Why would he?
                        [/p]
                        [p]
                            For several minutes, all you know is agony.
                            Every square inch of your body not covered by your cowering arms is viciously attacked.
                            Your ribs are smashed, your nose is broken,
                            your back is [b]pelted[/b] with a barrage of blows and even the back of your head endures more than a few beatings.
                            As a final kick hits you square in the stomach, on the exact same spot as the very first one did, you are sent into a coughing fit,
                            hacking up bloody phlegms from the depths of your injured lungs.
                        [/p]
                        [p]
                            The three burly men are panting.
                            Exhausted, they wipe your own blood off onto your scales.
                            One of them cruelly spits in your face.
                            They're done, [b]for now[/b].
                        [/p]
                        [p]
                            Your abusers hurl various insults at you as they walk away, but you're too far [i]gone[/i] to really pay them much mind.
                            Teetering on the edge of consciousness, on the very brink of [b]death[/b],
                            you try to escape the crippling pain clouding your mind by focusing on your breathing,
                            but even the slight motion of your chest inflating and deflates sends spikes of [b]torment[/b] down your spine.
                            You've been hurt before, but never [b]this[/b] bad.
                            Blood oozes from your battered muzzle as you quietly sob away the immense anguish.
                        [/p]
                        ${global.ui.condFail(`Insufficient health and endurance`)}
                        [p]
                            About half an hour later, you manage to open one of your eyelids anew, the swelling slowly fading.
                            Into your blurry vision walks a man.
                            A big man.
                            A pig, wearing an off-white, completely bloodied apron.
                            A wicked grin is plastered across his plump face.
                            "[color=ppink]Fresh meat. Already tenderized.[/color]" he speaks in a bass voice.
                        [/p]
                        [p]
                            The pig leans down and takes hold of your tail.
                            The last thing you remember is being dragged across the street, leaving a trail of blood in your wake.
                            As the man halts before a large, wooden building with sharp metallic noises coming from inside, you finally slip into sweet unconsciousness,
                            never to wake up again.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    Pig end. Oink.
                                [/b]
                            [/c]
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" style="margin-left: auto; position: relative; float: right;" id="${continueId}">`;
        choice += `&lt;&lt Revert`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400, function () {
            var main = $("#main");
            main.scrollTop(main[0].scrollHeight - main.height() - 5); // Fix scrolling
        });
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.misc.unsetAll();
            ctrl.flee.undo();
            ctrl.flee.unset();
            ctrl.runin.unset();
            ctrl.stage.unset();
            for (var i = 0; i < idsToRemove.length; i++) {
                $("#" + idsToRemove[i]).remove();
            }
            loadNextPart(new NightTwoStage(ctrl));
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        registerContinueListener();
    };
}

// END scenes\Varanar\NightTwo\flee\run\fail\die\NightTwoRunDeath.js

// scenes\Varanar\NightTwo\flee\run\fail\live\NightTwoRunSurvivalContinue.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoRunSurvivalContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Half an hour later, you manage to open one of your eyelids anew, the swelling slowly fading.
                            Just in time to witness the sun slowly starting to rise above the rooftops.
                            The sunrise inspires you.
                            You steel your mind.
                            You aren't dead yet.
                            You can't stay here.
                            There's some things you need to [b]fix[/b] before you can get up, though.
                            This is going to hurt.
                        [/p]
                        [p]
                            Bringing both your hands to your snout, you forcibly [b]twist[/b] your broken nose-bones back into place,
                            a sickening [b]snap[/b] accompanying the horrifying [b]pain[/b] of two of your bones grinding against one another.
                            You let out a shrill shriek of suffering, although self-inflicted this time around.
                            Your cryiing and your breathing both intensify, but as the new pain begins to fade, your muzzle feels slightly [b]better[/b] off as a whole.
                            Like it was [b]bent out of place[/b] during the intense beatdown, and you [b]set it straight[/b] just now.
                            It is a [b]good[/b] pain.
                        [/p]
                        [p]
                            After wallowing in your own misery for a little while longer, you wearily roll onto your heavily bruised stomach,
                            swiftly raising yourself to your knees to avoid having to lie on top of fresh injuries.
                            Your legs were mostly spared during the beating.
                            Leaning back, resting your weight on top of your calves, you inspect the damage done to your body.
                            You are bruised all over, with several small, but lengthy cuts along your side and your chest.
                            You think one of your ribs is broken, but you aren't sure.
                        [/p]
                        [p]
                            Your stomach hurts significantly.
                            You think you may have severe internal injuries, but you're no doctor.
                            And even if you were, you doubt there's a cure for [i]ruptured stomach[/i].
                            But you're a kobold.
                            Your kind has the uncanny ability to live through even the most grievous of injuries and come back swinging.
                            And even if your injuries [b]won't[/b] heal, you simply [b]refuse[/b] to die [b]here[/b], in the middle of the street.
                        [/p]
                        [p]
                            No.
                            You're [b]not[/b] going to die.
                            That's what [b]they[/b] -- the humans that beat you -- would want.
                            You are [b]tough[/b].
                            You're going to [b]live[/b], if just because your very existance serves to [b]taunt[/b] the people who did this to you.
                        [/p]
                        [p]
                            Driven by [b]grit[/b] and [b]determination[/b], by the sheer [b]will[/b] to live, you crawl to your feet.
                            With slow, unsteady paces, you begin to wobble in the general direction of the town square.
                            Every step you take seems to invigorate you further, empowering your sheer [b]stubbornness[/b], your [b]refusal[/b] to let go of life itself.
                            While your wounds remain severe, your mind is stalwartly set on overcoming them.
                        [/p]
                        [p]
                            The twinkle of magic you felt at midnight has all but faded.
                            You now realize you are but a lone, weak,
                            small kobold in a big city filled with people who can and [b]will[/b] maim, hurt, kill and perhaps even [b]rape[/b] you.
                            You can't let your guard down, ever.
                            Next time you do, it might end up costing you your life, as it already nearly has.
                            You need to stay tough if you want to perservere in the big city.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.runSurvive.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.runSurvive.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\run\fail\live\NightTwoRunSurvivalContinue.js

// scenes\Varanar\NightTwo\flee\run\fail\live\NightTwoRunSurvival.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoRunSurvival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoRunSurvivalContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.nightTwo.beatUp = true;
        var toReturn = `[p]
                            Desperate to get away from the men, you burst into a sprint the moment you turn a corner,
                            hoping to be able to outrun them before they even [i]know[/i] you're escaping.
                            While your gambit does pay off in the form of a head start, sadly, you don't manage to shake the group from the get-go.
                            Noticing their lizard-prey running away,
                            the men drop their tight-packed group-cohesion and begin to outright [b]chase[/b] after you.
                        [/p]
                        [p]
                            You run from street to street, periodically checking behind you to make sure they haven't given up yet.
                            While far bigger than you, and probably able to outrun you whilst sober, their heavy inebriation hinders the belligerent group significantly.
                            A couple of pursuers tumble over their own feet, falling to the ground, being left behind by their peers.
                            A few of the men get too sick to keep pursuing you, retreating into alleyways to empty their bellies of food and drink.
                            Some of them are simply too weary to keep up pursuit.
                            Initially numbering high in the single-digits,
                            it takes only a few solid minutes of running for their [i]pack[/i] to be reduced to three hard-boiled men.
                        [/p]
                        [p]
                            But still, even [b]one[/b] of them can [b]cave your head in[/b] if they manage to catch you.
                            You aren't out of the woods until you can shake off [b]all three[/b] of them.
                            And you get the feeling that all this running is making them [b]even more frustrated[/b].
                            There's no doubt in your mind: your very life is at stake here.
                            And the humans chasing you are [b]determined[/b] to rid their city of another [b]pest[/b] infesting it.
                        [/p]
                        [p]
                            You push your body to its very limits, sprinting away as fast as you can.
                            You can't hold out for much longer.
                            You can feel the end of your rope approaching.
                            You hope they're every bit as tired as you are.
                            That their bodies are as much on the verge of collapsing as your tiny, scaled frame is.
                        [/p]
                        [p]
                            Crossing into a more dignified area of Varanar, one of the few districts with cobblestone roads,
                            you feel the last of your energy leaving your body.
                            You try to keep running, you [b]really[/b] do, but your legs simply can no longer carry you.
                            As the flat dirt under your feet suddenly makes way for a more uneven, rock-like surface,
                            you lose your balance and plummet to the ground.
                            Time itself seems to slow down as your body falls limply onto the cold, hard street, completely exhausted.
                        [/p]
                        ${global.ui.condFail(`Insufficient stamina`)}
                        [p]
                            You roll onto your side right on time to witness one of your pursuers rapidly approaching.
                            Using the hefty momentum from his drunken running as a catalyst, he cruelly plants the tip of his heavy boot into your stomach,
                            kicking you [b]so[/b] hard your leathery body is lifted off the ground as a whole.
                            You scream in agony as you are hurled through the air, your panicked shout coming to an end as you fall down once more,
                            your frame heavily impacting with the stone street a second time, painfully skimming across it for a few more inches before finally coming to rest.
                        [/p]
                        [p]
                            Seriously hurt, you clutch your abdomen and slowly curl up into a ball.
                            You're pretty certain the [b]immense[/b] force behind the kick [b]broke[/b] something inside of you.
                            As you whimper and cry, the other two men that were chasing you catch up with their peer.
                            They advance upon you, and waste no time in resuming their brutal clobbering.
                        [/p]
                        [p]
                            One of them leans down, and mercilessly begins to [b]pound[/b] your side, battering your ribs with blow after blow of his powerful, clenched fist.
                            A second human begins to repeatedly kick your back with all of his might, your spine on the verge of [b]snapping[/b] in half.
                            The third and final abuser concerns himself with your face,
                            alternating between using his hands and his shoe-clad feet to batter every part of your visage from your forehead to your snout.
                        [/p]
                        [p]
                            You [b]beg[/b] for mercy, but the humans seem determined not to stop until you're [b]dead[/b].
                            Your eyes are beaten shut to the point where you can no longer [b]see[/b],
                            but you [b]hear[/b] someone walking by over the gut-wrenching sounds of your bones breaking and your scales bruising.
                            You cry out for help, hoping it's a kind-hearted guard, yet the clacking of spurred boots on stone don't cease, not even for an instant.
                            Help a kobold?
                            Why would he?
                        [/p]
                        [p]
                            For several minutes, all you know is agony.
                            Every square inch of your body not covered by your cowering arms is viciously attacked.
                            Your ribs are smashed, your nose is broken,
                            your back is [b]pelted[/b] with a barrage of blows and even the back of your head endures more than a few beatings.
                            As a final kick hits you square in the stomach, on the exact same spot as the very first one did, you are sent into a coughing fit,
                            hacking up bloody phlegms from the depths of your injured lungs.
                        [/p]
                        [p]
                            The three burly men are panting.
                            Exhausted, they wipe your own blood off onto your scales.
                            One of them cruelly spits in your face.
                            They're done, [b]for now[/b].
                        [/p]
                        [p]
                            Your abusers hurl various insults at you as they walk away, but you're too far [i]gone[/i] to really pay them much mind.
                            Teetering on the edge of consciousness, on the very brink of [b]death[/b],
                            you try to escape the crippling pain clouding your mind by focusing on your breathing,
                            but even the slight motion of your chest inflating and deflates sends spikes of [b]torment[/b] down your spine.
                            You've been hurt before, but never [b]this[/b] bad.
                            Blood oozes from your battered muzzle as you quietly sob away the immense anguish.
                        [/p]
                        ${global.ui.condSucc(`Sufficient health and endurance`)}`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.runSurvive.continued() ? "disabled" : "";
        var isSelected = ctrl.runSurvive.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.runSurvive.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.runSurvive.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.runSurvive.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\flee\run\fail\live\NightTwoRunSurvival.js

// scenes\Varanar\NightTwo\flee\run\fail\NightTwoRunBeatdown.js

/**
 * Fleeing section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoRunBeatdown(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new NightTwoRunDeath(ctrl);
    var toSuccess = new NightTwoRunSurvival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        var health = ctrl.misc.health();
        var endurance = ctrl.misc.endurance();
        var damage = Math.floor(400 / endurance);
        return health > damage;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\run\fail\NightTwoRunBeatdown.js

// scenes\Varanar\NightTwo\flee\run\succeed\NightTwoRunGetawayContinue.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoRunGetawayContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You've overworked yourself once more.
                            Your lungs and your legs both ache [b]so[/b] much it feels like they're [b]on fire[/b].
                            You take a series of deep breaths in an attempt to calm yourself.
                            It takes effort and willpower [b]not[/b] to faint, but you've been through worse.
                            Those humans didn't kill you, so neither will some bodily strain.
                        [/p]
                        [p]
                            After a couple of minutes spent recovering, you begin to get your bearings.
                            You think you recognize this area.
                            You [i]explored[/i] it before the [i]unfortunate run-in[/i] with those [b]ruffians[/b].
                            Slowly, your unsteady legs begin to carry you in the direction of the city's market square once more.
                        [/p]
                        [p]
                            The further you walk, the more you start to regain your composure.
                            Luckily, you're not [b]too[/b] far away from the plaza where your urban adventure began.
                            As you begin to near it, you notice the sun rising above the rooftops.
                            You must have lost track of time while exploring the city.
                            Or maybe that frantic chase took longer than you can remember.
                            Either way, you guess you won't be sleeping tonight.
                        [/p]
                        [p]
                            The twinkle of magic you felt at midnight has all but faded, giving way to a sense of [b]vulnerability[/b].
                            You are but a lone, weak, small, kobold in a big city filled with people who can and [b]want[/b] to maim,
                            hurt, kill and perhaps even [b]rape[/b] you.
                            If you let your guard down for even one moment, it could cost you your money, your life or both.
                            And next time, the men chasing you might [i]not[/i] be drunk off their asses.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.runGetaway.continued()) {
            return "";
        }

        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.endChange.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.runGetaway.continued()) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\run\succeed\NightTwoRunGetawayContinue.js

// scenes\Varanar\NightTwo\flee\run\succeed\NightTwoRunGetaway.js

/**
 * Flee section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoRunGetaway(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoRunGetawayContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Desperate to get away from the men, you burst into a sprint the moment you turn a corner,
                            hoping to be able to outrun them before they even [i]know[/i] you're escaping.
                            While your gambit does pay off in the form of a head start, sadly, you don't manage to shake the group from the get-go.
                            Noticing their lizard-prey running away,
                            the men drop their tight-packed group-cohesion and begin to outright [b]chase[/b] after you.
                        [/p]
                        [p]
                            You run from street to street, periodically checking behind you to make sure they haven't given up yet.
                            While far bigger than you, and probably able to outrun you whilst sober, their heavy inebriation hinders the belligerent group significantly.
                            A couple of pursuers tumble over their own feet, falling to the ground, being left behind by their peers.
                            A few of the men get too sick to keep pursuing you, retreating into alleyways to empty their bellies of food and drink.
                            Some of them are simply too weary to keep up pursuit.
                            Initially numbering high in the single-digits,
                            it takes only a few solid minutes of running for their [i]pack[/i] to be reduced to three hard-boiled men.
                        [/p]
                        [p]
                            But still, even just [b]one[/b] of them can [b]cave your head in[/b] if they manage to catch you.
                            You aren't out of the woods until you can shake off [b]all three[/b] of them.
                            And you get the feeling that all this running is making them [b]even more frustrated[/b].
                            There's no doubt in your mind: your very life is at stake here.
                            And the humans chasing you are [b]determined[/b] to rid their city of another [b]pest[/b] infesting it.
                        [/p]
                        [p]
                            You push your body to its very limits, sprinting away as fast as you can.
                            Your legs are rapidly growing weary, but so are theirs, you hope.
                            Crossing into a more dignified area of Varanar, one of the few districts with cobblestone roads,
                            you almost lose your balance as the flat dirt under your feet suddenly makes way for a more uneven, rock-like surface.
                            [i]Almost[/i].
                        [/p]
                        ${global.ui.condSucc(`Sufficient stamina`)}
                        [p]
                            Luckily, it seems like the men chasing you, too, have difficulty adapting to the new enviroment.
                            Looking behind you, you can see one of them taking a particularly [b]nasty[/b] dive, his forehead [b]slamming[/b] into the rocky ground.
                            You're pretty sure you can hear something [b]crack[/b], although you don't know whether it's his skull or the street he fell atop.
                            Either way, the man's [i]friends[/i] cease their pursuit in order to tend to their fallen comrade, who seems [b]seriously[/b] hurt.
                            Serves him right for chasing you!
                        [/p]
                        [p]
                            While no longer being tailed, you maintain a hasty pace,
                            crossing a few more streets on sheer adrenaline alone to ensure the hoodlums won't catch you.
                            Not tonight, at least.
                            Once you're [b]completely[/b] certain you've lost them, you slowly come to a halt,
                            leaning against a nearby building to keep your overexerted body upright.
                            You catch your breath, your loud panting resonating across the otherwise empty street.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.runGetaway.continued() ? "disabled" : "";
        var isSelected = ctrl.runGetaway.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.runGetaway.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.runGetaway.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.runGetaway.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\flee\run\succeed\NightTwoRunGetaway.js

// scenes\Varanar\NightTwo\flee\run\NightTwoKeepRunning.js

/**
 * Fleeing section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoKeepRunning(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var toFail = new NightTwoRunBeatdown(ctrl);
    var toSuccess = new NightTwoRunGetaway(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.stamina() > 70;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.flee.get() !== NightTwoFleeChoice.RUN) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.flee.get() !== NightTwoFleeChoice.RUN) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\NightTwo\flee\run\NightTwoKeepRunning.js

// scenes\Varanar\NightTwo\flee\NightTwoFlee.js

/**
 * Fleeing section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoFlee(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var runId = arguments.callee.name + "ButtonOne";
    var hideId = arguments.callee.name + "ButtonTwo";

    var toRun = new NightTwoKeepRunning(ctrl);
    var toHide = new NightTwoAlley(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You wisely decide to head the [i]other[/i] way, which unfortunately leads you back into the [i]brothel district[/i].
                            You aren't sure yet whether they're actually [i]following[/i] you or not, but the men seem to be heading this way as well.
                            They're getting more [b]rowdy[/b] by the minute.
                            You vaguely overhear them cracking rude jokes about...[i]kobolds[/i].
                            To be more precise: they're joking [b]about you[/b].
                            And the things they're saying are far from being nice.
                            While jolly at first, the mood is growing more grim with every passing second.
                        [/p]
                        [p]
                            To see whether you're just a victim of bad luck, or if the group is actively [i]chasing[/i] you,
                            you take a few random turns to nowhere in particular.
                            Still, the men remain on your tail.
                            Like they have some unsettled beef with you, or maybe with your kind as a whole.
                            You pick up your walking pace.
                            So do they.
                            You start jogging.
                            They resort to drunkenly and menacingly speedwalking,
                            their size allowing them to keep up with your relatively higher pace without needing to [i]run[/i].
                        [/p]
                        [p]
                            Your mind is racing a million miles an hour.
                            You feel anxious.
                            Like a rat trapped in a maze, pursued by bunch of hungry cats.
                            The men seem to be gaining on you, despite your best efforts.
                            If you don't do something quick, they'll catch you.
                            And who knows what a gang of angry, drunken humans'll do once they get their hands on a dirty kobold thief like yourself?
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildRunChoice() {
        var disabled = ctrl.flee.has() ? "disabled" : "";
        var isSelected = ctrl.flee.get() === NightTwoFleeChoice.RUN;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${runId}" ${disabled}>Run</button>`;

        var tooltip = global.parser.parse(`Run away as fast as you can. Hope they won't be able to keep up.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildHideChoice() {
        var disabled = ctrl.flee.has() ? "disabled" : "";
        var isSelected = ctrl.flee.get() === NightTwoFleeChoice.HIDE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${hideId}" ${disabled}>Hide</button>`;

        var tooltip = global.parser.parse(`Dart into a nearby alleyway and pray they don't find you.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildRunChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildHideChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + runId), $("#" + hideId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerRunListener() {
        $("#" + runId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.flee.set(NightTwoFleeChoice.RUN);
            loadNextPart(toRun);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerHideListener() {
        $("#" + hideId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.flee.set(NightTwoFleeChoice.HIDE);
            loadNextPart(toHide);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (ctrl.runin.get() !== NightTwoRunInChoice.FLEE) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.flee.change();
        part += buildChoice();
        part += `</div>`;
        part += toRun.build();
        part += toHide.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (ctrl.runin.get() !== NightTwoRunInChoice.FLEE) {
            return;
        }

        toRun.registerListeners();
        toHide.registerListeners();

        if (ctrl.flee.has()) {
            return;
        }

        registerRunListener();
        registerHideListener();
    };
}

var NightTwoFleeChoice = {
    RUN: 1,
    HIDE: 2
};

// END scenes\Varanar\NightTwo\flee\NightTwoFlee.js

// scenes\Varanar\NightTwo\arrival\NightTwoRunIn.js

/**
 * Arrival section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoRunIn(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var ignoreId = arguments.callee.name + "ButtonOne";
    var fleeId = arguments.callee.name + "ButtonTwo";

    var toIgnore = new NightTwoConfront(ctrl);
    var toFlee = new NightTwoFlee(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            After resting your feet for a while, you decide it's best to get a lay of the land before catching some shut-eye.
                            The night is still young, and you know pretty much nothing of the cityscape outside of the grand avenue.
                            Actually, you [i]do[/i] know how to make your way to the hideout of the Three-Tailed Rat Clan,
                            nested deep within the kobold district of Varanar,
                            but only because you [i]memorized[/i] the route between there and the eastern city gate.
                            Hoping to get a clearer picture of what road leads where, you pick a street at random,
                            and begin your journey of exploration throughout the city.
                        [/p]
                        [p]
                            You wander through the mercantile district and the residential area,
                            you pause briefly to inspect the magnificent architecture of the local church,
                            you walk alongside the river that splits the capital in two,
                            and you wisely avoid the kobold district at this time of night.
                            Less than two hours after you left the town square,
                            you manage to form a pretty decent idea on how to get around Varanar.
                        [/p]
                        [p]
                            Still, there's plenty more avenues left to explore, and you don't feel like going to bed quite yet,
                            too enthralled by all the sights to see to stop now.
                            Ultimately, after another few hours of discovery and mental mapmaking, you find yourself in a less [i]savory[/i] part of town.
                            You aren't sure whether you should nickname this area the [i]brothel district[/i] or the [i]tavern zone[/i].
                            Houses in this neighbourhood are as poorly built and maintained as they are in the kobold quarter,
                            although they are, admittedly, a lot [i]bigger[/i].
                        [/p]
                        [p]
                            While most of the city's streets are well-patrolled and well-lit, this section seems mostly exempt from sentries and guards.
                            Or maybe it's just getting so late that most of them have already gone to sleep.
                            You, too, are starting to get sleepy.
                            You begin to make your way back to the city square in hopes of finding an inn,
                            a bale of hay or at least an alleyway that [i]doesn't[/i] smell like someone died in it.
                        [/p]
                        [p]
                            As you turn a corner, you notice a small group of burly human males headed your way.
                            They seem drunk and unsavory, probably returning home after a night full of beer and wenches.
                            One of them crudely points his finger your way while loudly chattering to his friends.
                            You can't quite understand them, but whatever he said must have been funny.
                            The group bursts out into laughter.
                            Their eyes become fixated on you.
                            Somehow, you don't feel very safe anymore.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildIgnoreChoice() {
        var disabled = ctrl.runin.has() ? "disabled" : "";
        var isSelected = ctrl.runin.get() === NightTwoRunInChoice.IGNORE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${ignoreId}" ${disabled}>Ignore them</button>`;

        var tooltip = global.parser.parse(`Walk past the group. Maybe they'll leave you alone. `);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 150px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFleeChoice() {
        var disabled = ctrl.runin.has() ? "disabled" : "";
        var isSelected = ctrl.runin.get() === NightTwoRunInChoice.FLEE;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fleeId}" ${disabled}>Flee</button>`;

        var tooltip = global.parser.parse(`Walk the other way. Keep a steady pace to stay ahead of them.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You: </div>`;
        choice += buildIgnoreChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFleeChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + ignoreId), $("#" + fleeId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerIgnoreListener() {
        $("#" + ignoreId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.runin.set(NightTwoRunInChoice.IGNORE);
            loadNextPart(toIgnore);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFleeListener() {
        $("#" + fleeId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.runin.set(NightTwoRunInChoice.FLEE);
            loadNextPart(toFlee);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.stage.continued()) {
            return "";
        }

        ctrl.bgm.loop();
        global.kobold.inventory.exitSafeZone();
        global.ui.imageBackground.hide();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += toIgnore.build();
        part += toFlee.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.stage.continued()) {
            return;
        }

        toIgnore.registerListeners();
        toFlee.registerListeners();

        if (ctrl.runin.has()) {
            return;
        }

        registerIgnoreListener();
        registerFleeListener();
    };
}

var NightTwoRunInChoice = {
    IGNORE: 1,
    FLEE: 2
};

// END scenes\Varanar\NightTwo\arrival\NightTwoRunIn.js

// scenes\Varanar\NightTwo\arrival\NightTwoStage.js

/**
 * Arrival section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoStage(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoRunIn(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            After passing the gate, you are greeted with the sight of Varanar's grand avenue,
                            a single road leading all the way from the eastern entrance to the massive town square.
                            Last time you were in town, you tried pickpocketing people on this very path during the bustling morning rush.
                            Now, the streets are nigh-empty, save for a few patrolling guardsmen.
                        [/p]
                        [p]
                            You idlemindedly walk down the road towards the market square.
                            The tall, human-made buildings on either side almost make you feel a little [i]claustrophobic[/i],
                            despite the street being wide enough to fit several carts side-by-side.
                            To you, this entire city seems like a bit of a [i]miracle[/i].
                            You can't even [b]begin[/b] to imagine the [b]huge[/b] amount of work it must have taken to build all of this.
                            While other races are banding together to construct vast cities,
                            your kind can barely stand one another for long enough to rob people together.
                        [/p]
                        [p]
                            After a brief hike, you arrive at the town square.
                            While the plaza is empty at the moment, the bi-weekly market has left a clear impression on the land.
                            You wouldn't call it completely riddled with garbage, but there is more than a bit of litter remaining.
                            The hard, near-frozen ground is completely [b]trampled[/b], a multitude of holes denoting the former position of the stalls.
                        [/p]
                        [p]
                            Suppressing your natural urge to sift through the trash, you walk up to the small stage on the far-end of the square.
                            This is where important decrees are shared, where traitors are hanged,
                            where stockades are set and where prestigious travelling shows entertain thousands of people.
                            But right now, for you, it is nothing but a place to rest your weary feet at.
                            You crawl up onto the wooden surface, and sit yourself down on the edge.
                        [/p]
                        [p]
                            The church bells ring midnight.
                            You take in a deep breath of cool evening air, staring upwards at the night sky.
                            Stars twinkle back at your inquisitive gaze, mirroring the sparkle in your eyes.
                            While not full moon yet, it will be in a couple of days.
                            There's something magical about this city.
                            It makes you feel [i]safe[/i].
                            Moreso than the forest, at least.
                        [/p]
                        [p]
                            [sneak]
                                You are in a safe zone.
                                You can use items from your inventory here.
                                It may be wise to prepare yourself for the trials up ahead,
                                lest you be found wanting.
                            [/sneak]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.stage.continued() ? "disabled" : "";
        var isSelected = ctrl.stage.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.stage.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (!ctrl.arrival.continued()) {
            return "";
        }

        ctrl.bgm.stopLoop();
        global.ui.applyTheme("varanar_night"); // Fix theme when reverting
        global.ui.imageBackground.show("scenes/Varanar/NightTwo/arrival/stars.gif");
        global.kobold.inventory.enterSafeZone();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (!ctrl.arrival.continued()) {
            return;
        }

        next.registerListeners();

        if (ctrl.stage.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\arrival\NightTwoStage.js

// scenes\Varanar\NightTwo\arrival\NightTwoArrival.js

/**
 * Arrival section of Varanar, Night Two.
 * @param {NightTwoController} controller The Night Two controller.
 */
function NightTwoArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new NightTwoStage(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[st]
                            Arrival
                        [/st]
                        [p]
                            After a boring, uneventful trek, you finally arrive at the great walls of Varanar.
                            Home to thousands, if not tens of thousands, it is the single largest city outside of the neighbouring Dragon Empire.
                            The capital of humans, one of the most prominent trading hubs in the entire world, and a melting pot for all walks of life,
                            from rich and snobby noblemen, to humble, rundown [b]thieves[/b] like yourself.
                        [/p]
                        [p]
                            The sun may have gone down, but in the darkness,
                            the myriad of flickering torchlights illuminating the battlements high up above showcase how [b]incredibly[/b] well protected the city is.
                            Nothing short of a determined and well-prepared besieging [b]army[/b] could ever hope to take the walls of Varanar.
                            While you've been here before, you've never seen them at night.
                            You always assumed the battlements were [i]empty[/i], that they'd be manned solely during times of war.
                            These fortifications were impressive and imposing enough [i]without[/i] realizing that they are constantly staffed,
                            a dozen archers and crossbowmen probably betting on whether or not they can hit that tiny little kobold 
                            approaching the gates at this very instant.
                        [/p]
                        [p]
                            While kobolds are generally [i]tolerated[/i] in public, there are no official laws protecting your kind.
                            You could be shot, stabbed or keelhauled in front of a crowd, and nobody would bat an eye.
                            The thing that frightens you the most is the fact that people can do [b]far worse[/b] things than simply hurting you, all according to law.
                            The guards rarely intervene, and when they do, you can be [b]darn[/b] sure that they [b]won't[/b] be picking [i]your[/i] side.
                        [/p]
                        [p]
                            But those are worries for later.
                            Right now, you focus on actually getting [b]in[/b] to the city proper.
                            The eastern gate, in front of which you stand, is currently clogged to the brim with people [b]leaving[/b] Varanar.
                            Must have been a long market day or something.
                            Leaning against a nearby road-fence, you decide to wait until the worst of the crowd has cleared.
                        [/p]
                        [p]
                            You had hoped you would only have to wait for a couple of minutes, but it takes almost an entire [b]hour[/b] for the huge mass of daytrippers,
                            marketeers, farmers, supply-running forest bandits and foreign trading caravans to siphon from the great city.
                            You pass the time by observing the huge variety of lifeforms on display.
                            Humans, horses, rodents, birds, cats both big and small, bears, reptiles and all sorts of canines walk on by.
                            Some carry wheelbarrows, others ride carts pulled by beasts of burden and a few are even rich enough to own their own chariots.
                        [/p]
                        [p]
                            Outside of a single kobold asking for directions, nobody really pays you much mind.
                            As the crowd slowly disperses in all directions of the land, you hop off of the fence, and proceed towards the gate.
                            While passing through the wall-opening, you instinctively gaze upwards at the portcullis,
                            ready to jump out of the way should it budge even an inch.
                            Luckily, this time at least, you are spared the wrath of the suspended metal.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.arrival.continued() ? "disabled" : "";
        var isSelected = ctrl.arrival.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.arrival.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        global.ui.applyTheme("varanar_night");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.arrival.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.arrival.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\NightTwo\arrival\NightTwoArrival.js

// scenes\Varanar\NightTwo\NightTwoController.js

/**
 * Controller for the Varanar Night Two scene.
 * @param {string} sceneId The id of the scene.
 */
function NightTwoController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.nextSceneReady = true;
    this.nextSceneName = "MorningThree";
    this.nextFolderName = "Varanar/DayThree";

    this.misc = new MiscController(id);

    this.arrival = new ChangeController(id, "arrival", function () {
        global.kobold.status.mental.increase(15, "Amazed by Varanar");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Amazement", "Sanity", +15)
        ]);
    });

    this.stage = new ContinueController(id, "stage");
    this.runin = new ChoiceController(id, "runin");

    this.flee = new UndoController(id, "flee", function change() {
        var sanityBacklog = new ChoiceController(id, "fleeSanityBL"); // Todo: Make framework for this
        var staminaBacklog = new ChoiceController(id, "fleeStaminaBL");
        sanityBacklog.set(global.kobold.status.mental.lasthit);
        staminaBacklog.set(global.kobold.status.stamina.lasthit);
        global.kobold.status.mental.decrease(5, "Being chased");
        global.kobold.status.stamina.decrease(5, "Light jogging");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Being pursued", "Sanity", -5),
            new Change("Fearful jogging", "Stamina", -5)
        ]);
    }, function undoChange() {
        var sanityBacklog = new ChoiceController(id, "fleeSanityBL"); // Todo: Make framework for this
        var staminaBacklog = new ChoiceController(id, "fleeStaminaBL");
        global.kobold.status.mental.increase(5, sanityBacklog.get());
        global.kobold.status.stamina.increase(5, staminaBacklog.get());
        sanityBacklog.unset();
        staminaBacklog.unset();
    });

    this.alley = new ChoiceController(id, "alley");
    this.hide = new ChoiceController(id, "hide");

    this.hideStay = new ChangeController(id, "hideStay", function change() {
        global.kobold.status.physical.decrease(5, "Breathing in miasma");
        global.kobold.status.mental.decrease(25, "Got pissed on");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Miasma", "Health", -5),
            new Change("Pissed on", "Sanity", -25)
        ]);
    });

    this.hideAttack = new ChangeController(id, "hideAttack", function change() {
        global.kobold.status.mental.increase(10, "Righteous justice");
        global.kobold.status.mental.decrease(25, "Manual neutering");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Righteous justice", "Sanity", +10),
            new Change("Manual neutering", "Sanity", -25)
        ]);
    });

    this.alleyRunSuccess = new ChangeController(id, "alleyRunSuccess", function change() {
        global.kobold.status.physical.increase(5, "Good workout");
        global.kobold.status.mental.increase(10, "Showcased skill");
        global.kobold.status.stamina.decrease(10, "Parkouring");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Good workout", "Health", +5),
            new Change("Skillful escape", "Sanity", +10),
            new Change("Parkouring", "Stamina", -10)
        ]);
    });

    this.alleyRunFail = new UndoController(id, "alleyRunFail", function change() {
        var healthBacklog = new ChoiceController(id, "alleyRunFailHealthBL");
        var sanityBacklog = new ChoiceController(id, "alleyRunFailSanityBL");
        var staminaBacklog = new ChoiceController(id, "alleyRunFailStaminaBL");
        healthBacklog.set(global.kobold.status.physical.lasthit);
        sanityBacklog.set(global.kobold.status.mental.lasthit);
        staminaBacklog.set(global.kobold.status.stamina.lasthit);
        global.kobold.status.physical.decrease(5, "Fell down");
        global.kobold.status.mental.decrease(10, "Plan failed");
        global.kobold.status.stamina.decrease(5, "Attempted climbing");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Fell down", "Health", -5),
            new Change("Plan failed", "Sanity", -5),
            new Change("Cornered", "Sanity", -5),
            new Change("Attempted parkouring", "Stamina", -5)
        ]);
    }, function undo() {
        var healthBacklog = new ChoiceController(id, "alleyRunFailHealthBL");
        var sanityBacklog = new ChoiceController(id, "alleyRunFailSanityBL"); // Todo: Make framework for this
        var staminaBacklog = new ChoiceController(id, "alleyRunFailStaminaBL");
        global.kobold.status.physical.increase(5, healthBacklog);
        global.kobold.status.mental.increase(10, sanityBacklog.get());
        global.kobold.status.stamina.increase(5, staminaBacklog.get());
        healthBacklog.unset();
        sanityBacklog.unset();
        staminaBacklog.unset();
    });

    this.alleyEager = new ChangeController(id, "alleyEager", function change() {

    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("TODO", "CHANGE", +1)
        ]);
    });

    this.alleyStay = new ChangeController(id, "alleyStay", function change() {
        if (global.kobold.inventory.equipment.shirt) {
            global.kobold.inventory.equipment.shirt.stain("Piss");
        }
        if (global.kobold.inventory.equipment.pants) {
            global.kobold.inventory.equipment.pants.stain("Urine");
        }
        if (global.kobold.inventory.equipment.underwear) {
            global.kobold.inventory.equipment.underwear.stain("Pee");
        }
        global.kobold.status.physical.decrease(5, "Unsanitary actions");
        global.kobold.status.mental.decrease(25, "Got pissed on");
        global.kobold.stats.decreaseStat(StatsEnum.CHARISMA, 1);
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Unsanitary", "Health", -5),
            new Change("Unwilling watersports", "Sanity", -25),
            new Change("Lasting fragrance", "Charisma", -1)
        ]);
    });

    this.runGetaway = new ChangeController(id, "runGetaway", function change() {
        global.kobold.status.physical.decrease(10, "Overexertion");
        global.kobold.status.mental.increase(10, "Escaped certain death");
        global.kobold.status.stamina.decrease(70, "Running for your life");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Overexertion", "Health", -10),
            new Change("Escaped certain death", "Sanity", +10),
            new Change("Out of breath", "Stamina", -70)
        ]);
    });

    this.runSurvive = new ChangeController(id, "runSurvive", function change() {
        var stamina = that.misc.stamina();
        var endurance = that.misc.endurance();
        var damage = Math.floor(400 / endurance);
        global.kobold.status.physical.decrease(damage, "Beat up by humans");
        global.kobold.status.mental.decrease(20, "Victim complex");
        global.kobold.status.stamina.decrease(stamina, "Ran for your life");
    }, function changeNote() {
        var stamina = that.misc.stamina();
        var endurance = that.misc.endurance();
        var damage = Math.floor(400 / endurance);
        return global.ui.change.buildNotification([
            new Change("Beatdown", "Health", -damage),
            new Change("Victim complex", "Sanity", -20),
            new Change("Gave it your all", "Stamina", -stamina)
        ]);
    });

    this.lewdCrude = new UndoController(id, "lewdCrude", function change() {
        var sanityBacklog = new ChoiceController(id, "lewdCrudeSanityBL");
        sanityBacklog.set(global.kobold.status.mental.lasthit);
        global.kobold.status.mental.decrease(5, "Growing concerned");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Insults", "Sanity", -2),
            new Change("Concern", "Sanity", -3)
        ]);
    }, function undo() {
        var sanityBacklog = new ChoiceController(id, "lewdCrudeSanityBL");
        global.kobold.status.mental.increase(5, sanityBacklog.get());
        sanityBacklog.unset();
    });

    this.rude = new UndoController(id, "rude", function change() {
        var healthBacklog = new ChoiceController(id, "rudeHealthBL");
        var sanityBacklog = new ChoiceController(id, "rudeSanityBL");
        healthBacklog.set(global.kobold.status.physical.lasthit);
        sanityBacklog.set(global.kobold.status.mental.lasthit);
        global.kobold.status.physical.decrease(5, "Arm twisted");
        global.kobold.status.mental.decrease(10, "Pain");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Arm twisted", "Health", -5),
            new Change("Panic", "Sanity", -5),
            new Change("Pain", "Sanity", -5)
        ]);
    }, function undo() {
        var healthBacklog = new ChoiceController(id, "rudeHealthBL");
        var sanityBacklog = new ChoiceController(id, "rudeSanityBL");
        global.kobold.status.mental.increase(5, healthBacklog.get());
        global.kobold.status.mental.increase(10, sanityBacklog.get());
        healthBacklog.unset();
        sanityBacklog.unset();
    });

    this.appease = new ChoiceController(id, "appease");

    this.crudeEager = new ChangeController(id, "crudeEager", function change() {

    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("TODO", "Change", +1)
        ]);
    });

    this.crudeReluctant = new ChangeController(id, "crudeReluctant", function change() {
        global.kobold.status.physical.decrease(2, "Unsanitary kiss");
        global.kobold.status.mental.decrease(20, "Disgust and humiliation");
        global.kobold.status.exertion.increase(5, "Disgust");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Unsanitary kiss", "Health", -2),
            new Change("Disgusted", "Sanity", -10),
            new Change("Humiliated", "Sanity", -10),
            new Change("Disgusted", "Sexual exhaustion", +5, true)
        ]);
    });

    this.crudeUnfit = new ChangeController(id, "crudeUnfit", function change() {
        global.kobold.status.mental.increase(15, "Escaped through wits");
        global.kobold.stats.increaseStat(StatsEnum.INTELLIGENCE, 1);
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Out-smarted humans", "Sanity", +15),
            new Change("Out-smarted humans", "Intelligence", +1)
        ]);
    });

    this.crudeStruggle = new ChangeController(id, "crudeStruggle", function change() {
        global.kobold.status.physical.decrease(10, "Pain and unhygienic behavior");
        global.kobold.status.mental.decrease(30, "Disgust and humiliation");
        global.kobold.status.exertion.increase(10, "Disgust");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Pain", "Health", -5),
            new Change("Unsanitary", "Health", -5),
            new Change("Forced", "Sanity", -10),
            new Change("Disgusted", "Sanity", -10),
            new Change("Humiliated", "Sanity", -10),
            new Change("Disgusted", "Sexual exhaustion", +10, true)
        ]);
    });

    this.lewdEager = new ChangeController(id, "lewdEager", function change() {

    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("TODO", "Change", +1)
        ]);
    });

    this.lewdReluctant = new ChangeController(id, "lewdReluctant", function change() {
        global.kobold.status.physical.decrease(10, "Filth ingestion");
        global.kobold.status.mental.decrease(20, "Disgust and humiliation");
        global.kobold.status.exertion.increase(10, "Disgust");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Ingested filth", "Health", -10),
            new Change("Disgusted", "Sanity", -10),
            new Change("Humiliated", "Sanity", -10),
            new Change("Disgusted", "Sexual exhaustion", +10, true)
        ]);
    });

    this.lewdUnfit = new ChangeController(id, "lewdUnfit", function change() {
        global.kobold.status.mental.increase(15, "Escaped through wits");
        global.kobold.stats.increaseStat(StatsEnum.INTELLIGENCE, 1);
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Out-smarted humans", "Sanity", +15),
            new Change("Out-smarted humans", "Intelligence", +1)
        ]);
    });

    this.lewdStruggle = new ChangeController(id, "lewdStruggle", function change() {
        global.kobold.status.physical.decrease(15, "Pain and unhygienic behavior");
        global.kobold.status.mental.decrease(30, "Disgust and humiliation");
        global.kobold.status.exertion.increase(10, "Disgust");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Pain", "Health", -5),
            new Change("Ingested filth", "Health", -10),
            new Change("Forced", "Sanity", -10),
            new Change("Disgusted", "Sanity", -10),
            new Change("Humiliated", "Sanity", -10),
            new Change("Disgusted", "Sexual exhaustion", +10, true)
        ]);
    });

    this.fightWin = new ChangeController(id, "fightWin", function change() {
        global.kobold.status.mental.increase(10, "Fighting humans");
        global.kobold.status.stamina.decrease(10, "Fighting humans");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Victory", "Sanity", +15),
            new Change("Hurt people", "Sanity", -10),
            new Change("They deserved it", "Sanity", +5),
            new Change("Fight", "Stamina", -10)
        ]);
    });

    this.endChange = new ChangeController(id, "endChange", function change() {
        global.kobold.status.mental.decrease(5, "Lack of sleep");
        global.kobold.status.stamina.decrease(10, "Lack of sleep");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Lack of sleep", "Sanity", -5),
            new Change("Lack of sleep", "Stamina", -10)
        ]);
    });
    
    this.bgm = new AudioController(id,"res/audio/varanar/desolated_field.mp3", 60);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

}

// END scenes\Varanar\NightTwo\NightTwoController.js

// scenes\Varanar\NightTwo\NightTwo.js

/**
 * First night in Varanar.
 */
function NightTwo() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "NightTwo";
    var sceneName = "Varanar, Night Two";
    var sceneSubfolder = "Varanar";
    var ctrl = new NightTwoController(sceneId);
    var arrival = new NightTwoArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]Varanar[/t]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new NightTwo();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\Varanar\NightTwo\NightTwo.js



// END SCENE NightTwo



// SCENE MorningThree

// scenes\Varanar\DayThree\MorningThree\decline\MorningThreeDeclineSex.js

/**
 * Decline section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeDeclineSex(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreePropositionChoice.DECLINE;
    var reqController = ctrl.proposition;

    var continueId = arguments.callee.name + "ButtonOne"

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.morning.declinedSex = true;
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var toReturn = `[p]
                            You say you're flattered, but you're really not [i]in the mood[/i] right now.
                            The cyan kobold's visage instantly shifts from euphoric and devious to sad and disappointed.
                            ${She} lets go of your hand, and gloomily looks away, like ${he} can't even bring ${her}self to look you in the eyes.
                            "[color=cyan]I-I'm sorry. I got a little carried away. I think it'd be best if I just...[i]leave[/i].[/color]"
                            Before you can stop ${her}, the friendliest person you've met in Varanar thus far disappears into the crowd of people.
                        [/p]
                        [p]
                            You really didn't mean to offend ${her} or anything.
                            You just didn't want to have sex with ${her}.
                            Surely ${he} was just overreacting, right?
                            You shake your head to clear your thoughts.
                            With your new friend missing, you decide to stick around and see what happens to the bird on stage.
                            The crowd's relentless hurling of spoilt food continues for the better half of an hour.
                            Then, with the mob running out of things to throw at him, one of the guardsmen, a heavily battle-scarred feline,
                            approaches the avian from behind.
                        [/p]
                        [p]
                            With a single [b]swipe[/b] of the cat's clawed hand, the falcon's pants are ripped to shreds.
                            Kicking apart the bird's legs, the guardsman lowers his own trousers.
                            You watch on in horror, out of sheer morbid curiosity, as the bird begins to [b]beg[/b] for mercy,
                            while the feline lines up his barbed member with the criminal's unlubed feathery butthole.
                            Then, under the exuberant cheers of the crowd, he thrusts forward,
                            the avian squawking in pain and dishonor as his last bit of diginity is publically torn away from him.
                        [/p]
                        [p]
                            While the cat begins to violently hump away at the prisoner's rear, the very pillory itself rocking back and forth with every thrust,
                            you notice a small line starting to form near the steps leading up to the stage.
                            People are paying the guards, for what you can only assume to be their own turn with the bound falcon's ass.
                        [/p]
                        [p]
                            You aren't sure how to feel about this.
                            Turning prisoners into unwilling prostitutes seems morally [b]wrong[/b].
                            But then again, perhaps the bird is a rapist himself.
                            Maybe they're just doing to him what he did to someone else.
                            Either way, you now begin to realize why the guards of Varanar are widely praised as being highly effective at what they do.
                            You wouldn't want to get caught stealing, knowing something like [b]this[/b] might happen to you.
                        [/p]
                        [p]
                            You stick around for a little bit longer, to confirm your suspicions.
                            Indeed, as the feline finishes up and steps aside, the next person in line gets up on stage.
                            A big, burly horse.
                            There won't be much of the falcon left for the rest to enjoy after [i]that guy[/i] is done with him.
                            You turn around and walk away as the equine lowers his pants,
                            not interested in watching half the town have their way with the poor guy who got caught.
                            After all, there [i]are[/i] other things to do on the market square, apart from fucking captive birds.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.declineSex.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\decline\MorningThreeDeclineSex.js

// scenes\Varanar\DayThree\MorningThree\decline\MorningThreeDeclineEggs.js

/**
 * Decline section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeDeclineEggs(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeEggsChoice.DECLINE;
    var reqController = ctrl.eggs;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var toReturn = `[p]
                            While you thank the cyan kobold for ${his} offer,
                            you don't think you're going to take ${her} up on it.
                            Throwing things at a man, without even knowing what he did to deserve this,
                            if anything at all, doesn't seem like a very [i]knightly[/i] thing to do.
                            Did you mention you want to be a knight one day?
                        [/p]
                        [p]
                            The other kobold stares at you with a troubled, doubting look on ${his} face,
                            as though ${he}'s lost all of the respect ${he} had for you at first.
                            "[color=cyan]You could have just said that you didn't want to throw the eggs.[/color]"
                            ${he} replies, ${his} tone a lot less friendly than before.
                            Snagging the eggs from your claws once more, the cyan kobold walks away,
                            disappearing into the crowd after casting one last unpleasant glare your way.
                        [/p]
                        [p]
                            Well, that certainly could have gone better.
                            You didn't think ${he}'d take offence to you refusing to partake in abusing the prisoner.
                            Or maybe ${he} just doesn't like knights, or people who want to be like one.
                            With the cyan kobold having left, you decide to stick around and see what happens to the bird on stage.
                            The crowd's relentless hurling of spoilt food continues for the better half of an hour.
                            Then, with the mob running out of things to throw at him, one of the guardsmen, a heavily battle-scarred feline,
                            approaches the avian from behind.
                        [/p]
                        [p]
                            With a single [b]swipe[/b] of the cat's clawed hand, the falcon's pants are ripped to shreds.
                            Kicking apart the bird's legs, the guardsman lowers his own trousers.
                            You watch on in horror, out of sheer morbid curiosity, as the bird begins to [b]beg[/b] for mercy,
                            while the feline lines up his barbed member with the criminal's unlubed feathery butthole.
                            Then, under the exuberant cheers of the crowd, he thrusts forward,
                            the avian squawking in pain and dishonor as his last bit of diginity is publically torn away from him.
                        [/p]
                        [p]
                            While the cat begins to violently hump away at the prisoner's rear, the very pillory itself rocking back and forth with every thrust,
                            you notice a small line starting to form near the steps leading up to the stage.
                            People are paying the guards, for what you can only assume to be their own turn with the bound falcon's ass.
                        [/p]
                        [p]
                            You aren't sure how to feel about this.
                            Turning prisoners into unwilling prostitutes seems morally [b]wrong[/b].
                            But then again, perhaps the bird is a rapist himself.
                            Maybe they're just doing to him what he did to someone else.
                            Either way, you now begin to realize why the guards of Varanar are widely praised as being highly effective at what they do.
                            You wouldn't want to get caught stealing, knowing something like [b]this[/b] might happen to you.
                        [/p]
                        [p]
                            You stick around for a little bit longer, to confirm your suspicions.
                            Indeed, as the feline finishes up and steps aside, the next person in line gets up on stage.
                            A big, burly horse.
                            There won't be much of the falcon left for the rest to enjoy after [i]that guy[/i] is done with him.
                            You turn around and walk away as the equine lowers his pants,
                            not interested in watching half the town have their way with the poor guy who got caught.
                            After all, there [i]are[/i] other things to do on the market square, apart from fucking captive birds.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.troubling.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\decline\MorningThreeDeclineEggs.js

// scenes\Varanar\DayThree\MorningThree\thievery\MorningThreeKeepEgg.js

/**
 * Thievery section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeKeepEgg(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeThieveryChoice.KEEP;
    var reqController = ctrl.thievery;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var dick = ctrl.misc.dick();
        var baseMale = ctrl.misc.baseGender() === GenderEnum.MALE;
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var toReturn = `[p]
                            You decide to keep the ${global.kobold.inventory.createItemWithTooltip(new RottenEgg())}${weapon ? `, quickly and silently 
                            disposing of your ${global.kobold.inventory.createItemWithTooltip(weapon)}` : ``}.
                            You glare at the cyan kobold, making it [b]very[/b] clear that you [b]know[/b] what ${he} was up to.
                            Threateningly looming over ${her}, you bare your teeth, a low growl building at the back of your throat.
                            In the most bass voice you can muster, you tell ${her} [b]never[/b] to mess with you again.
                            Scared shitless, ${he} gives you a meek nod, before scampering off into the crowd.
                            ${She}'s probably going to go find someone else to scam.
                            Such is life for a kobold in Varanar.
                        [/p]
                        [p]
                            That was [b]fun[/b], being the tough ${baseMale ? `guy` : `girl`} for once, instead of being the meek thief getting scared off.
                            This is how it must have felt for all those people catching you while you were trying to steal their stuff.
                            With the thief having vanished, you decide to stick around and see what happens to the bird on stage.
                            The crowd's relentless hurling of spoilt food continues for the better half of an hour.
                            Then, with the mob running out of things to throw at him, one of the guardsmen, a heavily battle-scarred feline,
                            approaches the avian from behind.
                        [/p]
                        [p]
                            With a single [b]swipe[/b] of the cat's clawed hand, the falcon's pants are ripped to shreds.
                            Kicking apart the bird's legs, the guardsman lowers his own trousers.
                            You watch on in horror, out of sheer morbid curiosity, as the bird begins to [b]beg[/b] for mercy,
                            while the feline lines up his barbed member with the criminal's unlubed feathery butthole.
                            Then, under the exuberant cheers of the crowd, he thrusts forward,
                            the avian squawking in pain and dishonor as his last bit of diginity is publically torn away from him.
                        [/p]
                        [p]
                            While the cat begins to violently hump away at the prisoner's rear, the very pillory itself rocking back and forth with every thrust,
                            you notice a small line starting to form near the steps leading up to the stage.
                            People are paying the guards, for what you can only assume to be their own turn with the bound falcon's ass.
                        [/p]
                        [p]
                            You aren't sure how to feel about this.
                            Turning prisoners into unwilling prostitutes seems morally [b]wrong[/b].
                            But then again, perhaps the bird is a rapist himself.
                            Maybe they're just doing to him what he did to someone else.
                            Either way, you now begin to realize why the guards of Varanar are widely praised as being highly effective at what they do.
                            You wouldn't want to get caught stealing, knowing something like [b]this[/b] might happen to you.
                        [/p]
                        [p]
                            You stick around for a little bit longer, to confirm your suspicions.
                            Indeed, as the feline finishes up and steps aside, the next person in line gets up on stage.
                            A big, burly horse.
                            There won't be much of the falcon left for the rest to enjoy after [i]that guy[/i] is done with him.
                            You turn around and walk away as the equine lowers his pants,
                            not interested in watching half the town have their way with the poor guy who got caught.
                            After all, there [i]are[/i] other things to do on the market square, apart from fucking captive birds.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.keepEgg.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\thievery\MorningThreeKeepEgg.js

// scenes\Varanar\DayThree\MorningThree\thievery\MorningThreeReturnEgg.js

/**
 * Thievery section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeReturnEgg(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeThieveryChoice.RETURN;
    var reqController = ctrl.thievery;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var toReturn = `[p]
                            You hand the egg back to the cyan kobold, staring ${her} down with the most disappointed, betrayed face you can muster.
                            ${She} covers ${his} face with ${his} arms, almost as if ${he} [b]expects[/b] you to hit ${her}.
                            ${Her} guard slowly lowers as ${he} realizes no blow is coming.
                            You tell ${her} you won't hit ${her}, but ${he}'d better think twice before trying to steal from you again.
                            You might not be so forgiving next time.
                        [/p]
                        [p]
                            Almost grovelling before you, the cyan kobold thanks you for your mercy.
                            After a dismissive gesture, ${he} disappears into the crowd, probably off to go scam someone else.
                            Is this how it feels like to be the one [i]catching[/i] a thief instead of being the thief getting caught?
                            The rush of [b]power[/b], the feeling of utter [b]dominance[/b] and [b]righteousness[/b], actually having the [b]moral highgrounds[/b],
                            is this how knights feel [b]all[/b] of the time?!
                            Darnit, you should have stopped one of those dragons you saw earlier, to ask them what it's like to be a knight.
                        [/p]
                        [p]
                            With the thief having vanished, you decide to stick around and see what happens to the bird on stage.
                            The crowd's relentless hurling of spoilt food continues for the better half of an hour.
                            Then, with the mob running out of things to throw at him, one of the guardsmen, a heavily battle-scarred feline,
                            approaches the avian from behind.
                        [/p]
                        [p]
                            With a single [b]swipe[/b] of the cat's clawed hand, the falcon's pants are ripped to shreds.
                            Kicking apart the bird's legs, the guardsman lowers his own trousers.
                            You watch on in horror, out of sheer morbid curiosity, as the bird begins to [b]beg[/b] for mercy,
                            while the feline lines up his barbed member with the criminal's unlubed feathery butthole.
                            Then, under the exuberant cheers of the crowd, he thrusts forward,
                            the avian squawking in pain and dishonor as his last bit of diginity is publically torn away from him.
                        [/p]
                        [p]
                            While the cat begins to violently hump away at the prisoner's rear, the very pillory itself rocking back and forth with every thrust,
                            you notice a small line starting to form near the steps leading up to the stage.
                            People are paying the guards, for what you can only assume to be their own turn with the bound falcon's ass.
                        [/p]
                        [p]
                            You aren't sure how to feel about this.
                            Turning prisoners into unwilling prostitutes seems morally [b]wrong[/b].
                            But then again, perhaps the bird is a rapist himself.
                            Maybe they're just doing to him what he did to someone else.
                            Either way, you now begin to realize why the guards of Varanar are widely praised as being highly effective at what they do.
                            You wouldn't want to get caught stealing, knowing something like [b]this[/b] might happen to you.
                        [/p]
                        [p]
                            You stick around for a little bit longer, to confirm your suspicions.
                            Indeed, as the feline finishes up and steps aside, the next person in line gets up on stage.
                            A big, burly horse.
                            There won't be much of the falcon left for the rest to enjoy after [i]that guy[/i] is done with him.
                            You turn around and walk away as the equine lowers his pants,
                            not interested in watching half the town have their way with the poor guy who got caught.
                            After all, there [i]are[/i] other things to do on the market square, apart from fucking captive birds.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.returnEgg.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\thievery\MorningThreeReturnEgg.js

// scenes\Varanar\DayThree\MorningThree\thievery\MorningThreeThrowEgg.js

/**
 * Thievery section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeThrowEgg(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeThieveryChoice.THROW;
    var reqController = ctrl.thievery;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.morning.eggedKobold = true;
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var toReturn = `[p]
                            Having caught ${her} red-handed, you suddenly feel [i]mischievous[/i].
                            The cyan kobold wanted you to throw this second egg, and throw it you will.
                            With a lightning fast motion of your arm, and just a small flick of your wrist, your hurl the rancid projectile ${his} way.
                            It impacts against ${his} face, shattering into a million little pieces,
                            its absolutely [b]rancid[/b] contents splattering all over the kobold's scales. 
                        [/p]
                        [p]
                            As soon as ${he} realizes what just happened, the kobold bursts into tears.
                            ${She} turns tail, and runs away, disappearing into the crowd,
                            probably towards the nearest public bath to wipe the disgusting muck off ${his} face.
                            ${She} got what ${he} deserved, trying to steal from you after [b]pretending[/b] to be your [b]friend[/b].
                        [/p]
                        [p]
                            With the thief having vanished, you decide to stick around and see what happens to the bird on stage.
                            The crowd's relentless hurling of spoilt food continues for the better half of an hour.
                            Then, with the mob running out of things to throw at him, one of the guardsmen, a heavily battle-scarred feline,
                            approaches the avian from behind.
                        [/p]
                        [p]
                            With a single [b]swipe[/b] of the cat's clawed hand, the falcon's pants are ripped to shreds.
                            Kicking apart the bird's legs, the guardsman lowers his own trousers.
                            You watch on in horror, out of sheer morbid curiosity, as the bird begins to [b]beg[/b] for mercy,
                            while the feline lines up his barbed member with the criminal's unlubed feathery butthole.
                            Then, under the exuberant cheers of the crowd, he thrusts forward,
                            the avian squawking in pain and dishonor as his last bit of diginity is publically torn away from him.
                        [/p]
                        [p]
                            While the cat begins to violently hump away at the prisoner's rear, the very pillory itself rocking back and forth with every thrust,
                            you notice a small line starting to form near the steps leading up to the stage.
                            People are paying the guards, for what you can only assume to be their own turn with the bound falcon's ass.
                        [/p]
                        [p]
                            You aren't sure how to feel about this.
                            Turning prisoners into unwilling prostitutes seems morally [b]wrong[/b].
                            But then again, perhaps the bird is a rapist himself.
                            Maybe they're just doing to him what he did to someone else.
                            Either way, you now begin to realize why the guards of Varanar are widely praised as being highly effective at what they do.
                            You wouldn't want to get caught stealing, knowing something like [b]this[/b] might happen to you.
                        [/p]
                        [p]
                            You stick around for a little bit longer, to confirm your suspicions.
                            Indeed, as the feline finishes up and steps aside, the next person in line gets up on stage.
                            A big, burly horse.
                            There won't be much of the falcon left for the rest to enjoy after [i]that guy[/i] is done with him.
                            You turn around and walk away as the equine lowers his pants,
                            not interested in watching half the town have their way with the poor guy who got caught.
                            After all, there [i]are[/i] other things to do on the market square, apart from fucking captive birds.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.throwEgg.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\thievery\MorningThreeThrowEgg.js

// scenes\Varanar\DayThree\MorningThree\thievery\MorningThreeThievery.js

/**
 * Throw one section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeThievery(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.thievery;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new MorningThreeKeepEgg(ctrl);
    var secondNext = new MorningThreeReturnEgg(ctrl);
    var thirdNext = new MorningThreeThrowEgg(ctrl);

    var firstChoice = MorningThreeThieveryChoice.KEEP;
    var secondChoice = MorningThreeThieveryChoice.RETURN;
    var thirdChoice = MorningThreeThieveryChoice.THROW;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.morning.thievingAttempt = true;
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var toReturn = `[p]
                            You [i]could[/i] use some cheering up after such a hellish night,
                            and you don't want to look [i]weird[/i] in front of your new friend.
                            Plus it'll be nice to get some [i]retribution[/i] on a non-kobold.
                            After all, knights are supposed to [b]punish[/b] evildoers,
                            and the bird on stage must have done [i]something[/i] to earn his place in the spotlight.
                            With an abundance of reasons to throw, and none to hold back, you decide to let loose and pelt away.
                        [/p]
                        [p]
                            With a small grunt, you chuck your first egg at the trapped avian.
                            The rotten would-be-breakfast soars through the air with a slight curve, before hitting the bird squarely in the face!
                            It cracks open, the disgusting goop contained within splattering all over the man's feathery face, most of the stuff flowing down along his beak.
                            Some of it manages to make its way into his mouth.
                            He gags, sputters and spits.
                            The cyan kobold burst out into laughter, and you join in.
                            "[color=cyan]Rotten like your wife's clutches![/color]"
                            ${She} turns to you.
                            "[color=cyan]Hit'em with the other one![/color]"
                        [/p]
                        [p]
                            You decline.
                            The man you're throwing at is [i]helpless[/i].
                            Bullying him may be fun, but you know first-hand that being on the receiving end is everything but.
                            Doing this, throwing these eggs, makes you no better than the humans and their ilk.
                            And while you [i]did[/i] let yourself go [i]a little[/i], you refuse to stoop yourself down to their level any further.
                        [/p]
                        [p]
                            You hold the rotten egg in-hand, and turn to face your new friend, just in time to witness ${his} hand reaching for your coinpouch.
                            ${She} immediately pulls back, but you are no fool.
                            In fact, as a thief, you've performed these kinds of scams yourself.
                            You demand to know what ${he} was reaching for.
                            "[color=cyan]J-just stretching my arms.[/color]" comes the stuttering reply,
                            a droplet of sweat running drown from the kobold's forehead, signalling ${his} nervousness.
                        [/p]
                        ${global.ui.condFail(`Charisma below 6`)}`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Keep it</button>`;

        var tooltip = global.parser.parse(`The egg could serve as a weapon. Keep it.`);

        if (ctrl.misc.weapon()) {
            tooltip += ` Your current weapon will be discarded.`;
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Return it</button>`;

        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();

        var tooltip = global.parser.parse(`Hand it back to the would-be-thief and send ${her} on ${his} way.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Throw it</button>`;

        var dick = ctrl.misc.dick();

        var tooltip = global.parser.parse(`Teach the thief a lesson ${dick ? "s" : ""}he won't soon forget.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader choicerow">What do you do with the egg?</div><br/>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.thievery.change();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var MorningThreeThieveryChoice = {
    KEEP: 1,
    RETURN: 2,
    THROW: 3
};

// END scenes\Varanar\DayThree\MorningThree\thievery\MorningThreeThievery.js

// scenes\Varanar\DayThree\MorningThree\robbed\MorningThreeRobbedContinue.js

/**
 * Robbed section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeRobbedContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With your friend missing, you decide to stick around and see what happens to the bird on stage.
                            The crowd's relentless hurling of spoilt food continues for the better half of an hour.
                            Then, with the mob running out of things to throw at him, one of the guardsmen, a heavily battle-scarred feline,
                            approaches the avian from behind.
                        [/p]
                        [p]
                            With a single [b]swipe[/b] of the cat's clawed hand, the falcon's pants are ripped to shreds.
                            Kicking apart the bird's legs, the guardsman lowers his own trousers.
                            You watch on in horror, out of sheer morbid curiosity, as the bird begins to [b]beg[/b] for mercy,
                            while the feline lines up his barbed member with the criminal's unlubed feathery butthole.
                            Then, under the exuberant cheers of the crowd, he thrusts forward,
                            the avian squawking in pain and dishonor as his last bit of diginity is publically torn away from him.
                        [/p]
                        [p]
                            While the cat begins to violently hump away at the prisoner's rear, the very pillory itself rocking back and forth with every thrust,
                            you notice a small line starting to form near the steps leading up to the stage.
                            People are paying the guards, for what you can only assume to be their own turn with the bound falcon's ass.
                        [/p]
                        [p]
                            You aren't sure how to feel about this.
                            Turning prisoners into unwilling prostitutes seems morally [b]wrong[/b].
                            But then again, perhaps the bird is a rapist himself.
                            Maybe they're just doing to him what he did to someone else.
                            Either way, you now begin to realize why the guards of Varanar are widely praised as being highly effective at what they do.
                            You wouldn't want to get caught stealing, knowing something like [b]this[/b] might happen to you.
                        [/p]
                        [p]
                            You stick around for a little bit longer, to confirm your suspicions.
                            Indeed, as the feline finishes up and steps aside, the next person in line gets up on stage.
                            A big, burly horse.
                            There won't be much of the falcon left for the rest to enjoy after [i]that guy[/i] is done with him.
                            You turn around and walk away as the equine lowers his pants,
                            not interested in watching half the town have their way with the poor guy who got caught.
                            After all, there [i]are[/i] other things to do on the market square, apart from fucking captive birds.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.troubling.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\robbed\MorningThreeRobbedContinue.js

// scenes\Varanar\DayThree\MorningThree\robbed\MorningThreeRobbed.js

/**
 * Robbed section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeRobbed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreeRobbedContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.morning.stolenFrom = true;
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var toReturn = `[p]
                            You [i]could[/i] use some cheering up after such a hellish night,
                            and you don't want to look [i]weird[/i] in front of your new friend.
                            Plus it'll be nice to get some [i]retribution[/i] on a non-kobold.
                            After all, knights are supposed to [b]punish[/b] evildoers,
                            and the bird on stage must have done [i]something[/i] to earn his place in the spotlight.
                            With an abundance of reasons to throw, and none to hold back, you decide to let loose and pelt away.
                        [/p]
                        [p]
                            With a small grunt, you chuck your first egg at the trapped avian.
                            The rotten would-be-breakfast soars through the air with a slight curve, before hitting the bird squarely in the face!
                            It cracks open, the disgusting goop contained within splattering all over the man's feathery face, most of the stuff flowing down along his beak.
                            Some of it manages to make its way into his mouth.
                            He gags, sputters and spits.
                            The cyan kobold burst out into laughter, and you join in.
                            "[color=cyan]Rotten like your wife's clutches![/color]"
                            ${She} turns to you.
                            "[color=cyan]Hit'em with the other one![/color]"
                        [/p]
                        [p]
                            You don't comply right away.
                            Instead, you hold on to the second egg for a few moments,
                            putting all your [b]fear[/b] and [b]anger[/b] and [b]frustration[/b] into it,
                            as if enchanting it with every [b]semblance[/b] of your misfortune you've [b]ever[/b] had to deal with,
                            all the while [i]aiming[/i] your shot to make sure it'll hit its mark.
                            Only after pouring your very heart and soul into the impending throw, with a surge of great fury and determination, do you let loose.
                        [/p]
                        [p]
                            Like the last, the rotten egg flies gracefully through the air towards its avian target.
                            It seems you may have spent just a little bit [i]too[/i] long aiming, however.
                            Unlike your nigh-perfect first pitch, the second projectile misses the bird's face completely,
                            instead hitting the pillory's confining wood.
                            Oh well, at least some of the rancid splatter still made it to the side of his face.
                        [/p]
                        [p]
                            That second throw made you feel a [b]lot[/b] better.
                            It's as though a weight has been lifted off of your shoulders.
                            You turn to face your new friend once more, only to find out ${he}'s nowhere to be found.
                            ${She} must have disappeared into the crowd.
                            Odd.
                        [/p]
                        ${global.ui.condFail(`Charisma below 6`)}`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.robbed.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\robbed\MorningThreeRobbed.js

// scenes\Varanar\DayThree\MorningThree\bucket\MorningThreeBucketContinue.js

/**
 * Bucket section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeBucketContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var clothed = !ctrl.misc.nude();
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var toReturn = `[p]
                            You ${clothed ? `redress, and` : ``} consider going back to the alley to look for the cyan kobold.
                            After getting so crudely interrupted, you doubt ${he}'d stick around, though.
                            Not to mention the boot ${he} got hit by.
                            You [i]could[/i] go look for ${her} on the other end of the alleyway, but you don't even know where it leads to.
                            ${She}'ll be long gone by the time you get there.
                        [/p]
                        [p]
                            Besides, who knows?
                            Maybe ${he}'s out looking for you [b]right now[/b]!
                            If you went searching for ${her}, you'd just end up missing one another, especially with how crowded the streets seem to be getting.
                            No, the most reasonable thing for you to do, would be to wait here.
                            After all, this is where the two of you met.
                            If there's one place to go that the both of you know, [b]this[/b] would be it.
                            Surely ${he}'ll think to check here, right?
                        [/p]
                        [p]
                            You find a quiet spot near the stage, and sit yourself down on the ground.
                            You wait for seconds, then minutes, which slowly turn into hours.
                            Come high noon, with tears in your eyes, you realize:
                            ${he}'s not coming.
                        [/p]
                        [p]
                            Once again, you are all alone.
                            With a heavy heart, you rise to your feet.
                            The only consolation you can think of, is that one day the two of you may meet again.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.sad.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\bucket\MorningThreeBucketContinue.js

// scenes\Varanar\DayThree\MorningThree\bucket\MorningThreeBucket.js

/**
 * Bucket section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeBucket(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preBucket;
    var subController = ctrl.preContinue;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreeBucketContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var clothed = !ctrl.misc.nude();
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var toReturn = `[st]
                            [color=blue]Splash![/color]
                        [/st]
                        [p]
                            You shiver in cold and discomfort as an entire [b]bucket[/b] of [b]freezing[/b] water splashes all over you and your partner!
                            The alley is filled with the sounds of your combined hissing and growling.
                            A booming, bass voice rings out from above:
                            "[color=orange]Don't do that shit here, you filthy beasts! Shoo! Git![/color]"
                        [/p]
                        [p]
                            A naked human stands at a second-floor window leading out into the alleyway.
                            He reaches down to grab a heavy boot off the ground, and [b]hurls[/b] it towards the two of you.
                            You barely manage to dodge it, shouting a vicious insult back at him.
                            Your partner isn't quite so lucky, the boot landing directly atop ${his} head with a heavy impact, prompting a pained yelp.
                            "[color=orange]Don't make me come down there![/color]"
                        [/p]
                        [p]
                            The cyan kobold quickly gathers ${his} gear, ${clothed ? `while you do the same,` : ``}.
                            The both of you [b]leg[/b] it, fleeing from the alley as fast as you can, fearing the human might actually [i]follow up[/i] on his threats.
                            You [b]rush[/b] back towards the market square.
                            It isn't until you're standing ${clothed ? `naked` : ``} in front of the stage where the two of you met,
                            that you realize your partner fled [b]the other way[/b]!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        ctrl.splashSound.playOnce();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.bucket.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\bucket\MorningThreeBucket.js

// scenes\Varanar\DayThree\MorningThree\female\oral\MorningThreeFemaleHimFirst.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeFemaleHimFirst(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeFemaleOralChoice.HIM;
    var reqController = ctrl.femaleOral;
    var subController = ctrl.preBucket;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreeBucket(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var virgin = ctrl.misc.oralVirgin();
        var toReturn = `[p]
                            You're bursting with pleasure, your dripping wet cunt [b]begging[/b] for attention.
                            For now, your fingers must do.
                            Two of them slide into your soaked folds, and diligently begin to pump in and out.
                            Staring down at your partner's throbbing member fuels your rampant masturbation.
                            But the sight of it is not enough.
                            You want more.
                            You [b]need[/b] more.
                        [/p]
                        [p]
                            The cyan kobold leans against the alley-wall behind him, quickly catching on to your intentions as you slink down to your knees.
                            He strokes himself, resting his cock atop your snout, like you're some kind of [b]dog[/b] and he's presenting you with a [b]treat[/b].
                            You breathe in deeply, his [b]manly[/b], almost [b]intoxicating[/b] scent filling your lungs.
                            Your digits up their pace, a small puddle of feminine fluids forming on the ground in between your legs.
                            With your other hand, you take hold of his member yourself, jerking him off in his stead,
                            a small bauble of precum dripping onto the top of your muzzle.
                        [/p]
                        [p]
                            Feeling his length [b]throb[/b] in between your finger excites you further.
                            You can no longer contain yourself.
                            You [b]need[/b] to taste him.
                            You pull back a little, positioning the tapered tip of his member at the front-end of your muzzle.
                            Parting your lips, you you slowly take his meat into your mouth, its hefty,
                            undeniably [b]masculine[/b] taste making you lust for [b]more[/b].
                        [/p]
                        [p]
                            The kobold whispers sweet nothings to you as you begin to bob up and down on his shaft,
                            taking in inch after inch of pure, unadultered [b]cock[/b].
                            You swirl your tongue around his girth, both to slather it with saliva,
                            and to [b]bask[/b] in as much of his [b]heavenly[/b] flavor as you possibly can.
                            Your efforts are rewarded by a [b]squirt[/b] of pre along the entire length of your tongue,
                            your tastebuds [b]saturated[/b] with the tasteful, sticky liquid.
                        [/p]
                        [p]
                            Your fingers, while slightly pleasuring, are [b]far[/b] from enough to sate your [b]lust[/b].
                            You [b]want[/b] his tongue, and you'd rather have it sooner than later.
                            The need for release drives you to to suck your partner off as best as you can,
                            hoping to finish him quickly so he can return the favor.
                        [/p]
                        [p]
                            You take as much of the cyan kobold's member into your mouth as you can, without letting it slip into your throat.
                            Your lips slide up and down over half of his length, pleasuring it to the best of your abilities.
                            Still, he wants [b]more[/b].
                            The kobold begins to hump against your face, rolling his hips forwards whenever you bottom out,
                            attempting, but failing to get you to take [b]deeper[/b].
                            You don't blame him for trying, but you aren't comfortable enough with deepthroating, and you don't feel like trying it out now.
                        [/p]
                        [p]
                            After attempting to shove his cock down your throat a few more times, your partner relents with a slightly frustrated,
                            but still satisfied sigh, returning to resting against the wall behind him while letting [b]you[/b] do all of the work.
                            With his motions coming to a halt, you are free once again to happily work your magic.
                            It doesn't take much more for his shaft to start [b]quivering[/b] and [b]trembling[/b] in your maw, a clear sign that he's reaching his end.
                            As if you didn't already know, or maybe because he's such a nice guy, the kobold lets you know:
                            "[color=cyan]I-I'm getting close. Get ready.[/color]"
                        [/p]
                        [p]
                            You keep bobbing up and down his member until the very last second.
                            As the cyan kobold lets out a pleasured groan, visibly on the [b]verge[/b] of climax, you pull back,
                            quickly replacing your mouth with your hand, rapidly stroking his spit-slick shaft, [b]jerking[/b] the orgasm out of him.
                            With a loud grunt, your partner finishes, the first rope of his spunk splattering against your forehead.
                            You open your maw as wide as it'll go, lolling out your tongue just in time to catch the next wave of cum.
                            You eagerly swirl the thick fluid around, making sure its taste will last, before gulping it down.
                        [/p]
                        [p]
                            The rest of the contents of the kobold's internal balls get squirted out onto your face,
                            covering you from the underside of your muzzle to the very top of your head.
                            As his orgasm slowly comes to an end, the last few bits of his seed drooping down from his tapered tip onto the ground,
                            you let go of his spent member.
                            While one of your eyes is covered in spunk, you open the other,
                            looking up at him with a satisfied, perhaps even slightly [i]smug[/i] smile on your face.
                        [/p]
                        [p]
                            After wiping your face of most of the stuff, having another [i]taste[/i] in the process, you slowly rise to your feet.
                            Your frantic masturbation, which continued throughout the entire blowjob, has turned your groin into a [b]sopping wet[/b] mess,
                            your thighs completely [b]drenched[/b] in feminine arousal.
                            Still, you have yet to reach your own end.
                            Laying a hand atop your recovering partner's head, you begin to push him down to his knees,
                            when all of a sudden, the two of you get interrupted.
                        [/p]`;

        if (virgin) {
            toReturn += `[p]
                            [c]
                                [b]
                                    You have lost your oral virginity.
                                [/b]
                            [/c]
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.himFirst.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\female\oral\MorningThreeFemaleHimFirst.js

// scenes\Varanar\DayThree\MorningThree\female\oral\MorningThreeFemaleYouFirst.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeFemaleYouFirst(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeFemaleOralChoice.YOU;
    var reqController = ctrl.femaleOral;
    var subController = ctrl.preBucket;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreeBucket(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var boobs = ctrl.misc.boobs();
        var toReturn = `[p]
                            You'd be lying if you said his cock didn't seem the most alluring thing in the world right now,
                            practically [b]begging[/b] to be [b]touched[/b] and [b]licked[/b] and [b]sucked[/b], but your own pleasure comes first.
                            You give your slit a few needy rubs.
                            He's really got you going, and, to your delight, he seems to understand what you're about to ask of him.
                            Without even [b]needing[/b] to ask, he leans forward, and presses his muzzle against your belly,
                            planting kisses as his mouth travels downwards.
                        [/p]
                        [p]
                            He drops to his knees, both of his hands reaching upwards to caress your ${boobs ? `` : `flat`} chest.
                            While his snout is mere inches away from your loins, he doesn't immediately get to work.
                            Instead, he blows hot air onto your folds, teasing you further, arousing lust in ways you didn't even know were [b]possible[/b].
                            No longer able to stand all this winding up, you place a hand at the back of his head, and [b]pull[/b] his muzzle in between your legs,
                            his mouth ending up pressing directly against your soaked slit.
                        [/p]
                        [p]
                            The kobold immediately gets to work.
                            You let out a loud moan as you feel the first lash of his tongue against your folds,
                            scooping up more than a little of the excitement covering your slit is [b]drenched[/b] with,
                            the fluid replaced with equally slick and messy saliva.
                        [/p]
                        [p]
                            You close your feminine arousal-covered thighs around his head, locking your partner in between your legs and your groin,
                            while simultaneously pushing him [b]deeper[/b] into folds.
                            He steadies himself by grabbing onto your hips, his fingers gently massaging your scaly flesh while he starts to eat you out [b]properly[/b],
                            his tongue [b]digging[/b] into your cunt in a series of exceedingly delightful licks.
                            You relax your entire body, sinking your weight down onto his face, your loud moans echoing throughout the alleyway.
                        [/p]
                        [p]
                            He knows [b]exactly[/b] what he's doing.
                            His tongue probes your loins, hitting [b]all[/b] the right spots while lapping at every inch of your folds.
                            His snout rubs back and forth against your clit [b]so much[/b] you can no longer write it off as a pleasurable coincidence.
                            He's intently moving his face from side to side,
                            nuzzling your groin just so his snout'll brush alongside your pleasure button, all to get a rise out of you.
                            Worst of all is: [b]it's working[/b].
                        [/p]
                        [p]
                            You're [b]wetter[/b] than you've [b]ever[/b] been, not because of his spit,
                            but because the kobold's [b]professional[/b] cunt-munching skills are [b]driving you wild[/b].
                            You arch your back, leaning against the wall behind you to give your partner a slightly better angle to work with, not that he needs it.
                            Having shown off his impressive [i]technique[/i], the male digs deeper, his tongue burying into your slit, lapping at your very innards.
                            You can feel your end rapidly approaching.
                            He's got you on the ropes, and [b]he knows it[/b].
                        [/p]
                        [p]
                            As you reach the very [b]edge[/b] of climax, the cyan kobold suddenly [b]stops[/b] dead in his tracks.
                            You look down in frustration, just in time to witness him repositioning his muzzle,
                            tilting his head just a little bit [i]upwards[/i].
                            Then, he wraps his lips around your clit, gently suckling it while his tongue gives it a [b]dozen[/b] licks.
                            The sudden attention to your most sentive part sends spikes of sheer [b]pleasure[/b] down your spine,
                            your mind going blank as a hefty orgasm short-circuits your brain.
                            Your hips start to [b]ride[/b] your partner's face all on their own,
                            slathering his scales with your feminine fluids, which [b]leak[/b] liberally from your convulsing cunt.
                        [/p]
                        [p]
                            Your entire body trembles and shakes.
                            You pant, huff and gasp for air.
                            That was unlike [b]anything[/b] you've [b]ever[/b] experienced.
                            Your thighs unwrap from the cyan kobold's head.
                            You rest your entire weight onto the alley-wall, your legs unable to keep you upright on their own.
                        [/p]
                        [p]
                            You tell him he's [b]incredible[/b].
                            "[color=cyan]I know.[/color]" comes his cocky response.
                            He rises to his feet, his face stained with more than a little of your excitement.
                            While still a little [i]out of it[/i], your attention is still pulled to your partner's groin,
                            his erect member jutting out towards you.
                            Just like he knew what you expected of him, so too, do you realize what he wants you to do.
                            Right as you are about to drop down to your knees to suck the kobold off, however, your heated get-together gets interrupted.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.youFirst.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\female\oral\MorningThreeFemaleYouFirst.js

// scenes\Varanar\DayThree\MorningThree\female\oral\MorningThreeFemaleOral.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeFemaleOral(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeFemaleForeplayChoice.ORAL;
    var reqController = ctrl.femaleForeplay;
    var subController = ctrl.femaleOral;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new MorningThreeFemaleYouFirst(ctrl);
    var secondNext = new MorningThreeFemaleHimFirst(ctrl);

    var firstChoice = MorningThreeFemaleOralChoice.YOU;
    var secondChoice = MorningThreeFemaleOralChoice.HIM;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.morning.oral = true;
        var toReturn = `[p]
                            You shake your head.
                            While you [i]are[/i] every bit as [i]excited[/i] as he is, you don't want to have sex right now.
                            You feel a shiver running down your spine, your mind fearing that he [i]might[/i] simply disregard your wishes, and fuck you anyway.
                            Before your imagination can take that particular scenario any further, the cyan kobold gives you a small nod,
                            reluctantly pulling his member away from you, taking a small step back.
                            You let out a relieved sigh.
                            "[color=cyan]If you [b]don't[/b] want to [i]make love[/i], then, uh, what [b]do[/b] you want to do?[/color]"
                            Well, there [i]are[/i] some [i]other ways[/i] to have [i]fun[/i] together
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>You first</button>`;

        var tooltip = global.parser.parse(`Your groin is [b]burning[/b] with desire. Make him quench the fire.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var exertionCheck = ctrl.misc.exertion() >= 50;
        var disabled = subController.has() || exertionCheck ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Him first</button>`;

        var tooltip = global.parser.parse(`You just kind of [i]denied[/i] him. It's only fair that he gets to go first.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">Who gets pleasured first?</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var MorningThreeFemaleOralChoice = {
    YOU: 1,
    HIM: 2
};

// END scenes\Varanar\DayThree\MorningThree\female\oral\MorningThreeFemaleOral.js

// scenes\Varanar\DayThree\MorningThree\female\sex\MorningThreeFemaleOutside.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeFemaleOutside(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeFemaleSexChoice.OUTSIDE;
    var reqController = ctrl.femaleSex;
    var subController = ctrl.preBucket;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreeBucket(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var virgin = ctrl.misc.cuntVirgin();
        var toReturn = `[p]
                            You shake your head no.
                            While you [b]really[/b] like him, you're simply [b]not ready[/b] to be a mother yet.
                            You don't want his child.
                            You don't want [b]anyone's[/b] child.
                            "[color=cyan]It's alright.[/color]" he responds, nodding comfortingly.
                            To your alarm, however, his member is still firmly lodged inside of you, and his hips continue to move,
                            pumping his shaft in and out of your slit.
                            "[color=cyan]I've got this...don't worry, I'll just...[/color]"
                        [/p]
                        [p]
                            You tell him to stop, to pull out and to finish [b]outside[/b],
                            but your weak protests are soon cut off as [b]rams[/b] into you once more, and starts humping like [b]crazy[/b],
                            fucking you with renewed vigor and zeal, pressing you against the wall while [b]pounding[/b] away like there's no tomorrow,
                            outright [b]forcing[/b] an orgasm out of your sensitive body.
                        [/p]
                        [p]
                            You [b]scream[/b] in ecstasy, your mind [b]overloaded[/b] with pleasure as you reach your end.
                            You wildly buck your hips back against the cyan kobold's member, your arousal squirting onto his abdomen,
                            your cunt tightening, gripping his member like a vice.
                            You enter an orgasm-induced euphoria, your brain simply [b]turning off[/b], refusing to work, your body moving all on its own,
                            your partner continuing to piston in and out of your contracting slit.
                        [/p]
                        [p]
                            Time itself seems to stop.
                            You have no idea for how long you are left convulsing in absolute bliss.
                            When you awaken from your lust-driven haze,
                            returning to your senses after discarding your inhibitions and riding his dick like a woman possessed,
                            you find yourself in a tender embrace, the cyan kobold lovingly cuddling you.
                        [/p]`;

        if (ctrl.misc.luck() > 5) {
            toReturn += `[p]
                            Slowly, the kobold pulls his throbbing shaft from your slit.
                            To your amazement, he seems to have lasted throughout that entire ordeal.
                            While clearly on the very edge of climaxing, he managed to refrain from cumming inside of you,
                            instead pulling out [b]exactly[/b] like you asked him to.
                            A wave of [b]relief[/b] washes over you as you realize you're [b]not[/b] getting pregnant, at least not from this romp.
                        [/p]
                        ${global.ui.condSucc(`Luck greater than 5`)}
                        [p]
                            While lost in thoughts, you completely forget about the needs of your partner.
                            "[color=cyan]H-help me out here, will you?[/color]" he speaks, snapping you back to the real world.
                            You're all too happy to help, especially after the [b]amazing[/b] feat he pulled, and the [b]incredible[/b] orgasm he gave you.
                            He has more than [b]earned[/b] his release.
                            Right as you are about to drop down to your knees to finish the kobold off, however, your heated get-together gets interrupted.
                        [/p]`;
        } else {
            global.kobold.history.shared.varanar.dayThree.morning.cameInside = true;
            toReturn += `[p]
                            Slowly, the kobold pulls his throbbing shaft from your slit,
                            a [b]torrent[/b] of spunk [b]gushing[/b] from your well-fucked cunt along with it, dripping down to the ground below.
                            Wait a minute...did he cum inside of you?!
                            Panic settling in, you wipe your groin with one of your hands,
                            just so you can stare in disbelief at the sticky white stuff clinging to your palm.
                            You reach in with two of your finger, in a vain attempt to [i]scoop out[/i] his seed, but it's too late.
                            The damage has been done.
                        [/p]
                        ${global.ui.condFail(`Luck below 6`)}
                        [p]
                            You turn your attention back towards your partner, a scolding look on your face.
                            How could he have done this?!
                            Especially after you asked him [b]not[/b] to!
                            The cyan kobold simply shrugs, although he clearly feels bad for what he did.
                            You kind of [i]did[/i] start [b]riding[/b] him out of the blue, like you [b]wanted[/b] his seed.
                            Luckily, for the both of you, your one-sided [i]argument[/i] gets cut short by an unwanted interruption.
                        [/p]`;
        }

        if (virgin) {
            toReturn += `[p]
                            [c]
                                [b]
                                    You are no longer a virgin.
                                [/b]
                            [/c]
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.femaleOutside.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\female\sex\MorningThreeFemaleOutside.js

// scenes\Varanar\DayThree\MorningThree\female\sex\MorningThreeFemaleInside.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeFemaleInside(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeFemaleSexChoice.INSIDE;
    var reqController = ctrl.femaleSex;
    var subController = ctrl.preBucket;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreeBucket(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.morning.cameInside = true;
        var virgin = ctrl.misc.cuntVirgin();
        var toReturn = `[p]
                            A million thoughts flash through your mind, the most pressing one being that you could very well end up pregnant if you let him finish inside.
                            You're not sure if you're ready to be a mother yet.
                            But at least, even if you [b]do[/b] get pregnant, it'll be the child of someone you like, and may even grow to [b]love[/b].
                            He stops his thrusting, impatiently awaiting an answer.
                            Wrapping your legs around his body, you pull him closer to you, locking him inside of your sopping wet folds.
                            You nod at him.
                            You are ready.
                        [/p]
                        [p]
                            A look of pure bliss takes over the cyan kobold's face.
                            He happily obliges, starting to hump you like [b]crazy[/b],
                            fucking you with renewed vigor and zeal, holding you close while [b]pounding[/b] you like there's no tomorrow.
                            You forego riding him, instead just sitting still, letting him have his way with you, letting him [b]fuck[/b] you to climax.
                        [/p]
                        [p]
                            It doesn't take much more for you to reach your end.
                            Your moans get louder and louder, ultimately culminating into a [b]scream[/b] of pleasure, a mind-wracking orgasm hitting your body hard.
                            Your feminine fluids [b]squirt[/b] all over his abdomen, while your innards quiver and spasm,
                            tightening around the kobold's cock, massaging its entire length, [b]milking[/b] it for its fertile semen.
                            The sensation proves to be too much for your partner.
                            After giving you a few more aggressive thrusts, he keeps himself hilted, his shaft [b]pulsing[/b] as he cums inside of you,
                            wave after wave of warm, fertile semen pumped directly into the depths of your cunt.
                        [/p]
                        [p]
                            As his orgasm begins to fade, the cyan kobold slowly pulls out of your cum-leaking slit.
                            You have no idea whether or not he got you pregnant, but part of you [b]hopes[/b] that he did.
                            His bountiful spunk oozes from your thoroughly bred groin, dripping down to the ground below,
                            mixing with the small puddle of your arousal that formed during the intense bout of lovemaking.
                        [/p]
                        [p]
                            Your partner lowers you to the ground.
                            The two of you stare longingly into one another's eyes, wordlessly connecting on a spiritual level.
                            Time seems to stop.
                            You wish you could stay like this forever.
                            But alas, your tranquil moment gets interrupted all too soon.
                        [/p]`;

        if (virgin) {
            toReturn += `[p]
                            [c]
                                [b]
                                    You are no longer a virgin.
                                [/b]
                            [/c]
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.femaleInside.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\female\sex\MorningThreeFemaleInside.js

// scenes\Varanar\DayThree\MorningThree\female\sex\MorningThreeFemaleSex.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeFemaleSex(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeFemaleForeplayChoice.SEX;
    var reqController = ctrl.femaleForeplay;
    var subController = ctrl.femaleSex;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new MorningThreeFemaleOutside(ctrl);
    var secondNext = new MorningThreeFemaleInside(ctrl);

    var firstChoice = MorningThreeFemaleSexChoice.OUTSIDE;
    var secondChoice = MorningThreeFemaleSexChoice.INSIDE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.morning.fuckedMaleKobold = true;
        var virgin = ctrl.misc.cuntVirgin();
        var toReturn = `[p]
                            You don't think you've [b]ever[/b] been [b]this[/b] [i]aroused[/i] before in your life. 
                            Your loins [b]burn[/b] with [b]lust[/b], a raging fire of desire that can only be quelled by [b]him[/b].
                            So far gone in [b]need[/b] are you, that you are completely lost for words to express how much you [b]want[/b] this.
                            You give the cyan kobold a simple nod, a gesture which succeeds where words failed you.
                            Lining up with your slick slit, he rolls his hips forwards, slowly parting your folds, lodging himself deep inside of you.
                        [/p]`;

        if (virgin) {
            toReturn += `[p]
                            You wince as you feel him penetrate you, somewhat [i]expecting[/i] this to hurt.
                            Your partner, however, is slow and careful.
                            He whispers soothing words, making you forget about the pain, and instead focus on the pleasure, while he continues to push forwards.
                            Your maiden barrier is broken.
                            A small amount of blood trickles from your loins, signifying the loss of your virginity.
                            You grunt in discomfort.
                            "[color=cyan]Are you okay?[/color]" the kobold immediately asks, in an endearingly concerned fashion.
                            You tell him you're fine, and motion for him to keep on going.
                            You don't need to tell him twice.
                        [/p]`;
        } else {
            toReturn += `[p]
                            A small moan escapes your lips, your partner sinking his entire length inside of you in a single, smooth thrust.
                            After all but [b]impaling[/b] you on his shaft, the kobold comes to a rest,
                            simply [b]basking[/b] in your warmth and tightness, while staring longingly into your eyes.
                            You embrace him in a tender hug, steadying yourself against the wall.
                            To encourage him further, you tell him he's the [b]best[/b] you've ever had.
                            A big smirk forms on his face.
                        [/p]`;
        }

        toReturn += `   [p]
                            The cyan kobold gently starts to rock back and forth, shallowly at first.
                            His pace steadily increases, as does the ferocity of his thrusts,
                            more and more of his member sliding in and out of you with every motion of his hips.
                            Grabbing your plump behind, he lifts your legs off the ground, the new angle allowing him to up his tempo yet again.
                            Before long, he's outright [b]mating[/b] you, fucking you against the wall,
                            the sounds of his lustful grunts and your pleasured moans reaching far beyond the confines of the alleyway.
                        [/p]
                        [p]
                            You begin to hump back at his every thrust, letting your body [b]sink[/b] onto his member further,
                            somewhat [b]riding[/b] him without interrupting his steady rhythm.
                            All the while, your partner whispers sweet compliments.
                            He tells you how [b]beautiful[/b] you are, how [b]tight[/b] your cunt is, how he [b]wishes[/b] he'd met you earlier.
                            In between moans, you respond to his flattery, telling him he's doing [b]great[/b], [b]begging[/b] him to [b]keep going[/b].
                        [/p]
                        [p]
                            And keep going he does.
                            You can feel your excitement rapidly building, a distinct [i]peak[/i] nearing,
                            getting closer and closer every time your slit wetly slaps against his groin, and his tapered dick bottoms out in your receptive hole.
                            He, too, is nearing his end.
                            Trying to restrain himself, he bites his lower lip.
                            His member twitches and trembles inside of you, clearly on the very [b]edge[/b].
                            Realizing he won't be able to last much longer, the kobold stammers out a critical question in the heat of the moment:
                            "[color=cyan]C-can I cum inside?[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>No</button>`;

        var tooltip = global.parser.parse(`You're not ready to have a baby.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var exertionCheck = ctrl.misc.exertion() >= 50;
        var disabled = subController.has() || exertionCheck ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Yes</button>`;

        var tooltip = global.parser.parse(`Rather him than someone else.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">Do you let him?</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var MorningThreeFemaleSexChoice = {
    OUTSIDE: 1,
    INSIDE: 2
};

// END scenes\Varanar\DayThree\MorningThree\female\sex\MorningThreeFemaleSex.js

// scenes\Varanar\DayThree\MorningThree\female\MorningThreeFemaleForeplay.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeFemaleForeplay(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.femaleForeplay;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new MorningThreeFemaleOral(ctrl);
    var secondNext = new MorningThreeFemaleSex(ctrl);

    var firstChoice = MorningThreeFemaleForeplayChoice.ORAL;
    var secondChoice = MorningThreeFemaleForeplayChoice.SEX;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var clothed = !ctrl.misc.nude();
        var boobs = ctrl.misc.boobs();
        var virgin = ctrl.misc.cuntVirgin();
        var toReturn = `[p]
                            All of the trials and tribulations you've faced these past few days are forgotten in an instant.
                            In the arms of a stranger you have only just met, your worries melt away.
                            Feeling his slight shivering, his heartbeat and his warm breath against your scales only serves to arouse you further.
                            The both of you mutually struggle and fumble to 
                            ${clothed ? `shed whatever clothes you're wearing` : `take off his meagre clothing`},
                            inspecting one another with wandering hands and eyes alike.
                        [/p]
                        [p]
                            He's [b]eager[/b].
                            His snout buries into your neck, taking in your scent.
                            One of his hands starts to ${boobs ? `caress your breasts` : `rub your flat chest`},
                            while the other makes its way downwards, towards your crotch.
                            You slightly spread your legs, giving him more room to [i]explore[/i], to prod and to [b]touch[/b],
                            one of his fingers running along the entire length of your ${virgin ? `virginal` : ``} slit before slipping inside of your folds,
                            [i]probing[/i] you, testing you, checking just how [b]tight[/b] his latest hookup really is.
                        [/p]`;

        if (virgin) {
            toReturn += `[p]
                            His eyes widen a little in surprise, as the kobold realizes you're still a [i]virgin[/i].
                            You can [b]see[/b] him wondering how in the hell you've managed to stay [i]undefiled[/i] for this long.
                            To quell his confusion, you mention you don't actually [i]live[/i] in Varanar.
                            You're from the forest.
                            A sly smirk forms on his face.
                            "[color=cyan]Are all girls living in the forest as pretty as you are?[/color]"
                            You can't help but blush, your scaly cheeks flushing red.
                            You press your body down against his fingers, wordlessly urging him to continue.
                        [/p]`;
        }

        toReturn += `   [p]
                            You can't say his touches have left you [i]unaffected[/i].
                            Your loins moisten, arousal building up inside of you.
                            One of your own hands reaches for his groin, to find his fully erect member [b]waiting[/b] to be touched.
                            He licks your neck affectionately as you begin to stroke his length, mutual [b]lust[/b] and [b]desire[/b] building between the two of you.
                        [/p]
                        [p]
                            Things quickly get heated.
                            A second finger slips into your slit.
                            Precum drips from the tapered tip of his length.
                            He gently pushes you against the alley-wall for support.
                            Removing his fingers from your well-prepared cunt,
                            he takes hold of his own shaft, guiding it to your entrance,
                            rubbing it against your slick folds, eagerly and impatiently.
                            You can [b]feel[/b] how much he [b]wants[/b] this, but still he doesn't put it in just yet.
                            Instead, the kobold gallantly asks:
                            "[color=cyan]Are you [b]sure[/b] you want this?[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>No</button>`;

        var virgin = ctrl.misc.cuntVirgin();
        var tooltip = global.parser.parse(`${virgin ? `You don't want to lose your purity.` : `You don't want to get fucked right now.`}`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var exertionCheck = ctrl.misc.exertion() >= 50;
        var disabled = subController.has() || exertionCheck ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Yes</button>`;

        var tooltip = global.parser.parse(`You don't [i]want[/i] this. You [b]need[/b] it.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You say:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var MorningThreeFemaleForeplayChoice = {
    ORAL: 1,
    SEX: 2
};

// END scenes\Varanar\DayThree\MorningThree\female\MorningThreeFemaleForeplay.js

// scenes\Varanar\DayThree\MorningThree\male\MorningThreeMaleSexFinish.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeMaleSexFinish(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.maleSex;
    var subController = ctrl.preBucket;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreeBucket(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var clothed = !ctrl.misc.nude();
        var virgin = ctrl.misc.dickVirgin();
        var toReturn = ``;

        if (dick < SizeEnum.AVERAGE) {
            toReturn += `[p]
                            Taking a deep breath, you roll your hips forwards forwards, first parting the cyan kobold's slit with your tapered tip,
                            then slickly sliding the rest of your member into her loose, wet cunt.
                            Your entire length easily slips inside,
                            her soaked folds coming to rest against your crotch-scales.
                            You let out a loud, pleasured sigh, your entire body shuddering as you feel her warmth envelop you in every way possible.
                        [/p]
                        [p]
                            You're glad to hear the enjoyment is mutual.
                            Your partner outright [b]squeals[/b] in [b]delight[/b], her legs wrapping around you,
                            tugging you [b]even closer[/b], pinning her entire weight on you and the wall, wordlessly urging you to [b]fuck her[/b] as hard as you can.
                        [/p]
                        [p]
                            Once again, you oblige.
                            You grab her rear to keep her steady, before halfway pulling out of her.
                            For a moment, you feel a deep sense of [i]loss[/i] and a tiny hint of cold enveloping your member,
                            only for it to vanish as you thrust back inside,
                            elliciting a loud moan from the kobold as her sopping wet cunt sloppily smacks against your groin once again.
                            You can feel her lukewarm feminine fluids leaking down along your thighs,
                            rolling downwards along your scaly legs before ending up on your discarded clothing.
                        [/p]
                        [p]
                            You faintly hear the hollering and hooting of the crowd on the market square,
                            but while hurling eggs at the captured avian was fun,
                            you've found something [b]much[/b] more enjoyable to do.
                            You rapidly pick up a steady rhythm, pulling out of the cyan kobold's snatch before [b]ramming[/b] back inside of her,
                            [b]pounding[/b] her slit with [b]all[/b] of the [b]ferocity[/b] and the [b]passion[/b] she [b]begged[/b] you for, and then some.
                        [/p]
                        [p]
                            Her wordless gasping and needy trembling only serve to encourage you further.
                            For these few minutes, you forget about your debt.
                            About the years of abuse you've suffered.
                            About [b]everything[/b] bad in your life.
                            All you can think about is [b]her[/b].
                        [/p]
                        [p]
                            You let out a lusty growl, nuzzling your snout against her smooth neck.
                            Your member [b]throbs[/b] inside of her, on the verge of blowing its load.
                            You're not the only one on the edge.
                            Her heavy breathing, loud moaning, and the way her cunt [b]clenches[/b]
                            around your cock are all clear signs that she's nearing her own climax.
                            She lets her body sink down on you every time you thrust upwards, humping back against your rocking motions.
                            "[color=cyan]Y-yes![/color]" she [b]screams[/b], her legs joining in, [b]spearing[/b] her deeper onto your shaft.
                            "[color=cyan]Keep going! Don't pull out! Fill me! Please![/color]"
                        [/p]
                        [p]
                            You couldn't deny her even if you wanted to.
                        [/p]
                        [p]
                            You give the kobold one final thrust, with all of your might.
                            Her cunt begins to [b]quiver[/b] and [b]contract[/b] around your length, [b]milking[/b] your member for all it's got.
                            She just [b]barely[/b] manages to plant her lips atop yours in time to muffle a passionate moan,
                            yanking you into a deep kiss as her naked form shudders and tenses,
                            her legs gripping around you like a vice, holding you firmly lodged inside of her.
                            Her claws all but [b]bury[/b] into your back, excitement [b]gushing[/b] forth from her tightening slit,
                            drenching your groin in sticky, fragrant, feminine hormone-laden fluids.
                        [/p]
                        [p]
                            The intense scent of female arousal spurring a primal [b]need[/b] to [b]breed[/b] inside of your lust-clouded mind, you, too, reach your end.
                            The tapered tip of your member [b]erupts[/b] in orgasm,
                            your internal balls churning out rope after rope of thick, fertile semen deep inside of the cyan kobold's slit,
                            painting her inner walls a pearly white, your cum all too soon starting to leak out alongside her own arousal.
                        [/p]
                        [p]
                            After your seed is fully spent, the long-lasting kiss comes to an end, leaving the both of you huffing and panting as your lips part.
                            You hold her, and she clings to you, the two of you revelling in post-orgasmic bliss.
                            You pull your length from your partner's spunk-oozing twat.
                            It quickly grows flaccid, receding back into your cockslit after a job well done.
                        [/p]
                        [p]
                            The cyan kobold still seems to be a bit [i]out of it[/i].
                            Your arms growing slightly weary, you ask her if she can [i]stand[/i].
                            Shaking her head to snap herself back to reality, she plants a small kiss on your cheek.
                            "[color=cyan]I'm fine.[/color]"
                            The kobold takes a deep breath, seemingly clearing her mind further.
                            "[color=cyan]You can put me down now.[/color]"
                        [/p]
                        [p]
                            Like the gallantly knight you strive to be, you gently lower her legs down to the ground.
                            For a few moments, you happily stare into one another's eyes,
                            until, suddenly, your tranquil moment gets interrupted.
                        [/p]`;
        } else if (dick < SizeEnum.LARGER) {
            toReturn += `[p]
                            Taking a deep breath, you roll your hips forwards forwards, first parting the cyan kobold's slit with your tapered tip,
                            then outright [b]impaling[/b] her on your manhood,
                            which is big enough to slightly [b]stretch[/b] her cunt just a little further than what it's used to,
                            filling her up to the [b]brim[/b] with kobold cock.
                            You let out a loud, pleasured sigh, your entire body shuddering as you feel her warmth envelop you in every way possible.
                        [/p]
                        [p]
                            You're glad to hear the enjoyment is mutual.
                            Your partner [b]squeals[/b] in [b]ecstasy[/b], her legs wrapping around you, her entire body [b]clamping[/b] on to yours.
                            While tight to begin with, her innards suddenly grip your member like a [b]vice[/b],
                            a [b]torrent[/b] of feminine fluids [b]gushing[/b] from her loins, [b]soaking[/b] your groin.
                            Did she just [i]cum[/i]?
                            Oh man.
                            [i]You're good[/i].
                        [/p]
                        [p]
                            But you're [b]far[/b] from done.
                            Even with the cyan kobold mid-climax, you slowly start to rock back and forth, shallowly pumping your member in and out of her,
                            sending spikes of pure [b]pleasure[/b] coursing down her spine, adding to her orgasm, making her [b]wild[/b] with [b]lust[/b].
                            A loud, drawn-out moan echoes throughout the alleyway.
                            The kobold steadies herself against the wall.
                            She stares longingly into your eyes, with a needy look that [b]screams[/b] how much she [b]wants[/b] this.
                            "[color=cyan]Stop teasing and [b]fuck[/b] me.[/color]"
                            Her wish is your command.
                        [/p]
                        [p]
                            With both your hands, you grab onto your partner's plump, scaly behind, holding her in place while copping a few lustful squeezes.
                            Then, after pulling your member out nearly [i]all[/i] the way, you [b]ram[/b] it back inside of her sopping wet cunt,
                            and immediately begin to [b]properly[/b] pound her.
                            Your fierce, rapid thrusts make the kobold [b]cry out[/b] in delight, her excitement [b]dripping[/b] to the ground,
                            her heated moans accompanied with the wet slapping of your scaly hips against her drenched slit.
                        [/p]
                        [p]
                            "[color=cyan]Incredible![/color]" the lust-addled kobold shouts out in pleasure, her satisfaction only serving to spur you on further.
                            You let out a lusty growl, nuzzling your snout against her smooth neck, giving her smooth scales a single lick.
                            In response, she lets her body sink down on you every time you thrust upwards, humping back against your rocking motions,
                            using her legs to spear herself deeper onto your manhood.
                            While she's already finished once, you can feel her getting close [i]again[/i].
                        [/p]
                        [p]
                            Feeling your own end approaching, you break your steady rhythm in favor of a series of more [b]carnal[/b] thrusts.
                            Your slightly [i]rougher[/i] treatment seemingly hits [b]all[/b] of the right buttons,
                            the cyan kobold's eyes rolling into the back of her head, and her tongue lolling out of her mouth,
                            as a second mind-wracking climax rocks her body.
                            "[color=cyan][b]Yes![/b][/color]" she screams out, pulling you even [b]closer[/b] to her using her legs.
                            Like a woman possessed, the kobold begins to [b]ride[/b] your cock as best as she can, her cunt practically [b]milking[/b] your entire length.
                            "[color=cyan]Keep going! Fuck me! [b]Breed me![/b][/color]"
                        [/p]
                        [p]
                            You couldn't deny her even if you wanted to.
                        [/p]
                        [p]
                            With one fierce final [b]thrust[/b], you slam yourself into your partner's quivering slit.
                            While she continues to ride you, the tapered tip of your member [b]erupts[/b] in orgasm,
                            your internal balls churning out rope after rope of thick, fertile semen deep inside of the cyan kobold's slit,
                            painting her inner walls a pearly white, your cum soon starting to leak out alongside her own arousal,
                            all the while she continues to coax [b]more[/b] and [b]more[/b] from your trembling member.
                        [/p]
                        [p]
                            Only after your seed is fully spent, does the other kobold seem to somewhat [i]regain[/i] her senses.
                            As her intense riding slows down to a crawl, you pull your length from her spunk-oozing twat.
                            You hold her, and she clings to you, the two of you revelling in post-orgasmic bliss.
                        [/p]
                        [p]
                            The cyan kobold still seems to be a bit [i]out of it[/i].
                            Your arms growing slightly weary, you ask her if she can [i]stand[/i].
                            Shaking her head to snap herself back to reality, she plants a small kiss on your cheek.
                            "[color=cyan]I'm fine.[/color]"
                            The kobold takes a deep breath, seemingly clearing her mind further.
                            "[color=cyan]You can put me down now.[/color]"
                        [/p]
                        [p]
                            Like the gallantly knight you strive to be, you gently lower her feet down to the ground.
                            Her legs still seem a bit shaky, but that's to be expected after creaming her brains out [b]twice[/b].
                            For a few moments, you happily stare into one another's eyes,
                            until, suddenly, your tranquil moment gets interrupted.
                        [/p]`;
        } else {
            toReturn += `[p]
                            Even as she speaks, her fingers begin to run up and down your length.
                            As her words fade to silence, the kobold starts to outright stroke you,
                            jerking you off while your member hardens under her touch, rapidly growing to its [b]full[/b] size.
                            "[color=cyan]I-I'll just...if that's okay with you...[/color]"
                            You hush her mumbling, saying that she can do [i]whatever[/i] she feels comfortable with.
                        [/p]
                        [p]
                            She seems hesitant to take up your [i]invitation[/i] to do as she pleases.
                            In order to coax her out of her shell, you begin to egg her on a bit.
                            You lean back against the alley-wall, and tell her her hand feels [b]great[/b].
                            To emphasize how [i]good[/i] she's doing, you let out a long, pleasured sigh.
                            Your encouragement seems to be working.
                            Your partner picks up the pace, her other hand drifting towards her own groin, where she starts to openly play with herself.
                        [/p]
                        [p]
                            She's [b]definitely[/b] growing more comfortable.
                            The precum oozing from the tapered tip of your member slickens her palm, smoothing the motions of her hand.
                            After a solid half-minute of slathering your own arousal across your shaft, the cyan kobold suddenly lets go of your length,
                            instead bringing her precum-covered palm to her snout,
                            loudly [b]sniffing[/b] the scent of your cock while smearing the sticky fluid all over her nose,
                            her loins visibly dampening in the process, her eyes rolling into the back of her head.
                        [/p]
                        [p]
                            Your partner drops to her knees.
                            She's had a small [i]whiff[/i], but she needs [b]more[/b].
                            Sliding a finger inside of her folds, she presses the front-end of her muzzle against your length,
                            nuzzling it while audibly huffing your masculine fragrance, clearly getting off to it.
                            Her slickened palm returns to your cock, and resumes its stroking,
                            coaxing another bauble of pre from the tip, which droops down onto her face,
                            rolling along her scales until it reaches her lips, where she scoops it up with her tongue,
                            your potent taste adding a new [b]dimension[/b] to her rapidly growing needs.
                        [/p]
                        [p]
                            You break the silence between the two of you, by asking if she's enjoying herself.
                            The kobold seems startled, as though she'd somehow forgotten you were even [b]there[/b].
                            "[color=cyan]Y-yes! You, uh. [sub]You smell [b]really[/b] good.[/sub][/color]"
                            While she sort of [i]mumbled[/i] that last part, you understood every word she said.
                            You tell her that if she likes your scent, why not try the [b]taste[/b]?
                            And that's exactly what she does.
                        [/p]
                        [p]
                            Extending her tongue, the cyan kobold stares up at your face while giving your length a single,
                            long lick all the way from the base of your member to the very tip.
                            A broad grin forms on her visage as she notices the absent-minded look of pure, relaxing [b]pleasure[/b] plastered across your face.
                            Her grin fades fast, however, as she rolls her tongue back into her muzzle,
                            your taste spreading throughout her mouth, reminding her of her own need.
                            Like a feral dog drinking from a bowl of water, she begins to [b]lap[/b] at your shaft, planting kisses and licks all over.
                            A second finger finds its way inside of her slit, a small puddle of female arousal forming on the ground in between her legs.
                        [/p]
                        [p]
                            While getting jerked off and licked at the same time is pleasurable, you need [b]more[/b].
                            With one of your hands, you softly push your scaly partner's head backwards,
                            guiding her out from underneath your member, just so you can lower your cock to her lips.
                            She quickly gets the hint, lapping off a freshly formed droplet of precum before taking the tapered tip of your member into her mouth,
                            endearingly suckling on one end while diligently stroking at the base.
                        [/p]
                        [p]
                            Moving your hand to the back of her head, you gently pull the kobold towards you again,
                            urging her to take you [b]deeper[/b] than just the tip.
                            Without needing any further encouragement, she does just that, slowly starting to bob up and down your length,
                            taking inch after inch in her obviously skilled mouth.
                            Before long, she's sucking you off properly,
                            her tongue swirling around your girth while her hand continues to work the lower half of your cock.
                            She's done this before, and it shows.
                        [/p]
                        [p]
                            While she doesn't quite use her [i]throat[/i] to pleasure you, she doesn't [b]need[/b] to.
                            She's [b]good[/b] at giving head, and she knows it.
                            But what arouses you far more than her hand, her mouth or the sight of her pretty little tits,
                            is the glaring fact that she, too, is [b]getting off[/b] to this,
                            the only thing wetter than the inside of her mouth being her audibly [b]soaked[/b] cunt,
                            which she continues to toy with.
                        [/p]
                        [p]
                            [b]Both[/b] of her hands -- the one on your shaft and the one in between her legs -- pick up the pace.
                            You're already nearing your end.
                            Just [b]thinking[/b] about [b]ramming[/b] your oversized prick into her overly tight,
                            sopping wet slit brings you [b]dangerously[/b] close to blowing your load.
                            Subconsciously, you begin to rock your hips back and forth, slightly thrusting in and out of her mouth without hampering her own movements.
                        [/p]
                        [p]
                            As the hormone-laden feminine fragrance of your partner's growing pool of excitement reaches your nose, you can no longer contain yourself.
                            Your hips begin to move on their own, humping away at the other kobold's face.
                            Letting out a loud grunt, you lean your head back in ecstasy, your member quivering and pulsing as a first wave of thick,
                            heady spunk floods from your tapered tip, only to be quickly downed by the cyan kobold.
                            After effortlessly swallowing the second and even the third squirt, she finally pulls back,
                            just so the remainder of your intense orgasm can cover her face, her horns, her bare chest, and most importantly: her lolling tongue.
                        [/p]
                        [p]
                            At the end of your climax, you find yourself panting loudly, your spent shaft rapidly receding back into your cockslit.
                            Your partner, covered in a considerable amount of jizz, is fervently fingerfucking herself,
                            trying, but failing to get off to the taste of your cum alone.
                            The entire alley reeks not only of your masculine seed, but it [b]stinks[/b] of female [b]need[/b],
                            the two scents combining into an arousing aroma of pure [b]sex[/b].
                        [/p]
                        [p]
                            Seeing you've recovered from blowing your load, the cyan kobold rises to her feet.
                            Turning around, she bends over and spreads her plump buttcheeks,
                            presenting her absolutely [b]soaked[/b] slit to you, her thighs sticky with strings of excitement halfway down to her knees.
                            She's [b]obviously[/b] needy, but with the tip of your member disappearing back into your cockslit,
                            your sex-fueled mind can't possibly fathom what she wants [b]you[/b] to do about her little [i]problem[/i].
                        [/p]
                        [p]
                            Noticing your confusion, she decides to [i]nudge[/i] your mind onto the right train of thoughts.
                            "[color=cyan]Time to return the favor, [i]big boy[/i]. Don't leave me hanging.[/color]"
                            Oh.
                            [b]Oh![/b]
                            A grin forms on your face once more.
                            She wants you to eat her out!
                            Why didn't she just say so to begin with?!
                        [/p]
                        [p]
                            Of course you'd [b]gladly[/b] finish her off.
                            It's the least you can do, after all she's done for you.
                            "[color=cyan]Get to licking![/color]" the kobold interrupts your inadvertent stalling,
                            yet right before you manage to drop down to your knees, your heated moment gets interrupted.
                        [/p]`;
        }

        if (virgin) {
            toReturn += `[p]
                            [c]
                                [b]
                                    You are no longer a virgin.
                                [/b]
                            [/c]
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.maleSex.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\male\MorningThreeMaleSexFinish.js

// scenes\Varanar\DayThree\MorningThree\male\MorningThreeMaleSexStart.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeMaleSexStart(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.maleSex;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreeMaleSexFinish(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.morning.fuckedFemaleKobold = true;
        var dick = ctrl.misc.dick();
        var clothed = !ctrl.misc.nude();
        var toReturn = `[p]
                            All of the trials and tribulations you've faced these past few days are forgotten in an instant.
                            In the arms of a stranger you have only just met, your worries melt away.
                            Feeling her slight shivering, her heartbeat and her warm breath against your scales only serves to arouse you further.
                            The both of you mutually struggle and fumble to 
                            ${clothed ? `shed whatever clothes you're wearing` : `take off her meagre clothing`},
                            inspecting one another with wandering hands and eyes alike.
                        [/p]
                        [p]
                            Your fingers dawdle a little as they wander up to her chest, groping her and carefully rubbing her nipples,
                            each hearbeat and every soft, pleasured whimper emboldening you, inviting you to explore her body further.
                            Your other hand wanders down to her crotch, and you begin to stroke her already wet folds,
                            one of your fingers probing its way inside to [b]bask[/b] in her [i]warmth[/i].
                            You can tell that she's fairly [i]loose[/i].
                            She's undeniably [i]slept around[/i], for sure, but in a big city like this, who hasn't?
                            As you slide a second finger in, you make some [sneak]hopefully[/sneak] romantic comments about how [i]pretty[/i] she is.
                            All you get in response is a teasing giggle.
                        [/p]`;
        if (dick < SizeEnum.AVERAGE) {
            toReturn += `[p]
                            Meanwhile, her own hands embrace you, pulling you closer as she presses her back against the alley-wall.
                            A set of fingers gingerly wrap around your still growing shaft.
                            Feeling up your package, the kobold lets out a lustful and somewhat [i]amused[/i] hum.
                            "[color=cyan]Not the [i]largest[/i] I've ever had, but...it'll do.[/color]"
                        [/p]
                        [p]
                            Your remove your fingers from her cunt as she guides your member to her slit.
                            You brace yourself against the wall behind her, preparing to give her your all.
                            Your member rests against her [b]dripping wet[/b] folds.
                            She [b]begs[/b] you to not keep her waiting.
                            You don't intend to.
                        [/p]`;
        } else if (dick < SizeEnum.LARGER) {
            toReturn += `[p]
                            Meanwhile, her own hands embrace you, pulling you closer as she presses her back against the alley-wall.
                            A set of fingers gingerly wrap around your still growing shaft.
                            Feeling up your package, the kobold lets out a lustful and somewhat [i]surprised[/i] hum.
                            "[color=cyan]Oh...! That'll do just fine.[/color]"
                        [/p]
                        [p]
                            Your remove your fingers from her cunt as she guides your member to her slit.
                            You brace yourself against the wall behind her, preparing to give her your all.
                            Your member rests against her [b]dripping wet[/b] folds.
                            She [b]begs[/b] you to not keep her waiting.
                            You don't intend to.
                        [/p]`;
        } else {
            toReturn += `[p]
                            Meanwhile, her own hands embrace you, pulling you closer as she presses her back against the alley-wall.
                            A set of fingers gingerly wrap around your still growing shaft.
                            You feel a small jolt of pride as those fingers [b]barely[/b] manage to wrap around your considerable girth,
                            the cyan kobold's eyes widening in shock as she looks down to see what she's working with.
                            "[color=cyan]O-oh...oh my.[/color]"
                        [/p]
                        [p]
                            You can't help but grin like a smug jerk.
                            You've clearly impressed the kobold.
                            Heck, you might have even [i]scared[/i] her a bit with the sheer [b]size[/b] of your member.
                            You coyly ask her if something's wrong.
                            She takes a deep breath.
                            "[color=cyan]I-I don't think I can, you know. [i]Take[/i] this. I'm flattered, but what I'm saying is...[/color]"
                            The kobold takes a moment to consider her choice of words,
                            trying but failing to come up with a more [i]tactful[/i] way of saying what she wants to get off her chest.
                            With a small sigh and a rolling of her eyes, she bluntly stammers out:
                            "[color=cyan]Your dick is [b]too big[/b].[/color]"
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\male\MorningThreeMaleSexStart.js

// scenes\Varanar\DayThree\MorningThree\proposition\MorningThreePropositionAgreeSelector.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreePropositionAgreeSelector(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.propositionAgree;

    var toFail = new MorningThreeMaleSexStart(ctrl);
    var toSuccess = new MorningThreeFemaleForeplay(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.dick() === 0;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\proposition\MorningThreePropositionAgreeSelector.js

// scenes\Varanar\DayThree\MorningThree\proposition\MorningThreePropositionAgree.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreePropositionAgree(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreePropositionChoice.AGREE;
    var reqController = ctrl.proposition;
    var subController = ctrl.propositionAgree;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreePropositionAgreeSelector(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var hers = dick ? `hers` : `his`;
        var toReturn = `[p]
                            Your heart is thumping like crazy.
                            Your hand shivers in ${hers} and your fingers grow sweaty.
                            You hope ${he} doesn't mind.
                            As you eye ${his} alluring frame up and down, you can feel your loins pleasantly stirring.
                            You've [i]rarely[/i] felt this way before.
                            Without a moment's hesitation, you stammer out something remotely affirmative in response to her question.
                            ${Her} grin widens further, and ${he} leads you away from the crowd.
                        [/p]
                        [p]
                            Not that you [i]need[/i] much leading.
                            Wanting this every bit as much as ${he} does, you match ${his} brisk pace, walking alongside your new friend to a spot of privacy.
                            You carelessly bump into some people along the way, too concerned with one another to pay much attention to where you're going,
                            but eventually you manage to reach the alleyway.
                            A place where the sun has yet to reach.
                            It's rather dark, and somewhat cold,
                            but that last part's all but forgotten as your new partner practically [b]jumps[/b] on top of you the moment you're somewhat out of sight.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.propositionAgree.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\proposition\MorningThreePropositionAgree.js

// scenes\Varanar\DayThree\MorningThree\proposition\MorningThreeProposition.js

/**
 * Proposition section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeProposition(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.proposition;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new MorningThreeDeclineSex(ctrl);
    var secondNext = new MorningThreePropositionAgree(ctrl);

    var firstChoice = MorningThreePropositionChoice.DECLINE;
    var secondChoice = MorningThreePropositionChoice.AGREE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var Him = dick ? `Her` : `Him`;
        var her = Him.toLowerCase();
        var toReturn = ``;
        if (ctrl.eggs.get() === MorningThreeEggsChoice.ONE) {
            toReturn += `[p]
                            You [i]could[/i] use some cheering up after such a hellish night,
                            and you don't want to look [i]weird[/i] in front of your new friend.
                            Plus it'll be nice to get some [i]retribution[/i] on a non-kobold.
                            After all, knights are supposed to [b]punish[/b] evildoers,
                            and the bird on stage must have done [i]something[/i] to earn his place in the spotlight.
                            With an abundance of reasons to throw, and none to hold back, you decide to let loose and pelt away.
                        [/p]
                        [p]
                            With a small grunt, you chuck your first egg at the trapped avian.
                            The rotten would-be-breakfast soars through the air with a slight curve, before hitting the bird squarely in the face!
                            It cracks open, the disgusting goop contained within splattering all over the man's feathery face, most of the stuff flowing down along his beak.
                            Some of it manages to make its way into his mouth.
                            He gags, sputters and spits.
                            The cyan kobold burst out into laughter, and you join in.
                            "[color=cyan]Rotten like your wife's clutches![/color]"
                            ${She} turns to you.
                            "[color=cyan]Hit'em with the other one![/color]"
                        [/p]
                        [p]
                            You decline.
                            The man you're throwing at is [i]helpless[/i].
                            Bullying him may be fun, but you know first-hand that being on the receiving end is everything but.
                            Doing this, throwing these eggs, makes you no better than the humans and their ilk.
                            And while you [i]did[/i] let yourself go [i]a little[/i], you refuse to stoop yourself down to their level any further.
                        [/p]
                        [p]
                            You drop the rotten egg to the ground, and turn to face your new friend.
                            ${She} seems to respect your decision, an impressed look on ${his} face.
                            In order to lighten the mood a bit, you mention that throwing the first egg [b]was[/b] a lot more fun than you thought it'd be.
                            A small smile forms on ${his} face.
                            You decide to keep up the smalltalk.
                        [/p]
                        [p]
                            You mention you had a really nice time looking at the stars yesterday night,
                            but your visit to Varanar has been kind of going downhill ever since then.
                            Until you met ${her}, that is!
                            The cyan kobold smirks.
                            ${She} reaches out, grabbing your hand, tugging you towards a nearby alley.
                            "[color=cyan]Wanna see some stars? I can help you with that.[/color]"
                            ${She} winks suggestively.
                        [/p]
                        ${global.ui.condSucc(`Charisma greater than 5`)}
                        [p]
                            ${Her} expression, ${his} gentle touch, the innuendo, the wordplay and the privacy ${he}'s seeking,
                            all make ${his} intentions painfully obvious.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You [i]could[/i] use some cheering up after such a hellish night,
                            and you don't want to look [i]weird[/i] in front of your new friend.
                            Plus it'll be nice to get some [i]retribution[/i] on a non-kobold.
                            After all, knights are supposed to [b]punish[/b] evildoers,
                            and the bird on stage must have done [i]something[/i] to earn his place in the spotlight.
                            With an abundance of reasons to throw, and none to hold back, you decide to let loose and pelt away.
                        [/p]
                        [p]
                            With a small grunt, you chuck your first egg at the trapped avian.
                            The rotten would-be-breakfast soars through the air with a slight curve, before hitting the bird squarely in the face!
                            It cracks open, the disgusting goop contained within splattering all over the man's feathery face, most of the stuff flowing down along his beak.
                            Some of it manages to make its way into his mouth.
                            He gags, sputters and spits.
                            The cyan kobold burst out into laughter, and you join in.
                            "[color=cyan]Rotten like your wife's clutches![/color]"
                            ${She} turns to you.
                            "[color=cyan]Hit'em with the other one![/color]"
                        [/p]
                        [p]
                            You don't comply right away.
                            Instead, you hold on to the second egg for a few moments,
                            putting all your [b]fear[/b] and [b]anger[/b] and [b]frustration[/b] into it,
                            as if enchanting it with every [b]semblance[/b] of your misfortune you've [b]ever[/b] had to deal with,
                            all the while [i]aiming[/i] your shot to make sure it'll hit its mark.
                            Only after pouring your very heart and soul into the impending throw, with a surge of great fury and determination, do you let loose.
                        [/p]
                        [p]
                            Like the last, the rotten egg flies gracefully through the air towards its avian target.
                            It seems you may have spent just a little bit [i]too[/i] long aiming, however.
                            Unlike your nigh-perfect first pitch, the second projectile misses the bird's face completely,
                            instead hitting the pillory's confining wood.
                            Oh well, at least some of the rancid splatter still made it to the side of his face.
                        [/p]
                        [p]
                            That second throw made you feel a [b]lot[/b] better.
                            It's as though a weight has been lifted off of your shoulders.
                            You turn towards your new friend, and the two of you wordlessly exchange sadistic grins.
                            You thank ${her} for the eggs.
                            You've rarely had this much fun.
                        [/p]
                        [p]
                            You mention you had a really nice time looking at the stars yesterday night,
                            but your visit to Varanar has been kind of going downhill ever since then.
                            Until you met ${her}, that is!
                            The cyan kobold smirks.
                            ${She} reaches out, grabbing your hand, tugging you towards a nearby alley.
                            "[color=cyan]Wanna see some stars? I can help you with that.[/color]"
                            ${She} winks suggestively.
                        [/p]
                        ${global.ui.condSucc(`Charisma greater than 5`)}
                        [p]
                            ${Her} expression, ${his} gentle touch, the innuendo, the wordplay and the privacy ${he}'s seeking,
                            all make ${his} intentions painfully obvious.
                        [/p]`;
        }
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Decline</button>`;

        var tooltip = global.parser.parse(`After last night, you're not in the mood.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var exertionCheck = ctrl.misc.exertion() >= 60;
        var disabled = subController.has() || exertionCheck ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Agree</button>`;

        var tooltipStart = ctrl.misc.dick() ? `She [i]is[/i] pretty cute.` : `He [i]is[/i] kind of handsome.`;
        var tooltip = global.parser.parse(`${tooltipStart} Accept the kobold's offer.`);

        if (exertionCheck) {
            tooltip = global.parser.parse(`You are too sexually exerted to take up the kobold's offer.`);
        }

        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.proposition.change();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var MorningThreePropositionChoice = {
    DECLINE: 1,
    AGREE: 2
};

// END scenes\Varanar\DayThree\MorningThree\proposition\MorningThreeProposition.js

// scenes\Varanar\DayThree\MorningThree\throw\MorningThreeThrowBoth.js

/**
 * Throw both section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeThrowBoth(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeEggsChoice.BOTH;
    var reqController = ctrl.eggs;

    var toFail = new MorningThreeRobbed(ctrl);
    var toSuccess = new MorningThreeProposition(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.charisma() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\throw\MorningThreeThrowBoth.js

// scenes\Varanar\DayThree\MorningThree\throw\MorningThreeThrowOne.js

/**
 * Throw one section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeThrowOne(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = MorningThreeEggsChoice.ONE;
    var reqController = ctrl.eggs;

    var toFail = new MorningThreeThievery(ctrl);
    var toSuccess = new MorningThreeProposition(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.charisma() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\DayThree\MorningThree\throw\MorningThreeThrowOne.js

// scenes\Varanar\DayThree\MorningThree\arrival\MorningThreeEggs.js

/**
 * Arrival section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeEggs(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.knights;
    var subController = ctrl.eggs;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new MorningThreeDeclineEggs(ctrl);
    var secondNext = new MorningThreeThrowOne(ctrl);
    var thirdNext = new MorningThreeThrowBoth(ctrl);

    var firstChoice = MorningThreeEggsChoice.DECLINE;
    var secondChoice = MorningThreeEggsChoice.ONE;
    var thirdChoice = MorningThreeEggsChoice.BOTH;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var health = ctrl.misc.health();
        var dick = ctrl.misc.dick();
        var She = dick ? `She` : `He`;
        var he = She.toLowerCase();
        var Her = dick ? `Her` : `His`;
        var his = Her.toLowerCase();
        var toReturn = `[p]
                            Arriving finally at the plaza, you notice a commotion at the stage you used as a resting spot last night.
                            A crowd consisting of all walks of life has formed in front of it.
                            There's a few noble-looking dragons, a lot of humans, pigs, dogs, foxes, wolves, cats, a couple of robed birds,
                            orcs taking a break from their manual labor jobs, an entire dwarven family, a bunch of kobolds and even an elf.
                            And that's just the ones you can see on the outskirts of the mob.
                        [/p]
                        [p]
                            Circling around, you see what the entire ordeal is all about.
                            A pillory has been erected on top of the stage.
                            Confined within the stockade is an avian, some kind of hawk, or [i]maybe[/i] a falcon.
                            You aren't sure which.
                            The cheering and insult-hurling crowd is throwing a myriad of rotten fruits and vegetables at the feathery criminal's beaked face.
                            Guards are standing by to make sure nobody [i]helps[/i] the offender.
                            Or maybe they're just there to prevent a [i]riot[/i].
                        [/p]
                        [p]
                            You don't know how to feel about this display.
                            This must be [i]awful[/i] for the bird.
                            You don't know [i]what[/i] he did, and whether or not he even [i]did[/i] what they [i]think[/i] he did.
                            He looks like an okay guy to you, but looks can be deceiving.
                            Maybe he [b]killed someone[/b] or burned down an orphanage or stole from the church or something.
                            Maybe he [b]does[/b] deserve this.
                        [/p]
                        [p]
                            You seem to have caught the attention of one of the kobolds in the crowd.
                            ${She} enthusiastically waves you over.
                            While wary of someone you've never seen before in the middle of a crowd suddenly [i]taking interest[/i] in you,
                            curiosity eventually trumps your initial suspicion.
                            After a bit of dawdling, you approach the cyan-colored kobold.
                        [/p]
                        [p]
                            Standing at about the same height as you, maybe just a [i]little bit[/i] smaller,
                            the kobold that beckoned you over is a rather skinny-looking thing with a big smile across ${his} face.
                            A resident of Varanar, 
                            or so you deduce by the myriad of tiny bruises and scabbed-over cut-wounds${health < 70 ? `, not unlike your own,` : ``}
                            lining ${his} scantily-clad upper body.
                            Still, despite ${his} minor wounds, ${he}'s pretty ${dick ? `cute` : `handsome`}.
                        [/p]
                        [p]
                            You ask the kobold what's going on.
                            What did the bird do to deserve this?
                            ${She} simply shrugs in response.
                            "[color=cyan]Who cares? It's not a kobold. Lighten up. Have some fun![/color]"
                            Before you can refuse, ${he} shoves a pair of smelly ${global.kobold.inventory.createItemWithTooltip(new RottenEgg())}s into your hands.
                            Confused, you look at the spoilt foodstuffs for a few moments, before casting a quizzical gaze your new friend's way.
                            "[color=cyan]Well? What are you waiting for? Throw![/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Decline</button>`;

        var tooltip = global.parser.parse(`These eggs [b]stink[/b]. You don't want to throw them.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Throw one</button>`;

        var tooltip = global.parser.parse(`Hurl an egg to please the kobold.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Throw both</button>`;

        var tooltip = global.parser.parse(`[b]Free[/b] tossing material? Hurl away!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("varanar_generic");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var MorningThreeEggsChoice = {
    DECLINE: 1,
    ONE: 2,
    BOTH: 3
};

// END scenes\Varanar\DayThree\MorningThree\arrival\MorningThreeEggs.js

// scenes\Varanar\DayThree\MorningThree\arrival\MorningThreeKnights.js

/**
 * Arrival section of Varanar, Morning Three.
 * @param {MorningThreeController} controller The Morning Three controller.
 */
function MorningThreeKnights(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new MorningThreeEggs(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dragonessRelation = global.kobold.history.shared.dragonessPrincess.relation;
        var toReturn = `[p]
                            That was the worst night you think you've [b]ever[/b] had.
                            With dawn breaking, the streets are slowly filling with people.
                            You're so sleepy, you could drop to the ground and instantly start slumbering,
                            but after last night, you fear what would happen to you if you did.
                            Your tired feet carry you back towards the plaza.
                        [/p]
                        [p]
                            Nearing the grand avenue, you hear the distinct noise of a bunch of horses trotting down the road.
                            Mounted guardsmen?
                            This early?
                            And [i]so many[/i] of them?
                            No way.
                            Your inquisitiveness energizing you, you jog towards the display.
                            Arriving at the avenue proper, you see a column of heavily armed and armoured [b]mounted dragon knights[/b] strutting into town.
                            At the head of the military formation rides a blue-scaled maiden, the only one [i]not[/i] wearing armor.
                        [/p]`;

        if (dragonessRelation === DragonessPrincessEnum.UNMET) {
            toReturn += `[p]
                            The column passes by without paying you any mind.
                            They head past the marketplace, presumably towards the keep.
                            You don't know much about politics, but you doubt a group of dragon knights showing up at the human capital is a [i]good[/i] thing.
                            Unless they were just escorts, of course.
                            The roads [i]are[/i] pretty unsafe these days, especially for a dragoness as pretty as the one you just saw.
                        [/p]`;
        }

        if (dragonessRelation === DragonessPrincessEnum.UNHELPED) {
            toReturn += `[p]
                            Wait a minute.
                            Isn't the [i]maiden[/i] that one dragoness you saw in the woods?
                            The one you kind of [i]didn't help[/i] when she was getting [b]raped[/b].
                            You can't quite recall whether or not she saw you there, but you believe it'd be best for the both of you if she [i]doesn't[/i] see you now.
                            Y'know, to prevent any [i]unfortunate recollections[/i] that might make her want to tell her knightly friends to [b]cut you down[/b].
                        [/p]
                        [p]
                            Luckily for you, the column passes by without paying you any mind.
                            They head past the marketplace, presumably towards the keep.
                            Part of you is glad to see that the dragoness made it out of the woods alive.
                            But then again, if she [i]does[/i] recall you being present during her [i]unpleasant run-in[/i],
                            then perhaps her being in the same town as you might be a [b]bad[/b] thing.
                        [/p]`;
        }

        if (dragonessRelation === DragonessPrincessEnum.HELPED) {
            toReturn += `[p]
                            Wait a minute.
                            Isn't the [i]maiden[/i] that one dragoness you saw in the woods?
                            You know, the one you couldn't quite [i]prevent[/i] from getting abused by that kobold gang.
                            At least you helped her out afterwards.
                            Surely that counts for something, right?
                        [/p]
                        [p]
                            You try to draw the lady's attention, but the column passes by without paying you any mind,
                            stepping horse-hooves preventing you from getting [i]too[/i] close.
                            They head past the marketplace, presumably towards the keep.
                            At least you're glad to have seen the dragoness again.
                            Now you know she made it out of the woods alive, if not completely [i]untouched[/i].
                        [/p]`;
        }

        if (dragonessRelation === DragonessPrincessEnum.SLOPPY_SECONDS) {
            toReturn += `[p]
                            Wait a minute.
                            Isn't the [i]maiden[/i] that one dragoness you, uh, [i]met[/i] in the woods?
                            The one you kind of [i]didn't help[/i] when she was getting [b]raped[/b].
                            And then you had some [i]fun[/i] with her cum-oozing body while she was nigh-catatonic.
                            Oh dear.
                            You can't quite recall whether or not she was [i]conscious[/i] enough to remember your face,
                            but you believe it'd be best for the both of you if she [i]doesn't[/i] see you now.
                            Y'know, to prevent any [i]unfortunate recollections[/i] that might make her want to tell her knightly friends to [b]cut you down[/b].
                        [/p]
                        [p]
                            Luckily for you, the column passes by without paying you any mind.
                            They head past the marketplace, presumably towards the keep.
                            Part of you is glad to see that the dragoness made it out of the woods alive.
                            But then again, if she [i]does[/i] recall what you did,
                            then perhaps her being in the same town as you might be a [b]bad[/b] thing.
                        [/p]`;
        }

        if (dragonessRelation === DragonessPrincessEnum.SAVED) {
            var dragonessSavedBy = global.kobold.history.shared.dragonessPrincess.savedBy;
            var needle = ctrl.misc.hasMisc(ItemEnum.GOLDENNEEDLE);
            var savedBy = ``;
            if (dragonessSavedBy === DragonessPrincessSavedBy.PAYING_GANG) {
                savedBy = `You paid off the gang harassing her.`;
            }
            if (dragonessSavedBy === DragonessPrincessSavedBy.SCARY_NOISE) {
                savedBy = `You scared off the gang harassing her.`;
            }
            toReturn += `[p]
                            Wait a minute.
                            Isn't the [i]maiden[/i] that one dragoness you [i]saved[/i] in the woods?
                            ${savedBy}
                            It is!
                            Saphira was her name, or so you believe.
                            She even gave you a
                            ${global.kobold.inventory.createItemWithTooltip(new GoldenNeedle())}${needle ? `, which you still have,` : ``} as a reward!
                            You're glad to see she made it out of the forest alive and well!
                        [/p]
                        [p]
                            You try to draw the lady's attention, but the column passes by without paying you any mind,
                            stepping horse-hooves preventing you from getting [i]too[/i] close.
                            They head past the marketplace, presumably towards the keep.
                            You feel kind of [i]bummed out[/i], first having been [i]declined[/i] as an escort,
                            and now getting flat-out [i]ignored[/i] by the princess,
                            but you suppose a little kobold like yourself doesn't matter much to a high and mighty woman like her.
                        [/p]`;
        }

        if (dragonessRelation === DragonessPrincessEnum.RAN_AWAY) {
            var tenpaw = global.kobold.history.shared.tenPawGang.relation;
            var raped = tenpaw === TenPawGangEnum.RELUCTANT_GANGBANG || tenpaw === TenPawGangEnum.GANGRAPE;
            toReturn += `[p]
                            Wait a minute.
                            Isn't the [i]maiden[/i] that one dragoness you [i]kind of[/i] saved in the woods?
                            You know, the one that [i]ran away[/i] while you were discussing things with the kobold gang.
                            At least you [i]helped her[/i] escape.
                            Surely that counts for something, right?
                        [/p]
                        [p]
                            You try to draw the lady's attention, but the column passes by without paying you any mind,
                            stepping horse-hooves preventing you from getting [i]too[/i] close.
                            They head past the marketplace, presumably towards the keep.
                            At least you're glad to have seen the dragoness again.
                            Now you know she made it out of the woods alive${raped ? `, in part due to the [b]hell[/b] you went through` : ``}.
                        [/p]`;
        }

        toReturn += `   [p]
                            You follow the trail of the knights, not to chase after them, but simply to make your way back to the town square.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = ctrl.knights.continued() ? "disabled" : "";
        var isSelected = ctrl.knights.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.knights.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        ctrl.bgm.loop();
        global.ui.applyTheme("varanar_morning");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.knights.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        next.registerListeners();

        if (ctrl.knights.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\MorningThree\arrival\MorningThreeKnights.js

// scenes\Varanar\DayThree\MorningThree\MorningThreeController.js

/**
 * Controller for the Varanar Morning Three scene.
 * @param {string} sceneId The id of the scene.
 */
function MorningThreeController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.nextSceneReady = true;
    this.nextSceneName = "VaranarSoupKitchen";
    this.nextFolderName = "Varanar/DayThree";

    this.misc = new MiscController(id);

    this.knights = new ChangeController(id, "knight", function () {
        var dragonessRelation = global.kobold.history.shared.dragonessPrincess.relation;

        if (dragonessRelation === DragonessPrincessEnum.UNMET) {
            global.kobold.status.mental.increase(5, "Awe-inspiring knights");
        }

        if (dragonessRelation === DragonessPrincessEnum.UNHELPED) {
            global.kobold.status.mental.increase(0, "Saw dragoness");
        }

        if (dragonessRelation === DragonessPrincessEnum.HELPED) {
            global.kobold.status.mental.increase(0, "Saw dragoness");
        }

        if (dragonessRelation === DragonessPrincessEnum.SLOPPY_SECONDS) {
            global.kobold.status.mental.increase(0, "Saw dragoness");
        }

        if (dragonessRelation === DragonessPrincessEnum.SAVED) {
            global.kobold.status.mental.increase(0, "Saw dragoness");
        }

        if (dragonessRelation === DragonessPrincessEnum.RAN_AWAY) {
            var tenpaw = global.kobold.history.shared.tenPawGang.relation;
            var raped = tenpaw === TenPawGangEnum.RELUCTANT_GANGBANG || tenpaw === TenPawGangEnum.GANGRAPE;
            if (raped) {
                global.kobold.status.mental.increase(0, "Saw dragoness");
            } else {
                global.kobold.status.mental.increase(10, "Saw dragoness");
            }
        }
    }, function () {
        var dragonessRelation = global.kobold.history.shared.dragonessPrincess.relation;
        if (dragonessRelation === DragonessPrincessEnum.UNMET) {
            return global.ui.change.buildNotification([
                new Change("Knight convoy", "Sanity", +5)
            ]);
        }

        if (dragonessRelation === DragonessPrincessEnum.UNHELPED) {
            return global.ui.change.buildNotification([
                new Change("Saw dragoness again", "Sanity", +5),
                new Change("She may remember you", "Sanity", -5)
            ]);
        }

        if (dragonessRelation === DragonessPrincessEnum.HELPED) {
            return global.ui.change.buildNotification([
                new Change("Saw dragoness again", "Sanity", +5),
                new Change("She ignored you", "Sanity", -5)
            ]);
        }

        if (dragonessRelation === DragonessPrincessEnum.SLOPPY_SECONDS) {
            return global.ui.change.buildNotification([
                new Change("Saw dragoness again", "Sanity", +5),
                new Change("She may remember you", "Sanity", -5)
            ]);
        }

        if (dragonessRelation === DragonessPrincessEnum.SAVED) {
            return global.ui.change.buildNotification([
                new Change("Saw dragoness again", "Sanity", +5),
                new Change("She ignored you", "Sanity", -5)
            ]);
        }

        if (dragonessRelation === DragonessPrincessEnum.RAN_AWAY) {
            var tenpaw = global.kobold.history.shared.tenPawGang.relation;
            var raped = tenpaw === TenPawGangEnum.RELUCTANT_GANGBANG || tenpaw === TenPawGangEnum.GANGRAPE;
            if (raped) {
                return global.ui.change.buildNotification([
                    new Change("Saw dragoness again", "Sanity", +5),
                    new Change("Bad memories", "Sanity", -5)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Saw dragoness again", "Sanity", +5),
                    new Change("Good memories", "Sanity", +5)
                ]);
            }
        }
    });

    this.eggs = new ChoiceController(id, "eggs");

    this.thievery = new ChangeController(id, "thievery", function () {
        global.kobold.status.mental.increase(5, "Stopped thief");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stopped thief", "Sanity", +5)
        ]);
    });

    this.keepEgg = new ChangeController(id, "keepEgg", function () {
        global.kobold.inventory.equipment.wear(new RottenEgg());
        global.kobold.status.mental.decrease(0, "Guard-sanctioned gangrape");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Bullied thief", "Sanity", +5),
            new Change("Poetic justice?", "Sanity", +5),
            new Change("Guard-sanctioned gangrape", "Sanity", -5),
            new Change("Fear of getting caught", "Sanity", -5),
            new Change("Kept egg", global.kobold.inventory.createItemWithTooltip(new RottenEgg(), true), +1)
        ]);
    });

    this.returnEgg = new ChangeController(id, "returnEgg", function () {
        global.kobold.status.mental.decrease(0, "Guard-sanctioned gangrape");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Moral highground", "Sanity", +5),
            new Change("Poetic justice?", "Sanity", +5),
            new Change("Guard-sanctioned gangrape", "Sanity", -5),
            new Change("Fear of getting caught", "Sanity", -5)
        ]);
    });

    this.throwEgg = new ChangeController(id, "throwEgg", function () {
        global.kobold.status.mental.decrease(0, "Guard-sanctioned gangrape");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Punished thief", "Sanity", +5),
            new Change("Poetic justice?", "Sanity", +5),
            new Change("Guard-sanctioned gangrape", "Sanity", -5),
            new Change("Fear of getting caught", "Sanity", -5)
        ]);
    });

    this.proposition = new ChangeController(id, "proposition", function () {
        if (that.eggs.get() === MorningThreeEggsChoice.ONE) {
            global.kobold.status.mental.increase(5, "Throwing an egg");
        } else {
            global.kobold.status.mental.increase(10, "Throwing eggs");
        }
    }, function () {
        if (that.eggs.get() === MorningThreeEggsChoice.ONE) {
            return global.ui.change.buildNotification([
                new Change("Threw an egg", "Sanity", +2),
                new Change("Had fun", "Sanity", +3)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Threw first egg", "Sanity", +2),
                new Change("Threw second egg", "Sanity", +5),
                new Change("Had fun", "Sanity", +3)
            ]);
        }
    });

    this.propositionAgree = new ChangeController(id, "propositionAgree", function () {
        global.kobold.status.mental.increase(5, "Imminent sex");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Imminent sex", "Sanity", +5)
        ]);
    });

    this.declineSex = new ChangeController(id, "declineSex", function () {
        global.kobold.status.mental.decrease(3, "Friend disappeared");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Threw first egg", "Sanity", +2),
            new Change("Threw second egg", "Sanity", +5),
            new Change("Made friend sad", "Sanity", -5),
            new Change("Friend disappeared", "Sanity", -5)
        ]);
    });

    this.maleSex = new ChangeController(id, "maleSex", function () {
        global.kobold.body.dickVirgin = false;
        var dick = that.misc.dick();
        if (dick < SizeEnum.AVERAGE) {
            global.kobold.status.mental.increase(20, "Had sex");
            global.kobold.status.stamina.decrease(5, "Had sex");
            global.kobold.status.exertion.increase(25, "Had sex");
        } else if (dick < SizeEnum.LARGER) {
            global.kobold.status.mental.increase(25, "Had sex");
            global.kobold.status.stamina.decrease(10, "Had sex");
            global.kobold.status.exertion.increase(30, "Had sex");
        } else {
            global.kobold.status.mental.increase(25, "Got a blowjob");
            global.kobold.status.exertion.increase(20, "Got a blowjob");
        }
    }, function () {
        var dick = that.misc.dick();
        if (dick < SizeEnum.AVERAGE) {
            return global.ui.change.buildNotification([
                new Change("Sex", "Sanity", +15),
                new Change("Came", "Sanity", +5),
                new Change("Fucking", "Stamina", -5),
                new Change("Sensual fucking", "Sexual exhaustion", +20, true),
                new Change("Orgasm", "Sexual exhaustion", +5, true)
            ]);
        } else if (dick < SizeEnum.LARGER) {
            return global.ui.change.buildNotification([
                new Change("Sex", "Sanity", +15),
                new Change("Came", "Sanity", +5),
                new Change("Fucked partner silly", "Sanity", +5),
                new Change("Intense sex", "Stamina", -10),
                new Change("Intense fucking", "Sexual exhaustion", +25, true),
                new Change("Orgasm", "Sexual exhaustion", +5, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Impressed partner", "Sanity", +5),
                new Change("Blowjob", "Sanity", +15),
                new Change("Came", "Sanity", +5),
                new Change("Got sucked off", "Sexual exhaustion", +15, true),
                new Change("Orgasm", "Sexual exhaustion", +5, true)
            ]);
        }
    });

    this.femaleForeplay = new ChoiceController(id, "femaleForeplay");

    this.femaleOral = new ChoiceController(id, "femaleOral");

    this.youFirst = new ChangeController(id, "youFirst", function () {
        global.kobold.status.mental.increase(15, "Getting eaten out");
        global.kobold.status.exertion.increase(20, "Getting eaten out");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Got eaten out", "Sanity", +10),
            new Change("Reached climax", "Sanity", +5),
            new Change("Got eaten out", "Sexual exhaustion", +15, true),
            new Change("Orgasm", "Sexual exhaustion", +5, true)
        ]);
    });

    this.himFirst = new ChangeController(id, "himFirst", function () {
        global.kobold.body.oralVirgin = false;
        global.kobold.status.mental.increase(10, "Giving a good blowjob");
        global.kobold.status.stamina.decrease(5, "Giving a blowjob");
        global.kobold.status.exertion.increase(15, "Giving a blowjob");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Gave a good blowjob", "Sanity", +5),
            new Change("About to get eaten out", "Sanity", +5),
            new Change("Gave a blowjob", "Stamina", -5),
            new Change("Gave a blowjob", "Sexual exhaustion", +15, true)
        ]);
    });


    this.femaleSex = new ChoiceController(id, "femaleSex");

    this.femaleOutside = new ChangeController(id, "femaleOutside", function () {
        global.kobold.body.vaginalVirgin = false;
        if (that.misc.luck() > 5) {
            global.kobold.status.mental.increase(25, "Sex with the cyan kobold");
            global.kobold.status.stamina.decrease(5, "Sex with the cyan kobold");
            global.kobold.status.exertion.increase(30, "Sex with the cyan kobold");
        } else {
            global.kobold.status.mental.increase(5, "Sex with the cyan kobold");
            global.kobold.status.stamina.decrease(5, "Sex with the cyan kobold");
            global.kobold.status.exertion.increase(30, "Sex with the cyan kobold");
        }
    }, function () {
        if (that.misc.luck() > 5) {
            return global.ui.change.buildNotification([
                new Change("Got fucked", "Sanity", +10),
                new Change("Reached climax", "Sanity", +5),
                new Change("Not impregnated", "Sanity", +10),
                new Change("Rode cock", "Stamina", -5),
                new Change("Intensive sex", "Sexual exhaustion", +25, true),
                new Change("Orgasm", "Sexual exhaustion", +5, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Got fucked", "Sanity", +10),
                new Change("Reached climax", "Sanity", +5),
                new Change("Impregnated?", "Sanity", -10),
                new Change("Rode cock", "Stamina", -5),
                new Change("Intensive sex", "Sexual exhaustion", +25, true),
                new Change("Orgasm", "Sexual exhaustion", +5, true)
            ]);
        }
    });

    this.femaleInside = new ChangeController(id, "femaleInside", function () {
        global.kobold.body.vaginalVirgin = false;
        global.kobold.status.mental.increase(25, "Sex with the cyan kobold");
        global.kobold.status.exertion.increase(30, "Sex with the cyan kobold");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Got fucked", "Sanity", +10),
            new Change("Reached climax", "Sanity", +5),
            new Change("Impregnated?", "Sanity", +10),
            new Change("Intensive sex", "Sexual exhaustion", +25, true),
            new Change("Orgasm", "Sexual exhaustion", +5, true)
        ]);
    });

    this.preBucket = new ContinueController(id, "preBucket");

    this.bucket = new ChangeController(id, "bucket", function () {
        if (that.femaleOral.get() === MorningThreeFemaleOralChoice.HIM) {
            global.kobold.status.physical.increase(5, "Refreshing water");
            global.kobold.status.mental.decrease(10, "Denied orgasm");
            global.kobold.status.stamina.decrease(5, "Ran from alley");
            global.kobold.status.exertion.decrease(5, "Refreshing water");
        } else {
            global.kobold.status.physical.increase(5, "Refreshing water");
            global.kobold.status.mental.decrease(5, "Intimacy interrupted");
            global.kobold.status.stamina.decrease(5, "Ran from alley");
            global.kobold.status.exertion.decrease(5, "Refreshing water");
        }
    }, function () {
        if (that.femaleOral.get() === MorningThreeFemaleOralChoice.HIM) {
            return global.ui.change.buildNotification([
                new Change("Refreshing water", "Health", +5),
                new Change("Intimacy interrupted", "Sanity", -5),
                new Change("Denied orgasm", "Sanity", -5),
                new Change("Ran", "Stamina", -5),
                new Change("Refreshing water", "Sexual exhaustion", -5, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Refreshing water", "Health", +5),
                new Change("Intimacy interrupted", "Sanity", -5),
                new Change("Ran", "Stamina", -5),
                new Change("Refreshing water", "Sexual exhaustion", -5, true)
            ]);
        }
    });

    this.preContinue = new ContinueController(id, "preContinue");

    this.sad = new ChangeController(id, "sad", function () {
        global.kobold.status.mental.decrease(15, "Heartbroken");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Heartbroken", "Sanity", -15)
        ]);
    });

    this.troubling = new ChangeController(id, "troubling", function () {
        global.kobold.status.mental.decrease(5, "Guard-sanctioned gangrape");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Poetic justice?", "Sanity", +5),
            new Change("Guard-sanctioned gangrape", "Sanity", -5),
            new Change("Fear of getting caught", "Sanity", -5)
        ]);
    });

    this.robbed = new ChangeController(id, "robbed", function change() {
        var coins = global.kobold.inventory.coins;
        if (coins >= 5) {
            global.kobold.inventory.coins = coins - 5;
        } else {
            global.kobold.inventory.coins = 0;
        }
        global.kobold.status.mental.increase(5, "Friend disappeared");
    }, function changeNote() {
        return global.ui.change.buildNotification([
            new Change("Threw first egg", "Sanity", +2),
            new Change("Threw second egg", "Sanity", +5),
            new Change("Friend disappeared", "Sanity", -2)
        ]);
    });

    this.declineEgg = new ChangeController(id, "declineEgg", function () {
        global.kobold.status.mental.decrease(10, "Guard-sanctioned gangrape");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Scared off new friend", "Sanity", -5),
            new Change("Poetic justice?", "Sanity", +5),
            new Change("Guard-sanctioned gangrape", "Sanity", -5),
            new Change("Fear of getting caught", "Sanity", -5)
        ]);
    });
    
    this.bgm = new AudioController(id, "res/audio/varanar/morningthree.mp3", 100);
    this.splashSound = new AudioController(id, "res/audio/effects/splash.mp3", 50);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

}

// END scenes\Varanar\DayThree\MorningThree\MorningThreeController.js

// scenes\Varanar\DayThree\MorningThree\MorningThree.js

/**
 * First morning in Varanar.
 */
function MorningThree() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "MorningThree";
    var sceneName = "Varanar, Morning Three";
    var sceneSubfolder = "Varanar/DayThree";
    var ctrl = new MorningThreeController(sceneId);
    var arrival = new MorningThreeKnights(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]Varanar[/t]`);
        tab += global.parser.parse(`[st]Morning three[/st]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new MorningThree();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\Varanar\DayThree\MorningThree\MorningThree.js



// END SCENE MorningThree



// SCENE NoonThreeStore

// scenes\Varanar\DayThree\NoonThreeStore\shop\NoonThreeStoreLeave.js

/**
 * Information about the class.
 */
function NoonThreeStoreLeave() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\shop\NoonThreeStoreLeave.js

// scenes\Varanar\DayThree\NoonThreeStore\shop\NoonThreeStoreReturn.js

/**
 * Information about the class.
 */
function NoonThreeStoreReturn() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\shop\NoonThreeStoreReturn.js

// scenes\Varanar\DayThree\NoonThreeStore\shop\NoonThreeStoreFinishSelector.js

/**
 * Information about the class.
 */
function NoonThreeStoreFinishSelector() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\shop\NoonThreeStoreFinishSelector.js

// scenes\Varanar\DayThree\NoonThreeStore\shop\NoonThreeStoreShop.js

/**
 * Information about the class.
 */
function NoonThreeStoreShop() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\shop\NoonThreeStoreShop.js

// scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexReluctant.js

/**
 * Information about the class.
 */
function NoonThreeStoreSexReluctant() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexReluctant.js

// scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexEager.js

/**
 * Information about the class.
 */
function NoonThreeStoreSexEager() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexEager.js

// scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexMoreMoney.js

/**
 * Information about the class.
 */
function NoonThreeStoreSexMoreMoney() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexMoreMoney.js

// scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexAccept.js

/**
 * Information about the class.
 */
function NoonThreeStoreSexAccept() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexAccept.js

// scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexDecline.js

/**
 * Information about the class.
 */
function NoonThreeStoreSexDecline() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\sex\NoonThreeStoreSexDecline.js

// scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobReluctant.js

/**
 * Information about the class.
 */
function NoonThreeStoreBlowjobReluctant() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobReluctant.js

// scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobEager.js

/**
 * Information about the class.
 */
function NoonThreeBlowjobEager() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobEager.js

// scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobMoreMoney.js

/**
 * Information about the class.
 */
function NoonThreeStoreBlowjobMoreMoney() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobMoreMoney.js

// scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobAccept.js

/**
 * Information about the class.
 */
function NoonThreeStoreBlowjobAccept() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobAccept.js

// scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobDecline.js

/**
 * Information about the class.
 */
function NoonThreeBlowjobDecline() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\blowjob\NoonThreeStoreBlowjobDecline.js

// scenes\Varanar\DayThree\NoonThreeStore\handjob\NoonThreeStoreHandjobReluctant.js

/**
 * Information about the class.
 */
function NoonThreeStoreHandjobReluctant() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\handjob\NoonThreeStoreHandjobReluctant.js

// scenes\Varanar\DayThree\NoonThreeStore\handjob\NoonThreeStoreHandjobEager.js

/**
 * Information about the class.
 */
function NoonThreeStoreHandjobEager() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\handjob\NoonThreeStoreHandjobEager.js

// scenes\Varanar\DayThree\NoonThreeStore\handjob\NoonThreeStoreHandjobMoreMoney.js

/**
 * Information about the class.
 */
function NoonThreeStoreHandjobMoreMoney() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\handjob\NoonThreeStoreHandjobMoreMoney.js

// scenes\Varanar\DayThree\NoonThreeStore\handjob\NoonThreeStoreHandjobAccept.js

/**
 * Information about the class.
 */
function NoonThreeStoreHandjobAccept() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\handjob\NoonThreeStoreHandjobAccept.js

// scenes\Varanar\DayThree\NoonThreeStore\decline\NoonThreeStoreDecline.js

/**
 * Information about the class.
 */
function NoonThreeStoreDecline() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\decline\NoonThreeStoreDecline.js

// scenes\Varanar\DayThree\NoonThreeStore\arrival\NoonThreeStoreProposal.js

/**
 * Information about the class.
 */
function NoonThreeStoreProposal() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\arrival\NoonThreeStoreProposal.js

// scenes\Varanar\DayThree\NoonThreeStore\arrival\NoonThreeStoreArrival.js

/**
 * Information about the class.
 */
function NoonThreeStoreArrival() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\arrival\NoonThreeStoreArrival.js

// scenes\Varanar\DayThree\NoonThreeStore\NoonThreeStoreController.js

/**
 * Information about the class.
 */
function NoonThreeStoreController() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\NoonThreeStoreController.js

// scenes\Varanar\DayThree\NoonThreeStore\NoonThreeStore.js

/**
 * Information about the class.
 */
function NoonThreeStore() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Varanar\DayThree\NoonThreeStore\NoonThreeStore.js



// END SCENE NoonThreeStore



// SCENE VaranarSoupKitchen

// scenes\Varanar\DayThree\VaranarSoupKitchen\fail\VaranarSoupKitchenHalfApple.js

/**
 * End section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenHalfApple(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;
    var changeController = ctrl.halfApple;

    var nextSceneReady = ctrl.nextSceneReady;
    var nextSceneName = ctrl.nextSceneName;
    var nextFolderName = ctrl.nextFolderName;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Denied your meal, and kicked back out onto the cold streets, you decide to go back to your [i]original[/i] plan.
                            You scrounge the town square, crawling through the dirt looking for anything left to eat.
                            Almost [b]everything[/b] has been picked clean already, probably by [i]other kobolds[/i].
                            Right as you were about to give up hope completely, you find a single half-eaten apple concealed underneath a dropped handkerchief.
                        [/p]
                        [p]
                            You pick up the day-old, half-eaten fruit.
                            No worms, no bird-pecking marks, a bit of grime, but you wipe most of it off using the handkerchief.
                            Better than nothing, and you've had [b]far[/b] worse to eat in the past.
                            You feel like you [b]need[/b] to eat something, or you'll pass out.
                            Here goes nothing.
                        [/p]
                        [p]
                            You take a first bite into the apple.
                            It's still quite [i]hard[/i], having retained much of its its firmness overnight due to the sub-zero temperatures out.
                            While the taste of [i]dirt[/i] is [b]very[/b] prevalent, underneath lies a satisfying, rather sweet treat.
                            Savoring the delightful flavor, you gobble down the rest of your unsanitary snack.
                            While not much, your midday meal gives you the energy boost you needed to keep on truckin'.
                        [/p]
                        [p]
                            Your hunger [i]somewhat[/i] sated, you decide to try your luck elsewhere.
                            Perhaps the grand avenue will hold some interesting sights to see.
                            Everything was already closed last night, but you're certain there'll be at least one store worth browsing.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(nextSceneName, nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("varanar_generic");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\fail\VaranarSoupKitchenHalfApple.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEatNoMention.js

/**
 * End section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenEatNoMention(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenEatOfferChoice.DONOT;
    var reqController = ctrl.offer;
    var changeController = ctrl.noMention;

    var nextSceneReady = ctrl.nextSceneReady;
    var nextSceneName = ctrl.nextSceneName;
    var nextFolderName = ctrl.nextFolderName;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var toReturn = `[p]
                            You lie and say the previous owner [i]gave it[/i] to you because she felt [b]bad[/b] for keeping kobolds as slaves.
                            You were actually on your way there to [b]free[/b] your enslaved kin, but hunger stopped you dead in your tracks.
                            And so you ended up here.
                        [/p]
                        [p]
                            You can tell that the sheep isn't buying you bullshit.
                            But part of him [b]wants[/b] to believe.
                            He closes his eyes, and his inquisitive gaze turns into a grateful smile.
                            "[color=ppink][i]Kuhbold[/i], ah cannae thank ye enough.
                            A home in the Hamlet is perfect, since that's where ah get all mah food from to begin with.[/color]"
                            He puts the bag with his wedding ring in it around his neck once more.
                        [/p]
                        [p]
                            "[color=ppink]Mah waif's work can continue.
                            Maybe, just maybe, the people of the world can someday truly go without hunger.
                            And if that day ever comes, you'll know you helped make it happen, [i]kuhbold[/i].[/color]"
                            The caprine rises from his seat, signalling for you to get to your feet as well.
                            The two of you walk to the door.
                        [/p]
                        [p]
                            "[color=ppink]You've done me a real favor today, ${male ? `lad` : `lass`}.
                            If ye ever ever need something, [b]anything[/b] at all, ye know where me.
                            For now, I'm going to go pack me things.
                            I'll take care of yer friends real good, don't you worry![/color]"
                            The two of you say goodbye.
                            Your chubby host opens the door for you,
                            and you stride back out into the coldness of the streets of Varanar, your heart heavy with mixed feelings.
                        [/p]
                        [p]
                            On one end, it feels good knowing that someone will look after those [i]enslaved[/i] kobolds Cassidy was keeping.
                            On the other end of the stick, you [i]kind of[/i] failed to tell him about the [b]dead body[/b] in the barn.
                            Who knows what the sheep'll do when he finds out you've handed him a dead person's home?
                            Come to think of it, you didn't even catch his name.
                            Oh well, you're certain things'll work out for the best.
                        [/p]
                        [p]
                            Your conscience somewhat cleared, your spirits lifted and your ego slightly boosted,
                            you set off for Varanar's grand evenue.
                            Last night, all the stores were closed,
                            but most of them should be open around now.
                            Time to see some sights!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(nextSceneName, nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("varanar_generic");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEatNoMention.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEatMention.js

/**
 * End section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenEatMention(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenEatOfferChoice.MENTION;
    var reqController = ctrl.offer;
    var changeController = ctrl.mention;

    var nextSceneReady = ctrl.nextSceneReady;
    var nextSceneName = ctrl.nextSceneName;
    var nextFolderName = ctrl.nextFolderName;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.noon.gaveHouseMentioned = true;
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var toReturn = `[p]
                            You look away and mention that you [i]might[/i] have had a [i]disagreement[/i] with the previous owner.
                            And, well, as far as you know, the psycho bitch's dead body is still somewhere in the barn.
                            The sheep glares at you, shaking his head.
                            You say you really didn't mean to kill her, 
                            but she had you [b]trapped[/b] on this [b]table[/b] and she wanted to all kinds of [b]bad things[/b] to you and--
                            "[color=ppink]Hush, [i]kuhbold[/i].[/color]" interrupts the caprine.
                            "[color=ppink]The less ah know 'bout this, the better.[/color]"
                        [/p]
                        [p]
                            The man sighs loudly, putting the bag with his wedding ring in it around his neck once more.
                            "[color=ppink]Listen here.
                            I'll take the house, I'll get rid of the body, and I'll feed yer friend.[/color]"
                            You try to mention that they aren't actually your [i]friends[/i],
                            but the sheep loudly clears his throat to indicate that you've said enough already.
                        [/p]
                        [p]
                            "[color=ppink]I'm doin' this only because me waif would have wanted it.
                            Ah thank ye for the house, but ye must understand that ah cannae approve of what ye did to obtain it yerself.
                            The guards'd have mah head if they knew ah was talkin' to ah [b]murderin'[/b] [i]kuhbold[/i].
                            Not turning ye in makes me an [i]accessory[/i], but I'll risk it for the sake of me waif's dream of a world without 'unger.[/color]"
                        [/p]
                        [p]
                            The fat man rises from his chair.
                            "[color=ppink]Thank ye once again.
                            This doesn't make up for what you've done,
                            but it at least puts ye on the path of redemption.
                            Don't go 'round 'killin' anyone else now, ye hear?
                            There's hope left for us all.
                            Even [b]murderers[/b] like you.[/color]"
                        [/p]
                        [p]
                            The sheep signals for you to rise to your feet.
                            The two of you walk to the front door.
                            "[color=ppink]I'm going upstairs to pack me things.
                            You've done a good deed today.
                            Let's keep the corpse between you and me, okay?[/color]"
                            You give the man a small nod.
                            "[color=ppink]Good ${male ? `lad` : `lass`}.
                            Off you go now.
                            Come visit some time, will ye?[/color]"
                        [/p]
                        [p]
                            Before you even have a chance to respond, the caprine opens the door,
                            and practically [b]pushes[/b] you out of the building, into the coldness of the streets.
                            You turn around to say goodbye, but he rudely slams the door in your face.
                            Ungrateful bastard.
                            Then again, he might just have been a little [i]shocked[/i] to know the person he took in and fed was a [b]murderer[/b].
                        [/p]
                        [p]
                            Oh well.
                            At least it feels good knowing that someone will look after those [i]enslaved[/i] kobolds Cassidy was keeping.
                            You believe you did a good thing, helping out the owner of the soup kitchen.
                            Come to think of it, you didn't even catch his name.
                            Oh well, he seemed like a reliable enough guy.
                            And hey, he managed to feed [b]you[/b], right?
                        [/p]
                        [p]
                            Your conscience clear, your spirits lifted and your ego slightly boosted,
                            you set off for Varanar's grand evenue.
                            Last night, all the stores were closed,
                            but most of them should be open around now.
                            Time to see some sights!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(nextSceneName, nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("varanar_generic");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEatMention.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEatOffer.js

/**
 * Serve section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenEatOffer(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenEatChoice.OFFER;
    var reqController = ctrl.eat;
    var subController = ctrl.offer;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new VaranarSoupKitchenEatMention(ctrl);
    var secondNext = new VaranarSoupKitchenEatNoMention(ctrl);

    var firstChoice = VaranarSoupKitchenEatOfferChoice.MENTION;
    var secondChoice = VaranarSoupKitchenEatOfferChoice.DONOT;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.noon.gaveHouse = true;
        var toReturn = `[p]
                            At first, you can't really think of a way to console the man.
                            Then, suddenly, something comes to mind.
                            You're carrying a thing that might help.
                            A key.
                            ${global.kobold.inventory.createItemWithTooltip(new CassidyHouseKey())}.
                            You produce the small, metal object and slide it over the surface of the table towards the caprine.
                        [/p]
                        [p]
                            "[color=ppink]What's this?[/color]"
                            he asks, curiously picking your gift up from the table.
                            You explain that it's a [i]key[/i].
                            "[color=ppink]Ah know what ah key is, ye smartass.
                            What's it for?[/color]"
                            A house in the hamlet outside of town, you reply.
                            A [i]mansion[/i], more like.
                            The man eyes you suspiciously.
                            "[color=ppink]And what's that got tah do with me?[/color]"
                        [/p]
                        [p]
                            You explain that you don't really [b]need[/b] a mansion.
                            You have a shack in the woods, and you're content with that.
                            You could [b]give him[/b] the house, and he could continue his business there, if he'd like.
                            On one [i]condition[/i], that is.
                            There's a bunch of kobolds trapped in a barn attached to the house.
                            If he promises to free them and feed them, then by all means, he can have the place.
                        [/p]
                        [p]
                            You notice a twinkle of [i]hope[/i] flare up in the sheep's eyes.
                            Still, he remains wary.
                            "[color=ppink]And how'd [b]you[/b] get this [i]mansion[/i]?[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Mention</button>`;

        var tooltip = global.parser.parse(`Mention there [i]might[/i] be a dead vixen present.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Do not</button>`;

        var tooltip = global.parser.parse(`He'll find out about Cassidy's corpse eventually. [i]If[/i] it's still there.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var VaranarSoupKitchenEatOfferChoice = {
    MENTION: 1,
    DONOT: 2
};

// END scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEatOffer.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEatLeave.js

/**
 * End section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenEatLeave(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenEatChoice.LEAVE;
    var reqController = ctrl.eat;
    var changeController = ctrl.eatLeave;

    var nextSceneReady = ctrl.nextSceneReady;
    var nextSceneName = ctrl.nextSceneName;
    var nextFolderName = ctrl.nextFolderName;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You don't really know how to [i]console[/i] him,
                            so you settle for thanking the man for the food.
                            You feel better now.
                            A [b]lot[/b] better.
                            You're certain his wife is proud of him, wherever she may be.
                        [/p]
                        [p]
                            The man sighs.
                            "[color=ppink]Aye, aye.
                            Ah tried me hardest.
                            'Ave yeh finished yer meal?[/color]"
                            You nod.
                            "[color=ppink]Ah hate tah be rude, but could yeh maybe see yerself out?
                            Ah think ah need a moment alone.
                            To mahself.[/color]"
                            You say you understand, but you obviously don't.
                        [/p]
                        [p]
                            Getting up off your seat, you solemnly walk out the building the way you came.
                            Closing the door behind you, you let out an exasperated sigh.
                            That poor man.
                            The outside air is as cold as your heart felt when walking out on him.
                            But there's nothing you can do to help.
                            You've got your own issues to worry about.
                            While you [b]wish[/b] you were some kind of [b]hero[/b] or a [b]knight[/b],
                            you simply don't have the time or the resources to aid everyone in Varanar.
                        [/p]
                        [p]
                            Swallowing deeply, you set aside your empathetic sadness.
                            To take your mind off of trying to appease dead relatives,
                            you decide to take a stroll down the grand avenue.
                            None of the stores were open last night, but this time around,
                            you're certain there'll be at least [b]one[/b] shop able to pull your mind away from gloomy thoughts.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(nextSceneName, nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("varanar_generic");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEatLeave.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEat.js

/**
 * Serve section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenEat(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenServedChoice.EAT;
    var reqController = ctrl.serve;
    var subController = ctrl.eat;
    var changeController = ctrl.eat;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new VaranarSoupKitchenEatLeave(ctrl);
    var secondNext = new VaranarSoupKitchenEatOffer(ctrl);

    var firstChoice = VaranarSoupKitchenEatChoice.LEAVE;
    var secondChoice = VaranarSoupKitchenEatChoice.OFFER;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.noon.ate = true;
        var toReturn = `[p]
                            You take one of the wooden spoons lying atop the table,
                            and stir the viscous broth in front of you.
                            [b]Yuck.[/b]
                            The texture is inconsistent, far thicker near the bottom of the plate.
                            You'd rather not eat the soup, but you're feeling [b]way[/b] too hungry to pass up an opportunity like this.
                            You [b]need[/b] something to eat, and it's not like you haven't had worse.
                        [/p]
                        [p]
                            Scooping up a mixture of thick and less-thick liquid, you bring your spoon to your mouth, and loudly [b]slurp[/b] down its contents.
                            Your face sours, like the caprine's visage did before.
                            This was [b]definitely[/b] tomato soup once upon a time,
                            but now it's a barely edible mixture of broken-down vegetable chunks floating amidst
                            separated layers of concentrated tomato juice and murky water.
                            Given that the soup is only a day old, you doubt the tomatoes were very [i]fresh[/i] when being put in the pot to begin with.
                        [/p]
                        [p]
                            The taste isn't [i]too[/i] terrible, but the texture and the consistency is [b]revolting[/b].
                            Like that one time you ate the innards of a deer carcass without [i]removing[/i] them first.
                            While unpleasant to wrangle down your throat, your stomach grumbles, [b]begging[/b] for more of the somewhat nutritious goop.
                            Reluctantly, you down spoon after spoon of old tomato soup, until your plate is empty, save for a small layer at the bottom.
                        [/p]
                        [p]
                            As you finish up, the fat sheep enters the room once more, wiping an angry look off his face.
                            "[color=ppink]Sorreh 'bout that.[/color]" he speaks, taking a seat opposite to you once more.
                            "[color=ppink]Farmers from the 'amlet deliver leftover foodstuffs to the alley out ba~ack,
                            and some [i]kuhbolds[/i] that are [b]unwelcome[/b] 'ere oft'n try tah steal their meal from there instead.
                            Too ba~ad for them, the shipment that came yesterday was the last one [b]ever[/b].[/color]"
                        [/p]
                        [p]
                            You can tell the man is upset by having [i]lost[/i] his profession.
                            You ask him if there's any particular reason why he can't pay rent anymore.
                            "[color=ppink]'Twas the king that used to let us live here for free for as long as we kept the streets clean of starvin' lizards.
                            Him and mah waif had uhh...[i]met[/i] in the past.[/color]"
                            You get the feeling they did [b]more[/b] than just [i]meet[/i].
                            "[color=ppink]But then the king went on his lil' [i]crusade[/i] fightin' 'longside the dragons,
                            and now the ba~aron's in charge of the town.
                            Alas the ol' king did'nay tell his [i]regent[/i] about our little [i]agreement[/i].[/color]"
                            Oh dear.
                        [/p]
                        [p]
                            "[color=ppink]Even after me waif died, ah spent every coin I had on rent, tah try and keep this place open, but it's no use.
                            Produce, the farmers give for free, but without a place tah cook and serve it in, it's useless.[/color]"
                            The caprine picks up the bag he put on the table before, and presses it to his chest.
                            "[color=ppink]I've failed ye, me precious Ewelyn.
                            Please fahrgive meh.[/color]"
                            The man snickers.
                            A single tear rolls down his face.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Leave</button>`;

        var tooltip = global.parser.parse(`This has gotten way more awkward than it should have been.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var hasItem = ctrl.misc.hasMisc(new CassidyHouseKey());
        var disabled = subController.has() || !hasItem ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Offer house</button>`;

        var tooltip = global.parser.parse(`Offer the man Cassidy's residence. She won't be needing it anymore. And he can take care of the kobolds.`);
        if (!hasItem) {
            tooltip = global.parser.parse(`You don't have a house to offer.`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 220px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var VaranarSoupKitchenEatChoice = {
    LEAVE: 1,
    OFFER: 2
};

// END scenes\Varanar\DayThree\VaranarSoupKitchen\served\eat\VaranarSoupKitchenEat.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\served\VaranarSoupKitchenLeave.js

/**
 * End section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenLeave(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenServedChoice.LEAVE;
    var reqController = ctrl.serve;
    var changeController = ctrl.leave;

    var nextSceneReady = ctrl.nextSceneReady;
    var nextSceneName = ctrl.nextSceneName;
    var nextFolderName = ctrl.nextFolderName;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You take one of the wooden spoons lying atop the table,
                            and stir the viscous broth in front of you.
                            [b]Yuck.[/b]
                            The texture is inconsistent, far thicker near the bottom of the plate.
                            You're not eating this.
                            You'd get [b]sick[/b] if you did.
                        [/p]
                        [p]
                            While you appreciate the sheep's hospitality,
                            you don't have the heart to tell him face-to-face that you'd rather not eat his [i]leftovers[/i].
                            After all, you made him get out of his bathtub and technically even [i]retirement[/i], just for you.
                            Instead of waiting for him to come back, you get up off your chair,
                            sneak towards the door you came in through, and quietly exit the building.
                        [/p]
                        [p]
                            Back onto the cold streets, you decide to go back to your [i]original[/i] plan.
                            You scrounge the town square, crawling through the dirt looking for anything left to eat.
                            Almost [b]everything[/b] has been picked clean already, probably by [i]other kobolds[/i].
                            Right as you were about to give up hope completely, you find a single half-eaten apple concealed underneath a dropped handkerchief.
                        [/p]
                        [p]
                            You pick up the day-old, half-eaten fruit.
                            No worms, no bird-pecking marks, a bit of grime, but you wipe most of it off using the handkerchief.
                            Better than nothing, and you've had [b]far[/b] worse to eat in the past.
                            You feel like you [b]need[/b] to eat something, or you'll pass out.
                            Here goes nothing.
                        [/p]
                        [p]
                            You take a first bite into the apple.
                            It's still quite [i]hard[/i], having retained much of its its firmness overnight due to the sub-zero temperatures out.
                            While the taste of [i]dirt[/i] is [b]very[/b] prevalent, underneath lies a satisfying, rather sweet treat.
                            Savoring the delightful flavor, you gobble down the rest of your unsanitary snack.
                            While not much, your midday meal gives you the energy boost you needed to keep on truckin'.
                        [/p]
                        [p]
                            Your hunger [i]somewhat[/i] sated, you decide to try your luck elsewhere.
                            Perhaps the grand avenue will hold some interesting sights to see.
                            Everything was already closed last night, but you're certain there'll be at least one store worth browsing.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(nextSceneName, nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("varanar_generic");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\served\VaranarSoupKitchenLeave.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\served\VaranarSoupKitchenSteal.js

/**
 * End section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenSteal(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenServedChoice.STEAL;
    var reqController = ctrl.serve;
    var changeController = ctrl.steal;

    var nextSceneReady = ctrl.nextSceneReady;
    var nextSceneName = ctrl.nextSceneName;
    var nextFolderName = ctrl.nextFolderName;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.varanar.dayThree.noon.stole = true;
        var toReturn = `[p]
                            While you appreciate the sheep's hospitality,
                            he made a [b]very[/b] big made mistake.
                            He left what you presume to be one of his most valuable possessions in the same room as a kobold, without any supervision whatsoever.
                            What did he [b]think[/b] was going to happen?
                            That you were going to sit here and eat this old broth while being [b]tempted[/b] by the promise of riches and gold?
                            What an [b]idiot[/b].
                        [/p]
                        [p]
                            The caprine doesn't have any excuse.
                            He's been working with your kind for [b]years[/b].
                            He should have known what would happen if he left his wedding ring alone with you.
                            Heck, you're surprised the fat man has managed to hold onto it for this long.
                            It being stolen at one point or another was inevitable.
                            You're just helping the natural order of things along.
                        [/p]
                        [p]
                            Your decision rationalized, you reach out for the small bag.
                            Opening it up, you discover a shiny ${global.kobold.inventory.createItemWithTooltip(new GoldRing())} within.
                            It looks brand new, despite obviously being rather old.
                            The sheep must have kept it in good condition ever since his wife died.
                            This must be the only thing he has to remember her by, especially now that the soup kitchen has closed down.
                            Almost kind of sad, but you really [b]need[/b] the money.
                            Rather the caprine upset, than you [b]dead[/b].
                            Your loot in hand, you get up off your chair,
                            sneak towards the door you came in through, and quietly exit the building.
                        [/p]
                        [p]
                            Back onto the cold streets, you immediately [b]bolt[/b] towards the grand avenue,
                            to hide within the masses of people on the streets, should the fat man come looking for you after he figures out what you took.
                            Adrenaline somewhat suppresses your hunger, making you forget all about your rumbling stomach.
                            With a golden ring as your haul, you can buy [b]proper[/b] food later.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(nextSceneName, nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("varanar_generic");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\served\VaranarSoupKitchenSteal.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\served\VaranarSoupKitchenServed.js

/**
 * Serve section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenServed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.serveChoice;
    var subController = ctrl.serve;
    var changeController = ctrl.serve;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new VaranarSoupKitchenSteal(ctrl);
    var secondNext = new VaranarSoupKitchenLeave(ctrl);
    var thirdNext = new VaranarSoupKitchenEat(ctrl);

    var firstChoice = VaranarSoupKitchenServedChoice.STEAL;
    var secondChoice = VaranarSoupKitchenServedChoice.LEAVE;
    var thirdChoice = VaranarSoupKitchenServedChoice.EAT;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = ``;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Steal</button>`;

        var tooltip = global.parser.parse(`Take the ring. You need it way more than he does.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Leave</button>`;

        var tooltip = global.parser.parse(`You're not eating this slop. Silently leave the way you came.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 220px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Eat</button>`;

        var tooltip = global.parser.parse(`It may not be very appetizing, but it's better than nothing.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 155px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var VaranarSoupKitchenServedChoice = {
    STEAL: 1,
    LEAVE: 2,
    EAT: 3
};

// END scenes\Varanar\DayThree\VaranarSoupKitchen\served\VaranarSoupKitchenServed.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\lie\VaranarSoupKitchenLieSucceed.js

/**
 * Invited section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenLieSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.serveChoice;
    var changeController = false;

    var autoContinue = true;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.serveChoices;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var toReturn = `[p]
                            You tell the man a lie.
                            The bag around your neck isn't filled with coins.
                            It's filled with trinkets passed down from generation to generation,
                            ultimately falling into your hands.
                            You [i]could[/i] pawn it off, but doing so would be a great insult to your ancestors,
                            and you'd rather not risk the wrath of ghosts trapped in the afterlife.
                            You have enough [i]living[/i] problems to worry about as-is.
                        [/p]
                        [p]
                            "[color=ppink]Hmph.
                            I've heard many a [i]kuhbold[/i] claim tah be more broke than they actually are.
                            [b]Lying.[/b]
                            That's one thing yer kind is good at, isn't it?
                            How do I know you're not tellin' meh a lie?
                            If that ba~ag you're carryin' really is filled with ancient ba~aubles,
                            then let's open'er up and 'ave a look![/color]"
                        [/p]
                        [p]
                            Uh-oh.
                            You didn't expect the caprine to be smarter than he looks.
                            But you aren't out of tricks yet.
                            You stare up at the sheep, an offended frown forming on your face.
                            You [b]angrily[/b] ask him how he even [b]dares[/b] suggest to open up a purse that hasn't been opened in over a hundred years.
                            Does he [b]hate[/b] kobold traditions, or is he just [b]trying[/b] to make you mad?!
                        [/p]
                        [p]
                            "[color=ppink]That ba~ag doesn't look a hundred years old to me, ${male ? `lad` : `lass`}.[/color]"
                            comes the man's retort.
                            Hmph.
                            Looks like you're going to have to bring out your best if you want to fool him.
                            Your family has kept the pouch in top condition for all these years,
                            only for some [b]mean sheep[/b] to [b]harass[/b] you about it.
                            A few fake tears well up in the corners of your eyes.
                            You let out a well-acted snicker.
                            If your mother was still alive to see you [b]begging[/b] for scraps,
                            she'd [b]cut off[/b] her own arm and feed it to you instead of having you reduced to [b]tears[/b] in return for a measily handout.
                        [/p]
                        [p]
                            "[color=ppink]Alright, alright.
                            Stop yer cryin'.[/color]"
                            the caprine responds.
                            "[color=ppink]I'll get yeh some soup, okay?
                            I'm sorry for doubtin' ye.
                            I've seen a lot of frauds in mah day.
                            And I can't feed [b]all[/b] of yeh, understand?[/color]"
                            Hah.
                            Got'em.
                            While continuing to pretend to be sad, you nod and meekly stammer out that it's okay.
                        [/p]
                        ${global.ui.condSucc(`Charisma greater than 6`)}
                        [p]
                            Patting you on the head, the fat man steps behind the counter,
                            opens up a cupboard, and produces a rusty metal pot half-filled with day-old soup.
                            Wiping your crocodillian tears, you take a seat at the table nearest to you.
                            The sheep uses a wooden ladle to scoop some soup onto a deep plate,
                            having a quick taste of his own culinary prowess in the process.
                            His face scrunches up, signalling his disgust.
                            If [b]he[/b] doesn't like it, you very much so doubt [b]you[/b] will.
                        [/p]
                        [p]
                            Your dinner prepared, your host brings it over to your table.
                            He crudely tosses the plate in front of you, the viscous, red liquid within barely contained, almost splashing out onto you.
                            It looks to be some kind of [i]tomato[/i] soup, with chunks of what you presume to once have been [i]vegetables[/i].
                            Having served you your meal, the half-naked caprine takes a seat opposite to you, chair creaking loudly under his weight.
                        [/p]
                        [p]
                            The sheep reaches into the wool around his neck, and fetches out something you hadn't even noticed before:
                            a small pouch of his own, similar to yours, almost [i]identical[/i].
                            He places the bag on the table.
                            "[color=ppink]Y'see this here?
                            This is my memento.
                            Wedding ring.
                            Belonged to mah waif.
                            Most beautiful ewe in all of the world, she was.
                            Not too [i]faithful[/i],
                            but I looked past those flaws to see the [b]diamond[/b] hidden underneath all the layers of foolin' 'round with the neighbours.
                            It was [b]her[/b] idea to start this here soup kitchen, and in her memory I continued the business.
                            'Till yesterday, that is.[/color]"
                        [/p]
                        [p]
                            A sudden loud noise from the back of the building interrupts the chubby man's sad story.
                            Sounds like something [i]falling[/i] and [b]breaking[/b].
                            The caprine rises to his hooves.
                            "[color=ppink]Baa~aah, it's probably those [b]filthy[/b] thieves again.
                            You enjoy yer meal, [i]kuhbold[/i].
                            I'll go scare'em off.[/color]"
                        [/p]
                        [p]
                            That being said, the sheep rushes behind the counter once more,
                            before disappearing into a side room, leaving you all alone with your plate of [i]slop[/i].
                            Soon after, you hear his loud, profanity-filled yelling along with angry bleating directed at whoever attempted to sneak inside.
                            Looks like the fat man forgot the bag with the ring in on the table.
                            It'd be a shame if someone [i]took it[/i].
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\lie\VaranarSoupKitchenLieSucceed.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\lie\VaranarSoupKitchenLieFail.js

/**
 * Invited section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenLieFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.lieFail;

    var autoContinue = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.apple;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You tell the man a lie.
                            The bag around your neck isn't filled with coins.
                            It's filled with trinkets passed down from generation to generation,
                            ultimately falling into your hands.
                            You [i]could[/i] pawn it off, but doing so would be a great insult to your ancestors,
                            and you'd rather not risk the wrath of ghosts trapped in the afterlife.
                            You have enough [i]living[/i] problems to worry about as-is.
                        [/p]
                        [p]
                            "[color=ppink]Hmph.
                            I've heard many a [i]kuhbold[/i] claim tah be more broke than they actually are.
                            [b]Lying.[/b]
                            That's one thing yer kind is good at, isn't it?
                            How do I know you're not tellin' meh a lie?
                            If that ba~ag you're carryin' really is filled with ancient ba~aubles,
                            then let's open'er up and 'ave a look![/color]"
                        [/p]
                        ${global.ui.condFail(`Charisma below 7`)}
                        [p]
                            Uh-oh.
                            You didn't expect the caprine to be smarter than he looks.
                            Unable to comply to his request, you begin to stammer out a bunch of stalling gibberish.
                            A trick he instantly sees through.
                            "[color=ppink]Hah.
                            You're a baa~aad liar, [i]kuhbold[/i].
                            Come now, we don't serve liars here.
                            Git.
                            Out the door.
                            Or do I need tah [b]throw[/b] yeh out?[/color]"
                        [/p]
                        [p]
                            No.
                            [b]No![/b]
                            You're telling the truth!
                            You [b]swear[/b] you are!
                            You aren't lying.
                            Opening the bag would just be [b]really[/b] uhh...[i]offensive[/i] to your forefathers!
                            "[color=ppink]Yeah right.
                            And me wife, Armok rest'er soul, was a faithful woman, too?
                            Do ye expect me tah believe everythin' without questionin', lizard?
                            Yer a dirty, no-good liar.
                            Now get out of mah soup kitchen.
                            We're [b]closed[/b].
                            [b]Forever[/b].[/color]"
                        [/p]
                        [p]
                            That being said, the fat sheep points towards the door.
                            You wordlessly beg him to let you stay.
                            He looks like he's made his choice.
                            You try to come up with something, [b]anything[/b] to say, but it's hopeless.
                            Your lack of persuasion has cost you a valuable opportunity.
                            Hanging your head, you walk out the door, back towards the market square.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\lie\VaranarSoupKitchenLieFail.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\lie\VaranarSoupKitchenLie.js

/**
 * Invited section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenLie(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenQuestionedChoice.LIE;
    var reqController = ctrl.questioned;

    var toFail = new VaranarSoupKitchenLieFail(ctrl);
    var toSuccess = new VaranarSoupKitchenLieSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.charisma() >= 7;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\lie\VaranarSoupKitchenLie.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\truth\VaranarSoupKitchenTruthSucceed.js

/**
 * Invited section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenTruthSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.serveChoice;
    var changeController = false;

    var autoContinue = true;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.serveChoices;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var paid = global.kobold.history.shared.debtCollector.paid;
        var toReturn = `[p]
                            You tell the man the truth.
                            You're ${paid ? `heavily` : ``} indebted to the Three-Tailed Rat Clan, and are trying to save up some money to pay them off.
                            You [i]could[/i] use your gold to buy food for yourself,
                            but you don't [b]dare[/b] think of the consequences should the clan catch you empty-handed.
                            Worse yet, if one of them finds you spending coins that should belong to [b]them[/b], they'd probably [b]torture[/b] you to death for that.
                            So, here you are, hoping for some food that [i]won't[/i] lead to your slow and painful [b]death[/b].
                        [/p]
                        [p]
                            "[color=ppink]Hmph.
                            I've heard many a [i]kuhbold[/i] claim tah be hounded by this [i]gang[/i].
                            [b]Lying.[/b]
                            That's one thing yer kind is good at, isn't it?
                            How do I know you're not tellin' meh a lie?
                            If you're carryin' that gold around for, surely you know exactly [b]who[/b] it is you owe money to, right?
                            Well then, who is the leader of this so-called [i]Three-Tailed Rat Clan[/i]?[/color]"
                        [/p]
                        ${global.ui.condSucc(`Intelligence greater than 5`)}
                        [p]
                            That's easy enough.
                            The Three-Tailed Rat is the leader of the clan.
                            That's why it's called the [i]Three-Tailed Rat Clan[/i].
                            Because it's lead by the Three-Tailed Rat.
                            "[color=ppink]That's a pretty stupid name.
                            Dumb enough to have been thought up by a [i]kuhbold[/i] like you.[/color]"
                            Another pointless accusation.
                        [/p]
                        [p]
                            You say you think the Three-Tailed Rat is [b]actually[/b] a rat with three tails, which would explain the name, but you're not really sure.
                            Nobody knows his true identity, and he rarely ever shows his face.
                            Besides, outside of the leader, the gang [b]is[/b] completely comprised of kobolds such as yourself.
                            It's not unreasonable to assume that [i]they[/i] came up with the name.
                            "[color=ppink]Alright, alright.
                            Cease yer yappin'.
                            You've convinced me.
                            Truth be told, the less ah know about these [i]clans[/i], the better.
                            Wouldn't want to rub anyone the wrong way in this city.
                            Could end up [b]dead[/b] before ah even get evicted.
                            Anyway, lemme get ye somethin' tah eat.[/color]"
                        [/p]
                        [p]
                            Patting you on the head, the fat man steps behind the counter,
                            opens up a cupboard, and produces a rusty metal pot half-filled with day-old soup.
                            You take a seat at the table nearest to you.
                            The sheep uses a wooden ladle to scoop some soup onto a deep plate,
                            having a quick taste of his own culinary prowess in the process.
                            His face scrunches up, signalling his disgust.
                            If [b]he[/b] doesn't like it, you very much so doubt [b]you[/b] will.
                        [/p]
                        [p]
                            Your dinner prepared, your host brings it over to your table.
                            He crudely tosses the plate in front of you, the viscous, red liquid within barely contained, almost splashing out onto you.
                            It looks to be some kind of [i]tomato[/i] soup, with chunks of what you presume to once have been [i]vegetables[/i].
                            Having served you your meal, the half-naked caprine takes a seat opposite to you, chair creaking loudly under his weight.
                        [/p]
                        [p]
                            The sheep reaches into the wool around his neck, and fetches out something you hadn't even noticed before:
                            a small pouch, similar to your coin-purse, almost [i]identical[/i].
                            He places the bag on the table.
                            "[color=ppink]Y'see this here?
                            It's a memento.
                            Wedding ring.
                            Belonged to mah waif.
                            Most beautiful ewe in all of the world, she was.
                            Not too [i]faithful[/i],
                            but I looked past those flaws to see the [b]diamond[/b] hidden underneath all the layers of foolin' 'round with the neighbours.
                            It was [b]her[/b] idea to start this here soup kitchen, and in her memory I continued the business.
                            'Till yesterday, that is.[/color]"
                        [/p]
                        [p]
                            A sudden loud noise from the back of the building interrupts the chubby man's sad story.
                            Sounds like something [i]falling[/i] and [b]breaking[/b].
                            The caprine rises to his hooves.
                            "[color=ppink]Baa~aah, it's probably those [b]filthy[/b] thieves again.
                            You enjoy yer meal, [i]kuhbold[/i].
                            I'll go scare'em off.[/color]"
                        [/p]
                        [p]
                            That being said, the sheep rushes behind the counter once more,
                            before disappearing into a side room, leaving you all alone with your plate of [i]slop[/i].
                            Soon after, you hear his loud, profanity-filled yelling along with angry bleating directed at whoever attempted to sneak inside.
                            Looks like the fat man forgot the bag with the ring in on the table.
                            It'd be a shame if someone [i]took it[/i].
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\truth\VaranarSoupKitchenTruthSucceed.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\truth\VaranarSoupKitchenTruthFail.js

/**
 * Invited section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenTruthFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.truthFail;

    var autoContinue = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.apple;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var name = ctrl.misc.name();
        var paid = global.kobold.history.shared.debtCollector.paid;
        var toReturn = `[p]
                            You tell the man the truth.
                            You're ${paid ? `heavily` : ``} indebted to the Three-Tailed Rat Clan, and are trying to save up some money to pay them off.
                            You [i]could[/i] use your gold to buy food for yourself,
                            but you don't [b]dare[/b] think of the consequences should the clan catch you empty-handed.
                            Worse yet, if one of them finds you spending coins that should belong to [b]them[/b], they'd probably [b]torture[/b] you to death for that.
                            So, here you are, hoping for some food that [i]won't[/i] lead to your slow and painful [b]death[/b].
                        [/p]
                        [p]
                            "[color=ppink]Hmph.
                            I've heard many a [i]kuhbold[/i] claim tah be hounded by this [i]gang[/i].
                            [b]Lying.[/b]
                            That's one thing yer kind is good at, isn't it?
                            How do I know you're not tellin' meh a lie?
                            If you're carryin' that gold around for, surely you know exactly [b]who[/b] it is you owe money to, right?
                            Well then, who is the leader of this so-called [i]Three-Tailed Rat Clan[/i]?[/color]"
                        [/p]
                        ${global.ui.condFail(`Intelligence below 6`)}
                        [p]
                            Uh-oh.
                            That's a tough question.
                            The leader of the Three-Tailed Rat Clan?
                            Come on, ${name}, you know this one.
                            Uhhh, it starts with a [b]T[/b], you think.
                            Or maybe an [b]R[/b]?
                            You honestly can't remember.
                            What the hell was his name?!
                            Wait, do you even know his name?
                        [/p]
                        [p]
                            Unable to respond to the caprine's question, you begin to stammer out a bunch of stalling gibberish.
                            A trick he instantly sees through.
                            "[color=ppink]Hah.
                            You're a bad liar, [i]kuhbold[/i].
                            Come now, we don't serve liars here.
                            Git.
                            Out the door.
                            Or do I need tah [b]throw[/b] yeh out?[/color]"
                        [/p]
                        [p]
                            No.
                            [b]No![/b]
                            You're telling the truth!
                            You [b]swear[/b] you are!
                            You aren't lying, really!
                            "[color=ppink]Yeah right.
                            And me wife, Armok rest'er soul, was a faithful woman, too?
                            Do ye expect me tah believe everythin' without questionin', lizard?
                            Yer a dirty, no-good liar.
                            Now get out of mah soup kitchen.
                            We're [b]closed[/b].
                            [b]Forever[/b].[/color]"
                        [/p]
                        [p]
                            That being said, the fat sheep points towards the door.
                            You wordlessly beg him to let you stay.
                            He looks like he's made his choice.
                            You try to come up with something, [b]anything[/b] to say, but it's hopeless.
                            Your stupidity and your inability to remember a simple detail have cost you a valuable opportunity.
                            Hanging your head, you walk out the door, back towards the market square.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\truth\VaranarSoupKitchenTruthFail.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\truth\VaranarSoupKitchenTruth.js

/**
 * Invited section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenTruth(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenQuestionedChoice.TRUTH;
    var reqController = ctrl.questioned;

    var toFail = new VaranarSoupKitchenTruthFail(ctrl);
    var toSuccess = new VaranarSoupKitchenTruthSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.intelligence() >= 6;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\truth\VaranarSoupKitchenTruth.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\VaranarSoupKitchenInvitedServed.js

/**
 * Serve section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenInvitedServed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.serveChoice;
    var changeController = false;

    var autoContinue = true;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.serveChoices;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You give the sheep a small nod.
                            You [i]had[/i] hoped for a [i]warm[/i] meal, but beggars can't be choosers.
                            "[color=ppink]Don't worry 'bout the baa~aath.
                            It'll still be there when you're done.
                            I prefer me water cold anyway.[/color]"
                            You can sense a hint of [i]sarcasm[/i] in his voice.
                            "[color=ppink]Anyway, lemme get ye somethin' tah eat.[/color]"
                        [/p]
                        [p]
                            Patting you on the head, the fat man steps behind the counter,
                            opens up a cupboard, and produces a rusty metal pot half-filled with day-old soup.
                            You take a seat at the table nearest to you.
                            The sheep uses a wooden ladle to scoop some soup onto a deep plate,
                            having a quick taste of his own culinary prowess in the process.
                            His face scrunches up, signalling his disgust.
                            If [b]he[/b] doesn't like it, you very much so doubt [b]you[/b] will.
                        [/p]
                        [p]
                            Your dinner prepared, your host brings it over to your table.
                            He crudely tosses the plate in front of you, the viscous, red liquid within barely contained, almost splashing out onto you.
                            It looks to be some kind of [i]tomato[/i] soup, with chunks of what you presume to once have been [i]vegetables[/i].
                            Having served you your meal, the half-naked caprine takes a seat opposite to you, chair creaking loudly under his weight.
                        [/p]
                        [p]
                            The sheep reaches into the wool around his neck, and fetches out something you hadn't even noticed before:
                            a small pouch, similar to your coin-purse, almost [i]identical[/i].
                            He places the bag on the table.
                            "[color=ppink]Y'see this here?
                            It's a memento.
                            Wedding ring.
                            Belonged to mah waif.
                            Most beautiful ewe in all of the world, she was.
                            Not too [i]faithful[/i],
                            but I looked past those flaws to see the [b]diamond[/b] hidden underneath all the layers of foolin' 'round with the neighbours.
                            It was [b]her[/b] idea to start this here soup kitchen, and in her memory I continued the business.
                            'Till yesterday, that is.[/color]"
                        [/p]
                        [p]
                            A sudden loud noise from the back of the building interrupts the chubby man's sad story.
                            Sounds like something [i]falling[/i] and [b]breaking[/b].
                            The caprine rises to his hooves.
                            "[color=ppink]Baa~aah, it's probably those [b]filthy[/b] thieves again.
                            You enjoy yer meal, [i]kuhbold[/i].
                            I'll go scare'em off.[/color]"
                        [/p]
                        [p]
                            That being said, the sheep rushes behind the counter once more,
                            before disappearing into a side room, leaving you all alone with your plate of [i]slop[/i].
                            Soon after, you hear his loud, profanity-filled yelling along with angry bleating directed at whoever attempted to sneak inside.
                            Looks like the fat man forgot the bag with the ring in on the table.
                            It'd be a shame if someone [i]took it[/i].
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\VaranarSoupKitchenInvitedServed.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\VaranarSoupKitchenQuestioned.js

/**
 * Invited section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenQuestioned(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.questioned;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new VaranarSoupKitchenLie(ctrl);
    var secondNext = new VaranarSoupKitchenTruth(ctrl);

    var firstChoice = VaranarSoupKitchenQuestionedChoice.LIE;
    var secondChoice = VaranarSoupKitchenQuestionedChoice.TRUTH;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You give the sheep a small nod.
                            You [i]had[/i] hoped for a [i]warm[/i] meal, but beggars can't be choosers.
                            "[color=ppink]Don't worry 'bout the baa~aath.
                            It'll still be there when you're done.
                            I prefer me water cold anyway.[/color]"
                            You can sense a hint of [i]sarcasm[/i] in his voice.
                        [/p]
                        [p]
                            The jolly look on the caprine's face is suddenly replaced by an inquisitive stare as he notices the small bag hanging off your neck.
                            "[color=ppink]Now wait just a minute.
                            What dah we have 'ere?[/color]"
                            He taps one of his fat fingers against your coin pouch, making the gold pieces within clatter against one another.
                            "[color=ppink]Stingy rich'un, aren't ye?
                            Trying tah score a free meal while havin' more moneh than ah do?!
                            Ye get nothin', and yah can be happeh that ah don't kick yeh out mahself![/color]"
                        [/p]
                        ${global.ui.condFail(`Too much gold`)}
                        [p]
                            The burly man's accent seems to grow [i]heavier[/i] the more angry he gets.
                            You need to come up with something fast if you want to get a free meal still.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>A lie</button>`;

        var tooltip = global.parser.parse(`You won't get any food if you let on that your pouch is filled with gold.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>The truth</button>`;

        var tooltip = global.parser.parse(`This guy seems honest enough. You have no reason to lie.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You tell:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var VaranarSoupKitchenQuestionedChoice = {
    LIE: 1,
    TRUTH: 2
};

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\VaranarSoupKitchenQuestioned.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\VaranarSoupKitchenQuestionedSelector.js

/**
 * Invited section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenQuestionedSelector(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.invited;

    var toFail = new VaranarSoupKitchenInvitedServed(ctrl);
    var toSuccess = new VaranarSoupKitchenQuestioned(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.gold() >= 50;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\VaranarSoupKitchenQuestionedSelector.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\invited\VaranarSoupKitchenInvited.js

/**
 * Invited section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenInvited(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.invited;
    var changeController = ctrl.invited;

    var autoContinue = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new VaranarSoupKitchenQuestionedSelector(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Gathering your courage, you walk up to the door, and slowly push it open.
                            Despite taking little effort to move, it creaks loudly, alerting whoever may be inside to your presence.
                            Swallowing deeply, you step inside of the dilapidated building, and close the door behind you.
                        [/p]
                        [p]
                            You find yourself alone in a dusky room that has clearly seen better days.
                            Apart from the chairs and tables arrayed across the wooden floor, you'd say this almost looks like an abandoned [i]store[/i].
                            At least it's warm in here, so the soup kitchen can't be [i]completely[/i] deserted.
                            You figure the stove must be behind the mercantile countertop, but you're not tall enough to see.
                            To your left is a set of stairs which you figure must lead to the proprietor's [i]above-ground[/i] home.
                        [/p]
                        [p]
                            Suddenly, you hear movement above.
                            Footsteps.
                            No wait.
                            Clacking hooves?
                            "[color=ppink]Yeah, ah heard yah! I'ma comin'![/color]" a bass voice rings out from the second floor.
                        [/p]
                        [p]
                            Down the stairs comes storming a big, wet, outright [b]fat[/b] anthromorphic sheep,
                            his bloated, wool-covered belly jiggling with every step he takes,
                            soapy water dripping liberally from his fleece.
                            The man isn't wearing much besides a towel, tied snugly around his waist.
                            "[color=ppink]Can't a sheep take a fockin' baa~aath wifout a beggar wanderin' in?![/color]"
                            At least you don't have to ask why he's wet and half-naked.
                            "[color=ppink]Well then, what'd ye want, yah skinky lizard?![/color]"
                        [/p]
                        [p]
                            The caprine comes to a rest merely a foot away from you, the wool on his big gut almost pressing into your snout.
                            You can't help but stare at his massive belly,
                            as you stammer out the dumbest question your mind could think of while looking at someone big enough to [b]eat[/b] you.
                            You ask whether or not the man [i]cooks[/i] kobolds here.
                            He leans over you threateningly, a droplet of water dripping from his chin down ontop of your forehead.
                            Glaring at you, he responds to your question in a grim tone.
                            "[color=ppink]Aye, but only the ones that think they're funny.
                            Or the ones that [b]stare at mah belly[/b].[/color]"
                        [/p]
                        [p]
                            His words remind you that you [i]are[/i] still leering at his bulky lower body instead of looking the man in the eyes.
                            Correcting your mistake, you quickly stammer out a meek apology.
                            The sheep suddenly bursts out into laughter.
                            "[color=ppink]Hahahahaha!
                            Scared ye, didn't ah?
                            Only vegetarian meals here, [i]kuhbold[/i].
                            Us sheep, we cannae eat meat.[/color]"
                            You let out an uneasy chuckle, a disturbed smile forming on your face.
                            At least you won't be getting cut up and put into a stew today.
                        [/p]
                        [p]
                            You nervously ask the [i]chef[/i] if you could maybe, if at all possible, if it's not a bother, get something tiny to eat, perhaps?
                            You know you're interrupting his bath-time, but you're kind of [b]starving[/b].
                            "[color=ppink]'Fraid not.[/color]" the caprine responds, with a hint of sadness in his voice.
                            "[color=ppink]I closed down shop yesterday.
                            Not enough money to pay rent for this shitehole.
                            I was half expectin' yah tah be here tah [b]kick me out[/b].[/color]"
                            Even his [i]endearing[/i] accent fails to mask the man's distress.
                        [/p]
                        [p]
                            The chubby sheep lets out a sigh, eyeing you up and down.
                            "[color=ppink]Ye look like a scraggy one.
                            Tell yah what, for you, I'll make an exception.
                            I've got some leftovers from yesterday.
                            Won't be warm, but it's better than nothin', aye?[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        global.ui.applyTheme("hamlet_shop");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\invited\VaranarSoupKitchenInvited.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\threaten\VaranarSoupKitchenThreaten.js

/**
 * Overdressed section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenThreaten(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenAccusedChoice.THREATEN;
    var reqController = ctrl.accused;
    var subController = ctrl.preContinue;
    var changeController = ctrl.threaten;

    var autoContinue = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.apple;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var weapon = ctrl.misc.weapon();
        var toReturn = `[p]
                            You threaten him, boldly stating that he'd better go get you some soup and quick, [b]or else[/b].
                            ${weapon ? `You have a weapon, and you're not afraid to use it.` : `You may not have a weapon, but you can [b]rake[/b] him with your claws.`}
                            Unless the sheep is looking for [b]trouble[/b], he'd best get to serving you, and make it snappy.
                            He wouldn't want to get [b]killed[/b] the day after retirement, now would he?
                        [/p]
                        [p]
                            "[color=ppink]You?
                            Kill [b]me[/b]?
                            Don't make me [b]laugh[/b], [i]kuhbold[/i].[/color]"
                            You growl loudly, dropping into a combat stance to show the caprine how [b]serious[/b] you are.
                            Before you even have the chance to issue a final warning,
                            the palm of the fat sheep's hand [b]slams[/b] into your jaw,
                            the big man's slap strong enough to almost knock you off your feet.
                            You recoil, clutching your muzzle, on the verge of crying.
                            At first, you think he might have knocked out a tooth, but upon tongue-inspection, everything appears to be still in place.
                        [/p]
                        [p]
                            "[color=ppink]Do ye really think you're first lizard tah threaten meh?
                            I've dealt with far more imposin' [i]kuhbolds[/i] than you.
                            Smacked'em to hell and ba~ack, wrestled'em to the ground and made'em [b]beg[/b] for mercy.
                            What chance do [b]you[/b] stand against [b]me[/b]?
                            If ah shave the wool off mah body, arse to head, it'll form a pile bigger than you are!
                            Now, unless you want'ay get slapped again, get [b]out[/b] of mah soup kitchen.
                            We're [b]closed[/b].
                            [b]Forever[/b].[/color]"
                        [/p]
                        [p]
                            That being said, the fat sheep points towards the door.
                            You consider trying to fight him still, but he's right.
                            There's no point in getting the shit kicked out of you needlessly.
                            You simply aren't built to fight a man of his size.
                            Hanging your head, you walk out the door, back towards the market square.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\threaten\VaranarSoupKitchenThreaten.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\bluff\VaranarSoupKitchenBluffSucceed.js

/**
 * Overdressed section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenBluffSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.serveChoice;
    var changeController = false;

    var autoContinue = true;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.serveChoices;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You bluff, saying that you are a member of the Three-Tailed Rat Clan.
                            This explains why you're so well-dressed and a bit more [i]wealthy[/i] than the other kobolds around town.
                            Unless the sheep is looking for [b]trouble[/b], he'd best get to serving you, and make it snappy.
                            The clan can make life [b]very[/b] difficult for [i]uncooperative[/i] people.
                            Surely he doesn't want to get on their -- you mean [b]your[/b] -- bad side one day after [i]retirement[/i], right?
                        [/p]
                        ${global.ui.condSucc(`Strength greater than 5`)}
                        [p]
                            "[color=ppink]Alright, alright.
                            I knew ye were bad luck from the moment yah walked in here.
                            No [i]kuhbold[/i] flaunts their wealth about like that, apart from the ones that mean [b]trouble[/b].
                            I'm not lookin' for a fight.
                            I don't want any [b]trouble[/b].
                            Lemme get ye somethin' tah eat, okay?
                            And we'll forget this whole [i]misunderstanding[/i].[/color]"
                        [/p]
                        [p]
                            You tell the sheep to stop [b]yapping[/b] and get started already.
                            You're [b]hungry[/b].
                            The fat man steps behind the counter,
                            opens up a cupboard, and produces a rusty metal pot half-filled with day-old soup.
                            You take a seat at the table nearest to you.
                            He uses a wooden ladle to scoop some soup onto a deep plate,
                            having a quick taste of his own culinary prowess in the process.
                            His face scrunches up, signalling his disgust.
                            If [b]he[/b] doesn't like it, you very much so doubt [b]you[/b] will.
                        [/p]
                        [p]
                            Your dinner prepared, your host brings it over to your table.
                            He crudely tosses the plate in front of you, the viscous, red liquid within barely contained, almost splashing out onto you.
                            It looks to be some kind of [i]tomato[/i] soup, with chunks of what you presume to once have been [i]vegetables[/i].
                            Having served you your meal, the half-naked caprine takes a seat opposite to you, chair creaking loudly under his weight.
                        [/p]
                        [p]
                            The sheep reaches into the wool around his neck, and fetches out something you hadn't even noticed before:
                            a small pouch, similar to your coin-purse, almost [i]identical[/i].
                            He places the bag on the table.
                            "[color=ppink]Y'see this here?
                            It's a memento.
                            Wedding ring.
                            Belonged to mah waif.
                            Most beautiful ewe in all of the world, she was.
                            Not too [i]faithful[/i],
                            but I looked past those flaws to see the [b]diamond[/b] hidden underneath all the layers of foolin' 'round with the neighbours.
                            It was [b]her[/b] idea to start this here soup kitchen, and in her memory I continued the business.
                            'Till yesterday, that is.[/color]"
                        [/p]
                        [p]
                            A sudden loud noise from the back of the building interrupts the chubby man's sad story.
                            Sounds like something [i]falling[/i] and [b]breaking[/b].
                            The caprine rises to his hooves.
                            "[color=ppink]Baa~aah, it's probably those [b]filthy[/b] thieves again.
                            You stay here, [i]kuhbold[/i] of clan whatever.
                            I'll go scare'em off.[/color]"
                        [/p]
                        [p]
                            That being said, the sheep rushes behind the counter once more,
                            before disappearing into a side room, leaving you all alone with your plate of [i]slop[/i].
                            Soon after, you hear his loud, profanity-filled yelling along with angry bleating directed at whoever attempted to sneak inside.
                            Looks like the fat man forgot the bag with the ring in on the table.
                            It'd be a shame if someone [i]took it[/i].
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\bluff\VaranarSoupKitchenBluffSucceed.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\bluff\VaranarSoupKitchenBluffFail.js

/**
 * Overdressed section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenBluffFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;
    var changeController = ctrl.bluffFail;

    var autoContinue = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.apple;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You bluff, saying that you are a member of the Three-Tailed Rat Clan.
                            This explains why you're so well-dressed and a bit more [i]wealthy[/i] than the other kobolds around town.
                            Unless the sheep is looking for [b]trouble[/b], he'd best get to serving you, and make it snappy.
                            The clan can make life [b]very[/b] difficult for [i]uncooperative[/i] people.
                            Surely he doesn't want to get on their -- you mean [b]your[/b] -- bad side one day after [i]retirement[/i], right?
                        [/p]
                        ${global.ui.condFail(`Strength below 6`)}
                        [p]
                            "[color=ppink]You?
                            A [b]gang member[/b]?
                            Don't make me [b]laugh[/b], [i]kuhbold[/i].
                            With [b]wimpy[/b] little arms like that, ye can be glad people don't mistake ye for an overgrown [b]grasshopper[/b].[/color]"
                            The man snorts, almost laughing at his own joke.
                            "[color=ppink]You're a bad liar, lizard.
                            Come now, we don't serve liars here.
                            Git.
                            Out the door.
                            Or do I need tah [b]throw[/b] yeh out?[/color]"
                        [/p]
                        [p]
                            No.
                            [b]No![/b]
                            Does he know what he's doing?!
                            You're part of a [b]crime syndicate[/b]!
                            You'll have him [b]killed[/b]!
                            "[color=ppink]Yeah right.
                            And me wife, Armok rest'er soul, was a faithful woman, too?
                            Do ye expect me tah believe everythin' without questionin', lizard?
                            Yer a dirty, no-good liar.
                            Now get out of mah soup kitchen.
                            We're [b]closed[/b].
                            [b]Forever[/b].[/color]"
                        [/p]
                        [p]
                            That being said, the fat sheep points towards the door.
                            You wordlessly continue you act, in a last ditch attempt to get him to let you stay.
                            He looks like he's made his choice.
                            You try to come up with something, [b]anything[/b] to say, but it's hopeless.
                            Your scrawny figure has cost you a valuable opportunity.
                            Hanging your head, you walk out the door, back towards the market square.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\bluff\VaranarSoupKitchenBluffFail.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\bluff\VaranarSoupKitchenBluff.js

/**
 * Overdressed section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenBluff(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = VaranarSoupKitchenAccusedChoice.BLUFF;
    var reqController = ctrl.accused;

    var toFail = new VaranarSoupKitchenBluffFail(ctrl);
    var toSuccess = new VaranarSoupKitchenBluffSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.strength() >= 6;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\bluff\VaranarSoupKitchenBluff.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\VaranarSoupKitchenOverdressedServed.js

/**
 * Serve section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenOverdressedServed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.serveChoice;
    var changeController = false;

    var autoContinue = true;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.serveChoices;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You give the sheep a small nod.
                            You [i]had[/i] hoped for a [i]warm[/i] meal, but beggars can't be choosers.
                            "[color=ppink]Don't worry 'bout the baa~aath.
                            It'll still be there when you're done.
                            I prefer me water cold anyway.[/color]"
                            You can sense a hint of [i]sarcasm[/i] in his voice.
                            "[color=ppink]Anyway, lemme get ye somethin' tah eat.[/color]"
                        [/p]
                        [p]
                            Patting you on the head, the fat man steps behind the counter,
                            opens up a cupboard, and produces a rusty metal pot half-filled with day-old soup.
                            You take a seat at the table nearest to you.
                            The sheep uses a wooden ladle to scoop some soup onto a deep plate,
                            having a quick taste of his own culinary prowess in the process.
                            His face scrunches up, signalling his disgust.
                            If [b]he[/b] doesn't like it, you very much so doubt [b]you[/b] will.
                        [/p]
                        [p]
                            Your dinner prepared, your host brings it over to your table.
                            He crudely tosses the plate in front of you, the viscous, red liquid within barely contained, almost splashing out onto you.
                            It looks to be some kind of [i]tomato[/i] soup, with chunks of what you presume to once have been [i]vegetables[/i].
                            Having served you your meal, the half-naked caprine takes a seat opposite to you, chair creaking loudly under his weight.
                        [/p]
                        [p]
                            The sheep reaches into the wool around his neck, and fetches out something you hadn't even noticed before:
                            a small pouch, similar to your coin-purse, almost [i]identical[/i].
                            He places the bag on the table.
                            "[color=ppink]Y'see this here?
                            It's a memento.
                            Wedding ring.
                            Belonged to mah waif.
                            Most beautiful ewe in all of the world, she was.
                            Not too [i]faithful[/i],
                            but I looked past those flaws to see the [b]diamond[/b] hidden underneath all the layers of foolin' 'round with the neighbours.
                            It was [b]her[/b] idea to start this here soup kitchen, and in her memory I continued the business.
                            'Till yesterday, that is.[/color]"
                        [/p]
                        [p]
                            A sudden loud noise from the back of the building interrupts the chubby man's sad story.
                            Sounds like something [i]falling[/i] and [b]breaking[/b].
                            The caprine rises to his hooves.
                            "[color=ppink]Baa~aah, it's probably those [b]filthy[/b] thieves again.
                            You enjoy yer meal, [i]kuhbold[/i].
                            I'll go scare'em off.[/color]"
                        [/p]
                        [p]
                            That being said, the sheep rushes behind the counter once more,
                            before disappearing into a side room, leaving you all alone with your plate of [i]slop[/i].
                            Soon after, you hear his loud, profanity-filled yelling along with angry bleating directed at whoever attempted to sneak inside.
                            Looks like the fat man forgot the bag with the ring in on the table.
                            It'd be a shame if someone [i]took it[/i].
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\VaranarSoupKitchenOverdressedServed.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\VaranarSoupKitchenAccused.js

/**
 * Overdressed section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenAccused(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.accused;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new VaranarSoupKitchenBluff(ctrl);
    var secondNext = new VaranarSoupKitchenThreaten(ctrl);

    var firstChoice = VaranarSoupKitchenAccusedChoice.BLUFF;
    var secondChoice = VaranarSoupKitchenAccusedChoice.THREATEN;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You give the sheep a small nod.
                            You [i]had[/i] hoped for a [i]warm[/i] meal, but beggars can't be choosers.
                            "[color=ppink]Don't worry 'bout the baa~aath.
                            It'll still be there when you're done.
                            I prefer me water cold anyway.[/color]"
                            You can sense a hint of [i]sarcasm[/i] in his voice.
                        [/p]
                        [p]
                            The jolly look on the caprine's face is suddenly replaced by an inquisitive stare as he notices the small bag hanging off your neck.
                            "[color=ppink]Now wait just a minute.
                            What dah we have 'ere?[/color]"
                            He taps one of his fat fingers against your coin pouch, making the gold pieces within clatter against one another.
                            "[color=ppink]Stingy rich'un, aren't yah?
                            Trying tah score a free meal while havin' more moneh than ah do?!
                            Ye get nothin', and yah can be happeh that ah don't kick yeh out mahself![/color]"
                        [/p]
                        ${global.ui.condFail(`Too much gold`)}
                        [p]
                            The burly man's accent seems to grow [i]heavier[/i] the more angry he gets.
                            You need to come up with something fast if you want to get a free meal still.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Bluff</button>`;

        var tooltip = global.parser.parse(`Claim to be part of the Ten Paw Gang. [i]Technically[/i], you are.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Threaten</button>`;

        var tooltip = global.parser.parse(`Who does this old man think he is? Tell him to get you some food, [b]or else[/b].`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var VaranarSoupKitchenAccusedChoice = {
    BLUFF: 1,
    THREATEN: 2
};

// END scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\VaranarSoupKitchenAccused.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\VaranarSoupKitchenAccuseSelector.js

/**
 * Overdressed section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenAccuseSelector(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.overdressed;

    var toFail = new VaranarSoupKitchenOverdressedServed(ctrl);
    var toSuccess = new VaranarSoupKitchenAccused(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.gold() >= 50;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\VaranarSoupKitchenAccuseSelector.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\VaranarSoupKitchenOverdressed.js

/**
 * Overdressed section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenOverdressed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.overdressed;
    var changeController = ctrl.overdressed;

    var autoContinue = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new VaranarSoupKitchenAccuseSelector(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var clothes = ctrl.misc.bra() || ctrl.misc.shirt() || ctrl.misc.pants() || ctrl.misc.underwear();
        var toReturn = `[p]
                            Gathering your courage, you walk up to the door, and slowly push it open.
                            Despite taking little effort to move, it creaks loudly, alerting whoever may be inside to your presence.
                            Swallowing deeply, you step inside of the dilapidated building, and close the door behind you.
                        [/p]
                        [p]
                            You find yourself alone in a dusky room that has clearly seen better days.
                            Apart from the chairs and tables arrayed across the wooden floor, you'd say this almost looks like an abandoned [i]store[/i].
                            At least it's warm in here, so the soup kitchen can't be [i]completely[/i] deserted.
                            You figure the stove must be behind the mercantile countertop, but you're not tall enough to see.
                            To your left is a set of stairs which you figure must lead to the proprietor's [i]above-ground[/i] home.
                        [/p]
                        [p]
                            Suddenly, you hear movement above.
                            Footsteps.
                            No wait.
                            Clacking hooves?
                            "[color=ppink]Yeah, ah heard yah! I'ma comin'![/color]" a bass voice rings out from the second floor.
                        [/p]
                        [p]
                            Down the stairs comes storming a big, wet, outright [b]fat[/b] anthromorphic sheep,
                            his bloated, wool-covered belly jiggling with every step he takes,
                            soapy water dripping liberally from his fleece.
                            The man isn't wearing much besides a towel, tied snugly around his waist.
                            "[color=ppink]Can't a sheep take a fockin' bath wifout a beggar wanderin' in?![/color]"
                            At least you don't have to ask why he's wet and half-naked.
                            "[color=ppink]Well then, what'd ye want, yah skinky lizard?![/color]"
                        [/p]
                        [p]
                            The caprine comes to a rest merely a foot away from you, the wool on his big gut almost pressing into your snout.
                            You can't help but stare at his massive belly,
                            as you stammer out the dumbest question your mind could think of while looking at someone big enough to [b]eat[/b] you.
                            You ask whether or not the man [i]cooks[/i] kobolds here.
                            He leans over you threateningly, a droplet of water dripping from his chin down ontop of your forehead.
                            Glaring at you, he responds to your question in a grim tone.
                            "[color=ppink]Aye, but only the ones that think they're funny.
                            Or the ones that [b]stare at mah belly[/b].[/color]"
                        [/p]
                        [p]
                            His words remind you that you [i]are[/i] still leering at his bulky lower body instead of looking the man in the eyes.
                            Correcting your mistake, you quickly stammer out a meek apology.
                            The sheep suddenly bursts out into laughter.
                            "[color=ppink]Hahahahaha!
                            Scared ye, didn't ah?
                            Only vegetarian meals here, [i]kuhbold[/i].
                            Us sheep, we cannae eat meat.[/color]"
                            You let out an uneasy chuckle, a disturbed smile forming on your face.
                            At least you won't be getting cut up and put into a stew today.
                        [/p]
                        [p]
                            You nervously ask the [i]chef[/i] if you could maybe, if at all possible, if it's not a bother, get something tiny to eat, perhaps?
                            You know you're [i]interrupting[/i] his bath-time, but you're kind of [b]starving[/b].
                            "[color=ppink]'Fraid not.[/color]" the caprine responds, with a hint of sadness in his voice.
                            "[color=ppink]I closed down shop yesterday.
                            Not enough money to pay rent for this shitehole.
                            I was half expectin' yah tah be here tah [b]kick me out[/b].[/color]"
                            Even his [i]endearing[/i] accent fails to mask the man's distress.
                        [/p]
                        [p]
                            The chubby sheep lets out a sigh, eyeing you up and down.
                            "[color=ppink]Fancy ${clothes ? `clothing` : `weapon`}.
                            Who'd ye steal that from?
                            Tell yah what, for a well-${clothes ? `dressed` : `armed`} lizard like you, I'll make an exception.
                            I've got some leftovers from yesterday.
                            Won't be warm, but it's better than nothin', aye?[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        global.ui.applyTheme("hamlet_shop");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\overdressed\VaranarSoupKitchenOverdressed.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\arrival\VaranarSoupKitchenEntrance.js

/**
 * Arrival section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenEntrance(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.arrival;

    var toFail = new VaranarSoupKitchenInvited(ctrl);
    var toSuccess = new VaranarSoupKitchenOverdressed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.equipmentWorth() >= 35;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\arrival\VaranarSoupKitchenEntrance.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\arrival\VaranarSoupKitchenArrival.js

/**
 * Arrival section of Varanar, Noon Three.
 * @param {VaranarSoupKitchenController} controller The Varanar, Noon Three controller.
 */
function VaranarSoupKitchenArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.arrival;
    var changeController = false;

    var autoContinue = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new VaranarSoupKitchenEntrance(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var gold = ctrl.misc.gold();
        var paid = global.kobold.history.shared.debtCollector.paid;
        var toReturn = `[p]
                            Lingering around the market square, you can feel your stomach begin to grumble.
                            It's been a while since you've last had [i]anything[/i] to eat.
                            ${gold > 30 ? `You have plenty of gold to spare,
                            but you'd rather save it to pay off ${paid ? `the rest of` : ``}
                            your debt.` : `${gold > 10 ? `You have a few coins,
                            but you'd rather save them to pay off ${paid ? `the rest of` : ``}
                            your debt.` : `You have no money to buy food with, though.`}`}
                            While considering [b]scrounging[/b] the town square for any scraps of food leftover by yesterday's market,
                            your eye catches a peculiar sign hanging above the doorway to a rundown-looking structure at the edge of the plaza.
                        [/p]
                        [p]
                            [color=brown][big][b]KOBOLD SOUP KITCHEN[/b][/big][/color]
                        [/p]
                        [p]
                            That could mean two things.
                            Either you can get something to eat for free within the decrepit place,
                            or they [i]cook[/i] and [i]serve[/i] [b]kobolds[/b] as food.
                            You hope it's the former, and not the latter.
                            Judging by the state of the building, business must not be very good either way.
                            How does someone offering free stuff even make a living?
                            You're too hungry to properly pursue that train of thought.
                        [/p]
                        [p]
                            You remotely recall a few [i]rumors[/i] you heard about a year or two ago, about a place like this,
                            where tiny lizards down on their luck can take a break from scrounging and enjoy a hot meal for once.
                            You figured it was just hearsay, that they were [i]lies[/i], [b]fairytales[/b],
                            that it was your koboldic forest [i]friends[/i] playing a [b]trick[/b] on you,
                            or even the Three-Tailed Rat Clan attempting to lure you into town for whatever reason.
                            Maybe you were told the truth after all.
                            Or you could be walking into a death trap.
                        [/p]
                        [p]
                            Standing in front of the building itself, not to mention with an empty stomach,
                            you are more inclined to [i]believe[/i] the stories you've been told.
                            You decide you'll take a quick peek inside, and if anything seems fishy to you,
                            you'll immediately leave and it'll be like nothing happened.
                            At best, you'll get a free meal.
                            At worst, you'll be [i]turned into[/i] a meal.
                            Not like you've much meat on your bones anyway.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        if (autoContinue) {
            subController.continue();
        }

        global.ui.applyTheme("varanar_generic");
        ctrl.bgm.loop();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += (!autoContinue && buildChoice()) || "";
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\arrival\VaranarSoupKitchenArrival.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\VaranarSoupKitchenController.js

/**
 * Controller for the Varanar, Noon Three scene.
 * @param {string} sceneId The id of the scene.
 */
function VaranarSoupKitchenController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.nextSceneReady = false;
    this.nextSceneName = "SlaveAfternoonTwo";
    this.nextFolderName = "Enslavement/DayTwo";

    this.misc = new MiscController(id, "misc");

    this.arrival = new ContinueController(id, "arrival");

    this.invited = new ChangeController(id, "invited", function () {
        global.kobold.status.mental.decrease(2, "Caprine intimidation");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Intimidated", "Sanity", -2)
        ]);
    });

    this.overdressed = new ChangeController(id, "overdressed", function () {
        global.kobold.status.mental.decrease(2, "Caprine intimidation");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Intimidated", "Sanity", -2)
        ]);
    });

    this.questioned = new ChoiceController(id, "questioned");
    this.accused = new ChoiceController(id, "accused");

    this.serveChoice = new ContinueController(id, "serveChoice");

    this.serve = new ChangeController(id, "serve", function () {
        global.kobold.status.stamina.increase(5, "Taking a seat");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Temporary respite", "Stamina", +5)
        ]);
    });

    this.steal = new ChangeController(id, "steal", function () {
        global.kobold.status.physical.decrease(10, "Starvation");
        global.kobold.status.mental.decrease(5, "Stealing wedding ring");
        global.kobold.status.stamina.decrease(2, "Short sprint");
        global.kobold.inventory.addItem(new GoldRing());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Starvation", "Health", -10),
            new Change("Keepsake taken", "Sanity", -10),
            new Change("Valuable loot", "Sanity", +5),
            new Change("Short sprint", "Stamina", -2),
            new Change("Loot", global.kobold.inventory.createItemWithTooltip(new GoldRing(), true), +1)
        ]);
    });

    this.leave = new ChangeController(id, "leave", function () {
        global.kobold.status.physical.increase(3, "Eating half an apple");
        global.kobold.status.mental.decrease(2, "Eating scraps");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Food", "Health", +5),
            new Change("Unsanitary", "Health", -2),
            new Change("Eating scraps", "Sanity", -2)
        ]);
    });

    this.eat = new ChangeController(id, "eat", function () {
        global.kobold.status.physical.increase(10, "Eating soup");
        global.kobold.status.mental.increase(10, "Eating soup");
        global.kobold.status.stamina.increase(10, "Eating soup");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Food", "Health", +10),
            new Change("Decent meal", "Sanity", +10),
            new Change("Ate", "Stamina", +10)
        ]);
    });

    this.eatLeave = new ChangeController(id, "eatLeave", function () {
        global.kobold.status.mental.decrease(8, "Failure to help");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Can't help", "Sanity", -4),
            new Change("Sadness", "Sanity", -4)
        ]);
    });

    this.offer = new ChoiceController(id, "offer");

    this.mention = new ChangeController(id, "mention", function () {
        global.kobold.status.mental.increase(17, "Gave home to sheep");
        global.kobold.inventory.bag.removeMisc(new CassidyHouseKey());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Helped sheep", "Sanity", +6),
            new Change("Helped kobolds", "Sanity", +6),
            new Change("Conscience cleared", "Sanity", +6),
            new Change("Rudeness", "Sanity", -1)
        ]);
    });

    this.noMention = new ChangeController(id, "noMention", function () {
        global.kobold.status.mental.increase(10, "Gave home to sheep");
        global.kobold.inventory.bag.removeMisc(new CassidyHouseKey());
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Helped sheep", "Sanity", +6),
            new Change("Helped kobolds", "Sanity", +6),
            new Change("Troubled conscience", "Sanity", -2)
        ]);
    });

    this.lieFail = new ChangeController(id, "lieFail", function () {
        global.kobold.status.mental.decrease(4, "Failing to lie");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Failed to lie", "Sanity", -4)
        ]);
    });

    this.bluffFail = new ChangeController(id, "bluffFail", function () {
        global.kobold.status.mental.decrease(4, "Failing to bluff");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Failed to bluff", "Sanity", -4)
        ]);
    });

    this.truthFail = new ChangeController(id, "truthFail", function () {
        global.kobold.status.mental.decrease(8, "Unbelieving sheep");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Truth not enough", "Sanity", -4),
            new Change("Frustration", "Sanity", -4)
        ]);
    });

    this.threaten = new ChangeController(id, "threaten", function () {
        global.kobold.status.physical.decrease(4, "Getting slapped");
        global.kobold.status.mental.decrease(4, "Failing to intimidate");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Heavily slapped", "Health", -4),
            new Change("Failed to intimidate", "Sanity", -4),
        ]);
    });

    this.preContinue = new ContinueController(id, "preContinue");

    this.halfApple = new ChangeController(id, "halfApple", function () {
        global.kobold.status.physical.increase(3, "Eating half an apple");
        global.kobold.status.mental.decrease(2, "Eating scraps");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Food", "Health", +5),
            new Change("Unsanitary", "Health", -2),
            new Change("Eating scraps", "Sanity", -2)
        ]);
    });

    this.fragments = {};
    this.fragments.apple = new VaranarSoupKitchenHalfApple(this);
    this.fragments.serveChoices = new VaranarSoupKitchenServed(this);

    this.bgm = new AudioController(id, "res/audio/varanar/decline.mp3", 70);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

}

// END scenes\Varanar\DayThree\VaranarSoupKitchen\VaranarSoupKitchenController.js

// scenes\Varanar\DayThree\VaranarSoupKitchen\VaranarSoupKitchen.js

/**
 * Noon three of Varanar.
 */
function VaranarSoupKitchen() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "VaranarSoupKitchen";
    var sceneName = "Varanar, Noon Three";
    var sceneSubfolder = "Varanar/DayThree";
    var ctrl = new VaranarSoupKitchenController(sceneId);
    var arrival = new VaranarSoupKitchenArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]Varanar[/t]`);
        tab += global.parser.parse(`[st]Noon three[/st]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new VaranarSoupKitchen();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\Varanar\DayThree\VaranarSoupKitchen\VaranarSoupKitchen.js



// END SCENE VaranarSoupKitchen



// SCENE CaravanConfinement

// scenes\Enslavement\CaravanConfinement\cart\break\CaravanConfinementBreakFail.js

/**
 * Break section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementBreakFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide it'd be best to at least [b]try[/b] to escape the bandits' clutches as soon as possible.
                            In fact, why not start straight away?
                            The rope tied around your wrists feels kind of [i]weak[/i].
                            Old and ragged.
                            It's probably spent [b]years[/b] withering away on the outside of this very cart.
                            You think you might be able to [b]break[/b] it, if you try hard enough.
                        [/p]
                        [p]
                            With all of your strength, you pull and you tug at your restraints, testing the limits of the rope.
                            For nearly half an hour, you strenuously attempt to break your bonds.
                            Alas, the rope seems stronger than you anticipated.
                            It's tied too low for you to reach with your claws, and you can't seem to rip it through sheer strength alone.
                            Realizing your attempt at escaping is an exercise in futility, you cease your incessant struggling.
                        [/p]
                        ${global.ui.condFail(`Strength below 6`)}
                        [p]
                            While you didn't manage to exactly [i]break[/i] your bonds,
                            it feels as though the mere act of [i]trying[/i] alone has strengthened your arms.
                            While not exactly powertraining, the botched breakout has left you feeling a bit more [i]powerful[/i].
                        [/p]
                        [p]
                            After such an exerting effort, you slowly drift to sleep.
                            You get the feeling you'll be needing every ounce of energy you can scrape together to survive the coming day.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.breakFail.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\break\CaravanConfinementBreakFail.js

// scenes\Enslavement\CaravanConfinement\cart\break\CaravanConfinementBreakSucceedContinue.js

/**
 * End section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementBreakSucceedContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;

    var continueId = arguments.callee.name + "ButtonOne"

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.nightOne.knockedOut = true;
        var toReturn = `[p]
                            After caressing your wrists for a few moments,
                            slightly easing the pain of the minor rope burn you sustained while struggling to break your bonds,
                            you rise to your feet within the cell-like wagon.
                            Bad idea.
                        [/p]
                        [p]
                            The equine guard stationed outside of your cart just so happened to be leering inside of your confined area the moment you decided to stand upright.
                            Noticing your unbound hands, he draws a hefty-looking ${global.kobold.inventory.createItemWithTooltip(new Club())} from his belt.
                            For a moment, he stares at you, and you stare at him.
                            It takes a few seconds for you to register the fact that you've been caught.
                            Then, panic begins to settle in.
                        [/p]
                        [p]
                            The horse-sentry calmly walks over to the cart's single door.
                            He unlocks and opens it.
                            This is your chance.
                            Your one shot.
                            If you can get past him, you can escape into the wilderness, and regain your freedom!
                        [/p]
                        [p]
                            Alas, instead of attacking the horse, bolting past him, reasoning with him or even [b]seducing[/b] him,
                            your indecisive mind causes your body to freeze up in fright.
                            Instead of making a heroic getaway, you cower in fear, like a whelp that knows it [b]fucked up[/b].
                            Still, the equine shows no mercy.
                            He throws an overhead swing with his ${global.kobold.inventory.createItemWithTooltip(new Club())},
                            the wood crashing against the top of your head.
                        [/p]
                        [p]
                            You hear something [b]crack[/b].
                            You doubt it's the wood.
                            Your vision goes black.
                            You slump down onto the floor, knocked unconscious in a single blow.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.breakSucceedContinue.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\break\CaravanConfinementBreakSucceedContinue.js

// scenes\Enslavement\CaravanConfinement\cart\break\CaravanConfinementBreakSucceed.js

/**
 * Break section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementBreakSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementBreakSucceedContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.nightOne.escapedBindings = true;
        var toReturn = `[p]
                            You decide it'd be best to at least [b]try[/b] to escape the bandits' clutches as soon as possible.
                            In fact, why not start straight away?
                            The rope tied around your wrists feels kind of [i]weak[/i].
                            Old and ragged.
                            It's probably spent [b]years[/b] withering away on the outside of this very cart.
                            You think you might be able to [b]break[/b] it, if you try hard enough.
                        [/p]
                        [p]
                            With all of your strength, you pull and you tug at your restraints, testing the limits of the rope.
                            For nearly half an hour, you strenuously attempt to break your bonds.
                            Right as you are about to give up hope, you can feel a single [i]fiber[/i] of the rope spring loose,
                            torn due to constant stress.
                            The cord is weakened by just a very little, but that smidgen of integrity-loss spurs you to continue your incessant struggling.
                        [/p]
                        ${global.ui.condSucc(`Strength greater than 5`)}
                        [p]
                            It seems as though that, with a single strand loosened, the cord becomes far easier to rip apart.
                            Every time you attempt to tug your wrists apart, a new string pops loose.
                            Slowly, but surely, the rope thins, more and more of it snapping with every yank,
                            until finally it has grown weak enough for you to break out completely.
                            With one last mighty show of force, you tear your wrists free from their bondage, the rope audibly snapping.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.breakSucceed.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\break\CaravanConfinementBreakSucceed.js

// scenes\Enslavement\CaravanConfinement\cart\break\CaravanConfinementBreak.js

/**
 * Break section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementBreak(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementCartChoice.BREAK;
    var reqController = ctrl.cart;

    var toFail = new CaravanConfinementBreakFail(ctrl);
    var toSuccess = new CaravanConfinementBreakSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.strength() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\break\CaravanConfinementBreak.js

// scenes\Enslavement\CaravanConfinement\cart\slip\CaravanConfinementSlipFail.js

/**
 * Slip section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementSlipFail(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide it'd be best to at least [b]try[/b] to escape the bandits' clutches as soon as possible.
                            In fact, why not start straight away?
                            The rope tied around your wrists feels kind of [i]loose[/i].
                            Improperly tied.
                            The wolf probably thought you weren't even going to bother trying to escape.
                            He was wrong.
                        [/p]
                        [p]
                            You wriggle and rotate your wrists,
                            you try to roll the rope alongside your scales by rubbing it against the floor,
                            and you contort your hands in ways you didn't even know were possible.
                            Almost succeeding to free yourself, you break out every single last trick in the book in an attempt to regain your freedom.
                            For nearly half an hour, you strenuously attempt to abscond your bonds.
                            Alas, the rope seems to be far more secure than you anticipated.
                            It's tied too low for you to reach with your claws, and you can't seem to rip it through sheer wrist-motion alone.
                            Realizing your attempt at escaping is an exercise in futility, you cease your incessant struggling.
                        [/p]
                        ${global.ui.condFail(`Agility below 6`)}
                        [p]
                            While you didn't manage to exactly [i]break[/i] your bonds,
                            it feels as though the mere act of [i]trying[/i] alone has made your arms a bit more flexible.
                            While not exactly a stretching drill, the botched breakout has left you feeling a bit more [i]agile[/i].
                        [/p]
                        [p]
                            After such a tiring effort, you slowly drift to sleep.
                            It's been a long day, and you get the feeling you'll be needing every ounce of energy you can scrape together to make it through tomorrow.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.slipFail.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\slip\CaravanConfinementSlipFail.js

// scenes\Enslavement\CaravanConfinement\cart\slip\CaravanConfinementSlipSucceedContinue.js

/**
 * End section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementSlipSucceedContinue(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preContinue;

    var continueId = arguments.callee.name + "ButtonOne"

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.nightOne.knockedOut = true;
        var toReturn = `[p]
                            After caressing your wrists for a few moments,
                            slightly easing the pain of the minor rope burn you sustained while struggling to break your bonds,
                            you rise to your feet within the cell-like wagon.
                            Bad idea.
                        [/p]
                        [p]
                            The equine guard stationed outside of your cart just so happened to be leering inside of your confined area the moment you decided to stand upright.
                            Noticing your unbound hands, he draws a hefty-looking ${global.kobold.inventory.createItemWithTooltip(new Club())} from his belt.
                            For a moment, he stares at you, and you stare at him.
                            It takes a few seconds for you to register the fact that you've been caught.
                            Then, panic begins to settle in.
                        [/p]
                        [p]
                            The horse-sentry calmly walks over to the cart's single door.
                            He unlocks and opens it.
                            This is your chance.
                            Your one shot.
                            If you can get past him, you can escape into the wilderness, and regain your freedom!
                        [/p]
                        [p]
                            Alas, instead of attacking the horse, bolting past him, reasoning with him or even [b]seducing[/b] him,
                            your indecisive mind causes your body to freeze up in fright.
                            Instead of making a heroic getaway, you cower in fear, like a whelp that knows it [b]fucked up[/b].
                            Still, the equine shows no mercy.
                            He throws an overhead swing with his ${global.kobold.inventory.createItemWithTooltip(new Club())},
                            the wood crashing against the top of your head.
                        [/p]
                        [p]
                            You hear something [b]crack[/b].
                            You doubt it's the wood.
                            Your vision goes black.
                            You slump down onto the floor, knocked unconscious in a single blow.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.slipSucceedContinue.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\slip\CaravanConfinementSlipSucceedContinue.js

// scenes\Enslavement\CaravanConfinement\cart\slip\CaravanConfinementSlipSucceed.js

/**
 * Slip section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementSlipSucceed(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preContinue;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementSlipSucceedContinue(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.nightOne.escapedBindings = true;
        var toReturn = `[p]
                            You decide it'd be best to at least [b]try[/b] to escape the bandits' clutches as soon as possible.
                            In fact, why not start straight away?
                            The rope tied around your wrists feels kind of [i]loose[/i].
                            Improperly tied.
                            The wolf probably thought you weren't even going to bother trying to escape.
                            He was wrong.
                        [/p]
                        [p]
                            You wriggle and rotate your wrists,
                            you try to roll the rope alongside your scales by rubbing it against the floor,
                            and you contort your hands in ways you didn't even know were possible.
                            Almost succeeding to free yourself, you break out every single last trick in the book in an attempt to regain your freedom.
                            For nearly half an hour, you strenuously attempt to abscond your bonds.
                            Just as you were about to give up, one last attempt at rubbing against the floor finally seems to pay off!
                            The rope rolls just a little higher up, closer to your actual hands.
                        [/p]
                        ${global.ui.condSucc(`Agility greater than 5`)}
                        [p]
                            Encouraged by the sense of progress, you reach down with your fingers.
                            While you can't quite escape your bindings yet, the rope is now in range of your claws!
                            You arduously begin to [i]file[/i] away at the cord, cutting through it with short, hasty finger-motions.
                            After another couple of minutes, the rope suddenly [b]snaps[/b] in half, freeing you from your bonds.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.slipSucceed.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\slip\CaravanConfinementSlipSucceed.js

// scenes\Enslavement\CaravanConfinement\cart\slip\CaravanConfinementSlip.js

/**
 * Break section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementSlip(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementCartChoice.SLIP;
    var reqController = ctrl.cart;

    var toFail = new CaravanConfinementSlipFail(ctrl);
    var toSuccess = new CaravanConfinementSlipSucceed(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return ctrl.misc.agility() > 5;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\slip\CaravanConfinementSlip.js

// scenes\Enslavement\CaravanConfinement\cart\bang\CaravanConfinementBang.js

/**
 * Bang section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementBang(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementCartChoice.BANG;
    var reqController = ctrl.cart;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You decide it'd be best to at least [b]try[/b] to escape the bandits' clutches as soon as possible.
                            In fact, why not start straight away?
                            The wood the floor of this cart is made out of looks kind of [i]weak[/i].
                            It's seen better days, and you doubt it was very sturdy to begin with.
                            You figure you might be able to break it.
                        [/p]
                        [p]
                            But with what?
                            Your arms are tied, and laying on the ground, your legs aren't of much use.
                            If you get up, a guard might see what you're doing.
                            That leaves only one option: your head.
                        [/p]
                        [p]
                            You sigh loudly.
                            You [b]need[/b] to get out of here, and this is the only way you're going to accomplish that.
                            You slowly lift your head off the floor, before [b]bashing[/b] it against the frail wood.
                            You barely refrain from yelping out in pain, tears forming at the corners of your eyes.
                            Your head feels like it cracked open as much as the wood did, if not more.
                        [/p]
                        [p]
                            While hurting, you need to press on.
                            You've already taken the first step.
                            You're already [b]in pain[/b].
                            It'd be an outright [b]waste[/b] to stop now.
                            Gritting your teeth, your rear your head back once more.
                            Again, you [b]slam[/b] it downwards, your already bruised forehead feeling like it's about to [b]explode[/b].
                            The wood creaks, but it does not crack.
                            Just [b]once more[/b].
                        [/p]
                        [p]
                            Your teeth chattering, you lift your head for a third time.
                            You close your eyes.
                            It hurts [b]so much[/b].
                            Your breathing is ragged.
                            You [b]need[/b] to do this.
                            Swallowing deeply, you [b]bash[/b] the floor with your forehead a third time.
                        [/p]
                        [p]
                            A loud [b]crack[/b] resonates throughout the cart.
                            You open your eyes.
                            Blood stains the floor, which seems to have held, despite your best efforts.
                            That must mean that thing that cracked must have been...your head?
                        [/p]
                        [p]
                            Suddenly, your sight goes blurry.
                            Soon after, it fades to black.
                            You slump down lifelessly onto the wooden wagon floor.
                            Your breathing slows to a crawl.
                            You pass out, the last thought on your mind being that
                            perhaps this wasn't the brightest idea after all.
                        [/p]`;

        return global.parser.parse(toReturn);
    }
    
    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.bang.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\bang\CaravanConfinementBang.js

// scenes\Enslavement\CaravanConfinement\cart\count\CaravanConfinementCount.js

/**
 * Count section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementCount(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementCartChoice.COUNT;
    var reqController = ctrl.cart;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With nothing better to do, you decide to pass the time by counting the iron bars lining your [i]mobile cell[/i].
                            Starting at the left-hand side, you count:
                        [/p]
                        [p]
                            [i]One...two...three...four...five...six...seven...[/i]
                            You yawn.
                            [i]Eight...nine...ten...eleven...twelve...[/i]
                            Your eyelids are growing heavy.
                            [i]Thirteen...fourteen...fifteen...[/i]
                            Slowly, but steadily, your counting stops.
                            Your eyes close, and you drift away into sweet unconsciousness.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.count.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\count\CaravanConfinementCount.js

// scenes\Enslavement\CaravanConfinement\cart\gaze\CaravanConfinementGaze.js

/**
 * Gaze section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementGaze(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementCartChoice.GAZE;
    var reqController = ctrl.cart;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With nothing better to do, you decide to pass the time by gazing at the stars,
                            which you can see clearly through the metal bars of your [i]cell[/i].
                            You've always been fascinated with celestial things.
                            Those twinkling torches high up in the night sky have been the object of your dreams for many years now.
                        [/p]
                        [p]
                            You roll onto your back, and stare into the void.
                            The pitch blackness dotted with sparkling stars is soothing and easing.
                            You've been through [b]hell[/b] today.
                            You're glad to finally have a quiet moment in which you can [i]unwind[/i] and [b]think[/b] about things.
                        [/p]
                        [p]
                            Sure, you've been captured, but this isn't the first time you've botched a heist.
                            You've never actually been in a [i]cell[/i] before, though.
                            In the past, you've always managed to slip out of the greasy palms of whatever guardsman or vigilante caught you [i]before[/i] getting imprisoned.
                            Oh well, first time for everything.
                            You'll find a way to escape yet.
                            Just need to wait for the perfect opportunity to come along.
                        [/p]
                        [p]
                            You begin to relax, your eyelids growing more weary with every passing second.
                            Suddenly, it seems as though your cell is illuminated a little bit more than usual.
                            Like the stars themselves are shining their light down upon you.
                            After a few moments, the sudden brightness fades, leaving nothing but a lasting impression.
                            Come to think of it, your eyes weren't even open when it happened.
                            Perhaps you were already dreaming.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.gaze.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\gaze\CaravanConfinementGaze.js

// scenes\Enslavement\CaravanConfinement\cart\talk\CaravanConfinementTalk.js

/**
 * Talk section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementTalk(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementCartChoice.TALK;
    var reqController = ctrl.cart;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With nothing better to do, you decide to pass the time by striking a conversation with the guard outside of your cell.
                            You whisper to him, yet there is no response.
                            Raising your voice a little, you casually begin to talk about how you're having a [b]terrible[/b] day.
                            You ask him how [b]his[/b] day is going.
                            There is no response.
                        [/p]
                        [p]
                            Taking his silence for attentivity,
                            you begin to chat about that one time you caught a [b]really[/b] large fish out of the river, with your bare hands!
                            You ask him if he's ever gone fishing.
                            You've seen the way humans and most anthros fish -- using poles and stuff.
                            That's not the way kobolds do it.
                            You just kind of jump in and try to [i]grab[/i] one.
                            You guess that would be kind of hard for someone lacking [i]claws[/i].
                        [/p]
                        [p]
                            That's not to say you wouldn't give the human method a try.
                            It seems relaxing, just sitting in a small boat, or at the river shore, waiting for something to tug on your line.
                            But then again, in [i]this[/i] forest, you doubt you'd be alone for long before attracting some [i]unwanted company[/i].
                            Besides, that fish you told him about?
                            Yeah, you got halfway through cooking that over a campfire before a gang jumped out of the bushes and made you run for your life.
                            That was the last time you tried to cook a catch instead of eating it raw.
                        [/p]
                        [p]
                            You ask the man whether or not he likes to eat fish.
                            You personally prefer meat.
                            You haven't had any proper meat in [b]ages[/b], though.
                            The stuff you usually eat is refuse left by [i]things[/i] [b]bigger[/b] and [b]meaner[/b] than you are.
                            Sometimes, hunters from the city don't even bother taking home their catch.
                            All the better for you, of course.
                        [/p]
                        [p]
                            The person you're talking to still hasn't said a single word.
                            You're starting to think they either [b]can't hear[/b] you,
                            or maybe they're simply not interested in talking to an annoying little kobold they're forced to [b]babysit[/b].
                            You awkwardly say it's been nice talking to them, whoever they are, and then you stop speaking to the wooden wall in front of you.
                        [/p]
                        [p]
                            Truth be told, this actually [b]was[/b] a pleasant experience, regardless of whether they even [b]heard you[/b] or not.
                            You got to reflect on some things and you remembered some pleasant memories.
                            With those in mind, you roll onto your back, close your eyes, and try to get some sleep.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(ctrl.nextSceneName, ctrl.nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.talk.change();
        part += buildChoice();
        if (!ctrl.nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (ctrl.nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\CaravanConfinement\cart\talk\CaravanConfinementTalk.js

// scenes\Enslavement\CaravanConfinement\cart\CaravanConfinementCart.js

/**
 * Cart section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementCart(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.postTight;
    var subController = ctrl.cart;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";
    var fourthId = arguments.callee.name + "ButtonFour";
    var fifthId = arguments.callee.name + "ButtonFive";
    var sixthId = arguments.callee.name + "ButtonSix";

    var firstNext = new CaravanConfinementBreak(ctrl);
    var secondNext = new CaravanConfinementSlip(ctrl);
    var thirdNext = new CaravanConfinementBang(ctrl);
    var fourthNext = new CaravanConfinementCount(ctrl);
    var fifthNext = new CaravanConfinementGaze(ctrl);
    var sixthNext = new CaravanConfinementTalk(ctrl);

    var firstChoice = CaravanConfinementCartChoice.BREAK;
    var secondChoice = CaravanConfinementCartChoice.SLIP;
    var thirdChoice = CaravanConfinementCartChoice.BANG;
    var fourthChoice = CaravanConfinementCartChoice.COUNT;
    var fifthChoice = CaravanConfinementCartChoice.GAZE;
    var sixthChoice = CaravanConfinementCartChoice.TALK;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var hole = "rear end";
        if (male && fucked) {
            hole = "cockslit";
        }
        if (!male && !fucked) {
            hole = "cunt";
        }
        var toReturn = `[p]
                            Satisfied with his [i]discovery[/i], the caravan master removes his finger from your ${hole}, and rises to his feet.
                            "[color=brown]That's all I needed to know.
                            Shove'em in one of the empty carts near the back.
                            Station an extra watchman to ensure our little [i]guest[/i] stays put during the night.
                            We'll split off from the main caravan once we get close enough to the city
                            for the guards to ensure your men won't go running off with the gold.[/color]"
                        [/p]
                        [p]
                            Fenrir, or so the human called him earlier, scoffs at the notion that his group would break a bargain without being told to do so.
                            After a bit of uninteresting back and forth idle chatter,
                            the wolf grabs you by the collar, much like before,
                            and sets off for the rear end of the convoy,
                            tugging you along with.
                        [/p]
                        [p]
                            Arriving at the second-to-last transport carriage, the canine comes to a halt.
                            "[color=gblue]This is your stop.[/color]" he speaks, unlocking and opening the cart.
                            Grabbing a piece of rope hanging off the side of the vehicle, the buff lupine deftly ties your wrists together behind your back.
                            You don't even bother [i]trying[/i] to resist.
                            Your hands safely secured, the wolf casually lifts you up off the ground.
                            In a surprisingly gentle fashion, he rests your body down on the wooden carriage floor.
                        [/p]
                        [p]
                            "[color=gblue]Try to get some sleep.
                            Wouldn't want you to look weary during the auction tomorrow, now would we?
                            Remember: if you don't sell, your ass is [b]mine[/b].
                            And I share [b]everything[/b] with the rest of my crew.[/color]"
                            His mixture between a [i]threat[/i] and a sarcastic [i]pseudo-invitation[/i] shared,
                            the wolf closes the cart's sole door, granting you some peace and quiet after a long and arduous day.
                        [/p]
                        [p]
                            You're tired and exhausted, both mentally and physically, but you think you have enough energy left for one final act before going to sleep.
                            Outside of the cart, you hear footsteps.
                            Probably your personal guard, here to ensure you won't be [i]leaving[/i].
                            As if you could escape with your hands tied.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled} style="width: 190px;">Break bindings</button>`;

        var tooltip = global.parser.parse(`Attempt to break the rope holding your hands together.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled} style="width: 190px;">Slip from bindings</button>`;

        var tooltip = global.parser.parse(`Try to remove your hands from the rope binding them together.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled} style="width: 190px;">Use your head</button>`;

        var tooltip = global.parser.parse(`Wantonly bang your forehead against the side of the cart. Maybe the wood'll break.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }
    /**
     * Builds and returns the choice button.
     */
    function buildFourthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === fourthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fourthId}" ${disabled} style="width: 190px;">Count bars</button>`;

        var tooltip = global.parser.parse(`Count your makeshift cell's iron bars to pass the time.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFifthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === fifthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fifthId}" ${disabled} style="width: 190px;">Stargaze</button>`;

        var tooltip = global.parser.parse(`You can see some of the stars through the cart's small, barred windows. Let the night sky calm you.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSixthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === sixthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${sixthId}" ${disabled} style="width: 190px;">Talk</button>`;

        var tooltip = global.parser.parse(`Strike a conversation with the sentry posted outside of your prison.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += `<div class="choicerow" style="margin-bottom: 10px;">`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        choice += `<div class="choicerow">`;
        choice += buildFourthChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFifthChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSixthChoice();
        choice += `</div>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId), $("#" + fourthId), $("#" + fifthId), $("#" + sixthId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFourthListener() {
        $("#" + fourthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fourthChoice);
            loadNextPart(fourthNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFifthListener() {
        $("#" + fifthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fifthChoice);
            loadNextPart(fifthNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSixthListener() {
        $("#" + sixthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(sixthChoice);
            loadNextPart(sixthNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        part += fourthNext.build();
        part += fifthNext.build();
        part += sixthNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();
        fourthNext.registerListeners();
        fifthNext.registerListeners();
        sixthNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
        registerFourthListener();
        registerFifthListener();
        registerSixthListener();
    };
}

var CaravanConfinementCartChoice = {
    BREAK: 1,
    SLIP: 2,
    BANG: 3,
    COUNT: 4,
    GAZE: 5,
    TALK: 6
};

// END scenes\Enslavement\CaravanConfinement\cart\CaravanConfinementCart.js

// scenes\Enslavement\CaravanConfinement\tight\CaravanConfinementClaw.js

/**
 * Tightness section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementClaw(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementTightnessChoice.CLAW;
    var reqController = ctrl.tight;
    var subController = ctrl.postTight;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementCart(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var toReturn = `[p]
                            You aren't going to let this filthy [b]slaver[/b] touch you!
                            You'd rather [b]get your neck snapped[/b] by the wolf standing behind you than let some disgusting,
                            mean, [b]rude[/b] human lay his hands upon you!
                            As he reaches for your groin, you raise your arm threateningly, preparing to [b]maim[/b] his face with your sharp claws.
                            Mere moments before you manage to bring your natural weapons to bear, your wrist is grabbed from behind by a powerful, furred hand.
                        [/p]
                        [p]
                            The wolf, still holding your tail, restrains you, preventing you from striking your target.
                            You aren't stopped that easily, though.
                            You lift your [i]other[/i] hand, determined to [b]hurt[/b] the caravan master.
                            A second time, however, you are interrupted mid-strike.
                        [/p]
                        [p]
                            Rather than taking hold of your other wrist as well, the canine chooses to simply [b]squeeze[/b] the small scaly arm already in his grasp.
                            You wince in pain, the bandit's powerful grip making it [b]very[/b] clear that he could snap your bones like [b]twigs[/b].
                            "[color=gblue][b]Don't[/b].[/color]" he states, and nothing more.
                            Whimpering and whining, you meekly lower your free hand once more, and relent to the humiliating inspection.
                        [/p]
                        [p]
                            Tears form at the corners of your eyes, not from pain, but from [b]shame[/b].
                            Shame that you are going to [b]let this happen[/b],
                            instead of struggling so hard they're forced to beat you to a [b]bloody pulp[/b] before being able to check your loins.
                        [/p]
                        [p]
                            A knight would fight to the very end before suffering indignation at the hands of ${male ? `his` : `her`} captors.
                            But, despite your best hopes and dreams, you are no knight.
                            You are a [b]thief[/b], and now a [b]slave[/b].
                            A [b]coward[/b], scared of getting hurt.
                            A weakling that yields in the face of adversity, instead of rising beyond and above.
                        [/p]
                        [p]
                            If you were stronger, more agile, smarter perhaps, there might have been a way out of this situation.
                            But for a frail kobold like yourself, all you can do is resign to your fate,
                            and hope that the human kneeling before you will stay true to his word instead of taking things [i]one step further[/i].
                        [/p]`;

        if (fucked && male) {
            toReturn += `[p]
                            As your tears drip to the ground, the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            First, it makes its way towards your rear, rubbing over your still-sensitive butthole.
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body won't be a prominent selling point,
                            at least not for a couple of days more, when it'll regain its tightness.
                        [/p]
                        [p]
                            He sighs loudly, before moving his finger a bit more to the front, pressing it against your virginal cockslit.
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            Then, he slips the tip of the chubby digit inside of your warm slit,
                            testing the tightness of an orifice that was never meant to be fucked.
                        [/p]
                        [p]
                            "[color=brown]You could pass for a maiden if there wasn't a dick in there.[/color]" the caravan master speaks,
                            elliciting a chuckle from the wolf.
                            "[color=gblue]Well, now I know which hole I'll be fucking if you don't sell, kobold.[/color]"
                        [/p]`;
        } else if (fucked && !male) {
            toReturn += `[p]
                            As your tears drip to the ground, the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            First, it travels along the length of your slit, parting your folds ever so slightly.
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body won't be a prominent selling point,
                            at least not for a couple of days more, when it'll regain its tightness.
                        [/p]
                        [p]
                            He sighs loudly, before moving his finger a bit more to the rear, pressing it against your virginal tailhole.
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            Then, he slips the tip of the chubby digit inside of your butt,
                            testing the tightness of an orifice that was never meant to be fucked.
                        [/p]
                        [p]
                            "[color=brown]Tighter than a monk's lips.[/color]" the caravan master speaks,
                            elliciting a chuckle from the wolf.
                            "[color=gblue]Well, now I know which hole I'll be fucking if you don't sell, kobold.[/color]"
                        [/p]`;
        } else if (!fucked && male) {
            toReturn += `[p]
                            As your tears drip to the ground, the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            It makes its way towards your rear, rubbing over your virginal butthole.
                        [/p]
                        [p]
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body could be a prominent selling point,
                            but he isn't content with just [b]touching[/b].
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            Then, he presses the tip of his chubby finger inside of your butt,
                            testing the tightness of a virginal kobold.
                        [/p]
                        [p]
                            "[color=brown]Looks like he wasn't lying after all.[/color]" the caravan master speaks.
                            "[color=brown]A virgin kobold, a rarity for sure.
                            I have no doubt you'll sell well.[/color]"
                            The wolf chuckles, quickly adding:
                            "[color=gblue]And if you don't, you won't be a virgin for much longer.[/color]"
                            The canine gives a teasing thrust from behind, bumping his clothed groin against the back of your head.
                        [/p]`;
        } else {
            toReturn += `[p]
                            As your tears drip to the ground, the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            It travls along the length of your slit, parting your folds ever so slightly.
                        [/p]
                        [p]
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body could be a prominent selling point,
                            but he isn't content with just [b]touching[/b].
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            Then, he presses the tip of his chubby finger inside of your cunt,
                            testing the tightness of a virginal kobold.
                        [/p]
                        [p]
                            "[color=brown]Looks like she wasn't lying after all.[/color]" the caravan master speaks.
                            "[color=brown]A virgin kobold, a rarity for sure.
                            I have no doubt you'll sell well.[/color]"
                            The wolf chuckles, quickly adding:
                            "[color=gblue]And if you don't, you won't be a virgin for much longer.[/color]"
                            The canine gives a teasing thrust from behind, bumping his clothed groin against the back of your head.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += subController.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\tight\CaravanConfinementClaw.js

// scenes\Enslavement\CaravanConfinement\tight\CaravanConfinementStandStill.js

/**
 * Tightness section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementStandStill(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementTightnessChoice.STILL;
    var reqController = ctrl.tight;
    var subController = ctrl.postTight;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementCart(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var toReturn = `[p]
                            You don't want to let this filthy [b]slaver[/b] touch you, but you don't have much of a choice, either.
                            The wolf standing behind you could easily [b]snap your neck[/b] if you so much as [b]tried[/b] to make a move.
                            As the caravan master reaches for your groin, you consider raising your arm threateningly, but ultimately decide to just [i]let this happen[/i],
                            hoping it'll be over soon, that he won't decide to take things [i]one step further[/i].
                        [/p]`;

        if (fucked && male) {
            toReturn += `[p]
                            Noticing your lack of [i]resistance[/i], the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            First, it makes its way towards your rear, rubbing over your still-sensitive butthole.
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body won't be a prominent selling point,
                            at least not for a couple of days more, when it'll regain its tightness.
                        [/p]
                        [p]
                            He sighs loudly, before moving his finger a bit more to the front, pressing it against your virginal cockslit.
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            Then, he slips the tip of the chubby digit inside of your warm slit,
                            testing the tightness of an orifice that was never meant to be fucked.
                        [/p]
                        [p]
                            "[color=brown]You could pass for a maiden if there wasn't a dick in there.[/color]" the caravan master speaks,
                            elliciting a chuckle from the wolf.
                            "[color=gblue]Well, now I know which hole I'll be fucking if you don't sell, kobold.[/color]"
                        [/p]`;
        } else if (fucked && !male) {
            toReturn += `[p]
                            Noticing your lack of [i]resistance[/i], the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            First, it travels along the length of your slit, parting your folds ever so slightly.
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body won't be a prominent selling point,
                            at least not for a couple of days more, when it'll regain its tightness.
                        [/p]
                        [p]
                            He sighs loudly, before moving his finger a bit more to the rear, pressing it against your virginal tailhole.
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            Then, he slips the tip of the chubby digit inside of your butt,
                            testing the tightness of an orifice that was never meant to be fucked.
                        [/p]
                        [p]
                            "[color=brown]Tighter than a monk's lips.[/color]" the caravan master speaks,
                            elliciting a chuckle from the wolf.
                            "[color=gblue]Well, now I know which hole I'll be fucking if you don't sell, kobold.[/color]"
                        [/p]`;
        } else if (!fucked && male) {
            toReturn += `[p]
                            Noticing your lack of [i]resistance[/i], the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            It makes its way towards your rear, rubbing over your virginal butthole.
                        [/p]
                        [p]
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body could be a prominent selling point,
                            but he isn't content with just [b]touching[/b].
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            Then, he presses the tip of his chubby finger inside of your butt,
                            testing the tightness of a virginal kobold.
                        [/p]
                        [p]
                            "[color=brown]Looks like he wasn't lying after all.[/color]" the caravan master speaks.
                            "[color=brown]A virgin kobold, a rarity for sure.
                            I have no doubt you'll sell well.[/color]"
                            The wolf chuckles, quickly adding:
                            "[color=gblue]And if you don't, you won't be a virgin for much longer.[/color]"
                            The canine gives a teasing thrust from behind, bumping his clothed groin against the back of your head.
                        [/p]`;
        } else {
            toReturn += `[p]
                            Noticing your lack of [i]resistance[/i], the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            It travels along the length of your slit, parting your folds ever so slightly.
                        [/p]
                        [p]
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body could be a prominent selling point,
                            but he isn't content with just [b]touching[/b].
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            Then, he presses the tip of his chubby finger inside of your cunt,
                            testing the tightness of a virginal kobold.
                        [/p]
                        [p]
                            "[color=brown]Looks like she wasn't lying after all.[/color]" the caravan master speaks.
                            "[color=brown]A virgin kobold, a rarity for sure.
                            I have no doubt you'll sell well.[/color]"
                            The wolf chuckles, quickly adding:
                            "[color=gblue]And if you don't, you won't be a virgin for much longer.[/color]"
                            The canine gives a teasing thrust from behind, bumping his clothed groin against the back of your head.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += subController.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\tight\CaravanConfinementStandStill.js

// scenes\Enslavement\CaravanConfinement\tight\CaravanConfinementMoan.js

/**
 * Tightness section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementMoan(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementTightnessChoice.MOAN;
    var reqController = ctrl.tight;
    var subController = ctrl.postTight;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementCart(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var toReturn = `[p]
                            While you aren't keen on getting touched by this filthy [b]slaver[/b],
                            you feel like this might be a good time to [i]play along[/i],
                            to show that you're a [b]good ${male ? `boy` : `girl`}[/b],
                            and maybe earn some favor points from either of the men in the process.
                            After all, if you [i]act[/i] the part of a slave,
                            then they'll have no reason to hurt or mistreat you.
                            If you humor them, they might eventually choose to [i]let you go[/i], or so you hope.
                            And even if they don't, at the very least you'll be treated marginally better if you do as they say instead of fighting them every step of the way.
                        [/p]`;

        if (fucked && male) {
            toReturn += `[p]
                            Noticing your lack of [i]resistance[/i], the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            First, it makes its way towards your rear, rubbing over your still-sensitive butthole.
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body won't be a prominent selling point,
                            at least not for a couple of days more, when it'll regain its tightness.
                        [/p]
                        [p]
                            He sighs loudly, before moving his finger a bit more to the front, pressing it against your virginal cockslit.
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            To his confusion, his smirk is met with an anticipating grin.
                            As he slips the tip of the chubby digit inside of your warm slit,
                            testing the tightness of an orifice that was never meant to be fucked,
                            you let out a small, pleased sigh, like a [b]cheap whore[/b] trying to entice a client.
                        [/p]
                        [p]
                            "[color=brown]Are you sure you're not a [b]girl[/b]?[/color]" the caravan master asks,
                            elliciting a chuckle from the wolf.
                            "[color=brown]A finger in your cockslit can't feel [b]that[/b] good.[/color]"
                            The canine interrupts:
                            "[color=gblue]I don't know. A thumb up my sheath always gets me going.[/color]"
                            The human seems unamused.
                        [/p]`;
        } else if (fucked && !male) {
            toReturn += `[p]
                            Noticing your lack of [i]resistance[/i], the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            First, it travels along the length of your slit, parting your folds ever so slightly.
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body won't be a prominent selling point,
                            at least not for a couple of days more, when it'll regain its tightness.
                        [/p]
                        [p]
                            He sighs loudly, before moving his finger a bit more to the rear, pressing it against your virginal tailhole.
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            To his confusion, his smirk is met with an anticipating grin.
                            As he slips the tip of the chubby digit inside of your ass,
                            testing the tightness of an orifice that was never meant to be fucked,
                            you let out a small, pleased sigh, like a [b]cheap whore[/b] trying to entice a client.
                        [/p]
                        [p]
                            "[color=brown]A real [b]buttslut[/b], huh?[/color]" the caravan master comments,
                            elliciting a chuckle from the wolf.
                            "[color=gblue]Well, now I know which hole I'll be fucking if you don't sell, kobold.[/color]"
                        [/p]`;
        } else if (!fucked && male) {
            toReturn += `[p]
                            Noticing your lack of [i]resistance[/i], the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            It makes its way towards your rear, rubbing over your virginal butthole.
                        [/p]
                        [p]
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body could be a prominent selling point,
                            but he isn't content with just [b]touching[/b].
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            To his confusion, his smirk is met with an anticipating grin.
                            As he presses the tip of his chubby finger inside of your butt,
                            testing the tightness of a virginal kobold,
                            you let out a small, pleased sigh, like a [b]cheap whore[/b] trying to entice a client.
                        [/p]
                        [p]
                            "[color=brown]Looks like he wasn't lying after all.[/color]" the caravan master speaks.
                            "[color=brown]An eager virgin kobold slave.
                            I have no doubt you'll sell well.[/color]"
                            The wolf grunts dejectedly.
                            "[color=gblue]If you're so lusty, why didn't you let me do you in the forest?[/color]"
                            Clearing his throat, the human answers for you:
                            "[color=brown]Be glad that he didn't.
                            Now he's worth twice as much.[/color]"
                        [/p]`;
        } else {
            toReturn += `[p]
                            Noticing your lack of [i]resistance[/i], the caravan master begins his [i]inspection[/i].
                            A single, probing finger slides over the scales of your nether region.
                            It travels along the length of your slit, parting your folds ever so slightly.
                        [/p]
                        [p]
                            Even without slipping the digit inside,
                            the caravan master can clearly tell that this part of your body could be a prominent selling point,
                            but he isn't content with just [b]touching[/b].
                            The human stares directly into your eyes.
                            A small smirk forms on his face.
                            To his confusion, his smirk is met with an anticipating grin.
                            As he presses the tip of his chubby finger inside of your cunt,
                            testing the tightness of a virginal kobold,
                            you let out a small, pleased sigh, like a [b]cheap whore[/b] trying to entice a client.
                        [/p]
                        [p]
                            "[color=brown]Looks like she wasn't lying after all.[/color]" the caravan master speaks.
                            "[color=brown]An eager virgin kobold slave.
                            I have no doubt you'll sell well.[/color]"
                            The wolf grunts dejectedly.
                            "[color=gblue]If you're so lusty, why didn't you let me do you in the forest?[/color]"
                            Clearing his throat, the human answers for you:
                            "[color=brown]Be glad that she didn't.
                            Now she's worth twice as much.[/color]"
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += subController.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\tight\CaravanConfinementMoan.js

// scenes\Enslavement\CaravanConfinement\tight\CaravanConfinementTightness.js

/**
 * Tight section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementTightness(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.postMouth;
    var subController = ctrl.tight;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new CaravanConfinementClaw(ctrl);
    var secondNext = new CaravanConfinementStandStill(ctrl);
    var thirdNext = new CaravanConfinementMoan(ctrl);

    var firstChoice = CaravanConfinementTightnessChoice.CLAW;
    var secondChoice = CaravanConfinementTightnessChoice.STILL;
    var thirdChoice = CaravanConfinementTightnessChoice.MOAN;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var toReturn = `[p]
                            "[color=brown]Anyway, there's one more thing we need to take a look at.[/color]"
                            The caravan master gives the canine a small nod.
                            The wolf, who seemingly understands the human's intentions, steps behind you.
                            Grabbing your tail, the strong lupine [b]yanks[/b] it out from in between your legs, exposing your loins.
                            The expert in slavery, meanwhile, drops to his knees, coming face-to-face with you.
                            [i]He smells.[/i]
                        [/p]`;
        
        if(fucked) {
            toReturn += `[p]
                            The man forcibly parts your thighs with one of his burly hands.
                            "[color=brown]Damnit, Fenrir!
                            Do you [b]really[/b] have to [b]fuck[/b] everything you get your filthy hands on?![/color]"
                            You hear the canine behind you chuckle.
                            "[color=gblue]What can I say?
                            I'm a wolf.
                            That's what we do.
                            We [b]fuck[/b] things.[/color]"
                            The human seems unimpressed by the bandit's lackluster explanation.
                            "[color=gblue]Besides, that nun in the village we passed was practically [b]begging[/b] for it.[/color]"
                        [/p]
                        [p]
                            "[color=brown]Should have brought [b]her[/b] back to the caravan instead of this scaly [i]thief[/i].
                            Let's hope the [i]other[/i] hole is still tight enough to attract a couple of potential buyers.
                            A good mouth alone won't draw in the [i]perverts[/i], and those are the people who [b]really[/b] [i]jack up[/i] the price.[/color]"
                        [/p]`;
        } else {
            toReturn += `[p]
                            The man forcibly parts your thighs with one of his burly hands.
                            "[color=gblue]It said it was a virgin, back when I caught it.[/color]" the wolf shares.
                            The human inquisitively raises an eyebrow.
                            "[color=brown]A virgin kobold?
                            And you [b]bought[/b] that lie?
                            Come on, Fenrir, I thought you were smarter than that.[/color]"
                            The canine squeezes your tail, wordlessly threatening to harm you should you have [i]lied[/i] to him.
                            "[color=brown]Let's find out if we have an [b]honest thief[/b] or a [b]rotten apple[/b] on our hands.[/color]"
                        [/p]`;
        }
        
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Scratch him</button>`;

        var tooltip = global.parser.parse(`Claw at the human, hopefully dissuading him from probing your nether region.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 200px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Stand still</button>`;

        var tooltip = global.parser.parse(`Let the man do his thing. It'll be over a lot faster if you don't resist.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Moan</button>`;

        var tooltip = global.parser.parse(`Show him you know what a slave is to do when someone touches you.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var CaravanConfinementTightnessChoice = {
    CLAW: 1,
    STILL: 2,
    MOAN: 3
};

// END scenes\Enslavement\CaravanConfinement\tight\CaravanConfinementTightness.js

// scenes\Enslavement\CaravanConfinement\mouth\CaravanConfinementBite.js

/**
 * Mouth section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementBite(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementMouthChoice.BITE;
    var reqController = ctrl.mouth;
    var subController = ctrl.postMouth;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementTightness(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You quickly snap your jaws, hoping to catch the tips of the human's fingers, but sadly,
                            you miss your goal, the caravan master's strategic finger-placement causing your teeth to barely [b]scrape[/b] against the edge of his skin.
                            Despite your bite missing its mark, the man still ends up retracting his hand from your mouth, probably in fear of your sharp fangs.
                            Victory!
                        [/p]
                        [p]
                            Alas, the smug smile that forms on your face is all too soon wiped away.
                            Using the very same hand you nearly bit, the human harshly [b]slaps[/b] the side of your face, with all of the strength in his beefy arm.
                            You yelp out in pain.
                            While you're still reeling from the slap, he grabs your throat, squeezing it shut, choking you for just long enough to get his message across:
                            "[color=brown]You filthy fucking [b]lizard[/b].
                            If I wasn't going to sell you, I'd knock out your teeth and shove them up your slit for that.
                            Now, you listen good.
                            I'm only going to say this [b]once[/b].
                            [b]Don't test my patience.[/b]
                            I'll skin you alive and make you eat your own [b]guts[/b] if you try to pull something like that again.[/color]"
                        [/p]
                        [p]
                            His words freshly burned into your oxygen-starved mind, the caravan master lets go of your neck.
                            You gasp for air, only barely managing to stay on your feet.
                            While you're busy recovering from asphyxiation, the man looks over the rest of your body.
                            "[color=brown]Some small cuts and bruises.
                            Nothing major.
                            I wouldn't call it [i]pristine[/i] condition, but good enough for sellin'.
                            Too bad about the shitty attitude, though.[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += subController.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\mouth\CaravanConfinementBite.js

// scenes\Enslavement\CaravanConfinement\mouth\CaravanConfinementTeeth.js

/**
 * Mouth section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementTeeth(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementMouthChoice.TEETH;
    var reqController = ctrl.mouth;
    var subController = ctrl.postMouth;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementTightness(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var toReturn = `[p]
                            You don't really want to [i]anger[/i] the man, but you don't feel comfortable with him probing around your mouth either.
                            Your keep your jaw closed, without actively [i]resisting[/i] the human's inspection.
                            He sighs, but he doesn't even [i]attempt[/i] to pry open your muzzle further.
                            Instead, he lowers his head to better gaze into your maw,
                            inspecting what little he can see of your gums, as well as your teeth, internally grading you like a piece of [b]meat[/b].
                        [/p]
                        [p]
                            Finished with his shallow inspection, the man removes his fingers from your muzzle.
                            "[color=brown]I didn't get a good look, but there aren't any apparent diseases.
                            Teeth might need some [b]filing[/b], but that's not [i]our[/i] job.
                            Lizard, next time I want to look in your mouth, you better [b]open up[/b] properly,
                            or I'll [b]hurt you[/b] until you do.
                            Got it?[/color]"
                            You don't bother dignifying his threat with a nod.
                            He knows darn well you heard him.
                        [/p]
                        [p]
                            Your maw graded, the caravan master runs his eyes along the rest your naked frame, examining you from top to bottom.
                            "[color=brown]Some small cuts and bruises.
                            Nothing major.
                            I wouldn't call it [i]pristine[/i] condition, but good enough for sellin'.
                            Attitude could use some adjusting, but some buyers like the ones that still have a little [i]spark[/i] to them.[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += subController.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\mouth\CaravanConfinementTeeth.js

// scenes\Enslavement\CaravanConfinement\mouth\CaravanConfinementCooperate.js

/**
 * Mouth section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementCooperate(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementMouthChoice.COOPERATE;
    var reqController = ctrl.mouth;
    var subController = ctrl.postMouth;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementTightness(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var further = ctrl.name.get() === CaravanConfinementNameChoice.STATE;
        var toReturn = `[p]
                            Not wanting to anger your captors, you open your mouth wide.
                            The human seems pleased with your decision to cooperate, his stern visage somewhat relaxing${further ? ` further` : ``}.
                            "[color=brown]'Atta ${male ? "boy" : "girl"}.[/color]" he says, lowering his head to better gaze into your maw,
                            inspecting your gums, your teeth and your tongue, internally grading you like a piece of [b]meat[/b].
                        [/p]
                        [p]
                            Finished with his inspection, the man removes his fingers from your muzzle.
                            "[color=brown]No apparent diseases.
                            Mouth seems clean enough.
                            Teeth might need some [b]filing[/b], but that's not [i]our[/i] job.[/color]"
                        [/p]
                        [p]
                            Your maw graded, the caravan master runs his eyes along the rest your naked frame, examining you from top to bottom.
                            "[color=brown]Some small cuts and bruises.
                            Nothing major.
                            I wouldn't call it [i]pristine[/i] condition, but good enough for sellin'.
                            Looks like we might not have to roast you over the open fire after all, kobold.[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += subController.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\mouth\CaravanConfinementCooperate.js

// scenes\Enslavement\CaravanConfinement\mouth\CaravanConfinementMouth.js

/**
 * Mouth section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementMouth(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.postName;
    var subController = ctrl.mouth;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new CaravanConfinementBite(ctrl);
    var secondNext = new CaravanConfinementTeeth(ctrl);
    var thirdNext = new CaravanConfinementCooperate(ctrl);

    var firstChoice = CaravanConfinementMouthChoice.BITE;
    var secondChoice = CaravanConfinementMouthChoice.TEETH;
    var thirdChoice = CaravanConfinementMouthChoice.COOPERATE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            "[color=brown]In any case, people won't buy a slave based on its name alone.
                            There's more [b]important[/b] parts that need checking.[/color]"
                            You instinctively move your tail in between your legs to cover up your groin.
                            You don't like where this is headed,
                            but at least the fat human seems less [i]sexually inclined[/i] than the wolf.
                            He strikes you as more of a [i]businessman[/i] than the bandit who caught you.
                            With some luck, he might [i]actually[/i] be interested in simply [i]checking[/i] your body for injuries and whatnot,
                            instead of attempting to take advantage of the situation.
                        [/p]
                        [p]
                            Letting go of your collar, the human grabs you by the top of your scaly head.
                            Through brute force, he janks your head backwards, tilting your visage up to face his own.
                            Using his other hand, he forcibly spreads your lips with a pair of thick, unwashed fingers.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Bite</button>`;

        var tooltip = global.parser.parse(`Attempt to bite the human's hand.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Close mouth</button>`;

        var tooltip = global.parser.parse(`Keep your teeth tightly pressed together to seal off the rest of your muzzle.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Cooperate</button>`;

        var tooltip = global.parser.parse(`Open your mouth wide, like a good little [i]pet[/i].`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var CaravanConfinementMouthChoice = {
    BITE: 1,
    TEETH: 2,
    COOPERATE: 3
};

// END scenes\Enslavement\CaravanConfinement\mouth\CaravanConfinementMouth.js

// scenes\Enslavement\CaravanConfinement\name\CaravanConfinementGlare.js

/**
 * Name section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementGlare(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementNameChoice.GLARE;
    var reqController = ctrl.name;
    var subController = ctrl.postName;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementMouth(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var begged = !global.kobold.history.shared.enslavement.captureFucked;
        var enjoyed = global.kobold.history.shared.enslavement.captureEnjoyed;
        var toReturn = `[p]
                            You may be his [i]slave[/i], but that doesn't mean you have to do what he says.
                            Well, okay, [i]maybe[/i] it does, but you're still not doing it!
                            What does he need your name for anyway?
                            To [b]engrave[/b] on your [b]collar[/b]?!
                            Screw that!
                            You're not telling him [b]anything[/b].
                            He can go to [b]hell[/b] for all you care!
                        [/p]
                        [p]
                            Your resolve bolstered, you glare up at the chubby human,
                            returning his hateful gaze with equal amounts of disdain.
                            Your muzzle remains tightly sealed, not a single sound escaping your lips.
                            The man is visibly growing annoyed.
                            He voices his anger in the form of a low-pitched growl.
                            You could almost mistake him for the wolf like this, if it weren't for his far less athletic physique.
                            "[color=brown]Don't you know how to talk, you miserable [b]mutt[/b]?![/color]"
                        [/p]
                        [p]
                            "[color=gblue]${male ? `He` : `She`} speaks our language alright.
                            You should have heard'em ${enjoyed ? `begging for more` : `${begged ? `begging for mercy` : `screaming for help`}`} earlier.[/color]"
                            the canine butts in.
                            "[color=brown]So instead of being dumb, the lizard's just being uppity.
                            Good.
                            Disobedience can be fixed.
                            Stupidity, however, can not.
                            If I had my way, I'd [b]beat[/b] some sense into it, but bruised merchandise doesn't sell.[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += subController.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\name\CaravanConfinementGlare.js

// scenes\Enslavement\CaravanConfinement\name\CaravanConfinementGrowl.js

/**
 * Name section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementGrowl(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementNameChoice.GROWL;
    var reqController = ctrl.name;
    var subController = ctrl.postName;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementMouth(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Who does this [b]filthy human[/b] think he is?!
                            Maybe if he lost some weight and showed you some [b]common courtesy[/b],
                            you'd [b]respect[/b] him enough to honor his request.
                            The sheer tone with which he asked for your name alone is enough to make you [b]not[/b] want to comply.
                        [/p]
                        [p]
                            You softly growl at the chubby human,
                            the rumbling noise emanating from the back of your throat growing louder and louder as you reflect on the way he's treated you thus far.
                            In little short of five minutes, he's managed to be [b]far[/b] more obnoxious than the canine who captured you ever was!
                            At least the wolf doesn't [b]insult[/b] and [b]berate[/b] you every chance he gets.
                        [/p]
                        [p]
                            The human notices your display of defiance.
                            He menacingly raises his hand, but manages to restrain himself for the time being,
                            instead venting his frustration at the wolf.
                            "[color=brown]It's growling.
                            Did you bring me a [b]feral[/b] kobold to sell?!
                            Might as well take it to the [b]butcher[/b] straight away![/color]"
                        [/p]
                        [p]
                            "[color=gblue]It's not one of those tribal kobolds.
                            It said something earlier, when I caught it.
                            At least I think it did?"[/color]"
                            The wolf responds, clearly upset, like [b]he[/b]'s the one being blamed for [b]your[/b] actions.
                            He continues his defensive rambling:
                            "[color=gblue]It doesn't [b]look[/b] feral anyway.
                            And why would a tribal lizard steal a bag of coins?
                            They don't even [b]live[/b] this close to Varanar![/color]"
                        [/p]
                        [p]
                            The human lowers his threatening hand, warily exchanging glances with both you and the canine.
                            "[color=brown]I hope for both your sakes the lizard knows how to speak.
                            Disobedience can be fixed, but there isn't a single buyer I can think of that would be interested in a [b]savage[/b] kobold.
                            Unless you want to keep it as a fucktoy, we might be better off just roasting it over the fire.[/color]"
                        [/p]
                        [p]
                            "[color=gblue]I said it knows how to speak.
                            Maybe it doesn't like humans.
                            Plenty of other buyers around, right?[/color]"
                            The wolf chuckles.
                            The human utters an angry harrumph.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += subController.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\name\CaravanConfinementGrowl.js

// scenes\Enslavement\CaravanConfinement\name\CaravanConfinementState.js

/**
 * Name section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementState(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = CaravanConfinementNameChoice.STATE;
    var reqController = ctrl.name;
    var subController = ctrl.postName;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementMouth(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.nightOne.nameStated = true;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            Terrified of the well-set man still holding your metal collar,
                            you quickly and meekly stammer out your name.
                            The human's angry grimace visibly mellows out, if just a little.
                            "[color=brown]${name}?
                            That's a lousy name for a slave, let alone a kobold.
                            Who gave it to you?
                            A court jester, a three year old, or a senile dragon?[/color]"
                        [/p]
                        [p]
                            The caravan guard chuckles at his own joke,
                            but his laughter comes to a halt when he realizes neither you nor the wolf found his beratement enjoyable.
                            After a sigh and some mumbling to himself, he continues.
                            "[color=brown]I'd give you a more suitable moniker myself,
                            but that pleasure belongs to whoever ends up buying your sorry ass.
                            At least you know how to speak.
                            That'll up your value some.[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += subController.change();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\name\CaravanConfinementState.js

// scenes\Enslavement\CaravanConfinement\name\CaravanConfinementName.js

/**
 * Name section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementName(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.arrival;
    var subController = ctrl.name;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new CaravanConfinementGrowl(ctrl);
    var secondNext = new CaravanConfinementGlare(ctrl);
    var thirdNext = new CaravanConfinementState(ctrl);

    var firstChoice = CaravanConfinementNameChoice.GROWL;
    var secondChoice = CaravanConfinementNameChoice.GLARE;
    var thirdChoice = CaravanConfinementNameChoice.STATE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            With the two men having reached an accord, the situation quickly diffuses.
                            Done butting heads, their attention drifts to you once more.
                            "[color=gblue]So, what price do you think the lizard'll fetch?[/color]" asks the wolf, whilst licking his lips,
                            though you know not whether it's a [i]lustful[/i] lick, or if he's just eager to hear how much of a [i]bonus[/i] he'll be getting.
                        [/p]
                        [p]
                            "[color=brown]It depends.[/color]"
                            The human speaks.
                            "[color=brown][i]Good[/i] slaves can fetch up to a hundred coins.
                            Far from nothing, right?
                            Bad ones, though.
                            You know, the ones that are worn out, used and abused,
                            the ones you can't even get a rise out of anymore despite [b]still[/b] being disobedient,
                            [b]nobody[/b] wants those.
                            Let's see how much this one is worth, shall we?[/color]"
                        [/p]
                        [p]
                            You [b]really[/b] don't like the thought of being [i]appraised[/i], but you strongly doubt you have a say in the matter.
                            The caravan master turns to face you.
                            Casting down a stern gaze, the man addresses you:
                            "[color=brown]Let's start off simple.
                            Kobold.
                            Your name.[/color]"
                        [/p]
                        [p]
                            Confused, and more than a little intimidated, you hesitate, your brain freezing up completely.
                            You open your muzzle, yet no words escape your mouth.
                            The human, lacking patience for someone who tried to [b]steal[/b] from him, repeats himself:
                            "[color=brown]Your name, [b]lizard[/b]. State. Your. Name.[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Growl</button>`;

        var tooltip = global.parser.parse(`Growl at the man. Who does he think he is, ordering you around?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 180px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Glare</button>`;

        var tooltip = global.parser.parse(`Angrily and silently stare at him in defiance.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Reply</button>`;

        var tooltip = global.parser.parse(`Comply with the human, and state your name.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 170px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var CaravanConfinementNameChoice = {
    GROWL: 1,
    GLARE: 2,
    STATE: 3
};

// END scenes\Enslavement\CaravanConfinement\name\CaravanConfinementName.js

// scenes\Enslavement\CaravanConfinement\arrival\CaravanConfinementArrival.js

/**
 * Arrival section of Enslavement, Night One.
 * @param {CaravanConfinementController} controller The Enslavement, Night One controller.
 */
function CaravanConfinementArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.arrival;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new CaravanConfinementName(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var male = ctrl.misc.baseGender() === GenderEnum.MALE;
        var toReturn = `[p]
                            After a long, humiliating walk, the two of you finally arrive back at the money transport.
                            The sun luckily went down about halfway through the trek, darkness somewhat concealing your
                            nudity${fucked ? ` [sneak]and the cum leaking from your still slightly gaping ${male ? `tailhole` : `cunt`}[/sneak]` : ``}.
                            A small campfire along the side of the road illuminates the area.
                            Combined with a couple of torches affixed to the carts themselves, the convoy is lit well enough to scare off any would-be midnight thieves.
                        [/p]
                        [p]
                            While some caravan guards are busy patrolling and standing watch, most of them are seated around the warming flames.
                            The mood is audibly jolly.
                            You aren't even close to them,
                            but you can already [b]reek[/b] the hefty smell of the copious amounts of poorly brewed alcohol they're gulping down.
                            Amongst the group is the human that was chasing you earlier, sitting next to his dog.
                            He is the first to notice the wolf's triumphant return.
                            After a joyous greeting, all eyes turn to you.
                        [/p]
                        [p]
                            While an entire group of unruly, armed and drunken men of various species staring down your naked form is intimidating, to say the least,
                            the worst of your imminent worries are quelled when a more [i]rational looking[/i] fellow emerges from the front of the convoy.
                            It's the caravan master!
                            The [i]one[/i] human around who [i]doesn't[/i] look like he was ripped straight out of a portrait depicting a band of bandits.
                            And, going by his fancy imperial tunic, he's in charge of this whole operation.
                            Surely, one of the king's sergeants won't stand for a [i]drunken gangrape[/i], right?
                        [/p]
                        [p]
                            "[color=brown]So [b]this[/b] is our little [b]thief[/b]?[/color]" the bulky man speaks,
                            adjusting his belt as though to assert his dominance over the wolf.
                            "[color=gblue]Eyup.[/color]" comes the canine's response.
                            "[color=gblue]Caught'em running through the woods. Would have been halfway to Salamar by now, if not for me and Kajar.[/color]"
                            You assume Kajar is the human that was chasing you earlier.
                        [/p]
                        [p]
                            The caravan master casts a glare your way.
                            The hateful stare sends shivers down your spine.
                            "[color=brown]And what about the money?
                            Did you get it back, or are we having kobold meat for dinner?[/color]"
                            He practically [b]spits[/b] out the world [i]kobold[/i], his voice heavy with disdain.
                        [/p]
                        [p]
                            The canine pats his pocket, rattling the hoard of coins contained within.
                            A smug smile forms on his face.
                            He's holding all of the cards in the palm of his hand, and he knows it.
                            "[color=gblue]Two conditions.[/color]"
                            The human is clearly irked, his head jerking to face the lupine once more.
                            "[color=brown][b]Conditions?![/b]
                            You work for me!
                            Now hand over the bloody coin, or I'll have your entire gang of misfits [b]beheaded[/b]![/color]"
                        [/p]
                        [p]
                            The wolf seems unfazed by the caravan master's threat.
                            Once again, he repeats himself, a little louder.
                            "[color=gblue][b]Two. Conditions.[/b][/color]"
                            Not interrupted by his superior this time around, the canine states his terms.
                            "[color=gblue]Double pay for me and Kajar.
                            On top of that, I get the lizard.[/color]"
                            Grabbing you by the side of your head, the furred bandit tugs your face towards his clothed groin,
                            holding you close to him, flaunting his [i]ownership[/i].
                        [/p]
                        [p]
                            The human takes a moment to consider his options.
                            "[color=brown]Fine, double your [b]measly[/b] pay.[/color]"
                            Reaching for your metal collar, the caravan master firmly [b]yanks[/b] you towards him,
                            tearing you from the lupine's grasp.
                            "[color=brown]But the kobold is [b]mine[/b] by law.
                            It stole from [b]me[/b], not [b]you[/b].[/color]"
                        [/p]
                        [p]
                            The lupine chuckles.
                            "[color=gblue]Law?
                            We may be close, but we're not in Varanar yet.
                            The only [b]law[/b] in these woods is that the guy with the biggest gang gets what he wants.
                            And right now, that guy just happens to be [b]me[/b].[/color]"
                            The wolf bares his teeth.
                            The human grits his.
                            They glare at one another.
                            A small standoff ensues.
                            Tension is tangibly rising.
                        [/p]
                        [p]
                            "[color=brown]Half.[/color]"
                            the caravan master speaks, breaking the unnerving silence.
                            "[color=brown]I get two tries to sell it.
                            Once at the hamlet, once in the city proper.
                            You get half of whatever people are willing to pay for this [b]miserable[/b] excuse of a [b]thief[/b].[/color]"
                            The wolf narrows his eyes.
                            "[color=gblue]And if you can't sell'em?[/color]"
                        [/p]
                        [p]
                            "[color=brown]A kobold that doesn't sell is [b]useless[/b] to me.
                            If nobody wants it, it'll be yours to keep.[/color]"
                            The wolf's turn to ponder his choices.
                            After exchanging angry stares for a few more harrowing seconds, he replies:
                            "[color=gblue]Deal,
                            but if you [b]dare[/b] doublecross us, neither the baron, nor the king himself will be able to save your sorry ass.[/color]"
                            The human snickers.
                            "[color=brown]Unlike you [b]thugs[/b], I am a man of my word.[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("forest_night");
        ctrl.bgm.loop();
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\CaravanConfinement\arrival\CaravanConfinementArrival.js

// scenes\Enslavement\CaravanConfinement\CaravanConfinementController.js

/**
 * Controller for the Caravan Confinement scene.
 * @param {string} sceneId The id of the scene.
 */
function CaravanConfinementController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.nextSceneReady = true;
    this.nextSceneName = "SlaveNoonTwo";
    this.nextFolderName = "Enslavement/DayTwo";

    this.misc = new MiscController(id, "misc");

    this.arrival = new ChoiceController(id, "arrival");

    this.name = new ChoiceController(id, "name");

    this.postName = new ChangeController(id, "postName", function () {
        var choice = that.name.get();
        if (choice === CaravanConfinementNameChoice.GROWL) {
            global.kobold.history.shared.enslavement.nightOne.complianceRating += 1;
            global.kobold.status.mental.decrease(3, "Threatened and insulted");
        } else if (choice === CaravanConfinementNameChoice.GLARE) {
            global.kobold.history.shared.enslavement.nightOne.complianceRating += 2;
            global.kobold.status.mental.decrease(2, "Getting yelled at");
        } else {
            global.kobold.history.shared.enslavement.nightOne.complianceRating += 3;
            global.kobold.status.mental.decrease(1, "Crude joke");
        }
    }, function () {
        var choice = that.name.get();
        if (choice === CaravanConfinementNameChoice.GROWL) {
            return global.ui.change.buildNotification([
                new Change("Called a feral", "Sanity", -1),
                new Change("Threatened slap", "Sanity", -2)
            ]);
        } else if (choice === CaravanConfinementNameChoice.GLARE) {
            return global.ui.change.buildNotification([
                new Change("Yelled at", "Sanity", -2)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Crude joke", "Sanity", -1)
            ]);
        }
    });

    this.mouth = new ChoiceController(id, "mouth");

    this.postMouth = new ChangeController(id, "postMouth", function () {
        var choice = that.mouth.get();
        if (choice === CaravanConfinementMouthChoice.BITE) {
            global.kobold.history.shared.enslavement.nightOne.complianceRating += 1;
            global.kobold.status.physical.decrease(8, "Minor abuse");
            global.kobold.status.mental.decrease(6, "Threatened and examined");
        } else if (choice === CaravanConfinementMouthChoice.TEETH) {
            global.kobold.history.shared.enslavement.nightOne.complianceRating += 2;
            global.kobold.status.mental.decrease(6, "Threatened and graded");
        } else {
            global.kobold.history.shared.enslavement.nightOne.complianceRating += 3;
            global.kobold.status.mental.decrease(3, "Examined and graded");
        }
    }, function () {
        var choice = that.name.get();
        if (choice === CaravanConfinementMouthChoice.BITE) {
            return global.ui.change.buildNotification([
                new Change("Slapped", "Health", -3),
                new Change("Choked", "Health", -5),
                new Change("Threatened", "Sanity", -5),
                new Change("Examined", "Sanity", -1)
            ]);
        } else if (choice === CaravanConfinementMouthChoice.TEETH) {
            return global.ui.change.buildNotification([
                new Change("Threatened", "Sanity", -3),
                new Change("Examined", "Sanity", -1),
                new Change("Graded", "Sanity", -2)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Examined", "Sanity", -1),
                new Change("Graded", "Sanity", -2)
            ]);
        }
    });

    this.tight = new ChoiceController(id, "tight");

    this.postTight = new ChangeController(id, "postTight", function () {
        var choice = that.tight.get();
        if (choice === CaravanConfinementTightnessChoice.CLAW) {
            global.kobold.history.shared.enslavement.nightOne.complianceRating += 1;
            global.kobold.status.mental.decrease(8, "Unwillingly touched");
            global.kobold.status.exertion.increase(5, "Loins probed");
        } else if (choice === CaravanConfinementTightnessChoice.STILL) {
            global.kobold.history.shared.enslavement.nightOne.complianceRating += 2;
            global.kobold.status.mental.decrease(8, "Reluctantly touched");
            global.kobold.status.exertion.increase(5, "Loins probed");
        } else {
            global.kobold.history.shared.enslavement.nightOne.complianceRating += 3;
            global.kobold.status.exertion.increase(5, "Loins probed");
        }
    }, function () {
        var choice = that.tight.get();
        if (choice === CaravanConfinementTightnessChoice.CLAW) {
            return global.ui.change.buildNotification([
                new Change("Unwillingly touched", "Sanity", -5),
                new Change("Meekness exposed", "Sanity", -3),
                new Change("Loins probed", "Sexual exhaustion", +5, true)
            ]);
        } else if (choice === CaravanConfinementTightnessChoice.STILL) {
            return global.ui.change.buildNotification([
                new Change("Unwillingly touched", "Sanity", -5),
                new Change("Loins probed", "Sexual exhaustion", +5, true)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Loins probed", "Sexual exhaustion", +5, true)
            ]);
        }
    });

    this.cart = new ChoiceController(id, "cart");

    this.bang = new ChangeController(id, "bang", function () {
        global.kobold.status.physical.decrease(23, "Bashed head against floor");
        global.kobold.status.mental.decrease(12, "Immense pain");
        global.kobold.stats.increaseStat(StatsEnum.ENDURANCE, 1);
        global.kobold.stats.decreaseStat(StatsEnum.INTELLIGENCE, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Hurt self", "Health", -18),
            new Change("Passed out", "Health", -5),
            new Change("Pain without gain", "Sanity", -12),
            new Change("Endured pain", "Endurance", +1),
            new Change("Did something stupid", "Intelligence", -1)
        ]);
    });

    this.count = new ChangeController(id, "count", function () {
        global.kobold.status.mental.increase(7, "Fell asleep");
        global.kobold.stats.increaseStat(StatsEnum.INTELLIGENCE, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Keeping busy", "Sanity", +2),
            new Change("Dozed off", "Sanity", +5),
            new Change("Practiced counting", "Intelligence", +1)
        ]);
    });

    this.gaze = new ChangeController(id, "gaze", function () {
        global.kobold.status.mental.increase(8, "Fell asleep");
        global.kobold.stats.increaseStat(StatsEnum.LUCK, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Tranquility", "Sanity", +3),
            new Change("Dozed off", "Sanity", +5),
            new Change("Star-blessed", "Luck", +1)
        ]);
    });

    this.talk = new ChangeController(id, "talk", function () {
        global.kobold.status.mental.increase(10, "Onesided conversation");
        global.kobold.stats.increaseStat(StatsEnum.CHARISMA, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Onesided conversation", "Sanity", +10),
            new Change("Practiced talking to people", "Charisma", +1)
        ]);
    });

    this.breakFail = new ChangeController(id, "breakFail", function () {
        global.kobold.status.stamina.decrease(5, "Struggling");
        global.kobold.stats.increaseStat(StatsEnum.STRENGTH, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Struggling", "Stamina", -5),
            new Change("Workout", "Strength", +1)
        ]);
    });

    this.breakSucceed = new ChangeController(id, "breakSucceed", function () {
        global.kobold.history.shared.enslavement.nightOne.complianceRating -= 2;
        global.kobold.status.physical.decrease(5, "Rope burn");
        global.kobold.status.stamina.decrease(5, "Struggling");
        global.kobold.status.mental.increase(10, "Newfound freedom");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Rope burn", "Health", -5),
            new Change("Struggling", "Stamina", -5),
            new Change("Hope", "Sanity", +5),
            new Change("Freedom", "Sanity", +5)
        ]);
    });

    this.breakSucceedContinue = new ChangeController(id, "breakSucceedContinue", function () {
        global.kobold.status.physical.decrease(20, "Head bashed");
        global.kobold.status.mental.decrease(15, "Hope shattered");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Head bashed", "Health", -15),
            new Change("Knocked out", "Health", -5),
            new Change("Hope shattered", "Sanity", -15)
        ]);
    });

    this.slipFail = new ChangeController(id, "breakFail", function () {
        global.kobold.status.stamina.decrease(5, "Struggling");
        global.kobold.stats.increaseStat(StatsEnum.AGILITY, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Struggling", "Stamina", -5),
            new Change("Workout", "Agility", +1)
        ]);
    });

    this.slipSucceed = new ChangeController(id, "breakSucceed", function () {
        global.kobold.history.shared.enslavement.nightOne.complianceRating -= 2;
        global.kobold.status.physical.decrease(5, "Rope burn");
        global.kobold.status.stamina.decrease(5, "Struggling");
        global.kobold.status.mental.increase(10, "Newfound freedom");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Rope burn", "Health", -5),
            new Change("Struggling", "Stamina", -5),
            new Change("Hope", "Sanity", +5),
            new Change("Freedom", "Sanity", +5)
        ]);
    });

    this.slipSucceedContinue = new ChangeController(id, "slipSucceedContinue", function () {
        global.kobold.status.physical.decrease(20, "Head bashed");
        global.kobold.status.mental.decrease(15, "Hope shattered");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Head bashed", "Health", -15),
            new Change("Knocked out", "Health", -5),
            new Change("Hope shattered", "Sanity", -15)
        ]);
    });

    this.preContinue = new ContinueController(id, "preContinue");

    this.bgm = new AudioController(id, "res/audio/enslavement/camp.mp3", 100);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

}

// END scenes\Enslavement\CaravanConfinement\CaravanConfinementController.js

// scenes\Enslavement\CaravanConfinement\CaravanConfinement.js

/**
 * First night in enslavement.
 */
function CaravanConfinement() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "CaravanConfinement";
    var sceneName = "Enslavement, Night One";
    var sceneSubfolder = "Enslavement";
    var ctrl = new CaravanConfinementController(sceneId);
    var arrival = new CaravanConfinementArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]Enslavement[/t]`);
        tab += global.parser.parse(`[st]Night one[/st]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new CaravanConfinement();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\Enslavement\CaravanConfinement\CaravanConfinement.js



// END SCENE CaravanConfinement



// SCENE SlaveNoonTwo

// scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoFarmer.js

/**
 * Result section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoFarmer(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoResultsChoice.FARMER;
    var reqController = ctrl.result;
    var changeController = ctrl.farmer;

    var nextSceneReady = ctrl.farmerSceneReady;
    var nextSceneName = ctrl.farmerSceneName;
    var nextFolderName = ctrl.farmerFolderName;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Farmer buys kobold. His kid looks very excited.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(nextSceneName, nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoFarmer.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoVixen.js

/**
 * Result section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoVixen(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoResultsChoice.VIXEN;
    var reqController = ctrl.result;

    var continueId = arguments.callee.name + "ButtonOne";

    var idsToRemove = [
        "SlaveNoonTwoVixen",
        "SlaveNoonTwoResults"
    ];

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var name = ctrl.misc.name();
        var dick = ctrl.resultBodyMisc.dick();
        var injected = ctrl.gag.get();
        var drank = ctrl.outside.get() === SlaveNoonTwoOutsideChoice.DRINK;
        var toReturn = `[p]
                            "[color=orange]One hundred.[/color]" a gentle, feminine voice rings out.
                            It's the half-naked vixen.
                            She's soon challenged by the goblin, whose stingy pockets seem far deeper than what you'd have imagined.
                            "[color=green]One hundred and twenty![/color]"
                            The kind-looking canine casts a wicked leer towards the small humanoid.
                            "[color=orange]One hundred and fifty. Or would you rather deal with a back-handed [b]greenskin[/b]?[/color]"
                            "[color=green][b]One fifty-one[/b]![/color]"
                        [/p]
                        [p]
                            Bartholomew seems conflicted.
                            For a moment, it feels as though a [b]fight[/b] is brewing between the two interested parties.
                            But then, right as temperatures are about to reach a boiling point, Fenrir steps in to settle the debate,
                            your captor taking over from the caravan master.
                            "[color=gblue]Us canines have to stick together.
                            [b]Sold[/b], to the lady in black![/color]"
                            The goblin grumbles and complains, but the wolf's decision is final.
                            The crowd parts and money exchanges hands.
                            Seki must have wandered off somewhere during the bidding.
                            He's nowhere to be seen.
                        [/p]
                        [p]
                            While Barto chats with the woman, Fenrir approaches you.
                            "[color=gblue]Well, ${name}, you're hers now.
                            Next time we meet, you can thank me for not letting that [b]filthy[/b] goblin have you.
                            I'm no psychic, but I've a feeling at least the bitch'll treat you a bit better.
                            Either way, good luck.
                            You're going to need it.[/color]"
                        [/p]
                        [p]
                            Having said his farewells, the big wolf pushes you forwards, nudging you off of the small stage, towards your new [i]mistress[/i].
                            The deal concluded, the human and the lupine walk back to the carriages, bickering on how to split their profits.
                            The vixen approaches you, leaning down to talk eye-to-eye.
                            "[color=orange]Hey there, little ${dick ? `buddy` : `girl`}.[/color]"
                            She flashes you a toothy grin.
                            "[color=orange]You're all [b]mine[/b] now.
                            But don't worry, I won't hurt you, as long as you do everything I say.
                            You're going to have lots of [b]fun[/b] staying with me, I'm sure of it.[/color]"
                        [/p]`;

        if (injected) {
            toReturn += `[p]
                            The vixen notices the gag in your mouth.
                            "[color=orange]Let's get that [i]thing[/i] out of your mouth, shall we?[/color]"
                            You give her an eager look.
                            She moves one of her hands towards the rope tied around your muzzle
                            - which Fenrir told you [b]not[/b] to touch -
                            and with a single [i]flick[/i] of her claws, she [b]surgically[/b] severs the cord.
                            You reach for the underwear in your mouth, but the female fox beats you to it,
                            gracefully tugging the unsanitary, ${drank ? `feral-and-` : ``}kobold-spit drenched underpants from your muzzle.
                        [/p]
                        [p]
                            Finally able to fully close your jaws once more, you spit on the dry ground to get the lingering taste of [b]wolf ass[/b] out of your mouth.
                            The vulpine, meanwhile, deeply [b]sniffs[/b] the saliva-saturated garment,
                            before giving it a tiny lick, the disgusting flavor seemingly [i]pleasing[/i] her.
                            "[color=orange]This is [b]his[/b], isn't it?[/color]"
                            You meekly nod in response, knowing perfectly well who she's talking about.
                            "[color=orange]I'd have paid double if I knew you came with [i]extras[/i].[/color]"
                        [/p]`;
        }

        toReturn += `   [p]
                            ${injected ? `Your gag removed, the` : `The`} busty woman grabs your hand.
                            "[color=orange]I don't live far from here.
                            Let's walk there together, shall we?[/color]"
                            She's so [b]nice[/b].
                            Her palm is [b]soft[/b] and [b]warm[/b].
                            Living with [b]her[/b] is becoming a more inviting idea with every passing second.
                        [/p]
                        [p]
                            The two of you casually stroll alongside one of the hamlet's two main roads.
                            A few dozen yards further, you continue your journey on a small dirt path,
                            the opposite way of Seki's workshop, towards a large [b]mansion[/b] with an attached barn.
                            Wow, she must be [b]rich[/b] or something!
                            For a moment, you get your hopes up,
                            thinking you'll be allowed to spend your days in the luxury of an actual [b]house[/b],
                            but alas, the canine guides you to the stables instead.
                        [/p]
                        [p]
                            No big deal, rather a comfortable shelter with lots of hay than being forced to sleep outside, right?
                            The moment the big barn doors close behind you, the vixen immediately takes off her shirt, exposing her nude body.
                            So she [b]wasn't[/b] wearing anything underneath.
                            "[color=orange][b]Much[/b] better.
                            I [b]hate[/b] clothing.
                            Far too [i]constricting[/i].[/color]"
                            She's a [b]nudist[/b]?!
                            You get the feeling you're going to [b]love[/b] it here, despite the [i]weird[/i], lingering smell that you can't quite place.
                        [/p]
                        [p]
                            "[color=orange]Welcome to your new home, my little [b]pet[/b].[/color]"
                            If you get to stare at her naked body all day long, you'll [b]gladly[/b] be her [i]servant[/i],
                            or whatever she meant by that.
                            "[color=orange]Let's get you oriented, shall we?[/color]"
                        [/p]
                        [p]
                            The two of you walk to the back of the building, where an inconspicuous solemn table stands.
                            "[color=orange]Stay here for a minute, okay?[/color]"
                            You nod to your new [i]mistress[/i], taking the time to inspect the barn.
                            It's divided into two rows of cubicles, pens for whatever creatures the canine is keeping,
                            with the path the two of you just walked in between the corrals.
                            Outside of some strewn about hay, there isn't really much to look at.
                        [/p]
                        [p]
                            Your gaze quickly returns to the vixen.
                            She bends over deeply, rummaging through a sack before fetching out something you can't quite make out.
                            Then, she steps into one of the pens.
                            After half a minute or so, she emerges once more, with a leash in her hand.
                            A leash attached to a collar around the neck of a crying, pink, female kobold, clearly [b]unhappy[/b] to be here.
                        [/p]
                        [p]
                            Dumbstruck, you stare at Cassidy, as she guides the kobold towards the table you're standing next to.
                            Your mistress harshly [b]pushes[/b] what you can only presume to be her [i]other[/i] pet back-first onto the table,
                            a set of hidden mechanisms instantly activating, metal bindings snapping the lizard's arms and legs into place.
                            She struggles against them, but fails to free herself.
                        [/p]
                        [p]
                            Right as you are about to ask what the [b]hell[/b] is going on here,
                            the canine roughly [b]grabs[/b] you by the back of your neck,
                            crudely bringing you to your knees in between the female kobold's forcibly spread legs.
                            You barely manage to utter a protest before she shoves your face into the lizard's feminine loins.
                            "[color=orange]Be a good ${dick ? `boy` : `girl`} and [b]lick[/b] for me, okay?
                            Get her nice and [b]ready[/b].
                            It'll all make sense soon.[/color]"
                        [/p]
                        [p]
                            Lick?!
                            [b]Hell no![/b]
                            The pink kobold was clearly [b]not[/b] okay with what is happening here!
                            And neither are you!
                            You struggle against Cassidy's grasp, but she firmly keeps you in place, using just one of her hands.
                            With the other, she manipulates the plunger of the syringe she'd fetched earlier,
                            drawing liquid from an ominously looking bottle.
                        [/p]
                        [p]
                            As your squirming intensifies, the vixen suddenly [b]stabs[/b] the needle of the syringe into your arm, injecting your with its contents.
                            You yelp out in pain.
                            ${dick ? `You start to feel [b]hot[/b] and [b]flushed[/b].
                            You begin to pant, not from exertion, but from [b]arousal[/b].
                            Your member rapidly hardens, emerging from your cockslit.` : `Instantly,
                            your breasts recede into your chest completely, your bosom flattened.
                            Your ovaries [b]painfully[/b] wither, instead replaced by fully-formed internal [b]testicles[/b].
                            Your cervix grows shut, your clit vanishes completely, and a small, tapered, koboldic cock springs forth inside of you.
                            [b]You are now a boy[/b].
                            Worse yet, you feel [b]hot[/b] and [b]flushed[/b].
                            You begin to pant, not from exertion, but from [b]arousal[/b].
                            Your small, brand new member rapidly hardens, emerging from your cunt-turned-cockslit.`}
                            [i]Oh no.[/i]
                        [/p]
                        [p]
                            An intense [b]need[/b] to [b]breed[/b] builds up inside of you.
                            You greedily begin to [b]huff[/b] the female pheromones wafting from the slit in front of your face.
                            It smells [b]so[/b] good.
                            You need [b]more[/b].
                            "[color=orange]There you go.
                            Now you're getting into it.
                            Come on, now.
                            [b]Lick[/b].[/color]"
                        [/p]
                        [p]
                            An [b]excellent[/b] suggestion.
                            One you can't possibly resist.
                            Driven by a mind of its own, your tongue snakes out of your mouth,
                            outright [b]lapping[/b] at the pink kobold's folds, [b]tasting[/b] her feminine flavor.
                            It's not nearly enough to quell your burning desires.
                            Cassidy lets go of your neck.
                            You're not going anywhere.
                            The girl squirms as much as her restrains will allow.
                            As you begin to [b]eat her out[/b] properly, she [b]cries[/b] for help, only to be quickly silenced by the vixen.
                            "[color=orange]Good boy.
                            Don't worry, she'll end up liking this as much as you will, eventually.
                            Now on your feet, my precious new [b]breeder[/b].
                            [b]Mount the bitch.[/b][/color]"
                        [/p]
                        [p]
                            Like a mindless husk, you do exactly as your mistress tells you.
                            You retract your tongue from the pink kobold's folds.
                            Despite being unwilling, her wetness stains the front of your muzzle.
                            Its rich scent and lingering, bittersweet taste spurs you on further.
                            You rise to your feet and lean over her.
                            Positioning the tapered tip of your throbbing ${dick ? `` : `new`} member in between her legs,
                            you [b]ram[/b] your hips forwards, spreading her folds, firmly lodging your cock inside of her.
                        [/p]
                        [p]
                            The kobold's struggles intensify.
                            A few drops of blood ooze from her slit.
                            A virgin?
                            Not anymore.
                            She screams out in desperation and in pain, but her frantic pleas are muffled by Cassidy's furred hands.
                            The vixen doesn't even need to tell you to get started.
                            Your hips move on their own, rapidly pistoning your member in and out of the lizard's overly-tight cunt,
                            your green scales loudly slapping against her pink ones as you [b]hump[/b] away to your heart's content.
                        [/p]
                        [p]
                            Your need intensifies.
                            Your tongue lolls out of your mouth obscenely.
                            You can already feel your end approaching.
                            You grip onto the kobold's shoulders, fucking her [b]harder[/b] and [b]faster[/b],
                            [b]pounding[/b] her against the table with [b]all[/b] of your might, like a possessed [b]animal[/b].
                            Then, barely a minute after the intense [b]rutting[/b] began,
                            you [b]force[/b] your length in as deep as it'll go, one final time.
                            With a loud [b]grunt[/b], you reach your end,
                            your ${dick ? `` : `freshly formed`} internal balls pumping out wave after wave of spunk deep inside of your reluctant [b]bitch[/b],
                            thoroughly [b]breeding[/b] her.
                        [/p]
                        [p]
                            Your cock hardens again mere seconds after finishing.
                            Overtaken by aphrodisiac, your mind goes blank, and you continue to fuck the pink kobold for [b]hours[/b] on end,
                            dumping more loads inside of her than you can [b]count[/b].
                            Cassidy doesn't even bother to stay present all the time.
                            She fetches tea, does household chores and tends to her other [i]pets[/i].
                            At the end of the day, once you're fully spent, she guides you to one of the corrals,
                            where you drift to sleep after masturbating once or twice more.                            
                        [/p]
                        [p]
                            Everything from thereonout is a blur.
                            You get administered more drugs on a daily basis,
                            keeping you in a constant rutting-high,
                            driven by nothing but your desire to plant your seed inside of a fertile womb,
                            or [b]any[/b] hole tight enough to squeeze an orgasm out of you.
                        [/p]
                        [p]
                            You spend [b]days[/b] fucking the other kobolds in your pen.
                            Males, females, you don't care.
                            Sometimes you return the favor, and get them off as well.
                            Other times, you pump them full and leave them hanging until they [b]beg[/b] for release.
                            You find that works especially well on the ones that are chained up for their [b]bad behavior[/b].
                            Your member steadily growing due to the constant influx of various potions, you establish yourself as your pen's [b]alpha male[/b].
                        [/p]
                        [p]
                            Some of the other kobolds are reluctant to let you fuck them.
                            They don't want to be here.
                            They [b]beg[/b] you not to fuck them.
                            A few even snarl or claw at you.
                            After [b]hours[/b] of insisting, [b]all[/b] of them eventually cave.
                            If they don't, you wait until they go to sleep, and then [b]force[/b] yourself upon them.
                            None have the audacity to strike you, fearing the mistress' reprisal.
                            The last one who dared to start a fight got taken to the [b]butcher[/b].
                        [/p]
                        [p]
                            It isn't long until Cassidy starts taking a liking to you.
                            You become her [b]favorite[/b] pet.
                            You get better food than the rest, the privilege of roaming around at your own leisure instead of being confined to any one corral,
                            and the first go at breaking in fresh virgins.
                            Eventually, you even get to fuck the vixen herself.
                            At first, in the barn.
                            Then in her house.
                            After nearly a month of fucking her brains out daily, she invites you to her bedroom.
                        [/p]
                        [p]
                            A week later, she lets you sleep in bed with her,
                            instead of sending you back to the barn after sating her needs.
                            You cuddle up tightly against her warm fur, your spent member dribbling cum onto her back.
                            She rolls around to face you, lovingly kisses you on the head, and then embraces you like a lover would.
                            Finally, for the first time in years, you are truly happy.
                            With your scaly muzzle nestled snugly in between her firm breasts, the both of you slowly drift to sleep.
                        [/p]
                        [p]
                            [c]
                                [b]
                                    Bull end.
                                [/b]
                            [/c]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" style="margin-left: auto; position: relative; float: right;" id="${continueId}">`;
        choice += `&lt;&lt Revert`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400, function () {
            var main = $("#main");
            main.scrollTop(main[0].scrollHeight - main.height() - 5); // Fix scrolling
        });
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.result.unset();
            for (var i = 0; i < idsToRemove.length; i++) {
                $("#" + idsToRemove[i]).remove();
            }
            loadNextPart(ctrl.fragments.results);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoVixen.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoNobody.js

/**
 * Result section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoNobody(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoResultsChoice.NOBODY;
    var reqController = ctrl.result;
    var changeController = ctrl.nobody;

    var nextSceneReady = ctrl.nextSceneReady;
    var nextSceneName = ctrl.nextSceneName;
    var nextFolderName = ctrl.nextFolderName;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Nobody buys the kobold. Gloomy mood. Get shoved back into cart.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(nextSceneName, nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoNobody.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoSeki.js

/**
 * Result section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoSeki(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoResultsChoice.SEKI;
    var reqController = ctrl.result;
    var changeController = ctrl.seki;

    var nextSceneReady = ctrl.sekiSceneReady;
    var nextSceneName = ctrl.sekiSceneName;
    var nextFolderName = ctrl.sekiFolderName;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.boughtBySeki = true;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            "[color=orange]One hundred.[/color]" a gentle, feminine voice rings out.
                            It's the half-naked vixen.
                            She's soon challenged by the goblin, whose stingy pockets seem far deeper than what you'd have imagined.
                            "[color=green]One hundred and twenty![/color]"
                            The kind-looking canine casts a wicked leer towards the small humanoid.
                            "[color=orange]One hundred and fifty. Or would you rather deal with a back-handed [b]greenskin[/b]?[/color]"
                            "[color=green][b]One fifty-one[/b]![/color]"
                        [/p]
                        [p]
                            Bartholomew seems conflicted.
                            For a moment, it feels as though a [b]fight[/b] is brewing between the two interested parties.
                            But then, right as temperatures are about to reach a boiling point, you cry out for your avian friend in the distance,
                            drawing attention from your captors and both the interested parties, effectively [b]defusing[/b] the situation.
                            The owl's head turns towards you.
                            His eyes go wide.
                            As fast as his old legs can carry him, he rushes over to the crowd.
                        [/p]
                        [p]
                            [color=brown][big][b]THREE HUNDRED![/b][/big][/color]
                        [/p]
                        [p]
                            Seki's voice [b]booms[/b] over the rest of the crowd, although he bursts into a coughing fit at the end of his shout.
                            "[color=gblue]Sold to the senile bird![/color]" Fenrir instantly jumps in.
                            The goblin and the vixen both grumble and complain, but the wolf's decision is final.
                            The crowd parts and money exchanges hands.
                        [/p]
                        [p]
                            While Barto attempts to chat with Seki, your avian protector seems more interested in getting you the hell out of there.
                            The elderly man held up for a few minutes out of sheer courtesy, Fenrir approaches you.
                            "[color=gblue]Well, ${name}, you're his now.
                            Next time we meet, you can thank me for not letting that [b]filthy[/b] goblin have you.
                            He'd have matched the old coot's offer if I hadn't stepped in, you know.
                            Either way, good luck out there.
                            You're going to need it.[/color]"
                        [/p]
                        [p]
                            Having said his farewells, the big wolf pushes you forwards, nudging you off of the small stage, towards your new [i]master[/i].
                            The deal concluded, the human and the lupine walk back to the carriages, bickering on how to split their profits.
                            Your old friend approaches you, leaning down to closely inspect your body.
                            "[color=brown]Thank goodness, you're not hurt. You [b]scared[/b] me, ${name}.[/color]"
                            Bringing a key to your neck, he takes off your slave collar, and angrily hurls it down the well.
                            "[color=brown]A money transport?
                            I thought I taught you better!
                            [b]Never[/b] go after high-profile targets![/color]"
                        [/p]
                        [p]
                            Noticing the prying eyes of a few lingering locals, Seki quite literally takes you under his wing.
                            [color=brown]Come now.
                            We'll talk more once we get inside.[/color]"
                            Those words said, the two of you walk towards his workshop, just a few steps away.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart(nextSceneName, nextFolderName);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        if (!nextSceneReady) {
            part += `<span class="todo">There is no content beyond this point.</span>`;
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        if (nextSceneReady) {
            registerContinueListener();
        }
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoSeki.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoResults.js

/**
 * Result section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoResults(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preResult;
    var subController = ctrl.result;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";
    var fourthId = arguments.callee.name + "ButtonFour";

    var firstNext = new SlaveNoonTwoNobody(ctrl);
    var secondNext = new SlaveNoonTwoSeki(ctrl);
    var thirdNext = new SlaveNoonTwoFarmer(ctrl);
    var fourthNext = new SlaveNoonTwoVixen(ctrl);

    var firstChoice = SlaveNoonTwoResultsChoice.NOBODY;
    var secondChoice = SlaveNoonTwoResultsChoice.SEKI;
    var thirdChoice = SlaveNoonTwoResultsChoice.FARMER;
    var fourthChoice = SlaveNoonTwoResultsChoice.VIXEN;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        if (!ctrl.resultComplianceStorage.has()) {
            ctrl.resultComplianceStorage.set(
                    global.kobold.history.shared.enslavement.totalComplianceRating()
                    + ctrl.resultBodyMisc.boobs()
                    - ctrl.resultBodyMisc.dick());
        }

        var bartoOnStage = ctrl.stage.get() === SlaveNoonTwoStageChoice.REFUSE;

        var toReturn = `[p]
                            ${bartoOnStage ? `The caravan master hops off the stage.` : ``}
                            The bidding begins, at a price obviously [b]far[/b] too low according to Bartholomew.
                            "[color=purple]Ten coins![/color]" one of the horses offers.
                            "[color=green]Fifteen![/color]" the goblin counters.
                            "[color=purple]Oh yeah? Well...[/color]" reaching into his pocket, one of the equines produces an extra pair of shiny coins.
                            "[color=purple][b]Seventeen![/b][/color]" he proudly exlaims.
                            The goblin grumbles.
                            "[color=green]Twenty![/color]" comes his offer.
                        [/p]
                        [p]
                            Barto sighs.
                            Realizing this is going [b]nowhere[/b] fast,
                            the human puts an end to the [i]paltry[/i] bidding war.
                            "[color=brown][b]Serious[/b] offers [b]only[/b]!
                            Bidding starts at [b]fifty[/b] gold, and not a coin less.[/color]"
                            His bold statement immediately puts a damper on the crowd's enthusiasm.
                            Still, more than a few of the people gathered are still [b]very[/b] interested in [i]acquiring[/i] a kobold slave of their own.
                        [/p]
                        [p]
                            Right as you thought all hope was lost, a door opens at the far end of the hamlet.
                            It's [b]Seki[/b], emerging from his workshop!
                            He doesn't seem to pay the gathering around the stage any mind.
                            He never was one to support slavery.
                            If you can draw his attention, he [i]might[/i] be able to rescue you!
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        disabled = "disabled";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled} style="width: 190px;">Nobody</button>`;

        var tooltip = global.parser.parse(`The caravan master's not going to like this. [todo]Unavailable[/todo]`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var fail = ctrl.gag.get();
        var disabled = subController.has() || fail ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled} style="width: 190px;">Seki</button>`;

        var tooltip = global.parser.parse(`Call out to the elderly bird.`);
        if (fail) {
            tooltip = global.parser.parse(`You are gagged, and can not call for help.`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, `width: 1${fail ? `6` : `9`}0px;`);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var succeed = ctrl.resultComplianceStorage.get() > 15;
        var disabled = subController.has() || !succeed ? "disabled" : "";
        disabled = "disabled";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled} style="width: 190px;">A farmer</button>`;

        var tooltip = global.parser.parse(`The farmer's son looks fit to nag his father until he relents to buying you. [todo]Unavailable[/todo]`);
        if (!succeed) {
            tooltip = global.parser.parse(`The farmer is looking for a good girl. You don't interest him a lot.`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }
    /**
     * Builds and returns the choice button.
     */
    function buildFourthChoice() {
        var succeed = ctrl.resultComplianceStorage.get() < 8;
        var disabled = subController.has() || !succeed ? "disabled" : "";
        var isSelected = subController.get() === fourthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fourthId}" ${disabled} style="width: 190px;">The vixen</button>`;

        var tooltip = global.parser.parse(`That female fox has a wicked grin on her face. You don't like it one bit.`);
        if (!succeed) {
            tooltip = global.parser.parse(`The female fox is looking for a bad boy. You don't interest her a lot.`);
        }
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, `width: 1${succeed ? `6` : `7`}0px;`);
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You are bought by:</div>`;
        choice += `<div class="choicerow" style="margin-bottom: 10px;">`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        choice += `<div class="choicerow">`;
        choice += buildThirdChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFourthChoice();
        choice += `</div>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId), $("#" + fourthId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFourthListener() {
        $("#" + fourthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fourthChoice);
            loadNextPart(fourthNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        part += fourthNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();
        fourthNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
        registerFourthListener();
    };
}

var SlaveNoonTwoResultsChoice = {
    NOBODY: 1,
    SEKI: 2,
    FARMER: 3,
    VIXEN: 4
};

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\results\SlaveNoonTwoResults.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\stage\actions\SlaveNoonTwoPose.js

/**
 * Stage section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoPose(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoStageChoice.POSE;
    var reqController = ctrl.stage;
    var subController = ctrl.preResult;
    var changeController = ctrl.pose;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.results;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.stagePosed = true;
        var breasts = ctrl.misc.boobs();
        var injected = ctrl.gag.has();
        var analVirgin = ctrl.misc.analVirgin();
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            Smile?
                            You can do better than that.
                            Time to give these people a show.
                        [/p]
                        [p]
                            First, you lean forwards, and wink to your audience,
                            ${breasts ? `using both your hands to push your breasts together, ` : ``}
                            opening your mouth as wide as it'll go,
                            ${injected ? `[b]proudly[/b] showcasing your
                            ${global.kobold.inventory.createItemWithTooltip(new UnderwearGag(), true)}.` : `
                            showcasing how long your tapered tongue is by lolling it out obscenely.`}
                            ${injected ? `A mixed response from the crowd.` : `The crowd seems to [b]love[/b] it, especially the young bird boy.`}
                            That's not the only trick you have up your sleeve.
                        [/p]
                        [p]
                            Time to bring out the big guns.
                            Turning around, you lewdly bend over, lifting your tail up high, exposing your hindquarters to your potential [i]buyers[/i].
                            As if that wasn't enough, you spread your buttcheeks using your hands,
                            giving [b]everyone[/b] a good look at your ${analVirgin ? `virginal ass` : `recently fucked ass`}.
                            To entice the crowd further, you teasingly run a trace a single finger along the length of your ${dick ? `cock` : ``}slit.
                        [/p]
                        [p]
                            Turning to face your onlookers once more,
                            you decide to give them one [b]final[/b] gesture,
                            to push them over the edge, to [b]ensure[/b] you'll be [b]bought[/b].
                            You grab your tail, and suggestively stroke your hand along its tip.
                            The teenaged avian in the audience can [b]barely[/b] refrain from touching himself at the sight,
                            his human mother scolding him with a quick slap to the back of his head.
                            You'd like to think you did well.
                            At least the caravan master can't say you didn't try.
                            He [b]looks[/b] pleased, at least.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\stage\actions\SlaveNoonTwoPose.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\stage\actions\SlaveNoonTwoRefuse.js

/**
 * Stage section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoRefuse(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoStageChoice.REFUSE;
    var reqController = ctrl.stage;
    var subController = ctrl.preResult;
    var changeController = ctrl.refuse;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.results;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Smile?
                            Hell no.
                            You have [b]no[/b] reason whatsoever to smile.
                            You have been [b]enslaved[/b] and are now being [b]sold[/b] to the highest bidder like a piece of [b]meat[/b].
                        [/p]
                        [p]
                            Instead of smiling, you cast an angry glare towards Bartholomew.
                            The human doesn't appreciate your lack of cooperation.
                            His rage just about boiling over, he strides atop the wooden stage,
                            and before you even know what's going on,
                            he roughly [b]grabs[/b] you by your throat,
                            squeezing your windpipe shut while lifting you up high enough to come face-to-face with him.
                        [/p]
                        [p]
                            "[color=brown][b]Normally,[/b][/color]" the human speaks, addressing the crowd without looking at them,
                            glaring into your eyes with an intense look of pure [b]hatred[/b] on his face.
                            "[color=brown]I'd beat the [b]shit[/b] out of this little [b]cretin[/b] for disobeying me.[/color]"
                            You squirm around in Barto's grasp, running slightly low on air, weakly batting at his arm to [b]let go[/b].
                            "[color=brown]But, today, I'll leave that pleasure to [b]you[/b], the people.[/color]"
                            With his free hand, he points towards the crowd to accentuate his speech.
                        [/p]
                        [p]
                            Then, his hand curls into a fist, one which he [b]painfully[/b] rams into your gut with all of his might,
                            knocking what little air remained out of your lungs, while simultaneously letting go of your throat,
                            letting you drop down to the ground.
                            Your legs unable to carry you, you crumple to the floor, curling up into a tiny little ball,
                            clutching your battered stomach.
                            It [b]hurts[/b].
                            The crowd [b]laughs[/b].
                            Your pain is their [b]amusement[/b].
                            You burst out into tears.
                        [/p]
                        [p]
                            [big][color=brown][b]On your feet, slave![/b][/color][/big]
                        [/p]
                        [p]
                            The caravan master rears his foot backwards, threatening to [b]kick[/b] you lest you comply with his demand.
                            Fearing more [b]beatings[/b], you fearfully scramble to your feet,
                            visibly shaken from the human's sudden outburst of anger.
                            "[color=brown]See? If you [b]hurt[/b] it enough, it'll start listening [b]eventually[/b].
                            And if not, you'll have had the pleasure of killing a [b]criminal[/b].[/color]"
                            More crude laughter.
                            More tears.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\stage\actions\SlaveNoonTwoRefuse.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\stage\actions\SlaveNoonTwoSmile.js

/**
 * Stage section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoSmile(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoStageChoice.SMILE;
    var reqController = ctrl.stage;
    var subController = ctrl.preResult;
    var changeController = ctrl.smile;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.results;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.stageSmiled = true;
        var injected = ctrl.gag.has();
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var toReturn = `[p]
                            Smile?
                            For these [i]oafs[/i] that, after having been riled up by Bartholomew, pretty much want you [b]dead[/b]?
                            ${eager ? `Well, it would probably make Fenrir happy` : `Well, it [b]will[/b] up your odds of being [i]sold[/i]`},
                            and showing off how [i]nice[/i] you are [b]might[/b] make whoever ends up buying you just a little bit less inclined to [b]cave your head in[/b].
                            If you [b]really[/b] have to, out of sheer self-preservation, you think you [i]might[/i] be able to force out a weak happy look.
                        [/p]
                        [p]
                            ${injected ? `Even with the hefty gag in your mouth,
                            you still manage to form something remotely managing a smile.` : `Slowly, a small, obviously forced smile forms on your face.`}
                            In order to [i]accentuate[/i] your [i]happiness[/i] further, you form the fingers of both your hands into peace signs.
                            "[color=brown]You see?
                            Totally compliant![/color]"
                            Bartholomew comments, the onlookers nodding contently, some more pleased than others.
                            The father of the family with the avian son, in particular, seems happy with your actions.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\stage\actions\SlaveNoonTwoSmile.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\stage\SlaveNoonTwoStage.js

/**
 * Stage section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoStage(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preStage;
    var subController = ctrl.stage;
    var changeController = ctrl.stage;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new SlaveNoonTwoRefuse(ctrl);
    var secondNext = new SlaveNoonTwoSmile(ctrl);
    var thirdNext = new SlaveNoonTwoPose(ctrl);

    var firstChoice = SlaveNoonTwoStageChoice.REFUSE;
    var secondChoice = SlaveNoonTwoStageChoice.SMILE;
    var thirdChoice = SlaveNoonTwoStageChoice.POSE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var name = ctrl.misc.name();
        var dick = ctrl.misc.dick();
        var injected = ctrl.gag.has();
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var toReturn = `[p]
                            Seemingly directing his anger for the lupine towards you,
                            the caravan master grabs you by the collar, pulling you along as he strides towards the wooden stage.
                            You can only [b]barely[/b] keep up with his upset pace, but fortunately the podium isn't too far off.
                            "[color=brown]Come, people, gather 'round!
                            The main event is about to begin![/color]"
                            The human announces, addressing the people on the streets.
                            You have a [b]bad[/b] feeling about this.
                        [/p]
                        [p]
                            A small crowd slowly forms in front of the stage, numbering more than a dozen heads.
                            This must be at least half the people that live here, if not more!
                            No sign of Seki, though.
                            Most of them are talking amongst themselves, taking the occasion as a big social gathering,
                            which doesn't seem to amuse the already annoyed caravan master very much.
                            "[color=brown]Hear ye, hear ye!
                            [b]Settle down![/b]
                            [small]Filthy peasants.[/small][/color]"
                            That last part was naught but a whisper, one you [b]hope[/b] only [b]you[/b] heard, lest this [i]auction[/i] turns into a [b]lynching[/b].
                        [/p]
                        [p]
                            The mob slowly goes silent, their attention snagged by the authoratively dressed human and your naked frame besides him.
                            You shyly try to conceal your nudity by standing [i]behind[/i] the caravan master, but he [b]firmly[/b] yanks you back into place at his side.
                            Worse yet, as the [b]show[/b] is about to start, he crudely picks you up,
                            effortlessly lifts you off the ground, and practically [b]throws[/b] you on top of the wooden stage.
                            You quickly rise to your feet atop the podium, using your hands and your tail both to cover your [i]private parts[/i].
                        [/p]
                        [p]
                            With nowhere to run and nowhere to hide, you are at the mercy of the crowd's leering stares.
                            You can feel them [b]inspecting[/b] you, [b]judging[/b] you, estimating your worth in gold,
                            whispering to one another to comment on your [i]humble[/i] appearance, or to share how much they're willing to bid for your [b]body[/b].
                        [/p]
                        [p]
                            Amongst the onlookers, a few stick out the most.
                            There's a voluptuous vixen dressed in naught but a black shirt, a cowboy hat and some spurred, leather boots.
                            The blouse barely covers her loins.
                            You wonder if she's wearing anything [i]underneath[/i].
                            Next to her, there are a pair of muscular anthro horses.
                            Probably farmers.
                            They look like siblings.
                            Brothers or twins or something, maybe?
                            You don't know.
                            A lot of horses look alike to you.
                        [/p]
                        [p]
                            To the side, there stands an odd family.
                            A human mother and father, with a distinctly [b]avian[/b] son.
                            You don't know much about birds, but you think he's an [b]eagle[/b].
                            He looks kind of like one of Seki's relatives that you met a long time ago.
                            You briefly wonder how a bird and human could [i]mix[/i], but you decide not to linger on it for too long.
                            A final person of note, standing amidst a few of the more common farmers, is a single [b]goblin[/b].
                            You don't think he's from around here.
                            Behind the rabble, Fenrir is leaning slickly against the side of a building,
                            now dressed and eager to see the auction play out.
                        [/p]
                        [p]
                            With the crowd reduced to whispers, the caravan master begins to address them.
                            "[color=brown]I am [b]Bartholomew Grabgold[/b], royal tax collector,
                            appointed by the [b]king[/b] himself, currently serving the baron-regent.
                            But make no mistake, today I am [b]not[/b] here to [b]take[/b] your money.[/color]"
                            The chubby man chuckles, trying to lighten his own mood.
                            "[color=brown]That doesn't mean I intend to leave here empty-handed, though.[/color]"
                            A small, awkward pause.
                            The farmers don't seem to understand the joke.
                            Heck, you're not sure if [b]you[/b] get it.
                        [/p]
                        [p]
                            Disdain and frustration a bit more clear in his voice, [i]Bartholomew[/i] extends his arm towards you,
                            directing attention to your naked form before continuing.
                            "[color=brown]Today, I have come before you with a [b]slave[/b] to sell.
                            And not just [b]any[/b] slave.
                            You see, this kobold was [b]stupid[/b] enough to attempt to [b]steal[/b] from an imperial money transport.
                            It wanted to take gold to be used for the betterment of [b]all[/b] of mankind, to use for its own [b]selfish[/b] needs.[/color]"
                            You can almost [b]feel[/b] the crowd grow more [i]hostile[/i] towards you.
                        [/p]
                        [p]
                            "[color=brown]Not the smartest thing, is it?
                            Don't worry, the dumb lizard can still be used for simple, menial labor.
                            But you shouldn't buy this slave only to make it clean your house or work the fields.
                            No, instead, you should spend your coin on little ${name} here to [b]punish[/b] ${dick ? `him` : `her`}.
                            To make an [b]example[/b] out of ${dick ? `him` : `her`}.
                            To show ${dick ? `him` : `her`}, and indeed, [b]all[/b] kobolds,
                            that [b]any[/b] transgressions against their [b]superiors[/b] is a one-way ticket to [b]hell[/b] on earth.[/color]"
                        [/p]
                        [p]
                            The human has the crowd eating out of his hand.
                            Still, these people are but simple farmers.
                            You doubt many -- if any -- of them have the [b]means[/b] to [i]buy[/i] you.
                            Being relatively [i]poor[/i], surely they won't spend a bunch of money only to [b]torture[/b] or outright [b]kill[/b] you, would they?
                        [/p]
                        [p]
                            "[color=brown]${injected ? `As you can tell by the gag in its mouth, this lizard is ill-mannered and plain-out [b]rude[/b].
                            Buy it to teach it a [b]lesson[/b].
                            Whip it until it passed out.
                            ${fucked ? `Make it wish it was never even born.` : `Take its virginity.`}
                            Fuck it until it can no longer walk.
                            Break all of its limbs, and then slice its throat.
                            Make it [b]suffer[/b] until its very last breath.
                            ` : `While not the sharpest tool in the shed, this lizard has proven to be a bit more [b]cooperative[/b] than the others.
                            Perfect for all you [b]lonely men[/b] out there.
                            I can't ensure it won't [b]bite[/b], but there's two other holes for you to [b]fuck[/b] if you don't want to risk it.
                            Or you can pull out its teeth for maximum security.
                            ${fucked ? `One of its holes might still be virginal.
                            Find out which one after the auction.` : `Did I mention it's a virgin?
                            [b]Extra tight[/b] on the first go-around.
                            Guaranteed to be a [i]suitable[/i] fit for even a child.
                            A perfect gift for your firstborn son.`}`}[/color]"
                        [/p]
                        [p]
                            The crowd seems [b]very[/b] enticed.
                            [b]Barto[/b] appears to have a silver tongue when it comes to selling things.
                            His audience excited, the human turns to you.
                            "[color=brown][b]Smile, slave![/b][/color]" comes the simple command.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Refuse</button>`;

        var tooltip = global.parser.parse(`You aren't happy. Why would you pretend to be?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Smile</button>`;

        var tooltip = global.parser.parse(`You don't want to get hurt. Do as he says.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Pose</button>`;

        var tooltip = global.parser.parse(`Go above and beyond what is requested of you.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var SlaveNoonTwoStageChoice = {
    REFUSE: 1,
    SMILE: 2,
    POSE: 3
};

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\stage\SlaveNoonTwoStage.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\refuse\SlaveNoonTwoOutsideRefuseGood.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutsideRefuseGood(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preStage;
    var changeController = ctrl.refuseGood;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.stage;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.goodRefuse = true;
        var injected = ctrl.gag.has();
        var neither = ctrl.potion.get() === SlaveNoonTwoGoodPotionChoice.NEITHER;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var kissedBalls = global.kobold.history.shared.enslavement.noonTwo.kissedBalls || eager;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            Darnit.
                            Why couldn't the wolf have brought you to the river instead?!
                            Do you really have to wash up in the [b]middle[/b] of the hamlet, with all these people watching?
                            In a shallow [b]trough[/b] used by beasts of burden to slake their thirst?
                            You ${injected ? `sigh internally` : `let out another sigh`}.
                            You'd rather have some [b]privacy[/b] and the comfort of [b]running[/b] or at least [b]fresh[/b] water,
                            instead of this stale puddle of [b]animal spit[/b].
                        [/p]
                        [p]
                            It may be too late for the two of you to head for the river,
                            but that doesn't mean you're going to [b]wallow[/b] in animal grool for the bandit's amusement.
                            ${fucked ? `${eager ? `You wear Fenrir's dried spunk with pride,
                            and you don't want it washed off.` : `While you're eager to rid yourself of Fenrir's dried spunk,
                            you're not going to degrade yourself to wash it off.`}` : `It would be nice to wash off all the dirt from the chase,
                            but you're not going to degrade yourself for it.`}
                            Besides, it's a watering trough.
                            If you clean yourself in it, where will all the animals drink from?
                            Plus, your captor said the auction was starting soon.
                            You won't be able to dry yourself in time if you [i]take a dip[/i].
                        [/p]
                        [p]
                            Tilting your head upwards, you meet the wolf's determined gaze with a resolute stare of your own.
                            Your nerves steeled, you ${injected ? `firmly shake your head no` : `boldly refuse his request for you to bathe`}.
                            "[color=gblue]We [b]don't[/b] have time for this.
                            [b]Get in the tub.[/b][/color]"
                            Like a young child refusing to take an order from its parents, you hold your ground, stubbornly denying the canine.
                            Fenrir lets out an exasperated groan.
                            "[color=gblue]Listen, I'm not going to force you to wash up.
                            The caravan master's going to be [b]pissed[/b] that you're still covered in ${fucked ? `jizz` : `dirt`}, but so be it.
                            I'll take the flak for you this [b]one[/b] time, just because I like you.
                            [b]Don't[/b] get used to-[/color]"
                            Suddenly, a booming voice interrupts the bandit.
                        [/p]
                        [p]
                            [big][color=brown][b]Fenrir!![/b][/color][/big]
                        [/p]
                        [p]
                            It's the caravan master!
                            He's approaching the two of you rapidly.
                            "[color=brown]What the [b]hell[/b] are you doing, standing around?!
                            I thought I told you to get the slave cleaned?![/color]"
                            The chubby man walks up to the wolf.
                            Despite the lupine being nearly an entire foot taller,
                            the human assumes a crude, threatening posture, as if looking for a fight.
                            You [i]want[/i] to come between the two to explain what happened, yet you fail to find the courage to do so.
                            ${injected ? `Besides, you can't speak through your ${global.kobold.inventory.createItemWithTooltip(new UnderwearGag())}` : ``}
                        [/p]
                        [p]
                            "[color=gblue]Kobold's clean enough to be sold.
                            Now [b]back off[/b].[/color]"
                            Your captor's calm, collected response only serves to anger his [sneak]temporary[/sneak] superior further.
                            "[color=brown]Back off?!
                            I asked [b]one[/b] simple thing of you!
                            But then again, you couldn't even guard the transport in the first place![/color]"
                        [/p]
                        [p]
                            The wolf tries, but fails to come up with an apt response.
                            Sighing discontently, he submissively takes a step backwards.
                            For a moment there, you thought this was going to end in a fight.
                            "[color=brown]That's right.
                            Know your place, [b]mutt[/b].
                            Now quit acting like the [b]animal[/b] you are, and go get dressed.[/color]"
                            After an intense stare-down, Fenrir finally turns around.
                            Audibly grumbling, he begins to make his way back to the wagons, leaving you alone with the human, who harrumphs angrily.
                            "[color=brown]The baron will hear of this.[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\refuse\SlaveNoonTwoOutsideRefuseGood.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\refuse\SlaveNoonTwoOutsideRefuseBad.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutsideRefuseBad(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preStage;
    var changeController = ctrl.refuseBad;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.stage;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.badRefuse = true;
        var injected = ctrl.gag.has();
        var neither = ctrl.potion.get() === SlaveNoonTwoGoodPotionChoice.NEITHER;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var kissedBalls = global.kobold.history.shared.enslavement.noonTwo.kissedBalls || eager;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            Darnit.
                            Why couldn't the wolf have brought you to the river instead?!
                            Do you really have to wash up in the [b]middle[/b] of the hamlet, with all these people watching?
                            In a shallow [b]trough[/b] used by beasts of burden to slake their thirst?
                            You ${injected ? `sigh internally` : `let out another sigh`}.
                            You'd rather have some [b]privacy[/b] and the comfort of [b]running[/b] or at least [b]fresh[/b] water,
                            instead of this stale puddle of [b]animal spit[/b].
                        [/p]
                        [p]
                            It may be too late for the two of you to head for the river,
                            but that doesn't mean you're going to [b]wallow[/b] in animal grool for the bandit's amusement.
                            ${fucked ? `${eager ? `You wear Fenrir's dried spunk with pride,
                            and you don't want it washed off.` : `While you're eager to rid yourself of Fenrir's dried spunk,
                            you're not going to degrade yourself to wash it off.`}` : `It would be nice to wash off all the dirt from the chase,
                            but you're not going to degrade yourself for it.`}
                            Besides, it's a watering trough.
                            If you clean yourself in it, where will all the animals drink from?
                            Plus, your captor said the auction was starting soon.
                            You won't be able to dry yourself in time if you [i]take a dip[/i].
                        [/p]
                        [p]
                            Tilting your head upwards, you meet the wolf's determined gaze with a resolute stare of your own.
                            Your nerves steeled, you ${injected ? `firmly shake your head no` : `boldly refuse his request for you to bathe`}.
                            "[color=gblue]We [b]don't[/b] have time for this.
                            [b]Get in the tub.[/b][/color]"
                            Like a young child refusing to take an order from its parents, you hold your ground, stubbornly denying the canine.
                            Fenrir's growing angry.
                            He bares his fangs, growling loudly.
                            "[color=gblue]Get [b]the fuck[/b] in, ${name}.
                            Final warning.[/color]"
                        [/p]
                        [p]
                            The wolf's furred hand slowly raises up high, poised to [b]strike[/b].
                            Not just a slap, either.
                            His claws are fully extended.
                            You are more than a little intimidated, but you don't back down.
                            You drop into a combat stance, bracing yourself for his strike.
                            He's bluffing.
                            He won't hit you.
                            Will he?
                            Suddenly, a booming voice interrupts your impending fight.
                        [/p]
                        [p]
                            [big][color=brown][b]Fenrir!![/b][/color][/big]
                        [/p]
                        [p]
                            It's the caravan master!
                            He's approaching the two of you rapidly.
                            "[color=brown]What the [b]hell[/b] are you doing?![/color]"
                            The chubby human crudely pushes the wolf away, stepping in between the two of you.
                            You immediately snap out of your combat stance, bringing out your best [i]puppy eyes[/i] to look as [b]innocent[/b] as possible.
                            "[color=gblue]Get out of my way, human.[/color]"
                            Your captor's response only serves to anger his [sneak]temporary[/sneak] superior further.
                            "[color=brown]Get out of your way?!
                            Half the bloody village is watching!
                            We're trying to [b]sell[/b] to these people,
                            and you [b]threatening[/b] our [i]merchandise[/i] in front of them [b]isn't[/b] helping.[/color]"
                        [/p]
                        [p]
                            The wolf tries, but fails to come up with an apt response.
                            Sighing discontently, he shifts out of his threatening posture.
                            For a moment there, you thought he was going to hit the caravan master.
                            "[color=brown]That's more like it.
                            Now quit acting like the [b]animal[/b] you are, and go get dressed.[/color]"
                            After an intense stare-down, Fenrir finally backs off.
                            Audibly grumbling, he begins to make his way back to the wagons, leaving you alone with the human, who harrumphs angrily.
                            "[color=brown]The baron will hear of this.[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\refuse\SlaveNoonTwoOutsideRefuseBad.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\refuse\SlaveNoonTwoOutsideRefuse.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutsideRefuse(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOutsideChoice.REFUSE;
    var reqController = ctrl.outside;

    var toFail = new SlaveNoonTwoOutsideRefuseBad(ctrl);
    var toSuccess = new SlaveNoonTwoOutsideRefuseGood(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        if (!ctrl.outsideComplianceStorage.has()) {
            ctrl.outsideComplianceStorage.set(global.kobold.history.shared.enslavement.totalComplianceRating());
        }
        return ctrl.outsideComplianceStorage.get() > 8;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\refuse\SlaveNoonTwoOutsideRefuse.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\drink\SlaveNoonTwoOutsideDrinkGood.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutsideDrinkGood(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preStage;
    var changeController = ctrl.drinkGood;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.stage;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.goodDrink = true;
        var injected = ctrl.gag.has();
        var neither = ctrl.potion.get() === SlaveNoonTwoGoodPotionChoice.NEITHER;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var kissedBalls = global.kobold.history.shared.enslavement.noonTwo.kissedBalls || eager;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            Washing up is the last thing on your mind right now.
                            If anything, the sight of nourishing liquid reminds you that your throat is [b]parched[/b].
                            You haven't had a drink since long before stealing from the money transport yesterday.
                            Since then, you've ran, you've hidden,
                            then you ran some more${fucked ? `, got fucked` : ``},
                            spent the night imprisoned${kissedBalls ? `, kissed the wolf's dry balls` : ``}${injected ? `, got the wolf's dry underwear shoved into your mouth` : ``}
                            and all of this without a single drop of water.
                            No wonder you woke up with a headache.
                            You're dehydrated to heck and back.
                        [/p]
                        [p]
                            Even as thirsty as you are, the water before you looks unappetizing.
                            Darnit.
                            Why couldn't the wolf have brought you to the river instead?!
                            You'd have had something [b]fresh[/b] to drink in relative [b]privacy[/b],
                            instead of being forced to lap stale [b]slobber[/b] from a trough in public.
                        [/p]
                        [p]
                            Still, you [b]need[/b] something, [b]anything[/b] to drink.
                            However disgusting the liquid may be, it'll have to do.
                            You drop to your knees, the foul scent of feral horse spit only now reaching your nose fully.
                            It [b]stinks[/b].
                            Gross!
                            Are you [b]really[/b] going to do this?
                        [/p]
                        [p]
                            Yes.
                            Yes you are.
                            You feel like you'll [b]die[/b] if you don't get any moisture in you.
                            Under Fenrir's watchful eye,
                            you lower your head into the trough, hovering your muzzle directly above the water's surface.
                            The [b]stench[/b] is nearly unbearable.
                            Out of [b]necessity[/b],
                            you snake your tongue out of your mouth,
                            ${injected ? `scraping the prehensile appendage against the underwear in your mouth in the process,` : ``}
                            and lap at the fetid fluid that was once water.
                        [/p]`;

        if (injected) {
            toReturn += `[p]
                            It's [b]thick[/b].
                            It's [b]warm[/b].
                            It's not even anything [b]remotely[/b] like water.
                            The top layer of the trough, in which your tongue dipped, is pure, fresh [b]equine saliva[/b].
                            You retch even [b]before[/b] you manage to get the slimey goop into your mouth.
                            Worse yet, as your tongue retracts into your muzzle, it [b]soaks[/b] into your underwear-gag,
                            washing out an entire new [b]wave[/b] of dirt and filth concealed within the fabric,
                            the horrid taste of animal drool overshadowed by the absolutely [b]abhorrent[/b] flavor of wolf groin-grime.
                        [/p]
                        [p]
                            Your stomach churns, barely able to hold its contents.
                            But you're not nearly done.
                            Despite being outright [b]sickening[/b] in flavor, a lot warmer and far more [b]viscous[/b] than you would have liked,
                            the slop you licked up [i]did[/i] end up [b]refreshing[/b] the dry front-end of your mouth.
                            You need [b]more[/b].
                            However much it [b]revolts[/b] you,
                            you need to [b]keep licking[/b] until the underwear becomes so soaked that some much-needed liquid can make its way to your throat.
                        [/p]
                        [p]
                            Trying to think of [b]anything[/b] but the vomit-inducing taste of the feral spit,
                            you lap again and again at the layer of stale saliva,
                            taking more and more of it into your muzzle,
                            letting it [b]soak[/b] into the briefs lodged into the back of your mouth.
                            Once finally [b]saturated[/b] with packbeast-dribble,
                            you tilt your head backwards and bite down on the garment,
                            [b]squeezing[/b] the repulsive liquid from it,
                            letting the drool, along with more filth rinsed from the wolf's underwear,
                            ooze down your throat into your stomach.
                        [/p]
                        [p]
                            While it doesn't nearly manage to quench your thirst,
                            the lukewarm fluid sticks to your gums and to the walls of your esophagus,
                            slickening the back of your mouth, leaving a lingering, unpleasant yet somewhat [i]reinvigorating[/i] feeling.
                            At least the permeating [b]dryness[/b] onset by canine briefs is gone.
                            And, pleasantly enough, the underwear filters out most of the [i]fur[/i] and [i]grass[/i] contained within the goop.
                        [/p]
                        [p]
                            While still dehydrated, you doubt drinking more of the spittle will help.
                            You don't think you can [b]stomach[/b] any more of it, regardless.
                            Removing your head from the trough, you notice Fenrir staring at you with a disapproving gaze.
                        [/p]
                        [p]
                            "[color=gblue]If I wanted my underpants cleaned, I'd have taken you to the river instead.[/color]"
                            How ironic.
                            You cast a glare the canine's way.
                            "[color=gblue]Hey, now, I didn't force you to drink that stuff.
                            That's on [b]you[/b].[/color]"
                            He's right.
                            But that doesn't prevent you from being upset.
                            Suddenly, a voice rings throughout the hamlet, interrupting your uppity streak.
                        [/p]`;
        } else {
            toReturn += `[p]
                            It's [b]thick[/b].
                            It's [b]warm[/b].
                            It's not even anything [b]remotely[/b] like water.
                            The top layer of the trough, in which your tongue dipped, is pure [b]equine saliva[/b].
                            You retch even [b]before[/b] you manage to get the slimey goop into your mouth.
                            As your tongue retracts into your muzzle, the horrid taste of animal drool spreads throughout your mouth.
                            You swallow the stuff as quickly as you can, but the taste lingers.
                        [/p]
                        [p]
                            Your stomach churns, barely able to hold its fresh contents.
                            But you're not nearly done.
                            Despite being outright [b]sickening[/b] in flavor, a lot warmer and far more [b]viscous[/b] than you would have liked,
                            the slop you licked up [i]did[/i] end up [b]refreshing[/b] your dry mouth.
                            You need [b]more[/b].
                            However much it [b]revolts[/b] you,
                            you need to [b]keep drinking[/b] until you've taken in enough to survive.
                        [/p]
                        [p]
                            Trying to think of [b]anything[/b] but the vomit-inducing taste of the feral spit,
                            you lap again and again at the layer of stale saliva,
                            taking more and more of it into your muzzle,
                            letting it pool up before swallowing entire mouthfuls at a time.
                            You almost [b]throw up[/b] on the spot, as you accidentally lap up a clump of fur,
                            immediately followed by some chewed grass.
                            Luckily, you manage to refrain from vomiting,
                            as that would only serve to make the trough's contents even [b]less[/b] bearable to drink.
                        [/p]
                        [p]
                            While it doesn't nearly manage to quench your thirst,
                            the lukewarm fluid sticks to your gums and to the walls of your esophagus,
                            slickening your mouth, leaving a lingering, unpleasant yet somewhat [i]refreshing[/i] feeling.
                            You'll undoubtedly have bad breath for days to come.
                        [/p]
                        [p]
                            While still dehydrated, you doubt drinking more of the spittle will help.
                            You don't think you can [b]stomach[/b] any more of it, regardless.
                            Removing your head from the trough, you notice Fenrir staring at you with a disapproving gaze.
                        [/p]
                        [p]
                            "[color=gblue]You know, I'd have fetched some water from the well, if you'd asked.[/color]"
                            How ironic.
                            You cast a glare the canine's way.
                            "[color=gblue]Hey, now, I didn't force you to drink that stuff.
                            That's on [b]you[/b].[/color]"
                            He's right.
                            But that doesn't prevent you from being upset.
                            Suddenly, a voice rings throughout the hamlet, interrupting your uppity streak.
                        [/p]`;
        }

        toReturn += `   [p]
                            [big][color=brown][b]Fenrir!![/b][/color][/big]
                        [/p]
                        [p]
                            It's the caravan master!
                            He's approaching the two of you rapidly.
                            "[color=brown]What the [b]hell[/b] are you doing, standing around?!
                            I thought I told you to get the slave cleaned?![/color]"
                            The chubby man walks up to the wolf.
                            Despite the lupine being nearly an entire foot taller,
                            the human assumes a crude, threatening posture, as if looking for a fight.
                            You [i]want[/i] to come between the two to explain what happened, yet you fail to find the courage to do so.
                            ${injected ? `Besides, you can't speak through your ${global.kobold.inventory.createItemWithTooltip(new UnderwearGag())}` : ``}
                        [/p]
                        [p]
                            "[color=gblue]Kobold was thirsty.
                            Besides, it's clean enough to be sold.
                            Now [b]back off[/b].[/color]"
                            Your captor's calm, collected response only serves to anger his [sneak]temporary[/sneak] superior further.
                            "[color=brown]Back off?!
                            I asked [b]one[/b] simple thing of you!
                            But then again, you couldn't even guard the transport in the first place![/color]"
                        [/p]
                        [p]
                            The wolf tries, but fails to come up with an apt response.
                            Sighing discontently, he submissively takes a step backwards.
                            For a moment there, you thought this was going to end in a fight.
                            "[color=brown]That's right.
                            Know your place, [b]mutt[/b].
                            Now quit acting like the [b]animal[/b] you are, and go get dressed.[/color]"
                            After an intense stare-down, Fenrir finally turns around.
                            Audibly grumbling, he begins to make his way back to the wagons, leaving you alone with the human, who harrumphs angrily.
                            "[color=brown]The baron will hear of this.[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\drink\SlaveNoonTwoOutsideDrinkGood.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\drink\SlaveNoonTwoOutsideDrinkBad.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutsideDrinkBad(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preStage;
    var changeController = ctrl.drinkBad;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.stage;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.badDrink = true;
        var injected = ctrl.gag.has();
        var neither = ctrl.potion.get() === SlaveNoonTwoGoodPotionChoice.NEITHER;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var kissedBalls = global.kobold.history.shared.enslavement.noonTwo.kissedBalls || eager;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            Washing up is the last thing on your mind right now.
                            If anything, the sight of nourishing liquid reminds you that your throat is [b]parched[/b].
                            You haven't had a drink since long before stealing from the money transport yesterday.
                            Since then, you've ran, you've hidden,
                            then you ran some more${fucked ? `, got fucked` : ``},
                            spent the night imprisoned${kissedBalls ? `, kissed the wolf's dry balls` : ``}${injected ? `, got the wolf's dry underwear shoved into your mouth` : ``}
                            and all of this without a single drop of water.
                            No wonder you woke up with a headache.
                            You're dehydrated to heck and back.
                        [/p]
                        [p]
                            Even as thirsty as you are, the water before you looks unappetizing.
                            Darnit.
                            Why couldn't the wolf have brought you to the river instead?!
                            You'd have had something [b]fresh[/b] to drink in relative [b]privacy[/b],
                            instead of being forced to lap stale [b]slobber[/b] from a trough in public.
                        [/p]
                        [p]
                            Still, you [b]need[/b] something, [b]anything[/b] to drink.
                            However disgusting the liquid may be, it'll have to do.
                            You drop to your knees, the foul scent of feral horse spit only now reaching your nose fully.
                            It [b]stinks[/b].
                            Gross!
                            Are you [b]really[/b] going to do this?
                        [/p]
                        [p]
                            Yes.
                            Yes you are.
                            You feel like you'll [b]die[/b] if you don't get any moisture in you.
                            Under Fenrir's watchful eye,
                            you lower your head into the trough, hovering your muzzle directly above the water's surface.
                            The [b]stench[/b] is nearly unbearable.
                            Out of [b]necessity[/b],
                            you snake your tongue out of your mouth,
                            ${injected ? `scraping the prehensile appendage against the underwear in your mouth in the process,` : ``}
                            and lap at the fetid fluid that was once water.
                        [/p]`;

        if (injected) {
            toReturn += `[p]
                            It's [b]thick[/b].
                            It's [b]warm[/b].
                            It's not even anything [b]remotely[/b] like water.
                            The top layer of the trough, in which your tongue dipped, is pure, fresh [b]equine saliva[/b].
                            You retch even [b]before[/b] you manage to get the slimey goop into your mouth.
                            Worse yet, as your tongue retracts into your muzzle, it [b]soaks[/b] into your underwear-gag,
                            washing out an entire new [b]wave[/b] of dirt and filth concealed within the fabric,
                            the horrid taste of animal drool overshadowed by the absolutely [b]abhorrent[/b] flavor of wolf groin-grime.
                        [/p]
                        [p]
                            Your stomach churns, barely able to hold its contents.
                            But you're not nearly done.
                            Despite being outright [b]sickening[/b] in flavor, a lot warmer and far more [b]viscous[/b] than you would have liked,
                            the slop you licked up [i]did[/i] end up [b]refreshing[/b] the dry front-end of your mouth.
                            You need [b]more[/b].
                            However much it [b]revolts[/b] you,
                            you need to [b]keep licking[/b] until the underwear becomes so soaked that some much-needed liquid can make its way to your throat.
                        [/p]
                        [p]
                            Trying to think of [b]anything[/b] but the vomit-inducing taste of the feral spit,
                            you lap again and again at the layer of stale saliva,
                            taking more and more of it into your muzzle,
                            letting it [b]soak[/b] into the briefs lodged into the back of your mouth.
                            Once finally [b]saturated[/b] with packbeast-dribble,
                            you tilt your head backwards and bite down on the garment,
                            [b]squeezing[/b] the repulsive liquid from it,
                            letting the drool, along with more filth rinsed from the wolf's underwear,
                            ooze down your throat into your stomach.
                        [/p]
                        [p]
                            While it doesn't nearly manage to quench your thirst,
                            the lukewarm fluid sticks to your gums and to the walls of your esophagus,
                            slickening the back of your mouth, leaving a lingering, unpleasant yet somewhat [i]reinvigorating[/i] feeling.
                            At least the permeating [b]dryness[/b] onset by canine briefs is gone.
                            And, pleasantly enough, the underwear filters out most of the [i]fur[/i] and [i]grass[/i] contained within the goop.
                        [/p]
                        [p]
                            You notice Fenrir shifting around behind you.
                            Before you even have a chance to pull back and see what's up,
                            he painfully [b]kicks[/b] your rear, sending you flying forwards into the trough.
                            Your head momentarily dips underwater, spit-laced goop [b]flooding[/b] into your open mouth.
                            Luckily, the lupine's underwear prevents you from swallowing most of the vile stuff.
                            You resurface, spitting out as much of it as you can,
                            before turning your attention towards your captor, casting an angry [b]glare[/b] at him.
                        [/p]
                        [p]
                            "[color=gblue]I told you to [b]wash up[/b], not to [b]drink[/b].
                            You can be glad that I chose to kick your ass instead of [b]crushing[/b] your head.
                            Next time you disobey me-[/color]"
                            Suddenly, a booming voice interrupts the bandit.
                        [/p]`;
        } else {
            toReturn += `[p]
                            It's [b]thick[/b].
                            It's [b]warm[/b].
                            It's not even anything [b]remotely[/b] like water.
                            The top layer of the trough, in which your tongue dipped, is pure [b]equine saliva[/b].
                            You retch even [b]before[/b] you manage to get the slimey goop into your mouth.
                            As your tongue retracts into your muzzle, the horrid taste of animal drool spreads throughout your mouth.
                            You swallow the stuff as quickly as you can, but the taste lingers.
                        [/p]
                        [p]
                            Your stomach churns, barely able to hold its fresh contents.
                            But you're not nearly done.
                            Despite being outright [b]sickening[/b] in flavor, a lot warmer and far more [b]viscous[/b] than you would have liked,
                            the slop you licked up [i]did[/i] end up [b]refreshing[/b] your dry mouth.
                            You need [b]more[/b].
                            However much it [b]revolts[/b] you,
                            you need to [b]keep drinking[/b] until you've taken in enough to survive.
                        [/p]
                        [p]
                            Trying to think of [b]anything[/b] but the vomit-inducing taste of the feral spit,
                            you lap again and again at the layer of stale saliva,
                            taking more and more of it into your muzzle,
                            letting it pool up before swallowing entire mouthfuls at a time.
                            You almost [b]throw up[/b] on the spot, as you accidentally lap up a clump of fur,
                            immediately followed by some chewed grass.
                            Luckily, you manage to refrain from vomiting,
                            as that would only serve to make the trough's contents even [b]less[/b] bearable to drink.
                        [/p]
                        [p]
                            While it doesn't nearly manage to quench your thirst,
                            the lukewarm fluid sticks to your gums and to the walls of your esophagus,
                            slickening your mouth, leaving a lingering, unpleasant yet somewhat [i]refreshing[/i] feeling.
                            You'll undoubtedly have bad breath for days to come.
                        [/p]
                        [p]
                            You notice Fenrir shifting around behind you.
                            Before you even have a chance to pull back and see what's up,
                            he painfully [b]kicks[/b] your rear, sending you flying forwards into the trough.
                            Your head momentarily dips underwater, spit-laced goop [b]flooding[/b] into your open mouth.
                            Surprised, you swallow more than a little of the vile stuff,
                            including a thick wad of what you can only presume to be [b]animal phlegm[/b].
                            You resurface, spitting out as much of it as you can, retching and heaving
                            before turning your attention towards your captor, casting an angry [b]glare[/b] at him.
                        [/p]
                        [p]
                            "[color=gblue]I told you to [b]wash up[/b], not to [b]drink[/b].
                            You can be glad that I chose to kick your ass instead of [b]crushing[/b] your head.
                            Next time you disobey me-[/color]"
                            Suddenly, a booming voice interrupts the bandit.
                        [/p]`;
        }

        toReturn += `   [p]
                            [big][color=brown][b]Fenrir!![/b][/color][/big]
                        [/p]
                        [p]
                            It's the caravan master!
                            He's approaching the two of you rapidly.
                            "[color=brown]What the [b]hell[/b] are you doing to the slave?![/color]"
                            The chubby human crudely pushes the wolf to the side.
                            You immediately crawl out of the trough, rising to your feet and scraping as much of the spit off as possible.
                            "[color=gblue]Relax, I'm just getting it cleaned like you told me to.[/color]"
                            Your captor's response only serves to anger his [sneak]temporary[/sneak] superior further.
                            "[color=brown][b]Relax?![/b]
                            Half the bloody village is watching you!
                            We're trying to [b]sell[/b] to these people,
                            and you [b]kicking[/b] our [i]merchandise[/i] in front of them [b]isn't[/b] helping.[/color]"
                        [/p]
                        [p]
                            The wolf tries, but fails to come up with an apt response.
                            He sighs discontently, muttering a half-meant apology.
                            "[color=brown]That's more like it.
                            Now quit acting like the [b]animal[/b] you are, and go get dressed.[/color]"
                            After an intense stare-down, Fenrir finally backs off.
                            Audibly grumbling, he begins to make his way back to the wagons, leaving you alone with the caravan master.
                            The human harrumphs angrily.
                            "[color=brown]The baron will hear of this.[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\drink\SlaveNoonTwoOutsideDrinkBad.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\drink\SlaveNoonTwoOutsideDrink.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutsideDrink(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOutsideChoice.DRINK;
    var reqController = ctrl.outside;

    var toFail = new SlaveNoonTwoOutsideDrinkBad(ctrl);
    var toSuccess = new SlaveNoonTwoOutsideDrinkGood(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        if (!ctrl.outsideComplianceStorage.has()) {
            ctrl.outsideComplianceStorage.set(global.kobold.history.shared.enslavement.totalComplianceRating());
        }
        return ctrl.outsideComplianceStorage.get() > 8;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\drink\SlaveNoonTwoOutsideDrink.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\bathe\SlaveNoonTwoOutsideBatheGood.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutsideBatheGood(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preStage;
    var changeController = ctrl.batheGood;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.stage;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.goodBathe = true;
        var injected = ctrl.gag.has();
        var neither = ctrl.potion.get() === SlaveNoonTwoGoodPotionChoice.NEITHER;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var kissedBalls = global.kobold.history.shared.enslavement.noonTwo.kissedBalls || eager;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            Darnit.
                            Why couldn't the wolf have brought you to the river instead?!
                            Do you really have to wash up in the [b]middle[/b] of the hamlet, with all these people watching?
                            In a shallow [b]trough[/b] used by beasts of burden to slake their thirst?
                            You ${injected ? `sigh internally` : `let out another sigh`}.
                            You'd rather have some [b]privacy[/b] and the comfort of [b]running[/b] or at least [b]fresh[/b] water,
                            instead of this stale puddle of [b]animal spit[/b].
                        [/p]
                        [p]
                            But you don't get a say in the matter.
                            Either you wash up here, or not at all.
                            ${fucked ? `${eager ? `While you wear Fenrir's dried spunk with pride,
                            it would still be nice to clean off the rest of the dirt.` : `You're eager to rid yourself of
                            Fenrir's dried spunk.`}` : `It would be nice to clean off all the dirt from the chase.`}
                            Besides, it's a drinking trough.
                            You're probably [i]overreacting[/i].
                            There'll certainly be plenty of fresh, cool water underneath the filthy [i]top layer[/i].
                            Either way, you doubt the canine will change his mind on forcing you to bathe here,
                            especially considering the [i]auction[/i] is supposed to start soon.
                        [/p]
                        [p]
                            Reluctantly, you lift your foot off the ground, move it towards the trough, and dip it into the liquid.
                            The fluid feels [b]thick[/b], [b]warm[/b] and [b]sticky[/b], not just on the surface, but for at least an entire [b]inch[/b] deep.
                            Like a whole [b]layer[/b] of revolting animal drool coating your foot.
                            While appalled, you're in too far to back down now.
                            You press on, planting your sole on the wooden bottom of the trough.
                            At least the water is a [i]bit[/i] less [b]viscous[/b] and a hint [b]cooler[/b] near the bottom.
                        [/p]
                        [p]
                            Disgust visible on your face, you gingerly place your second foot into the makeshift tub.
                            A faint, unpleasant smell wafts forth from the fluid at your feet.
                            You look back at Fenrir, wordlessly pleading with him.
                            Do you [b]really[/b] have to do this?
                            The wolf's stern glare is unrelenting.
                            "[color=gblue]Hurry up. People are waiting.[/color]"
                        [/p]
                        [p]
                            With great hesitation, you lower your body into the water, your bare butt sinking through the slimy spit to the bottom of the trough.
                            You can feel the sickening feral saliva clinging to your loins, even when submerged into the clear part of the water,
                            along with more than a few mushy bits of half-digested grass.
                            Wanting to get this over with as fast as possible, if only to sate the wolf,
                            you dip your hands into the liquid, reaching as low as possible to collect as much of the [i]actual[/i] water as you can.
                            Still, you scoop up more [b]drool[/b] and [b]fur[/b] than anything else.
                        [/p]
                        [p]
                            Your stomach is churning, but you endure, smearing the slimy goop along your arms, [i]washing[/i] yourself with horse slobber.
                            You look behind you.
                            Fenrir seems pleased.
                            "[color=gblue]Carry on.
                            I'm going to leave you to it while I put on some clothes.
                            Try to run, if you [b]dare[/b].
                            We'll find you before you manage to reach the city [b]or[/b] the forest.
                            And when we do, the only thing we'll be selling'll be your [b]dried meat[/b].[/color]"
                        [/p]
                        [p]
                            With those words, the wolf turns tail, and casually begins to walk away.
                            Before he turns the corner back to the caravan, he lingers for a bit to stare at you, signalling for you to continue washing.
                            Dumbstruck, you gather more spit-water, and splatter it over your chest.
                            The saliva-ridden fluid is thick enough for you to [b]smear[/b] it over your body.
                            The act seems to satisfy the canine's lax suspicions.
                            He returns to the caravan, leaving you all to yourself, sitting naked in a watering trough in the middle of the hamlet.
                        [/p]
                        [p]
                            In a weird way, it's like Fenrir somewhat [i]trusts[/i] you not to try to run away or something.
                            Then again, he's right.
                            You have no way to escape.
                            If the farmers leering at you don't stop you themselves,
                            then they'll surely be able to tell the [b]bandits[/b] that'll be chasing after you where you went.
                        [/p]
                        [p]
                            Even [i]if[/i] you managed to make it past the farmers,
                            there's little cover between here and the city.
                            A few patches of woodland, none big enough to hide you for long.
                            Unless you fancy swimming across the wide river, there's only three ways to get back to the forest,
                            all three of which they'll undoubtedly guard.
                            If it were summer, at least you'd be able to conceal yourself in the growing grain.
                            But it's winter.
                            The fields are barren.
                        [/p]
                        [p]
                            You can't outrun them, as the wolf demonstrated yesterday, and there's nowhere to hide around here.
                            Fleeing would be a bad idea.
                            But that doesn't mean you have to sit here in this [i]slop[/i].
                            To make sure Fenrir can't say you didn't at least [i]try[/i] to clean yourself,
                            you slather some of the saliva-mixture onto your back, then you rise to your feet as quickly as you can, and you step out of the tub.
                        [/p]
                        [p]
                            Spit slowly oozes from your body, the viscous liquid inching its way down your scales, onto the ground.
                            To your surprise, it's actually quite [i]efficient[/i] at removing the dried up dirt${fucked ? `, dust and cum` : ` and dust`} from your body.
                            Even the bits of grass and fur get swept away.
                            The smell of the stuff, sadly, lingers on your scales, but it's mild enough to be passed off as [i]body odor[/i].
                            While it was one of the most unpleasant things you've ever done,
                            ${injected ? `surpassed only by taking the wolf's filthy underwear into your mouth,` : ``}
                            you've actually managed to get out of the slobber-rich drinking trough slightly more clean than you were before.
                        [/p]
                        [p]
                            Contemplating your next move, a familiar figure suddenly rushes towards you.
                            The caravan master.
                            "[color=brown]Ah, there you are, slave.
                            I was beginning to fear Fenrir had run off with you.
                            Where the hell is that mangy pooch anyway?
                            Did he leave you all alone?![/color]"
                            The human harrumphs angrily.
                            You don't dare interrupt him.
                            "[color=brown]The baron will hear of this.[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\bathe\SlaveNoonTwoOutsideBatheGood.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\bathe\SlaveNoonTwoOutsideBatheBad.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutsideBatheBad(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.preStage;
    var changeController = ctrl.batheBad;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.stage;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.badBathe = true;
        var injected = ctrl.gag.has();
        var neither = ctrl.potion.get() === SlaveNoonTwoGoodPotionChoice.NEITHER;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var kissedBalls = global.kobold.history.shared.enslavement.noonTwo.kissedBalls || eager;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            Judging by the look on Fenrir's face, you have a choice between either getting in the tub, or being [b]forced[/b] into it.
                            While you [b]really[/b] don't want to do this, if you [b]absolutely[/b] have to, then you'd prefer to do so on your own terms.
                        [/p]
                        [p]
                            Darnit.
                            Why couldn't the wolf have brought you to the river instead?!
                            Washing up in the [b]middle[/b] of the hamlet, with all these people watching, is more than a little degrading.
                            In a shallow [b]trough[/b] used by beasts of burden to slake their thirst, no less.
                            You ${injected ? `sigh internally` : `let out another sigh`}.
                            You'd rather have some [b]privacy[/b] and the comfort of [b]running[/b] or at least [b]fresh[/b] water,
                            instead of this stale puddle of [b]animal spit[/b].
                        [/p]
                        [p]
                            But you don't get a say in the matter.
                            ${fucked ? `${eager ? `While you wear Fenrir's dried spunk with pride,
                            it [i]would[/i] be kind of nice to clean off the rest of the dirt.` : `While you don't let it show,
                            you kind of appreciate that you get to 
                            rid yourself of Fenrir's dried spunk.`}` : `It's actually kind of nice of Fenrir,
                            to allow you to clean off all the dirt from the chase.`}
                            Besides, it's a drinking trough.
                            You're probably [i]overreacting[/i].
                            There'll certainly be plenty of fresh, cool water underneath the filthy [i]top layer[/i].
                            Either way, the canine looks about ready to [b]toss[/b] you in if you don't get moving.
                        [/p]
                        [p]
                            Reluctantly, you lift your foot off the ground, move it towards the trough, and dip it into the liquid.
                            The fluid feels [b]thick[/b], [b]warm[/b] and [b]sticky[/b], not just on the surface, but for at least an entire [b]inch[/b] deep.
                            Like a whole [b]layer[/b] of revolting animal drool coating your foot.
                            While appalled, you're in too far to back down now.
                            You press on, planting your sole on the wooden bottom of the trough.
                            At least the water is a [i]bit[/i] less [b]viscous[/b] and a hint [b]cooler[/b] near the bottom.
                        [/p]
                        [p]
                            Disgust visible on your face, you gingerly place your second foot into the makeshift tub.
                            A faint, unpleasant smell wafts forth from the fluid at your feet.
                            You look back at Fenrir, wordlessly pleading with him.
                            Do you [b]really[/b] have to do this?
                            The wolf's stern glare is unrelenting.
                            "[color=gblue]Hurry up, slave. People are waiting.[/color]"
                        [/p]
                        [p]
                            With great hesitation, you lower your body into the water, your bare butt sinking through the slimy spit to the bottom of the trough.
                            You can feel the sickening feral saliva clinging to your loins, even when submerged into the clear part of the water,
                            along with more than a few mushy bits of half-digested grass.
                            Wanting to get this over with as fast as possible, if only to sate the wolf,
                            you dip your hands into the liquid, reaching as low as possible to collect as much of the [i]actual[/i] water as you can.
                            Still, you scoop up more [b]drool[/b] and [b]fur[/b] than anything else.
                        [/p]
                        [p]
                            Your stomach is churning, but you endure, smearing the slimy goop along your arms, [i]washing[/i] yourself with horse slobber.
                            You look behind you,
                            only to find the wolf fondling his groin, enticing the tapered tip of his canine member from its sheath.
                            Is he [b]getting off[/b] to this?!
                            Outraged, you try to scramble to your feet,
                            but the big lupine places his free hand on your shoulder, easily keeping you pinned down.
                            "[color=gblue]Keep going, [b]slave[/b].
                            Don't make me ask again.[/color]"
                        [/p]
                        [p]
                            He squeezes your tender frame tightly, threatening to puncture your scales with his sharp claws.
                            Fearfully, you gather more spit-water, and splatter it over your chest.
                            The saliva-ridden fluid is thick enough for you to [b]smear[/b] it over your body.
                            "[color=gblue]That's it.
                            [b]Lather up.[/b]
                            I'll be able to [b]smell[/b] you from [b]miles[/b] away.[/color]"
                            Fenrir has begun to outright [b]stroke[/b] his length.
                            A droplet of precum drops directly from the tip into the trough, splashing down right behind you.
                        [/p]
                        [p]
                            What kind of [b]freak[/b] would even jerk off to a kobold bathing in spit?!
                            What's he getting a rise out of?
                            Is it the power he holds over you?
                            The control?
                            The washing itself?
                            The humiliation?
                            Or the [b]stench[/b]?
                            The canine growls loudly.
                            His claws dig deep enough to almost draw blood.
                            "[color=gblue]I said: [b]keep going[/b].[/color]"
                            Almost reduced to tears,
                            you half-heartedly slather some of the saliva-mixture onto your back.
                            It is quickly joined by another squirt of pre, this time hitting the back of your neck.
                            Suddenly, a booming voice rings out from behind, saving you from further indignation at the hands of the lupine.
                        [/p]
                        [p]
                            [big][color=brown][b]Fenrir!![/b][/color][/big]
                        [/p]
                        [p]
                            It's the caravan master!
                            He's approaching the two of you rapidly.
                            "[color=brown]What the [b]hell[/b] are you doing to the slave?![/color]"
                            The chubby human crudely pushes the wolf to the side.
                            You immediately crawl out of the trough, rising to your feet and scraping as much of the spit off as possible.
                            "[color=gblue]Relax, I'm just having some [b]fun[/b].[/color]"
                            Your captor's response only serves to anger his [sneak]temporary[/sneak] superior further.
                            "[color=brown][b]Fun?![/b]
                            Half the bloody village is watching you!
                            We're trying to [b]sell[/b] to these people, and you jerking your dick in front of them [b]isn't[/b] helping.[/color]"
                        [/p]
                        [p]
                            The wolf tries, but fails to come up with an apt response.
                            Sighing discontently, he lets go of his half-erect cock.
                            "[color=brown]That's more like it.
                            Now quit acting like the [b]animal[/b] you are, and go get dressed.[/color]"
                            After an intense stare-down, Fenrir finally backs off.
                            Audibly grumbling, he begins to make his way back to the wagons, leaving you alone with the caravan master.
                            The human harrumphs angrily.
                            "[color=brown]The baron will hear of this.[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\bathe\SlaveNoonTwoOutsideBatheBad.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\bathe\SlaveNoonTwoOutsideBathe.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutsideBathe(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOutsideChoice.BATHE;
    var reqController = ctrl.outside;

    var toFail = new SlaveNoonTwoOutsideBatheBad(ctrl);
    var toSuccess = new SlaveNoonTwoOutsideBatheGood(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        if (!ctrl.outsideComplianceStorage.has()) {
            ctrl.outsideComplianceStorage.set(global.kobold.history.shared.enslavement.totalComplianceRating());
        }
        return ctrl.outsideComplianceStorage.get() > 8;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\bathe\SlaveNoonTwoOutsideBathe.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\SlaveNoonTwoOutside.js

/**
 * Outside section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoOutside(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preOutside;
    var subController = ctrl.outside;
    var changeController = ctrl.outside;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new SlaveNoonTwoOutsideRefuse(ctrl);
    var secondNext = new SlaveNoonTwoOutsideDrink(ctrl);
    var thirdNext = new SlaveNoonTwoOutsideBathe(ctrl);

    var firstChoice = SlaveNoonTwoOutsideChoice.REFUSE;
    var secondChoice = SlaveNoonTwoOutsideChoice.DRINK;
    var thirdChoice = SlaveNoonTwoOutsideChoice.BATHE;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var injected = ctrl.gag.has();
        var neither = ctrl.potion.get() === SlaveNoonTwoGoodPotionChoice.NEITHER;
        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var kissedBalls = global.kobold.history.shared.enslavement.noonTwo.kissedBalls || eager;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            With your body ${injected ? `forcibly altered` : `${neither ? `left unaltered` : `changed for the better`}`},
                            Fenrir takes hold of the collar around your neck.
                            His [i]slave[/i] in tow, he proudly strides away from the caravan, into town.
                            While not a whole lot of people call the hamlet home, there's still quite a few farmers loitering about.
                            Probably getting some lunch, or just taking a break from work for a bit.
                            Perhaps they've [i]heard[/i] of the [b]slave auction[/b] about to take place.
                            The moment the two of you come into view, all of their eyes instantly shift towards the bandit and his prize.
                        [/p]
                        [p]
                            You can't tell who's drawing more attention: the naked wolf or you yourself.
                            While you'd [b]hoped[/b] that the both of you being nude would somewhat [b]lessen[/b] your shame,
                            it has quite the opposite effect.
                            It's as if the canine is broadcasting his ownership to the entire village,
                            strutting about the fact that
                            ${fucked ? `he fucked you yesterday` : `${kissedBalls ? `you just kissed his nuts` : `he's about to sell you`}`}.
                            ${eager ? `You can't help but admire his dominant attitude.` : `You [b]hate[/b] how he thinks he [b]owns[/b] you.`}
                        [/p]
                        [p]
                            Despite Fenrir catching some of their glares,
                            you can't help but feel the farmers' eyes staring at you, inpecting your body,
                            ${eager ? `envisioning you in every lewd situation imaginable` : `judging you ability to work as their maid,
                            their cook or their farmhand`}.
                            Being [b]paraded[/b] around like this is shameful, to say the least${eager ? `, but also kind of arousing` : ``}.
                            You can't help but wonder what other kind of ${eager ? `alluring` : `degrading`} things the lupine has in store for you.
                        [/p]
                        [p]
                            Walking over to one of the buildings, your captor suddenly comes to a halt.
                            Stepping besides him, you notice he's stopped in front of a wide, long wooden animal drinking trough.
                            You can immediately tell it's been [b]thoroughly[/b] used, [b]very[/b] recently.
                            The water is slightly murky, bits of fur and leaves of chewed grass floating at the surface, alongside a great amount of undiluted [b]spit[/b],
                            so much of the stuff that you're starting to wonder how [i]little[/i] of the liquid left is actual [b]water[/b].
                            But then again, the feral saliva might just be a [i]surface layer[/i] of sorts.
                        [/p]
                        [p]
                            "[color=gblue]Here we are, ${injected ? `slave` : name}.
                            Time to [b]wash up[/b].
                            You need to look presentable.
                            ${fucked ? `Can't have you go on stage with cum splatters on your thighs` : `
                            Can't have you go on stage all dirty from running through the forest the other day`}.[/color]"
                            Stage?
                            Casting your gaze to the crossroads at the center of the hamlet,
                            you notice a small wooden podium built right besides the well.
                            ${eager ? `` : `[i]Oh no[/i].`}
                            They're going to make a [b]big deal[/b] out of this, aren't they?
                        [/p]
                        [p]
                            ${eager ? `` : `You sigh loudly.`}
                            ${injected ? `Your mouth still gagged, you give Fenrir a small nod,
                            followed by an inquisitive gaze.` : `You tell Fenrir that you'll gladly accept the opportunity to wash up.`}
                            Did he bring a tub, or is he going to draw water from the well?
                            "[color=gblue]${injected ? `What are you waiting for?
                            Get in.` : `A tub?
                            You're staring at it, dumbass.`}[/color]"
                            Wait, the trough?
                            He has to be kidding.
                            "[color=gblue]If it's good enough for our pack animals to drink from,
                            then it's also good enough for a kobold like you to clean yourself in.
                            Now get started.
                            There's only a couple of minutes left before the auction's supposed to begin.[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Refuse</button>`;

        var tooltip = global.parser.parse(`You are [b]not[/b] going to wash up in a trough full of [b]slobber[/b].`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 205px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Drink</button>`;

        var tooltip = global.parser.parse(`It may be filthy, but it's the only water you've seen in more than a day.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>Bathe</button>`;

        var tooltip = global.parser.parse(`Disgusting or not, you need to be clean if you want to get [i]bought[/i].`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var SlaveNoonTwoOutsideChoice = {
    REFUSE: 1,
    DRINK: 2,
    BATHE: 3
};

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\outside\SlaveNoonTwoOutside.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\inject\SlaveNoonTwoBadInject.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadInject(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preInject;
    var subController = ctrl.preOutside;
    var changeController = ctrl.inject;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.outside;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.injected = true;
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            Fenrir lets you get [i]acquainted[/i] with your new gag for a minute more,
                            taking in the disgusted look on your face, ensuring he'll remember this moment for the rest of his life.
                            You know [b]you[/b] will.
                            How could you ever forget a taste this [b]foul[/b]?
                            You get the feeling you'll be [b]wishing[/b] you could in the future.
                        [/p]
                        [p]
                            His sadistic voyeurism sated,
                            the wolf crudely grabs you by the wrist, and begins to tug you towards the cart's open entryway.
                            "[color=gblue]While I could watch you [b]pout[/b] all day long,
                            you and I have business to attend to.
                            In case you hadn't realized yet: you're being [b]sold[/b].
                            [b]Today[/b].
                            And it's [b]my[/b] job to ensure you look at least [i]somewhat[/i] presentable.[/color]"
                        [/p]
                        [p]
                            He's not taking [i]no[/i] for an answer,
                            yanking you forwards hard enough to nearly pull your arm off.
                            Already having been [i]punished[/i] once, you don't resist much,
                            lest he decide to shove a sock up your ass next.
                            A few steps and a small hop later, you find yourself in the outside world once more.
                        [/p]`;

        if (ctrl.arrival.get() !== SlaveNoonTwoArrivalChoice.LOOKOUT) {
            toReturn += `[p]
                            You recognize this area.
                            You've been here before.
                            It's a place known simply as [i]the hamlet[/i],
                            a small village close to the human capital of Varanar.
                        [/p]
                        [p]
                            An old friend of yours lives here.
                            Seki, the alchemist-owl, who also happens to happily buy whatever stolen goods you bring him.
                            Besides being your [i]fence[/i], he's also the closest thing to a [i]father[/i] you've ever had.
                            Maybe [b]he[/b] can help you get out of this [b]mess[/b].
                        [/p]`;
        }

        toReturn += `   [p]
                            Luckily, the caravan stopped at the outskirts of the settlement,
                            your nudity and improper gag not made public quite yet.
                            And, if it's any [i]consolation[/i], your captor is every bit as naked as you are.
                            Walking alongside your prison carriage's exterior,
                            whilst dragging you with him, the canine comes to an unexpected halt,
                            your face ending up unpleasantly [i]bumping[/i] into his bare rear.
                            Reaching down, he fetches a small, brown satchel, conveniently concealed directly underneath the cart.
                        [/p]
                        [p]
                            "[color=gblue]I almost forgot.
                            I [i]had[/i] prepared a treat for you,
                            should you have chosen to [b]behave[/b].
                            But you were a bad ${dick ? `boy` : `girl`}.
                            And bad ${dick ? `boy` : `girl`}s don't [b]deserve[/b] any treats.[/color]"
                        [/p]
                        [p]
                            Rummaging through the tiny bag, Fenrir produces a pair of oddly shaped bottles, along with a large, ominous-looking syringe.
                            ${global.kobold.inventory.createItemWithTooltip(new FemPlus(), true)}
                            and ${global.kobold.inventory.createItemWithTooltip(new mPlus(), true)}.
                            Seki makes that sort of stuff.
                            "[color=gblue]It feels like such a [b]waste[/b] to let these go unused, though.
                            What do you think, slave?
                            Should I pick one for you?[/color]"
                        [/p]
                        [p]
                            The wolf wants to [b]inject[/b] you with body-altering [b]drugs[/b]?!
                            No way!
                            Knowing him, he'll probably pick the [b]worst one[/b]!
                            You wildly shake your head no, while audibly complaining through your underwear-gag.
                            Your muffled shouts for him to put the vials back in the bag seem to have persuaded the lupine.
                        [/p]
                        [p]
                            "[color=gblue]Alright, alright.
                            I'll put'em back.[/color]"
                            Fenrir turns his rear to you,
                            seemingly messing around with the fragile, glass containers.
                            Probably making sure they don't [i]break[/i] while returning them to the satchel.
                        [/p]
                        [p]
                            Suddenly, your captor turns to face you once more.
                            In one of his furred hands: a filled syringe.
                            Before you even know what's going on,
                            the burly canine [b]rams[/b] the needle into your upper arm,
                            the sharp piece of hollow metal [b]piercing[/b] your scales with ease.
                            While firmly holding you in place with his free hand,
                            he begins to lower the plunger with his thumb,
                            forcing the syringe's contents into your veins.
                        [/p]
                        [p]
                            You yelp out in pain, immediately starting to struggle and squirm,
                            trying your best to get away from the bandit's grasp [b]before[/b] he can inject the entirety of the potion into your body.
                            "[color=gblue]Stand [b]the fuck[/b] still!
                            Do you [b]want[/b] the needle to [b]break off[/b]?![/color]"
                            You fail to heed the wicked man's warnings, continuing your fierce resistance,
                            yet his grasp is far stronger than every muscle in your body combined.
                        [/p]`;

        if (ctrl.misc.dick()) {
            global.kobold.history.shared.enslavement.noonTwo.tookMPlus = true;
            toReturn += `[p]
                            Fenrir easily manages to keep you put while administering the drug.
                            You can already feel its effects starting to kick in.
                            Your members stirs within your loins, tangibly growing larger,
                            your cockslit becoming a lot more [b]cramped[/b].
                            Luckily, as the last of the liquid is squirted into your bloodstream, the effects swiftly come to a halt.
                        [/p]
                        [p]
                            [c]
                                [b]You have grown more masculine.[/b]
                            [/c]
                        [/p]
                        [p]
                            "[color=gblue]There you go.[/color]" the wolf speaks, retracting the needle from your arm,
                            before tossing it to the ground, breaking it.
                            You glare at him, with all the hatred and anger that you can muster.
                            He [b]changed[/b] your [b]body[/b] without your consent.
                            A [b]major[/b] crime, [i]if[/i] kobolds -- or slaves, for that matter -- had [i]any[/i] legal rights.
                        [/p]
                        [p]
                            "[color=gblue]I figured you could stand to be a bit [i]bigger[/i] where it counts.
                            It'll make you more attractive to the ladies shopping for a little lizard of their own.
                            And before you get all [b]uppity[/b] about not getting your way, remember:
                            [b]you[/b] are the [b]slave[/b] here.
                            I am your [b]master[/b].
                            [b]I[/b] decide what is good for you, and what is not.
                            And [b]your[/b] opinion on the matter is [b]irrelevant[/b].
                            You [b]live[/b] to [b]serve[/b] me, after all.[/color]"
                        [/p]`;
        } else {
            global.kobold.history.shared.enslavement.noonTwo.tookFemPlus = true;
            toReturn += `[p]
                            Fenrir easily manages to keep you put while administering the drug.
                            You can already feel its effects starting to kick in.
                            Your breasts tingle and throb, visibly growing bigger,
                            your upper chest feeling heavier with every passing second.
                            Luckily, as the last of the liquid is squirted into your bloodstream, the effects swiftly come to a halt,
                            leaving your bare tits a single size larger than what they used to be.
                        [/p]
                         [p]
                            [c]
                                [b]You have grown more feminine.[/b]
                            [/c]
                        [/p]
                        [p]
                            "[color=gblue]There you go.[/color]" the wolf speaks, retracting the needle from your arm,
                            before tossing it to the ground, breaking it.
                            You glare at him, with all the hatred and anger that you can muster.
                            He [b]changed[/b] your [b]body[/b] without your consent.
                            A [b]major[/b] crime, [i]if[/i] kobolds -- or slaves, for that matter -- had [i]any[/i] legal rights.
                        [/p]
                        [p]
                            "[color=gblue]I figured you could stand to be a bit [i]bigger[/i] where it counts.
                            It'll make you more attractive to the men shopping for a little lizard of their own.
                            And before you get all [b]uppity[/b] about not being allowed to pick your poison, remember:
                            [b]you[/b] are the [b]slave[/b] here.
                            I am your [b]master[/b].
                            [b]I[/b] decide what is good for you, and what is not.
                            And [b]your[/b] opinion on the matter is [b]irrelevant[/b].
                            You [b]live[/b] to [b]serve[/b] me, after all.[/color]"
                        [/p]`;
        }

        toReturn += `[p]
                        You hate this.
                        You hate [b]everything[/b] about this.
                        How your body is being warped to better suit Fenrir's [i]needs[/i],
                        how he's acting like he's [b]already[/b] your [i]master[/i],
                        how he thinks altering your body would be okay, even [b]if[/b] you were his slave,
                        it all [b]aggrevates[/b] you to no end!
                        And to make matters even worse,
                        you can't even [b]voice[/b] your disdain for what he's doing, due to the sickening [b]worn underpants[/b] gagging your mouth.
                    [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\inject\SlaveNoonTwoBadInject.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\gag\SlaveNoonTwoBadGagOpen.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadGagOpen(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoBadGagChoice.OPEN;
    var reqController = ctrl.gag;
    var subController = ctrl.preInject;
    var changeController = ctrl.gagOpen;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new SlaveNoonTwoBadInject(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.gagOpened = true;
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            While you loathe the sickening garment being dangled in front of your face,
                            there's no point in resisting what is about to happen.
                            The big wolf would force your mouth open with ease,
                            and ram it in regardless of how much you struggle.
                        [/p]
                        [p]
                            Besides, this is a matter of [b]pride[/b].
                            If you refuse to take your punishment, you'll look like a [b]child[/b].
                            But if you do as he asks, and reluctantly take it like an adult,
                            you'll come off looking a lot better.
                            And you think Fenrir, despite your disagreements, will be able to appreciate that.
                        [/p]
                        [p]
                            Glaring at the canine with hatred in your eyes, you slowly open your muzzle, spreading your jaws as far apart as they'll go.
                            Your captor seems pleasantly surprised.
                            A big grin forms on his face.
                            "[color=gblue]Good ${dick ? `boy` : `girl`}.
                            Suddenly growing obedient, are we?
                            Or maybe you just [b]wanted[/b] to suckle the sweat out of my week-old underwear from the get-go.[/color]"
                        [/p]
                        [p]
                            You resist the urge to tell Fenrir to go fuck himself.
                            It takes every ounce of willpower you can muster [b]not[/b] to go back on your decision to play along.
                            Deciding not to waste any more time with idle banter,
                            the wolf slowly moves the disgusting piece of cloth closer and closer to your mouth, as if to gauge your reaction.
                            A sparkle of hope remaining that this is all just a [b]test[/b],
                            a [b]bluff[/b] to see whether or not you'd try to bite him or something,
                            you remain completely motionless.
                        [/p]
                        [p]
                            As the first bit of the foul garment gets pressed into your open muzzle,
                            you realize this is [b]not[/b] a joke or a trial.
                            Fenrir is actually [b]serious[/b] about gagging you with his rancid underpants!
                            The coarse fabric drags along your tongue, leaving a trail of revolting filth in its wake,
                            its gut-wrenching, absolutely [b]revolting[/b] flavor spread across your tastebuds.
                            You thought the [b]smell[/b] was bad, but the taste is [b]ten times[/b] worse.
                        [/p]
                        [p]
                            On the verge of throwing up, you attempt to close your mouth,
                            to stop the wolf from taking this any further, but it's too late.
                            Too much of his underwear is already in your muzzle, blocking your jaws, preventing you from closing them.
                            And he isn't done yet.
                            The psychopathic bandit pushes his ass-and-crotch-sweat-drenched briefs deeper and deeper into your maw,
                            cramming the entirety of the repulsive loincloth down your face.
                        [/p]
                        [p]
                            Bits of dirt and grime mix with spit at the back of your mouth, rapidly spreading throughout the entirety of your muzzle,
                            tainting every nook and cranny that the unsanitary garment itself failed to reach.
                            Various unidentifiable specks of crud stick to your teeth and to your gums,
                            ensuring the vile taste will linger for [b]hours[/b], if not [b]days[/b] to come,
                            even [b]if[/b] your [i]gag[/i] is removed.
                        [/p]
                        [p]
                            A small patch of filthy grey fur parts from the fabric it was contained in for over a [b]week[/b],
                            only to agonizingly find its way down your throat instead,
                            unpleasantly scraping against the inner walls of your esophagus for its entire trip into your stomach.
                            Luckily, you haven't eaten in quite a while.
                            Otherwise, vomit would join the mixture of nauseating flavors and textures trapped within your muzzle.
                        [/p]
                        [p]
                            The more the underwear gets drenched in your spit,
                            the more prominent the horrifying taste of what you can only assume to be [b]wolf ass[/b] gets.
                            The large amount of stale sweat contained within the cloth is slowly being [b]rinsed out[/b] by your saliva,
                            causing a constant trickle of pungent, age-old canine perspiration to ooze onto your tongue,
                            along with even [b]more[/b] ancient filth, buried and entangled deep within the textile.
                            Has the garment [b]ever[/b] been properly washed?
                            You're starting to think not.
                        [/p]
                        [p]
                            Satisfied with his work, Fenrir retracts his hand from your maw,
                            while using the other to forcibly shut your jaws as best he can,
                            his wretched loincloth preventing them from closing fully.
                            Keeping your mouth firmly sealed over the horrid piece of clothing,
                            he deftly ties his piece of rope around your muzzle, completing your makeshift
                            ${global.kobold.inventory.createItemWithTooltip(new UnderwearGag())}.
                            With a cruel chuckle, the lupine takes a step back to admire his work.
                        [/p]
                        [p]
                            You try your best to ignore Fenrir, and instead focus on attempting to cope with your new [i]restraints[/i].
                            While you manage to force some of the grime out through the front and the sides of your muzzle using your flexible tongue,
                            a lot of it lingers in the back of your mouth, blocked by the big pair of underpants.
                            It can't keep piling up forever, and it has only one way to go.
                            Tears flowing liberally from your eyes,
                            you reluctantly swallow the disgusting, goopy build-up of mixed saliva, sweat and filth,
                            undoubtedly only the first of many such loads before the [i]gag[/i] is removed.
                            You only barely manage to hold it down.
                        [/p]
                        [p]
                            "[color=gblue][b]Hey![/b][/color]" the wolf draws your attention.
                            "[color=gblue]Listen up, [b]slave[/b].
                            You touch that rope around your maw, and you [b]die[/b].
                            Got it?[/color]"
                            Still not able to fully cope with the foul flavour, you half-heartedly give the canine a small nod.
                            You're hardly even listening to him.
                            "[color=gblue]If you're having this much trouble [i]enduring[/i] my underwear,
                            then I hope for your sake that someone's willing to buy your lousy ass.
                            Because if I get my hands on you, a bit of sweat and dirt will be the [b]least[/b] of your concerns.[/color]"
                        [/p]
                        [p]
                            You doubt there's even [b]worse[/b] things he can inflict on you.
                            But then again, this is Fenrir we're talking about.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\gag\SlaveNoonTwoBadGagOpen.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\gag\SlaveNoonTwoBadGagRefuse.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadGagRefuse(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoBadGagChoice.REFUSE;
    var reqController = ctrl.gag;
    var subController = ctrl.preInject;
    var changeController = ctrl.gagRefuse;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = new SlaveNoonTwoBadInject(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            You [b]loathe[/b] the sickening garment being dangled in front of your face.
                            There's [b]no way[/b] you're going to let the big wolf shove that thing inside of your mouth.
                            If he wants to get it in, then he's welcome to [b]try[/b].
                            You glare defiantly at Fenrir, your lips firmly sealed.
                            [i]It's on[/i].
                        [/p]
                        [p]
                            "[color=gblue]Why do you have to make things so [b]hard[/b] on yourself?[/color]"
                            comes the canine's melodramatic question.
                            You bluntly tell him to go [b]fuck himself[/b].
                            He doesn't seem to like that.
                            His free hand curls up, and in the flash of an eye, Fenrir [b]rams[/b] his fist into your side,
                            [b]pummeling[/b] your lower back, delivering a fast and [i]cheap[/i] kidney-shot,
                            painful enough to make you [b]shriek[/b] out in agony.
                        [/p]
                        [p]
                            The bandit leader takes advantage of your mouth opening to scream,
                            quickly [b]forcing[/b] the foul garment into your mouth, muffling your pathetic cry,
                            silencing you with his unwashed underpants.
                            The coarse fabric drags along your tongue, leaving a trail of revolting filth in its wake,
                            its gut-wrenching, absolutely [b]revolting[/b] flavor spread across your tastebuds.
                            You thought the [b]smell[/b] was bad, but the taste is [b]ten times[/b] worse.
                            Combined with the pain and the indignation, the sensation is almost strong enough to make you lose consciousness.
                        [/p]
                        [p]
                            On the verge of throwing up, you attempt to close your mouth, to [b]bite down[/b] on Fenrir's hand, but it's too late.
                            Too much of his underwear is already in your muzzle, blocking your jaws, preventing you from closing them.
                            And he isn't done yet.
                            The psychopathic bandit pushes his ass-and-crotch-sweat-drenched briefs deeper and deeper into your maw,
                            cramming the entirety of the repulsive loincloth down your face, making you [b]wish[/b] you actually did pass out.
                        [/p]
                        [p]
                            Bits of dirt and grime mix with spit at the back of your mouth, rapidly spreading throughout the entirety of your muzzle,
                            tainting every nook and cranny that the unsanitary garment itself failed to reach.
                            Various unidentifiable specks of crud stick to your teeth and to your gums,
                            ensuring the vile taste will linger for [b]hours[/b], if not [b]days[/b] to come,
                            even [b]if[/b] your [i]gag[/i] is removed.
                        [/p]
                        [p]
                            A small patch of filthy grey fur parts from the fabric it was contained in for over a [b]week[/b],
                            only to agonizingly find its way down your throat instead,
                            unpleasantly scraping against the inner walls of your esophagus for its entire trip into your stomach.
                            Luckily, you haven't eaten in quite a while.
                            Otherwise, vomit would join the mixture of nauseating flavors and textures trapped within your muzzle.
                        [/p]
                        [p]
                            The more the underwear gets drenched in your spit,
                            the more prominent the horrifying taste of what you can only assume to be [b]wolf ass[/b] gets.
                            The large amount of stale sweat contained within the cloth is slowly being [b]rinsed out[/b] by your saliva,
                            causing a constant trickle of pungent, age-old canine perspiration to ooze onto your tongue,
                            along with even [b]more[/b] ancient filth, buried and entangled deep within the textile.
                            Has the garment [b]ever[/b] been properly washed?
                            You're starting to think not.
                        [/p]
                        [p]
                            Satisfied with his work, Fenrir retracts his hand from your maw,
                            while using the other to forcibly shut your jaws as best he can,
                            his wretched loincloth preventing them from closing fully.
                            Keeping your mouth firmly sealed over the horrid piece of clothing,
                            he deftly ties his piece of rope around your muzzle, completing your makeshift
                            ${global.kobold.inventory.createItemWithTooltip(new UnderwearGag())}.
                            With a cruel chuckle, the lupine takes a step back to admire his work.
                        [/p]
                        [p]
                            You try your best to ignore Fenrir, and instead focus on attempting to cope with your new [i]restraints[/i].
                            While you manage to force some of the grime out through the front and the sides of your muzzle using your flexible tongue,
                            a lot of it lingers in the back of your mouth, blocked by the big pair of underpants.
                            It can't keep piling up forever, and it has only one way to go.
                            Tears flowing liberally from your eyes,
                            you reluctantly swallow the disgusting, goopy build-up of mixed saliva, sweat and filth,
                            undoubtedly only the first of many such loads before the gag is removed.
                            You only barely manage to hold it down.
                        [/p]
                        [p]
                            "[color=gblue][b]Hey![/b][/color]" the wolf draws your attention.
                            "[color=gblue]Listen up, [b]slave[/b].
                            You touch that rope around your maw, and you [b]die[/b].
                            Got it?[/color]"
                            Still not able to fully cope with the foul flavour, you half-heartedly give the canine a small nod.
                            You're hardly even listening to him.
                            "[color=gblue]If you're having this much trouble [i]enduring[/i] my underwear,
                            then I hope for your sake that someone's willing to buy your lousy ass.
                            Because if I get my hands on you, a bit of sweat and dirt will be the [b]least[/b] of your concerns.[/color]"
                        [/p]
                        [p]
                            You doubt there's even [b]worse[/b] things he can inflict on you.
                            But then again, this is Fenrir we're talking about.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\gag\SlaveNoonTwoBadGagRefuse.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\gag\SlaveNoonTwoBadGag.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadGag(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preGag;
    var subController = ctrl.gag;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new SlaveNoonTwoBadGagRefuse(ctrl);
    var secondNext = new SlaveNoonTwoBadGagOpen(ctrl);

    var firstChoice = SlaveNoonTwoBadGagChoice.REFUSE;
    var secondChoice = SlaveNoonTwoBadGagChoice.OPEN;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.gagged = true;
        var name = ctrl.misc.name();
        var declined = ctrl.blackmail.get() === SlaveNoonTwoOfferChoice.DECLINE;
        var toReturn = `[p]
                            "[color=gblue]Well, ${name}, I'm nothing if not a man of my word.[/color]"
                            The wolf ominously states, taking the piece of rope in one hand, and his underwear in the other.
                            You take a${declined ? `nother` : ``} small step backwards, ending up with your back pressed against the far-end of your cell.
                            Trapped between hard, unyielding, splintery wood and a maniacal canine intent on shoving his dirty laundry down your throat.
                            "[color=gblue]Remember: [b]you[/b] chose this path.
                            Not me.[/color]"
                        [/p]
                        [p]
                            Panic wells up inside of you as Fenrir bridges the gap between the two of you.
                            You consider fleeing, but you doubt you'd get far.
                            Even [i]if[/i] you managed to get past the lupine somehow,
                            either him or his gang would [b]hunt you down[/b] a second time.
                            And you don't think they'll make the mistake of trying to [i]sell[/i] you [b]twice[/b].
                        [/p]
                        [p]
                            Fearing for your very life, you stay still as the big bandit holds his disgusting linen underpants in front of your face.
                            He seems to be getting a cruel kick out of letting you take a [b]good look[/b] at the vile thing he's about to force into your mouth.
                            His briefs are visibly worn, small scraps of fur and many thick, grey hairs intertwined with the sloppily stitched-together fabric.
                            They [b]reek[/b] of stale sweat, and there are many bits of [b]filth[/b] bunched up around the leg-holes and the waistband.
                        [/p]
                        [p]
                            The seams are partially [i]discolored[/i] to a revolting, slightly yellowish tint.
                            And that's just what you can [b]see[/b].
                            You can only [b]imagine[/b] how dirty the insides must be.
                            The [b]stench[/b] alone is making your eyes water.
                            He was wearing this for [b]far[/b] longer than yesterday alone, that much is [b]very[/b] clear.
                            For the entire journey, perhaps, if not even longer.
                            You almost begin to [b]retch[/b], but manage to retain your composure through pride,
                            determination and the [b]will[/b] not to show weakness in front of Fenrir.
                        [/p]
                        [p]
                            "[color=gblue]Don't make this worse for yourself. Open your muzzle.[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Refuse</button>`;

        var tooltip = global.parser.parse(`Fenrir can't gag you if you don't open your mouth.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Open up</button>`;

        var tooltip = global.parser.parse(`Don't bother to resist.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var SlaveNoonTwoBadGagChoice = {
    REFUSE: 1,
    OPEN: 2
};

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\gag\SlaveNoonTwoBadGag.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\eager\SlaveNoonTwoBadEager.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadEager(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.EAGER;
    var reqController = ctrl.blackmail;
    var subController = ctrl.preGag;
    var changeController = ctrl.badEager;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.notGagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager = true;
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, there'll be [b]hell[/b] to pay.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll undoubtedly lead to him asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something.
                            And for some reason, that kind of [i]arouses[/i] you.
                        [/p]
                        [p]
                            You never had a problem with Fenrir.
                            It was the caravan master who was acting like a real [b]prick[/b] last night.
                            While you don't want to get [b]killed[/b] by his [i]Furskin Mountain-Crew[/i],
                            as the canine threatened might happen,
                            you figure life with them could still turn out to be quite [i]nice[/i],
                            as long you serve them [b]properly[/b].
                            Surely they'll have no reason to hurt you if you obey their every wish, right?
                        [/p]
                        [p]
                            Becoming the big wolf's [b]plaything[/b] is an alluring prospect by itself.
                            Combined with the equally enticing big, furred orbs swinging underneath his plump sheath,
                            Fenrir's offer is practically [b]irresistible[/b] to you.
                            How could [b]anyone[/b] be faced with such a [b]hunk[/b] of a man,
                            and [b]not[/b] want to submit to him completely?
                            Should you end up not being [i]bought[/i],
                            then you'll get to live a comfortable life with your new lupine [i]master[/i],
                            hopefully far away from the chubby human that accosted you yesterday.
                        [/p]
                        [p]
                            Noticing the bandit growing impatient,
                            you lustily lick your lips and begin to inch your face closer to his balls.
                            The strong scent of canine masculinity intensifies,
                            male pheromones [b]wafting[/b] liberally from Fenrir's loins,
                            [b]clogging[/b] your nostrils, [b]saturating[/b] every breath you take.
                            You greedily huff the olfactory nectar, savoring every lungful,
                            wishing you could keep your face buried in his balls [b]forever[/b].
                        [/p]
                        [p]
                            While your own need is rising, you aren't about to get [i]distracted[/i] by it.
                            You set out to [i]please[/i] your master, and you intend to [b]exceed[/b] his every expectation.
                            As your snout sinks into the grey fur covering the wolf's sack,
                            you pucker your lips, and eagerly begin to plant kiss after kiss atop of his nuts.
                            Taking one of his balls into your mouth slightly, you gently suckle on it, just a little.
                            He groans in delight, his tense, authorative stand visibly relaxing.
                            [b]Excellent[/b].
                        [/p]
                        [p]
                            With Fenrir starting to mellow out, you up the intensity of your loving [b]admiration[/b].
                            Your tongue snakes out of your muzzle, your rampant kissing interrupted by occasional [b]licks[/b].
                            Before long, you find yourself outright [b]lapping[/b] along the coarse surface of the canine's sack,
                            wetting his fur with your saliva, enjoying the salty,
                            undeniably [b]masculine[/b] flavour even [b]despite[/b] the dried up sweat and the small bits of dirt and grime.
                        [/p]
                        [p]
                            The wolf begins to pant, his tail wagging excitedly.
                            Slowly, but surely, you work your way upwards, towards his sheath.
                            Right as you were about to goad his member from its slumber by kissing its resting place directly,
                            Fenrir suddenly presses his palm against your forehead, softly pushing you away.
                        [/p]
                        [p]
                            "[color=gblue]That's enough for now, you little [b]cocktease[/b].
                            Quite the apology.
                            If you were trying to impress me, it's a job well done.
                            Continue to act like that, and I might need to consider buying you myself,
                            to make sure nobody else gets to run off with that lovely muzzle of yours.[/color]"
                            Your captor takes a deep breath, seemingly washing the impure thoughts from his mind.
                            "[color=gblue]There'll be plenty of time for me to fuck your face later.
                            And since you like my balls so much,
                            you'll get to [b]worship[/b] them every single day, once you're finally [b]mine[/b].[/color]"
                            You can hardly wait.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\eager\SlaveNoonTwoBadEager.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\reluctant\SlaveNoonTwoBadReluctant.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadReluctant(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.RELUCTANT;
    var reqController = ctrl.blackmail;
    var subController = ctrl.preGag;
    var changeController = ctrl.badReluctant;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.notGagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.kissedBalls = true;
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, there'll be [b]hell[/b] to pay.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll just lead to Fenrir asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            And while you [b]hate[/b] the idea of becoming his [i]sex toy[/i],
                            right now, relenting seems like a better option than getting his [b]disgusting underpants[/b] lodged in your mouth.
                        [/p]
                        [p]
                            You [b]really[/b] don't want to do this, but the wolf has left you with little choice.
                            You'd rather give his sack a single, quick kiss than have his loincloth soak in your mouth for the rest of the afternoon.
                            Should you end up not being able to escape, and not being [i]bought[/i],
                            then you'd prefer to stay on good terms with Fenrir instead of [i]pissing him off[/i] even further.
                            His threats might [i]sound[/i] a bit exaggerated, but you're fairly certain he's done this kind of thing before,
                            so he might [i]not[/i] be lying about what his men do to slaves.
                            Maybe [b]that's[/b] why he wants [b]you[/b].
                            The mere thought makes your stomach churn.
                        [/p]
                        [p]
                            Noticing the bandit growing impatient, you roll your eyes and let out a defeated sigh.
                            Slowly, but steadily, you inch your face closer to his balls.
                            The strong scent of canine masculinity intensifies,
                            male pheromones [b]wafting[/b] liberally from Fenrir's loins,
                            [b]clogging[/b] your nostrils, [b]saturating[/b] every breath you take.
                            As your snout sinks into the fur covering his sack,
                            you pucker your lips, and plant a single, tender kiss in between the wolf's large, dangling testicles.
                        [/p]
                        [p]
                            As soon as your half of the [i]bargain[/i] is complete,
                            you immediately pull back from Fenrir's groin.
                            You turn your face to the side,
                            deeply breathing in some less [i]heady[/i] air,
                            while wiping your tongue and your snout both, to symbolize your disdain.
                            You consider [i]spitting[/i] for dramatic effect,
                            but you figure you might have to spend another night in this cart,
                            so you'd rather not filthy the floor.
                        [/p]
                        [p]
                            "[color=gblue]Good ${dick ? `boy` : `girl`}.[/color]" he patronizes,
                            his voice heavy with smugness.
                            "[color=gblue]The sooner you get used to that taste, the better.
                            You'll be [b]bathing[/b] my nuts every morning, once you're finally [b]mine[/b].[/color]"
                            Arrogant prick.
                            At least he doesn't have to gag you now, right?
                            You think that goes without saying.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\reluctant\SlaveNoonTwoBadReluctant.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\attack\SlaveNoonTwoBadAttack.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadAttack(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.ATTACK;
    var reqController = ctrl.blackmail;
    var subController = ctrl.preGag;
    var changeController = ctrl.badAttack;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.gagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.attacked = true;
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, there'll be [b]hell[/b] to pay.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll just lead to Fenrir asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            And you don't want that.
                        [/p]
                        [p]
                            Consequences be damned, you're [b]not[/b] doing this.
                            He can threaten to [b]kill you[/b] all he'd like.
                            At least that way, you won't have to be his [b]slave[/b] for the rest of your undoubtedly [i]miserable[/i] life.
                            While you [i]are[/i] sort of still frightened of what his gang might do to you,
                            and you [b]definitely[/b] hope he was [b]bluffing[/b] when he said he was going to put his underwear into your mouth,
                            you think you've found the courage to stand up to the big, imposing predator.
                        [/p]
                        [p]
                            Anger wells up inside of you.
                            You aren't backing down.
                            Not this time.
                            In fact, with Fenrir's nude groin so close to your face, you spot an [i]opportunity[/i].
                            You could easily [b]bite down[/b] on his most sensitive bits,
                            and then, while he's writhing in pain, you can make a quick getaway.
                            [i]Excellent idea.[/i]
                        [/p]
                        [p]
                            Staring upwards at the lupine with a mischievous grin on your face,
                            you open your muzzle wide, and slowly approach his hefty balls with your snout.
                            Instead of going in for a kiss, you bring your sharp teeth to bear, 
                            intent on [b]chomping[/b] off a piece of the bandit's nutsack.
                            The wolf is not as ignorant as you believed him to be, however.
                            Or maybe you were too [i]eager[/i] to properly hide your true intentions.
                            Right before your fangs can sink into his loins,
                            he slams your forehead with the flat of his large palm,
                            harshly pushing you backwards, your muzzle snapping shut,
                            catching nothing but [i]air[/i] in between your teeth.                       
                        [/p]
                        [p]
                            Surprised by your plan's failure, you are suddenly filled with rage.
                            You close your eyes and begin to wildly [b]flail around[/b],
                            swinging your claws about every which way, hoping you'll strike a lucky hit.
                            Fenrir keeps you at bay by simply holding you by your forehead,
                            keeping you at arm's length like a parent would its unruly child.
                            That is, after all, what you are acting like: a [b]child[/b],
                            lashing out because it did something bad, one that doesn't [i]want[/i] to face the consequences of its actions.
                        [/p]
                        [p]
                            After nearly half a minute of hitting nothing but air, your angry, impotent rampage slowly comes to an end.
                            You open your eyes once more, a few frustrated tears rolling down your cheeks.
                            All the while, a sadistic grin is plastered across the canine's face.
                            "[color=gblue]You could have just said [i]no[/i], instead of [b]embarrassing[/b] yourself.
                            If you were mine already, I'd beat the [b]shit[/b] out of you for trying to lay a hand on me.
                            Luckily for you, bruises don't sell.
                            Anyway, you've made your choice.
                            The [b]hard way[/b] it is.[/color]"
                            [i]Oh dear.[/i]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\attack\SlaveNoonTwoBadAttack.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\insult\SlaveNoonTwoBadInsult.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadInsult(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.INSULT;
    var reqController = ctrl.blackmail;
    var subController = ctrl.preGag;
    var changeController = ctrl.badInsult;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.gagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.insulted = true;
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, there'll be [b]hell[/b] to pay.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll just lead to Fenrir asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            And you don't want that.
                        [/p]
                        [p]
                            Consequences be damned, you're [b]not[/b] doing this.
                            He can threaten to [b]kill you[/b] all he'd like.
                            At least that way, you won't have to be his [b]slave[/b] for the rest of your undoubtedly [i]miserable[/i] life.
                            While you [i]are[/i] sort of still frightened of what his gang might do to you,
                            and you [b]definitely[/b] hope he was [b]bluffing[/b] when he said he was going to put his underwear into your mouth,
                            you think you've found the courage to stand up to the big, imposing predator.
                        [/p]
                        [p]
                            You answer Fenrir's stern gaze with a determined glare of your own.
                            Anger wells up inside of you.
                            You aren't backing down.
                            Not this time.
                            You take in a deep breath, raise your voice, and let loose an angry [b]tirade[/b] of insults.
                            He's a filthy [b]slaver[/b].
                            A pathetic [b]loser[/b] that couldn't carve out an honest life for himself,
                            and is now stuck [b]bullying[/b] innocent kobolds like yourself!
                            You know you're not [i]really[/i] innocent, but it seemed like an [i]apt[/i] choice of words.
                        [/p]
                        [p]
                            Furthermore, he's [b]wrong[/b] about nobody wanting to buy you.
                            You're a [b]good[/b] kobold, and there'll surely be [b]dozens[/b] of people lining up to get a [b]chance[/b] at owning you!
                            You'll [b]never[/b] be [b]his[/b] slave!
                            Heck, you'll never be [b]anyone[/b]'s slave!
                            You're going to escape the first chance you get, and there's [b]nothing[/b] Fenrir can do to stop you.
                            And if the wolf so much as lays a hand on you, you'll tell the caravan master,
                            and the human'll yank the [b]big bitch[/b] that the bandit is back into submission in [b]no time[/b]!
                        [/p]
                        [p]
                            You find yourself panting slightly, your angry tantrum having left you out of breath.
                            The look on the canine's face transitions rapidly from disbelief, to anger, to something resembling a lack of understanding.
                            Finally, he settles on a sadistic grin.
                            "[color=gblue]Is that all?
                            You know, you could have just said [i]no[/i].
                            Anyway, suit yourself.
                            The [b]hard way[/b] it is.[/color]"
                            [i]Oh dear.[/i]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\insult\SlaveNoonTwoBadInsult.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\decline\SlaveNoonTwoBadDecline.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadDecline(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.DECLINE;
    var reqController = ctrl.blackmail;
    var subController = ctrl.preGag;
    var changeController = ctrl.badDecline;

    var autoContinue = false;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.gagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, there'll be [b]hell[/b] to pay.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll just lead to Fenrir asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            And you don't want that.
                        [/p]
                        [p]
                            Consequences be damned, you're [b]not[/b] doing this.
                            He can threaten to [b]kill you[/b] all he'd like.
                            At least that way, you won't have to be his [b]slave[/b] for the rest of your undoubtedly [i]miserable[/i] life.
                            While you [i]are[/i] sort of still frightened of what his gang might do to you,
                            and you [b]definitely[/b] hope he was [b]bluffing[/b] when he said he was going to put his underwear into your mouth,
                            you think you've found the courage to stand up to the big, imposing predator.
                        [/p]
                        [p]
                            Answering Fenrir's stern gaze with a determined glare of your own,
                            you firmly shake your head no, and take a step back.
                            The look on his face transitions rapidly from disbelief, to anger, to something resembling a lack of understanding.
                            Finally, he settles on a sadistic grin.
                            "[color=gblue]Defiant to the very end, huh?
                            Suit yourself.
                            The [b]hard way[/b] it is.[/color]"
                            [i]Oh dear.[/i]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    /**
     * Skips choice generation and immediately continues.
     */
    function skipChoice() {
        subController.continue();
        return "";
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += autoContinue ? skipChoice() : buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued() || autoContinue) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\decline\SlaveNoonTwoBadDecline.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\nothing\SlaveNoonTwoBadNothing.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadNothing(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.NOTHING;
    var reqController = ctrl.blackmail;
    var subController = ctrl.preGag;
    var changeController = ctrl.badNothing;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.gagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, there'll be [b]hell[/b] to pay.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll undoubtedly lead to Fenrir asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            And you don't want that.
                        [/p]
                        [p]
                            You find yourself locked in mental conflict,
                            internally debating on whether or not you should accept the wolf's offer.
                            The question the canine has posed is not an easy one.
                            Ultimately, it boils down to whether you want to live in agony,
                            or accept the sweet mercy of [b]death[/b].
                            Being a kobold, philosophy was never your strong suit.
                            You simply [i]lack[/i] the mental capacity needed to think something like this through.
                        [/p]
                        [p]
                            Why couldn't he have just grabbed your head and shoved you into his nuts?!
                            That would have made things [b]so[/b] much easier.
                            That way, your body and your natural instincts could have made this decision for you.
                            Fenrir seems to be getting impatient, but you haven't reached a conclusion yet.
                            Part of you wants to submit.
                            You're scared of what might happen if you don't.
                            But doing this means forsaking whatever dignity you have left.
                            And you can't do that.
                            Not out of your own volition, at least.
                        [/p]
                        [p]
                            You fail to heed the lupine's request.
                            The look on his face transitions rapidly from boredom, to anger, to something resembling a lack of understanding.
                            Finally, he settles on a sadistic grin.
                            "[color=gblue]Not going to do it, huh?
                            Suit yourself.
                            The [b]hard way[/b] it is.[/color]"
                            [i]Oh dear.[/i]
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\nothing\SlaveNoonTwoBadNothing.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\SlaveNoonTwoBadOffer.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoBadOffer(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.blackmail;
    var changeController = ctrl.blackmail;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";
    var fourthId = arguments.callee.name + "ButtonFour";
    var fifthId = arguments.callee.name + "ButtonFive";
    var sixthId = arguments.callee.name + "ButtonSix";

    var firstNext = new SlaveNoonTwoBadNothing(ctrl);
    var secondNext = new SlaveNoonTwoBadDecline(ctrl);
    var thirdNext = new SlaveNoonTwoBadInsult(ctrl);
    var fourthNext = new SlaveNoonTwoBadAttack(ctrl);
    var fifthNext = new SlaveNoonTwoBadReluctant(ctrl);
    var sixthNext = new SlaveNoonTwoBadEager(ctrl);

    var firstChoice = SlaveNoonTwoOfferChoice.NOTHING;
    var secondChoice = SlaveNoonTwoOfferChoice.DECLINE;
    var thirdChoice = SlaveNoonTwoOfferChoice.INSULT;
    var fourthChoice = SlaveNoonTwoOfferChoice.ATTACK;
    var fifthChoice = SlaveNoonTwoOfferChoice.RELUCTANT;
    var sixthChoice = SlaveNoonTwoOfferChoice.EAGER;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.blackmailed = true;

        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var enjoyed = global.kobold.history.shared.enslavement.captureEnjoyed;
        var bindings = !global.kobold.history.shared.enslavement.nightOne.escapedBindings;
        var nameKnown = global.kobold.history.shared.enslavement.nightOne.nameStated;
        var dick = ctrl.misc.dick();
        var name = ctrl.misc.name();

        var toReturn = ``;

        if (ctrl.arrival.get() === SlaveNoonTwoArrivalChoice.LOOKOUT) {
            toReturn += `[p]
                            The big wolf struggles to stand fully upright inside of your cell,
                            his head bumping against the ceiling.
                            After grumbling something about how they should really make these carts a bit bigger, a smirk forms on his muzzle.
                            "[color=gblue]Already up and about, are you?
                            ${!bindings ? `[i]Impressive[/i].` : `[i]Good[/i].`}
                            You have a big day ahead of you.[/color]"
                        [/p]`;
            if (bindings) {
                toReturn += `
                        [p]
                            "[color=gblue]Face away, slave.
                            Let me get that rope off of you.[/color]"
                            While you don't really [i]want[/i] to [b]turn around[/b] for the fully naked lupine,
                            the premise of having your hands freed sounds too good to pass up.
                            Besides, he's clearly [i]not[/i] aroused, so you doubt he'll [i]abuse[/i] the situation.
                        [/p]
                        [p]
                            Keeping your wary eyes on the wolf for as long as you can manage, you slowly turn your back towards him.
                            With a few heavy steps, the canine approaches you from behind, the wood creaking underneath his broad footpaws.
                            He comes to a halt directly behind you, but is nice enough to leave half a foot orso in between your body and his.
                            The bandit chief reaches for your wrists, still bound behind your back, and with a simple [b]slice[/b] of a single claw,
                            he severs the cord keeping your hands together.
                        [/p]
                        [p]
                            You clutch each of your wrists in turn, rubbing them to ease the minor ropeburn.
                            After giving you a moment to enjoy your newfound freedom, the well-built man grabs your shoulder,
                            and turns your body around once more, your muzzle ending up mere inches away from his bare groin,
                            a whiff of its potent scent already reaching your nostrils.
                            Fenrir stares down at you, a lecherous look on his face.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            "[color=gblue]So Guilliman [i]wasn't[/i] lying about you escaping your bindings.
                            I assumed he caved your head in because he [b]felt[/b] like it.
                            I know I sure did, after last night.[/color]"
                            You frown at the thought of being beaten by the bandit.
                            "[color=gblue]Oh well.
                            It's not like he [b]needs[/b] full pay.
                            His family back home is rich as fuck.
                            Figure he's [i]rebelling[/i] against'em or something.
                            And he [b]did[/b] still damage my [i]most valuable[/i] little slave, now didn't he?[/color]"
                        [/p]
                        [p]
                            You can't help but sort of [i]agree[/i] with Fenrir on that last point.
                            There was absolutely [b]no[/b] reason for that horse to hit you.
                            Sure, you were planning to escape,
                            and you'd have bolted past him or tackled him or tried to [b]maim[/b] him if you hadn't frozen up,
                            but still.
                            What he did was plain [b]rude[/b].
                            He could have easily overpowered you [i]without[/i] using his club.
                        [/p]
                        [p]
                            "[color=gblue]Let's have a look.[/color]"
                            the canine continues, crossing the distance between you with a few heavy steps, coming to rest about half a foot away from you,
                            your muzzle mere inches away from his bare groin,
                            a whiff of its potent scent already reaching your nostrils.
                            Acting like it's no big deal, he grabs your head and tilts it downwards,
                            nearly shoving your face into his crotch in the process.
                        [/p]
                        [p]
                            "[color=gblue]It's only a small [i]bruise[/i].
                            At worst, it'll swell up a bit.
                            If any of the buyers ask, tell them that's how you were captured.[/color]"
                        [/p]
                        [p]
                            His inspection finished, the wolf lets go of your cranium.
                            You quickly pull yourself away from his loins, expecting him to depart.
                            Fenrir, however, seems to have something else in mind.
                            He stares down at you, with a lecherous look on his face.
                        [/p]`;
            }

        } else {
            toReturn += `[p]
                            The big wolf struggles to stand fully upright inside of your cell,
                            his head bumping against the ceiling.
                            He casts his gaze downwards, staring at you like a disgruntled parent would at their child.
                            After grumbling something about how they should really make these carts a bit bigger, a faint smirk forms on his muzzle.
                            "[color=gblue]Rough night, huh?[/color]"
                            A rhetorical question.
                        [/p]
                        [p]
                            With a few heavy steps, the canine approaches your downed frame, the wood creaking underneath his broad footpaws.
                            Grabbing you by the back of your neck, he hoists you upright.
                            ${bindings ? `Reaching behind you, the wolf then speedily snaps the length of rope binding your hands together.` : ``}
                            "[color=gblue]On your feet.
                            There we go.
                            You have a big day ahead of you.[/color]"
                        [/p]`;
            if (bindings) {
                toReturn += `
                        [p]
                            You clutch each of your wrists in turn, rubbing them to ease the minor ropeburn.
                            After enjoying your newfound freedom for a moment, you tilt your muzzle upwards,
                            only to find yourself mere inches away from the lupine's naked groin,
                            a whiff of its potent scent already reaching your nostrils.
                            Fenrir stares down at you, a lecherous look on his face.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            "[color=gblue]So Guilliman [i]wasn't[/i] lying about you escaping your bindings.
                            I assumed he caved your head in because he [b]felt[/b] like it.
                            I know I sure did, after last night.[/color]"
                            You frown at the thought of a remorseless bandit hitting you [i]for fun[/i].
                            "[color=gblue]Oh well.
                            It's not like he [b]needs[/b] full pay.
                            His family back home is rich as fuck.
                            Figure he's [i]rebelling[/i] against'em or something.
                            And he [b]did[/b] still damage my [i]most valuable[/i] little slave, now didn't he?[/color]"
                        [/p]
                        [p]
                            While you hate the way it was worded, you can't help but sort of [i]agree[/i] with Fenrir on that last point.
                            There was absolutely [b]no[/b] reason for that horse to hit you.
                            Sure, you were planning to escape,
                            and you'd have bolted past him or tackled him or tried to [b]maim[/b] him if you hadn't frozen up,
                            but still.
                            What he did was plain [b]rude[/b].
                            He could have easily overpowered you [i]without[/i] using his club.
                        [/p]
                        [p]
                            "[color=gblue]Let's have a look.[/color]"
                            the canine continues.
                            Acting like it's no big deal, he grabs your head and tilts it downwards once more,
                            tugging you a little closer, nearly shoving your face into his crotch in the process.
                        [/p]
                        [p]
                            "[color=gblue]It's only a small [i]bruise[/i].
                            At worst, it'll swell up a bit.
                            If any of the buyers ask, tell them that's how you were captured.[/color]"
                        [/p]
                        [p]
                            His inspection finished, the wolf lets go of your cranium.
                            You quickly pull yourself away from his loins, expecting him to depart.
                            Fenrir, however, seems to have something else in mind.
                            He stares down at you, with a lecherous look on his face.
                        [/p]`;
            }
        }

        if (!nameKnown) {
            toReturn += `[p]
                            "[color=gblue]I had a few of my men ask around during the night.
                            The forest has loose lips, if you know where to pry.[/color]"
                            Huh?
                            He sent out [i]spies[/i]?
                            "[color=gblue]${name}.
                            That's your name, right?[/color]"
                            You stare up at the lupine in amazement and disbelief.
                            Barely [b]anyone[/b] around here knows your name.
                            He chuckles.
                            "[color=gblue]It fits you.[/color]"
                        [/p]`;
        }

        toReturn += `   [p]
                            The wolf moves one of his hands to his groin.
                            He boldly, brazenly and crudely cups the hefty, furred nuts that swing underneath his thick sheath,
                            propping them forwards as if to [i]entice[/i] you.
                            "[color=gblue]Look, ${name}, you acted like a real [b]pest[/b] yesterday.
                            ${fucked ? `In the forest, ${enjoyed ? `you sounded like you'd make a good slave,
                            but then when we got to the camp, you showed your true colors.` : `I had to [b]force[/b] myself upon you,
                            and then when we got back to camp, you acted all uppity.`}` : `First you blueballed me in the forest,
                            and then you failed to obey some simple commands in front of the caravan master.`}
                            You embarrassed me in front of that [b]moronic human[/b].
                            And that's not something I can appreciate.[/color]"
                        [/p]
                        [p]
                            You become aware that, in his free hand, the canine is holding a rope, and something you can't quite make out.
                            Some kind of [i]fabric[/i]?
                            "[color=gblue]From the moment you ran away,
                            I [b]knew[/b] you were different from all the other kobolds.
                            Instead of being [b]meek[/b] and [b]compliant[/b], you're [b]rebellious[/b] and [b]uncooperative[/b].
                            You remind me of my son.
                            But unlike him, I [i]can[/i] wring [b]your[/b] neck without repercussions.[/color]"
                        [/p]
                        [p]
                            Starting to think that Fenrir came here to [b]hang[/b] you with that rope,
                            you somewhat begin to panic.
                            "[color=gblue]I can make your life [b]a lot[/b] easier, you know.
                            I can bring you some leftovers to eat at night,
                            I can tell the gang not to inflict too much [i]permanent[/i] damage on you,
                            I might even be able to convince them that you'd be better off [b]keeping[/b] your teeth.[/color]"
                        [/p]
                        [p]
                            Hardly reassuring benefits, but you're [b]very[/b] certain that even those small [i]favors[/i] have a [b]price[/b].
                            Bandits like the wolf never hand out [i]freebies[/i].
                            "[color=gblue]Let's face the facts here.
                            That [b]fat fuck[/b] thinks he can [i]sell[/i] you.
                            But both you and me know that there's [b]plenty[/b] of slaves to go around.
                            It's an [i]oversaturated[/i] market.[/color]"
                            Fancy words, for a brigand.
                            Then again, he [i]is[/i] their leader.
                            "[color=gblue]You aren't the best product around, either.
                            People would much rather have a well-trained servant than an uppity [b]${dick ? `jerk` : `cunt`}[/b].
                            And, as you've no-doubt overheard, the caravan master only has two tries to auction you off.
                            After that, you'll be [b]mine[/b].[/color]"
                        [/p]
                        [p]
                            You repress the urge to ask the lupine about the rope.
                            You sense that he's slowly getting to that.
                            "[color=gblue]I won't lie.
                            You're going to be fucked, gangbanged, spitroasted and ran a train on by every single last member of my group.
                            Including the dog.
                            We're going to beat you, defile you in every way possible,
                            and make sure that nobody else will even [b]want[/b] you after we're done.
                            This [b]is[/b] going to happen, whether you like it or not.
                            But if you play along, then things don't [b]need[/b] to get [i]out of hand[/i].
                            If you stay unwilling, well...
                            Let's just say you won't be leaving our company alive.
                            And it won't be a pretty death, either.[/color]"
                        [/p]
                        [p]
                            You get the feeling he's not kidding about that threat.
                            In fact, Fenrir seems [b]dead serious[/b] about all of this.
                            He stares into your eyes, probing your reaction,
                            giving you a moment to process all of this, before resuming his long-winded speech.
                            "[color=gblue]On top of all of this,
                            since you acted so piss-poorly yesterday,
                            the caravan master has instructed me to gag your pretty little mouth,
                            so you don't go insulting buyers during the auction.[/color]"
                        [/p]
                        [p]
                            The wolf unfurls his hand, revealing that the thing he was clutching was a garment.
                            Underwear?
                            "[color=gblue]We don't carry any ballgags,
                            and the dog didn't want to let go of its favorite toy, so yeah.
                            This is the pair I wore yesterday, when [b]you[/b] made me run halfway through the forest.
                            ${fucked ? `Still has the dried spunk on it from when I fucked your tight little ${dick ? `ass` : `slit`}, even.` : ``}[/color]"
                        [/p]
                        [p]
                            Wait a minute, he isn't [i]actually[/i] considering using his sweaty briefs as a gag, is he?
                            He [b]must[/b] be bluffing, right?
                            "[color=gblue]I'm going to put this as simple as I can,
                            so even a dumb kobold like yourself can understand.
                            I want you to [b]submit[/b] to me.
                            If you don't, I'm going to make your life a living [b]hell[/b] from now on.
                            And I'll start by [b]shoving[/b] my well-worn underwear so far down your throat you'll barely be able to [b]breathe[/b].
                            Understand?[/color]"
                        [/p]
                        [p]
                            Awestruck and more than a little terrified of the consequences, you give Fenrir a small nod.
                            "[color=gblue]Good.
                            All I want is for you to recognize me as your new master.
                            I won't make you say it.
                            I know you're far too [i]proud[/i] for that.
                            Instead, just give my balls a single [b]good morning kiss[/b],
                            and then we can forget all about this whole gag thing.[/color]"
                        [/p]
                        [p]
                            As if to accentuate his statement, the wolf taps his furred nuts,
                            rolling his hips forwards ever so slightly to hold them a little closer to your mouth.
                            Part of your mind is honestly [i]considering[/i] Fenrir's offer.
                            Should both the auctions fail, at least you won't end up [b]dead[/b].
                            But then again, you plan on escaping long before you become the lupine's property.
                        [/p]
                        [p] 
                            "[color=gblue]Nobody has to know.
                            This'll be between [b]you[/b] and [b]me[/b].
                            A small peck is all I ask for.
                            A simple [b]sign[/b] of [b]submission[/b].
                            So what'll it be, slave?
                            Are we going to do this the [i]easy[/i] way, or the [b]hard[/b] way?[/color]"
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled} style="width: 190px;">Do nothing</button>`;

        var tooltip = global.parser.parse(`Remain in place. Do not act. If you don't do anything, you can't do something wrong.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled} style="width: 190px;">Decline</button>`;

        var tooltip = global.parser.parse(`You don't want to get anywhere [b]near[/b] the wolf's loins.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled} style="width: 190px;">Insult</button>`;

        var tooltip = global.parser.parse(`Who the hell does Fenrir think he is?! Tell him to fuck off.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }
    /**
     * Builds and returns the choice button.
     */
    function buildFourthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === fourthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fourthId}" ${disabled} style="width: 190px;">Attack</button>`;

        var tooltip = global.parser.parse(`Charge the wolf! Freedom or death!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 145px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFifthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === fifthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fifthId}" ${disabled} style="width: 190px;">Reluctantly kiss</button>`;

        var tooltip = global.parser.parse(`Ugh. Do you [i]really[/i] have to do this?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSixthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === sixthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${sixthId}" ${disabled} style="width: 190px;">Eagerly kiss</button>`;

        var tooltip = global.parser.parse(`Sweaty wolf nuts are to die for!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 225px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += `<div class="choicerow" style="margin-bottom: 10px;">`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        choice += `<div class="choicerow">`;
        choice += buildFourthChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFifthChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSixthChoice();
        choice += `</div>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId), $("#" + fourthId), $("#" + fifthId), $("#" + sixthId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFourthListener() {
        $("#" + fourthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fourthChoice);
            loadNextPart(fourthNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFifthListener() {
        $("#" + fifthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fifthChoice);
            loadNextPart(fifthNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSixthListener() {
        $("#" + sixthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(sixthChoice);
            loadNextPart(sixthNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        part += fourthNext.build();
        part += fifthNext.build();
        part += sixthNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();
        fourthNext.registerListeners();
        fifthNext.registerListeners();
        sixthNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
        registerFourthListener();
        registerFifthListener();
        registerSixthListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\bad\SlaveNoonTwoBadOffer.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\potion\SlaveNoonTwoGoodFemPlus.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodFemPlus(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoGoodPotionChoice.FEMPLUS;
    var reqController = ctrl.potion;
    var subController = ctrl.preOutside;
    var changeController = ctrl.femPlus;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.outside;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.tookFemPlus = true;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            You've steered clear of [i]body-altering[/i] drugs so far,
                            partly because you like yourself the way you are,
                            and partly because you've never really had the [i]money[/i] to buy one.
                            Now that an opportunity is presenting itself, however,
                            you can't say you're not a teensy bit [i]interested[/i].
                            You might like your body even [b]better[/b] with a few [i]enhancements[/i].
                            And you doubt you'll ever get a chance like this again.
                        [/p]
                        [p]
                            After a bit of mental deliberation, you gingerly reach out for the [color=pink][b]pink[/b][/color] bottle.
                            ${global.kobold.inventory.createItemWithTooltip(new FemPlus(), true)}.
                            This should make you more of a [i]woman[/i]${dick ? `, if at the cost of some masculinity` : ``}.
                            "[color=gblue]Girl juice, huh? Good choice. Drink up.[/color]"
                        [/p]
                        [p]
                            You've thought about this plenty.
                            Perhaps even considered it once or twice.
                            But you never would have believed that the opportunity would have presented itself so [i]casually[/i].
                            You didn't have to [b]steal[/b] the potion, or [b]beg[/b] Seki for one.
                            While Fenrir may or may not realize it, this [i]present[/i] means a lot more to you than he thinks.
                            Plus you know these concoctions don't come cheap.
                            Maybe he's not that bad of a person after all.
                            Or perhaps he hopes it'll make you [b]sell[/b] for more.
                        [/p]
                        [p]
                            You pop the lid off the potion, and gulp down its contents.
                            The rather watery liquid tastes like an odd mixture of strawberries and milk.
                            The moment you swallow the first mouthful, you can already feel the stuff beginning to affect your body.
                        `;
        if (dick) {
            toReturn += `   Your member stirs within your loins, tangibly growing smaller,
                            your cockslit becoming a lot more [b]roomy[/b].
                            After a few satisfying seconds, the effects come to an end.`;
        } else {
            toReturn += `   Your breasts tingle and throb, visibly growing bigger.
                            Your upper chest feels heavier with every passing second, until finally,
                            after a few satisfying seconds, the effects to an end.`;
        }

        toReturn += `   [/p]
                        [p]
                            [c]
                                [b]You have grown more feminine.[/b]
                            [/c]
                        [/p]
                        [p]
                            "[color=gblue]You look a lot better already.[/color]" comments the wolf.
                            You can't help but [i]blush[/i] a little.
                            You [b]do[/b] look [i]prettier[/i] than before.
                            And it's all thanks to Fenrir.
                            ${eager ? `You are going to [b]love[/b] being his slave.` : `The two of you may not see eye to eye,
                            but maybe being his [i]slave[/i] won't be so bad after all.
                            Then again, this might have all been a [b]trick[/b] to get you to trust him more.
                            With bandits, you can never know for sure.`}
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\potion\SlaveNoonTwoGoodFemPlus.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\potion\SlaveNoonTwoGoodMPlus.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodMPlus(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoGoodPotionChoice.MPLUS;
    var reqController = ctrl.potion;
    var subController = ctrl.preOutside;
    var changeController = ctrl.mPlus;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.outside;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.tookMPlus = true;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            You've steered clear of [i]body-altering[/i] drugs so far,
                            partly because you like yourself the way you are,
                            and partly because you've never really had the [i]money[/i] to buy one.
                            Now that an opportunity is presenting itself, however,
                            you can't say you're not a teensy bit [i]interested[/i].
                            You might like your body even [b]better[/b] with a few [i]enhancements[/i].
                            And you doubt you'll ever get a chance like this again.
                        [/p]
                        [p]
                            After a bit of mental deliberation, you gingerly reach out for the [color=blue][b]blue[/b][/color] bottle.
                            ${global.kobold.inventory.createItemWithTooltip(new mPlus(), true)}.
                            This should make you more of a [i]man[/i]${dick ? `` : `, if at the cost of some femininity`}.
                            "[color=gblue]Boy juice, huh? Good choice. Drink up.[/color]"
                        [/p]
                        [p]
                            You've thought about this plenty.
                            Perhaps even considered it once or twice.
                            But you never would have believed that the opportunity would have presented itself so [i]casually[/i].
                            You didn't have to [b]steal[/b] the potion, or [b]beg[/b] Seki for one.
                            While Fenrir may or may not realize it, this [i]present[/i] means a lot more to you than he thinks.
                            Plus you know these concoctions don't come cheap.
                            Maybe he's not that bad of a person after all.
                            Or perhaps he hopes it'll make you [b]sell[/b] for more.
                        [/p]
                        [p]
                            You pop the lid off the potion, and gulp down its contents.
                            The rather watery liquid tastes like blueberries.
                            The moment you swallow the first mouthful, you can already feel the stuff beginning to affect your body.
                        `;
        if (dick) {
            toReturn += `   Your member stirs within your loins, tangibly growing larger,
                            your cockslit becoming a lot more [b]cramped[/b].
                            After a few satisfying seconds, the effects come to an end.`;
        } else {
            toReturn += `   Your breasts tingle and throb, retracting into your chest, visibly growing smaller.
                            Your upper chest feels lighter with every passing second, until finally,
                            after a few satisfying seconds, the effects to an end.`;
        }

        toReturn += `   [/p]
                        [p]
                            [c]
                                [b]You have grown more masculine.[/b]
                            [/c]
                        [/p]
                        [p]
                            "[color=gblue]You look a lot better already.[/color]" comments the wolf.
                            You can't help but [i]blush[/i] a little.
                            You [b]do[/b] look a bit more [i]handsome[/i] than before.
                            And it's all thanks to Fenrir.
                            ${eager ? `You are going to [b]love[/b] being his slave.` : `The two of you may not see eye to eye,
                            but maybe being his [i]slave[/i] won't be so bad after all.
                            Then again, this might have all been a [b]trick[/b] to get you to trust him more.
                            With bandits, you can never know for sure.`}
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\potion\SlaveNoonTwoGoodMPlus.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\potion\SlaveNoonTwoGoodNeither.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodNeither(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoGoodPotionChoice.NEITHER;
    var reqController = ctrl.potion;
    var subController = ctrl.preOutside;
    var changeController = ctrl.neither;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.outside;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var knockedOut = global.kobold.history.shared.enslavement.nightOne.knockedOut;
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var toReturn = `[p]
                            You've steered clear of [i]body-altering[/i] drugs so far,
                            partly because you like yourself the way you are,
                            and partly because you've never really had the [i]money[/i] to buy one.
                            Now that an opportunity is presenting itself, however,
                            you can't say you're not a teensy bit [i]interested[/i].
                            You might like your body even [b]better[/b] with a few [i]enhancements[/i].
                            But that's a [b]dangerous[/b] road to go down.
                        [/p]
                        [p]
                            You've thought about this plenty.
                            Perhaps even considered it once or twice.
                            But now that the opportunity has presented itself,
                            you somehow don't feel very inclined to take it.
                            Besides, you're not even sure that's [i]actually[/i]
                            ${global.kobold.inventory.createItemWithTooltip(new mPlus(), true)}
                            and 
                            ${global.kobold.inventory.createItemWithTooltip(new FemPlus(), true)}
                            in those bottles.
                            Who knows what the wolf put in them?
                        [/p]
                        [p]
                            After a bit of mental deliberation, you decide not to take either potion.
                            You give Fenrir an unhappy look, and shake your head no.
                            "[color=gblue]Neither?
                            Picky little thing, aren't you?
                            Well, suit yourself.
                            Don't expect me to hang on to these until you change your mind, though.[/color]"
                        [/p]
                        [p]
                            While you didn't end up [i]accepting[/i] his gift,
                            you can't help but feel slightly [i]grateful[/i] for the canine to offer it in the first place.
                            You know those concoctions don't come cheap, and you assume he went out of his way to get them for you.
                            A nice gesture, although he could have [b]asked[/b] you first.
                            But you [b]do[/b] suppose you were ${knockedOut ? `unconscious` : `asleep`} until not too long ago.
                            ${eager ? `` : `Then again, this entire ordeal might have all been a [b]trick[/b] to get you to trust him more.
                            With bandits, you can never know for sure.`}
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\potion\SlaveNoonTwoGoodNeither.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\potion\SlaveNoonTwoGoodPotion.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodPotion(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preGag;
    var subController = ctrl.potion;
    var changeController = false;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";

    var firstNext = new SlaveNoonTwoGoodNeither(ctrl);
    var secondNext = new SlaveNoonTwoGoodMPlus(ctrl);
    var thirdNext = new SlaveNoonTwoGoodFemPlus(ctrl);

    var firstChoice = SlaveNoonTwoGoodPotionChoice.NEITHER;
    var secondChoice = SlaveNoonTwoGoodPotionChoice.MPLUS;
    var thirdChoice = SlaveNoonTwoGoodPotionChoice.FEMPLUS;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var eager = global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager;
        var name = ctrl.misc.name();
        var toReturn = `[p]
                            Your private business concluded,
                            the wolf crudely grabs you by the wrist, and begins to tug you towards the cart's open entryway.
                            "[color=gblue]You and I have business to attend to.
                            In case you hadn't realized yet: you're being [b]sold[/b].
                            [b]Today[/b].
                            And it's [b]my[/b] job to ensure you look at least [i]somewhat[/i] presentable.[/color]"
                        [/p]
                        [p]
                            He's not taking [i]no[/i] for an answer,
                            yanking you forwards hard enough to nearly pull your arm off.
                            Guess he's in a rush, despite wasting time with his [i]proposition[/i].
                            You don't resist much, instead trying to keep up with his pace as best you can.
                            A few steps and a small hop later, you find yourself in the outside world once more.
                        [/p]`;

        if (ctrl.arrival.get() !== SlaveNoonTwoArrivalChoice.LOOKOUT) {
            toReturn += `[p]
                            You recognize this area.
                            You've been here before.
                            It's a place known simply as [i]the hamlet[/i],
                            a small village close to the human capital of Varanar.
                        [/p]
                        [p]
                            An old friend of yours lives here.
                            Seki, the alchemist-owl, who also happens to happily buy whatever stolen goods you bring him.
                            Besides being your [i]fence[/i], he's also the closest thing to a [i]father[/i] you've ever had.
                            Maybe [b]he[/b] can help you get out of this [b]mess[/b].
                        [/p]`;
        }

        toReturn += `   [p]
                            Luckily, the caravan stopped at the outskirts of the settlement,
                            your nudity not made public quite yet.
                            If it's any [i]consolation[/i], your captor is every bit as naked as you are.
                            Walking alongside your prison carriage's exterior,
                            whilst dragging you with him, the canine comes to an unexpected halt,
                            your face ending up unpleasantly [i]bumping[/i] into his bare rear.
                            Reaching down, he fetches a small, brown satchel, conveniently concealed directly underneath the cart.
                        [/p]
                        [p]
                            "[color=gblue]I almost forgot.
                            I've prepared a [i]treat[/i] for you.
                            A little [b]present[/b], to help your odds at the auction.
                            I have no doubt you'll be making back every coin I spent on these and more.[/color]"
                            You don't know whether you should feel [i]flattered[/i], or [b]insulted[/b].
                            "[color=gblue]If not through being auctioned off,
                            then through all the gold I'll be saving from fucking you instead of visiting a brothel.[/color]"
                            Definitely [b]${eager ? `flattered` : `insulted`}[/b].
                        [/p]
                        [p]
                            Rummaging through the tiny bag, Fenrir produces a pair of oddly shaped bottles, along with a large, ominous-looking syringe.
                            ${global.kobold.inventory.createItemWithTooltip(new FemPlus(), true)}
                            and ${global.kobold.inventory.createItemWithTooltip(new mPlus(), true)}.
                            Seki makes that sort of stuff.
                            "[color=gblue]Well, here they are.
                            These things are expensive,
                            so you only get [b]one[/b].
                            Well, ${name}, which'll it be?
                            The blue one, or the pink one?[/color]"
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Neither</button>`;

        var tooltip = global.parser.parse(`You don't need to change.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>mPlus</button>`;

        var tooltip = global.parser.parse(`Maybe you could stand to be a bit more [i]masculine[/i].`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled}>FemPlus</button>`;

        var tooltip = global.parser.parse(`The more girly, the better.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 190px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
    };
}

var SlaveNoonTwoGoodPotionChoice = {
    NEITHER: 1,
    MPLUS: 2,
    FEMPLUS: 3
};

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\potion\SlaveNoonTwoGoodPotion.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\eager\SlaveNoonTwoGoodEager.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodEager(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.EAGER;
    var reqController = ctrl.offer;
    var subController = ctrl.preGag;
    var changeController = ctrl.goodEager;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.notGagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.kissedBallsEager = true;
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, you won't gain the wolf's favor.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll undoubtedly lead to him asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something.
                            And for some reason, that kind of [i]arouses[/i] you.
                        [/p]
                        [p]
                            Becoming the big wolf's [b]plaything[/b] is an alluring prospect by itself.
                            Combined with his tempting proposition,
                            and the equally enticing big, furred orbs swinging underneath his plump sheath,
                            Fenrir's offer is practically [b]irresistible[/b] to you.
                            How could [b]anyone[/b] be faced with such a [b]hunk[/b] of a man,
                            and [b]not[/b] want to submit to him completely?
                            Should you end up not being [i]bought[/i],
                            then you'll get to live a comfortable life with your new lupine [i]master[/i],
                            and his group.
                        [/p]
                        [p]
                            Noticing the bandit growing impatient,
                            you lustily lick your lips and begin to inch your face closer to his balls.
                            The strong scent of canine masculinity intensifies,
                            male pheromones [b]wafting[/b] liberally from Fenrir's loins,
                            [b]clogging[/b] your nostrils, [b]saturating[/b] every breath you take.
                            You greedily huff the olfactory nectar, savoring every lungful,
                            wishing you could keep your face buried in his balls [b]forever[/b].
                        [/p]
                        [p]
                            While your own need is rising, you aren't about to get [i]distracted[/i] by it.
                            You set out to [i]please[/i] your master, and you intend to [b]exceed[/b] his every expectation.
                            As your snout sinks into the grey fur covering the wolf's sack,
                            you pucker your lips, and eagerly begin to plant kiss after kiss atop of his nuts.
                            Taking one of his balls into your mouth slightly, you gently suckle on it, just a little.
                            He groans in delight, his tense, authorative stand visibly relaxing.
                            [b]Excellent[/b].
                        [/p]
                        [p]
                            With Fenrir starting to mellow out, you up the intensity of your loving [b]admiration[/b].
                            Your tongue snakes out of your muzzle, your rampant kissing interrupted by occasional [b]licks[/b].
                            Before long, you find yourself outright [b]lapping[/b] along the coarse surface of the canine's sack,
                            wetting his fur with your saliva, enjoying the salty,
                            undeniably [b]masculine[/b] flavour even [b]despite[/b] the dried up sweat and the small bits of dirt and grime.
                        [/p]
                        [p]
                            The wolf begins to pant, his tail wagging excitedly.
                            Slowly, but surely, you work your way upwards, towards his sheath.
                            Right as you were about to goad his member from its slumber by kissing its resting place directly,
                            Fenrir suddenly presses his palm against your forehead, softly pushing you away.
                        [/p]
                        [p]
                            "[color=gblue]That's enough for now, you little [b]cocktease[/b].
                            Continue to act like that, and I might need to consider buying you myself,
                            to make sure nobody else gets to run off with that lovely muzzle of yours.[/color]"
                            Your captor takes a deep breath, seemingly washing the impure thoughts from his mind.
                            "[color=gblue]There'll be plenty of time for me to fuck your face later.
                            And since you like my balls so much,
                            you'll get to [b]worship[/b] them every single day, once you're finally [b]mine[/b].[/color]"
                            You can hardly wait.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\eager\SlaveNoonTwoGoodEager.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\reluctant\SlaveNoonTwoGoodReluctant.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodReluctant(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.RELUCTANT;
    var reqController = ctrl.offer;
    var subController = ctrl.preGag;
    var changeController = ctrl.goodReluctant;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.notGagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.kissedBalls = true;
        var dick = ctrl.misc.dick();
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, you won't gain the wolf's favor.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll undoubtedly lead to him asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            And while you don't particularly enjoy the idea of becoming his [i]sex toy[/i],
                            right now, the long term benefits seem to [b]far[/b] outweigh the indignance of kissing his nuts a single time.
                        [/p]
                        [p]
                            You [b]really[/b] don't want to do this, but the wolf's offer is way too tempting to pass up on.
                            You'd rather give his sack a single, quick kiss,
                            than face the possibility of getting abused by his men for the rest of your life.
                            Should you end up not being able to escape, and not being [i]bought[/i],
                            then you'd prefer to stay on good terms with Fenrir instead of [i]upsetting him[/i].
                        [/p]
                        [p]
                            Noticing the bandit growing impatient, you roll your eyes and let out a defeated sigh.
                            Slowly, but steadily, you inch your face closer to his balls.
                            The strong scent of canine masculinity intensifies,
                            male pheromones [b]wafting[/b] liberally from Fenrir's loins,
                            [b]clogging[/b] your nostrils, [b]saturating[/b] every breath you take.
                            As your snout sinks into the fur covering his sack,
                            you pucker your lips, and plant a single, tender kiss right in between the wolf's large, dangling testicles.
                        [/p]
                        [p]
                            As soon as your half of the [i]bargain[/i] is complete,
                            you immediately pull back from Fenrir's groin.
                            You turn your face to the side,
                            deeply breathing in some less [i]heady[/i] air,
                            while wiping your tongue and your snout both, to symbolize your disdain.
                            You consider [i]spitting[/i] for dramatic effect,
                            but you figure you might have to spend another night in this cart,
                            so you'd rather not filthy the floor.
                        [/p]
                        [p]
                            "[color=gblue]Good ${dick ? `boy` : `girl`}.[/color]" he patronizes,
                            his voice heavy with smugness.
                            "[color=gblue]The sooner you get used to that taste, the better.
                            You'll be [b]kissing[/b] my [b]nuts[/b] every morning, once you're finally [b]mine[/b].[/color]"
                            Arrogant prick.
                            But, while you didn't enjoy it one bit, at least the lupine seems somewhat [i]pleased[/i] with your reluctant compliance.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\reluctant\SlaveNoonTwoGoodReluctant.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\attack\SlaveNoonTwoGoodAttack.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodAttack(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.ATTACK;
    var reqController = ctrl.offer;
    var subController = ctrl.preGag;
    var changeController = ctrl.goodAttack;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.notGagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.attacked = true;
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, you won't gain the wolf's favor.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll just lead to Fenrir asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            You don't want that.
                        [/p]
                        [p]
                            Consequences be damned, you're [b]not[/b] doing this.
                            He can threaten to have his men [b]violate[/b] you all he'd like.
                            At least that way, you won't have to be his [b]personal[/b] fucktoy.
                            While you [i]are[/i] sort of still frightened of what his gang might do to you,
                            and you [b]definitely[/b] hope he was [b]bluffing[/b] when he said that thing about the [i]dog[/i],
                            you think you've found the courage to stand up to the big, imposing predator.
                        [/p]
                        [p]
                            Anger wells up inside of you.
                            You aren't backing down.
                            Not this time.
                            In fact, with Fenrir's nude groin so close to your face, you spot an [i]opportunity[/i].
                            You could easily [b]bite down[/b] on his most sensitive bits,
                            and then, while he's writhing in pain, you can make a quick getaway.
                            [i]Excellent idea.[/i]
                        [/p]
                        [p]
                            Staring upwards at the lupine with a mischievous grin on your face,
                            you open your muzzle wide, and slowly approach his hefty balls with your snout.
                            Instead of going in for a kiss, you bring your sharp teeth to bear, 
                            intent on [b]chomping[/b] off a piece of the bandit's nutsack.
                            The wolf is not as ignorant as you believed him to be, however.
                            Or maybe you were too [i]eager[/i] to properly hide your true intentions.
                            Right before your fangs can sink into his loins,
                            he slams your forehead with the flat of his large palm,
                            harshly pushing you backwards, your muzzle snapping shut,
                            catching nothing but [i]air[/i] in between your teeth.                       
                        [/p]
                        [p]
                            Surprised by your plan's failure, you are suddenly filled with rage.
                            You close your eyes and begin to wildly [b]flail around[/b],
                            swinging your claws about every which way, hoping you'll strike a lucky hit.
                            Fenrir keeps you at bay by simply holding you by your forehead,
                            keeping you at arm's length like a parent would its unruly child.
                            That is, after all, what you are acting like: a [b]child[/b],
                            lashing out because it did something bad, one that doesn't [i]want[/i] to face the consequences of its actions.
                        [/p]
                        [p]
                            After nearly half a minute of hitting nothing but air, your angry, impotent rampage slowly comes to an end.
                            You open your eyes once more, a few frustrated tears rolling down your cheeks.
                            All the while, a sadistic grin is plastered across the canine's face.
                            "[color=gblue]You could have just said [i]no[/i], instead of [b]embarrassing[/b] yourself.
                            I'll let it slip this time, because you're new to being a slave and all.
                            Try to hurt me again, though, I'll return the favor.
                            And I can assure you [b]my[/b] fangs are [b]much[/b] sharper than yours.
                            Anyway, you've made your choice.
                            If you're [b]lucky[/b], I might see fit to ask you again later.
                            But a little [i]kiss[/i] won't suffice next time.[/color]"
                            His words worry you far more than they should.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\attack\SlaveNoonTwoGoodAttack.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\insult\SlaveNoonTwoGoodInsult.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodInsult(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.INSULT;
    var reqController = ctrl.offer;
    var subController = ctrl.preGag;
    var changeController = ctrl.goodInsult;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.notGagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.insulted = true;
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, you won't gain the wolf's favor.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll just lead to Fenrir asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            You don't want that.
                        [/p]
                        [p]
                            Consequences be damned, you're [b]not[/b] doing this.
                            He can threaten to have his men [b]violate[/b] you all he'd like.
                            At least that way, you won't have to be his [b]personal[/b] fucktoy.
                            While you [i]are[/i] sort of still frightened of what his gang might do to you,
                            and you [b]definitely[/b] hope he was [b]bluffing[/b] when he said that thing about the [i]dog[/i],
                            you think you've found the courage to stand up to the big, imposing predator.
                        [/p]
                        [p]
                            You answer Fenrir's gaze with a determined glare of your own.
                            Anger wells up inside of you.
                            You aren't backing down.
                            Not this time.
                            What gives [b]him[/b] the right to push you around?!
                            You take in a deep breath, raise your voice, and let loose an angry [b]tirade[/b] of insults.
                            He's a filthy [b]slaver[/b].
                            A pathetic [b]loser[/b] that couldn't carve out an honest life for himself,
                            and is now stuck [b]bullying[/b] innocent kobolds like yourself!
                            You know you're not [i]really[/i] innocent, but it seemed like an [i]apt[/i] choice of words.
                        [/p]
                        [p]
                            Furthermore, he's [b]wrong[/b] about nobody wanting to buy you.
                            You're a [b]good[/b] kobold, and there'll surely be [b]dozens[/b] of people lining up to get a [b]chance[/b] at owning you!
                            You'll [b]never[/b] be [b]his[/b] slave!
                            Heck, you'll never be [b]anyone[/b]'s slave!
                            You're going to escape the first chance you get, and there's [b]nothing[/b] Fenrir can do to stop you.
                            And if the wolf so much as lays a hand on you, you'll tell the caravan master,
                            and the human'll yank the [b]big bitch[/b] that the bandit is back into submission in [b]no time[/b]!
                        [/p]
                        [p]
                            You find yourself panting slightly, your angry tantrum having left you out of breath.
                            The look on the canine's face transitions rapidly from disbelief, to anger, to something resembling a lack of understanding.
                            Finally, he settles on a sadistic grin, probably trying to hide his true feelings.
                            "[color=gblue]You could have just said [i]no[/i].
                            This was a one time offer, and a good one at that.
                            If you're [b]lucky[/b], I might see fit to ask you again later.
                            But a little [i]kiss[/i] won't suffice next time.
                            And you better start watching your mouth.
                            [b]Or else[/b].[/color]"
                            His words worry you far more than they should.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\insult\SlaveNoonTwoGoodInsult.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\decline\SlaveNoonTwoGoodDecline.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodDecline(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.DECLINE;
    var reqController = ctrl.offer;
    var subController = ctrl.preGag;
    var changeController = ctrl.goodDecline;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.notGagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, you won't gain the wolf's favor.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll just lead to Fenrir asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            You don't want that.
                        [/p]
                        [p]
                            Consequences be damned, you're [b]not[/b] doing this.
                            He can threaten to have his men [b]violate[/b] you all he'd like.
                            At least that way, you won't have to be his [b]personal[/b] fucktoy.
                            While you [i]are[/i] sort of still frightened of what his gang might do to you,
                            and you [b]definitely[/b] hope he was [b]bluffing[/b] when he said that thing about the [i]dog[/i],
                            you think you've found the courage to stand up to the big, imposing predator.
                        [/p]
                        [p]
                            Answering Fenrir's stern gaze with a determined glare of your own,
                            you firmly shake your head no, and take a step back.
                            The look on his face transitions rapidly from disbelief, to anger, to something resembling a lack of understanding.
                            Finally, he settles on a sadistic grin, probably trying to hide his true feelings.
                            "[color=gblue]Not going to do it, huh?
                            Suit yourself.
                            This was a one time offer, and a good one at that.
                            If you're [b]lucky[/b], I might see fit to ask you again later.
                            But a little [i]kiss[/i] won't suffice next time.[/color]"
                            His words worry you far more than they should.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\decline\SlaveNoonTwoGoodDecline.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\nothing\SlaveNoonTwoGoodNothing.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodNothing(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoOfferChoice.NOTHING;
    var reqController = ctrl.offer;
    var subController = ctrl.preGag;
    var changeController = ctrl.goodNothing;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.notGagged;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            A million thoughts run through your head.
                            If you decline, you won't gain the wolf's favor.
                            But you are [b]very[/b] certain that [i]if[/i] you do this,
                            it'll undoubtedly lead to him asking [b]more[/b] and [b]more[/b] of you.
                            If you give in now, he'll keep pushing and pushing until you're his personal [b]concubine[/b] or something equally distasteful.
                            You don't want that.
                        [/p]
                        [p]
                            You find yourself locked in mental conflict,
                            internally debating on whether or not you should accept the wolf's offer.
                            The question the canine has posed is not an easy one.
                            Ultimately, it boils down to whether you want to please him,
                            and keep him as a friend,
                            or if you'd rather take your chances going at this alone.
                            The thing is: you don't know whether or not you [i]want[/i] a friend like Fenrir.
                            Especially not at the cost of your own pride.
                        [/p]
                        [p]
                            Why couldn't he have just grabbed your head and shoved you into his nuts?!
                            That would have made things [b]so[/b] much easier.
                            That way, your body and your natural instincts could have made this decision for you.
                            Fenrir seems to be getting impatient, but you haven't reached a conclusion yet.
                            Part of you wants to submit.
                            You're scared of what might happen if you don't.
                            But doing this means forsaking whatever dignity you have left.
                            And you can't do that.
                            Not out of your own volition, at least.
                        [/p]
                        [p]
                            You fail to heed the lupine's request.
                            The look on his face transitions rapidly from boredom, to anger, to something resembling a lack of understanding.
                            Finally, he settles on a sadistic grin, probably trying to hide his true feelings.
                            "[color=gblue]Not going to do it, huh?
                            Suit yourself.
                            This was a one time offer, and a good one at that.
                            If you're [b]lucky[/b], I might see fit to ask you again later.
                            But a little [i]kiss[/i] won't suffice next time.[/color]"
                            His words worry you far more than they should.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\nothing\SlaveNoonTwoGoodNothing.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\SlaveNoonTwoSimpleOffer.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoSimpleOffer(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.offer;
    var changeController = ctrl.offer;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";
    var fourthId = arguments.callee.name + "ButtonFour";
    var fifthId = arguments.callee.name + "ButtonFive";
    var sixthId = arguments.callee.name + "ButtonSix";

    var firstNext = new SlaveNoonTwoGoodNothing(ctrl);
    var secondNext = new SlaveNoonTwoGoodDecline(ctrl);
    var thirdNext = new SlaveNoonTwoGoodInsult(ctrl);
    var fourthNext = new SlaveNoonTwoGoodAttack(ctrl);
    var fifthNext = new SlaveNoonTwoGoodReluctant(ctrl);
    var sixthNext = new SlaveNoonTwoGoodEager(ctrl);

    var firstChoice = SlaveNoonTwoOfferChoice.NOTHING;
    var secondChoice = SlaveNoonTwoOfferChoice.DECLINE;
    var thirdChoice = SlaveNoonTwoOfferChoice.INSULT;
    var fourthChoice = SlaveNoonTwoOfferChoice.ATTACK;
    var fifthChoice = SlaveNoonTwoOfferChoice.RELUCTANT;
    var sixthChoice = SlaveNoonTwoOfferChoice.EAGER;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.simpleOffer = true;

        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var bindings = !global.kobold.history.shared.enslavement.nightOne.escapedBindings;
        var nameKnown = global.kobold.history.shared.enslavement.nightOne.nameStated;
        var dick = ctrl.misc.dick();
        var name = ctrl.misc.name();

        var toReturn = ``;

        if (ctrl.arrival.get() === SlaveNoonTwoArrivalChoice.LOOKOUT) {
            toReturn += `[p]
                            The big wolf struggles to stand fully upright inside of your cell,
                            his head bumping against the ceiling.
                            After grumbling something about how they should really make these carts a bit bigger, a smirk forms on his muzzle.
                            "[color=gblue]Already up and about, are you?
                            ${!bindings ? `[i]Impressive[/i].` : `[i]Good[/i].`}
                            You have a big day ahead of you.[/color]"
                        [/p]`;
            if (bindings) {
                toReturn += `
                        [p]
                            "[color=gblue]Face away for a second.
                            Let me get that rope off of you.[/color]"
                            While you don't really [i]want[/i] to [b]turn around[/b] for the fully naked lupine,
                            the premise of having your hands freed sounds too good to pass up.
                            Besides, he's clearly [i]not[/i] aroused, so you doubt he'll [i]abuse[/i] the situation.
                        [/p]
                        [p]
                            Keeping your wary eyes on the wolf for as long as you can manage, you slowly turn your back towards him.
                            With a few heavy steps, the canine approaches you from behind, the wood creaking underneath his broad footpaws.
                            He comes to a halt directly behind you, but is nice enough to leave half a foot orso in between your body and his.
                            The bandit chief reaches for your wrists, still bound behind your back, and with a simple [b]slice[/b] of a single claw,
                            he severs the cord keeping your hands together.
                        [/p]
                        [p]
                            You clutch each of your wrists in turn, rubbing them to ease the minor ropeburn.
                            After giving you a moment to enjoy your newfound freedom, the well-built man grabs your shoulder,
                            and turns your body around once more, your muzzle ending up mere inches away from his bare groin,
                            a whiff of its potent scent already reaching your nostrils.
                            Fenrir stares down at you, a lecherous look on his face.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            "[color=gblue]So Guilliman [i]wasn't[/i] lying about you escaping your bindings.
                            Heh.
                            I assumed he caved your head in because he [b]felt[/b] like it.[/color]"
                            You frown at the thought of a remorseless bandit hitting you [i]for fun[/i].
                            "[color=gblue]Oh well.
                            It's not like he [b]needs[/b] full pay.
                            His family back home is rich as fuck.
                            Figure he's [i]rebelling[/i] against'em or something.
                            And he [b]did[/b] still damage my [i]most precious[/i] little lizard, now didn't he?[/color]"
                        [/p]
                        [p]
                            You can't help but sort of [i]agree[/i] with Fenrir on that last point.
                            There was absolutely [b]no[/b] reason for that horse to hit you.
                            Sure, you were planning to escape,
                            and you'd have bolted past him or tackled him or tried to [b]maim[/b] him if you hadn't frozen up,
                            but still.
                            What he did was plain [b]rude[/b].
                            He could have easily overpowered you [i]without[/i] using his club.
                        [/p]
                        [p]
                            "[color=gblue]Let me have a look.[/color]"
                            the canine continues, crossing the distance between you with a few heavy steps, coming to rest about half a foot away from you,
                            your muzzle mere inches away from his bare groin,
                            a whiff of its potent scent already reaching your nostrils.
                            Acting like it's no big deal, he grabs your head and tilts it downwards,
                            nearly shoving your face into his crotch in the process.
                        [/p]
                        [p]
                            "[color=gblue]It's only a small [i]bruise[/i].
                            At worst, it'll swell up a bit.
                            If any of the buyers ask, tell them that's how you were captured.[/color]"
                        [/p]
                        [p]
                            His inspection finished, the wolf lets go of your cranium.
                            You quickly pull yourself away from his loins, expecting him to depart.
                            Fenrir, however, seems to have something else in mind.
                            He stares down at you, with a lecherous look on his face.
                        [/p]`;
            }

        } else {
            toReturn += `[p]
                            The big wolf struggles to stand fully upright inside of your cell,
                            his head bumping against the ceiling.
                            He casts his gaze downwards, staring at you like a disgruntled parent would at their child.
                            After grumbling something about how they should really make these carts a bit bigger, a faint smirk forms on his muzzle.
                            "[color=gblue]Rough night, huh?[/color]"
                            A rhetorical question.
                        [/p]
                        [p]
                            With a few heavy steps, the canine approaches your downed frame, the wood creaking underneath his broad footpaws.
                            Grabbing you by the back of your neck, he hoists you upright.
                            ${bindings ? `Reaching behind you, the wolf then speedily snaps the length of rope binding your hands together.` : ``}
                            "[color=gblue]On your feet.
                            There we go.
                            You have a big day ahead of you.[/color]"
                        [/p]`;
            if (bindings) {
                toReturn += `
                        [p]
                            You clutch each of your wrists in turn, rubbing them to ease the minor ropeburn.
                            After enjoying your newfound freedom for a moment, you tilt your muzzle upwards,
                            only to find yourself mere inches away from the lupine's naked groin,
                            a whiff of its potent scent already reaching your nostrils.
                            Fenrir stares down at you, a lecherous look on his face.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            "[color=gblue]So Guilliman [i]wasn't[/i] lying about you escaping your bindings.
                            Heh.
                            I assumed he caved your head in because he [b]felt[/b] like it.[/color]"
                            You frown at the thought of a remorseless bandit hitting you [i]for fun[/i].
                            "[color=gblue]Oh well.
                            It's not like he [b]needs[/b] full pay.
                            His family back home is rich as fuck.
                            Figure he's [i]rebelling[/i] against'em or something.
                            And he [b]did[/b] still damage my [i]most precious[/i] little lizard, now didn't he?[/color]"
                        [/p]
                        [p]
                            You can't help but sort of [i]agree[/i] with Fenrir on that last point.
                            There was absolutely [b]no[/b] reason for that horse to hit you.
                            Sure, you were planning to escape,
                            and you'd have bolted past him or tackled him or tried to [b]maim[/b] him if you hadn't frozen up,
                            but still.
                            What he did was plain [b]rude[/b].
                            He could have easily overpowered you [i]without[/i] using his club.
                        [/p]
                        [p]
                            "[color=gblue]Let me have a look.[/color]"
                            the canine continues.
                            Acting like it's no big deal, he grabs your head and tilts it downwards once more,
                            tugging you a little closer, nearly shoving your face into his crotch in the process.
                        [/p]
                        [p]
                            "[color=gblue]It's only a small [i]bruise[/i].
                            At worst, it'll swell up a bit.
                            If any of the buyers ask, tell them that's how you were captured.[/color]"
                        [/p]
                        [p]
                            His inspection finished, the wolf lets go of your cranium.
                            You quickly pull yourself away from his loins, expecting him to depart.
                            Fenrir, however, seems to have something else in mind.
                            He stares down at you, with a lecherous look on his face.
                        [/p]`;
            }
        }

        if (!nameKnown) {
            toReturn += `[p]
                            "[color=gblue]I had a few of my men ask around during the night.
                            The forest has loose lips, if you know where to pry.[/color]"
                            Huh?
                            He sent out [i]spies[/i]?
                            "[color=gblue]${name}.
                            That's your name, right?[/color]"
                            You stare up at the lupine in amazement and disbelief.
                            Barely [b]anyone[/b] around here knows your name.
                            He chuckles.
                            "[color=gblue]It fits you.[/color]"
                        [/p]`;
        }

        toReturn += `   [p]
                            The wolf moves one of his hands to his groin.
                            He boldly, brazenly and crudely cups the hefty, furred nuts that swing underneath his thick sheath,
                            propping them forwards as if to [i]entice[/i] you.
                            "[color=gblue]You've been a good ${dick ? `boy` : `girl`} so far, haven't you, ${name}?
                            At first, when I caught you, I thought you'd turn out to be a little [i]troublemaker[/i].
                            But you've proven me more than wrong.
                            I like that.[/color]"
                        [/p]
                        [p]
                            The canine licks his lips lustfully.
                            You hope he [b]isn't[/b] going to ask you to do what you [b]think[/b] he's going to ask of you.
                            "[color=gblue]From the moment ${fucked ? `I took your virginity` : `you told me you were a virgin`},
                            I [b]knew[/b] you were different from all the other kobolds.
                            Unlike the others, you are a [b]reasonable[/b] little creature.
                            You understand that not [i]all[/i] of life's problems can be solved through [b]violence[/b],
                            or by acting [b]uppity[/b] and [b]crying[/b] until you get your way.
                            I can respect that.[/color]"
                        [/p]
                        [p]
                            Oh boy.
                            You can [i]already[/i] tell where this is going, but your mind doesn't want to hear it.
                            You'd prefer to be blissfully ignorant until the words spill from Fenrir's own mouth.
                            "[color=gblue]I can make life [b]very[/b] easy for you, you know.
                            I can bring you some of the [b]good[/b] food at night,
                            I can tell the gang to [i]take it easy[/i] on you,
                            I can even make sure that [b]filthy[/b] human,
                            who I am [b]temporarily[/b] [i]indentured[/i] to,
                            won't lay even a [b]single[/b] hand on you.[/color]"
                        [/p]
                        [p]
                            A lot of benefits, but you're [b]very[/b] certain that all of this has a [b]price[/b].
                            Bandits like the wolf never hand out favors for [b]free[/b].
                            "[color=gblue]Let's face the facts here.
                            That [b]fat fuck[/b] thinks he can [i]sell[/i] you.
                            But both you and me know that there's [b]plenty[/b] of slaves to go around.
                            It's an [i]oversaturated[/i] market.[/color]"
                            Fancy words, for a brigand.
                            Then again, he [i]is[/i] their leader.
                            "[color=gblue]And, as you've no-doubt overheard, he only has two tries to auction you off.
                            After that, you'll be [b]mine[/b].[/color]"
                        [/p]
                        [p]
                            You repress the urge to tell the lupine to get to the point already.
                            It seems you've [i]placated[/i] him, and you'd hate to break his good demeanor by telling him to rush things along.
                            "[color=gblue]I won't lie.
                            You're going to be fucked, gangbanged, spitroasted and ran a train on by every single last member of my group.
                            Including the dog.
                            But that doesn't need to be a [i]bad[/i] thing.
                            If you play along, if you keep on being as nice as you've been thus far, then none of us will have any reason to treat you poorly.
                            And besides keeping us pleased, you won't have a thing in the world to worry about.
                            If you prove your worth, I might even see fit to [i]eventually[/i] consider [b]recruiting[/b] you, somewhere down the road.
                            Make you a [b]proper[/b] part of our band.[/color]"
                        [/p]
                        [p]
                            You don't get it.
                            Is Fenrir offering you a [i]job[/i]?
                            "[color=gblue]The Furskin Mountain-Crew could use a small and nimble thief like yourself.
                            And by the time we're [i]done[/i] with you, you'll be a [b]master[/b] of seduction and other [i]less common[/i] talents.[/color]"
                            You don't like the sound of that last part.
                            "[color=gblue]All I ask in return, is a simple sign of subservience.
                            You won't [i]officially[/i] be [b]my[/b] slave until the second auction fails,
                            but that doesn't mean we can't start our [i]friendship[/i] a little [i]early[/i], now does it?[/color]"
                        [/p]
                        [p]
                            You knew it.
                            The wolf may use fancy words and beat around the bush a lot, but in the end he's no better than a simple highwayman,
                            only instead of coercing money out of you, he's trying to extort your very [i]diginity[/i] itself.
                            "[color=gblue]So how about it, [b]friend[/b]?
                            All I want is a little good morning kiss.
                            Right on top of the balls that
                            ${fucked ? `squirted their contents all over your body` : `you denied release`} yesterday.[/color]"
                        [/p]
                        [p]
                            As if to accentuate his statement, the wolf taps his furred nuts,
                            rolling his hips forwards ever so slightly to hold them a little closer to your mouth.
                            Part of your mind is honestly [i]considering[/i] Fenrir's offer.
                            A life without worries [i]does[/i] seem [b]compelling[/b], to say the least.
                            And you [b]really[/b] don't want that human to get anywhere close to you.
                            But then again, you don't want to degrade yourself any more than you already have.
                            Besides, you plan on escaping long before you become the lupine's property.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled} style="width: 190px;">Do nothing</button>`;

        var tooltip = global.parser.parse(`Remain in place. Do not act. Let the canine draw his own conclusions.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled} style="width: 190px;">Decline</button>`;

        var tooltip = global.parser.parse(`You don't want to be anywhere [b]near[/b] the wolf's loins.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled} style="width: 190px;">Insult</button>`;

        var tooltip = global.parser.parse(`Who the hell does Fenrir think he is?! Tell him to get lost.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }
    /**
     * Builds and returns the choice button.
     */
    function buildFourthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === fourthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fourthId}" ${disabled} style="width: 190px;">Attack</button>`;

        var tooltip = global.parser.parse(`Fight the wolf. Freedom or death!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 145px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFifthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === fifthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fifthId}" ${disabled} style="width: 190px;">Reluctantly kiss</button>`;

        var tooltip = global.parser.parse(`Ugh. Do you [i]really[/i] have to do this?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSixthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === sixthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${sixthId}" ${disabled} style="width: 190px;">Eagerly kiss</button>`;

        var tooltip = global.parser.parse(`You get to kiss sweaty wolf nuts and live a life of nonstop sex? You wish you'd met him earlier!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 225px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += `<div class="choicerow" style="margin-bottom: 10px;">`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        choice += `<div class="choicerow">`;
        choice += buildFourthChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFifthChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSixthChoice();
        choice += `</div>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId), $("#" + fourthId), $("#" + fifthId), $("#" + sixthId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFourthListener() {
        $("#" + fourthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fourthChoice);
            loadNextPart(fourthNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFifthListener() {
        $("#" + fifthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fifthChoice);
            loadNextPart(fifthNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSixthListener() {
        $("#" + sixthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(sixthChoice);
            loadNextPart(sixthNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        part += fourthNext.build();
        part += fifthNext.build();
        part += sixthNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();
        fourthNext.registerListeners();
        fifthNext.registerListeners();
        sixthNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
        registerFourthListener();
        registerFifthListener();
        registerSixthListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\SlaveNoonTwoSimpleOffer.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\SlaveNoonTwoMakeUpOffer.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoMakeUpOffer(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.offer;
    var changeController = ctrl.offer;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";
    var thirdId = arguments.callee.name + "ButtonThree";
    var fourthId = arguments.callee.name + "ButtonFour";
    var fifthId = arguments.callee.name + "ButtonFive";
    var sixthId = arguments.callee.name + "ButtonSix";

    var firstNext = new SlaveNoonTwoGoodNothing(ctrl);
    var secondNext = new SlaveNoonTwoGoodDecline(ctrl);
    var thirdNext = new SlaveNoonTwoGoodInsult(ctrl);
    var fourthNext = new SlaveNoonTwoGoodAttack(ctrl);
    var fifthNext = new SlaveNoonTwoGoodReluctant(ctrl);
    var sixthNext = new SlaveNoonTwoGoodEager(ctrl);

    var firstChoice = SlaveNoonTwoOfferChoice.NOTHING;
    var secondChoice = SlaveNoonTwoOfferChoice.DECLINE;
    var thirdChoice = SlaveNoonTwoOfferChoice.INSULT;
    var fourthChoice = SlaveNoonTwoOfferChoice.ATTACK;
    var fifthChoice = SlaveNoonTwoOfferChoice.RELUCTANT;
    var sixthChoice = SlaveNoonTwoOfferChoice.EAGER;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        global.kobold.history.shared.enslavement.noonTwo.makeUpOffer = true;

        var fucked = global.kobold.history.shared.enslavement.captureFucked;
        var enjoyed = global.kobold.history.shared.enslavement.captureEnjoyed;
        var bindings = !global.kobold.history.shared.enslavement.nightOne.escapedBindings;
        var nameKnown = global.kobold.history.shared.enslavement.nightOne.nameStated;
        var dick = ctrl.misc.dick();
        var name = ctrl.misc.name();

        var toReturn = ``;

        if (ctrl.arrival.get() === SlaveNoonTwoArrivalChoice.LOOKOUT) {
            toReturn += `[p]
                            The big wolf struggles to stand fully upright inside of your cell,
                            his head bumping against the ceiling.
                            After grumbling something about how they should really make these carts a bit bigger, a smirk forms on his muzzle.
                            "[color=gblue]Already up and about, are you?
                            ${!bindings ? `[i]Impressive[/i].` : `[i]Good[/i].`}
                            You have a big day ahead of you.[/color]"
                        [/p]`;
            if (bindings) {
                toReturn += `
                        [p]
                            "[color=gblue]Face away for a second.
                            Let me get that rope off of you.[/color]"
                            While you don't really [i]want[/i] to [b]turn around[/b] for the fully naked lupine,
                            the premise of having your hands freed sounds too good to pass up.
                            Besides, he's clearly [i]not[/i] aroused, so you doubt he'll [i]abuse[/i] the situation.
                        [/p]
                        [p]
                            Keeping your wary eyes on the wolf for as long as you can manage, you slowly turn your back towards him.
                            With a few heavy steps, the canine approaches you from behind, the wood creaking underneath his broad footpaws.
                            He comes to a halt directly behind you, but is nice enough to leave half a foot orso in between your body and his.
                            The bandit chief reaches for your wrists, still bound behind your back, and with a simple [b]slice[/b] of a single claw,
                            he severs the cord keeping your hands together.
                        [/p]
                        [p]
                            You clutch each of your wrists in turn, rubbing them to ease the minor ropeburn.
                            After giving you a moment to enjoy your newfound freedom, the well-built man grabs your shoulder,
                            and turns your body around once more, your muzzle ending up mere inches away from his bare groin,
                            a whiff of its potent scent already reaching your nostrils.
                            Fenrir stares down at you, a lecherous look on his face.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            "[color=gblue]So Guilliman [i]wasn't[/i] lying about you escaping your bindings.
                            Heh.
                            I assumed he caved your head in because he [b]felt[/b] like it.[/color]"
                            You frown at the thought of a remorseless bandit hitting you [i]for fun[/i].
                            "[color=gblue]Oh well.
                            It's not like he [b]needs[/b] full pay.
                            His family back home is rich as fuck.
                            Figure he's [i]rebelling[/i] against'em or something.
                            And he [b]did[/b] still damage my [i]most precious[/i] little lizard, now didn't he?[/color]"
                        [/p]
                        [p]
                            You can't help but sort of [i]agree[/i] with Fenrir on that last point.
                            There was absolutely [b]no[/b] reason for that horse to hit you.
                            Sure, you were planning to escape,
                            and you'd have bolted past him or tackled him or tried to [b]maim[/b] him if you hadn't frozen up,
                            but still.
                            What he did was plain [b]rude[/b].
                            He could have easily overpowered you [i]without[/i] using his club.
                        [/p]
                        [p]
                            "[color=gblue]Let me have a look.[/color]"
                            the canine continues, crossing the distance between you with a few heavy steps, coming to rest about half a foot away from you,
                            your muzzle mere inches away from his bare groin,
                            a whiff of its potent scent already reaching your nostrils.
                            Acting like it's no big deal, he grabs your head and tilts it downwards,
                            nearly shoving your face into his crotch in the process.
                        [/p]
                        [p]
                            "[color=gblue]It's only a small [i]bruise[/i].
                            At worst, it'll swell up a bit.
                            If any of the buyers ask, tell them that's how you were captured.[/color]"
                        [/p]
                        [p]
                            His inspection finished, the wolf lets go of your cranium.
                            You quickly pull yourself away from his loins, expecting him to depart.
                            Fenrir, however, seems to have something else in mind.
                            He stares down at you, with a lecherous look on his face.
                        [/p]`;
            }

        } else {
            toReturn += `[p]
                            The big wolf struggles to stand fully upright inside of your cell,
                            his head bumping against the ceiling.
                            He casts his gaze downwards, staring at you like a disgruntled parent would at their child.
                            After grumbling something about how they should really make these carts a bit bigger, a faint smirk forms on his muzzle.
                            "[color=gblue]Rough night, huh?[/color]"
                            A rhetorical question.
                        [/p]
                        [p]
                            With a few heavy steps, the canine approaches your downed frame, the wood creaking underneath his broad footpaws.
                            Grabbing you by the back of your neck, he hoists you upright.
                            ${bindings ? `Reaching behind you, the wolf then speedily snaps the length of rope binding your hands together.` : ``}
                            "[color=gblue]On your feet.
                            There we go.
                            You have a big day ahead of you.[/color]"
                        [/p]`;
            if (bindings) {
                toReturn += `
                        [p]
                            You clutch each of your wrists in turn, rubbing them to ease the minor ropeburn.
                            After enjoying your newfound freedom for a moment, you tilt your muzzle upwards,
                            only to find yourself mere inches away from the lupine's naked groin,
                            a whiff of its potent scent already reaching your nostrils.
                            Fenrir stares down at you, a lecherous look on his face.
                        [/p]`;
            } else {
                toReturn += `
                        [p]
                            "[color=gblue]So Guilliman [i]wasn't[/i] lying about you escaping your bindings.
                            Heh.
                            I assumed he caved your head in because he [b]felt[/b] like it.[/color]"
                            You frown at the thought of a remorseless bandit hitting you [i]for fun[/i].
                            "[color=gblue]Oh well.
                            It's not like he [b]needs[/b] full pay.
                            His family back home is rich as fuck.
                            Figure he's [i]rebelling[/i] against'em or something.
                            And he [b]did[/b] still damage my [i]most precious[/i] little lizard, now didn't he?[/color]"
                        [/p]
                        [p]
                            You can't help but sort of [i]agree[/i] with Fenrir on that last point.
                            There was absolutely [b]no[/b] reason for that horse to hit you.
                            Sure, you were planning to escape,
                            and you'd have bolted past him or tackled him or tried to [b]maim[/b] him if you hadn't frozen up,
                            but still.
                            What he did was plain [b]rude[/b].
                            He could have easily overpowered you [i]without[/i] using his club.
                        [/p]
                        [p]
                            "[color=gblue]Let me have a look.[/color]"
                            the canine continues.
                            Acting like it's no big deal, he grabs your head and tilts it downwards once more,
                            tugging you a little closer, nearly shoving your face into his crotch in the process.
                        [/p]
                        [p]
                            "[color=gblue]It's only a small [i]bruise[/i].
                            At worst, it'll swell up a bit.
                            If any of the buyers ask, tell them that's how you were captured.[/color]"
                        [/p]
                        [p]
                            His inspection finished, the wolf lets go of your cranium.
                            You quickly pull yourself away from his loins, expecting him to depart.
                            Fenrir, however, seems to have something else in mind.
                            He stares down at you, with a lecherous look on his face.
                        [/p]`;
            }
        }

        if (!nameKnown) {
            toReturn += `[p]
                            "[color=gblue]I had a few of my men ask around during the night.
                            The forest has loose lips, if you know where to pry.[/color]"
                            Huh?
                            He sent out [i]spies[/i]?
                            "[color=gblue]${name}.
                            That's your name, right?[/color]"
                            You stare up at the lupine in amazement and disbelief.
                            Barely [b]anyone[/b] around here knows your name.
                            He chuckles.
                            "[color=gblue]It fits you.[/color]"
                        [/p]`;
        }

        toReturn += `   [p]
                            The wolf moves one of his hands to his groin.
                            He boldly, brazenly and crudely cups the hefty, furred nuts that swing underneath his thick sheath,
                            propping them forwards as if to [i]entice[/i] you.
                            "[color=gblue]Look, ${name}, I know we got off on a bad start.
                            ${fucked ? `In the forest, ${enjoyed ? `you sounded like you'd make a good slave,
                            but then when we got to the camp, your enthusiasm vanished.` : `I had to [b]force[/b] myself upon you,
                            and then when we got back to camp, you refused to follow even the simplest of commands.`}` : `First you blueballed me in the forest,
                            and then you failed to obey some simple commands in front of the caravan master.`}
                            Granted, you're not [i]as[/i] bad as I first made you out to be, but still, you embarrassed me in front of that [b]moronic human[/b].
                            And that's not something I can appreciate.[/color]"
                        [/p]
                        [p]
                            The canine licks his lips lustfully, his stern glare contrasting sharply with his body language.
                            For a moment, you ${enjoyed ? `reckon` : `fear`} he might try to force himself on you again,
                            but you don't know.
                            He's sending you some seriously mixed signals.
                            "[color=gblue]From the moment ${fucked ? `I took your virginity` : `you told me you were a virgin`},
                            I [b]knew[/b] you were different from all the other kobolds.
                            Unlike the others, you understand that not [i]all[/i] of life's problems can be solved through [b]violence[/b],
                            or by [b]crying[/b] until you get your way.
                            But being [b]uppity[/b] and [b]refusing[/b] to cooperate is [i]almost[/i] every bit as bad.[/color]"
                        [/p]
                        [p]
                            You don't understand.
                            Are you being scolded, or is he about to ask you to suck his dick or something?
                            "[color=gblue]I can make life [b]a lot[/b] easier for you, you know.
                            I can bring you some [i]actual[/i] food at night,
                            I can tell the gang to [i]take it easy[/i] on you,
                            I can even ask that [b]filthy[/b] human,
                            who I am [b]temporarily[/b] [i]indentured[/i] to,
                            to keep things relatively [i]tame[/i] with you.[/color]"
                        [/p]
                        [p]
                            A bunch of benefits, but you're [b]very[/b] certain that all of this has a [b]price[/b].
                            Bandits like the wolf never hand out favors for [b]free[/b].
                            "[color=gblue]Let's face the facts here.
                            That [b]fat fuck[/b] thinks he can [i]sell[/i] you.
                            But both you and me know that there's [b]plenty[/b] of slaves to go around.
                            It's an [i]oversaturated[/i] market.[/color]"
                            Fancy words, for a brigand.
                            Then again, he [i]is[/i] their leader.
                            "[color=gblue]You aren't the best product around, either.
                            People would much rather have a well-trained servant than an uppity [b]${dick ? `jerk` : `cunt`}[/b].
                            And, as you've no-doubt overheard, the caravan master only has two tries to auction you off.
                            After that, you'll be [b]mine[/b].[/color]"
                        [/p]
                        [p]
                            You repress the urge to tell the lupine to get to the point already.
                            It seems you've [i]irked[/i] him, and you'd hate to upset him further.
                            "[color=gblue]I won't lie.
                            You're going to be fucked, gangbanged, spitroasted and ran a train on by every single last member of my group.
                            Including the dog.
                            This [b]is[/b] going to happen, whether you like it or not.
                            But if you play along, then none of us will have any reason to treat you poorly.
                            If you don't, well...
                            Let's just say we could always use a punching bag for training on.[/color]"
                        [/p]
                        [p]
                            You get the feeling he's not kidding about that threat.
                            "[color=gblue]The Furskin Mountain-Crew will treat you fairly, for a kobold.
                            You'll be fed scraps, you'll have a tent or a cart or at worst a dog-house to sleep in,
                            and you'll be kept safe [i]for the most part[/i].[/color]"
                            You don't like the sound of that last bit.
                            "[color=gblue]All I ask in return, is a simple sign of subservience.
                            You won't [i]officially[/i] be [b]my[/b] slave until the second auction fails.
                            Consider this your [i]probational[/i] period.[/color]"
                        [/p]
                        [p]
                            You knew it.
                            The wolf may use fancy words and beat around the bush a lot, but in the end he's no better than a simple highwayman,
                            only instead of coercing money out of you, he's trying to extort your very [i]diginity[/i] itself.
                            "[color=gblue]So how about it, [b]slave[/b]?
                            All I want is a little good morning kiss.
                            Right on top of the balls that
                            ${fucked ? `squirted their contents all over your body` : `you denied release`} yesterday.[/color]"
                        [/p]
                        [p]
                            As if to accentuate his statement, the wolf taps his furred nuts,
                            rolling his hips forwards ever so slightly to hold them a little closer to your mouth.
                            Part of your mind is honestly [i]considering[/i] Fenrir's offer.
                            Should both the auctions fail, at least you won't end up as a [i]training dummy[/i].
                            And it's just a simple kiss.
                            But then again, you don't want to degrade yourself any more than you already have.
                            Besides, you plan on escaping long before you become the lupine's property.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled} style="width: 190px;">Do nothing</button>`;

        var tooltip = global.parser.parse(`Remain in place. Do not act. Let the canine draw his own conclusions.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled} style="width: 190px;">Decline</button>`;

        var tooltip = global.parser.parse(`You don't want to be anywhere [b]near[/b] the wolf's loins.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 175px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildThirdChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === thirdChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${thirdId}" ${disabled} style="width: 190px;">Insult</button>`;

        var tooltip = global.parser.parse(`Who the hell does Fenrir think he is?! Tell him to get lost.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }
    /**
     * Builds and returns the choice button.
     */
    function buildFourthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === fourthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fourthId}" ${disabled} style="width: 190px;">Attack</button>`;

        var tooltip = global.parser.parse(`Fight the wolf. Freedom or death!`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 145px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFifthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === fifthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${fifthId}" ${disabled} style="width: 190px;">Reluctantly kiss</button>`;

        var tooltip = global.parser.parse(`Ugh. Do you [i]really[/i] have to do this?`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSixthChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === sixthChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${sixthId}" ${disabled} style="width: 190px;">Eagerly kiss</button>`;

        var tooltip = global.parser.parse(`You really don't want to get on this guy's bad side. Better show him you mean it.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 225px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice" style="width: 190px;">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += `<div class="choicerow" style="margin-bottom: 10px;">`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildThirdChoice();
        choice += `</div>`;
        choice += `<div class="choicerow">`;
        choice += buildFourthChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildFifthChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSixthChoice();
        choice += `</div>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId), $("#" + thirdId), $("#" + fourthId), $("#" + fifthId), $("#" + sixthId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerThirdListener() {
        $("#" + thirdId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(thirdChoice);
            loadNextPart(thirdNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFourthListener() {
        $("#" + fourthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fourthChoice);
            loadNextPart(fourthNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerFifthListener() {
        $("#" + fifthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(fifthChoice);
            loadNextPart(fifthNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSixthListener() {
        $("#" + sixthId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(sixthChoice);
            loadNextPart(sixthNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        part += thirdNext.build();
        part += fourthNext.build();
        part += fifthNext.build();
        part += sixthNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();
        thirdNext.registerListeners();
        fourthNext.registerListeners();
        fifthNext.registerListeners();
        sixthNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
        registerThirdListener();
        registerFourthListener();
        registerFifthListener();
        registerSixthListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\SlaveNoonTwoMakeUpOffer.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\SlaveNoonTwoGoodOffer.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoGoodOffer(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;

    var toFail = new SlaveNoonTwoMakeUpOffer(ctrl);
    var toSuccess = new SlaveNoonTwoSimpleOffer(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return global.kobold.history.shared.enslavement.nightOne.complianceRating > 6;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\good\SlaveNoonTwoGoodOffer.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\SlaveNoonTwoProposition.js

/**
 * Proposition section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoProposition(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = ctrl.preBlackmail;

    var toFail = new SlaveNoonTwoBadOffer(ctrl);
    var toSuccess = new SlaveNoonTwoGoodOffer(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Checks whether the choice succeeded or failed.
     */
    function successCheck() {
        return global.kobold.history.shared.enslavement.nightOne.complianceRating > 3;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        if (successCheck()) {
            part += toSuccess.build();
        } else {
            part += toFail.build();
        }
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (successCheck()) {
            toSuccess.registerListeners();
        } else {
            toFail.registerListeners();
        }
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\proposition\SlaveNoonTwoProposition.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\arrival\SlaveNoonTwoLookOut.js

/**
 * Arrival section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoLookOut(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoArrivalChoice.LOOKOUT;
    var reqController = ctrl.arrival;
    var subController = ctrl.preBlackmail;
    var changeController = ctrl.preBlackmail;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.proposition;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = ``;
        var bindings = !global.kobold.history.shared.enslavement.nightOne.escapedBindings;

        if (global.kobold.history.shared.enslavement.nightOne.knockedOut) {
            toReturn += `[p]
                            ${bindings ? `You try to cover your eyes with your hands,
                            but find them still bound from the night before.
                            Instead, you roll over onto your side,
                            hiding your visage from the beaming sunlight,
                            letting it warm your back instead.` : `You cover your eyes with your hands,
                            protecting your delicate sight from the beaming sunlight.`}
                        [/p]
                        [p]
                            Despite having been out for what you can only presume to be over half a day,
                            you still don't feel very refreshed.
                            In fact, you feel more drained than [i]before[/i] you passed out.
                            You could use a few extra hours of peace and quiet,
                            but since you're awake, you might as well start trying to figure out where you are,
                            and more importantly: how you're going to [b]escape[/b].
                        [/p]
                        [p]
                            Your heavy eyelids slowly open.
                            ${bindings ? `You wish you could wipe some of the crud from your eyes.` : `You wipe some of the crud from your eyes.`}
                            Your vision is blurry, but it rapidly sharpens.
                            Despite not feeling like you're ready to face the light,
                            you decide to give it a go anyway.
                        [/p]
                        [p]
                            ${bindings ? `Rolling onto your back once again` : `Removing your hands from your head`},
                            you accidentally stare directly into the sun,
                            scorching your retinas for a couple of milliseconds before you manage to avert your gaze.
                            Your eyes hurt.
                            Your headache flares up.
                            This was a mistake.
                        [/p]
                        [p]
                            Keeping your eyelids closed, you slowly scramble to your feet${bindings ? `, hindered, but not stopped by your bindings` : ``}.
                            After resting against the side of the cart for a moment, you slowly open your eyes once again.
                            Stepping on your tippy-toes, you peer through the iron bars lining the wooden walls.
                        [/p]
                        [p]
                            The brightly lit outside world is painful to look at, but not nearly as harmful as the sun itself.
                            You see a few shabby houses alongside a dirt road, followed by large stretches of farmland in their backyards.
                            You recognize this area.
                            You've been here before.
                            It's a place known simply as [i]the hamlet[/i],
                            a small village close to the human capital of Varanar.
                        [/p]
                        [p]
                            An old friend of yours lives here.
                            Seki, the alchemist-owl, who also happens to happily buy whatever stolen goods you bring him.
                            Besides being your [i]fence[/i], he's also the closest thing to a [i]father[/i] you've ever had.
                            Maybe [b]he[/b] can help you get out of this [b]mess[/b].
                        [/p]
                        [p]
                            A subtle [i]click[/i] resonates throughout your cell.
                            The door swings open widly, banging against the exterior wall,
                            the loud noise like a [b]dagger[/b] sinking into your skull, worsening your killer headache.
                            Turning to face the entryway, you witness Fenrir, the wolf that caught you the other day,
                            entering the jail-like carriage.
                            He is completely [b]naked[/b].
                        [/p]
                        [p]
                            "[color=gblue]Good morning, [b]lizard[/b]![/color]" the wolf booms,
                            his loud voice unknowingly [b]torturing[/b] your sleep-deprived brain.
                            You sigh internally.
                            You wish [b]all[/b] of yesterday had been part of your nightmare.
                            But alas, no such luck.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You try to cover your eyes with your hands,
                            but find them still bound from the night before.
                            Instead, you roll over onto your side,
                            hiding your visage from the beaming sunlight,
                            letting it warm your back instead.
                        [/p]
                        [p]
                            Despite having slept for what you can only presume be over half a day,
                            you still don't feel very refreshed.
                            In fact, you feel more drained than [i]before[/i] your nap.
                            You could use a few extra hours of peace and quiet,
                            but since you're already awake, you might as well start trying to figure out where you are,
                            and more importantly: how you're going to [b]escape[/b].
                        [/p]
                        [p]
                            Your heavy eyelids slowly open.
                            You wish you could wipe some of the crud from your eyes.
                            Your vision is blurry, but it rapidly sharpens.
                            Despite not feeling like you're ready to face the light,
                            you decide to give it a go anyway.
                        [/p]
                        [p]
                            Rolling onto your back once again,
                            you accidentally stare directly into the sun,
                            scorching your retinas for a couple of milliseconds before you manage to avert your gaze,
                            hurting your eyes.
                            Your budding headache flares up.
                            This was a mistake.
                        [/p]
                        [p]
                            Keeping your eyelids closed, you slowly scramble to your feet, hindered, but not stopped by your bindings.
                            After resting against the side of the cart for a moment, you slowly open your eyes once again.
                            Stepping on your tippy-toes, you gaze through the iron bars lining the wooden walls.
                        [/p]
                        [p]
                            The brightly lit outside world is painful to look at, but not nearly as harmful as the sun itself.
                            You see a few shabby houses alongside a dirt road, followed by large stretches of farmland in their backyards.
                            You recognize this place.
                            You've been here before.
                            While asleep, you were taken to a place known simply as [i]the hamlet[/i],
                            a small village close to the human capital of Varanar.
                        [/p]
                        [p]
                            An old friend of yours lives here.
                            Seki, the alchemist-owl, who also happens to happily buy whatever stolen goods you bring him.
                            Besides being your [i]fence[/i], he's also the closest thing to a [i]father[/i] you've ever had.
                            Maybe [b]he[/b] can help you get out of this [b]mess[/b].
                        [/p]
                        [p]
                            A subtle [i]click[/i] resonates throughout your cell.
                            The door swings open widly, banging against the exterior wall,
                            the loud noise like a [b]dagger[/b] sinking into your skull, worsening your headache.
                            Turning to face the entryway, you witness Fenrir, the wolf that caught you the other day,
                            entering the jail-like carriage.
                            He is completely [b]naked[/b].
                        [/p]
                        [p]
                            "[color=gblue]Good morning, [b]lizard[/b]![/color]" the wolf booms,
                            his loud voice unknowingly [b]torturing[/b] your sleep-deprived brain.
                            You sigh internally.
                            You wish [b]all[/b] of yesterday had been part of your nightmare.
                            But alas, no such luck.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\arrival\SlaveNoonTwoLookOut.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\arrival\SlaveNoonTwoRest.js

/**
 * Arrival section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoRest(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = SlaveNoonTwoArrivalChoice.REST;
    var reqController = ctrl.arrival;
    var subController = ctrl.preBlackmail;
    var changeController = ctrl.preBlackmail;

    var continueId = arguments.callee.name + "ButtonOne";
    var next = ctrl.fragments.proposition;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = ``;
        var bindings = !global.kobold.history.shared.enslavement.nightOne.escapedBindings;

        if (global.kobold.history.shared.enslavement.nightOne.knockedOut) {
            toReturn += `[p]
                            You wearily roll over onto your side, hiding your visage from the beaming sunlight, letting it warm your back instead.
                            Despite having been out for what you can only presume to be over half a day,
                            you still don't feel very refreshed.
                            In fact, you feel more drained than [i]before[/i] you passed out.
                            You could use a few extra hours of peace and quiet.
                            You can worry about escaping this cell later.
                        [/p]
                        [p]
                            Right as you could feel reality starting to slip away once again,
                            a subtle [i]click[/i] resonates throughout your cell.
                            The wooden door widly swings open, banging against the exterior wall,
                            the loud noise like a [b]dagger[/b] sinking into your skull, worsening your killer headache.
                            Footsteps.
                            Your half-conscious mind distantly realizes someone is entering the cart.
                        [/p]
                        [p]
                            [big][color=gblue][b]Rise and shine, slave![/b][/color][/big]
                        [/p]
                        [p]
                            The booming voice is [b]agonizingly[/b] loud,
                            your head feeling like it's about to [b]burst[/b].
                            ${bindings ? `You instinctively try to move your hands to your forehead,
                            but find them still bound from the night before.` : `You instinctively move your hands to your forehead
                            to [b]try[/b] to cope with the horrible sensation.`}
                            Your heavy eyelids slowly open.
                            In your blurry vision you can see a pair of grey-furred feet standing in the doorway.
                            You slept so poorly, you almost feel like you have a bad hangover.
                        [/p]
                        [p]
                            You try to [b]beg[/b] for just a few more minutes of rest, but nothing intelligent escapes your muzzle.
                            Instead, you let out a mixture between a whine, a groan and a growl.
                            "[color=gblue]I said [b]get up[/b], lizard.[/color]" the wolf repeats.
                            In an effort to force you awake, he repeatedly bangs his foot against the floor,
                            unknowingly [b]torturing[/b] your sleep-deprived brain.
                        [/p]
                        [p]
                            Almost reduced to tears from the intense pain, your aching mind snaps to life.
                            While your sight is still hazy, you open your eyes further, and tilt your gaze up to meet his.
                            Fenrir.
                            That was his name.
                            The wolf that caught you the other day.
                            And he's [i]naked[/i], this time around.
                            You sigh internally.
                            You wish [b]all[/b] of yesterday had been part of your nightmare.
                            But alas, no such luck.
                        [/p]`;
        } else {
            toReturn += `[p]
                            You wearily roll over onto your side, hiding your visage from the beaming sunlight, letting it warm your back instead.
                            Despite having slept for what you can only presume be over half a day,
                            you still don't feel very refreshed.
                            In fact, you feel more drained than [i]before[/i] you passed out.
                            You could use a few extra hours of peace and quiet.
                            You can worry about escaping this cell later.
                        [/p]
                        [p]
                            Right as you could feel reality starting to slip away once again,
                            a subtle [i]click[/i] resonates throughout your cell.
                            The wooden door widly swings open, banging against the exterior wall,
                            the loud noise like a [b]dagger[/b] piercing your unconsciousness, worsening your headache.
                            Footsteps.
                            Your half-asleep mind distantly realizes someone is entering the cart.
                        [/p]
                        [p]
                            [big][color=gblue][b]Rise and shine, slave![/b][/color][/big]
                        [/p]
                        [p]
                            The booming voice is [b]agonizingly[/b] loud,
                            your head feeling like worse with every passing second.
                            You instinctively try to move your hands to your forehead,
                            but find them still bound from the night before.
                            Your heavy eyelids slowly open.
                            In your blurry vision you can see a pair of grey-furred feet standing in the doorway.
                            You slept so poorly, you almost feel like you have a hangover.
                        [/p]
                        [p]
                            You try to [b]beg[/b] for just a few more minutes of rest, but nothing intelligent escapes your muzzle.
                            Instead, you let out a mixture between a whine, a groan and a growl.
                            "[color=gblue]I said [b]get up[/b], lizard.[/color]" the wolf repeats.
                            In an effort to force you awake, he repeatedly bangs his foot against the floor,
                            unknowingly [b]torturing[/b] your sleep-deprived brain.
                        [/p]
                        [p]
                            Almost reduced to tears from the intensifying pain, your aching mind snaps to life.
                            While your sight is still hazy, you open your eyes a further, and tilt your gaze up to meet his.
                            Fenrir.
                            That was his name.
                            The wolf that caught you the other day.
                            And he's [i]naked[/i], this time around.
                            You sigh internally.
                            You wish [b]all[/b] of yesterday had been part of your nightmare.
                            But alas, no such luck.
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildNextChoice() {
        var disabled = subController.continued() ? "disabled" : "";
        var isSelected = subController.continued();
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${continueId}" ${disabled}>Continue</button>`;
        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer">`;
        choice += buildNextChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerNextListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.continue();
            loadNextPart(next);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += next.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        next.registerListeners();

        if (subController.continued()) {
            return;
        }

        registerNextListener();
    };
}

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\arrival\SlaveNoonTwoRest.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\arrival\SlaveNoonTwoArrival.js

/**
 * Arrival section of Enslavement, Noon Two.
 * @param {SlaveNoonTwoController} controller The Enslavement, Noon Two controller.
 */
function SlaveNoonTwoArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.arrival;
    var changeController = ctrl.arrival;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new SlaveNoonTwoLookOut(ctrl);
    var secondNext = new SlaveNoonTwoRest(ctrl);

    var firstChoice = SlaveNoonTwoArrivalChoice.LOOKOUT;
    var secondChoice = SlaveNoonTwoArrivalChoice.REST;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = ``;

        if (global.kobold.history.shared.enslavement.nightOne.knockedOut) {
            toReturn += `[p]
                            After getting knocked out, you find yourself trapped in ever-more-troubled slumber.
                            Bad dreams come and go,
                            some of the nightmares [i]lifelike[/i] enough to make you toss and tumble around the floor of your wooden cell.
                            The distant sensation of the caravan [i]moving[/i] isn't helping either.
                            Like an insidious poison, it seeps into your mind to corrupt and twist your feverish near-hallucinations further.
                        [/p]
                        [p]
                            The slow, rhythmic clopping of hooves,
                            the slight creaking of wood,
                            the clattering of the accompanying guards' armor,
                            and the lively sounds of the awakening forest,
                            all of these make your brain think back to the act that got you into this situation in the first place.
                            Robbing the transport.
                            That's where it all went wrong.
                        [/p]
                        [p]
                            You dream of how much [b]worse[/b] things could have gone.
                            How the ruffians protecting the transport could have seen you and killed you on the spot.
                            How they might have [b]tortured[/b] you to [b]death[/b].
                            Every scene seems more lifelike than the last.
                            Getting impaled by a spear.
                            Having your loins [b]mutilated[/b] and your tail cut off.
                            Them flaying your scales and feeding your flesh to their dog.
                            Luckily, right as the imaginary caravan master is about to bring a knife to bear on your forcibly held-open eyes,
                            one of the wheels of the cart you are in hits a rock, and the bump is enough to tear you from your hellish imagination.
                        [/p]
                        [p]
                            You let out an exasperated groan.
                            Your head [b]hurts[/b].
                            Your mind slowly rouses from unconsciousness.
                            A bright light is shining into your still-closed eyes, worsening your heavy migraine.
                            The sun?
                            Feels hot enough to be [b]noon[/b].
                            Is it [b]that[/b] late?
                        [/p]`;
        } else {
            toReturn += `[p]
                            Drifting asleep, you find yourself trapped in ever-more-troubled slumber.
                            Bad dreams come and go,
                            some of the nightmares [i]lifelike[/i] enough to make you toss and tumble around the floor of your wooden cell.
                            The distant sensation of the caravan [i]moving[/i] isn't helping either.
                            Like an insidious poison, it seeps into your mind to corrupt and twist your feverish near-hallucinations further.
                        [/p]
                        [p]
                            The slow, rhythmic clopping of hooves,
                            the slight creaking of wood,
                            the clattering of the accompanying guards' armor,
                            and the lively sounds of the awakening forest,
                            all of these make your brain think back to the act that got you into this situation in the first place.
                            Robbing the transport.
                            That's where it all went wrong.
                        [/p]
                        [p]
                            You dream of how much [b]worse[/b] things could have gone.
                            How the ruffians protecting the transport could have seen you and killed you on the spot.
                            How they might have [b]tortured[/b] you to [b]death[/b].
                            Every scene seems more lifelike than the last.
                            Getting impaled by a spear.
                            Having your loins [b]mutilated[/b] and your tail cut off.
                            Them flaying your scales and feeding your flesh to their dog.
                            Luckily, right as the imaginary caravan master is about to bring a knife to bear on your forcibly held-open eyes,
                            one of the wheels of the cart you are in hits a rock, and the bump is enough to tear you from your hellish imagination.
                        [/p]
                        [p]
                            You let out an exasperated groan.
                            You have a slight headache.
                            Your mind slowly rouses from unconsciousness.
                            A bright light is shining into your still-closed eyes, worsening your budding migraine.
                            The sun?
                            Feels hot enough to be [b]noon[/b].
                            Is it [b]that[/b] late?
                        [/p]`;
        }

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Look outside</button>`;

        var tooltip = global.parser.parse(`Rise and shine. Find out where you are.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 150px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Rest more</button>`;

        var tooltip = global.parser.parse(`It's too early. You need some more sleep.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("hamlet_day");
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += (changeController && changeController.change()) || "";
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        registerSecondListener();
    };
}

var SlaveNoonTwoArrivalChoice = {
    LOOKOUT: 1,
    REST: 2
};

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\arrival\SlaveNoonTwoArrival.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\SlaveNoonTwoController.js

/**
 * Controller for the Enslavement, Noon Two scene.
 * @param {string} sceneId The id of the scene.
 */
function SlaveNoonTwoController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.nextSceneReady = false;
    this.nextSceneName = "SlaveAfternoonTwo";
    this.nextFolderName = "Enslavement/DayTwo";

    this.sekiSceneReady = true;
    this.sekiSceneName = "SekiWorkshop";
    this.sekiFolderName = "HamletDay2";

    this.farmerSceneReady = false;
    this.farmerSceneName = "FarmerSlave";
    this.farmerFolderName = "Enslavement";

    this.misc = new MiscController(id, "misc");

    this.arrival = new ChangeController(id, "arrival", function () {
        if (global.kobold.history.shared.enslavement.nightOne.knockedOut) {
            global.kobold.status.physical.increase(4, "Troubled sleep");
            global.kobold.status.mental.decrease(8, "Nightmares");
            global.kobold.status.stamina.increase(6, "Troubled sleep");
        } else {
            global.kobold.status.physical.increase(4, "Troubled sleep");
            global.kobold.status.mental.increase(4, "Nightmares");
            global.kobold.status.stamina.increase(6, "Troubled sleep");
        }
    }, function () {
        if (global.kobold.history.shared.enslavement.nightOne.knockedOut) {
            return global.ui.change.buildNotification([
                new Change("Troubled sleep", "Health", +4),
                new Change("Long unconsciousness", "Sanity", -4),
                new Change("Nightmares", "Sanity", -4),
                new Change("Meagre rest", "Stamina", +6)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Troubled sleep", "Health", +4),
                new Change("Slept", "Sanity", +8),
                new Change("Nightmares", "Sanity", -4),
                new Change("Meagre rest", "Stamina", +6)
            ]);
        }
    });

    this.preBlackmail = new ChangeController(id, "preBlackmail", function () {
        if (that.arrival.get() === SlaveNoonTwoArrivalChoice.LOOKOUT) {
            if (global.kobold.history.shared.enslavement.nightOne.knockedOut) {
                global.kobold.status.physical.decrease(7, "Killer headache");
                global.kobold.status.mental.decrease(2, "Woke up early");
            } else {
                global.kobold.status.physical.decrease(5, "Headache");
                global.kobold.status.mental.decrease(2, "Woke up early");
            }
        } else {
            if (global.kobold.history.shared.enslavement.nightOne.knockedOut) {
                global.kobold.status.physical.decrease(5, "Killer headache");
                global.kobold.status.mental.decrease(4, "Crude awakening");
            } else {
                global.kobold.status.physical.decrease(3, "Headache");
                global.kobold.status.mental.decrease(4, "Crude awakening");
            }
        }
    }, function () {
        if (that.arrival.get() === SlaveNoonTwoArrivalChoice.LOOKOUT) {
            if (global.kobold.history.shared.enslavement.nightOne.knockedOut) {
                return global.ui.change.buildNotification([
                    new Change("Stared into the sun", "Health", -2),
                    new Change("Killer headache", "Health", -5),
                    new Change("Waking up early", "Sanity", -2)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Stared into the sun", "Health", -2),
                    new Change("Headache", "Health", -3),
                    new Change("Waking up early", "Sanity", -2)
                ]);
            }
        } else {
            if (global.kobold.history.shared.enslavement.nightOne.knockedOut) {
                return global.ui.change.buildNotification([
                    new Change("Killer headache", "Health", -5),
                    new Change("Crude awakening", "Sanity", -4)
                ]);
            } else {
                return global.ui.change.buildNotification([
                    new Change("Headache", "Health", -3),
                    new Change("Crude awakening", "Sanity", -4)
                ]);
            }
        }
    });

    this.offer = new ChangeController(id, "offer", function () {
        if (global.kobold.history.shared.enslavement.nightOne.escapedBindings) {
            global.kobold.status.mental.decrease(2, "Indecent proposal");
        } else {
            global.kobold.status.mental.increase(2, "Indecent proposal");
        }
    }, function () {
        if (global.kobold.history.shared.enslavement.nightOne.escapedBindings) {
            return global.ui.change.buildNotification([
                new Change("Indecent proposal", "Sanity", -1),
                new Change("Considering it", "Sanity", -1)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Hands freed", "Sanity", +4),
                new Change("Indecent proposal", "Sanity", -1),
                new Change("Considering it", "Sanity", -1)
            ]);
        }
    });

    this.blackmail = new ChangeController(id, "blackmail", function () {
        if (global.kobold.history.shared.enslavement.nightOne.escapedBindings) {
            global.kobold.status.mental.decrease(5, "Indecent proposal");
        } else {
            global.kobold.status.mental.decrease(1, "Indecent proposal");
        }
    }, function () {
        if (global.kobold.history.shared.enslavement.nightOne.escapedBindings) {
            return global.ui.change.buildNotification([
                new Change("Grim future", "Sanity", -3),
                new Change("Indecent proposal", "Sanity", -1),
                new Change("Considering it", "Sanity", -1)
            ]);
        } else {
            return global.ui.change.buildNotification([
                new Change("Hands freed", "Sanity", +4),
                new Change("Grim future", "Sanity", -3),
                new Change("Indecent proposal", "Sanity", -1),
                new Change("Considering it", "Sanity", -1)
            ]);
        }
    });

    this.goodNothing = new ChangeController(id, "goodNothing", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+4);
        global.kobold.status.mental.decrease(2, "Inability to decide");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Inability to decide", "Sanity", -2)
        ]);
    });

    this.goodDecline = new ChangeController(id, "goodDecline", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+3);
        global.kobold.status.mental.increase(5, "Stood up for yourself");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stood up for yourself", "Sanity", +5)
        ]);
    });

    this.goodInsult = new ChangeController(id, "goodInsult", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+2);
        global.kobold.status.mental.increase(4, "Stood up for yourself");
        global.kobold.status.stamina.decrease(2, "Excessive insulting");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stood up for yourself", "Sanity", +5),
            new Change("Went a little overboard", "Sanity", -1),
            new Change("Excessive insulting", "Stamina", -2)
        ]);
    });

    this.goodAttack = new ChangeController(id, "goodAttack", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+1);
        global.kobold.status.mental.increase(1, "Blind rage");
        global.kobold.status.stamina.decrease(5, "Wild flailing");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stood up for yourself", "Sanity", +5),
            new Change("Resorted to violence", "Sanity", -2),
            new Change("Blind rage", "Sanity", -2),
            new Change("Wild flailing", "Stamina", -5)
        ]);
    });

    this.goodReluctant = new ChangeController(id, "goodReluctant", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+5);
        global.kobold.status.mental.decrease(7, "Kissing balls");
        global.kobold.status.exertion.increase(5, "Kissing balls");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Humiliation", "Sanity", -10),
            new Change("Amused Fenrir", "Sanity", +3),
            new Change("Kissing balls", "Sexual exhaustion", +5, true)
        ]);
    });

    this.goodEager = new ChangeController(id, "goodEager", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+6);
        global.kobold.status.mental.increase(3, "Pleasing Fenrir");
        global.kobold.status.exertion.increase(10, "Worshipping balls");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stockholm syndrome", "Sanity", -2),
            new Change("Pleased Fenrir", "Sanity", +5),
            new Change("Eager ball-licking", "Sexual exhaustion", +10, true)
        ]);
    });

    this.badNothing = new ChangeController(id, "badNothing", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+4);
        global.kobold.status.mental.decrease(5, "Impending doom");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Inability to decide", "Sanity", -2),
            new Change("Impending doom", "Sanity", -3)
        ]);
    });

    this.badDecline = new ChangeController(id, "badDecline", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+3);
        global.kobold.status.mental.increase(2, "Impending doom");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stood up for yourself", "Sanity", +5),
            new Change("Impending doom", "Sanity", -3)
        ]);
    });

    this.badInsult = new ChangeController(id, "badInsult", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+2);
        global.kobold.status.mental.increase(1, "Impending doom");
        global.kobold.status.stamina.decrease(2, "Excessive insulting");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stood up for yourself", "Sanity", +5),
            new Change("Went a little overboard", "Sanity", -1),
            new Change("Impending doom", "Sanity", -3),
            new Change("Excessive insulting", "Stamina", -2)
        ]);
    });

    this.badAttack = new ChangeController(id, "badAttack", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+1);
        global.kobold.status.mental.decrease(2, "Impending doom");
        global.kobold.status.stamina.decrease(5, "Wild flailing");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stood up for yourself", "Sanity", +5),
            new Change("Resorted to violence", "Sanity", -2),
            new Change("Blind rage", "Sanity", -2),
            new Change("Impending doom", "Sanity", -3),
            new Change("Wild flailing", "Stamina", -5)
        ]);
    });

    this.badReluctant = new ChangeController(id, "badReluctant", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+5);
        global.kobold.status.mental.decrease(5, "Kissing balls");
        global.kobold.status.exertion.increase(5, "Kissing balls");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Humiliation", "Sanity", -10),
            new Change("Avoided gag", "Sanity", +5),
            new Change("Kissing balls", "Sexual exhaustion", +5, true)
        ]);
    });

    this.badEager = new ChangeController(id, "badEager", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+6);
        global.kobold.status.mental.increase(8, "Pleasing Fenrir");
        global.kobold.status.exertion.increase(10, "Worshipping balls");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Stockholm syndrome", "Sanity", -2),
            new Change("Pleased Fenrir", "Sanity", +5),
            new Change("Avoided gag", "Sanity", +5),
            new Change("Eager ball-licking", "Sexual exhaustion", +10, true)
        ]);
    });

    this.preGag = new ChoiceController(id, "preGag");

    this.gag = new ChoiceController(id, "gag");

    this.gagRefuse = new ChangeController(id, "gagRefuse", function () {
        global.kobold.inventory.equipment.wear(new UnderwearGag());
        global.kobold.status.physical.decrease(10, "Underwear gag");
        global.kobold.status.mental.decrease(14, "Underwear gag");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Kidney punched", "Health", -7),
            new Change("Unhygienic gag", "Health", -3),
            new Change("Rancid taste", "Sanity", -10),
            new Change("Humiliation", "Sanity", -4)
        ]);
    });

    this.gagOpen = new ChangeController(id, "gagOpen", function () {
        global.kobold.inventory.equipment.wear(new UnderwearGag());
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+2);
        global.kobold.status.physical.decrease(3, "Underwear gag");
        global.kobold.status.mental.decrease(9, "Underwear gag");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Unhygienic gag", "Health", -3),
            new Change("Kept pride", "Sanity", +5),
            new Change("Rancid taste", "Sanity", -10),
            new Change("Humiliation", "Sanity", -4)
        ]);
    });

    this.preInject = new ChoiceController(id, "preInject");

    this.inject = new ChangeController(id, "inject", function () {
        if (that.misc.dick()) {
            global.kobold.body.growDick(global.kobold.body.dick + 1);
        } else {
            global.kobold.body.growBreasts(global.kobold.body.breasts + 1)
        }
        global.kobold.status.physical.decrease(4, "Pricked by needle");
        global.kobold.status.mental.decrease(12, "Forced transformation");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Pricked by needle", "Health", -4),
            new Change("Forced transformation", "Sanity", -8),
            new Change("Unvented frustration", "Sanity", -4)
        ]);
    });

    this.potion = new ChoiceController(id, "potion");

    this.neither = new ChangeController(id, "neither", function () {
        global.kobold.status.mental.increase(4, "Kind gesture");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Kind gesture", "Sanity", +6)
        ]);
    });

    this.mPlus = new ChangeController(id, "mPlus", function () {
        if (that.misc.dick()) {
            global.kobold.body.growDick(global.kobold.body.dick + 1);
        } else {
            global.kobold.body.growBreasts(global.kobold.body.breasts - 1);
        }
        global.kobold.status.mental.increase(4, "Kind present");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Received present", "Sanity", +6),
            new Change("Transformation", "Sanity", -2)
        ]);
    });

    this.femPlus = new ChangeController(id, "femPlus", function () {
        if (that.misc.dick()) {
            global.kobold.body.growDick(global.kobold.body.dick - 1);
        } else {
            global.kobold.body.growBreasts(global.kobold.body.breasts + 1);
        }
        global.kobold.status.mental.increase(4, "Kind present");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Received present", "Sanity", +6),
            new Change("Transformation", "Sanity", -2)
        ]);
    });

    this.preOutside = new ChangeController(id, "preOutside");

    this.outside = new ChangeController(id, "outside", function () {
        global.kobold.status.mental.decrease(4, "Public nudity");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Public nudity", "Sanity", -4)
        ]);
    });

    this.batheBad = new ChangeController(id, "batheBad", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+2);
        global.kobold.status.physical.decrease(3, "Terrible hygiene");
        global.kobold.status.mental.decrease(15, "Bathing in animal spit");
        global.kobold.stats.decreaseStat(StatsEnum.CHARISMA, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Terrible hygiene", "Health", -3),
            new Change("Bathing in spit", "Sanity", -6),
            new Change("Fenrir's masturbation", "Sanity", -4),
            new Change("Captor standoff", "Sanity", -2),
            new Change("Lingering smell", "Charisma", -1)
        ]);
    });

    this.batheGood = new ChangeController(id, "batheGood", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+2);
        global.kobold.status.physical.decrease(3, "Terrible hygiene");
        global.kobold.status.mental.decrease(4, "Bathing in animal spit");
        global.kobold.stats.decreaseStat(StatsEnum.CHARISMA, 1);
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Terrible hygiene", "Health", -3),
            new Change("Bathing in spit", "Sanity", -6),
            new Change("Pleased Fenrir", "Sanity", +2),
            new Change("Lingering smell", "Charisma", -1)
        ]);
    });

    this.drinkBad = new ChangeController(id, "drinkBad", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+1);
        global.kobold.status.physical.decrease(5, "Drinking animal spit");
        global.kobold.status.mental.decrease(5, "Drinking animal spit");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Hydration", "Health", +3),
            new Change("Drinking spit", "Health", -3),
            new Change("Butt kicked", "Health", -2),
            new Change("Bathing in spit", "Health", -3),
            new Change("Hydration", "Sanity", +3),
            new Change("Drinking spit", "Sanity", -4),
            new Change("Pain", "Sanity", -2),
            new Change("Captor standoff", "Sanity", -2)
        ]);
    });

    this.drinkGood = new ChangeController(id, "drinkGood", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(+1);
        global.kobold.status.physical.increase(0, "Drinking animal spit");
        global.kobold.status.mental.decrease(4, "Drinking animal spit");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Hydration", "Health", +3),
            new Change("Drinking spit", "Health", -3),
            new Change("Hydration", "Sanity", +3),
            new Change("Drinking spit", "Sanity", -4),
            new Change("Irony", "Sanity", -1),
            new Change("Captor standoff", "Sanity", -2)
        ]);
    });

    this.refuseBad = new ChangeController(id, "refuseBad", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(0);
        global.kobold.status.mental.decrease(3, "Standoff with Fenrir");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Intense standoff", "Sanity", -4),
            new Change("Captor standoff", "Sanity", -2),
            new Change("Avoided spit-trough", "Sanity", +3)
        ]);
    });

    this.refuseGood = new ChangeController(id, "refuseGood", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(0);
        global.kobold.status.mental.increase(5, "Fenrir's favor");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Avoided spit-trough", "Sanity", +3),
            new Change("Fenrir's favor", "Sanity", +4),
            new Change("Captor standoff", "Sanity", -2)
        ]);
    });

    this.preStage = new ChoiceController(id, "preStage");

    this.stage = new ChangeController(id, "stage", function () {
        global.kobold.status.physical.decrease(2, "Thrown onto stage");
        global.kobold.status.mental.decrease(6, "Safety concerns");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Thrown onto stage", "Health", -2),
            new Change("Naked in front of crowd", "Sanity", -2),
            new Change("Insulted and degraded", "Sanity", -2),
            new Change("Fearing for safety", "Sanity", -2)
        ]);
    });

    this.pose = new ChangeController(id, "pose", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(2);
        global.kobold.status.mental.decrease(2, "Willing exhibition");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Willing exhibition", "Sanity", -4),
            new Change("Pleased caravan master", "Sanity", +2)
        ]);
    });

    this.refuse = new ChangeController(id, "refuse", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(0);
        global.kobold.status.physical.decrease(14, "Gut-punch");
        global.kobold.status.mental.decrease(12, "Pain");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Choked", "Health", -6),
            new Change("Gut-punched", "Health", -8),
            new Change("Pain", "Sanity", -6),
            new Change("Crowd schadenfreude", "Sanity", -6)
        ]);
    });

    this.smile = new ChangeController(id, "smile", function () {
        global.kobold.history.shared.enslavement.noonTwo.alterComplianceRating(1);
        global.kobold.status.mental.increase(0, "Reluctant smile");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Reluctant smile", "Sanity", -2),
            new Change("Pleased caravan master", "Sanity", +2)
        ]);
    });

    this.preResult = new ChoiceController(id, "preResult");

    this.result = new ChoiceController(id, "result");

    this.nobody = new ChangeController(id, "nobody", function () {

    }, function () {
        return global.ui.change.buildNotification([
            new Change("TODO", "CHANGE", +1)
        ]);
    });

    this.seki = new ChangeController(id, "seki", function () {
        global.kobold.status.mental.increase(5, "Saved by Seki");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Saved", "Sanity", +10),
            new Change("Scolded", "Sanity", -5)
        ]);
    });

    this.farmer = new ChangeController(id, "farmer", function () {

    }, function () {
        return global.ui.change.buildNotification([
            new Change("TODO", "CHANGE", +1)
        ]);
    });

    this.outsideComplianceStorage = new ChoiceController(id, "outsideComplianceStorage");
    this.resultComplianceStorage = new ChoiceController(id, "resultComplianceStorage");
    this.resultBodyMisc = new MiscController(id, "resultBody");

    this.fragments = {};
    this.fragments.results = new SlaveNoonTwoResults(this);
    this.fragments.stage = new SlaveNoonTwoStage(this);
    this.fragments.outside = new SlaveNoonTwoOutside(this);
    this.fragments.gagged = new SlaveNoonTwoBadGag(this);
    this.fragments.notGagged = new SlaveNoonTwoGoodPotion(this);
    this.fragments.proposition = new SlaveNoonTwoProposition(this);

    //this.bgm = new AudioController(id, "res/audio/effects/cave.mp3", 70);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

}

var SlaveNoonTwoOfferChoice = {
    NOTHING: 1,
    DECLINE: 2,
    INSULT: 3,
    ATTACK: 4,
    RELUCTANT: 5,
    EAGER: 6
};

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\SlaveNoonTwoController.js

// scenes\Enslavement\DayTwo\SlaveNoonTwo\SlaveNoonTwo.js

/**
 * Noon two of enslavement.
 */
function SlaveNoonTwo() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "SlaveNoonTwo";
    var sceneName = "Enslavement, Noon Two";
    var sceneSubfolder = "Enslavement/DayTwo";
    var ctrl = new SlaveNoonTwoController(sceneId);
    var arrival = new SlaveNoonTwoArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]Enslavement[/t]`);
        tab += global.parser.parse(`[st]Noon two[/st]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new SlaveNoonTwo();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\Enslavement\DayTwo\SlaveNoonTwo\SlaveNoonTwo.js



// END SCENE SlaveNoonTwo



// SCENE RiversideMarsh

// scenes\Deepwoods\RiversideMarsh\upstream\RiversideMarshUpstream.js

/**
 * Information about the class.
 */
function RiversideMarshUpstream() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
}

// END scenes\Deepwoods\RiversideMarsh\upstream\RiversideMarshUpstream.js

// scenes\Deepwoods\RiversideMarsh\downstream\RiversideMarshDownstream.js

/**
 * Downstream section of the Riverside Marsh.
 * @param {RiversideMarshController} controller The Riverside Marsh controller.
 */
function RiversideMarshDownstream(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = RiversideMarshArrivalChoice.DOWNSTREAM;
    var reqController = ctrl.arrival;

    var continueId = arguments.callee.name + "ButtonOne";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            You're saved!
                            You thought you were never going to get out of this damp [b]hellhole[/b].
                            With renewed determination, you walk alongside the river's edge back towards the habitable part of the forest.
                        [/p]
                        [p]
                            Navigating the riverbank is a lot easier than making your way through the deepwoods.
                            There may be a bunch of treacherous rocks in your way, but without the forest roof to block it,
                            the moon's radiating light illuminates your path rather nicely.
                            You can actually [i]see[/i] where you're going, which allows you to make a lot more progress than your earlier clumsy stumbling.
                        [/p]
                        [p]
                            In less than an hour, you find yourself back in the civilized side of the woods.
                            Recognizing your surroundings, you split off from the riverside, and head straight towards your home.
                        [/p]`;

        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the continue button.
     */
    function buildChoice() {
        var disabled = ctrl.continued();
        var choice = `<div class="choicecontainer">`;
        choice += `<button class="button bigbutton continuebutton" id="${continueId}" ${disabled ? "disabled" : ""}>`;
        choice += `Continue &gt;&gt`;
        choice += `</button>`;
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + continueId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next scene.
     * @param {string} sceneName Name of the scene to load.
     * @param {string} folder Folder of the scene to load.
     */
    function loadNextPart(sceneName, folder) {
        global.ui.fadeOutContent(function () {
            global.sceneEnd();
            global.sceneLoader.loadScene(sceneName, folder);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerContinueListener() {
        $("#" + continueId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            ctrl.continue();
            loadNextPart("Home", "");
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.downstream.change();
        part += buildChoice();
        part += `</div>`;
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        if (ctrl.continued()) {
            return;
        }

        registerContinueListener();
    };
}

// END scenes\Deepwoods\RiversideMarsh\downstream\RiversideMarshDownstream.js

// scenes\Deepwoods\RiversideMarsh\arrival\RiversideMarshArrival.js

/**
 * Arrival section of the Riverside Marsh.
 * @param {RiversideMarshController} controller The Riverside Marsh controller.
 */
function RiversideMarshArrival(controller) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var ctrl = controller;

    var reqChoice = false;
    var reqController = false;
    var subController = ctrl.arrival;

    var firstId = arguments.callee.name + "ButtonOne";
    var secondId = arguments.callee.name + "ButtonTwo";

    var firstNext = new RiversideMarshDownstream(ctrl);
    var secondNext = new RiversideMarshUpstream(ctrl);

    var firstChoice = RiversideMarshArrivalChoice.DOWNSTREAM;
    var secondChoice = RiversideMarshArrivalChoice.UPSTREAM;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the content of this section.
     * @returns {string} The content.
     */
    function buildContent() {
        var toReturn = `[p]
                            Traversing the deepwoods in near complete pitch-blackness isn't an easy task.
                            Luckily, your superb nightvision helps you more than a little, amplifying what little light manages to shine through the dense canopy.
                            Despite not being completely [i]blind[/i], you are in constant danger of tripping and falling.
                            Your feet are repeatedly getting caught on roots of varying shapes and sizes,
                            and when you finally find a place to plop down one of your scaly paws,
                            you still have barely visible low-hanging branches to worry about.
                            Flexible as you are, however, you succeed in making some form of progress.
                        [/p]
                        [p]
                            After nearly half an hour of slow and cautious trudging through the darkness, the ground beneath your feet begins to soften.
                            With every step, the soil [i]dampens[/i] further, and before long,
                            the previously solid dirt you were walking on is replaced by thick, sticky and slippery [b]mud[/b].
                            Coming to a halt for a moment, you become aware of a faint, continuous splashing in the distance.
                            Could you be nearing the river?
                        [/p]
                        [p]
                            Emboldened by the premise of a landmark you recognize, you press onwards.
                            After just a couple of minutes more of struggling to pass the terrain, the tightly packed trees give way to a clearing.
                            You were right!
                            It's the river!
                            You find yourself on top of the riverbank, gazing out onto the slowly streaming waterway, sparkling stars reflected by the water's surface.
                        [/p]
                        [p]
                            While you still have no idea where exactly you are, you know that Varanar is downstream from your home.
                            That, in turn, must mean that you are currently somewhere upstream.
                            Closer to the river's source than you've ever been.
                            If you follow the waterflow, it should lead you straight back to parts of the forest that you know.
                            Or you could head the opposite way, further away from civilization.
                        [/p]`;
        return global.parser.parse(toReturn);
    }

    /**
     * Builds and returns the choice button.
     */
    function buildFirstChoice() {
        var disabled = subController.has() ? "disabled" : "";
        var isSelected = subController.get() === firstChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${firstId}" ${disabled}>Downstream</button>`;

        var tooltip = global.parser.parse(`You need to get back home as soon as you can.`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice button.
     */
    function buildSecondChoice() {
        var disabled = subController.has() ? "disabled" : "";
        disabled = "disabled";
        var isSelected = subController.get() === secondChoice;
        var selected = isSelected ? "selected" : "";
        var button = `<button class="button inlinechoice ${selected}" id="${secondId}" ${disabled}>Upstream</button>`;

        var tooltip = global.parser.parse(`Your adventure has barely even begun. Why turn back now? <span class="todo">Unavailable.</span>`);
        var popupContainer = global.ui.makeToolTip(button, tooltip, false, false, "width: 160px;");
        // Re-wrap with proper positioning.
        button = `<div class="inlinechoice">`;
        button += popupContainer;
        button += `</div>`;

        return button;
    }

    /**
     * Builds and returns the choice of this section.
     * @returns {string} The choice of this section.
     */
    function buildChoice() {
        var choice = `<div class="choicecontainer"><div class="inlinechoiceheader">You go:</div>`;
        choice += buildFirstChoice();
        choice += `<hr class="inlinechoiceseparator">`;
        choice += buildSecondChoice();
        choice += `</div>`;
        return choice;
    }

    /**
     * Disables all three buttons.
     */
    function disableButtons() {
        var buttons = [$("#" + firstId), $("#" + secondId)];
        buttons.forEach(function (button) {
            button.unbind();
            button.attr("disabled", true);
        });
    }

    /**
     * Loads and displays the next part.
     * @param {object} part Which part to load.
     */
    function loadNextPart(part) {
        var built = part.build(true);
        global.ui.appendBeforeFinisher(built, global.ui.genericHeaderId);
        $("#" + part.id).fadeIn(400);
        part.registerListeners();
    }

    /**
     * Adds a listener to the button.
     */
    function registerFirstListener() {
        $("#" + firstId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(firstChoice);
            loadNextPart(firstNext);
        });
    }

    /**
     * Adds a listener to the button.
     */
    function registerSecondListener() {
        $("#" + secondId).click(function () {
            $(this).addClass("selected");
            disableButtons();
            subController.set(secondChoice);
            loadNextPart(secondNext);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = arguments.callee.name;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns this part.
     * @param {boolean} invisible If passed and true, will give the entire section a display:none style.
     */
    this.build = function (invisible) {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return "";
        }

        global.ui.applyTheme("forest_night");
        if (!subController.has()) {
            ctrl.riverAmbience.playSettingsDependent();
        }
        var part = `<div id="${that.id}" ${invisible ? 'style="display:none;"' : ""}>`;
        part += buildContent();
        part += ctrl.arrival.change();
        part += buildChoice();
        part += `</div>`;
        part += firstNext.build();
        //part += secondNext.build();
        return part;
    };

    /**
     * Registers all the required button listeners for this part.
     */
    this.registerListeners = function () {
        if (reqController && ((reqChoice && reqController.get() !== reqChoice) || (!reqChoice && !reqController.continued()))) {
            return;
        }

        firstNext.registerListeners();
        //secondNext.registerListeners();

        if (subController.has()) {
            return;
        }

        registerFirstListener();
        //registerSecondListener();
    };
}

var RiversideMarshArrivalChoice = {
    DOWNSTREAM: 1,
    UPSTREAM: 2
};

// END scenes\Deepwoods\RiversideMarsh\arrival\RiversideMarshArrival.js

// scenes\Deepwoods\RiversideMarsh\RiversideMarshController.js

/**
 * Controller for the Caravan Confinement scene.
 * @param {string} sceneId The id of the scene.
 */
function RiversideMarshController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var id = sceneId;
    SceneController.call(this, id);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.nextSceneReady = false;
    this.nextSceneName = "HamletAuction";
    this.nextFolderName = "Enslavement";

    this.misc = new MiscController(id, "misc");

    this.arrival = new ChangeController(id, "arrival", function () {
        global.kobold.status.mental.increase(10, "Finding a landmark");
        global.kobold.status.stamina.decrease(5, "Rough terrain traversal");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Landmark", "Sanity", +5),
            new Change("A way out?", "Sanity", +5),
            new Change("Traversing rough terrain", "Stamina", -5)
        ]);
    });

    this.downstream = new ChangeController(id, "downstream", function () {
        global.kobold.status.mental.increase(5, "Going home");
        global.kobold.status.stamina.decrease(5, "Walking on tired feet");
    }, function () {
        return global.ui.change.buildNotification([
            new Change("Going home", "Sanity", +5),
            new Change("More hiking", "Stamina", -5)
        ]);
    });

    this.preContinue = new ContinueController(id, "preContinue");

    this.riverAmbience = new AudioController(id, "res/audio/effects/river.mp3", 100);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

}

// END scenes\Deepwoods\RiversideMarsh\RiversideMarshController.js

// scenes\Deepwoods\RiversideMarsh\RiversideMarsh.js

/**
 * First night in the deepwoods.
 */
function RiversideMarsh() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;
    var sceneId = "RiversideMarsh";
    var sceneName = "Deepwoods, Riverside";
    var sceneSubfolder = "Deepwoods";
    var ctrl = new RiversideMarshController(sceneId);
    var arrival = new RiversideMarshArrival(ctrl);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets all the required global variables such as scene id, scene name,
     * scene stack and theme.
     */
    function setGlobals() {
        global.currentScene = that.id;
        global.currentSceneName = that.name;
        global.currentSceneFolder = that.subfolder ? that.subfolder : "";
        global.kobold.history.addToSceneStack(that.id);
    }

    /**
     * Builds the tab as a HTML string.
     * @returns {string} The tab, as a HTML string.
     */
    function buildTab() {
        var tab = global.ui.buildGenericHeader();
        tab += global.parser.parse(`[t]Deepwoods[/t]`);
        tab += global.parser.parse(`[st]Riverside[/st]`);
        tab += arrival.build();
        tab += global.ui.buildGenericFooter();
        return tab;
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.id = sceneId;
    this.name = sceneName;
    this.subfolder = sceneSubfolder;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this scene.
     */
    this.load = function () {
        setGlobals();
        var tab = buildTab();
        global.ui.loadToMain(tab);
        arrival.registerListeners();
        ctrl.autosave();
    };
}

global.sceneLoader.queueRegister(function () {
    var scene = new RiversideMarsh();
    global.sceneLoader.registerScene(scene.id, scene);
});

// END scenes\Deepwoods\RiversideMarsh\RiversideMarsh.js



// END SCENE RiversideMarsh

